#define GE_USE_THREADS

/*
	description:

		"C declarations for the Gobo Eiffel runtime."

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2005-2020, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef GE_EIFFEL_H
#define GE_EIFFEL_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

/* Class name mapping as defined in the FreeELKS library. */
#ifndef EIF_INTEGER
#define EIF_INTEGER EIF_INTEGER_32
#endif
#ifndef EIF_CHARACTER
#define EIF_CHARACTER EIF_CHARACTER_8
#endif
#ifndef EIF_REAL
#define EIF_REAL EIF_REAL_32
#endif
#ifndef EIF_DOUBLE
#define EIF_DOUBLE EIF_REAL_64
#endif
#ifndef GE_ms
#if EIF_CHARACTER == EIF_CHARACTER_8
#define GE_ms(s,c) GE_ms8((s),(c))
#else
#define GE_ms(s,c) GE_ms32((s),(c))
#endif
#endif

#if defined(__USE_POSIX) || defined(__unix__) || defined(_POSIX_C_SOURCE)
#include <unistd.h>
#endif
#if !defined(WIN32) && \
	(defined(WINVER) || defined(_WIN32_WINNT) || defined(_WIN32) || \
	defined(__WIN32__) || defined(__TOS_WIN__) || defined(_MSC_VER) || \
	defined(__MINGW32__))
#define WIN32 1
#endif
#ifdef WIN32
#define EIF_WINDOWS 1
#include <windows.h>
#endif

/* See https://sourceforge.net/p/predef/wiki/OperatingSystems/ */
#if (defined(macintosh) || defined(Macintosh))
#define EIF_MAC 1
#define EIF_MACOSX 1
#endif
#if (defined(__APPLE__) && defined(__MACH__))
/* Apparently ISE does not define EIF_MASOSX for Mac OS X >=10.4 (see EXECUTION_ENVIRONMENT.available_cpu_count) */
#define EIF_MAC 1
#endif

#if (defined(VMS) || defined(__VMS))
#define EIF_VMS 1
#endif

#if (defined(__VXWORKS__) || defined(__vxworks))
#define EIF_VXWORKS 1
#endif

#define BYTEORDER 0x1234

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <stddef.h>

#define EIF_OS_WINNT 	1
#define EIF_OS_LINUX 	2
#define EIF_OS_DARWIN 	4
#define EIF_OS_VXWORKS	11
#define EIF_OS_VMS	12

/* Platform definition */
/* Unix definition */
#define EIF_IS_UNIX EIF_TRUE
#define EIF_OS EIF_OS_LINUX
/* Windows definition */
#ifdef EIF_WINDOWS
#define EIF_IS_WINDOWS EIF_TRUE
#undef EIF_OS
#define EIF_OS EIF_OS_WINNT
#undef EIF_IS_UNIX
#define EIF_IS_UNIX EIF_FALSE
#else
#define EIF_IS_WINDOWS EIF_FALSE
#endif
/* VMS definition */
#ifdef EIF_VMS
#define EIF_IS_VMS EIF_TRUE
#undef EIF_OS
#define EIF_OS EIF_OS_VMS
#undef EIF_IS_UNIX
#define EIF_IS_UNIX EIF_FALSE
#else
#define EIF_IS_VMS EIF_FALSE
#endif
/* MAC definition */
#ifdef EIF_MAC
#define EIF_IS_MAC EIF_TRUE
#undef EIF_OS
#define EIF_OS EIF_OS_DARWIN
#undef EIF_IS_UNIX
#define EIF_IS_UNIX EIF_FALSE
#else
#define EIF_IS_MAC EIF_FALSE
#endif
/* VxWorks definition */
#ifdef EIF_VXWORKS
#define EIF_IS_VXWORKS EIF_TRUE
#undef EIF_OS
#define EIF_OS EIF_OS_VXWORKS
#undef EIF_IS_UNIX
#define EIF_IS_UNIX EIF_FALSE
#else
#define EIF_IS_VXWORKS EIF_FALSE
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifdef _MSC_VER /* MSVC */
typedef signed char int8_t;
typedef signed short int16_t;
typedef signed int int32_t;
typedef signed __int64 int64_t;
typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef unsigned int uint32_t;
typedef unsigned __int64 uint64_t;
#else
#if defined (__BORLANDC__) && (__BORLANDC__ < 0x600) /* Borland before 6.0 */
typedef signed char int8_t;
typedef signed short int16_t;
typedef signed long int int32_t;
typedef signed __int64 int64_t;
typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef unsigned long int uint32_t;
typedef unsigned __int64 uint64_t;
#else
#include <inttypes.h>
#endif
#endif

/* Portable integer pointers */
#ifdef EIF_WINDOWS
#ifndef _INTPTR_T_DEFINED
#define _INTPTR_T_DEFINED
#ifdef _WIN64
typedef __int64 intptr_t;
#else
typedef int intptr_t;
#endif
#endif
#ifndef _UINTPTR_T_DEFINED
#define _UINTPTR_T_DEFINED
#ifdef _WIN64
  typedef unsigned __int64 uintptr_t;
#else
  typedef unsigned int uintptr_t;
#endif
#endif
#endif

/* C type for underlying integer type identifying object's dynamic type. */
typedef uint16_t EIF_TYPE_INDEX;

/*
 * Abstraction representing an Eiffel type.
 * It is made of a compiler type-id,
 * and of some annotations (attached/detachable/separate/variant/frozen).
 */
typedef struct eif_type {
	EIF_TYPE_INDEX id;
	EIF_TYPE_INDEX annotations;
} EIF_TYPE;

/*
 * Since EIF_TYPE and EIF_ENCODED_TYPE have the same size, the encoded version
 * is basically a memcpy version of the EIF_TYPE representation.
 * It is used to provide backward compatibility to most Eiffel and
 * C APIs manipulating types as an INTEGER.
 */
typedef int32_t EIF_ENCODED_TYPE;
typedef EIF_ENCODED_TYPE EIF_TYPE_ID;
#define EIF_NO_TYPE (EIF_TYPE_ID)(-1)

/* Basic Eiffel types */
typedef char EIF_BOOLEAN;
typedef unsigned char EIF_CHARACTER_8;
typedef uint32_t EIF_CHARACTER_32;
typedef int8_t EIF_INTEGER_8;
typedef int16_t EIF_INTEGER_16;
typedef int32_t EIF_INTEGER_32;
typedef int64_t EIF_INTEGER_64;
typedef uint8_t EIF_NATURAL_8;
typedef uint16_t EIF_NATURAL_16;
typedef uint32_t EIF_NATURAL_32;
typedef uint64_t EIF_NATURAL_64;
typedef void* EIF_POINTER;
typedef float EIF_REAL_32;
typedef double EIF_REAL_64;
typedef struct {EIF_TYPE_INDEX id; uint16_t flags;} EIF_ANY;
typedef EIF_ANY* EIF_REFERENCE;
typedef struct {EIF_TYPE_INDEX id; uint16_t flags; EIF_REFERENCE area; EIF_INTEGER count;} EIF_STRING;
typedef struct {EIF_TYPE_INDEX id; uint16_t flags; uint32_t offset; EIF_INTEGER count; EIF_INTEGER capacity;} EIF_SPECIAL;

#ifdef EIF_WINDOWS
typedef wchar_t EIF_NATIVE_CHAR;
#else
typedef char EIF_NATIVE_CHAR;
#endif
typedef EIF_NATIVE_CHAR* EIF_FILENAME;

#define EIF_VOID ((EIF_REFERENCE)0)
#define EIF_FALSE ((EIF_BOOLEAN)'\0')
#define EIF_TRUE ((EIF_BOOLEAN)'\1')
#define EIF_TEST(x) ((x) ? EIF_TRUE : EIF_FALSE)

#define EIF_IS_WORKBENCH EIF_FALSE
#define EIF_POINTER_DISPLAY "lX"

/* For INTEGER and NATURAL manifest constants */
#define GE_int8(x) x
#define GE_nat8(x) x
#define GE_int16(x) x
#define GE_nat16(x) x
#define GE_int32(x) x##L
#define GE_nat32(x) x##U
#if defined (_MSC_VER) && (_MSC_VER < 1400) /* MSC older than v8 */
#define GE_int64(x) x##i64
#define GE_nat64(x) x##ui64
#elif defined(__BORLANDC__) && (__BORLANDC__ < 0x600) /* Borland before 6.0 */
#define GE_int64(x) x##i64
#define GE_nat64(x) x##ui64
#else /* ISO C 99 */
#define GE_int64(x) x##LL
#define GE_nat64(x) x##ULL
#endif
#ifdef __LCC__
/* lcc-win32 reports a constant overflow for -21474836478. */
#define GE_min_int32 (-GE_int32(2147483647)-GE_int32(1))
#else
#define GE_min_int32 GE_int32(-2147483648)
#endif
#define GE_max_int32 GE_int32(2147483647)
#if defined(__LCC__) || defined(__GNUC__) || defined(__MINGW32__)
/* lcc-win32 reports a constant overflow for -9223372036854775808. */
/* gcc and mingw-win64 warn that integer constant is so large that it is unsigned. */
#define GE_min_int64 (-GE_int64(9223372036854775807)-GE_int64(1))
#else
#define GE_min_int64 GE_int64(-9223372036854775808)
#endif
#if defined(__LCC__) && !defined(_WIN64)
/* lcc-win32 does not consider 64 bit constants as constants in case statement. */
#define GE_case_int64(x) ((int32_t)(x))
#define GE_case_nat64(x) ((uint32_t)(x))
#else
#define GE_case_int64(x) (x)
#define GE_case_nat64(x) (x)
#endif

#ifdef _WIN64
#define GE_IS_64_BITS EIF_TRUE
#else
#define GE_IS_64_BITS EIF_TEST(sizeof(void*)==64)
#endif

/* Posix threads */
#if !defined(EIF_WINDOWS)
#define GE_USE_POSIX_THREADS
#endif

#ifdef _MSC_VER /* MSVC */
/* MSVC does not support ISO C 99's 'snprintf' from stdio.h */
#define snprintf(a,b,c,d) sprintf(a,c,d)
#endif

/*
 * Gobo compiler version.
 * Starts with 6080 (looks like GOBO) followed by 5 digits.
 */
#define GE_compiler_version() 608000001

/*
	Interoperability with ISE.
*/
#define RTI64C(x) GE_int64(x)
#define EIF_OBJECT EIF_REFERENCE
#define EIF_OBJ EIF_OBJECT
#define OVERHEAD sizeof(EIF_ANY)
/* Function pointer call to make sure all arguments are correctly pushed onto stack. */
/* FUNCTION_CAST is for standard C calls. */
/* FUNCTION_CAST_TYPE is for non-standard C calls. */
#define FUNCTION_CAST(r_type,arg_types) (r_type (*) arg_types)
#define FUNCTION_CAST_TYPE(r_type,call_type,arg_types) (r_type (call_type *) arg_types)
#define SIGBLOCK
#define SIGRESUME
#define rt_public				/* default C scope */
#define rt_private static		/* static outside a block means private */
#define rt_shared				/* data shared between modules, but not public */
typedef intptr_t rt_int_ptr;
typedef uintptr_t rt_uint_ptr;
#define RTMS(s) GE_str8(s)
#define RTMS_EX(s,c) GE_ms8((s),(c))

#ifdef __cplusplus
}
#endif

#endif

/*
	description:

		"C types used to implement class THREAD and related threading facilities"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef GE_THREAD_TYPES_H
#define GE_THREAD_TYPES_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifdef GE_USE_THREADS

#define EIF_THREADS
#ifdef GE_USE_POSIX_THREADS
#define EIF_POSIX_THREADS
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#ifdef GE_USE_POSIX_THREADS
#include <pthread.h>
#include <semaphore.h>
#elif defined EIF_WINDOWS
#include <windows.h>
#include <process.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifdef GE_USE_POSIX_THREADS

#define EIF_THR_TYPE            pthread_t
#define EIF_CS_TYPE				pthread_mutex_t
#define EIF_MUTEX_TYPE          pthread_mutex_t
#define EIF_COND_TYPE			pthread_cond_t
#define EIF_SEM_TYPE			sem_t
#define EIF_RWL_TYPE			pthread_rwlock_t

#elif defined EIF_WINDOWS

#define EIF_THR_TYPE            HANDLE
#define EIF_CS_TYPE				CRITICAL_SECTION
#define EIF_MUTEX_TYPE          CRITICAL_SECTION
#define EIF_SEM_TYPE            HANDLE

typedef struct {
		/* Semaphore used to queue up threads waiting for the condition to become signaled. */
	EIF_SEM_TYPE* semaphore;
		/* Serialize access to fields of Current. */
	EIF_CS_TYPE* csection;
		/* Number of waiters. */
	unsigned long num_waiting;
		/* Number of already awoken. */
	unsigned long num_wake;
		/* Number of time we signaled/broadcasted for improving fairness.
		 * This ensures one thread won't steal wakeups from other threads in queue. */
	unsigned long generation;
} EIF_COND_TYPE;

typedef struct {
	EIF_MUTEX_TYPE* m; /* Internal monitor lock. */
	int rwlock; /* >0 = # readers, <0 = writer, 0 = none */
	EIF_COND_TYPE* readers_ok; /* Start waiting readers. */
	unsigned int waiting_writers; /* Number of waiting writers. */
	EIF_COND_TYPE* writers_ok; /* Start a waiting writer. */
} EIF_RWL_TYPE;

#endif

typedef struct {
	unsigned int priority;
	unsigned int stack_size;
} EIF_THR_ATTR_TYPE;

/* Struct for thread context. */
typedef struct GE_thread_context_struct GE_thread_context;
struct GE_thread_context_struct {
	EIF_THR_TYPE thread_id; /* Thread identifier for associated thread. */
	EIF_REFERENCE current; /* Eiffel root object. */
	void (*routine)(EIF_REFERENCE); /* Eiffel routine. */
	void (*set_terminated)(EIF_REFERENCE,EIF_BOOLEAN); /* Eiffel routine to set {THREAD}.terminated. */
	unsigned int initial_priority; /* Initial priority. */
	EIF_THR_TYPE last_thread_id; /* Last thread created from current thread. */
	volatile int n_children; /* Number of direct thread children. */
	EIF_MUTEX_TYPE* children_mutex; /* Mutex to wait for thread children. */
	EIF_COND_TYPE* children_cond; /* Condition variable to wait for thread children. */
	GE_thread_context* parent_context;	/* Context of parent thread, NULL if main thread. */
	int thread_exiting; /* Has current thread already called GE_thread_exit? */
	volatile int is_alive; /* Is Current thread still alive? */
};

#ifdef __cplusplus
}
#endif

#endif
#endif

/*
	description:

		"C functions used to implement once features"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef GE_ONCE_H
#define GE_ONCE_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Struct to keep track of the call status
 * and results of once features.
 */
typedef struct {
	EIF_BOOLEAN* boolean_value;
	EIF_REFERENCE* boolean_exception;
	unsigned char* boolean_status;
	unsigned int boolean_count;
	EIF_CHARACTER_8* character_8_value;
	EIF_REFERENCE* character_8_exception;
	unsigned char* character_8_status;
	unsigned int character_8_count;
	EIF_CHARACTER_32* character_32_value;
	EIF_REFERENCE* character_32_exception;
	unsigned char* character_32_status;
	unsigned int character_32_count;
	EIF_INTEGER_8* integer_8_value;
	EIF_REFERENCE* integer_8_exception;
	unsigned char* integer_8_status;
	unsigned int integer_8_count;
	EIF_INTEGER_16* integer_16_value;
	EIF_REFERENCE* integer_16_exception;
	unsigned char* integer_16_status;
	unsigned int integer_16_count;
	EIF_INTEGER_32* integer_32_value;
	EIF_REFERENCE* integer_32_exception;
	unsigned char* integer_32_status;
	unsigned int integer_32_count;
	EIF_INTEGER_64* integer_64_value;
	EIF_REFERENCE* integer_64_exception;
	unsigned char* integer_64_status;
	unsigned int integer_64_count;
	EIF_NATURAL_8* natural_8_value;
	EIF_REFERENCE* natural_8_exception;
	unsigned char* natural_8_status;
	unsigned int natural_8_count;
	EIF_NATURAL_16* natural_16_value;
	EIF_REFERENCE* natural_16_exception;
	unsigned char* natural_16_status;
	unsigned int natural_16_count;
	EIF_NATURAL_32* natural_32_value;
	EIF_REFERENCE* natural_32_exception;
	unsigned char* natural_32_status;
	unsigned int natural_32_count;
	EIF_NATURAL_64* natural_64_value;
	EIF_REFERENCE* natural_64_exception;
	unsigned char* natural_64_status;
	unsigned int natural_64_count;
	EIF_POINTER* pointer_value;
	EIF_REFERENCE* pointer_exception;
	unsigned char* pointer_status;
	unsigned int pointer_count;
	EIF_REAL_32* real_32_value;
	EIF_REFERENCE* real_32_exception;
	unsigned char* real_32_status;
	unsigned int real_32_count;
	EIF_REAL_64* real_64_value;
	EIF_REFERENCE* real_64_exception;
	unsigned char* real_64_status;
	unsigned int real_64_count;
	EIF_REFERENCE* reference_value;
	EIF_REFERENCE* reference_exception;
	unsigned char* reference_status;
	unsigned int reference_count;
	EIF_REFERENCE* procedure_exception;
	unsigned char* procedure_status;
	unsigned int procedure_count;
} GE_onces;

/*
 * Variable to keep track of the call status
 * and results of once-per-process features.
 */
extern GE_onces* GE_process_onces;

/*
 * Initialize `GE_process_onces'.
 */
extern void GE_init_onces(
	unsigned int a_boolean_count,
	unsigned int a_character_8_count,
	unsigned int a_character_32_count,
	unsigned int a_integer_8_count,
	unsigned int a_integer_16_count,
	unsigned int a_integer_32_count,
	unsigned int a_integer_64_count,
	unsigned int a_natural_8_count,
	unsigned int a_natural_16_count,
	unsigned int a_natural_32_count,
	unsigned int a_natural_64_count,
	unsigned int a_pointer_count,
	unsigned int a_real_32_count,
	unsigned int a_real_64_count,
	unsigned int a_reference_count,
	unsigned int a_procedure_count);

/*
 * Create a new 'GE_onces' struct which can deal with the
 * numbers of once features passed as argument.
 */
extern GE_onces* GE_new_onces(
	unsigned int a_boolean_count,
	unsigned int a_character_8_count,
	unsigned int a_character_32_count,
	unsigned int a_integer_8_count,
	unsigned int a_integer_16_count,
	unsigned int a_integer_32_count,
	unsigned int a_integer_64_count,
	unsigned int a_natural_8_count,
	unsigned int a_natural_16_count,
	unsigned int a_natural_32_count,
	unsigned int a_natural_64_count,
	unsigned int a_pointer_count,
	unsigned int a_real_32_count,
	unsigned int a_real_64_count,
	unsigned int a_reference_count,
	unsigned int a_procedure_count);

/*
 * Free memory allocated by `a_onces'.
 */
extern void GE_free_onces(GE_onces* a_onces);

#ifdef __cplusplus
}
#endif

#endif

/*
	description:

		"C functions used to implement class EXCEPTION"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2007-2018, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef GE_EXCEPTION_H
#define GE_EXCEPTION_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif
#ifndef GE_ONCE_H
#include "ge_once.h"
#endif
#ifdef GE_USE_THREADS
#ifndef GE_THREAD_TYPES_H
#include "ge_thread_types.h"
#endif
#endif

#include <setjmp.h>

/*
 * On Linux glibc systems, we need to use sig* versions of jmp_buf,
 * setjmp and longjmp to preserve the signal handling context.
 * One way to detect this is if _SIGSET_H_types has
 * been defined in /usr/include/setjmp.h.
 * NOTE: ANSI only recognizes the non-sig versions.
 */
#if (defined(_SIGSET_H_types) && !defined(__STRICT_ANSI__))
#define GE_jmp_buf sigjmp_buf
#define GE_setjmp(x) sigsetjmp((x),1)
#define GE_longjmp(x,y) siglongjmp((x),(y))
#else
#define GE_jmp_buf jmp_buf
#define GE_setjmp(x) setjmp((x))
#define GE_longjmp(x,y) longjmp((x),(y))
#endif

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Predefined exception codes.
 */
#define GE_EX_VOID		1			/* Feature applied to void reference */
#define GE_EX_MEM		2			/* No more memory */
#define GE_EX_PRE		3			/* Pre-condition violated */
#define GE_EX_POST		4			/* Post-condition violated */
#define GE_EX_FLOAT		5			/* Floating point exception (signal SIGFPE) */
#define GE_EX_CINV		6			/* Class invariant violated */
#define GE_EX_CHECK		7			/* Check instruction violated */
#define GE_EX_FAIL		8			/* Routine failure */
#define GE_EX_WHEN		9			/* Unmatched inspect value */
#define GE_EX_VAR		10			/* Non-decreasing loop variant */
#define GE_EX_LINV		11			/* Loop invariant violated */
#define GE_EX_SIG		12			/* Operating system signal */
#define GE_EX_BYE		13			/* Eiffel run-time panic */
#define GE_EX_RESC		14			/* Exception in rescue clause */
#define GE_EX_OMEM		15			/* Out of memory (cannot be ignored) */
#define GE_EX_RES		16			/* Resumption failed (retry did not succeed) */
#define GE_EX_CDEF		17			/* Create on deferred */
#define GE_EX_EXT		18			/* External event */
#define GE_EX_VEXP		19			/* Void assigned to expanded */
#define GE_EX_HDLR		20			/* Exception in signal handler */
#define GE_EX_IO		21			/* I/O error */
#define GE_EX_SYS		22			/* Operating system error */
#define GE_EX_RETR		23			/* Retrieval error */
#define GE_EX_PROG		24			/* Developer exception */
#define GE_EX_FATAL		25			/* Eiffel run-time fatal error */
#define GE_EX_DOL		26			/* $ applied to melted feature */
#define GE_EX_ISE_IO	27			/* I/O error raised by the ISE Eiffel runtime */
#define GE_EX_COM		28			/* COM error raised by EiffelCOM runtime */
#define GE_EX_RT_CHECK	29			/* Runtime check error such as out-of-bound array access */
#define GE_EX_OLD		30			/* Old violation */
#define GE_EX_SEL		31			/* Serialization failure */
#define GE_EX_DIRTY		32			/* SCOOP processor dirty exception. */
#define GE_EX_NEX		32			/* Number of internal exceptions */

/*
 * String buffer used to build the exception trace.
 */
typedef struct GE_exception_trace_buffer_struct GE_exception_trace_buffer;
struct GE_exception_trace_buffer_struct {
	char* area;
	uint32_t count;
	uint32_t capacity;
};

/*
 * Information about the feature being executed.
 */
typedef struct GE_call_struct GE_call;
struct GE_call_struct {
#ifdef GE_USE_CURRENT_IN_EXCEPTION_TRACE
	void* object; /* Current object */
#endif
	const char* class_name;
	const char* feature_name;
	GE_call* caller; /* previous feature in the call chain */
};

/*
 * Context of features containing a rescue clause.
 */
typedef struct GE_rescue_struct GE_rescue;
struct GE_rescue_struct {
	GE_jmp_buf jb;
	GE_rescue* previous; /* previous context in the call chain */
};

/*
 * Information about the execution context.
 * One such struct per thread.
 */
typedef struct GE_context_struct GE_context;
struct GE_context_struct {
	GE_call* call; /* Call stack */
	uint32_t in_assertion; /* Is an assertion evaluated? */
	GE_rescue* last_rescue; /* Context of last feature entered containing a rescue clause */
	uint32_t in_rescue; /* Number of rescue clauses currently being executed */
	EIF_REFERENCE exception_manager; /* Exception manager */
	char raising_exception; /* Is an exception currently being raised? */
	char exception_trace_enabled; /* Should exception trace be displayed? */
	long exception_code; /* Code of the exception currently being raised, 0 otherwise */
	const char* exception_tag; /* Tag of the exception currently being raised, NULL otherwise */
	GE_exception_trace_buffer exception_trace_buffer; /* String buffer used to build the exception trace */
	GE_exception_trace_buffer last_exception_trace; /* Last non-routine-failure exception trace */
	int pre_ecma_mapping_status; /* Do we map old names to new name? (i.e. STRING to STRING_8, INTEGER to INTEGER_32, ...). */
#ifdef GE_USE_THREADS
	GE_thread_context* thread; /* Thread context */
	GE_onces* process_onces; /* Cache for status and results of onces-per-process */
	GE_onces* thread_onces; /* Status and results of onces-per-thread */
	void* wel_per_thread_data; /* WEL private data */
#endif
};

/*
 * Default initialization for `GE_context'.
 */
extern GE_context GE_default_context;

/*
 * Execution context of main thread.
 */
extern GE_context* GE_main_context;

/*
 * Execution context of current thread.
 */
extern GE_context* GE_current_context(void);

/*
 * Initialization of exception handling.
 */
extern void GE_init_exception(GE_context* context);

/*
 * Free memory allocated in `a_context' for exception handling.
 */
extern void GE_free_exception(GE_context* a_context);

/*
 * Pointer to function to create a new exception manager object
 * (of type ISE_EXCEPTION_MANAGER).
 */
extern EIF_REFERENCE (*GE_new_exception_manager)(EIF_BOOLEAN);

/*
 * Pointer to Eiffel routine ISE_EXCEPTION_MANAGER.init_exception_manager.
 */
extern void (*GE_init_exception_manager)(GE_context*, EIF_REFERENCE);

/*
 * Pointer to Eiffel routine ISE_EXCEPTION_MANAGER.last_exception.
 */
extern EIF_REFERENCE (*GE_last_exception)(GE_context*, EIF_REFERENCE);

/*
 * Pointer to Eiffel routine ISE_EXCEPTION_MANAGER.once_raise.
 */
extern void (*GE_once_raise)(GE_context*, EIF_REFERENCE, EIF_REFERENCE);

/*
 * Pointer to Eiffel routine ISE_EXCEPTION_MANAGER.set_exception_data.
 */
extern void (*GE_set_exception_data)(GE_context*, EIF_REFERENCE, EIF_INTEGER_32, EIF_BOOLEAN, EIF_INTEGER_32, EIF_INTEGER_32, EIF_REFERENCE, EIF_REFERENCE, EIF_REFERENCE, EIF_REFERENCE, EIF_REFERENCE, EIF_REFERENCE, EIF_INTEGER_32, EIF_BOOLEAN);

/*
 * Exception tag associated with `a_code'.
 */
extern char* GE_exception_tag(long a_code);

/*
 * Raise an exception with code `a_code'.
 */
extern void GE_raise(long a_code);

/*
 * Raise an exception with code `a_code' and message `msg'.
 */
extern void GE_raise_with_message(long a_code, const char* msg);

/*
 * Raise an exception from EXCEPTION_MANAGER.
 */
extern void GE_developer_raise(long a_code, char* a_meaning, char* a_message);

/*
 * Raise exception which was raised the first time a once routine
 * was executed when executing it again.
 */
extern void GE_raise_once_exception(GE_context* a_context, EIF_REFERENCE a_exception);

/*
 * Exception, if any, which was last raised in `a_context'.
 */
extern EIF_REFERENCE GE_last_exception_raised(GE_context* a_context);

/*
 * Jump to execute the rescue of the last routine with a rescue
 * in the call stack.
 */
extern void GE_jump_to_last_rescue(GE_context* a_context);

/*
 * Set `in_assertion' to 'not b'.
 * Return the opposite of previous value.
 */
extern EIF_BOOLEAN GE_check_assert(EIF_BOOLEAN b);

/*
 * Check whether the type id of `obj' is not in `type_ids'.
 * If it is, then raise a CAT-call exception. Don't do anything if `obj' is Void.
 * `nb' is the number of ids in `type_ids' and is expected to be >0.
 * `type_ids' is sorted in increasing order.
 * Return `obj'.
 */
#define GE_catcall(obj,type_ids,nb) GE_check_catcall((obj),(type_ids),(nb))
extern EIF_REFERENCE GE_check_catcall(EIF_REFERENCE obj, EIF_TYPE_INDEX type_ids[], int nb);

/*
 * Check whether `obj' is Void.
 * If it is, then raise a call-on-void-target exception.
 * Return `obj'.
 */
#define GE_void(obj) (!(obj)?GE_check_void(obj):(obj))
extern EIF_REFERENCE GE_check_void(EIF_REFERENCE obj);

/*
 * Check whether `ptr' is a null pointer.
 * If it is, then raise a no-more-memory exception.
 * Return `ptr'.
 */
#define GE_null(ptr) GE_check_null(ptr)
extern void* GE_check_null(void* ptr);

#ifdef EIF_WINDOWS
/*
 * Set default exception handler.
 */
extern void GE_set_windows_exception_filter(void);
#endif

#ifdef __cplusplus
}
#endif

#endif

/*
	description:

		"C functions used to implement class THREAD and related threading facilities"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2016-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef GE_THREAD_H
#define GE_THREAD_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifdef GE_USE_THREADS

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif
#ifndef GE_THREAD_TYPES_H
#include "ge_thread_types.h"
#endif
#ifndef GE_EXCEPTION_H
#include "ge_exception.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifdef GE_USE_POSIX_THREADS
#include <unistd.h>
#endif

#ifdef GE_USE_POSIX_THREADS
#	define EIF_TSD_VAL_TYPE        void*
#	define EIF_TSD_TYPE            pthread_key_t
#	define EIF_TSD_CREATE(key,msg) \
		if (pthread_key_create(&(key),NULL)) \
	        GE_raise_with_message(GE_EX_EXT, msg)
#	define EIF_TSD_SET(key,val,msg) \
		if (pthread_setspecific((key), (EIF_TSD_VAL_TYPE)(val))) \
			GE_raise_with_message(GE_EX_EXT, msg)
#	define EIF_TSD_GET0(val_type,key,val) (val = pthread_getspecific(key))
#	define EIF_TSD_GET(val_type,key,val,msg) \
		if (EIF_TSD_GET0(val_type,key,val) == (void*) 0) GE_raise_with_message(GE_EX_EXT, msg)
#	define EIF_TSD_DESTROY(key,msg) if (pthread_key_delete(key)) GE_raise_with_message(GE_EX_EXT, msg)
#elif defined EIF_WINDOWS
#	define EIF_TSD_VAL_TYPE        LPVOID
#	define EIF_TSD_TYPE            DWORD
#	define EIF_TSD_CREATE(key,msg) \
		if ((key=TlsAlloc())==TLS_OUT_OF_INDEXES) GE_raise_with_message(GE_EX_EXT, msg)
#	define EIF_TSD_SET(key,val,msg) \
		if (!TlsSetValue((key),(EIF_TSD_VAL_TYPE)(val))) GE_raise_with_message(GE_EX_EXT, msg)
#	define EIF_TSD_GET0(val_type,key,val) \
		val=(val_type) TlsGetValue(key)
#	define EIF_TSD_GET(val_type,key,val,msg) \
		EIF_TSD_GET0(val_type,key,val); \
		if (GetLastError() != NO_ERROR) GE_raise_with_message(GE_EX_EXT, msg)
#	define EIF_TSD_DESTROY(key,msg) \
		if (!TlsFree(key)) GE_raise_with_message(GE_EX_EXT, msg)
#endif

/* Thread priority levels. */
#define EIF_MIN_THR_PRIORITY			0L
#define EIF_BELOW_NORMAL_THR_PRIORITY	100L
#define EIF_DEFAULT_THR_PRIORITY		127L
#define EIF_ABOVE_NORMAL_THR_PRIORITY	154L
#define EIF_MAX_THR_PRIORITY			255L

/*
 * Mutexes used to protect the calls to once-per-process features.
 */
typedef struct {
	EIF_POINTER* boolean_mutex;
	EIF_POINTER* character_8_mutex;
	EIF_POINTER* character_32_mutex;
	EIF_POINTER* integer_8_mutex;
	EIF_POINTER* integer_16_mutex;
	EIF_POINTER* integer_32_mutex;
	EIF_POINTER* integer_64_mutex;
	EIF_POINTER* natural_8_mutex;
	EIF_POINTER* natural_16_mutex;
	EIF_POINTER* natural_32_mutex;
	EIF_POINTER* natural_64_mutex;
	EIF_POINTER* pointer_mutex;
	EIF_POINTER* real_32_mutex;
	EIF_POINTER* real_64_mutex;
	EIF_POINTER* reference_mutex;
	EIF_POINTER* procedure_mutex;
} GE_once_mutexes;

/*
 * Mutexes used to protect the calls to once-per-process features.
 */
extern GE_once_mutexes* GE_process_once_mutexes;

/*
 * Create a new 'GE_once_mutexes' struct which can deal with the
 * numbers of once features passed as argument.
 */
extern GE_once_mutexes* GE_new_once_mutexes(
	unsigned int a_boolean_count,
	unsigned int a_character_8_count,
	unsigned int a_character_32_count,
	unsigned int a_integer_8_count,
	unsigned int a_integer_16_count,
	unsigned int a_integer_32_count,
	unsigned int a_integer_64_count,
	unsigned int a_natural_8_count,
	unsigned int a_natural_16_count,
	unsigned int a_natural_32_count,
	unsigned int a_natural_64_count,
	unsigned int a_pointer_count,
	unsigned int a_real_32_count,
	unsigned int a_real_64_count,
	unsigned int a_reference_count,
	unsigned int a_procedure_count);

/*
 * Keep track of the numbers of once-per-thread features.
 */
extern void GE_thread_onces_set_counts(
	unsigned int a_boolean_count,
	unsigned int a_character_8_count,
	unsigned int a_character_32_count,
	unsigned int a_integer_8_count,
	unsigned int a_integer_16_count,
	unsigned int a_integer_32_count,
	unsigned int a_integer_64_count,
	unsigned int a_natural_8_count,
	unsigned int a_natural_16_count,
	unsigned int a_natural_32_count,
	unsigned int a_natural_64_count,
	unsigned int a_pointer_count,
	unsigned int a_real_32_count,
	unsigned int a_real_64_count,
	unsigned int a_reference_count,
	unsigned int a_procedure_count);

/* Global mutex to protect creation of once-per-object data. */
extern EIF_POINTER GE_once_per_object_data_mutex;

/*
 * Initialize data to handle threads.
 * To be called at the beginning of the main function
 * on the main thread.
 */
extern void GE_init_thread(GE_context* a_context);

/*
 * Create a new thread with attributes `attr' and execute
 * Eiffel routine `routine' on object `current'.
 */
extern void GE_thread_create_with_attr(EIF_REFERENCE current, void (*routine)(EIF_REFERENCE), void (*set_terminated)(EIF_REFERENCE,EIF_BOOLEAN), EIF_THR_ATTR_TYPE* attr);

/*
 * Execution context of current thread.
 */
extern GE_context* GE_thread_current_context(void);

/*
 * Thead ID of current thread.
 */
extern EIF_POINTER GE_thread_id(void);

/*
 * Thread ID of last thread created from current thread.
 */
extern EIF_POINTER GE_last_thread_created(void);

#ifdef EIF_WINDOWS
/*
 * Support for Windows GUI that requires that all GUI operations are performed in the same thread.
 * Allocate new structure of the given size `a_size', assign it to `wel_per_thread_data'.
 * Return newly allocated memory block. It will be freed automatically on thread termination.
 */
extern void* GE_thread_create_wel_per_thread_data(size_t a_size);
#endif

/*
 * Waits until a child thread sets `terminated' from `obj' to True,
 * which means it is terminated. The calling thread must be the
 * direct parent of the thread, or the function might loop indefinitely.
 */
extern void GE_thread_wait(EIF_REFERENCE obj, EIF_BOOLEAN (*get_terminated)(EIF_REFERENCE));

/*
 * Waits until a child thread sets `terminated' from `obj' to True,
 * which means it is terminated, or reaching `a_timeout_ms'.
 * The calling thread must be the direct parent of the thread,
 * or the function might loop indefinitely.
 */
extern EIF_BOOLEAN GE_thread_wait_with_timeout(EIF_REFERENCE obj, EIF_BOOLEAN (*get_terminated)(EIF_REFERENCE), EIF_NATURAL_64 a_timeout_ms);

/*
 * Yields execution to other threads.
 */
extern void GE_thread_yield(void);

/*
 * The calling thread waits for all other children threads to terminate.
 */
extern void GE_thread_join_all(void);

/*
 * Function called to terminate a thread launched by Eiffel with `GE_thread_create_with_attr'.
 * This function must be called from the thread itself (not the parent).
 */
extern void GE_thread_exit(void);

/*
 * Default thread priority level.
 */
extern EIF_INTEGER GE_thread_default_priority(void);

/*
 * Minimum thread priority level.
 */
extern EIF_INTEGER GE_thread_min_priority(void);

/*
 * Maximum thread priority level.
 */
extern EIF_INTEGER GE_thread_max_priority(void);

/*
 * Create a new mutex.
 */
extern EIF_POINTER GE_mutex_create(void);

/*
 * Lock mutex.
 */
extern void GE_mutex_lock(EIF_POINTER a_mutex);

/*
 * Try to lock mutex. Return True on success.
 */
extern EIF_BOOLEAN GE_mutex_try_lock(EIF_POINTER a_mutex);

/*
 * Unlock mutex.
 */
extern void GE_mutex_unlock(EIF_POINTER a_mutex);

/*
 * Destroy and free all resources used by mutex.
 */
extern void GE_mutex_destroy(EIF_POINTER a_mutex);

/*
 * Create a new semaphore allowing `a_count' threads
 * to go into a critical section.
 */
extern EIF_POINTER GE_semaphore_create(EIF_INTEGER a_count);

/*
 * Decrement semaphore count, waiting if necessary
 * until that becomes possible.
 */
extern void GE_semaphore_wait(EIF_POINTER a_semaphore);

/*
 * Has client been successful in decrementing semaphore
 * count without waiting?
 */
extern EIF_BOOLEAN GE_semaphore_try_wait(EIF_POINTER a_semaphore);

/*
 * Increment semaphore count.
 */
extern void GE_semaphore_post(EIF_POINTER a_semaphore);

/*
 * Destroy and free all resources used by semaphore.
 */
extern void GE_semaphore_destroy(EIF_POINTER a_semaphore);

/*
 * Create a new condition variable.
 */
extern EIF_POINTER GE_condition_variable_create(void);

/*
 * Unblock all threads blocked on condition variable.
 */
extern void GE_condition_variable_broadcast(EIF_POINTER a_condition_variable);

/*
 * Unblock one thread blocked on condition variable.
 */
extern void GE_condition_variable_signal(EIF_POINTER a_condition_variable);

/*
 * Block calling thread on condition variable.
 */
extern void GE_condition_variable_wait(EIF_POINTER a_condition_variable, EIF_POINTER a_mutex);

/*
 * Block calling thread on condition variable for at most `a_timeout' milliseconds.
 * Return 1 is we got the condition variable on time, otherwise return 0.
 */
extern EIF_INTEGER GE_condition_variable_wait_with_timeout(EIF_POINTER a_condition_variable, EIF_POINTER a_mutex, EIF_INTEGER a_timeout);

/*
 * Destroy and free all resources used by condition variable.
 */
extern void GE_condition_variable_destroy(EIF_POINTER a_condition_variable);

/*
 * Create a new read-write lock.
 */
extern EIF_POINTER GE_read_write_lock_create(void);

/*
 * Acquire a read lock. Multiple readers can go if there are no writer.
 */
extern void GE_read_write_lock_read_lock(EIF_POINTER a_read_write_lock);

/*
 * Acquire a write lock. Only a single write can proceed.
 */
extern void GE_read_write_lock_write_lock(EIF_POINTER a_read_write_lock);

/*
 * Unlock a read or write lock.
 */
extern void GE_read_write_lock_unlock(EIF_POINTER a_read_write_lock);

/*
 * Destroy and free all resources used by read-write lock.
 */
extern void GE_read_write_lock_destroy(EIF_POINTER a_read_write_lock);

#ifdef __cplusplus
}
#endif

#endif
#endif

/*
	description:

		"C functions used to manipulate native strings"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2013-2018, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef GE_NATIVE_STRING_H
#define GE_NATIVE_STRING_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#include <string.h>
#ifdef EIF_WINDOWS
#ifdef __LCC__
/* With lcc-win32, stat.h should be included before wchar.h. */
#include <sys/stat.h>
#endif
#include <wchar.h>
#else
#include <sys/types.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifdef EIF_WINDOWS

/* Macro used to manipulate native strings, i.e: (wchar_t*) */
#define GE_nstrlen wcslen /* size of string */
#define GE_nstrncpy wcsncpy /* Copy n characters of one string to another */
#define GE_nstrcpy wcscpy /* Copy one string to another */
#define GE_nstrncat wcsncat /* Append characters of a string */
#define GE_nstrcat wcscat /* Append a string */
#define GE_nstrstr wcsstr /* Return a pointer to the first occurrence of a search string in a string. */
#define GE_nmakestr(quote) L##quote /* Manifest Native string declaration */
#define GE_nstr_fopen _wfopen /* Open file using native string name */
#define GE_nstrcmp wcscmp /* Compare two strings. */
#define GE_nstrdup _wcsdup /* Duplicate string. */
#define GE_nstr_cat_ascii(dest, src) {							\
		int i;													\
		size_t dest_len, src_len;								\
		dest_len = rt_nstrlen (dest);							\
		src_len = strlen (src);									\
		for (i = 0; i < src_len; i++) {							\
			dest[dest_len + i] = (EIF_NATIVE_CHAR)src[i];		\
		}														\
		dest[dest_len + src_len] = (EIF_NATIVE_CHAR)0;			\
	}

#else /* not EIF_WINDOWS */

/* Macro used to manipulate native strings, i.e: (char*) */
#define GE_nstrlen strlen /* size of string */
#define GE_nstrncpy strncpy /* Copy n characters of one string to another */
#define GE_nstrcpy strcpy /* Copy one string to another */
#define GE_nstrncat strncat /* Append characters of a string */
#define GE_nstrcat strcat /* Append a string */
#define GE_nstrstr strstr /* Return a pointer to the first occurrence of a search string in a string. */
#define GE_nmakestr(quote) quote /* Manifest Native string declaration */
#define GE_nstr_fopen fopen /* Open file using native string name */
#define GE_nstrcmp strcmp /* Compare two strings. */
#define GE_nstrdup strdup /* Duplicate string. */
#define GE_nstr_cat_ascii strcat

#endif


#ifdef __cplusplus
}
#endif

#endif

/*
	description:

		"C functions used to implement class ARGUMENTS"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2007-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef GE_ARGUMENTS_H
#define GE_ARGUMENTS_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

extern int GE_argc;
extern EIF_NATIVE_CHAR** GE_argv;

#ifdef __cplusplus
}
#endif

#endif

/*
	description:

		"C functions used to implement type information"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2016-2018, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef GE_TYPES_H
#define GE_TYPES_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif
#ifndef GE_EXCEPTION_H
#include "ge_exception.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Type annotations.
 * When a type has no annotation, it means a detachable, non-separate, variant type.
 * In all other cases, there will be an annotation.
 */
#define ANNOTATION_MASK			0x007F	/* All possible annotations. */
#define ATTACHED_FLAG			0x0001
#define DETACHABLE_FLAG			0x0002	/* Only present when overriding an attached type. */
#define SEPARATE_FLAG			0x0004
#define VARIANT_FLAG			0x0008	/* Only present when overriding a frozen/poly type. */
#define UNUSABLE_FLAG			0x0010	/* Reserved for backward compatibility for storables. */
#define FROZEN_FLAG				0x0020
#define POLY_FLAG				0x0040

/*
 * Type flags.
 */
#define GE_TYPE_FLAG_SPECIAL		0x0010
#define GE_TYPE_FLAG_TUPLE			0x0020
#define GE_TYPE_FLAG_EXPANDED		0x0040
#define GE_TYPE_FLAG_DEFERRED		0x0080
#define GE_TYPE_FLAG_NONE			0x0100
#define GE_TYPE_FLAG_BASIC_MASK		0x000F /* One of "BOOLEAN", "CHARACTER_8", "CHARACTER_32", "INTEGER_8", "INTEGER_16", "INTEGER_32", "INTEGER_64", "NATURAL_8", "NATURAL_16", "NATURAL_32", "NATURAL_64", "POINTER", "REAL_32", "REAL_64" */
#define GE_TYPE_FLAG_BOOLEAN		0x0001
#define GE_TYPE_FLAG_CHARACTER_8	0x0002
#define GE_TYPE_FLAG_CHARACTER_32	0x0003
#define GE_TYPE_FLAG_INTEGER_8		0x0004
#define GE_TYPE_FLAG_INTEGER_16		0x0005
#define GE_TYPE_FLAG_INTEGER_32		0x0006
#define GE_TYPE_FLAG_INTEGER_64		0x0007
#define GE_TYPE_FLAG_NATURAL_8		0x0008
#define GE_TYPE_FLAG_NATURAL_16		0x0009
#define GE_TYPE_FLAG_NATURAL_32		0x000A
#define GE_TYPE_FLAG_NATURAL_64		0x000B
#define GE_TYPE_FLAG_POINTER		0x000C
#define GE_TYPE_FLAG_REAL_32		0x000D
#define GE_TYPE_FLAG_REAL_64		0x000E

/*
 * Convention for attribute types.
 * The values are in sync with REFLECTOR_CONSTANTS.
 */
#define GE_TYPE_KIND_INVALID		-1
#define GE_TYPE_KIND_POINTER		0
#define GE_TYPE_KIND_REFERENCE		1
#define GE_TYPE_KIND_CHARACTER_8	2
#define GE_TYPE_KIND_BOOLEAN		3
#define GE_TYPE_KIND_INTEGER_32		4
#define GE_TYPE_KIND_REAL_32		5
#define GE_TYPE_KIND_REAL_64		6
#define GE_TYPE_KIND_EXPANDED		7
#define GE_TYPE_KIND_INTEGER_8		9
#define GE_TYPE_KIND_INTEGER_16		10
#define GE_TYPE_KIND_INTEGER_64 	11
#define GE_TYPE_KIND_CHARACTER_32	12
#define GE_TYPE_KIND_NATURAL_8		13
#define GE_TYPE_KIND_NATURAL_16		14
#define GE_TYPE_KIND_NATURAL_32 	15
#define GE_TYPE_KIND_NATURAL_64 	16

/*
 * Object flags.
 */
#define GE_OBJECT_FLAG_MARKED		0x0001

/*
 * Ancestor relationship between two types X and Y.
 */
#ifdef GE_USE_ANCESTORS
typedef struct {
	EIF_TYPE_INDEX type_id; /* Type id of Y */
	EIF_BOOLEAN conforms; /* Does X conform to Y? */
	void (**qualified_calls)(); /* Function pointers, indexed by call id, when the static type of the target is Y and the dynamic type is X */
} GE_ancestor;
#endif

/*
 * Attribute.
 */
#ifdef GE_USE_ATTRIBUTES
typedef struct {
#ifdef GE_USE_ATTRIBUTE_NAME
	const char* name; /* Attribute name */
#endif
#ifdef GE_USE_ATTRIBUTE_TYPE_ID
	EIF_ENCODED_TYPE type_id; /* Static type id */
#endif
#ifdef GE_USE_ATTRIBUTE_OFFSET
	uint32_t offset; /* Address offset in object */
#endif
} GE_attribute;
#endif

/*
 * Type information.
 */
typedef struct {
	EIF_TYPE_INDEX type_id;
	uint16_t flags;
#ifdef GE_USE_TYPE_GENERATOR
	const char* generator; /* Generator class name */
#endif
#ifdef GE_USE_TYPE_NAME
	const char* name; /* Full type name */
#endif
#ifdef GE_USE_TYPE_GENERIC_PARAMETERS
	EIF_ENCODED_TYPE* generic_parameters;
	uint32_t generic_parameter_count;
#endif
#ifdef GE_USE_ANCESTORS
	GE_ancestor** ancestors;
	uint32_t ancestor_count;
#endif
#ifdef GE_USE_ATTRIBUTES
	GE_attribute** attributes;
	uint32_t attribute_count;
#endif
#ifdef GE_USE_TYPE_OBJECT_SIZE
	uint64_t object_size;
#endif
	EIF_REFERENCE (*new_instance)();
	void (*dispose)(GE_context*, EIF_REFERENCE);
} GE_type_info;

typedef struct {
	EIF_TYPE_INDEX id; /* Type id of the "TYPE [X]" object */
	EIF_INTEGER type_id; /* Type id of the type "X" */
	EIF_BOOLEAN is_special;
	void (*dispose)(GE_context*, EIF_REFERENCE);
	EIF_REFERENCE a1; /* internal_name */
	EIF_REFERENCE a2; /* internal_name_32 */
} EIF_TYPE_OBJ;

/*
 * Types indexed by type id.
 * Generated by the compiler.
 */
extern EIF_TYPE_OBJ GE_types[][2];
extern GE_type_info GE_type_infos[];

/*
 * Number of type infos in `GE_type_infos'.
 * Do not take into account the fake item at index 0.
 */
extern int GE_type_info_count;

/*
 * Encode a EIF_TYPE into a EIF_ENCODED_TYPE.
 * The lower part of EIF_ENCODED_TYPE contains the .id field,
 * and the upper part the .annotations.
 */
extern EIF_ENCODED_TYPE GE_encoded_type(EIF_TYPE a_type);

/*
 * Decode a EIF_ENCODED_TYPE into a EIF_TYPE.
 * The lower part of EIF_ENCODED_TYPE contains the .id field,
 * and the upper part the .annotations.
 */
extern EIF_TYPE GE_decoded_type(EIF_ENCODED_TYPE a_type);

/*
 * Type with `a_id' and `a_annotations'.
 */
extern EIF_TYPE GE_new_type(EIF_TYPE_INDEX a_id, EIF_TYPE_INDEX a_annotations);

/*
 * Type of object `obj'.
 */
#define GE_object_type(obj)	GE_new_type(((EIF_REFERENCE)(obj))->id, 0x0)
#define GE_object_encoded_type(obj) GE_encoded_type(GE_object_type(obj))

/*
 * Attachment status of `a_type'.
 */
#define GE_is_attached_type(a_type) EIF_TEST(((a_type).annotations & ATTACHED_FLAG) || GE_is_expanded_type_index((a_type).id))
#define GE_is_attached_encoded_type(a_type) GE_is_attached_type(GE_decoded_type(a_type))

/*
 * Associated detachable type of `a_type' if any,
 * otherwise `a_type'.
 */
extern EIF_TYPE GE_non_attached_type(EIF_TYPE a_type);
#define GE_non_attached_encoded_type(a_type) GE_encoded_type(GE_non_attached_type(GE_decoded_type(a_type)))

/*
 * Associated attached type of `a_type' if any,
 * otherwise `a_type'.
 */
extern EIF_TYPE GE_attached_type(EIF_TYPE a_type);
#define GE_attached_encoded_type(t) GE_encoded_type(GE_attached_type(GE_decoded_type(t)))

/*
 * Is `a_type' a SPECIAL type?
 */
#define GE_is_special_type_index(a_type) EIF_TEST(GE_type_infos[a_type].flags & GE_TYPE_FLAG_SPECIAL)
#define GE_is_special_encoded_type(a_type) GE_is_special_type_index(GE_decoded_type(a_type).id)
#define GE_is_special_object(obj) GE_is_special_type_index(((EIF_REFERENCE)(obj))->id)

/*
 * Is `a_type' a SPECIAL type of user-defined expanded type?
 */
extern EIF_BOOLEAN GE_is_special_of_expanded_type_index(EIF_TYPE_INDEX a_type);
#define GE_is_special_of_expanded_encoded_type(a_type) GE_is_special_of_expanded_type_index(GE_decoded_type(a_type).id)
#define GE_is_special_of_expanded_object(obj) GE_is_special_of_expanded_type_index(((EIF_REFERENCE)(obj))->id)

/*
 * Is `a_type' a SPECIAL type of reference type?
 */
extern EIF_BOOLEAN GE_is_special_of_reference_type_index(EIF_TYPE_INDEX a_type);
#define GE_is_special_of_reference_encoded_type(a_type) GE_is_special_of_reference_type_index(GE_decoded_type(a_type).id)
#define GE_is_special_of_reference_object(obj) GE_is_special_of_reference_type_index(((EIF_REFERENCE)(obj))->id)

/*
 * Is `a_type' a SPECIAL type of reference type or basic expanded type?
 * (Note that user-defined expanded types are excluded.)
 */
extern EIF_BOOLEAN GE_is_special_of_reference_or_basic_expanded_type_index(EIF_TYPE_INDEX a_type);
#define GE_is_special_of_reference_or_basic_expanded_encoded_type(a_type) GE_is_special_of_reference_or_basic_expanded_type_index(GE_decoded_type(a_type).id)
#define GE_is_special_of_reference_or_basic_expanded_object(obj) GE_is_special_of_reference_or_basic_expanded_type_index(((EIF_REFERENCE)(obj))->id)

/*
 * Is `a_type' a TUPLE type?
 */
#define GE_is_tuple_type_index(a_type) EIF_TEST(GE_type_infos[a_type].flags & GE_TYPE_FLAG_TUPLE)
#define GE_is_tuple_encoded_type(a_type) GE_is_tuple_type_index(GE_decoded_type(a_type).id)
#define GE_is_tuple_object(obj) GE_is_tuple_type_index(((EIF_REFERENCE)(obj))->id)

/*
 * Is `a_type' an expanded type?
 */
#define GE_is_expanded_type_index(a_type) EIF_TEST(GE_type_infos[a_type].flags & GE_TYPE_FLAG_EXPANDED)
#define GE_is_expanded_encoded_type(a_type) GE_is_expanded_type_index(GE_decoded_type(a_type).id)
#define GE_is_expanded_object(obj) GE_is_expanded_type_index(((EIF_REFERENCE)(obj))->id)

/*
 * Is `a_type' a type whose base class is deferred?
 */
#define GE_is_deferred_type_index(a_type) EIF_TEST(GE_type_infos[a_type].flags & GE_TYPE_FLAG_DEFERRED)
#define GE_is_deferred_encoded_type(a_type) GE_is_deferred_type_index(GE_decoded_type(a_type).id)

/*
 * Does `i'-th field of `a_object + a_physical_offset' (which is expected to be reference)
 * denote a reference with copy semantics?
 */
extern EIF_BOOLEAN GE_is_copy_semantics_field(EIF_INTEGER i, EIF_POINTER a_object, EIF_INTEGER a_physical_offset);

/*
 * Does `i'-th item of special `a_object' (which is expected to be reference)
 * denote a reference with copy semantics?
 */
extern EIF_BOOLEAN GE_is_special_copy_semantics_item(EIF_INTEGER i, EIF_POINTER a_object);

/*
 * Generator class name of `a_type'.
 */
extern EIF_REFERENCE GE_generator_of_type_index(EIF_TYPE_INDEX a_type);
#define GE_generator_of_encoded_type(a_type) GE_generator_of_type_index(GE_decoded_type(a_type).id)
extern EIF_REFERENCE GE_generator_8_of_type_index(EIF_TYPE_INDEX a_type);
#define GE_generator_8_of_encoded_type(a_type) GE_generator_8_of_type_index(GE_decoded_type(a_type).id)

/*
 * Full name of `a_type'.
 */
extern EIF_REFERENCE GE_generating_type_of_encoded_type(EIF_ENCODED_TYPE a_type);
extern EIF_REFERENCE GE_generating_type_8_of_encoded_type(EIF_ENCODED_TYPE a_type);

/*
 * Encoded type whose name is `a_name'.
 * -1 if no such type.
 */
extern EIF_ENCODED_TYPE GE_encoded_type_from_name(EIF_POINTER a_name);

/*
 * Does `a_type_1' conform to `a_type_2'?
 */
extern EIF_BOOLEAN GE_encoded_type_conforms_to(EIF_ENCODED_TYPE a_type_1, EIF_ENCODED_TYPE a_type_2);

/*
 * Number of generic parameters.
 */
extern EIF_INTEGER GE_generic_parameter_count_of_type_index(EIF_TYPE_INDEX a_type);
#define GE_generic_parameter_count_of_encoded_type(a_type) GE_generic_parameter_count_of_type_index(GE_decoded_type(a_type).id)

/*
 * Type of `i'-th generic parameter of `a_type'.
 */
extern EIF_INTEGER GE_generic_parameter_of_type_index(EIF_TYPE_INDEX a_type, EIF_INTEGER i);
#define GE_generic_parameter_of_encoded_type(a_type,i) GE_generic_parameter_of_type_index(GE_decoded_type(a_type).id, (i))

/*
 * Number of fields of an object of dynamic type `a_type'.
 */
extern EIF_INTEGER GE_field_count_of_type_index(EIF_TYPE_INDEX a_type);
#define GE_field_count_of_encoded_type(a_type) GE_field_count_of_type_index(GE_decoded_type(a_type).id)

/*
 * Physical offset of the `i'-th field for an object of dynamic type `a_type'.
 */
extern EIF_INTEGER GE_field_offset_of_type_index(EIF_INTEGER i, EIF_TYPE_INDEX a_type);
#define GE_field_offset_of_encoded_type(i, a_type) GE_field_offset_of_type_index((i), GE_decoded_type(a_type).id)

/*
 * Name of the `i'-th field for an object of dynamic type `a_type'.
 */
extern EIF_POINTER GE_field_name_of_type_index(EIF_INTEGER i, EIF_TYPE_INDEX a_type);
#define GE_field_name_of_encoded_type(i, a_type) GE_field_name_of_type_index((i), GE_decoded_type(a_type).id)

/*
 * Static type of the `i'-th field for an object of dynamic type `a_type'.
 */
extern EIF_INTEGER GE_field_static_type_of_type_index(EIF_INTEGER i, EIF_TYPE_INDEX a_type);
#define GE_field_static_type_of_encoded_type(i, a_type) GE_field_static_type_of_type_index((i), GE_decoded_type(a_type).id)

/*
 * Kind of type of the `i'-th field for an object of dynamic type `a_type'.
 */
extern EIF_INTEGER GE_field_type_kind_of_type_index(EIF_INTEGER i, EIF_TYPE_INDEX a_type);
#define GE_field_type_kind_of_encoded_type(i, a_type) GE_field_type_kind_of_type_index((i), GE_decoded_type(a_type).id)

/*
 * Physical size of `a_object'.
 */
extern EIF_NATURAL_64 GE_object_size(EIF_POINTER a_object);

/*
 * Is `i'-th field of objects of type `a_type' a user-defined expanded attribute?
 */
extern EIF_BOOLEAN GE_is_field_expanded_of_type_index(EIF_INTEGER i, EIF_TYPE_INDEX a_type);
#define GE_is_field_expanded_of_encoded_type(i, a_type) GE_is_field_expanded_of_type_index((i), GE_decoded_type(a_type).id)

#define GE_field_address_at(a_field_offset, a_object, a_physical_offset) ((char*)(a_object) + (a_physical_offset) + (a_field_offset))
#define GE_object_at_offset(a_enclosing, a_physical_offset) (EIF_REFERENCE)(GE_field_address_at(0, (a_enclosing), (a_physical_offset)))
#define GE_raw_object_at_offset(a_enclosing, a_physical_offset) (EIF_POINTER)(GE_field_address_at(0, (a_enclosing), (a_physical_offset)))
#define GE_object_encoded_type_at_offset(a_enclosing, a_physical_offset) GE_object_encoded_type(GE_raw_object_at_offset((a_enclosing), (a_physical_offset)))
#define GE_boolean_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_BOOLEAN*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_character_8_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_CHARACTER_8*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_character_32_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_CHARACTER_32*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_integer_8_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_INTEGER_8*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_integer_16_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_INTEGER_16*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_integer_32_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_INTEGER_32*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_integer_64_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_INTEGER_64*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_natural_8_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_NATURAL_8*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_natural_16_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_NATURAL_16*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_natural_32_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_NATURAL_32*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_natural_64_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_NATURAL_64*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_pointer_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_POINTER*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_real_32_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_REAL_32*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_real_64_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_REAL_64*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_raw_reference_field_at(a_field_offset, a_object, a_physical_offset) (EIF_POINTER)*(EIF_REFERENCE*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_reference_field_at(a_field_offset, a_object, a_physical_offset) *(EIF_REFERENCE*)(GE_field_address_at((a_field_offset), (a_object), (a_physical_offset)))
#define GE_set_boolean_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_boolean_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_character_8_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_character_8_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_character_32_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_character_32_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_integer_8_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_integer_8_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_integer_16_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_integer_16_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_integer_32_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_integer_32_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_integer_64_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_integer_64_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_natural_8_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_natural_8_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_natural_16_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_natural_16_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_natural_32_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_natural_32_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_natural_64_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_natural_64_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_pointer_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_pointer_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_real_32_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_real_32_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_real_64_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_real_64_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value
#define GE_set_reference_field_at(a_field_offset, a_object, a_physical_offset, a_value) GE_reference_field_at((a_field_offset), (a_object), (a_physical_offset)) = a_value

#if defined(GE_USE_ATTRIBUTES) && defined(GE_USE_ATTRIBUTE_OFFSET)
#define GE_field_address(i, a_object, a_physical_offset) GE_field_address_at(GE_type_infos[((EIF_REFERENCE)(a_object))->id].attributes[(i) - 1]->offset, (a_object), (a_physical_offset))
#define GE_boolean_field(i, a_object, a_physical_offset) *(EIF_BOOLEAN*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_character_8_field(i, a_object, a_physical_offset) *(EIF_CHARACTER_8*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_character_32_field(i, a_object, a_physical_offset) *(EIF_CHARACTER_32*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_integer_8_field(i, a_object, a_physical_offset) *(EIF_INTEGER_8*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_integer_16_field(i, a_object, a_physical_offset) *(EIF_INTEGER_16*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_integer_32_field(i, a_object, a_physical_offset) *(EIF_INTEGER_32*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_integer_64_field(i, a_object, a_physical_offset) *(EIF_INTEGER_64*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_natural_8_field(i, a_object, a_physical_offset) *(EIF_NATURAL_8*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_natural_16_field(i, a_object, a_physical_offset) *(EIF_NATURAL_16*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_natural_32_field(i, a_object, a_physical_offset) *(EIF_NATURAL_32*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_natural_64_field(i, a_object, a_physical_offset) *(EIF_NATURAL_64*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_pointer_field(i, a_object, a_physical_offset) *(EIF_POINTER*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_real_32_field(i, a_object, a_physical_offset) *(EIF_REAL_32*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_real_64_field(i, a_object, a_physical_offset) *(EIF_REAL_64*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_reference_field(i, a_object, a_physical_offset) *(EIF_REFERENCE*)(GE_field_address((i), (a_object), (a_physical_offset)))
#define GE_set_boolean_field(i, a_object, a_physical_offset, a_value) GE_boolean_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_character_8_field(i, a_object, a_physical_offset, a_value) GE_character_8_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_character_32_field(i, a_object, a_physical_offset, a_value) GE_character_32_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_integer_8_field(i, a_object, a_physical_offset, a_value) GE_integer_8_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_integer_16_field(i, a_object, a_physical_offset, a_value) GE_integer_16_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_integer_32_field(i, a_object, a_physical_offset, a_value) GE_integer_32_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_integer_64_field(i, a_object, a_physical_offset, a_value) GE_integer_64_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_natural_8_field(i, a_object, a_physical_offset, a_value) GE_natural_8_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_natural_16_field(i, a_object, a_physical_offset, a_value) GE_natural_16_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_natural_32_field(i, a_object, a_physical_offset, a_value) GE_natural_32_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_natural_64_field(i, a_object, a_physical_offset, a_value) GE_natural_64_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_pointer_field(i, a_object, a_physical_offset, a_value) GE_pointer_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_real_32_field(i, a_object, a_physical_offset, a_value) GE_real_32_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_real_64_field(i, a_object, a_physical_offset, a_value) GE_real_64_field((i), (a_object), (a_physical_offset)) = (a_value)
#define GE_set_reference_field(i, a_object, a_physical_offset, a_value) GE_reference_field((i), (a_object), (a_physical_offset)) = (a_value)
#else
#define GE_boolean_field(i, a_object, a_physical_offset) (EIF_BOOLEAN)0
#define GE_character_8_field(i, a_object, a_physical_offset) (EIF_CHARACTER_8)0
#define GE_character_32_field(i, a_object, a_physical_offset) (EIF_CHARACTER_32)0
#define GE_integer_8_field(i, a_object, a_physical_offset) (EIF_INTEGER_8)0
#define GE_integer_16_field(i, a_object, a_physical_offset) (EIF_INTEGER_16)0
#define GE_integer_32_field(i, a_object, a_physical_offset) (EIF_INTEGER_32)0
#define GE_integer_64_field(i, a_object, a_physical_offset) (EIF_INTEGER_64)0
#define GE_natural_8_field(i, a_object, a_physical_offset) (EIF_NATURAL_8)0
#define GE_natural_16_field(i, a_object, a_physical_offset) (EIF_NATURAL_16)0
#define GE_natural_32_field(i, a_object, a_physical_offset) (EIF_NATURAL_32)0
#define GE_natural_64_field(i, a_object, a_physical_offset) (EIF_NATURAL_64)0
#define GE_pointer_field(i, a_object, a_physical_offset) (EIF_POINTER)0
#define GE_real_32_field(i, a_object, a_physical_offset) (EIF_REAL_32)0
#define GE_real_64_field(i, a_object, a_physical_offset) (EIF_REAL_64)0
#define GE_reference_field(i, a_object, a_physical_offset) (EIF_REFERENCE)0
#define GE_set_boolean_field(i, a_object, a_physical_offset, a_value)
#define GE_set_character_8_field(i, a_object, a_physical_offset, a_value)
#define GE_set_character_32_field(i, a_object, a_physical_offset, a_value)
#define GE_set_integer_8_field(i, a_object, a_physical_offset, a_value)
#define GE_set_integer_16_field(i, a_object, a_physical_offset, a_value)
#define GE_set_integer_32_field(i, a_object, a_physical_offset, a_value)
#define GE_set_integer_64_field(i, a_object, a_physical_offset, a_value)
#define GE_set_natural_8_field(i, a_object, a_physical_offset, a_value)
#define GE_set_natural_16_field(i, a_object, a_physical_offset, a_value)
#define GE_set_natural_32_field(i, a_object, a_physical_offset, a_value)
#define GE_set_natural_64_field(i, a_object, a_physical_offset, a_value)
#define GE_set_pointer_field(i, a_object, a_physical_offset, a_value)
#define GE_set_real_32_field(i, a_object, a_physical_offset, a_value)
#define GE_set_real_64_field(i, a_object, a_physical_offset, a_value)
#define GE_set_reference_field(i, a_object, a_physical_offset, a_value)
#endif

/*
 * Number of non-transient fields of an object of dynamic type `a_type'.
 * TODO: storable not implemented yet.
 */
#define GE_persistent_field_count_of_type_index(a_type) GE_field_count_of_type_index(a_type)
#define GE_persistent_field_count_of_encoded_type(a_type) GE_persistent_field_count_of_type_index(GE_decoded_type(a_type).id)

/*
 * Is `i'-th field of objects of type `a_type' a transient field?
 * TODO: storable not implemented yet.
 */
#define GE_is_field_transient_of_type_index(i, a_type) EIF_FALSE
#define GE_is_field_transient_of_encoded_type(i, a_type) GE_is_field_transient_of_type_index((i), GE_decoded_type(a_type).id)

/*
 * Storable version of objects of type `a_type'.
 * TODO: storable not implemented yet.
 */
#define GE_storable_version_of_type_index(a_type) EIF_VOID
#define GE_storable_version_of_encoded_type(a_type) GE_storable_version_of_type_index(GE_decoded_type(a_type).id)

/*
 * Get a lock on `GE_mark_object' and `GE_unmark_object' routines so that
 * 2 threads cannot `GE_mark_object' and `GE_unmark_object' at the same time.
 */
extern void GE_lock_marking(void);

/*
 * Release a lock on `GE_mark_object' and `GE_unmark_object', so that another
 * thread can use `GE_mark_object' and `GE_unmark_object'.
 */
extern void GE_unlock_marking(void);

/*
 * Is `obj' marked?
 */
extern EIF_BOOLEAN GE_is_object_marked(EIF_POINTER obj);

/*
 * Mark `obj'.
 */
extern void GE_mark_object(EIF_POINTER obj);

/*
 * Unmark `obj'.
 */
extern void GE_unmark_object(EIF_POINTER obj);

/*
 * New instance of dynamic `a_type'.
 * Note: returned object is not initialized and may
 * hence violate its invariant.
 * `a_type' cannot represent a SPECIAL type, use
 * `GE_new_special_of_reference_instance_of_type_index' instead.
 */
extern EIF_REFERENCE GE_new_instance_of_type_index(EIF_TYPE_INDEX a_type);
#define GE_new_instance_of_encoded_type(a_type) GE_new_instance_of_type_index(GE_decoded_type(a_type).id)

/*
 * New instance of dynamic `a_type' that represents
 * a SPECIAL with can contain `a_capacity' elements of reference type.
 * To create a SPECIAL of basic type, use class SPECIAL directly.
 */
extern EIF_REFERENCE GE_new_special_of_reference_instance_of_type_index(EIF_TYPE_INDEX a_type, EIF_INTEGER a_capacity);
#define GE_new_special_of_reference_instance_of_encoded_type(a_type, a_capacity) GE_new_special_of_reference_instance_of_type_index(GE_decoded_type(a_type).id, (a_capacity))

/*
 * New instance of tuple of type `a_type'.
 * Note: returned object is not initialized and may
 * hence violate its invariant.
 */
#define GE_new_tuple_instance_of_type_index(a_type) GE_new_instance_of_type_index(a_type)
#define GE_new_tuple_instance_of_encoded_type(a_type) GE_new_tuple_instance_of_type_index(GE_decoded_type(a_type).id)

/*
 * New instance of TYPE for object of type `a_type'.
 */
extern EIF_REFERENCE GE_new_type_instance_of_encoded_type(EIF_ENCODED_TYPE a_type);

#ifdef __cplusplus
}
#endif

#endif

/*
	description:

		"C functions used to manipulate strings"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2016-2019, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef GE_STRING_H
#define GE_STRING_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

/*
 * New Eiffel empty string of type "STRING_8" with can
 * contain `c' characters.
 * Note: The implementation of this function is generated
 * by the Eiffel compiler.
 */
extern EIF_REFERENCE GE_new_str8(EIF_INTEGER c);

/*
 * New Eiffel empty string of type "IMMUTABLE_STRING_8" with can
 * contain `c' characters.
 * Note: The implementation of this function is generated
 * by the Eiffel compiler.
 */
extern EIF_REFERENCE GE_new_istr8(EIF_INTEGER c);

/*
 * New Eiffel empty string of type "STRING_32" with can
 * contain `c' characters.
 * Note: The implementation of this function is generated
 * by the Eiffel compiler.
 */
extern EIF_REFERENCE GE_new_str32(EIF_INTEGER c);

/*
 * New Eiffel empty string of type "IMMUTABLE_STRING_32" with can
 * contain `c' characters.
 * Note: The implementation of this function is generated
 * by the Eiffel compiler.
 */
extern EIF_REFERENCE GE_new_istr32(EIF_INTEGER c);

/*
 * New Eiffel string of type "STRING_8" containing the
 * first `c' characters found in ISO 8859-1 string `s'.
 */
extern EIF_REFERENCE GE_ms8(const char* s, EIF_INTEGER c);

/*
 * New Eiffel string of type "STRING_8" containing all
 * characters found in the null-terminated ISO 8859-1 string `s'.
 */
extern EIF_REFERENCE GE_str8(const char* s);

/*
 * New Eiffel string of type "IMMUTABLE_STRING_8" containing the
 * first `c' characters found in ISO 8859-1 string `s'.
 */
extern EIF_REFERENCE GE_ims8(const char* s, EIF_INTEGER c);

/*
 * New Eiffel string of type "STRING_32" containing the
 * first `c' characters found in ISO 8859-1 string `s'.
 */
extern EIF_REFERENCE GE_ms32(const char* s, EIF_INTEGER c);

/*
 * New Eiffel string of type "STRING_32" containing the
 * first `c' 32-bit characters built from `s' by reading
 * groups of four bytes with little-endian byte order.
 */
extern EIF_REFERENCE GE_ms32_from_utf32le(const char* s, EIF_INTEGER c);

/*
 * New Eiffel string of type "STRING_32" containing all
 * characters found in the null-terminated ISO 8859-1 string `s'.
 */
extern EIF_REFERENCE GE_str32(const char* s);

/*
 * New Eiffel string of type "IMMUTABLE_STRING_32" containing
 * the first `c' characters found in ISO 8859-1 string `s'.
 */
extern EIF_REFERENCE GE_ims32(const char* s, EIF_INTEGER c);

/*
 * New Eiffel string of type "IMMUTABLE_STRING_32" containing the
 * first `c' 32-bit characters built from `s' by reading
 * groups of four bytes with little-endian byte order.
 */
extern EIF_REFERENCE GE_ms32_from_utf32le(const char* s, EIF_INTEGER c);

/*
 * New Eiffel string of type "IMMUTABLE_STRING_32" containing all
 * characters found in the null-terminated ISO 8859-1 string `s'.
 */
extern EIF_REFERENCE GE_istr32(const char* s);

/*
 * New Eiffel string of type "IMMUTABLE_STRING_32" containing the
 * first `n' native characters found in native string `s'.
 * Invalid native characters are escaped.
 */
extern EIF_REFERENCE GE_ims32_from_nstr(EIF_NATIVE_CHAR* s, EIF_INTEGER n);

/*
 * New Eiffel string of type "IMMUTABLE_STRING_32" containing all
 * characters found in the null-terminated native string `s'.
 * Invalid native characters are escaped.
 */
extern EIF_REFERENCE GE_istr32_from_nstr(EIF_NATIVE_CHAR* s);

/*
 * New Eiffel string of type "STRING" containing all
 * characters found in the null-terminated ISO 8859-1 string `s'
 */
extern EIF_REFERENCE GE_str(const char* s);

#ifdef __cplusplus
}
#endif

#endif

/*
	description:

		"C functions used to implement class CONSOLE"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2007-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef GE_CONSOLE_H
#define GE_CONSOLE_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Initialize mutex to determine whether a new
 * console needs to be created.
 */
#ifdef EIF_WINDOWS
extern void GE_init_console(void);
#else
#define GE_init_console()
#endif

/*
 * Create a new DOS console if needed (i.e. in case of a Windows application).
 */
#ifdef EIF_WINDOWS
extern void GE_show_console(void);
#else
#define GE_show_console()
#endif

#ifdef __cplusplus
}
#endif

#endif

/*
	description:

		"C functions used to implement the program initialization"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2007-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef GE_MAIN_H
#define GE_MAIN_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

extern int GE_main(int argc, EIF_NATIVE_CHAR** argv);

/*
 * System name.
 */
extern char* GE_system_name;

/*
 * Root class name.
 */
extern char* GE_root_class_name;

#ifdef EIF_WINDOWS

#include <windows.h>

/*
 * Used in WEL.
 */
extern HINSTANCE eif_hInstance;
extern HINSTANCE eif_hPrevInstance;
extern LPWSTR eif_lpCmdLine;
extern int eif_nCmdShow;

/*
 * Main entry point when compiling a Windows application.
 * See:
 *    http://en.wikipedia.org/wiki/WinMain
 *    http://msdn2.microsoft.com/en-us/library/ms633559.aspx
 */
extern int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow);

#endif

#ifdef __cplusplus
}
#endif

#endif

/*
	description:

		"C functions used to access garbage collector facilities"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2007-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef GE_GC_H
#define GE_GC_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EXCEPTION_H
#include "ge_exception.h"
#endif

#ifdef GE_USE_BOEHM_GC
/*
 *	Use the Boehm garbage collector.
 *	See:
 *		http://en.wikipedia.org/wiki/Boehm_GC
 *		http://www.hpl.hp.com/personal/Hans_Boehm/gc/
 */
#define GC_IGNORE_WARN
#ifdef GE_USE_THREADS
#define GC_THREADS
#undef GC_NO_THREAD_REDIRECTS
#endif
#include "gc.h"
#endif

/*
 * GC initialization.
 */

#ifdef GE_USE_BOEHM_GC
#ifdef GE_USE_THREADS
#define GE_init_gc() \
	GC_INIT(); \
	GC_allow_register_threads(); \
	GC_enable_incremental()
#else
#define GE_init_gc() \
	GC_INIT(); \
	GC_enable_incremental()
#endif
#else /* No GC */
#define GE_init_gc() /* do nothing */
#endif

/*
 * Memory allocation.
 */

/*
 * Allocate memory that can contain pointers to collectable objects.
 * The allocated memory is not necessarily zeroed (unless `GE_malloc_cleared' is_defined).
 * The allocated object is itself collectable.
 * Raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_malloc(size) GE_null(GC_MALLOC(size))
#else /* No GC */
#define GE_malloc(size) GE_null(malloc(size))
#endif

/*
 * Allocate memory that can contain pointers to collectable objects.
 * The allocated memory is not necessarily zeroed (unless `GE_malloc_cleared' is_defined).
 * The allocated object is itself collectable.
 * Do not raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_unprotected_malloc(size) GC_MALLOC(size)
#else /* No GC */
#define GE_unprotected_malloc(size) malloc(size)
#endif

/*
 * Allocate memory that does not contain pointers to collectable objects.
 * The allocated memory is not necessarily zeroed (unless `GE_malloc_atomic_cleared' is_defined).
 * The allocated object is itself collectable.
 * Raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_malloc_atomic(size) GE_null(GC_MALLOC_ATOMIC(size))
#else /* No GC */
#define GE_malloc_atomic(size) GE_null(malloc(size))
#endif

/*
 * Allocate memory that does not contain pointers to collectable objects.
 * The allocated memory is not necessarily zeroed (unless `GE_malloc_atomic_cleared' is_defined).
 * The allocated object is itself collectable.
 * Do not raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_unprotected_malloc_atomic(size) GC_MALLOC_ATOMIC(size)
#else /* No GC */
#define GE_unprotected_malloc_atomic(size) malloc(size)
#endif

/*
 * Allocate memory that can contain pointers to collectable objects.
 * The allocated memory is zeroed.
 * The allocated object is itself collectable.
 * Raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_calloc(nelem, elsize) GE_null(GC_MALLOC((nelem) * (elsize)))
#else /* No GC */
#define GE_calloc(nelem, elsize) GE_null(calloc((nelem), (elsize)))
#endif

/*
 * Allocate memory that can contain pointers to collectable objects.
 * The allocated memory is zeroed.
 * The allocated object is itself collectable.
 * Do not raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_unprotected_calloc(nelem, elsize) GC_MALLOC((nelem) * (elsize))
#else /* No GC */
#define GE_unprotected_calloc(nelem, elsize) calloc((nelem), (elsize))
#endif

/*
 * Allocate memory that does not contain pointers to collectable objects.
 * The allocated memory is zeroed.
 * The allocated object is itself collectable.
 * Raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_calloc_atomic(nelem, elsize) memset(GE_null(GC_MALLOC_ATOMIC((nelem) * (elsize))), 0, (nelem) * (elsize))
#else /* No GC */
#define GE_calloc(nelem, elsize) GE_null(calloc((nelem), (elsize)))
#endif

/*
 * Allocate memory that can contain pointers to collectable objects.
 * The allocated memory is not necessarily zeroed (unless `GE_malloc_cleared' is_defined).
 * The allocated object is itself not collectable.
 * Raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_malloc_uncollectable(size) GE_null(GC_MALLOC_UNCOLLECTABLE(size))
#else /* No GC */
#define GE_malloc_uncollectable(size) GE_null(malloc(size))
#endif

/*
 * Allocate memory that can contain pointers to collectable objects.
 * The allocated memory is not necessarily zeroed (unless `GE_malloc_cleared' is_defined).
 * The allocated object is itself not collectable.
 * Do not raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_unprotected_malloc_uncollectable(size) GC_MALLOC_UNCOLLECTABLE(size)
#else /* No GC */
#define GE_unprotected_malloc_uncollectable(size) malloc(size)
#endif

/*
 * Allocate memory that does not contain pointers to collectable objects.
 * The allocated memory is not necessarily zeroed (unless `GE_malloc_atomic_cleared' is_defined).
 * The allocated object is itself not collectable.
 * Raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_malloc_atomic_uncollectable(size) GE_null(GC_malloc_atomic_uncollectable(size))
#else /* No GC */
#define GE_malloc_atomic_uncollectable(size) GE_null(malloc(size))
#endif

/*
 * Allocate memory that does not contain pointers to collectable objects.
 * The allocated memory is not necessarily zeroed (unless `GE_malloc_atomic_cleared' is_defined).
 * The allocated object is itself not collectable.
 * Do not raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_unprotected_malloc_atomic_uncollectable(size) GC_malloc_atomic_uncollectable(size)
#else /* No GC */
#define GE_unprotected_malloc_atomic_uncollectable(size) malloc(size)
#endif

/*
 * Allocate memory that can contain pointers to collectable objects.
 * The allocated memory is zeroed.
 * The allocated object is itself not collectable.
 * Raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_calloc_uncollectable(nelem, elsize) GE_null(GC_MALLOC_UNCOLLECTABLE((nelem) * (elsize)))
#else /* No GC */
#define GE_calloc_uncollectable(nelem, elsize) GE_null(calloc((nelem), (elsize)))
#endif

/*
 * Allocate memory that can contain pointers to collectable objects.
 * The allocated memory is zeroed.
 * The allocated object is itself not collectable.
 * Do not raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_unprotected_calloc_uncollectable(nelem, elsize) GC_MALLOC_UNCOLLECTABLE((nelem) * (elsize))
#else /* No GC */
#define GE_unprotected_calloc_uncollectable(nelem, elsize) calloc((nelem), (elsize))
#endif

/*
 * Allocate memory that does not contain pointers to collectable objects.
 * The allocated memory is zeroed.
 * The allocated object is itself not collectable.
 * Raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_calloc_atomic_uncollectable(nelem, elsize) memset(GE_null(GC_malloc_atomic_uncollectable((nelem) * (elsize))), 0, (nelem) * (elsize))
#else /* No GC */
#define GE_calloc_atomic_uncollectable(nelem, elsize) GE_null(calloc((nelem), (elsize)))
#endif

/*
 * Allocate memory that does not contain pointers to collectable objects.
 * The allocated memory is zeroed.
 * The allocated object is itself not collectable.
 * Do not raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
extern void* GE_unprotected_calloc_atomic_uncollectable(size_t nelem, size_t elsize);
#else /* No GC */
#define GE_unprotected_calloc_atomic_uncollectable(nelem, elsize) calloc((nelem), (elsize))
#endif

/*
 * Allocate more memory for the given pointer.
 * The reallocated pointer keeps the same properties (e.g. atomic or not, collectable or not).
 * The extra allocated memory is not necessarily zeroed.
 * Raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_realloc(p, size) GE_null(GC_REALLOC((p), (size)))
#else /* No GC */
#define GE_realloc(p, size) GE_null(realloc((p), (size)))
#endif

/*
 * Allocate more memory for the given pointer.
 * The reallocated pointer keeps the same properties (e.g. atomic or not, collectable or not).
 * The extra allocated memory is not necessarily zeroed.
 * Do not raise an exception when no-more-memory.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_unprotected_realloc(p, size) GC_REALLOC((p), (size))
#else /* No GC */
#define GE_unprotected_realloc(p, size) realloc((p), (size))
#endif

/*
 * Allocate more memory for the given pointer.
 * The reallocated pointer keeps the same properties (e.g. atomic or not, collectable or not).
 * The extra allocated memory is zeroed.
 * Raise an exception when no-more-memory.
 */
extern void* GE_recalloc(void* p, size_t old_nelem, size_t new_nelem, size_t elsize);

/*
 * Allocate more memory for the given pointer.
 * The reallocated pointer keeps the same properties (e.g. atomic or not, collectable or not).
 * The extra allocated memory is zeroed.
 * Do not raise an exception when no-more-memory.
 */
extern void* GE_unprotected_recalloc(void* p, size_t old_nelem, size_t new_nelem, size_t elsize);

/*
 * Explicitly deallocate an object.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_free(p) GC_FREE(p)
#else /* No GC */
#define GE_free(p) free(p)
#endif

/*
 * When defined, `GE_malloc_cleared' means that `GE_malloc' and
 * `GE_malloc_uncollectable' make sure that the allocated memory
 * is zeroed.
 */
#ifdef GE_USE_BOEHM_GC
#define GE_malloc_cleared
#else /* No GC */
#endif

/*
 * When defined, `GE_malloc_atomic_cleared' means that `GE_malloc_atomic'
 * and `GE_malloc_atomic_uncollectable' make sure that the allocated memory
 * is zeroed.
 */
#ifdef GE_USE_BOEHM_GC
/* #define GE_malloc_atomic_cleared */
#else /* No GC */
#endif

/*
 * Dispose
 */

/*
 * Register dispose routine `disp' to be called on object `obj' when it will be collected.
 */
#ifdef GE_USE_BOEHM_GC
extern void GE_boehm_dispose(void* C, void* disp); /* Call dispose routine on object `C'. */
#define GE_register_dispose(obj, disp) GC_REGISTER_FINALIZER_NO_ORDER((void*)(obj), (void (*) (void*, void*)) &GE_boehm_dispose, (void*)(disp), NULL, NULL)
#else /* No GC */
#define GE_register_dispose(obj, disp) /* do nothing */
#endif

/*
 * Register dispose routine `disp' to be called on once-per-object `data' when it will be collected.
 */
#ifdef GE_USE_BOEHM_GC
extern void GE_boehm_dispose_once_per_object_data(void* data, void* disp); /* Call dispose routine `disp' on once-per-object data `data'. */
#define GE_register_dispose_once_per_object_data(data, disp) GC_REGISTER_FINALIZER_NO_ORDER((void*)(data), (void (*) (void*, void*)) &GE_boehm_dispose_once_per_object_data, (void*)(disp), NULL, NULL)
#else /* No GC */
#define GE_register_dispose_once_per_object_data(data, disp) /* do nothing */
#endif

/*
 * Access to objects, useful with GCs which move objects in memory.
 * This is not the case here, since the Boehm GC is not a moving GC.
 */

/* Access object through hector. */
#define eif_access(obj) (obj)
/* Freeze memory address. */
#define eif_freeze(obj) (obj)
/* The C side adopts an object. */
#define eif_adopt(obj) (obj)
/* The C side protects an object. */
#define eif_protect(obj) (obj)
/* The C side weans adopted object. */
#define eif_wean(obj) (obj)
/* Forget a frozen memory address. */
#define eif_unfreeze(obj)
/* Always frozen since they do not move. */
#define eif_frozen(obj) 1
/* Always frozen since they do not move. */
#define spfrozen(obj) 1

#endif

/*
	description:

		"C functions used to implement class IDENTIFIED"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2007-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef GE_IDENTIFIED_H
#define GE_IDENTIFIED_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Initialize data to keep track of object ids.
 */
extern void GE_init_identified(void);

/*
 * Get a new id for `object', assuming it is NOT in the stack.
 */
extern EIF_INTEGER_32 GE_object_id(EIF_OBJECT object);

/*
 * Return the object associated with `id'.
 */
extern EIF_REFERENCE GE_id_object(EIF_INTEGER_32 id);

/*
 * Remove the object associated with `id' from the stack.
 */
extern void GE_object_id_free(EIF_INTEGER_32 id);

#ifdef __cplusplus
}
#endif

#endif

#ifdef __cplusplus
extern "C" {
#endif

#define T0 EIF_ANY

/* CHARACTER */
#define EIF_CHARACTER EIF_CHARACTER_8

/* WIDE_CHARACTER */
#define EIF_WIDE_CHAR EIF_CHARACTER_32

/* INTEGER */
#define EIF_INTEGER EIF_INTEGER_32

/* NATURAL */
#define EIF_NATURAL EIF_NATURAL_32

/* REAL */
#define EIF_REAL EIF_REAL_32

/* DOUBLE */
#define EIF_DOUBLE EIF_REAL_64

/* BOOLEAN */
#define T1 EIF_BOOLEAN
extern T0* GE_boxed1(T1 a1);
extern T0* GE_boxed_pointer1(T1* a1);
typedef struct Sb1 Tb1;

/* CHARACTER_8 */
#define T2 EIF_CHARACTER_8
extern T0* GE_boxed2(T2 a1);
extern T0* GE_boxed_pointer2(T2* a1);
typedef struct Sb2 Tb2;

/* CHARACTER_32 */
#define T3 EIF_CHARACTER_32
extern T0* GE_boxed3(T3 a1);
extern T0* GE_boxed_pointer3(T3* a1);
typedef struct Sb3 Tb3;

/* INTEGER_8 */
#define T4 EIF_INTEGER_8
extern T0* GE_boxed4(T4 a1);
extern T0* GE_boxed_pointer4(T4* a1);
typedef struct Sb4 Tb4;

/* INTEGER_16 */
#define T5 EIF_INTEGER_16
extern T0* GE_boxed5(T5 a1);
extern T0* GE_boxed_pointer5(T5* a1);
typedef struct Sb5 Tb5;

/* INTEGER_32 */
#define T6 EIF_INTEGER_32
extern T0* GE_boxed6(T6 a1);
extern T0* GE_boxed_pointer6(T6* a1);
typedef struct Sb6 Tb6;

/* INTEGER_64 */
#define T7 EIF_INTEGER_64
extern T0* GE_boxed7(T7 a1);
extern T0* GE_boxed_pointer7(T7* a1);
typedef struct Sb7 Tb7;

/* NATURAL_8 */
#define T8 EIF_NATURAL_8
extern T0* GE_boxed8(T8 a1);
extern T0* GE_boxed_pointer8(T8* a1);
typedef struct Sb8 Tb8;

/* NATURAL_16 */
#define T9 EIF_NATURAL_16
extern T0* GE_boxed9(T9 a1);
extern T0* GE_boxed_pointer9(T9* a1);
typedef struct Sb9 Tb9;

/* NATURAL_32 */
#define T10 EIF_NATURAL_32
extern T0* GE_boxed10(T10 a1);
extern T0* GE_boxed_pointer10(T10* a1);
typedef struct Sb10 Tb10;

/* NATURAL_64 */
#define T11 EIF_NATURAL_64
extern T0* GE_boxed11(T11 a1);
extern T0* GE_boxed_pointer11(T11* a1);
typedef struct Sb11 Tb11;

/* REAL_32 */
#define T12 EIF_REAL_32
extern T0* GE_boxed12(T12 a1);
extern T0* GE_boxed_pointer12(T12* a1);
typedef struct Sb12 Tb12;

/* REAL_64 */
#define T13 EIF_REAL_64
extern T0* GE_boxed13(T13 a1);
extern T0* GE_boxed_pointer13(T13* a1);
typedef struct Sb13 Tb13;

/* POINTER */
#define T14 EIF_POINTER
extern T0* GE_boxed14(T14 a1);
extern T0* GE_boxed_pointer14(T14* a1);
typedef struct Sb14 Tb14;

/* [detachable] SPECIAL [CHARACTER_8] */
typedef struct S15 T15;

/* [detachable] SPECIAL [CHARACTER_32] */
typedef struct S16 T16;

/* [detachable] STRING_8 */
typedef struct S17 T17;

/* [detachable] STRING_32 */
typedef struct S18 T18;

/* [detachable] IMMUTABLE_STRING_32 */
typedef struct S20 T20;

/* [detachable] ISE_EXCEPTION_MANAGER */
typedef struct S21 T21;

/* [detachable] GEC */
typedef struct S25 T25;

/* [detachable] CELL [detachable EXCEPTION] */
typedef struct S26 T26;

/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32] */
typedef struct S27 T27;

/* [detachable] CELL [detachable TUPLE [[attached] INTEGER_32, [attached] INTEGER_32, [attached] INTEGER_32, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] INTEGER_32, [attached] BOOLEAN]] */
typedef struct S28 T28;

/* [detachable] CELL [[attached] NO_MORE_MEMORY] */
typedef struct S29 T29;

/* detachable C_STRING */
typedef struct S30 T30;

/* [detachable] TUPLE [[attached] INTEGER_32, [attached] INTEGER_32, [attached] INTEGER_32, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] INTEGER_32, [attached] BOOLEAN] */
typedef struct S31 T31;

/* [detachable] KL_ARGUMENTS */
typedef struct S32 T32;

/* [detachable] ARRAY [[attached] STRING_8] */
typedef struct S33 T33;

/* [detachable] SPECIAL [[attached] STRING_8] */
typedef struct S34 T34;

/* [detachable] KL_EXCEPTIONS */
typedef struct S35 T35;

/* [detachable] ET_ERROR_HANDLER */
typedef struct S36 T36;

/* [detachable] KL_TEXT_INPUT_FILE */
typedef struct S37 T37;

/* [detachable] ET_ISE_VARIABLES */
typedef struct S38 T38;

/* [detachable] AP_FLAG */
typedef struct S39 T39;

/* detachable ET_SYSTEM */
typedef struct S40 T40;

/* [detachable] AP_PARSER */
typedef struct S42 T42;

/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST */
typedef struct S43 T43;

/* [detachable] AP_STRING_OPTION */
typedef struct S44 T44;

/* [detachable] UT_VERSION */
typedef struct S46 T46;

/* [detachable] AP_ENUMERATION_OPTION */
typedef struct S47 T47;

/* [detachable] AP_BOOLEAN_OPTION */
typedef struct S48 T48;

/* [detachable] AP_INTEGER_OPTION */
typedef struct S49 T49;

/* [detachable] ET_NULL_ERROR_HANDLER */
typedef struct S53 T53;

/* [detachable] ET_ECF_SYSTEM_PARSER */
typedef struct S55 T55;

/* [detachable] ET_ECF_ERROR_HANDLER */
typedef struct S56 T56;

/* detachable ET_ECF_SETTINGS */
typedef struct S57 T57;

/* [detachable] ET_ECF_SYSTEM */
typedef struct S58 T58;

/* [detachable] ET_ECF_TARGET */
typedef struct S59 T59;

/* detachable ET_ECF_CAPABILITIES */
typedef struct S61 T61;

/* detachable ET_ECF_VARIABLES */
typedef struct S62 T62;

/* [detachable] ET_DYNAMIC_SYSTEM */
typedef struct S63 T63;

/* [detachable] ET_SYSTEM_PROCESSOR */
typedef struct S64 T64;

/* [detachable] ET_C_GENERATOR */
typedef struct S67 T67;

/* detachable DT_DATE_TIME */
typedef struct S68 T68;

/* detachable DS_HASH_SET [[attached] STRING_8] */
typedef struct S70 T70;

/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER */
typedef struct S71 T71;

/* [detachable] ET_CLASS_TYPE */
typedef struct S72 T72;

/* [detachable] ET_TOKEN_CONSTANTS */
typedef struct S73 T73;

/* [detachable] ET_CLASS */
typedef struct S74 T74;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S76 T76;

/* [detachable] UT_CANNOT_READ_FILE_ERROR */
typedef struct S77 T77;

/* [detachable] UT_VERSION_NUMBER */
typedef struct S79 T79;

/* [detachable] UT_MESSAGE */
typedef struct S80 T80;

/* [detachable] RX_PCRE_REGULAR_EXPRESSION */
typedef struct S81 T81;

/* [detachable] KL_STRING_ROUTINES */
typedef struct S82 T82;

/* [detachable] KL_EXECUTION_ENVIRONMENT */
typedef struct S86 T86;

/* [detachable] AP_ERROR */
typedef struct S87 T87;

/* [detachable] KL_SHELL_COMMAND */
typedef struct S88 T88;

/* [detachable] GECC */
typedef struct S89 T89;

/* [detachable] KL_OPERATING_SYSTEM */
typedef struct S90 T90;

/* [detachable] VOID_TARGET */
typedef struct S93 T93;

/* [detachable] TYPE [[attached] VOID_TARGET] */
#define T94 EIF_TYPE_OBJ

/* [detachable] ROUTINE_FAILURE */
typedef struct S95 T95;

/* [detachable] TYPE [[attached] ROUTINE_FAILURE] */
#define T96 EIF_TYPE_OBJ

/* [detachable] OLD_VIOLATION */
typedef struct S97 T97;

/* [detachable] TYPE [[attached] OLD_VIOLATION] */
#define T98 EIF_TYPE_OBJ

/* [detachable] NO_MORE_MEMORY */
typedef struct S99 T99;

/* [detachable] INVARIANT_VIOLATION */
typedef struct S100 T100;

/* [detachable] OPERATING_SYSTEM_SIGNAL_FAILURE */
typedef struct S101 T101;

/* [detachable] IO_FAILURE */
typedef struct S102 T102;

/* [detachable] OPERATING_SYSTEM_FAILURE */
typedef struct S103 T103;

/* [detachable] COM_FAILURE */
typedef struct S104 T104;

/* [detachable] EIFFEL_RUNTIME_PANIC */
typedef struct S105 T105;

/* [detachable] PRECONDITION_VIOLATION */
typedef struct S107 T107;

/* [detachable] POSTCONDITION_VIOLATION */
typedef struct S108 T108;

/* [detachable] FLOATING_POINT_FAILURE */
typedef struct S109 T109;

/* [detachable] CHECK_VIOLATION */
typedef struct S110 T110;

/* [detachable] BAD_INSPECT_VALUE */
typedef struct S111 T111;

/* [detachable] VARIANT_VIOLATION */
typedef struct S112 T112;

/* [detachable] LOOP_INVARIANT_VIOLATION */
typedef struct S113 T113;

/* [detachable] RESCUE_FAILURE */
typedef struct S114 T114;

/* [detachable] RESUMPTION_FAILURE */
typedef struct S115 T115;

/* [detachable] CREATE_ON_DEFERRED */
typedef struct S116 T116;

/* [detachable] EXTERNAL_FAILURE */
typedef struct S117 T117;

/* [detachable] VOID_ASSIGNED_TO_EXPANDED */
typedef struct S118 T118;

/* [detachable] EXCEPTION_IN_SIGNAL_HANDLER_FAILURE */
typedef struct S119 T119;

/* [detachable] MISMATCH_FAILURE */
typedef struct S120 T120;

/* [detachable] DEVELOPER_EXCEPTION */
typedef struct S121 T121;

/* [detachable] ADDRESS_APPLIED_TO_MELTED_FEATURE */
typedef struct S122 T122;

/* [detachable] SERIALIZATION_FAILURE */
typedef struct S123 T123;

/* [detachable] EXECUTION_ENVIRONMENT */
typedef struct S124 T124;

/* [detachable] KL_WINDOWS_FILE_SYSTEM */
typedef struct S125 T125;

/* [detachable] KL_UNIX_FILE_SYSTEM */
typedef struct S126 T126;

/* [detachable] PRIMES */
typedef struct S127 T127;

/* [detachable] SPECIAL [[attached] INTEGER_32] */
typedef struct S128 T128;

/* [detachable] SPECIAL [[attached] BOOLEAN] */
typedef struct S129 T129;

/* [detachable] ARGUMENTS_32 */
typedef struct S130 T130;

/* [detachable] MUTEX */
typedef struct S131 T131;

/* [detachable] UT_ERROR_HANDLER */
typedef struct S132 T132;

/* [detachable] KL_STANDARD_FILES */
typedef struct S133 T133;

/* [detachable] KL_STDERR_FILE */
typedef struct S134 T134;

/* [detachable] KL_STDOUT_FILE */
typedef struct S135 T135;

/* detachable KL_LINKABLE [[attached] CHARACTER_8] */
typedef struct S137 T137;

/* [detachable] MANAGED_POINTER */
typedef struct S140 T140;

/* [detachable] FILE_INFO */
typedef struct S141 T141;

/* [detachable] ST_WORD_WRAPPER */
typedef struct S142 T142;

/* [detachable] AP_DISPLAY_HELP_FLAG */
typedef struct S144 T144;

/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION] */
typedef struct S145 T145;

/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
typedef struct S146 T146;

/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8] */
typedef struct S147 T147;

/* [detachable] AP_ERROR_HANDLER */
typedef struct S148 T148;

/* [detachable] DS_LINKED_LIST_CURSOR [[attached] AP_OPTION] */
typedef struct S149 T149;

/* [detachable] DS_ARRAYED_LIST [detachable STRING_8] */
typedef struct S150 T150;

/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8] */
typedef struct S151 T151;

/* [detachable] DS_LINKED_LIST [[attached] STRING_8] */
typedef struct S152 T152;

/* [detachable] KL_STRING_EQUALITY_TESTER */
typedef struct S153 T153;

/* detachable KL_EQUALITY_TESTER [[attached] STRING_8] */
typedef struct S154 T154;

/* [detachable] DS_LINKED_LIST [[attached] BOOLEAN] */
typedef struct S157 T157;

/* [detachable] DS_LINKED_LIST [[attached] INTEGER_32] */
typedef struct S160 T160;

/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM */
typedef struct S161 T161;

/* [detachable] ET_ECF_AST_FACTORY */
typedef struct S162 T162;

/* [detachable] DS_CELL [detachable ET_ECF_SYSTEM] */
typedef struct S163 T163;

/* [detachable] TUPLE [[attached] ET_ECF_TARGET] */
typedef struct S164 T164;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_TARGET]] */
typedef struct S165 T165;

/* detachable TUPLE */
typedef struct S166 T166;

/* [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER] */
typedef struct S167 T167;

/* [detachable] XM_ELEMENT */
typedef struct S168 T168;

/* detachable XM_POSITION_TABLE */
typedef struct S169 T169;

/* [detachable] TUPLE [[attached] XM_ELEMENT, detachable XM_POSITION_TABLE, [attached] STRING_8] */
typedef struct S170 T170;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] XM_ELEMENT, detachable XM_POSITION_TABLE, [attached] STRING_8]] */
typedef struct S171 T171;

/* [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER, [detachable] STRING_8, [attached] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_TARGET]], [attached] DS_CELL [detachable ET_ECF_SYSTEM]] */
typedef struct S172 T172;

/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8] */
typedef struct S174 T174;

/* [detachable] KL_CASE_INSENSITIVE_STRING_EQUALITY_TESTER */
typedef struct S175 T175;

/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8] */
typedef struct S176 T176;

/* [detachable] XM_EIFFEL_PARSER */
typedef struct S178 T178;

/* [detachable] XM_TREE_CALLBACKS_PIPE */
typedef struct S179 T179;

/* [detachable] XM_CALLBACKS_TO_TREE_FILTER */
typedef struct S182 T182;

/* [detachable] ET_ECF_STATE */
typedef struct S183 T183;

/* [detachable] ET_IDENTIFIER */
typedef struct S184 T184;

/* [detachable] ET_ECF_SYSTEM_CONFIG */
typedef struct S185 T185;

/* [detachable] ET_ECF_LIBRARY */
typedef struct S187 T187;

/* [detachable] ET_ADAPTED_LIBRARIES */
typedef struct S188 T188;

/* [detachable] TUPLE [[attached] ET_ADAPTED_LIBRARY] */
typedef struct S190 T190;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ADAPTED_LIBRARY]] */
typedef struct S191 T191;

/* [detachable] XM_DOCUMENT */
typedef struct S192 T192;

/* [detachable] ET_COMPRESSED_POSITION */
typedef struct S193 T193;

/* [detachable] XM_STOP_ON_ERROR_FILTER */
typedef struct S195 T195;

/* detachable XM_ATTRIBUTE */
typedef struct S196 T196;

/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8] */
typedef struct S198 T198;

/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8] */
typedef struct S199 T199;

/* detachable ET_ECF_TARGET_PARENT */
typedef struct S200 T200;

/* [detachable] DS_CELL [detachable ET_ECF_SYSTEM_CONFIG] */
typedef struct S201 T201;

/* [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER, [attached] ET_ECF_INTERNAL_UNIVERSE, [attached] DS_CELL [detachable ET_ECF_SYSTEM_CONFIG]] */
typedef struct S202 T202;

/* detachable ET_ECF_TARGETS */
typedef struct S203 T203;

/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
typedef struct S204 T204;

/* [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
typedef struct S205 T205;

/* [detachable] DS_CELL [detachable ET_ECF_LIBRARY] */
typedef struct S206 T206;

/* [detachable] ET_ECF_ADAPTED_LIBRARY */
typedef struct S207 T207;

/* [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER, [attached] ET_ECF_ADAPTED_LIBRARY, [attached] DS_CELL [detachable ET_ECF_LIBRARY]] */
typedef struct S208 T208;

/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES */
typedef struct S210 T210;

/* [detachable] ET_ECF_DOTNET_ASSEMBLY */
typedef struct S211 T211;

/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY */
typedef struct S213 T213;

/* [detachable] ET_ECF_ERROR */
typedef struct S214 T214;

/* [detachable] KL_AGENT_HASH_FUNCTION [[attached] STRING_8] */
typedef struct S215 T215;

/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8] */
typedef struct S216 T216;

/* [detachable] TUPLE [[attached] STRING_8] */
typedef struct S217 T217;

/* [detachable] FUNCTION [[attached] TUPLE [[attached] STRING_8], [attached] INTEGER_32] */
typedef struct S218 T218;

/* [detachable] TUPLE [[attached] KL_STRING_ROUTINES] */
typedef struct S219 T219;

/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8] */
typedef struct S221 T221;

/* [detachable] ET_DYNAMIC_PRIMARY_TYPE */
typedef struct S223 T223;

/* [detachable] ET_DYNAMIC_FEATURE_LIST */
typedef struct S224 T224;

/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME] */
typedef struct S225 T225;

/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8] */
typedef struct S226 T226;

/* [detachable] TUPLE [[attached] ET_CLASS] */
typedef struct S227 T227;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]] */
typedef struct S228 T228;

/* [detachable] TUPLE [INTEGER_32] */
typedef struct S229 T229;

/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER */
typedef struct S230 T230;

/* [detachable] TUPLE [[attached] ET_DYNAMIC_SYSTEM] */
typedef struct S231 T231;

/* detachable ET_DYNAMIC_FEATURE */
typedef struct S236 T236;

/* [detachable] ET_ACTUAL_PARAMETER_LIST */
typedef struct S240 T240;

/* [detachable] ET_NESTED_TYPE_CONTEXT */
typedef struct S241 T241;

/* [detachable] ET_FORMAL_PARAMETER_TYPE */
typedef struct S242 T242;

/* [detachable] ARRAY [[attached] ET_TYPE] */
typedef struct S244 T244;

/* [detachable] SPECIAL [[attached] ET_TYPE] */
typedef struct S245 T245;

/* [detachable] ET_EXTERNAL_FUNCTION */
typedef struct S253 T253;

/* [detachable] ET_TUPLE_TYPE */
typedef struct S254 T254;

/* detachable ET_FORMAL_ARGUMENT_LIST */
typedef struct S255 T255;

/* [detachable] ET_FORMAL_ARGUMENT */
typedef struct S256 T256;

/* [detachable] ET_QUERY_LIST */
typedef struct S257 T257;

/* [detachable] ET_PROCEDURE_LIST */
typedef struct S258 T258;

/* [detachable] ET_DYNAMIC_TYPE_SET_LIST */
typedef struct S259 T259;

/* [detachable] KL_STRING_OUTPUT_STREAM */
typedef struct S260 T260;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER] */
typedef struct S261 T261;

/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S262 T262;

/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32] */
typedef struct S263 T263;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_UNQUALIFIED_CALL_EXPRESSION] */
typedef struct S264 T264;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUALIFIED_CALL_EXPRESSION] */
typedef struct S265 T265;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACTUAL_ARGUMENT_LIST] */
typedef struct S266 T266;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_RESULT] */
typedef struct S267 T267;

/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST */
typedef struct S268 T268;

/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET */
typedef struct S269 T269;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET] */
typedef struct S270 T270;

/* [detachable] DS_ARRAYED_STACK [[attached] ET_EXPRESSION] */
typedef struct S271 T271;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION] */
typedef struct S272 T272;

/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32] */
typedef struct S273 T273;

/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST */
typedef struct S274 T274;

/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S275 T275;

/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S276 T276;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST] */
typedef struct S277 T277;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
typedef struct S278 T278;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION] */
typedef struct S279 T279;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT] */
typedef struct S280 T280;

/* [detachable] ET_CURRENT */
typedef struct S282 T282;

/* [detachable] ET_ACTUAL_ARGUMENT_LIST */
typedef struct S283 T283;

/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION */
typedef struct S284 T284;

/* [detachable] ET_QUALIFIED_CALL_EXPRESSION */
typedef struct S287 T287;

/* [detachable] ET_MANIFEST_TUPLE */
typedef struct S288 T288;

/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE] */
typedef struct S289 T289;

/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE] */
typedef struct S290 T290;

/* [detachable] ARRAY [[attached] INTEGER_32] */
typedef struct S291 T291;

/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE] */
typedef struct S292 T292;

/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT] */
typedef struct S293 T293;

/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S294 T294;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE] */
typedef struct S295 T295;

/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8] */
typedef struct S296 T296;

/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER] */
typedef struct S297 T297;

/* [detachable] ET_IDENTIFIER_TESTER */
typedef struct S298 T298;

/* [detachable] KL_TEXT_OUTPUT_FILE */
typedef struct S300 T300;

/* [detachable] ET_IMPLICIT_TYPE_MARK */
typedef struct S301 T301;

/* [detachable] ET_DYNAMIC_TUPLE_TYPE */
typedef struct S302 T302;

/* [detachable] DS_STRING_HASH_TABLE */
typedef struct S303 T303;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8] */
typedef struct S304 T304;

/* [detachable] UT_CANNOT_WRITE_TO_FILE_ERROR */
typedef struct S305 T305;

/* detachable ET_DYNAMIC_PRECURSOR */
typedef struct S306 T306;

/* detachable ET_DYNAMIC_PRECURSOR_LIST */
typedef struct S307 T307;

/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S308 T308;

/* [detachable] ET_RESULT */
typedef struct S309 T309;

/* [detachable] ET_OBJECT_TEST */
typedef struct S311 T311;

/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION */
typedef struct S312 T312;

/* [detachable] ET_EQUALITY_EXPRESSION */
typedef struct S313 T313;

/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL */
typedef struct S314 T314;

/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL */
typedef struct S318 T318;

/* [detachable] ET_DYNAMIC_SPECIAL_TYPE */
typedef struct S319 T319;

/* [detachable] DS_QUICK_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S322 T322;

/* detachable ET_DYNAMIC_SECONDARY_TYPE */
typedef struct S323 T323;

/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_COMPARATOR_BY_ID */
typedef struct S324 T324;

/* [detachable] ET_CREATE_EXPRESSION */
typedef struct S328 T328;

/* [detachable] ET_QUALIFIED_CALL */
typedef struct S329 T329;

/* [detachable] KL_DIRECTORY */
typedef struct S333 T333;

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S334 T334;

/* [detachable] ET_SYMBOL */
typedef struct S341 T341;

/* [detachable] DS_QUICK_SORTER [[attached] INTEGER_32] */
typedef struct S344 T344;

/* [detachable] PROCEDURE [[attached] TUPLE] */
typedef struct S346 T346;

/* [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_IDENTIFIER, [attached] ET_CURRENT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
typedef struct S347 T347;

/* [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_IDENTIFIER, [attached] ET_RESULT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
typedef struct S348 T348;

/* [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_DYNAMIC_FEATURE, [attached] ET_RESULT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
typedef struct S349 T349;

/* [detachable] TUPLE [[attached] ET_C_GENERATOR, INTEGER_32, [attached] ET_RESULT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
typedef struct S350 T350;

/* [detachable] UT_INTEGER_FORMATTER */
typedef struct S351 T351;

/* detachable ET_COMPOUND */
typedef struct S355 T355;

/* [detachable] TUPLE [[attached] ET_C_GENERATOR, INTEGER_32, [attached] ET_EXPRESSION, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
typedef struct S357 T357;

/* [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_EXPRESSION, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S358 T358;

/* [detachable] ET_EXTENDED_ATTRIBUTE */
typedef struct S359 T359;

/* [detachable] ET_ATTRIBUTE */
typedef struct S360 T360;

/* [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
typedef struct S361 T361;

/* detachable ET_ITERATION_COMPONENT_LIST */
typedef struct S365 T365;

/* [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_DYNAMIC_FEATURE, [attached] ET_EXPRESSION, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
typedef struct S367 T367;

/* [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_DYNAMIC_FEATURE, INTEGER_32, INTEGER_32] */
typedef struct S368 T368;

/* [detachable] ET_CONSTANT_ATTRIBUTE */
typedef struct S369 T369;

/* [detachable] ET_UNIQUE_ATTRIBUTE */
typedef struct S371 T371;

/* [detachable] ET_REGULAR_INTEGER_CONSTANT */
typedef struct S372 T372;

/* [detachable] ET_SYSTEM_MULTIPROCESSOR */
typedef struct S375 T375;

/* [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8] */
typedef struct S376 T376;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_8] */
typedef struct S377 T377;

/* [detachable] TYPE [[attached] STRING_8] */
#define T378 EIF_TYPE_OBJ

/* [detachable] KL_SPECIAL_ROUTINES [[attached] INTEGER_32] */
typedef struct S379 T379;

/* [detachable] ET_FEATURE_CHECKER */
typedef struct S380 T380;

/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE] */
typedef struct S381 T381;

/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE] */
typedef struct S382 T382;

/* [detachable] ET_TYPE_CHECKER */
typedef struct S384 T384;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT] */
typedef struct S386 T386;

/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
typedef struct S387 T387;

/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
typedef struct S388 T388;

/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
typedef struct S389 T389;

/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
typedef struct S390 T390;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT] */
typedef struct S391 T391;

/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST] */
typedef struct S392 T392;

/* [detachable] ET_OBJECT_TEST_SCOPE */
typedef struct S393 T393;

/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER */
typedef struct S394 T394;

/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT] */
typedef struct S395 T395;

/* [detachable] ET_ITERATION_ITEM_SCOPE */
typedef struct S396 T396;

/* [detachable] ET_ATTACHMENT_SCOPE */
typedef struct S397 T397;

/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER */
typedef struct S398 T398;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE] */
typedef struct S399 T399;

/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE] */
typedef struct S400 T400;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM] */
typedef struct S401 T401;

/* [detachable] ET_CLIENT_LIST */
typedef struct S402 T402;

/* [detachable] ET_CLIENT */
typedef struct S403 T403;

/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
typedef struct S404 T404;

/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER */
typedef struct S405 T405;

/* detachable ET_PRECONDITIONS */
typedef struct S406 T406;

/* detachable ET_FEATURE_LIST */
typedef struct S408 T408;

/* [detachable] ET_UNKNOWN_GROUP */
typedef struct S410 T410;

/* [detachable] ET_BASE_TYPE_LIST */
typedef struct S412 T412;

/* [detachable] ET_KEYWORD */
typedef struct S413 T413;

/* [detachable] ET_CLASS_CODES */
typedef struct S414 T414;

/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S415 T415;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S416 T416;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S417 T417;

/* [detachable] RX_BYTE_CODE */
typedef struct S422 T422;

/* [detachable] RX_CHARACTER_SET */
typedef struct S423 T423;

/* [detachable] RX_CASE_MAPPING */
typedef struct S424 T424;

/* [detachable] UC_UNICODE_ROUTINES */
typedef struct S426 T426;

/* [detachable] ARRAY [[attached] RX_CHARACTER_SET] */
typedef struct S427 T427;

/* [detachable] SPECIAL [[attached] RX_CHARACTER_SET] */
typedef struct S428 T428;

/* [detachable] UC_STRING */
typedef struct S429 T429;

/* [detachable] NATIVE_STRING */
typedef struct S430 T430;

/* [detachable] TUPLE [[attached] GECC, [attached] DS_ARRAYED_LIST [[attached] STRING_8], INTEGER_32, [attached] MUTEX] */
typedef struct S431 T431;

/* [detachable] WORKER_THREAD */
typedef struct S432 T432;

/* [detachable] UC_UTF8_ROUTINES */
typedef struct S434 T434;

/* [detachable] PATH */
typedef struct S435 T435;

/* [detachable] STRING_TO_INTEGER_CONVERTOR */
typedef struct S436 T436;

/* [detachable] ET_ECF_OPTIONS */
typedef struct S444 T444;

/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT] */
typedef struct S445 T445;

/* [detachable] XM_NAMESPACE */
typedef struct S446 T446;

/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE] */
typedef struct S447 T447;

/* [detachable] ET_LIKE_CURRENT */
typedef struct S449 T449;

/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE */
typedef struct S450 T450;

/* [detachable] ET_DYNAMIC_FUNCTION_TYPE */
typedef struct S451 T451;

/* [detachable] DT_SHARED_SYSTEM_CLOCK */
typedef struct S452 T452;

/* [detachable] DT_SYSTEM_CLOCK */
typedef struct S453 T453;

/* [detachable] ET_EIFFEL_PREPARSER */
typedef struct S454 T454;

/* [detachable] ET_MASTER_CLASS_CHECKER */
typedef struct S455 T455;

/* [detachable] ET_EIFFEL_PARSER */
typedef struct S456 T456;

/* [detachable] ET_PROVIDER_CHECKER */
typedef struct S457 T457;

/* [detachable] ET_ANCESTOR_BUILDER */
typedef struct S458 T458;

/* [detachable] ET_FEATURE_FLATTENER */
typedef struct S459 T459;

/* [detachable] ET_INTERFACE_CHECKER */
typedef struct S460 T460;

/* [detachable] ET_IMPLEMENTATION_CHECKER */
typedef struct S461 T461;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS] */
typedef struct S462 T462;

/* [detachable] TUPLE [[attached] DS_ARRAYED_LIST [[attached] ET_CLASS]] */
typedef struct S463 T463;

/* [detachable] ET_AST_NULL_PROCESSOR */
typedef struct S464 T464;

/* [detachable] ET_AST_FACTORY */
typedef struct S465 T465;

/* [detachable] ET_DOTNET_ASSEMBLY_CLASSIC_CONSUMER */
typedef struct S467 T467;

/* [detachable] DT_DATE_TIME_DURATION */
typedef struct S468 T468;

/* [detachable] ET_DO_PROCEDURE */
typedef struct S469 T469;

/* [detachable] UT_CONFIG_PARSER */
typedef struct S471 T471;

/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] INTEGER_32] */
typedef struct S472 T472;

/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET */
typedef struct S475 T475;

/* [detachable] ET_DEFERRED_PROCEDURE */
typedef struct S476 T476;

/* [detachable] KL_PLATFORM */
typedef struct S477 T477;

/* UTF_CONVERTER */
typedef struct S478 T478;
extern T0* GE_boxed478(T478 a1);
extern T0* GE_boxed_pointer478(T478* a1);
typedef struct Sb478 Tb478;

/* [detachable] CELL [[attached] INTEGER_32] */
typedef struct S479 T479;

/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32] */
typedef struct S480 T480;

/* [detachable] KL_ANY_ROUTINES */
typedef struct S481 T481;

/* [detachable] KL_PATHNAME */
typedef struct S483 T483;

/* [detachable] ARRAY [[attached] IMMUTABLE_STRING_32] */
typedef struct S484 T484;

/* [detachable] SPECIAL [[attached] IMMUTABLE_STRING_32] */
typedef struct S485 T485;

/* [detachable] EXCEPTIONS */
typedef struct S486 T486;

/* [detachable] SPECIAL [[attached] NATURAL_8] */
typedef struct S487 T487;

/* TYPED_POINTER [[attached] SPECIAL [[attached] NATURAL_8]] */
typedef struct S488 T488;
extern T0* GE_boxed488(T488 a1);
extern T0* GE_boxed_pointer488(T488* a1);
typedef struct Sb488 Tb488;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_OPTION] */
typedef struct S489 T489;

/* [detachable] SPECIAL [[attached] AP_OPTION] */
typedef struct S491 T491;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] AP_OPTION] */
typedef struct S492 T492;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
typedef struct S493 T493;

/* [detachable] SPECIAL [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
typedef struct S494 T494;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
typedef struct S495 T495;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] STRING_8] */
typedef struct S496 T496;

/* [detachable] SPECIAL [detachable STRING_8] */
typedef struct S497 T497;

/* [detachable] KL_SPECIAL_ROUTINES [detachable STRING_8] */
typedef struct S498 T498;

/* [detachable] DS_LINKED_LIST_CURSOR [[attached] STRING_8] */
typedef struct S499 T499;

/* detachable DS_LINKABLE [[attached] STRING_8] */
typedef struct S500 T500;

/* detachable DS_LINKABLE [[attached] BOOLEAN] */
typedef struct S501 T501;

/* [detachable] DS_LINKED_LIST_CURSOR [[attached] BOOLEAN] */
typedef struct S502 T502;

/* detachable DS_LINKABLE [[attached] INTEGER_32] */
typedef struct S503 T503;

/* [detachable] DS_LINKED_LIST_CURSOR [[attached] INTEGER_32] */
typedef struct S504 T504;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_LIBRARY, [attached] STRING_8] */
typedef struct S506 T506;

/* [detachable] SPECIAL [[attached] ET_ECF_LIBRARY] */
typedef struct S507 T507;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_LIBRARY] */
typedef struct S510 T510;

/* [detachable] SPECIAL [[attached] ET_ECF_DOTNET_ASSEMBLY] */
typedef struct S511 T511;

/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8] */
typedef struct S514 T514;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_DOTNET_ASSEMBLY] */
typedef struct S515 T515;

/* [detachable] XM_EIFFEL_SCANNER */
typedef struct S516 T516;

/* [detachable] XM_DEFAULT_POSITION */
typedef struct S517 T517;

/* [detachable] DS_BILINKED_LIST [[attached] XM_POSITION] */
typedef struct S519 T519;

/* [detachable] DS_LINKED_STACK [[attached] XM_EIFFEL_SCANNER] */
typedef struct S520 T520;

/* [detachable] XM_CALLBACKS_NULL */
typedef struct S521 T521;

/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8] */
typedef struct S522 T522;

/* [detachable] XM_NULL_EXTERNAL_RESOLVER */
typedef struct S524 T524;

/* [detachable] XM_DTD_CALLBACKS_NULL */
typedef struct S526 T526;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ANY] */
typedef struct S527 T527;

/* [detachable] SPECIAL [detachable ANY] */
typedef struct S528 T528;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_PARSER_NAME] */
typedef struct S529 T529;

/* [detachable] XM_EIFFEL_PARSER_NAME */
typedef struct S530 T530;

/* [detachable] SPECIAL [[attached] XM_EIFFEL_PARSER_NAME] */
typedef struct S531 T531;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]] */
typedef struct S532 T532;

/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME] */
typedef struct S533 T533;

/* [detachable] SPECIAL [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]] */
typedef struct S534 T534;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_EXTERNAL_ID] */
typedef struct S535 T535;

/* [detachable] XM_DTD_EXTERNAL_ID */
typedef struct S536 T536;

/* [detachable] SPECIAL [[attached] XM_DTD_EXTERNAL_ID] */
typedef struct S537 T537;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ELEMENT_CONTENT] */
typedef struct S538 T538;

/* [detachable] XM_DTD_ELEMENT_CONTENT */
typedef struct S539 T539;

/* [detachable] SPECIAL [[attached] XM_DTD_ELEMENT_CONTENT] */
typedef struct S540 T540;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]] */
typedef struct S541 T541;

/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
typedef struct S542 T542;

/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]] */
typedef struct S543 T543;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
typedef struct S544 T544;

/* [detachable] XM_DTD_ATTRIBUTE_CONTENT */
typedef struct S545 T545;

/* [detachable] SPECIAL [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
typedef struct S546 T546;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] STRING_8]] */
typedef struct S547 T547;

/* [detachable] DS_BILINKED_LIST [[attached] STRING_8] */
typedef struct S548 T548;

/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] STRING_8]] */
typedef struct S549 T549;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] BOOLEAN] */
typedef struct S550 T550;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_DECLARATION] */
typedef struct S551 T551;

/* [detachable] XM_EIFFEL_DECLARATION */
typedef struct S552 T552;

/* [detachable] SPECIAL [[attached] XM_EIFFEL_DECLARATION] */
typedef struct S553 T553;

/* [detachable] XM_EIFFEL_ENTITY_DEF */
typedef struct S556 T556;

/* [detachable] XM_EIFFEL_SCANNER_DTD */
typedef struct S557 T557;

/* [detachable] XM_EIFFEL_PE_ENTITY_DEF */
typedef struct S559 T559;

/* [detachable] XM_NAMESPACE_RESOLVER */
typedef struct S560 T560;

/* [detachable] ARRAY [[attached] XM_CALLBACKS_FILTER] */
typedef struct S561 T561;

/* [detachable] SPECIAL [[attached] XM_CALLBACKS_FILTER] */
typedef struct S562 T562;

/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE] */
typedef struct S563 T563;

/* [detachable] ET_CLUSTERS */
typedef struct S564 T564;

/* [detachable] ET_CLASS_NAME_TESTER */
typedef struct S565 T565;

/* [detachable] ET_PARENT */
typedef struct S567 T567;

/* [detachable] ET_PARENT_LIST */
typedef struct S568 T568;

/* [detachable] ET_BUILTIN_CONVERT_FEATURE */
typedef struct S569 T569;

/* [detachable] ET_MASTER_CLASS */
typedef struct S570 T570;

/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD */
typedef struct S571 T571;

/* detachable ET_RENAME_LIST */
typedef struct S572 T572;

/* detachable ET_EXPORT_LIST */
typedef struct S573 T573;

/* detachable ET_KEYWORD_FEATURE_NAME_LIST */
typedef struct S574 T574;

/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS */
typedef struct S577 T577;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY] */
typedef struct S578 T578;

/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE] */
typedef struct S579 T579;

/* [detachable] SPECIAL [[attached] ET_ECF_SYSTEM_CONFIG] */
typedef struct S582 T582;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_SYSTEM_CONFIG] */
typedef struct S584 T584;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8] */
typedef struct S585 T585;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_TARGET, [attached] STRING_8] */
typedef struct S586 T586;

/* [detachable] SPECIAL [[attached] ET_ECF_TARGET] */
typedef struct S587 T587;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET] */
typedef struct S589 T589;

/* detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
typedef struct S591 T591;

/* [detachable] KL_SPECIAL_ROUTINES [detachable RX_REGULAR_EXPRESSION] */
typedef struct S592 T592;

/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION] */
typedef struct S593 T593;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
typedef struct S594 T594;

/* [detachable] ET_DOTNET_ASSEMBLIES */
typedef struct S595 T595;

/* [detachable] ET_DOTNET_ASSEMBLY */
typedef struct S596 T596;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_IDENTIFIER] */
typedef struct S599 T599;

/* [detachable] SPECIAL [[attached] ET_IDENTIFIER] */
typedef struct S600 T600;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8] */
typedef struct S601 T601;

/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32] */
typedef struct S602 T602;

/* detachable ET_FORMAL_PARAMETER_LIST */
typedef struct S603 T603;

/* detachable ET_FEATURE_IDS */
typedef struct S604 T604;

/* [detachable] SPECIAL [[attached] ET_DYNAMIC_FEATURE] */
typedef struct S605 T605;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_FEATURE] */
typedef struct S606 T606;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME] */
typedef struct S609 T609;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MASTER_CLASS] */
typedef struct S610 T610;

/* [detachable] SPECIAL [[attached] ET_MASTER_CLASS] */
typedef struct S611 T611;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS_NAME] */
typedef struct S612 T612;

/* [detachable] SPECIAL [[attached] ET_CLASS_NAME] */
typedef struct S613 T613;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8] */
typedef struct S615 T615;

/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME] */
typedef struct S616 T616;

/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME] */
typedef struct S617 T617;

/* [detachable] SPECIAL [[attached] ET_ACTUAL_PARAMETER_ITEM] */
typedef struct S621 T621;

/* [detachable] ET_BRACKET_SYMBOL */
typedef struct S622 T622;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACTUAL_PARAMETER_ITEM] */
typedef struct S624 T624;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE] */
typedef struct S625 T625;

/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE_SET] */
typedef struct S626 T626;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE_SET] */
typedef struct S627 T627;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_IDENTIFIER] */
typedef struct S628 T628;

/* [detachable] SPECIAL [detachable ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S629 T629;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S630 T630;

/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S631 T631;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] INTEGER_32] */
typedef struct S632 T632;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_UNQUALIFIED_CALL_EXPRESSION] */
typedef struct S634 T634;

/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION */
typedef struct S635 T635;

/* [detachable] SPECIAL [[attached] ET_UNQUALIFIED_CALL_EXPRESSION] */
typedef struct S636 T636;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_UNQUALIFIED_CALL_EXPRESSION] */
typedef struct S637 T637;

/* [detachable] SPECIAL [[attached] ET_QUALIFIED_CALL_EXPRESSION] */
typedef struct S638 T638;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_QUALIFIED_CALL_EXPRESSION] */
typedef struct S639 T639;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUALIFIED_CALL_EXPRESSION] */
typedef struct S640 T640;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACTUAL_ARGUMENT_LIST] */
typedef struct S641 T641;

/* [detachable] SPECIAL [[attached] ET_ACTUAL_ARGUMENT_LIST] */
typedef struct S642 T642;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ACTUAL_ARGUMENT_LIST] */
typedef struct S643 T643;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RESULT] */
typedef struct S644 T644;

/* [detachable] SPECIAL [[attached] ET_RESULT] */
typedef struct S645 T645;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_RESULT] */
typedef struct S646 T646;

/* [detachable] SPECIAL [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET] */
typedef struct S647 T647;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET] */
typedef struct S648 T648;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET] */
typedef struct S649 T649;

/* [detachable] SPECIAL [[attached] ET_EXPRESSION] */
typedef struct S650 T650;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION] */
typedef struct S651 T651;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EXPRESSION] */
typedef struct S652 T652;

/* detachable KL_EQUALITY_TESTER [[attached] INTEGER_32] */
typedef struct S653 T653;

/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32] */
typedef struct S654 T654;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S655 T655;

/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST] */
typedef struct S657 T657;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST] */
typedef struct S658 T658;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_TEST] */
typedef struct S659 T659;

/* [detachable] SPECIAL [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
typedef struct S660 T660;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
typedef struct S661 T661;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
typedef struct S662 T662;

/* [detachable] SPECIAL [[attached] ET_EQUALITY_EXPRESSION] */
typedef struct S663 T663;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EQUALITY_EXPRESSION] */
typedef struct S664 T664;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EQUALITY_EXPRESSION] */
typedef struct S665 T665;

/* [detachable] SPECIAL [[attached] ET_AGENT] */
typedef struct S666 T666;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT] */
typedef struct S667 T667;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_AGENT] */
typedef struct S668 T668;

/* [detachable] SPECIAL [[attached] ET_EXPRESSION_ITEM] */
typedef struct S669 T669;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION_ITEM] */
typedef struct S671 T671;

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_TUPLE_TYPE] */
typedef struct S672 T672;

/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TUPLE_TYPE] */
typedef struct S673 T673;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TUPLE_TYPE] */
typedef struct S674 T674;

/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_FEATURE] */
typedef struct S676 T676;

/* [detachable] SPECIAL [[attached] ET_FEATURE] */
typedef struct S677 T677;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE] */
typedef struct S678 T678;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CONSTANT, [attached] ET_FEATURE] */
typedef struct S679 T679;

/* [detachable] SPECIAL [[attached] ET_CONSTANT] */
typedef struct S680 T680;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTANT] */
typedef struct S682 T682;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT] */
typedef struct S683 T683;

/* [detachable] SPECIAL [[attached] ET_INLINE_CONSTANT] */
typedef struct S684 T684;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_CONSTANT] */
typedef struct S686 T686;

/* [detachable] SPECIAL [detachable ET_DYNAMIC_FEATURE] */
typedef struct S687 T687;

/* detachable DS_HASH_TABLE_CURSOR [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S689 T689;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_FEATURE] */
typedef struct S690 T690;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_FEATURE] */
typedef struct S691 T691;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] STRING_8] */
typedef struct S692 T692;

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_IDENTIFIER] */
typedef struct S695 T695;

/* [detachable] TYPE [[attached] ET_IDENTIFIER] */
#define T696 EIF_TYPE_OBJ

/* TYPED_POINTER [[attached] ANY] */
typedef struct S697 T697;
extern T0* GE_boxed697(T697 a1);
extern T0* GE_boxed_pointer697(T697* a1);
typedef struct Sb697 Tb697;

/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S698 T698;

/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]] */
typedef struct S699 T699;

/* [detachable] ARRAY [[attached] BOOLEAN] */
typedef struct S700 T700;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_SYSTEM_PROCESSOR] */
typedef struct S703 T703;

/* [detachable] TUPLE [[attached] ET_SYSTEM_PROCESSOR, [attached] DS_ARRAYED_LIST [[attached] ET_CLASS]] */
typedef struct S704 T704;

/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE] */
typedef struct S707 T707;

/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE] */
typedef struct S708 T708;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE] */
typedef struct S709 T709;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS] */
typedef struct S710 T710;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_AGENT] */
typedef struct S711 T711;

/* [detachable] SPECIAL [[attached] ET_INLINE_AGENT] */
typedef struct S712 T712;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INLINE_AGENT] */
typedef struct S713 T713;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
typedef struct S714 T714;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE] */
typedef struct S715 T715;

/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
typedef struct S716 T716;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
typedef struct S717 T717;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
typedef struct S718 T718;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY] */
typedef struct S719 T719;

/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
typedef struct S720 T720;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
typedef struct S721 T721;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
typedef struct S722 T722;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE] */
typedef struct S723 T723;

/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
typedef struct S724 T724;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
typedef struct S725 T725;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
typedef struct S726 T726;

/* [detachable] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT] */
typedef struct S727 T727;

/* [detachable] SPECIAL [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
typedef struct S728 T728;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
typedef struct S729 T729;

/* [detachable] SPECIAL [[attached] ET_NESTED_TYPE_CONTEXT] */
typedef struct S730 T730;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NESTED_TYPE_CONTEXT] */
typedef struct S731 T731;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT] */
typedef struct S732 T732;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST] */
typedef struct S733 T733;

/* [detachable] ET_NAMED_OBJECT_TEST */
typedef struct S736 T736;

/* [detachable] SPECIAL [[attached] ET_NAMED_OBJECT_TEST] */
typedef struct S737 T737;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_OBJECT_TEST] */
typedef struct S738 T738;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST] */
typedef struct S739 T739;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT] */
typedef struct S740 T740;

/* [detachable] SPECIAL [[attached] ET_ITERATION_COMPONENT] */
typedef struct S741 T741;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ITERATION_COMPONENT] */
typedef struct S743 T743;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ITERATION_COMPONENT] */
typedef struct S744 T744;

/* [detachable] DS_HASH_SET [[attached] INTEGER_32] */
typedef struct S745 T745;

/* [detachable] SPECIAL [[attached] ET_ATTACHMENT_SCOPE] */
typedef struct S747 T747;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ATTACHMENT_SCOPE] */
typedef struct S748 T748;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ATTACHMENT_SCOPE] */
typedef struct S749 T749;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ASSERTIONS, [attached] ET_FEATURE] */
typedef struct S750 T750;

/* [detachable] SPECIAL [[attached] ET_ASSERTIONS] */
typedef struct S751 T751;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTIONS] */
typedef struct S753 T753;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM] */
typedef struct S754 T754;

/* [detachable] SPECIAL [[attached] ET_INDEXING_TERM] */
typedef struct S756 T756;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INDEXING_TERM] */
typedef struct S757 T757;

/* [detachable] SPECIAL [[attached] ET_CLIENT_ITEM] */
typedef struct S758 T758;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_ITEM] */
typedef struct S759 T759;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
typedef struct S760 T760;

/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
typedef struct S761 T761;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
typedef struct S762 T762;

/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME] */
typedef struct S763 T763;

/* [detachable] SPECIAL [[attached] NATURAL_32] */
typedef struct S764 T764;

/* [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET] */
typedef struct S765 T765;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_32] */
typedef struct S766 T766;

/* [detachable] SPECIAL [[attached] NATURAL_64] */
typedef struct S767 T767;

/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32] */
typedef struct S768 T768;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32] */
typedef struct S769 T769;

/* [detachable] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]] */
typedef struct S770 T770;

/* [detachable] SPECIAL [[attached] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]]] */
typedef struct S771 T771;

/* [detachable] KL_INTEGER_ROUTINES */
typedef struct S772 T772;

/* [detachable] UC_UTF8_STRING */
typedef struct S775 T775;

/* [detachable] THREAD_ATTRIBUTES */
typedef struct S776 T776;

/* [detachable] CELL [[attached] BOOLEAN] */
typedef struct S777 T777;

/* [detachable] CHARACTER_PROPERTY */
typedef struct S780 T780;

/* [detachable] STRING_8_SEARCHER */
typedef struct S781 T781;

/* detachable ARRAYED_LIST [[attached] INTEGER_32] */
typedef struct S782 T782;

/* [detachable] ET_SYSTEM_ERROR */
typedef struct S783 T783;

/* [detachable] ET_INTERNAL_ERROR */
typedef struct S784 T784;

/* [detachable] ET_VALIDITY_ERROR */
typedef struct S785 T785;

/* detachable ET_ECF_NOTE_ELEMENT */
typedef struct S786 T786;

/* detachable ET_ECF_CLUSTERS */
typedef struct S787 T787;

/* detachable ET_ECF_ADAPTED_LIBRARIES */
typedef struct S788 T788;

/* detachable ET_ECF_ADAPTED_DOTNET_ASSEMBLIES */
typedef struct S789 T789;

/* detachable ET_ECF_FILE_RULES */
typedef struct S790 T790;

/* detachable ET_ECF_EXTERNAL_CFLAGS */
typedef struct S791 T791;

/* detachable ET_ECF_EXTERNAL_INCLUDES */
typedef struct S792 T792;

/* detachable ET_ECF_EXTERNAL_LIBRARIES */
typedef struct S793 T793;

/* detachable ET_ECF_EXTERNAL_LINKER_FLAGS */
typedef struct S794 T794;

/* detachable ET_ECF_EXTERNAL_MAKES */
typedef struct S795 T795;

/* detachable ET_ECF_EXTERNAL_OBJECTS */
typedef struct S796 T796;

/* detachable ET_ECF_EXTERNAL_RESOURCES */
typedef struct S797 T797;

/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_ACTION] */
typedef struct S798 T798;

/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY */
typedef struct S799 T799;

/* detachable ET_ECF_VERSION */
typedef struct S801 T801;

/* detachable ET_ECF_CLUSTER */
typedef struct S802 T802;

/* detachable ET_ECF_EXTERNAL_CFLAG */
typedef struct S803 T803;

/* detachable ET_ECF_EXTERNAL_INCLUDE */
typedef struct S804 T804;

/* detachable ET_ECF_EXTERNAL_LIBRARY */
typedef struct S805 T805;

/* detachable ET_ECF_EXTERNAL_LINKER_FLAG */
typedef struct S806 T806;

/* detachable ET_ECF_EXTERNAL_MAKE */
typedef struct S807 T807;

/* detachable ET_ECF_EXTERNAL_OBJECT */
typedef struct S808 T808;

/* detachable ET_ECF_EXTERNAL_RESOURCE */
typedef struct S809 T809;

/* detachable ET_ECF_FILE_RULE */
typedef struct S810 T810;

/* detachable ET_ECF_ACTION */
typedef struct S811 T811;

/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8] */
typedef struct S812 T812;

/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_VISIBLE_CLASS] */
typedef struct S813 T813;

/* [detachable] TUPLE [[attached] ET_ECF_OPTIONS] */
typedef struct S814 T814;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_OPTIONS]] */
typedef struct S815 T815;

/* detachable ET_ECF_ROOT_CLASS */
typedef struct S816 T816;

/* [detachable] ET_ECF_ROOT_ALL_CLASSES */
typedef struct S817 T817;

/* [detachable] ET_ECF_ORED_CONDITIONS */
typedef struct S818 T818;

/* detachable ET_ECF_ANDED_CONDITIONS */
typedef struct S819 T819;

/* detachable ET_ECF_VISIBLE_CLASS */
typedef struct S820 T820;

/* detachable ET_ECF_BUILD_CONDITION */
typedef struct S822 T822;

/* detachable ET_ECF_CONCURRENCY_CONDITION */
typedef struct S823 T823;

/* detachable ET_ECF_VOID_SAFETY_CONDITION */
typedef struct S824 T824;

/* detachable ET_ECF_CUSTOM_CONDITION */
typedef struct S825 T825;

/* detachable ET_ECF_DOTNET_CONDITION */
typedef struct S826 T826;

/* detachable ET_ECF_DYNAMIC_RUNTIME_CONDITION */
typedef struct S827 T827;

/* detachable ET_ECF_PLATFORM_CONDITION */
typedef struct S828 T828;

/* [detachable] ET_ECF_COMPILER_VERSION_CONDITION */
typedef struct S830 T830;

/* [detachable] ET_ECF_MSIL_CLR_VERSION_CONDITION */
typedef struct S831 T831;

/* [detachable] UT_COUNTER */
typedef struct S833 T833;

/* [detachable] KL_AGENT_ROUTINES [[attached] ET_CLASS] */
typedef struct S834 T834;

/* [detachable] TUPLE [[attached] UT_COUNTER] */
typedef struct S835 T835;

/* [detachable] TUPLE [[attached] KL_AGENT_ROUTINES [[attached] ET_CLASS], [attached] PROCEDURE [[attached] TUPLE]] */
typedef struct S836 T836;

/* [detachable] TUPLE [[attached] ET_UNIVERSE] */
typedef struct S837 T837;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_UNIVERSE]] */
typedef struct S838 T838;

/* [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]]] */
typedef struct S839 T839;

/* [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]], [attached] FUNCTION [[attached] TUPLE, [attached] BOOLEAN]] */
typedef struct S840 T840;

/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY] */
typedef struct S842 T842;

/* [detachable] TUPLE [[attached] ET_DOTNET_ASSEMBLY] */
typedef struct S843 T843;

/* [detachable] PREDICATE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]] */
typedef struct S844 T844;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]] */
typedef struct S845 T845;

/* [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY]] */
typedef struct S846 T846;

/* [detachable] TUPLE [[attached] ET_INTERNAL_UNIVERSE] */
typedef struct S847 T847;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_INTERNAL_UNIVERSE]] */
typedef struct S848 T848;

/* [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]], [attached] PREDICATE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]]] */
typedef struct S849 T849;

/* [detachable] TUPLE [[attached] ET_DOTNET_ASSEMBLIES] */
typedef struct S850 T850;

/* [detachable] TUPLE [[attached] ET_SYSTEM_PROCESSOR] */
typedef struct S851 T851;

/* [detachable] ET_LIBRARY */
typedef struct S852 T852;

/* [detachable] TUPLE [[attached] ET_LIBRARY] */
typedef struct S853 T853;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_LIBRARY]] */
typedef struct S854 T854;

/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE] */
typedef struct S855 T855;

/* [detachable] TUPLE [[attached] ET_CLUSTER] */
typedef struct S857 T857;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_CLUSTER]] */
typedef struct S858 T858;

/* [detachable] TUPLE [[attached] ET_ECF_SYSTEM] */
typedef struct S859 T859;

/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE] */
typedef struct S860 T860;

/* [detachable] TUPLE [[attached] ET_AST_PROCESSOR] */
typedef struct S861 T861;

/* [detachable] TUPLE [[attached] ET_MASTER_CLASS] */
typedef struct S862 T862;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]] */
typedef struct S863 T863;

/* [detachable] TUPLE [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
typedef struct S864 T864;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ADAPTED_DOTNET_ASSEMBLY]] */
typedef struct S865 T865;

/* [detachable] ET_ANCESTORS_STATUS_CHECKER */
typedef struct S866 T866;

/* [detachable] ET_FLATTENING_STATUS_CHECKER */
typedef struct S867 T867;

/* [detachable] ET_INTERFACE_STATUS_CHECKER */
typedef struct S868 T868;

/* [detachable] ET_IMPLEMENTATION_STATUS_CHECKER */
typedef struct S869 T869;

/* [detachable] PREDICATE [[attached] TUPLE [[attached] ET_CLASS]] */
typedef struct S870 T870;

/* [detachable] TUPLE [[attached] ET_ANCESTORS_STATUS_CHECKER] */
typedef struct S872 T872;

/* [detachable] TUPLE [[attached] ET_FLATTENING_STATUS_CHECKER] */
typedef struct S873 T873;

/* [detachable] TUPLE [[attached] ET_INTERFACE_STATUS_CHECKER] */
typedef struct S874 T874;

/* [detachable] TUPLE [[attached] ET_IMPLEMENTATION_STATUS_CHECKER] */
typedef struct S875 T875;

/* [detachable] TUPLE [BOOLEAN] */
typedef struct S876 T876;

/* [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_UNIVERSE]] */
typedef struct S877 T877;

/* [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLUSTER]]] */
typedef struct S880 T880;

/* [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE]] */
typedef struct S884 T884;

/* [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]]] */
typedef struct S885 T885;

/* [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]], [attached] FUNCTION [[attached] TUPLE [[attached] ET_CLASS], [attached] BOOLEAN]] */
typedef struct S886 T886;

/* [detachable] ET_NONE_GROUP */
typedef struct S887 T887;

/* detachable ET_PARENTHESIS_EXPRESSION */
typedef struct S888 T888;

/* detachable DS_ARRAYED_LIST [detachable ET_FORMAL_PARAMETER_TYPE] */
typedef struct S890 T890;

/* [detachable] ET_FORMAL_PARAMETER */
typedef struct S891 T891;

/* [detachable] STD_FILES */
typedef struct S892 T892;

/* TYPED_POINTER [[attached] NATURAL_8] */
typedef struct S893 T893;
extern T0* GE_boxed893(T893 a1);
extern T0* GE_boxed_pointer893(T893* a1);
typedef struct Sb893 Tb893;

/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE] */
typedef struct S894 T894;

/* [detachable] UC_STRING_EQUALITY_TESTER */
typedef struct S896 T896;

/* [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]] */
typedef struct S898 T898;

/* [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]] */
typedef struct S899 T899;

/* [detachable] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE] */
typedef struct S900 T900;

/* [detachable] XM_EIFFEL_INPUT_STREAM */
typedef struct S901 T901;

/* [detachable] KL_ARRAY_ROUTINES [[attached] INTEGER_32] */
typedef struct S902 T902;

/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY] */
typedef struct S903 T903;

/* [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_LIBRARY]] */
typedef struct S904 T904;

/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE] */
typedef struct S905 T905;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET] */
typedef struct S906 T906;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE] */
typedef struct S907 T907;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT] */
typedef struct S909 T909;

/* [detachable] SPECIAL [[attached] ET_QUERY] */
typedef struct S910 T910;

/* detachable ET_ALIAS_NAME_LIST */
typedef struct S911 T911;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_QUERY] */
typedef struct S912 T912;

/* [detachable] SPECIAL [[attached] ET_PROCEDURE] */
typedef struct S913 T913;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PROCEDURE] */
typedef struct S914 T914;

/* [detachable] TYPE [[attached] INTEGER_32] */
#define T919 EIF_TYPE_OBJ

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE] */
typedef struct S921 T921;

/* [detachable] SPECIAL [[attached] ET_BASE_TYPE] */
typedef struct S923 T923;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE] */
typedef struct S924 T924;

/* [detachable] INTEGER_OVERFLOW_CHECKER */
typedef struct S925 T925;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_NOTE_ELEMENT] */
typedef struct S926 T926;

/* [detachable] SPECIAL [[attached] ET_ECF_NOTE_ELEMENT] */
typedef struct S927 T927;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_NOTE_ELEMENT] */
typedef struct S928 T928;

/* detachable C_DATE */
typedef struct S929 T929;

/* [detachable] YY_FILE_BUFFER */
typedef struct S930 T930;

/* [detachable] KL_STDIN_FILE */
typedef struct S931 T931;

/* [detachable] YY_BUFFER */
typedef struct S932 T932;

/* [detachable] YY_UNICODE_FILE_BUFFER */
typedef struct S936 T936;

/* [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32] */
typedef struct S937 T937;

/* [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST] */
typedef struct S938 T938;

/* [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST] */
typedef struct S939 T939;

/* [detachable] DS_ARRAYED_STACK [detachable ET_KEYWORD] */
typedef struct S940 T940;

/* [detachable] DS_ARRAYED_STACK [detachable ET_SYMBOL] */
typedef struct S941 T941;

/* [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST] */
typedef struct S942 T942;

/* [detachable] DS_ARRAYED_STACK [[attached] ET_OBJECT_TEST_LIST] */
typedef struct S943 T943;

/* [detachable] DS_ARRAYED_STACK [detachable ET_ITERATION_COMPONENT_LIST] */
typedef struct S944 T944;

/* [detachable] DS_ARRAYED_STACK [[attached] ET_ITERATION_COMPONENT_LIST] */
typedef struct S945 T945;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM] */
typedef struct S946 T946;

/* [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE] */
typedef struct S947 T947;

/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS] */
typedef struct S948 T948;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD] */
typedef struct S949 T949;

/* [detachable] SPECIAL [detachable ET_KEYWORD] */
typedef struct S950 T950;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_KEYWORD] */
typedef struct S951 T951;

/* detachable ET_AGENT_KEYWORD */
typedef struct S952 T952;

/* [detachable] SPECIAL [detachable ET_AGENT_KEYWORD] */
typedef struct S953 T953;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECURSOR_KEYWORD] */
typedef struct S954 T954;

/* detachable ET_PRECURSOR_KEYWORD */
typedef struct S955 T955;

/* [detachable] SPECIAL [detachable ET_PRECURSOR_KEYWORD] */
typedef struct S956 T956;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL] */
typedef struct S957 T957;

/* [detachable] SPECIAL [detachable ET_SYMBOL] */
typedef struct S958 T958;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSITION] */
typedef struct S959 T959;

/* [detachable] SPECIAL [detachable ET_POSITION] */
typedef struct S960 T960;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BOOLEAN_CONSTANT] */
typedef struct S961 T961;

/* [detachable] SPECIAL [detachable ET_BOOLEAN_CONSTANT] */
typedef struct S963 T963;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BREAK] */
typedef struct S964 T964;

/* [detachable] SPECIAL [detachable ET_BREAK] */
typedef struct S966 T966;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHARACTER_CONSTANT] */
typedef struct S967 T967;

/* [detachable] SPECIAL [detachable ET_CHARACTER_CONSTANT] */
typedef struct S969 T969;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CURRENT] */
typedef struct S970 T970;

/* [detachable] SPECIAL [detachable ET_CURRENT] */
typedef struct S971 T971;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FREE_OPERATOR] */
typedef struct S972 T972;

/* detachable ET_FREE_OPERATOR */
typedef struct S973 T973;

/* [detachable] SPECIAL [detachable ET_FREE_OPERATOR] */
typedef struct S974 T974;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IDENTIFIER] */
typedef struct S975 T975;

/* [detachable] SPECIAL [detachable ET_IDENTIFIER] */
typedef struct S976 T976;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INTEGER_CONSTANT] */
typedef struct S977 T977;

/* [detachable] SPECIAL [detachable ET_INTEGER_CONSTANT] */
typedef struct S979 T979;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_OPERATOR] */
typedef struct S980 T980;

/* detachable ET_KEYWORD_OPERATOR */
typedef struct S981 T981;

/* [detachable] SPECIAL [detachable ET_KEYWORD_OPERATOR] */
typedef struct S982 T982;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING] */
typedef struct S983 T983;

/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING] */
typedef struct S984 T984;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_REAL_CONSTANT] */
typedef struct S985 T985;

/* [detachable] SPECIAL [detachable ET_REAL_CONSTANT] */
typedef struct S987 T987;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RESULT] */
typedef struct S988 T988;

/* [detachable] SPECIAL [detachable ET_RESULT] */
typedef struct S989 T989;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RETRY_INSTRUCTION] */
typedef struct S990 T990;

/* detachable ET_RETRY_INSTRUCTION */
typedef struct S991 T991;

/* [detachable] SPECIAL [detachable ET_RETRY_INSTRUCTION] */
typedef struct S992 T992;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL_OPERATOR] */
typedef struct S993 T993;

/* detachable ET_SYMBOL_OPERATOR */
typedef struct S994 T994;

/* [detachable] SPECIAL [detachable ET_SYMBOL_OPERATOR] */
typedef struct S995 T995;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_VOID] */
typedef struct S996 T996;

/* detachable ET_VOID */
typedef struct S997 T997;

/* [detachable] SPECIAL [detachable ET_VOID] */
typedef struct S998 T998;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SEMICOLON_SYMBOL] */
typedef struct S999 T999;

/* detachable ET_SEMICOLON_SYMBOL */
typedef struct S1000 T1000;

/* [detachable] SPECIAL [detachable ET_SEMICOLON_SYMBOL] */
typedef struct S1001 T1001;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_SYMBOL] */
typedef struct S1002 T1002;

/* [detachable] SPECIAL [detachable ET_BRACKET_SYMBOL] */
typedef struct S1003 T1003;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUESTION_MARK_SYMBOL] */
typedef struct S1004 T1004;

/* detachable ET_QUESTION_MARK_SYMBOL */
typedef struct S1005 T1005;

/* [detachable] SPECIAL [detachable ET_QUESTION_MARK_SYMBOL] */
typedef struct S1006 T1006;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_EXPRESSION] */
typedef struct S1007 T1007;

/* detachable ET_ACROSS_EXPRESSION */
typedef struct S1008 T1008;

/* [detachable] SPECIAL [detachable ET_ACROSS_EXPRESSION] */
typedef struct S1009 T1009;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_INSTRUCTION] */
typedef struct S1010 T1010;

/* detachable ET_ACROSS_INSTRUCTION */
typedef struct S1011 T1011;

/* [detachable] SPECIAL [detachable ET_ACROSS_INSTRUCTION] */
typedef struct S1012 T1012;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_ARGUMENT_LIST] */
typedef struct S1013 T1013;

/* [detachable] SPECIAL [detachable ET_ACTUAL_ARGUMENT_LIST] */
typedef struct S1014 T1014;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_ITEM] */
typedef struct S1015 T1015;

/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_ITEM] */
typedef struct S1016 T1016;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_LIST] */
typedef struct S1017 T1017;

/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_LIST] */
typedef struct S1018 T1018;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND] */
typedef struct S1019 T1019;

/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND] */
typedef struct S1020 T1020;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM] */
typedef struct S1021 T1021;

/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM] */
typedef struct S1023 T1023;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_LIST] */
typedef struct S1024 T1024;

/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST */
typedef struct S1025 T1025;

/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_LIST] */
typedef struct S1026 T1026;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_TARGET] */
typedef struct S1027 T1027;

/* [detachable] SPECIAL [detachable ET_AGENT_TARGET] */
typedef struct S1028 T1028;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ALIAS_NAME] */
typedef struct S1029 T1029;

/* detachable ET_ALIAS_NAME */
typedef struct S1030 T1030;

/* [detachable] SPECIAL [detachable ET_ALIAS_NAME] */
typedef struct S1031 T1031;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ALIAS_NAME_LIST] */
typedef struct S1032 T1032;

/* [detachable] SPECIAL [detachable ET_ALIAS_NAME_LIST] */
typedef struct S1033 T1033;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ASSIGNER] */
typedef struct S1034 T1034;

/* [detachable] SPECIAL [detachable ET_ASSIGNER] */
typedef struct S1036 T1036;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_EXPRESSION] */
typedef struct S1037 T1037;

/* detachable ET_BRACKET_EXPRESSION */
typedef struct S1038 T1038;

/* [detachable] SPECIAL [detachable ET_BRACKET_EXPRESSION] */
typedef struct S1039 T1039;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CALL_AGENT] */
typedef struct S1040 T1040;

/* detachable ET_CALL_AGENT */
typedef struct S1041 T1041;

/* [detachable] SPECIAL [detachable ET_CALL_AGENT] */
typedef struct S1042 T1042;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_CALL_EXPRESSION] */
typedef struct S1043 T1043;

/* [detachable] SPECIAL [detachable ET_QUALIFIED_CALL_EXPRESSION] */
typedef struct S1044 T1044;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHECK_INSTRUCTION] */
typedef struct S1045 T1045;

/* detachable ET_CHECK_INSTRUCTION */
typedef struct S1046 T1046;

/* [detachable] SPECIAL [detachable ET_CHECK_INSTRUCTION] */
typedef struct S1047 T1047;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE] */
typedef struct S1048 T1048;

/* [detachable] SPECIAL [detachable ET_CHOICE] */
typedef struct S1050 T1050;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_CONSTANT] */
typedef struct S1051 T1051;

/* [detachable] SPECIAL [detachable ET_CHOICE_CONSTANT] */
typedef struct S1053 T1053;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_ITEM] */
typedef struct S1054 T1054;

/* [detachable] SPECIAL [detachable ET_CHOICE_ITEM] */
typedef struct S1056 T1056;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_LIST] */
typedef struct S1057 T1057;

/* detachable ET_CHOICE_LIST */
typedef struct S1058 T1058;

/* [detachable] SPECIAL [detachable ET_CHOICE_LIST] */
typedef struct S1059 T1059;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLASS] */
typedef struct S1060 T1060;

/* [detachable] SPECIAL [detachable ET_CLASS] */
typedef struct S1061 T1061;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENT_ITEM] */
typedef struct S1062 T1062;

/* [detachable] SPECIAL [detachable ET_CLIENT_ITEM] */
typedef struct S1063 T1063;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENTS] */
typedef struct S1064 T1064;

/* detachable ET_CLIENTS */
typedef struct S1065 T1065;

/* [detachable] SPECIAL [detachable ET_CLIENTS] */
typedef struct S1066 T1066;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_COMPOUND] */
typedef struct S1067 T1067;

/* [detachable] SPECIAL [detachable ET_COMPOUND] */
typedef struct S1068 T1068;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONDITIONAL] */
typedef struct S1069 T1069;

/* [detachable] SPECIAL [detachable ET_CONDITIONAL] */
typedef struct S1071 T1071;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTANT] */
typedef struct S1072 T1072;

/* [detachable] SPECIAL [detachable ET_CONSTANT] */
typedef struct S1073 T1073;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
typedef struct S1074 T1074;

/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
typedef struct S1076 T1076;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST] */
typedef struct S1077 T1077;

/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST */
typedef struct S1078 T1078;

/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST] */
typedef struct S1079 T1079;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_CREATOR] */
typedef struct S1080 T1080;

/* detachable ET_CONSTRAINT_CREATOR */
typedef struct S1081 T1081;

/* [detachable] SPECIAL [detachable ET_CONSTRAINT_CREATOR] */
typedef struct S1082 T1082;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_RENAME_LIST] */
typedef struct S1083 T1083;

/* detachable ET_CONSTRAINT_RENAME_LIST */
typedef struct S1084 T1084;

/* [detachable] SPECIAL [detachable ET_CONSTRAINT_RENAME_LIST] */
typedef struct S1085 T1085;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_TYPE] */
typedef struct S1086 T1086;

/* [detachable] SPECIAL [detachable ET_CONSTRAINT_TYPE] */
typedef struct S1088 T1088;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE] */
typedef struct S1089 T1089;

/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE] */
typedef struct S1091 T1091;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_ITEM] */
typedef struct S1092 T1092;

/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_ITEM] */
typedef struct S1094 T1094;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_LIST] */
typedef struct S1095 T1095;

/* detachable ET_CONVERT_FEATURE_LIST */
typedef struct S1096 T1096;

/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_LIST] */
typedef struct S1097 T1097;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATE_EXPRESSION] */
typedef struct S1098 T1098;

/* [detachable] SPECIAL [detachable ET_CREATE_EXPRESSION] */
typedef struct S1099 T1099;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATION_REGION] */
typedef struct S1100 T1100;

/* detachable ET_CREATION_REGION */
typedef struct S1101 T1101;

/* [detachable] SPECIAL [detachable ET_CREATION_REGION] */
typedef struct S1102 T1102;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR] */
typedef struct S1103 T1103;

/* detachable ET_CREATOR */
typedef struct S1104 T1104;

/* [detachable] SPECIAL [detachable ET_CREATOR] */
typedef struct S1105 T1105;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR_LIST] */
typedef struct S1106 T1106;

/* detachable ET_CREATOR_LIST */
typedef struct S1107 T1107;

/* [detachable] SPECIAL [detachable ET_CREATOR_LIST] */
typedef struct S1108 T1108;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DEBUG_INSTRUCTION] */
typedef struct S1109 T1109;

/* detachable ET_DEBUG_INSTRUCTION */
typedef struct S1110 T1110;

/* [detachable] SPECIAL [detachable ET_DEBUG_INSTRUCTION] */
typedef struct S1111 T1111;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION] */
typedef struct S1112 T1112;

/* detachable ET_ELSEIF_EXPRESSION */
typedef struct S1113 T1113;

/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION] */
typedef struct S1114 T1114;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION_LIST] */
typedef struct S1115 T1115;

/* detachable ET_ELSEIF_EXPRESSION_LIST */
typedef struct S1116 T1116;

/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION_LIST] */
typedef struct S1117 T1117;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART] */
typedef struct S1118 T1118;

/* detachable ET_ELSEIF_PART */
typedef struct S1119 T1119;

/* [detachable] SPECIAL [detachable ET_ELSEIF_PART] */
typedef struct S1120 T1120;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART_LIST] */
typedef struct S1121 T1121;

/* detachable ET_ELSEIF_PART_LIST */
typedef struct S1122 T1122;

/* [detachable] SPECIAL [detachable ET_ELSEIF_PART_LIST] */
typedef struct S1123 T1123;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT] */
typedef struct S1124 T1124;

/* [detachable] SPECIAL [detachable ET_EXPORT] */
typedef struct S1126 T1126;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT_LIST] */
typedef struct S1127 T1127;

/* [detachable] SPECIAL [detachable ET_EXPORT_LIST] */
typedef struct S1128 T1128;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION] */
typedef struct S1129 T1129;

/* [detachable] SPECIAL [detachable ET_EXPRESSION] */
typedef struct S1130 T1130;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION_ITEM] */
typedef struct S1131 T1131;

/* [detachable] SPECIAL [detachable ET_EXPRESSION_ITEM] */
typedef struct S1132 T1132;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTENDED_FEATURE_NAME] */
typedef struct S1133 T1133;

/* [detachable] SPECIAL [detachable ET_EXTENDED_FEATURE_NAME] */
typedef struct S1134 T1134;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTERNAL_ALIAS] */
typedef struct S1135 T1135;

/* [detachable] SPECIAL [detachable ET_EXTERNAL_ALIAS] */
typedef struct S1137 T1137;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE] */
typedef struct S1138 T1138;

/* detachable ET_FEATURE_CLAUSE */
typedef struct S1139 T1139;

/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE] */
typedef struct S1140 T1140;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE_LIST] */
typedef struct S1141 T1141;

/* detachable ET_FEATURE_CLAUSE_LIST */
typedef struct S1142 T1142;

/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE_LIST] */
typedef struct S1143 T1143;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_EXPORT] */
typedef struct S1144 T1144;

/* detachable ET_FEATURE_EXPORT */
typedef struct S1145 T1145;

/* [detachable] SPECIAL [detachable ET_FEATURE_EXPORT] */
typedef struct S1146 T1146;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME] */
typedef struct S1147 T1147;

/* [detachable] SPECIAL [detachable ET_FEATURE_NAME] */
typedef struct S1148 T1148;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME_ITEM] */
typedef struct S1149 T1149;

/* [detachable] SPECIAL [detachable ET_FEATURE_NAME_ITEM] */
typedef struct S1151 T1151;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT] */
typedef struct S1152 T1152;

/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT] */
typedef struct S1153 T1153;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_ITEM] */
typedef struct S1154 T1154;

/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_ITEM] */
typedef struct S1156 T1156;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_LIST] */
typedef struct S1157 T1157;

/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_LIST] */
typedef struct S1158 T1158;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER] */
typedef struct S1159 T1159;

/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER] */
typedef struct S1160 T1160;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_ITEM] */
typedef struct S1161 T1161;

/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_ITEM] */
typedef struct S1163 T1163;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_LIST] */
typedef struct S1164 T1164;

/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_LIST] */
typedef struct S1165 T1165;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_EXPRESSION] */
typedef struct S1166 T1166;

/* detachable ET_IF_EXPRESSION */
typedef struct S1167 T1167;

/* [detachable] SPECIAL [detachable ET_IF_EXPRESSION] */
typedef struct S1168 T1168;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_INSTRUCTION] */
typedef struct S1169 T1169;

/* detachable ET_IF_INSTRUCTION */
typedef struct S1170 T1170;

/* [detachable] SPECIAL [detachable ET_IF_INSTRUCTION] */
typedef struct S1171 T1171;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_LIST] */
typedef struct S1172 T1172;

/* detachable ET_INDEXING_LIST */
typedef struct S1173 T1173;

/* [detachable] SPECIAL [detachable ET_INDEXING_LIST] */
typedef struct S1174 T1174;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_ITEM] */
typedef struct S1175 T1175;

/* [detachable] SPECIAL [detachable ET_INDEXING_ITEM] */
typedef struct S1177 T1177;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM] */
typedef struct S1178 T1178;

/* [detachable] SPECIAL [detachable ET_INDEXING_TERM] */
typedef struct S1179 T1179;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_ITEM] */
typedef struct S1180 T1180;

/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_ITEM] */
typedef struct S1182 T1182;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_LIST] */
typedef struct S1183 T1183;

/* detachable ET_INDEXING_TERM_LIST */
typedef struct S1184 T1184;

/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_LIST] */
typedef struct S1185 T1185;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INLINE_AGENT] */
typedef struct S1186 T1186;

/* [detachable] SPECIAL [detachable ET_INLINE_AGENT] */
typedef struct S1187 T1187;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSPECT_EXPRESSION] */
typedef struct S1188 T1188;

/* detachable ET_INSPECT_EXPRESSION */
typedef struct S1189 T1189;

/* [detachable] SPECIAL [detachable ET_INSPECT_EXPRESSION] */
typedef struct S1190 T1190;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSPECT_INSTRUCTION] */
typedef struct S1191 T1191;

/* detachable ET_INSPECT_INSTRUCTION */
typedef struct S1192 T1192;

/* [detachable] SPECIAL [detachable ET_INSPECT_INSTRUCTION] */
typedef struct S1193 T1193;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSTRUCTION] */
typedef struct S1194 T1194;

/* [detachable] SPECIAL [detachable ET_INSTRUCTION] */
typedef struct S1196 T1196;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INVARIANTS] */
typedef struct S1197 T1197;

/* detachable ET_INVARIANTS */
typedef struct S1198 T1198;

/* [detachable] SPECIAL [detachable ET_INVARIANTS] */
typedef struct S1199 T1199;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_FEATURE_NAME_LIST] */
typedef struct S1200 T1200;

/* [detachable] SPECIAL [detachable ET_KEYWORD_FEATURE_NAME_LIST] */
typedef struct S1201 T1201;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LIKE_TYPE] */
typedef struct S1202 T1202;

/* [detachable] SPECIAL [detachable ET_LIKE_TYPE] */
typedef struct S1204 T1204;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE] */
typedef struct S1205 T1205;

/* detachable ET_LOCAL_VARIABLE */
typedef struct S1206 T1206;

/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE] */
typedef struct S1207 T1207;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_ITEM] */
typedef struct S1208 T1208;

/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_ITEM] */
typedef struct S1210 T1210;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_LIST] */
typedef struct S1211 T1211;

/* detachable ET_LOCAL_VARIABLE_LIST */
typedef struct S1212 T1212;

/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_LIST] */
typedef struct S1213 T1213;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOOP_INVARIANTS] */
typedef struct S1214 T1214;

/* detachable ET_LOOP_INVARIANTS */
typedef struct S1215 T1215;

/* [detachable] SPECIAL [detachable ET_LOOP_INVARIANTS] */
typedef struct S1216 T1216;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_ARRAY] */
typedef struct S1217 T1217;

/* detachable ET_MANIFEST_ARRAY */
typedef struct S1218 T1218;

/* [detachable] SPECIAL [detachable ET_MANIFEST_ARRAY] */
typedef struct S1219 T1219;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_ITEM] */
typedef struct S1220 T1220;

/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_ITEM] */
typedef struct S1222 T1222;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_LIST] */
typedef struct S1223 T1223;

/* detachable ET_MANIFEST_STRING_LIST */
typedef struct S1224 T1224;

/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_LIST] */
typedef struct S1225 T1225;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_TUPLE] */
typedef struct S1226 T1226;

/* [detachable] SPECIAL [detachable ET_MANIFEST_TUPLE] */
typedef struct S1227 T1227;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBSOLETE] */
typedef struct S1228 T1228;

/* [detachable] SPECIAL [detachable ET_OBSOLETE] */
typedef struct S1230 T1230;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENTHESIZED_EXPRESSION] */
typedef struct S1231 T1231;

/* detachable ET_PARENTHESIZED_EXPRESSION */
typedef struct S1232 T1232;

/* [detachable] SPECIAL [detachable ET_PARENTHESIZED_EXPRESSION] */
typedef struct S1233 T1233;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT] */
typedef struct S1234 T1234;

/* [detachable] SPECIAL [detachable ET_PARENT] */
typedef struct S1235 T1235;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_CLAUSE_LIST] */
typedef struct S1236 T1236;

/* detachable ET_PARENT_CLAUSE_LIST */
typedef struct S1237 T1237;

/* [detachable] SPECIAL [detachable ET_PARENT_CLAUSE_LIST] */
typedef struct S1238 T1238;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_ITEM] */
typedef struct S1239 T1239;

/* [detachable] SPECIAL [detachable ET_PARENT_ITEM] */
typedef struct S1240 T1240;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_LIST] */
typedef struct S1241 T1241;

/* [detachable] SPECIAL [detachable ET_PARENT_LIST] */
typedef struct S1242 T1242;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSTCONDITIONS] */
typedef struct S1243 T1243;

/* detachable ET_POSTCONDITIONS */
typedef struct S1244 T1244;

/* [detachable] SPECIAL [detachable ET_POSTCONDITIONS] */
typedef struct S1245 T1245;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECONDITIONS] */
typedef struct S1246 T1246;

/* [detachable] SPECIAL [detachable ET_PRECONDITIONS] */
typedef struct S1247 T1247;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PROCEDURE] */
typedef struct S1248 T1248;

/* [detachable] SPECIAL [detachable ET_PROCEDURE] */
typedef struct S1249 T1249;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_LIKE_IDENTIFIER] */
typedef struct S1250 T1250;

/* [detachable] SPECIAL [detachable ET_QUALIFIED_LIKE_IDENTIFIER] */
typedef struct S1252 T1252;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUANTIFIER_EXPRESSION] */
typedef struct S1253 T1253;

/* detachable ET_QUANTIFIER_EXPRESSION */
typedef struct S1254 T1254;

/* [detachable] SPECIAL [detachable ET_QUANTIFIER_EXPRESSION] */
typedef struct S1255 T1255;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUERY] */
typedef struct S1256 T1256;

/* [detachable] SPECIAL [detachable ET_QUERY] */
typedef struct S1257 T1257;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_ITEM] */
typedef struct S1258 T1258;

/* [detachable] SPECIAL [detachable ET_RENAME_ITEM] */
typedef struct S1260 T1260;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_LIST] */
typedef struct S1261 T1261;

/* [detachable] SPECIAL [detachable ET_RENAME_LIST] */
typedef struct S1262 T1262;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_REPEAT_INSTRUCTION] */
typedef struct S1263 T1263;

/* detachable ET_REPEAT_INSTRUCTION */
typedef struct S1264 T1264;

/* [detachable] SPECIAL [detachable ET_REPEAT_INSTRUCTION] */
typedef struct S1265 T1265;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_STATIC_CALL_EXPRESSION] */
typedef struct S1266 T1266;

/* detachable ET_STATIC_CALL_EXPRESSION */
typedef struct S1267 T1267;

/* [detachable] SPECIAL [detachable ET_STATIC_CALL_EXPRESSION] */
typedef struct S1268 T1268;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_STRIP_EXPRESSION] */
typedef struct S1269 T1269;

/* detachable ET_STRIP_EXPRESSION */
typedef struct S1270 T1270;

/* [detachable] SPECIAL [detachable ET_STRIP_EXPRESSION] */
typedef struct S1271 T1271;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE] */
typedef struct S1272 T1272;

/* [detachable] SPECIAL [detachable ET_TYPE] */
typedef struct S1273 T1273;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT] */
typedef struct S1274 T1274;

/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT] */
typedef struct S1276 T1276;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_ITEM] */
typedef struct S1277 T1277;

/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_ITEM] */
typedef struct S1279 T1279;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_LIST] */
typedef struct S1280 T1280;

/* detachable ET_TYPE_CONSTRAINT_LIST */
typedef struct S1281 T1281;

/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_LIST] */
typedef struct S1282 T1282;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_ITEM] */
typedef struct S1283 T1283;

/* [detachable] SPECIAL [detachable ET_TYPE_ITEM] */
typedef struct S1285 T1285;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_LIST] */
typedef struct S1286 T1286;

/* [detachable] SPECIAL [detachable ET_TYPE_LIST] */
typedef struct S1288 T1288;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_VARIANT] */
typedef struct S1289 T1289;

/* detachable ET_VARIANT */
typedef struct S1290 T1290;

/* [detachable] SPECIAL [detachable ET_VARIANT] */
typedef struct S1291 T1291;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_EXPRESSION] */
typedef struct S1292 T1292;

/* detachable ET_WHEN_EXPRESSION */
typedef struct S1293 T1293;

/* [detachable] SPECIAL [detachable ET_WHEN_EXPRESSION] */
typedef struct S1294 T1294;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_EXPRESSION_LIST] */
typedef struct S1295 T1295;

/* detachable ET_WHEN_EXPRESSION_LIST */
typedef struct S1296 T1296;

/* [detachable] SPECIAL [detachable ET_WHEN_EXPRESSION_LIST] */
typedef struct S1297 T1297;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART] */
typedef struct S1298 T1298;

/* detachable ET_WHEN_PART */
typedef struct S1299 T1299;

/* [detachable] SPECIAL [detachable ET_WHEN_PART] */
typedef struct S1300 T1300;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART_LIST] */
typedef struct S1301 T1301;

/* detachable ET_WHEN_PART_LIST */
typedef struct S1302 T1302;

/* [detachable] SPECIAL [detachable ET_WHEN_PART_LIST] */
typedef struct S1303 T1303;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WRITABLE] */
typedef struct S1304 T1304;

/* [detachable] SPECIAL [detachable ET_WRITABLE] */
typedef struct S1305 T1305;

/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS] */
typedef struct S1306 T1306;

/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS] */
typedef struct S1307 T1307;

/* [detachable] ET_PARENT_CHECKER1 */
typedef struct S1308 T1308;

/* [detachable] ET_FORMAL_PARAMETER_CHECKER1 */
typedef struct S1309 T1309;

/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME] */
typedef struct S1310 T1310;

/* [detachable] ET_FEATURE_NAME_TESTER */
typedef struct S1311 T1311;

/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME] */
typedef struct S1313 T1313;

/* [detachable] ET_ALIAS_NAME_TESTER */
typedef struct S1314 T1314;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST] */
typedef struct S1316 T1316;

/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS] */
typedef struct S1317 T1317;

/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER */
typedef struct S1318 T1318;

/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER */
typedef struct S1319 T1319;

/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER */
typedef struct S1320 T1320;

/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1 */
typedef struct S1321 T1321;

/* [detachable] ET_ANCHORED_TYPE_CHECKER */
typedef struct S1322 T1322;

/* [detachable] ET_SIGNATURE_CHECKER */
typedef struct S1323 T1323;

/* [detachable] ET_PARENT_CHECKER2 */
typedef struct S1324 T1324;

/* [detachable] ET_FORMAL_PARAMETER_CHECKER2 */
typedef struct S1325 T1325;

/* [detachable] ET_BUILTIN_FEATURE_CHECKER */
typedef struct S1326 T1326;

/* [detachable] ET_PRECURSOR_CHECKER */
typedef struct S1327 T1327;

/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32] */
typedef struct S1328 T1328;

/* [detachable] DS_HASH_SET [[attached] ET_CLASS] */
typedef struct S1329 T1329;

/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER */
typedef struct S1330 T1330;

/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2 */
typedef struct S1331 T1331;

/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME] */
typedef struct S1332 T1332;

/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME] */
typedef struct S1333 T1333;

/* [detachable] ET_RENAME */
typedef struct S1337 T1337;

/* [detachable] ET_PARENT_CHECKER3 */
typedef struct S1338 T1338;

/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE] */
typedef struct S1339 T1339;

/* [detachable] DS_HASH_SET [[attached] ET_QUERY] */
typedef struct S1340 T1340;

/* [detachable] ET_SUPPLIER_BUILDER */
typedef struct S1341 T1341;

/* detachable ET_PARENT_FEATURE */
typedef struct S1344 T1344;

/* [detachable] ET_INHERITED_FEATURE */
typedef struct S1345 T1345;

/* [detachable] SPECIAL [[attached] ET_CLASS] */
typedef struct S1346 T1346;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS] */
typedef struct S1347 T1347;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS] */
typedef struct S1348 T1348;

/* [detachable] DS_ARRAYED_STACK [[attached] YY_BUFFER] */
typedef struct S1349 T1349;

/* [detachable] UT_SYNTAX_ERROR */
typedef struct S1350 T1350;

/* [detachable] UT_TOO_MANY_INCLUDES_ERROR */
typedef struct S1351 T1351;

/* detachable ET_DYNAMIC_TARGET_LIST */
typedef struct S1353 T1353;

/* [detachable] SPECIAL [[attached] NATIVE_STRING] */
typedef struct S1354 T1354;

/* [detachable] ARRAY [detachable STRING_8] */
typedef struct S1355 T1355;

/* [detachable] XM_EIFFEL_CHARACTER_ENTITY */
typedef struct S1356 T1356;

/* detachable DS_BILINKABLE [[attached] XM_POSITION] */
typedef struct S1358 T1358;

/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_POSITION] */
typedef struct S1359 T1359;

/* detachable DS_LINKABLE [[attached] XM_EIFFEL_SCANNER] */
typedef struct S1360 T1360;

/* [detachable] SPECIAL [[attached] XM_EIFFEL_ENTITY_DEF] */
typedef struct S1361 T1361;

/* detachable DS_HASH_TABLE_CURSOR [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8] */
typedef struct S1364 T1364;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_ENTITY_DEF] */
typedef struct S1365 T1365;

/* detachable KL_EQUALITY_TESTER [[attached] XM_EIFFEL_PARSER_NAME] */
typedef struct S1368 T1368;

/* [detachable] DS_HASH_SET_CURSOR [[attached] XM_EIFFEL_PARSER_NAME] */
typedef struct S1369 T1369;

/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ELEMENT_CONTENT] */
typedef struct S1370 T1370;

/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
typedef struct S1371 T1371;

/* detachable DS_BILINKABLE [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
typedef struct S1372 T1372;

/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] STRING_8] */
typedef struct S1373 T1373;

/* detachable DS_BILINKABLE [[attached] STRING_8] */
typedef struct S1374 T1374;

/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT */
typedef struct S1375 T1375;

/* [detachable] DS_LINKED_QUEUE [detachable STRING_8] */
typedef struct S1378 T1378;

/* [detachable] DS_LINKED_QUEUE [[attached] STRING_8] */
typedef struct S1379 T1379;

/* detachable KL_EQUALITY_TESTER [[attached] XM_NAMESPACE] */
typedef struct S1380 T1380;

/* [detachable] DS_HASH_SET_CURSOR [[attached] XM_NAMESPACE] */
typedef struct S1381 T1381;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_NAMESPACE] */
typedef struct S1382 T1382;

/* [detachable] SPECIAL [[attached] XM_NAMESPACE] */
typedef struct S1383 T1383;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLUSTER] */
typedef struct S1384 T1384;

/* [detachable] SPECIAL [[attached] ET_PARENT_ITEM] */
typedef struct S1386 T1386;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_ITEM] */
typedef struct S1387 T1387;

/* [detachable] ET_BRACED_TYPE_LIST */
typedef struct S1388 T1388;

/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS] */
typedef struct S1389 T1389;

/* [detachable] SPECIAL [[attached] ET_ADAPTED_LIBRARY] */
typedef struct S1390 T1390;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_LIBRARY] */
typedef struct S1391 T1391;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_LIBRARY] */
typedef struct S1392 T1392;

/* detachable DS_LINKABLE [[attached] XM_DOCUMENT_NODE] */
typedef struct S1393 T1393;

/* [detachable] SPECIAL [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
typedef struct S1395 T1395;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
typedef struct S1396 T1396;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
typedef struct S1397 T1397;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY] */
typedef struct S1398 T1398;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32] */
typedef struct S1400 T1400;

/* [detachable] SPECIAL [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL] */
typedef struct S1402 T1402;

/* [detachable] SPECIAL [[attached] ET_CALL_NAME] */
typedef struct S1404 T1404;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL] */
typedef struct S1406 T1406;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CALL_NAME] */
typedef struct S1407 T1407;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME] */
typedef struct S1408 T1408;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME] */
typedef struct S1410 T1410;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL] */
typedef struct S1411 T1411;

/* [detachable] SPECIAL [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL] */
typedef struct S1412 T1412;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S1413 T1413;

/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]] */
typedef struct S1414 T1414;

/* [detachable] KL_SPECIAL_ROUTINES [detachable DS_LINKABLE [[attached] INTEGER_32]] */
typedef struct S1415 T1415;

/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable DS_LINKABLE [[attached] INTEGER_32]] */
typedef struct S1416 T1416;

/* [detachable] SPECIAL [[attached] ET_SYSTEM_PROCESSOR] */
typedef struct S1418 T1418;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_SYSTEM_PROCESSOR] */
typedef struct S1419 T1419;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_SYSTEM_PROCESSOR] */
typedef struct S1420 T1420;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_CLASS] */
typedef struct S1421 T1421;

/* [detachable] SPECIAL [[attached] ET_ADAPTED_CLASS] */
typedef struct S1423 T1423;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_CLASS] */
typedef struct S1424 T1424;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PROCEDURE] */
typedef struct S1425 T1425;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUERY] */
typedef struct S1426 T1426;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NAMED_OBJECT_TEST] */
typedef struct S1427 T1427;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ITERATION_COMPONENT] */
typedef struct S1428 T1428;

/* [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32] */
typedef struct S1429 T1429;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_8] */
typedef struct S1433 T1433;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_8, [attached] ET_CLASS_NAME] */
typedef struct S1434 T1434;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] RX_CHARACTER_SET] */
typedef struct S1435 T1435;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] RX_CHARACTER_SET] */
typedef struct S1436 T1436;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_64] */
typedef struct S1443 T1443;

/* [detachable] TYPE [[attached] NATURAL_64] */
#define T1444 EIF_TYPE_OBJ

/* [detachable] TYPE [[attached] NATURAL_32] */
#define T1445 EIF_TYPE_OBJ

/* [detachable] KL_CHARACTER_BUFFER */
typedef struct S1446 T1446;

/* [detachable] UT_TRISTATE */
typedef struct S1448 T1448;

/* [detachable] KL_AGENT_ROUTINES [[attached] ANY] */
typedef struct S1449 T1449;

/* [detachable] TUPLE [[attached] UT_TRISTATE] */
typedef struct S1450 T1450;

/* [detachable] TUPLE [[attached] KL_AGENT_ROUTINES [[attached] ANY], [attached] PROCEDURE [[attached] TUPLE]] */
typedef struct S1451 T1451;

/* [detachable] PREDICATE [[attached] TUPLE [[attached] ET_MASTER_CLASS]] */
typedef struct S1452 T1452;

/* [detachable] PREDICATE [[attached] TUPLE] */
typedef struct S1453 T1453;

/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE] */
typedef struct S1455 T1455;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE] */
typedef struct S1456 T1456;

/* [detachable] TUPLE [[attached] ET_ECF_SYSTEM, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
typedef struct S1457 T1457;

/* [detachable] KL_AGENT_ROUTINES [[attached] ET_UNIVERSE] */
typedef struct S1458 T1458;

/* [detachable] PREDICATE [[attached] TUPLE [[attached] ET_UNIVERSE]] */
typedef struct S1459 T1459;

/* [detachable] TUPLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE]] */
typedef struct S1460 T1460;

/* [detachable] TUPLE [[attached] KL_AGENT_ROUTINES [[attached] ET_UNIVERSE], [attached] PREDICATE [[attached] TUPLE [[attached] ET_UNIVERSE]]] */
typedef struct S1461 T1461;

/* [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]], [attached] FUNCTION [[attached] TUPLE, [attached] BOOLEAN]] */
typedef struct S1463 T1463;

/* [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]], [attached] FUNCTION [[attached] TUPLE [[attached] ET_MASTER_CLASS], [attached] BOOLEAN]] */
typedef struct S1464 T1464;

/* [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]], [attached] FUNCTION [[attached] TUPLE [[attached] ET_MASTER_CLASS], [attached] BOOLEAN], [attached] FUNCTION [[attached] TUPLE, [attached] BOOLEAN]] */
typedef struct S1465 T1465;

/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET */
typedef struct S1466 T1466;

/* TYPED_POINTER [[attached] NATURAL_16] */
typedef struct S1468 T1468;
extern T0* GE_boxed1468(T1468 a1);
extern T0* GE_boxed_pointer1468(T1468* a1);
typedef struct Sb1468 Tb1468;

/* [detachable] XM_COMMENT */
typedef struct S1469 T1469;

/* [detachable] XM_PROCESSING_INSTRUCTION */
typedef struct S1470 T1470;

/* [detachable] XM_CHARACTER_DATA */
typedef struct S1471 T1471;

/* [detachable] ET_LIKE_N */
typedef struct S1474 T1474;

/* [detachable] KL_STRING_INPUT_STREAM */
typedef struct S1477 T1477;

/* [detachable] KL_UNICODE_CHARACTER_BUFFER */
typedef struct S1479 T1479;

/* detachable ET_OBJECT_TEST_LIST */
typedef struct S1481 T1481;

/* detachable ET_C3_CHARACTER_CONSTANT */
typedef struct S1482 T1482;

/* detachable ET_REGULAR_MANIFEST_STRING */
typedef struct S1483 T1483;

/* detachable ET_SPECIAL_MANIFEST_STRING */
typedef struct S1484 T1484;

/* detachable ET_VERBATIM_STRING */
typedef struct S1485 T1485;

/* detachable ET_UNDERSCORED_INTEGER_CONSTANT */
typedef struct S1486 T1486;

/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT */
typedef struct S1487 T1487;

/* detachable ET_OCTAL_INTEGER_CONSTANT */
typedef struct S1488 T1488;

/* detachable ET_BINARY_INTEGER_CONSTANT */
typedef struct S1489 T1489;

/* detachable ET_REGULAR_REAL_CONSTANT */
typedef struct S1490 T1490;

/* detachable ET_UNDERSCORED_REAL_CONSTANT */
typedef struct S1491 T1491;

/* detachable ET_TRUE_CONSTANT */
typedef struct S1493 T1493;

/* detachable ET_FALSE_CONSTANT */
typedef struct S1494 T1494;

/* detachable ET_C1_CHARACTER_CONSTANT */
typedef struct S1495 T1495;

/* detachable ET_C2_CHARACTER_CONSTANT */
typedef struct S1496 T1496;

/* detachable ET_TAGGED_INDEXING */
typedef struct S1498 T1498;

/* detachable ET_INDEXING */
typedef struct S1499 T1499;

/* detachable ET_CUSTOM_ATTRIBUTE */
typedef struct S1500 T1500;

/* detachable ET_CONSTRAINED_FORMAL_PARAMETER */
typedef struct S1501 T1501;

/* detachable ET_TYPE_RENAME_CONSTRAINT */
typedef struct S1502 T1502;

/* detachable ET_CONSTRAINT_NAMED_TYPE */
typedef struct S1503 T1503;

/* detachable ET_CONSTRAINT_LABELED_ACTUAL_PARAMETER */
typedef struct S1504 T1504;

/* detachable ET_CONSTRAINT_LABELED_COMMA_ACTUAL_PARAMETER */
typedef struct S1505 T1505;

/* detachable ET_ALL_EXPORT */
typedef struct S1506 T1506;

/* detachable ET_CONVERT_FUNCTION */
typedef struct S1508 T1508;

/* detachable ET_CONVERT_PROCEDURE */
typedef struct S1509 T1509;

/* detachable ET_DO_FUNCTION */
typedef struct S1511 T1511;

/* detachable ET_ONCE_FUNCTION */
typedef struct S1512 T1512;

/* detachable ET_DEFERRED_FUNCTION */
typedef struct S1513 T1513;

/* detachable ET_ONCE_PROCEDURE */
typedef struct S1515 T1515;

/* detachable ET_EXTERNAL_PROCEDURE */
typedef struct S1516 T1516;

/* detachable ET_ALIASED_FEATURE_NAME */
typedef struct S1517 T1517;

/* detachable ET_ALIAS_FREE_NAME */
typedef struct S1518 T1518;

/* detachable ET_CLASS_ASSERTION */
typedef struct S1521 T1521;

/* detachable ET_LABELED_ACTUAL_PARAMETER */
typedef struct S1522 T1522;

/* detachable ET_LIKE_FEATURE */
typedef struct S1524 T1524;

/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE */
typedef struct S1525 T1525;

/* detachable ET_QUALIFIED_LIKE_TYPE */
typedef struct S1526 T1526;

/* detachable ET_ASSIGNER_INSTRUCTION */
typedef struct S1527 T1527;

/* detachable ET_ASSIGNMENT */
typedef struct S1528 T1528;

/* detachable ET_ASSIGNMENT_ATTEMPT */
typedef struct S1529 T1529;

/* detachable ET_LOOP_INSTRUCTION */
typedef struct S1530 T1530;

/* detachable ET_BANG_INSTRUCTION */
typedef struct S1532 T1532;

/* detachable ET_CREATE_INSTRUCTION */
typedef struct S1533 T1533;

/* detachable ET_CHOICE_RANGE */
typedef struct S1534 T1534;

/* detachable ET_PRECURSOR_INSTRUCTION */
typedef struct S1535 T1535;

/* detachable ET_STATIC_CALL_INSTRUCTION */
typedef struct S1537 T1537;

/* detachable ET_PRECURSOR_EXPRESSION */
typedef struct S1538 T1538;

/* detachable ET_FEATURE_ADDRESS */
typedef struct S1539 T1539;

/* detachable ET_CURRENT_ADDRESS */
typedef struct S1540 T1540;

/* detachable ET_RESULT_ADDRESS */
typedef struct S1541 T1541;

/* detachable ET_EXPRESSION_ADDRESS */
typedef struct S1542 T1542;

/* detachable ET_INFIX_EXPRESSION */
typedef struct S1543 T1543;

/* detachable ET_INFIX_AND_THEN_OPERATOR */
typedef struct S1544 T1544;

/* detachable ET_INFIX_OR_ELSE_OPERATOR */
typedef struct S1545 T1545;

/* detachable ET_MANIFEST_TYPE */
typedef struct S1546 T1546;

/* detachable ET_PREFIX_EXPRESSION */
typedef struct S1547 T1547;

/* detachable ET_OLD_EXPRESSION */
typedef struct S1548 T1548;

/* detachable ET_OLD_OBJECT_TEST */
typedef struct S1549 T1549;

/* detachable ET_ONCE_MANIFEST_STRING */
typedef struct S1550 T1550;

/* detachable ET_ITERATION_CURSOR */
typedef struct S1551 T1551;

/* detachable ET_DO_FUNCTION_INLINE_AGENT */
typedef struct S1552 T1552;

/* detachable ET_ONCE_FUNCTION_INLINE_AGENT */
typedef struct S1553 T1553;

/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT */
typedef struct S1554 T1554;

/* detachable ET_DO_PROCEDURE_INLINE_AGENT */
typedef struct S1555 T1555;

/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT */
typedef struct S1556 T1556;

/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT */
typedef struct S1557 T1557;

/* detachable ET_AGENT_OPEN_TARGET */
typedef struct S1558 T1558;

/* detachable ET_AGENT_TYPED_OPEN_ARGUMENT */
typedef struct S1559 T1559;

/* detachable ET_TAGGED_ASSERTION */
typedef struct S1562 T1562;

/* [detachable] ET_FILE_POSITION */
typedef struct S1564 T1564;

/* [detachable] DS_QUICK_SORTER [[attached] ET_QUERY] */
typedef struct S1566 T1566;

/* [detachable] DS_QUICK_SORTER [[attached] ET_PROCEDURE] */
typedef struct S1567 T1567;

/* [detachable] ET_REDECLARED_FEATURE */
typedef struct S1569 T1569;

/* [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER */
typedef struct S1571 T1571;

/* [detachable] TUPLE [[attached] ET_MASTER_CLASS, [attached] ET_SYSTEM_PROCESSOR] */
typedef struct S1573 T1573;

/* [detachable] SPECIAL [[attached] NATURAL_16] */
typedef struct S1576 T1576;

/* [detachable] ARRAY [[attached] NATURAL_16] */
typedef struct S1577 T1577;

/* [detachable] ARRAY [[attached] NATURAL_32] */
typedef struct S1578 T1578;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER] */
typedef struct S1579 T1579;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY] */
typedef struct S1580 T1580;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
typedef struct S1581 T1581;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE] */
typedef struct S1582 T1582;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG] */
typedef struct S1583 T1583;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE] */
typedef struct S1584 T1584;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY] */
typedef struct S1585 T1585;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
typedef struct S1586 T1586;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE] */
typedef struct S1587 T1587;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT] */
typedef struct S1588 T1588;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE] */
typedef struct S1589 T1589;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ACTION] */
typedef struct S1590 T1590;

/* [detachable] SPECIAL [[attached] ET_ECF_ACTION] */
typedef struct S1591 T1591;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ACTION] */
typedef struct S1592 T1592;

/* detachable DS_ARRAYED_LIST [[attached] RX_PCRE_REGULAR_EXPRESSION] */
typedef struct S1593 T1593;

/* [detachable] SPECIAL [[attached] ET_ECF_OPTIONS] */
typedef struct S1594 T1594;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_OPTIONS] */
typedef struct S1597 T1597;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_OPTIONS, [attached] STRING_8] */
typedef struct S1598 T1598;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_VISIBLE_CLASS] */
typedef struct S1599 T1599;

/* [detachable] SPECIAL [[attached] ET_ECF_VISIBLE_CLASS] */
typedef struct S1600 T1600;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_VISIBLE_CLASS] */
typedef struct S1601 T1601;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS] */
typedef struct S1602 T1602;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CONDITION_ITEM] */
typedef struct S1603 T1603;

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DOTNET_ASSEMBLY] */
typedef struct S1604 T1604;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_ASSEMBLY] */
typedef struct S1605 T1605;

/* [detachable] SPECIAL [[attached] ET_DOTNET_ASSEMBLY] */
typedef struct S1606 T1606;

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_UNIVERSE] */
typedef struct S1608 T1608;

/* [detachable] SPECIAL [[attached] ET_UNIVERSE] */
typedef struct S1609 T1609;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_UNIVERSE] */
typedef struct S1611 T1611;

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_INTERNAL_UNIVERSE] */
typedef struct S1612 T1612;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INTERNAL_UNIVERSE] */
typedef struct S1614 T1614;

/* [detachable] SPECIAL [[attached] ET_INTERNAL_UNIVERSE] */
typedef struct S1615 T1615;

/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1 */
typedef struct S1616 T1616;

/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2 */
typedef struct S1617 T1617;

/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER */
typedef struct S1618 T1618;

/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3 */
typedef struct S1619 T1619;

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS] */
typedef struct S1620 T1620;

/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_TYPE] */
typedef struct S1621 T1621;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_TYPE] */
typedef struct S1622 T1622;

/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_FORMAL_PARAMETER_TYPE] */
typedef struct S1623 T1623;

/* [detachable] CONSOLE */
typedef struct S1624 T1624;

/* [detachable] DS_LINKED_QUEUE [[attached] CHARACTER_8] */
typedef struct S1626 T1626;

/* [detachable] UC_UTF16_ROUTINES */
typedef struct S1627 T1627;

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_LIBRARY] */
typedef struct S1628 T1628;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIBRARY] */
typedef struct S1629 T1629;

/* [detachable] SPECIAL [[attached] ET_LIBRARY] */
typedef struct S1630 T1630;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_TARGET] */
typedef struct S1631 T1631;

/* [detachable] SPECIAL [[attached] ET_ADAPTED_UNIVERSE] */
typedef struct S1632 T1632;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_UNIVERSE] */
typedef struct S1633 T1633;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_UNIVERSE] */
typedef struct S1634 T1634;

/* [detachable] SPECIAL [[attached] ET_ECF_TARGET_PARENT] */
typedef struct S1635 T1635;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET_PARENT] */
typedef struct S1636 T1636;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_TARGET_PARENT] */
typedef struct S1637 T1637;

/* [detachable] YY_UNICODE_BUFFER */
typedef struct S1638 T1638;

/* [detachable] SPECIAL [detachable ET_OBJECT_TEST_LIST] */
typedef struct S1639 T1639;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBJECT_TEST_LIST] */
typedef struct S1640 T1640;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST_LIST] */
typedef struct S1641 T1641;

/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST_LIST] */
typedef struct S1642 T1642;

/* [detachable] SPECIAL [detachable ET_ITERATION_COMPONENT_LIST] */
typedef struct S1643 T1643;

/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ITERATION_COMPONENT_LIST] */
typedef struct S1644 T1644;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ITERATION_COMPONENT_LIST] */
typedef struct S1645 T1645;

/* [detachable] SPECIAL [[attached] ET_ITERATION_COMPONENT_LIST] */
typedef struct S1646 T1646;

/* [detachable] SPECIAL [[attached] ET_ASSERTION_ITEM] */
typedef struct S1647 T1647;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTION_ITEM] */
typedef struct S1648 T1648;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ASSERTION_ITEM] */
typedef struct S1649 T1649;

/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_CONSTRAINT_TYPE] */
typedef struct S1650 T1650;

/* [detachable] SPECIAL [[attached] ET_NAMED_CLASS] */
typedef struct S1651 T1651;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_CLASS] */
typedef struct S1653 T1653;

/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS] */
typedef struct S1655 T1655;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_BASE_TYPE, [attached] ET_CLASS] */
typedef struct S1656 T1656;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT] */
typedef struct S1659 T1659;

/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32] */
typedef struct S1660 T1660;

/* [detachable] ET_BASE_TYPE_CONSTRAINT_LIST */
typedef struct S1661 T1661;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME] */
typedef struct S1662 T1662;

/* [detachable] SPECIAL [[attached] ET_FLATTENED_FEATURE] */
typedef struct S1663 T1663;

/* [detachable] SPECIAL [[attached] ET_FEATURE_NAME] */
typedef struct S1666 T1666;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FLATTENED_FEATURE] */
typedef struct S1667 T1667;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME] */
typedef struct S1668 T1668;

/* [detachable] SPECIAL [[attached] ET_ALIAS_NAME] */
typedef struct S1669 T1669;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ALIAS_NAME] */
typedef struct S1671 T1671;

/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME] */
typedef struct S1672 T1672;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_LIST] */
typedef struct S1673 T1673;

/* [detachable] SPECIAL [[attached] ET_CLIENT_LIST] */
typedef struct S1674 T1674;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLIENT_LIST] */
typedef struct S1675 T1675;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS] */
typedef struct S1677 T1677;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT] */
typedef struct S1678 T1678;

/* [detachable] SPECIAL [[attached] ET_CLIENT] */
typedef struct S1679 T1679;

/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME] */
typedef struct S1680 T1680;

/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME] */
typedef struct S1681 T1681;

/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32] */
typedef struct S1682 T1682;

/* [detachable] ET_REPLICATED_FEATURE */
typedef struct S1684 T1684;

/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE] */
typedef struct S1685 T1685;

/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE] */
typedef struct S1686 T1686;

/* [detachable] ET_DOTNET_SIGNATURE_TESTER */
typedef struct S1687 T1687;

/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE] */
typedef struct S1688 T1688;

/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE] */
typedef struct S1690 T1690;

/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE] */
typedef struct S1692 T1692;

/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE] */
typedef struct S1693 T1693;

/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE] */
typedef struct S1694 T1694;

/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_PARENT_FEATURE] */
typedef struct S1695 T1695;

/* detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE] */
typedef struct S1696 T1696;

/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8] */
typedef struct S1697 T1697;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE, [attached] INTEGER_32] */
typedef struct S1698 T1698;

/* [detachable] TYPE [[attached] ET_FEATURE] */
#define T1699 EIF_TYPE_OBJ

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_CLASS] */
typedef struct S1701 T1701;

/* [detachable] TYPE [[attached] ET_CLASS] */
#define T1702 EIF_TYPE_OBJ

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_FEATURE_NAME] */
typedef struct S1703 T1703;

/* [detachable] SPECIAL [[attached] ET_RENAME] */
typedef struct S1704 T1704;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME] */
typedef struct S1707 T1707;

/* [detachable] TYPE [[attached] ET_RENAME] */
#define T1708 EIF_TYPE_OBJ

/* [detachable] TYPE [[attached] ET_FEATURE_NAME] */
#define T1709 EIF_TYPE_OBJ

/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_ALIAS_NAME] */
typedef struct S1711 T1711;

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_PROCEDURE] */
typedef struct S1712 T1712;

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_QUERY] */
typedef struct S1713 T1713;

/* [detachable] SPECIAL [[attached] YY_BUFFER] */
typedef struct S1714 T1714;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] YY_BUFFER] */
typedef struct S1715 T1715;

/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TARGET] */
typedef struct S1716 T1716;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TARGET] */
typedef struct S1717 T1717;

/* [detachable] TYPE [detachable STRING_8] */
#define T1718 EIF_TYPE_OBJ

/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ELEMENT_CONTENT] */
typedef struct S1719 T1719;

/* detachable DS_BILINKABLE [[attached] XM_DTD_ELEMENT_CONTENT] */
typedef struct S1720 T1720;

/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]] */
typedef struct S1722 T1722;

/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]] */
typedef struct S1723 T1723;

/* detachable DS_LINKABLE [detachable STRING_8] */
typedef struct S1724 T1724;

/* [detachable] SPECIAL [[attached] ET_CLUSTER] */
typedef struct S1725 T1725;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLUSTER] */
typedef struct S1726 T1726;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLUSTER] */
typedef struct S1727 T1727;

/* [detachable] SPECIAL [[attached] ET_TYPE_ITEM] */
typedef struct S1728 T1728;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_ITEM] */
typedef struct S1729 T1729;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_MASTER_CLASS] */
typedef struct S1730 T1730;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DOTNET_ASSEMBLY] */
typedef struct S1731 T1731;

/* [detachable] TUPLE [[attached] ET_ECF_CLUSTER] */
typedef struct S1732 T1732;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_CLUSTER]] */
typedef struct S1733 T1733;

/* [detachable] TUPLE [[attached] ET_ECF_TARGET, [attached] STRING_8] */
typedef struct S1734 T1734;

/* [detachable] TUPLE [[attached] ET_ECF_ADAPTED_LIBRARY] */
typedef struct S1736 T1736;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_ADAPTED_LIBRARY]] */
typedef struct S1737 T1737;

/* [detachable] TUPLE [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
typedef struct S1738 T1738;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY]] */
typedef struct S1739 T1739;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] STRING_8]] */
typedef struct S1742 T1742;

/* [detachable] TUPLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8] */
typedef struct S1743 T1743;

/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8] */
typedef struct S1744 T1744;

/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8] */
typedef struct S1745 T1745;

/* detachable DS_LINKABLE [[attached] XM_ELEMENT_NODE] */
typedef struct S1750 T1750;

/* detachable ET_CONSTRAINT_GENERIC_NAMED_TYPE */
typedef struct S1755 T1755;

/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION */
typedef struct S1756 T1756;

/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT */
typedef struct S1758 T1758;

/* [detachable] ET_SEEDED_QUERY_COMPARATOR */
typedef struct S1759 T1759;

/* [detachable] ET_SEEDED_PROCEDURE_COMPARATOR */
typedef struct S1762 T1762;

/* [detachable] TUPLE [[attached] ET_CLIENT_LIST] */
typedef struct S1766 T1766;

/* [detachable] ET_FORMAL_COMMA_ARGUMENT */
typedef struct S1767 T1767;

/* [detachable] ET_LOCAL_COMMA_VARIABLE */
typedef struct S1768 T1768;

/* [detachable] ET_LABELED_COMMA_ACTUAL_PARAMETER */
typedef struct S1769 T1769;

/* [detachable] ET_KEYWORD_EXPRESSION */
typedef struct S1771 T1771;

/* [detachable] SPECIAL [[attached] ET_RENAME_ITEM] */
typedef struct S1776 T1776;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME_ITEM] */
typedef struct S1777 T1777;

/* [detachable] SPECIAL [[attached] ET_EXPORT] */
typedef struct S1779 T1779;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPORT] */
typedef struct S1780 T1780;

/* [detachable] SPECIAL [[attached] ET_FEATURE_NAME_ITEM] */
typedef struct S1782 T1782;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME_ITEM] */
typedef struct S1783 T1783;

/* [detachable] SPECIAL [[attached] ET_FORMAL_PARAMETER_ITEM] */
typedef struct S1784 T1784;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_PARAMETER_ITEM] */
typedef struct S1785 T1785;

/* [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_CFLAG] */
typedef struct S1788 T1788;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_CFLAG]] */
typedef struct S1789 T1789;

/* [detachable] TUPLE [[attached] ET_ECF_TARGET, [attached] ET_ECF_STATE] */
typedef struct S1790 T1790;

/* [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_INCLUDE] */
typedef struct S1791 T1791;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_INCLUDE]] */
typedef struct S1792 T1792;

/* [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_LIBRARY] */
typedef struct S1793 T1793;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_LIBRARY]] */
typedef struct S1794 T1794;

/* [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
typedef struct S1795 T1795;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_LINKER_FLAG]] */
typedef struct S1796 T1796;

/* [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_MAKE] */
typedef struct S1797 T1797;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_MAKE]] */
typedef struct S1798 T1798;

/* [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_OBJECT] */
typedef struct S1799 T1799;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_OBJECT]] */
typedef struct S1800 T1800;

/* [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_RESOURCE] */
typedef struct S1801 T1801;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_RESOURCE]] */
typedef struct S1802 T1802;

/* detachable DS_LINKABLE [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]] */
typedef struct S1806 T1806;

/* [detachable] SPECIAL [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM] */
typedef struct S1809 T1809;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM] */
typedef struct S1810 T1810;

/* [detachable] ET_AGENT_IMPLICIT_CURRENT_TARGET */
typedef struct S1811 T1811;

/* [detachable] SPECIAL [[attached] ET_CHOICE_ITEM] */
typedef struct S1813 T1813;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CHOICE_ITEM] */
typedef struct S1814 T1814;

/* [detachable] SPECIAL [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
typedef struct S1815 T1815;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
typedef struct S1817 T1817;

/* [detachable] SPECIAL [[attached] ET_CONVERT_FEATURE_ITEM] */
typedef struct S1819 T1819;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONVERT_FEATURE_ITEM] */
typedef struct S1820 T1820;

/* [detachable] SPECIAL [[attached] ET_CREATOR] */
typedef struct S1821 T1821;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CREATOR] */
typedef struct S1822 T1822;

/* [detachable] SPECIAL [[attached] ET_ELSEIF_EXPRESSION] */
typedef struct S1823 T1823;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_EXPRESSION] */
typedef struct S1824 T1824;

/* [detachable] SPECIAL [[attached] ET_ELSEIF_PART] */
typedef struct S1825 T1825;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_PART] */
typedef struct S1826 T1826;

/* [detachable] SPECIAL [[attached] ET_FEATURE_CLAUSE] */
typedef struct S1827 T1827;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_CLAUSE] */
typedef struct S1828 T1828;

/* [detachable] SPECIAL [[attached] ET_INDEXING_ITEM] */
typedef struct S1830 T1830;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_ITEM] */
typedef struct S1831 T1831;

/* [detachable] SPECIAL [[attached] ET_INDEXING_TERM_ITEM] */
typedef struct S1832 T1832;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM_ITEM] */
typedef struct S1833 T1833;

/* [detachable] SPECIAL [[attached] ET_LOCAL_VARIABLE_ITEM] */
typedef struct S1834 T1834;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LOCAL_VARIABLE_ITEM] */
typedef struct S1836 T1836;

/* [detachable] SPECIAL [[attached] ET_MANIFEST_STRING_ITEM] */
typedef struct S1838 T1838;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MANIFEST_STRING_ITEM] */
typedef struct S1839 T1839;

/* [detachable] SPECIAL [[attached] ET_PARENT_LIST] */
typedef struct S1840 T1840;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_LIST] */
typedef struct S1841 T1841;

/* [detachable] SPECIAL [[attached] ET_TYPE_CONSTRAINT_ITEM] */
typedef struct S1843 T1843;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_CONSTRAINT_ITEM] */
typedef struct S1844 T1844;

/* [detachable] SPECIAL [[attached] ET_WHEN_EXPRESSION] */
typedef struct S1845 T1845;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_WHEN_EXPRESSION] */
typedef struct S1846 T1846;

/* [detachable] SPECIAL [[attached] ET_WHEN_PART] */
typedef struct S1847 T1847;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_WHEN_PART] */
typedef struct S1848 T1848;

/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT */
typedef struct S1850 T1850;

/* detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS] */
typedef struct S1853 T1853;

/* [detachable] ET_STANDARD_ONCE_KEYS */
typedef struct S1857 T1857;

/* [detachable] SPECIAL [[attached] ET_ECF_CLUSTER] */
typedef struct S1859 T1859;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CLUSTER] */
typedef struct S1860 T1860;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_CLUSTER] */
typedef struct S1861 T1861;

/* [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_LIBRARY] */
typedef struct S1862 T1862;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_LIBRARY] */
typedef struct S1863 T1863;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ADAPTED_LIBRARY] */
typedef struct S1864 T1864;

/* [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
typedef struct S1865 T1865;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
typedef struct S1866 T1866;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
typedef struct S1867 T1867;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_FILE_RULE] */
typedef struct S1868 T1868;

/* [detachable] SPECIAL [[attached] ET_ECF_FILE_RULE] */
typedef struct S1869 T1869;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_FILE_RULE] */
typedef struct S1870 T1870;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_CFLAG] */
typedef struct S1871 T1871;

/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_CFLAG] */
typedef struct S1872 T1872;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_CFLAG] */
typedef struct S1873 T1873;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_INCLUDE] */
typedef struct S1874 T1874;

/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_INCLUDE] */
typedef struct S1875 T1875;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_INCLUDE] */
typedef struct S1876 T1876;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LIBRARY] */
typedef struct S1877 T1877;

/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LIBRARY] */
typedef struct S1878 T1878;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_LIBRARY] */
typedef struct S1879 T1879;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
typedef struct S1880 T1880;

/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
typedef struct S1881 T1881;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
typedef struct S1882 T1882;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_MAKE] */
typedef struct S1883 T1883;

/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_MAKE] */
typedef struct S1884 T1884;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_MAKE] */
typedef struct S1885 T1885;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_OBJECT] */
typedef struct S1886 T1886;

/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_OBJECT] */
typedef struct S1887 T1887;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_OBJECT] */
typedef struct S1888 T1888;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_RESOURCE] */
typedef struct S1889 T1889;

/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_RESOURCE] */
typedef struct S1890 T1890;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_RESOURCE] */
typedef struct S1891 T1891;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] RX_PCRE_REGULAR_EXPRESSION] */
typedef struct S1892 T1892;

/* [detachable] SPECIAL [[attached] RX_PCRE_REGULAR_EXPRESSION] */
typedef struct S1893 T1893;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] RX_PCRE_REGULAR_EXPRESSION] */
typedef struct S1894 T1894;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ANDED_CONDITIONS] */
typedef struct S1895 T1895;

/* [detachable] SPECIAL [[attached] ET_ECF_ANDED_CONDITIONS] */
typedef struct S1896 T1896;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ANDED_CONDITIONS] */
typedef struct S1897 T1897;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CONDITION_ITEM] */
typedef struct S1898 T1898;

/* [detachable] SPECIAL [[attached] ET_ECF_CONDITION_ITEM] */
typedef struct S1899 T1899;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_CONDITION_ITEM] */
typedef struct S1900 T1900;

/* detachable DS_LINKABLE [[attached] CHARACTER_8] */
typedef struct S1902 T1902;

/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_CLASS] */
typedef struct S1903 T1903;

/* [detachable] SPECIAL [[attached] ET_BASE_TYPE_CONSTRAINT] */
typedef struct S1904 T1904;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE_CONSTRAINT] */
typedef struct S1905 T1905;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_BASE_TYPE_CONSTRAINT] */
typedef struct S1906 T1906;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] NATURAL_32] */
typedef struct S1907 T1907;

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_FEATURE_NAME] */
typedef struct S1908 T1908;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] ET_FEATURE_NAME] */
typedef struct S1909 T1909;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32] */
typedef struct S1911 T1911;

/* [detachable] SPECIAL [[attached] ET_REPLICABLE_FEATURE] */
typedef struct S1912 T1912;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_REPLICABLE_FEATURE] */
typedef struct S1914 T1914;

/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_ADAPTED_FEATURE] */
typedef struct S1915 T1915;

/* detachable DS_LINKABLE [[attached] ET_ADAPTED_FEATURE] */
typedef struct S1916 T1916;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_FEATURE] */
typedef struct S1917 T1917;

/* [detachable] SPECIAL [[attached] ET_PARENT_FEATURE] */
typedef struct S1918 T1918;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE] */
typedef struct S1919 T1919;

/* [detachable] SPECIAL [[attached] ET_DOTNET_FEATURE] */
typedef struct S1920 T1920;

/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DOTNET_FEATURE] */
typedef struct S1921 T1921;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_FEATURE] */
typedef struct S1922 T1922;

/* [detachable] SPECIAL [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]] */
typedef struct S1923 T1923;

/* detachable DS_HASH_TABLE_CURSOR [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE] */
typedef struct S1926 T1926;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]] */
typedef struct S1927 T1927;

/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_DOTNET_FEATURE] */
typedef struct S1928 T1928;

/* detachable DS_LINKABLE [[attached] ET_DOTNET_FEATURE] */
typedef struct S1929 T1929;

/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE] */
typedef struct S1930 T1930;

/* [detachable] SPECIAL [[attached] ET_LIKE_FEATURE] */
typedef struct S1931 T1931;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIKE_FEATURE] */
typedef struct S1932 T1932;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE] */
typedef struct S1933 T1933;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8] */
typedef struct S1935 T1935;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]] */
typedef struct S1936 T1936;

/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME] */
typedef struct S1937 T1937;

/* [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]] */
typedef struct S1938 T1938;

/* detachable DS_BILINKABLE [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]] */
typedef struct S1939 T1939;

/* [detachable] ET_CLUSTER_ERROR */
typedef struct S1940 T1940;

/* [detachable] ET_SYNTAX_ERROR */
typedef struct S1941 T1941;

/* [detachable] ET_DOTNET_ASSEMBLY_ERROR */
typedef struct S1942 T1942;

/* [detachable] AP_OPTION_COMPARATOR */
typedef struct S1943 T1943;

/* [detachable] DS_QUICK_SORTER [[attached] AP_OPTION] */
typedef struct S1944 T1944;

/* [detachable] ST_SPLITTER */
typedef struct S1953 T1953;

/* [detachable] KL_CHARACTER_ROUTINES */
typedef struct S1954 T1954;

/* [detachable] TUPLE [[attached] ET_C_GENERATOR, INTEGER_32, [attached] ET_CURRENT, [attached] ET_DYNAMIC_TUPLE_TYPE, BOOLEAN] */
typedef struct S1955 T1955;

/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS */
typedef struct S1956 T1956;

/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST */
typedef struct S1957 T1957;

/* [detachable] TYPE [[detachable] DEVELOPER_EXCEPTION] */
#define T1958 EIF_TYPE_OBJ

/* [detachable] SPECIAL [[attached] ET_FORMAL_ARGUMENT_ITEM] */
typedef struct S1960 T1960;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_ARGUMENT_ITEM] */
typedef struct S1962 T1962;

/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRECURSOR] */
typedef struct S1964 T1964;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRECURSOR] */
typedef struct S1965 T1965;

/* [detachable] SPECIAL [[attached] ET_INSTRUCTION] */
typedef struct S1966 T1966;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INSTRUCTION] */
typedef struct S1968 T1968;

/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST */
typedef struct S1977 T1977;

/* [detachable] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8] */
typedef struct S1978 T1978;

/* [detachable] ET_INFIX_CAST_EXPRESSION */
typedef struct S1980 T1980;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] STRING_8] */
typedef struct S1984 T1984;

/* [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]] */
typedef struct S1985 T1985;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]] */
typedef struct S1988 T1988;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8] */
typedef struct S1989 T1989;

/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS] */
typedef struct S1990 T1990;

/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE] */
typedef struct S1993 T1993;

/* [detachable] SPECIAL [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]] */
typedef struct S1994 T1994;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]] */
typedef struct S1997 T1997;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME] */
typedef struct S1998 T1998;

/* [detachable] ET_UNIVERSE_ERROR */
typedef struct S1999 T1999;

/* [detachable] TUPLE [[attached] ET_SYSTEM, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
typedef struct S2000 T2000;

/* detachable ET_PARENTHESIS_INSTRUCTION */
typedef struct S2002 T2002;

/* [detachable] TYPE [detachable RX_REGULAR_EXPRESSION] */
#define T2006 EIF_TYPE_OBJ

/* [detachable] ET_PARENTHESIS_SYMBOL */
typedef struct S2020 T2020;

/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT_LIST */
typedef struct S2026 T2026;

/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT */
typedef struct S2027 T2027;

/* [detachable] TYPE [[attached] ET_NESTED_TYPE_CONTEXT] */
#define T2030 EIF_TYPE_OBJ

/* [detachable] TYPE [[attached] ET_ITERATION_COMPONENT] */
#define T2031 EIF_TYPE_OBJ

/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
typedef struct S2034 T2034;

/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8] */
typedef struct S2040 T2040;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8] */
typedef struct S2041 T2041;

/* [detachable] ET_DYNAMIC_EQUALITY_EXPRESSION */
typedef struct S2042 T2042;

/* [detachable] ET_DYNAMIC_OBJECT_EQUALITY_EXPRESSION */
typedef struct S2043 T2043;

/* [detachable] TUPLE [[attached] ET_ECF_LIBRARY] */
typedef struct S2044 T2044;

/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION */
typedef struct S2046 T2046;

/* [detachable] ET_CONVERT_FROM_EXPRESSION */
typedef struct S2047 T2047;

/* [detachable] ET_CONVERT_TO_EXPRESSION */
typedef struct S2048 T2048;

/* detachable DS_CELL [detachable ET_CLASS] */
typedef struct S2051 T2051;

/* [detachable] TUPLE [[attached] DS_CELL [detachable ET_CLASS]] */
typedef struct S2052 T2052;

/* [detachable] TUPLE [[attached] ET_EIFFEL_PREPARSER, [attached] DS_ARRAYED_LIST [[attached] STRING_8]] */
typedef struct S2054 T2054;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_FEATURE] */
typedef struct S2056 T2056;

/* [detachable] TUPLE [[attached] ET_ECF_FILE_RULE] */
typedef struct S2057 T2057;

/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_FILE_RULE]] */
typedef struct S2058 T2058;

/* [detachable] TUPLE [[attached] ET_ECF_FILE_RULES] */
typedef struct S2059 T2059;

/* [detachable] PREDICATE [[attached] TUPLE [[attached] ET_ECF_FILE_RULE]] */
typedef struct S2060 T2060;

/* [detachable] TUPLE [[attached] ET_ECF_STATE] */
typedef struct S2061 T2061;

/* [detachable] PREDICATE [[attached] TUPLE [[attached] STRING_8]] */
typedef struct S2063 T2063;

/* [detachable] TUPLE [[attached] KL_STRING_ROUTINES, [attached] STRING_8] */
typedef struct S2064 T2064;

/* [detachable] TUPLE [[attached] DS_HASH_SET [[attached] STRING_8]] */
typedef struct S2066 T2066;

/* [detachable] LX_DFA_WILDCARD */
typedef struct S2068 T2068;

/* detachable DS_LINKABLE [[attached] ET_PARENT_FEATURE] */
typedef struct S2071 T2071;

/* [detachable] TUPLE [[attached] LX_WILDCARD] */
typedef struct S2073 T2073;

/* [detachable] PREDICATE [[attached] TUPLE [[attached] LX_WILDCARD]] */
typedef struct S2074 T2074;

/* [detachable] TUPLE [[attached] ET_CLUSTER_DEPENDENCE_CONSTRAINT, [attached] ET_GROUP, [attached] STRING_8] */
typedef struct S2075 T2075;

/* [detachable] SPECIAL [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT] */
typedef struct S2077 T2077;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT] */
typedef struct S2078 T2078;

/* [detachable] SPECIAL [[attached] LX_WILDCARD] */
typedef struct S2079 T2079;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_WILDCARD] */
typedef struct S2082 T2082;

/* [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
typedef struct S2086 T2086;

/* [detachable] TUPLE [[attached] ET_ECF_LIBRARY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
typedef struct S2089 T2089;

/* [detachable] TUPLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
typedef struct S2090 T2090;

/* [detachable] TUPLE [[attached] ET_DOTNET_ASSEMBLY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
typedef struct S2091 T2091;

/* [detachable] TUPLE [[attached] ET_LIBRARY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
typedef struct S2092 T2092;

/* [detachable] ET_AST_NULL_LEAF */
typedef struct S2094 T2094;

/* [detachable] LX_WILDCARD_PARSER */
typedef struct S2095 T2095;

/* [detachable] LX_DESCRIPTION */
typedef struct S2096 T2096;

/* [detachable] LX_FULL_DFA */
typedef struct S2097 T2097;

/* [detachable] TUPLE [[attached] ET_GROUP] */
typedef struct S2099 T2099;

/* [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]], [attached] PREDICATE [[attached] TUPLE [[attached] ET_CLASS]]] */
typedef struct S2100 T2100;

/* [detachable] TUPLE [[attached] RX_PCRE_REGULAR_EXPRESSION] */
typedef struct S2101 T2101;

/* [detachable] PREDICATE [[attached] TUPLE [[attached] RX_PCRE_REGULAR_EXPRESSION]] */
typedef struct S2102 T2102;

/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE] */
typedef struct S2105 T2105;

/* [detachable] LX_START_CONDITIONS */
typedef struct S2106 T2106;

/* [detachable] LX_ACTION_FACTORY */
typedef struct S2107 T2107;

/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS] */
typedef struct S2108 T2108;

/* [detachable] LX_SYMBOL_CLASS */
typedef struct S2109 T2109;

/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8] */
typedef struct S2110 T2110;

/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS] */
typedef struct S2111 T2111;

/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32] */
typedef struct S2112 T2112;

/* [detachable] KL_EQUALITY_TESTER [[attached] STRING_32] */
typedef struct S2113 T2113;

/* [detachable] DS_ARRAYED_STACK [[attached] BOOLEAN] */
typedef struct S2114 T2114;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_32] */
typedef struct S2115 T2115;

/* [detachable] SPECIAL [[attached] STRING_32] */
typedef struct S2116 T2116;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_SYMBOL_CLASS] */
typedef struct S2117 T2117;

/* [detachable] SPECIAL [[attached] LX_SYMBOL_CLASS] */
typedef struct S2118 T2118;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA] */
typedef struct S2119 T2119;

/* [detachable] LX_NFA */
typedef struct S2120 T2120;

/* [detachable] SPECIAL [[attached] LX_NFA] */
typedef struct S2121 T2121;

/* detachable LX_RULE */
typedef struct S2122 T2122;

/* [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS] */
typedef struct S2123 T2123;

/* [detachable] LX_EQUIVALENCE_CLASSES */
typedef struct S2124 T2124;

/* [detachable] LX_UNRECOGNIZED_RULE_ERROR */
typedef struct S2125 T2125;

/* [detachable] LX_INVALID_UNICODE_CHARACTER_ERROR */
typedef struct S2126 T2126;

/* [detachable] LX_MISSING_QUOTE_ERROR */
typedef struct S2127 T2127;

/* [detachable] LX_BAD_CHARACTER_CLASS_ERROR */
typedef struct S2128 T2128;

/* [detachable] LX_BAD_CHARACTER_ERROR */
typedef struct S2129 T2129;

/* [detachable] LX_FULL_AND_META_ERROR */
typedef struct S2130 T2130;

/* [detachable] LX_FULL_AND_REJECT_ERROR */
typedef struct S2131 T2131;

/* [detachable] LX_FULL_AND_VARIABLE_TRAILING_CONTEXT_ERROR */
typedef struct S2132 T2132;

/* [detachable] LX_CHARACTER_OUT_OF_RANGE_ERROR */
typedef struct S2133 T2133;

/* [detachable] ARRAY [[attached] LX_RULE] */
typedef struct S2134 T2134;

/* [detachable] SPECIAL [[attached] LX_RULE] */
typedef struct S2135 T2135;

/* [detachable] LX_DFA_STATE */
typedef struct S2136 T2136;

/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE] */
typedef struct S2137 T2137;

/* [detachable] DS_ARRAYED_LIST [[attached] LX_DFA_STATE] */
typedef struct S2138 T2138;

/* detachable LX_SYMBOL_PARTITIONS */
typedef struct S2139 T2139;

/* [detachable] KL_ARRAY_ROUTINES [[attached] LX_RULE] */
typedef struct S2140 T2140;

/* [detachable] ARRAY [detachable LX_RULE] */
typedef struct S2141 T2141;

/* [detachable] SPECIAL [detachable LX_RULE] */
typedef struct S2142 T2142;

/* [detachable] LX_START_CONDITION */
typedef struct S2143 T2143;

/* [detachable] LX_TRANSITION_TABLE [[attached] LX_DFA_STATE] */
typedef struct S2144 T2144;

/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA] */
typedef struct S2145 T2145;

/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32] */
typedef struct S2146 T2146;

/* [detachable] LX_NFA_STATE */
typedef struct S2147 T2147;

/* detachable ARRAY [detachable LX_SYMBOL_CLASS] */
typedef struct S2148 T2148;

/* [detachable] SPECIAL [detachable LX_SYMBOL_CLASS] */
typedef struct S2149 T2149;

/* detachable ARRAY [detachable DS_HASH_SET [[attached] INTEGER_32]] */
typedef struct S2150 T2150;

/* [detachable] SPECIAL [detachable DS_HASH_SET [[attached] INTEGER_32]] */
typedef struct S2151 T2151;

/* detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
typedef struct S2152 T2152;

/* [detachable] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS] */
typedef struct S2153 T2153;

/* [detachable] LX_NEGATIVE_RANGE_IN_CHARACTER_CLASS_ERROR */
typedef struct S2154 T2154;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_RULE] */
typedef struct S2155 T2155;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_RULE] */
typedef struct S2156 T2156;

/* [detachable] KL_ARRAY_ROUTINES [[attached] STRING_8] */
typedef struct S2157 T2157;

/* [detachable] SPECIAL [[attached] LX_START_CONDITION] */
typedef struct S2158 T2158;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_START_CONDITION] */
typedef struct S2159 T2159;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_START_CONDITION] */
typedef struct S2160 T2160;

/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]] */
typedef struct S2161 T2161;

/* detachable KL_EQUALITY_TESTER [[attached] LX_SYMBOL_CLASS] */
typedef struct S2162 T2162;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]] */
typedef struct S2164 T2164;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS] */
typedef struct S2165 T2165;

/* detachable SPECIAL [detachable SPECIAL [[attached] NATURAL_64]] */
typedef struct S2166 T2166;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_32, [attached] STRING_8] */
typedef struct S2168 T2168;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_SYMBOL_CLASS, [attached] STRING_32] */
typedef struct S2170 T2170;

/* [detachable] LX_EPSILON_TRANSITION [[attached] LX_NFA_STATE] */
typedef struct S2171 T2171;

/* [detachable] LX_SYMBOL_CLASS_TRANSITION [[attached] LX_NFA_STATE] */
typedef struct S2174 T2174;

/* [detachable] LX_SYMBOL_TRANSITION [[attached] LX_NFA_STATE] */
typedef struct S2175 T2175;

/* [detachable] LX_ACTION */
typedef struct S2177 T2177;

/* [detachable] ARRAY [[attached] DS_BILINKABLE [[attached] INTEGER_32]] */
typedef struct S2178 T2178;

/* [detachable] DS_BILINKABLE [[attached] INTEGER_32] */
typedef struct S2179 T2179;

/* [detachable] SPECIAL [[attached] DS_BILINKABLE [[attached] INTEGER_32]] */
typedef struct S2180 T2180;

/* [detachable] DS_BUBBLE_SORTER [[attached] LX_NFA_STATE] */
typedef struct S2181 T2181;

/* [detachable] DS_BUBBLE_SORTER [[attached] LX_RULE] */
typedef struct S2183 T2183;

/* [detachable] SPECIAL [[attached] LX_NFA_STATE] */
typedef struct S2185 T2185;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA_STATE] */
typedef struct S2187 T2187;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE] */
typedef struct S2188 T2188;

/* [detachable] SPECIAL [[attached] LX_DFA_STATE] */
typedef struct S2190 T2190;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_DFA_STATE] */
typedef struct S2191 T2191;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_DFA_STATE] */
typedef struct S2192 T2192;

/* [detachable] KL_ARRAY [[attached] LX_RULE] */
typedef struct S2193 T2193;

/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32] */
typedef struct S2194 T2194;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA] */
typedef struct S2195 T2195;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_NFA, [attached] INTEGER_32] */
typedef struct S2197 T2197;

/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_RULE] */
typedef struct S2201 T2201;

/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_NFA_STATE] */
typedef struct S2204 T2204;

/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
typedef struct S2207 T2207;

/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
typedef struct S2208 T2208;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
typedef struct S2209 T2209;

/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_SYMBOL_CLASS] */
typedef struct S2210 T2210;

/* [detachable] KL_ARRAY [[attached] STRING_8] */
typedef struct S2211 T2211;

/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_DFA_STATE, [attached] INTEGER_32] */
typedef struct S2213 T2213;

/* Struct for boxed version of type BOOLEAN */
struct Sb1 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1* p1; /* pointer */
	T1 z1; /* item */
};

/* Struct for boxed version of type CHARACTER_8 */
struct Sb2 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2* p1; /* pointer */
	T2 z1; /* item */
};

/* Struct for boxed version of type CHARACTER_32 */
struct Sb3 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T3* p1; /* pointer */
	T3 z1; /* item */
};

/* Struct for boxed version of type INTEGER_8 */
struct Sb4 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T4* p1; /* pointer */
	T4 z1; /* item */
};

/* Struct for boxed version of type INTEGER_16 */
struct Sb5 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T5* p1; /* pointer */
	T5 z1; /* item */
};

/* Struct for boxed version of type INTEGER_32 */
struct Sb6 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6* p1; /* pointer */
	T6 z1; /* item */
};

/* Struct for boxed version of type INTEGER_64 */
struct Sb7 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T7* p1; /* pointer */
	T7 z1; /* item */
};

/* Struct for boxed version of type NATURAL_8 */
struct Sb8 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T8* p1; /* pointer */
	T8 z1; /* item */
};

/* Struct for boxed version of type NATURAL_16 */
struct Sb9 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T9* p1; /* pointer */
	T9 z1; /* item */
};

/* Struct for boxed version of type NATURAL_32 */
struct Sb10 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T10* p1; /* pointer */
	T10 z1; /* item */
};

/* Struct for boxed version of type NATURAL_64 */
struct Sb11 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T11* p1; /* pointer */
	T11 z1; /* item */
};

/* Struct for boxed version of type REAL_32 */
struct Sb12 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T12* p1; /* pointer */
	T12 z1; /* item */
};

/* Struct for boxed version of type REAL_64 */
struct Sb13 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T13* p1; /* pointer */
	T13 z1; /* item */
};

/* Struct for boxed version of type POINTER */
struct Sb14 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14* p1; /* pointer */
	T14 z1; /* item */
};

/* Struct for type UTF_CONVERTER */
struct S478 {
	char dummy;
};

/* Struct for boxed version of type UTF_CONVERTER */
struct Sb478 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T478* p1; /* pointer */
	T478 z1; /* item */
};

/* Struct for type TYPED_POINTER [[attached] NATURAL_16] */
struct S1468 {
	T14 a1; /* to_pointer */
};

/* Struct for boxed version of type TYPED_POINTER [[attached] NATURAL_16] */
struct Sb1468 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1468* p1; /* pointer */
	T1468 z1; /* item */
};

/* Struct for type TYPED_POINTER [[attached] NATURAL_8] */
struct S893 {
	T14 a1; /* to_pointer */
};

/* Struct for boxed version of type TYPED_POINTER [[attached] NATURAL_8] */
struct Sb893 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T893* p1; /* pointer */
	T893 z1; /* item */
};

/* Struct for type TYPED_POINTER [[attached] ANY] */
struct S697 {
	T14 a1; /* to_pointer */
};

/* Struct for boxed version of type TYPED_POINTER [[attached] ANY] */
struct Sb697 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T697* p1; /* pointer */
	T697 z1; /* item */
};

/* Struct for type TYPED_POINTER [[attached] SPECIAL [[attached] NATURAL_8]] */
struct S488 {
	T14 a1; /* to_pointer */
};

/* Struct for boxed version of type TYPED_POINTER [[attached] SPECIAL [[attached] NATURAL_8]] */
struct Sb488 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T488* p1; /* pointer */
	T488 z1; /* item */
};

/* Struct for type [detachable] SPECIAL [CHARACTER_8] */
struct S15 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T2 z2[1]; /* item */
};

/* Struct for type [detachable] SPECIAL [CHARACTER_32] */
struct S16 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T3 z2[1]; /* item */
};

/* Struct for type [detachable] STRING_8 */
struct S17 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* count */
	T6 a3; /* internal_hash_code */
	T6 a4; /* internal_case_insensitive_hash_code */
};

/* Struct for type [detachable] STRING_32 */
struct S18 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* count */
	T6 a3; /* internal_hash_code */
	T6 a4; /* internal_case_insensitive_hash_code */
};

/* Struct for type [detachable] IMMUTABLE_STRING_32 */
struct S20 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* count */
	T6 a3; /* internal_hash_code */
	T6 a4; /* area_lower */
};

/* Struct for type [detachable] ISE_EXCEPTION_MANAGER */
struct S21 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] GEC */
struct S25 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* exit_code */
	T0* a2; /* error_handler */
	T0* a3; /* version_flag */
	T0* a4; /* ecf_filename */
	T0* a5; /* last_system */
	T0* a6; /* target_option */
	T0* a7; /* setting_option */
	T0* a8; /* capability_option */
	T0* a9; /* variable_option */
	T0* a10; /* finalize_flag */
	T0* a11; /* gelint_flag */
	T0* a12; /* ise_option */
	T0* a13; /* catcall_option */
	T0* a14; /* c_compile_option */
	T0* a15; /* split_option */
	T0* a16; /* split_size_option */
	T0* a17; /* new_instance_types_option */
	T0* a18; /* gc_option */
	T0* a19; /* thread_option */
	T0* a20; /* silent_flag */
	T0* a21; /* no_benchmark_flag */
	T0* a22; /* nested_benchmark_flag */
	T0* a23; /* metrics_flag */
	T0* a24; /* verbose_flag */
	T0* a25; /* ise_version */
	T0* a26; /* override_settings */
	T0* a27; /* override_capabilities */
	T0* a28; /* override_variables */
	T1 a29; /* ecf_gelint_option */
	T0* a30; /* new_instance_types */
	T6 a31; /* split_size */
};

/* Struct for type [detachable] CELL [detachable EXCEPTION] */
struct S26 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
};

/* Struct for type [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32] */
struct S27 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* capacity */
	T0* a2; /* content */
	T0* a3; /* keys */
	T0* a4; /* deleted_marks */
	T0* a5; /* indexes_map */
	T6 a6; /* iteration_position */
	T6 a7; /* count */
	T6 a8; /* deleted_item_position */
	T6 a9; /* control */
	T6 a10; /* found_item */
	T1 a11; /* has_default */
	T6 a12; /* item_position */
	T6 a13; /* ht_lowest_deleted_position */
	T6 a14; /* ht_deleted_item */
	T6 a15; /* ht_deleted_key */
	T1 a16; /* object_comparison */
};

/* Struct for type [detachable] CELL [detachable TUPLE [[attached] INTEGER_32, [attached] INTEGER_32, [attached] INTEGER_32, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] INTEGER_32, [attached] BOOLEAN]] */
struct S28 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
};

/* Struct for type [detachable] CELL [[attached] NO_MORE_MEMORY] */
struct S29 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
};

/* Struct for type detachable C_STRING */
struct S30 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* managed_data */
	T6 a2; /* count */
};

/* Struct for type [detachable] TUPLE [[attached] INTEGER_32, [attached] INTEGER_32, [attached] INTEGER_32, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] INTEGER_32, [attached] BOOLEAN] */
struct S31 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 z1;
	T6 z2;
	T6 z3;
	T0* z4;
	T0* z5;
	T0* z6;
	T0* z7;
	T0* z8;
	T0* z9;
	T6 z10;
	T1 z11;
};

/* Struct for type [detachable] KL_ARGUMENTS */
struct S32 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* program_name */
};

/* Struct for type [detachable] ARRAY [[attached] STRING_8] */
struct S33 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] SPECIAL [[attached] STRING_8] */
struct S34 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_EXCEPTIONS */
struct S35 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_ERROR_HANDLER */
struct S36 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_error */
	T0* a2; /* info_file */
	T0* a3; /* mutex */
	T0* a4; /* error_file */
	T0* a5; /* warning_file */
	T1 a6; /* is_ise */
	T1 a7; /* is_verbose */
	T1 a8; /* has_eiffel_error */
	T1 a9; /* has_internal_error */
	T1 a10; /* is_ge */
};

/* Struct for type [detachable] KL_TEXT_INPUT_FILE */
struct S37 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* end_of_file */
	T0* a2; /* last_string */
	T6 a3; /* mode */
	T0* a4; /* name */
	T0* a5; /* character_buffer */
	T2 a6; /* last_character */
	T14 a7; /* file_pointer */
	T1 a8; /* descriptor_available */
	T0* a9; /* last_string_32 */
	T0* a10; /* internal_name */
	T0* a11; /* internal_detachable_name_pointer */
};

/* Struct for type [detachable] ET_ISE_VARIABLES */
struct S38 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] AP_FLAG */
struct S39 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* occurrences */
	T0* a2; /* description */
	T0* a3; /* long_form */
	T2 a4; /* short_form */
	T1 a5; /* has_short_form */
	T1 a6; /* is_hidden */
	T1 a7; /* is_mandatory */
	T6 a8; /* maximum_occurrences */
};

/* Struct for type detachable ET_SYSTEM */
struct S40 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* implicit_attachment_type_mark */
	T0* a2; /* boolean_type */
	T0* a3; /* character_type */
	T0* a4; /* character_32_type */
	T0* a5; /* integer_type */
	T0* a6; /* current_system */
	T0* a7; /* tuple_type */
	T0* a8; /* pointer_type */
	T0* a9; /* system_object_parents */
	T0* a10; /* any_parents */
	T0* a11; /* system_object_type */
	T0* a12; /* any_type */
	T0* a13; /* register_class_mutex */
	T0* a14; /* external_include_pathnames */
	T0* a15; /* external_object_pathnames */
	T0* a16; /* external_library_pathnames */
	T0* a17; /* external_make_pathnames */
	T0* a18; /* external_resource_pathnames */
	T0* a19; /* external_cflags */
	T0* a20; /* external_linker_flags */
	T1 a21; /* console_application_mode */
	T1 a22; /* total_order_on_reals_mode */
	T1 a23; /* attachment_type_conformance_mode */
	T1 a24; /* scoop_mode */
	T6 a25; /* default_create_seed */
	T0* a26; /* universe */
	T0* a27; /* name */
	T0* a28; /* libraries */
	T0* a29; /* clusters */
	T0* a30; /* dotnet_assemblies */
	T0* a31; /* master_classes */
	T0* a32; /* master_class_mutex */
	T0* a33; /* system_name */
	T0* a34; /* detachable_any_type */
	T0* a35; /* detachable_separate_any_type */
	T0* a36; /* any_parent */
	T0* a37; /* any_clients */
	T0* a38; /* detachable_tuple_type */
	T0* a39; /* tuple_identity_type */
	T0* a40; /* unfolded_empty_tuple_actual_parameters */
	T0* a41; /* array_any_type */
	T0* a42; /* array_detachable_any_type */
	T0* a43; /* array_none_type */
	T0* a44; /* array_identity_type */
	T0* a45; /* character_8_type */
	T0* a46; /* character_8_convert_feature */
	T0* a47; /* character_32_convert_feature */
	T0* a48; /* double_type */
	T0* a49; /* exception_type */
	T0* a50; /* detachable_exception_type */
	T0* a51; /* exception_manager_type */
	T0* a52; /* function_identity_any_type */
	T0* a53; /* immutable_string_8_type */
	T0* a54; /* immutable_string_32_type */
	T0* a55; /* integer_8_type */
	T0* a56; /* integer_8_convert_feature */
	T0* a57; /* integer_16_type */
	T0* a58; /* integer_16_convert_feature */
	T0* a59; /* integer_32_type */
	T0* a60; /* integer_32_convert_feature */
	T0* a61; /* integer_64_type */
	T0* a62; /* integer_64_convert_feature */
	T0* a63; /* ise_exception_manager_type */
	T0* a64; /* iterable_detachable_separate_any_type */
	T0* a65; /* natural_type */
	T0* a66; /* natural_8_type */
	T0* a67; /* natural_8_convert_feature */
	T0* a68; /* natural_16_type */
	T0* a69; /* natural_16_convert_feature */
	T0* a70; /* natural_32_type */
	T0* a71; /* natural_32_convert_feature */
	T0* a72; /* natural_64_type */
	T0* a73; /* natural_64_convert_feature */
	T0* a74; /* none_type */
	T0* a75; /* detachable_none_type */
	T0* a76; /* predicate_identity_type */
	T0* a77; /* procedure_identity_type */
	T0* a78; /* real_type */
	T0* a79; /* real_32_type */
	T0* a80; /* real_32_convert_feature */
	T0* a81; /* real_64_type */
	T0* a82; /* real_64_convert_feature */
	T0* a83; /* routine_identity_type */
	T0* a84; /* special_any_type */
	T0* a85; /* special_detachable_any_type */
	T0* a86; /* special_identity_type */
	T0* a87; /* string_type */
	T0* a88; /* detachable_string_type */
	T0* a89; /* string_8_type */
	T0* a90; /* detachable_string_8_type */
	T0* a91; /* string_8_convert_feature */
	T0* a92; /* string_32_type */
	T0* a93; /* string_32_convert_feature */
	T0* a94; /* system_string_type */
	T0* a95; /* type_detachable_any_type */
	T0* a96; /* detachable_type_exception_type */
	T0* a97; /* type_detachable_exception_type */
	T0* a98; /* type_detachable_like_current_type */
	T0* a99; /* type_identity_type */
	T0* a100; /* typed_pointer_identity_type */
	T0* a101; /* wide_character_type */
	T1 a102; /* is_read_only */
	T6 a103; /* registered_class_count */
	T1 a104; /* use_obsolete_syntax_mode */
	T1 a105; /* obsolete_iteration_mode */
	T6 a106; /* copy_seed */
	T6 a107; /* twin_seed */
	T6 a108; /* is_equal_seed */
	T6 a109; /* routine_call_seed */
	T6 a110; /* function_item_seed */
	T6 a111; /* dispose_seed */
	T6 a112; /* iterable_new_cursor_seed */
	T6 a113; /* iteration_cursor_item_seed */
	T6 a114; /* iteration_cursor_after_seed */
	T6 a115; /* iteration_cursor_forth_seed */
	T6 a116; /* special_item_seed */
	T1 a117; /* target_type_attachment_mode */
	T1 a118; /* obsolete_routine_type_mode */
	T1 a119; /* is_dotnet */
};

/* Struct for type [detachable] AP_PARSER */
struct S42 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* options */
	T0* a2; /* alternative_options_lists */
	T0* a3; /* parameters */
	T0* a4; /* parameters_description */
	T0* a5; /* help_option */
	T0* a6; /* error_handler */
	T0* a7; /* application_description */
	T0* a8; /* current_options */
	T1 a9; /* is_first_option */
	T0* a10; /* last_option_parameter */
};

/* Struct for type [detachable] AP_ALTERNATIVE_OPTIONS_LIST */
struct S43 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* introduction_option */
	T0* a2; /* parameters_description */
	T0* a3; /* internal_cursor */
	T0* a4; /* first_cell */
};

/* Struct for type [detachable] AP_STRING_OPTION */
struct S44 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
	T0* a2; /* default_parameter */
	T0* a3; /* description */
	T0* a4; /* parameter_description */
	T1 a5; /* has_default_parameter */
	T1 a6; /* needs_parameter */
	T0* a7; /* long_form */
	T1 a8; /* is_hidden */
	T1 a9; /* is_mandatory */
	T6 a10; /* maximum_occurrences */
	T1 a11; /* has_short_form */
	T2 a12; /* short_form */
};

/* Struct for type [detachable] UT_VERSION */
struct S46 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* internal_major */
	T6 a2; /* internal_minor */
	T6 a3; /* internal_revision */
	T6 a4; /* internal_build */
};

/* Struct for type [detachable] AP_ENUMERATION_OPTION */
struct S47 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
	T0* a2; /* default_parameter */
	T0* a3; /* description */
	T0* a4; /* possible_values */
	T0* a5; /* parameter_description */
	T1 a6; /* needs_parameter */
	T0* a7; /* long_form */
	T1 a8; /* is_hidden */
	T1 a9; /* is_mandatory */
	T6 a10; /* maximum_occurrences */
	T1 a11; /* has_short_form */
	T2 a12; /* short_form */
};

/* Struct for type [detachable] AP_BOOLEAN_OPTION */
struct S48 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
	T1 a2; /* default_parameter */
	T0* a3; /* description */
	T0* a4; /* parameter_description */
	T1 a5; /* needs_parameter */
	T0* a6; /* long_form */
	T1 a7; /* is_hidden */
	T1 a8; /* is_mandatory */
	T6 a9; /* maximum_occurrences */
	T1 a10; /* has_short_form */
	T2 a11; /* short_form */
};

/* Struct for type [detachable] AP_INTEGER_OPTION */
struct S49 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
	T6 a2; /* default_parameter */
	T0* a3; /* description */
	T0* a4; /* parameter_description */
	T1 a5; /* needs_parameter */
	T0* a6; /* long_form */
	T2 a7; /* short_form */
	T1 a8; /* is_hidden */
	T1 a9; /* is_mandatory */
	T6 a10; /* maximum_occurrences */
	T1 a11; /* has_short_form */
};

/* Struct for type [detachable] ET_NULL_ERROR_HANDLER */
struct S53 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_error */
	T0* a2; /* info_file */
	T0* a3; /* mutex */
	T0* a4; /* error_file */
	T0* a5; /* warning_file */
	T1 a6; /* is_ise */
	T1 a7; /* is_verbose */
	T1 a8; /* has_eiffel_error */
	T1 a9; /* has_internal_error */
	T1 a10; /* is_ge */
};

/* Struct for type [detachable] ET_ECF_SYSTEM_PARSER */
struct S55 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* last_system */
	T0* a2; /* ast_factory */
	T0* a3; /* override_settings */
	T0* a4; /* override_capabilities */
	T0* a5; /* override_variables */
	T0* a6; /* parsed_libraries */
	T0* a7; /* parsed_dotnet_assemblies */
	T0* a8; /* redirected_locations */
	T0* a9; /* xml_parser */
	T0* a10; /* tree_pipe */
	T0* a11; /* error_handler */
	T0* a12; /* ise_version */
};

/* Struct for type [detachable] ET_ECF_ERROR_HANDLER */
struct S56 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* error_count */
	T0* a2; /* error_file */
	T0* a3; /* warning_file */
	T0* a4; /* info_file */
};

/* Struct for type detachable ET_ECF_SETTINGS */
struct S57 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* primary_settings */
	T0* a2; /* secondary_settings */
};

/* Struct for type [detachable] ET_ECF_SYSTEM */
struct S58 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* filename */
	T0* a2; /* targets */
	T0* a3; /* library_target */
	T0* a4; /* libraries */
	T0* a5; /* selected_target */
	T0* a6; /* dotnet_assemblies */
	T0* a7; /* master_classes */
	T0* a8; /* root_type */
	T0* a9; /* none_type */
	T0* a10; /* any_type */
	T0* a11; /* root_creation */
	T6 a12; /* default_create_seed */
	T0* a13; /* boolean_type */
	T0* a14; /* character_8_type */
	T0* a15; /* character_32_type */
	T0* a16; /* integer_8_type */
	T0* a17; /* integer_16_type */
	T0* a18; /* integer_32_type */
	T0* a19; /* integer_64_type */
	T0* a20; /* natural_8_type */
	T0* a21; /* natural_16_type */
	T0* a22; /* natural_32_type */
	T0* a23; /* natural_64_type */
	T0* a24; /* real_32_type */
	T0* a25; /* real_64_type */
	T0* a26; /* pointer_type */
	T0* a27; /* special_any_type */
	T0* a28; /* universe */
	T0* a29; /* string_8_type */
	T0* a30; /* string_32_type */
	T0* a31; /* immutable_string_8_type */
	T0* a32; /* immutable_string_32_type */
	T0* a33; /* array_any_type */
	T0* a34; /* special_identity_type */
	T0* a35; /* typed_pointer_identity_type */
	T0* a36; /* routine_identity_type */
	T0* a37; /* detachable_tuple_type */
	T0* a38; /* ise_exception_manager_type */
	T0* a39; /* detachable_none_type */
	T1 a40; /* attachment_type_conformance_mode */
	T1 a41; /* use_boehm_gc */
	T0* a42; /* external_include_pathnames */
	T0* a43; /* external_cflags */
	T0* a44; /* external_linker_flags */
	T0* a45; /* external_library_pathnames */
	T0* a46; /* external_object_pathnames */
	T0* a47; /* character_type */
	T0* a48; /* wide_character_type */
	T0* a49; /* integer_type */
	T0* a50; /* natural_type */
	T0* a51; /* real_type */
	T0* a52; /* double_type */
	T0* a53; /* external_make_pathnames */
	T0* a54; /* external_resource_pathnames */
	T6 a55; /* is_equal_seed */
	T6 a56; /* routine_call_seed */
	T6 a57; /* function_item_seed */
	T6 a58; /* dispose_seed */
	T1 a59; /* trace_mode */
	T1 a60; /* exception_trace_mode */
	T1 a61; /* multithreaded_mode */
	T1 a62; /* total_order_on_reals_mode */
	T6 a63; /* twin_seed */
	T1 a64; /* console_application_mode */
	T0* a65; /* system_name */
	T6 a66; /* copy_seed */
	T0* a67; /* name */
	T0* a68; /* ecf_namespace */
	T0* a69; /* ecf_version */
	T0* a70; /* description */
	T0* a71; /* notes */
	T1 a72; /* is_read_only */
	T0* a73; /* uuid */
	T0* a74; /* register_class_mutex */
	T0* a75; /* current_system */
	T0* a76; /* tuple_type */
	T0* a77; /* detachable_exception_type */
	T0* a78; /* exception_type */
	T0* a79; /* string_type */
	T1 a80; /* is_preparsed */
	T0* a81; /* clusters */
	T0* a82; /* implicit_attachment_type_mark */
	T0* a83; /* master_class_mutex */
	T0* a84; /* detachable_any_type */
	T0* a85; /* detachable_separate_any_type */
	T0* a86; /* any_parent */
	T0* a87; /* any_parents */
	T0* a88; /* any_clients */
	T0* a89; /* tuple_identity_type */
	T0* a90; /* unfolded_empty_tuple_actual_parameters */
	T0* a91; /* array_detachable_any_type */
	T0* a92; /* array_none_type */
	T0* a93; /* array_identity_type */
	T0* a94; /* character_8_convert_feature */
	T0* a95; /* character_32_convert_feature */
	T0* a96; /* exception_manager_type */
	T0* a97; /* function_identity_any_type */
	T0* a98; /* integer_8_convert_feature */
	T0* a99; /* integer_16_convert_feature */
	T0* a100; /* integer_32_convert_feature */
	T0* a101; /* integer_64_convert_feature */
	T0* a102; /* iterable_detachable_separate_any_type */
	T0* a103; /* natural_8_convert_feature */
	T0* a104; /* natural_16_convert_feature */
	T0* a105; /* natural_32_convert_feature */
	T0* a106; /* natural_64_convert_feature */
	T0* a107; /* predicate_identity_type */
	T0* a108; /* procedure_identity_type */
	T0* a109; /* real_32_convert_feature */
	T0* a110; /* real_64_convert_feature */
	T0* a111; /* special_detachable_any_type */
	T0* a112; /* detachable_string_type */
	T0* a113; /* detachable_string_8_type */
	T0* a114; /* string_8_convert_feature */
	T0* a115; /* string_32_convert_feature */
	T0* a116; /* system_object_type */
	T0* a117; /* system_object_parents */
	T0* a118; /* system_string_type */
	T0* a119; /* type_detachable_any_type */
	T0* a120; /* detachable_type_exception_type */
	T0* a121; /* type_detachable_exception_type */
	T0* a122; /* type_detachable_like_current_type */
	T0* a123; /* type_identity_type */
	T0* a124; /* scm_read_mapping_builder */
	T0* a125; /* scm_write_mapping_builder */
	T6 a126; /* registered_class_count */
	T1 a127; /* scoop_mode */
	T1 a128; /* use_obsolete_syntax_mode */
	T1 a129; /* obsolete_iteration_mode */
	T1 a130; /* is_dotnet */
	T1 a131; /* obsolete_routine_type_mode */
	T1 a132; /* line_generation_mode */
	T1 a133; /* target_type_attachment_mode */
	T6 a134; /* iterable_new_cursor_seed */
	T6 a135; /* iteration_cursor_item_seed */
	T6 a136; /* iteration_cursor_after_seed */
	T6 a137; /* iteration_cursor_forth_seed */
	T6 a138; /* special_item_seed */
};

/* Struct for type [detachable] ET_ECF_TARGET */
struct S59 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* settings */
	T0* a2; /* capabilities */
	T0* a3; /* options */
	T0* a4; /* system_config */
	T0* a5; /* root */
	T0* a6; /* variables */
	T0* a7; /* name */
	T0* a8; /* parent */
	T0* a9; /* description */
	T0* a10; /* precompiled_library */
	T0* a11; /* version */
	T1 a12; /* is_abstract */
	T0* a13; /* clusters */
	T0* a14; /* libraries */
	T0* a15; /* dotnet_assemblies */
	T0* a16; /* file_rules */
	T0* a17; /* class_mappings */
	T0* a18; /* external_cflags */
	T0* a19; /* external_includes */
	T0* a20; /* external_libraries */
	T0* a21; /* external_linker_flags */
	T0* a22; /* external_makes */
	T0* a23; /* external_objects */
	T0* a24; /* external_resources */
	T0* a25; /* pre_compile_actions */
	T0* a26; /* post_compile_actions */
	T0* a27; /* notes */
};

/* Struct for type detachable ET_ECF_CAPABILITIES */
struct S61 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* primary_use_capabilities */
	T0* a2; /* primary_support_capabilities */
	T0* a3; /* secondary_capabilities */
};

/* Struct for type detachable ET_ECF_VARIABLES */
struct S62 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* secondary_variables */
	T0* a2; /* primary_variables */
};

/* Struct for type [detachable] ET_DYNAMIC_SYSTEM */
struct S63 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* dynamic_types */
	T1 a3; /* catcall_error_mode */
	T0* a4; /* current_system */
	T0* a5; /* system_processor */
	T0* a6; /* dynamic_generic_types_by_name */
	T0* a7; /* null_dynamic_type_set_builder */
	T1 a8; /* full_class_checking */
	T1 a9; /* catcall_warning_mode */
	T0* a10; /* new_instance_types */
	T0* a11; /* dynamic_type_set_builder */
	T0* a12; /* unknown_type */
	T0* a13; /* boolean_type */
	T0* a14; /* character_8_type */
	T0* a15; /* character_32_type */
	T0* a16; /* immutable_string_8_type */
	T0* a17; /* immutable_string_32_type */
	T0* a18; /* integer_8_type */
	T0* a19; /* integer_16_type */
	T0* a20; /* integer_32_type */
	T0* a21; /* integer_64_type */
	T0* a22; /* natural_8_type */
	T0* a23; /* natural_16_type */
	T0* a24; /* natural_32_type */
	T0* a25; /* natural_64_type */
	T0* a26; /* real_32_type */
	T0* a27; /* real_64_type */
	T0* a28; /* pointer_type */
	T0* a29; /* string_8_type */
	T0* a30; /* string_32_type */
	T0* a31; /* special_character_8_type */
	T0* a32; /* special_character_32_type */
	T0* a33; /* ise_exception_manager_type */
	T0* a34; /* any_type */
	T0* a35; /* none_type */
	T0* a36; /* root_type */
	T0* a37; /* root_creation_procedure */
	T0* a38; /* special_count_feature */
	T0* a39; /* special_capacity_feature */
	T0* a40; /* special_item_feature */
	T0* a41; /* array_area_feature */
	T0* a42; /* array_lower_feature */
	T0* a43; /* array_upper_feature */
	T0* a44; /* typed_pointer_to_pointer_feature */
	T0* a45; /* routine_closed_operands_feature */
	T0* a46; /* routine_rout_disp_feature */
	T0* a47; /* routine_set_rout_disp_final_feature */
	T0* a48; /* ise_exception_manager_init_exception_manager_feature */
	T0* a49; /* ise_exception_manager_last_exception_feature */
	T0* a50; /* ise_exception_manager_once_raise_feature */
	T0* a51; /* ise_exception_manager_set_exception_data_feature */
	T0* a52; /* ise_runtime_type_conforms_to_feature */
	T1 a53; /* in_dynamic_primary_type */
	T0* a54; /* ise_runtime_new_type_instance_of_feature */
	T1 a55; /* in_create_meta_type */
	T1 a56; /* all_attributes_used */
	T0* a57; /* ise_runtime_reference_field_feature */
	T0* a58; /* ise_runtime_set_reference_field_feature */
};

/* Struct for type [detachable] ET_SYSTEM_PROCESSOR */
struct S64 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* interface_checker */
	T0* a2; /* stop_request */
	T0* a3; /* eiffel_parser */
	T0* a4; /* error_handler */
	T0* a5; /* implementation_checker */
	T1 a6; /* benchmark_shown */
	T0* a7; /* ise_version */
	T0* a8; /* eiffel_preparser */
	T0* a9; /* master_class_checker */
	T0* a10; /* provider_checker */
	T0* a11; /* ancestor_builder */
	T0* a12; /* feature_flattener */
	T0* a13; /* processed_class_count_stack */
	T0* a14; /* ast_factory */
	T0* a15; /* dotnet_assembly_consumer */
	T0* a16; /* ecma_version */
	T1 a17; /* nested_benchmark_shown */
	T1 a18; /* metrics_shown */
	T1 a19; /* flat_mode */
	T1 a20; /* flat_dbc_mode */
	T1 a21; /* preparse_shallow_mode */
	T1 a22; /* preparse_single_mode */
	T1 a23; /* preparse_multiple_mode */
	T1 a24; /* unknown_builtin_reported */
	T1 a25; /* qualified_anchored_types_cycle_detection_enabled */
	T1 a26; /* use_attribute_keyword */
	T1 a27; /* use_note_keyword */
	T1 a28; /* use_reference_keyword */
	T1 a29; /* use_attached_keyword */
	T1 a30; /* use_detachable_keyword */
	T1 a31; /* cluster_dependence_enabled */
	T6 a32; /* postponed_class_count */
	T6 a33; /* processed_class_count */
	T1 a34; /* suppliers_enabled */
	T1 a35; /* preparse_override_mode */
	T1 a36; /* preparse_readonly_mode */
	T1 a37; /* providers_enabled */
	T1 a38; /* use_cluster_dependence_pathnames */
};

/* Struct for type [detachable] ET_C_GENERATOR */
struct S67 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* never_void_target_count */
	T6 a2; /* can_be_void_target_count */
	T1 a3; /* has_fatal_error */
	T0* a4; /* system_processor */
	T0* a5; /* current_dynamic_system */
	T0* a6; /* dynamic_types */
	T1 a7; /* short_names */
	T1 a8; /* split_mode */
	T6 a9; /* split_threshold */
	T1 a10; /* use_boehm_gc */
	T0* a11; /* system_name */
	T0* a12; /* current_file */
	T0* a13; /* header_file */
	T0* a14; /* current_type */
	T0* a15; /* current_feature */
	T0* a16; /* extra_dynamic_type_sets */
	T0* a17; /* current_dynamic_type_sets */
	T0* a18; /* current_function_header_buffer */
	T0* a19; /* current_function_body_buffer */
	T0* a20; /* temp_variables */
	T0* a21; /* used_temp_variables */
	T0* a22; /* free_temp_variables */
	T0* a23; /* frozen_temp_variables */
	T0* a24; /* unused_unqualified_call_expressions */
	T0* a25; /* unused_qualified_call_expressions */
	T0* a26; /* unused_actual_argument_lists */
	T0* a27; /* unused_result_expressions */
	T0* a28; /* unused_identifiers */
	T0* a29; /* conforming_types */
	T0* a30; /* non_conforming_types */
	T0* a31; /* equality_type_set */
	T0* a32; /* equality_common_types */
	T0* a33; /* unused_standalone_type_sets */
	T0* a34; /* operand_stack */
	T0* a35; /* call_operands */
	T0* a36; /* attachment_dynamic_type_ids */
	T0* a37; /* target_dynamic_type_ids */
	T0* a38; /* target_dynamic_types */
	T0* a39; /* standalone_type_sets */
	T0* a40; /* deep_twin_types */
	T0* a41; /* deep_equal_types */
	T0* a42; /* deep_feature_target_type_sets */
	T0* a43; /* current_object_tests */
	T0* a44; /* current_object_equalities */
	T0* a45; /* current_equalities */
	T0* a46; /* current_agents */
	T0* a47; /* agent_open_operands */
	T0* a48; /* agent_closed_operands */
	T0* a49; /* agent_target */
	T0* a50; /* agent_arguments */
	T0* a51; /* agent_instruction */
	T0* a52; /* agent_expression */
	T0* a53; /* agent_closed_operands_type */
	T0* a54; /* agent_manifest_tuple */
	T0* a55; /* formal_arguments */
	T0* a56; /* manifest_array_types */
	T0* a57; /* big_manifest_array_types */
	T0* a58; /* manifest_tuple_types */
	T0* a59; /* once_features */
	T0* a60; /* once_per_process_counts */
	T0* a61; /* once_per_thread_counts */
	T0* a62; /* constant_features */
	T0* a63; /* inline_constants */
	T0* a64; /* dispose_procedures */
	T0* a65; /* dynamic_type_id_set_names */
	T0* a66; /* called_features */
	T0* a67; /* called_static_features */
	T0* a68; /* included_header_filenames */
	T0* a69; /* included_cpp_header_filenames */
	T0* a70; /* included_runtime_header_files */
	T0* a71; /* included_runtime_c_files */
	T0* a72; /* c_filenames */
	T1 a73; /* finalize_mode */
	T0* a74; /* locals_written_in_body */
	T0* a75; /* locals_written_in_rescue */
	T0* a76; /* locals_written */
	T0* a77; /* locals_read_in_body */
	T0* a78; /* locals_read_in_rescue */
	T0* a79; /* locals_read */
	T0* a80; /* external_c_regexp */
	T0* a81; /* external_c_macro_regexp */
	T0* a82; /* external_c_struct_regexp */
	T0* a83; /* external_c_inline_regexp */
	T0* a84; /* old_external_c_regexp */
	T0* a85; /* old_external_c_macro_regexp */
	T0* a86; /* old_external_c_struct_regexp */
	T0* a87; /* external_cpp_regexp */
	T0* a88; /* external_cpp_macro_regexp */
	T0* a89; /* external_cpp_struct_regexp */
	T0* a90; /* external_cpp_inline_regexp */
	T0* a91; /* external_dllwin_regexp */
	T0* a92; /* old_external_dllwin32_regexp */
	T1 a93; /* in_static_feature */
	T1 a94; /* type_info_ancestors_used */
	T1 a95; /* type_info_attributes_used */
	T1 a96; /* type_info_attribute_name_used */
	T1 a97; /* type_info_attribute_type_id_used */
	T1 a98; /* type_info_attribute_offset_used */
	T1 a99; /* type_info_generator_used */
	T1 a100; /* type_info_name_used */
	T1 a101; /* type_info_generic_parameters_used */
	T1 a102; /* type_info_object_size_used */
	T0* a103; /* c_file */
	T6 a104; /* c_file_size */
	T0* a105; /* cpp_file */
	T6 a106; /* cpp_file_size */
	T6 a107; /* indentation */
	T0* a108; /* call_target_type */
	T0* a109; /* assignment_target */
	T1 a110; /* in_operand */
	T1 a111; /* in_procedure_call_target */
	T0* a112; /* current_agent */
	T1 a113; /* call_target_check_void */
	T1 a114; /* has_rescue */
	T1 a115; /* result_written */
	T1 a116; /* result_read */
	T1 a117; /* has_retry */
};

/* Struct for type detachable DT_DATE_TIME */
struct S68 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* time_storage */
	T6 a2; /* date_storage */
};

/* Struct for type detachable DS_HASH_SET [[attached] STRING_8] */
struct S70 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* internal_cursor */
	T6 a3; /* count */
	T0* a4; /* item_storage */
	T0* a5; /* equality_tester */
	T6 a6; /* capacity */
	T6 a7; /* slots_position */
	T6 a8; /* free_slot */
	T6 a9; /* last_position */
	T6 a10; /* modulus */
	T6 a11; /* clashes_previous_position */
	T6 a12; /* found_position */
	T0* a13; /* special_item_routines */
	T0* a14; /* clashes */
	T0* a15; /* slots */
	T0* a16; /* hash_function */
};

/* Struct for type [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER */
struct S71 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* dynamic_type_set_count */
	T1 a2; /* has_fatal_error */
	T0* a3; /* current_dynamic_system */
	T0* a4; /* feature_checker */
	T0* a5; /* current_dynamic_type */
	T0* a6; /* current_dynamic_feature */
	T0* a7; /* dynamic_type_sets */
	T0* a8; /* object_id_dynamic_type_set */
	T0* a9; /* constant_indexes */
	T1 a10; /* catcall_error_mode */
	T1 a11; /* catcall_warning_mode */
	T1 a12; /* no_debug */
	T1 a13; /* no_assertion */
	T1 a14; /* is_built */
	T0* a15; /* alive_conforming_descendants_per_type */
	T0* a16; /* type_checker */
	T0* a17; /* current_type */
	T0* a18; /* current_class */
	T0* a19; /* current_feature */
	T0* a20; /* current_feature_impl */
	T0* a21; /* current_class_impl */
	T0* a22; /* enclosing_inline_agents */
	T0* a23; /* unused_overloaded_procedures_list */
	T0* a24; /* unused_overloaded_queries_list */
	T0* a25; /* unused_overloaded_features_list */
	T0* a26; /* unused_call_infos */
	T0* a27; /* unused_contexts */
	T0* a28; /* current_context */
	T0* a29; /* current_target_type */
	T0* a30; /* current_object_test_types */
	T0* a31; /* current_object_test_scope */
	T0* a32; /* object_test_scope_builder */
	T0* a33; /* current_iteration_cursor_types */
	T0* a34; /* current_iteration_item_types */
	T0* a35; /* current_iteration_item_scope */
	T0* a36; /* current_initialization_scope */
	T0* a37; /* current_attachment_scope */
	T0* a38; /* attachment_scope_builder */
	T0* a39; /* unused_attachment_scopes */
	T0* a40; /* assertions_by_feature */
	T0* a41; /* common_ancestor_type_list */
	T0* a42; /* indexing_term_list */
	T0* a43; /* default_creation_call_name */
	T0* a44; /* default_creation_call */
	T0* a45; /* vape_non_descendant_clients */
	T0* a46; /* vape_creation_clients */
	T0* a47; /* vape_client */
	T0* a48; /* unused_adapted_base_classes */
	T0* a49; /* adapted_base_class_checker */
	T0* a50; /* system_processor */
	T6 a51; /* current_index */
	T6 a52; /* result_index */
	T6 a53; /* attached_result_index */
	T6 a54; /* void_index */
	T1 a55; /* in_precursor */
	T1 a56; /* in_static_feature */
	T1 a57; /* in_rescue */
	T0* a58; /* current_inline_agent */
	T0* a59; /* supplier_handler */
	T1 a60; /* in_invariant */
	T1 a61; /* in_precondition */
	T1 a62; /* in_postcondition */
	T1 a63; /* in_check_instruction */
	T0* a64; /* precursor_queries */
	T0* a65; /* precursor_procedures */
};

/* Struct for type [detachable] ET_CLASS_TYPE */
struct S72 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type_mark */
	T0* a2; /* name */
	T0* a3; /* actual_parameters */
	T0* a4; /* named_base_class */
	T1 a5; /* tuple_actual_parameters_unfolded_1 */
	T1 a6; /* tuple_actual_parameters_unfolded_2 */
};

/* Struct for type [detachable] ET_TOKEN_CONSTANTS */
struct S73 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_CLASS */
struct S74 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* group */
	T0* a2; /* status_mutex */
	T1 a3; /* unprotected_is_parsed */
	T1 a4; /* unprotected_has_syntax_error */
	T1 a5; /* unprotected_ancestors_built */
	T1 a6; /* unprotected_has_ancestors_error */
	T1 a7; /* unprotected_features_flattened */
	T1 a8; /* unprotected_has_flattening_error */
	T1 a9; /* unprotected_interface_checked */
	T1 a10; /* unprotected_has_interface_error */
	T1 a11; /* unprotected_implementation_checked */
	T1 a12; /* unprotected_has_implementation_error */
	T0* a13; /* name */
	T6 a14; /* id */
	T0* a15; /* ancestors */
	T0* a16; /* conforming_ancestors */
	T0* a17; /* queries */
	T0* a18; /* procedures */
	T0* a19; /* class_keyword */
	T0* a20; /* end_keyword */
	T0* a21; /* processing_mutex */
	T0* a22; /* formal_parameter_types_mutex */
	T0* a23; /* named_base_class */
	T6 a24; /* time_stamp */
	T8 a25; /* class_code */
	T1 a26; /* tuple_actual_parameters_unfolded_1 */
	T1 a27; /* tuple_actual_parameters_unfolded_2 */
	T1 a28; /* is_checking_implementation */
	T0* a29; /* formal_parameters */
	T6 a30; /* index */
	T0* a31; /* class_mark */
	T1 a32; /* has_deferred_features */
	T0* a33; /* formal_parameter_types */
	T0* a34; /* parent_clauses */
	T0* a35; /* suppliers */
	T1 a36; /* redeclared_signatures_checked */
	T0* a37; /* invariants */
	T0* a38; /* filename */
	T0* a39; /* creators */
	T0* a40; /* convert_features */
	T1 a41; /* is_ignored */
	T0* a42; /* providers */
	T1 a43; /* is_interface */
	T1 a44; /* unprotected_is_marked */
	T1 a45; /* has_utf8_bom */
	T6 a46; /* registered_feature_count */
	T0* a47; /* external_keyword */
	T0* a48; /* frozen_keyword */
	T0* a49; /* feature_clauses */
	T0* a50; /* first_indexing */
	T0* a51; /* second_indexing */
	T6 a52; /* tuple_constraint_position */
	T0* a53; /* obsolete_message */
	T6 a54; /* registered_inline_constant_count */
	T0* a55; /* leading_break */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S76 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] UT_CANNOT_READ_FILE_ERROR */
struct S77 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] UT_VERSION_NUMBER */
struct S79 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] UT_MESSAGE */
struct S80 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] RX_PCRE_REGULAR_EXPRESSION */
struct S81 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* match_count */
	T0* a2; /* subject */
	T6 a3; /* subject_end */
	T6 a4; /* subject_start */
	T0* a5; /* offset_vector */
	T6 a6; /* offset_vector_count */
	T6 a7; /* brastart_capacity */
	T0* a8; /* brastart_vector */
	T6 a9; /* brastart_lower */
	T6 a10; /* brastart_count */
	T6 a11; /* eptr_capacity */
	T0* a12; /* eptr_vector */
	T6 a13; /* eptr_lower */
	T6 a14; /* eptr_upper */
	T0* a15; /* byte_code */
	T0* a16; /* internal_start_bits */
	T0* a17; /* pattern_buffer */
	T6 a18; /* subexpression_count */
	T0* a19; /* pattern */
	T6 a20; /* pattern_count */
	T6 a21; /* pattern_position */
	T6 a22; /* code_index */
	T6 a23; /* maxbackrefs */
	T10 a24; /* optchanged */
	T7 a25; /* first_character */
	T7 a26; /* required_character */
	T6 a27; /* regexp_countlits */
	T0* a28; /* start_bits */
	T0* a29; /* error_message */
	T1 a30; /* is_anchored */
	T0* a31; /* character_case_mapping */
	T0* a32; /* word_set */
	T6 a33; /* subject_next_start */
	T6 a34; /* error_code */
	T6 a35; /* error_position */
	T1 a36; /* is_startline */
	T1 a37; /* is_caseless */
	T1 a38; /* is_bol */
	T1 a39; /* is_eol */
	T1 a40; /* is_multiline */
	T1 a41; /* is_dotall */
	T1 a42; /* is_extended */
	T1 a43; /* is_empty_allowed */
	T1 a44; /* is_dollar_endonly */
	T1 a45; /* is_greedy */
	T1 a46; /* is_strict */
	T1 a47; /* is_ichanged */
	T6 a48; /* first_matched_index */
	T6 a49; /* eptr */
	T6 a50; /* offset_top */
	T1 a51; /* is_matching_caseless */
	T1 a52; /* is_matching_multiline */
	T1 a53; /* is_matching_dotall */
};

/* Struct for type [detachable] KL_STRING_ROUTINES */
struct S82 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_EXECUTION_ENVIRONMENT */
struct S86 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] AP_ERROR */
struct S87 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
	T0* a2; /* default_template */
	T0* a3; /* code */
};

/* Struct for type [detachable] KL_SHELL_COMMAND */
struct S88 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* exit_code */
	T0* a2; /* command */
	T6 a3; /* return_code */
	T1 a4; /* is_system_code */
};

/* Struct for type [detachable] GECC */
struct S89 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* exit_code */
	T0* a2; /* error_handler */
	T0* a3; /* version_flag */
	T0* a4; /* script_filename */
	T0* a5; /* thread_option */
};

/* Struct for type [detachable] KL_OPERATING_SYSTEM */
struct S90 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] VOID_TARGET */
struct S93 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] ROUTINE_FAILURE */
struct S95 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* routine_name */
	T0* a2; /* class_name */
	T0* a3; /* throwing_exception */
	T0* a4; /* internal_trace */
	T0* a5; /* c_description */
	T0* a6; /* recipient_name */
	T0* a7; /* type_name */
};

/* Struct for type [detachable] OLD_VIOLATION */
struct S97 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] NO_MORE_MEMORY */
struct S99 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_trace */
	T6 a2; /* internal_code */
	T0* a3; /* throwing_exception */
	T0* a4; /* c_description */
	T0* a5; /* recipient_name */
	T0* a6; /* type_name */
};

/* Struct for type [detachable] INVARIANT_VIOLATION */
struct S100 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T1 a2; /* is_entry */
	T0* a3; /* internal_trace */
	T0* a4; /* c_description */
	T0* a5; /* recipient_name */
	T0* a6; /* type_name */
};

/* Struct for type [detachable] OPERATING_SYSTEM_SIGNAL_FAILURE */
struct S101 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T6 a2; /* signal_code */
	T0* a3; /* internal_trace */
	T0* a4; /* c_description */
	T0* a5; /* recipient_name */
	T0* a6; /* type_name */
};

/* Struct for type [detachable] IO_FAILURE */
struct S102 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* internal_code */
	T0* a2; /* throwing_exception */
	T6 a3; /* error_code */
	T0* a4; /* internal_trace */
	T0* a5; /* c_description */
	T0* a6; /* recipient_name */
	T0* a7; /* type_name */
};

/* Struct for type [detachable] OPERATING_SYSTEM_FAILURE */
struct S103 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T6 a2; /* error_code */
	T0* a3; /* internal_trace */
	T0* a4; /* c_description */
	T0* a5; /* recipient_name */
	T0* a6; /* type_name */
};

/* Struct for type [detachable] COM_FAILURE */
struct S104 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T6 a2; /* hresult_code */
	T0* a3; /* exception_information */
	T6 a4; /* hresult */
	T0* a5; /* internal_trace */
	T0* a6; /* c_description */
	T0* a7; /* recipient_name */
	T0* a8; /* type_name */
};

/* Struct for type [detachable] EIFFEL_RUNTIME_PANIC */
struct S105 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* internal_code */
	T0* a2; /* throwing_exception */
	T0* a3; /* internal_trace */
	T0* a4; /* c_description */
	T0* a5; /* recipient_name */
	T0* a6; /* type_name */
};

/* Struct for type [detachable] PRECONDITION_VIOLATION */
struct S107 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] POSTCONDITION_VIOLATION */
struct S108 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] FLOATING_POINT_FAILURE */
struct S109 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] CHECK_VIOLATION */
struct S110 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] BAD_INSPECT_VALUE */
struct S111 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] VARIANT_VIOLATION */
struct S112 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] LOOP_INVARIANT_VIOLATION */
struct S113 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] RESCUE_FAILURE */
struct S114 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] RESUMPTION_FAILURE */
struct S115 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] CREATE_ON_DEFERRED */
struct S116 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] EXTERNAL_FAILURE */
struct S117 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] VOID_ASSIGNED_TO_EXPANDED */
struct S118 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] EXCEPTION_IN_SIGNAL_HANDLER_FAILURE */
struct S119 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] MISMATCH_FAILURE */
struct S120 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] DEVELOPER_EXCEPTION */
struct S121 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] ADDRESS_APPLIED_TO_MELTED_FEATURE */
struct S122 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] SERIALIZATION_FAILURE */
struct S123 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* throwing_exception */
	T0* a2; /* internal_trace */
	T0* a3; /* c_description */
	T0* a4; /* recipient_name */
	T0* a5; /* type_name */
};

/* Struct for type [detachable] EXECUTION_ENVIRONMENT */
struct S124 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* return_code */
};

/* Struct for type [detachable] KL_WINDOWS_FILE_SYSTEM */
struct S125 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_UNIX_FILE_SYSTEM */
struct S126 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] PRIMES */
struct S127 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] INTEGER_32] */
struct S128 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T6 z2[1]; /* item */
};

/* Struct for type [detachable] SPECIAL [[attached] BOOLEAN] */
struct S129 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T1 z2[1]; /* item */
};

/* Struct for type [detachable] ARGUMENTS_32 */
struct S130 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] MUTEX */
struct S131 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* owner_thread_id */
	T14 a2; /* mutex_pointer */
};

/* Struct for type [detachable] UT_ERROR_HANDLER */
struct S132 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* error_file */
	T0* a2; /* warning_file */
	T0* a3; /* info_file */
};

/* Struct for type [detachable] KL_STANDARD_FILES */
struct S133 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_STDERR_FILE */
struct S134 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_STDOUT_FILE */
struct S135 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable KL_LINKABLE [[attached] CHARACTER_8] */
struct S137 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 a1; /* item */
	T0* a2; /* right */
};

/* Struct for type [detachable] MANAGED_POINTER */
struct S140 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* item */
	T1 a2; /* is_shared */
	T6 a3; /* count */
};

/* Struct for type [detachable] FILE_INFO */
struct S141 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* buffered_file_info */
	T1 a2; /* is_following_symlinks */
	T1 a3; /* exists */
	T0* a4; /* internal_file_name */
	T0* a5; /* internal_name_pointer */
};

/* Struct for type [detachable] ST_WORD_WRAPPER */
struct S142 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* new_line_indentation */
	T6 a2; /* broken_words */
	T6 a3; /* maximum_text_width */
};

/* Struct for type [detachable] AP_DISPLAY_HELP_FLAG */
struct S144 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* description */
	T2 a2; /* short_form */
	T1 a3; /* has_short_form */
	T0* a4; /* long_form */
	T6 a5; /* occurrences */
	T1 a6; /* is_hidden */
	T1 a7; /* is_mandatory */
	T6 a8; /* maximum_occurrences */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION] */
struct S145 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T6 a2; /* count */
	T0* a3; /* equality_tester */
	T0* a4; /* storage */
	T0* a5; /* special_routines */
	T6 a6; /* capacity */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
struct S146 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T6 a5; /* count */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] STRING_8] */
struct S147 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* internal_cursor */
	T0* a4; /* special_routines */
	T6 a5; /* capacity */
	T0* a6; /* equality_tester */
};

/* Struct for type [detachable] AP_ERROR_HANDLER */
struct S148 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_error */
	T0* a2; /* error_file */
	T0* a3; /* warning_file */
	T0* a4; /* info_file */
};

/* Struct for type [detachable] DS_LINKED_LIST_CURSOR [[attached] AP_OPTION] */
struct S149 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T1 a2; /* before */
	T1 a3; /* after */
	T0* a4; /* current_cell */
	T0* a5; /* next_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [detachable STRING_8] */
struct S150 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8] */
struct S151 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* container */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] DS_LINKED_LIST [[attached] STRING_8] */
struct S152 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T0* a2; /* last_cell */
	T6 a3; /* count */
	T0* a4; /* first_cell */
	T0* a5; /* equality_tester */
};

/* Struct for type [detachable] KL_STRING_EQUALITY_TESTER */
struct S153 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable KL_EQUALITY_TESTER [[attached] STRING_8] */
struct S154 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_LINKED_LIST [[attached] BOOLEAN] */
struct S157 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* last_cell */
	T0* a3; /* internal_cursor */
	T0* a4; /* first_cell */
};

/* Struct for type [detachable] DS_LINKED_LIST [[attached] INTEGER_32] */
struct S160 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* last_cell */
	T0* a3; /* internal_cursor */
	T0* a4; /* first_cell */
};

/* Struct for type [detachable] KL_NULL_TEXT_OUTPUT_STREAM */
struct S161 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
};

/* Struct for type [detachable] ET_ECF_AST_FACTORY */
struct S162 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_CELL [detachable ET_ECF_SYSTEM] */
struct S163 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_TARGET] */
struct S164 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_TARGET]] */
struct S165 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type detachable TUPLE */
struct S166 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER] */
struct S167 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] XM_ELEMENT */
struct S168 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* namespace */
	T0* a3; /* parent */
	T0* a4; /* children */
};

/* Struct for type detachable XM_POSITION_TABLE */
struct S169 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* table */
};

/* Struct for type [detachable] TUPLE [[attached] XM_ELEMENT, detachable XM_POSITION_TABLE, [attached] STRING_8] */
struct S170 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] XM_ELEMENT, detachable XM_POSITION_TABLE, [attached] STRING_8]] */
struct S171 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER, [detachable] STRING_8, [attached] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_TARGET]], [attached] DS_CELL [detachable ET_ECF_SYSTEM]] */
struct S172 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
	T0* z4;
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8] */
struct S174 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T6 a2; /* found_position */
	T0* a3; /* item_storage */
	T0* a4; /* key_equality_tester */
	T0* a5; /* internal_keys */
	T6 a6; /* count */
	T6 a7; /* last_position */
	T6 a8; /* free_slot */
	T6 a9; /* position */
	T6 a10; /* capacity */
	T0* a11; /* equality_tester */
	T0* a12; /* key_storage */
	T0* a13; /* clashes */
	T6 a14; /* modulus */
	T0* a15; /* slots */
	T6 a16; /* slots_position */
	T6 a17; /* clashes_previous_position */
	T0* a18; /* special_item_routines */
	T0* a19; /* special_key_routines */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] KL_CASE_INSENSITIVE_STRING_EQUALITY_TESTER */
struct S175 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8] */
struct S176 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T0* a3; /* key_equality_tester */
	T0* a4; /* internal_keys */
	T6 a5; /* count */
	T6 a6; /* last_position */
	T6 a7; /* free_slot */
	T6 a8; /* position */
	T6 a9; /* capacity */
	T0* a10; /* equality_tester */
	T0* a11; /* internal_cursor */
	T0* a12; /* key_storage */
	T0* a13; /* clashes */
	T6 a14; /* modulus */
	T0* a15; /* slots */
	T6 a16; /* slots_position */
	T6 a17; /* clashes_previous_position */
	T0* a18; /* special_item_routines */
	T0* a19; /* special_key_routines */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] XM_EIFFEL_PARSER */
struct S178 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_last_error_description */
	T0* a2; /* scanner */
	T0* a3; /* error_positions */
	T0* a4; /* scanners */
	T1 a5; /* in_external_dtd */
	T0* a6; /* callbacks */
	T0* a7; /* entities */
	T0* a8; /* pe_entities */
	T0* a9; /* dtd_resolver */
	T0* a10; /* entity_resolver */
	T1 a11; /* use_namespaces */
	T6 a12; /* string_mode */
	T0* a13; /* dtd_callbacks */
	T0* a14; /* last_string_value */
	T0* a15; /* yyss */
	T0* a16; /* yyspecial_routines1 */
	T6 a17; /* yyvsc1 */
	T0* a18; /* yyvs1 */
	T0* a19; /* yyspecial_routines2 */
	T6 a20; /* yyvsc2 */
	T0* a21; /* yyvs2 */
	T0* a22; /* yyspecial_routines3 */
	T6 a23; /* yyvsc3 */
	T0* a24; /* yyvs3 */
	T0* a25; /* yyspecial_routines4 */
	T6 a26; /* yyvsc4 */
	T0* a27; /* yyvs4 */
	T0* a28; /* yyspecial_routines5 */
	T6 a29; /* yyvsc5 */
	T0* a30; /* yyvs5 */
	T0* a31; /* yyspecial_routines6 */
	T6 a32; /* yyvsc6 */
	T0* a33; /* yyvs6 */
	T0* a34; /* yyspecial_routines7 */
	T6 a35; /* yyvsc7 */
	T0* a36; /* yyvs7 */
	T0* a37; /* yyspecial_routines8 */
	T6 a38; /* yyvsc8 */
	T0* a39; /* yyvs8 */
	T0* a40; /* yyspecial_routines9 */
	T6 a41; /* yyvsc9 */
	T0* a42; /* yyvs9 */
	T0* a43; /* yyspecial_routines10 */
	T6 a44; /* yyvsc10 */
	T0* a45; /* yyvs10 */
	T0* a46; /* yyspecial_routines11 */
	T6 a47; /* yyvsc11 */
	T0* a48; /* yyvs11 */
	T0* a49; /* yytranslate */
	T0* a50; /* yyr1 */
	T0* a51; /* yytypes1 */
	T0* a52; /* yytypes2 */
	T0* a53; /* yydefact */
	T0* a54; /* yydefgoto */
	T0* a55; /* yypact */
	T0* a56; /* yypgoto */
	T0* a57; /* yytable */
	T0* a58; /* yycheck */
	T6 a59; /* yy_parsing_status */
	T6 a60; /* yy_suspended_yystacksize */
	T6 a61; /* yy_suspended_yystate */
	T6 a62; /* yy_suspended_yyn */
	T6 a63; /* yy_suspended_yychar1 */
	T6 a64; /* yy_suspended_index */
	T6 a65; /* yy_suspended_yyss_top */
	T6 a66; /* yy_suspended_yy_goto */
	T6 a67; /* yyssp */
	T6 a68; /* error_count */
	T1 a69; /* yy_lookahead_needed */
	T6 a70; /* yyerrstatus */
	T6 a71; /* last_token */
	T6 a72; /* yyvsp1 */
	T6 a73; /* yyvsp2 */
	T6 a74; /* yyvsp3 */
	T6 a75; /* yyvsp4 */
	T6 a76; /* yyvsp5 */
	T6 a77; /* yyvsp6 */
	T6 a78; /* yyvsp7 */
	T6 a79; /* yyvsp8 */
	T6 a80; /* yyvsp9 */
	T6 a81; /* yyvsp10 */
	T6 a82; /* yyvsp11 */
	T0* a83; /* last_detachable_any_value */
};

/* Struct for type [detachable] XM_TREE_CALLBACKS_PIPE */
struct S179 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* start */
	T0* a2; /* tree */
	T0* a3; /* error */
	T0* a4; /* last */
};

/* Struct for type [detachable] XM_CALLBACKS_TO_TREE_FILTER */
struct S182 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* last_position_table */
	T0* a2; /* document */
	T0* a3; /* source_parser */
	T0* a4; /* current_element */
	T0* a5; /* namespace_cache */
	T0* a6; /* next */
};

/* Struct for type [detachable] ET_ECF_STATE */
struct S183 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target */
	T0* a2; /* ise_version */
};

/* Struct for type [detachable] ET_IDENTIFIER */
struct S184 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* seed */
	T6 a2; /* index */
	T0* a3; /* name */
	T2 a4; /* status_code */
	T6 a5; /* compressed_position */
	T6 a6; /* hash_code */
};

/* Struct for type [detachable] ET_ECF_SYSTEM_CONFIG */
struct S185 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* filename */
	T0* a3; /* universe */
	T0* a4; /* ecf_namespace */
	T0* a5; /* ecf_version */
	T0* a6; /* description */
	T0* a7; /* notes */
	T0* a8; /* targets */
	T1 a9; /* is_read_only */
	T0* a10; /* uuid */
	T0* a11; /* library_target */
};

/* Struct for type [detachable] ET_ECF_LIBRARY */
struct S187 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* filename */
	T0* a3; /* ecf_namespace */
	T0* a4; /* ecf_version */
	T0* a5; /* description */
	T0* a6; /* notes */
	T0* a7; /* targets */
	T1 a8; /* is_read_only */
	T0* a9; /* uuid */
	T0* a10; /* library_target */
	T0* a11; /* current_system */
	T0* a12; /* library */
	T0* a13; /* libraries */
	T0* a14; /* clusters */
	T0* a15; /* dotnet_assemblies */
	T0* a16; /* implicit_attachment_type_mark */
	T0* a17; /* master_classes */
	T0* a18; /* master_class_mutex */
	T0* a19; /* any_type */
	T0* a20; /* detachable_any_type */
	T0* a21; /* detachable_separate_any_type */
	T0* a22; /* any_parent */
	T0* a23; /* any_parents */
	T0* a24; /* any_clients */
	T0* a25; /* tuple_type */
	T0* a26; /* detachable_tuple_type */
	T0* a27; /* tuple_identity_type */
	T0* a28; /* unfolded_empty_tuple_actual_parameters */
	T0* a29; /* array_any_type */
	T0* a30; /* array_detachable_any_type */
	T0* a31; /* array_none_type */
	T0* a32; /* array_identity_type */
	T0* a33; /* boolean_type */
	T0* a34; /* character_type */
	T0* a35; /* character_8_type */
	T0* a36; /* character_8_convert_feature */
	T0* a37; /* character_32_type */
	T0* a38; /* character_32_convert_feature */
	T0* a39; /* double_type */
	T0* a40; /* exception_type */
	T0* a41; /* detachable_exception_type */
	T0* a42; /* exception_manager_type */
	T0* a43; /* function_identity_any_type */
	T0* a44; /* immutable_string_8_type */
	T0* a45; /* immutable_string_32_type */
	T0* a46; /* integer_type */
	T0* a47; /* integer_8_type */
	T0* a48; /* integer_8_convert_feature */
	T0* a49; /* integer_16_type */
	T0* a50; /* integer_16_convert_feature */
	T0* a51; /* integer_32_type */
	T0* a52; /* integer_32_convert_feature */
	T0* a53; /* integer_64_type */
	T0* a54; /* integer_64_convert_feature */
	T0* a55; /* ise_exception_manager_type */
	T0* a56; /* iterable_detachable_separate_any_type */
	T0* a57; /* natural_type */
	T0* a58; /* natural_8_type */
	T0* a59; /* natural_8_convert_feature */
	T0* a60; /* natural_16_type */
	T0* a61; /* natural_16_convert_feature */
	T0* a62; /* natural_32_type */
	T0* a63; /* natural_32_convert_feature */
	T0* a64; /* natural_64_type */
	T0* a65; /* natural_64_convert_feature */
	T0* a66; /* none_type */
	T0* a67; /* detachable_none_type */
	T0* a68; /* pointer_type */
	T0* a69; /* predicate_identity_type */
	T0* a70; /* procedure_identity_type */
	T0* a71; /* real_type */
	T0* a72; /* real_32_type */
	T0* a73; /* real_32_convert_feature */
	T0* a74; /* real_64_type */
	T0* a75; /* real_64_convert_feature */
	T0* a76; /* routine_identity_type */
	T0* a77; /* special_any_type */
	T0* a78; /* special_detachable_any_type */
	T0* a79; /* special_identity_type */
	T0* a80; /* string_type */
	T0* a81; /* detachable_string_type */
	T0* a82; /* string_8_type */
	T0* a83; /* detachable_string_8_type */
	T0* a84; /* string_8_convert_feature */
	T0* a85; /* string_32_type */
	T0* a86; /* string_32_convert_feature */
	T0* a87; /* system_object_type */
	T0* a88; /* system_object_parents */
	T0* a89; /* system_string_type */
	T0* a90; /* type_detachable_any_type */
	T0* a91; /* detachable_type_exception_type */
	T0* a92; /* type_detachable_exception_type */
	T0* a93; /* type_detachable_like_current_type */
	T0* a94; /* type_identity_type */
	T0* a95; /* typed_pointer_identity_type */
	T0* a96; /* wide_character_type */
	T0* a97; /* selected_target */
	T1 a98; /* obsolete_iteration_mode */
	T1 a99; /* obsolete_routine_type_mode */
	T1 a100; /* is_preparsed */
	T1 a101; /* use_obsolete_syntax_mode */
};

/* Struct for type [detachable] ET_ADAPTED_LIBRARIES */
struct S188 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* libraries */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ADAPTED_LIBRARY] */
struct S190 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ADAPTED_LIBRARY]] */
struct S191 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] XM_DOCUMENT */
struct S192 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* root_element */
	T0* a2; /* children */
};

/* Struct for type [detachable] ET_COMPRESSED_POSITION */
struct S193 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* compressed_position */
};

/* Struct for type [detachable] XM_STOP_ON_ERROR_FILTER */
struct S195 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_error */
	T0* a2; /* last_error */
	T0* a3; /* next */
};

/* Struct for type detachable XM_ATTRIBUTE */
struct S196 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* namespace */
	T0* a3; /* value */
	T0* a4; /* parent */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8] */
struct S198 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T6 a3; /* position */
	T6 a4; /* last_position */
	T6 a5; /* capacity */
	T6 a6; /* slots_position */
	T6 a7; /* count */
	T0* a8; /* key_equality_tester */
	T6 a9; /* modulus */
	T6 a10; /* clashes_previous_position */
	T0* a11; /* special_item_routines */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T0* a14; /* special_key_routines */
	T0* a15; /* key_storage */
	T0* a16; /* equality_tester */
	T6 a17; /* free_slot */
	T0* a18; /* internal_cursor */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8] */
struct S199 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* position */
	T0* a3; /* internal_cursor */
	T0* a4; /* key_storage */
	T0* a5; /* item_storage */
	T6 a6; /* last_position */
	T6 a7; /* capacity */
	T6 a8; /* slots_position */
	T6 a9; /* modulus */
	T6 a10; /* clashes_previous_position */
	T0* a11; /* key_equality_tester */
	T6 a12; /* found_position */
	T0* a13; /* special_item_routines */
	T0* a14; /* clashes */
	T0* a15; /* slots */
	T0* a16; /* special_key_routines */
	T0* a17; /* equality_tester */
	T6 a18; /* free_slot */
	T0* a19; /* hash_function */
};

/* Struct for type detachable ET_ECF_TARGET_PARENT */
struct S200 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name_id */
	T0* a2; /* location_id */
	T0* a3; /* target */
};

/* Struct for type [detachable] DS_CELL [detachable ET_ECF_SYSTEM_CONFIG] */
struct S201 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER, [attached] ET_ECF_INTERNAL_UNIVERSE, [attached] DS_CELL [detachable ET_ECF_SYSTEM_CONFIG]] */
struct S202 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
};

/* Struct for type detachable ET_ECF_TARGETS */
struct S203 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* targets */
};

/* Struct for type [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
struct S204 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* key_equality_tester */
	T0* a2; /* internal_keys */
	T0* a3; /* hash_function */
	T6 a4; /* position */
	T6 a5; /* last_position */
	T6 a6; /* capacity */
	T6 a7; /* slots_position */
	T6 a8; /* count */
	T0* a9; /* equality_tester */
	T6 a10; /* found_position */
	T6 a11; /* modulus */
	T6 a12; /* clashes_previous_position */
	T0* a13; /* special_item_routines */
	T0* a14; /* item_storage */
	T0* a15; /* clashes */
	T0* a16; /* slots */
	T0* a17; /* special_key_routines */
	T0* a18; /* key_storage */
	T6 a19; /* free_slot */
	T0* a20; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
struct S205 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* container */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] DS_CELL [detachable ET_ECF_LIBRARY] */
struct S206 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
};

/* Struct for type [detachable] ET_ECF_ADAPTED_LIBRARY */
struct S207 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name_id */
	T0* a2; /* filename_id */
	T0* a3; /* universe */
	T0* a4; /* target */
	T0* a5; /* description */
	T0* a6; /* classname_prefix */
	T1 a7; /* is_read_only */
	T1 a8; /* use_application_options */
	T0* a9; /* options */
	T0* a10; /* class_options */
	T0* a11; /* class_renamings */
	T0* a12; /* visible_classes */
	T0* a13; /* conditions */
	T0* a14; /* notes */
	T0* a15; /* name */
	T0* a16; /* library */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER, [attached] ET_ECF_ADAPTED_LIBRARY, [attached] DS_CELL [detachable ET_ECF_LIBRARY]] */
struct S208 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
};

/* Struct for type [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES */
struct S210 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* dotnet_assemblies */
};

/* Struct for type [detachable] ET_ECF_DOTNET_ASSEMBLY */
struct S211 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* pathname */
	T0* a2; /* referenced_assemblies */
	T0* a3; /* current_system */
	T0* a4; /* dotnet_assembly */
	T0* a5; /* name */
	T0* a6; /* implicit_attachment_type_mark */
	T0* a7; /* master_classes */
	T0* a8; /* master_class_mutex */
	T0* a9; /* any_type */
	T0* a10; /* detachable_any_type */
	T0* a11; /* detachable_separate_any_type */
	T0* a12; /* any_parent */
	T0* a13; /* any_parents */
	T0* a14; /* any_clients */
	T0* a15; /* tuple_type */
	T0* a16; /* detachable_tuple_type */
	T0* a17; /* tuple_identity_type */
	T0* a18; /* unfolded_empty_tuple_actual_parameters */
	T0* a19; /* array_any_type */
	T0* a20; /* array_detachable_any_type */
	T0* a21; /* array_none_type */
	T0* a22; /* array_identity_type */
	T0* a23; /* boolean_type */
	T0* a24; /* character_type */
	T0* a25; /* character_8_type */
	T0* a26; /* character_8_convert_feature */
	T0* a27; /* character_32_type */
	T0* a28; /* character_32_convert_feature */
	T0* a29; /* double_type */
	T0* a30; /* exception_type */
	T0* a31; /* detachable_exception_type */
	T0* a32; /* exception_manager_type */
	T0* a33; /* function_identity_any_type */
	T0* a34; /* immutable_string_8_type */
	T0* a35; /* immutable_string_32_type */
	T0* a36; /* integer_type */
	T0* a37; /* integer_8_type */
	T0* a38; /* integer_8_convert_feature */
	T0* a39; /* integer_16_type */
	T0* a40; /* integer_16_convert_feature */
	T0* a41; /* integer_32_type */
	T0* a42; /* integer_32_convert_feature */
	T0* a43; /* integer_64_type */
	T0* a44; /* integer_64_convert_feature */
	T0* a45; /* ise_exception_manager_type */
	T0* a46; /* iterable_detachable_separate_any_type */
	T0* a47; /* natural_type */
	T0* a48; /* natural_8_type */
	T0* a49; /* natural_8_convert_feature */
	T0* a50; /* natural_16_type */
	T0* a51; /* natural_16_convert_feature */
	T0* a52; /* natural_32_type */
	T0* a53; /* natural_32_convert_feature */
	T0* a54; /* natural_64_type */
	T0* a55; /* natural_64_convert_feature */
	T0* a56; /* none_type */
	T0* a57; /* detachable_none_type */
	T0* a58; /* pointer_type */
	T0* a59; /* predicate_identity_type */
	T0* a60; /* procedure_identity_type */
	T0* a61; /* real_type */
	T0* a62; /* real_32_type */
	T0* a63; /* real_32_convert_feature */
	T0* a64; /* real_64_type */
	T0* a65; /* real_64_convert_feature */
	T0* a66; /* routine_identity_type */
	T0* a67; /* special_any_type */
	T0* a68; /* special_detachable_any_type */
	T0* a69; /* special_identity_type */
	T0* a70; /* string_type */
	T0* a71; /* detachable_string_type */
	T0* a72; /* string_8_type */
	T0* a73; /* detachable_string_8_type */
	T0* a74; /* string_8_convert_feature */
	T0* a75; /* string_32_type */
	T0* a76; /* string_32_convert_feature */
	T0* a77; /* system_object_type */
	T0* a78; /* system_object_parents */
	T0* a79; /* system_string_type */
	T0* a80; /* type_detachable_any_type */
	T0* a81; /* detachable_type_exception_type */
	T0* a82; /* type_detachable_exception_type */
	T0* a83; /* type_detachable_like_current_type */
	T0* a84; /* type_identity_type */
	T0* a85; /* typed_pointer_identity_type */
	T0* a86; /* wide_character_type */
	T1 a87; /* is_read_only */
	T1 a88; /* is_preparsed */
};

/* Struct for type [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY */
struct S213 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name_id */
	T0* a2; /* filename_id */
	T0* a3; /* universe */
	T0* a4; /* target */
	T0* a5; /* description */
	T0* a6; /* assembly_culture */
	T0* a7; /* assembly_key */
	T0* a8; /* assembly_name */
	T0* a9; /* assembly_version */
	T0* a10; /* classname_prefix */
	T1 a11; /* is_read_only */
	T0* a12; /* options */
	T0* a13; /* class_options */
	T0* a14; /* class_renamings */
	T0* a15; /* conditions */
	T0* a16; /* notes */
	T0* a17; /* name */
	T0* a18; /* dotnet_assembly */
};

/* Struct for type [detachable] ET_ECF_ERROR */
struct S214 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* default_template */
	T0* a2; /* parameters */
	T0* a3; /* system_config */
	T0* a4; /* position */
	T0* a5; /* code */
};

/* Struct for type [detachable] KL_AGENT_HASH_FUNCTION [[attached] STRING_8] */
struct S215 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* hash_agent */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8] */
struct S216 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* internal_cursor */
	T6 a3; /* found_position */
	T0* a4; /* item_storage */
	T0* a5; /* key_storage */
	T0* a6; /* key_equality_tester */
	T0* a7; /* internal_keys */
	T0* a8; /* hash_function */
	T6 a9; /* position */
	T6 a10; /* last_position */
	T6 a11; /* capacity */
	T6 a12; /* slots_position */
	T6 a13; /* free_slot */
	T0* a14; /* equality_tester */
	T6 a15; /* modulus */
	T6 a16; /* clashes_previous_position */
	T0* a17; /* special_item_routines */
	T0* a18; /* clashes */
	T0* a19; /* slots */
	T0* a20; /* special_key_routines */
};

/* Struct for type [detachable] TUPLE [[attached] STRING_8] */
struct S217 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] FUNCTION [[attached] TUPLE [[attached] STRING_8], [attached] INTEGER_32] */
struct S218 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] KL_STRING_ROUTINES] */
struct S219 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8] */
struct S221 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* key_equality_tester */
	T0* a2; /* internal_keys */
	T0* a3; /* hash_function */
	T6 a4; /* position */
	T6 a5; /* last_position */
	T6 a6; /* capacity */
	T6 a7; /* slots_position */
	T6 a8; /* count */
	T0* a9; /* equality_tester */
	T6 a10; /* found_position */
	T6 a11; /* modulus */
	T6 a12; /* clashes_previous_position */
	T0* a13; /* special_item_routines */
	T0* a14; /* item_storage */
	T0* a15; /* clashes */
	T0* a16; /* slots */
	T0* a17; /* special_key_routines */
	T0* a18; /* key_storage */
	T6 a19; /* free_slot */
	T0* a20; /* internal_cursor */
};

/* Struct for type [detachable] ET_DYNAMIC_PRIMARY_TYPE */
struct S223 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* attribute_count */
	T0* a2; /* queries */
	T1 a3; /* is_alive */
	T6 a4; /* id */
	T0* a5; /* base_type */
	T0* a6; /* attached_type */
	T0* a7; /* base_class */
	T0* a8; /* meta_type */
	T0* a9; /* conforming_ancestors */
	T0* a10; /* queries_by_seed */
	T0* a11; /* procedures_by_seed */
	T0* a12; /* procedures */
	T1 a13; /* has_redefined_copy_routine */
	T1 a14; /* has_redefined_is_equal_routine */
	T1 a15; /* has_once_per_object_routines */
	T6 a16; /* hash_code */
	T0* a17; /* next_type */
	T1 a18; /* has_reference_attributes */
	T0* a19; /* query_calls */
	T1 a20; /* has_static */
	T0* a21; /* procedure_calls */
	T0* a22; /* equality_expressions */
	T0* a23; /* object_equality_expressions */
};

/* Struct for type [detachable] ET_DYNAMIC_FEATURE_LIST */
struct S224 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME] */
struct S225 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* key_equality_tester */
	T0* a2; /* internal_keys */
	T0* a3; /* equality_tester */
	T6 a4; /* capacity */
	T6 a5; /* modulus */
	T6 a6; /* last_position */
	T6 a7; /* free_slot */
	T6 a8; /* position */
	T0* a9; /* special_item_routines */
	T0* a10; /* item_storage */
	T0* a11; /* special_key_routines */
	T0* a12; /* key_storage */
	T0* a13; /* clashes */
	T0* a14; /* slots */
	T6 a15; /* found_position */
	T0* a16; /* internal_cursor */
	T6 a17; /* count */
	T6 a18; /* slots_position */
	T6 a19; /* clashes_previous_position */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8] */
struct S226 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T0* a3; /* key_equality_tester */
	T6 a4; /* position */
	T6 a5; /* last_position */
	T6 a6; /* capacity */
	T6 a7; /* slots_position */
	T6 a8; /* count */
	T0* a9; /* equality_tester */
	T6 a10; /* modulus */
	T6 a11; /* clashes_previous_position */
	T0* a12; /* special_item_routines */
	T0* a13; /* clashes */
	T0* a14; /* slots */
	T0* a15; /* special_key_routines */
	T0* a16; /* key_storage */
	T6 a17; /* free_slot */
	T0* a18; /* internal_cursor */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] TUPLE [[attached] ET_CLASS] */
struct S227 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]] */
struct S228 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [INTEGER_32] */
struct S229 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 z1;
};

/* Struct for type [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER */
struct S230 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* current_dynamic_system */
	T1 a3; /* catcall_error_mode */
	T1 a4; /* catcall_warning_mode */
	T1 a5; /* no_debug */
	T1 a6; /* no_assertion */
	T0* a7; /* alive_conforming_descendants_per_type */
};

/* Struct for type [detachable] TUPLE [[attached] ET_DYNAMIC_SYSTEM] */
struct S231 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type detachable ET_DYNAMIC_FEATURE */
struct S236 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* static_feature */
	T0* a2; /* dynamic_type_sets */
	T6 a3; /* id */
	T0* a4; /* target_type */
	T1 a5; /* is_generated */
	T0* a6; /* result_type_set */
	T8 a7; /* builtin_class_code */
	T8 a8; /* builtin_feature_code */
	T1 a9; /* is_is_equal_routine */
	T1 a10; /* is_copy_routine */
	T0* a11; /* query_calls */
	T0* a12; /* procedure_calls */
	T1 a13; /* is_regular */
	T0* a14; /* first_precursor */
	T1 a15; /* is_creation */
	T0* a16; /* other_precursors */
	T1 a17; /* is_built */
	T1 a18; /* is_address */
	T1 a19; /* is_static_generated */
};

/* Struct for type [detachable] ET_ACTUAL_PARAMETER_LIST */
struct S240 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* left_bracket */
	T0* a4; /* right_bracket */
};

/* Struct for type [detachable] ET_NESTED_TYPE_CONTEXT */
struct S241 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* root_context */
	T6 a2; /* count */
	T0* a3; /* storage */
};

/* Struct for type [detachable] ET_FORMAL_PARAMETER_TYPE */
struct S242 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type_mark */
	T0* a2; /* name */
	T6 a3; /* index */
	T0* a4; /* implementation_class */
};

/* Struct for type [detachable] ARRAY [[attached] ET_TYPE] */
struct S244 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_TYPE] */
struct S245 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] ET_EXTERNAL_FUNCTION */
struct S253 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* first_seed */
	T0* a2; /* iteration_components */
	T0* a3; /* implementation_class */
	T0* a4; /* implementation_feature */
	T0* a5; /* other_seeds */
	T0* a6; /* first_precursor */
	T0* a7; /* other_precursors */
	T0* a8; /* preconditions */
	T6 a9; /* hash_code */
	T8 a10; /* builtin_class_code */
	T8 a11; /* builtin_feature_code */
	T0* a12; /* arguments */
	T0* a13; /* postconditions */
	T0* a14; /* assigner */
	T0* a15; /* extended_name */
	T0* a16; /* declared_type */
	T0* a17; /* frozen_keyword */
	T0* a18; /* language */
	T0* a19; /* obsolete_message */
	T0* a20; /* alias_clause */
	T0* a21; /* clients */
	T0* a22; /* is_keyword */
	T0* a23; /* end_keyword */
	T0* a24; /* semicolon */
	T0* a25; /* feature_clause */
	T0* a26; /* first_indexing */
	T0* a27; /* object_tests */
	T1 a28; /* validity_checked */
	T1 a29; /* has_validity_error */
	T6 a30; /* id */
	T6 a31; /* version */
	T0* a32; /* synonym */
};

/* Struct for type [detachable] ET_TUPLE_TYPE */
struct S254 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* actual_parameters */
	T0* a2; /* type_mark */
	T0* a3; /* named_base_class */
	T0* a4; /* tuple_keyword */
};

/* Struct for type detachable ET_FORMAL_ARGUMENT_LIST */
struct S255 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* left_parenthesis */
	T0* a4; /* right_parenthesis */
};

/* Struct for type [detachable] ET_FORMAL_ARGUMENT */
struct S256 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name_item */
	T0* a2; /* declared_type */
	T6 a3; /* index */
	T6 a4; /* attached_index */
	T1 a5; /* is_used */
};

/* Struct for type [detachable] ET_QUERY_LIST */
struct S257 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* declared_count */
	T0* a3; /* storage */
};

/* Struct for type [detachable] ET_PROCEDURE_LIST */
struct S258 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* declared_count */
	T0* a3; /* storage */
};

/* Struct for type [detachable] ET_DYNAMIC_TYPE_SET_LIST */
struct S259 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] KL_STRING_OUTPUT_STREAM */
struct S260 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* string */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER] */
struct S261 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* a3; /* storage */
	T0* a4; /* special_routines */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE] */
struct S262 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32] */
struct S263 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* a3; /* storage */
	T0* a4; /* special_routines */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_UNQUALIFIED_CALL_EXPRESSION] */
struct S264 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_QUALIFIED_CALL_EXPRESSION] */
struct S265 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ACTUAL_ARGUMENT_LIST] */
struct S266 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_RESULT] */
struct S267 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST */
struct S268 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* clashes */
	T0* a4; /* slots */
};

/* Struct for type [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET */
struct S269 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* static_type */
	T6 a2; /* count */
	T0* a3; /* dynamic_types */
	T1 a4; /* is_never_void */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET] */
struct S270 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_STACK [[attached] ET_EXPRESSION] */
struct S271 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION] */
struct S272 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* capacity */
	T6 a2; /* count */
	T0* a3; /* storage */
	T0* a4; /* special_routines */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32] */
struct S273 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* item_storage */
	T6 a3; /* last_position */
	T6 a4; /* capacity */
	T6 a5; /* count */
	T6 a6; /* free_slot */
	T6 a7; /* slots_position */
	T6 a8; /* modulus */
	T6 a9; /* clashes_previous_position */
	T0* a10; /* key_equality_tester */
	T0* a11; /* equality_tester */
	T6 a12; /* found_position */
	T0* a13; /* clashes */
	T0* a14; /* slots */
	T0* a15; /* special_item_routines */
	T0* a16; /* special_key_routines */
	T0* a17; /* key_storage */
	T0* a18; /* internal_cursor */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST */
struct S274 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S275 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T6 a2; /* count */
	T6 a3; /* position */
	T0* a4; /* item_storage */
	T6 a5; /* capacity */
	T6 a6; /* modulus */
	T6 a7; /* last_position */
	T6 a8; /* free_slot */
	T6 a9; /* slots_position */
	T6 a10; /* clashes_previous_position */
	T0* a11; /* special_item_routines */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T6 a14; /* found_position */
	T0* a15; /* hash_function */
	T0* a16; /* equality_tester */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S276 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* internal_cursor */
	T6 a3; /* found_position */
	T0* a4; /* item_storage */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T6 a8; /* capacity */
	T0* a9; /* equality_tester */
	T0* a10; /* key_equality_tester */
	T0* a11; /* key_storage */
	T0* a12; /* clashes */
	T6 a13; /* modulus */
	T0* a14; /* slots */
	T6 a15; /* slots_position */
	T6 a16; /* clashes_previous_position */
	T0* a17; /* special_item_routines */
	T0* a18; /* special_key_routines */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST] */
struct S277 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
struct S278 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION] */
struct S279 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT] */
struct S280 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] ET_CURRENT */
struct S282 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T2 a2; /* code */
	T0* a3; /* text */
	T6 a4; /* compressed_position */
};

/* Struct for type [detachable] ET_ACTUAL_ARGUMENT_LIST */
struct S283 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* left_symbol */
	T0* a4; /* right_symbol */
};

/* Struct for type [detachable] ET_QUALIFIED_CALL_INSTRUCTION */
struct S284 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target */
	T0* a2; /* qualified_name */
	T0* a3; /* arguments */
	T0* a4; /* parenthesis_call */
};

/* Struct for type [detachable] ET_QUALIFIED_CALL_EXPRESSION */
struct S287 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target */
	T6 a2; /* index */
	T0* a3; /* arguments */
	T0* a4; /* qualified_name */
	T0* a5; /* parenthesis_call */
};

/* Struct for type [detachable] ET_MANIFEST_TUPLE */
struct S288 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* left_symbol */
	T0* a4; /* right_symbol */
	T6 a5; /* index */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE] */
struct S289 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T0* a2; /* item_storage */
	T6 a3; /* capacity */
	T6 a4; /* modulus */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T6 a8; /* count */
	T0* a9; /* special_item_routines */
	T0* a10; /* clashes */
	T0* a11; /* slots */
	T6 a12; /* found_position */
	T6 a13; /* slots_position */
	T6 a14; /* clashes_previous_position */
	T0* a15; /* hash_function */
	T0* a16; /* equality_tester */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE] */
struct S290 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T6 a2; /* found_position */
	T0* a3; /* item_storage */
	T6 a4; /* count */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* capacity */
	T6 a8; /* slots_position */
	T6 a9; /* modulus */
	T6 a10; /* clashes_previous_position */
	T0* a11; /* key_equality_tester */
	T0* a12; /* equality_tester */
	T0* a13; /* internal_cursor */
	T0* a14; /* key_storage */
	T0* a15; /* clashes */
	T0* a16; /* slots */
	T0* a17; /* special_item_routines */
	T0* a18; /* special_key_routines */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] ARRAY [[attached] INTEGER_32] */
struct S291 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE] */
struct S292 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T0* a2; /* key_storage */
	T0* a3; /* item_storage */
	T6 a4; /* count */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T6 a8; /* capacity */
	T6 a9; /* slots_position */
	T0* a10; /* equality_tester */
	T0* a11; /* key_equality_tester */
	T6 a12; /* found_position */
	T0* a13; /* clashes */
	T6 a14; /* modulus */
	T0* a15; /* slots */
	T6 a16; /* clashes_previous_position */
	T0* a17; /* special_item_routines */
	T0* a18; /* special_key_routines */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT] */
struct S293 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T0* a2; /* key_storage */
	T0* a3; /* item_storage */
	T6 a4; /* count */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T0* a8; /* equality_tester */
	T0* a9; /* key_equality_tester */
	T6 a10; /* found_position */
	T0* a11; /* clashes */
	T6 a12; /* modulus */
	T0* a13; /* slots */
	T6 a14; /* capacity */
	T0* a15; /* special_item_routines */
	T0* a16; /* special_key_routines */
	T6 a17; /* slots_position */
	T6 a18; /* clashes_previous_position */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S294 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T6 a3; /* count */
	T6 a4; /* last_position */
	T6 a5; /* free_slot */
	T6 a6; /* position */
	T6 a7; /* capacity */
	T0* a8; /* equality_tester */
	T0* a9; /* key_equality_tester */
	T0* a10; /* internal_cursor */
	T0* a11; /* key_storage */
	T0* a12; /* clashes */
	T6 a13; /* modulus */
	T0* a14; /* slots */
	T6 a15; /* slots_position */
	T6 a16; /* clashes_previous_position */
	T0* a17; /* special_item_routines */
	T0* a18; /* special_key_routines */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE] */
struct S295 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8] */
struct S296 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T6 a2; /* position */
	T0* a3; /* item_storage */
	T0* a4; /* key_storage */
	T0* a5; /* key_equality_tester */
	T0* a6; /* internal_keys */
	T6 a7; /* count */
	T6 a8; /* last_position */
	T6 a9; /* free_slot */
	T6 a10; /* capacity */
	T6 a11; /* slots_position */
	T6 a12; /* modulus */
	T6 a13; /* clashes_previous_position */
	T0* a14; /* equality_tester */
	T6 a15; /* found_position */
	T0* a16; /* clashes */
	T0* a17; /* slots */
	T0* a18; /* special_item_routines */
	T0* a19; /* special_key_routines */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER] */
struct S297 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T6 a2; /* capacity */
	T6 a3; /* modulus */
	T6 a4; /* last_position */
	T6 a5; /* free_slot */
	T0* a6; /* equality_tester */
	T6 a7; /* slots_position */
	T6 a8; /* count */
	T6 a9; /* clashes_previous_position */
	T0* a10; /* special_item_routines */
	T0* a11; /* item_storage */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T6 a14; /* found_position */
	T0* a15; /* internal_cursor */
	T0* a16; /* hash_function */
};

/* Struct for type [detachable] ET_IDENTIFIER_TESTER */
struct S298 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_TEXT_OUTPUT_FILE */
struct S300 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T6 a2; /* mode */
	T14 a3; /* file_pointer */
	T1 a4; /* descriptor_available */
	T0* a5; /* last_string_32 */
	T0* a6; /* last_string */
	T0* a7; /* internal_name */
	T0* a8; /* internal_detachable_name_pointer */
};

/* Struct for type [detachable] ET_IMPLICIT_TYPE_MARK */
struct S301 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* is_expanded_mark */
	T1 a2; /* is_attached_mark */
	T1 a3; /* is_reference_mark */
	T1 a4; /* is_separate_mark */
	T1 a5; /* is_detachable_mark */
};

/* Struct for type [detachable] ET_DYNAMIC_TUPLE_TYPE */
struct S302 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* id */
	T0* a2; /* base_class */
	T0* a3; /* base_type */
	T1 a4; /* is_alive */
	T0* a5; /* attached_type */
	T6 a6; /* attribute_count */
	T0* a7; /* queries */
	T0* a8; /* meta_type */
	T0* a9; /* conforming_ancestors */
	T0* a10; /* queries_by_seed */
	T0* a11; /* procedures_by_seed */
	T0* a12; /* procedures */
	T1 a13; /* has_redefined_copy_routine */
	T0* a14; /* item_type_sets */
	T1 a15; /* has_redefined_is_equal_routine */
	T1 a16; /* has_once_per_object_routines */
	T1 a17; /* has_reference_attributes */
	T6 a18; /* hash_code */
	T0* a19; /* next_type */
	T0* a20; /* query_calls */
	T1 a21; /* has_static */
	T0* a22; /* procedure_calls */
	T0* a23; /* equality_expressions */
	T0* a24; /* object_equality_expressions */
};

/* Struct for type [detachable] DS_STRING_HASH_TABLE */
struct S303 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T6 a2; /* position */
	T0* a3; /* item_storage */
	T6 a4; /* modulus */
	T0* a5; /* key_equality_tester */
	T0* a6; /* slots */
	T0* a7; /* hash_function */
	T0* a8; /* key_storage */
	T0* a9; /* clashes */
	T6 a10; /* count */
	T6 a11; /* capacity */
	T6 a12; /* slots_position */
	T6 a13; /* free_slot */
	T6 a14; /* last_position */
	T0* a15; /* internal_keys */
	T6 a16; /* clashes_previous_position */
	T0* a17; /* equality_tester */
	T0* a18; /* special_item_routines */
	T0* a19; /* special_key_routines */
	T0* a20; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8] */
struct S304 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] UT_CANNOT_WRITE_TO_FILE_ERROR */
struct S305 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type detachable ET_DYNAMIC_PRECURSOR */
struct S306 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* is_static_generated */
	T0* a2; /* target_type */
	T0* a3; /* static_feature */
	T0* a4; /* dynamic_type_sets */
	T0* a5; /* current_feature */
	T6 a6; /* id */
	T0* a7; /* result_type_set */
	T1 a8; /* is_regular */
	T1 a9; /* is_creation */
	T1 a10; /* is_address */
	T8 a11; /* builtin_class_code */
	T0* a12; /* first_precursor */
	T0* a13; /* other_precursors */
	T1 a14; /* is_built */
	T8 a15; /* builtin_feature_code */
	T1 a16; /* is_is_equal_routine */
	T1 a17; /* is_generated */
	T1 a18; /* is_copy_routine */
	T0* a19; /* parent_type */
	T0* a20; /* query_calls */
	T0* a21; /* procedure_calls */
};

/* Struct for type detachable ET_DYNAMIC_PRECURSOR_LIST */
struct S307 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S308 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* sorted_items */
	T0* a2; /* cycle */
	T0* a3; /* indexes */
	T0* a4; /* items */
	T0* a5; /* counts */
	T0* a6; /* successors */
};

/* Struct for type [detachable] ET_RESULT */
struct S309 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T2 a2; /* code */
	T0* a3; /* text */
	T6 a4; /* compressed_position */
};

/* Struct for type [detachable] ET_OBJECT_TEST */
struct S311 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* attached_keyword */
	T0* a3; /* declared_type */
	T0* a4; /* expression */
};

/* Struct for type [detachable] ET_OBJECT_EQUALITY_EXPRESSION */
struct S312 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* left */
	T0* a3; /* operator */
	T0* a4; /* right */
};

/* Struct for type [detachable] ET_EQUALITY_EXPRESSION */
struct S313 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* left */
	T0* a3; /* operator */
	T0* a4; /* right */
};

/* Struct for type detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL */
struct S314 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* current_feature */
	T0* a2; /* static_call */
	T0* a3; /* target_type_set */
	T0* a4; /* result_type_set */
	T0* a5; /* current_type */
	T0* a6; /* next */
	T1 a7; /* is_generated */
	T1 a8; /* force_result_boxing */
};

/* Struct for type detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL */
struct S318 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target_type_set */
	T1 a2; /* is_generated */
	T0* a3; /* next */
	T0* a4; /* static_call */
	T0* a5; /* current_feature */
	T0* a6; /* current_type */
};

/* Struct for type [detachable] ET_DYNAMIC_SPECIAL_TYPE */
struct S319 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* id */
	T0* a2; /* base_class */
	T0* a3; /* base_type */
	T1 a4; /* is_alive */
	T0* a5; /* attached_type */
	T6 a6; /* attribute_count */
	T0* a7; /* queries */
	T0* a8; /* meta_type */
	T0* a9; /* conforming_ancestors */
	T0* a10; /* queries_by_seed */
	T0* a11; /* procedures_by_seed */
	T0* a12; /* procedures */
	T0* a13; /* item_type_set */
	T1 a14; /* has_redefined_copy_routine */
	T1 a15; /* has_redefined_is_equal_routine */
	T1 a16; /* has_once_per_object_routines */
	T1 a17; /* has_reference_attributes */
	T6 a18; /* hash_code */
	T0* a19; /* next_type */
	T0* a20; /* query_calls */
	T1 a21; /* has_static */
	T0* a22; /* procedure_calls */
	T0* a23; /* equality_expressions */
	T0* a24; /* object_equality_expressions */
};

/* Struct for type [detachable] DS_QUICK_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S322 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* comparator */
};

/* Struct for type detachable ET_DYNAMIC_SECONDARY_TYPE */
struct S323 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* primary_type */
	T0* a2; /* meta_type */
	T0* a3; /* type_mark */
	T0* a4; /* internal_base_type */
};

/* Struct for type [detachable] ET_DYNAMIC_PRIMARY_TYPE_COMPARATOR_BY_ID */
struct S324 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_CREATE_EXPRESSION */
struct S328 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* creation_call */
	T0* a3; /* creation_type */
	T0* a4; /* create_keyword */
	T0* a5; /* creation_region */
};

/* Struct for type [detachable] ET_QUALIFIED_CALL */
struct S329 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* arguments */
	T0* a2; /* qualified_name */
};

/* Struct for type [detachable] KL_DIRECTORY */
struct S333 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* end_of_input */
	T0* a2; /* last_entry */
	T6 a3; /* mode */
	T0* a4; /* name */
	T0* a5; /* string_name */
	T0* a6; /* entry_buffer */
	T1 a7; /* old_end_of_input */
	T14 a8; /* directory_pointer */
	T14 a9; /* last_entry_pointer */
	T0* a10; /* lastentry */
	T0* a11; /* internal_name */
	T0* a12; /* internal_detachable_name_pointer */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S334 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] ET_SYMBOL */
struct S341 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 a1; /* code */
	T6 a2; /* compressed_position */
	T0* a3; /* break */
};

/* Struct for type [detachable] DS_QUICK_SORTER [[attached] INTEGER_32] */
struct S344 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* comparator */
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE] */
struct S346 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_IDENTIFIER, [attached] ET_CURRENT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
struct S347 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
	T0* z4;
	T1 z5;
};

/* Struct for type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_IDENTIFIER, [attached] ET_RESULT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
struct S348 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
	T0* z4;
	T1 z5;
};

/* Struct for type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_DYNAMIC_FEATURE, [attached] ET_RESULT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
struct S349 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
	T0* z4;
	T1 z5;
};

/* Struct for type [detachable] TUPLE [[attached] ET_C_GENERATOR, INTEGER_32, [attached] ET_RESULT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
struct S350 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T6 z2;
	T0* z3;
	T0* z4;
	T1 z5;
};

/* Struct for type [detachable] UT_INTEGER_FORMATTER */
struct S351 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_COMPOUND */
struct S355 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* keyword */
};

/* Struct for type [detachable] TUPLE [[attached] ET_C_GENERATOR, INTEGER_32, [attached] ET_EXPRESSION, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
struct S357 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T6 z2;
	T0* z3;
	T0* z4;
	T1 z5;
};

/* Struct for type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_EXPRESSION, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S358 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
};

/* Struct for type [detachable] ET_EXTENDED_ATTRIBUTE */
struct S359 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* first_seed */
	T0* a2; /* iteration_components */
	T0* a3; /* implementation_class */
	T0* a4; /* implementation_feature */
	T0* a5; /* other_seeds */
	T0* a6; /* first_precursor */
	T0* a7; /* other_precursors */
	T0* a8; /* preconditions */
	T6 a9; /* hash_code */
	T0* a10; /* postconditions */
	T0* a11; /* assigner */
	T0* a12; /* locals */
	T0* a13; /* rescue_clause */
	T0* a14; /* compound */
	T0* a15; /* extended_name */
	T0* a16; /* declared_type */
	T0* a17; /* frozen_keyword */
	T0* a18; /* obsolete_message */
	T0* a19; /* clients */
	T0* a20; /* end_keyword */
	T0* a21; /* semicolon */
	T0* a22; /* feature_clause */
	T0* a23; /* first_indexing */
	T0* a24; /* object_tests */
	T1 a25; /* validity_checked */
	T1 a26; /* has_validity_error */
	T6 a27; /* id */
	T6 a28; /* version */
	T0* a29; /* synonym */
};

/* Struct for type [detachable] ET_ATTRIBUTE */
struct S360 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* first_seed */
	T0* a2; /* iteration_components */
	T0* a3; /* implementation_class */
	T0* a4; /* implementation_feature */
	T0* a5; /* other_seeds */
	T0* a6; /* first_precursor */
	T0* a7; /* other_precursors */
	T6 a8; /* hash_code */
	T0* a9; /* assigner */
	T0* a10; /* extended_name */
	T0* a11; /* declared_type */
	T0* a12; /* frozen_keyword */
	T0* a13; /* clients */
	T0* a14; /* semicolon */
	T0* a15; /* feature_clause */
	T0* a16; /* first_indexing */
	T0* a17; /* object_tests */
	T1 a18; /* validity_checked */
	T1 a19; /* has_validity_error */
	T6 a20; /* id */
	T6 a21; /* version */
	T0* a22; /* synonym */
};

/* Struct for type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
struct S361 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
	T1 z4;
};

/* Struct for type detachable ET_ITERATION_COMPONENT_LIST */
struct S365 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_DYNAMIC_FEATURE, [attached] ET_EXPRESSION, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
struct S367 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
	T0* z4;
	T1 z5;
};

/* Struct for type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_DYNAMIC_FEATURE, INTEGER_32, INTEGER_32] */
struct S368 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T6 z3;
	T6 z4;
};

/* Struct for type [detachable] ET_CONSTANT_ATTRIBUTE */
struct S369 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* first_seed */
	T0* a2; /* iteration_components */
	T0* a3; /* implementation_class */
	T0* a4; /* constant */
	T0* a5; /* implementation_feature */
	T0* a6; /* other_seeds */
	T0* a7; /* first_precursor */
	T0* a8; /* other_precursors */
	T6 a9; /* hash_code */
	T0* a10; /* assigner */
	T0* a11; /* extended_name */
	T0* a12; /* declared_type */
	T0* a13; /* frozen_keyword */
	T0* a14; /* clients */
	T0* a15; /* is_keyword */
	T0* a16; /* semicolon */
	T0* a17; /* feature_clause */
	T0* a18; /* first_indexing */
	T0* a19; /* object_tests */
	T1 a20; /* validity_checked */
	T1 a21; /* has_validity_error */
	T6 a22; /* id */
	T6 a23; /* version */
	T0* a24; /* synonym */
};

/* Struct for type [detachable] ET_UNIQUE_ATTRIBUTE */
struct S371 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* constant */
	T6 a2; /* id */
	T6 a3; /* first_seed */
	T0* a4; /* iteration_components */
	T0* a5; /* implementation_class */
	T0* a6; /* implementation_feature */
	T0* a7; /* other_seeds */
	T0* a8; /* first_precursor */
	T0* a9; /* other_precursors */
	T6 a10; /* hash_code */
	T0* a11; /* assigner */
	T0* a12; /* extended_name */
	T0* a13; /* declared_type */
	T0* a14; /* frozen_keyword */
	T0* a15; /* clients */
	T0* a16; /* is_keyword */
	T0* a17; /* unique_keyword */
	T0* a18; /* semicolon */
	T0* a19; /* feature_clause */
	T0* a20; /* first_indexing */
	T0* a21; /* object_tests */
	T1 a22; /* validity_checked */
	T1 a23; /* has_validity_error */
	T6 a24; /* version */
	T0* a25; /* synonym */
};

/* Struct for type [detachable] ET_REGULAR_INTEGER_CONSTANT */
struct S372 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T11 a1; /* value */
	T0* a2; /* sign */
	T0* a3; /* literal */
	T1 a4; /* has_overflow */
	T6 a5; /* compressed_position */
	T6 a6; /* index */
	T0* a7; /* cast_type */
	T0* a8; /* type */
};

/* Struct for type [detachable] ET_SYSTEM_MULTIPROCESSOR */
struct S375 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* interface_checker */
	T0* a2; /* stop_request */
	T0* a3; /* eiffel_parser */
	T0* a4; /* error_handler */
	T0* a5; /* implementation_checker */
	T1 a6; /* benchmark_shown */
	T0* a7; /* ise_version */
	T0* a8; /* other_processors */
	T0* a9; /* eiffel_preparser */
	T0* a10; /* master_class_checker */
	T0* a11; /* provider_checker */
	T0* a12; /* ancestor_builder */
	T0* a13; /* feature_flattener */
	T0* a14; /* ecma_version */
	T1 a15; /* nested_benchmark_shown */
	T1 a16; /* metrics_shown */
	T1 a17; /* flat_mode */
	T1 a18; /* flat_dbc_mode */
	T0* a19; /* processed_class_count_stack */
	T0* a20; /* ast_factory */
	T0* a21; /* dotnet_assembly_consumer */
	T1 a22; /* use_attribute_keyword */
	T1 a23; /* use_note_keyword */
	T1 a24; /* use_reference_keyword */
	T1 a25; /* use_attached_keyword */
	T1 a26; /* use_detachable_keyword */
	T1 a27; /* preparse_shallow_mode */
	T1 a28; /* preparse_single_mode */
	T1 a29; /* preparse_multiple_mode */
	T1 a30; /* unknown_builtin_reported */
	T1 a31; /* qualified_anchored_types_cycle_detection_enabled */
	T1 a32; /* cluster_dependence_enabled */
	T6 a33; /* postponed_class_count */
	T6 a34; /* processed_class_count */
	T1 a35; /* suppliers_enabled */
	T1 a36; /* preparse_override_mode */
	T1 a37; /* preparse_readonly_mode */
	T1 a38; /* providers_enabled */
	T1 a39; /* use_cluster_dependence_pathnames */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8] */
struct S376 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_8] */
struct S377 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] INTEGER_32] */
struct S379 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_FEATURE_CHECKER */
struct S380 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type_checker */
	T0* a2; /* current_type */
	T0* a3; /* current_class */
	T0* a4; /* current_feature */
	T0* a5; /* current_feature_impl */
	T0* a6; /* current_class_impl */
	T0* a7; /* enclosing_inline_agents */
	T0* a8; /* unused_overloaded_procedures_list */
	T0* a9; /* unused_overloaded_queries_list */
	T0* a10; /* unused_overloaded_features_list */
	T0* a11; /* unused_call_infos */
	T0* a12; /* unused_contexts */
	T0* a13; /* current_context */
	T0* a14; /* current_target_type */
	T0* a15; /* current_object_test_types */
	T0* a16; /* current_object_test_scope */
	T0* a17; /* object_test_scope_builder */
	T0* a18; /* current_iteration_cursor_types */
	T0* a19; /* current_iteration_item_types */
	T0* a20; /* current_iteration_item_scope */
	T0* a21; /* current_initialization_scope */
	T0* a22; /* current_attachment_scope */
	T0* a23; /* attachment_scope_builder */
	T0* a24; /* unused_attachment_scopes */
	T0* a25; /* assertions_by_feature */
	T0* a26; /* common_ancestor_type_list */
	T0* a27; /* indexing_term_list */
	T0* a28; /* default_creation_call_name */
	T0* a29; /* default_creation_call */
	T0* a30; /* vape_non_descendant_clients */
	T0* a31; /* vape_creation_clients */
	T0* a32; /* vape_client */
	T0* a33; /* unused_adapted_base_classes */
	T0* a34; /* adapted_base_class_checker */
	T0* a35; /* system_processor */
	T1 a36; /* has_fatal_error */
	T1 a37; /* in_static_feature */
	T1 a38; /* in_precondition */
	T0* a39; /* supplier_handler */
	T1 a40; /* in_invariant */
	T0* a41; /* precursor_queries */
	T0* a42; /* precursor_procedures */
	T1 a43; /* in_postcondition */
	T1 a44; /* in_precursor */
	T1 a45; /* in_rescue */
	T0* a46; /* current_inline_agent */
	T1 a47; /* in_check_instruction */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE] */
struct S381 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* last_position */
	T6 a3; /* free_slot */
	T6 a4; /* position */
	T0* a5; /* equality_tester */
	T0* a6; /* key_equality_tester */
	T0* a7; /* internal_cursor */
	T6 a8; /* found_position */
	T0* a9; /* item_storage */
	T0* a10; /* key_storage */
	T0* a11; /* clashes */
	T6 a12; /* modulus */
	T0* a13; /* slots */
	T6 a14; /* capacity */
	T0* a15; /* special_item_routines */
	T0* a16; /* special_key_routines */
	T6 a17; /* slots_position */
	T6 a18; /* clashes_previous_position */
	T0* a19; /* hash_function */
};

/* Struct for type detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE] */
struct S382 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T6 a2; /* found_position */
	T0* a3; /* item_storage */
	T6 a4; /* position */
	T6 a5; /* last_position */
	T6 a6; /* capacity */
	T6 a7; /* count */
	T6 a8; /* modulus */
	T6 a9; /* slots_position */
	T6 a10; /* clashes_previous_position */
	T0* a11; /* key_equality_tester */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T0* a14; /* special_item_routines */
	T0* a15; /* special_key_routines */
	T0* a16; /* key_storage */
	T0* a17; /* equality_tester */
	T6 a18; /* free_slot */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] ET_TYPE_CHECKER */
struct S384 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* current_context */
	T0* a2; /* current_class_impl */
	T0* a3; /* current_feature_impl */
	T0* a4; /* constraint_context */
	T0* a5; /* current_class */
	T0* a6; /* target_context */
	T0* a7; /* other_context */
	T0* a8; /* adapted_base_class_checker */
	T0* a9; /* adapted_base_classes */
	T0* a10; /* system_processor */
	T0* a11; /* supplier_handler */
	T1 a12; /* has_fatal_error */
	T1 a13; /* class_interface_error_only */
	T1 a14; /* in_qualified_anchored_type */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT] */
struct S386 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
struct S387 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
struct S388 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
struct S389 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
struct S390 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT] */
struct S391 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST] */
struct S392 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T0* a2; /* item_storage */
	T6 a3; /* count */
	T6 a4; /* last_position */
	T6 a5; /* free_slot */
	T6 a6; /* position */
	T0* a7; /* equality_tester */
	T0* a8; /* key_equality_tester */
	T6 a9; /* found_position */
	T0* a10; /* key_storage */
	T0* a11; /* clashes */
	T6 a12; /* modulus */
	T0* a13; /* slots */
	T6 a14; /* capacity */
	T0* a15; /* special_item_routines */
	T0* a16; /* special_key_routines */
	T6 a17; /* slots_position */
	T6 a18; /* clashes_previous_position */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] ET_OBJECT_TEST_SCOPE */
struct S393 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* object_tests */
	T6 a2; /* hidden_count */
};

/* Struct for type [detachable] ET_OBJECT_TEST_SCOPE_BUILDER */
struct S394 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* current_class */
	T0* a3; /* system_processor */
	T0* a4; /* scope */
	T1 a5; /* is_negated */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT] */
struct S395 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T0* a2; /* item_storage */
	T0* a3; /* key_storage */
	T6 a4; /* count */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T0* a8; /* equality_tester */
	T0* a9; /* key_equality_tester */
	T6 a10; /* found_position */
	T0* a11; /* clashes */
	T6 a12; /* modulus */
	T0* a13; /* slots */
	T6 a14; /* capacity */
	T0* a15; /* special_item_routines */
	T0* a16; /* special_key_routines */
	T6 a17; /* slots_position */
	T6 a18; /* clashes_previous_position */
	T0* a19; /* internal_keys */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] ET_ITERATION_ITEM_SCOPE */
struct S396 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* hidden_count */
	T0* a2; /* iteration_components */
};

/* Struct for type [detachable] ET_ATTACHMENT_SCOPE */
struct S397 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* locals_attached */
	T0* a2; /* arguments_attached */
	T0* a3; /* attributes_attached */
	T1 a4; /* result_attached */
	T1 a5; /* is_code_unreachable */
};

/* Struct for type [detachable] ET_ATTACHMENT_SCOPE_BUILDER */
struct S398 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* scope */
	T1 a2; /* is_negated */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE] */
struct S399 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE] */
struct S400 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* internal_cursor */
	T0* a3; /* clashes */
	T0* a4; /* item_storage */
	T6 a5; /* last_position */
	T6 a6; /* position */
	T6 a7; /* capacity */
	T6 a8; /* slots_position */
	T6 a9; /* free_slot */
	T0* a10; /* equality_tester */
	T0* a11; /* key_equality_tester */
	T6 a12; /* found_position */
	T6 a13; /* modulus */
	T6 a14; /* clashes_previous_position */
	T0* a15; /* special_item_routines */
	T0* a16; /* slots */
	T0* a17; /* special_key_routines */
	T0* a18; /* key_storage */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM] */
struct S401 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] ET_CLIENT_LIST */
struct S402 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] ET_CLIENT */
struct S403 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* named_base_class */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
struct S404 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] ET_ADAPTED_BASE_CLASS_CHECKER */
struct S405 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* current_class_impl */
	T0* a2; /* current_class */
	T0* a3; /* system_processor */
	T1 a4; /* feature_flattening_error_only */
	T1 a5; /* has_fatal_error */
	T1 a6; /* class_interface_error_only */
};

/* Struct for type detachable ET_PRECONDITIONS */
struct S406 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* validity_checked */
	T1 a2; /* has_validity_error */
	T6 a3; /* count */
	T0* a4; /* storage */
	T0* a5; /* require_keyword */
	T0* a6; /* else_keyword */
};

/* Struct for type detachable ET_FEATURE_LIST */
struct S408 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* storage */
	T6 a2; /* count */
};

/* Struct for type [detachable] ET_UNKNOWN_GROUP */
struct S410 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T1 a2; /* use_obsolete_syntax */
	T0* a3; /* cached_absolute_pathname */
};

/* Struct for type [detachable] ET_BASE_TYPE_LIST */
struct S412 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] ET_KEYWORD */
struct S413 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 a1; /* code */
	T0* a2; /* text */
	T6 a3; /* compressed_position */
};

/* Struct for type [detachable] ET_CLASS_CODES */
struct S414 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S415 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S416 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S417 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] RX_BYTE_CODE */
struct S422 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* byte_code */
	T0* a3; /* character_sets */
	T6 a4; /* capacity */
};

/* Struct for type [detachable] RX_CHARACTER_SET */
struct S423 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* is_empty */
	T1 a2; /* is_negated */
	T11 a3; /* first_set */
	T11 a4; /* second_set */
	T11 a5; /* third_set */
	T11 a6; /* fourth_set */
	T0* a7; /* other_sets */
	T1 a8; /* is_reverted */
};

/* Struct for type [detachable] RX_CASE_MAPPING */
struct S424 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* lower_table */
	T0* a2; /* flip_table */
};

/* Struct for type [detachable] UC_UNICODE_ROUTINES */
struct S426 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ARRAY [[attached] RX_CHARACTER_SET] */
struct S427 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] SPECIAL [[attached] RX_CHARACTER_SET] */
struct S428 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] UC_STRING */
struct S429 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* area */
	T6 a3; /* byte_count */
	T6 a4; /* internal_hash_code */
	T6 a5; /* last_byte_index_input */
	T6 a6; /* last_byte_index_result */
	T6 a7; /* internal_case_insensitive_hash_code */
};

/* Struct for type [detachable] NATIVE_STRING */
struct S430 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* managed_data */
	T6 a2; /* unit_count */
};

/* Struct for type [detachable] TUPLE [[attached] GECC, [attached] DS_ARRAYED_LIST [[attached] STRING_8], INTEGER_32, [attached] MUTEX] */
struct S431 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T6 z3;
	T0* z4;
};

/* Struct for type [detachable] WORKER_THREAD */
struct S432 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* thread_procedure */
	T0* a2; /* launch_mutex */
	T14 a3; /* thread_id */
	T1 a4; /* terminated */
};

/* Struct for type [detachable] UC_UTF8_ROUTINES */
struct S434 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] PATH */
struct S435 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_name */
	T0* a2; /* storage */
	T1 a3; /* is_normalized */
};

/* Struct for type [detachable] STRING_TO_INTEGER_CONVERTOR */
struct S436 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* sign */
	T11 a2; /* part1 */
	T11 a3; /* part2 */
	T6 a4; /* last_state */
	T1 a5; /* internal_overflowed */
	T0* a6; /* leading_separators */
	T0* a7; /* trailing_separators */
	T1 a8; /* leading_separators_acceptable */
	T1 a9; /* trailing_separators_acceptable */
	T6 a10; /* conversion_type */
};

/* Struct for type [detachable] ET_ECF_OPTIONS */
struct S444 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* primary_assertions */
	T0* a2; /* primary_debugs */
	T0* a3; /* primary_warnings */
	T0* a4; /* primary_options */
	T0* a5; /* description */
	T0* a6; /* secondary_options */
};

/* Struct for type detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT] */
struct S445 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] XM_NAMESPACE */
struct S446 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* uri */
	T0* a2; /* ns_prefix */
};

/* Struct for type [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE] */
struct S447 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* after */
	T0* a2; /* current_cell */
	T0* a3; /* container */
	T1 a4; /* before */
	T0* a5; /* next_cursor */
};

/* Struct for type [detachable] ET_LIKE_CURRENT */
struct S449 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type_mark */
	T0* a2; /* like_keyword */
	T0* a3; /* current_keyword */
};

/* Struct for type [detachable] ET_DYNAMIC_PROCEDURE_TYPE */
struct S450 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* id */
	T0* a2; /* base_class */
	T0* a3; /* base_type */
	T1 a4; /* is_alive */
	T0* a5; /* attached_type */
	T6 a6; /* attribute_count */
	T0* a7; /* queries */
	T0* a8; /* meta_type */
	T0* a9; /* conforming_ancestors */
	T1 a10; /* has_redefined_is_equal_routine */
	T1 a11; /* has_redefined_copy_routine */
	T0* a12; /* next_type */
	T1 a13; /* has_once_per_object_routines */
	T0* a14; /* procedures */
	T1 a15; /* has_static */
	T0* a16; /* query_calls */
	T0* a17; /* procedure_calls */
	T0* a18; /* open_operand_type_sets */
	T6 a19; /* hash_code */
	T0* a20; /* queries_by_seed */
	T1 a21; /* has_reference_attributes */
	T0* a22; /* open_operand_tuple_type_set */
	T0* a23; /* set_rout_disp_final_feature */
	T0* a24; /* procedures_by_seed */
	T0* a25; /* equality_expressions */
	T0* a26; /* object_equality_expressions */
};

/* Struct for type [detachable] ET_DYNAMIC_FUNCTION_TYPE */
struct S451 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* id */
	T0* a2; /* base_class */
	T0* a3; /* base_type */
	T1 a4; /* is_alive */
	T0* a5; /* attached_type */
	T6 a6; /* attribute_count */
	T0* a7; /* queries */
	T0* a8; /* meta_type */
	T0* a9; /* conforming_ancestors */
	T1 a10; /* has_redefined_is_equal_routine */
	T1 a11; /* has_redefined_copy_routine */
	T0* a12; /* next_type */
	T1 a13; /* has_once_per_object_routines */
	T0* a14; /* procedures */
	T1 a15; /* has_static */
	T0* a16; /* query_calls */
	T0* a17; /* procedure_calls */
	T0* a18; /* open_operand_type_sets */
	T0* a19; /* result_type_set */
	T6 a20; /* hash_code */
	T0* a21; /* queries_by_seed */
	T1 a22; /* has_reference_attributes */
	T0* a23; /* procedures_by_seed */
	T0* a24; /* open_operand_tuple_type_set */
	T0* a25; /* set_rout_disp_final_feature */
	T0* a26; /* equality_expressions */
	T0* a27; /* object_equality_expressions */
};

/* Struct for type [detachable] DT_SHARED_SYSTEM_CLOCK */
struct S452 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DT_SYSTEM_CLOCK */
struct S453 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* second */
	T6 a2; /* year */
	T6 a3; /* month */
	T6 a4; /* day */
	T6 a5; /* hour */
	T6 a6; /* minute */
	T6 a7; /* millisecond */
	T0* a8; /* local_clock */
};

/* Struct for type [detachable] ET_EIFFEL_PREPARSER */
struct S454 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* eiffel_buffer */
	T0* a2; /* system_processor */
	T6 a3; /* last_text_count */
	T6 a4; /* last_literal_start */
	T0* a5; /* filename */
	T0* a6; /* group */
	T0* a7; /* verbatim_marker */
	T6 a8; /* verbatim_marker_count */
	T0* a9; /* verbatim_open_white_characters */
	T0* a10; /* verbatim_close_white_characters */
	T0* a11; /* yy_pushed_start_conditions */
	T0* a12; /* input_buffer */
	T0* a13; /* yy_state_stack */
	T6 a14; /* yy_start_state */
	T6 a15; /* yy_line */
	T6 a16; /* yy_column */
	T6 a17; /* yy_position */
	T6 a18; /* yy_start */
	T6 a19; /* yy_end */
	T6 a20; /* line */
	T6 a21; /* column */
	T6 a22; /* position */
	T0* a23; /* yy_nxt */
	T0* a24; /* yy_chk */
	T0* a25; /* yy_base */
	T0* a26; /* yy_def */
	T0* a27; /* yy_ec */
	T0* a28; /* yy_meta */
	T0* a29; /* yy_accept */
	T0* a30; /* yy_acclist */
	T0* a31; /* yy_content */
	T0* a32; /* yy_content_area */
	T0* a33; /* yy_unicode_content_area */
	T0* a34; /* last_classname */
	T1 a35; /* class_keyword_found */
	T6 a36; /* last_token */
	T1 a37; /* yy_more_flag */
	T6 a38; /* yy_more_len */
	T6 a39; /* yy_state_count */
	T6 a40; /* yy_last_accepting_state */
	T6 a41; /* yy_last_accepting_cpos */
	T6 a42; /* yy_lp */
	T6 a43; /* yy_looking_for_trail_begin */
	T6 a44; /* yy_full_match */
	T6 a45; /* yy_full_state */
	T6 a46; /* yy_full_lp */
	T1 a47; /* yy_rejected */
	T6 a48; /* last_literal_end */
	T6 a49; /* pushed_start_condition_count */
	T6 a50; /* ms_column */
	T6 a51; /* ms_line */
	T6 a52; /* last_break_end */
	T6 a53; /* last_comment_end */
	T3 a54; /* verbatim_opening_character */
};

/* Struct for type [detachable] ET_MASTER_CLASS_CHECKER */
struct S455 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* current_class */
	T0* a2; /* system_processor */
	T1 a3; /* has_fatal_error */
};

/* Struct for type [detachable] ET_EIFFEL_PARSER */
struct S456 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* eiffel_buffer */
	T0* a2; /* counters */
	T0* a3; /* last_formal_arguments_stack */
	T0* a4; /* last_local_variables_stack */
	T0* a5; /* last_keywords */
	T0* a6; /* last_symbols */
	T0* a7; /* last_object_tests_stack */
	T0* a8; /* last_object_tests_pool */
	T0* a9; /* last_iteration_components_stack */
	T0* a10; /* last_iteration_components_pool */
	T0* a11; /* assertions */
	T0* a12; /* assertion_counters */
	T0* a13; /* assertion_kinds */
	T6 a14; /* assertion_kind */
	T0* a15; /* queries */
	T0* a16; /* procedures */
	T0* a17; /* constraints */
	T0* a18; /* providers */
	T0* a19; /* current_class */
	T0* a20; /* system_processor */
	T6 a21; /* last_text_count */
	T6 a22; /* last_literal_start */
	T0* a23; /* filename */
	T0* a24; /* group */
	T0* a25; /* verbatim_marker */
	T6 a26; /* verbatim_marker_count */
	T0* a27; /* verbatim_open_white_characters */
	T0* a28; /* verbatim_close_white_characters */
	T0* a29; /* yyss */
	T0* a30; /* yy_pushed_start_conditions */
	T0* a31; /* input_buffer */
	T0* a32; /* yyspecial_routines1 */
	T6 a33; /* yyvsc1 */
	T0* a34; /* yyvs1 */
	T0* a35; /* yyspecial_routines2 */
	T6 a36; /* yyvsc2 */
	T0* a37; /* yyvs2 */
	T0* a38; /* yyspecial_routines3 */
	T6 a39; /* yyvsc3 */
	T0* a40; /* yyvs3 */
	T0* a41; /* yyspecial_routines4 */
	T6 a42; /* yyvsc4 */
	T0* a43; /* yyvs4 */
	T0* a44; /* yyspecial_routines5 */
	T6 a45; /* yyvsc5 */
	T0* a46; /* yyvs5 */
	T0* a47; /* yyspecial_routines6 */
	T6 a48; /* yyvsc6 */
	T0* a49; /* yyvs6 */
	T0* a50; /* yyspecial_routines7 */
	T6 a51; /* yyvsc7 */
	T0* a52; /* yyvs7 */
	T0* a53; /* yyspecial_routines8 */
	T6 a54; /* yyvsc8 */
	T0* a55; /* yyvs8 */
	T0* a56; /* yyspecial_routines9 */
	T6 a57; /* yyvsc9 */
	T0* a58; /* yyvs9 */
	T0* a59; /* yyspecial_routines10 */
	T6 a60; /* yyvsc10 */
	T0* a61; /* yyvs10 */
	T0* a62; /* yyspecial_routines11 */
	T6 a63; /* yyvsc11 */
	T0* a64; /* yyvs11 */
	T0* a65; /* yyspecial_routines12 */
	T6 a66; /* yyvsc12 */
	T0* a67; /* yyvs12 */
	T0* a68; /* yyspecial_routines13 */
	T6 a69; /* yyvsc13 */
	T0* a70; /* yyvs13 */
	T0* a71; /* yyspecial_routines14 */
	T6 a72; /* yyvsc14 */
	T0* a73; /* yyvs14 */
	T0* a74; /* yyspecial_routines15 */
	T6 a75; /* yyvsc15 */
	T0* a76; /* yyvs15 */
	T0* a77; /* yyspecial_routines16 */
	T6 a78; /* yyvsc16 */
	T0* a79; /* yyvs16 */
	T0* a80; /* yyspecial_routines17 */
	T6 a81; /* yyvsc17 */
	T0* a82; /* yyvs17 */
	T0* a83; /* yyspecial_routines18 */
	T6 a84; /* yyvsc18 */
	T0* a85; /* yyvs18 */
	T0* a86; /* yyspecial_routines19 */
	T6 a87; /* yyvsc19 */
	T0* a88; /* yyvs19 */
	T0* a89; /* yyspecial_routines20 */
	T6 a90; /* yyvsc20 */
	T0* a91; /* yyvs20 */
	T0* a92; /* yyspecial_routines21 */
	T6 a93; /* yyvsc21 */
	T0* a94; /* yyvs21 */
	T0* a95; /* yyspecial_routines22 */
	T6 a96; /* yyvsc22 */
	T0* a97; /* yyvs22 */
	T0* a98; /* yyspecial_routines23 */
	T6 a99; /* yyvsc23 */
	T0* a100; /* yyvs23 */
	T0* a101; /* yyspecial_routines24 */
	T6 a102; /* yyvsc24 */
	T0* a103; /* yyvs24 */
	T0* a104; /* yyspecial_routines25 */
	T6 a105; /* yyvsc25 */
	T0* a106; /* yyvs25 */
	T0* a107; /* yyspecial_routines26 */
	T6 a108; /* yyvsc26 */
	T0* a109; /* yyvs26 */
	T0* a110; /* yyspecial_routines27 */
	T6 a111; /* yyvsc27 */
	T0* a112; /* yyvs27 */
	T0* a113; /* yyspecial_routines28 */
	T6 a114; /* yyvsc28 */
	T0* a115; /* yyvs28 */
	T0* a116; /* yyspecial_routines29 */
	T6 a117; /* yyvsc29 */
	T0* a118; /* yyvs29 */
	T0* a119; /* yyspecial_routines30 */
	T6 a120; /* yyvsc30 */
	T0* a121; /* yyvs30 */
	T0* a122; /* yyspecial_routines31 */
	T6 a123; /* yyvsc31 */
	T0* a124; /* yyvs31 */
	T0* a125; /* yyspecial_routines32 */
	T6 a126; /* yyvsc32 */
	T0* a127; /* yyvs32 */
	T0* a128; /* yyspecial_routines33 */
	T6 a129; /* yyvsc33 */
	T0* a130; /* yyvs33 */
	T0* a131; /* yyspecial_routines34 */
	T6 a132; /* yyvsc34 */
	T0* a133; /* yyvs34 */
	T0* a134; /* yyspecial_routines35 */
	T6 a135; /* yyvsc35 */
	T0* a136; /* yyvs35 */
	T0* a137; /* yyspecial_routines36 */
	T6 a138; /* yyvsc36 */
	T0* a139; /* yyvs36 */
	T0* a140; /* yyspecial_routines37 */
	T6 a141; /* yyvsc37 */
	T0* a142; /* yyvs37 */
	T0* a143; /* yyspecial_routines38 */
	T6 a144; /* yyvsc38 */
	T0* a145; /* yyvs38 */
	T0* a146; /* yyspecial_routines39 */
	T6 a147; /* yyvsc39 */
	T0* a148; /* yyvs39 */
	T0* a149; /* yyspecial_routines40 */
	T6 a150; /* yyvsc40 */
	T0* a151; /* yyvs40 */
	T0* a152; /* yyspecial_routines41 */
	T6 a153; /* yyvsc41 */
	T0* a154; /* yyvs41 */
	T0* a155; /* yyspecial_routines42 */
	T6 a156; /* yyvsc42 */
	T0* a157; /* yyvs42 */
	T0* a158; /* yyspecial_routines43 */
	T6 a159; /* yyvsc43 */
	T0* a160; /* yyvs43 */
	T0* a161; /* yyspecial_routines44 */
	T6 a162; /* yyvsc44 */
	T0* a163; /* yyvs44 */
	T0* a164; /* yyspecial_routines45 */
	T6 a165; /* yyvsc45 */
	T0* a166; /* yyvs45 */
	T0* a167; /* yyspecial_routines46 */
	T6 a168; /* yyvsc46 */
	T0* a169; /* yyvs46 */
	T0* a170; /* yyspecial_routines47 */
	T6 a171; /* yyvsc47 */
	T0* a172; /* yyvs47 */
	T0* a173; /* yyspecial_routines48 */
	T6 a174; /* yyvsc48 */
	T0* a175; /* yyvs48 */
	T0* a176; /* yyspecial_routines49 */
	T6 a177; /* yyvsc49 */
	T0* a178; /* yyvs49 */
	T0* a179; /* yyspecial_routines50 */
	T6 a180; /* yyvsc50 */
	T0* a181; /* yyvs50 */
	T0* a182; /* yyspecial_routines51 */
	T6 a183; /* yyvsc51 */
	T0* a184; /* yyvs51 */
	T0* a185; /* yyspecial_routines52 */
	T6 a186; /* yyvsc52 */
	T0* a187; /* yyvs52 */
	T0* a188; /* yyspecial_routines53 */
	T6 a189; /* yyvsc53 */
	T0* a190; /* yyvs53 */
	T0* a191; /* yyspecial_routines54 */
	T6 a192; /* yyvsc54 */
	T0* a193; /* yyvs54 */
	T0* a194; /* yyspecial_routines55 */
	T6 a195; /* yyvsc55 */
	T0* a196; /* yyvs55 */
	T0* a197; /* yyspecial_routines56 */
	T6 a198; /* yyvsc56 */
	T0* a199; /* yyvs56 */
	T0* a200; /* yyspecial_routines57 */
	T6 a201; /* yyvsc57 */
	T0* a202; /* yyvs57 */
	T0* a203; /* yyspecial_routines58 */
	T6 a204; /* yyvsc58 */
	T0* a205; /* yyvs58 */
	T0* a206; /* yyspecial_routines59 */
	T6 a207; /* yyvsc59 */
	T0* a208; /* yyvs59 */
	T0* a209; /* yyspecial_routines60 */
	T6 a210; /* yyvsc60 */
	T0* a211; /* yyvs60 */
	T0* a212; /* yyspecial_routines61 */
	T6 a213; /* yyvsc61 */
	T0* a214; /* yyvs61 */
	T0* a215; /* yyspecial_routines62 */
	T6 a216; /* yyvsc62 */
	T0* a217; /* yyvs62 */
	T0* a218; /* yyspecial_routines63 */
	T6 a219; /* yyvsc63 */
	T0* a220; /* yyvs63 */
	T0* a221; /* yyspecial_routines64 */
	T6 a222; /* yyvsc64 */
	T0* a223; /* yyvs64 */
	T0* a224; /* yyspecial_routines65 */
	T6 a225; /* yyvsc65 */
	T0* a226; /* yyvs65 */
	T0* a227; /* yyspecial_routines66 */
	T6 a228; /* yyvsc66 */
	T0* a229; /* yyvs66 */
	T0* a230; /* yyspecial_routines67 */
	T6 a231; /* yyvsc67 */
	T0* a232; /* yyvs67 */
	T0* a233; /* yyspecial_routines68 */
	T6 a234; /* yyvsc68 */
	T0* a235; /* yyvs68 */
	T0* a236; /* yyspecial_routines69 */
	T6 a237; /* yyvsc69 */
	T0* a238; /* yyvs69 */
	T0* a239; /* yyspecial_routines70 */
	T6 a240; /* yyvsc70 */
	T0* a241; /* yyvs70 */
	T0* a242; /* yyspecial_routines71 */
	T6 a243; /* yyvsc71 */
	T0* a244; /* yyvs71 */
	T0* a245; /* yyspecial_routines72 */
	T6 a246; /* yyvsc72 */
	T0* a247; /* yyvs72 */
	T0* a248; /* yyspecial_routines73 */
	T6 a249; /* yyvsc73 */
	T0* a250; /* yyvs73 */
	T0* a251; /* yyspecial_routines74 */
	T6 a252; /* yyvsc74 */
	T0* a253; /* yyvs74 */
	T0* a254; /* yyspecial_routines75 */
	T6 a255; /* yyvsc75 */
	T0* a256; /* yyvs75 */
	T0* a257; /* yyspecial_routines76 */
	T6 a258; /* yyvsc76 */
	T0* a259; /* yyvs76 */
	T0* a260; /* yyspecial_routines77 */
	T6 a261; /* yyvsc77 */
	T0* a262; /* yyvs77 */
	T0* a263; /* yyspecial_routines78 */
	T6 a264; /* yyvsc78 */
	T0* a265; /* yyvs78 */
	T0* a266; /* yyspecial_routines79 */
	T6 a267; /* yyvsc79 */
	T0* a268; /* yyvs79 */
	T0* a269; /* yyspecial_routines80 */
	T6 a270; /* yyvsc80 */
	T0* a271; /* yyvs80 */
	T0* a272; /* yyspecial_routines81 */
	T6 a273; /* yyvsc81 */
	T0* a274; /* yyvs81 */
	T0* a275; /* yyspecial_routines82 */
	T6 a276; /* yyvsc82 */
	T0* a277; /* yyvs82 */
	T0* a278; /* yyspecial_routines83 */
	T6 a279; /* yyvsc83 */
	T0* a280; /* yyvs83 */
	T0* a281; /* yyspecial_routines84 */
	T6 a282; /* yyvsc84 */
	T0* a283; /* yyvs84 */
	T0* a284; /* yyspecial_routines85 */
	T6 a285; /* yyvsc85 */
	T0* a286; /* yyvs85 */
	T0* a287; /* yyspecial_routines86 */
	T6 a288; /* yyvsc86 */
	T0* a289; /* yyvs86 */
	T0* a290; /* yyspecial_routines87 */
	T6 a291; /* yyvsc87 */
	T0* a292; /* yyvs87 */
	T0* a293; /* yyspecial_routines88 */
	T6 a294; /* yyvsc88 */
	T0* a295; /* yyvs88 */
	T0* a296; /* yyspecial_routines89 */
	T6 a297; /* yyvsc89 */
	T0* a298; /* yyvs89 */
	T0* a299; /* yyspecial_routines90 */
	T6 a300; /* yyvsc90 */
	T0* a301; /* yyvs90 */
	T0* a302; /* yyspecial_routines91 */
	T6 a303; /* yyvsc91 */
	T0* a304; /* yyvs91 */
	T0* a305; /* yyspecial_routines92 */
	T6 a306; /* yyvsc92 */
	T0* a307; /* yyvs92 */
	T0* a308; /* yyspecial_routines93 */
	T6 a309; /* yyvsc93 */
	T0* a310; /* yyvs93 */
	T0* a311; /* yyspecial_routines94 */
	T6 a312; /* yyvsc94 */
	T0* a313; /* yyvs94 */
	T0* a314; /* yyspecial_routines95 */
	T6 a315; /* yyvsc95 */
	T0* a316; /* yyvs95 */
	T0* a317; /* yyspecial_routines96 */
	T6 a318; /* yyvsc96 */
	T0* a319; /* yyvs96 */
	T0* a320; /* yyspecial_routines97 */
	T6 a321; /* yyvsc97 */
	T0* a322; /* yyvs97 */
	T0* a323; /* yyspecial_routines98 */
	T6 a324; /* yyvsc98 */
	T0* a325; /* yyvs98 */
	T0* a326; /* yyspecial_routines99 */
	T6 a327; /* yyvsc99 */
	T0* a328; /* yyvs99 */
	T0* a329; /* yyspecial_routines100 */
	T6 a330; /* yyvsc100 */
	T0* a331; /* yyvs100 */
	T0* a332; /* yyspecial_routines101 */
	T6 a333; /* yyvsc101 */
	T0* a334; /* yyvs101 */
	T0* a335; /* yyspecial_routines102 */
	T6 a336; /* yyvsc102 */
	T0* a337; /* yyvs102 */
	T0* a338; /* yyspecial_routines103 */
	T6 a339; /* yyvsc103 */
	T0* a340; /* yyvs103 */
	T0* a341; /* yyspecial_routines104 */
	T6 a342; /* yyvsc104 */
	T0* a343; /* yyvs104 */
	T0* a344; /* yyspecial_routines105 */
	T6 a345; /* yyvsc105 */
	T0* a346; /* yyvs105 */
	T0* a347; /* yyspecial_routines106 */
	T6 a348; /* yyvsc106 */
	T0* a349; /* yyvs106 */
	T0* a350; /* yyspecial_routines107 */
	T6 a351; /* yyvsc107 */
	T0* a352; /* yyvs107 */
	T0* a353; /* yyspecial_routines108 */
	T6 a354; /* yyvsc108 */
	T0* a355; /* yyvs108 */
	T0* a356; /* yyspecial_routines109 */
	T6 a357; /* yyvsc109 */
	T0* a358; /* yyvs109 */
	T0* a359; /* yyspecial_routines110 */
	T6 a360; /* yyvsc110 */
	T0* a361; /* yyvs110 */
	T0* a362; /* yyspecial_routines111 */
	T6 a363; /* yyvsc111 */
	T0* a364; /* yyvs111 */
	T0* a365; /* yyspecial_routines112 */
	T6 a366; /* yyvsc112 */
	T0* a367; /* yyvs112 */
	T0* a368; /* yyspecial_routines113 */
	T6 a369; /* yyvsc113 */
	T0* a370; /* yyvs113 */
	T0* a371; /* yyspecial_routines114 */
	T6 a372; /* yyvsc114 */
	T0* a373; /* yyvs114 */
	T0* a374; /* yyspecial_routines115 */
	T6 a375; /* yyvsc115 */
	T0* a376; /* yyvs115 */
	T0* a377; /* yyspecial_routines116 */
	T6 a378; /* yyvsc116 */
	T0* a379; /* yyvs116 */
	T0* a380; /* yyspecial_routines117 */
	T6 a381; /* yyvsc117 */
	T0* a382; /* yyvs117 */
	T0* a383; /* yyspecial_routines118 */
	T6 a384; /* yyvsc118 */
	T0* a385; /* yyvs118 */
	T0* a386; /* yyspecial_routines119 */
	T6 a387; /* yyvsc119 */
	T0* a388; /* yyvs119 */
	T0* a389; /* yyspecial_routines120 */
	T6 a390; /* yyvsc120 */
	T0* a391; /* yyvs120 */
	T0* a392; /* yyspecial_routines121 */
	T6 a393; /* yyvsc121 */
	T0* a394; /* yyvs121 */
	T0* a395; /* yyspecial_routines122 */
	T6 a396; /* yyvsc122 */
	T0* a397; /* yyvs122 */
	T0* a398; /* yyspecial_routines123 */
	T6 a399; /* yyvsc123 */
	T0* a400; /* yyvs123 */
	T0* a401; /* yyspecial_routines124 */
	T6 a402; /* yyvsc124 */
	T0* a403; /* yyvs124 */
	T0* a404; /* yyspecial_routines125 */
	T6 a405; /* yyvsc125 */
	T0* a406; /* yyvs125 */
	T0* a407; /* yyspecial_routines126 */
	T6 a408; /* yyvsc126 */
	T0* a409; /* yyvs126 */
	T0* a410; /* yyspecial_routines127 */
	T6 a411; /* yyvsc127 */
	T0* a412; /* yyvs127 */
	T0* a413; /* yyspecial_routines128 */
	T6 a414; /* yyvsc128 */
	T0* a415; /* yyvs128 */
	T0* a416; /* yyspecial_routines129 */
	T6 a417; /* yyvsc129 */
	T0* a418; /* yyvs129 */
	T0* a419; /* yyspecial_routines130 */
	T6 a420; /* yyvsc130 */
	T0* a421; /* yyvs130 */
	T0* a422; /* yyspecial_routines131 */
	T6 a423; /* yyvsc131 */
	T0* a424; /* yyvs131 */
	T0* a425; /* yyspecial_routines132 */
	T6 a426; /* yyvsc132 */
	T0* a427; /* yyvs132 */
	T0* a428; /* yyspecial_routines133 */
	T6 a429; /* yyvsc133 */
	T0* a430; /* yyvs133 */
	T0* a431; /* yyspecial_routines134 */
	T6 a432; /* yyvsc134 */
	T0* a433; /* yyvs134 */
	T0* a434; /* yytranslate */
	T0* a435; /* yyr1 */
	T0* a436; /* yytypes1 */
	T0* a437; /* yytypes2 */
	T0* a438; /* yydefact */
	T0* a439; /* yydefgoto */
	T0* a440; /* yypact */
	T0* a441; /* yypgoto */
	T0* a442; /* yytable */
	T0* a443; /* yycheck */
	T0* a444; /* yy_state_stack */
	T6 a445; /* yy_start_state */
	T6 a446; /* yy_line */
	T6 a447; /* yy_column */
	T6 a448; /* yy_position */
	T6 a449; /* yy_start */
	T6 a450; /* yy_end */
	T6 a451; /* line */
	T6 a452; /* column */
	T6 a453; /* position */
	T0* a454; /* yy_nxt */
	T0* a455; /* yy_chk */
	T0* a456; /* yy_base */
	T0* a457; /* yy_def */
	T0* a458; /* yy_ec */
	T0* a459; /* yy_meta */
	T0* a460; /* yy_accept */
	T0* a461; /* yy_acclist */
	T0* a462; /* yy_content */
	T0* a463; /* yy_content_area */
	T0* a464; /* yy_unicode_content_area */
	T6 a465; /* time_stamp */
	T0* a466; /* last_class */
	T6 a467; /* yy_parsing_status */
	T6 a468; /* yy_suspended_yystacksize */
	T6 a469; /* yy_suspended_yystate */
	T6 a470; /* yy_suspended_yyn */
	T6 a471; /* yy_suspended_yychar1 */
	T6 a472; /* yy_suspended_index */
	T6 a473; /* yy_suspended_yyss_top */
	T6 a474; /* yy_suspended_yy_goto */
	T6 a475; /* yyssp */
	T6 a476; /* error_count */
	T1 a477; /* yy_lookahead_needed */
	T6 a478; /* yyerrstatus */
	T6 a479; /* last_token */
	T0* a480; /* last_clients */
	T0* a481; /* last_export_clients */
	T0* a482; /* last_feature_clause */
	T6 a483; /* last_literal_end */
	T1 a484; /* yy_rejected */
	T6 a485; /* yy_state_count */
	T6 a486; /* yy_full_match */
	T6 a487; /* yy_lp */
	T6 a488; /* yy_looking_for_trail_begin */
	T6 a489; /* yy_full_lp */
	T6 a490; /* yy_full_state */
	T6 a491; /* pushed_start_condition_count */
	T1 a492; /* yy_more_flag */
	T6 a493; /* yy_more_len */
	T6 a494; /* yy_last_accepting_state */
	T6 a495; /* yy_last_accepting_cpos */
	T6 a496; /* yyvsp1 */
	T6 a497; /* yyvsp2 */
	T6 a498; /* yyvsp3 */
	T6 a499; /* yyvsp4 */
	T6 a500; /* yyvsp5 */
	T6 a501; /* yyvsp6 */
	T6 a502; /* yyvsp7 */
	T6 a503; /* yyvsp8 */
	T6 a504; /* yyvsp9 */
	T6 a505; /* yyvsp10 */
	T6 a506; /* yyvsp11 */
	T6 a507; /* yyvsp12 */
	T6 a508; /* yyvsp13 */
	T6 a509; /* yyvsp14 */
	T6 a510; /* yyvsp15 */
	T6 a511; /* yyvsp16 */
	T6 a512; /* yyvsp17 */
	T6 a513; /* yyvsp18 */
	T6 a514; /* yyvsp19 */
	T6 a515; /* yyvsp20 */
	T6 a516; /* yyvsp21 */
	T6 a517; /* yyvsp22 */
	T6 a518; /* yyvsp23 */
	T6 a519; /* yyvsp24 */
	T6 a520; /* yyvsp25 */
	T6 a521; /* yyvsp26 */
	T6 a522; /* yyvsp27 */
	T6 a523; /* yyvsp28 */
	T6 a524; /* yyvsp29 */
	T6 a525; /* yyvsp30 */
	T6 a526; /* yyvsp31 */
	T6 a527; /* yyvsp32 */
	T6 a528; /* yyvsp33 */
	T6 a529; /* yyvsp34 */
	T6 a530; /* yyvsp35 */
	T6 a531; /* yyvsp36 */
	T6 a532; /* yyvsp37 */
	T6 a533; /* yyvsp38 */
	T6 a534; /* yyvsp39 */
	T6 a535; /* yyvsp40 */
	T6 a536; /* yyvsp41 */
	T6 a537; /* yyvsp42 */
	T6 a538; /* yyvsp43 */
	T6 a539; /* yyvsp44 */
	T6 a540; /* yyvsp45 */
	T6 a541; /* yyvsp46 */
	T6 a542; /* yyvsp47 */
	T6 a543; /* yyvsp48 */
	T6 a544; /* yyvsp49 */
	T6 a545; /* yyvsp50 */
	T6 a546; /* yyvsp51 */
	T6 a547; /* yyvsp52 */
	T6 a548; /* yyvsp53 */
	T6 a549; /* yyvsp54 */
	T6 a550; /* yyvsp55 */
	T6 a551; /* yyvsp56 */
	T6 a552; /* yyvsp57 */
	T6 a553; /* yyvsp58 */
	T6 a554; /* yyvsp59 */
	T6 a555; /* yyvsp60 */
	T6 a556; /* yyvsp61 */
	T6 a557; /* yyvsp62 */
	T6 a558; /* yyvsp63 */
	T6 a559; /* yyvsp64 */
	T6 a560; /* yyvsp65 */
	T6 a561; /* yyvsp66 */
	T6 a562; /* yyvsp67 */
	T6 a563; /* yyvsp68 */
	T6 a564; /* yyvsp69 */
	T6 a565; /* yyvsp70 */
	T6 a566; /* yyvsp71 */
	T6 a567; /* yyvsp72 */
	T6 a568; /* yyvsp73 */
	T6 a569; /* yyvsp74 */
	T6 a570; /* yyvsp75 */
	T6 a571; /* yyvsp76 */
	T6 a572; /* yyvsp77 */
	T6 a573; /* yyvsp78 */
	T6 a574; /* yyvsp79 */
	T6 a575; /* yyvsp80 */
	T6 a576; /* yyvsp81 */
	T6 a577; /* yyvsp82 */
	T6 a578; /* yyvsp83 */
	T6 a579; /* yyvsp84 */
	T6 a580; /* yyvsp85 */
	T6 a581; /* yyvsp86 */
	T6 a582; /* yyvsp87 */
	T6 a583; /* yyvsp88 */
	T6 a584; /* yyvsp89 */
	T6 a585; /* yyvsp90 */
	T6 a586; /* yyvsp91 */
	T6 a587; /* yyvsp92 */
	T6 a588; /* yyvsp93 */
	T6 a589; /* yyvsp94 */
	T6 a590; /* yyvsp95 */
	T6 a591; /* yyvsp96 */
	T6 a592; /* yyvsp97 */
	T6 a593; /* yyvsp98 */
	T6 a594; /* yyvsp99 */
	T6 a595; /* yyvsp100 */
	T6 a596; /* yyvsp101 */
	T6 a597; /* yyvsp102 */
	T6 a598; /* yyvsp103 */
	T6 a599; /* yyvsp104 */
	T6 a600; /* yyvsp105 */
	T6 a601; /* yyvsp106 */
	T6 a602; /* yyvsp107 */
	T6 a603; /* yyvsp108 */
	T6 a604; /* yyvsp109 */
	T6 a605; /* yyvsp110 */
	T6 a606; /* yyvsp111 */
	T6 a607; /* yyvsp112 */
	T6 a608; /* yyvsp113 */
	T6 a609; /* yyvsp114 */
	T6 a610; /* yyvsp115 */
	T6 a611; /* yyvsp116 */
	T6 a612; /* yyvsp117 */
	T6 a613; /* yyvsp118 */
	T6 a614; /* yyvsp119 */
	T6 a615; /* yyvsp120 */
	T6 a616; /* yyvsp121 */
	T6 a617; /* yyvsp122 */
	T6 a618; /* yyvsp123 */
	T6 a619; /* yyvsp124 */
	T6 a620; /* yyvsp125 */
	T6 a621; /* yyvsp126 */
	T6 a622; /* yyvsp127 */
	T6 a623; /* yyvsp128 */
	T6 a624; /* yyvsp129 */
	T6 a625; /* yyvsp130 */
	T6 a626; /* yyvsp131 */
	T6 a627; /* yyvsp132 */
	T6 a628; /* yyvsp133 */
	T6 a629; /* yyvsp134 */
	T0* a630; /* last_detachable_any_value */
	T0* a631; /* last_detachable_et_keyword_value */
	T0* a632; /* last_detachable_et_agent_keyword_value */
	T0* a633; /* last_detachable_et_precursor_keyword_value */
	T0* a634; /* last_detachable_et_symbol_value */
	T0* a635; /* last_detachable_et_position_value */
	T0* a636; /* last_detachable_et_boolean_constant_value */
	T0* a637; /* last_detachable_et_break_value */
	T0* a638; /* last_detachable_et_character_constant_value */
	T0* a639; /* last_detachable_et_current_value */
	T0* a640; /* last_detachable_et_free_operator_value */
	T0* a641; /* last_detachable_et_identifier_value */
	T0* a642; /* last_detachable_et_integer_constant_value */
	T0* a643; /* last_detachable_et_keyword_operator_value */
	T0* a644; /* last_detachable_et_manifest_string_value */
	T0* a645; /* last_detachable_et_real_constant_value */
	T0* a646; /* last_detachable_et_result_value */
	T0* a647; /* last_detachable_et_retry_instruction_value */
	T0* a648; /* last_detachable_et_symbol_operator_value */
	T0* a649; /* last_detachable_et_void_value */
	T0* a650; /* last_detachable_et_semicolon_symbol_value */
	T0* a651; /* last_detachable_et_bracket_symbol_value */
	T0* a652; /* last_detachable_et_question_mark_symbol_value */
	T0* a653; /* last_formal_arguments */
	T0* a654; /* last_local_variables */
	T0* a655; /* last_object_tests */
	T0* a656; /* last_iteration_components */
	T6 a657; /* last_break_end */
	T6 a658; /* last_comment_end */
	T6 a659; /* ms_column */
	T6 a660; /* ms_line */
	T6 a661; /* break_kind */
	T3 a662; /* verbatim_opening_character */
	T1 a663; /* has_natural_64_overflow */
	T11 a664; /* last_natural_64 */
};

/* Struct for type [detachable] ET_PROVIDER_CHECKER */
struct S457 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* current_class */
	T0* a2; /* system_processor */
};

/* Struct for type [detachable] ET_ANCESTOR_BUILDER */
struct S458 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* class_sorter */
	T0* a2; /* ancestors */
	T0* a3; /* conforming_ancestors */
	T0* a4; /* parent_checker */
	T0* a5; /* formal_parameter_checker */
	T0* a6; /* parent_context */
	T0* a7; /* current_class */
	T0* a8; /* system_processor */
};

/* Struct for type [detachable] ET_FEATURE_FLATTENER */
struct S459 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* named_features */
	T0* a2; /* queries */
	T0* a3; /* procedures */
	T0* a4; /* aliased_features */
	T0* a5; /* clients_list */
	T0* a6; /* client_classes */
	T0* a7; /* feature_adaptation_resolver */
	T0* a8; /* dotnet_feature_adaptation_resolver */
	T0* a9; /* identifier_type_resolver */
	T0* a10; /* unfolded_tuple_actual_parameters_resolver */
	T0* a11; /* anchored_type_checker */
	T0* a12; /* signature_checker */
	T0* a13; /* parent_checker */
	T0* a14; /* formal_parameter_checker */
	T0* a15; /* builtin_feature_checker */
	T0* a16; /* precursor_checker */
	T0* a17; /* precursors */
	T0* a18; /* current_class */
	T0* a19; /* system_processor */
	T1 a20; /* has_signature_error */
};

/* Struct for type [detachable] ET_INTERFACE_CHECKER */
struct S460 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* classes_to_be_processed */
	T0* a2; /* qualified_anchored_type_checker */
	T0* a3; /* unfolded_tuple_actual_parameters_resolver */
	T0* a4; /* old_name_rename_table */
	T0* a5; /* new_name_rename_table */
	T0* a6; /* new_alias_name_rename_table */
	T0* a7; /* current_class */
	T0* a8; /* system_processor */
};

/* Struct for type [detachable] ET_IMPLEMENTATION_CHECKER */
struct S461 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* current_class */
	T0* a2; /* system_processor */
	T0* a3; /* feature_checker */
	T0* a4; /* named_features */
	T0* a5; /* feature_adaptation_resolver */
	T0* a6; /* dotnet_feature_adaptation_resolver */
	T0* a7; /* signature_checker */
	T0* a8; /* parent_checker3 */
	T0* a9; /* precursor_procedures */
	T0* a10; /* precursor_queries */
	T0* a11; /* supplier_builder */
	T0* a12; /* no_suppliers */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS] */
struct S462 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
	T0* a6; /* equality_tester */
};

/* Struct for type [detachable] TUPLE [[attached] DS_ARRAYED_LIST [[attached] ET_CLASS]] */
struct S463 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] ET_AST_NULL_PROCESSOR */
struct S464 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_AST_FACTORY */
struct S465 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_DOTNET_ASSEMBLY_CLASSIC_CONSUMER */
struct S467 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* system_processor */
};

/* Struct for type [detachable] DT_DATE_TIME_DURATION */
struct S468 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* year */
	T6 a2; /* month */
	T6 a3; /* day */
	T6 a4; /* hour */
	T6 a5; /* minute */
	T6 a6; /* second */
	T6 a7; /* millisecond */
};

/* Struct for type [detachable] ET_DO_PROCEDURE */
struct S469 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* iteration_components */
	T0* a2; /* implementation_class */
	T0* a3; /* implementation_feature */
	T6 a4; /* first_seed */
	T0* a5; /* other_seeds */
	T0* a6; /* arguments */
	T0* a7; /* compound */
	T0* a8; /* extended_name */
	T6 a9; /* hash_code */
	T0* a10; /* end_keyword */
	T0* a11; /* clients */
	T1 a12; /* validity_checked */
	T1 a13; /* has_validity_error */
	T0* a14; /* preconditions */
	T0* a15; /* first_precursor */
	T0* a16; /* other_precursors */
	T0* a17; /* postconditions */
	T0* a18; /* rescue_clause */
	T0* a19; /* locals */
	T0* a20; /* object_tests */
	T0* a21; /* frozen_keyword */
	T0* a22; /* obsolete_message */
	T0* a23; /* is_keyword */
	T0* a24; /* semicolon */
	T0* a25; /* feature_clause */
	T0* a26; /* first_indexing */
	T6 a27; /* id */
	T6 a28; /* version */
	T0* a29; /* synonym */
};

/* Struct for type [detachable] UT_CONFIG_PARSER */
struct S471 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_error */
	T0* a2; /* config_values */
	T0* a3; /* error_handler */
	T0* a4; /* defined_values */
	T0* a5; /* include_stack */
	T0* a6; /* line_nb_stack */
	T0* a7; /* last_string_value */
	T6 a8; /* line_nb */
	T0* a9; /* yyss */
	T0* a10; /* input_buffer */
	T6 a11; /* yy_end */
	T6 a12; /* yy_position */
	T6 a13; /* yy_line */
	T6 a14; /* yy_column */
	T6 a15; /* yy_parsing_status */
	T6 a16; /* yy_suspended_yystacksize */
	T6 a17; /* yy_suspended_yystate */
	T6 a18; /* yy_suspended_yyn */
	T6 a19; /* yy_suspended_yychar1 */
	T6 a20; /* yy_suspended_index */
	T6 a21; /* yy_suspended_yyss_top */
	T6 a22; /* yy_suspended_yy_goto */
	T0* a23; /* yyr1 */
	T6 a24; /* yyssp */
	T0* a25; /* yypgoto */
	T0* a26; /* yycheck */
	T0* a27; /* yytable */
	T0* a28; /* yydefgoto */
	T6 a29; /* error_count */
	T1 a30; /* yy_lookahead_needed */
	T6 a31; /* yyerrstatus */
	T0* a32; /* yypact */
	T6 a33; /* last_token */
	T0* a34; /* yytranslate */
	T0* a35; /* yydefact */
	T0* a36; /* yy_pushed_start_conditions */
	T0* a37; /* yyspecial_routines1 */
	T6 a38; /* yyvsc1 */
	T0* a39; /* yyvs1 */
	T0* a40; /* yyspecial_routines2 */
	T6 a41; /* yyvsc2 */
	T0* a42; /* yyvs2 */
	T0* a43; /* yyspecial_routines3 */
	T6 a44; /* yyvsc3 */
	T0* a45; /* yyvs3 */
	T0* a46; /* yytypes1 */
	T0* a47; /* yytypes2 */
	T6 a48; /* yy_start */
	T6 a49; /* yyvsp1 */
	T6 a50; /* yyvsp2 */
	T6 a51; /* yyvsp3 */
	T1 a52; /* yy_more_flag */
	T6 a53; /* yy_more_len */
	T6 a54; /* line */
	T6 a55; /* column */
	T6 a56; /* position */
	T6 a57; /* yy_start_state */
	T0* a58; /* yy_state_stack */
	T6 a59; /* yy_state_count */
	T0* a60; /* yy_content_area */
	T0* a61; /* yy_unicode_content_area */
	T0* a62; /* yy_accept */
	T6 a63; /* yy_last_accepting_state */
	T6 a64; /* yy_last_accepting_cpos */
	T0* a65; /* yy_content */
	T0* a66; /* yy_ec */
	T0* a67; /* yy_chk */
	T0* a68; /* yy_base */
	T0* a69; /* yy_def */
	T0* a70; /* yy_meta */
	T0* a71; /* yy_nxt */
	T6 a72; /* yy_lp */
	T0* a73; /* yy_acclist */
	T6 a74; /* yy_looking_for_trail_begin */
	T6 a75; /* yy_full_match */
	T6 a76; /* yy_full_state */
	T6 a77; /* yy_full_lp */
	T1 a78; /* yy_rejected */
	T0* a79; /* last_detachable_any_value */
	T6 a80; /* if_level */
	T6 a81; /* ignored_level */
};

/* Struct for type [detachable] KL_COMPARABLE_COMPARATOR [[attached] INTEGER_32] */
struct S472 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_DYNAMIC_PUSH_TYPE_SET */
struct S475 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* static_type */
	T6 a2; /* count */
	T1 a3; /* is_never_void */
	T0* a4; /* dynamic_types */
	T0* a5; /* targets */
	T1 a6; /* is_dynamic_types_readonly */
};

/* Struct for type [detachable] ET_DEFERRED_PROCEDURE */
struct S476 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* implementation_feature */
	T0* a2; /* iteration_components */
	T0* a3; /* implementation_class */
	T6 a4; /* first_seed */
	T0* a5; /* other_seeds */
	T0* a6; /* arguments */
	T0* a7; /* extended_name */
	T0* a8; /* deferred_keyword */
	T6 a9; /* hash_code */
	T0* a10; /* end_keyword */
	T0* a11; /* clients */
	T1 a12; /* validity_checked */
	T1 a13; /* has_validity_error */
	T0* a14; /* preconditions */
	T0* a15; /* first_precursor */
	T0* a16; /* other_precursors */
	T0* a17; /* postconditions */
	T0* a18; /* object_tests */
	T0* a19; /* frozen_keyword */
	T0* a20; /* obsolete_message */
	T0* a21; /* is_keyword */
	T0* a22; /* semicolon */
	T0* a23; /* feature_clause */
	T0* a24; /* first_indexing */
	T6 a25; /* id */
	T6 a26; /* version */
	T0* a27; /* synonym */
};

/* Struct for type [detachable] KL_PLATFORM */
struct S477 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] CELL [[attached] INTEGER_32] */
struct S479 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* item */
};

/* Struct for type [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32] */
struct S480 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* capacity */
	T0* a2; /* content */
	T0* a3; /* keys */
	T0* a4; /* deleted_marks */
	T0* a5; /* indexes_map */
	T6 a6; /* iteration_position */
	T6 a7; /* count */
	T6 a8; /* deleted_item_position */
	T6 a9; /* control */
	T0* a10; /* found_item */
	T1 a11; /* has_default */
	T6 a12; /* item_position */
	T6 a13; /* ht_lowest_deleted_position */
	T0* a14; /* ht_deleted_item */
	T0* a15; /* ht_deleted_key */
	T1 a16; /* object_comparison */
};

/* Struct for type [detachable] KL_ANY_ROUTINES */
struct S481 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_PATHNAME */
struct S483 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* drive */
	T0* a3; /* hostname */
	T0* a4; /* sharename */
	T1 a5; /* is_relative */
	T0* a6; /* components */
};

/* Struct for type [detachable] ARRAY [[attached] IMMUTABLE_STRING_32] */
struct S484 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
	T1 a4; /* object_comparison */
};

/* Struct for type [detachable] SPECIAL [[attached] IMMUTABLE_STRING_32] */
struct S485 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] EXCEPTIONS */
struct S486 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] NATURAL_8] */
struct S487 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T8 z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_OPTION] */
struct S489 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] AP_OPTION] */
struct S491 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] AP_OPTION] */
struct S492 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
struct S493 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
struct S494 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
struct S495 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] STRING_8] */
struct S496 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [detachable STRING_8] */
struct S497 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable STRING_8] */
struct S498 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_LINKED_LIST_CURSOR [[attached] STRING_8] */
struct S499 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* after */
	T0* a2; /* current_cell */
	T0* a3; /* next_cursor */
	T0* a4; /* container */
	T1 a5; /* before */
};

/* Struct for type detachable DS_LINKABLE [[attached] STRING_8] */
struct S500 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
};

/* Struct for type detachable DS_LINKABLE [[attached] BOOLEAN] */
struct S501 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* item */
	T0* a2; /* right */
};

/* Struct for type [detachable] DS_LINKED_LIST_CURSOR [[attached] BOOLEAN] */
struct S502 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T1 a2; /* before */
};

/* Struct for type detachable DS_LINKABLE [[attached] INTEGER_32] */
struct S503 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* item */
	T0* a2; /* right */
};

/* Struct for type [detachable] DS_LINKED_LIST_CURSOR [[attached] INTEGER_32] */
struct S504 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T1 a2; /* before */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_LIBRARY, [attached] STRING_8] */
struct S506 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_LIBRARY] */
struct S507 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_LIBRARY] */
struct S510 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_DOTNET_ASSEMBLY] */
struct S511 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type detachable DS_HASH_TABLE_CURSOR [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8] */
struct S514 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_DOTNET_ASSEMBLY] */
struct S515 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] XM_EIFFEL_SCANNER */
struct S516 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* last_token */
	T0* a2; /* last_value */
	T0* a3; /* input_name */
	T6 a4; /* position */
	T6 a5; /* column */
	T6 a6; /* line */
	T0* a7; /* input_filter */
	T6 a8; /* yy_start_state */
	T0* a9; /* character_entity */
	T0* a10; /* input_stream */
	T0* a11; /* input_resolver */
	T1 a12; /* yy_more_flag */
	T6 a13; /* yy_more_len */
	T6 a14; /* yy_end */
	T6 a15; /* yy_start */
	T6 a16; /* yy_line */
	T6 a17; /* yy_column */
	T6 a18; /* yy_position */
	T0* a19; /* yy_state_stack */
	T6 a20; /* yy_state_count */
	T0* a21; /* yy_content_area */
	T0* a22; /* yy_unicode_content_area */
	T0* a23; /* yy_accept */
	T6 a24; /* yy_last_accepting_state */
	T6 a25; /* yy_last_accepting_cpos */
	T0* a26; /* yy_content */
	T0* a27; /* yy_ec */
	T0* a28; /* yy_chk */
	T0* a29; /* yy_base */
	T0* a30; /* yy_def */
	T0* a31; /* yy_meta */
	T0* a32; /* yy_nxt */
	T6 a33; /* yy_lp */
	T0* a34; /* yy_acclist */
	T6 a35; /* yy_looking_for_trail_begin */
	T6 a36; /* yy_full_match */
	T6 a37; /* yy_full_state */
	T6 a38; /* yy_full_lp */
	T0* a39; /* input_buffer */
	T1 a40; /* yy_rejected */
	T0* a41; /* yy_pushed_start_conditions */
	T0* a42; /* last_string_value */
	T0* a43; /* last_error */
	T6 a44; /* pushed_start_condition_count */
};

/* Struct for type [detachable] XM_DEFAULT_POSITION */
struct S517 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* row */
	T6 a2; /* column */
	T0* a3; /* source_name */
	T6 a4; /* byte_index */
};

/* Struct for type [detachable] DS_BILINKED_LIST [[attached] XM_POSITION] */
struct S519 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* first_cell */
	T0* a2; /* last_cell */
	T6 a3; /* count */
	T0* a4; /* internal_cursor */
};

/* Struct for type [detachable] DS_LINKED_STACK [[attached] XM_EIFFEL_SCANNER] */
struct S520 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* first_cell */
};

/* Struct for type [detachable] XM_CALLBACKS_NULL */
struct S521 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8] */
struct S522 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* item_storage */
	T6 a3; /* count */
	T6 a4; /* last_position */
	T6 a5; /* free_slot */
	T6 a6; /* capacity */
	T0* a7; /* key_equality_tester */
	T0* a8; /* internal_keys */
	T6 a9; /* modulus */
	T6 a10; /* slots_position */
	T6 a11; /* clashes_previous_position */
	T0* a12; /* internal_cursor */
	T6 a13; /* found_position */
	T0* a14; /* key_storage */
	T0* a15; /* clashes */
	T0* a16; /* slots */
	T0* a17; /* special_item_routines */
	T0* a18; /* special_key_routines */
	T0* a19; /* equality_tester */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] XM_NULL_EXTERNAL_RESOLVER */
struct S524 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] XM_DTD_CALLBACKS_NULL */
struct S526 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ANY] */
struct S527 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ANY] */
struct S528 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_PARSER_NAME] */
struct S529 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] XM_EIFFEL_PARSER_NAME */
struct S530 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* use_namespaces */
	T6 a2; /* count */
	T0* a3; /* first */
	T0* a4; /* second */
	T0* a5; /* tail */
};

/* Struct for type [detachable] SPECIAL [[attached] XM_EIFFEL_PARSER_NAME] */
struct S531 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]] */
struct S532 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME] */
struct S533 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* equality_tester */
	T6 a3; /* count */
	T6 a4; /* capacity */
	T6 a5; /* free_slot */
	T6 a6; /* last_position */
	T6 a7; /* modulus */
	T6 a8; /* slots_position */
	T6 a9; /* clashes_previous_position */
	T6 a10; /* found_position */
	T0* a11; /* clashes */
	T0* a12; /* slots */
	T0* a13; /* special_item_routines */
	T0* a14; /* item_storage */
	T0* a15; /* internal_cursor */
	T0* a16; /* hash_function */
};

/* Struct for type [detachable] SPECIAL [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]] */
struct S534 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_EXTERNAL_ID] */
struct S535 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] XM_DTD_EXTERNAL_ID */
struct S536 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* system_id */
	T0* a2; /* public_id */
};

/* Struct for type [detachable] SPECIAL [[attached] XM_DTD_EXTERNAL_ID] */
struct S537 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ELEMENT_CONTENT] */
struct S538 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] XM_DTD_ELEMENT_CONTENT */
struct S539 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* items */
	T2 a2; /* repetition */
	T0* a3; /* name */
	T2 a4; /* type */
	T1 a5; /* is_character_data_allowed */
};

/* Struct for type [detachable] SPECIAL [[attached] XM_DTD_ELEMENT_CONTENT] */
struct S540 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]] */
struct S541 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
struct S542 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* last_cell */
	T6 a2; /* count */
	T0* a3; /* first_cell */
	T0* a4; /* internal_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]] */
struct S543 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
struct S544 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] XM_DTD_ATTRIBUTE_CONTENT */
struct S545 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* default_value */
	T2 a3; /* type */
	T1 a4; /* is_list_type */
	T0* a5; /* enumeration_list */
	T2 a6; /* value */
};

/* Struct for type [detachable] SPECIAL [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
struct S546 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] STRING_8]] */
struct S547 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_BILINKED_LIST [[attached] STRING_8] */
struct S548 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* last_cell */
	T6 a2; /* count */
	T0* a3; /* first_cell */
	T0* a4; /* equality_tester */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] STRING_8]] */
struct S549 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] BOOLEAN] */
struct S550 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_DECLARATION] */
struct S551 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] XM_EIFFEL_DECLARATION */
struct S552 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* encoding */
	T0* a2; /* version */
	T1 a3; /* stand_alone */
};

/* Struct for type [detachable] SPECIAL [[attached] XM_EIFFEL_DECLARATION] */
struct S553 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] XM_EIFFEL_ENTITY_DEF */
struct S556 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* literal_name */
	T0* a2; /* value */
	T0* a3; /* resolver */
	T0* a4; /* external_id */
	T0* a5; /* character_entity */
	T0* a6; /* yy_pushed_start_conditions */
	T0* a7; /* input_buffer */
	T1 a8; /* in_use */
	T0* a9; /* input_name */
	T0* a10; /* last_value */
	T0* a11; /* last_string_value */
	T0* a12; /* last_error */
	T1 a13; /* yy_rejected */
	T6 a14; /* yy_state_count */
	T6 a15; /* yy_full_match */
	T6 a16; /* yy_lp */
	T6 a17; /* yy_looking_for_trail_begin */
	T6 a18; /* yy_full_lp */
	T6 a19; /* yy_full_state */
	T6 a20; /* pushed_start_condition_count */
	T6 a21; /* yy_start_state */
	T6 a22; /* yy_line */
	T6 a23; /* yy_column */
	T6 a24; /* yy_position */
	T6 a25; /* yy_start */
	T6 a26; /* yy_end */
	T6 a27; /* line */
	T6 a28; /* column */
	T6 a29; /* position */
	T1 a30; /* yy_more_flag */
	T6 a31; /* yy_more_len */
	T6 a32; /* yy_last_accepting_state */
	T6 a33; /* yy_last_accepting_cpos */
	T0* a34; /* yy_state_stack */
	T0* a35; /* yy_nxt */
	T0* a36; /* yy_chk */
	T0* a37; /* yy_base */
	T0* a38; /* yy_def */
	T0* a39; /* yy_ec */
	T0* a40; /* yy_meta */
	T0* a41; /* yy_accept */
	T0* a42; /* yy_content */
	T0* a43; /* yy_content_area */
	T0* a44; /* yy_unicode_content_area */
	T6 a45; /* last_token */
	T0* a46; /* input_filter */
	T0* a47; /* input_stream */
	T0* a48; /* input_resolver */
	T0* a49; /* yy_acclist */
};

/* Struct for type [detachable] XM_EIFFEL_SCANNER_DTD */
struct S557 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* last_token */
	T0* a2; /* last_value */
	T0* a3; /* input_name */
	T6 a4; /* position */
	T6 a5; /* column */
	T6 a6; /* line */
	T0* a7; /* input_filter */
	T6 a8; /* yy_start_state */
	T0* a9; /* character_entity */
	T0* a10; /* input_stream */
	T0* a11; /* input_resolver */
	T1 a12; /* decl_start_sent */
	T1 a13; /* decl_end_sent */
	T1 a14; /* yy_more_flag */
	T6 a15; /* yy_more_len */
	T6 a16; /* yy_end */
	T6 a17; /* yy_start */
	T6 a18; /* yy_line */
	T6 a19; /* yy_column */
	T6 a20; /* yy_position */
	T0* a21; /* yy_state_stack */
	T6 a22; /* yy_state_count */
	T0* a23; /* yy_content_area */
	T0* a24; /* yy_unicode_content_area */
	T0* a25; /* yy_accept */
	T6 a26; /* yy_last_accepting_state */
	T6 a27; /* yy_last_accepting_cpos */
	T0* a28; /* yy_content */
	T0* a29; /* yy_ec */
	T0* a30; /* yy_chk */
	T0* a31; /* yy_base */
	T0* a32; /* yy_def */
	T0* a33; /* yy_meta */
	T0* a34; /* yy_nxt */
	T6 a35; /* yy_lp */
	T0* a36; /* yy_acclist */
	T6 a37; /* yy_looking_for_trail_begin */
	T6 a38; /* yy_full_match */
	T6 a39; /* yy_full_state */
	T6 a40; /* yy_full_lp */
	T0* a41; /* input_buffer */
	T1 a42; /* yy_rejected */
	T0* a43; /* yy_pushed_start_conditions */
	T0* a44; /* last_string_value */
	T0* a45; /* last_error */
	T6 a46; /* pushed_start_condition_count */
};

/* Struct for type [detachable] XM_EIFFEL_PE_ENTITY_DEF */
struct S559 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* resolver */
	T0* a2; /* external_id */
	T0* a3; /* literal_name */
	T0* a4; /* value */
	T0* a5; /* character_entity */
	T0* a6; /* yy_pushed_start_conditions */
	T0* a7; /* input_buffer */
	T1 a8; /* in_use */
	T0* a9; /* input_name */
	T0* a10; /* last_value */
	T0* a11; /* last_string_value */
	T0* a12; /* last_error */
	T1 a13; /* yy_rejected */
	T6 a14; /* yy_state_count */
	T6 a15; /* yy_full_match */
	T6 a16; /* yy_lp */
	T6 a17; /* yy_looking_for_trail_begin */
	T6 a18; /* yy_full_lp */
	T6 a19; /* yy_full_state */
	T6 a20; /* pushed_start_condition_count */
	T6 a21; /* yy_start_state */
	T6 a22; /* yy_line */
	T6 a23; /* yy_column */
	T6 a24; /* yy_position */
	T6 a25; /* yy_start */
	T6 a26; /* yy_end */
	T6 a27; /* line */
	T6 a28; /* column */
	T6 a29; /* position */
	T1 a30; /* yy_more_flag */
	T6 a31; /* yy_more_len */
	T6 a32; /* yy_last_accepting_state */
	T6 a33; /* yy_last_accepting_cpos */
	T0* a34; /* yy_state_stack */
	T1 a35; /* pre_sent */
	T1 a36; /* post_sent */
	T0* a37; /* yy_nxt */
	T0* a38; /* yy_chk */
	T0* a39; /* yy_base */
	T0* a40; /* yy_def */
	T0* a41; /* yy_ec */
	T0* a42; /* yy_meta */
	T0* a43; /* yy_accept */
	T0* a44; /* yy_content */
	T0* a45; /* yy_content_area */
	T0* a46; /* yy_unicode_content_area */
	T6 a47; /* last_token */
	T0* a48; /* input_filter */
	T0* a49; /* input_stream */
	T0* a50; /* input_resolver */
	T0* a51; /* yy_acclist */
};

/* Struct for type [detachable] XM_NAMESPACE_RESOLVER */
struct S560 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next */
	T0* a2; /* element_local_part */
	T0* a3; /* element_prefix */
	T0* a4; /* context */
	T1 a5; /* forward_xmlns */
	T0* a6; /* attributes_prefix */
	T0* a7; /* attributes_local_part */
	T0* a8; /* attributes_value */
};

/* Struct for type [detachable] ARRAY [[attached] XM_CALLBACKS_FILTER] */
struct S561 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] SPECIAL [[attached] XM_CALLBACKS_FILTER] */
struct S562 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] XM_NAMESPACE] */
struct S563 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* equality_tester */
	T6 a2; /* capacity */
	T6 a3; /* modulus */
	T6 a4; /* last_position */
	T6 a5; /* free_slot */
	T6 a6; /* position */
	T0* a7; /* special_item_routines */
	T0* a8; /* item_storage */
	T0* a9; /* clashes */
	T0* a10; /* slots */
	T6 a11; /* found_position */
	T0* a12; /* internal_cursor */
	T6 a13; /* slots_position */
	T6 a14; /* count */
	T6 a15; /* clashes_previous_position */
	T0* a16; /* hash_function */
};

/* Struct for type [detachable] ET_CLUSTERS */
struct S564 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* clusters */
};

/* Struct for type [detachable] ET_CLASS_NAME_TESTER */
struct S565 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_PARENT */
struct S567 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type */
	T0* a2; /* renames */
	T0* a3; /* exports */
	T0* a4; /* undefines */
	T0* a5; /* redefines */
	T0* a6; /* selects */
	T0* a7; /* end_keyword */
};

/* Struct for type [detachable] ET_PARENT_LIST */
struct S568 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* inherit_keyword */
	T6 a2; /* count */
	T0* a3; /* storage */
	T0* a4; /* clients_clause */
};

/* Struct for type [detachable] ET_BUILTIN_CONVERT_FEATURE */
struct S569 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* types */
	T0* a3; /* type */
};

/* Struct for type [detachable] ET_MASTER_CLASS */
struct S570 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* mapped_class */
	T0* a2; /* first_overriding_class */
	T0* a3; /* intrinsic_class */
	T0* a4; /* name */
	T0* a5; /* universe */
	T0* a6; /* status_mutex */
	T0* a7; /* processing_mutex */
	T1 a8; /* unprotected_is_marked */
	T0* a9; /* first_imported_class */
	T0* a10; /* other_imported_classes */
	T0* a11; /* first_local_override_class */
	T0* a12; /* first_local_non_override_class */
	T0* a13; /* first_local_ignored_class */
	T0* a14; /* first_local_hidden_class */
	T1 a15; /* is_modified */
	T0* a16; /* other_local_override_classes */
	T0* a17; /* other_local_non_override_classes */
	T0* a18; /* other_local_ignored_classes */
	T0* a19; /* other_overriding_classes */
	T0* a20; /* other_local_hidden_classes */
};

/* Struct for type [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD */
struct S571 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* attachment_mark */
	T0* a2; /* separateness_keyword */
};

/* Struct for type detachable ET_RENAME_LIST */
struct S572 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* rename_keyword */
	T6 a2; /* count */
	T0* a3; /* storage */
};

/* Struct for type detachable ET_EXPORT_LIST */
struct S573 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* export_keyword */
	T6 a2; /* count */
	T0* a3; /* storage */
};

/* Struct for type detachable ET_KEYWORD_FEATURE_NAME_LIST */
struct S574 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* keyword */
	T6 a2; /* count */
	T0* a3; /* storage */
};

/* Struct for type [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS */
struct S577 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* tuple_type */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY] */
struct S578 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE] */
struct S579 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* last_cell */
	T0* a3; /* first_cell */
	T0* a4; /* internal_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_SYSTEM_CONFIG] */
struct S582 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_SYSTEM_CONFIG] */
struct S584 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8] */
struct S585 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_TARGET, [attached] STRING_8] */
struct S586 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_TARGET] */
struct S587 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET] */
struct S589 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
struct S591 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* table */
	T0* a2; /* equality_tester */
	T0* a3; /* internal_cursor */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable RX_REGULAR_EXPRESSION] */
struct S592 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION] */
struct S593 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
struct S594 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] ET_DOTNET_ASSEMBLIES */
struct S595 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* dotnet_assemblies */
};

/* Struct for type [detachable] ET_DOTNET_ASSEMBLY */
struct S596 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* pathname */
	T0* a2; /* referenced_assemblies */
	T0* a3; /* current_system */
	T0* a4; /* dotnet_assembly */
	T0* a5; /* name */
	T0* a6; /* implicit_attachment_type_mark */
	T0* a7; /* master_classes */
	T0* a8; /* master_class_mutex */
	T0* a9; /* any_type */
	T0* a10; /* detachable_any_type */
	T0* a11; /* detachable_separate_any_type */
	T0* a12; /* any_parent */
	T0* a13; /* any_parents */
	T0* a14; /* any_clients */
	T0* a15; /* tuple_type */
	T0* a16; /* detachable_tuple_type */
	T0* a17; /* tuple_identity_type */
	T0* a18; /* unfolded_empty_tuple_actual_parameters */
	T0* a19; /* array_any_type */
	T0* a20; /* array_detachable_any_type */
	T0* a21; /* array_none_type */
	T0* a22; /* array_identity_type */
	T0* a23; /* boolean_type */
	T0* a24; /* character_type */
	T0* a25; /* character_8_type */
	T0* a26; /* character_8_convert_feature */
	T0* a27; /* character_32_type */
	T0* a28; /* character_32_convert_feature */
	T0* a29; /* double_type */
	T0* a30; /* exception_type */
	T0* a31; /* detachable_exception_type */
	T0* a32; /* exception_manager_type */
	T0* a33; /* function_identity_any_type */
	T0* a34; /* immutable_string_8_type */
	T0* a35; /* immutable_string_32_type */
	T0* a36; /* integer_type */
	T0* a37; /* integer_8_type */
	T0* a38; /* integer_8_convert_feature */
	T0* a39; /* integer_16_type */
	T0* a40; /* integer_16_convert_feature */
	T0* a41; /* integer_32_type */
	T0* a42; /* integer_32_convert_feature */
	T0* a43; /* integer_64_type */
	T0* a44; /* integer_64_convert_feature */
	T0* a45; /* ise_exception_manager_type */
	T0* a46; /* iterable_detachable_separate_any_type */
	T0* a47; /* natural_type */
	T0* a48; /* natural_8_type */
	T0* a49; /* natural_8_convert_feature */
	T0* a50; /* natural_16_type */
	T0* a51; /* natural_16_convert_feature */
	T0* a52; /* natural_32_type */
	T0* a53; /* natural_32_convert_feature */
	T0* a54; /* natural_64_type */
	T0* a55; /* natural_64_convert_feature */
	T0* a56; /* none_type */
	T0* a57; /* detachable_none_type */
	T0* a58; /* pointer_type */
	T0* a59; /* predicate_identity_type */
	T0* a60; /* procedure_identity_type */
	T0* a61; /* real_type */
	T0* a62; /* real_32_type */
	T0* a63; /* real_32_convert_feature */
	T0* a64; /* real_64_type */
	T0* a65; /* real_64_convert_feature */
	T0* a66; /* routine_identity_type */
	T0* a67; /* special_any_type */
	T0* a68; /* special_detachable_any_type */
	T0* a69; /* special_identity_type */
	T0* a70; /* string_type */
	T0* a71; /* detachable_string_type */
	T0* a72; /* string_8_type */
	T0* a73; /* detachable_string_8_type */
	T0* a74; /* string_8_convert_feature */
	T0* a75; /* string_32_type */
	T0* a76; /* string_32_convert_feature */
	T0* a77; /* system_object_type */
	T0* a78; /* system_object_parents */
	T0* a79; /* system_string_type */
	T0* a80; /* type_detachable_any_type */
	T0* a81; /* detachable_type_exception_type */
	T0* a82; /* type_detachable_exception_type */
	T0* a83; /* type_detachable_like_current_type */
	T0* a84; /* type_identity_type */
	T0* a85; /* typed_pointer_identity_type */
	T0* a86; /* wide_character_type */
	T1 a87; /* is_read_only */
	T1 a88; /* is_preparsed */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_IDENTIFIER] */
struct S599 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_IDENTIFIER] */
struct S600 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8] */
struct S601 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32] */
struct S602 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* modulus */
	T0* a2; /* key_equality_tester */
	T0* a3; /* item_storage */
	T0* a4; /* slots */
	T0* a5; /* hash_function */
	T0* a6; /* key_storage */
	T0* a7; /* clashes */
	T6 a8; /* position */
	T6 a9; /* last_position */
	T6 a10; /* capacity */
	T6 a11; /* slots_position */
	T6 a12; /* count */
	T0* a13; /* equality_tester */
	T6 a14; /* found_position */
	T6 a15; /* clashes_previous_position */
	T0* a16; /* special_item_routines */
	T0* a17; /* special_key_routines */
	T6 a18; /* free_slot */
	T0* a19; /* internal_cursor */
};

/* Struct for type detachable ET_FORMAL_PARAMETER_LIST */
struct S603 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* left_bracket */
	T0* a4; /* right_bracket */
};

/* Struct for type detachable ET_FEATURE_IDS */
struct S604 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* feature_ids */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_DYNAMIC_FEATURE] */
struct S605 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_FEATURE] */
struct S606 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME] */
struct S609 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MASTER_CLASS] */
struct S610 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_MASTER_CLASS] */
struct S611 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS_NAME] */
struct S612 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CLASS_NAME] */
struct S613 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8] */
struct S615 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME] */
struct S616 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* modulus */
	T0* a2; /* key_equality_tester */
	T0* a3; /* item_storage */
	T0* a4; /* slots */
	T0* a5; /* hash_function */
	T0* a6; /* key_storage */
	T0* a7; /* clashes */
	T6 a8; /* position */
	T6 a9; /* count */
	T6 a10; /* capacity */
	T6 a11; /* slots_position */
	T6 a12; /* free_slot */
	T6 a13; /* last_position */
	T0* a14; /* equality_tester */
	T6 a15; /* found_position */
	T6 a16; /* clashes_previous_position */
	T0* a17; /* special_item_routines */
	T0* a18; /* special_key_routines */
	T0* a19; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME] */
struct S617 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* equality_tester */
	T0* a2; /* key_equality_tester */
	T6 a3; /* capacity */
	T6 a4; /* modulus */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T0* a8; /* special_item_routines */
	T0* a9; /* item_storage */
	T0* a10; /* special_key_routines */
	T0* a11; /* key_storage */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T6 a14; /* found_position */
	T0* a15; /* internal_cursor */
	T0* a16; /* hash_function */
	T6 a17; /* count */
	T6 a18; /* slots_position */
	T6 a19; /* clashes_previous_position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ACTUAL_PARAMETER_ITEM] */
struct S621 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] ET_BRACKET_SYMBOL */
struct S622 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* compressed_position */
	T2 a2; /* code */
	T0* a3; /* break */
	T6 a4; /* seed */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACTUAL_PARAMETER_ITEM] */
struct S624 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE] */
struct S625 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE_SET] */
struct S626 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE_SET] */
struct S627 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_IDENTIFIER] */
struct S628 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [detachable ET_DYNAMIC_PRIMARY_TYPE] */
struct S629 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_PRIMARY_TYPE] */
struct S630 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_DYNAMIC_PRIMARY_TYPE] */
struct S631 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] INTEGER_32] */
struct S632 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_UNQUALIFIED_CALL_EXPRESSION] */
struct S634 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_UNQUALIFIED_CALL_EXPRESSION */
struct S635 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* name */
	T0* a3; /* arguments */
	T0* a4; /* parenthesis_call */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_UNQUALIFIED_CALL_EXPRESSION] */
struct S636 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_UNQUALIFIED_CALL_EXPRESSION] */
struct S637 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_QUALIFIED_CALL_EXPRESSION] */
struct S638 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_QUALIFIED_CALL_EXPRESSION] */
struct S639 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUALIFIED_CALL_EXPRESSION] */
struct S640 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACTUAL_ARGUMENT_LIST] */
struct S641 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ACTUAL_ARGUMENT_LIST] */
struct S642 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ACTUAL_ARGUMENT_LIST] */
struct S643 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RESULT] */
struct S644 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_RESULT] */
struct S645 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_RESULT] */
struct S646 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET] */
struct S647 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET] */
struct S648 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET] */
struct S649 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_EXPRESSION] */
struct S650 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION] */
struct S651 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EXPRESSION] */
struct S652 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type detachable KL_EQUALITY_TESTER [[attached] INTEGER_32] */
struct S653 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32] */
struct S654 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S655 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_OBJECT_TEST] */
struct S657 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST] */
struct S658 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_TEST] */
struct S659 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
struct S660 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
struct S661 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
struct S662 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_EQUALITY_EXPRESSION] */
struct S663 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EQUALITY_EXPRESSION] */
struct S664 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EQUALITY_EXPRESSION] */
struct S665 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_AGENT] */
struct S666 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT] */
struct S667 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_AGENT] */
struct S668 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_EXPRESSION_ITEM] */
struct S669 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION_ITEM] */
struct S671 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_TUPLE_TYPE] */
struct S672 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_DYNAMIC_TUPLE_TYPE] */
struct S673 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TUPLE_TYPE] */
struct S674 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_FEATURE] */
struct S676 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_FEATURE] */
struct S677 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE] */
struct S678 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CONSTANT, [attached] ET_FEATURE] */
struct S679 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CONSTANT] */
struct S680 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTANT] */
struct S682 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT] */
struct S683 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_INLINE_CONSTANT] */
struct S684 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_CONSTANT] */
struct S686 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_DYNAMIC_FEATURE] */
struct S687 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type detachable DS_HASH_TABLE_CURSOR [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S689 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_FEATURE] */
struct S690 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_FEATURE] */
struct S691 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] STRING_8] */
struct S692 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_IDENTIFIER] */
struct S695 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S698 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T6 a2; /* found_position */
	T0* a3; /* item_storage */
	T6 a4; /* modulus */
	T6 a5; /* last_position */
	T6 a6; /* capacity */
	T6 a7; /* free_slot */
	T6 a8; /* count */
	T6 a9; /* slots_position */
	T6 a10; /* clashes_previous_position */
	T0* a11; /* key_equality_tester */
	T0* a12; /* equality_tester */
	T0* a13; /* slots */
	T0* a14; /* clashes */
	T0* a15; /* special_item_routines */
	T0* a16; /* key_storage */
	T0* a17; /* special_key_routines */
	T0* a18; /* internal_cursor */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]] */
struct S699 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* storage */
	T0* a2; /* special_routines */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] ARRAY [[attached] BOOLEAN] */
struct S700 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_SYSTEM_PROCESSOR] */
struct S703 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] TUPLE [[attached] ET_SYSTEM_PROCESSOR, [attached] DS_ARRAYED_LIST [[attached] ET_CLASS]] */
struct S704 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE] */
struct S707 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE] */
struct S708 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE] */
struct S709 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS] */
struct S710 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_AGENT] */
struct S711 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_INLINE_AGENT] */
struct S712 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INLINE_AGENT] */
struct S713 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
struct S714 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE] */
struct S715 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
struct S716 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
struct S717 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
struct S718 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY] */
struct S719 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
struct S720 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
struct S721 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
struct S722 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE] */
struct S723 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* equality_tester */
	T0* a4; /* special_routines */
	T6 a5; /* capacity */
	T0* a6; /* internal_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
struct S724 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
struct S725 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
struct S726 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT] */
struct S727 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
};

/* Struct for type [detachable] SPECIAL [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
struct S728 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
struct S729 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_NESTED_TYPE_CONTEXT] */
struct S730 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NESTED_TYPE_CONTEXT] */
struct S731 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT] */
struct S732 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST] */
struct S733 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] ET_NAMED_OBJECT_TEST */
struct S736 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* name */
	T0* a3; /* attached_keyword */
	T0* a4; /* declared_type */
	T0* a5; /* expression */
	T0* a6; /* as_keyword */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_NAMED_OBJECT_TEST] */
struct S737 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_OBJECT_TEST] */
struct S738 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST] */
struct S739 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* special_routines */
	T0* a3; /* storage */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT] */
struct S740 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ITERATION_COMPONENT] */
struct S741 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ITERATION_COMPONENT] */
struct S743 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ITERATION_COMPONENT] */
struct S744 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] INTEGER_32] */
struct S745 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T6 a2; /* position */
	T0* a3; /* item_storage */
	T6 a4; /* capacity */
	T6 a5; /* modulus */
	T6 a6; /* last_position */
	T6 a7; /* free_slot */
	T6 a8; /* count */
	T6 a9; /* slots_position */
	T6 a10; /* clashes_previous_position */
	T0* a11; /* special_item_routines */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T6 a14; /* found_position */
	T0* a15; /* equality_tester */
	T0* a16; /* hash_function */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ATTACHMENT_SCOPE] */
struct S747 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ATTACHMENT_SCOPE] */
struct S748 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ATTACHMENT_SCOPE] */
struct S749 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ASSERTIONS, [attached] ET_FEATURE] */
struct S750 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ASSERTIONS] */
struct S751 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTIONS] */
struct S753 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM] */
struct S754 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_INDEXING_TERM] */
struct S756 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INDEXING_TERM] */
struct S757 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CLIENT_ITEM] */
struct S758 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_ITEM] */
struct S759 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
struct S760 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
struct S761 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
struct S762 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME] */
struct S763 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T6 a3; /* position */
	T0* a4; /* key_equality_tester */
	T0* a5; /* internal_keys */
	T6 a6; /* last_position */
	T6 a7; /* capacity */
	T6 a8; /* slots_position */
	T6 a9; /* count */
	T0* a10; /* equality_tester */
	T6 a11; /* modulus */
	T6 a12; /* clashes_previous_position */
	T0* a13; /* special_item_routines */
	T0* a14; /* clashes */
	T0* a15; /* slots */
	T0* a16; /* special_key_routines */
	T0* a17; /* key_storage */
	T6 a18; /* free_slot */
	T0* a19; /* internal_cursor */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] SPECIAL [[attached] NATURAL_32] */
struct S764 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T10 z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET] */
struct S765 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_32] */
struct S766 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] NATURAL_64] */
struct S767 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T11 z2[1]; /* item */
};

/* Struct for type detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32] */
struct S768 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* position */
	T0* a3; /* internal_cursor */
	T6 a4; /* found_position */
	T0* a5; /* item_storage */
	T6 a6; /* modulus */
	T0* a7; /* key_equality_tester */
	T0* a8; /* key_storage */
	T0* a9; /* slots */
	T0* a10; /* hash_function */
	T0* a11; /* clashes */
	T0* a12; /* internal_keys */
	T6 a13; /* capacity */
	T6 a14; /* free_slot */
	T6 a15; /* last_position */
	T6 a16; /* slots_position */
	T6 a17; /* clashes_previous_position */
	T0* a18; /* equality_tester */
	T0* a19; /* special_item_routines */
	T0* a20; /* special_key_routines */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32] */
struct S769 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]] */
struct S770 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] SPECIAL [[attached] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]]] */
struct S771 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_INTEGER_ROUTINES */
struct S772 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] UC_UTF8_STRING */
struct S775 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* area */
	T6 a3; /* byte_count */
	T6 a4; /* internal_hash_code */
	T6 a5; /* last_byte_index_input */
	T6 a6; /* last_byte_index_result */
	T6 a7; /* internal_case_insensitive_hash_code */
};

/* Struct for type [detachable] THREAD_ATTRIBUTES */
struct S776 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* shared */
	T14 a2; /* internal_item */
	T0* a3; /* managed_pointer */
};

/* Struct for type [detachable] CELL [[attached] BOOLEAN] */
struct S777 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* item */
};

/* Struct for type [detachable] CHARACTER_PROPERTY */
struct S780 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] STRING_8_SEARCHER */
struct S781 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* deltas */
};

/* Struct for type detachable ARRAYED_LIST [[attached] INTEGER_32] */
struct S782 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area_v2 */
	T6 a2; /* index */
};

/* Struct for type [detachable] ET_SYSTEM_ERROR */
struct S783 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* default_template */
	T0* a2; /* parameters */
	T0* a3; /* code */
	T0* a4; /* etl_code */
};

/* Struct for type [detachable] ET_INTERNAL_ERROR */
struct S784 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* default_template */
	T0* a2; /* parameters */
	T0* a3; /* code */
	T0* a4; /* etl_code */
};

/* Struct for type [detachable] ET_VALIDITY_ERROR */
struct S785 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* ise_reported */
	T1 a2; /* ge_reported */
	T0* a3; /* default_template */
	T0* a4; /* parameters */
	T0* a5; /* current_class */
	T0* a6; /* class_impl */
	T0* a7; /* position */
	T0* a8; /* code */
	T0* a9; /* etl_code */
	T1 a10; /* ise_fatal */
	T1 a11; /* ge_fatal */
};

/* Struct for type detachable ET_ECF_NOTE_ELEMENT */
struct S786 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* attributes */
	T0* a3; /* elements */
	T0* a4; /* content */
};

/* Struct for type detachable ET_ECF_CLUSTERS */
struct S787 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* clusters */
};

/* Struct for type detachable ET_ECF_ADAPTED_LIBRARIES */
struct S788 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* libraries */
};

/* Struct for type detachable ET_ECF_ADAPTED_DOTNET_ASSEMBLIES */
struct S789 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* dotnet_assemblies */
};

/* Struct for type detachable ET_ECF_FILE_RULES */
struct S790 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* file_rules */
};

/* Struct for type detachable ET_ECF_EXTERNAL_CFLAGS */
struct S791 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* external_values */
};

/* Struct for type detachable ET_ECF_EXTERNAL_INCLUDES */
struct S792 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* external_values */
};

/* Struct for type detachable ET_ECF_EXTERNAL_LIBRARIES */
struct S793 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* external_values */
};

/* Struct for type detachable ET_ECF_EXTERNAL_LINKER_FLAGS */
struct S794 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* external_values */
};

/* Struct for type detachable ET_ECF_EXTERNAL_MAKES */
struct S795 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* external_values */
};

/* Struct for type detachable ET_ECF_EXTERNAL_OBJECTS */
struct S796 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* external_values */
};

/* Struct for type detachable ET_ECF_EXTERNAL_RESOURCES */
struct S797 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* external_values */
};

/* Struct for type detachable DS_ARRAYED_LIST [[attached] ET_ECF_ACTION] */
struct S798 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY */
struct S799 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name_id */
	T0* a2; /* filename_id */
	T0* a3; /* universe */
	T0* a4; /* target */
	T0* a5; /* description */
	T0* a6; /* eifgens_location */
	T0* a7; /* classname_prefix */
	T1 a8; /* is_read_only */
	T1 a9; /* use_application_options */
	T0* a10; /* options */
	T0* a11; /* class_options */
	T0* a12; /* class_renamings */
	T0* a13; /* visible_classes */
	T0* a14; /* conditions */
	T0* a15; /* name */
	T0* a16; /* library */
};

/* Struct for type detachable ET_ECF_VERSION */
struct S801 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* internal_major */
	T6 a2; /* internal_minor */
	T6 a3; /* internal_release */
	T6 a4; /* internal_build */
	T0* a5; /* product */
	T0* a6; /* company */
	T0* a7; /* copyright */
};

/* Struct for type detachable ET_ECF_CLUSTER */
struct S802 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* pathname */
	T1 a3; /* is_relative */
	T0* a4; /* universe */
	T0* a5; /* target */
	T0* a6; /* description */
	T1 a7; /* is_hidden */
	T0* a8; /* classname_prefix */
	T1 a9; /* is_read_only */
	T1 a10; /* is_recursive */
	T0* a11; /* conditioned_subclusters */
	T0* a12; /* conditioned_file_rules */
	T0* a13; /* options */
	T0* a14; /* class_options */
	T0* a15; /* class_renamings */
	T0* a16; /* class_mappings */
	T0* a17; /* visible_classes */
	T0* a18; /* provider_groups */
	T0* a19; /* conditions */
	T0* a20; /* notes */
	T1 a21; /* is_override */
	T0* a22; /* overridden_group */
	T1 a23; /* overridden_constraint_enabled */
	T1 a24; /* scm_mapping_constraint_enabled */
	T0* a25; /* subclusters */
	T1 a26; /* use_obsolete_syntax */
	T0* a27; /* scm_read_mapping */
	T0* a28; /* scm_write_mapping */
	T1 a29; /* is_implicit */
	T0* a30; /* file_rules */
	T0* a31; /* parent */
	T0* a32; /* provider_constraint */
	T0* a33; /* dependant_constraint */
	T1 a34; /* is_preparsed */
	T1 a35; /* is_abstract */
	T0* a36; /* cached_absolute_pathname */
};

/* Struct for type detachable ET_ECF_EXTERNAL_CFLAG */
struct S803 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* flag */
	T0* a2; /* description */
	T0* a3; /* conditions */
};

/* Struct for type detachable ET_ECF_EXTERNAL_INCLUDE */
struct S804 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* pathname */
	T0* a2; /* description */
	T0* a3; /* conditions */
};

/* Struct for type detachable ET_ECF_EXTERNAL_LIBRARY */
struct S805 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* pathname */
	T0* a2; /* description */
	T0* a3; /* conditions */
};

/* Struct for type detachable ET_ECF_EXTERNAL_LINKER_FLAG */
struct S806 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* flag */
	T0* a2; /* description */
	T0* a3; /* conditions */
};

/* Struct for type detachable ET_ECF_EXTERNAL_MAKE */
struct S807 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* pathname */
	T0* a2; /* description */
	T0* a3; /* conditions */
};

/* Struct for type detachable ET_ECF_EXTERNAL_OBJECT */
struct S808 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* pathname */
	T0* a2; /* description */
	T0* a3; /* conditions */
};

/* Struct for type detachable ET_ECF_EXTERNAL_RESOURCE */
struct S809 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* pathname */
	T0* a2; /* description */
	T0* a3; /* conditions */
};

/* Struct for type detachable ET_ECF_FILE_RULE */
struct S810 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* exclude */
	T0* a2; /* include */
	T0* a3; /* exclude_regexp */
	T0* a4; /* include_regexp */
	T0* a5; /* conditions */
	T0* a6; /* description */
};

/* Struct for type detachable ET_ECF_ACTION */
struct S811 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* command_name */
	T0* a2; /* description */
	T1 a3; /* must_succeed */
	T0* a4; /* working_directory */
	T0* a5; /* conditions */
};

/* Struct for type detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8] */
struct S812 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T6 a3; /* last_position */
	T6 a4; /* position */
	T6 a5; /* capacity */
	T6 a6; /* slots_position */
	T6 a7; /* count */
	T0* a8; /* key_equality_tester */
	T0* a9; /* internal_keys */
	T0* a10; /* equality_tester */
	T6 a11; /* modulus */
	T6 a12; /* clashes_previous_position */
	T0* a13; /* special_item_routines */
	T0* a14; /* clashes */
	T0* a15; /* slots */
	T0* a16; /* special_key_routines */
	T0* a17; /* key_storage */
	T6 a18; /* free_slot */
	T0* a19; /* internal_cursor */
	T0* a20; /* hash_function */
};

/* Struct for type detachable DS_ARRAYED_LIST [[attached] ET_ECF_VISIBLE_CLASS] */
struct S813 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_OPTIONS] */
struct S814 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_OPTIONS]] */
struct S815 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type detachable ET_ECF_ROOT_CLASS */
struct S816 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* class_name */
	T0* a2; /* creation_procedure_name */
	T0* a3; /* cluster_name */
};

/* Struct for type [detachable] ET_ECF_ROOT_ALL_CLASSES */
struct S817 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_ECF_ORED_CONDITIONS */
struct S818 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* conditions */
};

/* Struct for type detachable ET_ECF_ANDED_CONDITIONS */
struct S819 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* conditions */
};

/* Struct for type detachable ET_ECF_VISIBLE_CLASS */
struct S820 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* class_name */
	T0* a2; /* feature_name */
	T0* a3; /* new_class_name */
	T0* a4; /* new_feature_name */
};

/* Struct for type detachable ET_ECF_BUILD_CONDITION */
struct S822 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* value */
	T1 a2; /* is_excluded */
};

/* Struct for type detachable ET_ECF_CONCURRENCY_CONDITION */
struct S823 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* value */
	T1 a2; /* is_excluded */
};

/* Struct for type detachable ET_ECF_VOID_SAFETY_CONDITION */
struct S824 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* value */
	T1 a2; /* is_excluded */
};

/* Struct for type detachable ET_ECF_CUSTOM_CONDITION */
struct S825 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* value */
	T0* a3; /* match */
	T1 a4; /* is_excluded */
};

/* Struct for type detachable ET_ECF_DOTNET_CONDITION */
struct S826 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* value */
};

/* Struct for type detachable ET_ECF_DYNAMIC_RUNTIME_CONDITION */
struct S827 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* value */
};

/* Struct for type detachable ET_ECF_PLATFORM_CONDITION */
struct S828 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* value */
	T1 a2; /* is_excluded */
};

/* Struct for type [detachable] ET_ECF_COMPILER_VERSION_CONDITION */
struct S830 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* min_value */
	T0* a2; /* max_value */
};

/* Struct for type [detachable] ET_ECF_MSIL_CLR_VERSION_CONDITION */
struct S831 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* min_value */
	T0* a2; /* max_value */
};

/* Struct for type [detachable] UT_COUNTER */
struct S833 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* item */
};

/* Struct for type [detachable] KL_AGENT_ROUTINES [[attached] ET_CLASS] */
struct S834 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] TUPLE [[attached] UT_COUNTER] */
struct S835 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] KL_AGENT_ROUTINES [[attached] ET_CLASS], [attached] PROCEDURE [[attached] TUPLE]] */
struct S836 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] TUPLE [[attached] ET_UNIVERSE] */
struct S837 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_UNIVERSE]] */
struct S838 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]]] */
struct S839 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]], [attached] FUNCTION [[attached] TUPLE, [attached] BOOLEAN]] */
struct S840 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY] */
struct S842 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* capacity */
	T6 a2; /* modulus */
	T6 a3; /* last_position */
	T6 a4; /* free_slot */
	T6 a5; /* position */
	T6 a6; /* slots_position */
	T6 a7; /* count */
	T0* a8; /* special_item_routines */
	T0* a9; /* item_storage */
	T0* a10; /* clashes */
	T0* a11; /* slots */
	T6 a12; /* found_position */
	T0* a13; /* internal_cursor */
	T6 a14; /* clashes_previous_position */
	T0* a15; /* hash_function */
	T0* a16; /* equality_tester */
};

/* Struct for type [detachable] TUPLE [[attached] ET_DOTNET_ASSEMBLY] */
struct S843 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PREDICATE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]] */
struct S844 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]] */
struct S845 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY]] */
struct S846 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] ET_INTERNAL_UNIVERSE] */
struct S847 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_INTERNAL_UNIVERSE]] */
struct S848 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]], [attached] PREDICATE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]]] */
struct S849 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] TUPLE [[attached] ET_DOTNET_ASSEMBLIES] */
struct S850 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] ET_SYSTEM_PROCESSOR] */
struct S851 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] ET_LIBRARY */
struct S852 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* current_system */
	T0* a2; /* library */
	T0* a3; /* name */
	T0* a4; /* libraries */
	T0* a5; /* clusters */
	T0* a6; /* dotnet_assemblies */
	T0* a7; /* implicit_attachment_type_mark */
	T0* a8; /* master_classes */
	T0* a9; /* master_class_mutex */
	T0* a10; /* any_type */
	T0* a11; /* detachable_any_type */
	T0* a12; /* detachable_separate_any_type */
	T0* a13; /* any_parent */
	T0* a14; /* any_parents */
	T0* a15; /* any_clients */
	T0* a16; /* tuple_type */
	T0* a17; /* detachable_tuple_type */
	T0* a18; /* tuple_identity_type */
	T0* a19; /* unfolded_empty_tuple_actual_parameters */
	T0* a20; /* array_any_type */
	T0* a21; /* array_detachable_any_type */
	T0* a22; /* array_none_type */
	T0* a23; /* array_identity_type */
	T0* a24; /* boolean_type */
	T0* a25; /* character_type */
	T0* a26; /* character_8_type */
	T0* a27; /* character_8_convert_feature */
	T0* a28; /* character_32_type */
	T0* a29; /* character_32_convert_feature */
	T0* a30; /* double_type */
	T0* a31; /* exception_type */
	T0* a32; /* detachable_exception_type */
	T0* a33; /* exception_manager_type */
	T0* a34; /* function_identity_any_type */
	T0* a35; /* immutable_string_8_type */
	T0* a36; /* immutable_string_32_type */
	T0* a37; /* integer_type */
	T0* a38; /* integer_8_type */
	T0* a39; /* integer_8_convert_feature */
	T0* a40; /* integer_16_type */
	T0* a41; /* integer_16_convert_feature */
	T0* a42; /* integer_32_type */
	T0* a43; /* integer_32_convert_feature */
	T0* a44; /* integer_64_type */
	T0* a45; /* integer_64_convert_feature */
	T0* a46; /* ise_exception_manager_type */
	T0* a47; /* iterable_detachable_separate_any_type */
	T0* a48; /* natural_type */
	T0* a49; /* natural_8_type */
	T0* a50; /* natural_8_convert_feature */
	T0* a51; /* natural_16_type */
	T0* a52; /* natural_16_convert_feature */
	T0* a53; /* natural_32_type */
	T0* a54; /* natural_32_convert_feature */
	T0* a55; /* natural_64_type */
	T0* a56; /* natural_64_convert_feature */
	T0* a57; /* none_type */
	T0* a58; /* detachable_none_type */
	T0* a59; /* pointer_type */
	T0* a60; /* predicate_identity_type */
	T0* a61; /* procedure_identity_type */
	T0* a62; /* real_type */
	T0* a63; /* real_32_type */
	T0* a64; /* real_32_convert_feature */
	T0* a65; /* real_64_type */
	T0* a66; /* real_64_convert_feature */
	T0* a67; /* routine_identity_type */
	T0* a68; /* special_any_type */
	T0* a69; /* special_detachable_any_type */
	T0* a70; /* special_identity_type */
	T0* a71; /* string_type */
	T0* a72; /* detachable_string_type */
	T0* a73; /* string_8_type */
	T0* a74; /* detachable_string_8_type */
	T0* a75; /* string_8_convert_feature */
	T0* a76; /* string_32_type */
	T0* a77; /* string_32_convert_feature */
	T0* a78; /* system_object_type */
	T0* a79; /* system_object_parents */
	T0* a80; /* system_string_type */
	T0* a81; /* type_detachable_any_type */
	T0* a82; /* detachable_type_exception_type */
	T0* a83; /* type_detachable_exception_type */
	T0* a84; /* type_detachable_like_current_type */
	T0* a85; /* type_identity_type */
	T0* a86; /* typed_pointer_identity_type */
	T0* a87; /* wide_character_type */
	T1 a88; /* is_read_only */
	T1 a89; /* is_preparsed */
};

/* Struct for type [detachable] TUPLE [[attached] ET_LIBRARY] */
struct S853 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_LIBRARY]] */
struct S854 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_UNIVERSE] */
struct S855 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T6 a2; /* position */
	T0* a3; /* item_storage */
	T6 a4; /* capacity */
	T6 a5; /* modulus */
	T6 a6; /* last_position */
	T6 a7; /* free_slot */
	T6 a8; /* slots_position */
	T6 a9; /* count */
	T6 a10; /* clashes_previous_position */
	T0* a11; /* special_item_routines */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T6 a14; /* found_position */
	T0* a15; /* hash_function */
	T0* a16; /* equality_tester */
};

/* Struct for type [detachable] TUPLE [[attached] ET_CLUSTER] */
struct S857 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_CLUSTER]] */
struct S858 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_SYSTEM] */
struct S859 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE] */
struct S860 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T6 a2; /* capacity */
	T6 a3; /* modulus */
	T6 a4; /* last_position */
	T6 a5; /* free_slot */
	T6 a6; /* slots_position */
	T6 a7; /* count */
	T6 a8; /* clashes_previous_position */
	T0* a9; /* special_item_routines */
	T0* a10; /* item_storage */
	T0* a11; /* clashes */
	T0* a12; /* slots */
	T6 a13; /* found_position */
	T0* a14; /* internal_cursor */
	T0* a15; /* equality_tester */
	T0* a16; /* hash_function */
};

/* Struct for type [detachable] TUPLE [[attached] ET_AST_PROCESSOR] */
struct S861 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] ET_MASTER_CLASS] */
struct S862 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]] */
struct S863 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
struct S864 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ADAPTED_DOTNET_ASSEMBLY]] */
struct S865 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] ET_ANCESTORS_STATUS_CHECKER */
struct S866 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* class_type_checker */
	T0* a2; /* current_class */
	T0* a3; /* system_processor */
};

/* Struct for type [detachable] ET_FLATTENING_STATUS_CHECKER */
struct S867 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* class_type_checker */
	T0* a2; /* current_class */
	T0* a3; /* system_processor */
};

/* Struct for type [detachable] ET_INTERFACE_STATUS_CHECKER */
struct S868 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* qualified_anchored_type_checker */
	T0* a2; /* class_type_checker */
	T0* a3; /* current_class */
	T0* a4; /* system_processor */
};

/* Struct for type [detachable] ET_IMPLEMENTATION_STATUS_CHECKER */
struct S869 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* current_class */
	T0* a2; /* system_processor */
};

/* Struct for type [detachable] PREDICATE [[attached] TUPLE [[attached] ET_CLASS]] */
struct S870 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ANCESTORS_STATUS_CHECKER] */
struct S872 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] ET_FLATTENING_STATUS_CHECKER] */
struct S873 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] ET_INTERFACE_STATUS_CHECKER] */
struct S874 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] ET_IMPLEMENTATION_STATUS_CHECKER] */
struct S875 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [BOOLEAN] */
struct S876 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 z1;
};

/* Struct for type [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_UNIVERSE]] */
struct S877 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLUSTER]]] */
struct S880 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE]] */
struct S884 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]]] */
struct S885 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]], [attached] FUNCTION [[attached] TUPLE [[attached] ET_CLASS], [attached] BOOLEAN]] */
struct S886 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] ET_NONE_GROUP */
struct S887 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* universe */
	T0* a2; /* name */
	T1 a3; /* use_obsolete_syntax */
	T0* a4; /* cached_absolute_pathname */
};

/* Struct for type detachable ET_PARENTHESIS_EXPRESSION */
struct S888 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target */
	T0* a2; /* name */
	T0* a3; /* arguments */
	T6 a4; /* index */
};

/* Struct for type detachable DS_ARRAYED_LIST [detachable ET_FORMAL_PARAMETER_TYPE] */
struct S890 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* a3; /* storage */
	T0* a4; /* special_routines */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] ET_FORMAL_PARAMETER */
struct S891 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T6 a2; /* index */
	T0* a3; /* implementation_class */
	T0* a4; /* type_mark */
};

/* Struct for type [detachable] STD_FILES */
struct S892 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* default_output */
};

/* Struct for type [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE] */
struct S894 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T0* a2; /* last_cell */
	T6 a3; /* count */
	T0* a4; /* first_cell */
};

/* Struct for type [detachable] UC_STRING_EQUALITY_TESTER */
struct S896 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]] */
struct S898 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* after */
	T0* a2; /* current_cell */
	T0* a3; /* container */
	T1 a4; /* before */
	T0* a5; /* next_cursor */
};

/* Struct for type [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]] */
struct S899 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T0* a2; /* last_cell */
	T6 a3; /* count */
	T0* a4; /* first_cell */
};

/* Struct for type [detachable] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE] */
struct S900 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* first */
	T0* a2; /* second */
};

/* Struct for type [detachable] XM_EIFFEL_INPUT_STREAM */
struct S901 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* last_string */
	T6 a2; /* encoding */
	T0* a3; /* impl */
	T0* a4; /* utf_queue */
};

/* Struct for type [detachable] KL_ARRAY_ROUTINES [[attached] INTEGER_32] */
struct S902 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_LIBRARY] */
struct S903 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* capacity */
	T6 a2; /* modulus */
	T6 a3; /* last_position */
	T6 a4; /* free_slot */
	T6 a5; /* position */
	T0* a6; /* special_item_routines */
	T0* a7; /* item_storage */
	T0* a8; /* clashes */
	T0* a9; /* slots */
	T6 a10; /* found_position */
	T0* a11; /* internal_cursor */
	T6 a12; /* slots_position */
	T6 a13; /* count */
	T6 a14; /* clashes_previous_position */
	T0* a15; /* hash_function */
	T0* a16; /* equality_tester */
};

/* Struct for type [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_LIBRARY]] */
struct S904 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE] */
struct S905 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* after */
	T0* a2; /* current_cell */
	T0* a3; /* container */
	T1 a4; /* before */
	T0* a5; /* next_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET] */
struct S906 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE] */
struct S907 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT] */
struct S909 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_QUERY] */
struct S910 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type detachable ET_ALIAS_NAME_LIST */
struct S911 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_QUERY] */
struct S912 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_PROCEDURE] */
struct S913 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PROCEDURE] */
struct S914 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE] */
struct S921 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_BASE_TYPE] */
struct S923 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE] */
struct S924 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] INTEGER_OVERFLOW_CHECKER */
struct S925 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* integer_overflow_state1 */
	T0* a2; /* integer_overflow_state2 */
	T0* a3; /* natural_overflow_state1 */
	T0* a4; /* natural_overflow_state2 */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_NOTE_ELEMENT] */
struct S926 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_NOTE_ELEMENT] */
struct S927 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_NOTE_ELEMENT] */
struct S928 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type detachable C_DATE */
struct S929 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* millisecond_now */
	T0* a2; /* internal_item */
	T1 a3; /* is_utc */
};

/* Struct for type [detachable] YY_FILE_BUFFER */
struct S930 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* content */
	T6 a2; /* index */
	T6 a3; /* line */
	T6 a4; /* column */
	T6 a5; /* position */
	T6 a6; /* count */
	T1 a7; /* filled */
	T0* a8; /* file */
	T1 a9; /* beginning_of_line */
	T6 a10; /* capacity */
	T1 a11; /* end_of_file */
	T1 a12; /* interactive */
};

/* Struct for type [detachable] KL_STDIN_FILE */
struct S931 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* end_of_file */
	T2 a2; /* last_character */
	T0* a3; /* last_string */
	T0* a4; /* character_buffer */
};

/* Struct for type [detachable] YY_BUFFER */
struct S932 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* content */
	T6 a2; /* index */
	T6 a3; /* line */
	T6 a4; /* column */
	T6 a5; /* position */
	T6 a6; /* count */
	T1 a7; /* filled */
	T1 a8; /* beginning_of_line */
	T6 a9; /* capacity */
};

/* Struct for type [detachable] YY_UNICODE_FILE_BUFFER */
struct S936 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* content */
	T6 a2; /* index */
	T6 a3; /* line */
	T6 a4; /* column */
	T6 a5; /* position */
	T1 a6; /* has_utf8_bom */
	T6 a7; /* count */
	T1 a8; /* filled */
	T6 a9; /* capacity */
	T1 a10; /* end_of_file */
	T1 a11; /* interactive */
	T6 a12; /* encoding */
	T0* a13; /* file */
	T6 a14; /* default_encoding */
	T1 a15; /* beginning_of_line */
};

/* Struct for type [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32] */
struct S937 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* storage */
	T6 a2; /* count */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
};

/* Struct for type [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST] */
struct S938 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
};

/* Struct for type [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST] */
struct S939 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
};

/* Struct for type [detachable] DS_ARRAYED_STACK [detachable ET_KEYWORD] */
struct S940 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
};

/* Struct for type [detachable] DS_ARRAYED_STACK [detachable ET_SYMBOL] */
struct S941 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
};

/* Struct for type [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST] */
struct S942 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
};

/* Struct for type [detachable] DS_ARRAYED_STACK [[attached] ET_OBJECT_TEST_LIST] */
struct S943 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
};

/* Struct for type [detachable] DS_ARRAYED_STACK [detachable ET_ITERATION_COMPONENT_LIST] */
struct S944 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
};

/* Struct for type [detachable] DS_ARRAYED_STACK [[attached] ET_ITERATION_COMPONENT_LIST] */
struct S945 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM] */
struct S946 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE] */
struct S947 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS] */
struct S948 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* internal_cursor */
	T0* a3; /* item_storage */
	T6 a4; /* capacity */
	T6 a5; /* modulus */
	T6 a6; /* last_position */
	T6 a7; /* free_slot */
	T6 a8; /* position */
	T6 a9; /* slots_position */
	T0* a10; /* special_item_routines */
	T0* a11; /* clashes */
	T0* a12; /* slots */
	T6 a13; /* found_position */
	T6 a14; /* clashes_previous_position */
	T0* a15; /* equality_tester */
	T0* a16; /* hash_function */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD] */
struct S949 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_KEYWORD] */
struct S950 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_KEYWORD] */
struct S951 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_AGENT_KEYWORD */
struct S952 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 a1; /* code */
	T6 a2; /* compressed_position */
	T0* a3; /* name */
};

/* Struct for type [detachable] SPECIAL [detachable ET_AGENT_KEYWORD] */
struct S953 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECURSOR_KEYWORD] */
struct S954 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_PRECURSOR_KEYWORD */
struct S955 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 a1; /* code */
	T6 a2; /* compressed_position */
	T0* a3; /* name */
	T6 a4; /* seed */
};

/* Struct for type [detachable] SPECIAL [detachable ET_PRECURSOR_KEYWORD] */
struct S956 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL] */
struct S957 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_SYMBOL] */
struct S958 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSITION] */
struct S959 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_POSITION] */
struct S960 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_BOOLEAN_CONSTANT] */
struct S961 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_BOOLEAN_CONSTANT] */
struct S963 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_BREAK] */
struct S964 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_BREAK] */
struct S966 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHARACTER_CONSTANT] */
struct S967 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CHARACTER_CONSTANT] */
struct S969 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CURRENT] */
struct S970 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CURRENT] */
struct S971 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FREE_OPERATOR] */
struct S972 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_FREE_OPERATOR */
struct S973 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 a1; /* code */
	T6 a2; /* hash_code */
	T6 a3; /* compressed_position */
	T0* a4; /* break */
	T0* a5; /* operator_name */
	T6 a6; /* seed */
};

/* Struct for type [detachable] SPECIAL [detachable ET_FREE_OPERATOR] */
struct S974 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_IDENTIFIER] */
struct S975 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_IDENTIFIER] */
struct S976 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INTEGER_CONSTANT] */
struct S977 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_INTEGER_CONSTANT] */
struct S979 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_OPERATOR] */
struct S980 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_KEYWORD_OPERATOR */
struct S981 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* compressed_position */
	T2 a2; /* code */
	T0* a3; /* text */
	T6 a4; /* seed */
};

/* Struct for type [detachable] SPECIAL [detachable ET_KEYWORD_OPERATOR] */
struct S982 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING] */
struct S983 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_MANIFEST_STRING] */
struct S984 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_REAL_CONSTANT] */
struct S985 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_REAL_CONSTANT] */
struct S987 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_RESULT] */
struct S988 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_RESULT] */
struct S989 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_RETRY_INSTRUCTION] */
struct S990 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_RETRY_INSTRUCTION */
struct S991 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 a1; /* code */
	T6 a2; /* compressed_position */
	T0* a3; /* text */
};

/* Struct for type [detachable] SPECIAL [detachable ET_RETRY_INSTRUCTION] */
struct S992 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL_OPERATOR] */
struct S993 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_SYMBOL_OPERATOR */
struct S994 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* break */
	T2 a2; /* code */
	T6 a3; /* compressed_position */
	T6 a4; /* seed */
};

/* Struct for type [detachable] SPECIAL [detachable ET_SYMBOL_OPERATOR] */
struct S995 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_VOID] */
struct S996 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_VOID */
struct S997 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T6 a2; /* compressed_position */
	T2 a3; /* code */
	T0* a4; /* text */
};

/* Struct for type [detachable] SPECIAL [detachable ET_VOID] */
struct S998 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_SEMICOLON_SYMBOL] */
struct S999 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_SEMICOLON_SYMBOL */
struct S1000 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 a1; /* code */
	T6 a2; /* compressed_position */
};

/* Struct for type [detachable] SPECIAL [detachable ET_SEMICOLON_SYMBOL] */
struct S1001 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_SYMBOL] */
struct S1002 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_BRACKET_SYMBOL] */
struct S1003 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUESTION_MARK_SYMBOL] */
struct S1004 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_QUESTION_MARK_SYMBOL */
struct S1005 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 a1; /* code */
	T6 a2; /* compressed_position */
	T6 a3; /* index */
};

/* Struct for type [detachable] SPECIAL [detachable ET_QUESTION_MARK_SYMBOL] */
struct S1006 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_EXPRESSION] */
struct S1007 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_ACROSS_EXPRESSION */
struct S1008 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* item_name */
	T0* a3; /* as_keyword */
	T0* a4; /* unfolded_cursor_name */
	T0* a5; /* across_keyword */
	T0* a6; /* iterable_expression */
	T0* a7; /* until_conditional */
	T0* a8; /* iteration_conditional */
	T0* a9; /* end_keyword */
	T1 a10; /* is_all */
	T1 a11; /* has_cursor_name */
	T0* a12; /* invariant_part */
	T0* a13; /* variant_part */
	T0* a14; /* new_cursor_expression */
	T0* a15; /* cursor_item_expression */
	T0* a16; /* cursor_after_expression */
	T0* a17; /* cursor_forth_instruction */
};

/* Struct for type [detachable] SPECIAL [detachable ET_ACROSS_EXPRESSION] */
struct S1009 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_INSTRUCTION] */
struct S1010 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_ACROSS_INSTRUCTION */
struct S1011 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item_name */
	T0* a2; /* as_keyword */
	T0* a3; /* unfolded_cursor_name */
	T0* a4; /* across_keyword */
	T0* a5; /* iterable_expression */
	T0* a6; /* from_compound */
	T0* a7; /* until_conditional */
	T0* a8; /* loop_compound */
	T0* a9; /* end_keyword */
	T1 a10; /* has_cursor_name */
	T0* a11; /* invariant_part */
	T0* a12; /* variant_part */
	T0* a13; /* new_cursor_expression */
	T0* a14; /* cursor_item_expression */
	T0* a15; /* cursor_after_expression */
	T0* a16; /* cursor_forth_instruction */
};

/* Struct for type [detachable] SPECIAL [detachable ET_ACROSS_INSTRUCTION] */
struct S1012 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_ARGUMENT_LIST] */
struct S1013 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_ACTUAL_ARGUMENT_LIST] */
struct S1014 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_ITEM] */
struct S1015 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_ITEM] */
struct S1016 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_LIST] */
struct S1017 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_LIST] */
struct S1018 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND] */
struct S1019 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND] */
struct S1020 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM] */
struct S1021 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM] */
struct S1023 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_LIST] */
struct S1024 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_AGENT_ARGUMENT_OPERAND_LIST */
struct S1025 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* left_parenthesis */
	T0* a2; /* right_parenthesis */
	T6 a3; /* count */
	T0* a4; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_LIST] */
struct S1026 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_TARGET] */
struct S1027 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_AGENT_TARGET] */
struct S1028 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ALIAS_NAME] */
struct S1029 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_ALIAS_NAME */
struct S1030 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* alias_keyword */
	T0* a2; /* alias_string */
	T2 a3; /* code */
	T0* a4; /* convert_keyword */
	T6 a5; /* seed */
};

/* Struct for type [detachable] SPECIAL [detachable ET_ALIAS_NAME] */
struct S1031 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ALIAS_NAME_LIST] */
struct S1032 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_ALIAS_NAME_LIST] */
struct S1033 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ASSIGNER] */
struct S1034 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_ASSIGNER] */
struct S1036 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_EXPRESSION] */
struct S1037 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_BRACKET_EXPRESSION */
struct S1038 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* target */
	T0* a3; /* name */
	T0* a4; /* arguments */
};

/* Struct for type [detachable] SPECIAL [detachable ET_BRACKET_EXPRESSION] */
struct S1039 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CALL_AGENT] */
struct S1040 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_CALL_AGENT */
struct S1041 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* agent_keyword */
	T0* a3; /* target */
	T0* a4; /* qualified_name */
	T0* a5; /* arguments */
	T0* a6; /* implicit_result */
};

/* Struct for type [detachable] SPECIAL [detachable ET_CALL_AGENT] */
struct S1042 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_CALL_EXPRESSION] */
struct S1043 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_QUALIFIED_CALL_EXPRESSION] */
struct S1044 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHECK_INSTRUCTION] */
struct S1045 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_CHECK_INSTRUCTION */
struct S1046 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* check_keyword */
	T0* a2; /* end_keyword */
	T6 a3; /* count */
	T0* a4; /* storage */
	T0* a5; /* then_compound */
};

/* Struct for type [detachable] SPECIAL [detachable ET_CHECK_INSTRUCTION] */
struct S1047 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE] */
struct S1048 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CHOICE] */
struct S1050 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_CONSTANT] */
struct S1051 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CHOICE_CONSTANT] */
struct S1053 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_ITEM] */
struct S1054 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CHOICE_ITEM] */
struct S1056 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_LIST] */
struct S1057 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_CHOICE_LIST */
struct S1058 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* when_keyword */
	T6 a2; /* count */
	T0* a3; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_CHOICE_LIST] */
struct S1059 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLASS] */
struct S1060 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CLASS] */
struct S1061 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENT_ITEM] */
struct S1062 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CLIENT_ITEM] */
struct S1063 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENTS] */
struct S1064 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_CLIENTS */
struct S1065 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* left_brace */
	T0* a2; /* right_brace */
	T6 a3; /* count */
	T0* a4; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_CLIENTS] */
struct S1066 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_COMPOUND] */
struct S1067 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_COMPOUND] */
struct S1068 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONDITIONAL] */
struct S1069 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CONDITIONAL] */
struct S1071 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTANT] */
struct S1072 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CONSTANT] */
struct S1073 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
struct S1074 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
struct S1076 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST] */
struct S1077 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST */
struct S1078 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* left_bracket */
	T0* a4; /* right_bracket */
};

/* Struct for type [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST] */
struct S1079 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_CREATOR] */
struct S1080 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_CONSTRAINT_CREATOR */
struct S1081 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* create_keyword */
	T0* a2; /* end_keyword */
	T6 a3; /* count */
	T0* a4; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_CONSTRAINT_CREATOR] */
struct S1082 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_RENAME_LIST] */
struct S1083 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_CONSTRAINT_RENAME_LIST */
struct S1084 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* end_keyword */
	T0* a2; /* rename_keyword */
	T6 a3; /* count */
	T0* a4; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_CONSTRAINT_RENAME_LIST] */
struct S1085 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_TYPE] */
struct S1086 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CONSTRAINT_TYPE] */
struct S1088 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE] */
struct S1089 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CONVERT_FEATURE] */
struct S1091 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_ITEM] */
struct S1092 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_ITEM] */
struct S1094 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_LIST] */
struct S1095 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_CONVERT_FEATURE_LIST */
struct S1096 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* convert_keyword */
	T6 a2; /* count */
	T0* a3; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_LIST] */
struct S1097 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATE_EXPRESSION] */
struct S1098 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_CREATE_EXPRESSION] */
struct S1099 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATION_REGION] */
struct S1100 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_CREATION_REGION */
struct S1101 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* class_name */
	T0* a2; /* left_symbol */
	T0* a3; /* right_symbol */
};

/* Struct for type [detachable] SPECIAL [detachable ET_CREATION_REGION] */
struct S1102 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR] */
struct S1103 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_CREATOR */
struct S1104 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* create_keyword */
	T0* a2; /* clients */
	T0* a3; /* storage */
	T6 a4; /* count */
};

/* Struct for type [detachable] SPECIAL [detachable ET_CREATOR] */
struct S1105 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR_LIST] */
struct S1106 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_CREATOR_LIST */
struct S1107 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_CREATOR_LIST] */
struct S1108 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_DEBUG_INSTRUCTION] */
struct S1109 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_DEBUG_INSTRUCTION */
struct S1110 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* keys */
	T0* a2; /* compound */
	T0* a3; /* end_keyword */
};

/* Struct for type [detachable] SPECIAL [detachable ET_DEBUG_INSTRUCTION] */
struct S1111 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION] */
struct S1112 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_ELSEIF_EXPRESSION */
struct S1113 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* conditional */
	T0* a2; /* then_keyword */
	T0* a3; /* then_expression */
};

/* Struct for type [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION] */
struct S1114 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION_LIST] */
struct S1115 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_ELSEIF_EXPRESSION_LIST */
struct S1116 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION_LIST] */
struct S1117 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART] */
struct S1118 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_ELSEIF_PART */
struct S1119 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* conditional */
	T0* a2; /* then_compound */
};

/* Struct for type [detachable] SPECIAL [detachable ET_ELSEIF_PART] */
struct S1120 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART_LIST] */
struct S1121 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_ELSEIF_PART_LIST */
struct S1122 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_ELSEIF_PART_LIST] */
struct S1123 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT] */
struct S1124 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_EXPORT] */
struct S1126 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT_LIST] */
struct S1127 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_EXPORT_LIST] */
struct S1128 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION] */
struct S1129 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_EXPRESSION] */
struct S1130 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION_ITEM] */
struct S1131 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_EXPRESSION_ITEM] */
struct S1132 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTENDED_FEATURE_NAME] */
struct S1133 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_EXTENDED_FEATURE_NAME] */
struct S1134 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTERNAL_ALIAS] */
struct S1135 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_EXTERNAL_ALIAS] */
struct S1137 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE] */
struct S1138 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_FEATURE_CLAUSE */
struct S1139 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* feature_keyword */
	T0* a2; /* clients */
};

/* Struct for type [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE] */
struct S1140 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE_LIST] */
struct S1141 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_FEATURE_CLAUSE_LIST */
struct S1142 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE_LIST] */
struct S1143 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_EXPORT] */
struct S1144 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_FEATURE_EXPORT */
struct S1145 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* clients_clause */
	T0* a2; /* storage */
	T6 a3; /* count */
};

/* Struct for type [detachable] SPECIAL [detachable ET_FEATURE_EXPORT] */
struct S1146 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME] */
struct S1147 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_FEATURE_NAME] */
struct S1148 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME_ITEM] */
struct S1149 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_FEATURE_NAME_ITEM] */
struct S1151 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT] */
struct S1152 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT] */
struct S1153 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_ITEM] */
struct S1154 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_ITEM] */
struct S1156 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_LIST] */
struct S1157 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_LIST] */
struct S1158 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER] */
struct S1159 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER] */
struct S1160 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_ITEM] */
struct S1161 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_ITEM] */
struct S1163 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_LIST] */
struct S1164 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_LIST] */
struct S1165 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_EXPRESSION] */
struct S1166 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_IF_EXPRESSION */
struct S1167 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* conditional */
	T0* a3; /* then_keyword */
	T0* a4; /* then_expression */
	T0* a5; /* else_keyword */
	T0* a6; /* else_expression */
	T0* a7; /* end_keyword */
	T0* a8; /* elseif_parts */
};

/* Struct for type [detachable] SPECIAL [detachable ET_IF_EXPRESSION] */
struct S1168 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_INSTRUCTION] */
struct S1169 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_IF_INSTRUCTION */
struct S1170 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* conditional */
	T0* a2; /* then_compound */
	T0* a3; /* end_keyword */
	T0* a4; /* elseif_parts */
	T0* a5; /* else_compound */
};

/* Struct for type [detachable] SPECIAL [detachable ET_IF_INSTRUCTION] */
struct S1171 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_LIST] */
struct S1172 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_INDEXING_LIST */
struct S1173 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* indexing_keyword */
	T6 a2; /* count */
	T0* a3; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_INDEXING_LIST] */
struct S1174 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_ITEM] */
struct S1175 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_INDEXING_ITEM] */
struct S1177 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM] */
struct S1178 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_INDEXING_TERM] */
struct S1179 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_ITEM] */
struct S1180 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_INDEXING_TERM_ITEM] */
struct S1182 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_LIST] */
struct S1183 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_INDEXING_TERM_LIST */
struct S1184 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_INDEXING_TERM_LIST] */
struct S1185 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INLINE_AGENT] */
struct S1186 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_INLINE_AGENT] */
struct S1187 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSPECT_EXPRESSION] */
struct S1188 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_INSPECT_EXPRESSION */
struct S1189 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* conditional */
	T0* a3; /* when_parts */
	T0* a4; /* end_keyword */
	T0* a5; /* else_part */
};

/* Struct for type [detachable] SPECIAL [detachable ET_INSPECT_EXPRESSION] */
struct S1190 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSPECT_INSTRUCTION] */
struct S1191 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_INSPECT_INSTRUCTION */
struct S1192 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* conditional */
	T0* a2; /* when_parts */
	T0* a3; /* end_keyword */
	T0* a4; /* else_compound */
};

/* Struct for type [detachable] SPECIAL [detachable ET_INSPECT_INSTRUCTION] */
struct S1193 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSTRUCTION] */
struct S1194 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_INSTRUCTION] */
struct S1196 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INVARIANTS] */
struct S1197 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_INVARIANTS */
struct S1198 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* implementation_class */
	T0* a2; /* invariant_keyword */
	T0* a3; /* storage */
	T6 a4; /* count */
	T0* a5; /* object_tests */
	T0* a6; /* iteration_components */
	T1 a7; /* validity_checked */
	T1 a8; /* has_validity_error */
};

/* Struct for type [detachable] SPECIAL [detachable ET_INVARIANTS] */
struct S1199 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_FEATURE_NAME_LIST] */
struct S1200 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_KEYWORD_FEATURE_NAME_LIST] */
struct S1201 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_LIKE_TYPE] */
struct S1202 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_LIKE_TYPE] */
struct S1204 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE] */
struct S1205 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_LOCAL_VARIABLE */
struct S1206 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name_item */
	T0* a2; /* declared_type */
	T6 a3; /* index */
	T6 a4; /* attached_index */
	T1 a5; /* is_used */
};

/* Struct for type [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE] */
struct S1207 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_ITEM] */
struct S1208 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_ITEM] */
struct S1210 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_LIST] */
struct S1211 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_LOCAL_VARIABLE_LIST */
struct S1212 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* local_keyword */
};

/* Struct for type [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_LIST] */
struct S1213 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOOP_INVARIANTS] */
struct S1214 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_LOOP_INVARIANTS */
struct S1215 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* invariant_keyword */
	T6 a2; /* count */
	T0* a3; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_LOOP_INVARIANTS] */
struct S1216 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_ARRAY] */
struct S1217 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_MANIFEST_ARRAY */
struct S1218 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* cast_type */
	T0* a3; /* left_symbol */
	T6 a4; /* count */
	T0* a5; /* storage */
	T0* a6; /* right_symbol */
};

/* Struct for type [detachable] SPECIAL [detachable ET_MANIFEST_ARRAY] */
struct S1219 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_ITEM] */
struct S1220 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_MANIFEST_STRING_ITEM] */
struct S1222 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_LIST] */
struct S1223 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_MANIFEST_STRING_LIST */
struct S1224 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* left_parenthesis */
	T0* a2; /* right_parenthesis */
	T6 a3; /* count */
	T0* a4; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_MANIFEST_STRING_LIST] */
struct S1225 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_TUPLE] */
struct S1226 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_MANIFEST_TUPLE] */
struct S1227 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBSOLETE] */
struct S1228 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_OBSOLETE] */
struct S1230 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENTHESIZED_EXPRESSION] */
struct S1231 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_PARENTHESIZED_EXPRESSION */
struct S1232 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* expression */
	T0* a3; /* left_parenthesis */
	T0* a4; /* right_parenthesis */
};

/* Struct for type [detachable] SPECIAL [detachable ET_PARENTHESIZED_EXPRESSION] */
struct S1233 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT] */
struct S1234 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_PARENT] */
struct S1235 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_CLAUSE_LIST] */
struct S1236 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_PARENT_CLAUSE_LIST */
struct S1237 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_PARENT_CLAUSE_LIST] */
struct S1238 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_ITEM] */
struct S1239 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_PARENT_ITEM] */
struct S1240 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_LIST] */
struct S1241 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_PARENT_LIST] */
struct S1242 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSTCONDITIONS] */
struct S1243 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_POSTCONDITIONS */
struct S1244 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* validity_checked */
	T1 a2; /* has_validity_error */
	T6 a3; /* count */
	T0* a4; /* storage */
	T0* a5; /* ensure_keyword */
	T0* a6; /* then_keyword */
};

/* Struct for type [detachable] SPECIAL [detachable ET_POSTCONDITIONS] */
struct S1245 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECONDITIONS] */
struct S1246 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_PRECONDITIONS] */
struct S1247 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PROCEDURE] */
struct S1248 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_PROCEDURE] */
struct S1249 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_LIKE_IDENTIFIER] */
struct S1250 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_QUALIFIED_LIKE_IDENTIFIER] */
struct S1252 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUANTIFIER_EXPRESSION] */
struct S1253 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_QUANTIFIER_EXPRESSION */
struct S1254 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* item_name */
	T0* a3; /* unfolded_cursor_name */
	T0* a4; /* quantifier_symbol */
	T0* a5; /* colon_symbol */
	T0* a6; /* iterable_expression */
	T0* a7; /* bar_symbol */
	T0* a8; /* iteration_expression */
	T1 a9; /* is_all */
	T0* a10; /* new_cursor_expression */
	T0* a11; /* cursor_item_expression */
	T0* a12; /* cursor_after_expression */
	T0* a13; /* cursor_forth_instruction */
};

/* Struct for type [detachable] SPECIAL [detachable ET_QUANTIFIER_EXPRESSION] */
struct S1255 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUERY] */
struct S1256 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_QUERY] */
struct S1257 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_ITEM] */
struct S1258 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_RENAME_ITEM] */
struct S1260 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_LIST] */
struct S1261 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_RENAME_LIST] */
struct S1262 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_REPEAT_INSTRUCTION] */
struct S1263 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_REPEAT_INSTRUCTION */
struct S1264 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item_name */
	T0* a2; /* unfolded_cursor_name */
	T0* a3; /* open_repeat_symbol */
	T0* a4; /* colon_symbol */
	T0* a5; /* iterable_expression */
	T0* a6; /* bar_symbol */
	T0* a7; /* loop_compound */
	T0* a8; /* close_repeat_symbol */
	T0* a9; /* new_cursor_expression */
	T0* a10; /* cursor_item_expression */
	T0* a11; /* cursor_after_expression */
	T0* a12; /* cursor_forth_instruction */
};

/* Struct for type [detachable] SPECIAL [detachable ET_REPEAT_INSTRUCTION] */
struct S1265 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_STATIC_CALL_EXPRESSION] */
struct S1266 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_STATIC_CALL_EXPRESSION */
struct S1267 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* feature_keyword */
	T0* a3; /* static_type */
	T0* a4; /* qualified_name */
	T0* a5; /* arguments */
	T0* a6; /* parenthesis_call */
};

/* Struct for type [detachable] SPECIAL [detachable ET_STATIC_CALL_EXPRESSION] */
struct S1268 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_STRIP_EXPRESSION] */
struct S1269 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_STRIP_EXPRESSION */
struct S1270 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* strip_keyword */
	T0* a3; /* left_parenthesis */
	T6 a4; /* count */
	T0* a5; /* storage */
	T0* a6; /* right_parenthesis */
};

/* Struct for type [detachable] SPECIAL [detachable ET_STRIP_EXPRESSION] */
struct S1271 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE] */
struct S1272 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_TYPE] */
struct S1273 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT] */
struct S1274 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT] */
struct S1276 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_ITEM] */
struct S1277 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_ITEM] */
struct S1279 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_LIST] */
struct S1280 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_TYPE_CONSTRAINT_LIST */
struct S1281 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* left_brace */
	T0* a2; /* right_brace */
	T6 a3; /* count */
	T0* a4; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_LIST] */
struct S1282 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_ITEM] */
struct S1283 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_TYPE_ITEM] */
struct S1285 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_LIST] */
struct S1286 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_TYPE_LIST] */
struct S1288 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_VARIANT] */
struct S1289 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_VARIANT */
struct S1290 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* variant_keyword */
	T0* a2; /* tag */
	T0* a3; /* expression */
};

/* Struct for type [detachable] SPECIAL [detachable ET_VARIANT] */
struct S1291 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_EXPRESSION] */
struct S1292 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_WHEN_EXPRESSION */
struct S1293 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* choices */
	T0* a2; /* then_keyword */
	T0* a3; /* then_expression */
};

/* Struct for type [detachable] SPECIAL [detachable ET_WHEN_EXPRESSION] */
struct S1294 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_EXPRESSION_LIST] */
struct S1295 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_WHEN_EXPRESSION_LIST */
struct S1296 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_WHEN_EXPRESSION_LIST] */
struct S1297 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART] */
struct S1298 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_WHEN_PART */
struct S1299 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* choices */
	T0* a2; /* then_compound */
};

/* Struct for type [detachable] SPECIAL [detachable ET_WHEN_PART] */
struct S1300 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART_LIST] */
struct S1301 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable ET_WHEN_PART_LIST */
struct S1302 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] SPECIAL [detachable ET_WHEN_PART_LIST] */
struct S1303 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_WRITABLE] */
struct S1304 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [detachable ET_WRITABLE] */
struct S1305 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS] */
struct S1306 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* sorted_items */
	T0* a2; /* cycle */
	T0* a3; /* indexes */
	T0* a4; /* counts */
	T0* a5; /* successors */
	T0* a6; /* items */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS] */
struct S1307 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* internal_cursor */
	T6 a3; /* found_position */
	T0* a4; /* item_storage */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T6 a8; /* capacity */
	T0* a9; /* equality_tester */
	T0* a10; /* key_equality_tester */
	T0* a11; /* key_storage */
	T0* a12; /* clashes */
	T6 a13; /* modulus */
	T0* a14; /* slots */
	T6 a15; /* slots_position */
	T6 a16; /* clashes_previous_position */
	T0* a17; /* special_item_routines */
	T0* a18; /* special_key_routines */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] ET_PARENT_CHECKER1 */
struct S1308 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* current_class */
	T0* a3; /* system_processor */
	T0* a4; /* current_parent */
};

/* Struct for type [detachable] ET_FORMAL_PARAMETER_CHECKER1 */
struct S1309 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* all_base_types */
	T0* a3; /* base_types */
	T0* a4; /* formal_dependencies */
	T0* a5; /* base_type_dependencies */
	T0* a6; /* recursive_formal_constraints */
	T0* a7; /* recursive_formal_constraints_to_be_processed */
	T0* a8; /* current_class */
	T0* a9; /* system_processor */
	T0* a10; /* current_formal */
	T0* a11; /* current_type_constraint */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME] */
struct S1310 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* internal_cursor */
	T6 a3; /* found_position */
	T6 a4; /* position */
	T0* a5; /* item_storage */
	T0* a6; /* key_equality_tester */
	T0* a7; /* internal_keys */
	T6 a8; /* last_position */
	T6 a9; /* free_slot */
	T6 a10; /* modulus */
	T6 a11; /* slots_position */
	T6 a12; /* clashes_previous_position */
	T0* a13; /* equality_tester */
	T0* a14; /* key_storage */
	T0* a15; /* clashes */
	T0* a16; /* slots */
	T6 a17; /* capacity */
	T0* a18; /* special_item_routines */
	T0* a19; /* special_key_routines */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] ET_FEATURE_NAME_TESTER */
struct S1311 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME] */
struct S1313 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T0* a3; /* key_storage */
	T0* a4; /* key_equality_tester */
	T0* a5; /* internal_keys */
	T6 a6; /* position */
	T6 a7; /* last_position */
	T6 a8; /* capacity */
	T6 a9; /* count */
	T6 a10; /* free_slot */
	T0* a11; /* equality_tester */
	T6 a12; /* modulus */
	T6 a13; /* slots_position */
	T6 a14; /* clashes_previous_position */
	T0* a15; /* clashes */
	T0* a16; /* slots */
	T0* a17; /* special_item_routines */
	T0* a18; /* special_key_routines */
	T0* a19; /* internal_cursor */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] ET_ALIAS_NAME_TESTER */
struct S1314 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST] */
struct S1316 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS] */
struct S1317 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* equality_tester */
	T0* a2; /* key_equality_tester */
	T6 a3; /* capacity */
	T6 a4; /* modulus */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T0* a8; /* special_item_routines */
	T0* a9; /* item_storage */
	T0* a10; /* special_key_routines */
	T0* a11; /* key_storage */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T6 a14; /* found_position */
	T0* a15; /* internal_cursor */
	T6 a16; /* count */
	T0* a17; /* internal_keys */
	T6 a18; /* slots_position */
	T6 a19; /* clashes_previous_position */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] ET_FEATURE_ADAPTATION_RESOLVER */
struct S1318 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* rename_table */
	T0* a3; /* export_table */
	T0* a4; /* undefine_table */
	T0* a5; /* redefine_table */
	T0* a6; /* select_table */
	T0* a7; /* replicable_features */
	T0* a8; /* current_class */
	T0* a9; /* system_processor */
	T0* a10; /* free_parent_feature */
	T0* a11; /* parent_feature_list */
	T0* a12; /* free_inherited_feature */
	T0* a13; /* inherited_feature_list */
	T0* a14; /* free_redeclared_feature */
	T0* a15; /* redeclared_feature_list */
};

/* Struct for type [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER */
struct S1319 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* dotnet_features */
	T0* a3; /* other_dotnet_features */
	T0* a4; /* current_class */
	T0* a5; /* system_processor */
	T0* a6; /* free_parent_feature */
	T0* a7; /* parent_feature_list */
	T0* a8; /* free_inherited_feature */
	T0* a9; /* inherited_feature_list */
	T0* a10; /* free_redeclared_feature */
	T0* a11; /* redeclared_feature_list */
};

/* Struct for type [detachable] ET_IDENTIFIER_TYPE_RESOLVER */
struct S1320 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* current_class */
	T0* a3; /* system_processor */
	T0* a4; /* current_feature */
};

/* Struct for type [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1 */
struct S1321 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* constraint_context */
	T0* a2; /* current_class */
	T0* a3; /* system_processor */
};

/* Struct for type [detachable] ET_ANCHORED_TYPE_CHECKER */
struct S1322 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* anchored_type_sorter */
	T0* a3; /* current_class */
	T0* a4; /* system_processor */
	T0* a5; /* current_anchored_type */
};

/* Struct for type [detachable] ET_SIGNATURE_CHECKER */
struct S1323 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* parent_context */
	T0* a3; /* current_class */
	T0* a4; /* system_processor */
	T6 a5; /* processing_mode */
};

/* Struct for type [detachable] ET_PARENT_CHECKER2 */
struct S1324 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* constraint_context */
	T0* a3; /* current_class */
	T0* a4; /* system_processor */
};

/* Struct for type [detachable] ET_FORMAL_PARAMETER_CHECKER2 */
struct S1325 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* constraint_context */
	T0* a3; /* current_class */
	T0* a4; /* system_processor */
};

/* Struct for type [detachable] ET_BUILTIN_FEATURE_CHECKER */
struct S1326 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* builtin_features */
	T0* a3; /* current_class */
	T0* a4; /* system_processor */
};

/* Struct for type [detachable] ET_PRECURSOR_CHECKER */
struct S1327 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* current_feature */
	T0* a3; /* current_class */
	T0* a4; /* system_processor */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32] */
struct S1328 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* internal_cursor */
	T0* a3; /* item_storage */
	T6 a4; /* position */
	T6 a5; /* last_position */
	T6 a6; /* capacity */
	T6 a7; /* slots_position */
	T6 a8; /* free_slot */
	T0* a9; /* equality_tester */
	T0* a10; /* key_equality_tester */
	T6 a11; /* found_position */
	T6 a12; /* modulus */
	T6 a13; /* clashes_previous_position */
	T0* a14; /* special_item_routines */
	T0* a15; /* clashes */
	T0* a16; /* slots */
	T0* a17; /* special_key_routines */
	T0* a18; /* key_storage */
	T0* a19; /* internal_keys */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_CLASS] */
struct S1329 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* last_position */
	T0* a3; /* clashes */
	T0* a4; /* item_storage */
	T6 a5; /* capacity */
	T6 a6; /* modulus */
	T6 a7; /* free_slot */
	T6 a8; /* position */
	T0* a9; /* special_item_routines */
	T0* a10; /* slots */
	T6 a11; /* found_position */
	T0* a12; /* internal_cursor */
	T6 a13; /* slots_position */
	T6 a14; /* clashes_previous_position */
	T0* a15; /* equality_tester */
	T0* a16; /* hash_function */
};

/* Struct for type [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER */
struct S1330 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* current_class_impl */
	T0* a3; /* current_class */
	T0* a4; /* adapted_base_class_checker */
	T0* a5; /* adapted_base_classes */
	T0* a6; /* target_context */
	T0* a7; /* other_context */
	T0* a8; /* classes_to_be_processed */
	T0* a9; /* system_processor */
	T1 a10; /* in_qualified_anchored_type */
};

/* Struct for type [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2 */
struct S1331 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* constraint_context */
	T0* a2; /* current_class */
	T0* a3; /* system_processor */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME] */
struct S1332 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* capacity */
	T6 a2; /* found_position */
	T6 a3; /* position */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
	T0* a6; /* item_storage */
	T0* a7; /* key_equality_tester */
	T0* a8; /* internal_keys */
	T6 a9; /* last_position */
	T6 a10; /* free_slot */
	T6 a11; /* modulus */
	T6 a12; /* slots_position */
	T6 a13; /* clashes_previous_position */
	T0* a14; /* equality_tester */
	T0* a15; /* key_storage */
	T0* a16; /* clashes */
	T0* a17; /* slots */
	T0* a18; /* special_item_routines */
	T0* a19; /* special_key_routines */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME] */
struct S1333 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* capacity */
	T6 a2; /* position */
	T6 a3; /* found_position */
	T0* a4; /* item_storage */
	T0* a5; /* key_storage */
	T0* a6; /* key_equality_tester */
	T0* a7; /* internal_keys */
	T6 a8; /* count */
	T6 a9; /* last_position */
	T6 a10; /* free_slot */
	T6 a11; /* modulus */
	T6 a12; /* slots_position */
	T6 a13; /* clashes_previous_position */
	T0* a14; /* equality_tester */
	T0* a15; /* internal_cursor */
	T0* a16; /* clashes */
	T0* a17; /* slots */
	T0* a18; /* special_item_routines */
	T0* a19; /* special_key_routines */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] ET_RENAME */
struct S1337 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* old_name */
	T0* a2; /* as_keyword */
	T0* a3; /* new_name */
};

/* Struct for type [detachable] ET_PARENT_CHECKER3 */
struct S1338 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* current_class */
	T0* a3; /* system_processor */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_PROCEDURE] */
struct S1339 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T0* a2; /* item_storage */
	T6 a3; /* capacity */
	T6 a4; /* modulus */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T6 a8; /* count */
	T0* a9; /* special_item_routines */
	T0* a10; /* clashes */
	T0* a11; /* slots */
	T6 a12; /* found_position */
	T6 a13; /* slots_position */
	T6 a14; /* clashes_previous_position */
	T0* a15; /* hash_function */
	T0* a16; /* equality_tester */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_QUERY] */
struct S1340 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T0* a2; /* item_storage */
	T6 a3; /* capacity */
	T6 a4; /* modulus */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T6 a8; /* count */
	T0* a9; /* special_item_routines */
	T0* a10; /* clashes */
	T0* a11; /* slots */
	T6 a12; /* found_position */
	T6 a13; /* slots_position */
	T6 a14; /* clashes_previous_position */
	T0* a15; /* hash_function */
	T0* a16; /* equality_tester */
};

/* Struct for type [detachable] ET_SUPPLIER_BUILDER */
struct S1341 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* supplier_classes */
	T0* a2; /* current_class */
};

/* Struct for type detachable ET_PARENT_FEATURE */
struct S1344 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next */
	T0* a2; /* parent */
	T0* a3; /* precursor_feature */
	T0* a4; /* new_name */
	T0* a5; /* undefine_name */
	T0* a6; /* redefine_name */
	T0* a7; /* select_name */
	T0* a8; /* merged_feature */
	T1 a9; /* has_other_deferred */
	T1 a10; /* has_other_effective */
};

/* Struct for type [detachable] ET_INHERITED_FEATURE */
struct S1345 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next */
	T0* a2; /* parent_feature */
	T0* a3; /* flattened_feature */
	T0* a4; /* flattened_parent */
	T6 a5; /* first_seed */
	T0* a6; /* other_seeds */
	T1 a7; /* is_selected */
	T0* a8; /* replicated_features */
	T0* a9; /* replicated_seeds */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CLASS] */
struct S1346 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS] */
struct S1347 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS] */
struct S1348 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] DS_ARRAYED_STACK [[attached] YY_BUFFER] */
struct S1349 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* a3; /* storage */
	T0* a4; /* special_routines */
};

/* Struct for type [detachable] UT_SYNTAX_ERROR */
struct S1350 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] UT_TOO_MANY_INCLUDES_ERROR */
struct S1351 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type detachable ET_DYNAMIC_TARGET_LIST */
struct S1353 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] SPECIAL [[attached] NATIVE_STRING] */
struct S1354 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] ARRAY [detachable STRING_8] */
struct S1355 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] XM_EIFFEL_CHARACTER_ENTITY */
struct S1356 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* code */
};

/* Struct for type detachable DS_BILINKABLE [[attached] XM_POSITION] */
struct S1358 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
	T0* a3; /* left */
};

/* Struct for type [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_POSITION] */
struct S1359 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T1 a2; /* before */
};

/* Struct for type detachable DS_LINKABLE [[attached] XM_EIFFEL_SCANNER] */
struct S1360 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
};

/* Struct for type [detachable] SPECIAL [[attached] XM_EIFFEL_ENTITY_DEF] */
struct S1361 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type detachable DS_HASH_TABLE_CURSOR [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8] */
struct S1364 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_ENTITY_DEF] */
struct S1365 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable KL_EQUALITY_TESTER [[attached] XM_EIFFEL_PARSER_NAME] */
struct S1368 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] XM_EIFFEL_PARSER_NAME] */
struct S1369 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ELEMENT_CONTENT] */
struct S1370 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* first_cell */
	T6 a2; /* count */
	T0* a3; /* last_cell */
	T0* a4; /* internal_cursor */
};

/* Struct for type [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
struct S1371 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* after */
	T0* a2; /* current_cell */
	T0* a3; /* container */
	T1 a4; /* before */
	T0* a5; /* next_cursor */
};

/* Struct for type detachable DS_BILINKABLE [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
struct S1372 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
	T0* a3; /* left */
};

/* Struct for type [detachable] DS_BILINKED_LIST_CURSOR [[attached] STRING_8] */
struct S1373 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* after */
	T0* a2; /* current_cell */
	T0* a3; /* container */
	T1 a4; /* before */
	T0* a5; /* next_cursor */
};

/* Struct for type detachable DS_BILINKABLE [[attached] STRING_8] */
struct S1374 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
	T0* a3; /* left */
};

/* Struct for type [detachable] XM_NAMESPACE_RESOLVER_CONTEXT */
struct S1375 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* context */
};

/* Struct for type [detachable] DS_LINKED_QUEUE [detachable STRING_8] */
struct S1378 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* first_cell */
	T6 a2; /* count */
	T0* a3; /* last_cell */
};

/* Struct for type [detachable] DS_LINKED_QUEUE [[attached] STRING_8] */
struct S1379 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* first_cell */
	T0* a2; /* last_cell */
	T6 a3; /* count */
};

/* Struct for type detachable KL_EQUALITY_TESTER [[attached] XM_NAMESPACE] */
struct S1380 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] XM_NAMESPACE] */
struct S1381 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_NAMESPACE] */
struct S1382 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] XM_NAMESPACE] */
struct S1383 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_CLUSTER] */
struct S1384 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_PARENT_ITEM] */
struct S1386 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_ITEM] */
struct S1387 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_BRACED_TYPE_LIST */
struct S1388 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* left_brace */
	T0* a2; /* right_brace */
	T6 a3; /* count */
	T0* a4; /* storage */
};

/* Struct for type detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS] */
struct S1389 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* equality_tester */
	T0* a6; /* internal_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ADAPTED_LIBRARY] */
struct S1390 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_LIBRARY] */
struct S1391 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_LIBRARY] */
struct S1392 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type detachable DS_LINKABLE [[attached] XM_DOCUMENT_NODE] */
struct S1393 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
struct S1395 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
struct S1396 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
struct S1397 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY] */
struct S1398 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
	T0* a6; /* equality_tester */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32] */
struct S1400 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL] */
struct S1402 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CALL_NAME] */
struct S1404 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL] */
struct S1406 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CALL_NAME] */
struct S1407 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME] */
struct S1408 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME] */
struct S1410 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL] */
struct S1411 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL] */
struct S1412 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S1413 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]] */
struct S1414 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable DS_LINKABLE [[attached] INTEGER_32]] */
struct S1415 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [detachable DS_LINKABLE [[attached] INTEGER_32]] */
struct S1416 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_SYSTEM_PROCESSOR] */
struct S1418 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_SYSTEM_PROCESSOR] */
struct S1419 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_SYSTEM_PROCESSOR] */
struct S1420 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_CLASS] */
struct S1421 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ADAPTED_CLASS] */
struct S1423 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_CLASS] */
struct S1424 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PROCEDURE] */
struct S1425 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T6 a2; /* position */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUERY] */
struct S1426 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T6 a2; /* position */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NAMED_OBJECT_TEST] */
struct S1427 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ITERATION_COMPONENT] */
struct S1428 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32] */
struct S1429 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_8] */
struct S1433 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_8, [attached] ET_CLASS_NAME] */
struct S1434 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] RX_CHARACTER_SET] */
struct S1435 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] RX_CHARACTER_SET] */
struct S1436 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_64] */
struct S1443 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_CHARACTER_BUFFER */
struct S1446 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* as_special */
	T0* a2; /* area */
};

/* Struct for type [detachable] UT_TRISTATE */
struct S1448 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 a1; /* value */
};

/* Struct for type [detachable] KL_AGENT_ROUTINES [[attached] ANY] */
struct S1449 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] TUPLE [[attached] UT_TRISTATE] */
struct S1450 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] KL_AGENT_ROUTINES [[attached] ANY], [attached] PROCEDURE [[attached] TUPLE]] */
struct S1451 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] PREDICATE [[attached] TUPLE [[attached] ET_MASTER_CLASS]] */
struct S1452 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] PREDICATE [[attached] TUPLE] */
struct S1453 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE] */
struct S1455 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T6 a2; /* last_position */
	T6 a3; /* capacity */
	T6 a4; /* slots_position */
	T6 a5; /* count */
	T6 a6; /* modulus */
	T6 a7; /* clashes_previous_position */
	T0* a8; /* key_equality_tester */
	T0* a9; /* equality_tester */
	T6 a10; /* found_position */
	T0* a11; /* special_item_routines */
	T0* a12; /* item_storage */
	T0* a13; /* clashes */
	T0* a14; /* slots */
	T0* a15; /* special_key_routines */
	T0* a16; /* key_storage */
	T6 a17; /* free_slot */
	T0* a18; /* internal_cursor */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE] */
struct S1456 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* container */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_SYSTEM, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
struct S1457 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T6 z3;
};

/* Struct for type [detachable] KL_AGENT_ROUTINES [[attached] ET_UNIVERSE] */
struct S1458 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] PREDICATE [[attached] TUPLE [[attached] ET_UNIVERSE]] */
struct S1459 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE]] */
struct S1460 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] KL_AGENT_ROUTINES [[attached] ET_UNIVERSE], [attached] PREDICATE [[attached] TUPLE [[attached] ET_UNIVERSE]]] */
struct S1461 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]], [attached] FUNCTION [[attached] TUPLE, [attached] BOOLEAN]] */
struct S1463 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]], [attached] FUNCTION [[attached] TUPLE [[attached] ET_MASTER_CLASS], [attached] BOOLEAN]] */
struct S1464 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]], [attached] FUNCTION [[attached] TUPLE [[attached] ET_MASTER_CLASS], [attached] BOOLEAN], [attached] FUNCTION [[attached] TUPLE, [attached] BOOLEAN]] */
struct S1465 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
};

/* Struct for type [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET */
struct S1466 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* static_type */
	T6 a2; /* count */
	T1 a3; /* is_never_void */
	T0* a4; /* dynamic_types */
	T0* a5; /* agent_type */
	T1 a6; /* is_dynamic_types_readonly */
};

/* Struct for type [detachable] XM_COMMENT */
struct S1469 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* data */
	T0* a2; /* parent */
};

/* Struct for type [detachable] XM_PROCESSING_INSTRUCTION */
struct S1470 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target */
	T0* a2; /* data */
	T0* a3; /* parent */
};

/* Struct for type [detachable] XM_CHARACTER_DATA */
struct S1471 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* content */
	T0* a2; /* parent */
};

/* Struct for type [detachable] ET_LIKE_N */
struct S1474 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* previous */
	T0* a3; /* type_mark */
	T0* a4; /* like_keyword */
};

/* Struct for type [detachable] KL_STRING_INPUT_STREAM */
struct S1477 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* end_of_input */
	T2 a2; /* last_character */
	T0* a3; /* last_string */
	T0* a4; /* string */
	T6 a5; /* location */
};

/* Struct for type [detachable] KL_UNICODE_CHARACTER_BUFFER */
struct S1479 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* as_unicode_special */
	T2 a2; /* invalid_character */
	T0* a3; /* unicode_area */
	T0* a4; /* area */
	T3 a5; /* invalid_unicode_character */
};

/* Struct for type detachable ET_OBJECT_TEST_LIST */
struct S1481 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type detachable ET_C3_CHARACTER_CONSTANT */
struct S1482 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* cast_type */
	T1 a3; /* has_invalid_code */
	T3 a4; /* value */
	T6 a5; /* compressed_position */
	T0* a6; /* literal */
	T0* a7; /* type */
};

/* Struct for type detachable ET_REGULAR_MANIFEST_STRING */
struct S1483 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* value */
	T6 a2; /* index */
	T0* a3; /* cast_type */
	T6 a4; /* compressed_position */
	T0* a5; /* type */
};

/* Struct for type detachable ET_SPECIAL_MANIFEST_STRING */
struct S1484 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* value */
	T6 a2; /* index */
	T0* a3; /* cast_type */
	T1 a4; /* has_invalid_code */
	T6 a5; /* compressed_position */
	T0* a6; /* literal */
	T0* a7; /* type */
};

/* Struct for type detachable ET_VERBATIM_STRING */
struct S1485 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* value */
	T6 a2; /* index */
	T0* a3; /* cast_type */
	T6 a4; /* compressed_position */
	T0* a5; /* literal */
	T0* a6; /* marker */
	T0* a7; /* open_white_characters */
	T0* a8; /* close_white_characters */
	T1 a9; /* is_left_aligned */
	T0* a10; /* type */
};

/* Struct for type detachable ET_UNDERSCORED_INTEGER_CONSTANT */
struct S1486 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* sign */
	T0* a3; /* cast_type */
	T11 a4; /* value */
	T1 a5; /* has_overflow */
	T6 a6; /* compressed_position */
	T0* a7; /* literal */
	T0* a8; /* type */
};

/* Struct for type detachable ET_HEXADECIMAL_INTEGER_CONSTANT */
struct S1487 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* sign */
	T0* a3; /* cast_type */
	T11 a4; /* value */
	T1 a5; /* has_overflow */
	T6 a6; /* compressed_position */
	T0* a7; /* literal */
	T0* a8; /* type */
};

/* Struct for type detachable ET_OCTAL_INTEGER_CONSTANT */
struct S1488 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* sign */
	T0* a3; /* cast_type */
	T11 a4; /* value */
	T1 a5; /* has_overflow */
	T6 a6; /* compressed_position */
	T0* a7; /* literal */
	T0* a8; /* type */
};

/* Struct for type detachable ET_BINARY_INTEGER_CONSTANT */
struct S1489 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* sign */
	T0* a3; /* cast_type */
	T11 a4; /* value */
	T1 a5; /* has_overflow */
	T6 a6; /* compressed_position */
	T0* a7; /* literal */
	T0* a8; /* type */
};

/* Struct for type detachable ET_REGULAR_REAL_CONSTANT */
struct S1490 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* sign */
	T6 a2; /* index */
	T0* a3; /* cast_type */
	T6 a4; /* compressed_position */
	T0* a5; /* literal */
	T0* a6; /* type */
};

/* Struct for type detachable ET_UNDERSCORED_REAL_CONSTANT */
struct S1491 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* sign */
	T6 a2; /* index */
	T0* a3; /* cast_type */
	T6 a4; /* compressed_position */
	T0* a5; /* literal */
	T0* a6; /* type */
};

/* Struct for type detachable ET_TRUE_CONSTANT */
struct S1493 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T6 a2; /* compressed_position */
	T2 a3; /* code */
	T0* a4; /* text */
};

/* Struct for type detachable ET_FALSE_CONSTANT */
struct S1494 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T6 a2; /* compressed_position */
	T2 a3; /* code */
	T0* a4; /* text */
};

/* Struct for type detachable ET_C1_CHARACTER_CONSTANT */
struct S1495 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* cast_type */
	T6 a3; /* compressed_position */
	T3 a4; /* value */
	T0* a5; /* type */
};

/* Struct for type detachable ET_C2_CHARACTER_CONSTANT */
struct S1496 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* cast_type */
	T6 a3; /* compressed_position */
	T3 a4; /* value */
	T0* a5; /* type */
};

/* Struct for type detachable ET_TAGGED_INDEXING */
struct S1498 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* tag */
	T0* a2; /* terms */
};

/* Struct for type detachable ET_INDEXING */
struct S1499 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* terms */
};

/* Struct for type detachable ET_CUSTOM_ATTRIBUTE */
struct S1500 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* creation_expression */
	T0* a2; /* settings */
	T0* a3; /* end_keyword */
};

/* Struct for type detachable ET_CONSTRAINED_FORMAL_PARAMETER */
struct S1501 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* arrow_symbol */
	T0* a2; /* constraint */
	T0* a3; /* creation_procedures */
	T0* a4; /* constraint_base_types */
	T0* a5; /* implementation_class */
	T0* a6; /* type_mark */
	T6 a7; /* index */
	T0* a8; /* name */
	T0* a9; /* recursive_formal_constraints */
};

/* Struct for type detachable ET_TYPE_RENAME_CONSTRAINT */
struct S1502 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type */
	T0* a2; /* renames */
};

/* Struct for type detachable ET_CONSTRAINT_NAMED_TYPE */
struct S1503 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type_mark */
	T0* a2; /* name */
};

/* Struct for type detachable ET_CONSTRAINT_LABELED_ACTUAL_PARAMETER */
struct S1504 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* label */
	T0* a2; /* colon */
	T0* a3; /* type */
};

/* Struct for type detachable ET_CONSTRAINT_LABELED_COMMA_ACTUAL_PARAMETER */
struct S1505 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* label */
	T0* a2; /* comma */
	T0* a3; /* type */
};

/* Struct for type detachable ET_ALL_EXPORT */
struct S1506 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* clients_clause */
	T0* a2; /* all_keyword */
};

/* Struct for type detachable ET_CONVERT_FUNCTION */
struct S1508 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* types */
	T0* a3; /* colon */
};

/* Struct for type detachable ET_CONVERT_PROCEDURE */
struct S1509 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* types */
	T0* a3; /* left_parenthesis */
	T0* a4; /* right_parenthesis */
};

/* Struct for type detachable ET_DO_FUNCTION */
struct S1511 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* first_seed */
	T0* a2; /* iteration_components */
	T0* a3; /* implementation_class */
	T0* a4; /* implementation_feature */
	T0* a5; /* other_seeds */
	T0* a6; /* first_precursor */
	T0* a7; /* other_precursors */
	T0* a8; /* preconditions */
	T6 a9; /* hash_code */
	T0* a10; /* arguments */
	T0* a11; /* postconditions */
	T0* a12; /* assigner */
	T0* a13; /* extended_name */
	T0* a14; /* declared_type */
	T0* a15; /* frozen_keyword */
	T0* a16; /* obsolete_message */
	T0* a17; /* locals */
	T0* a18; /* compound */
	T0* a19; /* rescue_clause */
	T0* a20; /* clients */
	T0* a21; /* is_keyword */
	T0* a22; /* end_keyword */
	T0* a23; /* semicolon */
	T0* a24; /* feature_clause */
	T0* a25; /* first_indexing */
	T0* a26; /* object_tests */
	T1 a27; /* validity_checked */
	T1 a28; /* has_validity_error */
	T6 a29; /* id */
	T6 a30; /* version */
	T0* a31; /* synonym */
};

/* Struct for type detachable ET_ONCE_FUNCTION */
struct S1512 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* first_seed */
	T0* a2; /* iteration_components */
	T0* a3; /* implementation_class */
	T0* a4; /* implementation_feature */
	T0* a5; /* other_seeds */
	T0* a6; /* first_precursor */
	T0* a7; /* other_precursors */
	T0* a8; /* preconditions */
	T6 a9; /* hash_code */
	T0* a10; /* arguments */
	T0* a11; /* postconditions */
	T0* a12; /* assigner */
	T0* a13; /* keys */
	T0* a14; /* extended_name */
	T0* a15; /* declared_type */
	T0* a16; /* first_indexing */
	T0* a17; /* frozen_keyword */
	T0* a18; /* obsolete_message */
	T0* a19; /* locals */
	T0* a20; /* compound */
	T0* a21; /* rescue_clause */
	T0* a22; /* clients */
	T0* a23; /* is_keyword */
	T0* a24; /* end_keyword */
	T0* a25; /* semicolon */
	T0* a26; /* feature_clause */
	T0* a27; /* object_tests */
	T1 a28; /* validity_checked */
	T1 a29; /* has_validity_error */
	T6 a30; /* id */
	T6 a31; /* version */
	T0* a32; /* synonym */
};

/* Struct for type detachable ET_DEFERRED_FUNCTION */
struct S1513 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* first_seed */
	T0* a2; /* iteration_components */
	T0* a3; /* implementation_class */
	T0* a4; /* implementation_feature */
	T0* a5; /* other_seeds */
	T0* a6; /* first_precursor */
	T0* a7; /* other_precursors */
	T0* a8; /* preconditions */
	T6 a9; /* hash_code */
	T0* a10; /* arguments */
	T0* a11; /* postconditions */
	T0* a12; /* assigner */
	T0* a13; /* extended_name */
	T0* a14; /* declared_type */
	T0* a15; /* frozen_keyword */
	T0* a16; /* obsolete_message */
	T0* a17; /* clients */
	T0* a18; /* is_keyword */
	T0* a19; /* deferred_keyword */
	T0* a20; /* end_keyword */
	T0* a21; /* semicolon */
	T0* a22; /* feature_clause */
	T0* a23; /* first_indexing */
	T0* a24; /* object_tests */
	T1 a25; /* validity_checked */
	T1 a26; /* has_validity_error */
	T6 a27; /* id */
	T6 a28; /* version */
	T0* a29; /* synonym */
};

/* Struct for type detachable ET_ONCE_PROCEDURE */
struct S1515 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* first_seed */
	T0* a2; /* iteration_components */
	T0* a3; /* implementation_class */
	T0* a4; /* implementation_feature */
	T0* a5; /* other_seeds */
	T0* a6; /* compound */
	T0* a7; /* first_precursor */
	T0* a8; /* other_precursors */
	T0* a9; /* preconditions */
	T6 a10; /* hash_code */
	T0* a11; /* arguments */
	T0* a12; /* postconditions */
	T0* a13; /* keys */
	T0* a14; /* extended_name */
	T0* a15; /* first_indexing */
	T0* a16; /* frozen_keyword */
	T0* a17; /* obsolete_message */
	T0* a18; /* locals */
	T0* a19; /* rescue_clause */
	T0* a20; /* clients */
	T0* a21; /* is_keyword */
	T0* a22; /* end_keyword */
	T0* a23; /* semicolon */
	T0* a24; /* feature_clause */
	T0* a25; /* object_tests */
	T1 a26; /* validity_checked */
	T1 a27; /* has_validity_error */
	T6 a28; /* id */
	T6 a29; /* version */
	T0* a30; /* synonym */
};

/* Struct for type detachable ET_EXTERNAL_PROCEDURE */
struct S1516 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* first_seed */
	T0* a2; /* iteration_components */
	T0* a3; /* implementation_class */
	T0* a4; /* implementation_feature */
	T0* a5; /* other_seeds */
	T0* a6; /* first_precursor */
	T0* a7; /* other_precursors */
	T0* a8; /* preconditions */
	T6 a9; /* hash_code */
	T8 a10; /* builtin_class_code */
	T8 a11; /* builtin_feature_code */
	T0* a12; /* arguments */
	T0* a13; /* postconditions */
	T0* a14; /* extended_name */
	T0* a15; /* frozen_keyword */
	T0* a16; /* language */
	T0* a17; /* obsolete_message */
	T0* a18; /* alias_clause */
	T0* a19; /* clients */
	T0* a20; /* is_keyword */
	T0* a21; /* end_keyword */
	T0* a22; /* semicolon */
	T0* a23; /* feature_clause */
	T0* a24; /* first_indexing */
	T0* a25; /* object_tests */
	T1 a26; /* validity_checked */
	T1 a27; /* has_validity_error */
	T6 a28; /* id */
	T6 a29; /* version */
	T0* a30; /* synonym */
};

/* Struct for type detachable ET_ALIASED_FEATURE_NAME */
struct S1517 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* feature_name */
	T0* a2; /* alias_names */
};

/* Struct for type detachable ET_ALIAS_FREE_NAME */
struct S1518 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* alias_keyword */
	T0* a2; /* alias_string */
	T2 a3; /* code */
	T6 a4; /* hash_code */
	T0* a5; /* convert_keyword */
	T6 a6; /* seed */
};

/* Struct for type detachable ET_CLASS_ASSERTION */
struct S1521 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* class_keyword */
};

/* Struct for type detachable ET_LABELED_ACTUAL_PARAMETER */
struct S1522 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* declared_type */
	T0* a2; /* label_item */
};

/* Struct for type detachable ET_LIKE_FEATURE */
struct S1524 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type_mark */
	T0* a2; /* like_keyword */
	T0* a3; /* name */
	T6 a4; /* seed */
	T1 a5; /* is_procedure */
};

/* Struct for type detachable ET_QUALIFIED_LIKE_BRACED_TYPE */
struct S1525 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type_mark */
	T0* a2; /* like_keyword */
	T0* a3; /* target_type */
	T0* a4; /* implementation_class */
	T0* a5; /* qualified_name */
	T0* a6; /* left_brace */
	T0* a7; /* right_brace */
};

/* Struct for type detachable ET_QUALIFIED_LIKE_TYPE */
struct S1526 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type_mark */
	T0* a2; /* target_type */
	T0* a3; /* implementation_class */
	T0* a4; /* qualified_name */
};

/* Struct for type detachable ET_ASSIGNER_INSTRUCTION */
struct S1527 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* call */
	T0* a2; /* source */
	T0* a3; /* assign_symbol */
	T0* a4; /* name */
};

/* Struct for type detachable ET_ASSIGNMENT */
struct S1528 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target */
	T0* a2; /* source */
	T0* a3; /* assign_symbol */
};

/* Struct for type detachable ET_ASSIGNMENT_ATTEMPT */
struct S1529 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target */
	T0* a2; /* source */
	T0* a3; /* assign_attempt_symbol */
};

/* Struct for type detachable ET_LOOP_INSTRUCTION */
struct S1530 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* from_compound */
	T0* a2; /* until_conditional */
	T0* a3; /* loop_compound */
	T0* a4; /* end_keyword */
	T0* a5; /* invariant_part */
	T0* a6; /* variant_part */
	T1 a7; /* has_old_variant_syntax */
};

/* Struct for type detachable ET_BANG_INSTRUCTION */
struct S1532 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type */
	T0* a2; /* target */
	T0* a3; /* creation_call */
	T0* a4; /* left_bang */
	T0* a5; /* right_bang */
};

/* Struct for type detachable ET_CREATE_INSTRUCTION */
struct S1533 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* creation_type */
	T0* a2; /* target */
	T0* a3; /* creation_call */
	T0* a4; /* create_keyword */
	T0* a5; /* creation_region */
};

/* Struct for type detachable ET_CHOICE_RANGE */
struct S1534 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* lower */
	T0* a2; /* upper */
	T0* a3; /* dotdot */
};

/* Struct for type detachable ET_PRECURSOR_INSTRUCTION */
struct S1535 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parent_name */
	T0* a2; /* arguments */
	T0* a3; /* precursor_keyword */
	T1 a4; /* is_parent_prefixed */
	T0* a5; /* parenthesis_call */
	T0* a6; /* parent_type */
};

/* Struct for type detachable ET_STATIC_CALL_INSTRUCTION */
struct S1537 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* static_type */
	T0* a2; /* feature_keyword */
	T0* a3; /* qualified_name */
	T0* a4; /* arguments */
	T0* a5; /* parenthesis_call */
};

/* Struct for type detachable ET_PRECURSOR_EXPRESSION */
struct S1538 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T1 a2; /* is_parent_prefixed */
	T0* a3; /* parent_name */
	T0* a4; /* precursor_keyword */
	T0* a5; /* arguments */
	T0* a6; /* parenthesis_call */
	T0* a7; /* parent_type */
};

/* Struct for type detachable ET_FEATURE_ADDRESS */
struct S1539 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* dollar */
	T0* a3; /* name */
};

/* Struct for type detachable ET_CURRENT_ADDRESS */
struct S1540 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* dollar */
	T0* a3; /* current_keyword */
};

/* Struct for type detachable ET_RESULT_ADDRESS */
struct S1541 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* dollar */
	T0* a3; /* result_keyword */
};

/* Struct for type detachable ET_EXPRESSION_ADDRESS */
struct S1542 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* dollar */
	T0* a3; /* expression */
};

/* Struct for type detachable ET_INFIX_EXPRESSION */
struct S1543 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* left */
	T0* a3; /* name */
	T0* a4; /* right */
	T1 a5; /* is_boolean_operator */
};

/* Struct for type detachable ET_INFIX_AND_THEN_OPERATOR */
struct S1544 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* and_keyword */
	T0* a2; /* then_keyword */
	T6 a3; /* seed */
};

/* Struct for type detachable ET_INFIX_OR_ELSE_OPERATOR */
struct S1545 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* or_keyword */
	T0* a2; /* else_keyword */
	T6 a3; /* seed */
};

/* Struct for type detachable ET_MANIFEST_TYPE */
struct S1546 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* left_brace */
	T0* a3; /* type */
	T0* a4; /* right_brace */
};

/* Struct for type detachable ET_PREFIX_EXPRESSION */
struct S1547 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* name */
	T0* a3; /* expression */
	T1 a4; /* is_boolean_operator */
};

/* Struct for type detachable ET_OLD_EXPRESSION */
struct S1548 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* old_keyword */
	T0* a3; /* expression */
};

/* Struct for type detachable ET_OLD_OBJECT_TEST */
struct S1549 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* name */
	T0* a3; /* left_brace */
	T0* a4; /* attached_keyword */
	T0* a5; /* as_keyword */
	T0* a6; /* colon */
	T0* a7; /* declared_type */
	T0* a8; /* right_brace */
	T0* a9; /* expression */
};

/* Struct for type detachable ET_ONCE_MANIFEST_STRING */
struct S1550 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* once_keyword */
	T0* a3; /* manifest_string */
	T6 a4; /* id */
};

/* Struct for type detachable ET_ITERATION_CURSOR */
struct S1551 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* at_symbol */
	T0* a3; /* item_name */
};

/* Struct for type detachable ET_DO_FUNCTION_INLINE_AGENT */
struct S1552 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* agent_keyword */
	T0* a3; /* formal_arguments */
	T0* a4; /* declared_type */
	T0* a5; /* end_keyword */
	T0* a6; /* implicit_result */
	T0* a7; /* object_tests */
	T0* a8; /* iteration_components */
	T0* a9; /* actual_arguments */
	T0* a10; /* preconditions */
	T0* a11; /* locals */
	T0* a12; /* compound */
	T0* a13; /* postconditions */
	T0* a14; /* rescue_clause */
	T0* a15; /* target */
	T6 a16; /* attached_result_index */
	T6 a17; /* result_index */
};

/* Struct for type detachable ET_ONCE_FUNCTION_INLINE_AGENT */
struct S1553 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* agent_keyword */
	T0* a3; /* formal_arguments */
	T0* a4; /* declared_type */
	T0* a5; /* end_keyword */
	T0* a6; /* implicit_result */
	T0* a7; /* object_tests */
	T0* a8; /* iteration_components */
	T0* a9; /* actual_arguments */
	T0* a10; /* preconditions */
	T0* a11; /* locals */
	T0* a12; /* keys */
	T0* a13; /* compound */
	T0* a14; /* postconditions */
	T0* a15; /* rescue_clause */
	T0* a16; /* target */
	T6 a17; /* attached_result_index */
	T6 a18; /* result_index */
};

/* Struct for type detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT */
struct S1554 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* agent_keyword */
	T0* a3; /* language */
	T0* a4; /* object_tests */
	T0* a5; /* iteration_components */
	T0* a6; /* actual_arguments */
	T0* a7; /* preconditions */
	T0* a8; /* alias_clause */
	T0* a9; /* postconditions */
	T0* a10; /* formal_arguments */
	T0* a11; /* declared_type */
	T0* a12; /* end_keyword */
	T0* a13; /* implicit_result */
	T0* a14; /* target */
	T6 a15; /* attached_result_index */
	T6 a16; /* result_index */
};

/* Struct for type detachable ET_DO_PROCEDURE_INLINE_AGENT */
struct S1555 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* agent_keyword */
	T0* a3; /* formal_arguments */
	T0* a4; /* end_keyword */
	T0* a5; /* object_tests */
	T0* a6; /* iteration_components */
	T0* a7; /* actual_arguments */
	T0* a8; /* preconditions */
	T0* a9; /* locals */
	T0* a10; /* compound */
	T0* a11; /* postconditions */
	T0* a12; /* rescue_clause */
	T0* a13; /* target */
	T6 a14; /* attached_result_index */
	T6 a15; /* result_index */
};

/* Struct for type detachable ET_ONCE_PROCEDURE_INLINE_AGENT */
struct S1556 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* agent_keyword */
	T0* a3; /* formal_arguments */
	T0* a4; /* end_keyword */
	T0* a5; /* object_tests */
	T0* a6; /* iteration_components */
	T0* a7; /* actual_arguments */
	T0* a8; /* preconditions */
	T0* a9; /* locals */
	T0* a10; /* keys */
	T0* a11; /* compound */
	T0* a12; /* postconditions */
	T0* a13; /* rescue_clause */
	T0* a14; /* target */
	T6 a15; /* attached_result_index */
	T6 a16; /* result_index */
};

/* Struct for type detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT */
struct S1557 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* agent_keyword */
	T0* a3; /* language */
	T0* a4; /* object_tests */
	T0* a5; /* iteration_components */
	T0* a6; /* actual_arguments */
	T0* a7; /* preconditions */
	T0* a8; /* alias_clause */
	T0* a9; /* postconditions */
	T0* a10; /* formal_arguments */
	T0* a11; /* end_keyword */
	T0* a12; /* target */
	T6 a13; /* attached_result_index */
	T6 a14; /* result_index */
};

/* Struct for type detachable ET_AGENT_OPEN_TARGET */
struct S1558 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* left_brace */
	T0* a2; /* type */
	T0* a3; /* right_brace */
	T6 a4; /* index */
};

/* Struct for type detachable ET_AGENT_TYPED_OPEN_ARGUMENT */
struct S1559 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type */
	T0* a2; /* left_brace */
	T0* a3; /* right_brace */
	T0* a4; /* question_mark */
	T6 a5; /* index */
};

/* Struct for type detachable ET_TAGGED_ASSERTION */
struct S1562 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* tag */
	T0* a2; /* untagged_assertion */
};

/* Struct for type [detachable] ET_FILE_POSITION */
struct S1564 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* compressed_position */
	T0* a2; /* filename */
};

/* Struct for type [detachable] DS_QUICK_SORTER [[attached] ET_QUERY] */
struct S1566 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* comparator */
};

/* Struct for type [detachable] DS_QUICK_SORTER [[attached] ET_PROCEDURE] */
struct S1567 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* comparator */
};

/* Struct for type [detachable] ET_REDECLARED_FEATURE */
struct S1569 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next */
	T0* a2; /* flattened_feature */
	T0* a3; /* parent_feature */
	T6 a4; /* first_seed */
	T0* a5; /* other_seeds */
	T1 a6; /* is_selected */
	T0* a7; /* replicated_features */
	T0* a8; /* replicated_seeds */
};

/* Struct for type [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER */
struct S1571 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* last_state */
	T1 a2; /* internal_overflowed */
	T11 a3; /* part1 */
	T11 a4; /* part2 */
	T0* a5; /* leading_separators */
	T0* a6; /* trailing_separators */
	T6 a7; /* sign */
	T6 a8; /* conversion_type */
	T2 a9; /* internal_lookahead */
	T1 a10; /* leading_separators_acceptable */
	T1 a11; /* trailing_separators_acceptable */
};

/* Struct for type [detachable] TUPLE [[attached] ET_MASTER_CLASS, [attached] ET_SYSTEM_PROCESSOR] */
struct S1573 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] SPECIAL [[attached] NATURAL_16] */
struct S1576 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T9 z2[1]; /* item */
};

/* Struct for type [detachable] ARRAY [[attached] NATURAL_16] */
struct S1577 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] ARRAY [[attached] NATURAL_32] */
struct S1578 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER] */
struct S1579 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY] */
struct S1580 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
struct S1581 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE] */
struct S1582 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG] */
struct S1583 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE] */
struct S1584 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY] */
struct S1585 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
struct S1586 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE] */
struct S1587 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT] */
struct S1588 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE] */
struct S1589 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ACTION] */
struct S1590 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_ACTION] */
struct S1591 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ACTION] */
struct S1592 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type detachable DS_ARRAYED_LIST [[attached] RX_PCRE_REGULAR_EXPRESSION] */
struct S1593 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_OPTIONS] */
struct S1594 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_OPTIONS] */
struct S1597 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_OPTIONS, [attached] STRING_8] */
struct S1598 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_VISIBLE_CLASS] */
struct S1599 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_VISIBLE_CLASS] */
struct S1600 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_VISIBLE_CLASS] */
struct S1601 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS] */
struct S1602 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CONDITION_ITEM] */
struct S1603 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_DOTNET_ASSEMBLY] */
struct S1604 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_ASSEMBLY] */
struct S1605 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_DOTNET_ASSEMBLY] */
struct S1606 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_UNIVERSE] */
struct S1608 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_UNIVERSE] */
struct S1609 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_UNIVERSE] */
struct S1611 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_INTERNAL_UNIVERSE] */
struct S1612 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INTERNAL_UNIVERSE] */
struct S1614 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_INTERNAL_UNIVERSE] */
struct S1615 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] ET_CLASS_TYPE_STATUS_CHECKER1 */
struct S1616 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
};

/* Struct for type [detachable] ET_CLASS_TYPE_STATUS_CHECKER2 */
struct S1617 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
};

/* Struct for type [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER */
struct S1618 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
	T0* a2; /* current_class */
	T0* a3; /* system_processor */
};

/* Struct for type [detachable] ET_CLASS_TYPE_STATUS_CHECKER3 */
struct S1619 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_fatal_error */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS] */
struct S1620 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_TYPE] */
struct S1621 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_TYPE] */
struct S1622 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_FORMAL_PARAMETER_TYPE] */
struct S1623 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] CONSOLE */
struct S1624 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* file_pointer */
	T0* a2; /* last_string_32 */
	T6 a3; /* mode */
	T0* a4; /* last_string */
	T0* a5; /* internal_name */
	T0* a6; /* internal_detachable_name_pointer */
	T2 a7; /* last_character */
};

/* Struct for type [detachable] DS_LINKED_QUEUE [[attached] CHARACTER_8] */
struct S1626 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* first_cell */
	T0* a3; /* last_cell */
};

/* Struct for type [detachable] UC_UTF16_ROUTINES */
struct S1627 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_LIBRARY] */
struct S1628 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIBRARY] */
struct S1629 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_LIBRARY] */
struct S1630 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_TARGET] */
struct S1631 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ADAPTED_UNIVERSE] */
struct S1632 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_UNIVERSE] */
struct S1633 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_UNIVERSE] */
struct S1634 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_TARGET_PARENT] */
struct S1635 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET_PARENT] */
struct S1636 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_TARGET_PARENT] */
struct S1637 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] YY_UNICODE_BUFFER */
struct S1638 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* content */
	T6 a2; /* index */
	T6 a3; /* line */
	T6 a4; /* column */
	T6 a5; /* position */
	T6 a6; /* count */
	T1 a7; /* filled */
	T1 a8; /* beginning_of_line */
	T6 a9; /* capacity */
};

/* Struct for type [detachable] SPECIAL [detachable ET_OBJECT_TEST_LIST] */
struct S1639 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBJECT_TEST_LIST] */
struct S1640 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST_LIST] */
struct S1641 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_OBJECT_TEST_LIST] */
struct S1642 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] SPECIAL [detachable ET_ITERATION_COMPONENT_LIST] */
struct S1643 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ITERATION_COMPONENT_LIST] */
struct S1644 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ITERATION_COMPONENT_LIST] */
struct S1645 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ITERATION_COMPONENT_LIST] */
struct S1646 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ASSERTION_ITEM] */
struct S1647 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTION_ITEM] */
struct S1648 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ASSERTION_ITEM] */
struct S1649 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T6 a2; /* position */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_CONSTRAINT_TYPE] */
struct S1650 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_NAMED_CLASS] */
struct S1651 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_CLASS] */
struct S1653 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS] */
struct S1655 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T6 a2; /* found_position */
	T0* a3; /* item_storage */
	T6 a4; /* count */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* modulus */
	T6 a8; /* capacity */
	T6 a9; /* slots_position */
	T6 a10; /* clashes_previous_position */
	T0* a11; /* key_equality_tester */
	T0* a12; /* equality_tester */
	T0* a13; /* internal_cursor */
	T0* a14; /* key_storage */
	T0* a15; /* clashes */
	T0* a16; /* slots */
	T0* a17; /* special_item_routines */
	T0* a18; /* special_key_routines */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_BASE_TYPE, [attached] ET_CLASS] */
struct S1656 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT] */
struct S1659 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32] */
struct S1660 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* storage */
	T0* a2; /* special_routines */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] ET_BASE_TYPE_CONSTRAINT_LIST */
struct S1661 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* left_brace */
	T0* a2; /* right_brace */
	T0* a3; /* storage */
	T6 a4; /* count */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME] */
struct S1662 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_FLATTENED_FEATURE] */
struct S1663 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_FEATURE_NAME] */
struct S1666 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FLATTENED_FEATURE] */
struct S1667 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME] */
struct S1668 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ALIAS_NAME] */
struct S1669 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ALIAS_NAME] */
struct S1671 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME] */
struct S1672 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_LIST] */
struct S1673 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CLIENT_LIST] */
struct S1674 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLIENT_LIST] */
struct S1675 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS] */
struct S1677 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT] */
struct S1678 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CLIENT] */
struct S1679 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME] */
struct S1680 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* capacity */
	T6 a2; /* found_position */
	T0* a3; /* internal_cursor */
	T6 a4; /* count */
	T0* a5; /* item_storage */
	T6 a6; /* modulus */
	T6 a7; /* last_position */
	T6 a8; /* free_slot */
	T6 a9; /* position */
	T0* a10; /* equality_tester */
	T0* a11; /* special_item_routines */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T6 a14; /* slots_position */
	T6 a15; /* clashes_previous_position */
	T0* a16; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME] */
struct S1681 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T6 a3; /* found_position */
	T0* a4; /* internal_cursor */
	T0* a5; /* key_storage */
	T0* a6; /* item_storage */
	T0* a7; /* key_equality_tester */
	T0* a8; /* internal_keys */
	T6 a9; /* position */
	T6 a10; /* last_position */
	T6 a11; /* free_slot */
	T6 a12; /* modulus */
	T0* a13; /* equality_tester */
	T6 a14; /* slots_position */
	T6 a15; /* clashes_previous_position */
	T0* a16; /* special_item_routines */
	T0* a17; /* clashes */
	T0* a18; /* slots */
	T0* a19; /* special_key_routines */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32] */
struct S1682 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* internal_cursor */
	T6 a2; /* found_position */
	T0* a3; /* item_storage */
	T0* a4; /* key_storage */
	T6 a5; /* count */
	T6 a6; /* last_position */
	T6 a7; /* free_slot */
	T6 a8; /* position */
	T6 a9; /* capacity */
	T0* a10; /* equality_tester */
	T0* a11; /* key_equality_tester */
	T0* a12; /* clashes */
	T6 a13; /* modulus */
	T0* a14; /* slots */
	T6 a15; /* slots_position */
	T6 a16; /* clashes_previous_position */
	T0* a17; /* special_item_routines */
	T0* a18; /* special_key_routines */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] ET_REPLICATED_FEATURE */
struct S1684 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* features */
	T6 a2; /* selected_count */
};

/* Struct for type [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE] */
struct S1685 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* last_cell */
	T6 a2; /* count */
	T0* a3; /* first_cell */
	T0* a4; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE] */
struct S1686 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] ET_DOTNET_SIGNATURE_TESTER */
struct S1687 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE] */
struct S1688 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* capacity */
	T6 a2; /* found_position */
	T0* a3; /* item_storage */
	T6 a4; /* modulus */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T0* a8; /* equality_tester */
	T6 a9; /* count */
	T6 a10; /* slots_position */
	T0* a11; /* special_item_routines */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T0* a14; /* internal_cursor */
	T6 a15; /* clashes_previous_position */
	T0* a16; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE] */
struct S1690 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T0* a3; /* key_equality_tester */
	T0* a4; /* internal_keys */
	T6 a5; /* count */
	T6 a6; /* last_position */
	T6 a7; /* free_slot */
	T6 a8; /* position */
	T6 a9; /* capacity */
	T0* a10; /* equality_tester */
	T0* a11; /* internal_cursor */
	T0* a12; /* key_storage */
	T0* a13; /* clashes */
	T6 a14; /* modulus */
	T0* a15; /* slots */
	T6 a16; /* slots_position */
	T6 a17; /* clashes_previous_position */
	T0* a18; /* special_item_routines */
	T0* a19; /* special_key_routines */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE] */
struct S1692 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* last_cell */
	T6 a2; /* count */
	T0* a3; /* first_cell */
	T0* a4; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE] */
struct S1693 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* cycle */
	T0* a2; /* counts */
	T0* a3; /* successors */
	T0* a4; /* items */
	T0* a5; /* sorted_items */
	T0* a6; /* indexes */
};

/* Struct for type detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE] */
struct S1694 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_PARENT_FEATURE] */
struct S1695 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* after */
	T0* a2; /* current_cell */
	T0* a3; /* container */
	T1 a4; /* before */
	T0* a5; /* next_cursor */
};

/* Struct for type detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE] */
struct S1696 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* last_cell */
	T6 a2; /* count */
	T0* a3; /* first_cell */
	T0* a4; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8] */
struct S1697 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* equality_tester */
	T0* a2; /* key_equality_tester */
	T6 a3; /* capacity */
	T6 a4; /* modulus */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T0* a8; /* special_item_routines */
	T0* a9; /* item_storage */
	T0* a10; /* special_key_routines */
	T0* a11; /* key_storage */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T6 a14; /* found_position */
	T0* a15; /* internal_cursor */
	T6 a16; /* slots_position */
	T6 a17; /* count */
	T6 a18; /* clashes_previous_position */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE, [attached] INTEGER_32] */
struct S1698 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_CLASS] */
struct S1701 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_FEATURE_NAME] */
struct S1703 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_RENAME] */
struct S1704 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME] */
struct S1707 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type detachable DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_ALIAS_NAME] */
struct S1711 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_PROCEDURE] */
struct S1712 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_QUERY] */
struct S1713 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] YY_BUFFER] */
struct S1714 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] YY_BUFFER] */
struct S1715 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_DYNAMIC_TARGET] */
struct S1716 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TARGET] */
struct S1717 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ELEMENT_CONTENT] */
struct S1719 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T1 a2; /* before */
};

/* Struct for type detachable DS_BILINKABLE [[attached] XM_DTD_ELEMENT_CONTENT] */
struct S1720 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
	T0* a3; /* left */
};

/* Struct for type [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]] */
struct S1722 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* last_cell */
	T0* a3; /* first_cell */
	T0* a4; /* internal_cursor */
};

/* Struct for type [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]] */
struct S1723 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* before */
	T0* a2; /* current_cell */
	T0* a3; /* next_cursor */
	T0* a4; /* container */
	T1 a5; /* after */
};

/* Struct for type detachable DS_LINKABLE [detachable STRING_8] */
struct S1724 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CLUSTER] */
struct S1725 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLUSTER] */
struct S1726 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLUSTER] */
struct S1727 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_TYPE_ITEM] */
struct S1728 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_ITEM] */
struct S1729 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_MASTER_CLASS] */
struct S1730 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DOTNET_ASSEMBLY] */
struct S1731 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_CLUSTER] */
struct S1732 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_CLUSTER]] */
struct S1733 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_TARGET, [attached] STRING_8] */
struct S1734 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_ADAPTED_LIBRARY] */
struct S1736 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_ADAPTED_LIBRARY]] */
struct S1737 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
struct S1738 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY]] */
struct S1739 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] STRING_8]] */
struct S1742 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8] */
struct S1743 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8] */
struct S1744 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* key_storage */
	T0* a3; /* item_storage */
	T0* a4; /* key_equality_tester */
	T0* a5; /* internal_keys */
	T0* a6; /* hash_function */
	T6 a7; /* position */
	T6 a8; /* last_position */
	T6 a9; /* capacity */
	T6 a10; /* slots_position */
	T6 a11; /* count */
	T6 a12; /* free_slot */
	T0* a13; /* equality_tester */
	T6 a14; /* modulus */
	T6 a15; /* clashes_previous_position */
	T0* a16; /* special_item_routines */
	T0* a17; /* clashes */
	T0* a18; /* slots */
	T0* a19; /* special_key_routines */
	T0* a20; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8] */
struct S1745 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T6 a3; /* position */
	T0* a4; /* key_equality_tester */
	T0* a5; /* internal_keys */
	T0* a6; /* hash_function */
	T6 a7; /* last_position */
	T6 a8; /* capacity */
	T6 a9; /* slots_position */
	T6 a10; /* count */
	T0* a11; /* equality_tester */
	T6 a12; /* modulus */
	T6 a13; /* clashes_previous_position */
	T0* a14; /* special_item_routines */
	T0* a15; /* clashes */
	T0* a16; /* slots */
	T0* a17; /* special_key_routines */
	T0* a18; /* key_storage */
	T6 a19; /* free_slot */
	T0* a20; /* internal_cursor */
};

/* Struct for type detachable DS_LINKABLE [[attached] XM_ELEMENT_NODE] */
struct S1750 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* right */
	T0* a2; /* item */
};

/* Struct for type detachable ET_CONSTRAINT_GENERIC_NAMED_TYPE */
struct S1755 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type_mark */
	T0* a2; /* name */
	T0* a3; /* actual_parameters */
};

/* Struct for type detachable ET_UNQUALIFIED_CALL_INSTRUCTION */
struct S1756 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* arguments */
	T0* a3; /* parenthesis_call */
};

/* Struct for type detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT */
struct S1758 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* current_cluster */
	T0* a2; /* group_names */
	T0* a3; /* group_pathnames */
	T0* a4; /* pathname_buffer */
};

/* Struct for type [detachable] ET_SEEDED_QUERY_COMPARATOR */
struct S1759 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_SEEDED_PROCEDURE_COMPARATOR */
struct S1762 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] TUPLE [[attached] ET_CLIENT_LIST] */
struct S1766 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] ET_FORMAL_COMMA_ARGUMENT */
struct S1767 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name_item */
	T0* a2; /* declared_type */
	T6 a3; /* index */
	T6 a4; /* attached_index */
	T1 a5; /* is_used */
};

/* Struct for type [detachable] ET_LOCAL_COMMA_VARIABLE */
struct S1768 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name_item */
	T0* a2; /* declared_type */
	T6 a3; /* index */
	T6 a4; /* attached_index */
	T1 a5; /* is_used */
};

/* Struct for type [detachable] ET_LABELED_COMMA_ACTUAL_PARAMETER */
struct S1769 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* declared_type */
	T0* a2; /* label_item */
};

/* Struct for type [detachable] ET_KEYWORD_EXPRESSION */
struct S1771 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* expression */
	T0* a2; /* keyword */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_RENAME_ITEM] */
struct S1776 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME_ITEM] */
struct S1777 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_EXPORT] */
struct S1779 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPORT] */
struct S1780 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_FEATURE_NAME_ITEM] */
struct S1782 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME_ITEM] */
struct S1783 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_FORMAL_PARAMETER_ITEM] */
struct S1784 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_PARAMETER_ITEM] */
struct S1785 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_CFLAG] */
struct S1788 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_CFLAG]] */
struct S1789 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_TARGET, [attached] ET_ECF_STATE] */
struct S1790 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_INCLUDE] */
struct S1791 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_INCLUDE]] */
struct S1792 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_LIBRARY] */
struct S1793 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_LIBRARY]] */
struct S1794 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
struct S1795 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_LINKER_FLAG]] */
struct S1796 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_MAKE] */
struct S1797 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_MAKE]] */
struct S1798 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_OBJECT] */
struct S1799 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_OBJECT]] */
struct S1800 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_RESOURCE] */
struct S1801 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_RESOURCE]] */
struct S1802 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type detachable DS_LINKABLE [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]] */
struct S1806 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM] */
struct S1809 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM] */
struct S1810 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_AGENT_IMPLICIT_CURRENT_TARGET */
struct S1811 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* agent_expression */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CHOICE_ITEM] */
struct S1813 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CHOICE_ITEM] */
struct S1814 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
struct S1815 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
struct S1817 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CONVERT_FEATURE_ITEM] */
struct S1819 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONVERT_FEATURE_ITEM] */
struct S1820 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_CREATOR] */
struct S1821 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CREATOR] */
struct S1822 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ELSEIF_EXPRESSION] */
struct S1823 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_EXPRESSION] */
struct S1824 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ELSEIF_PART] */
struct S1825 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_PART] */
struct S1826 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_FEATURE_CLAUSE] */
struct S1827 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_CLAUSE] */
struct S1828 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_INDEXING_ITEM] */
struct S1830 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_ITEM] */
struct S1831 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_INDEXING_TERM_ITEM] */
struct S1832 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM_ITEM] */
struct S1833 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_LOCAL_VARIABLE_ITEM] */
struct S1834 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LOCAL_VARIABLE_ITEM] */
struct S1836 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_MANIFEST_STRING_ITEM] */
struct S1838 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MANIFEST_STRING_ITEM] */
struct S1839 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_PARENT_LIST] */
struct S1840 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_LIST] */
struct S1841 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_TYPE_CONSTRAINT_ITEM] */
struct S1843 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_CONSTRAINT_ITEM] */
struct S1844 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_WHEN_EXPRESSION] */
struct S1845 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_WHEN_EXPRESSION] */
struct S1846 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_WHEN_PART] */
struct S1847 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_WHEN_PART] */
struct S1848 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT */
struct S1850 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* type */
	T0* a2; /* renames */
};

/* Struct for type detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS] */
struct S1853 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* table */
	T0* a2; /* equality_tester */
	T0* a3; /* internal_cursor */
};

/* Struct for type [detachable] ET_STANDARD_ONCE_KEYS */
struct S1857 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_CLUSTER] */
struct S1859 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CLUSTER] */
struct S1860 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_CLUSTER] */
struct S1861 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_LIBRARY] */
struct S1862 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_LIBRARY] */
struct S1863 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ADAPTED_LIBRARY] */
struct S1864 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
struct S1865 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
struct S1866 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
struct S1867 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_FILE_RULE] */
struct S1868 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_FILE_RULE] */
struct S1869 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_FILE_RULE] */
struct S1870 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_CFLAG] */
struct S1871 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_CFLAG] */
struct S1872 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_CFLAG] */
struct S1873 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_INCLUDE] */
struct S1874 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_INCLUDE] */
struct S1875 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_INCLUDE] */
struct S1876 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LIBRARY] */
struct S1877 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LIBRARY] */
struct S1878 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_LIBRARY] */
struct S1879 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
struct S1880 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
struct S1881 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
struct S1882 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_MAKE] */
struct S1883 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_MAKE] */
struct S1884 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_MAKE] */
struct S1885 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_OBJECT] */
struct S1886 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_OBJECT] */
struct S1887 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_OBJECT] */
struct S1888 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_RESOURCE] */
struct S1889 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_RESOURCE] */
struct S1890 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_RESOURCE] */
struct S1891 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] RX_PCRE_REGULAR_EXPRESSION] */
struct S1892 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] RX_PCRE_REGULAR_EXPRESSION] */
struct S1893 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] RX_PCRE_REGULAR_EXPRESSION] */
struct S1894 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ANDED_CONDITIONS] */
struct S1895 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_ANDED_CONDITIONS] */
struct S1896 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ANDED_CONDITIONS] */
struct S1897 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CONDITION_ITEM] */
struct S1898 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_ECF_CONDITION_ITEM] */
struct S1899 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_CONDITION_ITEM] */
struct S1900 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type detachable DS_LINKABLE [[attached] CHARACTER_8] */
struct S1902 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T2 a1; /* item */
	T0* a2; /* right */
};

/* Struct for type detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_CLASS] */
struct S1903 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_BASE_TYPE_CONSTRAINT] */
struct S1904 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE_CONSTRAINT] */
struct S1905 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_BASE_TYPE_CONSTRAINT] */
struct S1906 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] NATURAL_32] */
struct S1907 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_FEATURE_NAME] */
struct S1908 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] ET_FEATURE_NAME] */
struct S1909 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32] */
struct S1911 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_REPLICABLE_FEATURE] */
struct S1912 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_REPLICABLE_FEATURE] */
struct S1914 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_ADAPTED_FEATURE] */
struct S1915 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T1 a2; /* before */
	T1 a3; /* after */
	T0* a4; /* current_cell */
	T0* a5; /* next_cursor */
};

/* Struct for type detachable DS_LINKABLE [[attached] ET_ADAPTED_FEATURE] */
struct S1916 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_FEATURE] */
struct S1917 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_PARENT_FEATURE] */
struct S1918 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE] */
struct S1919 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_DOTNET_FEATURE] */
struct S1920 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] ET_DOTNET_FEATURE] */
struct S1921 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_FEATURE] */
struct S1922 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]] */
struct S1923 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type detachable DS_HASH_TABLE_CURSOR [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE] */
struct S1926 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]] */
struct S1927 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_DOTNET_FEATURE] */
struct S1928 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T1 a2; /* before */
	T1 a3; /* after */
	T0* a4; /* current_cell */
	T0* a5; /* next_cursor */
};

/* Struct for type detachable DS_LINKABLE [[attached] ET_DOTNET_FEATURE] */
struct S1929 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE] */
struct S1930 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* last_position */
	T6 a3; /* free_slot */
	T6 a4; /* position */
	T6 a5; /* modulus */
	T6 a6; /* capacity */
	T0* a7; /* equality_tester */
	T0* a8; /* key_equality_tester */
	T0* a9; /* internal_cursor */
	T6 a10; /* found_position */
	T0* a11; /* item_storage */
	T0* a12; /* key_storage */
	T0* a13; /* clashes */
	T0* a14; /* slots */
	T0* a15; /* special_item_routines */
	T0* a16; /* special_key_routines */
	T0* a17; /* hash_function */
	T6 a18; /* slots_position */
	T6 a19; /* clashes_previous_position */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_LIKE_FEATURE] */
struct S1931 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIKE_FEATURE] */
struct S1932 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE] */
struct S1933 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T6 a2; /* position */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8] */
struct S1935 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]] */
struct S1936 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME] */
struct S1937 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T0* a3; /* key_equality_tester */
	T0* a4; /* internal_keys */
	T6 a5; /* position */
	T6 a6; /* last_position */
	T6 a7; /* capacity */
	T6 a8; /* slots_position */
	T6 a9; /* count */
	T0* a10; /* equality_tester */
	T6 a11; /* modulus */
	T6 a12; /* clashes_previous_position */
	T0* a13; /* special_item_routines */
	T0* a14; /* clashes */
	T0* a15; /* slots */
	T0* a16; /* special_key_routines */
	T0* a17; /* key_storage */
	T6 a18; /* free_slot */
	T0* a19; /* internal_cursor */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]] */
struct S1938 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type detachable DS_BILINKABLE [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]] */
struct S1939 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* left */
	T0* a3; /* right */
};

/* Struct for type [detachable] ET_CLUSTER_ERROR */
struct S1940 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* default_template */
	T0* a2; /* parameters */
	T0* a3; /* code */
	T0* a4; /* etl_code */
	T0* a5; /* cluster */
};

/* Struct for type [detachable] ET_SYNTAX_ERROR */
struct S1941 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* default_template */
	T0* a2; /* parameters */
	T0* a3; /* code */
	T0* a4; /* etl_code */
	T0* a5; /* filename */
	T0* a6; /* position */
	T1 a7; /* ise_reported */
	T1 a8; /* ise_fatal */
	T1 a9; /* ge_reported */
	T1 a10; /* ge_fatal */
};

/* Struct for type [detachable] ET_DOTNET_ASSEMBLY_ERROR */
struct S1942 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* default_template */
	T0* a2; /* parameters */
	T0* a3; /* code */
	T0* a4; /* etl_code */
	T0* a5; /* assembly */
};

/* Struct for type [detachable] AP_OPTION_COMPARATOR */
struct S1943 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_QUICK_SORTER [[attached] AP_OPTION] */
struct S1944 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* comparator */
};

/* Struct for type [detachable] ST_SPLITTER */
struct S1953 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* has_escape_character */
	T2 a2; /* escape_character */
	T0* a3; /* separator_codes */
	T0* a4; /* separators */
};

/* Struct for type [detachable] KL_CHARACTER_ROUTINES */
struct S1954 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] TUPLE [[attached] ET_C_GENERATOR, INTEGER_32, [attached] ET_CURRENT, [attached] ET_DYNAMIC_TUPLE_TYPE, BOOLEAN] */
struct S1955 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T6 z2;
	T0* z3;
	T0* z4;
	T1 z5;
};

/* Struct for type [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS */
struct S1956 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* actual_parameters */
	T6 a3; /* tuple_position */
	T0* a4; /* tuple_type */
};

/* Struct for type [detachable] ET_ACTUAL_PARAMETER_SUBLIST */
struct S1957 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* actual_parameters */
	T6 a2; /* upper */
	T6 a3; /* lower */
};

/* Struct for type [detachable] SPECIAL [[attached] ET_FORMAL_ARGUMENT_ITEM] */
struct S1960 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_ARGUMENT_ITEM] */
struct S1962 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_DYNAMIC_PRECURSOR] */
struct S1964 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRECURSOR] */
struct S1965 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_INSTRUCTION] */
struct S1966 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INSTRUCTION] */
struct S1968 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST */
struct S1977 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* actual_arguments */
	T0* a3; /* left_symbol */
	T0* a4; /* storage */
	T0* a5; /* right_symbol */
};

/* Struct for type [detachable] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8] */
struct S1978 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T8 z3;
};

/* Struct for type [detachable] ET_INFIX_CAST_EXPRESSION */
struct S1980 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* expression */
	T0* a3; /* type */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] STRING_8] */
struct S1984 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]] */
struct S1985 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]] */
struct S1988 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8] */
struct S1989 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS] */
struct S1990 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* table_cursor */
	T0* a2; /* container */
};

/* Struct for type detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE] */
struct S1993 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T0* a2; /* container */
	T6 a3; /* position */
};

/* Struct for type [detachable] SPECIAL [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]] */
struct S1994 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]] */
struct S1997 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME] */
struct S1998 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] ET_UNIVERSE_ERROR */
struct S1999 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* default_template */
	T0* a2; /* parameters */
	T0* a3; /* code */
	T0* a4; /* etl_code */
	T0* a5; /* universe */
};

/* Struct for type [detachable] TUPLE [[attached] ET_SYSTEM, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
struct S2000 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T6 z3;
};

/* Struct for type detachable ET_PARENTHESIS_INSTRUCTION */
struct S2002 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target */
	T0* a2; /* name */
	T0* a3; /* arguments */
};

/* Struct for type [detachable] ET_PARENTHESIS_SYMBOL */
struct S2020 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* seed */
	T6 a2; /* compressed_position */
	T2 a3; /* code */
};

/* Struct for type [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT_LIST */
struct S2026 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
};

/* Struct for type [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT */
struct S2027 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* agent_expression */
	T6 a2; /* argument_index */
	T6 a3; /* index */
};

/* Struct for type [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
struct S2034 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* table_cursor */
	T0* a2; /* container */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8] */
struct S2040 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* last_position */
	T0* a3; /* clashes */
	T0* a4; /* item_storage */
	T0* a5; /* key_equality_tester */
	T0* a6; /* internal_keys */
	T6 a7; /* position */
	T6 a8; /* capacity */
	T6 a9; /* slots_position */
	T0* a10; /* equality_tester */
	T6 a11; /* found_position */
	T6 a12; /* modulus */
	T6 a13; /* clashes_previous_position */
	T0* a14; /* special_item_routines */
	T0* a15; /* slots */
	T0* a16; /* special_key_routines */
	T0* a17; /* key_storage */
	T6 a18; /* free_slot */
	T0* a19; /* internal_cursor */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8] */
struct S2041 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] ET_DYNAMIC_EQUALITY_EXPRESSION */
struct S2042 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target_type_set */
	T0* a2; /* static_equality */
	T0* a3; /* current_feature */
	T0* a4; /* current_type */
	T0* a5; /* next */
};

/* Struct for type [detachable] ET_DYNAMIC_OBJECT_EQUALITY_EXPRESSION */
struct S2043 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target_type_set */
	T0* a2; /* static_equality */
	T0* a3; /* current_feature */
	T0* a4; /* current_type */
	T0* a5; /* next */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_LIBRARY] */
struct S2044 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] ET_CONVERT_BUILTIN_EXPRESSION */
struct S2046 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* index */
	T0* a2; /* expression */
	T0* a3; /* convert_feature */
	T0* a4; /* type */
};

/* Struct for type [detachable] ET_CONVERT_FROM_EXPRESSION */
struct S2047 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T0* a2; /* type */
	T6 a3; /* index */
	T0* a4; /* expression */
	T0* a5; /* convert_feature */
};

/* Struct for type [detachable] ET_CONVERT_TO_EXPRESSION */
struct S2048 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T6 a2; /* index */
	T0* a3; /* expression */
	T0* a4; /* convert_feature */
};

/* Struct for type detachable DS_CELL [detachable ET_CLASS] */
struct S2051 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
};

/* Struct for type [detachable] TUPLE [[attached] DS_CELL [detachable ET_CLASS]] */
struct S2052 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] ET_EIFFEL_PREPARSER, [attached] DS_ARRAYED_LIST [[attached] STRING_8]] */
struct S2054 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_FEATURE] */
struct S2056 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* next_cursor */
	T6 a2; /* position */
	T0* a3; /* container */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_FILE_RULE] */
struct S2057 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_FILE_RULE]] */
struct S2058 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_FILE_RULES] */
struct S2059 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PREDICATE [[attached] TUPLE [[attached] ET_ECF_FILE_RULE]] */
struct S2060 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_STATE] */
struct S2061 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PREDICATE [[attached] TUPLE [[attached] STRING_8]] */
struct S2063 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] KL_STRING_ROUTINES, [attached] STRING_8] */
struct S2064 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] TUPLE [[attached] DS_HASH_SET [[attached] STRING_8]] */
struct S2066 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] LX_DFA_WILDCARD */
struct S2068 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* yy_nxt */
	T0* a2; /* yy_accept */
	T6 a3; /* subject_start */
	T6 a4; /* subject_end */
	T6 a5; /* match_count */
	T6 a6; /* matched_start */
	T6 a7; /* matched_end */
	T0* a8; /* yy_ec */
	T6 a9; /* yyNull_equiv_class */
	T6 a10; /* yyMax_symbol_equiv_class */
	T6 a11; /* yyNb_rows */
	T0* a12; /* pattern */
	T1 a13; /* is_case_insensitive */
	T0* a14; /* subject */
};

/* Struct for type detachable DS_LINKABLE [[attached] ET_PARENT_FEATURE] */
struct S2071 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* item */
	T0* a2; /* right */
};

/* Struct for type [detachable] TUPLE [[attached] LX_WILDCARD] */
struct S2073 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PREDICATE [[attached] TUPLE [[attached] LX_WILDCARD]] */
struct S2074 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] TUPLE [[attached] ET_CLUSTER_DEPENDENCE_CONSTRAINT, [attached] ET_GROUP, [attached] STRING_8] */
struct S2075 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
};

/* Struct for type [detachable] SPECIAL [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT] */
struct S2077 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT] */
struct S2078 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] LX_WILDCARD] */
struct S2079 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_WILDCARD] */
struct S2082 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
struct S2086 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T0* z3;
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_LIBRARY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
struct S2089 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T6 z3;
};

/* Struct for type [detachable] TUPLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
struct S2090 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T6 z3;
};

/* Struct for type [detachable] TUPLE [[attached] ET_DOTNET_ASSEMBLY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
struct S2091 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T6 z3;
};

/* Struct for type [detachable] TUPLE [[attached] ET_LIBRARY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
struct S2092 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
	T6 z3;
};

/* Struct for type [detachable] ET_AST_NULL_LEAF */
struct S2094 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* compressed_position */
};

/* Struct for type [detachable] LX_WILDCARD_PARSER */
struct S2095 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* successful */
	T0* a2; /* pending_rules */
	T0* a3; /* start_condition_stack */
	T0* a4; /* action_factory */
	T0* a5; /* old_singleton_lines */
	T0* a6; /* old_singleton_columns */
	T0* a7; /* old_singleton_counts */
	T0* a8; /* old_regexp_lines */
	T0* a9; /* old_regexp_columns */
	T0* a10; /* old_regexp_counts */
	T0* a11; /* unions_of_concatenations_of_symbol_classes_by_utf8_character_class */
	T0* a12; /* buffer */
	T6 a13; /* maximum_used_symbol */
	T0* a14; /* description */
	T0* a15; /* last_string_value */
	T0* a16; /* last_string_32_value */
	T0* a17; /* last_lx_symbol_class_value */
	T0* a18; /* error_handler */
	T0* a19; /* name_definitions */
	T0* a20; /* character_classes */
	T0* a21; /* character_classes_by_name */
	T0* a22; /* equiv_character_classes */
	T0* a23; /* utf8_character_classes_by_name */
	T0* a24; /* utf8_mode */
	T6 a25; /* line_nb */
	T0* a26; /* eiffel_verbatim_marker */
	T0* a27; /* yyss */
	T0* a28; /* input_buffer */
	T6 a29; /* yy_end */
	T6 a30; /* yy_position */
	T6 a31; /* yy_line */
	T6 a32; /* yy_column */
	T6 a33; /* yy_parsing_status */
	T6 a34; /* yy_suspended_yystacksize */
	T6 a35; /* yy_suspended_yystate */
	T6 a36; /* yy_suspended_yyn */
	T6 a37; /* yy_suspended_yychar1 */
	T6 a38; /* yy_suspended_index */
	T6 a39; /* yy_suspended_yyss_top */
	T6 a40; /* yy_suspended_yy_goto */
	T0* a41; /* yyr1 */
	T6 a42; /* yyssp */
	T0* a43; /* yypgoto */
	T0* a44; /* yycheck */
	T0* a45; /* yytable */
	T0* a46; /* yydefgoto */
	T6 a47; /* error_count */
	T1 a48; /* yy_lookahead_needed */
	T6 a49; /* yyerrstatus */
	T0* a50; /* yypact */
	T6 a51; /* last_token */
	T0* a52; /* yytranslate */
	T0* a53; /* yydefact */
	T0* a54; /* yy_pushed_start_conditions */
	T0* a55; /* yyspecial_routines1 */
	T6 a56; /* yyvsc1 */
	T0* a57; /* yyvs1 */
	T0* a58; /* yyspecial_routines2 */
	T6 a59; /* yyvsc2 */
	T0* a60; /* yyvs2 */
	T0* a61; /* yyspecial_routines3 */
	T6 a62; /* yyvsc3 */
	T0* a63; /* yyvs3 */
	T0* a64; /* yyspecial_routines4 */
	T6 a65; /* yyvsc4 */
	T0* a66; /* yyvs4 */
	T0* a67; /* yyspecial_routines5 */
	T6 a68; /* yyvsc5 */
	T0* a69; /* yyvs5 */
	T0* a70; /* yytypes1 */
	T0* a71; /* yytypes2 */
	T6 a72; /* yy_start */
	T6 a73; /* yyvsp1 */
	T6 a74; /* yyvsp2 */
	T6 a75; /* yyvsp3 */
	T6 a76; /* yyvsp4 */
	T6 a77; /* yyvsp5 */
	T1 a78; /* yy_more_flag */
	T6 a79; /* yy_more_len */
	T6 a80; /* line */
	T6 a81; /* column */
	T6 a82; /* position */
	T6 a83; /* yy_start_state */
	T0* a84; /* yy_state_stack */
	T6 a85; /* yy_state_count */
	T0* a86; /* yy_content_area */
	T0* a87; /* yy_unicode_content_area */
	T0* a88; /* yy_accept */
	T6 a89; /* yy_last_accepting_state */
	T6 a90; /* yy_last_accepting_cpos */
	T0* a91; /* yy_content */
	T0* a92; /* yy_ec */
	T0* a93; /* yy_chk */
	T0* a94; /* yy_base */
	T0* a95; /* yy_def */
	T0* a96; /* yy_meta */
	T0* a97; /* yy_nxt */
	T6 a98; /* yy_lp */
	T0* a99; /* yy_acclist */
	T6 a100; /* yy_looking_for_trail_begin */
	T6 a101; /* yy_full_match */
	T6 a102; /* yy_full_state */
	T6 a103; /* yy_full_lp */
	T1 a104; /* yy_rejected */
	T0* a105; /* last_detachable_any_value */
	T6 a106; /* last_integer_value */
	T1 a107; /* in_trail_context */
	T0* a108; /* rule */
	T1 a109; /* has_trail_context */
	T6 a110; /* head_count */
	T6 a111; /* head_line */
	T6 a112; /* head_column */
	T6 a113; /* trail_count */
	T6 a114; /* i_ */
	T6 a115; /* rule_line_nb */
	T1 a116; /* has_bol_context */
};

/* Struct for type [detachable] LX_DESCRIPTION */
struct S2096 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* maximum_symbol */
	T1 a2; /* utf8_mode */
	T1 a3; /* equiv_classes_used */
	T1 a4; /* full_table */
	T1 a5; /* meta_equiv_classes_used */
	T1 a6; /* reject_used */
	T1 a7; /* variable_trail_context */
	T0* a8; /* rules */
	T0* a9; /* start_conditions */
	T6 a10; /* array_size */
	T1 a11; /* line_pragma */
	T1 a12; /* inspect_used */
	T0* a13; /* eof_rules */
	T0* a14; /* eiffel_header */
	T1 a15; /* case_insensitive */
	T0* a16; /* equiv_classes */
	T1 a17; /* bol_needed */
	T0* a18; /* input_filename */
	T1 a19; /* has_utf8_enconding */
	T1 a20; /* actions_separated */
	T0* a21; /* eiffel_code */
	T1 a22; /* pre_action_used */
	T1 a23; /* post_action_used */
	T1 a24; /* pre_eof_action_used */
	T1 a25; /* post_eof_action_used */
	T1 a26; /* line_used */
	T1 a27; /* position_used */
};

/* Struct for type [detachable] LX_FULL_DFA */
struct S2097 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* yy_nxt */
	T0* a2; /* yy_accept */
	T0* a3; /* yy_ec */
	T6 a4; /* yyNb_rows */
	T6 a5; /* yyNull_equiv_class */
	T6 a6; /* yyMax_symbol_equiv_class */
	T0* a7; /* input_filename */
	T1 a8; /* has_utf8_enconding */
	T6 a9; /* array_size */
	T1 a10; /* inspect_used */
	T1 a11; /* actions_separated */
	T0* a12; /* eiffel_code */
	T0* a13; /* eiffel_header */
	T1 a14; /* bol_needed */
	T1 a15; /* pre_action_used */
	T1 a16; /* post_action_used */
	T1 a17; /* pre_eof_action_used */
	T1 a18; /* post_eof_action_used */
	T1 a19; /* line_pragma */
	T0* a20; /* yy_start_conditions */
	T6 a21; /* yyNb_rules */
	T0* a22; /* yy_rules */
	T6 a23; /* yyEnd_of_buffer */
	T1 a24; /* yyLine_used */
	T1 a25; /* yyPosition_used */
	T6 a26; /* minimum_symbol */
	T6 a27; /* maximum_symbol */
	T0* a28; /* states */
	T6 a29; /* backing_up_count */
	T0* a30; /* partitions */
	T6 a31; /* start_states_count */
	T1 a32; /* yyBacking_up */
	T0* a33; /* yy_eof_rules */
};

/* Struct for type [detachable] TUPLE [[attached] ET_GROUP] */
struct S2099 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]], [attached] PREDICATE [[attached] TUPLE [[attached] ET_CLASS]]] */
struct S2100 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
	T0* z2;
};

/* Struct for type [detachable] TUPLE [[attached] RX_PCRE_REGULAR_EXPRESSION] */
struct S2101 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* z1;
};

/* Struct for type [detachable] PREDICATE [[attached] TUPLE [[attached] RX_PCRE_REGULAR_EXPRESSION]] */
struct S2102 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T14 a1; /* rout_disp */
	T0* a2; /* closed_operands */
	T14 a3; /* encaps_rout_disp */
	T14 a4; /* calc_rout_addr */
	T1 a5; /* is_target_closed */
	T6 a6; /* open_count */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] LX_RULE] */
struct S2105 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] LX_START_CONDITIONS */
struct S2106 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] LX_ACTION_FACTORY */
struct S2107 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS] */
struct S2108 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T0* a3; /* key_equality_tester */
	T6 a4; /* position */
	T6 a5; /* count */
	T6 a6; /* capacity */
	T6 a7; /* free_slot */
	T6 a8; /* last_position */
	T0* a9; /* equality_tester */
	T6 a10; /* modulus */
	T6 a11; /* slots_position */
	T6 a12; /* clashes_previous_position */
	T0* a13; /* clashes */
	T0* a14; /* slots */
	T0* a15; /* special_item_routines */
	T0* a16; /* special_key_routines */
	T0* a17; /* key_storage */
	T0* a18; /* internal_cursor */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] LX_SYMBOL_CLASS */
struct S2109 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* upper */
	T1 a2; /* is_empty */
	T1 a3; /* is_negated */
	T11 a4; /* first_set */
	T11 a5; /* second_set */
	T11 a6; /* third_set */
	T11 a7; /* fourth_set */
	T0* a8; /* other_sets */
	T6 a9; /* lower */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8] */
struct S2110 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* key_equality_tester */
	T0* a2; /* internal_keys */
	T0* a3; /* equality_tester */
	T6 a4; /* capacity */
	T6 a5; /* modulus */
	T6 a6; /* last_position */
	T6 a7; /* free_slot */
	T6 a8; /* position */
	T0* a9; /* special_item_routines */
	T0* a10; /* item_storage */
	T0* a11; /* special_key_routines */
	T0* a12; /* key_storage */
	T0* a13; /* clashes */
	T0* a14; /* slots */
	T6 a15; /* found_position */
	T0* a16; /* internal_cursor */
};

/* Struct for type [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS] */
struct S2111 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T0* a3; /* equality_tester */
	T6 a4; /* capacity */
	T6 a5; /* modulus */
	T6 a6; /* last_position */
	T6 a7; /* free_slot */
	T6 a8; /* position */
	T6 a9; /* count */
	T6 a10; /* slots_position */
	T0* a11; /* special_item_routines */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T0* a14; /* internal_cursor */
	T6 a15; /* clashes_previous_position */
	T0* a16; /* hash_function */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32] */
struct S2112 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* found_position */
	T0* a2; /* item_storage */
	T0* a3; /* key_equality_tester */
	T0* a4; /* internal_keys */
	T6 a5; /* position */
	T6 a6; /* count */
	T6 a7; /* capacity */
	T6 a8; /* slots_position */
	T6 a9; /* free_slot */
	T6 a10; /* last_position */
	T0* a11; /* equality_tester */
	T6 a12; /* modulus */
	T6 a13; /* clashes_previous_position */
	T0* a14; /* special_item_routines */
	T0* a15; /* clashes */
	T0* a16; /* slots */
	T0* a17; /* special_key_routines */
	T0* a18; /* key_storage */
	T0* a19; /* internal_cursor */
	T0* a20; /* hash_function */
};

/* Struct for type [detachable] KL_EQUALITY_TESTER [[attached] STRING_32] */
struct S2113 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_STACK [[attached] BOOLEAN] */
struct S2114 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* storage */
	T6 a2; /* count */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_32] */
struct S2115 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] STRING_32] */
struct S2116 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_SYMBOL_CLASS] */
struct S2117 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] LX_SYMBOL_CLASS] */
struct S2118 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA] */
struct S2119 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] LX_NFA */
struct S2120 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* in_trail_context */
	T0* a2; /* states */
};

/* Struct for type [detachable] SPECIAL [[attached] LX_NFA] */
struct S2121 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type detachable LX_RULE */
struct S2122 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* id */
	T0* a2; /* pattern */
	T0* a3; /* action */
	T6 a4; /* head_count */
	T6 a5; /* trail_count */
	T6 a6; /* line_count */
	T6 a7; /* column_count */
	T6 a8; /* line_nb */
	T1 a9; /* has_trail_context */
	T1 a10; /* is_useful */
};

/* Struct for type [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS] */
struct S2123 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* container */
	T0* a3; /* next_cursor */
};

/* Struct for type [detachable] LX_EQUIVALENCE_CLASSES */
struct S2124 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* new_upper */
	T0* a2; /* storage */
};

/* Struct for type [detachable] LX_UNRECOGNIZED_RULE_ERROR */
struct S2125 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] LX_INVALID_UNICODE_CHARACTER_ERROR */
struct S2126 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] LX_MISSING_QUOTE_ERROR */
struct S2127 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] LX_BAD_CHARACTER_CLASS_ERROR */
struct S2128 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] LX_BAD_CHARACTER_ERROR */
struct S2129 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] LX_FULL_AND_META_ERROR */
struct S2130 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] LX_FULL_AND_REJECT_ERROR */
struct S2131 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] LX_FULL_AND_VARIABLE_TRAILING_CONTEXT_ERROR */
struct S2132 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] LX_CHARACTER_OUT_OF_RANGE_ERROR */
struct S2133 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] ARRAY [[attached] LX_RULE] */
struct S2134 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] SPECIAL [[attached] LX_RULE] */
struct S2135 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] LX_DFA_STATE */
struct S2136 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* accepted_rules */
	T0* a2; /* states */
	T0* a3; /* transitions */
	T0* a4; /* accepted_head_rules */
	T6 a5; /* code */
	T6 a6; /* id */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE] */
struct S2137 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* equality_tester */
	T0* a4; /* special_routines */
	T6 a5; /* capacity */
	T0* a6; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] LX_DFA_STATE] */
struct S2138 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* a3; /* storage */
	T0* a4; /* special_routines */
	T0* a5; /* internal_cursor */
};

/* Struct for type detachable LX_SYMBOL_PARTITIONS */
struct S2139 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* symbols */
	T0* a2; /* storage */
	T6 a3; /* new_upper */
};

/* Struct for type [detachable] KL_ARRAY_ROUTINES [[attached] LX_RULE] */
struct S2140 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] ARRAY [detachable LX_RULE] */
struct S2141 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] SPECIAL [detachable LX_RULE] */
struct S2142 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] LX_START_CONDITION */
struct S2143 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* name */
	T6 a2; /* id */
	T1 a3; /* is_exclusive */
	T0* a4; /* patterns */
	T0* a5; /* bol_patterns */
};

/* Struct for type [detachable] LX_TRANSITION_TABLE [[attached] LX_DFA_STATE] */
struct S2144 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* lower */
	T6 a2; /* upper */
	T6 a3; /* minimum_label */
	T6 a4; /* second_minimum_label */
	T6 a5; /* maximum_label */
	T0* a6; /* transitions */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] LX_NFA] */
struct S2145 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T0* a4; /* internal_cursor */
	T6 a5; /* count */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32] */
struct S2146 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* item_storage */
	T0* a3; /* key_equality_tester */
	T6 a4; /* count */
	T6 a5; /* capacity */
	T6 a6; /* slots_position */
	T6 a7; /* free_slot */
	T6 a8; /* last_position */
	T6 a9; /* modulus */
	T6 a10; /* clashes_previous_position */
	T0* a11; /* equality_tester */
	T6 a12; /* found_position */
	T0* a13; /* special_item_routines */
	T0* a14; /* clashes */
	T0* a15; /* slots */
	T0* a16; /* special_key_routines */
	T0* a17; /* key_storage */
	T0* a18; /* internal_cursor */
	T0* a19; /* hash_function */
};

/* Struct for type [detachable] LX_NFA_STATE */
struct S2147 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T1 a1; /* in_trail_context */
	T0* a2; /* transition */
	T0* a3; /* epsilon_transition */
	T6 a4; /* id */
	T0* a5; /* accepted_rule */
};

/* Struct for type detachable ARRAY [detachable LX_SYMBOL_CLASS] */
struct S2148 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] SPECIAL [detachable LX_SYMBOL_CLASS] */
struct S2149 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type detachable ARRAY [detachable DS_HASH_SET [[attached] INTEGER_32]] */
struct S2150 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] SPECIAL [detachable DS_HASH_SET [[attached] INTEGER_32]] */
struct S2151 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
struct S2152 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* count */
	T0* a2; /* storage */
	T0* a3; /* special_routines */
	T6 a4; /* capacity */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS] */
struct S2153 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* special_routines */
	T0* a2; /* storage */
	T6 a3; /* capacity */
	T6 a4; /* count */
	T0* a5; /* internal_cursor */
};

/* Struct for type [detachable] LX_NEGATIVE_RANGE_IN_CHARACTER_CLASS_ERROR */
struct S2154 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* parameters */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_RULE] */
struct S2155 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_RULE] */
struct S2156 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] KL_ARRAY_ROUTINES [[attached] STRING_8] */
struct S2157 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] LX_START_CONDITION] */
struct S2158 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_START_CONDITION] */
struct S2159 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_START_CONDITION] */
struct S2160 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]] */
struct S2161 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type detachable KL_EQUALITY_TESTER [[attached] LX_SYMBOL_CLASS] */
struct S2162 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]] */
struct S2164 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS] */
struct S2165 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type detachable SPECIAL [detachable SPECIAL [[attached] NATURAL_64]] */
struct S2166 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_32, [attached] STRING_8] */
struct S2168 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_SYMBOL_CLASS, [attached] STRING_32] */
struct S2170 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] LX_EPSILON_TRANSITION [[attached] LX_NFA_STATE] */
struct S2171 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* target */
};

/* Struct for type [detachable] LX_SYMBOL_CLASS_TRANSITION [[attached] LX_NFA_STATE] */
struct S2174 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* label */
	T0* a2; /* target */
};

/* Struct for type [detachable] LX_SYMBOL_TRANSITION [[attached] LX_NFA_STATE] */
struct S2175 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* label */
	T0* a2; /* target */
};

/* Struct for type [detachable] LX_ACTION */
struct S2177 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* text */
};

/* Struct for type [detachable] ARRAY [[attached] DS_BILINKABLE [[attached] INTEGER_32]] */
struct S2178 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] DS_BILINKABLE [[attached] INTEGER_32] */
struct S2179 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* item */
	T0* a2; /* left */
	T0* a3; /* right */
};

/* Struct for type [detachable] SPECIAL [[attached] DS_BILINKABLE [[attached] INTEGER_32]] */
struct S2180 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] DS_BUBBLE_SORTER [[attached] LX_NFA_STATE] */
struct S2181 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* comparator */
};

/* Struct for type [detachable] DS_BUBBLE_SORTER [[attached] LX_RULE] */
struct S2183 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* comparator */
};

/* Struct for type [detachable] SPECIAL [[attached] LX_NFA_STATE] */
struct S2185 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA_STATE] */
struct S2187 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE] */
struct S2188 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T6 a1; /* position */
	T0* a2; /* next_cursor */
	T0* a3; /* container */
};

/* Struct for type [detachable] SPECIAL [[attached] LX_DFA_STATE] */
struct S2190 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_DFA_STATE] */
struct S2191 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_DFA_STATE] */
struct S2192 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_ARRAY [[attached] LX_RULE] */
struct S2193 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32] */
struct S2194 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* equality_tester */
	T0* a2; /* key_equality_tester */
	T6 a3; /* capacity */
	T6 a4; /* modulus */
	T6 a5; /* last_position */
	T6 a6; /* free_slot */
	T6 a7; /* position */
	T0* a8; /* special_item_routines */
	T0* a9; /* item_storage */
	T0* a10; /* special_key_routines */
	T0* a11; /* key_storage */
	T0* a12; /* clashes */
	T0* a13; /* slots */
	T6 a14; /* found_position */
	T0* a15; /* internal_cursor */
	T0* a16; /* hash_function */
	T6 a17; /* count */
	T6 a18; /* slots_position */
	T6 a19; /* clashes_previous_position */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA] */
struct S2195 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_NFA, [attached] INTEGER_32] */
struct S2197 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_RULE] */
struct S2201 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_NFA_STATE] */
struct S2204 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
struct S2207 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	uint32_t offset;
	T6 a1; /* count */
	T6 a2; /* capacity */
	T0* z2[1]; /* item */
};

/* Struct for type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
struct S2208 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
struct S2209 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_SYMBOL_CLASS] */
struct S2210 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};

/* Struct for type [detachable] KL_ARRAY [[attached] STRING_8] */
struct S2211 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* area */
	T6 a2; /* lower */
	T6 a3; /* upper */
};

/* Struct for type [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_DFA_STATE, [attached] INTEGER_32] */
struct S2213 {
	EIF_TYPE_INDEX id;
	uint16_t flags;
	T0* a1; /* container */
	T6 a2; /* position */
};


extern T1 GE_default1;
extern T2 GE_default2;
extern T3 GE_default3;
extern T4 GE_default4;
extern T5 GE_default5;
extern T6 GE_default6;
extern T7 GE_default7;
extern T8 GE_default8;
extern T9 GE_default9;
extern T10 GE_default10;
extern T11 GE_default11;
extern T12 GE_default12;
extern T13 GE_default13;
extern T14 GE_default14;
extern T15 GE_default15;
extern T16 GE_default16;
extern T17 GE_default17;
extern T18 GE_default18;
extern T20 GE_default20;
extern T21 GE_default21;
extern T25 GE_default25;
extern T26 GE_default26;
extern T27 GE_default27;
extern T28 GE_default28;
extern T29 GE_default29;
extern T30 GE_default30;
extern T31 GE_default31;
extern T32 GE_default32;
extern T33 GE_default33;
extern T34 GE_default34;
extern T35 GE_default35;
extern T36 GE_default36;
extern T37 GE_default37;
extern T38 GE_default38;
extern T39 GE_default39;
extern T40 GE_default40;
extern T42 GE_default42;
extern T43 GE_default43;
extern T44 GE_default44;
extern T46 GE_default46;
extern T47 GE_default47;
extern T48 GE_default48;
extern T49 GE_default49;
extern T53 GE_default53;
extern T55 GE_default55;
extern T56 GE_default56;
extern T57 GE_default57;
extern T58 GE_default58;
extern T59 GE_default59;
extern T61 GE_default61;
extern T62 GE_default62;
extern T63 GE_default63;
extern T64 GE_default64;
extern T67 GE_default67;
extern T68 GE_default68;
extern T70 GE_default70;
extern T71 GE_default71;
extern T72 GE_default72;
extern T73 GE_default73;
extern T74 GE_default74;
extern T76 GE_default76;
extern T77 GE_default77;
extern T79 GE_default79;
extern T80 GE_default80;
extern T81 GE_default81;
extern T82 GE_default82;
extern T86 GE_default86;
extern T87 GE_default87;
extern T88 GE_default88;
extern T89 GE_default89;
extern T90 GE_default90;
extern T93 GE_default93;
extern T95 GE_default95;
extern T97 GE_default97;
extern T99 GE_default99;
extern T100 GE_default100;
extern T101 GE_default101;
extern T102 GE_default102;
extern T103 GE_default103;
extern T104 GE_default104;
extern T105 GE_default105;
extern T107 GE_default107;
extern T108 GE_default108;
extern T109 GE_default109;
extern T110 GE_default110;
extern T111 GE_default111;
extern T112 GE_default112;
extern T113 GE_default113;
extern T114 GE_default114;
extern T115 GE_default115;
extern T116 GE_default116;
extern T117 GE_default117;
extern T118 GE_default118;
extern T119 GE_default119;
extern T120 GE_default120;
extern T121 GE_default121;
extern T122 GE_default122;
extern T123 GE_default123;
extern T124 GE_default124;
extern T125 GE_default125;
extern T126 GE_default126;
extern T127 GE_default127;
extern T128 GE_default128;
extern T129 GE_default129;
extern T130 GE_default130;
extern T131 GE_default131;
extern T132 GE_default132;
extern T133 GE_default133;
extern T134 GE_default134;
extern T135 GE_default135;
extern T137 GE_default137;
extern T140 GE_default140;
extern T141 GE_default141;
extern T142 GE_default142;
extern T144 GE_default144;
extern T145 GE_default145;
extern T146 GE_default146;
extern T147 GE_default147;
extern T148 GE_default148;
extern T149 GE_default149;
extern T150 GE_default150;
extern T151 GE_default151;
extern T152 GE_default152;
extern T153 GE_default153;
extern T154 GE_default154;
extern T157 GE_default157;
extern T160 GE_default160;
extern T161 GE_default161;
extern T162 GE_default162;
extern T163 GE_default163;
extern T164 GE_default164;
extern T165 GE_default165;
extern T166 GE_default166;
extern T167 GE_default167;
extern T168 GE_default168;
extern T169 GE_default169;
extern T170 GE_default170;
extern T171 GE_default171;
extern T172 GE_default172;
extern T174 GE_default174;
extern T175 GE_default175;
extern T176 GE_default176;
extern T178 GE_default178;
extern T179 GE_default179;
extern T182 GE_default182;
extern T183 GE_default183;
extern T184 GE_default184;
extern T185 GE_default185;
extern T187 GE_default187;
extern T188 GE_default188;
extern T190 GE_default190;
extern T191 GE_default191;
extern T192 GE_default192;
extern T193 GE_default193;
extern T195 GE_default195;
extern T196 GE_default196;
extern T198 GE_default198;
extern T199 GE_default199;
extern T200 GE_default200;
extern T201 GE_default201;
extern T202 GE_default202;
extern T203 GE_default203;
extern T204 GE_default204;
extern T205 GE_default205;
extern T206 GE_default206;
extern T207 GE_default207;
extern T208 GE_default208;
extern T210 GE_default210;
extern T211 GE_default211;
extern T213 GE_default213;
extern T214 GE_default214;
extern T215 GE_default215;
extern T216 GE_default216;
extern T217 GE_default217;
extern T218 GE_default218;
extern T219 GE_default219;
extern T221 GE_default221;
extern T223 GE_default223;
extern T224 GE_default224;
extern T225 GE_default225;
extern T226 GE_default226;
extern T227 GE_default227;
extern T228 GE_default228;
extern T229 GE_default229;
extern T230 GE_default230;
extern T231 GE_default231;
extern T236 GE_default236;
extern T240 GE_default240;
extern T241 GE_default241;
extern T242 GE_default242;
extern T244 GE_default244;
extern T245 GE_default245;
extern T253 GE_default253;
extern T254 GE_default254;
extern T255 GE_default255;
extern T256 GE_default256;
extern T257 GE_default257;
extern T258 GE_default258;
extern T259 GE_default259;
extern T260 GE_default260;
extern T261 GE_default261;
extern T262 GE_default262;
extern T263 GE_default263;
extern T264 GE_default264;
extern T265 GE_default265;
extern T266 GE_default266;
extern T267 GE_default267;
extern T268 GE_default268;
extern T269 GE_default269;
extern T270 GE_default270;
extern T271 GE_default271;
extern T272 GE_default272;
extern T273 GE_default273;
extern T274 GE_default274;
extern T275 GE_default275;
extern T276 GE_default276;
extern T277 GE_default277;
extern T278 GE_default278;
extern T279 GE_default279;
extern T280 GE_default280;
extern T282 GE_default282;
extern T283 GE_default283;
extern T284 GE_default284;
extern T287 GE_default287;
extern T288 GE_default288;
extern T289 GE_default289;
extern T290 GE_default290;
extern T291 GE_default291;
extern T292 GE_default292;
extern T293 GE_default293;
extern T294 GE_default294;
extern T295 GE_default295;
extern T296 GE_default296;
extern T297 GE_default297;
extern T298 GE_default298;
extern T300 GE_default300;
extern T301 GE_default301;
extern T302 GE_default302;
extern T303 GE_default303;
extern T304 GE_default304;
extern T305 GE_default305;
extern T306 GE_default306;
extern T307 GE_default307;
extern T308 GE_default308;
extern T309 GE_default309;
extern T311 GE_default311;
extern T312 GE_default312;
extern T313 GE_default313;
extern T314 GE_default314;
extern T318 GE_default318;
extern T319 GE_default319;
extern T322 GE_default322;
extern T323 GE_default323;
extern T324 GE_default324;
extern T328 GE_default328;
extern T329 GE_default329;
extern T333 GE_default333;
extern T334 GE_default334;
extern T341 GE_default341;
extern T344 GE_default344;
extern T346 GE_default346;
extern T347 GE_default347;
extern T348 GE_default348;
extern T349 GE_default349;
extern T350 GE_default350;
extern T351 GE_default351;
extern T355 GE_default355;
extern T357 GE_default357;
extern T358 GE_default358;
extern T359 GE_default359;
extern T360 GE_default360;
extern T361 GE_default361;
extern T365 GE_default365;
extern T367 GE_default367;
extern T368 GE_default368;
extern T369 GE_default369;
extern T371 GE_default371;
extern T372 GE_default372;
extern T375 GE_default375;
extern T376 GE_default376;
extern T377 GE_default377;
extern T379 GE_default379;
extern T380 GE_default380;
extern T381 GE_default381;
extern T382 GE_default382;
extern T384 GE_default384;
extern T386 GE_default386;
extern T387 GE_default387;
extern T388 GE_default388;
extern T389 GE_default389;
extern T390 GE_default390;
extern T391 GE_default391;
extern T392 GE_default392;
extern T393 GE_default393;
extern T394 GE_default394;
extern T395 GE_default395;
extern T396 GE_default396;
extern T397 GE_default397;
extern T398 GE_default398;
extern T399 GE_default399;
extern T400 GE_default400;
extern T401 GE_default401;
extern T402 GE_default402;
extern T403 GE_default403;
extern T404 GE_default404;
extern T405 GE_default405;
extern T406 GE_default406;
extern T408 GE_default408;
extern T410 GE_default410;
extern T412 GE_default412;
extern T413 GE_default413;
extern T414 GE_default414;
extern T415 GE_default415;
extern T416 GE_default416;
extern T417 GE_default417;
extern T422 GE_default422;
extern T423 GE_default423;
extern T424 GE_default424;
extern T426 GE_default426;
extern T427 GE_default427;
extern T428 GE_default428;
extern T429 GE_default429;
extern T430 GE_default430;
extern T431 GE_default431;
extern T432 GE_default432;
extern T434 GE_default434;
extern T435 GE_default435;
extern T436 GE_default436;
extern T444 GE_default444;
extern T445 GE_default445;
extern T446 GE_default446;
extern T447 GE_default447;
extern T449 GE_default449;
extern T450 GE_default450;
extern T451 GE_default451;
extern T452 GE_default452;
extern T453 GE_default453;
extern T454 GE_default454;
extern T455 GE_default455;
extern T456 GE_default456;
extern T457 GE_default457;
extern T458 GE_default458;
extern T459 GE_default459;
extern T460 GE_default460;
extern T461 GE_default461;
extern T462 GE_default462;
extern T463 GE_default463;
extern T464 GE_default464;
extern T465 GE_default465;
extern T467 GE_default467;
extern T468 GE_default468;
extern T469 GE_default469;
extern T471 GE_default471;
extern T472 GE_default472;
extern T475 GE_default475;
extern T476 GE_default476;
extern T477 GE_default477;
extern T478 GE_default478;
extern T479 GE_default479;
extern T480 GE_default480;
extern T481 GE_default481;
extern T483 GE_default483;
extern T484 GE_default484;
extern T485 GE_default485;
extern T486 GE_default486;
extern T487 GE_default487;
extern T488 GE_default488;
extern T489 GE_default489;
extern T491 GE_default491;
extern T492 GE_default492;
extern T493 GE_default493;
extern T494 GE_default494;
extern T495 GE_default495;
extern T496 GE_default496;
extern T497 GE_default497;
extern T498 GE_default498;
extern T499 GE_default499;
extern T500 GE_default500;
extern T501 GE_default501;
extern T502 GE_default502;
extern T503 GE_default503;
extern T504 GE_default504;
extern T506 GE_default506;
extern T507 GE_default507;
extern T510 GE_default510;
extern T511 GE_default511;
extern T514 GE_default514;
extern T515 GE_default515;
extern T516 GE_default516;
extern T517 GE_default517;
extern T519 GE_default519;
extern T520 GE_default520;
extern T521 GE_default521;
extern T522 GE_default522;
extern T524 GE_default524;
extern T526 GE_default526;
extern T527 GE_default527;
extern T528 GE_default528;
extern T529 GE_default529;
extern T530 GE_default530;
extern T531 GE_default531;
extern T532 GE_default532;
extern T533 GE_default533;
extern T534 GE_default534;
extern T535 GE_default535;
extern T536 GE_default536;
extern T537 GE_default537;
extern T538 GE_default538;
extern T539 GE_default539;
extern T540 GE_default540;
extern T541 GE_default541;
extern T542 GE_default542;
extern T543 GE_default543;
extern T544 GE_default544;
extern T545 GE_default545;
extern T546 GE_default546;
extern T547 GE_default547;
extern T548 GE_default548;
extern T549 GE_default549;
extern T550 GE_default550;
extern T551 GE_default551;
extern T552 GE_default552;
extern T553 GE_default553;
extern T556 GE_default556;
extern T557 GE_default557;
extern T559 GE_default559;
extern T560 GE_default560;
extern T561 GE_default561;
extern T562 GE_default562;
extern T563 GE_default563;
extern T564 GE_default564;
extern T565 GE_default565;
extern T567 GE_default567;
extern T568 GE_default568;
extern T569 GE_default569;
extern T570 GE_default570;
extern T571 GE_default571;
extern T572 GE_default572;
extern T573 GE_default573;
extern T574 GE_default574;
extern T577 GE_default577;
extern T578 GE_default578;
extern T579 GE_default579;
extern T582 GE_default582;
extern T584 GE_default584;
extern T585 GE_default585;
extern T586 GE_default586;
extern T587 GE_default587;
extern T589 GE_default589;
extern T591 GE_default591;
extern T592 GE_default592;
extern T593 GE_default593;
extern T594 GE_default594;
extern T595 GE_default595;
extern T596 GE_default596;
extern T599 GE_default599;
extern T600 GE_default600;
extern T601 GE_default601;
extern T602 GE_default602;
extern T603 GE_default603;
extern T604 GE_default604;
extern T605 GE_default605;
extern T606 GE_default606;
extern T609 GE_default609;
extern T610 GE_default610;
extern T611 GE_default611;
extern T612 GE_default612;
extern T613 GE_default613;
extern T615 GE_default615;
extern T616 GE_default616;
extern T617 GE_default617;
extern T621 GE_default621;
extern T622 GE_default622;
extern T624 GE_default624;
extern T625 GE_default625;
extern T626 GE_default626;
extern T627 GE_default627;
extern T628 GE_default628;
extern T629 GE_default629;
extern T630 GE_default630;
extern T631 GE_default631;
extern T632 GE_default632;
extern T634 GE_default634;
extern T635 GE_default635;
extern T636 GE_default636;
extern T637 GE_default637;
extern T638 GE_default638;
extern T639 GE_default639;
extern T640 GE_default640;
extern T641 GE_default641;
extern T642 GE_default642;
extern T643 GE_default643;
extern T644 GE_default644;
extern T645 GE_default645;
extern T646 GE_default646;
extern T647 GE_default647;
extern T648 GE_default648;
extern T649 GE_default649;
extern T650 GE_default650;
extern T651 GE_default651;
extern T652 GE_default652;
extern T653 GE_default653;
extern T654 GE_default654;
extern T655 GE_default655;
extern T657 GE_default657;
extern T658 GE_default658;
extern T659 GE_default659;
extern T660 GE_default660;
extern T661 GE_default661;
extern T662 GE_default662;
extern T663 GE_default663;
extern T664 GE_default664;
extern T665 GE_default665;
extern T666 GE_default666;
extern T667 GE_default667;
extern T668 GE_default668;
extern T669 GE_default669;
extern T671 GE_default671;
extern T672 GE_default672;
extern T673 GE_default673;
extern T674 GE_default674;
extern T676 GE_default676;
extern T677 GE_default677;
extern T678 GE_default678;
extern T679 GE_default679;
extern T680 GE_default680;
extern T682 GE_default682;
extern T683 GE_default683;
extern T684 GE_default684;
extern T686 GE_default686;
extern T687 GE_default687;
extern T689 GE_default689;
extern T690 GE_default690;
extern T691 GE_default691;
extern T692 GE_default692;
extern T695 GE_default695;
extern T697 GE_default697;
extern T698 GE_default698;
extern T699 GE_default699;
extern T700 GE_default700;
extern T703 GE_default703;
extern T704 GE_default704;
extern T707 GE_default707;
extern T708 GE_default708;
extern T709 GE_default709;
extern T710 GE_default710;
extern T711 GE_default711;
extern T712 GE_default712;
extern T713 GE_default713;
extern T714 GE_default714;
extern T715 GE_default715;
extern T716 GE_default716;
extern T717 GE_default717;
extern T718 GE_default718;
extern T719 GE_default719;
extern T720 GE_default720;
extern T721 GE_default721;
extern T722 GE_default722;
extern T723 GE_default723;
extern T724 GE_default724;
extern T725 GE_default725;
extern T726 GE_default726;
extern T727 GE_default727;
extern T728 GE_default728;
extern T729 GE_default729;
extern T730 GE_default730;
extern T731 GE_default731;
extern T732 GE_default732;
extern T733 GE_default733;
extern T736 GE_default736;
extern T737 GE_default737;
extern T738 GE_default738;
extern T739 GE_default739;
extern T740 GE_default740;
extern T741 GE_default741;
extern T743 GE_default743;
extern T744 GE_default744;
extern T745 GE_default745;
extern T747 GE_default747;
extern T748 GE_default748;
extern T749 GE_default749;
extern T750 GE_default750;
extern T751 GE_default751;
extern T753 GE_default753;
extern T754 GE_default754;
extern T756 GE_default756;
extern T757 GE_default757;
extern T758 GE_default758;
extern T759 GE_default759;
extern T760 GE_default760;
extern T761 GE_default761;
extern T762 GE_default762;
extern T763 GE_default763;
extern T764 GE_default764;
extern T765 GE_default765;
extern T766 GE_default766;
extern T767 GE_default767;
extern T768 GE_default768;
extern T769 GE_default769;
extern T770 GE_default770;
extern T771 GE_default771;
extern T772 GE_default772;
extern T775 GE_default775;
extern T776 GE_default776;
extern T777 GE_default777;
extern T780 GE_default780;
extern T781 GE_default781;
extern T782 GE_default782;
extern T783 GE_default783;
extern T784 GE_default784;
extern T785 GE_default785;
extern T786 GE_default786;
extern T787 GE_default787;
extern T788 GE_default788;
extern T789 GE_default789;
extern T790 GE_default790;
extern T791 GE_default791;
extern T792 GE_default792;
extern T793 GE_default793;
extern T794 GE_default794;
extern T795 GE_default795;
extern T796 GE_default796;
extern T797 GE_default797;
extern T798 GE_default798;
extern T799 GE_default799;
extern T801 GE_default801;
extern T802 GE_default802;
extern T803 GE_default803;
extern T804 GE_default804;
extern T805 GE_default805;
extern T806 GE_default806;
extern T807 GE_default807;
extern T808 GE_default808;
extern T809 GE_default809;
extern T810 GE_default810;
extern T811 GE_default811;
extern T812 GE_default812;
extern T813 GE_default813;
extern T814 GE_default814;
extern T815 GE_default815;
extern T816 GE_default816;
extern T817 GE_default817;
extern T818 GE_default818;
extern T819 GE_default819;
extern T820 GE_default820;
extern T822 GE_default822;
extern T823 GE_default823;
extern T824 GE_default824;
extern T825 GE_default825;
extern T826 GE_default826;
extern T827 GE_default827;
extern T828 GE_default828;
extern T830 GE_default830;
extern T831 GE_default831;
extern T833 GE_default833;
extern T834 GE_default834;
extern T835 GE_default835;
extern T836 GE_default836;
extern T837 GE_default837;
extern T838 GE_default838;
extern T839 GE_default839;
extern T840 GE_default840;
extern T842 GE_default842;
extern T843 GE_default843;
extern T844 GE_default844;
extern T845 GE_default845;
extern T846 GE_default846;
extern T847 GE_default847;
extern T848 GE_default848;
extern T849 GE_default849;
extern T850 GE_default850;
extern T851 GE_default851;
extern T852 GE_default852;
extern T853 GE_default853;
extern T854 GE_default854;
extern T855 GE_default855;
extern T857 GE_default857;
extern T858 GE_default858;
extern T859 GE_default859;
extern T860 GE_default860;
extern T861 GE_default861;
extern T862 GE_default862;
extern T863 GE_default863;
extern T864 GE_default864;
extern T865 GE_default865;
extern T866 GE_default866;
extern T867 GE_default867;
extern T868 GE_default868;
extern T869 GE_default869;
extern T870 GE_default870;
extern T872 GE_default872;
extern T873 GE_default873;
extern T874 GE_default874;
extern T875 GE_default875;
extern T876 GE_default876;
extern T877 GE_default877;
extern T880 GE_default880;
extern T884 GE_default884;
extern T885 GE_default885;
extern T886 GE_default886;
extern T887 GE_default887;
extern T888 GE_default888;
extern T890 GE_default890;
extern T891 GE_default891;
extern T892 GE_default892;
extern T893 GE_default893;
extern T894 GE_default894;
extern T896 GE_default896;
extern T898 GE_default898;
extern T899 GE_default899;
extern T900 GE_default900;
extern T901 GE_default901;
extern T902 GE_default902;
extern T903 GE_default903;
extern T904 GE_default904;
extern T905 GE_default905;
extern T906 GE_default906;
extern T907 GE_default907;
extern T909 GE_default909;
extern T910 GE_default910;
extern T911 GE_default911;
extern T912 GE_default912;
extern T913 GE_default913;
extern T914 GE_default914;
extern T921 GE_default921;
extern T923 GE_default923;
extern T924 GE_default924;
extern T925 GE_default925;
extern T926 GE_default926;
extern T927 GE_default927;
extern T928 GE_default928;
extern T929 GE_default929;
extern T930 GE_default930;
extern T931 GE_default931;
extern T932 GE_default932;
extern T936 GE_default936;
extern T937 GE_default937;
extern T938 GE_default938;
extern T939 GE_default939;
extern T940 GE_default940;
extern T941 GE_default941;
extern T942 GE_default942;
extern T943 GE_default943;
extern T944 GE_default944;
extern T945 GE_default945;
extern T946 GE_default946;
extern T947 GE_default947;
extern T948 GE_default948;
extern T949 GE_default949;
extern T950 GE_default950;
extern T951 GE_default951;
extern T952 GE_default952;
extern T953 GE_default953;
extern T954 GE_default954;
extern T955 GE_default955;
extern T956 GE_default956;
extern T957 GE_default957;
extern T958 GE_default958;
extern T959 GE_default959;
extern T960 GE_default960;
extern T961 GE_default961;
extern T963 GE_default963;
extern T964 GE_default964;
extern T966 GE_default966;
extern T967 GE_default967;
extern T969 GE_default969;
extern T970 GE_default970;
extern T971 GE_default971;
extern T972 GE_default972;
extern T973 GE_default973;
extern T974 GE_default974;
extern T975 GE_default975;
extern T976 GE_default976;
extern T977 GE_default977;
extern T979 GE_default979;
extern T980 GE_default980;
extern T981 GE_default981;
extern T982 GE_default982;
extern T983 GE_default983;
extern T984 GE_default984;
extern T985 GE_default985;
extern T987 GE_default987;
extern T988 GE_default988;
extern T989 GE_default989;
extern T990 GE_default990;
extern T991 GE_default991;
extern T992 GE_default992;
extern T993 GE_default993;
extern T994 GE_default994;
extern T995 GE_default995;
extern T996 GE_default996;
extern T997 GE_default997;
extern T998 GE_default998;
extern T999 GE_default999;
extern T1000 GE_default1000;
extern T1001 GE_default1001;
extern T1002 GE_default1002;
extern T1003 GE_default1003;
extern T1004 GE_default1004;
extern T1005 GE_default1005;
extern T1006 GE_default1006;
extern T1007 GE_default1007;
extern T1008 GE_default1008;
extern T1009 GE_default1009;
extern T1010 GE_default1010;
extern T1011 GE_default1011;
extern T1012 GE_default1012;
extern T1013 GE_default1013;
extern T1014 GE_default1014;
extern T1015 GE_default1015;
extern T1016 GE_default1016;
extern T1017 GE_default1017;
extern T1018 GE_default1018;
extern T1019 GE_default1019;
extern T1020 GE_default1020;
extern T1021 GE_default1021;
extern T1023 GE_default1023;
extern T1024 GE_default1024;
extern T1025 GE_default1025;
extern T1026 GE_default1026;
extern T1027 GE_default1027;
extern T1028 GE_default1028;
extern T1029 GE_default1029;
extern T1030 GE_default1030;
extern T1031 GE_default1031;
extern T1032 GE_default1032;
extern T1033 GE_default1033;
extern T1034 GE_default1034;
extern T1036 GE_default1036;
extern T1037 GE_default1037;
extern T1038 GE_default1038;
extern T1039 GE_default1039;
extern T1040 GE_default1040;
extern T1041 GE_default1041;
extern T1042 GE_default1042;
extern T1043 GE_default1043;
extern T1044 GE_default1044;
extern T1045 GE_default1045;
extern T1046 GE_default1046;
extern T1047 GE_default1047;
extern T1048 GE_default1048;
extern T1050 GE_default1050;
extern T1051 GE_default1051;
extern T1053 GE_default1053;
extern T1054 GE_default1054;
extern T1056 GE_default1056;
extern T1057 GE_default1057;
extern T1058 GE_default1058;
extern T1059 GE_default1059;
extern T1060 GE_default1060;
extern T1061 GE_default1061;
extern T1062 GE_default1062;
extern T1063 GE_default1063;
extern T1064 GE_default1064;
extern T1065 GE_default1065;
extern T1066 GE_default1066;
extern T1067 GE_default1067;
extern T1068 GE_default1068;
extern T1069 GE_default1069;
extern T1071 GE_default1071;
extern T1072 GE_default1072;
extern T1073 GE_default1073;
extern T1074 GE_default1074;
extern T1076 GE_default1076;
extern T1077 GE_default1077;
extern T1078 GE_default1078;
extern T1079 GE_default1079;
extern T1080 GE_default1080;
extern T1081 GE_default1081;
extern T1082 GE_default1082;
extern T1083 GE_default1083;
extern T1084 GE_default1084;
extern T1085 GE_default1085;
extern T1086 GE_default1086;
extern T1088 GE_default1088;
extern T1089 GE_default1089;
extern T1091 GE_default1091;
extern T1092 GE_default1092;
extern T1094 GE_default1094;
extern T1095 GE_default1095;
extern T1096 GE_default1096;
extern T1097 GE_default1097;
extern T1098 GE_default1098;
extern T1099 GE_default1099;
extern T1100 GE_default1100;
extern T1101 GE_default1101;
extern T1102 GE_default1102;
extern T1103 GE_default1103;
extern T1104 GE_default1104;
extern T1105 GE_default1105;
extern T1106 GE_default1106;
extern T1107 GE_default1107;
extern T1108 GE_default1108;
extern T1109 GE_default1109;
extern T1110 GE_default1110;
extern T1111 GE_default1111;
extern T1112 GE_default1112;
extern T1113 GE_default1113;
extern T1114 GE_default1114;
extern T1115 GE_default1115;
extern T1116 GE_default1116;
extern T1117 GE_default1117;
extern T1118 GE_default1118;
extern T1119 GE_default1119;
extern T1120 GE_default1120;
extern T1121 GE_default1121;
extern T1122 GE_default1122;
extern T1123 GE_default1123;
extern T1124 GE_default1124;
extern T1126 GE_default1126;
extern T1127 GE_default1127;
extern T1128 GE_default1128;
extern T1129 GE_default1129;
extern T1130 GE_default1130;
extern T1131 GE_default1131;
extern T1132 GE_default1132;
extern T1133 GE_default1133;
extern T1134 GE_default1134;
extern T1135 GE_default1135;
extern T1137 GE_default1137;
extern T1138 GE_default1138;
extern T1139 GE_default1139;
extern T1140 GE_default1140;
extern T1141 GE_default1141;
extern T1142 GE_default1142;
extern T1143 GE_default1143;
extern T1144 GE_default1144;
extern T1145 GE_default1145;
extern T1146 GE_default1146;
extern T1147 GE_default1147;
extern T1148 GE_default1148;
extern T1149 GE_default1149;
extern T1151 GE_default1151;
extern T1152 GE_default1152;
extern T1153 GE_default1153;
extern T1154 GE_default1154;
extern T1156 GE_default1156;
extern T1157 GE_default1157;
extern T1158 GE_default1158;
extern T1159 GE_default1159;
extern T1160 GE_default1160;
extern T1161 GE_default1161;
extern T1163 GE_default1163;
extern T1164 GE_default1164;
extern T1165 GE_default1165;
extern T1166 GE_default1166;
extern T1167 GE_default1167;
extern T1168 GE_default1168;
extern T1169 GE_default1169;
extern T1170 GE_default1170;
extern T1171 GE_default1171;
extern T1172 GE_default1172;
extern T1173 GE_default1173;
extern T1174 GE_default1174;
extern T1175 GE_default1175;
extern T1177 GE_default1177;
extern T1178 GE_default1178;
extern T1179 GE_default1179;
extern T1180 GE_default1180;
extern T1182 GE_default1182;
extern T1183 GE_default1183;
extern T1184 GE_default1184;
extern T1185 GE_default1185;
extern T1186 GE_default1186;
extern T1187 GE_default1187;
extern T1188 GE_default1188;
extern T1189 GE_default1189;
extern T1190 GE_default1190;
extern T1191 GE_default1191;
extern T1192 GE_default1192;
extern T1193 GE_default1193;
extern T1194 GE_default1194;
extern T1196 GE_default1196;
extern T1197 GE_default1197;
extern T1198 GE_default1198;
extern T1199 GE_default1199;
extern T1200 GE_default1200;
extern T1201 GE_default1201;
extern T1202 GE_default1202;
extern T1204 GE_default1204;
extern T1205 GE_default1205;
extern T1206 GE_default1206;
extern T1207 GE_default1207;
extern T1208 GE_default1208;
extern T1210 GE_default1210;
extern T1211 GE_default1211;
extern T1212 GE_default1212;
extern T1213 GE_default1213;
extern T1214 GE_default1214;
extern T1215 GE_default1215;
extern T1216 GE_default1216;
extern T1217 GE_default1217;
extern T1218 GE_default1218;
extern T1219 GE_default1219;
extern T1220 GE_default1220;
extern T1222 GE_default1222;
extern T1223 GE_default1223;
extern T1224 GE_default1224;
extern T1225 GE_default1225;
extern T1226 GE_default1226;
extern T1227 GE_default1227;
extern T1228 GE_default1228;
extern T1230 GE_default1230;
extern T1231 GE_default1231;
extern T1232 GE_default1232;
extern T1233 GE_default1233;
extern T1234 GE_default1234;
extern T1235 GE_default1235;
extern T1236 GE_default1236;
extern T1237 GE_default1237;
extern T1238 GE_default1238;
extern T1239 GE_default1239;
extern T1240 GE_default1240;
extern T1241 GE_default1241;
extern T1242 GE_default1242;
extern T1243 GE_default1243;
extern T1244 GE_default1244;
extern T1245 GE_default1245;
extern T1246 GE_default1246;
extern T1247 GE_default1247;
extern T1248 GE_default1248;
extern T1249 GE_default1249;
extern T1250 GE_default1250;
extern T1252 GE_default1252;
extern T1253 GE_default1253;
extern T1254 GE_default1254;
extern T1255 GE_default1255;
extern T1256 GE_default1256;
extern T1257 GE_default1257;
extern T1258 GE_default1258;
extern T1260 GE_default1260;
extern T1261 GE_default1261;
extern T1262 GE_default1262;
extern T1263 GE_default1263;
extern T1264 GE_default1264;
extern T1265 GE_default1265;
extern T1266 GE_default1266;
extern T1267 GE_default1267;
extern T1268 GE_default1268;
extern T1269 GE_default1269;
extern T1270 GE_default1270;
extern T1271 GE_default1271;
extern T1272 GE_default1272;
extern T1273 GE_default1273;
extern T1274 GE_default1274;
extern T1276 GE_default1276;
extern T1277 GE_default1277;
extern T1279 GE_default1279;
extern T1280 GE_default1280;
extern T1281 GE_default1281;
extern T1282 GE_default1282;
extern T1283 GE_default1283;
extern T1285 GE_default1285;
extern T1286 GE_default1286;
extern T1288 GE_default1288;
extern T1289 GE_default1289;
extern T1290 GE_default1290;
extern T1291 GE_default1291;
extern T1292 GE_default1292;
extern T1293 GE_default1293;
extern T1294 GE_default1294;
extern T1295 GE_default1295;
extern T1296 GE_default1296;
extern T1297 GE_default1297;
extern T1298 GE_default1298;
extern T1299 GE_default1299;
extern T1300 GE_default1300;
extern T1301 GE_default1301;
extern T1302 GE_default1302;
extern T1303 GE_default1303;
extern T1304 GE_default1304;
extern T1305 GE_default1305;
extern T1306 GE_default1306;
extern T1307 GE_default1307;
extern T1308 GE_default1308;
extern T1309 GE_default1309;
extern T1310 GE_default1310;
extern T1311 GE_default1311;
extern T1313 GE_default1313;
extern T1314 GE_default1314;
extern T1316 GE_default1316;
extern T1317 GE_default1317;
extern T1318 GE_default1318;
extern T1319 GE_default1319;
extern T1320 GE_default1320;
extern T1321 GE_default1321;
extern T1322 GE_default1322;
extern T1323 GE_default1323;
extern T1324 GE_default1324;
extern T1325 GE_default1325;
extern T1326 GE_default1326;
extern T1327 GE_default1327;
extern T1328 GE_default1328;
extern T1329 GE_default1329;
extern T1330 GE_default1330;
extern T1331 GE_default1331;
extern T1332 GE_default1332;
extern T1333 GE_default1333;
extern T1337 GE_default1337;
extern T1338 GE_default1338;
extern T1339 GE_default1339;
extern T1340 GE_default1340;
extern T1341 GE_default1341;
extern T1344 GE_default1344;
extern T1345 GE_default1345;
extern T1346 GE_default1346;
extern T1347 GE_default1347;
extern T1348 GE_default1348;
extern T1349 GE_default1349;
extern T1350 GE_default1350;
extern T1351 GE_default1351;
extern T1353 GE_default1353;
extern T1354 GE_default1354;
extern T1355 GE_default1355;
extern T1356 GE_default1356;
extern T1358 GE_default1358;
extern T1359 GE_default1359;
extern T1360 GE_default1360;
extern T1361 GE_default1361;
extern T1364 GE_default1364;
extern T1365 GE_default1365;
extern T1368 GE_default1368;
extern T1369 GE_default1369;
extern T1370 GE_default1370;
extern T1371 GE_default1371;
extern T1372 GE_default1372;
extern T1373 GE_default1373;
extern T1374 GE_default1374;
extern T1375 GE_default1375;
extern T1378 GE_default1378;
extern T1379 GE_default1379;
extern T1380 GE_default1380;
extern T1381 GE_default1381;
extern T1382 GE_default1382;
extern T1383 GE_default1383;
extern T1384 GE_default1384;
extern T1386 GE_default1386;
extern T1387 GE_default1387;
extern T1388 GE_default1388;
extern T1389 GE_default1389;
extern T1390 GE_default1390;
extern T1391 GE_default1391;
extern T1392 GE_default1392;
extern T1393 GE_default1393;
extern T1395 GE_default1395;
extern T1396 GE_default1396;
extern T1397 GE_default1397;
extern T1398 GE_default1398;
extern T1400 GE_default1400;
extern T1402 GE_default1402;
extern T1404 GE_default1404;
extern T1406 GE_default1406;
extern T1407 GE_default1407;
extern T1408 GE_default1408;
extern T1410 GE_default1410;
extern T1411 GE_default1411;
extern T1412 GE_default1412;
extern T1413 GE_default1413;
extern T1414 GE_default1414;
extern T1415 GE_default1415;
extern T1416 GE_default1416;
extern T1418 GE_default1418;
extern T1419 GE_default1419;
extern T1420 GE_default1420;
extern T1421 GE_default1421;
extern T1423 GE_default1423;
extern T1424 GE_default1424;
extern T1425 GE_default1425;
extern T1426 GE_default1426;
extern T1427 GE_default1427;
extern T1428 GE_default1428;
extern T1429 GE_default1429;
extern T1433 GE_default1433;
extern T1434 GE_default1434;
extern T1435 GE_default1435;
extern T1436 GE_default1436;
extern T1443 GE_default1443;
extern T1446 GE_default1446;
extern T1448 GE_default1448;
extern T1449 GE_default1449;
extern T1450 GE_default1450;
extern T1451 GE_default1451;
extern T1452 GE_default1452;
extern T1453 GE_default1453;
extern T1455 GE_default1455;
extern T1456 GE_default1456;
extern T1457 GE_default1457;
extern T1458 GE_default1458;
extern T1459 GE_default1459;
extern T1460 GE_default1460;
extern T1461 GE_default1461;
extern T1463 GE_default1463;
extern T1464 GE_default1464;
extern T1465 GE_default1465;
extern T1466 GE_default1466;
extern T1468 GE_default1468;
extern T1469 GE_default1469;
extern T1470 GE_default1470;
extern T1471 GE_default1471;
extern T1474 GE_default1474;
extern T1477 GE_default1477;
extern T1479 GE_default1479;
extern T1481 GE_default1481;
extern T1482 GE_default1482;
extern T1483 GE_default1483;
extern T1484 GE_default1484;
extern T1485 GE_default1485;
extern T1486 GE_default1486;
extern T1487 GE_default1487;
extern T1488 GE_default1488;
extern T1489 GE_default1489;
extern T1490 GE_default1490;
extern T1491 GE_default1491;
extern T1493 GE_default1493;
extern T1494 GE_default1494;
extern T1495 GE_default1495;
extern T1496 GE_default1496;
extern T1498 GE_default1498;
extern T1499 GE_default1499;
extern T1500 GE_default1500;
extern T1501 GE_default1501;
extern T1502 GE_default1502;
extern T1503 GE_default1503;
extern T1504 GE_default1504;
extern T1505 GE_default1505;
extern T1506 GE_default1506;
extern T1508 GE_default1508;
extern T1509 GE_default1509;
extern T1511 GE_default1511;
extern T1512 GE_default1512;
extern T1513 GE_default1513;
extern T1515 GE_default1515;
extern T1516 GE_default1516;
extern T1517 GE_default1517;
extern T1518 GE_default1518;
extern T1521 GE_default1521;
extern T1522 GE_default1522;
extern T1524 GE_default1524;
extern T1525 GE_default1525;
extern T1526 GE_default1526;
extern T1527 GE_default1527;
extern T1528 GE_default1528;
extern T1529 GE_default1529;
extern T1530 GE_default1530;
extern T1532 GE_default1532;
extern T1533 GE_default1533;
extern T1534 GE_default1534;
extern T1535 GE_default1535;
extern T1537 GE_default1537;
extern T1538 GE_default1538;
extern T1539 GE_default1539;
extern T1540 GE_default1540;
extern T1541 GE_default1541;
extern T1542 GE_default1542;
extern T1543 GE_default1543;
extern T1544 GE_default1544;
extern T1545 GE_default1545;
extern T1546 GE_default1546;
extern T1547 GE_default1547;
extern T1548 GE_default1548;
extern T1549 GE_default1549;
extern T1550 GE_default1550;
extern T1551 GE_default1551;
extern T1552 GE_default1552;
extern T1553 GE_default1553;
extern T1554 GE_default1554;
extern T1555 GE_default1555;
extern T1556 GE_default1556;
extern T1557 GE_default1557;
extern T1558 GE_default1558;
extern T1559 GE_default1559;
extern T1562 GE_default1562;
extern T1564 GE_default1564;
extern T1566 GE_default1566;
extern T1567 GE_default1567;
extern T1569 GE_default1569;
extern T1571 GE_default1571;
extern T1573 GE_default1573;
extern T1576 GE_default1576;
extern T1577 GE_default1577;
extern T1578 GE_default1578;
extern T1579 GE_default1579;
extern T1580 GE_default1580;
extern T1581 GE_default1581;
extern T1582 GE_default1582;
extern T1583 GE_default1583;
extern T1584 GE_default1584;
extern T1585 GE_default1585;
extern T1586 GE_default1586;
extern T1587 GE_default1587;
extern T1588 GE_default1588;
extern T1589 GE_default1589;
extern T1590 GE_default1590;
extern T1591 GE_default1591;
extern T1592 GE_default1592;
extern T1593 GE_default1593;
extern T1594 GE_default1594;
extern T1597 GE_default1597;
extern T1598 GE_default1598;
extern T1599 GE_default1599;
extern T1600 GE_default1600;
extern T1601 GE_default1601;
extern T1602 GE_default1602;
extern T1603 GE_default1603;
extern T1604 GE_default1604;
extern T1605 GE_default1605;
extern T1606 GE_default1606;
extern T1608 GE_default1608;
extern T1609 GE_default1609;
extern T1611 GE_default1611;
extern T1612 GE_default1612;
extern T1614 GE_default1614;
extern T1615 GE_default1615;
extern T1616 GE_default1616;
extern T1617 GE_default1617;
extern T1618 GE_default1618;
extern T1619 GE_default1619;
extern T1620 GE_default1620;
extern T1621 GE_default1621;
extern T1622 GE_default1622;
extern T1623 GE_default1623;
extern T1624 GE_default1624;
extern T1626 GE_default1626;
extern T1627 GE_default1627;
extern T1628 GE_default1628;
extern T1629 GE_default1629;
extern T1630 GE_default1630;
extern T1631 GE_default1631;
extern T1632 GE_default1632;
extern T1633 GE_default1633;
extern T1634 GE_default1634;
extern T1635 GE_default1635;
extern T1636 GE_default1636;
extern T1637 GE_default1637;
extern T1638 GE_default1638;
extern T1639 GE_default1639;
extern T1640 GE_default1640;
extern T1641 GE_default1641;
extern T1642 GE_default1642;
extern T1643 GE_default1643;
extern T1644 GE_default1644;
extern T1645 GE_default1645;
extern T1646 GE_default1646;
extern T1647 GE_default1647;
extern T1648 GE_default1648;
extern T1649 GE_default1649;
extern T1650 GE_default1650;
extern T1651 GE_default1651;
extern T1653 GE_default1653;
extern T1655 GE_default1655;
extern T1656 GE_default1656;
extern T1659 GE_default1659;
extern T1660 GE_default1660;
extern T1661 GE_default1661;
extern T1662 GE_default1662;
extern T1663 GE_default1663;
extern T1666 GE_default1666;
extern T1667 GE_default1667;
extern T1668 GE_default1668;
extern T1669 GE_default1669;
extern T1671 GE_default1671;
extern T1672 GE_default1672;
extern T1673 GE_default1673;
extern T1674 GE_default1674;
extern T1675 GE_default1675;
extern T1677 GE_default1677;
extern T1678 GE_default1678;
extern T1679 GE_default1679;
extern T1680 GE_default1680;
extern T1681 GE_default1681;
extern T1682 GE_default1682;
extern T1684 GE_default1684;
extern T1685 GE_default1685;
extern T1686 GE_default1686;
extern T1687 GE_default1687;
extern T1688 GE_default1688;
extern T1690 GE_default1690;
extern T1692 GE_default1692;
extern T1693 GE_default1693;
extern T1694 GE_default1694;
extern T1695 GE_default1695;
extern T1696 GE_default1696;
extern T1697 GE_default1697;
extern T1698 GE_default1698;
extern T1701 GE_default1701;
extern T1703 GE_default1703;
extern T1704 GE_default1704;
extern T1707 GE_default1707;
extern T1711 GE_default1711;
extern T1712 GE_default1712;
extern T1713 GE_default1713;
extern T1714 GE_default1714;
extern T1715 GE_default1715;
extern T1716 GE_default1716;
extern T1717 GE_default1717;
extern T1719 GE_default1719;
extern T1720 GE_default1720;
extern T1722 GE_default1722;
extern T1723 GE_default1723;
extern T1724 GE_default1724;
extern T1725 GE_default1725;
extern T1726 GE_default1726;
extern T1727 GE_default1727;
extern T1728 GE_default1728;
extern T1729 GE_default1729;
extern T1730 GE_default1730;
extern T1731 GE_default1731;
extern T1732 GE_default1732;
extern T1733 GE_default1733;
extern T1734 GE_default1734;
extern T1736 GE_default1736;
extern T1737 GE_default1737;
extern T1738 GE_default1738;
extern T1739 GE_default1739;
extern T1742 GE_default1742;
extern T1743 GE_default1743;
extern T1744 GE_default1744;
extern T1745 GE_default1745;
extern T1750 GE_default1750;
extern T1755 GE_default1755;
extern T1756 GE_default1756;
extern T1758 GE_default1758;
extern T1759 GE_default1759;
extern T1762 GE_default1762;
extern T1766 GE_default1766;
extern T1767 GE_default1767;
extern T1768 GE_default1768;
extern T1769 GE_default1769;
extern T1771 GE_default1771;
extern T1776 GE_default1776;
extern T1777 GE_default1777;
extern T1779 GE_default1779;
extern T1780 GE_default1780;
extern T1782 GE_default1782;
extern T1783 GE_default1783;
extern T1784 GE_default1784;
extern T1785 GE_default1785;
extern T1788 GE_default1788;
extern T1789 GE_default1789;
extern T1790 GE_default1790;
extern T1791 GE_default1791;
extern T1792 GE_default1792;
extern T1793 GE_default1793;
extern T1794 GE_default1794;
extern T1795 GE_default1795;
extern T1796 GE_default1796;
extern T1797 GE_default1797;
extern T1798 GE_default1798;
extern T1799 GE_default1799;
extern T1800 GE_default1800;
extern T1801 GE_default1801;
extern T1802 GE_default1802;
extern T1806 GE_default1806;
extern T1809 GE_default1809;
extern T1810 GE_default1810;
extern T1811 GE_default1811;
extern T1813 GE_default1813;
extern T1814 GE_default1814;
extern T1815 GE_default1815;
extern T1817 GE_default1817;
extern T1819 GE_default1819;
extern T1820 GE_default1820;
extern T1821 GE_default1821;
extern T1822 GE_default1822;
extern T1823 GE_default1823;
extern T1824 GE_default1824;
extern T1825 GE_default1825;
extern T1826 GE_default1826;
extern T1827 GE_default1827;
extern T1828 GE_default1828;
extern T1830 GE_default1830;
extern T1831 GE_default1831;
extern T1832 GE_default1832;
extern T1833 GE_default1833;
extern T1834 GE_default1834;
extern T1836 GE_default1836;
extern T1838 GE_default1838;
extern T1839 GE_default1839;
extern T1840 GE_default1840;
extern T1841 GE_default1841;
extern T1843 GE_default1843;
extern T1844 GE_default1844;
extern T1845 GE_default1845;
extern T1846 GE_default1846;
extern T1847 GE_default1847;
extern T1848 GE_default1848;
extern T1850 GE_default1850;
extern T1853 GE_default1853;
extern T1857 GE_default1857;
extern T1859 GE_default1859;
extern T1860 GE_default1860;
extern T1861 GE_default1861;
extern T1862 GE_default1862;
extern T1863 GE_default1863;
extern T1864 GE_default1864;
extern T1865 GE_default1865;
extern T1866 GE_default1866;
extern T1867 GE_default1867;
extern T1868 GE_default1868;
extern T1869 GE_default1869;
extern T1870 GE_default1870;
extern T1871 GE_default1871;
extern T1872 GE_default1872;
extern T1873 GE_default1873;
extern T1874 GE_default1874;
extern T1875 GE_default1875;
extern T1876 GE_default1876;
extern T1877 GE_default1877;
extern T1878 GE_default1878;
extern T1879 GE_default1879;
extern T1880 GE_default1880;
extern T1881 GE_default1881;
extern T1882 GE_default1882;
extern T1883 GE_default1883;
extern T1884 GE_default1884;
extern T1885 GE_default1885;
extern T1886 GE_default1886;
extern T1887 GE_default1887;
extern T1888 GE_default1888;
extern T1889 GE_default1889;
extern T1890 GE_default1890;
extern T1891 GE_default1891;
extern T1892 GE_default1892;
extern T1893 GE_default1893;
extern T1894 GE_default1894;
extern T1895 GE_default1895;
extern T1896 GE_default1896;
extern T1897 GE_default1897;
extern T1898 GE_default1898;
extern T1899 GE_default1899;
extern T1900 GE_default1900;
extern T1902 GE_default1902;
extern T1903 GE_default1903;
extern T1904 GE_default1904;
extern T1905 GE_default1905;
extern T1906 GE_default1906;
extern T1907 GE_default1907;
extern T1908 GE_default1908;
extern T1909 GE_default1909;
extern T1911 GE_default1911;
extern T1912 GE_default1912;
extern T1914 GE_default1914;
extern T1915 GE_default1915;
extern T1916 GE_default1916;
extern T1917 GE_default1917;
extern T1918 GE_default1918;
extern T1919 GE_default1919;
extern T1920 GE_default1920;
extern T1921 GE_default1921;
extern T1922 GE_default1922;
extern T1923 GE_default1923;
extern T1926 GE_default1926;
extern T1927 GE_default1927;
extern T1928 GE_default1928;
extern T1929 GE_default1929;
extern T1930 GE_default1930;
extern T1931 GE_default1931;
extern T1932 GE_default1932;
extern T1933 GE_default1933;
extern T1935 GE_default1935;
extern T1936 GE_default1936;
extern T1937 GE_default1937;
extern T1938 GE_default1938;
extern T1939 GE_default1939;
extern T1940 GE_default1940;
extern T1941 GE_default1941;
extern T1942 GE_default1942;
extern T1943 GE_default1943;
extern T1944 GE_default1944;
extern T1953 GE_default1953;
extern T1954 GE_default1954;
extern T1955 GE_default1955;
extern T1956 GE_default1956;
extern T1957 GE_default1957;
extern T1960 GE_default1960;
extern T1962 GE_default1962;
extern T1964 GE_default1964;
extern T1965 GE_default1965;
extern T1966 GE_default1966;
extern T1968 GE_default1968;
extern T1977 GE_default1977;
extern T1978 GE_default1978;
extern T1980 GE_default1980;
extern T1984 GE_default1984;
extern T1985 GE_default1985;
extern T1988 GE_default1988;
extern T1989 GE_default1989;
extern T1990 GE_default1990;
extern T1993 GE_default1993;
extern T1994 GE_default1994;
extern T1997 GE_default1997;
extern T1998 GE_default1998;
extern T1999 GE_default1999;
extern T2000 GE_default2000;
extern T2002 GE_default2002;
extern T2020 GE_default2020;
extern T2026 GE_default2026;
extern T2027 GE_default2027;
extern T2034 GE_default2034;
extern T2040 GE_default2040;
extern T2041 GE_default2041;
extern T2042 GE_default2042;
extern T2043 GE_default2043;
extern T2044 GE_default2044;
extern T2046 GE_default2046;
extern T2047 GE_default2047;
extern T2048 GE_default2048;
extern T2051 GE_default2051;
extern T2052 GE_default2052;
extern T2054 GE_default2054;
extern T2056 GE_default2056;
extern T2057 GE_default2057;
extern T2058 GE_default2058;
extern T2059 GE_default2059;
extern T2060 GE_default2060;
extern T2061 GE_default2061;
extern T2063 GE_default2063;
extern T2064 GE_default2064;
extern T2066 GE_default2066;
extern T2068 GE_default2068;
extern T2071 GE_default2071;
extern T2073 GE_default2073;
extern T2074 GE_default2074;
extern T2075 GE_default2075;
extern T2077 GE_default2077;
extern T2078 GE_default2078;
extern T2079 GE_default2079;
extern T2082 GE_default2082;
extern T2086 GE_default2086;
extern T2089 GE_default2089;
extern T2090 GE_default2090;
extern T2091 GE_default2091;
extern T2092 GE_default2092;
extern T2094 GE_default2094;
extern T2095 GE_default2095;
extern T2096 GE_default2096;
extern T2097 GE_default2097;
extern T2099 GE_default2099;
extern T2100 GE_default2100;
extern T2101 GE_default2101;
extern T2102 GE_default2102;
extern T2105 GE_default2105;
extern T2106 GE_default2106;
extern T2107 GE_default2107;
extern T2108 GE_default2108;
extern T2109 GE_default2109;
extern T2110 GE_default2110;
extern T2111 GE_default2111;
extern T2112 GE_default2112;
extern T2113 GE_default2113;
extern T2114 GE_default2114;
extern T2115 GE_default2115;
extern T2116 GE_default2116;
extern T2117 GE_default2117;
extern T2118 GE_default2118;
extern T2119 GE_default2119;
extern T2120 GE_default2120;
extern T2121 GE_default2121;
extern T2122 GE_default2122;
extern T2123 GE_default2123;
extern T2124 GE_default2124;
extern T2125 GE_default2125;
extern T2126 GE_default2126;
extern T2127 GE_default2127;
extern T2128 GE_default2128;
extern T2129 GE_default2129;
extern T2130 GE_default2130;
extern T2131 GE_default2131;
extern T2132 GE_default2132;
extern T2133 GE_default2133;
extern T2134 GE_default2134;
extern T2135 GE_default2135;
extern T2136 GE_default2136;
extern T2137 GE_default2137;
extern T2138 GE_default2138;
extern T2139 GE_default2139;
extern T2140 GE_default2140;
extern T2141 GE_default2141;
extern T2142 GE_default2142;
extern T2143 GE_default2143;
extern T2144 GE_default2144;
extern T2145 GE_default2145;
extern T2146 GE_default2146;
extern T2147 GE_default2147;
extern T2148 GE_default2148;
extern T2149 GE_default2149;
extern T2150 GE_default2150;
extern T2151 GE_default2151;
extern T2152 GE_default2152;
extern T2153 GE_default2153;
extern T2154 GE_default2154;
extern T2155 GE_default2155;
extern T2156 GE_default2156;
extern T2157 GE_default2157;
extern T2158 GE_default2158;
extern T2159 GE_default2159;
extern T2160 GE_default2160;
extern T2161 GE_default2161;
extern T2162 GE_default2162;
extern T2164 GE_default2164;
extern T2165 GE_default2165;
extern T2166 GE_default2166;
extern T2168 GE_default2168;
extern T2170 GE_default2170;
extern T2171 GE_default2171;
extern T2174 GE_default2174;
extern T2175 GE_default2175;
extern T2177 GE_default2177;
extern T2178 GE_default2178;
extern T2179 GE_default2179;
extern T2180 GE_default2180;
extern T2181 GE_default2181;
extern T2183 GE_default2183;
extern T2185 GE_default2185;
extern T2187 GE_default2187;
extern T2188 GE_default2188;
extern T2190 GE_default2190;
extern T2191 GE_default2191;
extern T2192 GE_default2192;
extern T2193 GE_default2193;
extern T2194 GE_default2194;
extern T2195 GE_default2195;
extern T2197 GE_default2197;
extern T2201 GE_default2201;
extern T2204 GE_default2204;
extern T2207 GE_default2207;
extern T2208 GE_default2208;
extern T2209 GE_default2209;
extern T2210 GE_default2210;
extern T2211 GE_default2211;
extern T2213 GE_default2213;

extern T0* GE_new_str8(T6 c);
extern T0* GE_new_str32(T6 c);
extern T0* GE_new_istr8(T6 c);
extern T0* GE_new_istr32(T6 c);
/* New instance of type [detachable] SPECIAL [CHARACTER_8] */
extern T0* GE_new15(T6 a1, T1 initialize);
/* New instance of type [detachable] SPECIAL [CHARACTER_32] */
extern T0* GE_new16(T6 a1, T1 initialize);
/* New instance of type [detachable] STRING_8 */
extern T0* GE_new17(T1 initialize);
/* New instance of type [detachable] STRING_32 */
extern T0* GE_new18(T1 initialize);
/* New instance of type [detachable] IMMUTABLE_STRING_32 */
extern T0* GE_new20(T1 initialize);
/* New instance of type [detachable] ISE_EXCEPTION_MANAGER */
extern T0* GE_new21(T1 initialize);
/* New instance of type [detachable] GEC */
extern T0* GE_new25(T1 initialize);
/* New instance of type [detachable] CELL [detachable EXCEPTION] */
extern T0* GE_new26(T1 initialize);
/* New instance of type [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32] */
extern T0* GE_new27(T1 initialize);
/* New instance of type [detachable] CELL [detachable TUPLE [[attached] INTEGER_32, [attached] INTEGER_32, [attached] INTEGER_32, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] INTEGER_32, [attached] BOOLEAN]] */
extern T0* GE_new28(T1 initialize);
/* New instance of type [detachable] CELL [[attached] NO_MORE_MEMORY] */
extern T0* GE_new29(T1 initialize);
/* New instance of type detachable C_STRING */
extern T0* GE_new30(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] INTEGER_32, [attached] INTEGER_32, [attached] INTEGER_32, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] INTEGER_32, [attached] BOOLEAN] */
extern T0* GE_new31(T1 initialize);
/* New instance of type [detachable] KL_ARGUMENTS */
extern T0* GE_new32(T1 initialize);
/* New instance of type [detachable] ARRAY [[attached] STRING_8] */
extern T0* GE_new33(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] STRING_8] */
extern T0* GE_new34(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_EXCEPTIONS */
extern T0* GE_new35(T1 initialize);
/* New instance of type [detachable] ET_ERROR_HANDLER */
extern T0* GE_new36(T1 initialize);
/* New instance of type [detachable] KL_TEXT_INPUT_FILE */
extern T0* GE_new37(T1 initialize);
/* New instance of type [detachable] ET_ISE_VARIABLES */
extern T0* GE_new38(T1 initialize);
/* New instance of type [detachable] AP_FLAG */
extern T0* GE_new39(T1 initialize);
/* New instance of type detachable ET_SYSTEM */
extern T0* GE_new40(T1 initialize);
/* New instance of type [detachable] AP_PARSER */
extern T0* GE_new42(T1 initialize);
/* New instance of type [detachable] AP_ALTERNATIVE_OPTIONS_LIST */
extern T0* GE_new43(T1 initialize);
/* New instance of type [detachable] AP_STRING_OPTION */
extern T0* GE_new44(T1 initialize);
/* New instance of type [detachable] UT_VERSION */
extern T0* GE_new46(T1 initialize);
/* New instance of type [detachable] AP_ENUMERATION_OPTION */
extern T0* GE_new47(T1 initialize);
/* New instance of type [detachable] AP_BOOLEAN_OPTION */
extern T0* GE_new48(T1 initialize);
/* New instance of type [detachable] AP_INTEGER_OPTION */
extern T0* GE_new49(T1 initialize);
/* New instance of type [detachable] ET_NULL_ERROR_HANDLER */
extern T0* GE_new53(T1 initialize);
/* New instance of type [detachable] ET_ECF_SYSTEM_PARSER */
extern T0* GE_new55(T1 initialize);
/* New instance of type [detachable] ET_ECF_ERROR_HANDLER */
extern T0* GE_new56(T1 initialize);
/* New instance of type detachable ET_ECF_SETTINGS */
extern T0* GE_new57(T1 initialize);
/* New instance of type [detachable] ET_ECF_SYSTEM */
extern T0* GE_new58(T1 initialize);
/* New instance of type [detachable] ET_ECF_TARGET */
extern T0* GE_new59(T1 initialize);
/* New instance of type detachable ET_ECF_CAPABILITIES */
extern T0* GE_new61(T1 initialize);
/* New instance of type detachable ET_ECF_VARIABLES */
extern T0* GE_new62(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_SYSTEM */
extern T0* GE_new63(T1 initialize);
/* New instance of type [detachable] ET_SYSTEM_PROCESSOR */
extern T0* GE_new64(T1 initialize);
/* New instance of type [detachable] ET_C_GENERATOR */
extern T0* GE_new67(T1 initialize);
/* New instance of type detachable DT_DATE_TIME */
extern T0* GE_new68(T1 initialize);
/* New instance of type detachable DS_HASH_SET [[attached] STRING_8] */
extern T0* GE_new70(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER */
extern T0* GE_new71(T1 initialize);
/* New instance of type [detachable] ET_CLASS_TYPE */
extern T0* GE_new72(T1 initialize);
/* New instance of type [detachable] ET_TOKEN_CONSTANTS */
extern T0* GE_new73(T1 initialize);
/* New instance of type [detachable] ET_CLASS */
extern T0* GE_new74(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new76(T1 initialize);
/* New instance of type [detachable] UT_CANNOT_READ_FILE_ERROR */
extern T0* GE_new77(T1 initialize);
/* New instance of type [detachable] UT_VERSION_NUMBER */
extern T0* GE_new79(T1 initialize);
/* New instance of type [detachable] UT_MESSAGE */
extern T0* GE_new80(T1 initialize);
/* New instance of type [detachable] RX_PCRE_REGULAR_EXPRESSION */
extern T0* GE_new81(T1 initialize);
/* New instance of type [detachable] KL_STRING_ROUTINES */
extern T0* GE_new82(T1 initialize);
/* New instance of type [detachable] KL_EXECUTION_ENVIRONMENT */
extern T0* GE_new86(T1 initialize);
/* New instance of type [detachable] AP_ERROR */
extern T0* GE_new87(T1 initialize);
/* New instance of type [detachable] KL_SHELL_COMMAND */
extern T0* GE_new88(T1 initialize);
/* New instance of type [detachable] GECC */
extern T0* GE_new89(T1 initialize);
/* New instance of type [detachable] KL_OPERATING_SYSTEM */
extern T0* GE_new90(T1 initialize);
/* New instance of type [detachable] VOID_TARGET */
extern T0* GE_new93(T1 initialize);
/* New instance of type [detachable] ROUTINE_FAILURE */
extern T0* GE_new95(T1 initialize);
/* New instance of type [detachable] OLD_VIOLATION */
extern T0* GE_new97(T1 initialize);
/* New instance of type [detachable] NO_MORE_MEMORY */
extern T0* GE_new99(T1 initialize);
/* New instance of type [detachable] INVARIANT_VIOLATION */
extern T0* GE_new100(T1 initialize);
/* New instance of type [detachable] OPERATING_SYSTEM_SIGNAL_FAILURE */
extern T0* GE_new101(T1 initialize);
/* New instance of type [detachable] IO_FAILURE */
extern T0* GE_new102(T1 initialize);
/* New instance of type [detachable] OPERATING_SYSTEM_FAILURE */
extern T0* GE_new103(T1 initialize);
/* New instance of type [detachable] COM_FAILURE */
extern T0* GE_new104(T1 initialize);
/* New instance of type [detachable] EIFFEL_RUNTIME_PANIC */
extern T0* GE_new105(T1 initialize);
/* New instance of type [detachable] PRECONDITION_VIOLATION */
extern T0* GE_new107(T1 initialize);
/* New instance of type [detachable] POSTCONDITION_VIOLATION */
extern T0* GE_new108(T1 initialize);
/* New instance of type [detachable] FLOATING_POINT_FAILURE */
extern T0* GE_new109(T1 initialize);
/* New instance of type [detachable] CHECK_VIOLATION */
extern T0* GE_new110(T1 initialize);
/* New instance of type [detachable] BAD_INSPECT_VALUE */
extern T0* GE_new111(T1 initialize);
/* New instance of type [detachable] VARIANT_VIOLATION */
extern T0* GE_new112(T1 initialize);
/* New instance of type [detachable] LOOP_INVARIANT_VIOLATION */
extern T0* GE_new113(T1 initialize);
/* New instance of type [detachable] RESCUE_FAILURE */
extern T0* GE_new114(T1 initialize);
/* New instance of type [detachable] RESUMPTION_FAILURE */
extern T0* GE_new115(T1 initialize);
/* New instance of type [detachable] CREATE_ON_DEFERRED */
extern T0* GE_new116(T1 initialize);
/* New instance of type [detachable] EXTERNAL_FAILURE */
extern T0* GE_new117(T1 initialize);
/* New instance of type [detachable] VOID_ASSIGNED_TO_EXPANDED */
extern T0* GE_new118(T1 initialize);
/* New instance of type [detachable] EXCEPTION_IN_SIGNAL_HANDLER_FAILURE */
extern T0* GE_new119(T1 initialize);
/* New instance of type [detachable] MISMATCH_FAILURE */
extern T0* GE_new120(T1 initialize);
/* New instance of type [detachable] DEVELOPER_EXCEPTION */
extern T0* GE_new121(T1 initialize);
/* New instance of type [detachable] ADDRESS_APPLIED_TO_MELTED_FEATURE */
extern T0* GE_new122(T1 initialize);
/* New instance of type [detachable] SERIALIZATION_FAILURE */
extern T0* GE_new123(T1 initialize);
/* New instance of type [detachable] EXECUTION_ENVIRONMENT */
extern T0* GE_new124(T1 initialize);
/* New instance of type [detachable] KL_WINDOWS_FILE_SYSTEM */
extern T0* GE_new125(T1 initialize);
/* New instance of type [detachable] KL_UNIX_FILE_SYSTEM */
extern T0* GE_new126(T1 initialize);
/* New instance of type [detachable] PRIMES */
extern T0* GE_new127(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] INTEGER_32] */
extern T0* GE_new128(T6 a1, T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] BOOLEAN] */
extern T0* GE_new129(T6 a1, T1 initialize);
/* New instance of type [detachable] ARGUMENTS_32 */
extern T0* GE_new130(T1 initialize);
/* New instance of type [detachable] MUTEX */
extern T0* GE_new131(T1 initialize);
/* New instance of type [detachable] UT_ERROR_HANDLER */
extern T0* GE_new132(T1 initialize);
/* New instance of type [detachable] KL_STANDARD_FILES */
extern T0* GE_new133(T1 initialize);
/* New instance of type [detachable] KL_STDERR_FILE */
extern T0* GE_new134(T1 initialize);
/* New instance of type [detachable] KL_STDOUT_FILE */
extern T0* GE_new135(T1 initialize);
/* New instance of type detachable KL_LINKABLE [[attached] CHARACTER_8] */
extern T0* GE_new137(T1 initialize);
/* New instance of type [detachable] MANAGED_POINTER */
extern T0* GE_new140(T1 initialize);
/* New instance of type [detachable] FILE_INFO */
extern T0* GE_new141(T1 initialize);
/* New instance of type [detachable] ST_WORD_WRAPPER */
extern T0* GE_new142(T1 initialize);
/* New instance of type [detachable] AP_DISPLAY_HELP_FLAG */
extern T0* GE_new144(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION] */
extern T0* GE_new145(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
extern T0* GE_new146(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] STRING_8] */
extern T0* GE_new147(T1 initialize);
/* New instance of type [detachable] AP_ERROR_HANDLER */
extern T0* GE_new148(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST_CURSOR [[attached] AP_OPTION] */
extern T0* GE_new149(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [detachable STRING_8] */
extern T0* GE_new150(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8] */
extern T0* GE_new151(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST [[attached] STRING_8] */
extern T0* GE_new152(T1 initialize);
/* New instance of type [detachable] KL_STRING_EQUALITY_TESTER */
extern T0* GE_new153(T1 initialize);
/* New instance of type detachable KL_EQUALITY_TESTER [[attached] STRING_8] */
extern T0* GE_new154(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST [[attached] BOOLEAN] */
extern T0* GE_new157(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST [[attached] INTEGER_32] */
extern T0* GE_new160(T1 initialize);
/* New instance of type [detachable] KL_NULL_TEXT_OUTPUT_STREAM */
extern T0* GE_new161(T1 initialize);
/* New instance of type [detachable] ET_ECF_AST_FACTORY */
extern T0* GE_new162(T1 initialize);
/* New instance of type [detachable] DS_CELL [detachable ET_ECF_SYSTEM] */
extern T0* GE_new163(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_TARGET] */
extern T0* GE_new164(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_TARGET]] */
extern T0* GE_new165(T1 initialize);
/* New instance of type detachable TUPLE */
extern T0* GE_new166(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER] */
extern T0* GE_new167(T1 initialize);
/* New instance of type [detachable] XM_ELEMENT */
extern T0* GE_new168(T1 initialize);
/* New instance of type detachable XM_POSITION_TABLE */
extern T0* GE_new169(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] XM_ELEMENT, detachable XM_POSITION_TABLE, [attached] STRING_8] */
extern T0* GE_new170(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] XM_ELEMENT, detachable XM_POSITION_TABLE, [attached] STRING_8]] */
extern T0* GE_new171(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER, [detachable] STRING_8, [attached] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_TARGET]], [attached] DS_CELL [detachable ET_ECF_SYSTEM]] */
extern T0* GE_new172(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8] */
extern T0* GE_new174(T1 initialize);
/* New instance of type [detachable] KL_CASE_INSENSITIVE_STRING_EQUALITY_TESTER */
extern T0* GE_new175(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8] */
extern T0* GE_new176(T1 initialize);
/* New instance of type [detachable] XM_EIFFEL_PARSER */
extern T0* GE_new178(T1 initialize);
/* New instance of type [detachable] XM_TREE_CALLBACKS_PIPE */
extern T0* GE_new179(T1 initialize);
/* New instance of type [detachable] XM_CALLBACKS_TO_TREE_FILTER */
extern T0* GE_new182(T1 initialize);
/* New instance of type [detachable] ET_ECF_STATE */
extern T0* GE_new183(T1 initialize);
/* New instance of type [detachable] ET_IDENTIFIER */
extern T0* GE_new184(T1 initialize);
/* New instance of type [detachable] ET_ECF_SYSTEM_CONFIG */
extern T0* GE_new185(T1 initialize);
/* New instance of type [detachable] ET_ECF_LIBRARY */
extern T0* GE_new187(T1 initialize);
/* New instance of type [detachable] ET_ADAPTED_LIBRARIES */
extern T0* GE_new188(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ADAPTED_LIBRARY] */
extern T0* GE_new190(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ADAPTED_LIBRARY]] */
extern T0* GE_new191(T1 initialize);
/* New instance of type [detachable] XM_DOCUMENT */
extern T0* GE_new192(T1 initialize);
/* New instance of type [detachable] ET_COMPRESSED_POSITION */
extern T0* GE_new193(T1 initialize);
/* New instance of type [detachable] XM_STOP_ON_ERROR_FILTER */
extern T0* GE_new195(T1 initialize);
/* New instance of type detachable XM_ATTRIBUTE */
extern T0* GE_new196(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8] */
extern T0* GE_new198(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8] */
extern T0* GE_new199(T1 initialize);
/* New instance of type detachable ET_ECF_TARGET_PARENT */
extern T0* GE_new200(T1 initialize);
/* New instance of type [detachable] DS_CELL [detachable ET_ECF_SYSTEM_CONFIG] */
extern T0* GE_new201(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER, [attached] ET_ECF_INTERNAL_UNIVERSE, [attached] DS_CELL [detachable ET_ECF_SYSTEM_CONFIG]] */
extern T0* GE_new202(T1 initialize);
/* New instance of type detachable ET_ECF_TARGETS */
extern T0* GE_new203(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
extern T0* GE_new204(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
extern T0* GE_new205(T1 initialize);
/* New instance of type [detachable] DS_CELL [detachable ET_ECF_LIBRARY] */
extern T0* GE_new206(T1 initialize);
/* New instance of type [detachable] ET_ECF_ADAPTED_LIBRARY */
extern T0* GE_new207(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_SYSTEM_PARSER, [attached] ET_ECF_ADAPTED_LIBRARY, [attached] DS_CELL [detachable ET_ECF_LIBRARY]] */
extern T0* GE_new208(T1 initialize);
/* New instance of type [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES */
extern T0* GE_new210(T1 initialize);
/* New instance of type [detachable] ET_ECF_DOTNET_ASSEMBLY */
extern T0* GE_new211(T1 initialize);
/* New instance of type [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY */
extern T0* GE_new213(T1 initialize);
/* New instance of type [detachable] ET_ECF_ERROR */
extern T0* GE_new214(T1 initialize);
/* New instance of type [detachable] KL_AGENT_HASH_FUNCTION [[attached] STRING_8] */
extern T0* GE_new215(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8] */
extern T0* GE_new216(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] STRING_8] */
extern T0* GE_new217(T1 initialize);
/* New instance of type [detachable] FUNCTION [[attached] TUPLE [[attached] STRING_8], [attached] INTEGER_32] */
extern T0* GE_new218(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] KL_STRING_ROUTINES] */
extern T0* GE_new219(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8] */
extern T0* GE_new221(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_PRIMARY_TYPE */
extern T0* GE_new223(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_FEATURE_LIST */
extern T0* GE_new224(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME] */
extern T0* GE_new225(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8] */
extern T0* GE_new226(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_CLASS] */
extern T0* GE_new227(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]] */
extern T0* GE_new228(T1 initialize);
/* New instance of type [detachable] TUPLE [INTEGER_32] */
extern T0* GE_new229(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER */
extern T0* GE_new230(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_DYNAMIC_SYSTEM] */
extern T0* GE_new231(T1 initialize);
/* New instance of type detachable ET_DYNAMIC_FEATURE */
extern T0* GE_new236(T1 initialize);
/* New instance of type [detachable] ET_ACTUAL_PARAMETER_LIST */
extern T0* GE_new240(T1 initialize);
/* New instance of type [detachable] ET_NESTED_TYPE_CONTEXT */
extern T0* GE_new241(T1 initialize);
/* New instance of type [detachable] ET_FORMAL_PARAMETER_TYPE */
extern T0* GE_new242(T1 initialize);
/* New instance of type [detachable] ARRAY [[attached] ET_TYPE] */
extern T0* GE_new244(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_TYPE] */
extern T0* GE_new245(T6 a1, T1 initialize);
/* New instance of type [detachable] ET_EXTERNAL_FUNCTION */
extern T0* GE_new253(T1 initialize);
/* New instance of type [detachable] ET_TUPLE_TYPE */
extern T0* GE_new254(T1 initialize);
/* New instance of type detachable ET_FORMAL_ARGUMENT_LIST */
extern T0* GE_new255(T1 initialize);
/* New instance of type [detachable] ET_FORMAL_ARGUMENT */
extern T0* GE_new256(T1 initialize);
/* New instance of type [detachable] ET_QUERY_LIST */
extern T0* GE_new257(T1 initialize);
/* New instance of type [detachable] ET_PROCEDURE_LIST */
extern T0* GE_new258(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_TYPE_SET_LIST */
extern T0* GE_new259(T1 initialize);
/* New instance of type [detachable] KL_STRING_OUTPUT_STREAM */
extern T0* GE_new260(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER] */
extern T0* GE_new261(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new262(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32] */
extern T0* GE_new263(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_UNQUALIFIED_CALL_EXPRESSION] */
extern T0* GE_new264(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_QUALIFIED_CALL_EXPRESSION] */
extern T0* GE_new265(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ACTUAL_ARGUMENT_LIST] */
extern T0* GE_new266(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_RESULT] */
extern T0* GE_new267(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST */
extern T0* GE_new268(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET */
extern T0* GE_new269(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET] */
extern T0* GE_new270(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [[attached] ET_EXPRESSION] */
extern T0* GE_new271(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION] */
extern T0* GE_new272(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32] */
extern T0* GE_new273(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST */
extern T0* GE_new274(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new275(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new276(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST] */
extern T0* GE_new277(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
extern T0* GE_new278(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION] */
extern T0* GE_new279(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT] */
extern T0* GE_new280(T1 initialize);
/* New instance of type [detachable] ET_CURRENT */
extern T0* GE_new282(T1 initialize);
/* New instance of type [detachable] ET_ACTUAL_ARGUMENT_LIST */
extern T0* GE_new283(T1 initialize);
/* New instance of type [detachable] ET_QUALIFIED_CALL_INSTRUCTION */
extern T0* GE_new284(T1 initialize);
/* New instance of type [detachable] ET_QUALIFIED_CALL_EXPRESSION */
extern T0* GE_new287(T1 initialize);
/* New instance of type [detachable] ET_MANIFEST_TUPLE */
extern T0* GE_new288(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE] */
extern T0* GE_new289(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE] */
extern T0* GE_new290(T1 initialize);
/* New instance of type [detachable] ARRAY [[attached] INTEGER_32] */
extern T0* GE_new291(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE] */
extern T0* GE_new292(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT] */
extern T0* GE_new293(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new294(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE] */
extern T0* GE_new295(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8] */
extern T0* GE_new296(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER] */
extern T0* GE_new297(T1 initialize);
/* New instance of type [detachable] ET_IDENTIFIER_TESTER */
extern T0* GE_new298(T1 initialize);
/* New instance of type [detachable] KL_TEXT_OUTPUT_FILE */
extern T0* GE_new300(T1 initialize);
/* New instance of type [detachable] ET_IMPLICIT_TYPE_MARK */
extern T0* GE_new301(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_TUPLE_TYPE */
extern T0* GE_new302(T1 initialize);
/* New instance of type [detachable] DS_STRING_HASH_TABLE */
extern T0* GE_new303(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8] */
extern T0* GE_new304(T1 initialize);
/* New instance of type [detachable] UT_CANNOT_WRITE_TO_FILE_ERROR */
extern T0* GE_new305(T1 initialize);
/* New instance of type detachable ET_DYNAMIC_PRECURSOR */
extern T0* GE_new306(T1 initialize);
/* New instance of type detachable ET_DYNAMIC_PRECURSOR_LIST */
extern T0* GE_new307(T1 initialize);
/* New instance of type [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new308(T1 initialize);
/* New instance of type [detachable] ET_RESULT */
extern T0* GE_new309(T1 initialize);
/* New instance of type [detachable] ET_OBJECT_TEST */
extern T0* GE_new311(T1 initialize);
/* New instance of type [detachable] ET_OBJECT_EQUALITY_EXPRESSION */
extern T0* GE_new312(T1 initialize);
/* New instance of type [detachable] ET_EQUALITY_EXPRESSION */
extern T0* GE_new313(T1 initialize);
/* New instance of type detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL */
extern T0* GE_new314(T1 initialize);
/* New instance of type detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL */
extern T0* GE_new318(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_SPECIAL_TYPE */
extern T0* GE_new319(T1 initialize);
/* New instance of type [detachable] DS_QUICK_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new322(T1 initialize);
/* New instance of type detachable ET_DYNAMIC_SECONDARY_TYPE */
extern T0* GE_new323(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_PRIMARY_TYPE_COMPARATOR_BY_ID */
extern T0* GE_new324(T1 initialize);
/* New instance of type [detachable] ET_CREATE_EXPRESSION */
extern T0* GE_new328(T1 initialize);
/* New instance of type [detachable] ET_QUALIFIED_CALL */
extern T0* GE_new329(T1 initialize);
/* New instance of type [detachable] KL_DIRECTORY */
extern T0* GE_new333(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new334(T1 initialize);
/* New instance of type [detachable] ET_SYMBOL */
extern T0* GE_new341(T1 initialize);
/* New instance of type [detachable] DS_QUICK_SORTER [[attached] INTEGER_32] */
extern T0* GE_new344(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE] */
extern T0* GE_new346(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_IDENTIFIER, [attached] ET_CURRENT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
extern T0* GE_new347(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_IDENTIFIER, [attached] ET_RESULT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
extern T0* GE_new348(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_DYNAMIC_FEATURE, [attached] ET_RESULT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
extern T0* GE_new349(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_C_GENERATOR, INTEGER_32, [attached] ET_RESULT, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
extern T0* GE_new350(T1 initialize);
/* New instance of type [detachable] UT_INTEGER_FORMATTER */
extern T0* GE_new351(T1 initialize);
/* New instance of type detachable ET_COMPOUND */
extern T0* GE_new355(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_C_GENERATOR, INTEGER_32, [attached] ET_EXPRESSION, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
extern T0* GE_new357(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_EXPRESSION, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new358(T1 initialize);
/* New instance of type [detachable] ET_EXTENDED_ATTRIBUTE */
extern T0* GE_new359(T1 initialize);
/* New instance of type [detachable] ET_ATTRIBUTE */
extern T0* GE_new360(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
extern T0* GE_new361(T1 initialize);
/* New instance of type detachable ET_ITERATION_COMPONENT_LIST */
extern T0* GE_new365(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_DYNAMIC_FEATURE, [attached] ET_EXPRESSION, [attached] ET_DYNAMIC_PRIMARY_TYPE, BOOLEAN] */
extern T0* GE_new367(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_DYNAMIC_FEATURE, INTEGER_32, INTEGER_32] */
extern T0* GE_new368(T1 initialize);
/* New instance of type [detachable] ET_CONSTANT_ATTRIBUTE */
extern T0* GE_new369(T1 initialize);
/* New instance of type [detachable] ET_UNIQUE_ATTRIBUTE */
extern T0* GE_new371(T1 initialize);
/* New instance of type [detachable] ET_REGULAR_INTEGER_CONSTANT */
extern T0* GE_new372(T1 initialize);
/* New instance of type [detachable] ET_SYSTEM_MULTIPROCESSOR */
extern T0* GE_new375(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8] */
extern T0* GE_new376(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_8] */
extern T0* GE_new377(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] INTEGER_32] */
extern T0* GE_new379(T1 initialize);
/* New instance of type [detachable] ET_FEATURE_CHECKER */
extern T0* GE_new380(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE] */
extern T0* GE_new381(T1 initialize);
/* New instance of type detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE] */
extern T0* GE_new382(T1 initialize);
/* New instance of type [detachable] ET_TYPE_CHECKER */
extern T0* GE_new384(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT] */
extern T0* GE_new386(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
extern T0* GE_new387(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
extern T0* GE_new388(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
extern T0* GE_new389(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
extern T0* GE_new390(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT] */
extern T0* GE_new391(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST] */
extern T0* GE_new392(T1 initialize);
/* New instance of type [detachable] ET_OBJECT_TEST_SCOPE */
extern T0* GE_new393(T1 initialize);
/* New instance of type [detachable] ET_OBJECT_TEST_SCOPE_BUILDER */
extern T0* GE_new394(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT] */
extern T0* GE_new395(T1 initialize);
/* New instance of type [detachable] ET_ITERATION_ITEM_SCOPE */
extern T0* GE_new396(T1 initialize);
/* New instance of type [detachable] ET_ATTACHMENT_SCOPE */
extern T0* GE_new397(T1 initialize);
/* New instance of type [detachable] ET_ATTACHMENT_SCOPE_BUILDER */
extern T0* GE_new398(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE] */
extern T0* GE_new399(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE] */
extern T0* GE_new400(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM] */
extern T0* GE_new401(T1 initialize);
/* New instance of type [detachable] ET_CLIENT_LIST */
extern T0* GE_new402(T1 initialize);
/* New instance of type [detachable] ET_CLIENT */
extern T0* GE_new403(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
extern T0* GE_new404(T1 initialize);
/* New instance of type [detachable] ET_ADAPTED_BASE_CLASS_CHECKER */
extern T0* GE_new405(T1 initialize);
/* New instance of type detachable ET_PRECONDITIONS */
extern T0* GE_new406(T1 initialize);
/* New instance of type detachable ET_FEATURE_LIST */
extern T0* GE_new408(T1 initialize);
/* New instance of type [detachable] ET_UNKNOWN_GROUP */
extern T0* GE_new410(T1 initialize);
/* New instance of type [detachable] ET_BASE_TYPE_LIST */
extern T0* GE_new412(T1 initialize);
/* New instance of type [detachable] ET_KEYWORD */
extern T0* GE_new413(T1 initialize);
/* New instance of type [detachable] ET_CLASS_CODES */
extern T0* GE_new414(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new415(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new416(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new417(T1 initialize);
/* New instance of type [detachable] RX_BYTE_CODE */
extern T0* GE_new422(T1 initialize);
/* New instance of type [detachable] RX_CHARACTER_SET */
extern T0* GE_new423(T1 initialize);
/* New instance of type [detachable] RX_CASE_MAPPING */
extern T0* GE_new424(T1 initialize);
/* New instance of type [detachable] UC_UNICODE_ROUTINES */
extern T0* GE_new426(T1 initialize);
/* New instance of type [detachable] ARRAY [[attached] RX_CHARACTER_SET] */
extern T0* GE_new427(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] RX_CHARACTER_SET] */
extern T0* GE_new428(T6 a1, T1 initialize);
/* New instance of type [detachable] UC_STRING */
extern T0* GE_new429(T1 initialize);
/* New instance of type [detachable] NATIVE_STRING */
extern T0* GE_new430(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] GECC, [attached] DS_ARRAYED_LIST [[attached] STRING_8], INTEGER_32, [attached] MUTEX] */
extern T0* GE_new431(T1 initialize);
/* New instance of type [detachable] WORKER_THREAD */
extern T0* GE_new432(T1 initialize);
/* New instance of type [detachable] UC_UTF8_ROUTINES */
extern T0* GE_new434(T1 initialize);
/* New instance of type [detachable] PATH */
extern T0* GE_new435(T1 initialize);
/* New instance of type [detachable] STRING_TO_INTEGER_CONVERTOR */
extern T0* GE_new436(T1 initialize);
/* New instance of type [detachable] ET_ECF_OPTIONS */
extern T0* GE_new444(T1 initialize);
/* New instance of type detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT] */
extern T0* GE_new445(T1 initialize);
/* New instance of type [detachable] XM_NAMESPACE */
extern T0* GE_new446(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE] */
extern T0* GE_new447(T1 initialize);
/* New instance of type [detachable] ET_LIKE_CURRENT */
extern T0* GE_new449(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_PROCEDURE_TYPE */
extern T0* GE_new450(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_FUNCTION_TYPE */
extern T0* GE_new451(T1 initialize);
/* New instance of type [detachable] DT_SHARED_SYSTEM_CLOCK */
extern T0* GE_new452(T1 initialize);
/* New instance of type [detachable] DT_SYSTEM_CLOCK */
extern T0* GE_new453(T1 initialize);
/* New instance of type [detachable] ET_EIFFEL_PREPARSER */
extern T0* GE_new454(T1 initialize);
/* New instance of type [detachable] ET_MASTER_CLASS_CHECKER */
extern T0* GE_new455(T1 initialize);
/* New instance of type [detachable] ET_EIFFEL_PARSER */
extern T0* GE_new456(T1 initialize);
/* New instance of type [detachable] ET_PROVIDER_CHECKER */
extern T0* GE_new457(T1 initialize);
/* New instance of type [detachable] ET_ANCESTOR_BUILDER */
extern T0* GE_new458(T1 initialize);
/* New instance of type [detachable] ET_FEATURE_FLATTENER */
extern T0* GE_new459(T1 initialize);
/* New instance of type [detachable] ET_INTERFACE_CHECKER */
extern T0* GE_new460(T1 initialize);
/* New instance of type [detachable] ET_IMPLEMENTATION_CHECKER */
extern T0* GE_new461(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS] */
extern T0* GE_new462(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] DS_ARRAYED_LIST [[attached] ET_CLASS]] */
extern T0* GE_new463(T1 initialize);
/* New instance of type [detachable] ET_AST_NULL_PROCESSOR */
extern T0* GE_new464(T1 initialize);
/* New instance of type [detachable] ET_AST_FACTORY */
extern T0* GE_new465(T1 initialize);
/* New instance of type [detachable] ET_DOTNET_ASSEMBLY_CLASSIC_CONSUMER */
extern T0* GE_new467(T1 initialize);
/* New instance of type [detachable] DT_DATE_TIME_DURATION */
extern T0* GE_new468(T1 initialize);
/* New instance of type [detachable] ET_DO_PROCEDURE */
extern T0* GE_new469(T1 initialize);
/* New instance of type [detachable] UT_CONFIG_PARSER */
extern T0* GE_new471(T1 initialize);
/* New instance of type [detachable] KL_COMPARABLE_COMPARATOR [[attached] INTEGER_32] */
extern T0* GE_new472(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_PUSH_TYPE_SET */
extern T0* GE_new475(T1 initialize);
/* New instance of type [detachable] ET_DEFERRED_PROCEDURE */
extern T0* GE_new476(T1 initialize);
/* New instance of type [detachable] KL_PLATFORM */
extern T0* GE_new477(T1 initialize);
/* New instance of type [detachable] CELL [[attached] INTEGER_32] */
extern T0* GE_new479(T1 initialize);
/* New instance of type [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32] */
extern T0* GE_new480(T1 initialize);
/* New instance of type [detachable] KL_ANY_ROUTINES */
extern T0* GE_new481(T1 initialize);
/* New instance of type [detachable] KL_PATHNAME */
extern T0* GE_new483(T1 initialize);
/* New instance of type [detachable] ARRAY [[attached] IMMUTABLE_STRING_32] */
extern T0* GE_new484(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] IMMUTABLE_STRING_32] */
extern T0* GE_new485(T6 a1, T1 initialize);
/* New instance of type [detachable] EXCEPTIONS */
extern T0* GE_new486(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] NATURAL_8] */
extern T0* GE_new487(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_OPTION] */
extern T0* GE_new489(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] AP_OPTION] */
extern T0* GE_new491(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] AP_OPTION] */
extern T0* GE_new492(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
extern T0* GE_new493(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
extern T0* GE_new494(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] AP_ALTERNATIVE_OPTIONS_LIST] */
extern T0* GE_new495(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] STRING_8] */
extern T0* GE_new496(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable STRING_8] */
extern T0* GE_new497(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable STRING_8] */
extern T0* GE_new498(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST_CURSOR [[attached] STRING_8] */
extern T0* GE_new499(T1 initialize);
/* New instance of type detachable DS_LINKABLE [[attached] STRING_8] */
extern T0* GE_new500(T1 initialize);
/* New instance of type detachable DS_LINKABLE [[attached] BOOLEAN] */
extern T0* GE_new501(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST_CURSOR [[attached] BOOLEAN] */
extern T0* GE_new502(T1 initialize);
/* New instance of type detachable DS_LINKABLE [[attached] INTEGER_32] */
extern T0* GE_new503(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST_CURSOR [[attached] INTEGER_32] */
extern T0* GE_new504(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_LIBRARY, [attached] STRING_8] */
extern T0* GE_new506(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_LIBRARY] */
extern T0* GE_new507(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_LIBRARY] */
extern T0* GE_new510(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_DOTNET_ASSEMBLY] */
extern T0* GE_new511(T6 a1, T1 initialize);
/* New instance of type detachable DS_HASH_TABLE_CURSOR [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8] */
extern T0* GE_new514(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_DOTNET_ASSEMBLY] */
extern T0* GE_new515(T1 initialize);
/* New instance of type [detachable] XM_EIFFEL_SCANNER */
extern T0* GE_new516(T1 initialize);
/* New instance of type [detachable] XM_DEFAULT_POSITION */
extern T0* GE_new517(T1 initialize);
/* New instance of type [detachable] DS_BILINKED_LIST [[attached] XM_POSITION] */
extern T0* GE_new519(T1 initialize);
/* New instance of type [detachable] DS_LINKED_STACK [[attached] XM_EIFFEL_SCANNER] */
extern T0* GE_new520(T1 initialize);
/* New instance of type [detachable] XM_CALLBACKS_NULL */
extern T0* GE_new521(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8] */
extern T0* GE_new522(T1 initialize);
/* New instance of type [detachable] XM_NULL_EXTERNAL_RESOLVER */
extern T0* GE_new524(T1 initialize);
/* New instance of type [detachable] XM_DTD_CALLBACKS_NULL */
extern T0* GE_new526(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ANY] */
extern T0* GE_new527(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ANY] */
extern T0* GE_new528(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_PARSER_NAME] */
extern T0* GE_new529(T1 initialize);
/* New instance of type [detachable] XM_EIFFEL_PARSER_NAME */
extern T0* GE_new530(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] XM_EIFFEL_PARSER_NAME] */
extern T0* GE_new531(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]] */
extern T0* GE_new532(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME] */
extern T0* GE_new533(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]] */
extern T0* GE_new534(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_EXTERNAL_ID] */
extern T0* GE_new535(T1 initialize);
/* New instance of type [detachable] XM_DTD_EXTERNAL_ID */
extern T0* GE_new536(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] XM_DTD_EXTERNAL_ID] */
extern T0* GE_new537(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ELEMENT_CONTENT] */
extern T0* GE_new538(T1 initialize);
/* New instance of type [detachable] XM_DTD_ELEMENT_CONTENT */
extern T0* GE_new539(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] XM_DTD_ELEMENT_CONTENT] */
extern T0* GE_new540(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]] */
extern T0* GE_new541(T1 initialize);
/* New instance of type [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
extern T0* GE_new542(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]] */
extern T0* GE_new543(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
extern T0* GE_new544(T1 initialize);
/* New instance of type [detachable] XM_DTD_ATTRIBUTE_CONTENT */
extern T0* GE_new545(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
extern T0* GE_new546(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] STRING_8]] */
extern T0* GE_new547(T1 initialize);
/* New instance of type [detachable] DS_BILINKED_LIST [[attached] STRING_8] */
extern T0* GE_new548(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] STRING_8]] */
extern T0* GE_new549(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] BOOLEAN] */
extern T0* GE_new550(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_DECLARATION] */
extern T0* GE_new551(T1 initialize);
/* New instance of type [detachable] XM_EIFFEL_DECLARATION */
extern T0* GE_new552(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] XM_EIFFEL_DECLARATION] */
extern T0* GE_new553(T6 a1, T1 initialize);
/* New instance of type [detachable] XM_EIFFEL_ENTITY_DEF */
extern T0* GE_new556(T1 initialize);
/* New instance of type [detachable] XM_EIFFEL_SCANNER_DTD */
extern T0* GE_new557(T1 initialize);
/* New instance of type [detachable] XM_EIFFEL_PE_ENTITY_DEF */
extern T0* GE_new559(T1 initialize);
/* New instance of type [detachable] XM_NAMESPACE_RESOLVER */
extern T0* GE_new560(T1 initialize);
/* New instance of type [detachable] ARRAY [[attached] XM_CALLBACKS_FILTER] */
extern T0* GE_new561(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] XM_CALLBACKS_FILTER] */
extern T0* GE_new562(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] XM_NAMESPACE] */
extern T0* GE_new563(T1 initialize);
/* New instance of type [detachable] ET_CLUSTERS */
extern T0* GE_new564(T1 initialize);
/* New instance of type [detachable] ET_CLASS_NAME_TESTER */
extern T0* GE_new565(T1 initialize);
/* New instance of type [detachable] ET_PARENT */
extern T0* GE_new567(T1 initialize);
/* New instance of type [detachable] ET_PARENT_LIST */
extern T0* GE_new568(T1 initialize);
/* New instance of type [detachable] ET_BUILTIN_CONVERT_FEATURE */
extern T0* GE_new569(T1 initialize);
/* New instance of type [detachable] ET_MASTER_CLASS */
extern T0* GE_new570(T1 initialize);
/* New instance of type [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD */
extern T0* GE_new571(T1 initialize);
/* New instance of type detachable ET_RENAME_LIST */
extern T0* GE_new572(T1 initialize);
/* New instance of type detachable ET_EXPORT_LIST */
extern T0* GE_new573(T1 initialize);
/* New instance of type detachable ET_KEYWORD_FEATURE_NAME_LIST */
extern T0* GE_new574(T1 initialize);
/* New instance of type [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS */
extern T0* GE_new577(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY] */
extern T0* GE_new578(T1 initialize);
/* New instance of type [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE] */
extern T0* GE_new579(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_SYSTEM_CONFIG] */
extern T0* GE_new582(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_SYSTEM_CONFIG] */
extern T0* GE_new584(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8] */
extern T0* GE_new585(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_TARGET, [attached] STRING_8] */
extern T0* GE_new586(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_TARGET] */
extern T0* GE_new587(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET] */
extern T0* GE_new589(T1 initialize);
/* New instance of type detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
extern T0* GE_new591(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable RX_REGULAR_EXPRESSION] */
extern T0* GE_new592(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION] */
extern T0* GE_new593(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
extern T0* GE_new594(T1 initialize);
/* New instance of type [detachable] ET_DOTNET_ASSEMBLIES */
extern T0* GE_new595(T1 initialize);
/* New instance of type [detachable] ET_DOTNET_ASSEMBLY */
extern T0* GE_new596(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_IDENTIFIER] */
extern T0* GE_new599(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_IDENTIFIER] */
extern T0* GE_new600(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8] */
extern T0* GE_new601(T1 initialize);
/* New instance of type detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32] */
extern T0* GE_new602(T1 initialize);
/* New instance of type detachable ET_FORMAL_PARAMETER_LIST */
extern T0* GE_new603(T1 initialize);
/* New instance of type detachable ET_FEATURE_IDS */
extern T0* GE_new604(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_DYNAMIC_FEATURE] */
extern T0* GE_new605(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_FEATURE] */
extern T0* GE_new606(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME] */
extern T0* GE_new609(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MASTER_CLASS] */
extern T0* GE_new610(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_MASTER_CLASS] */
extern T0* GE_new611(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS_NAME] */
extern T0* GE_new612(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CLASS_NAME] */
extern T0* GE_new613(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8] */
extern T0* GE_new615(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME] */
extern T0* GE_new616(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME] */
extern T0* GE_new617(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ACTUAL_PARAMETER_ITEM] */
extern T0* GE_new621(T6 a1, T1 initialize);
/* New instance of type [detachable] ET_BRACKET_SYMBOL */
extern T0* GE_new622(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACTUAL_PARAMETER_ITEM] */
extern T0* GE_new624(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE] */
extern T0* GE_new625(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE_SET] */
extern T0* GE_new626(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE_SET] */
extern T0* GE_new627(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_IDENTIFIER] */
extern T0* GE_new628(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new629(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new630(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new631(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] INTEGER_32] */
extern T0* GE_new632(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_UNQUALIFIED_CALL_EXPRESSION] */
extern T0* GE_new634(T1 initialize);
/* New instance of type [detachable] ET_UNQUALIFIED_CALL_EXPRESSION */
extern T0* GE_new635(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_UNQUALIFIED_CALL_EXPRESSION] */
extern T0* GE_new636(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_UNQUALIFIED_CALL_EXPRESSION] */
extern T0* GE_new637(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_QUALIFIED_CALL_EXPRESSION] */
extern T0* GE_new638(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_QUALIFIED_CALL_EXPRESSION] */
extern T0* GE_new639(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUALIFIED_CALL_EXPRESSION] */
extern T0* GE_new640(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACTUAL_ARGUMENT_LIST] */
extern T0* GE_new641(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ACTUAL_ARGUMENT_LIST] */
extern T0* GE_new642(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ACTUAL_ARGUMENT_LIST] */
extern T0* GE_new643(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RESULT] */
extern T0* GE_new644(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_RESULT] */
extern T0* GE_new645(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_RESULT] */
extern T0* GE_new646(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET] */
extern T0* GE_new647(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET] */
extern T0* GE_new648(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET] */
extern T0* GE_new649(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_EXPRESSION] */
extern T0* GE_new650(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION] */
extern T0* GE_new651(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EXPRESSION] */
extern T0* GE_new652(T1 initialize);
/* New instance of type detachable KL_EQUALITY_TESTER [[attached] INTEGER_32] */
extern T0* GE_new653(T1 initialize);
/* New instance of type detachable DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32] */
extern T0* GE_new654(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new655(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_OBJECT_TEST] */
extern T0* GE_new657(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST] */
extern T0* GE_new658(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_TEST] */
extern T0* GE_new659(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
extern T0* GE_new660(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
extern T0* GE_new661(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_EQUALITY_EXPRESSION] */
extern T0* GE_new662(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_EQUALITY_EXPRESSION] */
extern T0* GE_new663(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EQUALITY_EXPRESSION] */
extern T0* GE_new664(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EQUALITY_EXPRESSION] */
extern T0* GE_new665(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_AGENT] */
extern T0* GE_new666(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT] */
extern T0* GE_new667(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_AGENT] */
extern T0* GE_new668(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_EXPRESSION_ITEM] */
extern T0* GE_new669(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION_ITEM] */
extern T0* GE_new671(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_TUPLE_TYPE] */
extern T0* GE_new672(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_DYNAMIC_TUPLE_TYPE] */
extern T0* GE_new673(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TUPLE_TYPE] */
extern T0* GE_new674(T1 initialize);
/* New instance of type detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_FEATURE] */
extern T0* GE_new676(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_FEATURE] */
extern T0* GE_new677(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE] */
extern T0* GE_new678(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CONSTANT, [attached] ET_FEATURE] */
extern T0* GE_new679(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CONSTANT] */
extern T0* GE_new680(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTANT] */
extern T0* GE_new682(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT] */
extern T0* GE_new683(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_INLINE_CONSTANT] */
extern T0* GE_new684(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_CONSTANT] */
extern T0* GE_new686(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_DYNAMIC_FEATURE] */
extern T0* GE_new687(T6 a1, T1 initialize);
/* New instance of type detachable DS_HASH_TABLE_CURSOR [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new689(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_FEATURE] */
extern T0* GE_new690(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_FEATURE] */
extern T0* GE_new691(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] STRING_8] */
extern T0* GE_new692(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_IDENTIFIER] */
extern T0* GE_new695(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new698(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]] */
extern T0* GE_new699(T1 initialize);
/* New instance of type [detachable] ARRAY [[attached] BOOLEAN] */
extern T0* GE_new700(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_SYSTEM_PROCESSOR] */
extern T0* GE_new703(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_SYSTEM_PROCESSOR, [attached] DS_ARRAYED_LIST [[attached] ET_CLASS]] */
extern T0* GE_new704(T1 initialize);
/* New instance of type detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE] */
extern T0* GE_new707(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE] */
extern T0* GE_new708(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE] */
extern T0* GE_new709(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS] */
extern T0* GE_new710(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_AGENT] */
extern T0* GE_new711(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_INLINE_AGENT] */
extern T0* GE_new712(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INLINE_AGENT] */
extern T0* GE_new713(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
extern T0* GE_new714(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE] */
extern T0* GE_new715(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
extern T0* GE_new716(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]] */
extern T0* GE_new717(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
extern T0* GE_new718(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY] */
extern T0* GE_new719(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
extern T0* GE_new720(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]] */
extern T0* GE_new721(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
extern T0* GE_new722(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE] */
extern T0* GE_new723(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
extern T0* GE_new724(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]] */
extern T0* GE_new725(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
extern T0* GE_new726(T1 initialize);
/* New instance of type [detachable] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT] */
extern T0* GE_new727(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
extern T0* GE_new728(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]] */
extern T0* GE_new729(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_NESTED_TYPE_CONTEXT] */
extern T0* GE_new730(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NESTED_TYPE_CONTEXT] */
extern T0* GE_new731(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT] */
extern T0* GE_new732(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST] */
extern T0* GE_new733(T1 initialize);
/* New instance of type [detachable] ET_NAMED_OBJECT_TEST */
extern T0* GE_new736(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_NAMED_OBJECT_TEST] */
extern T0* GE_new737(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_OBJECT_TEST] */
extern T0* GE_new738(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST] */
extern T0* GE_new739(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT] */
extern T0* GE_new740(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ITERATION_COMPONENT] */
extern T0* GE_new741(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ITERATION_COMPONENT] */
extern T0* GE_new743(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ITERATION_COMPONENT] */
extern T0* GE_new744(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] INTEGER_32] */
extern T0* GE_new745(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ATTACHMENT_SCOPE] */
extern T0* GE_new747(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ATTACHMENT_SCOPE] */
extern T0* GE_new748(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ATTACHMENT_SCOPE] */
extern T0* GE_new749(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ASSERTIONS, [attached] ET_FEATURE] */
extern T0* GE_new750(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ASSERTIONS] */
extern T0* GE_new751(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTIONS] */
extern T0* GE_new753(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM] */
extern T0* GE_new754(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_INDEXING_TERM] */
extern T0* GE_new756(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INDEXING_TERM] */
extern T0* GE_new757(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CLIENT_ITEM] */
extern T0* GE_new758(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_ITEM] */
extern T0* GE_new759(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
extern T0* GE_new760(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
extern T0* GE_new761(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]] */
extern T0* GE_new762(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME] */
extern T0* GE_new763(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] NATURAL_32] */
extern T0* GE_new764(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET] */
extern T0* GE_new765(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_32] */
extern T0* GE_new766(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] NATURAL_64] */
extern T0* GE_new767(T6 a1, T1 initialize);
/* New instance of type detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32] */
extern T0* GE_new768(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32] */
extern T0* GE_new769(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]] */
extern T0* GE_new770(T6 a1, T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]]] */
extern T0* GE_new771(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_INTEGER_ROUTINES */
extern T0* GE_new772(T1 initialize);
/* New instance of type [detachable] UC_UTF8_STRING */
extern T0* GE_new775(T1 initialize);
/* New instance of type [detachable] THREAD_ATTRIBUTES */
extern T0* GE_new776(T1 initialize);
/* New instance of type [detachable] CELL [[attached] BOOLEAN] */
extern T0* GE_new777(T1 initialize);
/* New instance of type [detachable] CHARACTER_PROPERTY */
extern T0* GE_new780(T1 initialize);
/* New instance of type [detachable] STRING_8_SEARCHER */
extern T0* GE_new781(T1 initialize);
/* New instance of type detachable ARRAYED_LIST [[attached] INTEGER_32] */
extern T0* GE_new782(T1 initialize);
/* New instance of type [detachable] ET_SYSTEM_ERROR */
extern T0* GE_new783(T1 initialize);
/* New instance of type [detachable] ET_INTERNAL_ERROR */
extern T0* GE_new784(T1 initialize);
/* New instance of type [detachable] ET_VALIDITY_ERROR */
extern T0* GE_new785(T1 initialize);
/* New instance of type detachable ET_ECF_NOTE_ELEMENT */
extern T0* GE_new786(T1 initialize);
/* New instance of type detachable ET_ECF_CLUSTERS */
extern T0* GE_new787(T1 initialize);
/* New instance of type detachable ET_ECF_ADAPTED_LIBRARIES */
extern T0* GE_new788(T1 initialize);
/* New instance of type detachable ET_ECF_ADAPTED_DOTNET_ASSEMBLIES */
extern T0* GE_new789(T1 initialize);
/* New instance of type detachable ET_ECF_FILE_RULES */
extern T0* GE_new790(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_CFLAGS */
extern T0* GE_new791(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_INCLUDES */
extern T0* GE_new792(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_LIBRARIES */
extern T0* GE_new793(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_LINKER_FLAGS */
extern T0* GE_new794(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_MAKES */
extern T0* GE_new795(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_OBJECTS */
extern T0* GE_new796(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_RESOURCES */
extern T0* GE_new797(T1 initialize);
/* New instance of type detachable DS_ARRAYED_LIST [[attached] ET_ECF_ACTION] */
extern T0* GE_new798(T1 initialize);
/* New instance of type detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY */
extern T0* GE_new799(T1 initialize);
/* New instance of type detachable ET_ECF_VERSION */
extern T0* GE_new801(T1 initialize);
/* New instance of type detachable ET_ECF_CLUSTER */
extern T0* GE_new802(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_CFLAG */
extern T0* GE_new803(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_INCLUDE */
extern T0* GE_new804(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_LIBRARY */
extern T0* GE_new805(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_LINKER_FLAG */
extern T0* GE_new806(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_MAKE */
extern T0* GE_new807(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_OBJECT */
extern T0* GE_new808(T1 initialize);
/* New instance of type detachable ET_ECF_EXTERNAL_RESOURCE */
extern T0* GE_new809(T1 initialize);
/* New instance of type detachable ET_ECF_FILE_RULE */
extern T0* GE_new810(T1 initialize);
/* New instance of type detachable ET_ECF_ACTION */
extern T0* GE_new811(T1 initialize);
/* New instance of type detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8] */
extern T0* GE_new812(T1 initialize);
/* New instance of type detachable DS_ARRAYED_LIST [[attached] ET_ECF_VISIBLE_CLASS] */
extern T0* GE_new813(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_OPTIONS] */
extern T0* GE_new814(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_OPTIONS]] */
extern T0* GE_new815(T1 initialize);
/* New instance of type detachable ET_ECF_ROOT_CLASS */
extern T0* GE_new816(T1 initialize);
/* New instance of type [detachable] ET_ECF_ROOT_ALL_CLASSES */
extern T0* GE_new817(T1 initialize);
/* New instance of type [detachable] ET_ECF_ORED_CONDITIONS */
extern T0* GE_new818(T1 initialize);
/* New instance of type detachable ET_ECF_ANDED_CONDITIONS */
extern T0* GE_new819(T1 initialize);
/* New instance of type detachable ET_ECF_VISIBLE_CLASS */
extern T0* GE_new820(T1 initialize);
/* New instance of type detachable ET_ECF_BUILD_CONDITION */
extern T0* GE_new822(T1 initialize);
/* New instance of type detachable ET_ECF_CONCURRENCY_CONDITION */
extern T0* GE_new823(T1 initialize);
/* New instance of type detachable ET_ECF_VOID_SAFETY_CONDITION */
extern T0* GE_new824(T1 initialize);
/* New instance of type detachable ET_ECF_CUSTOM_CONDITION */
extern T0* GE_new825(T1 initialize);
/* New instance of type detachable ET_ECF_DOTNET_CONDITION */
extern T0* GE_new826(T1 initialize);
/* New instance of type detachable ET_ECF_DYNAMIC_RUNTIME_CONDITION */
extern T0* GE_new827(T1 initialize);
/* New instance of type detachable ET_ECF_PLATFORM_CONDITION */
extern T0* GE_new828(T1 initialize);
/* New instance of type [detachable] ET_ECF_COMPILER_VERSION_CONDITION */
extern T0* GE_new830(T1 initialize);
/* New instance of type [detachable] ET_ECF_MSIL_CLR_VERSION_CONDITION */
extern T0* GE_new831(T1 initialize);
/* New instance of type [detachable] UT_COUNTER */
extern T0* GE_new833(T1 initialize);
/* New instance of type [detachable] KL_AGENT_ROUTINES [[attached] ET_CLASS] */
extern T0* GE_new834(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] UT_COUNTER] */
extern T0* GE_new835(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] KL_AGENT_ROUTINES [[attached] ET_CLASS], [attached] PROCEDURE [[attached] TUPLE]] */
extern T0* GE_new836(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_UNIVERSE] */
extern T0* GE_new837(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_UNIVERSE]] */
extern T0* GE_new838(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]]] */
extern T0* GE_new839(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]], [attached] FUNCTION [[attached] TUPLE, [attached] BOOLEAN]] */
extern T0* GE_new840(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY] */
extern T0* GE_new842(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_DOTNET_ASSEMBLY] */
extern T0* GE_new843(T1 initialize);
/* New instance of type [detachable] PREDICATE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]] */
extern T0* GE_new844(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]] */
extern T0* GE_new845(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY]] */
extern T0* GE_new846(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_INTERNAL_UNIVERSE] */
extern T0* GE_new847(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_INTERNAL_UNIVERSE]] */
extern T0* GE_new848(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]], [attached] PREDICATE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]]] */
extern T0* GE_new849(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_DOTNET_ASSEMBLIES] */
extern T0* GE_new850(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_SYSTEM_PROCESSOR] */
extern T0* GE_new851(T1 initialize);
/* New instance of type [detachable] ET_LIBRARY */
extern T0* GE_new852(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_LIBRARY] */
extern T0* GE_new853(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_LIBRARY]] */
extern T0* GE_new854(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_UNIVERSE] */
extern T0* GE_new855(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_CLUSTER] */
extern T0* GE_new857(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_CLUSTER]] */
extern T0* GE_new858(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_SYSTEM] */
extern T0* GE_new859(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE] */
extern T0* GE_new860(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_AST_PROCESSOR] */
extern T0* GE_new861(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_MASTER_CLASS] */
extern T0* GE_new862(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]] */
extern T0* GE_new863(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
extern T0* GE_new864(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ADAPTED_DOTNET_ASSEMBLY]] */
extern T0* GE_new865(T1 initialize);
/* New instance of type [detachable] ET_ANCESTORS_STATUS_CHECKER */
extern T0* GE_new866(T1 initialize);
/* New instance of type [detachable] ET_FLATTENING_STATUS_CHECKER */
extern T0* GE_new867(T1 initialize);
/* New instance of type [detachable] ET_INTERFACE_STATUS_CHECKER */
extern T0* GE_new868(T1 initialize);
/* New instance of type [detachable] ET_IMPLEMENTATION_STATUS_CHECKER */
extern T0* GE_new869(T1 initialize);
/* New instance of type [detachable] PREDICATE [[attached] TUPLE [[attached] ET_CLASS]] */
extern T0* GE_new870(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ANCESTORS_STATUS_CHECKER] */
extern T0* GE_new872(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_FLATTENING_STATUS_CHECKER] */
extern T0* GE_new873(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_INTERFACE_STATUS_CHECKER] */
extern T0* GE_new874(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_IMPLEMENTATION_STATUS_CHECKER] */
extern T0* GE_new875(T1 initialize);
/* New instance of type [detachable] TUPLE [BOOLEAN] */
extern T0* GE_new876(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_UNIVERSE]] */
extern T0* GE_new877(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLUSTER]]] */
extern T0* GE_new880(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE]] */
extern T0* GE_new884(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]]] */
extern T0* GE_new885(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]], [attached] FUNCTION [[attached] TUPLE [[attached] ET_CLASS], [attached] BOOLEAN]] */
extern T0* GE_new886(T1 initialize);
/* New instance of type [detachable] ET_NONE_GROUP */
extern T0* GE_new887(T1 initialize);
/* New instance of type detachable ET_PARENTHESIS_EXPRESSION */
extern T0* GE_new888(T1 initialize);
/* New instance of type detachable DS_ARRAYED_LIST [detachable ET_FORMAL_PARAMETER_TYPE] */
extern T0* GE_new890(T1 initialize);
/* New instance of type [detachable] ET_FORMAL_PARAMETER */
extern T0* GE_new891(T1 initialize);
/* New instance of type [detachable] STD_FILES */
extern T0* GE_new892(T1 initialize);
/* New instance of type [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE] */
extern T0* GE_new894(T1 initialize);
/* New instance of type [detachable] UC_STRING_EQUALITY_TESTER */
extern T0* GE_new896(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]] */
extern T0* GE_new898(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]] */
extern T0* GE_new899(T1 initialize);
/* New instance of type [detachable] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE] */
extern T0* GE_new900(T1 initialize);
/* New instance of type [detachable] XM_EIFFEL_INPUT_STREAM */
extern T0* GE_new901(T1 initialize);
/* New instance of type [detachable] KL_ARRAY_ROUTINES [[attached] INTEGER_32] */
extern T0* GE_new902(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_LIBRARY] */
extern T0* GE_new903(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] DS_HASH_SET [[attached] ET_LIBRARY]] */
extern T0* GE_new904(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE] */
extern T0* GE_new905(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET] */
extern T0* GE_new906(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE] */
extern T0* GE_new907(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT] */
extern T0* GE_new909(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_QUERY] */
extern T0* GE_new910(T6 a1, T1 initialize);
/* New instance of type detachable ET_ALIAS_NAME_LIST */
extern T0* GE_new911(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_QUERY] */
extern T0* GE_new912(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_PROCEDURE] */
extern T0* GE_new913(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PROCEDURE] */
extern T0* GE_new914(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE] */
extern T0* GE_new921(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_BASE_TYPE] */
extern T0* GE_new923(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE] */
extern T0* GE_new924(T1 initialize);
/* New instance of type [detachable] INTEGER_OVERFLOW_CHECKER */
extern T0* GE_new925(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_NOTE_ELEMENT] */
extern T0* GE_new926(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_NOTE_ELEMENT] */
extern T0* GE_new927(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_NOTE_ELEMENT] */
extern T0* GE_new928(T1 initialize);
/* New instance of type detachable C_DATE */
extern T0* GE_new929(T1 initialize);
/* New instance of type [detachable] YY_FILE_BUFFER */
extern T0* GE_new930(T1 initialize);
/* New instance of type [detachable] KL_STDIN_FILE */
extern T0* GE_new931(T1 initialize);
/* New instance of type [detachable] YY_BUFFER */
extern T0* GE_new932(T1 initialize);
/* New instance of type [detachable] YY_UNICODE_FILE_BUFFER */
extern T0* GE_new936(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32] */
extern T0* GE_new937(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST] */
extern T0* GE_new938(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST] */
extern T0* GE_new939(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [detachable ET_KEYWORD] */
extern T0* GE_new940(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [detachable ET_SYMBOL] */
extern T0* GE_new941(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST] */
extern T0* GE_new942(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [[attached] ET_OBJECT_TEST_LIST] */
extern T0* GE_new943(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [detachable ET_ITERATION_COMPONENT_LIST] */
extern T0* GE_new944(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [[attached] ET_ITERATION_COMPONENT_LIST] */
extern T0* GE_new945(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM] */
extern T0* GE_new946(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE] */
extern T0* GE_new947(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS] */
extern T0* GE_new948(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD] */
extern T0* GE_new949(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_KEYWORD] */
extern T0* GE_new950(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_KEYWORD] */
extern T0* GE_new951(T1 initialize);
/* New instance of type detachable ET_AGENT_KEYWORD */
extern T0* GE_new952(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_AGENT_KEYWORD] */
extern T0* GE_new953(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECURSOR_KEYWORD] */
extern T0* GE_new954(T1 initialize);
/* New instance of type detachable ET_PRECURSOR_KEYWORD */
extern T0* GE_new955(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_PRECURSOR_KEYWORD] */
extern T0* GE_new956(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL] */
extern T0* GE_new957(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_SYMBOL] */
extern T0* GE_new958(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSITION] */
extern T0* GE_new959(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_POSITION] */
extern T0* GE_new960(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_BOOLEAN_CONSTANT] */
extern T0* GE_new961(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_BOOLEAN_CONSTANT] */
extern T0* GE_new963(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_BREAK] */
extern T0* GE_new964(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_BREAK] */
extern T0* GE_new966(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHARACTER_CONSTANT] */
extern T0* GE_new967(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CHARACTER_CONSTANT] */
extern T0* GE_new969(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CURRENT] */
extern T0* GE_new970(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CURRENT] */
extern T0* GE_new971(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FREE_OPERATOR] */
extern T0* GE_new972(T1 initialize);
/* New instance of type detachable ET_FREE_OPERATOR */
extern T0* GE_new973(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FREE_OPERATOR] */
extern T0* GE_new974(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_IDENTIFIER] */
extern T0* GE_new975(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_IDENTIFIER] */
extern T0* GE_new976(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INTEGER_CONSTANT] */
extern T0* GE_new977(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_INTEGER_CONSTANT] */
extern T0* GE_new979(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_OPERATOR] */
extern T0* GE_new980(T1 initialize);
/* New instance of type detachable ET_KEYWORD_OPERATOR */
extern T0* GE_new981(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_KEYWORD_OPERATOR] */
extern T0* GE_new982(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING] */
extern T0* GE_new983(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_MANIFEST_STRING] */
extern T0* GE_new984(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_REAL_CONSTANT] */
extern T0* GE_new985(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_REAL_CONSTANT] */
extern T0* GE_new987(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_RESULT] */
extern T0* GE_new988(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_RESULT] */
extern T0* GE_new989(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_RETRY_INSTRUCTION] */
extern T0* GE_new990(T1 initialize);
/* New instance of type detachable ET_RETRY_INSTRUCTION */
extern T0* GE_new991(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_RETRY_INSTRUCTION] */
extern T0* GE_new992(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL_OPERATOR] */
extern T0* GE_new993(T1 initialize);
/* New instance of type detachable ET_SYMBOL_OPERATOR */
extern T0* GE_new994(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_SYMBOL_OPERATOR] */
extern T0* GE_new995(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_VOID] */
extern T0* GE_new996(T1 initialize);
/* New instance of type detachable ET_VOID */
extern T0* GE_new997(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_VOID] */
extern T0* GE_new998(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_SEMICOLON_SYMBOL] */
extern T0* GE_new999(T1 initialize);
/* New instance of type detachable ET_SEMICOLON_SYMBOL */
extern T0* GE_new1000(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_SEMICOLON_SYMBOL] */
extern T0* GE_new1001(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_SYMBOL] */
extern T0* GE_new1002(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_BRACKET_SYMBOL] */
extern T0* GE_new1003(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUESTION_MARK_SYMBOL] */
extern T0* GE_new1004(T1 initialize);
/* New instance of type detachable ET_QUESTION_MARK_SYMBOL */
extern T0* GE_new1005(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_QUESTION_MARK_SYMBOL] */
extern T0* GE_new1006(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_EXPRESSION] */
extern T0* GE_new1007(T1 initialize);
/* New instance of type detachable ET_ACROSS_EXPRESSION */
extern T0* GE_new1008(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ACROSS_EXPRESSION] */
extern T0* GE_new1009(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_INSTRUCTION] */
extern T0* GE_new1010(T1 initialize);
/* New instance of type detachable ET_ACROSS_INSTRUCTION */
extern T0* GE_new1011(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ACROSS_INSTRUCTION] */
extern T0* GE_new1012(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_ARGUMENT_LIST] */
extern T0* GE_new1013(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ACTUAL_ARGUMENT_LIST] */
extern T0* GE_new1014(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_ITEM] */
extern T0* GE_new1015(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_ITEM] */
extern T0* GE_new1016(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_LIST] */
extern T0* GE_new1017(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_LIST] */
extern T0* GE_new1018(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND] */
extern T0* GE_new1019(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND] */
extern T0* GE_new1020(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM] */
extern T0* GE_new1021(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM] */
extern T0* GE_new1023(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_LIST] */
extern T0* GE_new1024(T1 initialize);
/* New instance of type detachable ET_AGENT_ARGUMENT_OPERAND_LIST */
extern T0* GE_new1025(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_LIST] */
extern T0* GE_new1026(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_TARGET] */
extern T0* GE_new1027(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_AGENT_TARGET] */
extern T0* GE_new1028(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ALIAS_NAME] */
extern T0* GE_new1029(T1 initialize);
/* New instance of type detachable ET_ALIAS_NAME */
extern T0* GE_new1030(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ALIAS_NAME] */
extern T0* GE_new1031(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ALIAS_NAME_LIST] */
extern T0* GE_new1032(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ALIAS_NAME_LIST] */
extern T0* GE_new1033(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ASSIGNER] */
extern T0* GE_new1034(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ASSIGNER] */
extern T0* GE_new1036(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_EXPRESSION] */
extern T0* GE_new1037(T1 initialize);
/* New instance of type detachable ET_BRACKET_EXPRESSION */
extern T0* GE_new1038(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_BRACKET_EXPRESSION] */
extern T0* GE_new1039(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CALL_AGENT] */
extern T0* GE_new1040(T1 initialize);
/* New instance of type detachable ET_CALL_AGENT */
extern T0* GE_new1041(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CALL_AGENT] */
extern T0* GE_new1042(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_CALL_EXPRESSION] */
extern T0* GE_new1043(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_QUALIFIED_CALL_EXPRESSION] */
extern T0* GE_new1044(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHECK_INSTRUCTION] */
extern T0* GE_new1045(T1 initialize);
/* New instance of type detachable ET_CHECK_INSTRUCTION */
extern T0* GE_new1046(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CHECK_INSTRUCTION] */
extern T0* GE_new1047(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE] */
extern T0* GE_new1048(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CHOICE] */
extern T0* GE_new1050(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_CONSTANT] */
extern T0* GE_new1051(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CHOICE_CONSTANT] */
extern T0* GE_new1053(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_ITEM] */
extern T0* GE_new1054(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CHOICE_ITEM] */
extern T0* GE_new1056(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_LIST] */
extern T0* GE_new1057(T1 initialize);
/* New instance of type detachable ET_CHOICE_LIST */
extern T0* GE_new1058(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CHOICE_LIST] */
extern T0* GE_new1059(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLASS] */
extern T0* GE_new1060(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CLASS] */
extern T0* GE_new1061(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENT_ITEM] */
extern T0* GE_new1062(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CLIENT_ITEM] */
extern T0* GE_new1063(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENTS] */
extern T0* GE_new1064(T1 initialize);
/* New instance of type detachable ET_CLIENTS */
extern T0* GE_new1065(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CLIENTS] */
extern T0* GE_new1066(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_COMPOUND] */
extern T0* GE_new1067(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_COMPOUND] */
extern T0* GE_new1068(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONDITIONAL] */
extern T0* GE_new1069(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CONDITIONAL] */
extern T0* GE_new1071(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTANT] */
extern T0* GE_new1072(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CONSTANT] */
extern T0* GE_new1073(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
extern T0* GE_new1074(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
extern T0* GE_new1076(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST] */
extern T0* GE_new1077(T1 initialize);
/* New instance of type detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST */
extern T0* GE_new1078(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST] */
extern T0* GE_new1079(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_CREATOR] */
extern T0* GE_new1080(T1 initialize);
/* New instance of type detachable ET_CONSTRAINT_CREATOR */
extern T0* GE_new1081(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CONSTRAINT_CREATOR] */
extern T0* GE_new1082(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_RENAME_LIST] */
extern T0* GE_new1083(T1 initialize);
/* New instance of type detachable ET_CONSTRAINT_RENAME_LIST */
extern T0* GE_new1084(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CONSTRAINT_RENAME_LIST] */
extern T0* GE_new1085(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_TYPE] */
extern T0* GE_new1086(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CONSTRAINT_TYPE] */
extern T0* GE_new1088(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE] */
extern T0* GE_new1089(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CONVERT_FEATURE] */
extern T0* GE_new1091(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_ITEM] */
extern T0* GE_new1092(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_ITEM] */
extern T0* GE_new1094(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_LIST] */
extern T0* GE_new1095(T1 initialize);
/* New instance of type detachable ET_CONVERT_FEATURE_LIST */
extern T0* GE_new1096(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_LIST] */
extern T0* GE_new1097(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATE_EXPRESSION] */
extern T0* GE_new1098(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CREATE_EXPRESSION] */
extern T0* GE_new1099(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATION_REGION] */
extern T0* GE_new1100(T1 initialize);
/* New instance of type detachable ET_CREATION_REGION */
extern T0* GE_new1101(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CREATION_REGION] */
extern T0* GE_new1102(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR] */
extern T0* GE_new1103(T1 initialize);
/* New instance of type detachable ET_CREATOR */
extern T0* GE_new1104(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CREATOR] */
extern T0* GE_new1105(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR_LIST] */
extern T0* GE_new1106(T1 initialize);
/* New instance of type detachable ET_CREATOR_LIST */
extern T0* GE_new1107(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_CREATOR_LIST] */
extern T0* GE_new1108(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_DEBUG_INSTRUCTION] */
extern T0* GE_new1109(T1 initialize);
/* New instance of type detachable ET_DEBUG_INSTRUCTION */
extern T0* GE_new1110(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_DEBUG_INSTRUCTION] */
extern T0* GE_new1111(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION] */
extern T0* GE_new1112(T1 initialize);
/* New instance of type detachable ET_ELSEIF_EXPRESSION */
extern T0* GE_new1113(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION] */
extern T0* GE_new1114(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION_LIST] */
extern T0* GE_new1115(T1 initialize);
/* New instance of type detachable ET_ELSEIF_EXPRESSION_LIST */
extern T0* GE_new1116(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION_LIST] */
extern T0* GE_new1117(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART] */
extern T0* GE_new1118(T1 initialize);
/* New instance of type detachable ET_ELSEIF_PART */
extern T0* GE_new1119(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ELSEIF_PART] */
extern T0* GE_new1120(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART_LIST] */
extern T0* GE_new1121(T1 initialize);
/* New instance of type detachable ET_ELSEIF_PART_LIST */
extern T0* GE_new1122(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ELSEIF_PART_LIST] */
extern T0* GE_new1123(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT] */
extern T0* GE_new1124(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_EXPORT] */
extern T0* GE_new1126(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT_LIST] */
extern T0* GE_new1127(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_EXPORT_LIST] */
extern T0* GE_new1128(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION] */
extern T0* GE_new1129(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_EXPRESSION] */
extern T0* GE_new1130(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION_ITEM] */
extern T0* GE_new1131(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_EXPRESSION_ITEM] */
extern T0* GE_new1132(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTENDED_FEATURE_NAME] */
extern T0* GE_new1133(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_EXTENDED_FEATURE_NAME] */
extern T0* GE_new1134(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTERNAL_ALIAS] */
extern T0* GE_new1135(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_EXTERNAL_ALIAS] */
extern T0* GE_new1137(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE] */
extern T0* GE_new1138(T1 initialize);
/* New instance of type detachable ET_FEATURE_CLAUSE */
extern T0* GE_new1139(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE] */
extern T0* GE_new1140(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE_LIST] */
extern T0* GE_new1141(T1 initialize);
/* New instance of type detachable ET_FEATURE_CLAUSE_LIST */
extern T0* GE_new1142(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE_LIST] */
extern T0* GE_new1143(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_EXPORT] */
extern T0* GE_new1144(T1 initialize);
/* New instance of type detachable ET_FEATURE_EXPORT */
extern T0* GE_new1145(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FEATURE_EXPORT] */
extern T0* GE_new1146(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME] */
extern T0* GE_new1147(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FEATURE_NAME] */
extern T0* GE_new1148(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME_ITEM] */
extern T0* GE_new1149(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FEATURE_NAME_ITEM] */
extern T0* GE_new1151(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT] */
extern T0* GE_new1152(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT] */
extern T0* GE_new1153(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_ITEM] */
extern T0* GE_new1154(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_ITEM] */
extern T0* GE_new1156(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_LIST] */
extern T0* GE_new1157(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_LIST] */
extern T0* GE_new1158(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER] */
extern T0* GE_new1159(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER] */
extern T0* GE_new1160(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_ITEM] */
extern T0* GE_new1161(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_ITEM] */
extern T0* GE_new1163(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_LIST] */
extern T0* GE_new1164(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_LIST] */
extern T0* GE_new1165(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_EXPRESSION] */
extern T0* GE_new1166(T1 initialize);
/* New instance of type detachable ET_IF_EXPRESSION */
extern T0* GE_new1167(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_IF_EXPRESSION] */
extern T0* GE_new1168(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_INSTRUCTION] */
extern T0* GE_new1169(T1 initialize);
/* New instance of type detachable ET_IF_INSTRUCTION */
extern T0* GE_new1170(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_IF_INSTRUCTION] */
extern T0* GE_new1171(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_LIST] */
extern T0* GE_new1172(T1 initialize);
/* New instance of type detachable ET_INDEXING_LIST */
extern T0* GE_new1173(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_INDEXING_LIST] */
extern T0* GE_new1174(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_ITEM] */
extern T0* GE_new1175(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_INDEXING_ITEM] */
extern T0* GE_new1177(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM] */
extern T0* GE_new1178(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_INDEXING_TERM] */
extern T0* GE_new1179(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_ITEM] */
extern T0* GE_new1180(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_INDEXING_TERM_ITEM] */
extern T0* GE_new1182(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_LIST] */
extern T0* GE_new1183(T1 initialize);
/* New instance of type detachable ET_INDEXING_TERM_LIST */
extern T0* GE_new1184(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_INDEXING_TERM_LIST] */
extern T0* GE_new1185(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INLINE_AGENT] */
extern T0* GE_new1186(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_INLINE_AGENT] */
extern T0* GE_new1187(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSPECT_EXPRESSION] */
extern T0* GE_new1188(T1 initialize);
/* New instance of type detachable ET_INSPECT_EXPRESSION */
extern T0* GE_new1189(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_INSPECT_EXPRESSION] */
extern T0* GE_new1190(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSPECT_INSTRUCTION] */
extern T0* GE_new1191(T1 initialize);
/* New instance of type detachable ET_INSPECT_INSTRUCTION */
extern T0* GE_new1192(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_INSPECT_INSTRUCTION] */
extern T0* GE_new1193(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSTRUCTION] */
extern T0* GE_new1194(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_INSTRUCTION] */
extern T0* GE_new1196(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_INVARIANTS] */
extern T0* GE_new1197(T1 initialize);
/* New instance of type detachable ET_INVARIANTS */
extern T0* GE_new1198(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_INVARIANTS] */
extern T0* GE_new1199(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_FEATURE_NAME_LIST] */
extern T0* GE_new1200(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_KEYWORD_FEATURE_NAME_LIST] */
extern T0* GE_new1201(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_LIKE_TYPE] */
extern T0* GE_new1202(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_LIKE_TYPE] */
extern T0* GE_new1204(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE] */
extern T0* GE_new1205(T1 initialize);
/* New instance of type detachable ET_LOCAL_VARIABLE */
extern T0* GE_new1206(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE] */
extern T0* GE_new1207(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_ITEM] */
extern T0* GE_new1208(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_ITEM] */
extern T0* GE_new1210(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_LIST] */
extern T0* GE_new1211(T1 initialize);
/* New instance of type detachable ET_LOCAL_VARIABLE_LIST */
extern T0* GE_new1212(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_LIST] */
extern T0* GE_new1213(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOOP_INVARIANTS] */
extern T0* GE_new1214(T1 initialize);
/* New instance of type detachable ET_LOOP_INVARIANTS */
extern T0* GE_new1215(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_LOOP_INVARIANTS] */
extern T0* GE_new1216(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_ARRAY] */
extern T0* GE_new1217(T1 initialize);
/* New instance of type detachable ET_MANIFEST_ARRAY */
extern T0* GE_new1218(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_MANIFEST_ARRAY] */
extern T0* GE_new1219(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_ITEM] */
extern T0* GE_new1220(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_MANIFEST_STRING_ITEM] */
extern T0* GE_new1222(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_LIST] */
extern T0* GE_new1223(T1 initialize);
/* New instance of type detachable ET_MANIFEST_STRING_LIST */
extern T0* GE_new1224(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_MANIFEST_STRING_LIST] */
extern T0* GE_new1225(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_TUPLE] */
extern T0* GE_new1226(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_MANIFEST_TUPLE] */
extern T0* GE_new1227(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBSOLETE] */
extern T0* GE_new1228(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_OBSOLETE] */
extern T0* GE_new1230(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENTHESIZED_EXPRESSION] */
extern T0* GE_new1231(T1 initialize);
/* New instance of type detachable ET_PARENTHESIZED_EXPRESSION */
extern T0* GE_new1232(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_PARENTHESIZED_EXPRESSION] */
extern T0* GE_new1233(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT] */
extern T0* GE_new1234(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_PARENT] */
extern T0* GE_new1235(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_CLAUSE_LIST] */
extern T0* GE_new1236(T1 initialize);
/* New instance of type detachable ET_PARENT_CLAUSE_LIST */
extern T0* GE_new1237(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_PARENT_CLAUSE_LIST] */
extern T0* GE_new1238(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_ITEM] */
extern T0* GE_new1239(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_PARENT_ITEM] */
extern T0* GE_new1240(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_LIST] */
extern T0* GE_new1241(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_PARENT_LIST] */
extern T0* GE_new1242(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSTCONDITIONS] */
extern T0* GE_new1243(T1 initialize);
/* New instance of type detachable ET_POSTCONDITIONS */
extern T0* GE_new1244(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_POSTCONDITIONS] */
extern T0* GE_new1245(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECONDITIONS] */
extern T0* GE_new1246(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_PRECONDITIONS] */
extern T0* GE_new1247(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_PROCEDURE] */
extern T0* GE_new1248(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_PROCEDURE] */
extern T0* GE_new1249(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_LIKE_IDENTIFIER] */
extern T0* GE_new1250(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_QUALIFIED_LIKE_IDENTIFIER] */
extern T0* GE_new1252(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUANTIFIER_EXPRESSION] */
extern T0* GE_new1253(T1 initialize);
/* New instance of type detachable ET_QUANTIFIER_EXPRESSION */
extern T0* GE_new1254(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_QUANTIFIER_EXPRESSION] */
extern T0* GE_new1255(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUERY] */
extern T0* GE_new1256(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_QUERY] */
extern T0* GE_new1257(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_ITEM] */
extern T0* GE_new1258(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_RENAME_ITEM] */
extern T0* GE_new1260(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_LIST] */
extern T0* GE_new1261(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_RENAME_LIST] */
extern T0* GE_new1262(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_REPEAT_INSTRUCTION] */
extern T0* GE_new1263(T1 initialize);
/* New instance of type detachable ET_REPEAT_INSTRUCTION */
extern T0* GE_new1264(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_REPEAT_INSTRUCTION] */
extern T0* GE_new1265(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_STATIC_CALL_EXPRESSION] */
extern T0* GE_new1266(T1 initialize);
/* New instance of type detachable ET_STATIC_CALL_EXPRESSION */
extern T0* GE_new1267(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_STATIC_CALL_EXPRESSION] */
extern T0* GE_new1268(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_STRIP_EXPRESSION] */
extern T0* GE_new1269(T1 initialize);
/* New instance of type detachable ET_STRIP_EXPRESSION */
extern T0* GE_new1270(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_STRIP_EXPRESSION] */
extern T0* GE_new1271(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE] */
extern T0* GE_new1272(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_TYPE] */
extern T0* GE_new1273(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT] */
extern T0* GE_new1274(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT] */
extern T0* GE_new1276(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_ITEM] */
extern T0* GE_new1277(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_ITEM] */
extern T0* GE_new1279(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_LIST] */
extern T0* GE_new1280(T1 initialize);
/* New instance of type detachable ET_TYPE_CONSTRAINT_LIST */
extern T0* GE_new1281(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_LIST] */
extern T0* GE_new1282(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_ITEM] */
extern T0* GE_new1283(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_TYPE_ITEM] */
extern T0* GE_new1285(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_LIST] */
extern T0* GE_new1286(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_TYPE_LIST] */
extern T0* GE_new1288(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_VARIANT] */
extern T0* GE_new1289(T1 initialize);
/* New instance of type detachable ET_VARIANT */
extern T0* GE_new1290(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_VARIANT] */
extern T0* GE_new1291(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_EXPRESSION] */
extern T0* GE_new1292(T1 initialize);
/* New instance of type detachable ET_WHEN_EXPRESSION */
extern T0* GE_new1293(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_WHEN_EXPRESSION] */
extern T0* GE_new1294(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_EXPRESSION_LIST] */
extern T0* GE_new1295(T1 initialize);
/* New instance of type detachable ET_WHEN_EXPRESSION_LIST */
extern T0* GE_new1296(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_WHEN_EXPRESSION_LIST] */
extern T0* GE_new1297(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART] */
extern T0* GE_new1298(T1 initialize);
/* New instance of type detachable ET_WHEN_PART */
extern T0* GE_new1299(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_WHEN_PART] */
extern T0* GE_new1300(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART_LIST] */
extern T0* GE_new1301(T1 initialize);
/* New instance of type detachable ET_WHEN_PART_LIST */
extern T0* GE_new1302(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_WHEN_PART_LIST] */
extern T0* GE_new1303(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_WRITABLE] */
extern T0* GE_new1304(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_WRITABLE] */
extern T0* GE_new1305(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS] */
extern T0* GE_new1306(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS] */
extern T0* GE_new1307(T1 initialize);
/* New instance of type [detachable] ET_PARENT_CHECKER1 */
extern T0* GE_new1308(T1 initialize);
/* New instance of type [detachable] ET_FORMAL_PARAMETER_CHECKER1 */
extern T0* GE_new1309(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME] */
extern T0* GE_new1310(T1 initialize);
/* New instance of type [detachable] ET_FEATURE_NAME_TESTER */
extern T0* GE_new1311(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME] */
extern T0* GE_new1313(T1 initialize);
/* New instance of type [detachable] ET_ALIAS_NAME_TESTER */
extern T0* GE_new1314(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST] */
extern T0* GE_new1316(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS] */
extern T0* GE_new1317(T1 initialize);
/* New instance of type [detachable] ET_FEATURE_ADAPTATION_RESOLVER */
extern T0* GE_new1318(T1 initialize);
/* New instance of type [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER */
extern T0* GE_new1319(T1 initialize);
/* New instance of type [detachable] ET_IDENTIFIER_TYPE_RESOLVER */
extern T0* GE_new1320(T1 initialize);
/* New instance of type [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1 */
extern T0* GE_new1321(T1 initialize);
/* New instance of type [detachable] ET_ANCHORED_TYPE_CHECKER */
extern T0* GE_new1322(T1 initialize);
/* New instance of type [detachable] ET_SIGNATURE_CHECKER */
extern T0* GE_new1323(T1 initialize);
/* New instance of type [detachable] ET_PARENT_CHECKER2 */
extern T0* GE_new1324(T1 initialize);
/* New instance of type [detachable] ET_FORMAL_PARAMETER_CHECKER2 */
extern T0* GE_new1325(T1 initialize);
/* New instance of type [detachable] ET_BUILTIN_FEATURE_CHECKER */
extern T0* GE_new1326(T1 initialize);
/* New instance of type [detachable] ET_PRECURSOR_CHECKER */
extern T0* GE_new1327(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32] */
extern T0* GE_new1328(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_CLASS] */
extern T0* GE_new1329(T1 initialize);
/* New instance of type [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER */
extern T0* GE_new1330(T1 initialize);
/* New instance of type [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2 */
extern T0* GE_new1331(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME] */
extern T0* GE_new1332(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME] */
extern T0* GE_new1333(T1 initialize);
/* New instance of type [detachable] ET_RENAME */
extern T0* GE_new1337(T1 initialize);
/* New instance of type [detachable] ET_PARENT_CHECKER3 */
extern T0* GE_new1338(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_PROCEDURE] */
extern T0* GE_new1339(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_QUERY] */
extern T0* GE_new1340(T1 initialize);
/* New instance of type [detachable] ET_SUPPLIER_BUILDER */
extern T0* GE_new1341(T1 initialize);
/* New instance of type detachable ET_PARENT_FEATURE */
extern T0* GE_new1344(T1 initialize);
/* New instance of type [detachable] ET_INHERITED_FEATURE */
extern T0* GE_new1345(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CLASS] */
extern T0* GE_new1346(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS] */
extern T0* GE_new1347(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS] */
extern T0* GE_new1348(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [[attached] YY_BUFFER] */
extern T0* GE_new1349(T1 initialize);
/* New instance of type [detachable] UT_SYNTAX_ERROR */
extern T0* GE_new1350(T1 initialize);
/* New instance of type [detachable] UT_TOO_MANY_INCLUDES_ERROR */
extern T0* GE_new1351(T1 initialize);
/* New instance of type detachable ET_DYNAMIC_TARGET_LIST */
extern T0* GE_new1353(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] NATIVE_STRING] */
extern T0* GE_new1354(T6 a1, T1 initialize);
/* New instance of type [detachable] ARRAY [detachable STRING_8] */
extern T0* GE_new1355(T1 initialize);
/* New instance of type [detachable] XM_EIFFEL_CHARACTER_ENTITY */
extern T0* GE_new1356(T1 initialize);
/* New instance of type detachable DS_BILINKABLE [[attached] XM_POSITION] */
extern T0* GE_new1358(T1 initialize);
/* New instance of type [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_POSITION] */
extern T0* GE_new1359(T1 initialize);
/* New instance of type detachable DS_LINKABLE [[attached] XM_EIFFEL_SCANNER] */
extern T0* GE_new1360(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] XM_EIFFEL_ENTITY_DEF] */
extern T0* GE_new1361(T6 a1, T1 initialize);
/* New instance of type detachable DS_HASH_TABLE_CURSOR [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8] */
extern T0* GE_new1364(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_ENTITY_DEF] */
extern T0* GE_new1365(T1 initialize);
/* New instance of type detachable KL_EQUALITY_TESTER [[attached] XM_EIFFEL_PARSER_NAME] */
extern T0* GE_new1368(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] XM_EIFFEL_PARSER_NAME] */
extern T0* GE_new1369(T1 initialize);
/* New instance of type [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ELEMENT_CONTENT] */
extern T0* GE_new1370(T1 initialize);
/* New instance of type [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
extern T0* GE_new1371(T1 initialize);
/* New instance of type detachable DS_BILINKABLE [[attached] XM_DTD_ATTRIBUTE_CONTENT] */
extern T0* GE_new1372(T1 initialize);
/* New instance of type [detachable] DS_BILINKED_LIST_CURSOR [[attached] STRING_8] */
extern T0* GE_new1373(T1 initialize);
/* New instance of type detachable DS_BILINKABLE [[attached] STRING_8] */
extern T0* GE_new1374(T1 initialize);
/* New instance of type [detachable] XM_NAMESPACE_RESOLVER_CONTEXT */
extern T0* GE_new1375(T1 initialize);
/* New instance of type [detachable] DS_LINKED_QUEUE [detachable STRING_8] */
extern T0* GE_new1378(T1 initialize);
/* New instance of type [detachable] DS_LINKED_QUEUE [[attached] STRING_8] */
extern T0* GE_new1379(T1 initialize);
/* New instance of type detachable KL_EQUALITY_TESTER [[attached] XM_NAMESPACE] */
extern T0* GE_new1380(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] XM_NAMESPACE] */
extern T0* GE_new1381(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] XM_NAMESPACE] */
extern T0* GE_new1382(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] XM_NAMESPACE] */
extern T0* GE_new1383(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_CLUSTER] */
extern T0* GE_new1384(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_PARENT_ITEM] */
extern T0* GE_new1386(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_ITEM] */
extern T0* GE_new1387(T1 initialize);
/* New instance of type [detachable] ET_BRACED_TYPE_LIST */
extern T0* GE_new1388(T1 initialize);
/* New instance of type detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS] */
extern T0* GE_new1389(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ADAPTED_LIBRARY] */
extern T0* GE_new1390(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_LIBRARY] */
extern T0* GE_new1391(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_LIBRARY] */
extern T0* GE_new1392(T1 initialize);
/* New instance of type detachable DS_LINKABLE [[attached] XM_DOCUMENT_NODE] */
extern T0* GE_new1393(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
extern T0* GE_new1395(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
extern T0* GE_new1396(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_DOTNET_ASSEMBLY] */
extern T0* GE_new1397(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY] */
extern T0* GE_new1398(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32] */
extern T0* GE_new1400(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL] */
extern T0* GE_new1402(T6 a1, T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CALL_NAME] */
extern T0* GE_new1404(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL] */
extern T0* GE_new1406(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CALL_NAME] */
extern T0* GE_new1407(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME] */
extern T0* GE_new1408(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME] */
extern T0* GE_new1410(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL] */
extern T0* GE_new1411(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL] */
extern T0* GE_new1412(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new1413(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]] */
extern T0* GE_new1414(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable DS_LINKABLE [[attached] INTEGER_32]] */
extern T0* GE_new1415(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [detachable DS_LINKABLE [[attached] INTEGER_32]] */
extern T0* GE_new1416(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_SYSTEM_PROCESSOR] */
extern T0* GE_new1418(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_SYSTEM_PROCESSOR] */
extern T0* GE_new1419(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_SYSTEM_PROCESSOR] */
extern T0* GE_new1420(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_CLASS] */
extern T0* GE_new1421(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ADAPTED_CLASS] */
extern T0* GE_new1423(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_CLASS] */
extern T0* GE_new1424(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PROCEDURE] */
extern T0* GE_new1425(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUERY] */
extern T0* GE_new1426(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NAMED_OBJECT_TEST] */
extern T0* GE_new1427(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ITERATION_COMPONENT] */
extern T0* GE_new1428(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32] */
extern T0* GE_new1429(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_8] */
extern T0* GE_new1433(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_8, [attached] ET_CLASS_NAME] */
extern T0* GE_new1434(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] RX_CHARACTER_SET] */
extern T0* GE_new1435(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] RX_CHARACTER_SET] */
extern T0* GE_new1436(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_64] */
extern T0* GE_new1443(T1 initialize);
/* New instance of type [detachable] KL_CHARACTER_BUFFER */
extern T0* GE_new1446(T1 initialize);
/* New instance of type [detachable] UT_TRISTATE */
extern T0* GE_new1448(T1 initialize);
/* New instance of type [detachable] KL_AGENT_ROUTINES [[attached] ANY] */
extern T0* GE_new1449(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] UT_TRISTATE] */
extern T0* GE_new1450(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] KL_AGENT_ROUTINES [[attached] ANY], [attached] PROCEDURE [[attached] TUPLE]] */
extern T0* GE_new1451(T1 initialize);
/* New instance of type [detachable] PREDICATE [[attached] TUPLE [[attached] ET_MASTER_CLASS]] */
extern T0* GE_new1452(T1 initialize);
/* New instance of type [detachable] PREDICATE [[attached] TUPLE] */
extern T0* GE_new1453(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE] */
extern T0* GE_new1455(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE] */
extern T0* GE_new1456(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_SYSTEM, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
extern T0* GE_new1457(T1 initialize);
/* New instance of type [detachable] KL_AGENT_ROUTINES [[attached] ET_UNIVERSE] */
extern T0* GE_new1458(T1 initialize);
/* New instance of type [detachable] PREDICATE [[attached] TUPLE [[attached] ET_UNIVERSE]] */
extern T0* GE_new1459(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE]] */
extern T0* GE_new1460(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] KL_AGENT_ROUTINES [[attached] ET_UNIVERSE], [attached] PREDICATE [[attached] TUPLE [[attached] ET_UNIVERSE]]] */
extern T0* GE_new1461(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]], [attached] FUNCTION [[attached] TUPLE, [attached] BOOLEAN]] */
extern T0* GE_new1463(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]], [attached] FUNCTION [[attached] TUPLE [[attached] ET_MASTER_CLASS], [attached] BOOLEAN]] */
extern T0* GE_new1464(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]], [attached] FUNCTION [[attached] TUPLE [[attached] ET_MASTER_CLASS], [attached] BOOLEAN], [attached] FUNCTION [[attached] TUPLE, [attached] BOOLEAN]] */
extern T0* GE_new1465(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET */
extern T0* GE_new1466(T1 initialize);
/* New instance of type [detachable] XM_COMMENT */
extern T0* GE_new1469(T1 initialize);
/* New instance of type [detachable] XM_PROCESSING_INSTRUCTION */
extern T0* GE_new1470(T1 initialize);
/* New instance of type [detachable] XM_CHARACTER_DATA */
extern T0* GE_new1471(T1 initialize);
/* New instance of type [detachable] ET_LIKE_N */
extern T0* GE_new1474(T1 initialize);
/* New instance of type [detachable] KL_STRING_INPUT_STREAM */
extern T0* GE_new1477(T1 initialize);
/* New instance of type [detachable] KL_UNICODE_CHARACTER_BUFFER */
extern T0* GE_new1479(T1 initialize);
/* New instance of type detachable ET_OBJECT_TEST_LIST */
extern T0* GE_new1481(T1 initialize);
/* New instance of type detachable ET_C3_CHARACTER_CONSTANT */
extern T0* GE_new1482(T1 initialize);
/* New instance of type detachable ET_REGULAR_MANIFEST_STRING */
extern T0* GE_new1483(T1 initialize);
/* New instance of type detachable ET_SPECIAL_MANIFEST_STRING */
extern T0* GE_new1484(T1 initialize);
/* New instance of type detachable ET_VERBATIM_STRING */
extern T0* GE_new1485(T1 initialize);
/* New instance of type detachable ET_UNDERSCORED_INTEGER_CONSTANT */
extern T0* GE_new1486(T1 initialize);
/* New instance of type detachable ET_HEXADECIMAL_INTEGER_CONSTANT */
extern T0* GE_new1487(T1 initialize);
/* New instance of type detachable ET_OCTAL_INTEGER_CONSTANT */
extern T0* GE_new1488(T1 initialize);
/* New instance of type detachable ET_BINARY_INTEGER_CONSTANT */
extern T0* GE_new1489(T1 initialize);
/* New instance of type detachable ET_REGULAR_REAL_CONSTANT */
extern T0* GE_new1490(T1 initialize);
/* New instance of type detachable ET_UNDERSCORED_REAL_CONSTANT */
extern T0* GE_new1491(T1 initialize);
/* New instance of type detachable ET_TRUE_CONSTANT */
extern T0* GE_new1493(T1 initialize);
/* New instance of type detachable ET_FALSE_CONSTANT */
extern T0* GE_new1494(T1 initialize);
/* New instance of type detachable ET_C1_CHARACTER_CONSTANT */
extern T0* GE_new1495(T1 initialize);
/* New instance of type detachable ET_C2_CHARACTER_CONSTANT */
extern T0* GE_new1496(T1 initialize);
/* New instance of type detachable ET_TAGGED_INDEXING */
extern T0* GE_new1498(T1 initialize);
/* New instance of type detachable ET_INDEXING */
extern T0* GE_new1499(T1 initialize);
/* New instance of type detachable ET_CUSTOM_ATTRIBUTE */
extern T0* GE_new1500(T1 initialize);
/* New instance of type detachable ET_CONSTRAINED_FORMAL_PARAMETER */
extern T0* GE_new1501(T1 initialize);
/* New instance of type detachable ET_TYPE_RENAME_CONSTRAINT */
extern T0* GE_new1502(T1 initialize);
/* New instance of type detachable ET_CONSTRAINT_NAMED_TYPE */
extern T0* GE_new1503(T1 initialize);
/* New instance of type detachable ET_CONSTRAINT_LABELED_ACTUAL_PARAMETER */
extern T0* GE_new1504(T1 initialize);
/* New instance of type detachable ET_CONSTRAINT_LABELED_COMMA_ACTUAL_PARAMETER */
extern T0* GE_new1505(T1 initialize);
/* New instance of type detachable ET_ALL_EXPORT */
extern T0* GE_new1506(T1 initialize);
/* New instance of type detachable ET_CONVERT_FUNCTION */
extern T0* GE_new1508(T1 initialize);
/* New instance of type detachable ET_CONVERT_PROCEDURE */
extern T0* GE_new1509(T1 initialize);
/* New instance of type detachable ET_DO_FUNCTION */
extern T0* GE_new1511(T1 initialize);
/* New instance of type detachable ET_ONCE_FUNCTION */
extern T0* GE_new1512(T1 initialize);
/* New instance of type detachable ET_DEFERRED_FUNCTION */
extern T0* GE_new1513(T1 initialize);
/* New instance of type detachable ET_ONCE_PROCEDURE */
extern T0* GE_new1515(T1 initialize);
/* New instance of type detachable ET_EXTERNAL_PROCEDURE */
extern T0* GE_new1516(T1 initialize);
/* New instance of type detachable ET_ALIASED_FEATURE_NAME */
extern T0* GE_new1517(T1 initialize);
/* New instance of type detachable ET_ALIAS_FREE_NAME */
extern T0* GE_new1518(T1 initialize);
/* New instance of type detachable ET_CLASS_ASSERTION */
extern T0* GE_new1521(T1 initialize);
/* New instance of type detachable ET_LABELED_ACTUAL_PARAMETER */
extern T0* GE_new1522(T1 initialize);
/* New instance of type detachable ET_LIKE_FEATURE */
extern T0* GE_new1524(T1 initialize);
/* New instance of type detachable ET_QUALIFIED_LIKE_BRACED_TYPE */
extern T0* GE_new1525(T1 initialize);
/* New instance of type detachable ET_QUALIFIED_LIKE_TYPE */
extern T0* GE_new1526(T1 initialize);
/* New instance of type detachable ET_ASSIGNER_INSTRUCTION */
extern T0* GE_new1527(T1 initialize);
/* New instance of type detachable ET_ASSIGNMENT */
extern T0* GE_new1528(T1 initialize);
/* New instance of type detachable ET_ASSIGNMENT_ATTEMPT */
extern T0* GE_new1529(T1 initialize);
/* New instance of type detachable ET_LOOP_INSTRUCTION */
extern T0* GE_new1530(T1 initialize);
/* New instance of type detachable ET_BANG_INSTRUCTION */
extern T0* GE_new1532(T1 initialize);
/* New instance of type detachable ET_CREATE_INSTRUCTION */
extern T0* GE_new1533(T1 initialize);
/* New instance of type detachable ET_CHOICE_RANGE */
extern T0* GE_new1534(T1 initialize);
/* New instance of type detachable ET_PRECURSOR_INSTRUCTION */
extern T0* GE_new1535(T1 initialize);
/* New instance of type detachable ET_STATIC_CALL_INSTRUCTION */
extern T0* GE_new1537(T1 initialize);
/* New instance of type detachable ET_PRECURSOR_EXPRESSION */
extern T0* GE_new1538(T1 initialize);
/* New instance of type detachable ET_FEATURE_ADDRESS */
extern T0* GE_new1539(T1 initialize);
/* New instance of type detachable ET_CURRENT_ADDRESS */
extern T0* GE_new1540(T1 initialize);
/* New instance of type detachable ET_RESULT_ADDRESS */
extern T0* GE_new1541(T1 initialize);
/* New instance of type detachable ET_EXPRESSION_ADDRESS */
extern T0* GE_new1542(T1 initialize);
/* New instance of type detachable ET_INFIX_EXPRESSION */
extern T0* GE_new1543(T1 initialize);
/* New instance of type detachable ET_INFIX_AND_THEN_OPERATOR */
extern T0* GE_new1544(T1 initialize);
/* New instance of type detachable ET_INFIX_OR_ELSE_OPERATOR */
extern T0* GE_new1545(T1 initialize);
/* New instance of type detachable ET_MANIFEST_TYPE */
extern T0* GE_new1546(T1 initialize);
/* New instance of type detachable ET_PREFIX_EXPRESSION */
extern T0* GE_new1547(T1 initialize);
/* New instance of type detachable ET_OLD_EXPRESSION */
extern T0* GE_new1548(T1 initialize);
/* New instance of type detachable ET_OLD_OBJECT_TEST */
extern T0* GE_new1549(T1 initialize);
/* New instance of type detachable ET_ONCE_MANIFEST_STRING */
extern T0* GE_new1550(T1 initialize);
/* New instance of type detachable ET_ITERATION_CURSOR */
extern T0* GE_new1551(T1 initialize);
/* New instance of type detachable ET_DO_FUNCTION_INLINE_AGENT */
extern T0* GE_new1552(T1 initialize);
/* New instance of type detachable ET_ONCE_FUNCTION_INLINE_AGENT */
extern T0* GE_new1553(T1 initialize);
/* New instance of type detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT */
extern T0* GE_new1554(T1 initialize);
/* New instance of type detachable ET_DO_PROCEDURE_INLINE_AGENT */
extern T0* GE_new1555(T1 initialize);
/* New instance of type detachable ET_ONCE_PROCEDURE_INLINE_AGENT */
extern T0* GE_new1556(T1 initialize);
/* New instance of type detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT */
extern T0* GE_new1557(T1 initialize);
/* New instance of type detachable ET_AGENT_OPEN_TARGET */
extern T0* GE_new1558(T1 initialize);
/* New instance of type detachable ET_AGENT_TYPED_OPEN_ARGUMENT */
extern T0* GE_new1559(T1 initialize);
/* New instance of type detachable ET_TAGGED_ASSERTION */
extern T0* GE_new1562(T1 initialize);
/* New instance of type [detachable] ET_FILE_POSITION */
extern T0* GE_new1564(T1 initialize);
/* New instance of type [detachable] DS_QUICK_SORTER [[attached] ET_QUERY] */
extern T0* GE_new1566(T1 initialize);
/* New instance of type [detachable] DS_QUICK_SORTER [[attached] ET_PROCEDURE] */
extern T0* GE_new1567(T1 initialize);
/* New instance of type [detachable] ET_REDECLARED_FEATURE */
extern T0* GE_new1569(T1 initialize);
/* New instance of type [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER */
extern T0* GE_new1571(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_MASTER_CLASS, [attached] ET_SYSTEM_PROCESSOR] */
extern T0* GE_new1573(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] NATURAL_16] */
extern T0* GE_new1576(T6 a1, T1 initialize);
/* New instance of type [detachable] ARRAY [[attached] NATURAL_16] */
extern T0* GE_new1577(T1 initialize);
/* New instance of type [detachable] ARRAY [[attached] NATURAL_32] */
extern T0* GE_new1578(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER] */
extern T0* GE_new1579(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY] */
extern T0* GE_new1580(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
extern T0* GE_new1581(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE] */
extern T0* GE_new1582(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG] */
extern T0* GE_new1583(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE] */
extern T0* GE_new1584(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY] */
extern T0* GE_new1585(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
extern T0* GE_new1586(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE] */
extern T0* GE_new1587(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT] */
extern T0* GE_new1588(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE] */
extern T0* GE_new1589(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ACTION] */
extern T0* GE_new1590(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_ACTION] */
extern T0* GE_new1591(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ACTION] */
extern T0* GE_new1592(T1 initialize);
/* New instance of type detachable DS_ARRAYED_LIST [[attached] RX_PCRE_REGULAR_EXPRESSION] */
extern T0* GE_new1593(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_OPTIONS] */
extern T0* GE_new1594(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_OPTIONS] */
extern T0* GE_new1597(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_OPTIONS, [attached] STRING_8] */
extern T0* GE_new1598(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_VISIBLE_CLASS] */
extern T0* GE_new1599(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_VISIBLE_CLASS] */
extern T0* GE_new1600(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_VISIBLE_CLASS] */
extern T0* GE_new1601(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS] */
extern T0* GE_new1602(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CONDITION_ITEM] */
extern T0* GE_new1603(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_DOTNET_ASSEMBLY] */
extern T0* GE_new1604(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_ASSEMBLY] */
extern T0* GE_new1605(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_DOTNET_ASSEMBLY] */
extern T0* GE_new1606(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_UNIVERSE] */
extern T0* GE_new1608(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_UNIVERSE] */
extern T0* GE_new1609(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_UNIVERSE] */
extern T0* GE_new1611(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_INTERNAL_UNIVERSE] */
extern T0* GE_new1612(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INTERNAL_UNIVERSE] */
extern T0* GE_new1614(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_INTERNAL_UNIVERSE] */
extern T0* GE_new1615(T6 a1, T1 initialize);
/* New instance of type [detachable] ET_CLASS_TYPE_STATUS_CHECKER1 */
extern T0* GE_new1616(T1 initialize);
/* New instance of type [detachable] ET_CLASS_TYPE_STATUS_CHECKER2 */
extern T0* GE_new1617(T1 initialize);
/* New instance of type [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER */
extern T0* GE_new1618(T1 initialize);
/* New instance of type [detachable] ET_CLASS_TYPE_STATUS_CHECKER3 */
extern T0* GE_new1619(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS] */
extern T0* GE_new1620(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_TYPE] */
extern T0* GE_new1621(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_TYPE] */
extern T0* GE_new1622(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_FORMAL_PARAMETER_TYPE] */
extern T0* GE_new1623(T1 initialize);
/* New instance of type [detachable] CONSOLE */
extern T0* GE_new1624(T1 initialize);
/* New instance of type [detachable] DS_LINKED_QUEUE [[attached] CHARACTER_8] */
extern T0* GE_new1626(T1 initialize);
/* New instance of type [detachable] UC_UTF16_ROUTINES */
extern T0* GE_new1627(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_LIBRARY] */
extern T0* GE_new1628(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIBRARY] */
extern T0* GE_new1629(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_LIBRARY] */
extern T0* GE_new1630(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_TARGET] */
extern T0* GE_new1631(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ADAPTED_UNIVERSE] */
extern T0* GE_new1632(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_UNIVERSE] */
extern T0* GE_new1633(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_UNIVERSE] */
extern T0* GE_new1634(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_TARGET_PARENT] */
extern T0* GE_new1635(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET_PARENT] */
extern T0* GE_new1636(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_TARGET_PARENT] */
extern T0* GE_new1637(T1 initialize);
/* New instance of type [detachable] YY_UNICODE_BUFFER */
extern T0* GE_new1638(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_OBJECT_TEST_LIST] */
extern T0* GE_new1639(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBJECT_TEST_LIST] */
extern T0* GE_new1640(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST_LIST] */
extern T0* GE_new1641(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_OBJECT_TEST_LIST] */
extern T0* GE_new1642(T6 a1, T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable ET_ITERATION_COMPONENT_LIST] */
extern T0* GE_new1643(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [detachable ET_ITERATION_COMPONENT_LIST] */
extern T0* GE_new1644(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ITERATION_COMPONENT_LIST] */
extern T0* GE_new1645(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ITERATION_COMPONENT_LIST] */
extern T0* GE_new1646(T6 a1, T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ASSERTION_ITEM] */
extern T0* GE_new1647(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTION_ITEM] */
extern T0* GE_new1648(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ASSERTION_ITEM] */
extern T0* GE_new1649(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_CONSTRAINT_TYPE] */
extern T0* GE_new1650(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_NAMED_CLASS] */
extern T0* GE_new1651(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_CLASS] */
extern T0* GE_new1653(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS] */
extern T0* GE_new1655(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_BASE_TYPE, [attached] ET_CLASS] */
extern T0* GE_new1656(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT] */
extern T0* GE_new1659(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32] */
extern T0* GE_new1660(T1 initialize);
/* New instance of type [detachable] ET_BASE_TYPE_CONSTRAINT_LIST */
extern T0* GE_new1661(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME] */
extern T0* GE_new1662(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_FLATTENED_FEATURE] */
extern T0* GE_new1663(T6 a1, T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_FEATURE_NAME] */
extern T0* GE_new1666(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FLATTENED_FEATURE] */
extern T0* GE_new1667(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME] */
extern T0* GE_new1668(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ALIAS_NAME] */
extern T0* GE_new1669(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ALIAS_NAME] */
extern T0* GE_new1671(T1 initialize);
/* New instance of type detachable DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME] */
extern T0* GE_new1672(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_LIST] */
extern T0* GE_new1673(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CLIENT_LIST] */
extern T0* GE_new1674(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLIENT_LIST] */
extern T0* GE_new1675(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS] */
extern T0* GE_new1677(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT] */
extern T0* GE_new1678(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CLIENT] */
extern T0* GE_new1679(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME] */
extern T0* GE_new1680(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME] */
extern T0* GE_new1681(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32] */
extern T0* GE_new1682(T1 initialize);
/* New instance of type [detachable] ET_REPLICATED_FEATURE */
extern T0* GE_new1684(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE] */
extern T0* GE_new1685(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE] */
extern T0* GE_new1686(T1 initialize);
/* New instance of type [detachable] ET_DOTNET_SIGNATURE_TESTER */
extern T0* GE_new1687(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE] */
extern T0* GE_new1688(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE] */
extern T0* GE_new1690(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE] */
extern T0* GE_new1692(T1 initialize);
/* New instance of type [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE] */
extern T0* GE_new1693(T1 initialize);
/* New instance of type detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE] */
extern T0* GE_new1694(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_PARENT_FEATURE] */
extern T0* GE_new1695(T1 initialize);
/* New instance of type detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE] */
extern T0* GE_new1696(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8] */
extern T0* GE_new1697(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE, [attached] INTEGER_32] */
extern T0* GE_new1698(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_CLASS] */
extern T0* GE_new1701(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_FEATURE_NAME] */
extern T0* GE_new1703(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_RENAME] */
extern T0* GE_new1704(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME] */
extern T0* GE_new1707(T1 initialize);
/* New instance of type detachable DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_ALIAS_NAME] */
extern T0* GE_new1711(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_PROCEDURE] */
extern T0* GE_new1712(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_QUERY] */
extern T0* GE_new1713(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] YY_BUFFER] */
extern T0* GE_new1714(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] YY_BUFFER] */
extern T0* GE_new1715(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_DYNAMIC_TARGET] */
extern T0* GE_new1716(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TARGET] */
extern T0* GE_new1717(T1 initialize);
/* New instance of type [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ELEMENT_CONTENT] */
extern T0* GE_new1719(T1 initialize);
/* New instance of type detachable DS_BILINKABLE [[attached] XM_DTD_ELEMENT_CONTENT] */
extern T0* GE_new1720(T1 initialize);
/* New instance of type [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]] */
extern T0* GE_new1722(T1 initialize);
/* New instance of type [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]] */
extern T0* GE_new1723(T1 initialize);
/* New instance of type detachable DS_LINKABLE [detachable STRING_8] */
extern T0* GE_new1724(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CLUSTER] */
extern T0* GE_new1725(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLUSTER] */
extern T0* GE_new1726(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLUSTER] */
extern T0* GE_new1727(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_TYPE_ITEM] */
extern T0* GE_new1728(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_ITEM] */
extern T0* GE_new1729(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_MASTER_CLASS] */
extern T0* GE_new1730(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DOTNET_ASSEMBLY] */
extern T0* GE_new1731(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_CLUSTER] */
extern T0* GE_new1732(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_CLUSTER]] */
extern T0* GE_new1733(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_TARGET, [attached] STRING_8] */
extern T0* GE_new1734(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_ADAPTED_LIBRARY] */
extern T0* GE_new1736(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_ADAPTED_LIBRARY]] */
extern T0* GE_new1737(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
extern T0* GE_new1738(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY]] */
extern T0* GE_new1739(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] STRING_8]] */
extern T0* GE_new1742(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8] */
extern T0* GE_new1743(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8] */
extern T0* GE_new1744(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8] */
extern T0* GE_new1745(T1 initialize);
/* New instance of type detachable DS_LINKABLE [[attached] XM_ELEMENT_NODE] */
extern T0* GE_new1750(T1 initialize);
/* New instance of type detachable ET_CONSTRAINT_GENERIC_NAMED_TYPE */
extern T0* GE_new1755(T1 initialize);
/* New instance of type detachable ET_UNQUALIFIED_CALL_INSTRUCTION */
extern T0* GE_new1756(T1 initialize);
/* New instance of type detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT */
extern T0* GE_new1758(T1 initialize);
/* New instance of type [detachable] ET_SEEDED_QUERY_COMPARATOR */
extern T0* GE_new1759(T1 initialize);
/* New instance of type [detachable] ET_SEEDED_PROCEDURE_COMPARATOR */
extern T0* GE_new1762(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_CLIENT_LIST] */
extern T0* GE_new1766(T1 initialize);
/* New instance of type [detachable] ET_FORMAL_COMMA_ARGUMENT */
extern T0* GE_new1767(T1 initialize);
/* New instance of type [detachable] ET_LOCAL_COMMA_VARIABLE */
extern T0* GE_new1768(T1 initialize);
/* New instance of type [detachable] ET_LABELED_COMMA_ACTUAL_PARAMETER */
extern T0* GE_new1769(T1 initialize);
/* New instance of type [detachable] ET_KEYWORD_EXPRESSION */
extern T0* GE_new1771(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_RENAME_ITEM] */
extern T0* GE_new1776(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME_ITEM] */
extern T0* GE_new1777(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_EXPORT] */
extern T0* GE_new1779(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPORT] */
extern T0* GE_new1780(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_FEATURE_NAME_ITEM] */
extern T0* GE_new1782(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME_ITEM] */
extern T0* GE_new1783(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_FORMAL_PARAMETER_ITEM] */
extern T0* GE_new1784(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_PARAMETER_ITEM] */
extern T0* GE_new1785(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_CFLAG] */
extern T0* GE_new1788(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_CFLAG]] */
extern T0* GE_new1789(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_TARGET, [attached] ET_ECF_STATE] */
extern T0* GE_new1790(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_INCLUDE] */
extern T0* GE_new1791(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_INCLUDE]] */
extern T0* GE_new1792(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_LIBRARY] */
extern T0* GE_new1793(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_LIBRARY]] */
extern T0* GE_new1794(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
extern T0* GE_new1795(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_LINKER_FLAG]] */
extern T0* GE_new1796(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_MAKE] */
extern T0* GE_new1797(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_MAKE]] */
extern T0* GE_new1798(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_OBJECT] */
extern T0* GE_new1799(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_OBJECT]] */
extern T0* GE_new1800(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_EXTERNAL_RESOURCE] */
extern T0* GE_new1801(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_RESOURCE]] */
extern T0* GE_new1802(T1 initialize);
/* New instance of type detachable DS_LINKABLE [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]] */
extern T0* GE_new1806(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM] */
extern T0* GE_new1809(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM] */
extern T0* GE_new1810(T1 initialize);
/* New instance of type [detachable] ET_AGENT_IMPLICIT_CURRENT_TARGET */
extern T0* GE_new1811(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CHOICE_ITEM] */
extern T0* GE_new1813(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CHOICE_ITEM] */
extern T0* GE_new1814(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
extern T0* GE_new1815(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM] */
extern T0* GE_new1817(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CONVERT_FEATURE_ITEM] */
extern T0* GE_new1819(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONVERT_FEATURE_ITEM] */
extern T0* GE_new1820(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_CREATOR] */
extern T0* GE_new1821(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CREATOR] */
extern T0* GE_new1822(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ELSEIF_EXPRESSION] */
extern T0* GE_new1823(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_EXPRESSION] */
extern T0* GE_new1824(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ELSEIF_PART] */
extern T0* GE_new1825(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_PART] */
extern T0* GE_new1826(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_FEATURE_CLAUSE] */
extern T0* GE_new1827(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_CLAUSE] */
extern T0* GE_new1828(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_INDEXING_ITEM] */
extern T0* GE_new1830(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_ITEM] */
extern T0* GE_new1831(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_INDEXING_TERM_ITEM] */
extern T0* GE_new1832(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM_ITEM] */
extern T0* GE_new1833(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_LOCAL_VARIABLE_ITEM] */
extern T0* GE_new1834(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LOCAL_VARIABLE_ITEM] */
extern T0* GE_new1836(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_MANIFEST_STRING_ITEM] */
extern T0* GE_new1838(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MANIFEST_STRING_ITEM] */
extern T0* GE_new1839(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_PARENT_LIST] */
extern T0* GE_new1840(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_LIST] */
extern T0* GE_new1841(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_TYPE_CONSTRAINT_ITEM] */
extern T0* GE_new1843(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_CONSTRAINT_ITEM] */
extern T0* GE_new1844(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_WHEN_EXPRESSION] */
extern T0* GE_new1845(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_WHEN_EXPRESSION] */
extern T0* GE_new1846(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_WHEN_PART] */
extern T0* GE_new1847(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_WHEN_PART] */
extern T0* GE_new1848(T1 initialize);
/* New instance of type [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT */
extern T0* GE_new1850(T1 initialize);
/* New instance of type detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS] */
extern T0* GE_new1853(T1 initialize);
/* New instance of type [detachable] ET_STANDARD_ONCE_KEYS */
extern T0* GE_new1857(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_CLUSTER] */
extern T0* GE_new1859(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CLUSTER] */
extern T0* GE_new1860(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_CLUSTER] */
extern T0* GE_new1861(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_LIBRARY] */
extern T0* GE_new1862(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_LIBRARY] */
extern T0* GE_new1863(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ADAPTED_LIBRARY] */
extern T0* GE_new1864(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
extern T0* GE_new1865(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
extern T0* GE_new1866(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY] */
extern T0* GE_new1867(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_FILE_RULE] */
extern T0* GE_new1868(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_FILE_RULE] */
extern T0* GE_new1869(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_FILE_RULE] */
extern T0* GE_new1870(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_CFLAG] */
extern T0* GE_new1871(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_CFLAG] */
extern T0* GE_new1872(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_CFLAG] */
extern T0* GE_new1873(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_INCLUDE] */
extern T0* GE_new1874(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_INCLUDE] */
extern T0* GE_new1875(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_INCLUDE] */
extern T0* GE_new1876(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LIBRARY] */
extern T0* GE_new1877(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LIBRARY] */
extern T0* GE_new1878(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_LIBRARY] */
extern T0* GE_new1879(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
extern T0* GE_new1880(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
extern T0* GE_new1881(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_LINKER_FLAG] */
extern T0* GE_new1882(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_MAKE] */
extern T0* GE_new1883(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_MAKE] */
extern T0* GE_new1884(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_MAKE] */
extern T0* GE_new1885(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_OBJECT] */
extern T0* GE_new1886(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_OBJECT] */
extern T0* GE_new1887(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_OBJECT] */
extern T0* GE_new1888(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_RESOURCE] */
extern T0* GE_new1889(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_RESOURCE] */
extern T0* GE_new1890(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_RESOURCE] */
extern T0* GE_new1891(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] RX_PCRE_REGULAR_EXPRESSION] */
extern T0* GE_new1892(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] RX_PCRE_REGULAR_EXPRESSION] */
extern T0* GE_new1893(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] RX_PCRE_REGULAR_EXPRESSION] */
extern T0* GE_new1894(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ANDED_CONDITIONS] */
extern T0* GE_new1895(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_ANDED_CONDITIONS] */
extern T0* GE_new1896(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ANDED_CONDITIONS] */
extern T0* GE_new1897(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CONDITION_ITEM] */
extern T0* GE_new1898(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_ECF_CONDITION_ITEM] */
extern T0* GE_new1899(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_CONDITION_ITEM] */
extern T0* GE_new1900(T1 initialize);
/* New instance of type detachable DS_LINKABLE [[attached] CHARACTER_8] */
extern T0* GE_new1902(T1 initialize);
/* New instance of type detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_CLASS] */
extern T0* GE_new1903(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_BASE_TYPE_CONSTRAINT] */
extern T0* GE_new1904(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE_CONSTRAINT] */
extern T0* GE_new1905(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_BASE_TYPE_CONSTRAINT] */
extern T0* GE_new1906(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] NATURAL_32] */
extern T0* GE_new1907(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_FEATURE_NAME] */
extern T0* GE_new1908(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] ET_FEATURE_NAME] */
extern T0* GE_new1909(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32] */
extern T0* GE_new1911(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_REPLICABLE_FEATURE] */
extern T0* GE_new1912(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_REPLICABLE_FEATURE] */
extern T0* GE_new1914(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_ADAPTED_FEATURE] */
extern T0* GE_new1915(T1 initialize);
/* New instance of type detachable DS_LINKABLE [[attached] ET_ADAPTED_FEATURE] */
extern T0* GE_new1916(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_FEATURE] */
extern T0* GE_new1917(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_PARENT_FEATURE] */
extern T0* GE_new1918(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE] */
extern T0* GE_new1919(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_DOTNET_FEATURE] */
extern T0* GE_new1920(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] ET_DOTNET_FEATURE] */
extern T0* GE_new1921(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_FEATURE] */
extern T0* GE_new1922(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]] */
extern T0* GE_new1923(T6 a1, T1 initialize);
/* New instance of type detachable DS_HASH_TABLE_CURSOR [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE] */
extern T0* GE_new1926(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]] */
extern T0* GE_new1927(T1 initialize);
/* New instance of type [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_DOTNET_FEATURE] */
extern T0* GE_new1928(T1 initialize);
/* New instance of type detachable DS_LINKABLE [[attached] ET_DOTNET_FEATURE] */
extern T0* GE_new1929(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE] */
extern T0* GE_new1930(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_LIKE_FEATURE] */
extern T0* GE_new1931(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIKE_FEATURE] */
extern T0* GE_new1932(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE] */
extern T0* GE_new1933(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8] */
extern T0* GE_new1935(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]] */
extern T0* GE_new1936(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME] */
extern T0* GE_new1937(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]] */
extern T0* GE_new1938(T6 a1, T1 initialize);
/* New instance of type detachable DS_BILINKABLE [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]] */
extern T0* GE_new1939(T1 initialize);
/* New instance of type [detachable] ET_CLUSTER_ERROR */
extern T0* GE_new1940(T1 initialize);
/* New instance of type [detachable] ET_SYNTAX_ERROR */
extern T0* GE_new1941(T1 initialize);
/* New instance of type [detachable] ET_DOTNET_ASSEMBLY_ERROR */
extern T0* GE_new1942(T1 initialize);
/* New instance of type [detachable] AP_OPTION_COMPARATOR */
extern T0* GE_new1943(T1 initialize);
/* New instance of type [detachable] DS_QUICK_SORTER [[attached] AP_OPTION] */
extern T0* GE_new1944(T1 initialize);
/* New instance of type [detachable] ST_SPLITTER */
extern T0* GE_new1953(T1 initialize);
/* New instance of type [detachable] KL_CHARACTER_ROUTINES */
extern T0* GE_new1954(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_C_GENERATOR, INTEGER_32, [attached] ET_CURRENT, [attached] ET_DYNAMIC_TUPLE_TYPE, BOOLEAN] */
extern T0* GE_new1955(T1 initialize);
/* New instance of type [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS */
extern T0* GE_new1956(T1 initialize);
/* New instance of type [detachable] ET_ACTUAL_PARAMETER_SUBLIST */
extern T0* GE_new1957(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_FORMAL_ARGUMENT_ITEM] */
extern T0* GE_new1960(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_ARGUMENT_ITEM] */
extern T0* GE_new1962(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_DYNAMIC_PRECURSOR] */
extern T0* GE_new1964(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRECURSOR] */
extern T0* GE_new1965(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_INSTRUCTION] */
extern T0* GE_new1966(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INSTRUCTION] */
extern T0* GE_new1968(T1 initialize);
/* New instance of type [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST */
extern T0* GE_new1977(T1 initialize);
/* New instance of type [detachable] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8] */
extern T0* GE_new1978(T1 initialize);
/* New instance of type [detachable] ET_INFIX_CAST_EXPRESSION */
extern T0* GE_new1980(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] STRING_8] */
extern T0* GE_new1984(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]] */
extern T0* GE_new1985(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]] */
extern T0* GE_new1988(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8] */
extern T0* GE_new1989(T1 initialize);
/* New instance of type [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS] */
extern T0* GE_new1990(T1 initialize);
/* New instance of type detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE] */
extern T0* GE_new1993(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]] */
extern T0* GE_new1994(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]] */
extern T0* GE_new1997(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME] */
extern T0* GE_new1998(T1 initialize);
/* New instance of type [detachable] ET_UNIVERSE_ERROR */
extern T0* GE_new1999(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_SYSTEM, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
extern T0* GE_new2000(T1 initialize);
/* New instance of type detachable ET_PARENTHESIS_INSTRUCTION */
extern T0* GE_new2002(T1 initialize);
/* New instance of type [detachable] ET_PARENTHESIS_SYMBOL */
extern T0* GE_new2020(T1 initialize);
/* New instance of type [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT_LIST */
extern T0* GE_new2026(T1 initialize);
/* New instance of type [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT */
extern T0* GE_new2027(T1 initialize);
/* New instance of type [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8] */
extern T0* GE_new2034(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8] */
extern T0* GE_new2040(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8] */
extern T0* GE_new2041(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_EQUALITY_EXPRESSION */
extern T0* GE_new2042(T1 initialize);
/* New instance of type [detachable] ET_DYNAMIC_OBJECT_EQUALITY_EXPRESSION */
extern T0* GE_new2043(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_LIBRARY] */
extern T0* GE_new2044(T1 initialize);
/* New instance of type [detachable] ET_CONVERT_BUILTIN_EXPRESSION */
extern T0* GE_new2046(T1 initialize);
/* New instance of type [detachable] ET_CONVERT_FROM_EXPRESSION */
extern T0* GE_new2047(T1 initialize);
/* New instance of type [detachable] ET_CONVERT_TO_EXPRESSION */
extern T0* GE_new2048(T1 initialize);
/* New instance of type detachable DS_CELL [detachable ET_CLASS] */
extern T0* GE_new2051(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] DS_CELL [detachable ET_CLASS]] */
extern T0* GE_new2052(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_EIFFEL_PREPARSER, [attached] DS_ARRAYED_LIST [[attached] STRING_8]] */
extern T0* GE_new2054(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_FEATURE] */
extern T0* GE_new2056(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_FILE_RULE] */
extern T0* GE_new2057(T1 initialize);
/* New instance of type [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_FILE_RULE]] */
extern T0* GE_new2058(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_FILE_RULES] */
extern T0* GE_new2059(T1 initialize);
/* New instance of type [detachable] PREDICATE [[attached] TUPLE [[attached] ET_ECF_FILE_RULE]] */
extern T0* GE_new2060(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_STATE] */
extern T0* GE_new2061(T1 initialize);
/* New instance of type [detachable] PREDICATE [[attached] TUPLE [[attached] STRING_8]] */
extern T0* GE_new2063(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] KL_STRING_ROUTINES, [attached] STRING_8] */
extern T0* GE_new2064(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] DS_HASH_SET [[attached] STRING_8]] */
extern T0* GE_new2066(T1 initialize);
/* New instance of type [detachable] LX_DFA_WILDCARD */
extern T0* GE_new2068(T1 initialize);
/* New instance of type detachable DS_LINKABLE [[attached] ET_PARENT_FEATURE] */
extern T0* GE_new2071(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] LX_WILDCARD] */
extern T0* GE_new2073(T1 initialize);
/* New instance of type [detachable] PREDICATE [[attached] TUPLE [[attached] LX_WILDCARD]] */
extern T0* GE_new2074(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_CLUSTER_DEPENDENCE_CONSTRAINT, [attached] ET_GROUP, [attached] STRING_8] */
extern T0* GE_new2075(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT] */
extern T0* GE_new2077(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT] */
extern T0* GE_new2078(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] LX_WILDCARD] */
extern T0* GE_new2079(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_WILDCARD] */
extern T0* GE_new2082(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_C_GENERATOR, [attached] ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE] */
extern T0* GE_new2086(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_LIBRARY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
extern T0* GE_new2089(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
extern T0* GE_new2090(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_DOTNET_ASSEMBLY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
extern T0* GE_new2091(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_LIBRARY, [attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE], INTEGER_32] */
extern T0* GE_new2092(T1 initialize);
/* New instance of type [detachable] ET_AST_NULL_LEAF */
extern T0* GE_new2094(T1 initialize);
/* New instance of type [detachable] LX_WILDCARD_PARSER */
extern T0* GE_new2095(T1 initialize);
/* New instance of type [detachable] LX_DESCRIPTION */
extern T0* GE_new2096(T1 initialize);
/* New instance of type [detachable] LX_FULL_DFA */
extern T0* GE_new2097(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] ET_GROUP] */
extern T0* GE_new2099(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]], [attached] PREDICATE [[attached] TUPLE [[attached] ET_CLASS]]] */
extern T0* GE_new2100(T1 initialize);
/* New instance of type [detachable] TUPLE [[attached] RX_PCRE_REGULAR_EXPRESSION] */
extern T0* GE_new2101(T1 initialize);
/* New instance of type [detachable] PREDICATE [[attached] TUPLE [[attached] RX_PCRE_REGULAR_EXPRESSION]] */
extern T0* GE_new2102(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] LX_RULE] */
extern T0* GE_new2105(T1 initialize);
/* New instance of type [detachable] LX_START_CONDITIONS */
extern T0* GE_new2106(T1 initialize);
/* New instance of type [detachable] LX_ACTION_FACTORY */
extern T0* GE_new2107(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS] */
extern T0* GE_new2108(T1 initialize);
/* New instance of type [detachable] LX_SYMBOL_CLASS */
extern T0* GE_new2109(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8] */
extern T0* GE_new2110(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS] */
extern T0* GE_new2111(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32] */
extern T0* GE_new2112(T1 initialize);
/* New instance of type [detachable] KL_EQUALITY_TESTER [[attached] STRING_32] */
extern T0* GE_new2113(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_STACK [[attached] BOOLEAN] */
extern T0* GE_new2114(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_32] */
extern T0* GE_new2115(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] STRING_32] */
extern T0* GE_new2116(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_SYMBOL_CLASS] */
extern T0* GE_new2117(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] LX_SYMBOL_CLASS] */
extern T0* GE_new2118(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA] */
extern T0* GE_new2119(T1 initialize);
/* New instance of type [detachable] LX_NFA */
extern T0* GE_new2120(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] LX_NFA] */
extern T0* GE_new2121(T6 a1, T1 initialize);
/* New instance of type detachable LX_RULE */
extern T0* GE_new2122(T1 initialize);
/* New instance of type [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS] */
extern T0* GE_new2123(T1 initialize);
/* New instance of type [detachable] LX_EQUIVALENCE_CLASSES */
extern T0* GE_new2124(T1 initialize);
/* New instance of type [detachable] LX_UNRECOGNIZED_RULE_ERROR */
extern T0* GE_new2125(T1 initialize);
/* New instance of type [detachable] LX_INVALID_UNICODE_CHARACTER_ERROR */
extern T0* GE_new2126(T1 initialize);
/* New instance of type [detachable] LX_MISSING_QUOTE_ERROR */
extern T0* GE_new2127(T1 initialize);
/* New instance of type [detachable] LX_BAD_CHARACTER_CLASS_ERROR */
extern T0* GE_new2128(T1 initialize);
/* New instance of type [detachable] LX_BAD_CHARACTER_ERROR */
extern T0* GE_new2129(T1 initialize);
/* New instance of type [detachable] LX_FULL_AND_META_ERROR */
extern T0* GE_new2130(T1 initialize);
/* New instance of type [detachable] LX_FULL_AND_REJECT_ERROR */
extern T0* GE_new2131(T1 initialize);
/* New instance of type [detachable] LX_FULL_AND_VARIABLE_TRAILING_CONTEXT_ERROR */
extern T0* GE_new2132(T1 initialize);
/* New instance of type [detachable] LX_CHARACTER_OUT_OF_RANGE_ERROR */
extern T0* GE_new2133(T1 initialize);
/* New instance of type [detachable] ARRAY [[attached] LX_RULE] */
extern T0* GE_new2134(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] LX_RULE] */
extern T0* GE_new2135(T6 a1, T1 initialize);
/* New instance of type [detachable] LX_DFA_STATE */
extern T0* GE_new2136(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE] */
extern T0* GE_new2137(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] LX_DFA_STATE] */
extern T0* GE_new2138(T1 initialize);
/* New instance of type detachable LX_SYMBOL_PARTITIONS */
extern T0* GE_new2139(T1 initialize);
/* New instance of type [detachable] KL_ARRAY_ROUTINES [[attached] LX_RULE] */
extern T0* GE_new2140(T1 initialize);
/* New instance of type [detachable] ARRAY [detachable LX_RULE] */
extern T0* GE_new2141(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable LX_RULE] */
extern T0* GE_new2142(T6 a1, T1 initialize);
/* New instance of type [detachable] LX_START_CONDITION */
extern T0* GE_new2143(T1 initialize);
/* New instance of type [detachable] LX_TRANSITION_TABLE [[attached] LX_DFA_STATE] */
extern T0* GE_new2144(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] LX_NFA] */
extern T0* GE_new2145(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32] */
extern T0* GE_new2146(T1 initialize);
/* New instance of type [detachable] LX_NFA_STATE */
extern T0* GE_new2147(T1 initialize);
/* New instance of type detachable ARRAY [detachable LX_SYMBOL_CLASS] */
extern T0* GE_new2148(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable LX_SYMBOL_CLASS] */
extern T0* GE_new2149(T6 a1, T1 initialize);
/* New instance of type detachable ARRAY [detachable DS_HASH_SET [[attached] INTEGER_32]] */
extern T0* GE_new2150(T1 initialize);
/* New instance of type [detachable] SPECIAL [detachable DS_HASH_SET [[attached] INTEGER_32]] */
extern T0* GE_new2151(T6 a1, T1 initialize);
/* New instance of type detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
extern T0* GE_new2152(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS] */
extern T0* GE_new2153(T1 initialize);
/* New instance of type [detachable] LX_NEGATIVE_RANGE_IN_CHARACTER_CLASS_ERROR */
extern T0* GE_new2154(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_RULE] */
extern T0* GE_new2155(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_RULE] */
extern T0* GE_new2156(T1 initialize);
/* New instance of type [detachable] KL_ARRAY_ROUTINES [[attached] STRING_8] */
extern T0* GE_new2157(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] LX_START_CONDITION] */
extern T0* GE_new2158(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_START_CONDITION] */
extern T0* GE_new2159(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_START_CONDITION] */
extern T0* GE_new2160(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]] */
extern T0* GE_new2161(T6 a1, T1 initialize);
/* New instance of type detachable KL_EQUALITY_TESTER [[attached] LX_SYMBOL_CLASS] */
extern T0* GE_new2162(T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]] */
extern T0* GE_new2164(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS] */
extern T0* GE_new2165(T1 initialize);
/* New instance of type detachable SPECIAL [detachable SPECIAL [[attached] NATURAL_64]] */
extern T0* GE_new2166(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_32, [attached] STRING_8] */
extern T0* GE_new2168(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_SYMBOL_CLASS, [attached] STRING_32] */
extern T0* GE_new2170(T1 initialize);
/* New instance of type [detachable] LX_EPSILON_TRANSITION [[attached] LX_NFA_STATE] */
extern T0* GE_new2171(T1 initialize);
/* New instance of type [detachable] LX_SYMBOL_CLASS_TRANSITION [[attached] LX_NFA_STATE] */
extern T0* GE_new2174(T1 initialize);
/* New instance of type [detachable] LX_SYMBOL_TRANSITION [[attached] LX_NFA_STATE] */
extern T0* GE_new2175(T1 initialize);
/* New instance of type [detachable] LX_ACTION */
extern T0* GE_new2177(T1 initialize);
/* New instance of type [detachable] ARRAY [[attached] DS_BILINKABLE [[attached] INTEGER_32]] */
extern T0* GE_new2178(T1 initialize);
/* New instance of type [detachable] DS_BILINKABLE [[attached] INTEGER_32] */
extern T0* GE_new2179(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_BILINKABLE [[attached] INTEGER_32]] */
extern T0* GE_new2180(T6 a1, T1 initialize);
/* New instance of type [detachable] DS_BUBBLE_SORTER [[attached] LX_NFA_STATE] */
extern T0* GE_new2181(T1 initialize);
/* New instance of type [detachable] DS_BUBBLE_SORTER [[attached] LX_RULE] */
extern T0* GE_new2183(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] LX_NFA_STATE] */
extern T0* GE_new2185(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA_STATE] */
extern T0* GE_new2187(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE] */
extern T0* GE_new2188(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] LX_DFA_STATE] */
extern T0* GE_new2190(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] LX_DFA_STATE] */
extern T0* GE_new2191(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_DFA_STATE] */
extern T0* GE_new2192(T1 initialize);
/* New instance of type [detachable] KL_ARRAY [[attached] LX_RULE] */
extern T0* GE_new2193(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32] */
extern T0* GE_new2194(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA] */
extern T0* GE_new2195(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_NFA, [attached] INTEGER_32] */
extern T0* GE_new2197(T1 initialize);
/* New instance of type [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_RULE] */
extern T0* GE_new2201(T1 initialize);
/* New instance of type [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_NFA_STATE] */
extern T0* GE_new2204(T1 initialize);
/* New instance of type [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
extern T0* GE_new2207(T6 a1, T1 initialize);
/* New instance of type [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
extern T0* GE_new2208(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]] */
extern T0* GE_new2209(T1 initialize);
/* New instance of type [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_SYMBOL_CLASS] */
extern T0* GE_new2210(T1 initialize);
/* New instance of type [detachable] KL_ARRAY [[attached] STRING_8] */
extern T0* GE_new2211(T1 initialize);
/* New instance of type [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_DFA_STATE, [attached] INTEGER_32] */
extern T0* GE_new2213(T1 initialize);
/* [detachable] GEC.execute */
extern T0* T25c61(GE_context* ac);
/* [detachable] GEC.execute_with_arguments */
extern void T25f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] GEC.execute_with_arguments_and_error_handler */
extern void T25f63(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] GEC.report_cannot_read_error */
extern void T25f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_error */
extern void T53f183(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_error_message */
extern void T53f185(GE_context* ac, T0* C, T0* a1);
/* [detachable] MUTEX.unlock */
extern void T131f12(GE_context* ac, T0* C);
/* [detachable] MUTEX.eif_thr_mutex_unlock */
extern void T131f15(GE_context* ac, T0* C, T14 a1);
/* [detachable] MUTEX.default_pointer */
extern T14 T131f3(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.set_has_error */
extern void T53f187(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_error_message */
extern void T53f185p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM.put_line */
extern void T161f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM.put_new_line */
extern void T161f7(GE_context* ac, T0* C);
/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM.put_string */
extern void T161f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] MUTEX.lock */
extern void T131f11(GE_context* ac, T0* C);
/* [detachable] MUTEX.current_thread_id */
extern T14 T131f4(GE_context* ac, T0* C);
/* [detachable] MUTEX.eif_thr_mutex_lock */
extern void T131f14(GE_context* ac, T0* C, T14 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.message */
extern T0* T53f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIVERSE_ERROR.default_message */
extern T0* T1999f12(GE_context* ac, T0* C);
/* [detachable] ET_UNIVERSE_ERROR.message */
extern T0* T1999f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIVERSE_ERROR.arguments */
extern T0* T1999f6(GE_context* ac, T0* C);
/* [detachable] KL_ARGUMENTS.make */
extern T0* T32c6(GE_context* ac);
/* [detachable] KL_ARGUMENTS.argument */
extern T0* T32f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] IMMUTABLE_STRING_32.as_string_8 */
extern T0* T20f9(GE_context* ac, T0* C);
/* [detachable] STRING_8.put_code */
extern void T17f65(GE_context* ac, T0* C, T10 a1, T6 a2);
/* [detachable] STRING_8.reset_hash_codes */
extern void T17f53(GE_context* ac, T0* C);
/* [detachable] STRING_8.valid_code */
extern T1 T17f39(GE_context* ac, T0* C, T10 a1);
/* NATURAL_32.is_less_equal */
extern T1 T10f18(GE_context* ac, T10* C, T10 a1);
/* INTEGER_32.to_natural_32 */
extern T10 T6f21(GE_context* ac, T6* C);
/* [detachable] IMMUTABLE_STRING_32.code */
extern T10 T20f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_8.set_count */
extern void T17f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_8.make */
extern void T17f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_8.make */
extern T0* T17c48(GE_context* ac, T6 a1);
/* [detachable] SPECIAL [CHARACTER_8].make_filled */
extern T0* T15c12(GE_context* ac, T2 a1, T6 a2);
/* [detachable] SPECIAL [CHARACTER_8].fill_with */
extern void T15f15(GE_context* ac, T0* C, T2 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [CHARACTER_8].extend */
extern void T15f19(GE_context* ac, T0* C, T2 a1);
/* INTEGER_32.min */
extern T6 T6f17(GE_context* ac, T6* C, T6 a1);
/* INTEGER_32.is_less_equal */
extern T1 T6f15(GE_context* ac, T6* C, T6 a1);
/* [detachable] SPECIAL [CHARACTER_8].make_empty */
extern void T15f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [CHARACTER_8].make_empty */
extern T0* T15c16(GE_context* ac, T6 a1);
/* [detachable] ARGUMENTS_32.argument */
extern T0* T130f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [[attached] IMMUTABLE_STRING_32].item */
extern T0* T484f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARGUMENTS_32.internal_argument_array */
extern T0* T130f3(GE_context* ac, T0* C);
/* [detachable] ARRAY [[attached] IMMUTABLE_STRING_32].put */
extern void T484f8(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ARRAY [[attached] IMMUTABLE_STRING_32].compare_objects */
extern void T484f7(GE_context* ac, T0* C);
/* [detachable] ARRAY [[attached] IMMUTABLE_STRING_32].make_filled */
extern T0* T484c6(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] ARRAY [[attached] IMMUTABLE_STRING_32].make_filled_area */
extern void T484f9(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] IMMUTABLE_STRING_32].make_filled */
extern T0* T485c6(GE_context* ac, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] IMMUTABLE_STRING_32].fill_with */
extern void T485f9(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] IMMUTABLE_STRING_32].extend */
extern void T485f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [[attached] IMMUTABLE_STRING_32].make_empty */
extern void T485f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] IMMUTABLE_STRING_32].make_empty */
extern T0* T485c4(GE_context* ac, T6 a1);
/* [detachable] IMMUTABLE_STRING_32.make_empty */
extern T0* T20c14(GE_context* ac);
/* [detachable] IMMUTABLE_STRING_32.make */
extern void T20f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] IMMUTABLE_STRING_32.make */
extern T0* T20c18(GE_context* ac, T6 a1);
/* [detachable] SPECIAL [CHARACTER_32].make_filled */
extern T0* T16c8(GE_context* ac, T3 a1, T6 a2);
/* [detachable] SPECIAL [CHARACTER_32].fill_with */
extern void T16f11(GE_context* ac, T0* C, T3 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [CHARACTER_32].extend */
extern void T16f15(GE_context* ac, T0* C, T3 a1);
/* [detachable] SPECIAL [CHARACTER_32].make_empty */
extern void T16f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [CHARACTER_32].make_empty */
extern T0* T16c12(GE_context* ac, T6 a1);
/* [detachable] KL_ARGUMENTS.internal_arguments */
extern T0* T32f4(GE_context* ac, T0* C);
/* [detachable] ARGUMENTS_32.default_create */
extern T0* T130c5(GE_context* ac);
/* [detachable] KL_STRING_ROUTINES.appended_string */
extern T0* T82f10(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] UC_UTF8_STRING.to_string_32 */
extern T0* T775f20(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.next_byte_index */
extern T6 T775f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_ROUTINES.encoded_byte_count */
extern T6 T434f4(GE_context* ac, T0* C, T2 a1);
/* CHARACTER_8.is_less_equal */
extern T1 T2f9(GE_context* ac, T2* C, T2 a1);
/* CHARACTER_8.is_less */
extern T1 T2f14(GE_context* ac, T2* C, T2 a1);
/* CHARACTER_8.natural_32_code */
extern T10 T2f10(GE_context* ac, T2* C);
/* [detachable] UC_UTF8_STRING.byte_item */
extern T2 T775f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.old_item */
extern T2 T775f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.set_count */
extern void T775f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.reset_byte_index_cache */
extern void T775f93(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.utf8 */
extern T0* T775f42(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_ROUTINES.default_create */
extern T0* T434c53(GE_context* ac);
/* [detachable] STRING_32.append_code */
extern void T18f25(GE_context* ac, T0* C, T10 a1);
/* [detachable] STRING_32.put_code */
extern void T18f31(GE_context* ac, T0* C, T10 a1, T6 a2);
/* [detachable] STRING_32.reset_hash_codes */
extern void T18f30(GE_context* ac, T0* C);
/* [detachable] STRING_32.set_count */
extern void T18f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_32.resize */
extern void T18f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [CHARACTER_32].aliased_resized_area_with_default */
extern T0* T16f4(GE_context* ac, T0* C, T3 a1, T6 a2);
/* [detachable] SPECIAL [CHARACTER_32].aliased_resized_area */
extern T0* T16f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_32.capacity */
extern T6 T18f5(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.item_code_at_byte_index */
extern T6 T775f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_ROUTINES.encoded_next_value */
extern T6 T434f5(GE_context* ac, T0* C, T2 a1);
/* [detachable] UC_UTF8_ROUTINES.encoded_first_value */
extern T6 T434f3(GE_context* ac, T0* C, T2 a1);
/* [detachable] STRING_32.make */
extern void T18f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_32.make */
extern T0* T18c23(GE_context* ac, T6 a1);
/* [detachable] UC_STRING.to_string_32 */
extern T0* T429f55(GE_context* ac, T0* C);
/* [detachable] UC_STRING.next_byte_index */
extern T6 T429f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.byte_item */
extern T2 T429f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.old_item */
extern T2 T429f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.set_count */
extern void T429f86(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.reset_byte_index_cache */
extern void T429f99(GE_context* ac, T0* C);
/* [detachable] UC_STRING.utf8 */
extern T0* T429f40(GE_context* ac, T0* C);
/* [detachable] UC_STRING.item_code_at_byte_index */
extern T6 T429f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_8.to_string_32 */
extern T0* T17f30(GE_context* ac, T0* C);
/* [detachable] STRING_8.code */
extern T10 T17f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.append_string_general */
extern void T775f73(GE_context* ac, T0* C, T0* a1);
extern T1 T775f73ot1(T0* a1, T0** a2);
/* [detachable] UC_UTF8_STRING.append_string_general */
extern void T775f73p1(GE_context* ac, T0* C, T0* a1);
extern T1 T775f73p1ot1(T0* a1, T0** a2);
/* [detachable] UC_UTF8_STRING.append */
extern void T775f73p0(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.reset_hash_codes */
extern void T775f96(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.old_set_count */
extern void T775f89(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.append_code */
extern void T775f78(GE_context* ac, T0* C, T10 a1);
/* [detachable] UC_UTF8_STRING.append_item_code */
extern void T775f102(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.put_code_at_byte_index */
extern void T775f109(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] UC_UTF8_STRING.put_byte */
extern void T775f110(GE_context* ac, T0* C, T2 a1, T6 a2);
/* [detachable] UC_UTF8_STRING.old_put */
extern void T775f98(GE_context* ac, T0* C, T2 a1, T6 a2);
/* [detachable] KL_INTEGER_ROUTINES.to_character */
extern T2 T772f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.integer_ */
extern T0* T775f56(GE_context* ac, T0* C);
/* [detachable] KL_INTEGER_ROUTINES.default_create */
extern T0* T772c5(GE_context* ac);
/* [detachable] UC_UTF8_STRING.resize_byte_storage */
extern void T775f97(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_ROUTINES.code_byte_count */
extern T6 T434f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_ROUTINES.natural_32_code_byte_count */
extern T6 T434f18(GE_context* ac, T0* C, T10 a1);
/* [detachable] STRING_32.code */
extern T10 T18f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.code */
extern T10 T775f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.item_code */
extern T6 T775f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.byte_index */
extern T6 T775f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.code */
extern T10 T429f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.item_code */
extern T6 T429f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.byte_index */
extern T6 T429f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.resize */
extern void T775f95(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [CHARACTER_8].aliased_resized_area_with_default */
extern T0* T15f4(GE_context* ac, T0* C, T2 a1, T6 a2);
/* [detachable] SPECIAL [CHARACTER_8].aliased_resized_area */
extern T0* T15f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.byte_capacity */
extern T6 T775f59(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.capacity */
extern T6 T775f59p1(GE_context* ac, T0* C);
/* [detachable] STRING_32.count */
extern T6 T18f2(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.count */
extern T6 T775f1(GE_context* ac, T0* C);
/* [detachable] UC_STRING.count */
extern T6 T429f1(GE_context* ac, T0* C);
/* [detachable] STRING_8.count */
extern T6 T17f2(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.append */
extern void T775f94(GE_context* ac, T0* C, T0* a1);
extern T1 T775f94ot1(T0* a1, T0** a2);
extern T1 T775f94ot2(T0* a1);
/* [detachable] UC_UTF8_STRING.dummy_uc_string */
extern T0* T775f63(GE_context* ac, T0* C);
/* [detachable] UC_STRING.make_empty */
extern T0* T429c106(GE_context* ac);
/* [detachable] UC_STRING.make */
extern void T429f80(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.make */
extern T0* T429c80(GE_context* ac, T6 a1);
/* [detachable] UC_STRING.old_set_count */
extern void T429f83(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.reset_hash_codes */
extern void T429f90(GE_context* ac, T0* C);
/* [detachable] UC_STRING.byte_capacity */
extern T6 T429f57(GE_context* ac, T0* C);
/* [detachable] UC_STRING.capacity */
extern T6 T429f57p1(GE_context* ac, T0* C);
/* [detachable] UC_STRING.make */
extern void T429f80p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.gobo_append_substring */
extern void T775f75(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] UC_UTF8_STRING.put_substring_at_byte_index */
extern void T775f100(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4, T6 a5);
extern T1 T775f100ot1(T0* a1, T0** a2);
extern T1 T775f100ot2(T0* a1, T0** a2);
extern T1 T775f100ot3(T0* a1, T0** a2);
/* NATURAL_32.to_integer_32 */
extern T6 T10f8(GE_context* ac, T10* C);
/* [detachable] UC_UTF8_STRING.put_character_at_byte_index */
extern void T775f99(GE_context* ac, T0* C, T2 a1, T6 a2, T6 a3);
/* [detachable] UC_UTF8_ROUTINES.character_byte_count */
extern T6 T434f6(GE_context* ac, T0* C, T2 a1);
/* [detachable] UC_UTF8_ROUTINES.character_8_byte_count */
extern T6 T434f11(GE_context* ac, T0* C, T2 a1);
/* [detachable] UC_STRING.item */
extern T2 T429f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.character_item_at_byte_index */
extern T2 T429f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.integer_ */
extern T0* T429f56(GE_context* ac, T0* C);
/* [detachable] KL_PLATFORM.maximum_character_code */
extern T6 T477f2(GE_context* ac, T0* C);
/* [detachable] UC_STRING.platform */
extern T0* T429f43(GE_context* ac, T0* C);
/* [detachable] KL_PLATFORM.default_create */
extern T0* T477c3(GE_context* ac);
/* [detachable] UC_UTF8_STRING.item */
extern T2 T775f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.character_item_at_byte_index */
extern T2 T775f43(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.platform */
extern T0* T775f45(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.cloned_string */
extern T0* T775f32(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.twin */
extern T0* T775f25(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.copy */
extern void T775f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.copy */
extern void T775f88p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [CHARACTER_8].copy_data */
extern void T15f14(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] SPECIAL [CHARACTER_8].force */
extern void T15f18(GE_context* ac, T0* C, T2 a1, T6 a2);
/* [detachable] SPECIAL [CHARACTER_8].move_data */
extern void T15f17(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [CHARACTER_8].overlapping_move */
extern void T15f21(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* INTEGER_32.is_greater_equal */
extern T1 T6f14(GE_context* ac, T6* C, T6 a1);
/* [detachable] SPECIAL [CHARACTER_8].non_overlapping_move */
extern void T15f20(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [CHARACTER_8].resized_area */
extern T0* T15f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.append */
extern void T775f94p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.area_lower */
extern T6 T775f47(GE_context* ac, T0* C);
/* [detachable] UC_STRING.area_lower */
extern T6 T429f17(GE_context* ac, T0* C);
/* [detachable] STRING_8.area_lower */
extern T6 T17f9(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.area */
extern T0* T775f2(GE_context* ac, T0* C);
/* [detachable] UC_STRING.area */
extern T0* T429f2(GE_context* ac, T0* C);
/* [detachable] STRING_8.area */
extern T0* T17f1(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.additional_space */
extern T6 T775f60(GE_context* ac, T0* C);
/* INTEGER_32.max */
extern T6 T6f16(GE_context* ac, T6* C, T6 a1);
/* [detachable] UC_UTF8_ROUTINES.substring_byte_count */
extern T6 T434f2(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
extern T1 T434f2ot1(T0* a1, T0** a2);
extern T1 T434f2ot2(T0* a1, T0** a2);
extern T1 T434f2ot3(T0* a1, T0** a2);
/* [detachable] UC_UTF8_ROUTINES.character_32_byte_count */
extern T6 T434f13(GE_context* ac, T0* C, T3 a1);
/* [detachable] STRING_32.item */
extern T3 T18f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.character_32_item */
extern T3 T775f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.character_32_item */
extern T3 T429f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_8.character_32_item */
extern T3 T17f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.shifted_byte_index */
extern T6 T775f12(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UC_STRING.shifted_byte_index */
extern T6 T429f62(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UC_UTF8_ROUTINES.dummy_uc_string */
extern T0* T434f12(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_ROUTINES.any_ */
extern T0* T434f9(GE_context* ac, T0* C);
/* [detachable] KL_ANY_ROUTINES.default_create */
extern T0* T481c2(GE_context* ac);
/* [detachable] KL_ANY_ROUTINES.same_types */
extern T1 T481f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].same_type */
extern T1 T2137f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_DFA_STATE.same_type */
extern T1 T2136f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].same_type */
extern T1 T745f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_32.same_type */
extern T1 T18f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.same_type */
extern T1 T775f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.same_type */
extern T1 T17f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_STRING.same_type */
extern T1 T429f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.same_type */
extern T1 T126f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.same_type */
extern T1 T125f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.any_ */
extern T0* T775f41(GE_context* ac, T0* C);
/* [detachable] UC_STRING.append_string_general */
extern void T429f74(GE_context* ac, T0* C, T0* a1);
extern T1 T429f74ot1(T0* a1, T0** a2);
/* [detachable] UC_STRING.append_string_general */
extern void T429f74p1(GE_context* ac, T0* C, T0* a1);
extern T1 T429f74p1ot1(T0* a1, T0** a2);
/* [detachable] UC_STRING.append */
extern void T429f74p0(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_STRING.append_code */
extern void T429f79(GE_context* ac, T0* C, T10 a1);
/* [detachable] UC_STRING.append_item_code */
extern void T429f98(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.put_code_at_byte_index */
extern void T429f104(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] UC_STRING.put_byte */
extern void T429f105(GE_context* ac, T0* C, T2 a1, T6 a2);
/* [detachable] UC_STRING.old_put */
extern void T429f92(GE_context* ac, T0* C, T2 a1, T6 a2);
/* [detachable] UC_STRING.resize_byte_storage */
extern void T429f91(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.resize */
extern void T429f89(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.append */
extern void T429f94(GE_context* ac, T0* C, T0* a1);
extern T1 T429f94ot1(T0* a1, T0** a2);
extern T1 T429f94ot2(T0* a1);
/* [detachable] UC_STRING.dummy_uc_string */
extern T0* T429f61(GE_context* ac, T0* C);
/* [detachable] UC_STRING.gobo_append_substring */
extern void T429f75(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] UC_STRING.put_substring_at_byte_index */
extern void T429f95(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4, T6 a5);
extern T1 T429f95ot1(T0* a1, T0** a2);
extern T1 T429f95ot2(T0* a1, T0** a2);
extern T1 T429f95ot3(T0* a1, T0** a2);
/* [detachable] UC_STRING.put_character_at_byte_index */
extern void T429f93(GE_context* ac, T0* C, T2 a1, T6 a2, T6 a3);
/* [detachable] UC_STRING.cloned_string */
extern T0* T429f46(GE_context* ac, T0* C);
/* [detachable] UC_STRING.twin */
extern T0* T429f54(GE_context* ac, T0* C);
/* [detachable] UC_STRING.copy */
extern void T429f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_STRING.copy */
extern void T429f82p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_STRING.append */
extern void T429f94p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_STRING.additional_space */
extern T6 T429f58(GE_context* ac, T0* C);
/* [detachable] UC_STRING.any_ */
extern T0* T429f48(GE_context* ac, T0* C);
/* [detachable] STRING_8.append_string_general */
extern void T17f70(GE_context* ac, T0* C, T0* a1);
extern T1 T17f70ot1(T0* a1, T0** a2);
/* [detachable] STRING_8.append */
extern void T17f70p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.append_code */
extern void T17f58(GE_context* ac, T0* C, T10 a1);
/* [detachable] STRING_8.resize */
extern void T17f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_8.capacity */
extern T6 T17f7(GE_context* ac, T0* C);
/* [detachable] STRING_8.append */
extern void T17f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.additional_space */
extern T6 T17f8(GE_context* ac, T0* C);
/* [detachable] KL_STRING_ROUTINES.as_readable_string_general_no_uc_string */
extern T0* T82f15(GE_context* ac, T0* C, T0* a1);
extern T1 T82f15ot1(T0* a1, T0** a2);
/* [detachable] UC_UTF8_STRING.as_string */
extern T0* T775f19(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.to_utf8 */
extern T0* T775f49(GE_context* ac, T0* C);
/* [detachable] UC_STRING.as_string */
extern T0* T429f9(GE_context* ac, T0* C);
/* [detachable] UC_STRING.to_utf8 */
extern T0* T429f39(GE_context* ac, T0* C);
/* [detachable] KL_STRING_ROUTINES.appended_string_8 */
extern T0* T82f16(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T82f16ot1(T0* a1, T0** a2);
extern T1 T82f16ot2(T0* a1, T0** a2);
/* [detachable] UC_UTF8_STRING.append_string */
extern void T775f77(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_STRING.append_string */
extern void T429f77(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.append_string */
extern void T17f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STRING_ROUTINES.concat_string_8 */
extern T0* T82f17(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T82f17ot1(T0* a1, T0** a2);
extern T1 T82f17ot2(T0* a1, T0** a2);
/* [detachable] UC_STRING.make_from_string_general */
extern T0* T429c71(GE_context* ac, T0* a1);
extern T1 T429f71ot1(T0* a1, T0** a2);
/* [detachable] UC_STRING.make_from_substring_general */
extern void T429f88(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
extern T1 T429f88ot1(T0* a1, T0** a2);
/* [detachable] UC_UTF8_STRING.plus */
extern T0* T775f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.make */
extern void T775f82(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.make */
extern T0* T775c82(GE_context* ac, T6 a1);
/* [detachable] UC_UTF8_STRING.make */
extern void T775f82p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_8.plus */
extern T0* T17f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.new_string */
extern T0* T17f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.plus */
extern T0* T429f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.to_string_8 */
extern T0* T775f31(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.put_code */
extern void T775f90(GE_context* ac, T0* C, T10 a1, T6 a2);
/* [detachable] UC_UTF8_STRING.put_item_code */
extern void T775f108(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UC_UTF8_STRING.move_bytes_right */
extern void T775f106(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UC_UTF8_STRING.move_bytes_left */
extern void T775f105(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UC_STRING.to_string_8 */
extern T0* T429f11(GE_context* ac, T0* C);
/* [detachable] UC_STRING.put_code */
extern void T429f84(GE_context* ac, T0* C, T10 a1, T6 a2);
/* [detachable] UC_STRING.put_item_code */
extern void T429f101(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UC_STRING.move_bytes_right */
extern void T429f103(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UC_STRING.move_bytes_left */
extern void T429f102(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] STRING_8.to_string_8 */
extern T0* T17f32(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.prefixed_string */
extern T0* T775f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_STRING.prefixed_string */
extern T0* T429f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.gobo_plus_general */
extern T0* T775f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_STRING.gobo_plus_general */
extern T0* T429f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ARRAY [[attached] STRING_8].item */
extern T0* T33f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [[attached] STRING_8].valid_index */
extern T1 T33f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.to_integer */
extern T6 T775f40(GE_context* ac, T0* C);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.parsed_integer */
extern T6 T436f13(GE_context* ac, T0* C);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.parse_string_with_type */
extern void T436f17(GE_context* ac, T0* C, T0* a1, T6 a2);
/* NATURAL_32.is_valid_character_8_code */
extern T1 T10f20(GE_context* ac, T10* C);
/* CHARACTER_32.is_character_8 */
extern T1 T3f10(GE_context* ac, T3* C);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.parse_character */
extern void T436f23(GE_context* ac, T0* C, T2 a1);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.overflowed */
extern T1 T436f15(GE_context* ac, T0* C);
/* [detachable] INTEGER_OVERFLOW_CHECKER.will_overflow */
extern T1 T925f7(GE_context* ac, T0* C, T11 a1, T11 a2, T6 a3, T6 a4);
/* NATURAL_64.is_greater */
extern T1 T11f14(GE_context* ac, T11* C, T11 a1);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.overflow_checker */
extern T0* T436f14(GE_context* ac, T0* C);
/* [detachable] INTEGER_OVERFLOW_CHECKER.make */
extern T0* T925c13(GE_context* ac);
/* NATURAL_64.to_natural_64 */
extern T11 T11f17(GE_context* ac, T11* C);
/* NATURAL_32.to_natural_64 */
extern T11 T10f32(GE_context* ac, T10* C);
/* NATURAL_16.to_natural_64 */
extern T11 T9f18(GE_context* ac, T9* C);
/* NATURAL_8.to_natural_64 */
extern T11 T8f19(GE_context* ac, T8* C);
/* INTEGER_64.to_natural_64 */
extern T11 T7f18(GE_context* ac, T7* C);
/* INTEGER_16.to_natural_64 */
extern T11 T5f6(GE_context* ac, T5* C);
/* INTEGER_8.to_natural_64 */
extern T11 T4f6(GE_context* ac, T4* C);
/* [detachable] SPECIAL [[attached] NATURAL_64].extend */
extern void T767f13(GE_context* ac, T0* C, T11 a1);
/* [detachable] SPECIAL [[attached] NATURAL_64].make_empty */
extern void T767f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] NATURAL_64].make_empty */
extern T0* T767c11(GE_context* ac, T6 a1);
/* [detachable] STRING_8.has */
extern T1 T17f44(GE_context* ac, T0* C, T2 a1);
/* INTEGER_32.to_natural_64 */
extern T11 T6f29(GE_context* ac, T6* C);
/* CHARACTER_8.is_digit */
extern T1 T2f26(GE_context* ac, T2* C);
/* NATURAL_8.is_greater */
extern T1 T8f5(GE_context* ac, T8* C, T8 a1);
/* CHARACTER_8.character_types */
extern T8 T2f12(GE_context* ac, T2* C, T6 a1);
/* CHARACTER_8.internal_character_types */
extern T0* T2f15(GE_context* ac, T2* C);
/* [detachable] SPECIAL [[attached] NATURAL_8].extend */
extern void T487f7(GE_context* ac, T0* C, T8 a1);
/* [detachable] SPECIAL [[attached] NATURAL_8].make_empty */
extern void T487f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] NATURAL_8].make_empty */
extern T0* T487c6(GE_context* ac, T6 a1);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.reset */
extern void T436f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.ctoi_convertor */
extern T0* T775f58(GE_context* ac, T0* C);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.set_trailing_separators_acceptable */
extern void T436f21(GE_context* ac, T0* C, T1 a1);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.set_leading_separators_acceptable */
extern void T436f20(GE_context* ac, T0* C, T1 a1);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.set_trailing_separators */
extern void T436f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.make_from_string */
extern T0* T17c81(GE_context* ac, T0* a1);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.set_leading_separators */
extern void T436f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.make */
extern T0* T436c16(GE_context* ac);
/* [detachable] UC_STRING.to_integer */
extern T6 T429f31(GE_context* ac, T0* C);
/* [detachable] UC_STRING.ctoi_convertor */
extern T0* T429f52(GE_context* ac, T0* C);
/* [detachable] STRING_8.to_integer */
extern T6 T17f23(GE_context* ac, T0* C);
/* [detachable] STRING_8.ctoi_convertor */
extern T0* T17f25(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.is_integer */
extern T1 T775f39(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.is_valid_integer_or_natural */
extern T1 T775f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_TO_INTEGER_CONVERTOR.is_integral_integer */
extern T1 T436f12(GE_context* ac, T0* C);
/* [detachable] UC_STRING.is_integer */
extern T1 T429f30(GE_context* ac, T0* C);
/* [detachable] UC_STRING.is_valid_integer_or_natural */
extern T1 T429f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_8.is_integer */
extern T1 T17f29(GE_context* ac, T0* C);
/* [detachable] STRING_8.is_valid_integer_or_natural */
extern T1 T17f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_STRING_ROUTINES.append_substring_to_string */
extern void T82f23(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
extern T1 T82f23ot1(T0* a1, T0** a2);
/* [detachable] KL_STRING_ROUTINES.appended_substring */
extern T0* T82f9(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
/* [detachable] UC_UTF8_STRING.append_substring_general */
extern void T775f80(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] UC_STRING.append_substring_general */
extern void T429f72(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] STRING_8.append_substring_general */
extern void T17f69(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] KL_STRING_ROUTINES.appended_substring_8 */
extern T0* T82f14(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
extern T1 T82f14ot1(T0* a1, T0** a2);
extern T1 T82f14ot2(T0* a1, T0** a2);
/* [detachable] UC_UTF8_STRING.append_substring */
extern void T775f79(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] UC_STRING.append_substring */
extern void T429f76(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] STRING_8.append_substring */
extern void T17f71(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] UC_UTF8_STRING.new_empty_string */
extern T0* T775f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.new_empty_string */
extern T0* T429f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.append_character */
extern void T775f74(GE_context* ac, T0* C, T2 a1);
/* [detachable] UC_STRING.append_character */
extern void T429f73(GE_context* ac, T0* C, T2 a1);
/* [detachable] STRING_8.append_character */
extern void T17f51(GE_context* ac, T0* C, T2 a1);
/* [detachable] STRING_8.item */
extern T2 T17f14(GE_context* ac, T0* C, T6 a1);
/* INTEGER_32.is_greater */
extern T1 T6f4(GE_context* ac, T6* C, T6 a1);
/* [detachable] KL_STRING_ROUTINES.new_empty_string */
extern T0* T82f2(GE_context* ac, T0* C, T0* a1, T6 a2);
extern T1 T82f2ot1(T0* a1, T0** a2);
/* [detachable] UC_UTF8_STRING.old_wipe_out */
extern void T775f84(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.wipe_out */
extern void T775f101(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.wipe_out */
extern void T775f84p1(GE_context* ac, T0* C);
/* [detachable] UC_STRING.old_wipe_out */
extern void T429f81(GE_context* ac, T0* C);
/* [detachable] UC_STRING.wipe_out */
extern void T429f96(GE_context* ac, T0* C);
/* [detachable] UC_STRING.wipe_out */
extern void T429f81p1(GE_context* ac, T0* C);
/* [detachable] STRING_8.wipe_out */
extern void T17f59(GE_context* ac, T0* C);
/* [detachable] STRING_8.twin */
extern T0* T17f10(GE_context* ac, T0* C);
/* [detachable] STRING_8.copy */
extern void T17f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIVERSE_ERROR.string_ */
extern T0* T1999f9(GE_context* ac, T0* C);
/* [detachable] KL_STRING_ROUTINES.default_create */
extern T0* T82c22(GE_context* ac);
/* [detachable] ET_DOTNET_ASSEMBLY_ERROR.default_message */
extern T0* T1942f12(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY_ERROR.message */
extern T0* T1942f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY_ERROR.arguments */
extern T0* T1942f6(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY_ERROR.string_ */
extern T0* T1942f9(GE_context* ac, T0* C);
/* [detachable] ET_CLUSTER_ERROR.default_message */
extern T0* T1940f12(GE_context* ac, T0* C);
/* [detachable] ET_CLUSTER_ERROR.message */
extern T0* T1940f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLUSTER_ERROR.arguments */
extern T0* T1940f6(GE_context* ac, T0* C);
/* [detachable] ET_CLUSTER_ERROR.string_ */
extern T0* T1940f9(GE_context* ac, T0* C);
/* [detachable] ET_SYNTAX_ERROR.default_message */
extern T0* T1941f11(GE_context* ac, T0* C);
/* [detachable] ET_SYNTAX_ERROR.message */
extern T0* T1941f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYNTAX_ERROR.arguments */
extern T0* T1941f12(GE_context* ac, T0* C);
/* [detachable] ET_SYNTAX_ERROR.string_ */
extern T0* T1941f16(GE_context* ac, T0* C);
/* [detachable] UT_TOO_MANY_INCLUDES_ERROR.default_message */
extern T0* T1351f5(GE_context* ac, T0* C);
/* [detachable] UT_TOO_MANY_INCLUDES_ERROR.message */
extern T0* T1351f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_TOO_MANY_INCLUDES_ERROR.arguments */
extern T0* T1351f6(GE_context* ac, T0* C);
/* [detachable] UT_TOO_MANY_INCLUDES_ERROR.string_ */
extern T0* T1351f4(GE_context* ac, T0* C);
/* [detachable] UT_SYNTAX_ERROR.default_message */
extern T0* T1350f5(GE_context* ac, T0* C);
/* [detachable] UT_SYNTAX_ERROR.message */
extern T0* T1350f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_SYNTAX_ERROR.arguments */
extern T0* T1350f6(GE_context* ac, T0* C);
/* [detachable] UT_SYNTAX_ERROR.string_ */
extern T0* T1350f4(GE_context* ac, T0* C);
/* [detachable] ET_VALIDITY_ERROR.default_message */
extern T0* T785f41(GE_context* ac, T0* C);
/* [detachable] ET_VALIDITY_ERROR.message */
extern T0* T785f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_VALIDITY_ERROR.arguments */
extern T0* T785f13(GE_context* ac, T0* C);
/* [detachable] ET_VALIDITY_ERROR.string_ */
extern T0* T785f40(GE_context* ac, T0* C);
/* [detachable] ET_INTERNAL_ERROR.default_message */
extern T0* T784f5(GE_context* ac, T0* C);
/* [detachable] ET_INTERNAL_ERROR.message */
extern T0* T784f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERNAL_ERROR.arguments */
extern T0* T784f6(GE_context* ac, T0* C);
/* [detachable] ET_INTERNAL_ERROR.string_ */
extern T0* T784f9(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_ERROR.default_message */
extern T0* T783f5(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_ERROR.message */
extern T0* T783f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_ERROR.arguments */
extern T0* T783f6(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_ERROR.string_ */
extern T0* T783f9(GE_context* ac, T0* C);
/* [detachable] UT_CANNOT_WRITE_TO_FILE_ERROR.default_message */
extern T0* T305f5(GE_context* ac, T0* C);
/* [detachable] UT_CANNOT_WRITE_TO_FILE_ERROR.message */
extern T0* T305f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_CANNOT_WRITE_TO_FILE_ERROR.arguments */
extern T0* T305f6(GE_context* ac, T0* C);
/* [detachable] UT_CANNOT_WRITE_TO_FILE_ERROR.string_ */
extern T0* T305f4(GE_context* ac, T0* C);
/* [detachable] UT_VERSION_NUMBER.default_message */
extern T0* T79f3(GE_context* ac, T0* C);
/* [detachable] UT_VERSION_NUMBER.message */
extern T0* T79f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_VERSION_NUMBER.arguments */
extern T0* T79f7(GE_context* ac, T0* C);
/* [detachable] UT_VERSION_NUMBER.string_ */
extern T0* T79f6(GE_context* ac, T0* C);
/* [detachable] AP_ERROR.default_message */
extern T0* T87f17(GE_context* ac, T0* C);
/* [detachable] AP_ERROR.message */
extern T0* T87f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ERROR.arguments */
extern T0* T87f20(GE_context* ac, T0* C);
/* [detachable] AP_ERROR.string_ */
extern T0* T87f19(GE_context* ac, T0* C);
/* [detachable] UT_MESSAGE.default_message */
extern T0* T80f3(GE_context* ac, T0* C);
/* [detachable] UT_MESSAGE.message */
extern T0* T80f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_MESSAGE.arguments */
extern T0* T80f7(GE_context* ac, T0* C);
/* [detachable] UT_MESSAGE.string_ */
extern T0* T80f6(GE_context* ac, T0* C);
/* [detachable] UT_CANNOT_READ_FILE_ERROR.default_message */
extern T0* T77f3(GE_context* ac, T0* C);
/* [detachable] UT_CANNOT_READ_FILE_ERROR.message */
extern T0* T77f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_CANNOT_READ_FILE_ERROR.arguments */
extern T0* T77f7(GE_context* ac, T0* C);
/* [detachable] UT_CANNOT_READ_FILE_ERROR.string_ */
extern T0* T77f6(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_error */
extern void T36f183(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_error_message */
extern void T36f185(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.set_has_error */
extern void T36f187(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ERROR_HANDLER.report_error_message */
extern void T36f185p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STDERR_FILE.put_line */
extern void T134f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STDERR_FILE.put_new_line */
extern void T134f8(GE_context* ac, T0* C);
/* [detachable] KL_STDERR_FILE.put_string */
extern void T134f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] CONSOLE.put_string */
extern void T1624f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] CONSOLE.console_ps */
extern void T1624f27(GE_context* ac, T0* C, T14 a1, T14 a2, T6 a3);
/* [detachable] KL_STRING_ROUTINES.as_readable_string_8_no_uc_string */
extern T0* T82f19(GE_context* ac, T0* C, T0* a1);
extern T1 T82f19ot1(T0* a1, T0** a2);
/* [detachable] KL_STDERR_FILE.string_ */
extern T0* T134f2(GE_context* ac, T0* C);
/* [detachable] KL_STDERR_FILE.console */
extern T0* T134f1(GE_context* ac, T0* C);
/* [detachable] STD_FILES.error */
extern T0* T892f3(GE_context* ac, T0* C);
/* [detachable] CONSOLE.make_open_stderr */
extern T0* T1624c21(GE_context* ac, T0* a1);
/* [detachable] CONSOLE.set_write_mode */
extern void T1624f26(GE_context* ac, T0* C);
/* [detachable] CONSOLE.console_def */
extern T14 T1624f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] CONSOLE.make_with_name */
extern void T1624f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_32.make_empty */
extern T0* T18c22(GE_context* ac);
/* [detachable] CONSOLE.make_with_name */
extern void T1624f25p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.make_empty */
extern T0* T17c54(GE_context* ac);
/* [detachable] CONSOLE.default_pointer */
extern T14 T1624f10(GE_context* ac, T0* C);
/* [detachable] CONSOLE.set_name */
extern void T1624f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] FILE_INFO.file_name_to_pointer */
extern T0* T141f11(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable C_STRING.set_string */
extern void T30f11(GE_context* ac, T0* C, T0* a1);
/* detachable C_STRING.set_substring */
extern void T30f13(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] MANAGED_POINTER.put_natural_8 */
extern void T140f15(GE_context* ac, T0* C, T8 a1, T6 a2);
/* POINTER.memory_copy */
extern void T14f19(GE_context* ac, T14* C, T14 a1, T6 a2);
/* POINTER.c_memcpy */
extern void T14f20(GE_context* ac, T14* C, T14 a1, T14 a2, T6 a3);
/* NATURAL_32.to_natural_8 */
extern T8 T10f19(GE_context* ac, T10* C);
/* detachable C_STRING.make_shared_from_pointer_and_count */
extern void T30f10(GE_context* ac, T0* C, T14 a1, T6 a2);
/* detachable C_STRING.make_shared_from_pointer_and_count */
extern T0* T30c10(GE_context* ac, T14 a1, T6 a2);
/* [detachable] MANAGED_POINTER.share_from_pointer */
extern T0* T140c14(GE_context* ac, T14 a1, T6 a2);
/* [detachable] MANAGED_POINTER.increment_counter */
extern void T140f12(GE_context* ac, T0* C);
/* UTF_CONVERTER.escaped_utf_32_substring_into_utf_8_0_pointer */
extern void T478f16(GE_context* ac, T478* C, T0* a1, T6 a2, T6 a3, T0* a4, T6 a5, T0* a6);
/* [detachable] CELL [[attached] INTEGER_32].put */
extern void T479f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] CELL [[attached] INTEGER_32].put */
extern T0* T479c2(GE_context* ac, T6 a1);
/* UTF_CONVERTER.to_natural_32 */
extern T10 T478f12(GE_context* ac, T478* C, T0* a1);
/* [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER.parsed_natural_32 */
extern T10 T1571f12(GE_context* ac, T0* C);
/* [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER.parse_string_with_type */
extern void T1571f19(GE_context* ac, T0* C, T0* a1, T6 a2);
extern T1 T1571f19ot1(T0* a1, T0** a2);
extern T1 T1571f19ot2(T0* a1, T0** a2);
/* [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER.parse_character */
extern void T1571f22(GE_context* ac, T0* C, T2 a1);
/* [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER.overflowed */
extern T1 T1571f16(GE_context* ac, T0* C);
/* [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER.overflow_checker */
extern T0* T1571f15(GE_context* ac, T0* C);
/* CHARACTER_8.is_greater_equal */
extern T1 T2f8(GE_context* ac, T2* C, T2 a1);
/* CHARACTER_8.is_hexa_digit */
extern T1 T2f32(GE_context* ac, T2* C);
/* [detachable] IMMUTABLE_STRING_32.count */
extern T6 T20f2(GE_context* ac, T0* C);
/* [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER.reset */
extern void T1571f18(GE_context* ac, T0* C, T6 a1);
/* UTF_CONVERTER.ctoi_convertor */
extern T0* T478f13(GE_context* ac, T478* C);
/* [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER.set_trailing_separators_acceptable */
extern void T1571f21(GE_context* ac, T0* C, T1 a1);
/* [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER.set_leading_separators_acceptable */
extern void T1571f20(GE_context* ac, T0* C, T1 a1);
/* [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER.make */
extern T0* T1571c17(GE_context* ac);
/* UTF_CONVERTER.is_hexa_decimal */
extern T1 T478f11(GE_context* ac, T478* C, T0* a1);
/* [detachable] HEXADECIMAL_STRING_TO_INTEGER_CONVERTER.is_integral_integer */
extern T1 T1571f14(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.substring */
extern T0* T775f22(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UC_UTF8_STRING.make_from_substring */
extern T0* T775c83(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] UC_UTF8_STRING.make_from_substring_general */
extern void T775f103(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] STRING_32.substring */
extern T0* T18f17(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] SPECIAL [CHARACTER_32].copy_data */
extern void T16f9(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] SPECIAL [CHARACTER_32].force */
extern void T16f14(GE_context* ac, T0* C, T3 a1, T6 a2);
/* [detachable] SPECIAL [CHARACTER_32].move_data */
extern void T16f13(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [CHARACTER_32].overlapping_move */
extern void T16f17(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [CHARACTER_32].non_overlapping_move */
extern void T16f16(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] STRING_32.new_string */
extern T0* T18f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.substring */
extern T0* T429f14(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UC_STRING.make_from_substring */
extern T0* T429c85(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] STRING_8.substring */
extern T0* T17f15(GE_context* ac, T0* C, T6 a1, T6 a2);
/* UTF_CONVERTER.utf_8_bytes_count */
extern T6 T478f8(GE_context* ac, T478* C, T0* a1, T6 a2, T6 a3);
/* INTEGER_32.do_nothing */
extern void T6f37(GE_context* ac, T6* C);
/* [detachable] FILE_INFO.multi_byte_to_utf_16 */
extern T6 T141f10(GE_context* ac, T0* C, T14 a1, T14 a2, T6 a3);
/* [detachable] FILE_INFO.default_pointer */
extern T14 T141f9(GE_context* ac, T0* C);
/* detachable C_STRING.item */
extern T14 T30f4(GE_context* ac, T0* C);
/* detachable C_STRING.make */
extern T0* T30c9(GE_context* ac, T0* a1);
/* detachable C_STRING.make_empty */
extern void T30f7(GE_context* ac, T0* C, T6 a1);
/* detachable C_STRING.make_empty */
extern T0* T30c7(GE_context* ac, T6 a1);
/* UTF_CONVERTER.escaped_utf_32_substring_into_utf_16_0_pointer */
extern void T478f15(GE_context* ac, T478* C, T0* a1, T6 a2, T6 a3, T0* a4, T6 a5, T0* a6);
/* [detachable] MANAGED_POINTER.put_natural_16 */
extern void T140f16(GE_context* ac, T0* C, T9 a1, T6 a2);
/* NATURAL_32.to_natural_16 */
extern T9 T10f25(GE_context* ac, T10* C);
/* NATURAL_32.is_greater */
extern T1 T10f6(GE_context* ac, T10* C, T10 a1);
/* UTF_CONVERTER.utf_16_bytes_count */
extern T6 T478f9(GE_context* ac, T478* C, T0* a1, T6 a2, T6 a3);
/* [detachable] MANAGED_POINTER.make */
extern T0* T140c10(GE_context* ac, T6 a1);
/* [detachable] EXCEPTIONS.raise */
extern void T486f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] DEVELOPER_EXCEPTION.raise */
extern void T121f16(GE_context* ac, T0* C);
/* [detachable] ISE_EXCEPTION_MANAGER.raise */
extern void T21f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DEVELOPER_EXCEPTION.code */
extern T6 T121f8(GE_context* ac, T0* C);
/* [detachable] ISE_EXCEPTION_MANAGER.default_pointer */
extern T14 T21f9(GE_context* ac, T0* C);
/* [detachable] ISE_EXCEPTION_MANAGER.set_last_exception */
extern void T21f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] CELL [detachable EXCEPTION].put */
extern void T26f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] CELL [detachable EXCEPTION].put */
extern T0* T26c2(GE_context* ac, T0* a1);
/* [detachable] ISE_EXCEPTION_MANAGER.last_exception_cell */
extern T0* T21f2(GE_context* ac, T0* C);
/* [detachable] DEVELOPER_EXCEPTION.set_throwing_exception */
extern void T121f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ISE_EXCEPTION_MANAGER.in_rescue */
extern T1 T21f8(GE_context* ac, T0* C);
/* [detachable] DEVELOPER_EXCEPTION.is_ignored */
extern T1 T121f7(GE_context* ac, T0* C);
/* [detachable] ISE_EXCEPTION_MANAGER.is_ignored */
extern T1 T21f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].has */
extern T1 T27f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].same_keys */
extern T1 T27f26(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].hash_code_of */
extern T6 T27f25(GE_context* ac, T0* C, T6 a1);
/* INTEGER_32.hash_code */
extern T6 T6f25(GE_context* ac, T6* C);
/* [detachable] ISE_EXCEPTION_MANAGER.ignored_exceptions */
extern T0* T21f3(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].make */
extern T0* T27c35(GE_context* ac, T6 a1);
/* [detachable] SPECIAL [[attached] INTEGER_32].make_filled */
extern T0* T128c9(GE_context* ac, T6 a1, T6 a2);
/* [detachable] SPECIAL [[attached] INTEGER_32].fill_with */
extern void T128f12(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] INTEGER_32].extend */
extern void T128f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] BOOLEAN].make_filled */
extern T0* T129c5(GE_context* ac, T1 a1, T6 a2);
/* [detachable] SPECIAL [[attached] BOOLEAN].fill_with */
extern void T129f8(GE_context* ac, T0* C, T1 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] BOOLEAN].extend */
extern void T129f10(GE_context* ac, T0* C, T1 a1);
/* [detachable] SPECIAL [[attached] BOOLEAN].make_empty */
extern void T129f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] BOOLEAN].make_empty */
extern T0* T129c7(GE_context* ac, T6 a1);
/* [detachable] SPECIAL [[attached] INTEGER_32].make_empty */
extern void T128f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] INTEGER_32].make_empty */
extern T0* T128c8(GE_context* ac, T6 a1);
/* [detachable] PRIMES.higher_prime */
extern T6 T127f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] PRIMES.is_prime */
extern T1 T127f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] PRIMES.default_create */
extern T0* T127c5(GE_context* ac);
/* [detachable] DEVELOPER_EXCEPTION.exception_manager */
extern T0* T121f6(GE_context* ac, T0* C);
/* [detachable] DEVELOPER_EXCEPTION.set_description */
extern void T121f13(GE_context* ac, T0* C, T0* a1);
/* detachable C_STRING.set_count */
extern void T30f8(GE_context* ac, T0* C, T6 a1);
/* UTF_CONVERTER.utf_32_string_into_utf_8_0_pointer */
extern void T478f14(GE_context* ac, T478* C, T0* a1, T0* a2, T6 a3, T0* a4);
/* [detachable] DEVELOPER_EXCEPTION.default_create */
extern T0* T121c10(GE_context* ac);
/* [detachable] EXCEPTIONS.default_create */
extern T0* T486c1(GE_context* ac);
/* [detachable] MANAGED_POINTER.default_pointer */
extern T14 T140f4(GE_context* ac, T0* C);
/* POINTER.memory_calloc */
extern T14 T14f5(GE_context* ac, T14* C, T6 a1, T6 a2);
/* POINTER.c_calloc */
extern T14 T14f8(GE_context* ac, T14* C, T6 a1, T6 a2);
/* [detachable] MANAGED_POINTER.resize */
extern void T140f11(GE_context* ac, T0* C, T6 a1);
/* POINTER.memory_set */
extern void T14f15(GE_context* ac, T14* C, T6 a1, T6 a2);
/* POINTER.c_memset */
extern void T14f18(GE_context* ac, T14* C, T14 a1, T6 a2, T6 a3);
/* POINTER.memory_realloc */
extern T14 T14f6(GE_context* ac, T14* C, T6 a1);
/* POINTER.c_realloc */
extern T14 T14f9(GE_context* ac, T14* C, T14 a1, T6 a2);
/* [detachable] CONSOLE.buffered_file_info */
extern T0* T1624f11(GE_context* ac, T0* C);
/* [detachable] FILE_INFO.make */
extern T0* T141c22(GE_context* ac);
/* [detachable] FILE_INFO.make_filled_area */
extern void T141f24(GE_context* ac, T0* C, T8 a1, T6 a2);
/* [detachable] SPECIAL [[attached] NATURAL_8].make_filled */
extern T0* T487c5(GE_context* ac, T8 a1, T6 a2);
/* [detachable] SPECIAL [[attached] NATURAL_8].fill_with */
extern void T487f8(GE_context* ac, T0* C, T8 a1, T6 a2, T6 a3);
/* [detachable] FILE_INFO.stat_size */
extern T6 T141f7(GE_context* ac, T0* C);
/* [detachable] KL_STDERR_FILE.io */
extern T0* T134f4(GE_context* ac, T0* C);
/* [detachable] STD_FILES.set_output_default */
extern void T892f6(GE_context* ac, T0* C);
/* [detachable] STD_FILES.output */
extern T0* T892f2(GE_context* ac, T0* C);
/* [detachable] CONSOLE.make_open_stdout */
extern T0* T1624c23(GE_context* ac, T0* a1);
/* [detachable] STD_FILES.default_create */
extern T0* T892c5(GE_context* ac);
/* [detachable] ET_ERROR_HANDLER.message */
extern T0* T36f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_CANNOT_READ_FILE_ERROR.make */
extern T0* T77c8(GE_context* ac, T0* a1);
/* [detachable] ARRAY [[attached] STRING_8].put */
extern void T33f8(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ARRAY [[attached] STRING_8].make_filled */
extern T0* T33c7(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] ARRAY [[attached] STRING_8].make_filled_area */
extern void T33f9(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] STRING_8].make_filled */
extern T0* T34c7(GE_context* ac, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] STRING_8].fill_with */
extern void T34f9(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] STRING_8].extend */
extern void T34f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [[attached] STRING_8].make_empty */
extern void T34f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] STRING_8].make_empty */
extern T0* T34c8(GE_context* ac, T6 a1);
/* [detachable] GEC.process_system */
extern void T25f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] GEC.compile_c_code */
extern void T25f76(GE_context* ac, T0* C, T0* a1);
/* [detachable] GECC.execute_with_arguments */
extern T0* T89c10(GE_context* ac, T0* a1);
/* [detachable] GECC.execute_with_arguments_and_error_handler */
extern void T89f11(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature [detachable] GECC.execute_with_arguments_and_error_handler */
extern T0* T89f11ac1(GE_context* ac, T0* a1, T0* a2, T6 a3, T0* a4);
/* [detachable] PROCEDURE [[attached] TUPLE].set_rout_disp_final */
extern void T346f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] GECC.report_cannot_read_error */
extern void T89f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_ERROR_HANDLER.report_error */
extern void T132f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_ERROR_HANDLER.report_error_message */
extern void T132f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_ERROR_HANDLER.message */
extern T0* T132f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_NEGATIVE_RANGE_IN_CHARACTER_CLASS_ERROR.default_message */
extern T0* T2154f5(GE_context* ac, T0* C);
/* [detachable] LX_NEGATIVE_RANGE_IN_CHARACTER_CLASS_ERROR.message */
extern T0* T2154f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_NEGATIVE_RANGE_IN_CHARACTER_CLASS_ERROR.arguments */
extern T0* T2154f6(GE_context* ac, T0* C);
/* [detachable] LX_NEGATIVE_RANGE_IN_CHARACTER_CLASS_ERROR.string_ */
extern T0* T2154f4(GE_context* ac, T0* C);
/* [detachable] LX_CHARACTER_OUT_OF_RANGE_ERROR.default_message */
extern T0* T2133f5(GE_context* ac, T0* C);
/* [detachable] LX_CHARACTER_OUT_OF_RANGE_ERROR.message */
extern T0* T2133f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_CHARACTER_OUT_OF_RANGE_ERROR.arguments */
extern T0* T2133f6(GE_context* ac, T0* C);
/* [detachable] LX_CHARACTER_OUT_OF_RANGE_ERROR.string_ */
extern T0* T2133f4(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_VARIABLE_TRAILING_CONTEXT_ERROR.default_message */
extern T0* T2132f5(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_VARIABLE_TRAILING_CONTEXT_ERROR.message */
extern T0* T2132f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_FULL_AND_VARIABLE_TRAILING_CONTEXT_ERROR.arguments */
extern T0* T2132f6(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_VARIABLE_TRAILING_CONTEXT_ERROR.string_ */
extern T0* T2132f4(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_REJECT_ERROR.default_message */
extern T0* T2131f5(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_REJECT_ERROR.message */
extern T0* T2131f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_FULL_AND_REJECT_ERROR.arguments */
extern T0* T2131f6(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_REJECT_ERROR.string_ */
extern T0* T2131f4(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_META_ERROR.default_message */
extern T0* T2130f5(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_META_ERROR.message */
extern T0* T2130f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_FULL_AND_META_ERROR.arguments */
extern T0* T2130f6(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_META_ERROR.string_ */
extern T0* T2130f4(GE_context* ac, T0* C);
/* [detachable] LX_BAD_CHARACTER_ERROR.default_message */
extern T0* T2129f5(GE_context* ac, T0* C);
/* [detachable] LX_BAD_CHARACTER_ERROR.message */
extern T0* T2129f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_BAD_CHARACTER_ERROR.arguments */
extern T0* T2129f6(GE_context* ac, T0* C);
/* [detachable] LX_BAD_CHARACTER_ERROR.string_ */
extern T0* T2129f4(GE_context* ac, T0* C);
/* [detachable] LX_BAD_CHARACTER_CLASS_ERROR.default_message */
extern T0* T2128f5(GE_context* ac, T0* C);
/* [detachable] LX_BAD_CHARACTER_CLASS_ERROR.message */
extern T0* T2128f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_BAD_CHARACTER_CLASS_ERROR.arguments */
extern T0* T2128f6(GE_context* ac, T0* C);
/* [detachable] LX_BAD_CHARACTER_CLASS_ERROR.string_ */
extern T0* T2128f4(GE_context* ac, T0* C);
/* [detachable] LX_MISSING_QUOTE_ERROR.default_message */
extern T0* T2127f5(GE_context* ac, T0* C);
/* [detachable] LX_MISSING_QUOTE_ERROR.message */
extern T0* T2127f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_MISSING_QUOTE_ERROR.arguments */
extern T0* T2127f6(GE_context* ac, T0* C);
/* [detachable] LX_MISSING_QUOTE_ERROR.string_ */
extern T0* T2127f4(GE_context* ac, T0* C);
/* [detachable] LX_INVALID_UNICODE_CHARACTER_ERROR.default_message */
extern T0* T2126f5(GE_context* ac, T0* C);
/* [detachable] LX_INVALID_UNICODE_CHARACTER_ERROR.message */
extern T0* T2126f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_INVALID_UNICODE_CHARACTER_ERROR.arguments */
extern T0* T2126f6(GE_context* ac, T0* C);
/* [detachable] LX_INVALID_UNICODE_CHARACTER_ERROR.string_ */
extern T0* T2126f4(GE_context* ac, T0* C);
/* [detachable] LX_UNRECOGNIZED_RULE_ERROR.default_message */
extern T0* T2125f5(GE_context* ac, T0* C);
/* [detachable] LX_UNRECOGNIZED_RULE_ERROR.message */
extern T0* T2125f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_UNRECOGNIZED_RULE_ERROR.arguments */
extern T0* T2125f6(GE_context* ac, T0* C);
/* [detachable] LX_UNRECOGNIZED_RULE_ERROR.string_ */
extern T0* T2125f4(GE_context* ac, T0* C);
/* [detachable] GECC.process_lines */
extern void T89f13(GE_context* ac, T0* C, T0* a1, T6 a2, T0* a3);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].remove_first */
extern void T147f27(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].remove */
extern void T147f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] STRING_8].keep_head */
extern void T34f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] STRING_8].set_count */
extern void T34f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].move_left */
extern void T147f37(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].move_cursors_left */
extern void T147f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] STRING_8].set_position */
extern void T496f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].remove_last */
extern void T147f35(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].move_last_cursors_after */
extern void T147f38(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] STRING_8].set_next_cursor */
extern void T496f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] STRING_8].set_after */
extern void T496f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].first */
extern T0* T147f7(GE_context* ac, T0* C);
/* [detachable] GECC.process */
extern void T89f14(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] WORKER_THREAD.launch */
extern void T432f10(GE_context* ac, T0* C);
/* [detachable] WORKER_THREAD.launch_with_attributes */
extern void T432f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] WORKER_THREAD.last_created_thread */
extern T14 T432f6(GE_context* ac, T0* C);
/* [detachable] WORKER_THREAD.thr_set_terminated */
extern void T432f14(GE_context* ac, T0* C, T1 a1);
/* [detachable] WORKER_THREAD.thr_set_terminated */
extern void T432a14(T0* C, T1 a1);
/* [detachable] WORKER_THREAD.thr_main */
extern void T432f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] WORKER_THREAD.thr_main */
extern void T432a13(T0* C, T6 a1);
/* [detachable] WORKER_THREAD.execute_procedure */
extern void T432f16(GE_context* ac, T0* C);
/* [detachable] WORKER_THREAD.default_pointer */
extern T14 T432f8(GE_context* ac, T0* C);
/* [detachable] WORKER_THREAD.create_thread_with_attr */
extern void T432f15(GE_context* ac, T0* C, T0* a1, T14 a2, T14 a3, T14 a4);
/* [detachable] THREAD_ATTRIBUTES.item */
extern T14 T776f4(GE_context* ac, T0* C);
/* [detachable] WORKER_THREAD.is_launchable */
extern T1 T432f5(GE_context* ac, T0* C);
/* [detachable] MUTEX.is_set */
extern T1 T131f5(GE_context* ac, T0* C);
/* [detachable] CELL [[attached] BOOLEAN].put */
extern void T777f2(GE_context* ac, T0* C, T1 a1);
/* [detachable] CELL [[attached] BOOLEAN].put */
extern T0* T777c2(GE_context* ac, T1 a1);
/* [detachable] WORKER_THREAD.is_last_launch_successful_cell */
extern T0* T432f7(GE_context* ac, T0* C);
/* [detachable] THREAD_ATTRIBUTES.make */
extern T0* T776c7(GE_context* ac);
/* [detachable] THREAD_ATTRIBUTES.set_priority */
extern void T776f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] THREAD_ATTRIBUTES.c_set_priority */
extern void T776f10(GE_context* ac, T0* C, T14 a1, T6 a2);
/* [detachable] THREAD_ATTRIBUTES.default_priority */
extern T6 T776f5(GE_context* ac, T0* C);
/* [detachable] THREAD_ATTRIBUTES.mem_make */
extern void T776f8(GE_context* ac, T0* C);
/* [detachable] THREAD_ATTRIBUTES.structure_size */
extern T6 T776f6(GE_context* ac, T0* C);
/* [detachable] WORKER_THREAD.make */
extern T0* T432c9(GE_context* ac, T0* a1);
/* [detachable] WORKER_THREAD.thread_make */
extern void T432f11(GE_context* ac, T0* C);
/* [detachable] GECC.thread_count */
extern T6 T89f7(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.parameter */
extern T6 T49f14(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] INTEGER_32].last */
extern T6 T160f7(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] INTEGER_32].is_empty */
extern T1 T160f6(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.was_found */
extern T1 T49f13(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.occurrences */
extern T6 T49f16(GE_context* ac, T0* C);
/* [detachable] MUTEX.make */
extern T0* T131c10(GE_context* ac);
/* [detachable] MUTEX.eif_thr_mutex_create */
extern T14 T131f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].is_empty */
extern T1 T147f19(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].force_last */
extern void T147f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_8].force */
extern void T377f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].resize */
extern void T147f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_8].aliased_resized_area */
extern T0* T377f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] STRING_8].aliased_resized_area */
extern T0* T34f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].new_capacity */
extern T6 T147f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].extendible */
extern T1 T147f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.ends_with */
extern T1 T775f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.area_upper */
extern T6 T17f22(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.area_upper */
extern T6 T775f51(GE_context* ac, T0* C);
/* [detachable] UC_STRING.ends_with */
extern T1 T429f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_STRING.area_upper */
extern T6 T429f45(GE_context* ac, T0* C);
/* [detachable] STRING_8.ends_with */
extern T1 T17f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.starts_with */
extern T1 T17f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.is_empty */
extern T1 T17f11(GE_context* ac, T0* C);
/* [detachable] STRING_8.adjust */
extern void T17f60(GE_context* ac, T0* C);
/* [detachable] STRING_8.right_adjust */
extern void T17f67(GE_context* ac, T0* C);
/* CHARACTER_8.is_space */
extern T1 T2f6(GE_context* ac, T2* C);
/* [detachable] STRING_8.left_adjust */
extern void T17f66(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.read_line */
extern void T37f64(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.unread_character */
extern void T37f69(GE_context* ac, T0* C, T2 a1);
/* detachable KL_LINKABLE [[attached] CHARACTER_8].put_right */
extern void T137f4(GE_context* ac, T0* C, T0* a1);
/* detachable KL_LINKABLE [[attached] CHARACTER_8].make */
extern T0* T137c3(GE_context* ac, T2 a1);
/* [detachable] KL_TEXT_INPUT_FILE.read_character */
extern void T37f68(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.old_read_character */
extern void T37f72(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.file_gc */
extern T2 T37f30(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_TEXT_INPUT_FILE.old_end_of_file */
extern T1 T37f26(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.file_feof */
extern T1 T37f32(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_STRING_ROUTINES.wipe_out */
extern void T82f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.keep_head */
extern void T17f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_TEXT_INPUT_FILE.string_ */
extern T0* T37f28(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].make */
extern void T147f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].make */
extern T0* T147c21(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].set_internal_cursor */
extern void T147f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].new_cursor */
extern T0* T147f14(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] STRING_8].make */
extern T0* T496c8(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_8].make */
extern T0* T377f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_8].default_create */
extern T0* T377c3(GE_context* ac);
/* [detachable] GECC.parse_arguments */
extern void T89f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_ERROR_HANDLER.report_info_message */
extern void T132f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STDOUT_FILE.put_line */
extern void T135f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_DISPLAY_HELP_FLAG.full_usage_instruction */
extern T0* T144f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_PARSER.full_usage_instruction */
extern T0* T42f23(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].forth */
extern void T146f17(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].cursor_forth */
extern void T146f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_ALTERNATIVE_OPTIONS_LIST].set_position */
extern void T493f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].add_traversing_cursor */
extern void T146f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_ALTERNATIVE_OPTIONS_LIST].set_next_cursor */
extern void T493f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].remove_traversing_cursor */
extern void T146f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_PARSER.alternative_usage_instruction */
extern T0* T42f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ST_WORD_WRAPPER.wrapped_string */
extern T0* T142f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ST_WORD_WRAPPER.canonify_whitespace */
extern void T142f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.put */
extern void T775f87(GE_context* ac, T0* C, T2 a1, T6 a2);
/* [detachable] UC_STRING.put */
extern void T429f108(GE_context* ac, T0* C, T2 a1, T6 a2);
/* [detachable] STRING_8.put */
extern void T17f73(GE_context* ac, T0* C, T2 a1, T6 a2);
/* [detachable] ST_WORD_WRAPPER.is_space */
extern T1 T142f8(GE_context* ac, T0* C, T2 a1);
/* [detachable] KL_STRING_ROUTINES.cloned_string */
extern T0* T82f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ST_WORD_WRAPPER.string_ */
extern T0* T142f4(GE_context* ac, T0* C);
/* [detachable] ST_WORD_WRAPPER.set_new_line_indentation */
extern void T142f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] AP_PARSER.wrapper */
extern T0* T42f12(GE_context* ac, T0* C);
/* [detachable] ST_WORD_WRAPPER.set_maximum_text_width */
extern void T142f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ST_WORD_WRAPPER.make */
extern T0* T142c9(GE_context* ac);
/* [detachable] STRING_8.make_filled */
extern T0* T17c56(GE_context* ac, T2 a1, T6 a2);
/* [detachable] STRING_8.fill_character */
extern void T17f63(GE_context* ac, T0* C, T2 a1);
/* [detachable] UC_UTF8_STRING.keep_head */
extern void T775f86(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.keep_head */
extern void T429f97(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.basename */
extern T0* T126f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.root_directory */
extern T0* T126f15(GE_context* ac, T0* C);
/* [detachable] KL_UNIX_FILE_SYSTEM.is_root_directory */
extern T1 T126f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.basename */
extern T0* T125f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.is_directory_separator */
extern T1 T125f18(GE_context* ac, T0* C, T2 a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.secondary_directory_separator */
extern T2 T125f26(GE_context* ac, T0* C);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.root_directory */
extern T0* T125f17(GE_context* ac, T0* C);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.is_root_directory */
extern T1 T125f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_PARSER.arguments */
extern T0* T42f27(GE_context* ac, T0* C);
/* [detachable] AP_PARSER.file_system */
extern T0* T42f26(GE_context* ac, T0* C);
/* [detachable] AP_PARSER.unix_file_system */
extern T0* T42f15(GE_context* ac, T0* C);
/* [detachable] KL_UNIX_FILE_SYSTEM.make */
extern T0* T126c31(GE_context* ac);
/* [detachable] KL_OPERATING_SYSTEM.is_unix */
extern T1 T90f2(GE_context* ac, T0* C);
/* [detachable] KL_OPERATING_SYSTEM.current_working_directory */
extern T0* T90f4(GE_context* ac, T0* C);
/* [detachable] PATH.utf_8_name */
extern T0* T435f6(GE_context* ac, T0* C);
/* UTF_CONVERTER.escaped_utf_32_string_to_utf_8_string_8 */
extern T0* T478f5(GE_context* ac, T478* C, T0* a1);
/* UTF_CONVERTER.escaped_utf_32_string_into_utf_8_string_8 */
extern void T478f18(GE_context* ac, T478* C, T0* a1, T0* a2);
/* [detachable] STRING_8.extend */
extern void T17f83(GE_context* ac, T0* C, T2 a1);
/* [detachable] IMMUTABLE_STRING_32.substring */
extern T0* T20f8(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] IMMUTABLE_STRING_32.empty_string */
extern T0* T20f10(GE_context* ac, T0* C);
/* [detachable] IMMUTABLE_STRING_32.make_from_area_and_bounds */
extern void T20f19(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] IMMUTABLE_STRING_32.make_from_area_and_bounds */
extern T0* T20c19(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] IMMUTABLE_STRING_32.item */
extern T3 T20f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_8.grow */
extern void T17f82(GE_context* ac, T0* C, T6 a1);
/* [detachable] PATH.name */
extern T0* T435f4(GE_context* ac, T0* C);
/* UTF_CONVERTER.utf_8_string_8_to_escaped_string_32 */
extern T0* T478f7(GE_context* ac, T478* C, T0* a1);
/* UTF_CONVERTER.utf_8_string_8_into_escaped_string_32 */
extern void T478f23(GE_context* ac, T478* C, T0* a1, T0* a2);
/* UTF_CONVERTER.escape_code_into */
extern void T478f26(GE_context* ac, T478* C, T0* a1, T9 a2);
/* NATURAL_16.to_hex_string */
extern T0* T9f13(GE_context* ac, T9* C);
/* INTEGER_32.to_hex_character */
extern T2 T6f31(GE_context* ac, T6* C);
/* CHARACTER_8.plus */
extern T2 T2f30(GE_context* ac, T2* C, T6 a1);
/* CHARACTER_8.minus */
extern T2 T2f29(GE_context* ac, T2* C, T6 a1);
/* NATURAL_16.to_integer_32 */
extern T6 T9f5(GE_context* ac, T9* C);
/* [detachable] STRING_32.append_string_general */
extern void T18f24(GE_context* ac, T0* C, T0* a1);
extern T1 T18f24ot1(T0* a1, T0** a2);
/* [detachable] STRING_32.append */
extern void T18f24p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_32.append */
extern void T18f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_32.area_lower */
extern T6 T18f9(GE_context* ac, T0* C);
/* [detachable] STRING_32.additional_space */
extern T6 T18f6(GE_context* ac, T0* C);
/* NATURAL_8.to_hex_string */
extern T0* T8f14(GE_context* ac, T8* C);
/* NATURAL_8.to_integer_32 */
extern T6 T8f16(GE_context* ac, T8* C);
/* NATURAL_16.is_less_equal */
extern T1 T9f11(GE_context* ac, T9* C, T9 a1);
/* [detachable] STRING_32.append_character */
extern void T18f32(GE_context* ac, T0* C, T3 a1);
/* NATURAL_8.to_natural_16 */
extern T9 T8f12(GE_context* ac, T8* C);
/* [detachable] STRING_32.extend */
extern void T18f36(GE_context* ac, T0* C, T3 a1);
/* NATURAL_8.is_less_equal */
extern T1 T8f9(GE_context* ac, T8* C, T8 a1);
/* [detachable] STRING_32.grow */
extern void T18f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] IMMUTABLE_STRING_32.make_from_string */
extern T0* T20c15(GE_context* ac, T0* a1);
/* UTF_CONVERTER.utf_16le_string_8_to_escaped_string_32 */
extern T0* T478f6(GE_context* ac, T478* C, T0* a1);
/* UTF_CONVERTER.utf_16le_string_8_into_escaped_string_32 */
extern void T478f22(GE_context* ac, T478* C, T0* a1, T0* a2);
/* NATURAL_32.is_greater_equal */
extern T1 T10f10(GE_context* ac, T10* C, T10 a1);
/* [detachable] EXECUTION_ENVIRONMENT.current_working_path */
extern T0* T124f3(GE_context* ac, T0* C);
/* [detachable] PATH.make_from_pointer */
extern T0* T435c17(GE_context* ac, T14 a1);
/* [detachable] PATH.reset_internal_data */
extern void T435f20(GE_context* ac, T0* C);
/* [detachable] PATH.normalize */
extern void T435f19(GE_context* ac, T0* C);
/* [detachable] STRING_8.remove_tail */
extern void T17f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] PATH.directory_separator */
extern T2 T435f13(GE_context* ac, T0* C);
/* [detachable] STRING_8.remove_head */
extern void T17f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_8.keep_tail */
extern void T17f77(GE_context* ac, T0* C, T6 a1);
/* [detachable] PATH.root_end_position */
extern T6 T435f12(GE_context* ac, T0* C);
/* [detachable] PATH.next_directory_separator */
extern T6 T435f15(GE_context* ac, T0* C, T6 a1);
/* CHARACTER_8.as_lower */
extern T2 T2f25(GE_context* ac, T2* C);
/* CHARACTER_8.lower_value */
extern T0* T2f27(GE_context* ac, T2* C);
/* [detachable] PATH.is_character */
extern T1 T435f8(GE_context* ac, T0* C, T0* a1, T6 a2, T2 a3);
/* detachable C_STRING.substring_8 */
extern T0* T30f3(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable C_STRING.read_substring_into */
extern void T30f12(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* NATURAL_8.to_natural_32 */
extern T10 T8f8(GE_context* ac, T8* C);
/* [detachable] MANAGED_POINTER.read_natural_8 */
extern T8 T140f5(GE_context* ac, T0* C, T6 a1);
/* TYPED_POINTER [[attached] NATURAL_8].memory_copy */
extern void T893f4(GE_context* ac, T893* C, T14 a1, T6 a2);
/* TYPED_POINTER [[attached] NATURAL_8].c_memcpy */
extern void T893f6(GE_context* ac, T893* C, T14 a1, T14 a2, T6 a3);
/* [detachable] PATH.unit_size */
extern T6 T435f5(GE_context* ac, T0* C);
/* [detachable] PATH.pointer_length_in_bytes */
extern T6 T435f7(GE_context* ac, T0* C, T14 a1);
/* NATURAL_64.to_integer_32 */
extern T6 T11f9(GE_context* ac, T11* C);
/* NATURAL_64.is_less_equal */
extern T1 T11f8(GE_context* ac, T11* C, T11 a1);
/* [detachable] PATH.c_pointer_length_in_bytes */
extern T11 T435f14(GE_context* ac, T0* C, T14 a1);
/* [detachable] PATH.make_from_string */
extern T0* T435c16(GE_context* ac, T0* a1);
/* [detachable] PATH.internal_append_into */
extern void T435f18(GE_context* ac, T0* C, T0* a1, T0* a2, T2 a3);
/* UTF_CONVERTER.escaped_utf_32_string_into_utf_16le_string_8 */
extern void T478f17(GE_context* ac, T478* C, T0* a1, T0* a2);
/* [detachable] EXECUTION_ENVIRONMENT.eif_dir_current */
extern T6 T124f5(GE_context* ac, T0* C, T14 a1, T6 a2);
/* [detachable] KL_OPERATING_SYSTEM.execution_environment */
extern T0* T90f5(GE_context* ac, T0* C);
/* [detachable] EXECUTION_ENVIRONMENT.default_create */
extern T0* T124c10(GE_context* ac);
/* [detachable] STRING_8.is_equal */
extern T1 T17f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [CHARACTER_8].same_items */
extern T1 T15f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] KL_OPERATING_SYSTEM.variable_value */
extern T0* T90f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_32.to_string_8 */
extern T0* T18f15(GE_context* ac, T0* C);
/* [detachable] STRING_32.is_valid_as_string_8 */
extern T1 T18f14(GE_context* ac, T0* C);
/* [detachable] EXECUTION_ENVIRONMENT.item */
extern T0* T124f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] NATIVE_STRING.string */
extern T0* T430f6(GE_context* ac, T0* C);
/* UTF_CONVERTER.utf_8_0_pointer_to_escaped_string_32 */
extern T0* T478f4(GE_context* ac, T478* C, T0* a1);
/* UTF_CONVERTER.utf_8_0_pointer_into_escaped_string_32 */
extern void T478f21(GE_context* ac, T478* C, T0* a1, T0* a2);
/* UTF_CONVERTER.utf_8_0_subpointer_into_escaped_string_32 */
extern void T478f25(GE_context* ac, T478* C, T0* a1, T6 a2, T6 a3, T1 a4, T0* a5);
/* UTF_CONVERTER.utf_16_0_pointer_to_escaped_string_32 */
extern T0* T478f3(GE_context* ac, T478* C, T0* a1);
/* UTF_CONVERTER.utf_16_0_pointer_into_escaped_string_32 */
extern void T478f20(GE_context* ac, T478* C, T0* a1, T0* a2);
/* UTF_CONVERTER.utf_16_0_subpointer_into_escaped_string_32 */
extern void T478f24(GE_context* ac, T478* C, T0* a1, T6 a2, T6 a3, T1 a4, T0* a5);
/* NATURAL_16.to_natural_32 */
extern T10 T9f10(GE_context* ac, T9* C);
/* [detachable] MANAGED_POINTER.read_natural_16 */
extern T9 T140f7(GE_context* ac, T0* C, T6 a1);
/* TYPED_POINTER [[attached] NATURAL_16].memory_copy */
extern void T1468f4(GE_context* ac, T1468* C, T14 a1, T6 a2);
/* TYPED_POINTER [[attached] NATURAL_16].c_memcpy */
extern void T1468f6(GE_context* ac, T1468* C, T14 a1, T14 a2, T6 a3);
/* [detachable] NATIVE_STRING.make_from_pointer */
extern T0* T430c10(GE_context* ac, T14 a1);
/* [detachable] MANAGED_POINTER.make_from_pointer */
extern T0* T140c13(GE_context* ac, T14 a1, T6 a2);
/* POINTER.memory_alloc */
extern T14 T14f11(GE_context* ac, T14* C, T6 a1);
/* POINTER.c_malloc */
extern T14 T14f12(GE_context* ac, T14* C, T6 a1);
/* [detachable] NATIVE_STRING.unit_size */
extern T6 T430f5(GE_context* ac, T0* C);
/* [detachable] NATIVE_STRING.pointer_length_in_bytes */
extern T6 T430f4(GE_context* ac, T0* C, T14 a1);
/* [detachable] NATIVE_STRING.c_pointer_length_in_bytes */
extern T11 T430f8(GE_context* ac, T0* C, T14 a1);
/* [detachable] EXECUTION_ENVIRONMENT.eif_getenv */
extern T14 T124f4(GE_context* ac, T0* C, T14 a1);
/* [detachable] NATIVE_STRING.item */
extern T14 T430f3(GE_context* ac, T0* C);
/* [detachable] NATIVE_STRING.make */
extern T0* T430c9(GE_context* ac, T0* a1);
/* [detachable] NATIVE_STRING.set_string */
extern void T430f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] NATIVE_STRING.set_substring */
extern void T430f13(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] NATIVE_STRING.upper_cell */
extern T0* T430f7(GE_context* ac, T0* C);
/* [detachable] NATIVE_STRING.make_empty */
extern void T430f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] AP_PARSER.windows_file_system */
extern T0* T42f14(GE_context* ac, T0* C);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.make */
extern T0* T125c34(GE_context* ac);
/* [detachable] AP_PARSER.operating_system */
extern T0* T42f13(GE_context* ac, T0* C);
/* [detachable] KL_OPERATING_SYSTEM.default_create */
extern T0* T90c6(GE_context* ac);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.forth */
extern void T43f17(GE_context* ac, T0* C);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.cursor_forth */
extern void T43f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.add_traversing_cursor */
extern void T43f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] AP_OPTION].set_next_cursor */
extern void T149f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.remove_traversing_cursor */
extern void T43f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] AP_OPTION].set */
extern void T149f8(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.item_for_iteration */
extern T0* T43f6(GE_context* ac, T0* C);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.cursor_item */
extern T0* T43f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.after */
extern T1 T43f5(GE_context* ac, T0* C);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.cursor_after */
extern T1 T43f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.start */
extern void T43f16(GE_context* ac, T0* C);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.cursor_start */
extern void T43f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.cursor_off */
extern T1 T43f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] AP_OPTION].off */
extern T1 T149f6(GE_context* ac, T0* C);
/* [detachable] AP_DISPLAY_HELP_FLAG.example */
extern T0* T144f14(GE_context* ac, T0* C);
/* CHARACTER_8.out */
extern T0* T2f4(GE_context* ac, T2* C);
/* [detachable] AP_FLAG.example */
extern T0* T39f11(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].item_for_iteration */
extern T0* T146f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].cursor_item */
extern T0* T146f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].after */
extern T1 T146f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].cursor_after */
extern T1 T146f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_ALTERNATIVE_OPTIONS_LIST].after */
extern T1 T493f5(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].start */
extern void T146f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].cursor_start */
extern void T146f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_ALTERNATIVE_OPTIONS_LIST].set_after */
extern void T493f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].is_empty */
extern T1 T146f13(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_ALTERNATIVE_OPTIONS_LIST].off */
extern T1 T493f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_ALTERNATIVE_OPTIONS_LIST].before */
extern T1 T493f6(GE_context* ac, T0* C);
/* [detachable] AP_PARSER.usage_instruction */
extern T0* T42f17(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].forth */
extern void T145f20(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].cursor_forth */
extern void T145f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_OPTION].set_position */
extern void T489f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].add_traversing_cursor */
extern void T145f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_OPTION].set_next_cursor */
extern void T489f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].remove_traversing_cursor */
extern void T145f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_INTEGER_OPTION.example */
extern T0* T49f15(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.example */
extern T0* T48f17(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.example */
extern T0* T47f13(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.example */
extern T0* T44f18(GE_context* ac, T0* C);
/* [detachable] AP_DISPLAY_HELP_FLAG.is_hidden */
extern T1 T144f6(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.is_hidden */
extern T1 T49f8(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.is_hidden */
extern T1 T48f7(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.is_hidden */
extern T1 T47f8(GE_context* ac, T0* C);
/* [detachable] AP_FLAG.is_hidden */
extern T1 T39f6(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.is_hidden */
extern T1 T44f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].item_for_iteration */
extern T0* T145f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].cursor_item */
extern T0* T145f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].after */
extern T1 T145f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].cursor_after */
extern T1 T145f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_OPTION].after */
extern T1 T489f5(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].start */
extern void T145f19(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].cursor_start */
extern void T145f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_OPTION].set_after */
extern void T489f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].is_empty */
extern T1 T145f15(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_OPTION].off */
extern T1 T489f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_OPTION].before */
extern T1 T489f6(GE_context* ac, T0* C);
/* [detachable] GECC.report_version_number */
extern void T89f16(GE_context* ac, T0* C);
/* [detachable] UT_ERROR_HANDLER.report_info */
extern void T132f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_VERSION_NUMBER.make */
extern T0* T79c8(GE_context* ac, T0* a1);
/* [detachable] AP_PARSER.parse_array */
extern void T42f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_PARSER.parse_list */
extern void T42f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_PARSER.final_error_action */
extern void T42f49(GE_context* ac, T0* C);
/* [detachable] AP_PARSER.exceptions */
extern T0* T42f31(GE_context* ac, T0* C);
/* [detachable] KL_EXCEPTIONS.default_create */
extern T0* T35c1(GE_context* ac);
/* [detachable] AP_ERROR_HANDLER.report_error_message */
extern void T148f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ERROR_HANDLER.report_error_message */
extern void T148f10p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_PARSER.check_options_after_parsing */
extern void T42f48(GE_context* ac, T0* C);
/* [detachable] AP_ERROR.make_surplus_option_error */
extern T0* T87c23(GE_context* ac, T0* a1);
/* [detachable] AP_DISPLAY_HELP_FLAG.name */
extern T0* T144f10(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.name */
extern T0* T49f17(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.name */
extern T0* T48f13(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.name */
extern T0* T47f15(GE_context* ac, T0* C);
/* [detachable] AP_FLAG.name */
extern T0* T39f10(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.name */
extern T0* T44f14(GE_context* ac, T0* C);
/* [detachable] AP_DISPLAY_HELP_FLAG.occurrences */
extern T6 T144f5(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.occurrences */
extern T6 T48f14(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.occurrences */
extern T6 T47f16(GE_context* ac, T0* C);
/* [detachable] AP_FLAG.occurrences */
extern T6 T39f1(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.occurrences */
extern T6 T44f17(GE_context* ac, T0* C);
/* [detachable] AP_DISPLAY_HELP_FLAG.maximum_occurrences */
extern T6 T144f8(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.maximum_occurrences */
extern T6 T49f10(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.maximum_occurrences */
extern T6 T48f9(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.maximum_occurrences */
extern T6 T47f10(GE_context* ac, T0* C);
/* [detachable] AP_FLAG.maximum_occurrences */
extern T6 T39f8(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.maximum_occurrences */
extern T6 T44f10(GE_context* ac, T0* C);
/* [detachable] AP_ERROR_HANDLER.report_error */
extern void T148f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ERROR_HANDLER.message */
extern T0* T148f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ERROR.make_missing_option_error */
extern T0* T87c22(GE_context* ac, T0* a1);
/* [detachable] AP_DISPLAY_HELP_FLAG.was_found */
extern T1 T144f13(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.was_found */
extern T1 T48f16(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.was_found */
extern T1 T44f13(GE_context* ac, T0* C);
/* [detachable] AP_DISPLAY_HELP_FLAG.is_mandatory */
extern T1 T144f7(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.is_mandatory */
extern T1 T49f9(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.is_mandatory */
extern T1 T48f8(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.is_mandatory */
extern T1 T47f9(GE_context* ac, T0* C);
/* [detachable] AP_FLAG.is_mandatory */
extern T1 T39f7(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.is_mandatory */
extern T1 T44f9(GE_context* ac, T0* C);
/* [detachable] AP_PARSER.parse_argument */
extern void T42f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_PARSER.parse_short */
extern void T42f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_DISPLAY_HELP_FLAG.record_occurrence */
extern void T144f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_DISPLAY_HELP_FLAG.display_help */
extern void T144f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_DISPLAY_HELP_FLAG.exceptions */
extern T0* T144f18(GE_context* ac, T0* C);
/* [detachable] AP_ERROR_HANDLER.report_info_message */
extern void T148f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_DISPLAY_HELP_FLAG.full_help_text */
extern T0* T144f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_PARSER.full_help_text */
extern T0* T42f35(GE_context* ac, T0* C);
/* [detachable] AP_PARSER.option_help_text */
extern T0* T42f40(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] AP_DISPLAY_HELP_FLAG.description */
extern T0* T144f1(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.description */
extern T0* T49f3(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.description */
extern T0* T48f3(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.description */
extern T0* T47f3(GE_context* ac, T0* C);
/* [detachable] AP_FLAG.description */
extern T0* T39f2(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.description */
extern T0* T44f3(GE_context* ac, T0* C);
/* [detachable] AP_DISPLAY_HELP_FLAG.names */
extern T0* T144f19(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.names */
extern T0* T49f21(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.has_long_form */
extern T1 T49f22(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.names */
extern T0* T49f21p1(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.names */
extern T0* T48f24(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.has_long_form */
extern T1 T48f25(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.names */
extern T0* T48f24p1(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.names */
extern T0* T47f23(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.has_long_form */
extern T1 T47f24(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.names */
extern T0* T47f23p1(GE_context* ac, T0* C);
/* [detachable] AP_FLAG.names */
extern T0* T39f16(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.names */
extern T0* T44f22(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.has_long_form */
extern T1 T44f23(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.names */
extern T0* T44f22p1(GE_context* ac, T0* C);
/* [detachable] DS_QUICK_SORTER [[attached] AP_OPTION].sort */
extern void T1944f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_QUICK_SORTER [[attached] AP_OPTION].sort_with_comparator */
extern void T1944f4(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_QUICK_SORTER [[attached] AP_OPTION].subsort_with_comparator */
extern void T1944f5(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
/* [detachable] ARRAY [[attached] INTEGER_32].force */
extern void T291f14(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] SPECIAL [[attached] INTEGER_32].move_data */
extern void T128f16(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] INTEGER_32].overlapping_move */
extern void T128f19(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] INTEGER_32].force */
extern void T128f11(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] SPECIAL [[attached] INTEGER_32].non_overlapping_move */
extern void T128f18(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] ARRAY [[attached] INTEGER_32].capacity */
extern T6 T291f7(GE_context* ac, T0* C);
/* [detachable] ARRAY [[attached] INTEGER_32].set_area */
extern void T291f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [[attached] INTEGER_32].aliased_resized_area */
extern T0* T128f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [[attached] INTEGER_32].make_empty_area */
extern void T291f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [[attached] INTEGER_32].additional_space */
extern T6 T291f6(GE_context* ac, T0* C);
/* [detachable] ARRAY [[attached] INTEGER_32].empty_area */
extern T1 T291f5(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].swap */
extern void T145f28(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].replace */
extern void T145f27(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] AP_OPTION_COMPARATOR.less_than */
extern T1 T1943f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_STRING_ROUTINES.three_way_comparison */
extern T6 T82f21(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] AP_OPTION_COMPARATOR.string_ */
extern T0* T1943f2(GE_context* ac, T0* C);
/* [detachable] AP_DISPLAY_HELP_FLAG.long_form */
extern T0* T144f4(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.long_form */
extern T0* T49f6(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.long_form */
extern T0* T48f6(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.long_form */
extern T0* T47f7(GE_context* ac, T0* C);
/* [detachable] AP_FLAG.long_form */
extern T0* T39f3(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.long_form */
extern T0* T44f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].item */
extern T0* T145f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [[attached] INTEGER_32].item */
extern T6 T291f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [[attached] INTEGER_32].put */
extern void T291f12(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ARRAY [[attached] INTEGER_32].make_filled */
extern T0* T291c10(GE_context* ac, T6 a1, T6 a2, T6 a3);
/* [detachable] ARRAY [[attached] INTEGER_32].make_filled_area */
extern void T291f13(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] AP_PARSER.all_options */
extern T0* T42f25(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].has */
extern T1 T145f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].make */
extern T0* T145c17(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].set_internal_cursor */
extern void T145f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].new_cursor */
extern T0* T145f12(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_OPTION].make */
extern T0* T489c8(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] AP_OPTION].make */
extern T0* T492f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] AP_OPTION].make_empty */
extern T0* T491c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] AP_OPTION].default_create */
extern T0* T492c3(GE_context* ac);
/* [detachable] DS_QUICK_SORTER [[attached] AP_OPTION].make */
extern T0* T1944c2(GE_context* ac, T0* a1);
/* [detachable] AP_OPTION_COMPARATOR.default_create */
extern T0* T1943c3(GE_context* ac);
/* [detachable] AP_INTEGER_OPTION.record_occurrence */
extern void T49f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ERROR.make_invalid_parameter_error */
extern T0* T87c21(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_LINKED_LIST [[attached] INTEGER_32].force_last */
extern void T160f10(GE_context* ac, T0* C, T6 a1);
/* detachable DS_LINKABLE [[attached] INTEGER_32].put_right */
extern void T503f4(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] INTEGER_32].make */
extern T0* T503c3(GE_context* ac, T6 a1);
/* [detachable] AP_BOOLEAN_OPTION.record_occurrence */
extern void T48f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_BOOLEAN_OPTION.false_strings */
extern T0* T48f22(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].put_last */
extern void T147f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].set_equality_tester */
extern void T147f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_BOOLEAN_OPTION.case_insensitive_string_equality_tester */
extern T0* T48f23(GE_context* ac, T0* C);
/* [detachable] KL_CASE_INSENSITIVE_STRING_EQUALITY_TESTER.default_create */
extern T0* T175c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].has */
extern T1 T147f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_CASE_INSENSITIVE_STRING_EQUALITY_TESTER.test */
extern T1 T175f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_STRING_ROUTINES.same_case_insensitive */
extern T1 T82f6(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T82f6ot1(T0* a1);
extern T1 T82f6ot2(T0* a1);
/* CHARACTER_32.lower */
extern T3 T3f5(GE_context* ac, T3* C);
/* [detachable] CHARACTER_PROPERTY.to_lower */
extern T3 T780f1(GE_context* ac, T0* C, T3 a1);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_16 */
extern T0* T780f19(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_15 */
extern T0* T780f18(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_14 */
extern T0* T780f17(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_13 */
extern T0* T780f16(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_12 */
extern T0* T780f15(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_11 */
extern T0* T780f14(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_10 */
extern T0* T780f13(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_9 */
extern T0* T780f12(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_8 */
extern T0* T780f11(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_7 */
extern T0* T780f10(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_6 */
extern T0* T780f9(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_5 */
extern T0* T780f8(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_4 */
extern T0* T780f7(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_3 */
extern T0* T780f6(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_2 */
extern T0* T780f5(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_lower_table_1 */
extern T0* T780f4(GE_context* ac, T0* C);
/* CHARACTER_32.plus */
extern T3 T3f13(GE_context* ac, T3* C, T10 a1);
/* CHARACTER_32.is_less_equal */
extern T1 T3f12(GE_context* ac, T3* C, T3 a1);
/* CHARACTER_32.is_less */
extern T1 T3f15(GE_context* ac, T3* C, T3 a1);
/* CHARACTER_32.properties */
extern T0* T3f9(GE_context* ac, T3* C);
/* [detachable] CHARACTER_PROPERTY.make */
extern T0* T780c37(GE_context* ac);
/* [detachable] UC_UNICODE_ROUTINES.lower_code */
extern T6 T426f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UNICODE_ROUTINES.lower_codes */
extern T0* T426f7(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.empty_lower_code_plane */
extern T0* T426f10(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]].make_filled */
extern T0* T770c4(GE_context* ac, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]].fill_with */
extern void T770f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]].extend */
extern void T770f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]].make_empty */
extern void T770f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UNICODE_ROUTINES.empty_lower_code_segment */
extern T0* T426f17(GE_context* ac, T0* C);
/* [detachable] KL_INTEGER_ROUTINES.to_integer */
extern T6 T772f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UNICODE_ROUTINES.integer_ */
extern T0* T426f28(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_1 */
extern T0* T426f9(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_1_segment_4 */
extern T0* T426f27(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]]].make_filled */
extern T0* T771c4(GE_context* ac, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]]].fill_with */
extern void T771f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]]].extend */
extern void T771f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [[attached] SPECIAL [[attached] ARRAY [[attached] INTEGER_32]]].make_empty */
extern void T771f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0 */
extern T0* T426f8(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_255 */
extern T0* T426f26(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_167 */
extern T0* T426f25(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_166 */
extern T0* T426f24(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_44 */
extern T0* T426f23(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_36 */
extern T0* T426f22(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_33 */
extern T0* T426f21(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_31 */
extern T0* T426f20(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_30 */
extern T0* T426f19(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_16 */
extern T0* T426f18(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_5 */
extern T0* T426f16(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_4 */
extern T0* T426f15(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_3 */
extern T0* T426f14(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_2 */
extern T0* T426f13(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_1 */
extern T0* T426f12(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.lower_code_plane_0_segment_0 */
extern T0* T426f11(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.valid_non_surrogate_natural_32_code */
extern T1 T426f1(GE_context* ac, T0* C, T10 a1);
/* [detachable] KL_STRING_ROUTINES.unicode */
extern T0* T82f12(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.default_create */
extern T0* T426c57(GE_context* ac);
/* [detachable] KL_CASE_INSENSITIVE_STRING_EQUALITY_TESTER.string_ */
extern T0* T175f2(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.true_strings */
extern T0* T48f21(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] BOOLEAN].force_last */
extern void T157f10(GE_context* ac, T0* C, T1 a1);
/* detachable DS_LINKABLE [[attached] BOOLEAN].put_right */
extern void T501f4(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] BOOLEAN].make */
extern T0* T501c3(GE_context* ac, T1 a1);
/* [detachable] AP_ENUMERATION_OPTION.record_occurrence */
extern void T47f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].has */
extern T1 T152f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STRING_EQUALITY_TESTER.test */
extern T1 T153f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] STRING_8.item_code */
extern T6 T17f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].force_last */
extern void T150f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable STRING_8].force */
extern void T498f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable STRING_8].fill_with */
extern void T497f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable STRING_8].extend */
extern void T497f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].resize */
extern void T150f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable STRING_8].aliased_resized_area */
extern T0* T498f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable STRING_8].aliased_resized_area */
extern T0* T497f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].new_capacity */
extern T6 T150f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].extendible */
extern T1 T150f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] AP_FLAG.record_occurrence */
extern void T39f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_STRING_OPTION.record_occurrence */
extern void T44f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ERROR.make_missing_parameter_error */
extern T0* T87c25(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].off */
extern T1 T147f13(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].cursor_off */
extern T1 T147f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] STRING_8].off */
extern T1 T496f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] STRING_8].before */
extern T1 T496f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] STRING_8].after */
extern T1 T496f5(GE_context* ac, T0* C);
/* [detachable] AP_DISPLAY_HELP_FLAG.needs_parameter */
extern T1 T144f9(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.needs_parameter */
extern T1 T49f5(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.needs_parameter */
extern T1 T48f5(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.needs_parameter */
extern T1 T47f6(GE_context* ac, T0* C);
/* [detachable] AP_FLAG.needs_parameter */
extern T1 T39f12(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.needs_parameter */
extern T1 T44f6(GE_context* ac, T0* C);
/* [detachable] AP_ERROR.make_unknown_option_error */
extern T0* T87c24(GE_context* ac, T0* a1);
/* [detachable] AP_DISPLAY_HELP_FLAG.short_form */
extern T2 T144f2(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.short_form */
extern T2 T49f7(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.short_form */
extern T2 T48f11(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.short_form */
extern T2 T47f12(GE_context* ac, T0* C);
/* [detachable] AP_FLAG.short_form */
extern T2 T39f4(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.short_form */
extern T2 T44f12(GE_context* ac, T0* C);
/* [detachable] AP_DISPLAY_HELP_FLAG.has_short_form */
extern T1 T144f3(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.has_short_form */
extern T1 T49f11(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.has_short_form */
extern T1 T48f10(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.has_short_form */
extern T1 T47f11(GE_context* ac, T0* C);
/* [detachable] AP_FLAG.has_short_form */
extern T1 T39f5(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.has_short_form */
extern T1 T44f11(GE_context* ac, T0* C);
/* [detachable] AP_PARSER.parse_long */
extern void T42f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ERROR.make_unnecessary_parameter_error */
extern T0* T87c26(GE_context* ac, T0* a1, T0* a2);
/* [detachable] AP_DISPLAY_HELP_FLAG.allows_parameter */
extern T1 T144f11(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.allows_parameter */
extern T1 T49f18(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.allows_parameter */
extern T1 T48f15(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.allows_parameter */
extern T1 T47f17(GE_context* ac, T0* C);
/* [detachable] AP_FLAG.allows_parameter */
extern T1 T39f13(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.allows_parameter */
extern T1 T44f16(GE_context* ac, T0* C);
/* [detachable] KL_STRING_ROUTINES.same_string */
extern T1 T82f1(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T82f1ot1(T0* a1, T0** a2);
extern T1 T82f1ot2(T0* a1, T0** a2);
/* [detachable] UC_UTF8_STRING.same_string_general */
extern T1 T775f36(GE_context* ac, T0* C, T0* a1);
extern T1 T775f36ot1(T0* a1, T0** a2);
/* [detachable] UC_STRING.same_string_general */
extern T1 T429f27(GE_context* ac, T0* C, T0* a1);
extern T1 T429f27ot1(T0* a1, T0** a2);
/* [detachable] STRING_8.same_string_general */
extern T1 T17f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.same_characters_general */
extern T1 T17f24(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] UC_UTF8_STRING.same_unicode_string */
extern T1 T775f35(GE_context* ac, T0* C, T0* a1);
extern T1 T775f35ot1(T0* a1, T0** a2);
/* [detachable] UC_STRING.same_unicode_string */
extern T1 T429f26(GE_context* ac, T0* C, T0* a1);
extern T1 T429f26ot1(T0* a1, T0** a2);
/* [detachable] AP_PARSER.string_ */
extern T0* T42f33(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.index_of */
extern T6 T775f29(GE_context* ac, T0* C, T2 a1, T6 a2);
/* [detachable] UC_UTF8_STRING.index_of_item_code */
extern T6 T775f52(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UC_STRING.index_of */
extern T6 T429f25(GE_context* ac, T0* C, T2 a1, T6 a2);
/* [detachable] UC_STRING.index_of_item_code */
extern T6 T429f49(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] STRING_8.index_of */
extern T6 T17f37(GE_context* ac, T0* C, T2 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].forth */
extern void T147f24(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].cursor_forth */
extern void T147f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].add_traversing_cursor */
extern void T147f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].remove_traversing_cursor */
extern void T147f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].item_for_iteration */
extern T0* T147f12(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].cursor_item */
extern T0* T147f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].after */
extern T1 T147f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].cursor_after */
extern T1 T147f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].start */
extern void T147f23(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].cursor_start */
extern void T147f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_PARSER.reset_parser */
extern void T42f46(GE_context* ac, T0* C);
/* [detachable] AP_DISPLAY_HELP_FLAG.reset */
extern void T144f25(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.reset */
extern void T49f28(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] INTEGER_32].make */
extern T0* T160c8(GE_context* ac);
/* [detachable] DS_LINKED_LIST [[attached] INTEGER_32].set_internal_cursor */
extern void T160f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] INTEGER_32].new_cursor */
extern T0* T160f5(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] INTEGER_32].make */
extern T0* T504c3(GE_context* ac, T0* a1);
/* [detachable] AP_BOOLEAN_OPTION.reset */
extern void T48f31(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] BOOLEAN].make */
extern T0* T157c8(GE_context* ac);
/* [detachable] DS_LINKED_LIST [[attached] BOOLEAN].set_internal_cursor */
extern void T157f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] BOOLEAN].new_cursor */
extern T0* T157f5(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] BOOLEAN].make */
extern T0* T502c3(GE_context* ac, T0* a1);
/* [detachable] AP_ENUMERATION_OPTION.reset */
extern void T47f31(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].make */
extern T0* T150c13(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].set_internal_cursor */
extern void T150f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].new_cursor */
extern T0* T150f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8].make */
extern T0* T151c9(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable STRING_8].make */
extern T0* T498f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable STRING_8].make_empty */
extern void T497f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable STRING_8].make_empty */
extern T0* T497c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable STRING_8].default_create */
extern T0* T498c3(GE_context* ac);
/* [detachable] AP_FLAG.reset */
extern void T39f22(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.reset */
extern void T44f30(GE_context* ac, T0* C);
/* [detachable] AP_ERROR_HANDLER.reset */
extern void T148f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].make_from_array */
extern T0* T147c22(GE_context* ac, T0* a1);
/* [detachable] ARRAY [[attached] STRING_8].count */
extern T6 T33f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].force_last */
extern void T146f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] AP_ALTERNATIVE_OPTIONS_LIST].force */
extern void T495f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] AP_ALTERNATIVE_OPTIONS_LIST].fill_with */
extern void T494f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] AP_ALTERNATIVE_OPTIONS_LIST].extend */
extern void T494f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].resize */
extern void T146f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] AP_ALTERNATIVE_OPTIONS_LIST].aliased_resized_area */
extern T0* T495f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] AP_ALTERNATIVE_OPTIONS_LIST].aliased_resized_area */
extern T0* T494f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].new_capacity */
extern T6 T146f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].extendible */
extern T1 T146f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.make */
extern T0* T43c11(GE_context* ac, T0* a1);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.old_make */
extern void T43f14(GE_context* ac, T0* C);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.set_internal_cursor */
extern void T43f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.new_cursor */
extern T0* T43f10(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] AP_OPTION].make */
extern T0* T149c7(GE_context* ac, T0* a1);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.set_parameters_description */
extern void T43f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ALTERNATIVE_OPTIONS_LIST.set_introduction_option */
extern void T43f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_FLAG.set_description */
extern void T39f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_FLAG.make */
extern T0* T39c19(GE_context* ac, T2 a1, T0* a2);
/* [detachable] AP_FLAG.set_long_form */
extern void T39f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_FLAG.set_short_form */
extern void T39f23(GE_context* ac, T0* C, T2 a1);
/* [detachable] AP_FLAG.initialize */
extern void T39f20(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].force_last */
extern void T145f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] AP_OPTION].force */
extern void T492f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] AP_OPTION].fill_with */
extern void T491f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] AP_OPTION].extend */
extern void T491f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].resize */
extern void T145f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] AP_OPTION].aliased_resized_area */
extern T0* T492f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] AP_OPTION].aliased_resized_area */
extern T0* T491f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].new_capacity */
extern T6 T145f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_OPTION].extendible */
extern T1 T145f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] AP_INTEGER_OPTION.set_parameter_description */
extern void T49f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_INTEGER_OPTION.set_description */
extern void T49f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_INTEGER_OPTION.make_with_long_form */
extern T0* T49c23(GE_context* ac, T0* a1);
/* [detachable] AP_INTEGER_OPTION.set_long_form */
extern void T49f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_INTEGER_OPTION.initialize */
extern void T49f26(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.initialize */
extern void T49f26p1(GE_context* ac, T0* C);
/* [detachable] AP_INTEGER_OPTION.initialize */
extern void T49f26p0(GE_context* ac, T0* C);
/* [detachable] AP_PARSER.set_parameters_description */
extern void T42f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_PARSER.set_application_description */
extern void T42f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_PARSER.make */
extern T0* T42c41(GE_context* ac);
/* [detachable] AP_DISPLAY_HELP_FLAG.set_description */
extern void T144f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_DISPLAY_HELP_FLAG.make */
extern T0* T144c20(GE_context* ac, T2 a1, T0* a2);
/* [detachable] AP_DISPLAY_HELP_FLAG.set_long_form */
extern void T144f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_DISPLAY_HELP_FLAG.set_short_form */
extern void T144f23(GE_context* ac, T0* C, T2 a1);
/* [detachable] AP_DISPLAY_HELP_FLAG.initialize */
extern void T144f22(GE_context* ac, T0* C);
/* [detachable] AP_PARSER.make */
extern void T42f41p1(GE_context* ac, T0* C);
/* [detachable] AP_ERROR_HANDLER.make_standard */
extern T0* T148c7(GE_context* ac);
/* [detachable] KL_STANDARD_FILES.output */
extern T0* T133f2(GE_context* ac, T0* C);
/* [detachable] KL_STDOUT_FILE.make */
extern T0* T135c5(GE_context* ac);
/* [detachable] KL_STANDARD_FILES.error */
extern T0* T133f1(GE_context* ac, T0* C);
/* [detachable] KL_STDERR_FILE.make */
extern T0* T134c5(GE_context* ac);
/* [detachable] AP_ERROR_HANDLER.std */
extern T0* T148f5(GE_context* ac, T0* C);
/* [detachable] KL_STANDARD_FILES.default_create */
extern T0* T133c4(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].make */
extern T0* T146c14(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].set_internal_cursor */
extern void T146f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] AP_ALTERNATIVE_OPTIONS_LIST].new_cursor */
extern T0* T146f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] AP_ALTERNATIVE_OPTIONS_LIST].make */
extern T0* T493c8(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] AP_ALTERNATIVE_OPTIONS_LIST].make */
extern T0* T495f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] AP_ALTERNATIVE_OPTIONS_LIST].make_empty */
extern T0* T494c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] AP_ALTERNATIVE_OPTIONS_LIST].default_create */
extern T0* T495c3(GE_context* ac);
/* [detachable] GECC.ise_variables */
extern T0* T89f8(GE_context* ac, T0* C);
/* [detachable] ET_ISE_VARIABLES.default_create */
extern T0* T38c17(GE_context* ac);
/* [detachable] GECC.arguments */
extern T0* T89f6(GE_context* ac, T0* C);
/* [detachable] UT_ERROR_HANDLER.make_standard */
extern T0* T132c7(GE_context* ac);
/* [detachable] UT_ERROR_HANDLER.std */
extern T0* T132f4(GE_context* ac, T0* C);
/* INTEGER_32.out */
extern T0* T6f7(GE_context* ac, T6* C);
/* [detachable] STRING_8.append_integer */
extern void T17f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] GEC.c_compile_using_gecc */
extern T1 T25f58(GE_context* ac, T0* C);
/* [detachable] GEC.string_ */
extern T0* T25f52(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.parameter */
extern T0* T47f19(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].last */
extern T0* T150f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].is_empty */
extern T1 T150f8(GE_context* ac, T0* C);
/* [detachable] GEC.c_compile_using_make */
extern T1 T25f57(GE_context* ac, T0* C);
/* [detachable] KL_SHELL_COMMAND.execute */
extern void T88f11(GE_context* ac, T0* C);
/* [detachable] KL_SHELL_COMMAND.operating_system */
extern T0* T88f5(GE_context* ac, T0* C);
/* [detachable] KL_SHELL_COMMAND.system */
extern void T88f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SHELL_COMMAND.system_call */
extern T6 T88f7(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_SHELL_COMMAND.default_shell */
extern T0* T88f6(GE_context* ac, T0* C);
/* [detachable] KL_SHELL_COMMAND.item */
extern T0* T88f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SHELL_COMMAND.eif_getenv */
extern T14 T88f9(GE_context* ac, T0* C, T14 a1);
/* [detachable] UC_UTF8_STRING.old_is_empty */
extern T1 T775f26(GE_context* ac, T0* C);
/* [detachable] UC_STRING.old_is_empty */
extern T1 T429f19(GE_context* ac, T0* C);
/* [detachable] KL_SHELL_COMMAND.make */
extern T0* T88c10(GE_context* ac, T0* a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.absolute_pathname */
extern T0* T126f1(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.pathname */
extern T0* T126f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_UNIX_FILE_SYSTEM.string_ */
extern T0* T126f17(GE_context* ac, T0* C);
/* [detachable] KL_UNIX_FILE_SYSTEM.cwd */
extern T0* T126f13(GE_context* ac, T0* C);
/* [detachable] KL_UNIX_FILE_SYSTEM.execution_environment */
extern T0* T126f23(GE_context* ac, T0* C);
/* [detachable] KL_UNIX_FILE_SYSTEM.is_absolute_pathname */
extern T1 T126f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.absolute_pathname */
extern T0* T125f1(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STRING_ROUTINES.concat */
extern T0* T82f18(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] STRING_32.plus */
extern T0* T18f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.as_string_32 */
extern T0* T775f21(GE_context* ac, T0* C);
/* [detachable] STRING_32.as_string_32 */
extern T0* T18f12(GE_context* ac, T0* C);
/* [detachable] UC_STRING.as_string_32 */
extern T0* T429f63(GE_context* ac, T0* C);
/* [detachable] STRING_8.as_string_32 */
extern T0* T17f42(GE_context* ac, T0* C);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.current_drive */
extern T0* T125f15(GE_context* ac, T0* C);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.string_ */
extern T0* T125f14(GE_context* ac, T0* C);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.pathname */
extern T0* T125f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.cwd */
extern T0* T125f13(GE_context* ac, T0* C);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.execution_environment */
extern T0* T125f25(GE_context* ac, T0* C);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.is_relative_pathname */
extern T1 T125f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.is_absolute_pathname */
extern T1 T125f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] GEC.file_system */
extern T0* T25f56(GE_context* ac, T0* C);
/* [detachable] GEC.unix_file_system */
extern T0* T25f60(GE_context* ac, T0* C);
/* [detachable] GEC.windows_file_system */
extern T0* T25f59(GE_context* ac, T0* C);
/* [detachable] GEC.c_compile_using_script */
extern T1 T25f55(GE_context* ac, T0* C);
/* [detachable] KL_OPERATING_SYSTEM.is_windows */
extern T1 T90f1(GE_context* ac, T0* C);
/* [detachable] GEC.operating_system */
extern T0* T25f54(GE_context* ac, T0* C);
/* [detachable] GEC.no_c_compile */
extern T1 T25f50(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.dynamic_feature_count */
extern T6 T63f72(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.procedures */
extern T0* T451f14(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.procedures */
extern T0* T450f14(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.procedures */
extern T0* T302f12(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.procedures */
extern T0* T319f12(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.procedures */
extern T0* T223f12(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.queries */
extern T0* T451f7(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.queries */
extern T0* T450f7(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.queries */
extern T0* T302f7(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.queries */
extern T0* T319f7(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.queries */
extern T0* T223f2(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].item */
extern T0* T76f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_SYSTEM.alive_dynamic_type_count */
extern T6 T63f73(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.is_alive */
extern T1 T451f4(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.is_alive */
extern T1 T450f4(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.is_alive */
extern T1 T302f4(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.is_alive */
extern T1 T319f4(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_alive */
extern T1 T223f3(GE_context* ac, T0* C);
/* [detachable] KL_STDOUT_FILE.put_new_line */
extern void T135f8(GE_context* ac, T0* C);
/* [detachable] KL_STDOUT_FILE.put_integer */
extern void T135f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_STDOUT_FILE.put_integer_64 */
extern void T135f11(GE_context* ac, T0* C, T7 a1);
/* [detachable] KL_STDOUT_FILE.put_character */
extern void T135f10(GE_context* ac, T0* C, T2 a1);
/* [detachable] CONSOLE.put_character */
extern void T1624f24(GE_context* ac, T0* C, T2 a1);
/* [detachable] CONSOLE.console_pc */
extern void T1624f28(GE_context* ac, T0* C, T14 a1, T2 a2);
/* [detachable] KL_STDOUT_FILE.console */
extern T0* T135f1(GE_context* ac, T0* C);
/* [detachable] KL_STDOUT_FILE.io */
extern T0* T135f4(GE_context* ac, T0* C);
/* INTEGER_32.to_integer_64 */
extern T7 T6f22(GE_context* ac, T6* C);
/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM.put_integer */
extern void T161f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM.put_integer_64 */
extern void T161f10(GE_context* ac, T0* C, T7 a1);
/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM.put_character */
extern void T161f8(GE_context* ac, T0* C, T2 a1);
/* [detachable] KL_STDOUT_FILE.put_string */
extern void T135f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STDOUT_FILE.string_ */
extern T0* T135f2(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.generate */
extern void T67f561(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].wipe_out */
extern void T216f52(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].slots_wipe_out */
extern void T216f69(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].clashes_wipe_out */
extern void T216f68(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].key_storage_wipe_out */
extern void T216f67(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].item_storage_wipe_out */
extern void T216f66(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].unset_found_item */
extern void T216f58(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].move_all_cursors_after */
extern void T216f65(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].set_next_cursor */
extern void T304f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].set_after */
extern void T304f17(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.generate_compilation_script */
extern void T67f565(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.generate_compilation_makefile */
extern void T67f604(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.report_cannot_write_error */
extern void T67f567(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.error_handler */
extern T0* T67f140(GE_context* ac, T0* C);
/* [detachable] UT_CANNOT_WRITE_TO_FILE_ERROR.make */
extern T0* T305c8(GE_context* ac, T0* a1);
/* [detachable] ET_C_GENERATOR.set_fatal_error */
extern void T67f566(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.close */
extern void T300f35(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.old_close */
extern void T300f40(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.default_pointer */
extern T14 T300f18(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.file_close */
extern void T300f48(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.put_line */
extern void T300f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.put_new_line */
extern void T300f34(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.put_string */
extern void T300f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.old_put_string */
extern void T300f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.file_ps */
extern void T300f50(GE_context* ac, T0* C, T14 a1, T14 a2, T6 a3);
/* [detachable] KL_TEXT_OUTPUT_FILE.string_ */
extern T0* T300f12(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.expanded_string */
extern T0* T303f23(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] DS_STRING_HASH_TABLE.value */
extern T0* T303f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_STRING_HASH_TABLE.item_storage_item */
extern T0* T303f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.position_of_key */
extern T6 T303f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_STRING_HASH_TABLE.clashes_item */
extern T6 T303f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.key_storage_item */
extern T0* T303f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.hash_position */
extern T6 T303f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.hash_code */
extern T6 T775f10(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.string */
extern T0* T775f38(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.hash_code */
extern T6 T775f10p1(GE_context* ac, T0* C);
/* [detachable] UC_STRING.hash_code */
extern T6 T429f13(GE_context* ac, T0* C);
/* [detachable] UC_STRING.string */
extern T0* T429f44(GE_context* ac, T0* C);
/* [detachable] UC_STRING.hash_code */
extern T6 T429f13p1(GE_context* ac, T0* C);
/* [detachable] STRING_8.hash_code */
extern T6 T17f28(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.slots_item */
extern T6 T303f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.string_ */
extern T0* T303f32(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.is_open_write */
extern T1 T300f21(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.old_is_open_write */
extern T1 T300f19(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.open_write */
extern void T300f30(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.is_closed */
extern T1 T300f17(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.old_is_closed */
extern T1 T300f20(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.old_open_write */
extern void T300f42(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.open_write */
extern void T300f42p1(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.file_open */
extern T14 T300f23(GE_context* ac, T0* C, T14 a1, T6 a2);
/* [detachable] KL_TEXT_OUTPUT_FILE.internal_name_pointer */
extern T0* T300f22(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.string_name */
extern T0* T300f14(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.as_string_8 */
extern T0* T775f27(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.valid_code */
extern T1 T775f50(GE_context* ac, T0* C, T10 a1);
/* [detachable] UC_STRING.as_string_8 */
extern T0* T429f12(GE_context* ac, T0* C);
/* [detachable] UC_STRING.valid_code */
extern T1 T429f41(GE_context* ac, T0* C, T10 a1);
/* [detachable] STRING_8.as_string_8 */
extern T0* T17f27(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.make */
extern T0* T300c29(GE_context* ac, T0* a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.old_make */
extern void T300f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.make_with_name */
extern void T300f41p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.set_name */
extern void T300f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.buffered_file_info */
extern T0* T300f26(GE_context* ac, T0* C);
/* [detachable] KL_STRING_ROUTINES.as_string_no_uc_string */
extern T0* T82f4(GE_context* ac, T0* C, T0* a1);
extern T1 T82f4ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.operating_system */
extern T0* T67f139(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.generate_compilation_shell_script */
extern void T67f603(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] KL_TEXT_OUTPUT_FILE.change_mode */
extern void T300f39(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.old_change_mode */
extern void T300f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.file_chmod */
extern void T300f52(GE_context* ac, T0* C, T14 a1, T6 a2);
/* [detachable] KL_TEXT_OUTPUT_FILE.old_exists */
extern T1 T300f16(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.file_exists */
extern T1 T300f27(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.file_exists */
extern T1 T126f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.exists */
extern T1 T37f33(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.is_plain */
extern T1 T37f35(GE_context* ac, T0* C);
/* [detachable] FILE_INFO.is_plain */
extern T1 T141f13(GE_context* ac, T0* C);
/* [detachable] FILE_INFO.file_info */
extern T6 T141f14(GE_context* ac, T0* C, T14 a1, T6 a2);
/* [detachable] KL_TEXT_INPUT_FILE.buffered_file_info */
extern T0* T37f29(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.set_buffer */
extern void T37f73(GE_context* ac, T0* C);
/* [detachable] FILE_INFO.fast_update */
extern void T141f23(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] FILE_INFO.eif_file_stat */
extern T6 T141f12(GE_context* ac, T0* C, T14 a1, T14 a2, T1 a3);
/* [detachable] KL_TEXT_INPUT_FILE.internal_name_pointer */
extern T0* T37f24(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.old_exists */
extern T1 T37f15(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.file_exists */
extern T1 T37f31(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_TEXT_INPUT_FILE.old_is_closed */
extern T1 T37f23(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.string_name */
extern T0* T37f13(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.reset */
extern void T37f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.tmp_file */
extern T0* T126f22(GE_context* ac, T0* C);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.file_exists */
extern T1 T125f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.tmp_file */
extern T0* T125f24(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.remove_tail */
extern void T775f76(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.remove_tail */
extern void T429f78(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.has_extension */
extern T1 T126f10(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.has_extension */
extern T1 T125f10(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.file_system */
extern T0* T67f126(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.unix_file_system */
extern T0* T67f488(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.windows_file_system */
extern T0* T67f487(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].item */
extern T0* T304f5(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.cursor_item */
extern T0* T303f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].cursor_item */
extern T0* T216f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].item_storage_item */
extern T0* T216f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ISE_VARIABLES.ise_platform_value */
extern T0* T38f7(GE_context* ac, T0* C);
/* [detachable] KL_EXECUTION_ENVIRONMENT.variable_value */
extern T0* T86f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_EXECUTION_ENVIRONMENT.string_ */
extern T0* T86f2(GE_context* ac, T0* C);
/* [detachable] KL_EXECUTION_ENVIRONMENT.environment_impl */
extern T0* T86f1(GE_context* ac, T0* C);
/* [detachable] ET_ISE_VARIABLES.execution_environment */
extern T0* T38f3(GE_context* ac, T0* C);
/* [detachable] KL_EXECUTION_ENVIRONMENT.default_create */
extern T0* T86c16(GE_context* ac);
/* [detachable] ET_C_GENERATOR.ise_variables */
extern T0* T67f207(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].forth */
extern void T304f16(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.cursor_forth */
extern void T303f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_STRING_HASH_TABLE.add_traversing_cursor */
extern void T303f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].set_position */
extern void T304f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.remove_traversing_cursor */
extern void T303f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].cursor_forth */
extern void T216f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].add_traversing_cursor */
extern void T216f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].remove_traversing_cursor */
extern void T216f77(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].clashes_item */
extern T6 T216f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].after */
extern T1 T304f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].start */
extern void T304f15(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.cursor_start */
extern void T303f80(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_STRING_HASH_TABLE.cursor_off */
extern T1 T303f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].off */
extern T1 T304f11(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.is_empty */
extern T1 T303f53(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].cursor_start */
extern void T216f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].cursor_off */
extern T1 T216f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].is_empty */
extern T1 T216f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].back */
extern void T304f14(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.cursor_back */
extern void T303f79(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].set_before */
extern void T304f20(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].cursor_back */
extern void T216f85(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].key */
extern T0* T304f9(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.cursor_key */
extern T0* T303f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].cursor_key */
extern T0* T216f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].key_storage_item */
extern T0* T216f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].is_last */
extern T1 T304f7(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.cursor_is_last */
extern T1 T303f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].cursor_is_last */
extern T1 T216f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].before */
extern T1 T304f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].finish */
extern void T304f13(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.cursor_finish */
extern void T303f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].cursor_finish */
extern void T216f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].new_cursor */
extern T0* T216f22(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_8, [attached] STRING_8].make */
extern T0* T304c12(GE_context* ac, T0* a1);
/* [detachable] DS_STRING_HASH_TABLE.item */
extern T0* T303f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_STRING_HASH_TABLE.search_position */
extern void T303f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.add_external_c_files */
extern void T67f602(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] KL_DIRECTORY.close */
extern void T333f36(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.old_close */
extern void T333f37(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.default_pointer */
extern T14 T333f19(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.dir_close */
extern void T333f41(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_STRING_ROUTINES.replaced_all_substrings */
extern T0* T82f8(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] KL_STRING_ROUTINES.substring_index */
extern T6 T82f13(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
extern T1 T82f13ot1(T0* a1, T0** a2);
/* [detachable] STRING_8.substring_index */
extern T6 T17f31(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] STRING_8_SEARCHER.substring_index */
extern T6 T781f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
/* [detachable] STRING_8_SEARCHER.substring_index_with_deltas */
extern T6 T781f2(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
/* [detachable] STRING_8_SEARCHER.internal_initialize_deltas */
extern void T781f8(GE_context* ac, T0* C, T0* a1, T6 a2, T0* a3);
/* [detachable] STRING_8.string_searcher */
extern T0* T17f41(GE_context* ac, T0* C);
/* [detachable] STRING_8_SEARCHER.make */
extern T0* T781c6(GE_context* ac);
/* [detachable] UC_UTF8_STRING.substring_index */
extern T6 T775f33(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] UC_STRING.substring_index */
extern T6 T429f59(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_TEXT_OUTPUT_FILE.put_character */
extern void T300f32(GE_context* ac, T0* C, T2 a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.old_put_character */
extern void T300f44(GE_context* ac, T0* C, T2 a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.file_pc */
extern void T300f51(GE_context* ac, T0* C, T14 a1, T2 a2);
/* [detachable] ET_C_GENERATOR.multithreaded_mode */
extern T1 T67f275(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].force_last */
extern void T216f51(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].key_storage_put */
extern void T216f64(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].slots_put */
extern void T216f63(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].clashes_put */
extern void T216f62(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].slots_item */
extern T6 T216f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].hash_position */
extern T6 T216f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_AGENT_HASH_FUNCTION [[attached] STRING_8].hash_code */
extern T6 T215f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].resize */
extern void T216f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].clashes_resize */
extern void T216f76(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] INTEGER_32].aliased_resized_area_with_default */
extern T0* T379f1(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] INTEGER_32].aliased_resized_area_with_default */
extern T0* T128f4(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].special_integer_ */
extern T0* T216f41(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] INTEGER_32].default_create */
extern T0* T379c6(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].key_storage_resize */
extern void T216f75(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].item_storage_resize */
extern void T216f74(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].slots_resize */
extern void T216f73(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].new_modulus */
extern T6 T216f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].new_capacity */
extern T6 T216f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].item_storage_put */
extern void T216f60(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].search_position */
extern void T216f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_STRING_EQUALITY_TESTER.test */
extern T1 T896f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] UC_STRING_EQUALITY_TESTER.string_ */
extern T0* T896f2(GE_context* ac, T0* C);
/* [detachable] KL_UNIX_FILE_SYSTEM.extension */
extern T0* T126f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.extension */
extern T0* T125f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_DIRECTORY.read_entry */
extern void T333f35(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.last_entry_8 */
extern T0* T333f17(GE_context* ac, T0* C);
/* [detachable] FILE_INFO.pointer_to_file_name_8 */
extern T0* T141f15(GE_context* ac, T0* C, T14 a1);
/* detachable C_STRING.string_8 */
extern T0* T30f5(GE_context* ac, T0* C);
/* detachable C_STRING.c_strlen */
extern T6 T30f6(GE_context* ac, T0* C, T14 a1);
/* detachable C_STRING.make_shared_from_pointer */
extern T0* T30c14(GE_context* ac, T14 a1);
/* [detachable] FILE_INFO.utf_16_to_multi_byte */
extern T6 T141f19(GE_context* ac, T0* C, T14 a1, T14 a2, T6 a3);
/* [detachable] KL_DIRECTORY.file_info */
extern T0* T333f26(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.last_entry_32 */
extern T0* T333f16(GE_context* ac, T0* C);
/* [detachable] FILE_INFO.pointer_to_file_name_32 */
extern T0* T141f16(GE_context* ac, T0* C, T14 a1);
/* [detachable] FILE_INFO.pointer_length_in_bytes */
extern T6 T141f20(GE_context* ac, T0* C, T14 a1);
/* [detachable] FILE_INFO.c_pointer_length_in_bytes */
extern T11 T141f21(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_DIRECTORY.readentry */
extern void T333f40(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.eif_dir_next */
extern T14 T333f24(GE_context* ac, T0* C, T14 a1);
/* [detachable] ET_C_GENERATOR.report_cannot_read_error */
extern void T67f675(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_DIRECTORY.is_open_read */
extern T1 T333f23(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.is_closed */
extern T1 T333f15(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.open_read */
extern void T333f34(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.old_open_read */
extern void T333f39(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.dir_open */
extern T14 T333f21(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_DIRECTORY.internal_name_pointer */
extern T0* T333f20(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.old_is_readable */
extern T1 T333f14(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.eif_dir_is_readable */
extern T1 T333f30(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_DIRECTORY.old_exists */
extern T1 T333f13(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.eif_dir_exists */
extern T1 T333f29(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_DIRECTORY.make */
extern void T333f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_DIRECTORY.make */
extern T0* T333c33(GE_context* ac, T0* a1);
/* [detachable] KL_DIRECTORY.old_make */
extern void T333f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_DIRECTORY.make_with_name */
extern void T333f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_DIRECTORY.set_name */
extern void T333f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_DIRECTORY.string_ */
extern T0* T333f25(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.captured_substring */
extern T0* T81f79(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.captured_end_position */
extern T6 T81f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.captured_start_position */
extern T6 T81f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.recognizes */
extern T1 T81f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.has_matched */
extern T1 T81f236(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].set_key_equality_tester */
extern void T216f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.string_equality_tester */
extern T0* T67f118(GE_context* ac, T0* C);
/* [detachable] KL_STRING_EQUALITY_TESTER.default_create */
extern T0* T153c2(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].make_map */
extern void T216f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].make_map */
extern T0* T216c48(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].make_with_equality_testers */
extern void T216f57(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].make_sparse_container */
extern void T216f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].set_internal_cursor */
extern void T216f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].make_slots */
extern void T216f82(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].make_clashes */
extern void T216f81(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].make_key_storage */
extern void T216f80(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].make_item_storage */
extern void T216f79(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.found_item */
extern T0* T303f31(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.found */
extern T1 T303f25(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.search */
extern void T303f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_STRING_HASH_TABLE.force */
extern void T303f56(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_STRING_HASH_TABLE.key_storage_put */
extern void T303f67(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_STRING_HASH_TABLE.slots_put */
extern void T303f66(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_STRING_HASH_TABLE.clashes_put */
extern void T303f65(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_STRING_HASH_TABLE.resize */
extern void T303f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.clashes_resize */
extern void T303f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.special_integer_ */
extern T0* T303f48(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.key_storage_resize */
extern void T303f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.item_storage_resize */
extern void T303f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.slots_resize */
extern void T303f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.new_modulus */
extern T6 T303f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.new_capacity */
extern T6 T303f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.item_storage_put */
extern void T303f63(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_STRING_HASH_TABLE.unset_found_item */
extern void T303f62(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.starts_with */
extern T1 T775f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_STRING.starts_with */
extern T1 T429f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_EXECUTION_ENVIRONMENT.interpreted_string */
extern T0* T86f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_EXECUTION_ENVIRONMENT.expanded_string */
extern T0* T86f5(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.replace_all */
extern T0* T81f232(GE_context* ac, T0* C, T0* a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.append_replace_all_to_string */
extern void T81f313(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_unbounded_substring */
extern void T81f327(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_substring */
extern void T81f325(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_it */
extern void T81f340(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_start */
extern T1 T81f151(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_internal */
extern T6 T81f237(GE_context* ac, T0* C, T6 a1, T1 a2, T1 a3);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_repeated_type */
extern T6 T81f303(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3, T1 a4);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_not_repeated_characters */
extern T6 T81f302(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3, T1 a4);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.infinity */
extern T6 T81f59(GE_context* ac, T0* C);
/* [detachable] KL_PLATFORM.maximum_integer */
extern T6 T477f1(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.platform */
extern T0* T81f194(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_repeated_characters */
extern T6 T81f301(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3, T1 a4);
/* [detachable] RX_BYTE_CODE.character_item */
extern T10 T422f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_repeated_classes */
extern T6 T81f300(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_BYTE_CODE.character_set_has */
extern T1 T422f5(GE_context* ac, T0* C, T6 a1, T10 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET].item */
extern T0* T765f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_repeated_refs */
extern T6 T81f299(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_ref */
extern T6 T81f305(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.space_set */
extern T0* T81f152(GE_context* ac, T0* C);
/* [detachable] RX_CHARACTER_SET.make */
extern T0* T423c16(GE_context* ac, T0* a1);
/* [detachable] RX_CHARACTER_SET.add_string */
extern void T423f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] RX_CHARACTER_SET.add_character */
extern void T423f14(GE_context* ac, T0* C, T10 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].force */
extern void T768f60(GE_context* ac, T0* C, T11 a1, T10 a2);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].key_storage_put */
extern void T768f78(GE_context* ac, T0* C, T10 a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_32].force */
extern void T766f6(GE_context* ac, T0* C, T0* a1, T10 a2, T6 a3);
/* [detachable] SPECIAL [[attached] NATURAL_32].fill_with */
extern void T764f9(GE_context* ac, T0* C, T10 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] NATURAL_32].extend */
extern void T764f11(GE_context* ac, T0* C, T10 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].slots_put */
extern void T768f77(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].clashes_put */
extern void T768f76(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].slots_item */
extern T6 T768f26(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].clashes_item */
extern T6 T768f29(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].hash_position */
extern T6 T768f28(GE_context* ac, T0* C, T10 a1);
/* NATURAL_32.hash_code */
extern T6 T10f29(GE_context* ac, T10* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].resize */
extern void T768f75(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].clashes_resize */
extern void T768f93(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].special_integer_ */
extern T0* T768f49(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].key_storage_resize */
extern void T768f92(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_32].aliased_resized_area */
extern T0* T766f3(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] NATURAL_32].aliased_resized_area */
extern T0* T764f5(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].item_storage_resize */
extern void T768f91(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_64].aliased_resized_area */
extern T0* T1443f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] NATURAL_64].aliased_resized_area */
extern T0* T767f5(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].key_storage_item */
extern T10 T768f24(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].slots_resize */
extern void T768f90(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].new_modulus */
extern T6 T768f47(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].new_capacity */
extern T6 T768f37(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].item_storage_put */
extern void T768f65(GE_context* ac, T0* C, T11 a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_64].force */
extern void T1443f4(GE_context* ac, T0* C, T0* a1, T11 a2, T6 a3);
/* [detachable] SPECIAL [[attached] NATURAL_64].fill_with */
extern void T767f12(GE_context* ac, T0* C, T11 a1, T6 a2, T6 a3);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].search_position */
extern void T768f61(GE_context* ac, T0* C, T10 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].unset_found_item */
extern void T768f64(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].value */
extern T11 T768f34(GE_context* ac, T0* C, T10 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].item_storage_item */
extern T11 T768f38(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].position_of_key */
extern T6 T768f22(GE_context* ac, T0* C, T10 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].has */
extern T1 T768f33(GE_context* ac, T0* C, T10 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].make_map */
extern T0* T768c50(GE_context* ac, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].make_with_equality_testers */
extern void T768f62(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].make_sparse_container */
extern void T768f83(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].set_internal_cursor */
extern void T768f69(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].new_cursor */
extern T0* T768f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32].make */
extern T0* T769c10(GE_context* ac, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].make_slots */
extern void T768f97(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].make_clashes */
extern void T768f96(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].make_key_storage */
extern void T768f95(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_32].make */
extern T0* T766f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] NATURAL_32].make_empty */
extern void T764f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] NATURAL_32].make_empty */
extern T0* T764c10(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_32].default_create */
extern T0* T766c5(GE_context* ac);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].make_item_storage */
extern void T768f94(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_64].make */
extern T0* T1443f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_64].default_create */
extern T0* T1443c3(GE_context* ac);
/* [detachable] RX_CHARACTER_SET.masks */
extern T0* T423f9(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] NATURAL_64].make_filled */
extern T0* T767c9(GE_context* ac, T11 a1, T6 a2);
/* [detachable] RX_CHARACTER_SET.make_empty */
extern void T423f11(GE_context* ac, T0* C);
/* [detachable] RX_CHARACTER_SET.make_empty */
extern T0* T423c11(GE_context* ac);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.digit_set */
extern T0* T81f167(GE_context* ac, T0* C);
/* [detachable] RX_CHARACTER_SET.has */
extern T1 T423f10(GE_context* ac, T0* C, T10 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].item */
extern T11 T768f36(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_ims_options */
extern void T81f338(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.is_option_dotall */
extern T1 T81f86(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.is_option_multiline */
extern T1 T81f85(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.is_option_caseless */
extern T1 T81f84(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_match_count */
extern void T81f348(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_next_start */
extern void T81f347(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_recursive */
extern T6 T81f298(GE_context* ac, T0* C, T6 a1, T1 a2, T1 a3);
/* [detachable] RX_BYTE_CODE.integer_item */
extern T10 T422f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.next_matching_alternate */
extern T6 T81f296(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match_additional_bracket */
extern T6 T81f277(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] RX_BYTE_CODE.opcode_item */
extern T10 T422f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.special_integer_ */
extern T0* T81f54(GE_context* ac, T0* C);
/* [detachable] RX_CASE_MAPPING.flip_case */
extern T10 T424f4(GE_context* ac, T0* C, T10 a1);
/* NATURAL_32.to_integer_64 */
extern T7 T10f11(GE_context* ac, T10* C);
/* [detachable] RX_CASE_MAPPING.to_lower */
extern T10 T424f3(GE_context* ac, T0* C, T10 a1);
/* INTEGER_64.to_natural_32 */
extern T10 T7f6(GE_context* ac, T7* C);
/* INTEGER_64.is_greater_equal */
extern T1 T7f5(GE_context* ac, T7* C, T7 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_subject */
extern void T81f345(GE_context* ac, T0* C, T0* a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.append_replacement_to_string */
extern void T81f326(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.append_captured_substring_to_string */
extern void T81f341(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.string_ */
extern T0* T81f70(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.execution_environment */
extern T0* T67f138(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.string_ */
extern T0* T67f137(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.match */
extern void T81f312(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].item */
extern T0* T147f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.current_system */
extern T0* T67f123(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_case_insensitive */
extern void T81f311(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_caseless */
extern void T81f324(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.compile */
extern void T81f310(GE_context* ac, T0* C, T0* a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.compile */
extern void T81f310p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_startline */
extern void T81f323(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.has_startline */
extern T1 T81f235(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.first_significant_code */
extern T10 T81f189(GE_context* ac, T0* C, T10 a1, T10 a2, T1 a3);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.find_firstchar */
extern T10 T81f234(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_anchored */
extern void T81f322(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.can_anchored */
extern T1 T81f233(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.ims_options */
extern T10 T81f149(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_option_dotall */
extern T10 T81f132(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_option_multiline */
extern T10 T81f130(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_option_caseless */
extern T10 T81f128(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.compile_regexp */
extern void T81f321(GE_context* ac, T0* C, T10 a1, T1 a2, T1 a3, T7 a4);
/* [detachable] RX_BYTE_CODE.set_count */
extern void T422f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.find_fixed_code_length */
extern T6 T81f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_BYTE_CODE.put_integer */
extern void T422f14(GE_context* ac, T0* C, T10 a1, T6 a2);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.compile_branch */
extern void T81f339(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.meta_set */
extern T0* T81f148(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.maxlit */
extern T6 T81f147(GE_context* ac, T0* C);
/* [detachable] RX_BYTE_CODE.append_character */
extern void T422f16(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_BYTE_CODE.put_character */
extern void T422f22(GE_context* ac, T0* C, T10 a1, T6 a2);
/* [detachable] RX_BYTE_CODE.resize_byte_code */
extern void T422f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_32].aliased_resized_area_with_default */
extern T0* T766f2(GE_context* ac, T0* C, T0* a1, T10 a2, T6 a3);
/* [detachable] SPECIAL [[attached] NATURAL_32].aliased_resized_area_with_default */
extern T0* T764f4(GE_context* ac, T0* C, T10 a1, T6 a2);
/* [detachable] RX_BYTE_CODE.special_natural_32_ */
extern T0* T422f9(GE_context* ac, T0* C);
/* INTEGER_64.is_less_equal */
extern T1 T7f10(GE_context* ac, T7* C, T7 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.scan_escape */
extern T7 T81f141(GE_context* ac, T0* C, T6 a1, T1 a2);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.scan_hex_number */
extern T10 T81f224(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.xdigit_set */
extern T0* T81f249(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.scan_octal_number */
extern T10 T81f223(GE_context* ac, T0* C, T6 a1);
/* INTEGER_64.to_integer_32 */
extern T6 T7f14(GE_context* ac, T7* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.escape_character */
extern T7 T81f218(GE_context* ac, T0* C, T10 a1);
/* INTEGER_64.is_greater */
extern T1 T7f7(GE_context* ac, T7* C, T7 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.to_option_ims */
extern T10 T81f135(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_ichanged */
extern void T81f337(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.unset_option_dotall */
extern T10 T81f133(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.unset_option_multiline */
extern T10 T81f131(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.unset_option_caseless */
extern T10 T81f129(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.scan_decimal_number */
extern T10 T81f110(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.compile_repeats */
extern void T81f344(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3, T7 a4, T6 a5);
/* [detachable] RX_BYTE_CODE.append_subcopy */
extern void T422f20(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] RX_BYTE_CODE.put_opcode */
extern void T422f19(GE_context* ac, T0* C, T10 a1, T6 a2);
/* [detachable] RX_BYTE_CODE.move_right */
extern void T422f18(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.compile_single_repeat */
extern void T81f346(GE_context* ac, T0* C, T10 a1, T6 a2, T6 a3, T6 a4, T10 a5, T10 a6);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.compile_character_class */
extern void T81f343(GE_context* ac, T0* C);
/* [detachable] RX_BYTE_CODE.append_character_set */
extern void T422f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET].force_last */
extern void T765f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] RX_CHARACTER_SET].force */
extern void T1435f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] RX_CHARACTER_SET].fill_with */
extern void T428f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] RX_CHARACTER_SET].extend */
extern void T428f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET].resize */
extern void T765f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] RX_CHARACTER_SET].aliased_resized_area */
extern T0* T1435f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] RX_CHARACTER_SET].aliased_resized_area */
extern T0* T428f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET].new_capacity */
extern T6 T765f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET].extendible */
extern T1 T765f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_CHARACTER_SET.set_negated */
extern void T423f15(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_CHARACTER_SET.add_set */
extern void T423f13(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].wipe_out */
extern void T768f59(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].slots_wipe_out */
extern void T768f82(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].clashes_wipe_out */
extern void T768f81(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].key_storage_wipe_out */
extern void T768f80(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] NATURAL_32].keep_head */
extern void T764f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] NATURAL_32].set_count */
extern void T764f14(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].item_storage_wipe_out */
extern void T768f79(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] NATURAL_64].keep_head */
extern void T767f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] NATURAL_64].set_count */
extern void T767f16(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].move_all_cursors_after */
extern void T768f67(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32].set_next_cursor */
extern void T769f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32].set_after */
extern void T769f13(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32].forth */
extern void T769f11(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].cursor_forth */
extern void T768f66(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].add_traversing_cursor */
extern void T768f85(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32].set_position */
extern void T769f14(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].remove_traversing_cursor */
extern void T768f84(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].force_new */
extern void T768f58(GE_context* ac, T0* C, T11 a1, T10 a2);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].replace_found_item */
extern void T768f57(GE_context* ac, T0* C, T11 a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32].item */
extern T11 T769f7(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].cursor_item */
extern T11 T768f44(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].found_item */
extern T11 T768f32(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].found */
extern T1 T768f31(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].search */
extern void T768f56(GE_context* ac, T0* C, T10 a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32].key */
extern T10 T769f6(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].cursor_key */
extern T10 T768f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32].after */
extern T1 T769f4(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].new_iterator */
extern T0* T768f23(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32].start */
extern void T769f12(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].cursor_start */
extern void T768f63(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].cursor_off */
extern T1 T768f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_64, [attached] NATURAL_32].off */
extern T1 T769f9(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].forth */
extern void T768f53(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].remove */
extern void T768f55(GE_context* ac, T0* C, T10 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].remove_position */
extern void T768f74(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].key_storage_keep_head */
extern void T768f89(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].item_storage_keep_head */
extern void T768f88(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].move_cursors_forth */
extern void T768f87(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].internal_set_key_equality_tester */
extern void T768f86(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].key_for_iteration */
extern T10 T768f30(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].after */
extern T1 T768f25(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].cursor_after */
extern T1 T768f42(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].start */
extern void T768f51(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].copy */
extern void T768f54(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].copy */
extern void T768f54p1(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].clone_clashes */
extern void T768f73(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] INTEGER_32].resized_area */
extern T0* T128f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] INTEGER_32].copy_data */
extern void T128f17(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].clone_slots */
extern void T768f72(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].clone_key_storage */
extern void T768f71(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] NATURAL_32].resized_area */
extern T0* T764f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] NATURAL_32].copy_data */
extern void T764f13(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] SPECIAL [[attached] NATURAL_32].force */
extern void T764f16(GE_context* ac, T0* C, T10 a1, T6 a2);
/* [detachable] SPECIAL [[attached] NATURAL_32].move_data */
extern void T764f15(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] NATURAL_32].overlapping_move */
extern void T764f18(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] NATURAL_32].non_overlapping_move */
extern void T764f17(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].clone_item_storage */
extern void T768f70(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] NATURAL_64].resized_area */
extern T0* T767f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] NATURAL_64].copy_data */
extern void T767f15(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] SPECIAL [[attached] NATURAL_64].force */
extern void T767f18(GE_context* ac, T0* C, T11 a1, T6 a2);
/* [detachable] SPECIAL [[attached] NATURAL_64].move_data */
extern void T767f17(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] NATURAL_64].overlapping_move */
extern void T767f20(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] NATURAL_64].non_overlapping_move */
extern void T767f19(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].valid_cursor */
extern T1 T768f35(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].twin */
extern T0* T768f48(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].is_empty */
extern T1 T768f41(GE_context* ac, T0* C);
/* [detachable] RX_CHARACTER_SET.add_negated_set */
extern void T423f12(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].replace */
extern void T768f52(GE_context* ac, T0* C, T11 a1, T10 a2);
/* detachable DS_HASH_TABLE [[attached] NATURAL_64, [attached] NATURAL_32].item_for_iteration */
extern T11 T768f27(GE_context* ac, T0* C);
/* [detachable] ARRAY [[attached] RX_CHARACTER_SET].item */
extern T0* T427f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.class_sets */
extern T0* T81f159(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.default_word_set */
extern T0* T81f76(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.punct_set */
extern T0* T81f248(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.print_set */
extern T0* T81f247(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.graph_set */
extern T0* T81f246(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.cntrl_set */
extern T0* T81f245(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.ascii_set */
extern T0* T81f244(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.alnum_set */
extern T0* T81f243(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.upper_set */
extern T0* T81f242(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.lower_set */
extern T0* T81f241(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.alpha_set */
extern T0* T81f239(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.check_posix_name */
extern T6 T81f157(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.class_names */
extern T0* T81f240(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.check_posix_syntax */
extern T6 T81f155(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.compile_counted_repeats */
extern T1 T81f88(GE_context* ac, T0* C, T6 a1, T7 a2, T6 a3);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.scan_comment */
extern void T81f342(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.is_option_undef */
extern T1 T81f61(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_BYTE_CODE.append_integer */
extern void T422f13(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_BYTE_CODE.append_opcode */
extern void T422f12(GE_context* ac, T0* C, T10 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_error */
extern void T81f320(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_default_internal_options */
extern void T81f319(GE_context* ac, T0* C);
/* [detachable] RX_BYTE_CODE.wipe_out */
extern void T422f11(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET].wipe_out */
extern void T765f11(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] RX_CHARACTER_SET].keep_head */
extern void T428f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] RX_CHARACTER_SET].set_count */
extern void T428f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET].move_all_cursors_after */
extern void T765f14(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] RX_CHARACTER_SET].set_next_cursor */
extern void T1436f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] RX_CHARACTER_SET].set_after */
extern void T1436f6(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.wipe_out */
extern void T81f318(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.wipe_out */
extern void T81f318p1(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.make */
extern T0* T81c309(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] INTEGER_32].make_filled */
extern T0* T379f2(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.make */
extern void T81f309p1(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_default_options */
extern void T81f317(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_strict */
extern void T81f336(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_greedy */
extern void T81f335(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_eol */
extern void T81f334(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_bol */
extern void T81f333(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_dollar_endonly */
extern void T81f332(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_empty_allowed */
extern void T81f331(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_extended */
extern void T81f330(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_dotall */
extern void T81f329(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_multiline */
extern void T81f328(GE_context* ac, T0* C, T1 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_word_set */
extern void T81f316(GE_context* ac, T0* C, T0* a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.set_character_case_mapping */
extern void T81f315(GE_context* ac, T0* C, T0* a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.default_character_case_mapping */
extern T0* T81f75(GE_context* ac, T0* C);
/* [detachable] RX_CASE_MAPPING.make */
extern T0* T424c6(GE_context* ac, T0* a1, T0* a2);
/* [detachable] RX_CASE_MAPPING.add */
extern void T424f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] RX_CASE_MAPPING.make_default */
extern void T424f7(GE_context* ac, T0* C);
/* [detachable] RX_CASE_MAPPING.clear */
extern void T424f9(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_32].make_filled */
extern T0* T766f1(GE_context* ac, T0* C, T10 a1, T6 a2);
/* [detachable] SPECIAL [[attached] NATURAL_32].make_filled */
extern T0* T764c7(GE_context* ac, T10 a1, T6 a2);
/* [detachable] RX_CASE_MAPPING.special_natural_32_ */
extern T0* T424f5(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.reset */
extern void T81f314(GE_context* ac, T0* C);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.reset */
extern void T81f314p1(GE_context* ac, T0* C);
/* [detachable] STRING_32.wipe_out */
extern void T18f26(GE_context* ac, T0* C);
/* [detachable] STRING_32.twin */
extern T0* T18f8(GE_context* ac, T0* C);
/* [detachable] STRING_32.copy */
extern void T18f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [CHARACTER_32].resized_area */
extern T0* T16f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.empty_pattern_buffer */
extern T0* T81f74(GE_context* ac, T0* C);
/* [detachable] RX_BYTE_CODE.make */
extern T0* T422c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET].make */
extern T0* T765c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET].set_internal_cursor */
extern void T765f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] RX_CHARACTER_SET].new_cursor */
extern T0* T765f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] RX_CHARACTER_SET].make */
extern T0* T1436c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] RX_CHARACTER_SET].make */
extern T0* T1435f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] RX_CHARACTER_SET].make_empty */
extern T0* T428c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] RX_CHARACTER_SET].default_create */
extern T0* T1435c3(GE_context* ac);
/* [detachable] ET_C_GENERATOR.c_config */
extern T0* T67f136(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.parse_file */
extern void T471f185(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_CONFIG_PARSER.parse */
extern void T471f189(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_pop_last_value */
extern void T471f202(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.yy_push_error_value */
extern void T471f201(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ANY].force */
extern void T527f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ANY].fill_with */
extern void T528f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ANY].extend */
extern void T528f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ANY].aliased_resized_area */
extern T0* T527f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ANY].aliased_resized_area */
extern T0* T528f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.yy_do_action */
extern void T471f200(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] BOOLEAN].force */
extern void T550f4(GE_context* ac, T0* C, T0* a1, T1 a2, T6 a3);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] BOOLEAN].aliased_resized_area */
extern T0* T550f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] BOOLEAN].aliased_resized_area */
extern T0* T129f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.is_defined */
extern T1 T471f140(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].has */
extern T1 T216f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].force */
extern void T216f86(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] UT_CONFIG_PARSER.process_include */
extern void T471f214(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_TOO_MANY_INCLUDES_ERROR.make */
extern T0* T1351c8(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] YY_BUFFER].put */
extern void T1349f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] YY_BUFFER].force */
extern void T1715f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] YY_BUFFER].fill_with */
extern void T1714f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] YY_BUFFER].extend */
extern void T1714f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32].put */
extern void T937f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.execution_environment */
extern T0* T471f166(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] YY_BUFFER].is_full */
extern T1 T1349f5(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.undefine_value */
extern void T471f213(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].remove */
extern void T216f87(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].remove_position */
extern void T216f88(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].key_storage_keep_head */
extern void T216f92(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].item_storage_keep_head */
extern void T216f91(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].move_cursors_forth */
extern void T216f90(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].internal_set_key_equality_tester */
extern void T216f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_CONFIG_PARSER.ignored */
extern T1 T471f137(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_push_last_value */
extern void T471f199(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.accept */
extern void T471f198(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_do_error_action */
extern void T471f196(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.report_error */
extern void T471f212(GE_context* ac, T0* C, T0* a1);
extern T1 T471f212ot1(T0* a1, T0** a2);
/* [detachable] UT_SYNTAX_ERROR.make */
extern T0* T1350c8(GE_context* ac, T0* a1, T6 a2);
/* [detachable] KL_STDIN_FILE.name */
extern T0* T931f6(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.name */
extern T0* T901f26(GE_context* ac, T0* C);
/* [detachable] KL_STRING_INPUT_STREAM.name */
extern T0* T1477f6(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.name */
extern T0* T37f4(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.report_eof_expected_error */
extern void T471f211(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.read_token */
extern void T471f195(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_execute_action */
extern void T471f210(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.yy_set_beginning_of_line */
extern void T471f219(GE_context* ac, T0* C);
/* [detachable] KL_CHARACTER_BUFFER.item */
extern T2 T1446f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] YY_FILE_BUFFER.set_beginning_of_line */
extern void T930f22(GE_context* ac, T0* C, T1 a1);
/* [detachable] YY_BUFFER.set_beginning_of_line */
extern void T932f16(GE_context* ac, T0* C, T1 a1);
/* [detachable] UT_CONFIG_PARSER.text_item */
extern T2 T471f165(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.text_substring */
extern T0* T471f159(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] KL_CHARACTER_BUFFER.substring */
extern T0* T1446f3(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UT_CONFIG_PARSER.text_count */
extern T6 T471f158(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.string_ */
extern T0* T471f154(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.text */
extern T0* T471f153(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.set_start_condition */
extern void T471f217(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.yy_execute_eof_action */
extern void T471f209(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.terminate */
extern void T471f218(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.wrap */
extern T1 T471f132(GE_context* ac, T0* C);
extern T1 T471f132ot1(T0* a1, T0** a2);
/* [detachable] KL_STDIN_FILE.close */
extern void T931f13(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.close */
extern void T901f38(GE_context* ac, T0* C);
/* [detachable] KL_STDIN_FILE.is_closable */
extern T1 T931f5(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.is_closable */
extern T1 T901f27(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.is_closable */
extern T1 T37f39(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] YY_BUFFER].remove */
extern void T1349f10(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] YY_BUFFER].keep_head */
extern void T1714f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] YY_BUFFER].set_count */
extern void T1714f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32].remove */
extern void T937f11(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] INTEGER_32].keep_head */
extern void T128f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] INTEGER_32].set_count */
extern void T128f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32].item */
extern T6 T937f5(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] YY_BUFFER].item */
extern T0* T1349f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] YY_BUFFER].is_empty */
extern T1 T1349f6(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_refill_input_buffer */
extern void T471f208(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_set_content */
extern void T471f205(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_CHARACTER_BUFFER.count */
extern T6 T1446f6(GE_context* ac, T0* C);
/* [detachable] KL_CHARACTER_BUFFER.as_unicode_special */
extern T0* T1446f5(GE_context* ac, T0* C);
/* [detachable] YY_FILE_BUFFER.fill */
extern void T930f21(GE_context* ac, T0* C);
/* [detachable] KL_CHARACTER_BUFFER.fill_from_stream */
extern T6 T1446f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] KL_STDIN_FILE.read_to_string */
extern T6 T931f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] STRING_8.subcopy */
extern void T17f84(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] CONSOLE.read_stream */
extern void T1624f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] CONSOLE.read_stream_thread_aware */
extern void T1624f34(GE_context* ac, T0* C, T6 a1);
/* detachable C_STRING.read_string_into */
extern void T30f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] CONSOLE.console_readstream */
extern T6 T1624f18(GE_context* ac, T0* C, T14 a1, T14 a2, T6 a3);
/* [detachable] CONSOLE.read_data_buffer */
extern T0* T1624f17(GE_context* ac, T0* C);
/* [detachable] KL_STDIN_FILE.any_ */
extern T0* T931f10(GE_context* ac, T0* C);
/* [detachable] CONSOLE.end_of_file */
extern T1 T1624f12(GE_context* ac, T0* C);
/* [detachable] CONSOLE.console_eof */
extern T1 T1624f14(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_STDIN_FILE.console */
extern T0* T931f7(GE_context* ac, T0* C);
/* [detachable] STD_FILES.input */
extern T0* T892f4(GE_context* ac, T0* C);
/* [detachable] CONSOLE.make_open_stdin */
extern T0* T1624c30(GE_context* ac, T0* a1);
/* [detachable] CONSOLE.set_read_mode */
extern void T1624f33(GE_context* ac, T0* C);
/* [detachable] KL_STDIN_FILE.io */
extern T0* T931f9(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.read_to_string */
extern T6 T901f28(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] XM_EIFFEL_INPUT_STREAM.read_to_string */
extern T6 T901f28p1(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] KL_STRING_INPUT_STREAM.read_to_string */
extern T6 T1477f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] KL_STRING_INPUT_STREAM.read_character */
extern void T1477f10(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.read_to_string */
extern T6 T37f38(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] KL_TEXT_INPUT_FILE.dummy_kl_character_buffer */
extern T0* T37f48(GE_context* ac, T0* C);
/* [detachable] KL_CHARACTER_BUFFER.make */
extern T0* T1446c11(GE_context* ac, T6 a1);
/* [detachable] KL_TEXT_INPUT_FILE.old_read_to_string */
extern T6 T37f37(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] KL_TEXT_INPUT_FILE.file_gss */
extern T6 T37f45(GE_context* ac, T0* C, T14 a1, T14 a2, T6 a3);
/* [detachable] SPECIAL [CHARACTER_8].item_address */
extern T14 T15f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_TEXT_INPUT_FILE.any_ */
extern T0* T37f46(GE_context* ac, T0* C);
/* [detachable] KL_CHARACTER_BUFFER.put */
extern void T1446f12(GE_context* ac, T0* C, T2 a1, T6 a2);
/* [detachable] KL_STDIN_FILE.last_character */
extern T2 T931f2(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.last_character */
extern T2 T901f19(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_QUEUE [[attached] CHARACTER_8].item */
extern T2 T1626f4(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.last_character */
extern T2 T37f6(GE_context* ac, T0* C);
/* [detachable] KL_STDIN_FILE.end_of_file */
extern T1 T931f1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.end_of_input */
extern T1 T901f18(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.end_of_input */
extern T1 T37f34(GE_context* ac, T0* C);
/* [detachable] KL_STDIN_FILE.read_character */
extern void T931f14(GE_context* ac, T0* C);
/* [detachable] CONSOLE.read_character */
extern void T1624f31(GE_context* ac, T0* C);
/* [detachable] CONSOLE.console_readchar */
extern T2 T1624f15(GE_context* ac, T0* C, T14 a1);
/* [detachable] XM_EIFFEL_INPUT_STREAM.read_character */
extern void T901f32(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_QUEUE [[attached] CHARACTER_8].remove */
extern void T1626f6(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_QUEUE [[attached] CHARACTER_8].wipe_out */
extern void T1626f8(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.noqueue_read_character */
extern void T901f34(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.utf16_read_character */
extern void T901f36(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.append_character */
extern void T901f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_LINKED_QUEUE [[attached] CHARACTER_8].force */
extern void T1626f7(GE_context* ac, T0* C, T2 a1);
/* detachable DS_LINKABLE [[attached] CHARACTER_8].put_right */
extern void T1902f4(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] CHARACTER_8].make */
extern T0* T1902c3(GE_context* ac, T2 a1);
/* [detachable] UC_UTF8_ROUTINES.append_code_to_utf8 */
extern void T434f56(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] UC_UTF8_ROUTINES.append_natural_32_code_to_utf8 */
extern void T434f55(GE_context* ac, T0* C, T0* a1, T10 a2);
/* [detachable] XM_EIFFEL_INPUT_STREAM.utf8 */
extern T0* T901f22(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.string_ */
extern T0* T901f17(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.utf8_buffer */
extern T0* T901f25(GE_context* ac, T0* C);
/* [detachable] UC_UTF16_ROUTINES.surrogate */
extern T6 T1627f7(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UC_UTF16_ROUTINES.is_low_surrogate */
extern T1 T1627f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF16_ROUTINES.least_10_bits */
extern T6 T1627f5(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UC_UTF16_ROUTINES.is_high_surrogate */
extern T1 T1627f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF16_ROUTINES.is_surrogate */
extern T1 T1627f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_INPUT_STREAM.utf16 */
extern T0* T901f20(GE_context* ac, T0* C);
/* [detachable] UC_UTF16_ROUTINES.default_create */
extern T0* T1627c16(GE_context* ac);
/* [detachable] XM_EIFFEL_INPUT_STREAM.least_significant */
extern T6 T901f24(GE_context* ac, T0* C, T2 a1, T2 a2);
/* [detachable] XM_EIFFEL_INPUT_STREAM.most_significant */
extern T6 T901f23(GE_context* ac, T0* C, T2 a1, T2 a2);
/* [detachable] XM_EIFFEL_INPUT_STREAM.latin1_read_character */
extern void T901f35(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.utf16_detect_read_character */
extern void T901f33(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_ROUTINES.is_endian_detection_character */
extern T1 T434f39(GE_context* ac, T0* C, T2 a1, T2 a2, T2 a3);
/* [detachable] UC_UTF8_ROUTINES.is_endian_detection_character_start */
extern T1 T434f40(GE_context* ac, T0* C, T2 a1, T2 a2);
/* [detachable] UC_UTF16_ROUTINES.is_endian_detection_character_least_first */
extern T1 T1627f2(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] UC_UTF16_ROUTINES.is_endian_detection_character_most_first */
extern T1 T1627f1(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] YY_FILE_BUFFER.compact_left */
extern void T930f24(GE_context* ac, T0* C);
/* [detachable] KL_CHARACTER_BUFFER.move_left */
extern void T1446f13(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] YY_FILE_BUFFER.resize */
extern void T930f26(GE_context* ac, T0* C);
/* [detachable] KL_CHARACTER_BUFFER.resize */
extern void T1446f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] YY_FILE_BUFFER.default_capacity */
extern T6 T930f13(GE_context* ac, T0* C);
/* [detachable] YY_BUFFER.fill */
extern void T932f14(GE_context* ac, T0* C);
/* [detachable] YY_FILE_BUFFER.set_index */
extern void T930f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] YY_BUFFER.set_index */
extern void T932f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.yy_null_trans_state */
extern T6 T471f131(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.yy_previous_state */
extern T6 T471f130(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.fatal_error */
extern void T471f207(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_MESSAGE.make */
extern T0* T80c8(GE_context* ac, T0* a1);
/* [detachable] UT_CONFIG_PARSER.report_invalid_unicode_character_error */
extern void T471f206(GE_context* ac, T0* C, T10 a1);
/* NATURAL_32.to_hex_string */
extern T0* T10f21(GE_context* ac, T10* C);
/* NATURAL_32.to_hex_character */
extern T2 T10f30(GE_context* ac, T10* C);
/* [detachable] KL_CHARACTER_BUFFER.item_code */
extern T10 T1446f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.yy_at_beginning_of_line */
extern T6 T471f99(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] INTEGER_32].force */
extern void T379f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] INTEGER_32].aliased_resized_area */
extern T0* T379f4(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] UT_CONFIG_PARSER.special_integer_ */
extern T0* T471f105(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_init_value_stacks */
extern void T471f194(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_clear_all */
extern void T471f203(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.clear_all */
extern void T471f215(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.clear_stacks */
extern void T471f220(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_clear_value_stacks */
extern void T471f221(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] BOOLEAN].keep_head */
extern void T129f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] BOOLEAN].set_count */
extern void T129f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ANY].keep_head */
extern void T528f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ANY].set_count */
extern void T528f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] UT_CONFIG_PARSER.abort */
extern void T471f197(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.set_input_buffer */
extern void T471f188(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_CONFIG_PARSER.yy_load_input_buffer */
extern void T471f193(GE_context* ac, T0* C);
/* [detachable] YY_FILE_BUFFER.set_position */
extern void T930f20(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] YY_BUFFER.set_position */
extern void T932f15(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] UT_CONFIG_PARSER.new_file_buffer */
extern T0* T471f104(GE_context* ac, T0* C, T0* a1);
/* [detachable] YY_FILE_BUFFER.make */
extern T0* T930c18(GE_context* ac, T0* a1);
/* [detachable] YY_FILE_BUFFER.make_with_size */
extern void T930f17(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] YY_FILE_BUFFER.make_with_size */
extern T0* T930c17(GE_context* ac, T0* a1, T6 a2);
/* [detachable] YY_FILE_BUFFER.set_file */
extern void T930f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] YY_FILE_BUFFER.set_file_with_size */
extern void T930f25(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] YY_FILE_BUFFER.flush */
extern void T930f27(GE_context* ac, T0* C);
/* [detachable] YY_FILE_BUFFER.new_default_buffer */
extern T0* T930f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.console_application_mode */
extern T1 T67f491(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.define_value */
extern void T471f184(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] UT_CONFIG_PARSER.make */
extern T0* T471c183(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32].make */
extern T0* T937c8(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] INTEGER_32].make */
extern T0* T379f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] YY_BUFFER].make */
extern T0* T1349c8(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] YY_BUFFER].make */
extern T0* T1715f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] YY_BUFFER].make_empty */
extern T0* T1714c4(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] YY_BUFFER].default_create */
extern T0* T1715c2(GE_context* ac);
/* [detachable] UT_CONFIG_PARSER.string_equality_tester */
extern T0* T471f97(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.make_parser_skeleton */
extern void T471f187(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_build_parser_tables */
extern void T471f192(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yycheck_template */
extern T0* T471f82(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yyfixed_array */
extern T0* T471f178(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] INTEGER_32].to_special */
extern T0* T379f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_CONFIG_PARSER.yytable_template */
extern T0* T471f139(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yypgoto_template */
extern T0* T471f138(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yypact_template */
extern T0* T471f136(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yydefgoto_template */
extern T0* T471f133(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yydefact_template */
extern T0* T471f127(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yytypes2_template */
extern T0* T471f126(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yytypes1_template */
extern T0* T471f123(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yyr1_template */
extern T0* T471f119(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yytranslate_template */
extern T0* T471f116(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yytranslate_template_2 */
extern void T471f223(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_CONFIG_PARSER.yyarray_subcopy */
extern void T471f225(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] KL_ARRAY_ROUTINES [[attached] INTEGER_32].subcopy */
extern void T902f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] ARRAY [[attached] INTEGER_32].subcopy */
extern void T291f17(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] UT_CONFIG_PARSER.integer_array_ */
extern T0* T471f182(GE_context* ac, T0* C);
/* [detachable] KL_ARRAY_ROUTINES [[attached] INTEGER_32].default_create */
extern T0* T902c2(GE_context* ac);
/* [detachable] UT_CONFIG_PARSER.yytranslate_template_1 */
extern void T471f222(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_CONFIG_PARSER.yy_create_value_stacks */
extern void T471f191(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] BOOLEAN].make */
extern T0* T550f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] BOOLEAN].default_create */
extern T0* T550c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ANY].make */
extern T0* T527f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ANY].make_empty */
extern T0* T528c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ANY].default_create */
extern T0* T527c3(GE_context* ac);
/* [detachable] UT_CONFIG_PARSER.make_scanner */
extern void T471f186(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.make_with_buffer */
extern void T471f190(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_CONFIG_PARSER.yy_initialize */
extern void T471f204(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_initialize */
extern void T471f204p1(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_build_tables */
extern void T471f216(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_accept_template */
extern T0* T471f173(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_fixed_array */
extern T0* T471f181(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_CONFIG_PARSER.yy_meta_template */
extern T0* T471f172(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_ec_template */
extern T0* T471f171(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_ec_template_1 */
extern void T471f224(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_CONFIG_PARSER.yy_array_subcopy */
extern void T471f226(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] UT_CONFIG_PARSER.yy_def_template */
extern T0* T471f170(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_base_template */
extern T0* T471f169(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_chk_template */
extern T0* T471f168(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.yy_nxt_template */
extern T0* T471f167(GE_context* ac, T0* C);
/* [detachable] UT_CONFIG_PARSER.empty_buffer */
extern T0* T471f83(GE_context* ac, T0* C);
/* [detachable] YY_BUFFER.make */
extern T0* T932c12(GE_context* ac, T0* a1);
/* [detachable] YY_BUFFER.make_from_buffer */
extern void T932f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_CHARACTER_BUFFER.fill_from_string */
extern void T1446f15(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_CHARACTER_BUFFER.fill_from_substring */
extern void T1446f17(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] YY_BUFFER.new_default_buffer */
extern T0* T932f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.put_new */
extern void T303f59(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_STRING_HASH_TABLE.set_key_equality_tester */
extern void T303f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_STRING_HASH_TABLE.make_map */
extern T0* T303c55(GE_context* ac, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.make_with_equality_testers */
extern void T303f61(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_STRING_HASH_TABLE.make_sparse_container */
extern void T303f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.set_internal_cursor */
extern void T303f77(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_STRING_HASH_TABLE.new_cursor */
extern T0* T303f47(GE_context* ac, T0* C);
/* [detachable] DS_STRING_HASH_TABLE.make_slots */
extern void T303f76(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.make_clashes */
extern void T303f75(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.make_key_storage */
extern void T303f74(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_STRING_HASH_TABLE.make_item_storage */
extern void T303f73(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_STRING_ROUTINES.right_adjust */
extern void T82f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STRING_ROUTINES.left_adjust */
extern void T82f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.remove_head */
extern void T775f85(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UTF8_STRING.keep_tail */
extern void T775f104(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.remove_head */
extern void T429f107(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.keep_tail */
extern void T429f110(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.nested_pathname */
extern T0* T126f3(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.nested_pathname */
extern T0* T125f3(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.generate_c_code */
extern void T67f564(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.dummy_feature */
extern T0* T67f120(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.make */
extern T0* T236c45(GE_context* ac, T0* a1, T0* a2, T0* a3);
extern T1 T236f45ot1(T0* a1, T0** a2);
/* detachable ET_EXTERNAL_PROCEDURE.has_seed */
extern T1 T1516f45(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FEATURE_IDS.has */
extern T1 T604f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EXTERNAL_FUNCTION.has_seed */
extern T1 T253f61(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_PROCEDURE.has_seed */
extern T1 T1515f45(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DEFERRED_FUNCTION.has_seed */
extern T1 T1513f45(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_FUNCTION.has_seed */
extern T1 T1512f47(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_FUNCTION.has_seed */
extern T1 T1511f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.has_seed */
extern T1 T371f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.has_seed */
extern T1 T369f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.has_seed */
extern T1 T359f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ATTRIBUTE.has_seed */
extern T1 T360f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DEFERRED_PROCEDURE.has_seed */
extern T1 T476f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DO_PROCEDURE.has_seed */
extern T1 T469f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_SYSTEM.set_ise_runtime_set_reference_field_feature */
extern void T63f108(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.use_all_attributes */
extern void T63f109(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.use_all_attributes */
extern void T451f76(GE_context* ac, T0* C, T0* a1);
extern T1 T451f76ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.is_builtin_attribute */
extern T1 T451f41(GE_context* ac, T0* C, T0* a1, T8 a2, T8 a3);
/* [detachable] ET_EXTERNAL_FUNCTION.is_builtin */
extern T1 T253f62(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.tokens */
extern T0* T253f52(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.default_create */
extern T0* T73c1545(GE_context* ac);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.dynamic_query */
extern T0* T451f29(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.put_function */
extern void T451f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.is_basic */
extern T1 T451f31(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_basic */
extern T1 T74f182(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_CODES.is_basic */
extern T1 T414f34(GE_context* ac, T0* C, T8 a1);
/* [detachable] ET_CLASS.class_codes */
extern T0* T74f123(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_CODES.default_create */
extern T0* T414c35(GE_context* ac);
/* detachable ET_DYNAMIC_PRECURSOR.is_builtin_any_is_equal */
extern T1 T306f31(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.is_builtin_any_is_equal */
extern T1 T236f21(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.is_once_per_object */
extern T1 T306f22(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_once_per_object */
extern T1 T1516f54(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_once_per_object */
extern T1 T476f47(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_once_per_object */
extern T1 T1515f31(GE_context* ac, T0* C);
/* [detachable] ET_STANDARD_ONCE_KEYS.has_object_key */
extern T1 T1857f1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_STANDARD_ONCE_KEYS.has_expected_key */
extern T1 T1857f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_STANDARD_ONCE_KEYS.is_expected_key */
extern T1 T1857f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_VERBATIM_STRING.value */
extern T0* T1485f1(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.value */
extern T0* T1484f1(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.value */
extern T0* T1483f1(GE_context* ac, T0* C);
/* [detachable] ET_STANDARD_ONCE_KEYS.string_ */
extern T0* T1857f9(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_STRING_LIST.manifest_string */
extern T0* T1224f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_VERBATIM_STRING.manifest_string */
extern T0* T1485f27(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.manifest_string */
extern T0* T1484f10(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.manifest_string */
extern T0* T1483f15(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_STRING_LIST.item */
extern T0* T1224f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_PROCEDURE.standard_once_keys */
extern T0* T1515f32(GE_context* ac, T0* C);
/* [detachable] ET_STANDARD_ONCE_KEYS.default_create */
extern T0* T1857c13(GE_context* ac);
/* [detachable] ET_DO_PROCEDURE.is_once_per_object */
extern T1 T469f47(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_once_per_object */
extern T1 T253f63(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_once_per_object */
extern T1 T1513f34(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_once_per_object */
extern T1 T1512f34(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.standard_once_keys */
extern T0* T1512f35(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_once_per_object */
extern T1 T1511f34(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_once_per_object */
extern T1 T371f52(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_once_per_object */
extern T1 T369f51(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_once_per_object */
extern T1 T359f48(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_once_per_object */
extern T1 T360f49(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.is_once_per_object */
extern T1 T236f37(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.force_last */
extern void T224f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_FEATURE].force */
extern void T606f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_FEATURE].fill_with */
extern void T605f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_FEATURE].extend */
extern void T605f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.fixed_array */
extern T0* T224f5(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_FEATURE].default_create */
extern T0* T606c3(GE_context* ac);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.resize */
extern void T224f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_FEATURE].aliased_resized_area */
extern T0* T606f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_FEATURE].aliased_resized_area */
extern T0* T605f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.new_capacity */
extern T6 T224f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.capacity */
extern T6 T224f6(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.put_attribute */
extern void T451f81(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.propagate_reference_field_dynamic_types */
extern void T230f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_reference_field_dynamic_types */
extern void T71f127(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_giaaa_error */
extern void T53f193(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_internal_error */
extern void T53f203(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.set_has_internal_error */
extern void T53f206(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_INTERNAL_ERROR.make_giaaa */
extern T0* T784c13(GE_context* ac);
/* [detachable] ET_ERROR_HANDLER.report_giaaa_error */
extern void T36f193(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_internal_error */
extern void T36f203(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.set_has_internal_error */
extern void T36f206(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.error_handler */
extern T0* T71f92(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.set_fatal_error */
extern void T71f115(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_builtin_result_dynamic_types */
extern void T71f125(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_builtin_dynamic_types */
extern void T71f126(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.put_target */
extern void T1466f19(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.put_target */
extern void T269f24(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.put_target */
extern void T475f18(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_OBJECT_EQUALITY_EXPRESSION.propagate_can_be_void */
extern void T2043f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_EQUALITY_EXPRESSION.propagate_can_be_void */
extern void T2042f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.propagate_can_be_void */
extern void T314f20(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.propagate_can_be_void */
extern void T318f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.propagate_can_be_void */
extern void T1466f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.propagate_can_be_void */
extern void T269f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.propagate_can_be_void */
extern void T475f19(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_TARGET_LIST.item */
extern T0* T1353f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.propagate_can_be_void */
extern void T451f79(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.propagate_can_be_void */
extern void T450f79(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.propagate_can_be_void */
extern void T302f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.propagate_can_be_void */
extern void T319f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.propagate_can_be_void */
extern void T223f79(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.propagate_can_be_void */
extern void T323f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_OBJECT_EQUALITY_EXPRESSION.put_type_from_type_set */
extern void T2043f8(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_OBJECT_EQUALITY_EXPRESSION.propagate_type */
extern void T2043f10(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.propagate_is_equal_argument_type */
extern void T230f30(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_is_equal_argument_type */
extern void T71f565(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_builtin_actual_argument_dynamic_types */
extern void T71f172(GE_context* ac, T0* C, T0* a1, T6 a2, T0* a3);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.conforms_to_primary_type */
extern T1 T451f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].has */
extern T1 T275f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].search_position */
extern void T275f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].hash_position */
extern T6 T275f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.hash_code */
extern T6 T451f20(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.hash_code */
extern T6 T450f19(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.hash_code */
extern T6 T302f18(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.hash_code */
extern T6 T319f18(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.hash_code */
extern T6 T223f16(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_item */
extern T0* T275f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_item */
extern T0* T275f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_item */
extern T6 T275f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].key_equality_tester */
extern T0* T275f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].slots_item */
extern T6 T275f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS.is_none */
extern T1 T74f179(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.is_none */
extern T1 T802f41(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.is_none */
extern T1 T887f7(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.is_none */
extern T1 T410f11(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.conforms_to_primary_type */
extern T1 T450f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.conforms_to_primary_type */
extern T1 T302f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.conforms_to_primary_type */
extern T1 T319f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.conforms_to_primary_type */
extern T1 T223f34(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_PRECURSOR.set_regular */
extern void T306f51(GE_context* ac, T0* C, T1 a1);
/* detachable ET_DYNAMIC_FEATURE.set_regular */
extern void T236f46(GE_context* ac, T0* C, T1 a1);
/* detachable ET_DYNAMIC_PRECURSOR_LIST.item */
extern T0* T307f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.error_handler */
extern T0* T230f12(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.error_handler */
extern T0* T63f76(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.set_fatal_error */
extern void T230f27(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.seeded_dynamic_query */
extern T0* T451f65(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_CLASS.seeded_query */
extern T0* T74f180(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_QUERY_LIST.seeded_feature */
extern T0* T257f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.seeded_dynamic_query */
extern T0* T450f27(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.put_function */
extern void T450f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.put_attribute */
extern void T450f81(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.register_feature_seeds */
extern void T450f80(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_FEATURE_IDS.item */
extern T6 T604f6(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_PROCEDURE.other_seeds */
extern T0* T1516f5(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.other_seeds */
extern T0* T253f5(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.other_seeds */
extern T0* T1515f5(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.other_seeds */
extern T0* T1513f5(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.other_seeds */
extern T0* T1512f5(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.other_seeds */
extern T0* T1511f5(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.other_seeds */
extern T0* T371f7(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.other_seeds */
extern T0* T369f6(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.other_seeds */
extern T0* T359f5(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.other_seeds */
extern T0* T360f5(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.other_seeds */
extern T0* T476f5(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.other_seeds */
extern T0* T469f5(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].force_last */
extern void T602f34(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].key_storage_put */
extern void T602f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].slots_put */
extern void T602f41(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].clashes_put */
extern void T602f40(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].slots_item */
extern T6 T602f23(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].hash_position */
extern T6 T602f26(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].resize */
extern void T602f39(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].clashes_resize */
extern void T602f47(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].special_integer_ */
extern T0* T602f31(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].key_storage_resize */
extern void T602f46(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].item_storage_resize */
extern void T602f45(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].key_storage_item */
extern T6 T602f28(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].clashes_item */
extern T6 T602f29(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].slots_resize */
extern void T602f44(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].new_modulus */
extern T6 T602f27(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].new_capacity */
extern T6 T602f32(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].item_storage_put */
extern void T602f38(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].search_position */
extern void T602f37(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].unset_found_item */
extern void T602f36(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.first_seed */
extern T6 T1516f1(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.first_seed */
extern T6 T1515f1(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.first_seed */
extern T6 T476f4(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.first_seed */
extern T6 T469f4(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.new_dynamic_query */
extern T0* T450f51(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.seeded_dynamic_query */
extern T0* T302f26(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.put_function */
extern void T302f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.put_attribute */
extern void T302f71(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.register_feature_seeds */
extern void T302f70(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.new_dynamic_query */
extern T0* T302f62(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.build_tuple_reference_item */
extern void T230f24(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.build_tuple_reference_item */
extern void T71f129(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.is_basic */
extern T1 T323f13(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.base_class */
extern T0* T323f14(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.item */
extern T0* T259f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_IDENTIFIER.same_feature_name */
extern T1 T184f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER.string_ */
extern T0* T184f52(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.reference_item_feature_name */
extern T0* T73f91(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.tokens */
extern T0* T302f29(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.is_builtin */
extern T1 T236f24(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.name */
extern T0* T253f48(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.feature_name */
extern T0* T184f32(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.name */
extern T0* T1513f37(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.name */
extern T0* T1512f37(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.name */
extern T0* T1511f37(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.name */
extern T0* T371f54(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.name */
extern T0* T369f55(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.name */
extern T0* T359f52(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.name */
extern T0* T360f51(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.new_dynamic_query */
extern T0* T302f62p1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.seeded_dynamic_query */
extern T0* T319f26(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.put_function */
extern void T319f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.put_attribute */
extern void T319f73(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.register_feature_seeds */
extern void T319f72(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.new_dynamic_query */
extern T0* T319f64(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_DYNAMIC_FEATURE.set_result_type_set */
extern void T236f50(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_FEATURE.is_builtin_special_item */
extern T1 T236f41(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.new_dynamic_query */
extern T0* T319f64p1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.seeded_dynamic_query */
extern T0* T223f28(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_function */
extern void T223f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_attribute */
extern void T223f70(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.register_feature_seeds */
extern void T223f69(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.new_dynamic_query */
extern T0* T223f29(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.is_basic */
extern T1 T450f31(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.is_basic */
extern T1 T302f37(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.is_basic */
extern T1 T319f34(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_basic */
extern T1 T223f45(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_EQUALITY_EXPRESSION.put_type_from_type_set */
extern void T2042f8(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_EQUALITY_EXPRESSION.propagate_type */
extern void T2042f10(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.put_type_from_type_set */
extern void T314f19(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.put_type_with_tuple_label */
extern void T314f22(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.propagate_tuple_label_expression_dynamic_types */
extern void T230f28(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_tuple_label_expression_dynamic_types */
extern void T71f501(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T71f501ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_tuple_label_result_dynamic_types */
extern void T71f520(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.valid_index */
extern T1 T259f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_BRACKET_SYMBOL.seed */
extern T6 T622f4(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.seed */
extern T6 T1545f3(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.seed */
extern T6 T1544f3(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.seed */
extern T6 T973f6(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.seed */
extern T6 T981f4(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.seed */
extern T6 T994f4(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.seed */
extern T6 T2020f1(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.seed */
extern T6 T184f1(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.name */
extern T0* T2048f1(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.name */
extern T0* T1041f7(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.name */
extern T0* T1038f3(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.name */
extern T0* T1547f2(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.name */
extern T0* T1543f3(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_EXPRESSION.name */
extern T0* T888f2(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.name */
extern T0* T287f8(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.put_type_with_feature */
extern void T314f21(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.put_type_with_feature */
extern void T314f21p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T314f21p1ot1(T0* a1, T0** a2);
extern T1 T314f21p1ot2(T0* a1, T0** a2);
extern T1 T314f21p1ot3(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.put_type */
extern void T1466f20(GE_context* ac, T0* C, T0* a1);
extern T1 T1466f20ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.force_last */
extern void T268f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRIMARY_TYPE].force */
extern void T416f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRIMARY_TYPE].fill_with */
extern void T415f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRIMARY_TYPE].extend */
extern void T415f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.fixed_array */
extern T0* T268f5(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRIMARY_TYPE].default_create */
extern T0* T416c3(GE_context* ac);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.hash_position */
extern T6 T268f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.modulus */
extern T6 T268f14(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.resize */
extern void T268f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.special_integer_ */
extern T0* T268f7(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRIMARY_TYPE].aliased_resized_area */
extern T0* T416f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRIMARY_TYPE].aliased_resized_area */
extern T0* T415f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.new_modulus */
extern T6 T268f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.new_capacity */
extern T6 T268f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.capacity */
extern T6 T268f10(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.put_last */
extern void T268f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.append_last */
extern void T268f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.dynamic_type */
extern T0* T1466f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.make_with_capacity */
extern T0* T268c17(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T416f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRIMARY_TYPE].make_empty */
extern T0* T415c5(GE_context* ac, T6 a1);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.has_type */
extern T1 T1466f13(GE_context* ac, T0* C, T0* a1);
extern T1 T1466f13ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.has_type */
extern T1 T1466f13p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.index_of */
extern T6 T1466f11(GE_context* ac, T0* C, T0* a1);
extern T1 T1466f11ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.index_of */
extern T6 T1466f11p1(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.index_of */
extern T6 T323f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.index_of */
extern T6 T451f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.index_of */
extern T6 T450f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.index_of */
extern T6 T302f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.index_of */
extern T6 T319f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.index_of */
extern T6 T223f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.index_of */
extern T6 T268f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.has_type */
extern T1 T268f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.has_type */
extern T1 T323f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.has_type */
extern T1 T451f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.has_type */
extern T1 T450f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.has_type */
extern T1 T302f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.has_type */
extern T1 T319f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_type */
extern T1 T223f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.dynamic_type */
extern T0* T268f6(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.dynamic_type */
extern T0* T323f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.dynamic_type */
extern T0* T451f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.dynamic_type */
extern T0* T450f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.dynamic_type */
extern T0* T302f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.dynamic_type */
extern T0* T319f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.dynamic_type */
extern T0* T223f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.count */
extern T6 T268f1(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.count */
extern T6 T323f17(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.count */
extern T6 T451f34(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.count */
extern T6 T450f34(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.count */
extern T6 T302f34(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.count */
extern T6 T319f35(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.count */
extern T6 T223f52(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.conforms_to_type */
extern T1 T323f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.is_type_attached_with_type_mark */
extern T1 T254f18(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_SYMBOL.is_detachable_mark */
extern T1 T341f6(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_question_mark */
extern T1 T341f23(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.tokens */
extern T0* T341f12(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_detachable */
extern T1 T341f22(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_detachable_mark */
extern T1 T1005f37(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_question_mark */
extern T1 T1005f14(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.tokens */
extern T0* T1005f15(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_detachable */
extern T1 T1005f13(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_detachable_mark */
extern T1 T994f72(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_question_mark */
extern T1 T994f19(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.tokens */
extern T0* T994f21(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_detachable */
extern T1 T994f18(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_detachable_mark */
extern T1 T622f11(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_question_mark */
extern T1 T622f24(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.tokens */
extern T0* T622f8(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_detachable */
extern T1 T622f23(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_detachable_mark */
extern T1 T571f12(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_detachable_mark */
extern T1 T413f7(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_question_mark */
extern T1 T413f17(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_detachable */
extern T1 T413f16(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.tokens */
extern T0* T413f4(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_detachable_mark */
extern T1 T301f5(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_attached_mark */
extern T1 T341f5(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_bang */
extern T1 T341f21(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_attached */
extern T1 T341f20(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_attached_mark */
extern T1 T1005f24(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_bang */
extern T1 T1005f12(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_attached */
extern T1 T1005f11(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_attached_mark */
extern T1 T994f22(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_bang */
extern T1 T994f17(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_attached */
extern T1 T994f16(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_attached_mark */
extern T1 T622f10(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_bang */
extern T1 T622f22(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_attached */
extern T1 T622f21(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_attached_mark */
extern T1 T571f9(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_attached_mark */
extern T1 T413f6(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_bang */
extern T1 T413f15(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_attached */
extern T1 T413f14(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_attached_mark */
extern T1 T301f2(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.is_attached */
extern T1 T254f23(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_attachment_mark */
extern T1 T571f8(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_attachment_mark */
extern T1 T994f9(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_attachment_mark */
extern T1 T622f9(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_attachment_mark */
extern T1 T1005f19(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_attachment_mark */
extern T1 T341f4(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_attachment_mark */
extern T1 T413f13(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_attachment_mark */
extern T1 T301f7(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.is_type_attached_with_type_mark */
extern T1 T72f23(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.is_expanded */
extern T1 T72f14(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_reference_mark */
extern T1 T994f11(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_reference */
extern T1 T994f74(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_reference_mark */
extern T1 T622f16(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_reference */
extern T1 T622f75(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_reference_mark */
extern T1 T1005f6(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_reference */
extern T1 T1005f17(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_reference_mark */
extern T1 T341f14(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_reference */
extern T1 T341f75(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_reference_mark */
extern T1 T571f4(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_reference */
extern T1 T571f11(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_reference_mark */
extern T1 T413f9(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_reference */
extern T1 T413f19(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_reference_mark */
extern T1 T301f3(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_expanded_mark */
extern T1 T994f10(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_expanded */
extern T1 T994f73(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_expanded_mark */
extern T1 T622f15(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_expanded */
extern T1 T622f74(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_expanded_mark */
extern T1 T1005f5(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_expanded */
extern T1 T1005f16(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_expanded_mark */
extern T1 T341f13(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_expanded */
extern T1 T341f74(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_expanded_mark */
extern T1 T571f3(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_expanded */
extern T1 T571f10(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_expanded_mark */
extern T1 T413f8(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_expanded */
extern T1 T413f18(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_expanded_mark */
extern T1 T301f1(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_expanded */
extern T1 T74f170(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.has_expanded_mark */
extern T1 T74f89(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.is_attached */
extern T1 T72f16(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_type_attached_with_type_mark */
extern T1 T74f173(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.is_attached */
extern T1 T74f90(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.current_system */
extern T0* T74f97(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.current_system */
extern T0* T187f11(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.current_system */
extern T0* T58f75(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.current_system */
extern T0* T40f6(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.implicit_attachment_type_mark */
extern T0* T187f16(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.implicit_attachment_type_mark */
extern T0* T58f82(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.implicit_attachment_type_mark */
extern T0* T40f1(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.type_mark */
extern T0* T323f3(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.type_mark */
extern T0* T451f47(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.tokens */
extern T0* T451f55(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.type_mark */
extern T0* T450f48(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.tokens */
extern T0* T450f58(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.type_mark */
extern T0* T319f50(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.tokens */
extern T0* T319f29(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.type_mark */
extern T0* T302f48(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.type_mark */
extern T0* T223f46(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.tokens */
extern T0* T223f33(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.conforms_to_type */
extern T1 T451f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.conforms_to_type */
extern T1 T450f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.conforms_to_type */
extern T1 T302f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.conforms_to_type */
extern T1 T319f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.conforms_to_type */
extern T1 T223f36(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_PRECURSOR.dynamic_type_set */
extern T0* T306f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT.index */
extern T6 T2027f3(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_CAST_EXPRESSION.index */
extern T6 T1980f1(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.index */
extern T6 T2046f1(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.index */
extern T6 T2048f2(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.index */
extern T6 T2047f3(GE_context* ac, T0* C);
/* detachable ET_AGENT_TYPED_OPEN_ARGUMENT.index */
extern T6 T1559f5(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.index */
extern T6 T1005f3(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.index */
extern T6 T372f6(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.index */
extern T6 T1489f1(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.index */
extern T6 T1488f1(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.index */
extern T6 T1487f1(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.index */
extern T6 T1486f1(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.index */
extern T6 T1485f2(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.index */
extern T6 T1484f2(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.index */
extern T6 T1482f1(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.index */
extern T6 T328f1(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.index */
extern T6 T313f1(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.index */
extern T6 T312f1(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.index */
extern T6 T311f1(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.index */
extern T6 T287f2(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.index */
extern T6 T1557f1(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.index */
extern T6 T1556f1(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.index */
extern T6 T1555f1(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.index */
extern T6 T1554f1(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.index */
extern T6 T1553f1(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.index */
extern T6 T1552f1(GE_context* ac, T0* C);
/* detachable ET_ITERATION_CURSOR.index */
extern T6 T1551f1(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.index */
extern T6 T1550f1(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.index */
extern T6 T1549f1(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.index */
extern T6 T1548f1(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.index */
extern T6 T1547f1(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.index */
extern T6 T1546f1(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.index */
extern T6 T1543f1(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.index */
extern T6 T1542f1(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.index */
extern T6 T1541f1(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.index */
extern T6 T1540f1(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.index */
extern T6 T1539f1(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.index */
extern T6 T1538f1(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.index */
extern T6 T1496f1(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.index */
extern T6 T1495f1(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.index */
extern T6 T1494f1(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.index */
extern T6 T1493f1(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.index */
extern T6 T1491f2(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.index */
extern T6 T1490f2(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.index */
extern T6 T1483f2(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.index */
extern T6 T1270f1(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.index */
extern T6 T1267f1(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.index */
extern T6 T1254f1(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.index */
extern T6 T1232f1(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.index */
extern T6 T1218f1(GE_context* ac, T0* C);
/* detachable ET_INSPECT_EXPRESSION.index */
extern T6 T1189f1(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.index */
extern T6 T1167f1(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.index */
extern T6 T1041f1(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.index */
extern T6 T1038f1(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.index */
extern T6 T1008f1(GE_context* ac, T0* C);
/* detachable ET_VOID.index */
extern T6 T997f1(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.index */
extern T6 T736f1(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.index */
extern T6 T635f1(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.index */
extern T6 T309f1(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.index */
extern T6 T282f1(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.index */
extern T6 T184f2(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.index */
extern T6 T288f5(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.dynamic_type_set */
extern T0* T236f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MANIFEST_TUPLE.expression */
extern T0* T288f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_INFIX_CAST_EXPRESSION.expression_item */
extern T0* T1980f13(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.expression_item */
extern T0* T2048f10(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.expression_item */
extern T0* T2047f11(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.expression_item */
extern T0* T2046f9(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.expression_item */
extern T0* T1489f35(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.expression_item */
extern T0* T1488f35(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.expression_item */
extern T0* T1487f35(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.expression_item */
extern T0* T1486f35(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.expression_item */
extern T0* T372f10(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.expression_item */
extern T0* T1485f20(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.expression_item */
extern T0* T1484f18(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.expression_item */
extern T0* T1482f8(GE_context* ac, T0* C);
/* detachable ET_ITERATION_CURSOR.expression_item */
extern T0* T1551f4(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.expression_item */
extern T0* T1254f15(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.expression_item */
extern T0* T1008f19(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.expression_item */
extern T0* T1550f11(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.expression_item */
extern T0* T736f7(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.expression_item */
extern T0* T1549f11(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.expression_item */
extern T0* T1539f4(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.expression_item */
extern T0* T635f7(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.expression_item */
extern T0* T1557f33(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.expression_item */
extern T0* T1556f39(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.expression_item */
extern T0* T1555f36(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.expression_item */
extern T0* T1554f17(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.expression_item */
extern T0* T1553f19(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.expression_item */
extern T0* T1552f19(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.expression_item */
extern T0* T1041f8(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.expression_item */
extern T0* T1270f9(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.expression_item */
extern T0* T1218f9(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.expression_item */
extern T0* T1232f8(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.expression_item */
extern T0* T1038f7(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.expression_item */
extern T0* T311f9(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.expression_item */
extern T0* T1548f4(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.expression_item */
extern T0* T1547f15(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.expression_item */
extern T0* T1546f8(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.expression_item */
extern T0* T312f9(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.expression_item */
extern T0* T313f9(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.expression_item */
extern T0* T1543f7(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.expression_item */
extern T0* T1542f4(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.expression_item */
extern T0* T1541f4(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.expression_item */
extern T0* T1540f4(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.expression_item */
extern T0* T1538f11(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.expression_item */
extern T0* T1267f7(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.expression_item */
extern T0* T287f11(GE_context* ac, T0* C);
/* detachable ET_INSPECT_EXPRESSION.expression_item */
extern T0* T1189f7(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.expression_item */
extern T0* T1167f12(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.expression_item */
extern T0* T328f10(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.expression_item */
extern T0* T1496f11(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.expression_item */
extern T0* T1495f11(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.expression_item */
extern T0* T1494f15(GE_context* ac, T0* C);
/* detachable ET_VOID.expression_item */
extern T0* T997f15(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.expression_item */
extern T0* T1493f15(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.expression_item */
extern T0* T1491f12(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.expression_item */
extern T0* T1490f12(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.expression_item */
extern T0* T1483f26(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.expression_item */
extern T0* T309f8(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.expression_item */
extern T0* T282f8(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.expression_item */
extern T0* T288f15(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.expression_item */
extern T0* T184f51(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.item */
extern T0* T288f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_INFIX_CAST_EXPRESSION.actual_argument */
extern T0* T1980f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CONVERT_TO_EXPRESSION.actual_argument */
extern T0* T2048f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.actual_argument */
extern T0* T2047f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.actual_argument */
extern T0* T2046f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT_LIST.actual_argument */
extern T0* T2026f5(GE_context* ac, T0* C, T6 a1);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.actual_argument */
extern T0* T1025f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_INFIX_CAST_EXPRESSION.agent_actual_argument */
extern T0* T1980f14(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.agent_actual_argument */
extern T0* T2048f14(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.agent_actual_argument */
extern T0* T2047f15(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.agent_actual_argument */
extern T0* T2046f13(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.agent_actual_argument */
extern T0* T184f86(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.agent_actual_argument */
extern T0* T1489f40(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.agent_actual_argument */
extern T0* T1488f40(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.agent_actual_argument */
extern T0* T1487f40(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.agent_actual_argument */
extern T0* T1486f40(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.agent_actual_argument */
extern T0* T372f30(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.agent_actual_argument */
extern T0* T1485f31(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.agent_actual_argument */
extern T0* T1484f27(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.agent_actual_argument */
extern T0* T1482f26(GE_context* ac, T0* C);
/* detachable ET_ITERATION_CURSOR.agent_actual_argument */
extern T0* T1551f14(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.agent_actual_argument */
extern T0* T1254f26(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.agent_actual_argument */
extern T0* T1008f28(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.agent_actual_argument */
extern T0* T1550f16(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.agent_actual_argument */
extern T0* T736f17(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.agent_actual_argument */
extern T0* T1549f20(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.agent_actual_argument */
extern T0* T1539f14(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.agent_actual_argument */
extern T0* T635f16(GE_context* ac, T0* C);
/* detachable ET_AGENT_TYPED_OPEN_ARGUMENT.agent_actual_argument */
extern T0* T1559f6(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.agent_actual_argument */
extern T0* T1557f25(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.agent_actual_argument */
extern T0* T1556f32(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.agent_actual_argument */
extern T0* T1555f28(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.agent_actual_argument */
extern T0* T1554f26(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.agent_actual_argument */
extern T0* T1553f33(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.agent_actual_argument */
extern T0* T1552f29(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.agent_actual_argument */
extern T0* T1041f18(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.agent_actual_argument */
extern T0* T1270f20(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.agent_actual_argument */
extern T0* T288f21(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.agent_actual_argument */
extern T0* T1218f20(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.agent_actual_argument */
extern T0* T1232f15(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.agent_actual_argument */
extern T0* T1038f15(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.agent_actual_argument */
extern T0* T311f14(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.agent_actual_argument */
extern T0* T1548f13(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.agent_actual_argument */
extern T0* T1547f16(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.agent_actual_argument */
extern T0* T1546f15(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.agent_actual_argument */
extern T0* T312f13(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.agent_actual_argument */
extern T0* T313f13(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.agent_actual_argument */
extern T0* T1543f16(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.agent_actual_argument */
extern T0* T1542f13(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.agent_actual_argument */
extern T0* T1541f14(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.agent_actual_argument */
extern T0* T1540f14(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.agent_actual_argument */
extern T0* T1538f17(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.agent_actual_argument */
extern T0* T1267f17(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.agent_actual_argument */
extern T0* T287f17(GE_context* ac, T0* C);
/* detachable ET_INSPECT_EXPRESSION.agent_actual_argument */
extern T0* T1189f15(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.agent_actual_argument */
extern T0* T1167f19(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.agent_actual_argument */
extern T0* T328f17(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.agent_actual_argument */
extern T0* T1496f26(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.agent_actual_argument */
extern T0* T1495f26(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.agent_actual_argument */
extern T0* T282f23(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.agent_actual_argument */
extern T0* T309f23(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.agent_actual_argument */
extern T0* T1494f23(GE_context* ac, T0* C);
/* detachable ET_VOID.agent_actual_argument */
extern T0* T997f22(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.agent_actual_argument */
extern T0* T1493f23(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.agent_actual_argument */
extern T0* T1005f35(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.agent_actual_argument */
extern T0* T1491f25(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.agent_actual_argument */
extern T0* T1490f25(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.agent_actual_argument */
extern T0* T1483f27(GE_context* ac, T0* C);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.item */
extern T0* T1025f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_IDENTIFIER.actual_argument */
extern T0* T184f61(GE_context* ac, T0* C, T6 a1);
/* detachable ET_BINARY_INTEGER_CONSTANT.actual_argument */
extern T0* T1489f33(GE_context* ac, T0* C, T6 a1);
/* detachable ET_OCTAL_INTEGER_CONSTANT.actual_argument */
extern T0* T1488f33(GE_context* ac, T0* C, T6 a1);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.actual_argument */
extern T0* T1487f33(GE_context* ac, T0* C, T6 a1);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.actual_argument */
extern T0* T1486f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.actual_argument */
extern T0* T372f14(GE_context* ac, T0* C, T6 a1);
/* detachable ET_VERBATIM_STRING.actual_argument */
extern T0* T1485f25(GE_context* ac, T0* C, T6 a1);
/* detachable ET_SPECIAL_MANIFEST_STRING.actual_argument */
extern T0* T1484f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_C3_CHARACTER_CONSTANT.actual_argument */
extern T0* T1482f21(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ITERATION_CURSOR.actual_argument */
extern T0* T1551f9(GE_context* ac, T0* C, T6 a1);
/* detachable ET_QUANTIFIER_EXPRESSION.actual_argument */
extern T0* T1254f19(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ACROSS_EXPRESSION.actual_argument */
extern T0* T1008f23(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_MANIFEST_STRING.actual_argument */
extern T0* T1550f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NAMED_OBJECT_TEST.actual_argument */
extern T0* T736f13(GE_context* ac, T0* C, T6 a1);
/* detachable ET_OLD_OBJECT_TEST.actual_argument */
extern T0* T1549f16(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FEATURE_ADDRESS.actual_argument */
extern T0* T1539f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.actual_argument */
extern T0* T635f11(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.actual_argument */
extern T0* T1557f18(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.actual_argument */
extern T0* T1556f20(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.actual_argument */
extern T0* T1555f19(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.actual_argument */
extern T0* T1554f20(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.actual_argument */
extern T0* T1553f22(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.actual_argument */
extern T0* T1552f21(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CALL_AGENT.actual_argument */
extern T0* T1041f13(GE_context* ac, T0* C, T6 a1);
/* detachable ET_STRIP_EXPRESSION.actual_argument */
extern T0* T1270f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_MANIFEST_TUPLE.actual_argument */
extern T0* T288f9(GE_context* ac, T0* C, T6 a1);
/* detachable ET_MANIFEST_ARRAY.actual_argument */
extern T0* T1218f16(GE_context* ac, T0* C, T6 a1);
/* detachable ET_PARENTHESIZED_EXPRESSION.actual_argument */
extern T0* T1232f10(GE_context* ac, T0* C, T6 a1);
/* detachable ET_BRACKET_EXPRESSION.actual_argument */
extern T0* T1038f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_OBJECT_TEST.actual_argument */
extern T0* T311f6(GE_context* ac, T0* C, T6 a1);
/* detachable ET_OLD_EXPRESSION.actual_argument */
extern T0* T1548f9(GE_context* ac, T0* C, T6 a1);
/* detachable ET_PREFIX_EXPRESSION.actual_argument */
extern T0* T1547f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_MANIFEST_TYPE.actual_argument */
extern T0* T1546f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.actual_argument */
extern T0* T312f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EQUALITY_EXPRESSION.actual_argument */
extern T0* T313f6(GE_context* ac, T0* C, T6 a1);
/* detachable ET_INFIX_EXPRESSION.actual_argument */
extern T0* T1543f9(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXPRESSION_ADDRESS.actual_argument */
extern T0* T1542f9(GE_context* ac, T0* C, T6 a1);
/* detachable ET_RESULT_ADDRESS.actual_argument */
extern T0* T1541f10(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CURRENT_ADDRESS.actual_argument */
extern T0* T1540f10(GE_context* ac, T0* C, T6 a1);
/* detachable ET_PRECURSOR_EXPRESSION.actual_argument */
extern T0* T1538f13(GE_context* ac, T0* C, T6 a1);
/* detachable ET_STATIC_CALL_EXPRESSION.actual_argument */
extern T0* T1267f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.actual_argument */
extern T0* T287f13(GE_context* ac, T0* C, T6 a1);
/* detachable ET_INSPECT_EXPRESSION.actual_argument */
extern T0* T1189f11(GE_context* ac, T0* C, T6 a1);
/* detachable ET_IF_EXPRESSION.actual_argument */
extern T0* T1167f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CREATE_EXPRESSION.actual_argument */
extern T0* T328f12(GE_context* ac, T0* C, T6 a1);
/* detachable ET_C2_CHARACTER_CONSTANT.actual_argument */
extern T0* T1496f17(GE_context* ac, T0* C, T6 a1);
/* detachable ET_C1_CHARACTER_CONSTANT.actual_argument */
extern T0* T1495f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CURRENT.actual_argument */
extern T0* T282f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_RESULT.actual_argument */
extern T0* T309f15(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FALSE_CONSTANT.actual_argument */
extern T0* T1494f18(GE_context* ac, T0* C, T6 a1);
/* detachable ET_VOID.actual_argument */
extern T0* T997f18(GE_context* ac, T0* C, T6 a1);
/* detachable ET_TRUE_CONSTANT.actual_argument */
extern T0* T1493f18(GE_context* ac, T0* C, T6 a1);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.actual_argument */
extern T0* T1491f19(GE_context* ac, T0* C, T6 a1);
/* detachable ET_REGULAR_REAL_CONSTANT.actual_argument */
extern T0* T1490f19(GE_context* ac, T0* C, T6 a1);
/* detachable ET_REGULAR_MANIFEST_STRING.actual_argument */
extern T0* T1483f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.actual_argument */
extern T0* T1977f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.item */
extern T0* T1977f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.actual_argument */
extern T0* T283f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.item */
extern T0* T283f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.is_agent_type */
extern T1 T451f63(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.is_agent_type */
extern T1 T450f63(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.is_agent_type */
extern T1 T302f60(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.is_agent_type */
extern T1 T319f62(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_agent_type */
extern T1 T223f59(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.is_builtin_function_item */
extern T1 T306f42(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.is_builtin_function_item */
extern T1 T236f42(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.is_builtin_procedure_call */
extern T1 T306f44(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.is_builtin_procedure_call */
extern T1 T236f44(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_CAST_EXPRESSION.actual_argument_count */
extern T6 T1980f10(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.actual_argument_count */
extern T6 T2048f11(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.actual_argument_count */
extern T6 T2047f12(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.actual_argument_count */
extern T6 T2046f10(GE_context* ac, T0* C);
/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT_LIST.count */
extern T6 T2026f1(GE_context* ac, T0* C);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.count */
extern T6 T1025f3(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.actual_argument_count */
extern T6 T184f60(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.actual_argument_count */
extern T6 T1489f23(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.actual_argument_count */
extern T6 T1488f23(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.actual_argument_count */
extern T6 T1487f23(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.actual_argument_count */
extern T6 T1486f23(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.actual_argument_count */
extern T6 T372f13(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.actual_argument_count */
extern T6 T1485f24(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.actual_argument_count */
extern T6 T1484f23(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.actual_argument_count */
extern T6 T1482f19(GE_context* ac, T0* C);
/* detachable ET_ITERATION_CURSOR.actual_argument_count */
extern T6 T1551f8(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.actual_argument_count */
extern T6 T1254f18(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.actual_argument_count */
extern T6 T1008f22(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.actual_argument_count */
extern T6 T1550f8(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.actual_argument_count */
extern T6 T736f12(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.actual_argument_count */
extern T6 T1549f15(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.actual_argument_count */
extern T6 T1539f9(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.actual_argument_count */
extern T6 T635f10(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.actual_argument_count */
extern T6 T1557f17(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.actual_argument_count */
extern T6 T1556f19(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.actual_argument_count */
extern T6 T1555f18(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.actual_argument_count */
extern T6 T1554f19(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.actual_argument_count */
extern T6 T1553f21(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.actual_argument_count */
extern T6 T1552f20(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.actual_argument_count */
extern T6 T1041f12(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.actual_argument_count */
extern T6 T1270f15(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.actual_argument_count */
extern T6 T288f7(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.actual_argument_count */
extern T6 T1218f15(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.actual_argument_count */
extern T6 T1232f9(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.actual_argument_count */
extern T6 T1038f10(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.actual_argument_count */
extern T6 T311f5(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.actual_argument_count */
extern T6 T1548f8(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.actual_argument_count */
extern T6 T1547f7(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.actual_argument_count */
extern T6 T1546f9(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.actual_argument_count */
extern T6 T312f5(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.actual_argument_count */
extern T6 T313f5(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.actual_argument_count */
extern T6 T1543f8(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.actual_argument_count */
extern T6 T1542f8(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.actual_argument_count */
extern T6 T1541f9(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.actual_argument_count */
extern T6 T1540f9(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.actual_argument_count */
extern T6 T1538f12(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.actual_argument_count */
extern T6 T1267f12(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.actual_argument_count */
extern T6 T287f12(GE_context* ac, T0* C);
/* detachable ET_INSPECT_EXPRESSION.actual_argument_count */
extern T6 T1189f10(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.actual_argument_count */
extern T6 T1167f14(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.actual_argument_count */
extern T6 T328f11(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.actual_argument_count */
extern T6 T1496f16(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.actual_argument_count */
extern T6 T1495f16(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.actual_argument_count */
extern T6 T282f14(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.actual_argument_count */
extern T6 T309f14(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.actual_argument_count */
extern T6 T1494f17(GE_context* ac, T0* C);
/* detachable ET_VOID.actual_argument_count */
extern T6 T997f17(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.actual_argument_count */
extern T6 T1493f17(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.actual_argument_count */
extern T6 T1491f18(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.actual_argument_count */
extern T6 T1490f18(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.actual_argument_count */
extern T6 T1483f16(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.count */
extern T6 T1977f1(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.count */
extern T6 T283f1(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.arguments */
extern T0* T2048f15(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.arguments */
extern T0* T1041f5(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.arguments */
extern T0* T1038f4(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.arguments */
extern T0* T1547f9(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.right */
extern T0* T1543f4(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_EXPRESSION.arguments */
extern T0* T888f3(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.arguments */
extern T0* T287f3(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.dynamic_feature */
extern T0* T314f13(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.is_tuple_label */
extern T1 T314f9(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.is_tuple_label */
extern T1 T2048f16(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_tuple_label */
extern T1 T184f38(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.is_tuple_label */
extern T1 T1041f25(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.is_tuple_label */
extern T1 T1038f8(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_tuple_label */
extern T1 T622f95(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.is_tuple_label */
extern T1 T1547f10(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_tuple_label */
extern T1 T973f24(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_tuple_label */
extern T1 T981f24(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_tuple_label */
extern T1 T994f95(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.is_tuple_label */
extern T1 T1543f10(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_tuple_label */
extern T1 T1545f15(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_tuple_label */
extern T1 T1544f15(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_EXPRESSION.is_tuple_label */
extern T1 T888f6(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_tuple_label */
extern T1 T2020f43(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.is_tuple_label */
extern T1 T287f6(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.put_type_from_type_set */
extern void T318f14(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.put_type_with_tuple_label */
extern void T318f18(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.propagate_tuple_label_setter_dynamic_types */
extern void T230f29(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_tuple_label_setter_dynamic_types */
extern void T71f502(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T71f502ot1(T0* a1, T0** a2);
extern T1 T71f502ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_tuple_label_argument_dynamic_types */
extern void T71f521(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ASSIGNER_INSTRUCTION.name */
extern T0* T1527f4(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_INSTRUCTION.name */
extern T0* T2002f2(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.name */
extern T0* T284f5(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.put_type_with_feature */
extern void T318f17(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T318f17ot1(T0* a1, T0** a2);
extern T1 T318f17ot2(T0* a1, T0** a2);
extern T1 T318f17ot3(T0* a1, T0** a2);
/* detachable ET_ASSIGNER_INSTRUCTION.actual_argument */
extern T0* T1527f13(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ASSIGNER_INSTRUCTION.count */
extern T6 T1527f12(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.arguments_count */
extern T6 T1038f19(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.arguments_count */
extern T6 T287f21(GE_context* ac, T0* C);
/* detachable ET_ASSIGNER_INSTRUCTION.arguments */
extern T0* T1527f9(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_INSTRUCTION.arguments */
extern T0* T2002f3(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.arguments */
extern T0* T284f3(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.dynamic_feature */
extern T0* T318f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.seeded_dynamic_procedure */
extern T0* T451f43(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.put_procedure */
extern void T451f84(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_PRECURSOR.is_builtin_any_copy */
extern T1 T306f36(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.is_builtin_any_copy */
extern T1 T236f23(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.new_dynamic_procedure */
extern T0* T451f54(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.seeded_procedure */
extern T0* T74f181(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_PROCEDURE_LIST.seeded_feature */
extern T0* T258f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.seeded_dynamic_procedure */
extern T0* T450f43(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.put_procedure */
extern void T450f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.new_dynamic_procedure */
extern T0* T450f57(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.build_agent_call */
extern void T230f26(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.build_agent_call */
extern void T71f132(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_DYNAMIC_FEATURE.set_dynamic_type_sets */
extern void T236f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.make */
extern T0* T1466c15(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.is_expanded */
extern T1 T323f6(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.is_expanded */
extern T1 T254f17(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.is_empty */
extern T1 T259f8(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.new_dynamic_procedure */
extern T0* T450f57p1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.seeded_dynamic_procedure */
extern T0* T302f46(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.put_procedure */
extern void T302f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.new_dynamic_procedure */
extern T0* T302f28(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.build_tuple_put_reference */
extern void T230f25(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.build_tuple_put_reference */
extern void T71f130(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.put_reference_feature_name */
extern T0* T73f92(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.name */
extern T0* T1516f35(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.name */
extern T0* T476f40(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.name */
extern T0* T1515f35(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.name */
extern T0* T469f33(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.new_dynamic_procedure */
extern T0* T302f28p1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.seeded_dynamic_procedure */
extern T0* T319f48(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.put_procedure */
extern void T319f76(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.new_dynamic_procedure */
extern T0* T319f28(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_DYNAMIC_FEATURE.is_builtin_special_extend */
extern T1 T236f33(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.is_builtin_special_put */
extern T1 T236f32(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.new_dynamic_procedure */
extern T0* T319f28p1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.seeded_dynamic_procedure */
extern T0* T223f35(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_procedure */
extern void T223f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.new_dynamic_procedure */
extern T0* T223f32(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.is_tuple_label */
extern T1 T318f10(GE_context* ac, T0* C);
/* detachable ET_ASSIGNER_INSTRUCTION.is_tuple_label */
extern T1 T1527f10(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_INSTRUCTION.is_tuple_label */
extern T1 T2002f5(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.is_tuple_label */
extern T1 T284f6(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.put_type_from_type_set */
extern void T1466f17(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T1466f17ot1(T0* a1, T0** a2);
extern T1 T1466f17ot2(T0* a1, T0** a2);
/* [detachable] ET_SYSTEM_PROCESSOR.is_ise */
extern T1 T64f42(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.is_ise */
extern T1 T375f44(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.put_type_from_type_set */
extern void T1466f17p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T1466f17p1ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.count */
extern T6 T475f2(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.dynamic_types */
extern T0* T475f4(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.dynamic_types */
extern T0* T323f19(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.dynamic_types */
extern T0* T451f60(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.dynamic_types */
extern T0* T450f60(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.dynamic_types */
extern T0* T302f57(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.dynamic_types */
extern T0* T319f59(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.dynamic_types */
extern T0* T223f56(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.put_type_from_type_set */
extern void T269f23(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.put_type */
extern void T269f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.put_type_from_type_set */
extern void T475f17(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.put_type_from_type_set */
extern void T475f17p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T475f17p1ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.has_type */
extern T1 T475f12(GE_context* ac, T0* C, T0* a1);
extern T1 T475f12ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.has_type */
extern T1 T475f12p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.index_of */
extern T6 T475f13(GE_context* ac, T0* C, T0* a1);
extern T1 T475f13ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.index_of */
extern T6 T475f13p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.put_type_from_type_set */
extern void T451f78(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.put_type_from_type_set */
extern void T450f78(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.put_type_from_type_set */
extern void T302f80(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.put_type_from_type_set */
extern void T319f82(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_type_from_type_set */
extern void T223f78(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.put_type_from_type_set */
extern void T323f28(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.dynamic_type */
extern T0* T475f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DYNAMIC_TARGET_LIST.force_last */
extern void T1353f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TARGET].force */
extern void T1717f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TARGET].fill_with */
extern void T1716f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TARGET].extend */
extern void T1716f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_TARGET_LIST.fixed_array */
extern T0* T1353f6(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TARGET].default_create */
extern T0* T1717c3(GE_context* ac);
/* detachable ET_DYNAMIC_TARGET_LIST.resize */
extern void T1353f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TARGET].aliased_resized_area */
extern T0* T1717f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TARGET].aliased_resized_area */
extern T0* T1716f4(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DYNAMIC_TARGET_LIST.new_capacity */
extern T6 T1353f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DYNAMIC_TARGET_LIST.capacity */
extern T6 T1353f7(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_TARGET_LIST.has */
extern T1 T1353f3(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_TARGET_LIST.index_of */
extern T6 T1353f5(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_TARGET_LIST.put_last */
extern void T1353f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_TARGET_LIST.make_with_capacity */
extern T0* T1353c9(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TARGET].make */
extern T0* T1717f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TARGET].make_empty */
extern T0* T1716c5(GE_context* ac, T6 a1);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.put_target */
extern void T323f29(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.is_never_void */
extern T1 T323f16(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.put_target */
extern void T451f77(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.is_never_void */
extern T1 T451f40(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.put_target */
extern void T450f77(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.is_never_void */
extern T1 T450f40(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.put_target */
extern void T302f79(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.is_never_void */
extern T1 T302f33(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.put_target */
extern void T319f81(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.is_never_void */
extern T1 T319f33(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_target */
extern void T223f74(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_never_void */
extern T1 T223f24(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.is_expanded */
extern T1 T451f58(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.is_expanded */
extern T1 T450f64(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.is_expanded */
extern T1 T302f61(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.is_expanded */
extern T1 T319f63(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_expanded */
extern T1 T223f41(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.static_type */
extern T0* T1466f1(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.static_type */
extern T0* T269f1(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.static_type */
extern T0* T475f1(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.static_type */
extern T0* T451f48(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.static_type */
extern T0* T450f50(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.static_type */
extern T0* T302f49(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.static_type */
extern T0* T319f51(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.static_type */
extern T0* T223f50(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.static_type */
extern T0* T323f7(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.put */
extern void T224f12(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ET_DYNAMIC_PRECURSOR.is_attribute */
extern T1 T306f29(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.is_builtin_attribute */
extern T1 T450f41(GE_context* ac, T0* C, T0* a1, T8 a2, T8 a3);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.is_builtin_attribute */
extern T1 T319f40(GE_context* ac, T0* C, T0* a1, T8 a2, T8 a3);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.is_builtin_attribute */
extern T1 T302f40(GE_context* ac, T0* C, T0* a1, T8 a2, T8 a3);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_builtin_attribute */
extern T1 T223f47(GE_context* ac, T0* C, T0* a1, T8 a2, T8 a3);
/* detachable ET_EXTERNAL_PROCEDURE.is_attribute */
extern T1 T1516f37(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_attribute */
extern T1 T476f43(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_attribute */
extern T1 T1515f38(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_attribute */
extern T1 T469f42(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.is_builtin */
extern T1 T306f26(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.is_attribute */
extern T1 T236f26(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.register_feature_seeds */
extern void T451f80(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.new_dynamic_query */
extern T0* T451f50(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.new_dynamic_query */
extern T0* T451f50p1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].value */
extern T0* T602f22(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].item_storage_item */
extern T0* T602f21(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].position_of_key */
extern T6 T602f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EXTERNAL_FUNCTION.first_seed */
extern T6 T253f1(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.first_seed */
extern T6 T1513f1(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.first_seed */
extern T6 T1512f1(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.first_seed */
extern T6 T1511f1(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.first_seed */
extern T6 T371f3(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.first_seed */
extern T6 T369f1(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.first_seed */
extern T6 T359f1(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.first_seed */
extern T6 T360f1(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].make_map */
extern T0* T602c33(GE_context* ac, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].make_with_equality_testers */
extern void T602f35(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].make_sparse_container */
extern void T602f43(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].set_internal_cursor */
extern void T602f52(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].new_cursor */
extern T0* T602f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].make */
extern T0* T1400c4(GE_context* ac, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].make_slots */
extern void T602f51(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].make_clashes */
extern void T602f50(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].make_key_storage */
extern void T602f49(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_FEATURE, [attached] INTEGER_32].make_item_storage */
extern void T602f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_FEATURE].make */
extern T0* T606f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_FEATURE].make_empty */
extern T0* T605c5(GE_context* ac, T6 a1);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.make_with_capacity */
extern T0* T224c10(GE_context* ac, T6 a1);
/* [detachable] ET_EXTERNAL_FUNCTION.is_attribute */
extern T1 T253f53(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_attribute */
extern T1 T1513f39(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_attribute */
extern T1 T1512f41(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_attribute */
extern T1 T1511f39(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_attribute */
extern T1 T371f34(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_attribute */
extern T1 T369f33(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_attribute */
extern T1 T359f62(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_attribute */
extern T1 T360f32(GE_context* ac, T0* C);
/* [detachable] ET_QUERY_LIST.item */
extern T0* T257f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.use_all_attributes */
extern void T450f76(GE_context* ac, T0* C, T0* a1);
extern T1 T450f76ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.dynamic_query */
extern T0* T450f29(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.use_all_attributes */
extern void T302f78(GE_context* ac, T0* C, T0* a1);
extern T1 T302f78ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.dynamic_query */
extern T0* T302f27(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.use_all_attributes */
extern void T319f80(GE_context* ac, T0* C, T0* a1);
extern T1 T319f80ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.dynamic_query */
extern T0* T319f27(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.use_all_attributes */
extern void T223f77(GE_context* ac, T0* C, T0* a1);
extern T1 T223f77ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.dynamic_query */
extern T0* T223f44(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.propagate_set_reference_field_dynamic_types */
extern void T230f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_set_reference_field_dynamic_types */
extern void T71f128(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_builtin_formal_argument_dynamic_types */
extern void T71f131(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.item */
extern T0* T224f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.attribute_count */
extern T6 T451f6(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.attribute_count */
extern T6 T450f6(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.attribute_count */
extern T6 T302f6(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.attribute_count */
extern T6 T319f6(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.attribute_count */
extern T6 T223f1(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.set_ise_runtime_reference_field_feature */
extern void T63f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.set_ise_runtime_new_type_instance_of_feature */
extern void T63f106(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.attached_type */
extern T0* T451f5(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.attached_type */
extern T0* T450f5(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.attached_type */
extern T0* T302f5(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.attached_type */
extern T0* T319f5(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.attached_type */
extern T0* T223f6(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.propagate_type_of_type_result_type */
extern void T63f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.propagate_type_of_type_result_type */
extern void T230f20(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_type_of_type_result_type */
extern void T71f124(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.meta_type */
extern T0* T451f8(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.meta_type */
extern T0* T450f8(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.meta_type */
extern T0* T302f8(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.meta_type */
extern T0* T319f8(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.meta_type */
extern T0* T223f8(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.meta_type */
extern T0* T323f2(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.create_meta_type */
extern void T63f104(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.meta_type */
extern T0* T63f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.set_meta_type */
extern void T451f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.set_meta_type */
extern void T450f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.set_meta_type */
extern void T302f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.set_meta_type */
extern void T319f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_meta_type */
extern void T223f65(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.set_meta_type */
extern void T323f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.dynamic_primary_type */
extern T0* T63f59(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_SYSTEM.initialize_dynamic_types */
extern void T63f93(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_DYNAMIC_SYSTEM.propagate_alive_conforming_descendants */
extern void T63f103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.propagate_alive_conforming_descendants */
extern void T230f21(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].forth */
extern void T382f42(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].cursor_forth */
extern void T382f47(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].add_traversing_cursor */
extern void T382f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].set_next_cursor */
extern void T921f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].set_position */
extern void T921f11(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].remove_traversing_cursor */
extern void T382f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].set_after */
extern void T921f9(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].clashes_item */
extern T6 T382f22(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].item_for_iteration */
extern T0* T382f24(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].cursor_item */
extern T0* T382f39(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].item_storage_item */
extern T0* T382f26(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].after */
extern T1 T382f32(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].cursor_after */
extern T1 T382f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].after */
extern T1 T921f4(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].start */
extern void T382f41(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].cursor_start */
extern void T382f46(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].cursor_off */
extern T1 T382f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].off */
extern T1 T921f7(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].is_empty */
extern T1 T382f20(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_alive_conforming_descendants */
extern void T71f109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.propagate_attribute_type_sets */
extern void T63f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.initialize_predicate_type */
extern void T63f101(GE_context* ac, T0* C, T0* a1);
extern T1 T63f101ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.set_set_rout_disp_final_feature */
extern void T451f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.set_set_rout_disp_final_feature */
extern void T450f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.set_attribute_position */
extern void T451f74(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.index_of */
extern T6 T224f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.set_attribute_position */
extern void T450f74(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.set_attribute_position */
extern void T302f77(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.set_attribute_position */
extern void T319f79(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_attribute_position */
extern void T223f76(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_CLASS.is_predicate_class */
extern T1 T74f78(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.initialize_function_type */
extern void T63f100(GE_context* ac, T0* C, T0* a1);
extern T1 T63f100ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.is_function_class */
extern T1 T74f77(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.initialize_procedure_type */
extern void T63f99(GE_context* ac, T0* C, T0* a1);
extern T1 T63f99ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.is_procedure_class */
extern T1 T74f76(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.initialize_typed_pointer_type */
extern void T63f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_typed_pointer_class */
extern T1 T74f75(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.initialize_array_type */
extern void T63f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_array_class */
extern T1 T74f74(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.initialize_special_type */
extern void T63f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_special_class */
extern T1 T74f73(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.runtime_name_to_text */
extern T0* T74f69(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.runtime_name_to_text */
extern T0* T254f21(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.runtime_name_to_text */
extern T0* T72f13(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.has_interface_error */
extern T1 T74f94(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.interface_checked */
extern T1 T74f91(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.set_fatal_error */
extern void T63f92(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.has_implementation_error */
extern T1 T74f119(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.implementation_checked */
extern T1 T74f108(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.process */
extern void T74f261(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.process_class */
extern void T1338f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.process_class_type */
extern void T1338f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.check_class_type_validity */
extern void T1338f9(GE_context* ac, T0* C, T0* a1);
extern T1 T1338f9ot1(T0* a1, T0** a2);
extern T1 T1338f9ot2(T0* a1, T0** a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.process */
extern void T1501f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.process_constrained_formal_parameter */
extern void T1325f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.process_constrained_formal_parameter */
extern void T1309f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.process_constrained_formal_parameter */
extern void T1338f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.process_constrained_formal_parameter */
extern void T1324f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.process_constrained_formal_parameter */
extern void T1308f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.process_constrained_formal_parameter */
extern void T1619f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.process_constrained_formal_parameter */
extern void T1618f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.process_constrained_formal_parameter */
extern void T1617f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.process_constrained_formal_parameter */
extern void T1616f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.process_constrained_formal_parameter */
extern void T1331f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.process_constrained_formal_parameter */
extern void T1330f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.process_constrained_formal_parameter */
extern void T1323f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.process_constrained_formal_parameter */
extern void T1322f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.process_constrained_formal_parameter */
extern void T1321f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.process_constrained_formal_parameter */
extern void T1320f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.process_constrained_formal_parameter */
extern void T384f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.process */
extern void T891f85(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.process_formal_parameter */
extern void T1338f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.process_formal_parameter */
extern void T1325f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.process_formal_parameter */
extern void T1324f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.process_formal_parameter */
extern void T1309f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.process_formal_parameter */
extern void T1308f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.process_formal_parameter */
extern void T1619f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.process_formal_parameter */
extern void T1618f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.process_formal_parameter */
extern void T1617f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.process_formal_parameter */
extern void T1616f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.process_formal_parameter */
extern void T1331f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.process_formal_parameter */
extern void T1330f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.process_formal_parameter */
extern void T1323f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.process_formal_parameter */
extern void T1322f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.process_formal_parameter */
extern void T1321f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.process_formal_parameter */
extern void T1320f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.process_formal_parameter */
extern void T384f43(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.process */
extern void T1526f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.process_qualified_like_type */
extern void T1325f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.process_qualified_like_type */
extern void T1309f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.process_like_type */
extern void T1309f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.check_like_type_constraint */
extern void T1309f46(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vggc1a_error */
extern void T53f538(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_validity_error */
extern void T53f204(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.set_has_eiffel_error */
extern void T53f205(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_info */
extern void T53f184(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_info_message */
extern void T53f186(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_VALIDITY_ERROR.make_vggc1a */
extern T0* T785c1160(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.set_compilers */
extern void T785f858(GE_context* ac, T0* C, T1 a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.to_text */
extern T0* T1526f16(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.append_to_string */
extern void T1526f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER.lower_name */
extern T0* T184f10(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.as_lower */
extern T0* T775f61(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.to_lower */
extern void T775f91(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.unicode */
extern T0* T775f62(GE_context* ac, T0* C);
/* [detachable] UC_STRING.as_lower */
extern T0* T429f20(GE_context* ac, T0* C);
/* [detachable] UC_STRING.to_lower */
extern void T429f87(GE_context* ac, T0* C);
/* [detachable] UC_STRING.unicode */
extern T0* T429f60(GE_context* ac, T0* C);
/* [detachable] STRING_8.as_lower */
extern T0* T17f33(GE_context* ac, T0* C);
/* [detachable] STRING_8.to_lower */
extern void T17f75(GE_context* ac, T0* C);
/* [detachable] STRING_8.to_lower_area */
extern void T17f78(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* CHARACTER_8.lower */
extern T2 T2f23(GE_context* ac, T2* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.name */
extern T0* T1526f39(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.append_to_string */
extern void T1524f89(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.append_to_string */
extern void T1525f89(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.name */
extern T0* T1525f35(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.append_to_string */
extern void T1501f87(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.upper_name */
extern T0* T1501f28(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.upper_name */
extern T0* T184f11(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.as_upper */
extern T0* T775f65(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.to_upper */
extern void T775f111(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code */
extern T6 T426f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_UNICODE_ROUTINES.upper_codes */
extern T0* T426f31(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.empty_upper_code_plane */
extern T0* T426f38(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.empty_upper_code_segment */
extern T0* T426f45(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_1 */
extern T0* T426f37(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_1_segment_4 */
extern T0* T426f56(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0 */
extern T0* T426f36(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_255 */
extern T0* T426f55(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_167 */
extern T0* T426f54(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_166 */
extern T0* T426f53(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_45 */
extern T0* T426f52(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_44 */
extern T0* T426f51(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_36 */
extern T0* T426f50(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_33 */
extern T0* T426f49(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_31 */
extern T0* T426f48(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_30 */
extern T0* T426f47(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_29 */
extern T0* T426f46(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_5 */
extern T0* T426f44(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_4 */
extern T0* T426f43(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_3 */
extern T0* T426f42(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_2 */
extern T0* T426f41(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_1 */
extern T0* T426f40(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.upper_code_plane_0_segment_0 */
extern T0* T426f39(GE_context* ac, T0* C);
/* [detachable] UC_STRING.as_upper */
extern T0* T429f66(GE_context* ac, T0* C);
/* [detachable] UC_STRING.to_upper */
extern void T429f109(GE_context* ac, T0* C);
/* [detachable] STRING_8.as_upper */
extern T0* T17f36(GE_context* ac, T0* C);
/* [detachable] STRING_8.to_upper */
extern void T17f76(GE_context* ac, T0* C);
/* [detachable] STRING_8.to_upper_area */
extern void T17f79(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* CHARACTER_8.upper */
extern T2 T2f24(GE_context* ac, T2* C);
/* CHARACTER_8.upper_value */
extern T0* T2f28(GE_context* ac, T2* C);
/* [detachable] ET_FORMAL_PARAMETER.append_to_string */
extern void T891f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.upper_name */
extern T0* T891f29(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.append_to_string */
extern void T74f266(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.append_to_string */
extern void T603f32(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.label */
extern T0* T1501f68(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.label */
extern T0* T891f70(GE_context* ac, T0* C);
/* detachable ET_FORMAL_PARAMETER_LIST.actual_parameter */
extern T0* T603f19(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.actual_parameter */
extern T0* T1501f37(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.type */
extern T0* T1501f27(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.formal_parameter */
extern T0* T1501f10(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.actual_parameter */
extern T0* T891f38(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.type */
extern T0* T891f25(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.formal_parameter */
extern T0* T891f8(GE_context* ac, T0* C);
/* detachable ET_FORMAL_PARAMETER_LIST.item */
extern T0* T603f5(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FORMAL_PARAMETER_LIST.is_empty */
extern T1 T603f11(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.append_to_string */
extern void T254f122(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.append_to_string */
extern void T1957f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LABELED_COMMA_ACTUAL_PARAMETER.label */
extern T0* T1769f5(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.identifier */
extern T0* T184f59(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.label */
extern T0* T1526f32(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.label */
extern T0* T1525f28(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.label */
extern T0* T1524f22(GE_context* ac, T0* C);
/* detachable ET_LABELED_ACTUAL_PARAMETER.label */
extern T0* T1522f5(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.label */
extern T0* T242f30(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.label */
extern T0* T449f24(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.label */
extern T0* T254f53(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.label */
extern T0* T74f142(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.label */
extern T0* T72f49(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.actual_parameter */
extern T0* T1957f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.actual_parameter */
extern T0* T577f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.actual_parameter */
extern T0* T1956f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.actual_parameter */
extern T0* T240f5(GE_context* ac, T0* C, T6 a1);
/* detachable ET_LIKE_FEATURE.actual_parameter */
extern T0* T1524f25(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.actual_parameter */
extern T0* T1526f38(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.actual_parameter */
extern T0* T1525f31(GE_context* ac, T0* C);
/* [detachable] ET_LABELED_COMMA_ACTUAL_PARAMETER.actual_parameter */
extern T0* T1769f6(GE_context* ac, T0* C);
/* detachable ET_LABELED_ACTUAL_PARAMETER.actual_parameter */
extern T0* T1522f6(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.actual_parameter */
extern T0* T242f11(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.actual_parameter */
extern T0* T449f14(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.actual_parameter */
extern T0* T254f44(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.actual_parameter */
extern T0* T74f128(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.actual_parameter */
extern T0* T72f44(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.item */
extern T0* T240f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.append_to_string */
extern void T240f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.is_empty */
extern T1 T1957f4(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.is_empty */
extern T1 T240f10(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.append_to_string */
extern void T242f80(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.upper_name */
extern T0* T242f36(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.append_to_string */
extern void T72f129(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.append_to_string */
extern void T577f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.append_to_string */
extern void T1956f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.is_empty */
extern T1 T577f5(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.is_empty */
extern T1 T1956f15(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.append_to_string */
extern void T449f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLICIT_TYPE_MARK.append_to_string_with_space */
extern void T301f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLICIT_TYPE_MARK.expandedness_text */
extern T0* T301f11(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.tokens */
extern T0* T301f13(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_explicit_expandedness_mark */
extern T1 T301f15(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_reference */
extern T1 T301f22(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_expanded */
extern T1 T301f21(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.separateness_text */
extern T0* T301f10(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_explicit_separateness_mark */
extern T1 T301f14(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_separate */
extern T1 T301f20(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.attachment_text */
extern T0* T301f9(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_explicit_attachment_mark */
extern T1 T301f12(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_question_mark */
extern T1 T301f19(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_bang */
extern T1 T301f18(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_detachable */
extern T1 T301f17(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_attached */
extern T1 T301f16(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.append_to_string_with_space */
extern void T571f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.expandedness_text */
extern T0* T571f15(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.tokens */
extern T0* T571f18(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_explicit_expandedness_mark */
extern T1 T571f20(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.separateness_text */
extern T0* T571f14(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_explicit_separateness_mark */
extern T1 T571f19(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_explicit_separateness_mark */
extern T1 T413f25(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_separate */
extern T1 T413f20(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_separate_mark */
extern T1 T571f5(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_separate_mark */
extern T1 T413f10(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.attachment_text */
extern T0* T571f13(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_explicit_attachment_mark */
extern T1 T571f17(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_explicit_attachment_mark */
extern T1 T994f15(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_explicit_attachment_mark */
extern T1 T622f20(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_explicit_attachment_mark */
extern T1 T1005f10(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_explicit_attachment_mark */
extern T1 T341f18(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_explicit_attachment_mark */
extern T1 T413f24(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.append_to_string_with_space */
extern void T994f150(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.expandedness_text */
extern T0* T994f82(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_explicit_expandedness_mark */
extern T1 T994f98(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.separateness_text */
extern T0* T994f81(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_explicit_separateness_mark */
extern T1 T994f97(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_separate */
extern T1 T994f75(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_separate_mark */
extern T1 T994f12(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.attachment_text */
extern T0* T994f80(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.append_to_string_with_space */
extern void T622f133(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BRACKET_SYMBOL.expandedness_text */
extern T0* T622f82(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_explicit_expandedness_mark */
extern T1 T622f98(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.separateness_text */
extern T0* T622f81(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_explicit_separateness_mark */
extern T1 T622f97(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_separate */
extern T1 T622f76(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_separate_mark */
extern T1 T622f17(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.attachment_text */
extern T0* T622f80(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.append_to_string_with_space */
extern void T1005f44(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUESTION_MARK_SYMBOL.expandedness_text */
extern T0* T1005f27(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_explicit_expandedness_mark */
extern T1 T1005f31(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.separateness_text */
extern T0* T1005f26(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_explicit_separateness_mark */
extern T1 T1005f30(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_separate */
extern T1 T1005f18(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_separate_mark */
extern T1 T1005f7(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.attachment_text */
extern T0* T1005f25(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.append_to_string_with_space */
extern void T341f105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYMBOL.expandedness_text */
extern T0* T341f85(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_explicit_expandedness_mark */
extern T1 T341f77(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.separateness_text */
extern T0* T341f84(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_explicit_separateness_mark */
extern T1 T341f78(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_separate */
extern T1 T341f76(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_separate_mark */
extern T1 T341f15(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.attachment_text */
extern T0* T341f83(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.append_to_string_with_space */
extern void T413f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_KEYWORD.expandedness_text */
extern T0* T413f23(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_explicit_expandedness_mark */
extern T1 T413f26(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.separateness_text */
extern T0* T413f22(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.attachment_text */
extern T0* T413f21(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.to_text */
extern T0* T1525f41(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.to_text */
extern T0* T1524f42(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.to_text */
extern T0* T449f38(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.upper_name */
extern T0* T74f174(GE_context* ac, T0* C);
/* detachable ET_RETRY_INSTRUCTION.column */
extern T6 T991f12(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.column */
extern T6 T2020f40(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.column */
extern T6 T372f33(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.column */
extern T6 T1489f30(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.column */
extern T6 T1488f30(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.column */
extern T6 T1487f30(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.column */
extern T6 T1486f30(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.column */
extern T6 T1485f16(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.column */
extern T6 T1484f16(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.column */
extern T6 T1482f14(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.column */
extern T6 T973f26(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.column */
extern T6 T955f12(GE_context* ac, T0* C);
/* detachable ET_AGENT_KEYWORD.column */
extern T6 T952f11(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.column */
extern T6 T981f12(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.column */
extern T6 T1496f7(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.column */
extern T6 T1495f7(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.column */
extern T6 T1494f6(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.column */
extern T6 T1493f6(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.column */
extern T6 T1491f8(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.column */
extern T6 T1490f8(GE_context* ac, T0* C);
/* detachable ET_VOID.column */
extern T6 T997f6(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.column */
extern T6 T309f11(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.column */
extern T6 T282f11(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.column */
extern T6 T1483f8(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.column */
extern T6 T1005f34(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.column */
extern T6 T413f35(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.column */
extern T6 T341f11(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.column */
extern T6 T994f8(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.column */
extern T6 T622f14(GE_context* ac, T0* C);
/* [detachable] ET_COMPRESSED_POSITION.column */
extern T6 T193f7(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.column */
extern T6 T184f36(GE_context* ac, T0* C);
/* detachable ET_RETRY_INSTRUCTION.line */
extern T6 T991f10(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.line */
extern T6 T2020f39(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.line */
extern T6 T372f32(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.line */
extern T6 T1489f29(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.line */
extern T6 T1488f29(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.line */
extern T6 T1487f29(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.line */
extern T6 T1486f29(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.line */
extern T6 T1485f15(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.line */
extern T6 T1484f15(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.line */
extern T6 T1482f13(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.line */
extern T6 T973f25(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.line */
extern T6 T955f11(GE_context* ac, T0* C);
/* detachable ET_AGENT_KEYWORD.line */
extern T6 T952f10(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.line */
extern T6 T981f11(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.line */
extern T6 T1496f6(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.line */
extern T6 T1495f6(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.line */
extern T6 T1494f5(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.line */
extern T6 T1493f5(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.line */
extern T6 T1491f7(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.line */
extern T6 T1490f7(GE_context* ac, T0* C);
/* detachable ET_VOID.line */
extern T6 T997f5(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.line */
extern T6 T309f10(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.line */
extern T6 T282f10(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.line */
extern T6 T1483f7(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.line */
extern T6 T1005f33(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.line */
extern T6 T413f34(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.line */
extern T6 T341f10(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.line */
extern T6 T994f7(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.line */
extern T6 T622f7(GE_context* ac, T0* C);
/* [detachable] ET_COMPRESSED_POSITION.line */
extern T6 T193f6(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.line */
extern T6 T184f35(GE_context* ac, T0* C);
/* [detachable] ET_VALIDITY_ERROR.filename */
extern T0* T785f21(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.is_cluster */
extern T1 T802f44(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.is_cluster */
extern T1 T887f8(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.is_cluster */
extern T1 T410f9(GE_context* ac, T0* C);
/* [detachable] ET_VALIDITY_ERROR.default_message_template */
extern T0* T785f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_VALIDITY_ERROR.null_position */
extern T0* T785f12(GE_context* ac, T0* C);
/* [detachable] ET_COMPRESSED_POSITION.make_default */
extern T0* T193c12(GE_context* ac);
/* [detachable] ET_VALIDITY_ERROR.template_code */
extern T0* T785f15(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.position */
extern T0* T1526f41(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_null_position */
extern T1 T1489f31(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_null_position */
extern T1 T1488f31(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_null_position */
extern T1 T1487f31(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_null_position */
extern T1 T1486f31(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_null_position */
extern T1 T372f34(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_null_position */
extern T1 T1485f17(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_null_position */
extern T1 T1484f17(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_null_position */
extern T1 T1482f15(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_null_position */
extern T1 T1496f8(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_null_position */
extern T1 T1495f8(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.is_null_position */
extern T1 T1494f7(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.is_null_position */
extern T1 T1493f7(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.is_null_position */
extern T1 T1491f9(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.is_null_position */
extern T1 T1490f9(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.is_null_position */
extern T1 T1483f9(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.is_null_position */
extern T1 T282f12(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_null_position */
extern T1 T184f53(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.like_keyword */
extern T0* T1526f34(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.like_keyword */
extern T0* T1524f2(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.like_keyword */
extern T0* T1525f2(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.like_keyword */
extern T0* T449f2(GE_context* ac, T0* C);
/* [detachable] ET_COMPRESSED_POSITION.is_null */
extern T1 T193f8(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_null_position */
extern T1 T1005f32(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_null_position */
extern T1 T413f33(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_null_position */
extern T1 T341f24(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_null_position */
extern T1 T994f20(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_null_position */
extern T1 T622f94(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.position */
extern T0* T301f25(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.null_position */
extern T0* T73f778(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.position */
extern T0* T571f22(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.position */
extern T0* T994f94(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.position */
extern T0* T622f5(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.position */
extern T0* T1005f29(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.position */
extern T0* T341f7(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.position */
extern T0* T413f27(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_implicit_mark */
extern T1 T301f24(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_implicit_mark */
extern T1 T571f21(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_implicit_mark */
extern T1 T994f83(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_implicit_mark */
extern T1 T622f83(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_implicit_mark */
extern T1 T1005f28(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_implicit_mark */
extern T1 T341f19(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_implicit_mark */
extern T1 T413f31(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.position */
extern T0* T1525f34(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.position */
extern T0* T1501f57(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.position */
extern T0* T184f31(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.position */
extern T0* T891f57(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.position */
extern T0* T254f67(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.position */
extern T0* T242f37(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.position */
extern T0* T1524f28(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.position */
extern T0* T449f32(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.position */
extern T0* T282f9(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vggc1_error */
extern T1 T53f165(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vggc1a_error */
extern void T36f538(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_validity_error */
extern void T36f204(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.set_has_eiffel_error */
extern void T36f205(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ERROR_HANDLER.report_info */
extern void T36f184(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_info_message */
extern void T36f186(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_vggc1_error */
extern T1 T36f165(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.error_handler */
extern T0* T1309f12(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.set_fatal_error */
extern void T1309f24(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER3.process_qualified_like_type */
extern void T1338f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.process_qualified_like_type */
extern void T1324f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.process_qualified_like_type */
extern void T1308f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.process_like_type */
extern void T1308f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.check_like_type_validity */
extern void T1308f23(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vhpr3a_error */
extern void T53f537(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vhpr3a */
extern T0* T785c1159(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vhpr3_error */
extern T1 T53f164(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vhpr3a_error */
extern void T36f537(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vhpr3_error */
extern T1 T36f164(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.error_handler */
extern T0* T1308f6(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER1.set_fatal_error */
extern void T1308f12(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.process_qualified_like_type */
extern void T1619f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.check_qualified_like_identifier_validity */
extern void T1619f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.process_qualified_like_type */
extern void T1616f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.check_qualified_like_identifier_validity */
extern void T1616f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.process_qualified_like_type */
extern void T384f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.check_qualified_like_identifier_validity */
extern void T384f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].wipe_out */
extern void T710f14(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_CLASS].keep_head */
extern void T1423f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_CLASS].set_count */
extern void T1423f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].move_all_cursors_after */
extern void T710f18(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_CLASS].set_next_cursor */
extern void T1424f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_CLASS].set_after */
extern void T1424f6(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.reset */
extern void T241f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.wipe_out */
extern void T241f55(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE].keep_head */
extern void T625f5(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_TYPE].keep_head */
extern void T245f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_TYPE].set_count */
extern void T245f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.fixed_array */
extern T0* T241f4(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE].default_create */
extern T0* T625c3(GE_context* ac);
/* [detachable] ET_TYPE_CHECKER.tokens */
extern T0* T384f19(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.keep_first */
extern void T241f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.copy_type_context */
extern void T241f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.item */
extern T0* T241f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.set_root_context */
extern void T241f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgmc0e_error */
extern void T53f398(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vgmc0e */
extern T0* T785c1028(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_EXTERNAL_FUNCTION.lower_name */
extern T0* T253f49(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.lower_name */
extern T0* T1513f38(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.lower_name */
extern T0* T1512f39(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.lower_name */
extern T0* T1511f38(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.lower_name */
extern T0* T371f56(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.lower_name */
extern T0* T369f52(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.lower_name */
extern T0* T359f53(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.lower_name */
extern T0* T360f53(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.to_text */
extern T0* T254f28(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.to_text */
extern T0* T72f27(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.to_text */
extern T0* T74f72(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.type_constraint_type */
extern T0* T254f104(GE_context* ac, T0* C);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.type */
extern T0* T1850f1(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.type_constraint_type */
extern T0* T72f96(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.type_constraint_type */
extern T0* T74f212(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.lower_name */
extern T0* T2020f35(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.name */
extern T0* T2020f42(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.tokens */
extern T0* T2020f38(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.lower_name */
extern T0* T622f124(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.name */
extern T0* T622f126(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.lower_name */
extern T0* T1545f40(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.name */
extern T0* T1545f41(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.tokens */
extern T0* T1545f14(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.lower_name */
extern T0* T1544f40(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.name */
extern T0* T1544f41(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.tokens */
extern T0* T1544f14(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.lower_name */
extern T0* T973f51(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.name */
extern T0* T973f52(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.lower_name */
extern T0* T981f48(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.name */
extern T0* T981f49(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.tokens */
extern T0* T981f7(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.lower_name */
extern T0* T994f125(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.name */
extern T0* T994f126(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.alias_not_symbol_name */
extern T0* T73f891(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.alias_xor_symbol_name */
extern T0* T73f888(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.alias_or_else_symbol_name */
extern T0* T73f883(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.alias_or_symbol_name */
extern T0* T73f881(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.alias_implies_symbol_name */
extern T0* T73f875(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.alias_and_then_symbol_name */
extern T0* T73f869(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.alias_and_symbol_name */
extern T0* T73f867(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.position */
extern T0* T2020f34(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.position */
extern T0* T1545f39(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.position */
extern T0* T1544f39(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.position */
extern T0* T973f12(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.position */
extern T0* T981f5(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vgmc_error */
extern T1 T53f105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vgmc0e_error */
extern void T36f398(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.reportable_vgmc_error */
extern T1 T36f105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.same_named_context */
extern T1 T241f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.same_named_context_with_type_marks */
extern T1 T241f41(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NESTED_TYPE_CONTEXT.same_named_type_with_type_marks */
extern T1 T241f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T241f15ot1(T0* a1, T0** a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.put_last */
extern void T241f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE].force */
extern void T625f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_TYPE].fill_with */
extern void T245f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_TYPE].extend */
extern void T245f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_named_type_with_type_marks */
extern T1 T1501f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1501f15ot1(T0* a1, T0** a2);
/* [detachable] ET_TUPLE_TYPE.root_context */
extern T0* T254f15(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.root_context */
extern T0* T72f21(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.root_context */
extern T0* T241f1(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.tokens */
extern T0* T1501f29(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.as_nested_type_context */
extern T0* T254f34(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.to_nested_type_context */
extern T0* T254f35(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.make_with_capacity */
extern T0* T241c51(GE_context* ac, T0* a1, T6 a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.make_ast_list_with_capacity */
extern void T241f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE].make */
extern T0* T625f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_TYPE].make_empty */
extern T0* T245c5(GE_context* ac, T6 a1);
/* [detachable] ET_CLASS_TYPE.as_nested_type_context */
extern T0* T72f36(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.to_nested_type_context */
extern T0* T72f37(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.as_nested_type_context */
extern T0* T74f83(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.to_nested_type_context */
extern T0* T74f84(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.as_nested_type_context */
extern T0* T241f7(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_named_formal_parameter_type_with_type_marks */
extern T1 T1501f14(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1501f14ot1(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.is_type_detachable_with_type_mark */
extern T1 T891f24(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T891f24ot1(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.tokens */
extern T0* T891f40(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_expanded */
extern T1 T1501f31(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.is_expanded */
extern T1 T891f32(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_type_detachable_with_type_mark */
extern T1 T1526f30(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1526f30ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.overridden_type_mark */
extern T0* T1526f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLICIT_TYPE_MARK.overridden_type_mark */
extern T0* T301f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_type_mark */
extern T0* T73f45(GE_context* ac, T0* C, T1 a1, T1 a2, T1 a3, T1 a4, T1 a5);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_no_type_mark */
extern T0* T73f83(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.make */
extern T0* T301c26(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_reference_type_mark */
extern T0* T73f82(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.set_reference_mark */
extern void T301f31(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_expanded_type_mark */
extern T0* T73f81(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.set_expanded_mark */
extern void T301f30(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_separate_type_mark */
extern T0* T73f80(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.set_separate_mark */
extern void T301f29(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_separate_reference_type_mark */
extern T0* T73f79(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_separate_expanded_type_mark */
extern T0* T73f78(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_detachable_reference_type_mark */
extern T0* T73f77(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.set_detachable_mark */
extern void T301f27(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_detachable_expanded_type_mark */
extern T0* T73f76(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_detachable_separate_type_mark */
extern T0* T73f75(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_detachable_separate_reference_type_mark */
extern T0* T73f74(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_detachable_separate_expanded_type_mark */
extern T0* T73f73(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_attached_type_mark */
extern T0* T73f23(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.set_attached_mark */
extern void T301f28(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_attached_reference_type_mark */
extern T0* T73f72(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_attached_expanded_type_mark */
extern T0* T73f71(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_attached_separate_type_mark */
extern T0* T73f70(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_attached_separate_reference_type_mark */
extern T0* T73f69(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_attached_separate_expanded_type_mark */
extern T0* T73f68(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_separateness_mark */
extern T1 T341f17(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_separateness_mark */
extern T1 T1005f9(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_separateness_mark */
extern T1 T994f14(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_separateness_mark */
extern T1 T622f19(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_separateness_mark */
extern T1 T571f7(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_separateness_mark */
extern T1 T413f12(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_separateness_mark */
extern T1 T301f6(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_expandedness_mark */
extern T1 T341f16(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_expandedness_mark */
extern T1 T1005f8(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_expandedness_mark */
extern T1 T994f13(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_expandedness_mark */
extern T1 T622f18(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.is_expandedness_mark */
extern T1 T571f6(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_expandedness_mark */
extern T1 T413f11(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_expandedness_mark */
extern T1 T301f8(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.overridden_type_mark */
extern T0* T571f16(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.overridden_type_mark */
extern T0* T994f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BRACKET_SYMBOL.overridden_type_mark */
extern T0* T622f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUESTION_MARK_SYMBOL.overridden_type_mark */
extern T0* T1005f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYMBOL.overridden_type_mark */
extern T0* T341f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_KEYWORD.overridden_type_mark */
extern T0* T413f5(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.adapted_base_class_with_seeded_feature */
extern T0* T1524f40(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_CLASS.adapted_base_class_with_seeded_feature */
extern T0* T74f208(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.adapted_base_class_with_seeded_feature */
extern T0* T254f71(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.adapted_base_class_with_seeded_feature */
extern T0* T242f41(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T242f41ot1(T0* a1, T0** a2);
extern T1 T242f41ot2(T0* a1, T0** a2);
extern T1 T242f41ot3(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.tokens */
extern T0* T242f8(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.adapted_base_class_with_seeded_feature */
extern T0* T1501f34(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T1501f34ot1(T0* a1, T0** a2);
extern T1 T1501f34ot2(T0* a1, T0** a2);
extern T1 T1501f34ot3(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.adapted_base_class_with_seeded_feature */
extern T0* T891f35(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T891f35ot1(T0* a1, T0** a2);
extern T1 T891f35ot2(T0* a1, T0** a2);
extern T1 T891f35ot3(T0* a1, T0** a2);
/* [detachable] ET_CLASS_TYPE.adapted_base_class_with_seeded_feature */
extern T0* T72f70(GE_context* ac, T0* C, T6 a1, T0* a2);
/* detachable ET_FORMAL_ARGUMENT_LIST.item */
extern T0* T255f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS.seeded_feature */
extern T0* T74f197(GE_context* ac, T0* C, T6 a1);
/* detachable ET_LIKE_FEATURE.index */
extern T6 T1524f31(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.is_like_argument */
extern T1 T1524f30(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_argument */
extern T1 T184f9(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.tokens */
extern T0* T1524f29(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.adapted_base_class_with_seeded_feature */
extern T0* T1526f6(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T1526f6ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.tokens */
extern T0* T1526f37(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.adapted_base_class_with_seeded_feature */
extern T0* T1525f39(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T1525f39ot1(T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.adapted_base_class_with_named_feature */
extern T0* T1501f33(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1501f33ot1(T0* a1, T0** a2);
extern T1 T1501f33ot2(T0* a1, T0** a2);
extern T1 T1501f33ot3(T0* a1, T0** a2);
/* [detachable] ET_TUPLE_TYPE.named_feature */
extern T0* T254f113(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.named_procedure */
extern T0* T254f116(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.named_procedure */
extern T0* T74f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PROCEDURE_LIST.named_feature */
extern T0* T258f5(GE_context* ac, T0* C, T0* a1);
extern T1 T258f5ot1(T0* a1, T0** a2);
/* detachable ET_ALIAS_NAME_LIST.has_call_name */
extern T1 T911f4(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALIAS_FREE_NAME.same_call_name */
extern T1 T1518f17(GE_context* ac, T0* C, T0* a1);
extern T1 T1518f17ot1(T0* a1, T0** a2);
/* detachable ET_ALIAS_FREE_NAME.string_ */
extern T0* T1518f7(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.operator_name */
extern T0* T1518f8(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_freeop */
extern T1 T973f50(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.tokens */
extern T0* T973f9(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_freeop */
extern T1 T1518f51(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.tokens */
extern T0* T1518f9(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.same_call_name */
extern T1 T1030f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_prefix_not_symbol */
extern T1 T2020f32(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_prefix_not_symbol */
extern T1 T622f123(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_prefix_not_symbol */
extern T1 T1545f38(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_prefix_not_symbol */
extern T1 T1544f38(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_prefix_not_symbol */
extern T1 T973f31(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_prefix_not_symbol */
extern T1 T981f29(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_prefix_not_symbol */
extern T1 T994f105(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_prefix_not_symbol */
extern T1 T1518f48(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_prefix_not_symbol */
extern T1 T1030f78(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.tokens */
extern T0* T1030f79(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_prefix_not_symbol */
extern T1 T184f82(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_prefix_not */
extern T1 T2020f31(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_prefix_not */
extern T1 T622f122(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_prefix_not */
extern T1 T1545f37(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_prefix_not */
extern T1 T1544f37(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_prefix_not */
extern T1 T973f30(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_prefix_not */
extern T1 T981f28(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_prefix_not */
extern T1 T994f104(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_prefix_not */
extern T1 T1518f47(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_prefix_not */
extern T1 T1030f77(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_prefix_not */
extern T1 T184f81(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_prefix_plus */
extern T1 T2020f30(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_prefix_plus */
extern T1 T622f121(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_prefix_plus */
extern T1 T1545f36(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_prefix_plus */
extern T1 T1544f36(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_prefix_plus */
extern T1 T973f49(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_prefix_plus */
extern T1 T981f47(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_prefix_plus */
extern T1 T994f124(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_prefix_plus */
extern T1 T1518f46(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_prefix_plus */
extern T1 T1030f76(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_prefix_plus */
extern T1 T184f80(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_prefix_minus */
extern T1 T2020f29(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_prefix_minus */
extern T1 T622f120(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_prefix_minus */
extern T1 T1545f35(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_prefix_minus */
extern T1 T1544f35(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_prefix_minus */
extern T1 T973f48(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_prefix_minus */
extern T1 T981f46(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_prefix_minus */
extern T1 T994f123(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_prefix_minus */
extern T1 T1518f45(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_prefix_minus */
extern T1 T1030f75(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_prefix_minus */
extern T1 T184f79(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_dotdot */
extern T1 T2020f28(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_dotdot */
extern T1 T622f119(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_dotdot */
extern T1 T1545f34(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_dotdot */
extern T1 T1544f34(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_dotdot */
extern T1 T973f47(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_dotdot */
extern T1 T981f45(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_dotdot */
extern T1 T994f122(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_dotdot */
extern T1 T1518f44(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_dotdot */
extern T1 T1030f74(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_dotdot */
extern T1 T184f78(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_xor_symbol */
extern T1 T2020f27(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_xor_symbol */
extern T1 T622f118(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_xor_symbol */
extern T1 T1545f33(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_xor_symbol */
extern T1 T1544f33(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_xor_symbol */
extern T1 T973f46(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_xor_symbol */
extern T1 T981f44(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_xor_symbol */
extern T1 T994f121(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_xor_symbol */
extern T1 T1518f43(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_xor_symbol */
extern T1 T1030f73(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_xor_symbol */
extern T1 T184f77(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_xor */
extern T1 T2020f26(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_xor */
extern T1 T622f117(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_xor */
extern T1 T1545f32(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_xor */
extern T1 T1544f32(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_xor */
extern T1 T973f45(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_xor */
extern T1 T981f43(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_xor */
extern T1 T994f120(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_xor */
extern T1 T1518f42(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_xor */
extern T1 T1030f72(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_xor */
extern T1 T184f76(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_times */
extern T1 T2020f25(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_times */
extern T1 T622f116(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_times */
extern T1 T1545f31(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_times */
extern T1 T1544f31(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_times */
extern T1 T973f44(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_times */
extern T1 T981f42(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_times */
extern T1 T994f119(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_times */
extern T1 T1518f41(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_times */
extern T1 T1030f71(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_times */
extern T1 T184f75(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_power */
extern T1 T2020f24(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_power */
extern T1 T622f115(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_power */
extern T1 T1545f30(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_power */
extern T1 T1544f30(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_power */
extern T1 T973f43(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_power */
extern T1 T981f41(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_power */
extern T1 T994f118(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_power */
extern T1 T1518f40(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_power */
extern T1 T1030f70(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_power */
extern T1 T184f74(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_plus */
extern T1 T2020f23(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_plus */
extern T1 T622f114(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_plus */
extern T1 T1545f29(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_plus */
extern T1 T1544f29(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_plus */
extern T1 T973f42(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_plus */
extern T1 T981f40(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_plus */
extern T1 T994f117(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_plus */
extern T1 T1518f39(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_plus */
extern T1 T1030f69(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_plus */
extern T1 T184f73(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_or_else_symbol */
extern T1 T2020f22(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_or_else_symbol */
extern T1 T622f90(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_or_else_symbol */
extern T1 T1545f10(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_or_else_symbol */
extern T1 T1544f10(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_or_else_symbol */
extern T1 T973f19(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_or_else_symbol */
extern T1 T981f19(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_or_else_symbol */
extern T1 T994f90(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_or_else_symbol */
extern T1 T1518f38(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_or_else_symbol */
extern T1 T1030f68(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_or_else_symbol */
extern T1 T184f45(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_or_else */
extern T1 T2020f21(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_or_else */
extern T1 T622f85(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_or_else */
extern T1 T1545f5(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_or_else */
extern T1 T1544f5(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_or_else */
extern T1 T973f14(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_or_else */
extern T1 T981f14(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_or_else */
extern T1 T994f85(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_or_else */
extern T1 T1518f37(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_or_else */
extern T1 T1030f67(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_or_else */
extern T1 T184f40(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_or_symbol */
extern T1 T2020f20(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_or_symbol */
extern T1 T622f93(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_or_symbol */
extern T1 T1545f13(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_or_symbol */
extern T1 T1544f13(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_or_symbol */
extern T1 T973f22(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_or_symbol */
extern T1 T981f22(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_or_symbol */
extern T1 T994f93(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_or_symbol */
extern T1 T1518f36(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_or_symbol */
extern T1 T1030f66(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_or_symbol */
extern T1 T184f48(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_or */
extern T1 T2020f19(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_or */
extern T1 T622f88(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_or */
extern T1 T1545f8(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_or */
extern T1 T1544f8(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_or */
extern T1 T973f17(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_or */
extern T1 T981f17(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_or */
extern T1 T994f88(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_or */
extern T1 T1518f35(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_or */
extern T1 T1030f65(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_or */
extern T1 T184f43(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_mod */
extern T1 T2020f18(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_mod */
extern T1 T622f113(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_mod */
extern T1 T1545f28(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_mod */
extern T1 T1544f28(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_mod */
extern T1 T973f41(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_mod */
extern T1 T981f39(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_mod */
extern T1 T994f116(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_mod */
extern T1 T1518f34(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_mod */
extern T1 T1030f64(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_mod */
extern T1 T184f72(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_minus */
extern T1 T2020f17(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_minus */
extern T1 T622f112(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_minus */
extern T1 T1545f27(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_minus */
extern T1 T1544f27(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_minus */
extern T1 T973f40(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_minus */
extern T1 T981f38(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_minus */
extern T1 T994f115(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_minus */
extern T1 T1518f33(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_minus */
extern T1 T1030f63(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_minus */
extern T1 T184f71(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_lt */
extern T1 T2020f16(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_lt */
extern T1 T622f111(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_lt */
extern T1 T1545f26(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_lt */
extern T1 T1544f26(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_lt */
extern T1 T973f39(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_lt */
extern T1 T981f37(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_lt */
extern T1 T994f114(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_lt */
extern T1 T1518f32(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_lt */
extern T1 T1030f62(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_lt */
extern T1 T184f70(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_le */
extern T1 T2020f15(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_le */
extern T1 T622f110(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_le */
extern T1 T1545f25(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_le */
extern T1 T1544f25(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_le */
extern T1 T973f38(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_le */
extern T1 T981f36(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_le */
extern T1 T994f113(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_le */
extern T1 T1518f31(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_le */
extern T1 T1030f61(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_le */
extern T1 T184f69(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_implies_symbol */
extern T1 T2020f14(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_implies_symbol */
extern T1 T622f91(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_implies_symbol */
extern T1 T1545f11(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_implies_symbol */
extern T1 T1544f11(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_implies_symbol */
extern T1 T973f20(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_implies_symbol */
extern T1 T981f20(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_implies_symbol */
extern T1 T994f91(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_implies_symbol */
extern T1 T1518f30(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_implies_symbol */
extern T1 T1030f60(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_implies_symbol */
extern T1 T184f46(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_implies */
extern T1 T2020f13(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_implies */
extern T1 T622f86(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_implies */
extern T1 T1545f6(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_implies */
extern T1 T1544f6(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_implies */
extern T1 T973f15(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_implies */
extern T1 T981f15(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_implies */
extern T1 T994f86(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_implies */
extern T1 T1518f29(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_implies */
extern T1 T1030f59(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_implies */
extern T1 T184f41(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_gt */
extern T1 T2020f12(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_gt */
extern T1 T622f109(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_gt */
extern T1 T1545f24(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_gt */
extern T1 T1544f24(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_gt */
extern T1 T973f37(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_gt */
extern T1 T981f35(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_gt */
extern T1 T994f112(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_gt */
extern T1 T1518f28(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_gt */
extern T1 T1030f58(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_gt */
extern T1 T184f68(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_ge */
extern T1 T2020f11(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_ge */
extern T1 T622f108(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_ge */
extern T1 T1545f23(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_ge */
extern T1 T1544f23(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_ge */
extern T1 T973f36(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_ge */
extern T1 T981f34(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_ge */
extern T1 T994f111(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_ge */
extern T1 T1518f27(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_ge */
extern T1 T1030f57(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_ge */
extern T1 T184f67(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_divide */
extern T1 T2020f10(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_divide */
extern T1 T622f107(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_divide */
extern T1 T1545f22(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_divide */
extern T1 T1544f22(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_divide */
extern T1 T973f35(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_divide */
extern T1 T981f33(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_divide */
extern T1 T994f110(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_divide */
extern T1 T1518f26(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_divide */
extern T1 T1030f56(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_divide */
extern T1 T184f66(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_div */
extern T1 T2020f9(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_div */
extern T1 T622f106(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_div */
extern T1 T1545f21(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_div */
extern T1 T1544f21(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_div */
extern T1 T973f34(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_div */
extern T1 T981f32(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_div */
extern T1 T994f109(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_div */
extern T1 T1518f25(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_div */
extern T1 T1030f55(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_div */
extern T1 T184f65(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_and_then_symbol */
extern T1 T2020f8(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_and_then_symbol */
extern T1 T622f89(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_and_then_symbol */
extern T1 T1545f9(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_and_then_symbol */
extern T1 T1544f9(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_and_then_symbol */
extern T1 T973f18(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_and_then_symbol */
extern T1 T981f18(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_and_then_symbol */
extern T1 T994f89(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_and_then_symbol */
extern T1 T1518f24(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_and_then_symbol */
extern T1 T1030f54(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_and_then_symbol */
extern T1 T184f44(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_and_then */
extern T1 T2020f7(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_and_then */
extern T1 T622f84(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_and_then */
extern T1 T1545f4(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_and_then */
extern T1 T1544f4(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_and_then */
extern T1 T973f13(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_and_then */
extern T1 T981f13(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_and_then */
extern T1 T994f84(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_and_then */
extern T1 T1518f23(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_and_then */
extern T1 T1030f53(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_and_then */
extern T1 T184f39(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_and_symbol */
extern T1 T2020f6(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_and_symbol */
extern T1 T622f92(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_and_symbol */
extern T1 T1545f12(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_and_symbol */
extern T1 T1544f12(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_and_symbol */
extern T1 T973f21(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_and_symbol */
extern T1 T981f21(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_and_symbol */
extern T1 T994f92(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_and_symbol */
extern T1 T1518f22(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_and_symbol */
extern T1 T1030f52(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_and_symbol */
extern T1 T184f47(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_infix_and */
extern T1 T2020f5(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_infix_and */
extern T1 T622f87(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_infix_and */
extern T1 T1545f7(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_infix_and */
extern T1 T1544f7(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_infix_and */
extern T1 T973f16(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_infix_and */
extern T1 T981f16(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_infix_and */
extern T1 T994f87(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix_and */
extern T1 T1518f21(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix_and */
extern T1 T1030f51(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_infix_and */
extern T1 T184f42(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_parenthesis */
extern T1 T2020f4(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_parenthesis */
extern T1 T622f105(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_parenthesis */
extern T1 T1545f20(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_parenthesis */
extern T1 T1544f20(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_parenthesis */
extern T1 T973f33(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_parenthesis */
extern T1 T981f31(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_parenthesis */
extern T1 T994f108(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_parenthesis */
extern T1 T1518f13(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_parenthesis */
extern T1 T1030f11(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_parenthesis */
extern T1 T184f64(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_bracket */
extern T1 T2020f45(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_bracket */
extern T1 T622f104(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_bracket */
extern T1 T1545f19(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_bracket */
extern T1 T1544f19(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_bracket */
extern T1 T973f32(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_bracket */
extern T1 T981f30(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_bracket */
extern T1 T994f107(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_bracket */
extern T1 T1518f12(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_bracket */
extern T1 T1030f10(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_bracket */
extern T1 T184f63(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.alias_names */
extern T0* T1516f53(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.alias_names */
extern T0* T184f54(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.alias_names */
extern T0* T476f35(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.alias_names */
extern T0* T1515f46(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.alias_names */
extern T0* T469f37(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.hash_code */
extern T6 T1516f9(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.hash_code */
extern T6 T476f9(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.hash_code */
extern T6 T1515f10(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.hash_code */
extern T6 T469f9(GE_context* ac, T0* C);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.named_feature */
extern T0* T1850f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.named_procedure */
extern T0* T1850f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINT_RENAME_LIST.index_of_old_name */
extern T6 T1084f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_RENAME.rename_pair */
extern T0* T1337f5(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_RENAME_LIST.rename_pair */
extern T0* T1084f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONSTRAINT_RENAME_LIST.item */
extern T0* T1084f10(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONSTRAINT_RENAME_LIST.index_of_new_name */
extern T6 T1084f8(GE_context* ac, T0* C, T0* a1);
extern T1 T1084f8ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.named_feature */
extern T0* T74f176(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.named_feature */
extern T0* T72f101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.named_procedure */
extern T0* T72f105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.adapted_base_class_with_named_feature */
extern T0* T891f34(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T891f34ot1(T0* a1, T0** a2);
extern T1 T891f34ot2(T0* a1, T0** a2);
extern T1 T891f34ot3(T0* a1, T0** a2);
/* [detachable] ET_CLASS.adapted_base_class_with_named_feature */
extern T0* T74f207(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.adapted_base_class_with_named_feature */
extern T0* T254f70(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.adapted_base_class_with_named_feature */
extern T0* T242f40(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T242f40ot1(T0* a1, T0** a2);
extern T1 T242f40ot2(T0* a1, T0** a2);
extern T1 T242f40ot3(T0* a1, T0** a2);
/* [detachable] ET_CLASS_TYPE.adapted_base_class_with_named_feature */
extern T0* T72f69(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.tokens */
extern T0* T1525f33(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.adapted_base_class_with_seeded_feature */
extern T0* T449f36(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.adapted_base_class_with_seeded_feature */
extern T0* T241f39(GE_context* ac, T0* C, T6 a1);
extern T1 T241f39ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.adapted_base_class_with_seeded_feature */
extern T0* T1474f25(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.valid_index */
extern T1 T241f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TUPLE_TYPE.context_adapted_base_class_with_seeded_feature */
extern T0* T254f83(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS_TYPE.context_adapted_base_class_with_seeded_feature */
extern T0* T72f85(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS.context_adapted_base_class_with_seeded_feature */
extern T0* T74f214(GE_context* ac, T0* C, T6 a1);
/* detachable ET_LIKE_FEATURE.adapted_base_class_with_named_feature */
extern T0* T1524f39(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.adapted_base_class_with_named_feature */
extern T0* T1526f5(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1526f5ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.adapted_base_class_with_named_feature */
extern T0* T1525f38(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1525f38ot1(T0* a1);
/* [detachable] ET_LIKE_CURRENT.adapted_base_class_with_named_feature */
extern T0* T449f35(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.adapted_base_class_with_named_feature */
extern T0* T241f38(GE_context* ac, T0* C, T0* a1);
extern T1 T241f38ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.adapted_base_class_with_named_feature */
extern T0* T1474f24(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.context_adapted_base_class_with_named_feature */
extern T0* T254f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.context_adapted_base_class_with_named_feature */
extern T0* T72f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.context_adapted_base_class_with_named_feature */
extern T0* T74f213(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_type_detachable_with_type_mark */
extern T1 T1525f23(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1525f23ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.overridden_type_mark */
extern T0* T1525f36(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.is_type_detachable_with_type_mark */
extern T1 T1524f33(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_LIKE_FEATURE.overridden_type_mark */
extern T0* T1524f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.is_type_detachable_with_type_mark */
extern T1 T449f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.is_type_detachable_with_type_mark */
extern T1 T241f34(GE_context* ac, T0* C, T0* a1);
extern T1 T241f34ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.is_type_detachable_with_type_mark */
extern T1 T1474f23(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIKE_N.tokens */
extern T0* T1474f14(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_N.overridden_type_mark */
extern T0* T1474f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.context_is_type_detachable_with_type_mark */
extern T1 T254f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.context_is_type_detachable_with_type_mark */
extern T1 T72f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.context_is_type_detachable_with_type_mark */
extern T1 T74f112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.overridden_type_mark */
extern T0* T449f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_type_detachable_with_type_mark */
extern T1 T74f129(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.is_type_detachable_with_type_mark */
extern T1 T254f48(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.is_type_detachable_with_type_mark */
extern T1 T72f45(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER.overridden_type_mark */
extern T0* T891f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_type_detachable_with_type_mark */
extern T1 T242f33(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T242f33ot1(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.overridden_type_mark */
extern T0* T242f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_type_detachable_with_type_mark */
extern T1 T1501f26(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1501f26ot1(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.is_type_attached_with_type_mark */
extern T1 T891f31(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T891f31ot1(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_type_attached_with_type_mark */
extern T1 T1526f52(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1526f52ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_type_attached_with_type_mark */
extern T1 T1525f55(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1525f55ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.is_type_attached_with_type_mark */
extern T1 T1524f53(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIKE_CURRENT.is_type_attached_with_type_mark */
extern T1 T449f52(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.context_is_type_attached_with_type_mark */
extern T1 T254f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.context_is_type_attached_with_type_mark */
extern T1 T72f112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.context_is_type_attached_with_type_mark */
extern T1 T74f241(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.is_type_attached_with_type_mark */
extern T1 T241f42(GE_context* ac, T0* C, T0* a1);
extern T1 T241f42ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.is_type_attached_with_type_mark */
extern T1 T1474f27(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_type_attached_with_type_mark */
extern T1 T242f32(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T242f32ot1(T0* a1, T0** a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_type_attached_with_type_mark */
extern T1 T1501f30(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1501f30ot1(T0* a1, T0** a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.attachment_type_conformance_mode */
extern T1 T241f5(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.attachment_type_conformance_mode */
extern T1 T254f32(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.attachment_type_conformance_mode */
extern T1 T72f33(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.attachment_type_conformance_mode */
extern T1 T74f81(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.same_named_formal_parameter_type_with_type_marks */
extern T1 T891f12(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T891f12ot1(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_named_formal_parameter_type_with_type_marks */
extern T1 T1526f14(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1526f14ot1(T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.to_nested_type_context */
extern T0* T241f8(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_named_formal_parameter_type_with_type_marks */
extern T1 T1525f12(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1525f12ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.same_named_formal_parameter_type_with_type_marks */
extern T1 T1524f12(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_named_formal_parameter_type_with_type_marks */
extern T1 T242f18(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T242f18ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.same_named_formal_parameter_type_with_type_marks */
extern T1 T449f28(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NESTED_TYPE_CONTEXT.same_named_formal_parameter_type_with_type_marks */
extern T1 T241f26(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T241f26ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.same_named_formal_parameter_type_with_type_marks */
extern T1 T1474f10(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.context_same_named_formal_parameter_type_with_type_marks */
extern T1 T254f63(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS_TYPE.context_same_named_formal_parameter_type_with_type_marks */
extern T1 T72f61(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS.context_same_named_formal_parameter_type_with_type_marks */
extern T1 T74f157(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS.same_named_formal_parameter_type_with_type_marks */
extern T1 T74f154(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.same_named_formal_parameter_type_with_type_marks */
extern T1 T254f60(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.same_named_formal_parameter_type_with_type_marks */
extern T1 T72f58(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.is_root_context */
extern T1 T254f14(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.is_root_context */
extern T1 T72f20(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_root_context */
extern T1 T74f126(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.is_root_context */
extern T1 T241f19(GE_context* ac, T0* C);
extern T1 T241f19ot1(T0* a1, T0** a2);
/* [detachable] ET_TUPLE_TYPE.context_base_type_actual */
extern T0* T254f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TUPLE_TYPE.base_type_actual */
extern T0* T254f54(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T254f54ot1(T0* a1, T0** a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.named_type */
extern T0* T1501f53(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.named_type_with_type_mark */
extern T0* T1501f16(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1501f16ot1(T0* a1, T0** a2);
extern T1 T1501f16ot2(T0* a1, T0** a2);
extern T1 T1501f16ot3(T0* a1, T0** a2);
/* [detachable] ET_CLASS.type_with_type_mark */
extern T0* T74f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.make */
extern T0* T72c127(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CLASS_TYPE.make_generic */
extern T0* T72c125(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS.overridden_type_mark */
extern T0* T74f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.tokens */
extern T0* T74f178(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_separate */
extern T1 T74f95(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.has_separate_mark */
extern T1 T74f107(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.type_with_type_mark */
extern T0* T254f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.set_tuple_keyword */
extern void T254f119(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.make */
extern T0* T254c118(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_TOKEN_CONSTANTS.tuple_keyword */
extern T0* T73f90(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.tokens */
extern T0* T254f24(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.overridden_type_mark */
extern T0* T254f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.is_separate */
extern T1 T254f22(GE_context* ac, T0* C);
/* [detachable] ET_IMPLICIT_TYPE_MARK.is_separate_mark */
extern T1 T301f4(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.type_with_type_mark */
extern T0* T72f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.overridden_type_mark */
extern T0* T72f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.tokens */
extern T0* T72f17(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.is_separate */
extern T1 T72f15(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.type_with_type_mark */
extern T0* T242f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.make */
extern T0* T242c77(GE_context* ac, T0* a1, T0* a2, T6 a3, T0* a4);
/* [detachable] ET_FORMAL_PARAMETER.type_with_type_mark */
extern T0* T891f41(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.type_with_type_mark */
extern T0* T1501f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.implementation_class */
extern T0* T242f4(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.implementation_class */
extern T0* T891f3(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.implementation_class */
extern T0* T1501f5(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.named_type_with_type_mark */
extern T0* T1526f8(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1526f8ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.named_type_with_type_mark */
extern T0* T1525f26(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1525f26ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.named_type_with_type_mark */
extern T0* T1524f36(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.named_type_with_type_mark */
extern T0* T242f6(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T242f6ot1(T0* a1, T0** a2);
extern T1 T242f6ot2(T0* a1, T0** a2);
extern T1 T242f6ot3(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.named_type_with_type_mark */
extern T0* T449f27(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.named_type_with_type_mark */
extern T0* T241f16(GE_context* ac, T0* C, T0* a1);
extern T1 T241f16ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.named_type_with_type_mark */
extern T0* T1474f16(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.context_named_type_with_type_mark */
extern T0* T72f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.context_named_type_with_type_mark */
extern T0* T74f156(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.context_named_type_with_type_mark */
extern T0* T254f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.named_type_with_type_mark */
extern T0* T74f145(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.named_type_with_type_mark */
extern T0* T254f45(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.named_type_with_type_mark */
extern T0* T72f50(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER.named_type_with_type_mark */
extern T0* T891f14(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T891f14ot1(T0* a1, T0** a2);
extern T1 T891f14ot2(T0* a1, T0** a2);
extern T1 T891f14ot3(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.named_type */
extern T0* T891f51(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.named_type */
extern T0* T1526f28(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.named_type */
extern T0* T1525f18(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.named_type */
extern T0* T1524f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.named_type */
extern T0* T242f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.named_type */
extern T0* T449f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.named_type */
extern T0* T74f140(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.named_type */
extern T0* T254f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.named_type */
extern T0* T72f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.context_base_type_actual */
extern T0* T72f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS_TYPE.base_type_actual */
extern T0* T72f62(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T72f62ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.context_base_type_actual */
extern T0* T74f148(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS.base_type_actual */
extern T0* T74f177(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.base_type_actual */
extern T0* T241f22(GE_context* ac, T0* C, T6 a1);
extern T1 T241f22ot1(T0* a1, T0** a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.base_type_actual */
extern T0* T1501f18(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T1501f18ot1(T0* a1, T0** a2);
extern T1 T1501f18ot2(T0* a1, T0** a2);
extern T1 T1501f18ot3(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.base_type_actual */
extern T0* T891f16(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T891f16ot1(T0* a1, T0** a2);
extern T1 T891f16ot2(T0* a1, T0** a2);
extern T1 T891f16ot3(T0* a1, T0** a2);
/* detachable ET_LIKE_FEATURE.base_type_actual */
extern T0* T1524f41(GE_context* ac, T0* C, T6 a1, T0* a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.base_type_actual */
extern T0* T1526f10(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T1526f10ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.base_type_actual */
extern T0* T1525f40(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T1525f40ot1(T0* a1);
/* [detachable] ET_LIKE_CURRENT.base_type_actual */
extern T0* T449f37(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_LIKE_N.base_type_actual */
extern T0* T1474f6(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.base_type_actual */
extern T0* T242f14(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T242f14ot1(T0* a1, T0** a2);
extern T1 T242f14ot2(T0* a1, T0** a2);
extern T1 T242f14ot3(T0* a1, T0** a2);
/* [detachable] ET_TUPLE_TYPE.context_base_type_actual_count */
extern T6 T254f46(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.base_type_actual_count */
extern T6 T254f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.actual_parameter_count */
extern T6 T254f49(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.context_base_type_actual_count */
extern T6 T72f51(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.base_type_actual_count */
extern T6 T72f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.actual_parameter_count */
extern T6 T72f46(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.context_base_type_actual_count */
extern T6 T74f147(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.base_type_actual_count */
extern T6 T74f136(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.formal_parameter_count */
extern T6 T74f144(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.base_type_actual_count */
extern T6 T241f21(GE_context* ac, T0* C);
extern T1 T241f21ot1(T0* a1, T0** a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.base_type_actual_count */
extern T6 T1501f17(GE_context* ac, T0* C, T0* a1);
extern T1 T1501f17ot1(T0* a1, T0** a2);
extern T1 T1501f17ot2(T0* a1, T0** a2);
extern T1 T1501f17ot3(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.base_type_actual_count */
extern T6 T891f15(GE_context* ac, T0* C, T0* a1);
extern T1 T891f15ot1(T0* a1, T0** a2);
extern T1 T891f15ot2(T0* a1, T0** a2);
extern T1 T891f15ot3(T0* a1, T0** a2);
/* detachable ET_LIKE_FEATURE.base_type_actual_count */
extern T6 T1524f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.base_type_actual_count */
extern T6 T1526f9(GE_context* ac, T0* C, T0* a1);
extern T1 T1526f9ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.base_type_actual_count */
extern T6 T1525f9(GE_context* ac, T0* C, T0* a1);
extern T1 T1525f9ot1(T0* a1);
/* [detachable] ET_LIKE_CURRENT.base_type_actual_count */
extern T6 T449f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_N.base_type_actual_count */
extern T6 T1474f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.base_type_actual_count */
extern T6 T242f13(GE_context* ac, T0* C, T0* a1);
extern T1 T242f13ot1(T0* a1, T0** a2);
extern T1 T242f13ot2(T0* a1, T0** a2);
extern T1 T242f13ot3(T0* a1, T0** a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.overridden_type_mark */
extern T0* T1501f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.ancestor */
extern T0* T74f146(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BASE_TYPE_LIST.base_type */
extern T0* T412f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_unknown */
extern T1 T74f99(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.is_unknown */
extern T1 T802f42(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.is_unknown */
extern T1 T887f10(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.is_unknown */
extern T1 T410f4(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.base_class */
extern T0* T241f20(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.actual_class */
extern T0* T570f22(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.actual_class */
extern T0* T74f92(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.named_base_class */
extern T0* T241f27(GE_context* ac, T0* C);
extern T1 T241f27ot1(T0* a1, T0** a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.named_base_class */
extern T0* T1501f20(GE_context* ac, T0* C, T0* a1);
extern T1 T1501f20ot1(T0* a1, T0** a2);
extern T1 T1501f20ot2(T0* a1, T0** a2);
extern T1 T1501f20ot3(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.named_base_class */
extern T0* T891f18(GE_context* ac, T0* C, T0* a1);
extern T1 T891f18ot1(T0* a1, T0** a2);
extern T1 T891f18ot2(T0* a1, T0** a2);
extern T1 T891f18ot3(T0* a1, T0** a2);
/* detachable ET_LIKE_FEATURE.named_base_class */
extern T0* T1524f26(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.named_base_class */
extern T0* T1526f15(GE_context* ac, T0* C, T0* a1);
extern T1 T1526f15ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.named_base_class */
extern T0* T1525f21(GE_context* ac, T0* C, T0* a1);
extern T1 T1525f21ot1(T0* a1);
/* [detachable] ET_LIKE_CURRENT.named_base_class */
extern T0* T449f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_N.named_base_class */
extern T0* T1474f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.named_base_class_in_context */
extern T0* T74f143(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.named_base_class */
extern T0* T242f19(GE_context* ac, T0* C, T0* a1);
extern T1 T242f19ot1(T0* a1, T0** a2);
extern T1 T242f19ot2(T0* a1, T0** a2);
extern T1 T242f19ot3(T0* a1, T0** a2);
/* [detachable] ET_TUPLE_TYPE.named_base_class_in_context */
extern T0* T254f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.named_base_class_in_context */
extern T0* T72f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.named_base_class */
extern T0* T254f3(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.named_base_class */
extern T0* T72f4(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.named_base_class */
extern T0* T74f23(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.same_named_type_with_type_marks */
extern T1 T891f13(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T891f13ot1(T0* a1, T0** a2);
/* detachable ET_LIKE_FEATURE.same_named_type_with_type_marks */
extern T1 T1524f34(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_named_type_with_type_marks */
extern T1 T1526f7(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1526f7ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_named_type_with_type_marks */
extern T1 T1525f24(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1525f24ot1(T0* a1);
/* [detachable] ET_LIKE_CURRENT.same_named_type_with_type_marks */
extern T1 T449f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.context_same_named_type_with_type_marks */
extern T1 T254f59(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS_TYPE.context_same_named_type_with_type_marks */
extern T1 T72f57(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS.context_same_named_type_with_type_marks */
extern T1 T74f153(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_LIKE_N.same_named_type_with_type_marks */
extern T1 T1474f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_named_type_with_type_marks */
extern T1 T242f5(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T242f5ot1(T0* a1, T0** a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.put */
extern void T241f60(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.cloned_type_context */
extern T0* T241f9(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.capacity */
extern T6 T241f10(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.remove_last */
extern void T241f57(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.like_0 */
extern T0* T73f233(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_N.make_zero */
extern T0* T1474c39(GE_context* ac, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.like_keyword */
extern T0* T73f240(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_like */
extern T0* T413c52(GE_context* ac);
/* [detachable] ET_KEYWORD.make_token */
extern void T413f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_KEYWORD.make_leaf */
extern void T413f45(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.last */
extern T0* T241f17(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.same_named_type_with_type_marks */
extern T1 T254f19(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_named_tuple_type_with_type_marks */
extern T1 T1501f13(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1501f13ot1(T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.same_named_tuple_type_with_type_marks */
extern T1 T891f11(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T891f11ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_named_tuple_type_with_type_marks */
extern T1 T1526f12(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1526f12ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_named_tuple_type_with_type_marks */
extern T1 T1525f11(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1525f11ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.same_named_tuple_type_with_type_marks */
extern T1 T1524f11(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_named_tuple_type_with_type_marks */
extern T1 T242f16(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T242f16ot1(T0* a1);
/* [detachable] ET_LIKE_CURRENT.same_named_tuple_type_with_type_marks */
extern T1 T449f17(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NESTED_TYPE_CONTEXT.same_named_tuple_type_with_type_marks */
extern T1 T241f24(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T241f24ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.same_named_tuple_type_with_type_marks */
extern T1 T1474f8(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.context_same_named_tuple_type_with_type_marks */
extern T1 T254f56(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS.context_same_named_tuple_type_with_type_marks */
extern T1 T74f150(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS_TYPE.context_same_named_tuple_type_with_type_marks */
extern T1 T72f54(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS.same_named_tuple_type_with_type_marks */
extern T1 T74f138(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.same_named_tuple_type_with_type_marks */
extern T1 T254f42(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.same_named_types */
extern T1 T1957f5(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_named_type */
extern T1 T1501f60(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FORMAL_PARAMETER.same_named_type */
extern T1 T891f60(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_named_type */
extern T1 T1526f23(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_named_type */
extern T1 T1525f13(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_LIKE_FEATURE.same_named_type */
extern T1 T1524f13(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_named_type */
extern T1 T242f24(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_LIKE_CURRENT.same_named_type */
extern T1 T449f18(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CLASS.same_named_type */
extern T1 T74f66(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_TUPLE_TYPE.same_named_type */
extern T1 T254f20(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.same_named_types */
extern T1 T240f12(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_TUPLE_TYPE.is_type_separate_with_type_mark */
extern T1 T254f50(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.scoop_mode */
extern T1 T241f6(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.scoop_mode */
extern T1 T254f33(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.scoop_mode */
extern T1 T74f82(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.scoop_mode */
extern T1 T72f34(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.same_named_tuple_type_with_type_marks */
extern T1 T72f30(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.is_generic */
extern T1 T254f25(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.same_named_type_with_type_marks */
extern T1 T74f185(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_named_class_type_with_type_marks */
extern T1 T1501f12(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1501f12ot1(T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.same_named_class_type_with_type_marks */
extern T1 T891f10(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T891f10ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_named_class_type_with_type_marks */
extern T1 T1526f11(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1526f11ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_named_class_type_with_type_marks */
extern T1 T1525f10(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1525f10ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.same_named_class_type_with_type_marks */
extern T1 T1524f10(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_named_class_type_with_type_marks */
extern T1 T242f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T242f15ot1(T0* a1);
/* [detachable] ET_LIKE_CURRENT.same_named_class_type_with_type_marks */
extern T1 T449f16(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NESTED_TYPE_CONTEXT.same_named_class_type_with_type_marks */
extern T1 T241f23(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T241f23ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.same_named_class_type_with_type_marks */
extern T1 T1474f7(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.context_same_named_class_type_with_type_marks */
extern T1 T254f55(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS.context_same_named_class_type_with_type_marks */
extern T1 T74f149(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS_TYPE.context_same_named_class_type_with_type_marks */
extern T1 T72f53(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS.same_named_class_type_with_type_marks */
extern T1 T74f137(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_FORMAL_PARAMETER_LIST.same_named_types */
extern T1 T603f12(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CLASS_TYPE.is_type_separate_with_type_mark */
extern T1 T72f35(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.is_type_separate_with_type_mark */
extern T1 T74f87(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.is_type_expanded_with_type_mark */
extern T1 T72f9(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.is_type_expanded_with_type_mark */
extern T1 T74f86(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.same_named_class_type_with_type_marks */
extern T1 T254f41(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.same_named_class_type_with_type_marks */
extern T1 T72f32(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.same_named_types */
extern T1 T577f6(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.same_named_types */
extern T1 T1956f5(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CLASS_TYPE.is_generic */
extern T1 T72f18(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.identity_type */
extern T0* T73f165(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.make */
extern T0* T449c79(GE_context* ac, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.current_keyword */
extern T0* T73f25(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.make */
extern T0* T282c32(GE_context* ac);
/* [detachable] ET_CURRENT.make_token */
extern void T282f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CURRENT.make_leaf */
extern void T282f35(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.tokens */
extern T0* T282f5(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.tokens */
extern T0* T449f11(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.tokens */
extern T0* T241f18(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.force_last */
extern void T241f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.resize */
extern void T241f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE].aliased_resized_area */
extern T0* T625f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_TYPE].aliased_resized_area */
extern T0* T245f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.new_capacity */
extern T6 T241f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER.reset_context_if_multiple_constraints */
extern void T405f11(GE_context* ac, T0* C, T1 a1, T0* a2, T0* a3);
/* [detachable] ET_TOKEN_CONSTANTS.detachable_like_current */
extern T0* T73f175(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.detachable_keyword */
extern T0* T73f162(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_detachable */
extern T0* T413c48(GE_context* ac);
/* [detachable] ET_NESTED_TYPE_CONTEXT.is_type_detachable */
extern T1 T241f37(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.attached_like_current */
extern T0* T73f21(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.attached_keyword */
extern T0* T73f61(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_attached */
extern T0* T413c43(GE_context* ac);
/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER.tokens */
extern T0* T405f7(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.is_type_attached */
extern T1 T241f36(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].item */
extern T0* T710f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].is_empty */
extern T1 T710f8(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vtat1c_error */
extern void T53f397(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vtat1c */
extern T0* T785c1027(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vtat1_error */
extern T1 T53f104(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vtat1c_error */
extern void T36f397(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vtat1_error */
extern T1 T36f104(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vtat2b_error */
extern void T53f394(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vtat2b */
extern T0* T785c1024(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_LIKE_FEATURE.name */
extern T0* T1524f3(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vtat2_error */
extern T1 T53f87(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vtat2b_error */
extern void T36f394(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vtat2_error */
extern T1 T36f87(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.depends_on_qualified_anchored_type */
extern T1 T1524f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.depends_on_qualified_anchored_type */
extern T1 T74f226(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.depends_on_qualified_anchored_type */
extern T1 T603f20(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.depends_on_qualified_anchored_type */
extern T1 T1501f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.depends_on_qualified_anchored_type */
extern T1 T891f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.depends_on_qualified_anchored_type */
extern T1 T254f79(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.depends_on_qualified_anchored_type */
extern T1 T1957f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.depends_on_qualified_anchored_type */
extern T1 T240f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.depends_on_qualified_anchored_type */
extern T1 T242f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.depends_on_qualified_anchored_type */
extern T1 T72f80(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.depends_on_qualified_anchored_type */
extern T1 T577f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.depends_on_qualified_anchored_type */
extern T1 T1956f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.depends_on_qualified_anchored_type */
extern T1 T1526f49(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.depends_on_qualified_anchored_type */
extern T1 T1525f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.depends_on_qualified_anchored_type */
extern T1 T449f46(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.resolve_identifier_type */
extern void T1526f89(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_IDENTIFIER.set_seed */
extern void T184f115(GE_context* ac, T0* C, T6 a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.resolve_identifier_type */
extern void T1525f93(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TUPLE_TYPE.named_query */
extern T0* T254f114(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.named_query */
extern T0* T1850f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.named_query */
extern T0* T72f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.named_query */
extern T0* T74f101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUERY_LIST.named_feature */
extern T0* T257f6(GE_context* ac, T0* C, T0* a1);
extern T1 T257f6ot1(T0* a1, T0** a2);
/* [detachable] ET_EXTERNAL_FUNCTION.alias_names */
extern T0* T253f71(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.alias_names */
extern T0* T1513f51(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.alias_names */
extern T0* T1512f51(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.alias_names */
extern T0* T1511f51(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.alias_names */
extern T0* T371f29(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.alias_names */
extern T0* T369f28(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.alias_names */
extern T0* T359f65(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.alias_names */
extern T0* T360f27(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.hash_code */
extern T6 T253f9(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.hash_code */
extern T6 T1513f9(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.hash_code */
extern T6 T1512f9(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.hash_code */
extern T6 T1511f9(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.hash_code */
extern T6 T371f10(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.hash_code */
extern T6 T369f9(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.hash_code */
extern T6 T359f9(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.hash_code */
extern T6 T360f8(GE_context* ac, T0* C);
/* [detachable] ET_TYPE_CHECKER.error_handler */
extern T0* T384f15(GE_context* ac, T0* C);
/* [detachable] ET_TYPE_CHECKER.set_fatal_error */
extern void T384f26(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].first */
extern T0* T710f7(GE_context* ac, T0* C);
/* [detachable] ET_TYPE_CHECKER.check_adapted_base_classes_validity */
extern void T384f38(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER.check_adapted_base_classes_validity */
extern void T405f13(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T405f13ot1(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgmc0a_error */
extern void T53f531(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vgmc0a */
extern T0* T785c1154(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vgmc0a_error */
extern void T36f531(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgmc0d_error */
extern void T53f530(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T6 a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vgmc0d */
extern T0* T785c1153(GE_context* ac, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T6 a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.report_vgmc0d_error */
extern void T36f530(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T6 a6, T0* a7);
/* [detachable] ET_TUPLE_TYPE.adapted_class_base_type_index_of_label */
extern T6 T254f115(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.index_of_label */
extern T6 T254f109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.index_of_label */
extern T6 T1957f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER.same_identifier */
extern T1 T184f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.index_of_label */
extern T6 T240f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.base_type_index_of_label */
extern T6 T1850f10(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.index_of_label */
extern T6 T74f248(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.index_of_label */
extern T6 T603f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.index_of_label */
extern T6 T72f117(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.index_of_label */
extern T6 T577f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.index_of_label */
extern T6 T1956f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.adapted_class_base_type_index_of_label */
extern T6 T72f115(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.adapted_class_base_type_index_of_label */
extern T6 T74f244(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.base_type_index_of_label */
extern T6 T241f49(GE_context* ac, T0* C, T0* a1);
extern T1 T241f49ot1(T0* a1, T0** a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.base_type_index_of_label */
extern T6 T1501f74(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1501f74ot1(T0* a1, T0** a2);
extern T1 T1501f74ot2(T0* a1, T0** a2);
extern T1 T1501f74ot3(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.base_type_index_of_label */
extern T6 T891f76(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T891f76ot1(T0* a1, T0** a2);
extern T1 T891f76ot2(T0* a1, T0** a2);
extern T1 T891f76ot3(T0* a1, T0** a2);
/* detachable ET_LIKE_FEATURE.base_type_index_of_label */
extern T6 T1524f81(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.base_type_index_of_label */
extern T6 T1526f78(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1526f78ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.base_type_index_of_label */
extern T6 T1525f81(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1525f81ot1(T0* a1);
/* [detachable] ET_LIKE_CURRENT.base_type_index_of_label */
extern T6 T449f74(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIKE_N.base_type_index_of_label */
extern T6 T1474f32(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.base_type_index_of_label */
extern T6 T74f250(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.base_type_index_of_label */
extern T6 T242f73(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T242f73ot1(T0* a1, T0** a2);
extern T1 T242f73ot2(T0* a1, T0** a2);
extern T1 T242f73ot3(T0* a1, T0** a2);
/* [detachable] ET_TUPLE_TYPE.base_type_index_of_label */
extern T6 T254f110(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.base_type_index_of_label */
extern T6 T72f122(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgmc0c_error */
extern void T53f529(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vgmc0c */
extern T0* T785c1152(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7);
/* detachable ET_EXTERNAL_PROCEDURE.lower_name */
extern T0* T1516f36(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.lower_name */
extern T0* T476f41(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.lower_name */
extern T0* T1515f37(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.lower_name */
extern T0* T469f35(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_vgmc0c_error */
extern void T36f529(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgmc0b_error */
extern void T53f528(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vgmc0b */
extern T0* T785c1151(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.report_vgmc0b_error */
extern void T36f528(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER.error_handler */
extern T0* T405f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].put_last */
extern void T710f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_CLASS].force */
extern void T1421f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_CLASS].fill_with */
extern void T1423f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_CLASS].extend */
extern void T1423f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].remove */
extern void T710f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].move_left */
extern void T710f21(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].move_cursors_left */
extern void T710f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_CLASS].set_position */
extern void T1424f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].remove_last */
extern void T710f19(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].move_last_cursors_after */
extern void T710f22(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].keep_first */
extern void T710f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER.set_fatal_error */
extern void T405f14(GE_context* ac, T0* C);
/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER.set_class_interface_error_only */
extern void T405f12(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.add_adapted_base_classes_to_list */
extern void T241f62(GE_context* ac, T0* C, T0* a1);
extern T1 T241f62ot1(T0* a1, T0** a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.add_adapted_base_classes_to_list */
extern void T1501f86(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1501f86ot1(T0* a1, T0** a2);
extern T1 T1501f86ot2(T0* a1, T0** a2);
extern T1 T1501f86ot3(T0* a1, T0** a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].force_last */
extern void T710f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].resize */
extern void T710f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_CLASS].aliased_resized_area */
extern T0* T1421f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_CLASS].aliased_resized_area */
extern T0* T1423f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].new_capacity */
extern T6 T710f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].extendible */
extern T1 T710f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FORMAL_PARAMETER.add_adapted_base_classes_to_list */
extern void T891f83(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T891f83ot1(T0* a1, T0** a2);
extern T1 T891f83ot2(T0* a1, T0** a2);
extern T1 T891f83ot3(T0* a1, T0** a2);
/* detachable ET_LIKE_FEATURE.add_adapted_base_classes_to_list */
extern void T1524f95(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.add_adapted_base_classes_to_list */
extern void T1526f90(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1526f90ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.add_adapted_base_classes_to_list */
extern void T1525f94(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1525f94ot1(T0* a1);
/* [detachable] ET_LIKE_CURRENT.add_adapted_base_classes_to_list */
extern void T449f87(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIKE_N.add_adapted_base_classes_to_list */
extern void T1474f41(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.add_adapted_base_classes_to_list */
extern void T74f326(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.add_adapted_base_classes_to_list */
extern void T242f84(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T242f84ot1(T0* a1, T0** a2);
extern T1 T242f84ot2(T0* a1, T0** a2);
extern T1 T242f84ot3(T0* a1, T0** a2);
/* [detachable] ET_TUPLE_TYPE.add_adapted_base_classes_to_list */
extern void T254f126(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.add_adapted_base_classes_to_list */
extern void T72f135(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.context_add_adapted_base_classes_to_list */
extern void T254f127(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.context_add_adapted_base_classes_to_list */
extern void T72f137(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.context_add_adapted_base_classes_to_list */
extern void T74f325(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.set */
extern void T241f59(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TYPE_CHECKER.report_qualified_anchored_type_supplier */
extern void T384f37(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_SUPPLIER_BUILDER.report_qualified_anchored_type_supplier */
extern void T1341f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_SUPPLIER_BUILDER.add_supplier */
extern void T1341f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].force_last */
extern void T948f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].slots_put */
extern void T948f59(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].clashes_put */
extern void T948f58(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].slots_item */
extern T6 T948f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].hash_position */
extern T6 T948f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.hash_code */
extern T6 T570f37(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.hash_code */
extern T6 T74f184(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].resize */
extern void T948f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].clashes_resize */
extern void T948f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].special_integer_ */
extern T0* T948f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].key_storage_resize */
extern void T948f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].item_storage_resize */
extern void T948f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_CLASS].aliased_resized_area */
extern T0* T1653f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_NAMED_CLASS].aliased_resized_area */
extern T0* T1651f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].key_storage_item */
extern T0* T948f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].item_storage_item */
extern T0* T948f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].clashes_item */
extern T6 T948f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].slots_resize */
extern void T948f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].new_modulus */
extern T6 T948f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].new_capacity */
extern T6 T948f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].item_storage_put */
extern void T948f56(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_CLASS].force */
extern void T1653f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_NAMED_CLASS].fill_with */
extern void T1651f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_NAMED_CLASS].extend */
extern void T1651f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].search_position */
extern void T948f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].key_equality_tester */
extern T0* T948f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].unset_found_item */
extern void T948f46(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.root_context */
extern T0* T74f159(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.seed */
extern T6 T1526f36(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.seed */
extern T6 T1525f32(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.process_qualified_like_type */
extern void T1618f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.check_qualified_like_identifier_validity */
extern void T1618f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.set_fatal_error */
extern void T1618f12(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.has_flattening_error */
extern T1 T570f39(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.has_flattening_error */
extern T1 T74f122(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.features_flattened */
extern T1 T570f38(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.features_flattened */
extern T1 T74f121(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.process_qualified_like_type */
extern void T1617f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.check_qualified_like_identifier_validity */
extern void T1617f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.process_qualified_like_type */
extern void T1330f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.check_qualified_like_identifier_validity */
extern void T1330f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.tokens */
extern T0* T1330f11(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.error_handler */
extern T0* T1330f12(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.set_fatal_error */
extern void T1330f27(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].force_last */
extern void T1329f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].slots_put */
extern void T1329f44(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].clashes_put */
extern void T1329f45(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].slots_item */
extern T6 T1329f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].hash_position */
extern T6 T1329f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].resize */
extern void T1329f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].clashes_resize */
extern void T1329f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].special_integer_ */
extern T0* T1329f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].key_storage_resize */
extern void T1329f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].item_storage_resize */
extern void T1329f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS].aliased_resized_area */
extern T0* T1347f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_CLASS].aliased_resized_area */
extern T0* T1346f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].key_storage_item */
extern T0* T1329f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].item_storage_item */
extern T0* T1329f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].clashes_item */
extern T6 T1329f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].slots_resize */
extern void T1329f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].new_modulus */
extern T6 T1329f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].new_capacity */
extern T6 T1329f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].item_storage_put */
extern void T1329f48(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS].force */
extern void T1347f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLASS].fill_with */
extern void T1346f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLASS].extend */
extern void T1346f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].search_position */
extern void T1329f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].key_equality_tester */
extern T0* T1329f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].unset_found_item */
extern void T1329f38(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.add_adapted_base_classes_to_list */
extern void T1330f28(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T1330f28ot1(T0* a1, T0** a2);
extern T1 T1330f28ot2(T0* a1, T0** a2);
/* detachable ET_EXTERNAL_PROCEDURE.implementation_class */
extern T0* T1516f3(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.implementation_class */
extern T0* T476f3(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.implementation_class */
extern T0* T1515f3(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.implementation_class */
extern T0* T469f2(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.implementation_class */
extern T0* T253f3(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.implementation_class */
extern T0* T1513f3(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.implementation_class */
extern T0* T1512f3(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.implementation_class */
extern T0* T1511f3(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.implementation_class */
extern T0* T371f5(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.implementation_class */
extern T0* T369f3(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.implementation_class */
extern T0* T359f3(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.implementation_class */
extern T0* T360f3(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.copy_to_type_context */
extern void T241f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.copy_to_type_context */
extern void T74f324(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.process_qualified_like_type */
extern void T1322f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.add_qualified_like_identifier_to_sorter */
extern void T1322f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.process_qualified_like_type */
extern void T1321f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.resolve_qualified_like_identifier */
extern void T1321f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.process_qualified_like_type */
extern void T1323f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.check_qualified_like_identifier_tuple_actual_parameters_unfolded */
extern void T1323f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.check_qualified_like_identifier_vtct_validity */
extern void T1323f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.process_qualified_like_type */
extern void T1320f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.resolve_qualified_like_identifier */
extern void T1320f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.resolve_type */
extern void T1320f8(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.process_qualified_like_type */
extern void T1331f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.resolve_qualified_like_identifier */
extern void T1331f15(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.process */
extern void T1525f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.process_qualified_like_braced_type */
extern void T1325f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.process_qualified_like_braced_type */
extern void T1309f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.process_qualified_like_braced_type */
extern void T1338f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.process_qualified_like_braced_type */
extern void T1324f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.process_qualified_like_braced_type */
extern void T1308f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.process_qualified_like_braced_type */
extern void T1619f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.process_qualified_like_braced_type */
extern void T1616f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.process_qualified_like_braced_type */
extern void T384f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.process_qualified_like_braced_type */
extern void T1618f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.process_qualified_like_braced_type */
extern void T1617f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.process_qualified_like_braced_type */
extern void T1330f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.process_qualified_like_braced_type */
extern void T1322f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.process_qualified_like_braced_type */
extern void T1321f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.process_qualified_like_braced_type */
extern void T1323f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.process_qualified_like_braced_type */
extern void T1320f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.process_qualified_like_braced_type */
extern void T1331f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.process */
extern void T1524f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.process_like_feature */
extern void T1325f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.process_like_feature */
extern void T1309f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.process_like_feature */
extern void T1338f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.process_like_feature */
extern void T1324f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.process_like_feature */
extern void T1308f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.process_like_feature */
extern void T1619f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.process_like_feature */
extern void T1616f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.process_like_feature */
extern void T384f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.check_like_feature_validity */
extern void T384f33(GE_context* ac, T0* C, T0* a1);
extern T1 T384f33ot1(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vtat1a_error */
extern void T53f396(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vtat1a */
extern T0* T785c1026(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vtat1a_error */
extern void T36f396(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vtat1b_error */
extern void T53f395(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vtat1b */
extern T0* T785c1025(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vtat1b_error */
extern void T36f395(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_LIKE_FEATURE.resolve_like_argument */
extern void T1524f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_procedure */
extern T1 T371f45(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_procedure */
extern T1 T369f44(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_procedure */
extern T1 T360f43(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_procedure */
extern T1 T359f44(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_procedure */
extern T1 T253f46(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_procedure */
extern T1 T1516f67(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_procedure */
extern T1 T1515f62(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_procedure */
extern T1 T1513f63(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_procedure */
extern T1 T1512f67(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_procedure */
extern T1 T1511f65(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_procedure */
extern T1 T469f60(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_procedure */
extern T1 T476f58(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.set_argument */
extern void T184f121(GE_context* ac, T0* C, T1 a1);
/* detachable ET_FORMAL_ARGUMENT_LIST.index_of */
extern T6 T255f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_ARGUMENT.formal_argument */
extern T0* T256f6(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_COMMA_ARGUMENT.formal_argument */
extern T0* T1767f6(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.resolve_like_feature */
extern void T1524f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.process_like_feature */
extern void T1618f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.process_like_feature */
extern void T1617f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.process_like_feature */
extern void T1330f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.check_like_feature_validity */
extern void T1330f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.process_like_feature */
extern void T1322f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.add_like_feature_to_sorter */
extern void T1322f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].force_relation */
extern void T1693f15(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].put_indexed_relation */
extern void T1693f17(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].replace */
extern void T699f14(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].item */
extern T0* T699f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].replace */
extern void T263f22(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].item */
extern T6 T263f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].reset */
extern void T1693f14(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].force */
extern void T1693f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].put */
extern void T1693f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].put_last */
extern void T699f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable DS_LINKABLE [[attached] INTEGER_32]].force */
extern void T1415f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].fill_with */
extern void T1414f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].extend */
extern void T1414f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].put_last */
extern void T263f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].put_new */
extern void T1930f36(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].key_storage_put */
extern void T1930f51(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIKE_FEATURE].force */
extern void T1932f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_LIKE_FEATURE].fill_with */
extern void T1931f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_LIKE_FEATURE].extend */
extern void T1931f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].item_storage_put */
extern void T1930f50(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].slots_put */
extern void T1930f45(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].clashes_put */
extern void T1930f46(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].slots_item */
extern T6 T1930f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].hash_position */
extern T6 T1930f32(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.hash_code */
extern T6 T1524f80(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].clashes_item */
extern T6 T1930f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].unset_found_item */
extern void T1930f39(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].put_last */
extern void T1694f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].resize */
extern void T699f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable DS_LINKABLE [[attached] INTEGER_32]].aliased_resized_area */
extern T0* T1415f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].aliased_resized_area */
extern T0* T1414f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].resize */
extern void T263f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].resize */
extern void T1930f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].clashes_resize */
extern void T1930f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].special_integer_ */
extern T0* T1930f23(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].key_storage_resize */
extern void T1930f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIKE_FEATURE].aliased_resized_area */
extern T0* T1932f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_LIKE_FEATURE].aliased_resized_area */
extern T0* T1931f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].item_storage_resize */
extern void T1930f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].key_storage_item */
extern T0* T1930f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].slots_resize */
extern void T1930f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].new_modulus */
extern T6 T1930f21(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].resize */
extern void T1694f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].capacity */
extern T6 T1693f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].count */
extern T6 T1693f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].index_of */
extern T6 T1693f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].found_item */
extern T6 T1930f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].item_storage_item */
extern T6 T1930f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].found */
extern T1 T1930f27(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].search */
extern void T1930f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].search_position */
extern void T1930f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.process_like_feature */
extern void T1321f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.process_like_feature */
extern void T1323f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.process_like_feature */
extern void T1320f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.resolve_like_feature */
extern void T1320f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.error_handler */
extern T0* T1320f6(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.set_fatal_error */
extern void T1320f19(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.process_like_feature */
extern void T1331f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.process */
extern void T242f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.process_formal_parameter_type */
extern void T1325f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.process_formal_parameter_type */
extern void T1309f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.check_formal_parameter_type_constraint */
extern void T1309f45(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_LIKE_CURRENT.type_constraint_type */
extern T0* T449f70(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.type_constraint_type */
extern T0* T242f69(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.type_constraint_type */
extern T0* T1526f72(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.type_constraint_type */
extern T0* T1525f75(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.type_constraint_type */
extern T0* T1524f73(GE_context* ac, T0* C);
/* detachable ET_TYPE_RENAME_CONSTRAINT.type */
extern T0* T1502f1(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER3.process_formal_parameter_type */
extern void T1338f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.process_formal_parameter_type */
extern void T1324f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.process_formal_parameter_type */
extern void T1308f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.process_formal_parameter_type */
extern void T1619f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.process_formal_parameter_type */
extern void T1616f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.process_formal_parameter_type */
extern void T384f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.process_formal_parameter_type */
extern void T1618f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.process_formal_parameter_type */
extern void T1617f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.process_formal_parameter_type */
extern void T1330f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.process_formal_parameter_type */
extern void T1322f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.process_formal_parameter_type */
extern void T1321f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.process_formal_parameter_type */
extern void T1323f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.process_formal_parameter_type */
extern void T1320f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.process_formal_parameter_type */
extern void T1331f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.process */
extern void T449f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.process_like_current */
extern void T1325f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.process_like_current */
extern void T1309f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.process_like_current */
extern void T1338f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.process_like_current */
extern void T1324f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.process_like_current */
extern void T1308f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.process_like_current */
extern void T1619f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.process_like_current */
extern void T1616f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.process_like_current */
extern void T384f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.check_like_current_validity */
extern void T384f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.process_like_current */
extern void T1618f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.process_like_current */
extern void T1617f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.process_like_current */
extern void T1330f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.process_like_current */
extern void T1322f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.process_like_current */
extern void T1321f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.process_like_current */
extern void T1323f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.process_like_current */
extern void T1320f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.process_like_current */
extern void T1331f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.process */
extern void T254f123(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.process_tuple_type */
extern void T1325f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.check_tuple_type_constraint */
extern void T1325f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.process_tuple_type */
extern void T1309f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.check_tuple_type_constraint */
extern void T1309f43(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.add_base_type_dependency */
extern void T1309f35(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.add_n_items_to_list */
extern void T1309f29(GE_context* ac, T0* C, T6 a1, T6 a2, T0* a3);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].force_last */
extern void T263f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].new_capacity */
extern T6 T263f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].extendible */
extern T1 T263f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT].force_last */
extern void T1659f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE_CONSTRAINT].force */
extern void T1905f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE_CONSTRAINT].fill_with */
extern void T1904f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE_CONSTRAINT].extend */
extern void T1904f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT].resize */
extern void T1659f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE_CONSTRAINT].aliased_resized_area */
extern T0* T1905f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE_CONSTRAINT].aliased_resized_area */
extern T0* T1904f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT].new_capacity */
extern T6 T1659f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT].extendible */
extern T1 T1659f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.make */
extern T0* T1850c12(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_LIKE_CURRENT.renames */
extern T0* T449f76(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.renames */
extern T0* T254f111(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.renames */
extern T0* T242f74(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.renames */
extern T0* T1526f77(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.renames */
extern T0* T1525f80(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.renames */
extern T0* T1524f79(GE_context* ac, T0* C);
/* detachable ET_TYPE_RENAME_CONSTRAINT.renames */
extern T0* T1502f2(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.renames */
extern T0* T72f97(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.renames */
extern T0* T74f225(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.is_named_type */
extern T1 T254f78(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_named_type */
extern T1 T1501f64(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.is_named_type */
extern T1 T891f64(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_named_type */
extern T1 T1526f48(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_named_type */
extern T1 T1525f51(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.is_named_type */
extern T1 T1524f49(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_named_type */
extern T1 T242f48(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.is_named_type */
extern T1 T449f45(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_named_type */
extern T1 T74f222(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.is_named_type */
extern T1 T72f79(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER3.process_tuple_type */
extern void T1338f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.check_tuple_type_validity */
extern void T1338f19(GE_context* ac, T0* C, T0* a1);
extern T1 T1338f19ot1(T0* a1, T0** a2);
/* [detachable] ET_PARENT_CHECKER2.process_tuple_type */
extern void T1324f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.check_tuple_type_validity */
extern void T1324f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.process_tuple_type */
extern void T1308f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.check_tuple_type_validity */
extern void T1308f21(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvhpr5a_error */
extern void T53f536(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_gvhpr5a */
extern T0* T785c1158(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_COMPRESSED_POSITION.to_text */
extern T0* T193f9(GE_context* ac, T0* C);
/* [detachable] ET_COMPRESSED_POSITION.append_to_string */
extern void T193f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_INTEGER_ROUTINES.append_decimal_integer */
extern void T772f6(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] ET_COMPRESSED_POSITION.integer_ */
extern T0* T193f10(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.position_to_text */
extern T0* T1005f38(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.append_position_to_string */
extern void T1005f48(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUESTION_MARK_SYMBOL.integer_ */
extern T0* T1005f40(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.position_to_text */
extern T0* T413f40(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.append_position_to_string */
extern void T413f101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_KEYWORD.integer_ */
extern T0* T413f41(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.position_to_text */
extern T0* T341f89(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.append_position_to_string */
extern void T341f121(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYMBOL.integer_ */
extern T0* T341f90(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.position_to_text */
extern T0* T184f96(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.append_position_to_string */
extern void T184f131(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER.integer_ */
extern T0* T184f101(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.position_to_text */
extern T0* T994f128(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.append_position_to_string */
extern void T994f158(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.integer_ */
extern T0* T994f129(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.position_to_text */
extern T0* T622f125(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.append_position_to_string */
extern void T622f136(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BRACKET_SYMBOL.integer_ */
extern T0* T622f127(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvhpr5_error */
extern T1 T53f163(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvhpr5a_error */
extern void T36f536(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_gvhpr5_error */
extern T1 T36f163(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.process_tuple_type */
extern void T1619f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.check_tuple_type_validity */
extern void T1619f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.set_fatal_error */
extern void T1619f6(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.process_tuple_type */
extern void T1616f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.check_tuple_type_validity */
extern void T1616f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.has_syntax_error */
extern T1 T74f115(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.has_syntax_error */
extern T1 T570f44(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_parsed */
extern T1 T74f114(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.is_parsed */
extern T1 T570f43(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.set_fatal_error */
extern void T1616f6(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_preparsed */
extern T1 T74f100(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.is_preparsed */
extern T1 T570f42(GE_context* ac, T0* C);
/* [detachable] ET_TYPE_CHECKER.process_tuple_type */
extern void T384f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.check_tuple_type_validity */
extern void T384f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.reset_fatal_error */
extern void T384f40(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.process_tuple_type */
extern void T1618f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.check_actual_parameters_validity */
extern void T1618f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.process_tuple_type */
extern void T1617f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.check_tuple_type_validity */
extern void T1617f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.set_fatal_error */
extern void T1617f14(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.has_ancestors_error */
extern T1 T74f120(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.has_ancestors_error */
extern T1 T570f46(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.ancestors_built */
extern T1 T74f113(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.ancestors_built */
extern T1 T570f45(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.process_tuple_type */
extern void T1330f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.check_actual_parameters_validity */
extern void T1330f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.reset_fatal_error */
extern void T1330f16(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.process_tuple_type */
extern void T1322f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.add_actual_parameters_to_sorter */
extern void T1322f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.process_tuple_type */
extern void T1321f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.resolve_tuple_type */
extern void T1321f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.process_tuple_type */
extern void T1323f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.check_tuple_type_tuple_actual_parameters_unfolded */
extern void T1323f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.check_tuple_type_vtct_validity */
extern void T1323f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.process_tuple_type */
extern void T1320f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.resolve_actual_parameters */
extern void T1320f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.reset_fatal_error */
extern void T1320f21(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.process_tuple_type */
extern void T1331f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.resolve_tuple_type */
extern void T1331f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.process */
extern void T72f130(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.process_class_type */
extern void T1325f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.check_class_type_constraint */
extern void T1325f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vtcg3a_error */
extern void T53f403(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vtcg3a */
extern T0* T785c1033(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.types_to_text */
extern T0* T254f107(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.append_types_to_string */
extern void T254f129(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.types_to_text */
extern T0* T242f72(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.append_types_to_string */
extern void T242f86(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.types_to_text */
extern T0* T449f73(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.append_types_to_string */
extern void T449f90(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.types_to_text */
extern T0* T1526f75(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.append_types_to_string */
extern void T1526f92(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.types_to_text */
extern T0* T1525f78(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.append_types_to_string */
extern void T1525f98(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.types_to_text */
extern T0* T1524f76(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.append_types_to_string */
extern void T1524f98(GE_context* ac, T0* C, T0* a1);
/* detachable ET_TYPE_CONSTRAINT_LIST.types_to_text */
extern T0* T1281f9(GE_context* ac, T0* C);
/* detachable ET_TYPE_CONSTRAINT_LIST.append_types_to_string */
extern void T1281f16(GE_context* ac, T0* C, T0* a1);
/* detachable ET_TYPE_RENAME_CONSTRAINT.types_to_text */
extern T0* T1502f5(GE_context* ac, T0* C);
/* detachable ET_TYPE_RENAME_CONSTRAINT.append_types_to_string */
extern void T1502f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.types_to_text */
extern T0* T74f245(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.append_types_to_string */
extern void T74f328(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.types_to_text */
extern T0* T72f95(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.append_types_to_string */
extern void T72f136(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.to_text */
extern T0* T1501f11(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.to_text */
extern T0* T891f9(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.to_text */
extern T0* T242f12(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vtcg3_error */
extern T1 T53f109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vtcg3a_error */
extern void T36f403(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vtcg3_error */
extern T1 T36f109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.error_handler */
extern T0* T1325f7(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.resolve_unfolded_tuple_actual_parameters_2 */
extern void T72f134(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.make */
extern T0* T1956c19(GE_context* ac, T0* a1, T0* a2, T6 a3);
/* [detachable] ET_IDENTIFIER.set_position */
extern void T184f118(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_ECF_LIBRARY.tuple_type */
extern T0* T187f25(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.tuple_type */
extern T0* T58f76(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.tuple_type */
extern T0* T40f7(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.make */
extern T0* T1957c22(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_CLASS.resolve_unfolded_tuple_actual_parameters_2 */
extern void T74f323(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_QUALIFIED_LIKE_TYPE.conforms_to_constraint */
extern T1 T1526f51(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_QUALIFIED_LIKE_TYPE.conforms_to_constraint_with_type_marks */
extern T1 T1526f60(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_QUALIFIED_LIKE_TYPE.conforms_to_type_with_type_marks */
extern T1 T1526f13(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1526f13ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.conforms_to_type_with_type_marks */
extern T1 T1525f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1525f25ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.conforms_to_type_with_type_marks */
extern T1 T1524f35(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_CLASS.conforms_to_type_with_type_marks */
extern T1 T74f163(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.conforms_from_class_type_with_type_marks */
extern T1 T1501f21(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1501f21ot1(T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_type_reference_with_type_mark */
extern T1 T1501f35(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1501f35ot1(T0* a1, T0** a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_reference */
extern T1 T1501f32(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.is_reference */
extern T1 T891f33(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_type_reference_with_type_mark */
extern T1 T1526f53(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1526f53ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_type_reference_with_type_mark */
extern T1 T1525f56(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1525f56ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.is_type_reference_with_type_mark */
extern T1 T1524f54(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_type_reference_with_type_mark */
extern T1 T242f35(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T242f35ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.is_type_reference_with_type_mark */
extern T1 T449f50(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.context_is_type_reference_with_type_mark */
extern T1 T254f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.context_is_type_reference_with_type_mark */
extern T1 T72f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.is_type_reference_with_type_mark */
extern T1 T241f40(GE_context* ac, T0* C, T0* a1);
extern T1 T241f40ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.is_type_reference_with_type_mark */
extern T1 T1474f26(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.context_is_type_reference_with_type_mark */
extern T1 T74f215(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_type_reference_with_type_mark */
extern T1 T74f191(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.is_type_reference_with_type_mark */
extern T1 T254f65(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.is_type_expanded_with_type_mark */
extern T1 T254f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.is_type_reference_with_type_mark */
extern T1 T72f64(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER.is_type_reference_with_type_mark */
extern T1 T891f36(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T891f36ot1(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.conforms_from_class_type_with_type_marks */
extern T1 T891f19(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T891f19ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.conforms_from_class_type_with_type_marks */
extern T1 T1526f24(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1526f24ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.conforms_from_class_type_with_type_marks */
extern T1 T1525f14(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1525f14ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.conforms_from_class_type_with_type_marks */
extern T1 T1524f14(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.conforms_from_class_type_with_type_marks */
extern T1 T242f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T242f25ot1(T0* a1);
/* [detachable] ET_LIKE_CURRENT.conforms_from_class_type_with_type_marks */
extern T1 T449f19(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_NESTED_TYPE_CONTEXT.conforms_from_class_type_with_type_marks */
extern T1 T241f28(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T241f28ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.conforms_from_class_type_with_type_marks */
extern T1 T1474f18(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_TUPLE_TYPE.context_conforms_from_class_type_with_type_marks */
extern T1 T254f57(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.context_conforms_from_class_type_with_type_marks */
extern T1 T72f55(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS.context_conforms_from_class_type_with_type_marks */
extern T1 T74f151(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.conforms_from_class_type_with_type_marks */
extern T1 T254f31(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_CLASS_TYPE.conforms_from_class_type_with_type_marks */
extern T1 T72f31(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_CLASS.is_dotnet */
extern T1 T74f106(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_in_dotnet_assembly */
extern T1 T74f109(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.is_dotnet_assembly */
extern T1 T802f43(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.is_dotnet_assembly */
extern T1 T887f22(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.is_dotnet_assembly */
extern T1 T410f5(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_system_object_class */
extern T1 T74f105(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.conforming_ancestor */
extern T0* T74f104(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.conforms_to_types */
extern T1 T577f7(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_TUPLE_TYPE.conforms_to_type */
extern T1 T254f43(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.conforms_to_types */
extern T1 T1956f6(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.conforms_to_type */
extern T1 T1501f61(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.conforms_to_type_with_type_marks */
extern T1 T1501f19(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1501f19ot1(T0* a1, T0** a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T1501f23(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1501f23ot1(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.recursive_formal_constraints */
extern T0* T891f30(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T891f21(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T891f21ot1(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T1526f26(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1526f26ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T1525f16(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1525f16ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T1524f16(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T242f31(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T242f31ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T449f30(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_NESTED_TYPE_CONTEXT.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T241f30(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T241f30ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T1474f20(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_TUPLE_TYPE.context_conforms_from_formal_parameter_type_with_type_marks */
extern T1 T254f66(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.context_conforms_from_formal_parameter_type_with_type_marks */
extern T1 T72f65(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS.context_conforms_from_formal_parameter_type_with_type_marks */
extern T1 T74f192(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T74f190(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.conforms_to_type_with_type_marks */
extern T1 T1850f6(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_BASE_TYPE_CONSTRAINT_LIST.conforms_to_type_with_type_marks */
extern T1 T1661f9(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_TUPLE_TYPE.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T254f64(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_CLASS_TYPE.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T72f63(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_FORMAL_PARAMETER.conforms_to_type_with_type_marks */
extern T1 T891f17(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T891f17ot1(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.conforms_to_type */
extern T1 T891f61(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_QUALIFIED_LIKE_TYPE.conforms_to_type */
extern T1 T1526f27(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.conforms_to_type */
extern T1 T1525f17(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_LIKE_FEATURE.conforms_to_type */
extern T1 T1524f17(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.conforms_to_type */
extern T1 T242f27(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_LIKE_CURRENT.conforms_to_type */
extern T1 T449f21(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS.conforms_to_type */
extern T1 T74f139(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS_TYPE.conforms_to_type */
extern T1 T72f43(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.conforms_to_types */
extern T1 T1957f6(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_FORMAL_PARAMETER_LIST.conforms_to_types */
extern T1 T603f13(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.conforms_to_types */
extern T1 T240f13(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS.conforms_from_class_type_with_type_marks */
extern T1 T74f80(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_TUPLE_TYPE.conforms_to_type_with_type_marks */
extern T1 T254f16(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.conforms_from_tuple_type_with_type_marks */
extern T1 T1501f22(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1501f22ot1(T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.conforms_from_tuple_type_with_type_marks */
extern T1 T891f20(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T891f20ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.conforms_from_tuple_type_with_type_marks */
extern T1 T1526f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1526f25ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.conforms_from_tuple_type_with_type_marks */
extern T1 T1525f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T1525f15ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.conforms_from_tuple_type_with_type_marks */
extern T1 T1524f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.conforms_from_tuple_type_with_type_marks */
extern T1 T242f26(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T242f26ot1(T0* a1);
/* [detachable] ET_LIKE_CURRENT.conforms_from_tuple_type_with_type_marks */
extern T1 T449f20(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_NESTED_TYPE_CONTEXT.conforms_from_tuple_type_with_type_marks */
extern T1 T241f29(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T241f29ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.conforms_from_tuple_type_with_type_marks */
extern T1 T1474f19(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_TUPLE_TYPE.context_conforms_from_tuple_type_with_type_marks */
extern T1 T254f58(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.context_conforms_from_tuple_type_with_type_marks */
extern T1 T72f56(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS.context_conforms_from_tuple_type_with_type_marks */
extern T1 T74f152(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.conforms_from_tuple_type_with_type_marks */
extern T1 T254f29(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.tuple_conforms_to_types */
extern T1 T1957f14(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.tuple_conforms_to_types */
extern T1 T240f14(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS_TYPE.conforms_from_tuple_type_with_type_marks */
extern T1 T72f28(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_CLASS.conforms_from_tuple_type_with_type_marks */
extern T1 T74f79(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.conforms_to_type_with_type_marks */
extern T1 T242f17(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
extern T1 T242f17ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS_TYPE.conforms_to_type_with_type_marks */
extern T1 T72f22(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_LIKE_CURRENT.conforms_to_type_with_type_marks */
extern T1 T449f26(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_NESTED_TYPE_CONTEXT.conforms_to_type_with_type_marks */
extern T1 T241f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T241f25ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.conforms_to_type_with_type_marks */
extern T1 T1474f9(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_TUPLE_TYPE.context_conforms_to_type_with_type_marks */
extern T1 T254f61(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.context_conforms_to_type_with_type_marks */
extern T1 T72f59(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS.context_conforms_to_type_with_type_marks */
extern T1 T74f155(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.type_constraint */
extern T0* T242f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_LIKE_CURRENT.type_constraint */
extern T0* T449f72(GE_context* ac, T0* C, T6 a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.type_constraint */
extern T0* T1526f74(GE_context* ac, T0* C, T6 a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.type_constraint */
extern T0* T1525f77(GE_context* ac, T0* C, T6 a1);
/* detachable ET_LIKE_FEATURE.type_constraint */
extern T0* T1524f75(GE_context* ac, T0* C, T6 a1);
/* detachable ET_TYPE_CONSTRAINT_LIST.type_constraint */
extern T0* T1281f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS_TYPE.current_type_constraint */
extern T0* T72f116(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.current_type_constraint */
extern T0* T254f103(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.current_type_constraint */
extern T0* T242f68(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.current_type_constraint */
extern T0* T449f69(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.current_type_constraint */
extern T0* T1526f71(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.current_type_constraint */
extern T0* T1525f74(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.current_type_constraint */
extern T0* T1524f72(GE_context* ac, T0* C);
/* detachable ET_TYPE_RENAME_CONSTRAINT.current_type_constraint */
extern T0* T1502f6(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.current_type_constraint */
extern T0* T74f246(GE_context* ac, T0* C);
/* detachable ET_TYPE_CONSTRAINT_LIST.item */
extern T0* T1281f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_TYPE_RENAME_CONSTRAINT.type_constraint */
extern T0* T1502f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.type_constraint_count */
extern T6 T242f70(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.type_constraint_count */
extern T6 T449f71(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.type_constraint_count */
extern T6 T1526f73(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.type_constraint_count */
extern T6 T1525f76(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.type_constraint_count */
extern T6 T1524f74(GE_context* ac, T0* C);
/* detachable ET_TYPE_CONSTRAINT_LIST.count */
extern T6 T1281f3(GE_context* ac, T0* C);
/* detachable ET_TYPE_RENAME_CONSTRAINT.count */
extern T6 T1502f3(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.conforms_to_constraint */
extern T1 T1525f54(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.conforms_to_constraint_with_type_marks */
extern T1 T1525f62(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_LIKE_FEATURE.conforms_to_constraint */
extern T1 T1524f52(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_LIKE_FEATURE.conforms_to_constraint_with_type_marks */
extern T1 T1524f59(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.conforms_to_constraint */
extern T1 T242f51(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.conforms_to_constraint_with_type_marks */
extern T1 T242f62(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_LIKE_CURRENT.conforms_to_constraint */
extern T1 T449f48(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_LIKE_CURRENT.conforms_to_constraint_with_type_marks */
extern T1 T449f51(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_CLASS.conforms_to_constraint */
extern T1 T74f211(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS.conforms_to_constraint_with_type_marks */
extern T1 T74f223(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_TUPLE_TYPE.conforms_to_constraint */
extern T1 T254f81(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_TUPLE_TYPE.conforms_to_constraint_with_type_marks */
extern T1 T254f96(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_CLASS_TYPE.conforms_to_constraint */
extern T1 T72f83(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS_TYPE.conforms_to_constraint_with_type_marks */
extern T1 T72f100(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.conforms_to_constraint */
extern T1 T1501f59(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.conforms_to_constraint_with_type_marks */
extern T1 T1501f71(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_FORMAL_PARAMETER.conforms_to_constraint */
extern T1 T891f59(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FORMAL_PARAMETER.conforms_to_constraint_with_type_marks */
extern T1 T891f73(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ECF_LIBRARY.detachable_any_type */
extern T0* T187f20(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.detachable_any_type */
extern T0* T58f84(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.detachable_any_type */
extern T0* T40f34(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.constraint */
extern T0* T891f26(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.set_fatal_error */
extern void T1325f14(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.resolve_unfolded_tuple_actual_parameters_1 */
extern void T72f133(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.unfolded_empty_tuple_actual_parameters */
extern T0* T187f28(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.unfolded_empty_tuple_actual_parameters */
extern T0* T58f90(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.unfolded_empty_tuple_actual_parameters */
extern T0* T40f40(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.position */
extern T0* T1957f15(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.tokens */
extern T0* T1957f19(GE_context* ac, T0* C);
/* [detachable] ET_LABELED_COMMA_ACTUAL_PARAMETER.position */
extern T0* T1769f8(GE_context* ac, T0* C);
/* detachable ET_LABELED_ACTUAL_PARAMETER.position */
extern T0* T1522f8(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_routine_class */
extern T1 T74f164(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.obsolete_routine_type_mode */
extern T1 T187f99(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.obsolete_routine_type_mode */
extern T1 T58f131(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.obsolete_routine_type_mode */
extern T1 T40f118(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.resolve_unfolded_tuple_actual_parameters_1 */
extern void T74f322(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.current_universe */
extern T0* T1325f6(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.process_class_type */
extern void T1309f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.check_class_type_constraint */
extern void T1309f33(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvtcg5a_error */
extern void T53f402(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_gvtcg5a */
extern T0* T785c1032(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvtcg5_error */
extern T1 T53f108(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvtcg5a_error */
extern void T36f402(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_gvtcg5_error */
extern T1 T36f108(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_type_reference */
extern T1 T1526f50(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_type_reference */
extern T1 T1525f53(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.is_type_reference */
extern T1 T1524f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_type_reference */
extern T1 T242f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.is_type_reference */
extern T1 T449f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_type_reference */
extern T1 T74f210(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.is_type_reference */
extern T1 T254f80(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.is_type_reference */
extern T1 T72f82(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_type_reference */
extern T1 T1501f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.is_type_reference */
extern T1 T891f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvtcg5b_error */
extern void T53f401(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_gvtcg5b */
extern T0* T785c1031(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_gvtcg5b_error */
extern void T36f401(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vtug2a_error */
extern void T53f400(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vtug2a */
extern T0* T785c1030(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vtug2_error */
extern T1 T53f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vtug2a_error */
extern void T36f400(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vtug2_error */
extern T1 T36f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.current_universe */
extern T0* T1309f15(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vtug1a_error */
extern void T53f399(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vtug1a */
extern T0* T785c1029(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.upper_name */
extern T0* T72f81(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vtug1_error */
extern T1 T53f106(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vtug1a_error */
extern void T36f399(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vtug1_error */
extern T1 T36f106(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vtct0a_error */
extern void T53f213(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vtct0a */
extern T0* T785c863(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vtct_error */
extern T1 T53f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vtct0a_error */
extern void T36f213(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vtct_error */
extern T1 T36f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.process_class_type */
extern void T1324f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.check_class_type_validity */
extern void T1324f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.error_handler */
extern T0* T1324f8(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER2.current_system */
extern T0* T1324f7(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER2.set_fatal_error */
extern void T1324f13(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER2.current_universe */
extern T0* T1324f6(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER1.process_class_type */
extern void T1308f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.check_class_type_validity */
extern void T1308f11(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_PARENT_CHECKER1.current_universe */
extern T0* T1308f7(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.process_class_type */
extern void T1619f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.check_class_type_validity */
extern void T1619f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.process_class_type */
extern void T1616f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.check_class_type_validity */
extern void T1616f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.process_class_type */
extern void T384f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.check_class_type_validity */
extern void T384f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.current_system */
extern T0* T384f18(GE_context* ac, T0* C);
/* [detachable] ET_TYPE_CHECKER.current_universe */
extern T0* T384f20(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.process_class_type */
extern void T1618f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.process_class_type */
extern void T1617f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.check_class_type_validity */
extern void T1617f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.process_class_type */
extern void T1330f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.process_class_type */
extern void T1322f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.process_class_type */
extern void T1321f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.resolve_class_type */
extern void T1321f16(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.has_identifier_anchored_type */
extern T1 T1501f63(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.has_identifier_anchored_type */
extern T1 T891f63(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.has_identifier_anchored_type */
extern T1 T1526f54(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.has_identifier_anchored_type */
extern T1 T1525f57(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.has_identifier_anchored_type */
extern T1 T1524f55(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.has_identifier_anchored_type */
extern T1 T242f52(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.has_identifier_anchored_type */
extern T1 T449f53(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.has_identifier_anchored_type */
extern T1 T74f218(GE_context* ac, T0* C);
/* detachable ET_FORMAL_PARAMETER_LIST.has_identifier_anchored_type */
extern T1 T603f21(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.has_identifier_anchored_type */
extern T1 T254f85(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.has_identifier_anchored_type */
extern T1 T1957f11(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.has_identifier_anchored_type */
extern T1 T240f19(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.has_identifier_anchored_type */
extern T1 T72f91(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.has_identifier_anchored_type */
extern T1 T577f11(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.has_identifier_anchored_type */
extern T1 T1956f11(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.current_universe */
extern T0* T1321f5(GE_context* ac, T0* C);
/* [detachable] ET_SIGNATURE_CHECKER.process_class_type */
extern void T1323f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.check_class_type_tuple_actual_parameters_unfolded */
extern void T1323f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.set_fatal_error */
extern void T1323f16(GE_context* ac, T0* C);
/* [detachable] ET_SIGNATURE_CHECKER.check_class_type_vtct_validity */
extern void T1323f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.error_handler */
extern T0* T1323f8(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.process_class_type */
extern void T1320f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.process_class_type */
extern void T1331f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.resolve_class_type */
extern void T1331f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.current_universe */
extern T0* T1331f5(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vtcg4a_error */
extern void T53f393(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vtcg4a */
extern T0* T785c1023(GE_context* ac, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vtcg4_error */
extern T1 T53f103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vtcg4a_error */
extern void T36f393(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.reportable_vtcg4_error */
extern T1 T36f103(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.is_creation_exported_to */
extern T1 T1516f71(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CLASS.is_deferred */
extern T1 T74f68(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.has_deferred_mark */
extern T1 T74f93(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_deferred */
extern T1 T413f32(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_creation_exported_to */
extern T1 T74f216(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CREATOR_LIST.is_exported_to */
extern T1 T1107f8(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CREATOR.is_exported_to */
extern T1 T1104f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CREATOR.has_feature_name */
extern T1 T1104f14(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLIENTS.has_descendant */
extern T1 T1065f11(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.has_ancestor */
extern T1 T74f240(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BASE_TYPE_LIST.has_class */
extern T1 T412f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLIENT.base_class */
extern T0* T403f3(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.ancestors_built_successfully */
extern T1 T74f161(GE_context* ac, T0* C);
/* detachable ET_CLIENTS.has_class */
extern T1 T1065f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLIENT_LIST.has_descendant */
extern T1 T402f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLIENT_LIST.has_class */
extern T1 T402f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CREATOR.is_empty */
extern T1 T1104f11(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_creation_exported_to */
extern T1 T476f60(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ONCE_PROCEDURE.is_creation_exported_to */
extern T1 T1515f64(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DO_PROCEDURE.is_creation_exported_to */
extern T1 T469f62(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vtcg4b_error */
extern void T53f392(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vtcg4b */
extern T0* T785c1022(GE_context* ac, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.report_vtcg4b_error */
extern void T36f392(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_CLASS.position */
extern T0* T74f209(GE_context* ac, T0* C);
/* detachable ET_INDEXING_LIST.position */
extern T0* T1173f9(GE_context* ac, T0* C);
/* detachable ET_INDEXING.position */
extern T0* T1499f3(GE_context* ac, T0* C);
/* detachable ET_INDEXING_TERM_LIST.position */
extern T0* T1184f6(GE_context* ac, T0* C);
/* detachable ET_INDEXING_TERM_LIST.tokens */
extern T0* T1184f8(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.position */
extern T0* T1489f20(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.position */
extern T0* T1488f20(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.position */
extern T0* T1487f20(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.position */
extern T0* T1486f20(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.position */
extern T0* T372f9(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.position */
extern T0* T1485f23(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.position */
extern T0* T1484f22(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.position */
extern T0* T1482f17(GE_context* ac, T0* C);
/* detachable ET_CUSTOM_ATTRIBUTE.position */
extern T0* T1500f6(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.position */
extern T0* T328f9(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.position */
extern T0* T1496f9(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.position */
extern T0* T1495f9(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.position */
extern T0* T1494f10(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.position */
extern T0* T1493f10(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.position */
extern T0* T1491f10(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.position */
extern T0* T1490f10(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.position */
extern T0* T1483f10(GE_context* ac, T0* C);
/* detachable ET_INDEXING_TERM_LIST.first */
extern T0* T1184f7(GE_context* ac, T0* C);
/* detachable ET_INDEXING_TERM_LIST.is_empty */
extern T1 T1184f5(GE_context* ac, T0* C);
/* detachable ET_INDEXING.tag */
extern T0* T1499f4(GE_context* ac, T0* C);
/* detachable ET_TAGGED_INDEXING.position */
extern T0* T1498f4(GE_context* ac, T0* C);
/* detachable ET_INDEXING_LIST.first */
extern T0* T1173f11(GE_context* ac, T0* C);
/* detachable ET_INDEXING_LIST.is_empty */
extern T1 T1173f10(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.position */
extern T0* T72f66(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_CREATOR.has_feature */
extern T1 T1081f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.features_flattened_successfully */
extern T1 T74f219(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.base_class */
extern T0* T254f11(GE_context* ac, T0* C);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.base_class */
extern T0* T1850f7(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.type_constraint */
extern T0* T254f106(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS.type_constraint */
extern T0* T74f221(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.type_constraint */
extern T0* T1850f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_BASE_TYPE_CONSTRAINT_LIST.type_constraint */
extern T0* T1661f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_BASE_TYPE_CONSTRAINT_LIST.item */
extern T0* T1661f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS_TYPE.type_constraint */
extern T0* T72f87(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONSTRAINT_CREATOR.feature_name */
extern T0* T1081f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONSTRAINT_CREATOR.item */
extern T0* T1081f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TUPLE_TYPE.type_constraint_count */
extern T6 T254f105(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.type_constraint_count */
extern T6 T74f220(GE_context* ac, T0* C);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.count */
extern T6 T1850f3(GE_context* ac, T0* C);
/* [detachable] ET_BASE_TYPE_CONSTRAINT_LIST.count */
extern T6 T1661f4(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.type_constraint_count */
extern T6 T72f86(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.constraint_base_types */
extern T0* T891f28(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER3.error_handler */
extern T0* T1338f5(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.index */
extern T6 T1501f7(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.index */
extern T6 T891f2(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.index */
extern T6 T242f3(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_CREATOR.is_empty */
extern T1 T1081f9(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.creation_procedures */
extern T0* T891f27(GE_context* ac, T0* C);
/* detachable ET_FORMAL_PARAMETER_LIST.formal_parameter */
extern T0* T603f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.type */
extern T0* T577f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.type */
extern T0* T1956f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.type */
extern T0* T1957f18(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FORMAL_PARAMETER_LIST.type */
extern T0* T603f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.type */
extern T0* T240f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_LIKE_FEATURE.type */
extern T0* T1524f24(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.type */
extern T0* T1526f35(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.type */
extern T0* T1525f30(GE_context* ac, T0* C);
/* [detachable] ET_LABELED_COMMA_ACTUAL_PARAMETER.type */
extern T0* T1769f3(GE_context* ac, T0* C);
/* detachable ET_LABELED_ACTUAL_PARAMETER.type */
extern T0* T1522f3(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.type */
extern T0* T242f10(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.type */
extern T0* T449f13(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.type */
extern T0* T254f30(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.type */
extern T0* T74f127(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.type */
extern T0* T72f29(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.count */
extern T6 T577f3(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.count */
extern T6 T1956f1(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.count */
extern T6 T1957f17(GE_context* ac, T0* C);
/* detachable ET_FORMAL_PARAMETER_LIST.count */
extern T6 T603f1(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.count */
extern T6 T240f1(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER3.set_fatal_error */
extern void T1338f10(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.interface_checked_successfully */
extern T1 T74f124(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.base_class */
extern T0* T74f70(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER2.process_class */
extern void T1324f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.process_class */
extern void T1308f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.process_class */
extern void T1325f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.process_class */
extern void T1309f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.process_class */
extern void T1619f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.process_class */
extern void T1618f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.process_class */
extern void T1617f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.process_class */
extern void T1616f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.process_class */
extern void T1331f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.process_class */
extern void T1330f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.process_class */
extern void T1323f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.process_class */
extern void T1322f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.process_class */
extern void T1321f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.process_class */
extern void T1320f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.process_class */
extern void T384f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PROVIDER_CHECKER.process_class */
extern void T457f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PROVIDER_CHECKER.internal_process_class */
extern void T457f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PROVIDER_CHECKER.check_cluster_dependence_constraints */
extern void T457f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS].forth */
extern void T1620f12(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].cursor_forth */
extern void T948f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].add_traversing_cursor */
extern void T948f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS].set_next_cursor */
extern void T1620f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS].set_position */
extern void T1620f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].remove_traversing_cursor */
extern void T948f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS].set_after */
extern void T1620f13(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gcdep_error */
extern void T53f409(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_cluster_error */
extern void T53f364(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLUSTER_ERROR.make_gcdep */
extern T0* T1940c25(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8].forth */
extern void T2041f11(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].cursor_forth */
extern void T2040f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].add_traversing_cursor */
extern void T2040f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8].set_next_cursor */
extern void T2041f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8].set_position */
extern void T2041f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].remove_traversing_cursor */
extern void T2040f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8].set_after */
extern void T2041f13(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].clashes_item */
extern T6 T2040f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8].key */
extern T0* T2041f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].cursor_key */
extern T0* T2040f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].key_storage_item */
extern T0* T2040f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8].after */
extern T1 T2041f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8].start */
extern void T2041f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].cursor_start */
extern void T2040f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].cursor_off */
extern T1 T2040f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8].off */
extern T1 T2041f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].new_cursor */
extern T0* T2040f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_WILDCARD, [attached] STRING_8].make */
extern T0* T2041c9(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].is_empty */
extern T1 T2040f29(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.full_lower_name */
extern T0* T887f12(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_NONE_GROUP.relative_lower_name */
extern T0* T887f13(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_NONE_GROUP.string_ */
extern T0* T887f15(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.relative_lower_name */
extern T0* T40f122(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].item */
extern T0* T907f9(GE_context* ac, T0* C, T6 a1);
/* detachable ET_SYSTEM.string_ */
extern T0* T40f127(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].is_empty */
extern T1 T907f10(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.shortest_path */
extern T0* T852f104(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_LIBRARY.shortest_path */
extern T0* T852f104ac1(GE_context* ac, T0* a1, T0* a2, T6 a3);
/* Creation of agent #2 in feature [detachable] ET_LIBRARY.shortest_path */
extern T0* T852f104ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_LIBRARY.shortest_path */
extern T0* T852f104ac3(GE_context* ac, T0* a1, T0* a2);
/* [detachable] KL_AGENT_ROUTINES [[attached] ET_UNIVERSE].negated */
extern T1 T1458f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] PREDICATE [[attached] TUPLE [[attached] ET_UNIVERSE]].set_rout_disp_final */
extern void T1459f8(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_UNIVERSE]].set_rout_disp_final */
extern void T838f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_LIBRARY.insert_in_hash_table */
extern void T852f170(GE_context* ac, T0* C, T0* a1, T6 a2, T0* a3);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].force_first */
extern void T907f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].put */
extern void T907f15(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].move_cursors_right */
extern void T907f18(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_UNIVERSE].set_position */
extern void T1634f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].move_right */
extern void T907f17(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_UNIVERSE].force */
extern void T1633f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_UNIVERSE].fill_with */
extern void T1632f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_UNIVERSE].extend */
extern void T1632f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].put_last */
extern void T907f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].resize */
extern void T907f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_UNIVERSE].aliased_resized_area */
extern T0* T1633f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_UNIVERSE].aliased_resized_area */
extern T0* T1632f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].new_capacity */
extern T6 T907f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].extendible */
extern T1 T907f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].before */
extern T1 T1456f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].back */
extern void T1456f16(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].cursor_back */
extern void T1455f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].add_traversing_cursor */
extern void T1455f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].set_next_cursor */
extern void T1456f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].set_position */
extern void T1456f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].remove_traversing_cursor */
extern void T1455f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].set_before */
extern void T1456f19(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].clashes_item */
extern T6 T1455f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].finish */
extern void T1456f15(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].cursor_finish */
extern void T1455f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].cursor_off */
extern T1 T1455f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].off */
extern T1 T1456f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].is_empty */
extern T1 T1455f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].forth */
extern void T1456f14(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].cursor_forth */
extern void T1455f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].set_after */
extern void T1456f17(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].go_after */
extern void T1456f13(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].cursor_go_after */
extern void T1455f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].has */
extern T1 T1455f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].search_position */
extern void T1455f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].hash_position */
extern T6 T1455f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.hash_code */
extern T6 T211f92(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.hash_code */
extern T6 T596f92(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.hash_code */
extern T6 T852f93(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.hash_code */
extern T6 T40f125(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.hash_code */
extern T6 T187f104(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.hash_code */
extern T6 T58f142(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].key_storage_item */
extern T0* T1455f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].slots_item */
extern T6 T1455f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.universes_do_if */
extern void T211f162(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_ASSEMBLIES.universes_do_if */
extern void T595f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY].item */
extern T0* T1398f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DOTNET_ASSEMBLY.universes_do_if */
extern void T596f162(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_LIBRARY.universes_do_if */
extern void T187f192(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.universes_do_if */
extern void T210f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].item */
extern T0* T594f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ADAPTED_LIBRARIES.universes_do_if */
extern void T188f11(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY].item */
extern T0* T578f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_LIBRARY.universes_do_if */
extern void T852f163(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_SYSTEM.universes_do_if */
extern void T40f206(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.universes_do_if */
extern void T58f280(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIBRARY.universe_actions */
extern T0* T852f105(GE_context* ac, T0* C);
/* [detachable] KL_AGENT_ROUTINES [[attached] ET_UNIVERSE].default_create */
extern T0* T1458c2(GE_context* ac);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].item */
extern T6 T1456f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].cursor_item */
extern T6 T1455f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].item_storage_item */
extern T6 T1455f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].key */
extern T0* T1456f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].cursor_key */
extern T0* T1455f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].after */
extern T1 T1456f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].start */
extern void T1456f12(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].cursor_start */
extern void T1455f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].new_cursor */
extern T0* T1455f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_UNIVERSE].make */
extern T0* T1456c11(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].force_last */
extern void T1455f37(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].key_storage_put */
extern void T1455f45(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_UNIVERSE].force */
extern void T1611f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_UNIVERSE].fill_with */
extern void T1609f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_UNIVERSE].extend */
extern void T1609f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].slots_put */
extern void T1455f44(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].clashes_put */
extern void T1455f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].resize */
extern void T1455f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].clashes_resize */
extern void T1455f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].special_integer_ */
extern T0* T1455f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].key_storage_resize */
extern void T1455f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_UNIVERSE].aliased_resized_area */
extern T0* T1611f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_UNIVERSE].aliased_resized_area */
extern T0* T1609f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].item_storage_resize */
extern void T1455f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].slots_resize */
extern void T1455f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].new_modulus */
extern T6 T1455f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].new_capacity */
extern T6 T1455f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].item_storage_put */
extern void T1455f41(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].unset_found_item */
extern void T1455f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].make_map */
extern T0* T1455c36(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].make_with_equality_testers */
extern void T1455f39(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].make_sparse_container */
extern void T1455f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].set_internal_cursor */
extern void T1455f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].make_slots */
extern void T1455f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].make_clashes */
extern void T1455f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].make_key_storage */
extern void T1455f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_UNIVERSE].make */
extern T0* T1611f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_UNIVERSE].make_empty */
extern T0* T1609c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_UNIVERSE].default_create */
extern T0* T1611c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_UNIVERSE].make_item_storage */
extern void T1455f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].make */
extern T0* T907c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].set_internal_cursor */
extern void T907f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_UNIVERSE].new_cursor */
extern T0* T907f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_UNIVERSE].make */
extern T0* T1634c4(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_UNIVERSE].make */
extern T0* T1633f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_UNIVERSE].make_empty */
extern T0* T1632c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_UNIVERSE].default_create */
extern T0* T1633c3(GE_context* ac);
/* [detachable] ET_DOTNET_ASSEMBLY.shortest_path */
extern T0* T596f103(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_DOTNET_ASSEMBLY.shortest_path */
extern T0* T596f103ac1(GE_context* ac, T0* a1, T0* a2, T6 a3);
/* Creation of agent #2 in feature [detachable] ET_DOTNET_ASSEMBLY.shortest_path */
extern T0* T596f103ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_DOTNET_ASSEMBLY.shortest_path */
extern T0* T596f103ac3(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_ASSEMBLY.insert_in_hash_table */
extern void T596f168(GE_context* ac, T0* C, T0* a1, T6 a2, T0* a3);
/* [detachable] ET_DOTNET_ASSEMBLY.universe_actions */
extern T0* T596f104(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.shortest_path */
extern T0* T211f100(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_DOTNET_ASSEMBLY.shortest_path */
extern T0* T211f100ac1(GE_context* ac, T0* a1, T0* a2, T6 a3);
/* Creation of agent #2 in feature [detachable] ET_ECF_DOTNET_ASSEMBLY.shortest_path */
extern T0* T211f100ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_ECF_DOTNET_ASSEMBLY.shortest_path */
extern T0* T211f100ac3(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.insert_in_hash_table */
extern void T211f168(GE_context* ac, T0* C, T0* a1, T6 a2, T0* a3);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.universe_actions */
extern T0* T211f103(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.shortest_path */
extern T0* T187f112(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_LIBRARY.shortest_path */
extern T0* T187f112ac1(GE_context* ac, T0* a1, T0* a2, T6 a3);
/* Creation of agent #2 in feature [detachable] ET_ECF_LIBRARY.shortest_path */
extern T0* T187f112ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_ECF_LIBRARY.shortest_path */
extern T0* T187f112ac3(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_LIBRARY.insert_in_hash_table */
extern void T187f199(GE_context* ac, T0* C, T0* a1, T6 a2, T0* a3);
/* [detachable] ET_ECF_LIBRARY.universe_actions */
extern T0* T187f116(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.shortest_path */
extern T0* T40f126(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature detachable ET_SYSTEM.shortest_path */
extern T0* T40f126ac1(GE_context* ac, T0* a1, T0* a2, T6 a3);
/* Creation of agent #2 in feature detachable ET_SYSTEM.shortest_path */
extern T0* T40f126ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature detachable ET_SYSTEM.shortest_path */
extern T0* T40f126ac3(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_SYSTEM.insert_in_hash_table */
extern void T40f207(GE_context* ac, T0* C, T0* a1, T6 a2, T0* a3);
/* detachable ET_SYSTEM.universe_actions */
extern T0* T40f128(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.shortest_path */
extern T0* T58f150(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.shortest_path */
extern T0* T58f150ac1(GE_context* ac, T0* a1, T0* a2, T6 a3);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.shortest_path */
extern T0* T58f150ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_ECF_SYSTEM.shortest_path */
extern T0* T58f150ac3(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.insert_in_hash_table */
extern void T58f243(GE_context* ac, T0* C, T0* a1, T6 a2, T0* a3);
/* [detachable] ET_ECF_SYSTEM.universe_actions */
extern T0* T58f152(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.lower_name */
extern T0* T211f95(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.lower_name */
extern T0* T596f96(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.lower_name */
extern T0* T213f23(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.lower_name */
extern T0* T207f21(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.adapted_universe */
extern T0* T852f94(GE_context* ac, T0* C, T0* a1);
extern T1 T852f94ot1(T0* a1, T0** a2);
extern T1 T852f94ot2(T0* a1, T0** a2);
/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.adapted_dotnet_assembly */
extern T0* T210f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ADAPTED_LIBRARIES.adapted_library */
extern T0* T188f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.adapted_universe */
extern T0* T596f94(GE_context* ac, T0* C, T0* a1);
extern T1 T596f94ot1(T0* a1, T0** a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY].has */
extern T1 T1398f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.adapted_universe */
extern T0* T211f94(GE_context* ac, T0* C, T0* a1);
extern T1 T211f94ot1(T0* a1, T0** a2);
/* [detachable] ET_ECF_LIBRARY.adapted_universe */
extern T0* T187f105(GE_context* ac, T0* C, T0* a1);
extern T1 T187f105ot1(T0* a1, T0** a2);
extern T1 T187f105ot2(T0* a1, T0** a2);
/* detachable ET_SYSTEM.adapted_universe */
extern T0* T40f124(GE_context* ac, T0* C, T0* a1);
extern T1 T40f124ot1(T0* a1, T0** a2);
extern T1 T40f124ot2(T0* a1, T0** a2);
/* [detachable] ET_ECF_SYSTEM.adapted_universe */
extern T0* T58f154(GE_context* ac, T0* C, T0* a1);
extern T1 T58f154ot1(T0* a1, T0** a2);
extern T1 T58f154ot2(T0* a1, T0** a2);
/* detachable ET_SYSTEM.lower_name */
extern T0* T40f123(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.relative_lower_name */
extern T0* T58f153(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_ECF_SYSTEM.string_ */
extern T0* T58f156(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.lower_name */
extern T0* T58f155(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.lower_name */
extern T0* T887f14(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.current_system */
extern T0* T887f11(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.full_lower_name */
extern T0* T410f13(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_UNKNOWN_GROUP.relative_lower_name */
extern T0* T410f15(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_UNKNOWN_GROUP.string_ */
extern T0* T410f17(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.lower_name */
extern T0* T410f16(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.universe */
extern T0* T410f23(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_system */
extern T0* T73f210(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.make */
extern T0* T40c138(GE_context* ac, T0* a1);
/* detachable ET_SYSTEM.set_kernel_types */
extern void T40f143(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_wide_character_type */
extern void T40f185(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.set_marked */
extern void T570f51(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_CLASS.set_marked */
extern void T74f284(GE_context* ac, T0* C, T1 a1);
/* detachable ET_SYSTEM.master_class */
extern T0* T40f121(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].force_last_new */
extern void T225f50(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].key_storage_put */
extern void T225f56(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS_NAME].force */
extern void T612f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLASS_NAME].fill_with */
extern void T613f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLASS_NAME].extend */
extern void T613f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].item_storage_put */
extern void T225f55(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MASTER_CLASS].force */
extern void T610f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_MASTER_CLASS].fill_with */
extern void T611f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_MASTER_CLASS].extend */
extern void T611f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].slots_put */
extern void T225f54(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].clashes_put */
extern void T225f53(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].slots_item */
extern T6 T225f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].hash_position */
extern T6 T225f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].resize */
extern void T225f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].clashes_resize */
extern void T225f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].special_integer_ */
extern T0* T225f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].key_storage_resize */
extern void T225f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS_NAME].aliased_resized_area */
extern T0* T612f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_CLASS_NAME].aliased_resized_area */
extern T0* T613f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].item_storage_resize */
extern void T225f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MASTER_CLASS].aliased_resized_area */
extern T0* T610f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_MASTER_CLASS].aliased_resized_area */
extern T0* T611f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].key_storage_item */
extern T0* T225f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].clashes_item */
extern T6 T225f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].slots_resize */
extern void T225f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].new_modulus */
extern T6 T225f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].new_capacity */
extern T6 T225f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].unset_found_item */
extern void T225f46(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.make */
extern T0* T570c50(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_MASTER_CLASS.tokens */
extern T0* T570f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].found_item */
extern T0* T225f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].item_storage_item */
extern T0* T225f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].found */
extern T1 T225f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].search */
extern void T225f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].search_position */
extern void T225f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_NAME_TESTER.test */
extern T1 T565f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_IDENTIFIER.same_class_name */
extern T1 T184f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.wide_character_class_name */
extern T0* T73f28(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.tokens */
extern T0* T40f130(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_typed_pointer_type */
extern void T40f184(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.put_first */
extern void T240f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACTUAL_PARAMETER_ITEM].force */
extern void T624f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ACTUAL_PARAMETER_ITEM].fill_with */
extern void T621f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ACTUAL_PARAMETER_ITEM].extend */
extern void T621f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.fixed_array */
extern T0* T240f11(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACTUAL_PARAMETER_ITEM].default_create */
extern T0* T624c3(GE_context* ac);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.make_with_capacity */
extern T0* T240c26(GE_context* ac, T6 a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.make_with_capacity */
extern void T240f26p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACTUAL_PARAMETER_ITEM].make */
extern T0* T624f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ACTUAL_PARAMETER_ITEM].make_empty */
extern T0* T621c5(GE_context* ac, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.right_bracket_symbol */
extern T0* T73f191(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_right_bracket */
extern T0* T341c92(GE_context* ac);
/* [detachable] ET_SYMBOL.make_leaf */
extern void T341f95(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.left_bracket_symbol */
extern T0* T73f190(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.make */
extern T0* T622c130(GE_context* ac);
/* [detachable] ET_BRACKET_SYMBOL.make_leaf */
extern void T622f131(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.tokens */
extern T0* T240f9(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.typed_pointer_class_name */
extern T0* T73f114(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_type_type */
extern void T40f183(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.type_class_name */
extern T0* T73f113(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_system_string_type */
extern void T40f182(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.system_string_class_name */
extern T0* T73f174(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_system_object_type */
extern void T40f181(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_LIST.put_first */
extern void T568f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_ITEM].force */
extern void T1387f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_PARENT_ITEM].fill_with */
extern void T1386f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_PARENT_ITEM].extend */
extern void T1386f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_LIST.fixed_array */
extern T0* T568f5(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_ITEM].default_create */
extern T0* T1387c2(GE_context* ac);
/* [detachable] ET_PARENT_LIST.make_with_capacity */
extern T0* T568c11(GE_context* ac, T6 a1);
/* [detachable] ET_PARENT_LIST.make_with_capacity */
extern void T568f11p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_ITEM].make */
extern T0* T1387f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_PARENT_ITEM].make_empty */
extern T0* T1386c4(GE_context* ac, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.inherit_keyword */
extern T0* T73f250(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_inherit */
extern T0* T413c54(GE_context* ac);
/* [detachable] ET_PARENT_LIST.tokens */
extern T0* T568f9(GE_context* ac, T0* C);
/* [detachable] ET_PARENT.make */
extern T0* T567c11(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_TOKEN_CONSTANTS.end_keyword */
extern T0* T73f89(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_end */
extern T0* T413c47(GE_context* ac);
/* [detachable] ET_PARENT.tokens */
extern T0* T567f9(GE_context* ac, T0* C);
/* [detachable] ET_PARENT.has_feature_adaptation */
extern T1 T567f8(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.system_object_class_name */
extern T0* T73f117(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_string_type */
extern void T40f180(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.string_class_name */
extern T0* T73f173(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_string_32_type */
extern void T40f179(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_CONVERT_FEATURE.make */
extern T0* T569c6(GE_context* ac, T0* a1);
/* [detachable] ET_BRACED_TYPE_LIST.make */
extern T0* T1388c8(GE_context* ac);
/* [detachable] ET_BRACED_TYPE_LIST.make */
extern void T1388f8p1(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_ITEM].make */
extern T0* T1729f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_TYPE_ITEM].make_empty */
extern T0* T1728c4(GE_context* ac, T6 a1);
/* [detachable] ET_BRACED_TYPE_LIST.fixed_array */
extern T0* T1388f6(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_ITEM].default_create */
extern T0* T1729c2(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.right_brace_symbol */
extern T0* T73f694(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_right_brace */
extern T0* T341c110(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.left_brace_symbol */
extern T0* T73f693(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_left_brace */
extern T0* T341c109(GE_context* ac);
/* [detachable] ET_BRACED_TYPE_LIST.tokens */
extern T0* T1388f5(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.string_32_class_name */
extern T0* T73f172(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_string_8_type */
extern void T40f178(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.string_8_class_name */
extern T0* T73f171(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_special_type */
extern void T40f177(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.special_class_name */
extern T0* T73f112(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_routine_type */
extern void T40f176(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.routine_class_name */
extern T0* T73f111(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_real_64_type */
extern void T40f175(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.real_64_class_name */
extern T0* T73f105(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_real_32_type */
extern void T40f174(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.real_32_class_name */
extern T0* T73f104(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_real_type */
extern void T40f173(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.real_class_name */
extern T0* T73f31(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_procedure_type */
extern void T40f172(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.procedure_class_name */
extern T0* T73f110(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_predicate_type */
extern void T40f171(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.predicate_class_name */
extern T0* T73f109(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_pointer_type */
extern void T40f170(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.pointer_class_name */
extern T0* T73f106(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_natural_64_type */
extern void T40f169(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_64_class_name */
extern T0* T73f103(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_natural_32_type */
extern void T40f168(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_32_class_name */
extern T0* T73f102(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_natural_16_type */
extern void T40f167(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_16_class_name */
extern T0* T73f101(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_natural_8_type */
extern void T40f166(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_8_class_name */
extern T0* T73f100(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_natural_type */
extern void T40f165(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_class_name */
extern T0* T73f30(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_iterable_type */
extern void T40f164(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.iterable_class_name */
extern T0* T73f119(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_ise_exception_manager_type */
extern void T40f163(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ise_exception_manager_class_name */
extern T0* T73f170(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_integer_64_type */
extern void T40f162(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_64_class_name */
extern T0* T73f99(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_integer_32_type */
extern void T40f161(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_32_class_name */
extern T0* T73f98(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_integer_16_type */
extern void T40f160(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_16_class_name */
extern T0* T73f97(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_integer_8_type */
extern void T40f159(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_8_class_name */
extern T0* T73f96(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_integer_type */
extern void T40f158(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_class_name */
extern T0* T73f29(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_immutable_string_32_type */
extern void T40f157(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.immutable_string_32_class_name */
extern T0* T73f169(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_immutable_string_8_type */
extern void T40f156(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.immutable_string_8_class_name */
extern T0* T73f168(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_function_type */
extern void T40f155(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.function_class_name */
extern T0* T73f108(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_exception_manager_type */
extern void T40f154(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.exception_manager_class_name */
extern T0* T73f167(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_exception_type */
extern void T40f153(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.exception_class_name */
extern T0* T73f166(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_double_type */
extern void T40f152(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.double_class_name */
extern T0* T73f32(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_character_32_type */
extern void T40f151(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_32_class_name */
extern T0* T73f95(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_character_8_type */
extern void T40f150(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_8_class_name */
extern T0* T73f94(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_character_type */
extern void T40f149(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_class_name */
extern T0* T73f27(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_boolean_type */
extern void T40f148(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.boolean_class_name */
extern T0* T73f93(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_array_type */
extern void T40f147(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.array_class_name */
extern T0* T73f115(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_tuple_type */
extern void T40f146(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.make */
extern T0* T577c17(GE_context* ac, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.tuple_class_name */
extern T0* T73f107(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_none_type */
extern void T40f145(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.add_first_local_class */
extern void T570f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.add_first_local_non_override_class */
extern void T570f79(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.update_intrinsic_class */
extern void T570f54(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.set_intrinsic_class */
extern void T570f55(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_MASTER_CLASS.set_intrinsic_class */
extern T0* T570f55ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_MASTER_CLASS.set_intrinsic_class */
extern T0* T570f55ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_MASTER_CLASS.set_intrinsic_class */
extern T0* T570f55ac3(GE_context* ac, T0* a1);
extern T1 T570f55ot1(T0* a1, T0** a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_MASTER_CLASS]].set_rout_disp_final */
extern void T863f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_MASTER_CLASS.is_marked */
extern T1 T570f21(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.mark_overridden */
extern void T570f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.add_last_overriding_class */
extern void T570f64(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].force_last */
extern void T1389f20(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].resize */
extern void T1389f22(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].new_capacity */
extern T6 T1389f10(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].extendible */
extern T1 T1389f9(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].make */
extern T0* T1389c14(GE_context* ac, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].set_internal_cursor */
extern void T1389f21(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].new_cursor */
extern T0* T1389f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_MASTER_CLASS].make */
extern T0* T1730c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MASTER_CLASS].make */
extern T0* T610f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_MASTER_CLASS].make_empty */
extern T0* T611c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MASTER_CLASS].default_create */
extern T0* T610c3(GE_context* ac);
/* [detachable] ET_MASTER_CLASS.is_override */
extern T1 T570f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_override */
extern T1 T74f168(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].do_all */
extern void T1389f16(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].item */
extern T0* T1389f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_MASTER_CLASS.unmark_overridden */
extern void T570f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.remove_overriding_class */
extern void T570f63(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].remove_first */
extern void T1389f19(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].remove */
extern void T1389f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_MASTER_CLASS].keep_head */
extern void T611f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_MASTER_CLASS].set_count */
extern void T611f10(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].move_left */
extern void T1389f26(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].move_cursors_left */
extern void T1389f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_MASTER_CLASS].set_position */
extern void T1730f6(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].remove_last */
extern void T1389f24(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].move_last_cursors_after */
extern void T1389f31(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_MASTER_CLASS].set_next_cursor */
extern void T1730f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_MASTER_CLASS].set_after */
extern void T1730f7(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].first */
extern T0* T1389f7(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].is_empty */
extern T1 T1389f12(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].delete */
extern void T1389f18(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].move_all_cursors_after */
extern void T1389f27(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_unknown_class */
extern T1 T74f169(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_class_name */
extern T0* T73f84(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].force_first */
extern void T462f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].put */
extern void T462f33(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].move_cursors_right */
extern void T462f36(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS].set_position */
extern void T1348f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].move_right */
extern void T462f35(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].put_last */
extern void T462f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].resize */
extern void T462f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].new_capacity */
extern T6 T462f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].extendible */
extern T1 T462f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].make */
extern T0* T462c15(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].set_internal_cursor */
extern void T462f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].new_cursor */
extern T0* T462f12(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS].make */
extern T0* T1348c9(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS].make */
extern T0* T1347f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CLASS].make_empty */
extern T0* T1346c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS].default_create */
extern T0* T1347c3(GE_context* ac);
/* [detachable] ET_MASTER_CLASS.add_first_local_override_class */
extern void T570f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_in_override_group */
extern T1 T74f171(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.is_override */
extern T1 T802f21(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.is_override */
extern T1 T887f9(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.is_override */
extern T1 T410f10(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.add_first_local_ignored_class */
extern void T570f77(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_group */
extern void T74f254(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NONE_GROUP.make */
extern T0* T887c30(GE_context* ac, T0* a1);
/* detachable ET_SYSTEM.register_class */
extern void T40f186(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_id */
extern void T74f263(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS.make */
extern void T74f260(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.make */
extern T0* T74c260(GE_context* ac, T0* a1);
/* [detachable] ET_CLASS_CODES.class_code */
extern T8 T414f1(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].found_item */
extern T8 T763f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].item_storage_item */
extern T8 T763f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].found */
extern T1 T763f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].search */
extern void T763f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].search_position */
extern void T763f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].hash_position */
extern T6 T763f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].key_storage_item */
extern T0* T763f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].clashes_item */
extern T6 T763f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].slots_item */
extern T6 T763f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS_CODES.codes_by_name */
extern T0* T414f2(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.iteration_cursor_class_name */
extern T0* T73f120(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.disposable_class_name */
extern T0* T73f118(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.any_class_name */
extern T0* T73f116(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].force_last */
extern void T763f37(GE_context* ac, T0* C, T8 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].key_storage_put */
extern void T763f45(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].slots_put */
extern void T763f44(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].clashes_put */
extern void T763f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].resize */
extern void T763f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].clashes_resize */
extern void T763f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].special_integer_ */
extern T0* T763f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].key_storage_resize */
extern void T763f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].item_storage_resize */
extern void T763f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_8].aliased_resized_area */
extern T0* T1433f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] NATURAL_8].aliased_resized_area */
extern T0* T487f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].slots_resize */
extern void T763f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].new_modulus */
extern T6 T763f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].new_capacity */
extern T6 T763f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].item_storage_put */
extern void T763f41(GE_context* ac, T0* C, T8 a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_8].force */
extern void T1433f4(GE_context* ac, T0* C, T0* a1, T8 a2, T6 a3);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].unset_found_item */
extern void T763f40(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_CODES.tokens */
extern T0* T414f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].set_key_equality_tester */
extern void T763f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_CODES.class_name_tester */
extern T0* T414f4(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_NAME_TESTER.default_create */
extern T0* T565c2(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].make_map */
extern T0* T763c34(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].make_with_equality_testers */
extern void T763f38(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].make_sparse_container */
extern void T763f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].set_internal_cursor */
extern void T763f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].new_cursor */
extern T0* T763f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] NATURAL_8, [attached] ET_CLASS_NAME].make */
extern T0* T1434c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].make_slots */
extern void T763f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].make_clashes */
extern void T763f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].make_key_storage */
extern void T763f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS_NAME].make */
extern T0* T612f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CLASS_NAME].make_empty */
extern T0* T613c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLASS_NAME].default_create */
extern T0* T612c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] NATURAL_8, [attached] ET_CLASS_NAME].make_item_storage */
extern void T763f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_8].make */
extern T0* T1433f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] NATURAL_8].default_create */
extern T0* T1433c3(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_group */
extern T0* T73f2(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.make */
extern T0* T410c31(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.class_keyword */
extern T0* T73f88(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_class */
extern T0* T413c46(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.empty_procedures */
extern T0* T73f87(GE_context* ac, T0* C);
/* [detachable] ET_PROCEDURE_LIST.make_with_capacity */
extern T0* T258c8(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PROCEDURE].make */
extern T0* T914f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_PROCEDURE].make_empty */
extern T0* T913c5(GE_context* ac, T6 a1);
/* [detachable] ET_PROCEDURE_LIST.fixed_array */
extern T0* T258f6(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PROCEDURE].default_create */
extern T0* T914c3(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.empty_queries */
extern T0* T73f86(GE_context* ac, T0* C);
/* [detachable] ET_QUERY_LIST.make_with_capacity */
extern T0* T257c8(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_QUERY].make */
extern T0* T912f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_QUERY].make_empty */
extern T0* T910c5(GE_context* ac, T6 a1);
/* [detachable] ET_QUERY_LIST.fixed_array */
extern T0* T257f7(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_QUERY].default_create */
extern T0* T912c3(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.empty_ancestors */
extern T0* T73f85(GE_context* ac, T0* C);
/* [detachable] ET_BASE_TYPE_LIST.make_with_capacity */
extern T0* T412c7(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE].make */
extern T0* T924f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE].make_empty */
extern T0* T923c5(GE_context* ac, T6 a1);
/* [detachable] ET_BASE_TYPE_LIST.fixed_array */
extern T0* T412f6(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE].default_create */
extern T0* T924c3(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.none_class_name */
extern T0* T73f164(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_any_type */
extern void T40f144(GE_context* ac, T0* C);
/* [detachable] ET_CLIENT_LIST.put_first */
extern void T402f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_ITEM].force */
extern void T759f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLIENT_ITEM].fill_with */
extern void T758f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLIENT_ITEM].extend */
extern void T758f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLIENT_LIST.fixed_array */
extern T0* T402f3(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_ITEM].default_create */
extern T0* T759c3(GE_context* ac);
/* [detachable] ET_CLIENT.make */
extern T0* T403c6(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_CLIENT_LIST.make_with_capacity */
extern T0* T402c13(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_ITEM].make */
extern T0* T759f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CLIENT_ITEM].make_empty */
extern T0* T758c5(GE_context* ac, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.detachable_separate_type_mark */
extern T0* T73f163(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_MARK_SEPARATE_KEYWORD.make */
extern T0* T571c23(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.separate_keyword */
extern T0* T73f220(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_separate */
extern T0* T413c51(GE_context* ac);
/* detachable ET_SYSTEM.make_adapted */
extern void T40f142(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_SYSTEM.default_read_only_value */
extern T1 T40f135(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_unknown_kernel_types */
extern void T40f141(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_convert_feature */
extern T0* T73f161(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_generic_class_type */
extern T0* T73f160(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.make */
extern T0* T240c31(GE_context* ac);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.make */
extern void T240f31p1(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.empty_actual_parameters */
extern T0* T73f159(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_tuple_type */
extern T0* T73f158(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.empty_clients */
extern T0* T73f157(GE_context* ac, T0* C);
/* [detachable] ET_CLIENT_LIST.make */
extern T0* T402c15(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_parents */
extern T0* T73f156(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_parent */
extern T0* T73f155(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_class_type */
extern T0* T73f154(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_system_name */
extern void T40f140(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.initialize */
extern void T40f139(GE_context* ac, T0* C);
/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.make_empty */
extern T0* T210c6(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].make */
extern T0* T594c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].set_internal_cursor */
extern void T594f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].new_cursor */
extern T0* T594f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].make */
extern T0* T1397c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].make */
extern T0* T1396f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].make_empty */
extern T0* T1395c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].default_create */
extern T0* T1396c3(GE_context* ac);
/* [detachable] ET_CLUSTERS.make_empty */
extern T0* T564c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLUSTER].make */
extern T0* T1384c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLUSTER].set_internal_cursor */
extern void T1384f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLUSTER].new_cursor */
extern T0* T1384f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLUSTER].make */
extern T0* T1727c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLUSTER].make */
extern T0* T1726f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CLUSTER].make_empty */
extern T0* T1725c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLUSTER].default_create */
extern T0* T1726c3(GE_context* ac);
/* [detachable] ET_ADAPTED_LIBRARIES.make_empty */
extern T0* T188c6(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY].make */
extern T0* T578c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY].set_internal_cursor */
extern void T578f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY].new_cursor */
extern T0* T578f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_LIBRARY].make */
extern T0* T1392c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_LIBRARY].make */
extern T0* T1391f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_LIBRARY].make_empty */
extern T0* T1390c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_LIBRARY].default_create */
extern T0* T1391c3(GE_context* ac);
/* detachable ET_SYSTEM.initialize */
extern void T40f139p1(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].set_key_equality_tester */
extern void T225f39(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.class_name_tester */
extern T0* T40f129(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].make_map */
extern T0* T225c38(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].make_with_equality_testers */
extern void T225f40(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].make_sparse_container */
extern void T225f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].set_internal_cursor */
extern void T225f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].new_cursor */
extern T0* T225f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].make */
extern T0* T609c10(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].make_slots */
extern void T225f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].make_clashes */
extern void T225f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].make_key_storage */
extern void T225f43(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].make_item_storage */
extern void T225f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNKNOWN_GROUP.tokens */
extern T0* T410f6(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.current_system */
extern T0* T410f12(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.full_lower_name */
extern T0* T802f48(GE_context* ac, T0* C, T2 a1);
/* detachable ET_ECF_CLUSTER.relative_lower_name */
extern T0* T802f58(GE_context* ac, T0* C, T0* a1, T2 a2);
/* detachable ET_ECF_CLUSTER.relative_lower_name */
extern T0* T802f58p1(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_ECF_LIBRARY.relative_lower_name */
extern T0* T187f113(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_ECF_LIBRARY.string_ */
extern T0* T187f118(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.lower_name */
extern T0* T187f117(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.string_ */
extern T0* T802f54(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.lower_name */
extern T0* T802f60(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.current_system */
extern T0* T802f45(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gcdep_error */
extern T1 T53f112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gcdep_error */
extern void T36f409(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_cluster_error */
extern void T36f364(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_gcdep_error */
extern T1 T36f112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gcpro_error */
extern void T53f408(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLUSTER_ERROR.make_gcpro */
extern T0* T1940c24(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gcpro_error */
extern T1 T53f111(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gcpro_error */
extern void T36f408(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_gcpro_error */
extern T1 T36f111(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.has_group */
extern T1 T1758f8(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.has_group */
extern T0* T1758f8ac1(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* Creation of agent #2 in feature detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.has_group */
extern T0* T1758f8ac2(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #3 in feature detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.has_group */
extern T0* T1758f8ac3(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #4 in feature detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.has_group */
extern T0* T1758f8ac4(GE_context* ac, T0* a1, T0* a2);
extern T1 T1758f8ot1(T0* a1, T0** a2);
/* [detachable] PREDICATE [[attached] TUPLE [[attached] STRING_8]].set_rout_disp_final */
extern void T2063f8(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] PREDICATE [[attached] TUPLE [[attached] LX_WILDCARD]].set_rout_disp_final */
extern void T2074f8(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.matches_pathname */
extern T1 T1758f6(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] LX_DFA_WILDCARD.recognizes */
extern T1 T2068f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_DFA_WILDCARD.longest_end_position */
extern T6 T2068f16(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] LX_DFA_WILDCARD.set_subject */
extern void T2068f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_DFA_WILDCARD.is_compiled */
extern T1 T2068f17(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.full_name */
extern T0* T187f111(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_ECF_LIBRARY.relative_name */
extern T0* T187f115(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.name */
extern T0* T211f5(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.name */
extern T0* T596f5(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.name */
extern T0* T213f17(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.name */
extern T0* T207f15(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.universe */
extern T0* T802f4(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.universe */
extern T0* T887f1(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.parent */
extern T0* T802f31(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.parent */
extern T0* T887f18(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.parent */
extern T0* T410f19(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].there_exists */
extern T1 T147f20(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.full_name */
extern T0* T802f62(GE_context* ac, T0* C, T2 a1);
/* detachable ET_ECF_CLUSTER.relative_name */
extern T0* T802f66(GE_context* ac, T0* C, T0* a1, T2 a2);
/* detachable ET_ECF_CLUSTER.relative_name */
extern T0* T802f66p1(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_ECF_SYSTEM.relative_name */
extern T0* T58f160(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_NONE_GROUP.full_name */
extern T0* T887f17(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_NONE_GROUP.relative_name */
extern T0* T887f21(GE_context* ac, T0* C, T0* a1, T2 a2);
/* detachable ET_SYSTEM.relative_name */
extern T0* T40f134(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_UNKNOWN_GROUP.full_name */
extern T0* T410f18(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_UNKNOWN_GROUP.relative_name */
extern T0* T410f22(GE_context* ac, T0* C, T0* a1, T2 a2);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.string_ */
extern T0* T1758f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].there_exists */
extern T1 T2040f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].item_storage_item */
extern T0* T2040f25(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.group_pathname */
extern T0* T1758f5(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.replace_backslashes */
extern void T1758f18(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.absolute_pathname */
extern T0* T802f63(GE_context* ac, T0* C);
/* [detachable] KL_UNIX_FILE_SYSTEM.canonical_pathname */
extern T0* T126f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.pathname_to_string */
extern T0* T126f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_PATHNAME.item */
extern T0* T483f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [detachable STRING_8].item */
extern T0* T1355f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_PATHNAME.is_parent */
extern T1 T483f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_PATHNAME.is_current */
extern T1 T483f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_PATHNAME.set_canonical */
extern void T483f21(GE_context* ac, T0* C);
/* [detachable] KL_PATHNAME.append_current */
extern void T483f23(GE_context* ac, T0* C);
/* [detachable] KL_PATHNAME.append_name */
extern void T483f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ARRAY [detachable STRING_8].force */
extern void T1355f11(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable STRING_8].move_data */
extern void T497f10(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable STRING_8].overlapping_move */
extern void T497f12(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable STRING_8].force */
extern void T497f13(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable STRING_8].non_overlapping_move */
extern void T497f11(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] ARRAY [detachable STRING_8].capacity */
extern T6 T1355f7(GE_context* ac, T0* C);
/* [detachable] ARRAY [detachable STRING_8].set_area */
extern void T1355f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ARRAY [detachable STRING_8].make_empty_area */
extern void T1355f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [detachable STRING_8].additional_space */
extern T6 T1355f6(GE_context* ac, T0* C);
/* [detachable] ARRAY [detachable STRING_8].empty_area */
extern T1 T1355f5(GE_context* ac, T0* C);
/* [detachable] ARRAY [detachable STRING_8].put */
extern void T1355f10(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_UNIX_FILE_SYSTEM.string_to_pathname */
extern T0* T126f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_PATHNAME.append_parent */
extern void T483f24(GE_context* ac, T0* C);
/* [detachable] KL_PATHNAME.set_relative */
extern void T483f22(GE_context* ac, T0* C, T1 a1);
/* [detachable] KL_PATHNAME.make */
extern T0* T483c20(GE_context* ac);
/* [detachable] ARRAY [detachable STRING_8].make_filled */
extern T0* T1355c9(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] ARRAY [detachable STRING_8].make_filled_area */
extern void T1355f12(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable STRING_8].make_filled */
extern T0* T497c8(GE_context* ac, T0* a1, T6 a2);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.canonical_pathname */
extern T0* T125f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.pathname_to_string */
extern T0* T125f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.string_to_pathname */
extern T0* T125f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_PATHNAME.set_drive */
extern void T483f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_PATHNAME.set_sharename */
extern void T483f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_PATHNAME.set_hostname */
extern void T483f25(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.file_system */
extern T0* T802f49(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.unix_file_system */
extern T0* T802f51(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.windows_file_system */
extern T0* T802f50(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.operating_system */
extern T0* T802f59(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.full_pathname */
extern T0* T802f38(GE_context* ac, T0* C);
/* [detachable] KL_UNIX_FILE_SYSTEM.dirname */
extern T0* T126f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.dirname */
extern T0* T125f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.filename */
extern T0* T187f2(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_CONFIG.filename */
extern T0* T185f2(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.filename */
extern T0* T58f1(GE_context* ac, T0* C);
/* detachable ET_ECF_SETTINGS.value */
extern T0* T57f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].found_item */
extern T0* T216f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].found */
extern T1 T216f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].search */
extern void T216f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.is_relative_pathname */
extern T1 T126f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.pathname_from_file_system */
extern T0* T126f4(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_UNIX_FILE_SYSTEM.any_ */
extern T0* T126f18(GE_context* ac, T0* C);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.pathname_from_file_system */
extern T0* T125f4(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.any_ */
extern T0* T125f20(GE_context* ac, T0* C);
/* detachable ET_ECF_VARIABLES.interpreted_string */
extern T0* T62f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VARIABLES.expanded_string */
extern T0* T62f9(GE_context* ac, T0* C, T0* a1, T1 a2);
/* detachable ET_ECF_VARIABLES.value */
extern T0* T62f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VARIABLES.string_ */
extern T0* T62f5(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.has */
extern T1 T775f70(GE_context* ac, T0* C, T2 a1);
/* [detachable] UC_STRING.has */
extern T1 T429f70(GE_context* ac, T0* C, T2 a1);
/* detachable ET_ECF_CLUSTER.execution_environment */
extern T0* T802f68(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.absolute_pathname */
extern T0* T887f19(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.file_system */
extern T0* T887f25(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.unix_file_system */
extern T0* T887f29(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.windows_file_system */
extern T0* T887f28(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.operating_system */
extern T0* T887f27(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.full_pathname */
extern T0* T887f24(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.pathname */
extern T0* T887f26(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.execution_environment */
extern T0* T887f23(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.absolute_pathname */
extern T0* T410f20(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.file_system */
extern T0* T410f26(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.unix_file_system */
extern T0* T410f30(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.windows_file_system */
extern T0* T410f29(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.operating_system */
extern T0* T410f28(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.full_pathname */
extern T0* T410f25(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.pathname */
extern T0* T410f27(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.execution_environment */
extern T0* T410f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS].item */
extern T0* T1620f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].cursor_item */
extern T0* T948f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS].after */
extern T1 T1620f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS].start */
extern void T1620f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].cursor_start */
extern void T948f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].cursor_off */
extern T1 T948f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS].off */
extern T1 T1620f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].is_empty */
extern T1 T948f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].new_cursor */
extern T0* T948f19(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS].make */
extern T0* T1620c9(GE_context* ac, T0* a1);
/* [detachable] ET_CLASS.first_non_override_overridden_class */
extern T0* T74f196(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].first */
extern T0* T462f11(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].is_empty */
extern T1 T462f10(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.master_class_in_universe */
extern T0* T74f202(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.master_class */
extern T0* T187f108(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.master_class */
extern T0* T58f148(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PROVIDER_CHECKER.scm_mapped_cluster */
extern T0* T457f5(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.scm_write_mapping_recursive */
extern T0* T802f57(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.cluster */
extern T0* T802f46(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.cluster */
extern T0* T887f5(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.cluster */
extern T0* T410f7(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_compilation_status */
extern void T53f208(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_compilation_status */
extern void T36f208(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_PROVIDER_CHECKER.set_fatal_error */
extern void T457f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PROVIDER_CHECKER.make */
extern T0* T457c6(GE_context* ac, T0* a1);
/* [detachable] ET_PROVIDER_CHECKER.tokens */
extern T0* T457f3(GE_context* ac, T0* C);
/* [detachable] ET_PROVIDER_CHECKER.error_handler */
extern T0* T457f4(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.process_class */
extern void T459f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.internal_process_class */
extern void T459f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.flatten_features */
extern void T459f38(GE_context* ac, T0* C);
extern T1 T459f38ot1(T0* a1, T0** a2);
extern T1 T459f38ot2(T0* a1, T0** a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].wipe_out */
extern void T1310f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].slots_wipe_out */
extern void T1310f62(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].clashes_wipe_out */
extern void T1310f61(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].key_storage_wipe_out */
extern void T1310f60(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_FEATURE_NAME].keep_head */
extern void T1666f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_FEATURE_NAME].set_count */
extern void T1666f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].item_storage_wipe_out */
extern void T1310f59(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_FLATTENED_FEATURE].keep_head */
extern void T1663f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_FLATTENED_FEATURE].set_count */
extern void T1663f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].unset_found_item */
extern void T1310f58(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].move_all_cursors_after */
extern void T1310f57(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].set_next_cursor */
extern void T1662f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].set_after */
extern void T1662f10(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.check_kernel_features_validity */
extern void T459f49(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_special_item_seed */
extern void T40f198(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_special_item_seed */
extern void T58f268(GE_context* ac, T0* C, T6 a1);
/* detachable ET_SYSTEM.set_iteration_cursor_forth_seed */
extern void T40f197(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_iteration_cursor_forth_seed */
extern void T58f267(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.forth_feature_name */
extern T0* T73f248(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_iteration_cursor_after_seed */
extern void T40f196(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_iteration_cursor_after_seed */
extern void T58f266(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.after_feature_name */
extern T0* T73f247(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_iteration_cursor_item_seed */
extern void T40f195(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_iteration_cursor_item_seed */
extern void T58f265(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS.is_iteration_cursor_class */
extern T1 T74f167(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_iterable_new_cursor_seed */
extern void T40f194(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_iterable_new_cursor_seed */
extern void T58f264(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.new_cursor_feature_name */
extern T0* T73f246(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_iterable_class */
extern T1 T74f166(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_dispose_seed */
extern void T40f193(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_dispose_seed */
extern void T58f263(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.dispose_feature_name */
extern T0* T73f245(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_disposable_class */
extern T1 T74f165(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_function_item_seed */
extern void T40f192(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_function_item_seed */
extern void T58f262(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.item_feature_name */
extern T0* T73f6(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_routine_call_seed */
extern void T40f191(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_routine_call_seed */
extern void T58f261(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.call_feature_name */
extern T0* T73f244(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_is_equal_seed */
extern void T40f190(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_is_equal_seed */
extern void T58f260(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.is_equal_feature_name */
extern T0* T73f243(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvkfe5a_error */
extern void T53f198(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_gvkfe5a */
extern T0* T785c854(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvkfe5_error */
extern T1 T53f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvkfe5a_error */
extern void T36f198(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_gvkfe5_error */
extern T1 T36f21(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.set_twin_seed */
extern void T40f189(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_twin_seed */
extern void T58f259(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_PROCEDURE.is_query */
extern T1 T1516f68(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_query */
extern T1 T1515f63(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_query */
extern T1 T469f61(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_query */
extern T1 T253f47(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_query */
extern T1 T1513f64(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_query */
extern T1 T1512f68(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_query */
extern T1 T1511f66(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_query */
extern T1 T371f46(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_query */
extern T1 T369f45(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_query */
extern T1 T359f45(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_query */
extern T1 T360f44(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_query */
extern T1 T476f59(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.twin_feature_name */
extern T0* T73f242(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.set_copy_seed */
extern void T40f188(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_copy_seed */
extern void T58f258(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.copy_feature_name */
extern T0* T73f241(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvkfe1a_error */
extern void T53f196(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_gvkfe1a */
extern T0* T785c852(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvkfe1_error */
extern T1 T53f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvkfe1a_error */
extern void T36f196(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_gvkfe1_error */
extern T1 T36f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvkfe4a_error */
extern void T53f200(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_gvkfe4a */
extern T0* T785c856(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvkfe4_error */
extern T1 T53f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvkfe4a_error */
extern void T36f200(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_gvkfe4_error */
extern T1 T36f23(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.set_default_create_seed */
extern void T40f187(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_default_create_seed */
extern void T58f257(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_FLATTENER.current_system */
extern T0* T459f27(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].found_item */
extern T0* T1310f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].item_storage_item */
extern T0* T1310f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].found */
extern T1 T1310f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].search */
extern void T1310f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].search_position */
extern void T1310f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].hash_position */
extern T6 T1310f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_NAME_TESTER.test */
extern T1 T1311f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].key_storage_item */
extern T0* T1310f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].clashes_item */
extern T6 T1310f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].slots_item */
extern T6 T1310f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.default_create_feature_name */
extern T0* T73f3(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.tokens */
extern T0* T459f21(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_any_class */
extern T1 T74f118(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.check_convert_validity */
extern void T459f48(GE_context* ac, T0* C);
/* detachable ET_CONVERT_FEATURE_LIST.convert_feature */
extern T0* T1096f6(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONVERT_PROCEDURE.convert_feature */
extern T0* T1509f6(GE_context* ac, T0* C);
/* detachable ET_CONVERT_FUNCTION.convert_feature */
extern T0* T1508f5(GE_context* ac, T0* C);
/* detachable ET_CONVERT_FEATURE_LIST.item */
extern T0* T1096f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_FLATTENER.check_creators_validity */
extern void T459f47(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcp2a_error */
extern void T53f310(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vgcp2a */
extern T0* T785c943(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vgcp2_error */
extern T1 T53f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vgcp2a_error */
extern void T36f310(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vgcp2_error */
extern T1 T36f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcp2b_error */
extern void T53f309(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vgcp2b */
extern T0* T785c942(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vgcp2b_error */
extern void T36f309(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcc6a_error */
extern void T53f308(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vgcc6a */
extern T0* T785c941(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vgcc6_error */
extern T1 T53f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vgcc6a_error */
extern void T36f308(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vgcc6_error */
extern T1 T36f58(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.is_once */
extern T1 T1516f40(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_once */
extern T1 T1515f39(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_once */
extern T1 T469f43(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_once */
extern T1 T476f44(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcp3b_error */
extern void T53f307(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vgcp3b */
extern T0* T785c940(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vgcp3_error */
extern T1 T53f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vgcp3b_error */
extern void T36f307(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vgcp3_error */
extern T1 T36f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcp3a_error */
extern void T53f306(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vgcp3a */
extern T0* T785c939(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vgcp3a_error */
extern void T36f306(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CREATOR.feature_name */
extern T0* T1104f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CREATOR.item */
extern T0* T1104f13(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CREATOR_LIST.item */
extern T0* T1107f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcp1a_error */
extern void T53f305(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vgcp1a */
extern T0* T785c938(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_CREATOR.position */
extern T0* T1104f7(GE_context* ac, T0* C);
/* detachable ET_CREATOR.first */
extern T0* T1104f12(GE_context* ac, T0* C);
/* detachable ET_CLIENTS.position */
extern T0* T1065f13(GE_context* ac, T0* C);
/* [detachable] ET_CLIENT.position */
extern T0* T403f5(GE_context* ac, T0* C);
/* detachable ET_CLIENTS.first */
extern T0* T1065f16(GE_context* ac, T0* C);
/* detachable ET_CLIENTS.is_empty */
extern T1 T1065f15(GE_context* ac, T0* C);
/* detachable ET_CREATOR.clients_clause */
extern T0* T1104f10(GE_context* ac, T0* C);
extern T1 T1104f10ot1(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vgcp1_error */
extern T1 T53f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vgcp1a_error */
extern void T36f305(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vgcp1_error */
extern T1 T36f57(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CREATOR_LIST.first */
extern T0* T1107f5(GE_context* ac, T0* C);
/* detachable ET_CREATOR_LIST.is_empty */
extern T1 T1107f4(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.check_assigners_validity */
extern void T459f46(GE_context* ac, T0* C);
extern T1 T459f46ot1(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfac4a_error */
extern void T53f304(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6);
/* [detachable] ET_VALIDITY_ERROR.make_vfac4a */
extern T0* T785c937(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vfac4_error */
extern T1 T53f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vfac4a_error */
extern void T36f304(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6);
/* [detachable] ET_ERROR_HANDLER.reportable_vfac4_error */
extern T1 T36f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfac3a_error */
extern void T53f303(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vfac3a */
extern T0* T785c936(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vfac3_error */
extern T1 T53f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vfac3a_error */
extern void T36f303(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vfac3_error */
extern T1 T36f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.arguments_count */
extern T6 T253f35(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.arguments_count */
extern T6 T1513f35(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.arguments_count */
extern T6 T1512f36(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.arguments_count */
extern T6 T1511f35(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.arguments_count */
extern T6 T371f53(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.arguments_count */
extern T6 T369f53(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.arguments_count */
extern T6 T359f50(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.arguments_count */
extern T6 T360f50(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfac2a_error */
extern void T53f302(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vfac2a */
extern T0* T785c935(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vfac2_error */
extern T1 T53f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vfac2a_error */
extern void T36f302(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vfac2_error */
extern T1 T36f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfac1b_error */
extern void T53f301(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfac1b */
extern T0* T785c934(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vfac1_error */
extern T1 T53f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vfac1b_error */
extern void T36f301(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vfac1_error */
extern T1 T36f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfac1a_error */
extern void T53f300(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfac1a */
extern T0* T785c933(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vfac1a_error */
extern void T36f300(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EXTERNAL_FUNCTION.assigner */
extern T0* T253f14(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.assigner */
extern T0* T1513f12(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.assigner */
extern T0* T1512f12(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.assigner */
extern T0* T1511f12(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.assigner */
extern T0* T371f11(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.assigner */
extern T0* T369f10(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.assigner */
extern T0* T359f11(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.assigner */
extern T0* T360f9(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vcch2a_error */
extern void T53f296(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_VALIDITY_ERROR.set_ise_reported */
extern void T785f929(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_VALIDITY_ERROR.make_vcch2a */
extern T0* T785c928(GE_context* ac, T0* a1);
/* [detachable] ET_VALIDITY_ERROR.ast_position */
extern T0* T785f198(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vcch2_error */
extern T1 T53f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vcch2a_error */
extern void T36f296(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_vcch2_error */
extern T1 T36f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vcch1a_error */
extern void T53f295(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vcch1a */
extern T0* T785c927(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vcch1_error */
extern T1 T53f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vcch1a_error */
extern void T36f295(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vcch1_error */
extern T1 T36f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vcch1b_error */
extern void T53f294(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vcch1b */
extern T0* T785c926(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_INHERITED_FEATURE.name */
extern T0* T1345f22(GE_context* ac, T0* C);
/* detachable ET_PARENT_FEATURE.name */
extern T0* T1344f13(GE_context* ac, T0* C);
/* detachable ET_PARENT_FEATURE.has_undefine */
extern T1 T1344f19(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_vcch1b_error */
extern void T36f294(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].go_after */
extern void T1310f52(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].cursor_go_after */
extern void T1310f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].remove_traversing_cursor */
extern void T1310f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].cursor_off */
extern T1 T1310f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].off */
extern T1 T1662f4(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_has_deferred_features */
extern void T74f282(GE_context* ac, T0* C, T1 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].wipe_out */
extern void T1313f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].slots_wipe_out */
extern void T1313f52(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].clashes_wipe_out */
extern void T1313f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].key_storage_wipe_out */
extern void T1313f50(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_ALIAS_NAME].keep_head */
extern void T1669f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ALIAS_NAME].set_count */
extern void T1669f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].item_storage_wipe_out */
extern void T1313f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].unset_found_item */
extern void T1313f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].move_all_cursors_after */
extern void T1313f48(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].set_next_cursor */
extern void T1672f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].set_after */
extern void T1672f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].forth */
extern void T1310f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].cursor_forth */
extern void T1310f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].add_traversing_cursor */
extern void T1310f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].set_position */
extern void T1662f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_FLATTENER.check_signature_validity */
extern void T459f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.check_signature_validity */
extern void T1323f11(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].forth */
extern void T1695f10(GE_context* ac, T0* C);
/* detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE].cursor_forth */
extern void T1696f11(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE].add_traversing_cursor */
extern void T1696f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].set_next_cursor */
extern void T1695f12(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE].remove_traversing_cursor */
extern void T1696f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].set */
extern void T1695f11(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] ET_SIGNATURE_CHECKER.check_selected_signature_validity */
extern void T1323f15(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
extern T1 T1323f15ot1(T0* a1);
extern T1 T1323f15ot2(T0* a1);
/* detachable ET_FORMAL_ARGUMENT_LIST.is_empty */
extern T1 T255f8(GE_context* ac, T0* C);
/* [detachable] ET_SIGNATURE_CHECKER.check_tuple_actual_parameters_unfolded */
extern void T1323f17(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.same_syntactical_type_with_type_marks */
extern T1 T74f233(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_LIKE_N.same_syntactical_class_type_with_type_marks */
extern T1 T1474f35(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_syntactical_class_type_with_type_marks */
extern T1 T1501f48(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1501f48ot1(T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.same_syntactical_class_type_with_type_marks */
extern T1 T891f49(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T891f49ot1(T0* a1);
/* [detachable] ET_CLASS.same_syntactical_class_type_with_type_marks */
extern T1 T74f229(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_FORMAL_PARAMETER_LIST.same_syntactical_types */
extern T1 T603f23(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_syntactical_type */
extern T1 T1501f66(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_syntactical_type_with_type_marks */
extern T1 T1501f44(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1501f44ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T1474f38(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T1501f49(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1501f49ot1(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T891f50(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T891f50ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T74f234(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_LIKE_FEATURE.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T1524f67(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T254f94(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T242f63(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T242f63ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS_TYPE.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T72f107(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T1526f66(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T1525f69(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_LIKE_CURRENT.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T449f62(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER.same_syntactical_type_with_type_marks */
extern T1 T891f45(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T891f45ot1(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.same_syntactical_type */
extern T1 T891f68(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_LIKE_FEATURE.same_syntactical_class_type_with_type_marks */
extern T1 T1524f64(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.same_syntactical_class_type_with_type_marks */
extern T1 T254f91(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_syntactical_class_type_with_type_marks */
extern T1 T242f58(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T242f58ot1(T0* a1);
/* [detachable] ET_CLASS_TYPE.same_syntactical_class_type_with_type_marks */
extern T1 T72f104(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.same_syntactical_types */
extern T1 T577f13(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_TUPLE_TYPE.same_syntactical_type */
extern T1 T254f86(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.same_syntactical_types */
extern T1 T1956f13(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_syntactical_type */
extern T1 T1526f55(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_syntactical_type */
extern T1 T1525f58(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_LIKE_FEATURE.same_syntactical_type */
extern T1 T1524f56(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_syntactical_type */
extern T1 T242f53(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_LIKE_CURRENT.same_syntactical_type */
extern T1 T449f54(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CLASS.same_syntactical_type */
extern T1 T74f230(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CLASS_TYPE.same_syntactical_type */
extern T1 T72f88(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.same_syntactical_types */
extern T1 T1957f13(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.same_syntactical_types */
extern T1 T240f21(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_syntactical_class_type_with_type_marks */
extern T1 T1526f63(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_syntactical_class_type_with_type_marks */
extern T1 T1525f66(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_LIKE_CURRENT.same_syntactical_class_type_with_type_marks */
extern T1 T449f59(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_LIKE_FEATURE.same_syntactical_type_with_type_marks */
extern T1 T1524f57(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_LIKE_N.same_syntactical_like_feature_with_type_marks */
extern T1 T1474f37(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_syntactical_like_feature_with_type_marks */
extern T1 T1501f45(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1501f45ot1(T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.same_syntactical_like_feature_with_type_marks */
extern T1 T891f46(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T891f46ot1(T0* a1);
/* [detachable] ET_CLASS.same_syntactical_like_feature_with_type_marks */
extern T1 T74f231(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_LIKE_FEATURE.same_syntactical_like_feature_with_type_marks */
extern T1 T1524f60(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_LIKE_FEATURE.same_attachment_marks_with_default */
extern T1 T1524f61(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_TUPLE_TYPE.same_syntactical_like_feature_with_type_marks */
extern T1 T254f88(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_syntactical_like_feature_with_type_marks */
extern T1 T242f55(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T242f55ot1(T0* a1);
/* [detachable] ET_CLASS_TYPE.same_syntactical_like_feature_with_type_marks */
extern T1 T72f93(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_syntactical_like_feature_with_type_marks */
extern T1 T1526f57(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_syntactical_like_feature_with_type_marks */
extern T1 T1525f60(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_LIKE_CURRENT.same_syntactical_like_feature_with_type_marks */
extern T1 T449f56(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.same_syntactical_type_with_type_marks */
extern T1 T254f87(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_LIKE_N.same_syntactical_tuple_type_with_type_marks */
extern T1 T1474f36(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_syntactical_tuple_type_with_type_marks */
extern T1 T1501f50(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1501f50ot1(T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.same_syntactical_tuple_type_with_type_marks */
extern T1 T891f65(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T891f65ot1(T0* a1);
/* [detachable] ET_CLASS.same_syntactical_tuple_type_with_type_marks */
extern T1 T74f236(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_LIKE_FEATURE.same_syntactical_tuple_type_with_type_marks */
extern T1 T1524f69(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.same_syntactical_tuple_type_with_type_marks */
extern T1 T254f97(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_syntactical_tuple_type_with_type_marks */
extern T1 T242f65(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T242f65ot1(T0* a1);
/* [detachable] ET_CLASS_TYPE.same_syntactical_tuple_type_with_type_marks */
extern T1 T72f108(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_syntactical_tuple_type_with_type_marks */
extern T1 T1526f68(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_syntactical_tuple_type_with_type_marks */
extern T1 T1525f71(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_LIKE_CURRENT.same_syntactical_tuple_type_with_type_marks */
extern T1 T449f64(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_syntactical_type_with_type_marks */
extern T1 T242f54(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T242f54ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS_TYPE.same_syntactical_type_with_type_marks */
extern T1 T72f92(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_syntactical_type_with_type_marks */
extern T1 T1526f56(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_LIKE_N.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T1474f34(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T1501f46(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1501f46ot1(T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T891f47(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T891f47ot1(T0* a1);
/* [detachable] ET_CLASS.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T74f232(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_LIKE_FEATURE.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T1524f62(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T254f89(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T242f56(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T242f56ot1(T0* a1);
/* [detachable] ET_CLASS_TYPE.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T72f94(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T1526f58(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_attachment_marks_with_default */
extern T1 T1526f61(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T1525f63(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_attachment_marks_with_default */
extern T1 T1525f64(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_LIKE_CURRENT.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T449f57(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_syntactical_type_with_type_marks */
extern T1 T1525f59(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_LIKE_CURRENT.same_syntactical_type_with_type_marks */
extern T1 T449f55(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_LIKE_N.same_syntactical_like_current_with_type_marks */
extern T1 T1474f33(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_syntactical_like_current_with_type_marks */
extern T1 T1501f51(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1501f51ot1(T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.same_syntactical_like_current_with_type_marks */
extern T1 T891f66(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T891f66ot1(T0* a1);
/* [detachable] ET_CLASS.same_syntactical_like_current_with_type_marks */
extern T1 T74f242(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_LIKE_FEATURE.same_syntactical_like_current_with_type_marks */
extern T1 T1524f71(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.same_syntactical_like_current_with_type_marks */
extern T1 T254f101(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_syntactical_like_current_with_type_marks */
extern T1 T242f67(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T242f67ot1(T0* a1);
/* [detachable] ET_CLASS_TYPE.same_syntactical_like_current_with_type_marks */
extern T1 T72f113(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_syntactical_like_current_with_type_marks */
extern T1 T1526f70(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_syntactical_like_current_with_type_marks */
extern T1 T1525f73(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_LIKE_CURRENT.same_syntactical_like_current_with_type_marks */
extern T1 T449f66(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_LIKE_CURRENT.same_attachment_marks_with_default */
extern T1 T449f68(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SIGNATURE_CHECKER.tokens */
extern T0* T1323f6(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd2c_error */
extern void T53f363(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd2c */
extern T0* T785c996(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdrd2_error */
extern T1 T53f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdrd2c_error */
extern void T36f363(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdrd2_error */
extern T1 T36f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd2d_error */
extern void T53f362(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd2d */
extern T0* T785c995(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdrd2d_error */
extern void T36f362(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].item */
extern T0* T1695f6(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].start */
extern void T1695f9(GE_context* ac, T0* C);
/* detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE].cursor_start */
extern void T1696f10(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE].cursor_off */
extern T1 T1696f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].off */
extern T1 T1695f7(GE_context* ac, T0* C);
/* detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE].new_cursor */
extern T0* T1696f5(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].make */
extern T0* T1695c8(GE_context* ac, T0* a1);
/* [detachable] ET_INHERITED_FEATURE.adapted_feature */
extern T0* T1345f12(GE_context* ac, T0* C);
/* [detachable] ET_REDECLARED_FEATURE.adapted_feature */
extern T0* T1569f11(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.flattened_adapted_feature */
extern T0* T1516f61(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.flattened_adapted_feature */
extern T0* T253f40(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.flattened_adapted_feature */
extern T0* T476f52(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.flattened_adapted_feature */
extern T0* T1515f56(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.flattened_adapted_feature */
extern T0* T469f54(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.flattened_adapted_feature */
extern T0* T1513f57(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.flattened_adapted_feature */
extern T0* T1512f61(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.flattened_adapted_feature */
extern T0* T1511f59(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.flattened_adapted_feature */
extern T0* T371f39(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.flattened_adapted_feature */
extern T0* T369f38(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.flattened_adapted_feature */
extern T0* T359f38(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.flattened_adapted_feature */
extern T0* T360f37(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.is_adapted */
extern T1 T1345f17(GE_context* ac, T0* C);
/* [detachable] ET_REDECLARED_FEATURE.is_adapted */
extern T1 T1569f16(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_flattened_adapted */
extern T1 T1516f66(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_flattened_adapted */
extern T1 T253f45(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_flattened_adapted */
extern T1 T476f57(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_flattened_adapted */
extern T1 T1515f61(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_flattened_adapted */
extern T1 T469f59(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_flattened_adapted */
extern T1 T1513f62(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_flattened_adapted */
extern T1 T1512f66(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_flattened_adapted */
extern T1 T1511f64(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_flattened_adapted */
extern T1 T371f44(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_flattened_adapted */
extern T1 T369f43(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_flattened_adapted */
extern T1 T359f43(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_flattened_adapted */
extern T1 T360f42(GE_context* ac, T0* C);
/* detachable ET_PARENT_FEATURE.is_deferred */
extern T1 T1344f14(GE_context* ac, T0* C);
/* [detachable] ET_SIGNATURE_CHECKER.check_joined_signature_validity */
extern void T1323f14(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdjr0b_error */
extern void T53f361(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4);
/* [detachable] ET_VALIDITY_ERROR.make_vdjr0b */
extern T0* T785c994(GE_context* ac, T0* a1, T0* a2, T0* a3, T6 a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdjr_error */
extern T1 T53f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdjr0b_error */
extern void T36f361(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vdjr_error */
extern T1 T36f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdjr0a_error */
extern void T53f360(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdjr0a */
extern T0* T785c993(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdjr0a_error */
extern void T36f360(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdjr0c_error */
extern void T53f359(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdjr0c */
extern T0* T785c992(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdjr0c_error */
extern void T36f359(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SIGNATURE_CHECKER.check_redeclared_signature_validity */
extern void T1323f13(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
extern T1 T1323f13ot1(T0* a1);
extern T1 T1323f13ot2(T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd6b_error */
extern void T53f358(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd6b */
extern T0* T785c991(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdrd6_error */
extern T1 T53f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdrd6b_error */
extern void T36f358(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdrd6_error */
extern T1 T36f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd2a_error */
extern void T53f357(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd2a */
extern T0* T785c990(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdrd2a_error */
extern void T36f357(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd2b_error */
extern void T53f356(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd2b */
extern T0* T785c989(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdrd2b_error */
extern void T36f356(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_INHERITED_FEATURE.redeclared_feature */
extern T0* T1345f14(GE_context* ac, T0* C);
/* [detachable] ET_REDECLARED_FEATURE.redeclared_feature */
extern T0* T1569f13(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.flattened_redeclared_feature */
extern T0* T1516f63(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.flattened_redeclared_feature */
extern T0* T253f42(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.flattened_redeclared_feature */
extern T0* T476f54(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.flattened_redeclared_feature */
extern T0* T1515f58(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.flattened_redeclared_feature */
extern T0* T469f56(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.flattened_redeclared_feature */
extern T0* T1513f59(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.flattened_redeclared_feature */
extern T0* T1512f63(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.flattened_redeclared_feature */
extern T0* T1511f61(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.flattened_redeclared_feature */
extern T0* T371f41(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.flattened_redeclared_feature */
extern T0* T369f40(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.flattened_redeclared_feature */
extern T0* T359f40(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.flattened_redeclared_feature */
extern T0* T360f39(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.is_redeclared */
extern T1 T1345f13(GE_context* ac, T0* C);
/* [detachable] ET_REDECLARED_FEATURE.is_redeclared */
extern T1 T1569f12(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_flattened_redeclared */
extern T1 T1516f62(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_flattened_redeclared */
extern T1 T253f41(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_flattened_redeclared */
extern T1 T476f53(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_flattened_redeclared */
extern T1 T1515f57(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_flattened_redeclared */
extern T1 T469f55(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_flattened_redeclared */
extern T1 T1513f58(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_flattened_redeclared */
extern T1 T1512f62(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_flattened_redeclared */
extern T1 T1511f60(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_flattened_redeclared */
extern T1 T371f40(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_flattened_redeclared */
extern T1 T369f39(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_flattened_redeclared */
extern T1 T359f39(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_flattened_redeclared */
extern T1 T360f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].force_last_new */
extern void T1313f38(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].key_storage_put */
extern void T1313f47(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ALIAS_NAME].force */
extern void T1671f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ALIAS_NAME].fill_with */
extern void T1669f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ALIAS_NAME].extend */
extern void T1669f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].item_storage_put */
extern void T1313f46(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FLATTENED_FEATURE].force */
extern void T1667f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FLATTENED_FEATURE].fill_with */
extern void T1663f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FLATTENED_FEATURE].extend */
extern void T1663f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].slots_put */
extern void T1313f45(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].clashes_put */
extern void T1313f44(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].slots_item */
extern T6 T1313f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].hash_position */
extern T6 T1313f28(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALIAS_NAME.hash_code */
extern T6 T1030f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].resize */
extern void T1313f43(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].clashes_resize */
extern void T1313f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].special_integer_ */
extern T0* T1313f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].key_storage_resize */
extern void T1313f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ALIAS_NAME].aliased_resized_area */
extern T0* T1671f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ALIAS_NAME].aliased_resized_area */
extern T0* T1669f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].item_storage_resize */
extern void T1313f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FLATTENED_FEATURE].aliased_resized_area */
extern T0* T1667f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_FLATTENED_FEATURE].aliased_resized_area */
extern T0* T1663f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].key_storage_item */
extern T0* T1313f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].clashes_item */
extern T6 T1313f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].slots_resize */
extern void T1313f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].new_modulus */
extern T6 T1313f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].new_capacity */
extern T6 T1313f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1f_error */
extern void T53f293(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1f */
extern T0* T785c925(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_ALIAS_FREE_NAME.alias_lower_name */
extern T0* T1518f10(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.alias_name */
extern T0* T1518f49(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.alias_lower_name */
extern T0* T1030f7(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.alias_name */
extern T0* T1030f19(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.position */
extern T0* T1518f11(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.position */
extern T0* T1030f8(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vfav1_error */
extern T1 T53f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vfav1f_error */
extern void T36f293(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vfav1_error */
extern T1 T36f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1c_error */
extern void T53f292(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1c */
extern T0* T785c924(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vfav1c_error */
extern void T36f292(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav3b_error */
extern void T53f291(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vfav3b */
extern T0* T785c923(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vfav3_error */
extern T1 T53f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vfav3b_error */
extern void T36f291(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vfav3_error */
extern T1 T36f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav2b_error */
extern void T53f290(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vfav2b */
extern T0* T785c922(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vfav2_error */
extern T1 T53f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vfav2b_error */
extern void T36f290(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vfav2_error */
extern T1 T36f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1g_error */
extern void T53f289(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1g */
extern T0* T785c921(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vfav1g_error */
extern void T36f289(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1d_error */
extern void T53f288(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1d */
extern T0* T785c920(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vfav1d_error */
extern void T36f288(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav3c_error */
extern void T53f287(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vfav3c */
extern T0* T785c919(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vfav3c_error */
extern void T36f287(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav2c_error */
extern void T53f286(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vfav2c */
extern T0* T785c918(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vfav2c_error */
extern void T36f286(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1h_error */
extern void T53f285(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1h */
extern T0* T785c917(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vfav1h_error */
extern void T36f285(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1e_error */
extern void T53f284(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1e */
extern T0* T785c916(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vfav1e_error */
extern void T36f284(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav3d_error */
extern void T53f283(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vfav3d */
extern T0* T785c915(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vfav3d_error */
extern void T36f283(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav2d_error */
extern void T53f282(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vfav2d */
extern T0* T785c914(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vfav2d_error */
extern void T36f282(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_INHERITED_FEATURE.inherited_feature */
extern T0* T1345f16(GE_context* ac, T0* C);
/* [detachable] ET_REDECLARED_FEATURE.inherited_feature */
extern T0* T1569f15(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.flattened_inherited_feature */
extern T0* T1516f65(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.flattened_inherited_feature */
extern T0* T253f44(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.flattened_inherited_feature */
extern T0* T476f56(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.flattened_inherited_feature */
extern T0* T1515f60(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.flattened_inherited_feature */
extern T0* T469f58(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.flattened_inherited_feature */
extern T0* T1513f61(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.flattened_inherited_feature */
extern T0* T1512f65(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.flattened_inherited_feature */
extern T0* T1511f63(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.flattened_inherited_feature */
extern T0* T371f43(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.flattened_inherited_feature */
extern T0* T369f42(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.flattened_inherited_feature */
extern T0* T359f42(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.flattened_inherited_feature */
extern T0* T360f41(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.is_inherited */
extern T1 T1345f15(GE_context* ac, T0* C);
/* [detachable] ET_REDECLARED_FEATURE.is_inherited */
extern T1 T1569f14(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_flattened_inherited */
extern T1 T1516f64(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_flattened_inherited */
extern T1 T253f43(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_flattened_inherited */
extern T1 T476f55(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_flattened_inherited */
extern T1 T1515f59(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_flattened_inherited */
extern T1 T469f57(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_flattened_inherited */
extern T1 T1513f60(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_flattened_inherited */
extern T1 T1512f64(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_flattened_inherited */
extern T1 T1511f62(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_flattened_inherited */
extern T1 T371f42(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_flattened_inherited */
extern T1 T369f41(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_flattened_inherited */
extern T1 T359f41(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_flattened_inherited */
extern T1 T360f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].found_key */
extern T0* T1313f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].found_item */
extern T0* T1313f23(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].item_storage_item */
extern T0* T1313f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].found */
extern T1 T1313f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].search */
extern void T1313f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].search_position */
extern void T1313f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ALIAS_NAME_TESTER.test */
extern T1 T1314f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].after */
extern T1 T1310f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].cursor_after */
extern T1 T1310f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].after */
extern T1 T1662f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].start */
extern void T1310f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].cursor_start */
extern void T1310f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].is_empty */
extern T1 T1310f31(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_redeclared_signatures_checked */
extern void T74f281(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_FEATURE_FLATTENER.check_anchored_signatures */
extern void T459f44(GE_context* ac, T0* C);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.check_signatures */
extern void T1322f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].wipe_out */
extern void T1693f12(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].wipe_out */
extern void T699f19(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].keep_head */
extern void T1414f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].set_count */
extern void T1414f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].move_all_cursors_after */
extern void T699f17(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable DS_LINKABLE [[attached] INTEGER_32]].set_next_cursor */
extern void T1416f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable DS_LINKABLE [[attached] INTEGER_32]].set_after */
extern void T1416f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].wipe_out */
extern void T263f17(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].move_all_cursors_after */
extern void T263f25(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] INTEGER_32].set_next_cursor */
extern void T632f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] INTEGER_32].set_after */
extern void T632f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].wipe_out */
extern void T1930f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].slots_wipe_out */
extern void T1930f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].clashes_wipe_out */
extern void T1930f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].key_storage_wipe_out */
extern void T1930f41(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_LIKE_FEATURE].keep_head */
extern void T1931f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_LIKE_FEATURE].set_count */
extern void T1931f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].item_storage_wipe_out */
extern void T1930f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].move_all_cursors_after */
extern void T1930f38(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].set_next_cursor */
extern void T1993f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].set_after */
extern void T1993f7(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].wipe_out */
extern void T1694f12(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].move_all_cursors_after */
extern void T1694f14(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE].set_next_cursor */
extern void T1933f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE].set_after */
extern void T1933f10(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vtat2a_error */
extern void T53f355(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vtat2a */
extern T0* T785c988(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE].forth */
extern void T1933f14(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].cursor_forth */
extern void T1694f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE].set_position */
extern void T1933f15(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].add_traversing_cursor */
extern void T1694f19(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].remove_traversing_cursor */
extern void T1694f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE].go_after */
extern void T1933f13(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].cursor_go_after */
extern void T1694f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE].off */
extern T1 T1933f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE].before */
extern T1 T1933f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE].item */
extern T0* T1933f6(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].cursor_item */
extern T0* T1694f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE].after */
extern T1 T1933f4(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE].start */
extern void T1933f12(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].cursor_start */
extern void T1694f16(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].new_cursor */
extern T0* T1694f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_LIKE_FEATURE].make */
extern T0* T1933c9(GE_context* ac, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vtat2a_error */
extern void T36f355(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.error_handler */
extern T0* T1322f6(GE_context* ac, T0* C);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.set_fatal_error */
extern void T1322f10(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].is_empty */
extern T1 T1694f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].sort */
extern void T1693f11(GE_context* ac, T0* C);
/* [detachable] ARRAY [[attached] BOOLEAN].item */
extern T1 T700f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [[attached] BOOLEAN].put */
extern void T700f6(GE_context* ac, T0* C, T1 a1, T6 a2);
/* [detachable] ARRAY [[attached] BOOLEAN].make_filled */
extern T0* T700c5(GE_context* ac, T1 a1, T6 a2, T6 a3);
/* [detachable] ARRAY [[attached] BOOLEAN].make_filled_area */
extern void T700f7(GE_context* ac, T0* C, T1 a1, T6 a2);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].item */
extern T0* T1694f8(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].make */
extern T0* T1694c10(GE_context* ac, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_LIKE_FEATURE].set_internal_cursor */
extern void T1694f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIKE_FEATURE].make */
extern T0* T1932f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_LIKE_FEATURE].make_empty */
extern T0* T1931c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIKE_FEATURE].default_create */
extern T0* T1932c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].cloned_object */
extern T0* T699f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].twin */
extern T0* T699f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].copy */
extern void T699f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].resized_area */
extern T0* T1414f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].copy_data */
extern void T1414f10(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].force */
extern void T1414f14(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].move_data */
extern void T1414f13(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].overlapping_move */
extern void T1414f16(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].non_overlapping_move */
extern void T1414f15(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable DS_LINKABLE [[attached] INTEGER_32]].make_empty */
extern T0* T1414c6(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].new_cursor */
extern T0* T699f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable DS_LINKABLE [[attached] INTEGER_32]].make */
extern T0* T1416c5(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].set_internal_cursor */
extern void T699f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].valid_cursor */
extern T1 T699f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].cloned_object */
extern T0* T263f12(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].twin */
extern T0* T263f13(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].copy */
extern void T263f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].new_cursor */
extern T0* T263f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] INTEGER_32].make */
extern T0* T632c5(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].set_internal_cursor */
extern void T263f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].valid_cursor */
extern T1 T263f14(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.is_parenthesisable */
extern T1 T1516f57(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_parenthesisable */
extern T1 T476f38(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_parenthesisable */
extern T1 T1515f51(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_parenthesisable */
extern T1 T469f40(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_frozen */
extern T1 T1516f56(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_frozen */
extern T1 T476f37(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_frozen */
extern T1 T1515f50(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_frozen */
extern T1 T469f39(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_deferred */
extern T1 T1516f55(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_deferred */
extern T1 T476f36(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_deferred */
extern T1 T1515f49(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_deferred */
extern T1 T469f38(GE_context* ac, T0* C);
/* [detachable] ET_PROCEDURE_LIST.item */
extern T0* T258f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vffd7b_error */
extern void T53f281(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vffd7b */
extern T0* T785c913(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vffd7_error */
extern T1 T53f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vffd7b_error */
extern void T36f281(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vffd7_error */
extern T1 T36f48(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.has_formal_types */
extern T1 T1526f18(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.has_formal_types */
extern T1 T1525f20(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.has_formal_types */
extern T1 T1501f52(GE_context* ac, T0* C, T0* a1);
extern T1 T1501f52ot1(T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.has_formal_types */
extern T1 T891f54(GE_context* ac, T0* C, T0* a1);
extern T1 T891f54ot1(T0* a1);
/* [detachable] ET_CLASS.has_formal_types */
extern T1 T74f206(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.has_formal_types */
extern T1 T603f16(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.has_formal_types */
extern T1 T1524f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.has_formal_types */
extern T1 T254f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.has_formal_types */
extern T1 T1957f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.has_formal_types */
extern T1 T240f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.has_formal_types */
extern T1 T242f39(GE_context* ac, T0* C, T0* a1);
extern T1 T242f39ot1(T0* a1);
/* [detachable] ET_CLASS_TYPE.has_formal_types */
extern T1 T72f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.has_formal_types */
extern T1 T577f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.has_formal_types */
extern T1 T1956f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.has_formal_types */
extern T1 T449f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vffd7a_error */
extern void T53f280(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vffd7a */
extern T0* T785c912(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vffd7a_error */
extern void T36f280(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.has_unqualified_anchored_type */
extern T1 T1526f17(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.has_unqualified_anchored_type */
extern T1 T1525f19(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.has_unqualified_anchored_type */
extern T1 T1501f62(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.has_unqualified_anchored_type */
extern T1 T891f62(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.has_unqualified_anchored_type */
extern T1 T74f205(GE_context* ac, T0* C);
/* detachable ET_FORMAL_PARAMETER_LIST.has_unqualified_anchored_type */
extern T1 T603f15(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.has_unqualified_anchored_type */
extern T1 T1524f19(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.has_unqualified_anchored_type */
extern T1 T254f68(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.has_unqualified_anchored_type */
extern T1 T1957f7(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.has_unqualified_anchored_type */
extern T1 T240f16(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.has_unqualified_anchored_type */
extern T1 T242f38(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.has_unqualified_anchored_type */
extern T1 T72f67(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.has_unqualified_anchored_type */
extern T1 T577f8(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.has_unqualified_anchored_type */
extern T1 T1956f7(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.has_unqualified_anchored_type */
extern T1 T449f33(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_once */
extern T1 T253f56(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_once */
extern T1 T1513f44(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_once */
extern T1 T1512f46(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_once */
extern T1 T1511f44(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_once */
extern T1 T371f47(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_once */
extern T1 T369f46(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_once */
extern T1 T359f33(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_once */
extern T1 T360f45(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav5a_error */
extern void T53f239(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vfav5a */
extern T0* T785c889(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_ALIAS_FREE_NAME.lower_name */
extern T0* T1518f16(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.name */
extern T0* T1518f50(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.lower_name */
extern T0* T1030f14(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.name */
extern T0* T1030f20(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vfav5_error */
extern T1 T53f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vfav5a_error */
extern void T36f239(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vfav5_error */
extern T1 T36f40(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALIAS_FREE_NAME.is_prefix */
extern T1 T1518f14(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_prefix */
extern T1 T1030f12(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infix */
extern T1 T1518f15(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infix */
extern T1 T1030f13(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1k_error */
extern void T53f279(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1k */
extern T0* T785c911(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vfav1k_error */
extern void T36f279(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1b_error */
extern void T53f278(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1b */
extern T0* T785c910(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vfav1b_error */
extern void T36f278(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav4d_error */
extern void T53f277(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfav4d */
extern T0* T785c909(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vfav4_error */
extern T1 T53f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vfav4d_error */
extern void T36f277(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vfav4_error */
extern T1 T36f36(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALIAS_FREE_NAME.set_infix */
extern void T1518f58(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.set_infix */
extern void T1030f111(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_infixable */
extern T1 T1518f19(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_infixable */
extern T1 T1030f17(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_infixable */
extern T1 T253f34(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_infixable */
extern T1 T1513f31(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_infixable */
extern T1 T1512f56(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_infixable */
extern T1 T1511f55(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_infixable */
extern T1 T371f33(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_infixable */
extern T1 T369f32(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_infixable */
extern T1 T359f32(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_infixable */
extern T1 T360f31(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1a_error */
extern void T53f276(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1a */
extern T0* T785c908(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vfav1a_error */
extern void T36f276(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav4c_error */
extern void T53f275(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfav4c */
extern T0* T785c907(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vfav4c_error */
extern void T36f275(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_FREE_NAME.set_prefix */
extern void T1518f57(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.set_prefix */
extern void T1030f110(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.is_prefixable */
extern T1 T1518f18(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.is_prefixable */
extern T1 T1030f16(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_prefixable */
extern T1 T253f33(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_prefixable */
extern T1 T1513f30(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_prefixable */
extern T1 T1512f55(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_prefixable */
extern T1 T1511f54(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_prefixable */
extern T1 T371f32(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_prefixable */
extern T1 T369f31(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_prefixable */
extern T1 T359f31(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_prefixable */
extern T1 T360f30(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav3a_error */
extern void T53f274(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfav3a */
extern T0* T785c906(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vfav3a_error */
extern void T36f274(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EXTERNAL_FUNCTION.is_parenthesisable */
extern T1 T253f73(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_parenthesisable */
extern T1 T1513f53(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_parenthesisable */
extern T1 T1512f54(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_parenthesisable */
extern T1 T1511f53(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_parenthesisable */
extern T1 T371f31(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_parenthesisable */
extern T1 T369f30(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_parenthesisable */
extern T1 T359f30(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_parenthesisable */
extern T1 T360f29(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav4b_error */
extern void T53f273(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfav4b */
extern T0* T785c905(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vfav4b_error */
extern void T36f273(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav2a_error */
extern void T53f272(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfav2a */
extern T0* T785c904(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vfav2a_error */
extern void T36f272(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EXTERNAL_FUNCTION.is_bracketable */
extern T1 T253f72(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_bracketable */
extern T1 T1513f52(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_bracketable */
extern T1 T1512f53(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_bracketable */
extern T1 T1511f52(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_bracketable */
extern T1 T371f30(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_bracketable */
extern T1 T369f29(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_bracketable */
extern T1 T359f66(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_bracketable */
extern T1 T360f28(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav4a_error */
extern void T53f271(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vfav4a */
extern T0* T785c903(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vfav4a_error */
extern void T36f271(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME_LIST.head_has_alias_name */
extern T1 T911f6(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ET_ALIAS_FREE_NAME.same_alias_name */
extern T1 T1518f20(GE_context* ac, T0* C, T0* a1);
extern T1 T1518f20ot1(T0* a1, T0** a2);
/* detachable ET_ALIAS_NAME.same_alias_name */
extern T1 T1030f18(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALIAS_NAME_LIST.item */
extern T0* T911f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vffd4a_error */
extern void T53f270(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vffd4a */
extern T0* T785c902(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vffd4_error */
extern T1 T53f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vffd4a_error */
extern void T36f270(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vffd4_error */
extern T1 T36f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.is_frozen */
extern T1 T253f70(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_frozen */
extern T1 T1513f50(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_frozen */
extern T1 T1512f50(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_frozen */
extern T1 T1511f48(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_frozen */
extern T1 T371f28(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_frozen */
extern T1 T369f27(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_frozen */
extern T1 T359f64(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_frozen */
extern T1 T360f26(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_deferred */
extern T1 T253f69(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_deferred */
extern T1 T1513f49(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_deferred */
extern T1 T1512f49(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_deferred */
extern T1 T1511f47(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_deferred */
extern T1 T371f27(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_deferred */
extern T1 T369f26(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_deferred */
extern T1 T359f63(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_deferred */
extern T1 T360f25(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.resolve_signature_unfolded_tuple_actual_parameters */
extern void T459f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.resolve_type */
extern void T1321f7(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_FLATTENER.check_signature_vtct_validity */
extern void T459f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.check_signature_vtct_validity */
extern void T1323f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_FLATTENER.resolve_identifier_signature */
extern void T459f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vrfa0a_error */
extern void T53f299(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vrfa0a */
extern T0* T785c932(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vrfa_error */
extern T1 T53f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vrfa0a_error */
extern void T36f299(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vrfa_error */
extern T1 T36f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].item */
extern T0* T1310f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].has */
extern T1 T1310f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vreg0a_error */
extern void T53f298(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vreg0a */
extern T0* T785c931(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vreg_error */
extern T1 T53f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vreg0a_error */
extern void T36f298(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vreg_error */
extern T1 T36f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].wipe_out */
extern void T715f17(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_PROCEDURE].keep_head */
extern void T913f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_PROCEDURE].set_count */
extern void T913f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].move_all_cursors_after */
extern void T715f23(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PROCEDURE].set_next_cursor */
extern void T1425f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PROCEDURE].set_after */
extern void T1425f10(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_procedures */
extern void T74f280(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PROCEDURE_LIST.set_declared_count */
extern void T258f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_PROCEDURE_LIST.put_first */
extern void T258f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PROCEDURE].force */
extern void T914f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_PROCEDURE].fill_with */
extern void T913f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_PROCEDURE].extend */
extern void T913f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].item */
extern T0* T715f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_QUICK_SORTER [[attached] ET_PROCEDURE].subsort */
extern void T1567f3(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] DS_QUICK_SORTER [[attached] ET_PROCEDURE].subsort_with_comparator */
extern void T1567f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].swap */
extern void T715f36(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].replace */
extern void T715f35(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_SEEDED_PROCEDURE_COMPARATOR.less_than */
extern T1 T1762f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_FLATTENER.procedure_sorter */
extern T0* T459f26(GE_context* ac, T0* C);
/* [detachable] DS_QUICK_SORTER [[attached] ET_PROCEDURE].make */
extern T0* T1567c2(GE_context* ac, T0* a1);
/* [detachable] ET_SEEDED_PROCEDURE_COMPARATOR.make */
extern T0* T1762c2(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].wipe_out */
extern void T719f17(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_QUERY].keep_head */
extern void T910f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_QUERY].set_count */
extern void T910f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].move_all_cursors_after */
extern void T719f23(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUERY].set_next_cursor */
extern void T1426f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUERY].set_after */
extern void T1426f10(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_queries */
extern void T74f279(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUERY_LIST.set_declared_count */
extern void T257f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_QUERY_LIST.put_first */
extern void T257f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_QUERY].force */
extern void T912f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_QUERY].fill_with */
extern void T910f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_QUERY].extend */
extern void T910f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].item */
extern T0* T719f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_QUICK_SORTER [[attached] ET_QUERY].subsort */
extern void T1566f3(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] DS_QUICK_SORTER [[attached] ET_QUERY].subsort_with_comparator */
extern void T1566f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].swap */
extern void T719f36(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].replace */
extern void T719f35(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_SEEDED_QUERY_COMPARATOR.less_than */
extern T1 T1759f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_FLATTENER.query_sorter */
extern T0* T459f25(GE_context* ac, T0* C);
/* [detachable] DS_QUICK_SORTER [[attached] ET_QUERY].make */
extern T0* T1566c2(GE_context* ac, T0* a1);
/* [detachable] ET_SEEDED_QUERY_COMPARATOR.make */
extern T0* T1759c2(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].back */
extern void T1310f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].cursor_back */
extern void T1310f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].set_before */
extern void T1662f13(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].force_last */
extern void T719f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].resize */
extern void T719f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_QUERY].aliased_resized_area */
extern T0* T912f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_QUERY].aliased_resized_area */
extern T0* T910f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].new_capacity */
extern T6 T719f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].extendible */
extern T1 T719f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].force_last */
extern void T715f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].resize */
extern void T715f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PROCEDURE].aliased_resized_area */
extern T0* T914f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_PROCEDURE].aliased_resized_area */
extern T0* T913f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].new_capacity */
extern T6 T715f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].extendible */
extern T1 T715f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_INHERITED_FEATURE.flattened_feature */
extern T0* T1345f3(GE_context* ac, T0* C);
/* [detachable] ET_REDECLARED_FEATURE.flattened_feature */
extern T0* T1569f2(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.flattened_feature */
extern T0* T1516f58(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.flattened_feature */
extern T0* T253f37(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.flattened_feature */
extern T0* T476f49(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.flattened_feature */
extern T0* T1515f53(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.flattened_feature */
extern T0* T469f51(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.flattened_feature */
extern T0* T1513f54(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.flattened_feature */
extern T0* T1512f58(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.flattened_feature */
extern T0* T1511f56(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.flattened_feature */
extern T0* T371f36(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.flattened_feature */
extern T0* T369f35(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.flattened_feature */
extern T0* T359f35(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.flattened_feature */
extern T0* T360f34(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.flatten_feature */
extern void T459f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.flatten_inherited_feature */
extern void T459f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdjr2b_error */
extern void T53f317(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdjr2b */
extern T0* T785c950(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.alias_lower_names */
extern T0* T785f254(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdjr2_error */
extern T1 T53f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdjr2b_error */
extern void T36f317(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdjr2_error */
extern T1 T36f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vmfn2b_error */
extern void T53f316(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vmfn2b */
extern T0* T785c949(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vmfn2_error */
extern T1 T53f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vmfn2b_error */
extern void T36f316(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vmfn2_error */
extern T1 T36f62(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALIAS_NAME_LIST.same_alias_names */
extern T1 T911f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALIAS_NAME_LIST.has_alias_name */
extern T1 T911f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALIAS_NAME_LIST.first */
extern T0* T911f9(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdjr2a_error */
extern void T53f315(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdjr2a */
extern T0* T785c948(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdjr2a_error */
extern void T36f315(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vmfn2a_error */
extern void T53f314(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vmfn2a */
extern T0* T785c947(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vmfn2a_error */
extern void T36f314(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_INHERITED_FEATURE.set_flattened_parent */
extern void T1345f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INHERITED_FEATURE.set_flattened_feature */
extern void T1345f32(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_other_seeds */
extern void T1516f121(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_other_seeds */
extern void T476f111(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_other_seeds */
extern void T1515f117(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_other_seeds */
extern void T469f114(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_other_seeds */
extern void T253f124(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_other_seeds */
extern void T1513f114(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_other_seeds */
extern void T1512f120(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_other_seeds */
extern void T1511f117(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_other_seeds */
extern void T371f112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_other_seeds */
extern void T369f109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_other_seeds */
extern void T359f117(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_other_seeds */
extern void T360f105(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_first_seed */
extern void T1516f112(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_first_seed */
extern void T476f102(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_PROCEDURE.set_first_seed */
extern void T1515f108(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DO_PROCEDURE.set_first_seed */
extern void T469f105(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_first_seed */
extern void T253f115(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DEFERRED_FUNCTION.set_first_seed */
extern void T1513f105(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_FUNCTION.set_first_seed */
extern void T1512f111(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_FUNCTION.set_first_seed */
extern void T1511f108(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_first_seed */
extern void T371f103(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_first_seed */
extern void T369f100(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_first_seed */
extern void T359f108(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ATTRIBUTE.set_first_seed */
extern void T360f96(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_clients */
extern void T1516f105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_clients */
extern void T476f95(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_clients */
extern void T1515f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_clients */
extern void T469f99(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_clients */
extern void T253f108(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_clients */
extern void T1513f98(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_clients */
extern void T1512f105(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_clients */
extern void T1511f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_clients */
extern void T371f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_clients */
extern void T369f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_clients */
extern void T359f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_clients */
extern void T360f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.process_replicated_seeds */
extern void T459f54(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_REDECLARED_FEATURE.set_other_seeds */
extern void T1569f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INHERITED_FEATURE.set_other_seeds */
extern void T1345f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REDECLARED_FEATURE.set_first_seed */
extern void T1569f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_INHERITED_FEATURE.set_first_seed */
extern void T1345f35(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FEATURE_IDS.replace */
extern void T604f10(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_FEATURE_IDS.first */
extern T6 T604f8(GE_context* ac, T0* C);
/* detachable ET_FEATURE_IDS.cloned_object */
extern T0* T604f4(GE_context* ac, T0* C);
/* detachable ET_FEATURE_IDS.twin */
extern T0* T604f3(GE_context* ac, T0* C);
/* detachable ET_FEATURE_IDS.copy */
extern void T604f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_ARRAY_ROUTINES [[attached] INTEGER_32].cloned_array */
extern T0* T902f1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ARRAY [[attached] INTEGER_32].twin */
extern T0* T291f9(GE_context* ac, T0* C);
/* [detachable] ARRAY [[attached] INTEGER_32].copy */
extern void T291f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [[attached] INTEGER_32].twin */
extern T0* T128f7(GE_context* ac, T0* C);
/* detachable ET_FEATURE_IDS.integer_array_ */
extern T0* T604f5(GE_context* ac, T0* C);
/* [detachable] ET_REDECLARED_FEATURE.is_other_seeds_shared */
extern T1 T1569f20(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.is_other_seeds_shared */
extern T1 T1345f21(GE_context* ac, T0* C);
/* detachable ET_FEATURE_IDS.make */
extern T0* T604c9(GE_context* ac, T6 a1);
/* detachable ET_FEATURE_IDS.put */
extern void T604f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_ARRAY_ROUTINES [[attached] INTEGER_32].resize_with_default */
extern void T902f4(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] ARRAY [[attached] INTEGER_32].conservative_resize_with_default */
extern void T291f18(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* detachable ET_FEATURE_IDS.remove */
extern void T604f11(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_PROCEDURE.id */
extern T6 T1516f28(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.id */
extern T6 T476f25(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.id */
extern T6 T1515f28(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.id */
extern T6 T469f27(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.id */
extern T6 T253f30(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.id */
extern T6 T1513f27(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.id */
extern T6 T1512f30(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.id */
extern T6 T1511f29(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.id */
extern T6 T371f2(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.id */
extern T6 T369f22(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.id */
extern T6 T359f27(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.id */
extern T6 T360f20(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.set_version */
extern void T1516f113(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_version */
extern void T476f103(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_PROCEDURE.set_version */
extern void T1515f109(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DO_PROCEDURE.set_version */
extern void T469f106(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_version */
extern void T253f116(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DEFERRED_FUNCTION.set_version */
extern void T1513f106(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_FUNCTION.set_version */
extern void T1512f112(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_FUNCTION.set_version */
extern void T1511f109(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_version */
extern void T371f104(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_version */
extern void T369f101(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_version */
extern void T359f109(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ATTRIBUTE.set_version */
extern void T360f97(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS.register_feature */
extern void T74f283(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_id */
extern void T1516f89(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_id */
extern void T253f91(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_id */
extern void T476f83(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_PROCEDURE.set_id */
extern void T1515f83(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DO_PROCEDURE.set_id */
extern void T469f84(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DEFERRED_FUNCTION.set_id */
extern void T1513f82(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_FUNCTION.set_id */
extern void T1512f85(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_FUNCTION.set_id */
extern void T1511f83(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_id */
extern void T371f83(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_id */
extern void T369f82(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_id */
extern void T359f84(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ATTRIBUTE.set_id */
extern void T360f79(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_PROCEDURE.renamed_feature */
extern T0* T1516f76(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_builtin_code */
extern void T1516f111(GE_context* ac, T0* C, T8 a1, T8 a2);
/* detachable ET_EXTERNAL_PROCEDURE.set_iteration_components */
extern void T1516f98(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_object_tests */
extern void T1516f97(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_first_indexing */
extern void T1516f106(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_feature_clause */
extern void T1516f109(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_semicolon */
extern void T1516f108(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_frozen_keyword */
extern void T1516f99(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_end_keyword */
extern void T1516f107(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_is_keyword */
extern void T1516f100(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_alias_clause */
extern void T1516f104(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_postconditions */
extern void T1516f103(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_preconditions */
extern void T1516f102(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_obsolete_message */
extern void T1516f101(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.make */
extern T0* T1516c88(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_EXTERNAL_PROCEDURE.tokens */
extern T0* T1516f47(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.renamed_feature */
extern T0* T476f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_iteration_components */
extern void T476f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_object_tests */
extern void T476f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_first_indexing */
extern void T476f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_feature_clause */
extern void T476f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_semicolon */
extern void T476f99(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_frozen_keyword */
extern void T476f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_end_keyword */
extern void T476f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_deferred_keyword */
extern void T476f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_is_keyword */
extern void T476f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_postconditions */
extern void T476f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_preconditions */
extern void T476f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_obsolete_message */
extern void T476f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.make */
extern T0* T476c79(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DEFERRED_PROCEDURE.make */
extern void T476f79p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_TOKEN_CONSTANTS.deferred_keyword */
extern T0* T73f249(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_deferred */
extern T0* T413c53(GE_context* ac);
/* [detachable] ET_DEFERRED_PROCEDURE.tokens */
extern T0* T476f42(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.renamed_feature */
extern T0* T1515f69(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_iteration_components */
extern void T1515f92(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_object_tests */
extern void T1515f91(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_first_indexing */
extern void T1515f103(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_feature_clause */
extern void T1515f106(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_semicolon */
extern void T1515f105(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_frozen_keyword */
extern void T1515f93(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_end_keyword */
extern void T1515f104(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_is_keyword */
extern void T1515f94(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_rescue_clause */
extern void T1515f101(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_postconditions */
extern void T1515f100(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_compound */
extern void T1515f99(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_locals */
extern void T1515f97(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_preconditions */
extern void T1515f96(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_obsolete_message */
extern void T1515f95(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.make */
extern T0* T1515c82(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_ONCE_PROCEDURE.tokens */
extern T0* T1515f36(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.renamed_feature */
extern T0* T469f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_iteration_components */
extern void T469f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_object_tests */
extern void T469f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_first_indexing */
extern void T469f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_feature_clause */
extern void T469f103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_semicolon */
extern void T469f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_frozen_keyword */
extern void T469f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_end_keyword */
extern void T469f101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_is_keyword */
extern void T469f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_rescue_clause */
extern void T469f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_postconditions */
extern void T469f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_compound */
extern void T469f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_locals */
extern void T469f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_preconditions */
extern void T469f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_obsolete_message */
extern void T469f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.renamed_feature */
extern T0* T253f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_builtin_code */
extern void T253f114(GE_context* ac, T0* C, T8 a1, T8 a2);
/* [detachable] ET_EXTERNAL_FUNCTION.set_iteration_components */
extern void T253f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_object_tests */
extern void T253f99(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_first_indexing */
extern void T253f109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_feature_clause */
extern void T253f112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_semicolon */
extern void T253f111(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_frozen_keyword */
extern void T253f101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_end_keyword */
extern void T253f110(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_is_keyword */
extern void T253f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_alias_clause */
extern void T253f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_postconditions */
extern void T253f106(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_preconditions */
extern void T253f105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_obsolete_message */
extern void T253f104(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_assigner */
extern void T253f103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.make */
extern T0* T253c90(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_DEFERRED_FUNCTION.renamed_feature */
extern T0* T1513f66(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_iteration_components */
extern void T1513f91(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_object_tests */
extern void T1513f90(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_first_indexing */
extern void T1513f99(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_feature_clause */
extern void T1513f103(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_semicolon */
extern void T1513f102(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_frozen_keyword */
extern void T1513f92(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_end_keyword */
extern void T1513f101(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_deferred_keyword */
extern void T1513f100(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_is_keyword */
extern void T1513f93(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_postconditions */
extern void T1513f97(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_preconditions */
extern void T1513f96(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_obsolete_message */
extern void T1513f95(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_assigner */
extern void T1513f94(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.make */
extern T0* T1513c81(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_DEFERRED_FUNCTION.make */
extern void T1513f81p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_DEFERRED_FUNCTION.tokens */
extern T0* T1513f48(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.renamed_feature */
extern T0* T1512f70(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_iteration_components */
extern void T1512f94(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_object_tests */
extern void T1512f93(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_first_indexing */
extern void T1512f106(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_feature_clause */
extern void T1512f109(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_semicolon */
extern void T1512f108(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_frozen_keyword */
extern void T1512f95(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_end_keyword */
extern void T1512f107(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_is_keyword */
extern void T1512f96(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_rescue_clause */
extern void T1512f104(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_postconditions */
extern void T1512f103(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_compound */
extern void T1512f102(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_locals */
extern void T1512f100(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_preconditions */
extern void T1512f99(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_obsolete_message */
extern void T1512f98(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_assigner */
extern void T1512f97(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.make */
extern T0* T1512c84(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ONCE_FUNCTION.tokens */
extern T0* T1512f40(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.renamed_feature */
extern T0* T1511f68(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_iteration_components */
extern void T1511f92(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_object_tests */
extern void T1511f91(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_first_indexing */
extern void T1511f103(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_feature_clause */
extern void T1511f106(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_semicolon */
extern void T1511f105(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_frozen_keyword */
extern void T1511f93(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_end_keyword */
extern void T1511f104(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_is_keyword */
extern void T1511f94(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_rescue_clause */
extern void T1511f101(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_postconditions */
extern void T1511f100(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_compound */
extern void T1511f99(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_locals */
extern void T1511f98(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_preconditions */
extern void T1511f97(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_obsolete_message */
extern void T1511f96(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_assigner */
extern void T1511f95(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.make */
extern T0* T1511c82(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_DO_FUNCTION.tokens */
extern T0* T1511f50(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.renamed_feature */
extern T0* T371f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_constant */
extern void T371f101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_iteration_components */
extern void T371f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_object_tests */
extern void T371f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_first_indexing */
extern void T371f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_feature_clause */
extern void T371f99(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_semicolon */
extern void T371f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_frozen_keyword */
extern void T371f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_unique_keyword */
extern void T371f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_is_keyword */
extern void T371f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_assigner */
extern void T371f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.make */
extern T0* T371c82(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.make */
extern T0* T372c63(GE_context* ac, T0* a1, T11 a2, T1 a3);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.make_leaf */
extern void T372f65(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.unique_keyword */
extern T0* T73f1537(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_unique */
extern T0* T413c62(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.is_keyword */
extern T0* T73f708(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_is */
extern T0* T413c69(GE_context* ac);
/* [detachable] ET_UNIQUE_ATTRIBUTE.tokens */
extern T0* T371f62(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.renamed_feature */
extern T0* T369f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_iteration_components */
extern void T369f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_object_tests */
extern void T369f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_first_indexing */
extern void T369f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_feature_clause */
extern void T369f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_semicolon */
extern void T369f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_frozen_keyword */
extern void T369f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_is_keyword */
extern void T369f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_assigner */
extern void T369f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.make */
extern T0* T369c81(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CONSTANT_ATTRIBUTE.tokens */
extern T0* T369f60(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.renamed_feature */
extern T0* T359f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_iteration_components */
extern void T359f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_object_tests */
extern void T359f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_first_indexing */
extern void T359f103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_feature_clause */
extern void T359f106(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_semicolon */
extern void T359f105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_frozen_keyword */
extern void T359f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_end_keyword */
extern void T359f104(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_rescue_clause */
extern void T359f101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_postconditions */
extern void T359f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_compound */
extern void T359f99(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_locals */
extern void T359f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_preconditions */
extern void T359f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_obsolete_message */
extern void T359f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_assigner */
extern void T359f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.make */
extern T0* T359c83(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EXTENDED_ATTRIBUTE.make */
extern void T359f83p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EXTENDED_ATTRIBUTE.tokens */
extern T0* T359f61(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.renamed_feature */
extern T0* T360f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_iteration_components */
extern void T360f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_object_tests */
extern void T360f87(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_first_indexing */
extern void T360f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_feature_clause */
extern void T360f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_semicolon */
extern void T360f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_frozen_keyword */
extern void T360f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_assigner */
extern void T360f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.make */
extern T0* T360c78(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ATTRIBUTE.tokens */
extern T0* T360f55(GE_context* ac, T0* C);
/* detachable ET_PARENT_FEATURE.extended_name */
extern T0* T1344f21(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.extended_name */
extern T0* T1516f14(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.extended_name */
extern T0* T476f7(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.extended_name */
extern T0* T1515f14(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.extended_name */
extern T0* T469f8(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.extended_name */
extern T0* T253f15(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.extended_name */
extern T0* T1513f13(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.extended_name */
extern T0* T1512f14(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.extended_name */
extern T0* T1511f13(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.extended_name */
extern T0* T371f12(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.extended_name */
extern T0* T369f11(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.extended_name */
extern T0* T359f15(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.extended_name */
extern T0* T360f10(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.set_other_precursors */
extern void T1516f120(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_other_precursors */
extern void T476f110(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_other_precursors */
extern void T1515f116(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_other_precursors */
extern void T469f113(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_other_precursors */
extern void T253f123(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_other_precursors */
extern void T1513f113(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_other_precursors */
extern void T1512f119(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_other_precursors */
extern void T1511f116(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_other_precursors */
extern void T371f111(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_other_precursors */
extern void T369f108(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_other_precursors */
extern void T359f116(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_other_precursors */
extern void T360f104(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_first_precursor */
extern void T1516f119(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_first_precursor */
extern void T476f109(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_first_precursor */
extern void T1515f115(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_first_precursor */
extern void T469f112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_first_precursor */
extern void T253f122(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_first_precursor */
extern void T1513f112(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_first_precursor */
extern void T1512f118(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_first_precursor */
extern void T1511f115(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_first_precursor */
extern void T371f110(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_first_precursor */
extern void T369f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_first_precursor */
extern void T359f115(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_first_precursor */
extern void T360f103(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_implementation_class */
extern void T1516f118(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_implementation_class */
extern void T476f108(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_implementation_class */
extern void T1515f114(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_implementation_class */
extern void T469f111(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_implementation_class */
extern void T253f121(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_implementation_class */
extern void T1513f111(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_implementation_class */
extern void T1512f117(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_implementation_class */
extern void T1511f114(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_implementation_class */
extern void T371f109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_implementation_class */
extern void T369f106(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_implementation_class */
extern void T359f114(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_implementation_class */
extern void T360f102(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_implementation_feature */
extern void T1516f117(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_implementation_feature */
extern void T476f107(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_implementation_feature */
extern void T1515f113(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_implementation_feature */
extern void T469f110(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_implementation_feature */
extern void T253f120(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_implementation_feature */
extern void T1513f110(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_implementation_feature */
extern void T1512f116(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_implementation_feature */
extern void T1511f113(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_implementation_feature */
extern void T371f108(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_implementation_feature */
extern void T369f105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_implementation_feature */
extern void T359f113(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_implementation_feature */
extern void T360f101(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.reset_postconditions */
extern void T1516f116(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.reset_postconditions */
extern void T476f106(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.reset_postconditions */
extern void T1515f112(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.reset_postconditions */
extern void T469f109(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.reset_postconditions */
extern void T253f119(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.reset_postconditions */
extern void T1513f109(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.reset_postconditions */
extern void T1512f115(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.reset_postconditions */
extern void T1511f112(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.reset_postconditions */
extern void T371f107(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.reset_postconditions */
extern void T369f104(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.reset_postconditions */
extern void T359f112(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.reset_postconditions */
extern void T360f100(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.reset_preconditions */
extern void T1516f115(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.reset_preconditions */
extern void T476f105(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.reset_preconditions */
extern void T1515f111(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.reset_preconditions */
extern void T469f108(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.reset_preconditions */
extern void T253f118(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.reset_preconditions */
extern void T1513f108(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.reset_preconditions */
extern void T1512f114(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.reset_preconditions */
extern void T1511f111(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.reset_preconditions */
extern void T371f106(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.reset_preconditions */
extern void T369f103(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.reset_preconditions */
extern void T359f111(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.reset_preconditions */
extern void T360f99(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.undefined_feature */
extern T0* T1516f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.undefined_feature */
extern T0* T476f64(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.undefined_feature */
extern T0* T1515f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.undefined_feature */
extern T0* T469f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.undefined_feature */
extern T0* T253f74(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.undefined_feature */
extern T0* T1513f65(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.undefined_feature */
extern T0* T1512f69(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.undefined_feature */
extern T0* T1511f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.undefined_feature */
extern T0* T371f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.postconditions */
extern T0* T371f26(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.preconditions */
extern T0* T371f55(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.obsolete_message */
extern T0* T371f71(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.undefined_feature */
extern T0* T369f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.postconditions */
extern T0* T369f25(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.preconditions */
extern T0* T369f54(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.obsolete_message */
extern T0* T369f70(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.undefined_feature */
extern T0* T359f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.undefined_feature */
extern T0* T360f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.postconditions */
extern T0* T360f24(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.preconditions */
extern T0* T360f52(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.obsolete_message */
extern T0* T360f67(GE_context* ac, T0* C);
/* detachable ET_ALIASED_FEATURE_NAME.make */
extern T0* T1517c6(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_ALIAS_NAME_LIST.is_empty */
extern T1 T911f7(GE_context* ac, T0* C);
/* detachable ET_PARENT_FEATURE.alias_names */
extern T0* T1344f20(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].wipe_out */
extern void T1328f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].slots_wipe_out */
extern void T1328f60(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].clashes_wipe_out */
extern void T1328f59(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].key_storage_wipe_out */
extern void T1328f58(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].item_storage_wipe_out */
extern void T1328f57(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_FEATURE].keep_head */
extern void T677f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_FEATURE].set_count */
extern void T677f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].unset_found_item */
extern void T1328f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].move_all_cursors_after */
extern void T1328f56(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE, [attached] INTEGER_32].set_next_cursor */
extern void T1698f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE, [attached] INTEGER_32].set_after */
extern void T1698f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].forth */
extern void T1328f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].cursor_forth */
extern void T1328f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].add_traversing_cursor */
extern void T1328f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE, [attached] INTEGER_32].set_position */
extern void T1698f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].remove_traversing_cursor */
extern void T1328f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].clashes_item */
extern T6 T1328f23(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FEATURE_LIST.put_first */
extern void T408f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE].force */
extern void T678f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FEATURE].fill_with */
extern void T677f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FEATURE].extend */
extern void T677f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FEATURE_LIST.fixed_array */
extern T0* T408f4(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE].default_create */
extern T0* T678c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].item_for_iteration */
extern T0* T1328f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].cursor_item */
extern T0* T1328f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].item_storage_item */
extern T0* T1328f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].after */
extern T1 T1328f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].cursor_after */
extern T1 T1328f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE, [attached] INTEGER_32].after */
extern T1 T1698f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].start */
extern void T1328f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].cursor_start */
extern void T1328f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].cursor_off */
extern T1 T1328f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE, [attached] INTEGER_32].off */
extern T1 T1698f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].is_empty */
extern T1 T1328f33(GE_context* ac, T0* C);
/* detachable ET_FEATURE_LIST.make_with_capacity */
extern T0* T408c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE].make */
extern T0* T678f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_FEATURE].make_empty */
extern T0* T677c5(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].remove */
extern void T1328f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].remove_position */
extern void T1328f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].key_storage_put */
extern void T1328f52(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].item_storage_put */
extern void T1328f48(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].key_storage_keep_head */
extern void T1328f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].item_storage_keep_head */
extern void T1328f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].clashes_put */
extern void T1328f50(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].slots_put */
extern void T1328f51(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].move_cursors_forth */
extern void T1328f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].internal_set_key_equality_tester */
extern void T1328f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].slots_item */
extern T6 T1328f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].hash_position */
extern T6 T1328f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].key_storage_item */
extern T6 T1328f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].search_position */
extern void T1328f47(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CLIENTS.same_clients */
extern T1 T1065f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLIENT_LIST.same_clients */
extern T1 T402f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENT_FEATURE.clients */
extern T0* T1344f18(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.clients */
extern T0* T1516f19(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.clients */
extern T0* T476f11(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.clients */
extern T0* T1515f20(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.clients */
extern T0* T469f11(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.clients */
extern T0* T253f21(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.clients */
extern T0* T1513f17(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.clients */
extern T0* T1512f22(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.clients */
extern T0* T1511f20(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.clients */
extern T0* T371f15(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.clients */
extern T0* T369f14(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.clients */
extern T0* T359f19(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.clients */
extern T0* T360f13(GE_context* ac, T0* C);
/* detachable ET_PARENT_FEATURE.other_seeds */
extern T0* T1344f12(GE_context* ac, T0* C);
/* detachable ET_PARENT_FEATURE.first_seed */
extern T6 T1344f11(GE_context* ac, T0* C);
/* detachable ET_PARENT_FEATURE.has_rename */
extern T1 T1344f17(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vmfn0c_error */
extern void T53f313(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vmfn0c */
extern T0* T785c946(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vmfn_error */
extern T1 T53f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vmfn0c_error */
extern void T36f313(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vmfn_error */
extern T1 T36f61(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENT_FEATURE.same_version */
extern T1 T1344f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].force_last */
extern void T1328f40(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].resize */
extern void T1328f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].clashes_resize */
extern void T1328f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].special_integer_ */
extern T0* T1328f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].key_storage_resize */
extern void T1328f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].item_storage_resize */
extern void T1328f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE].aliased_resized_area */
extern T0* T678f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_FEATURE].aliased_resized_area */
extern T0* T677f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].slots_resize */
extern void T1328f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].new_modulus */
extern T6 T1328f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].new_capacity */
extern T6 T1328f22(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_PROCEDURE.version */
extern T6 T1516f29(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.version */
extern T6 T476f26(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.version */
extern T6 T1515f29(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.version */
extern T6 T469f28(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.version */
extern T6 T253f31(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.version */
extern T6 T1513f28(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.version */
extern T6 T1512f31(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.version */
extern T6 T1511f30(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.version */
extern T6 T371f24(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.version */
extern T6 T369f23(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.version */
extern T6 T359f28(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.version */
extern T6 T360f21(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.check_no_redeclaration_validity */
extern void T459f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrs4a_error */
extern void T53f328(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vdrs4a */
extern T0* T785c962(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.feature_lower_name */
extern T0* T785f208(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdrs4_error */
extern T1 T53f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdrs4a_error */
extern void T36f328(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vdrs4_error */
extern T1 T36f66(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENT_FEATURE.has_redefine */
extern T1 T1344f15(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.check_redefine_clause_validity */
extern void T459f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrs2b_error */
extern void T53f337(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrs2b */
extern T0* T785c971(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdrs2_error */
extern T1 T53f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdrs2b_error */
extern void T36f337(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdrs2_error */
extern T1 T36f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrs2a_error */
extern void T53f336(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrs2a */
extern T0* T785c970(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdrs2a_error */
extern void T36f336(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_FLATTENER.check_undefine_clause_validity */
extern void T459f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdus2b_error */
extern void T53f335(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdus2b */
extern T0* T785c969(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdus2_error */
extern T1 T53f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdus2b_error */
extern void T36f335(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdus2_error */
extern T1 T36f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdus2a_error */
extern void T53f334(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdus2a */
extern T0* T785c968(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdus2a_error */
extern void T36f334(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdus3a_error */
extern void T53f333(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdus3a */
extern T0* T785c967(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdus3_error */
extern T1 T53f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdus3a_error */
extern void T36f333(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdus3_error */
extern T1 T36f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.check_rename_clause_validity */
extern void T459f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1p_error */
extern void T53f238(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1p */
extern T0* T785c888(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.upper_name */
extern T0* T254f108(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_vfav1p_error */
extern void T36f238(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1n_error */
extern void T53f236(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1n */
extern T0* T785c886(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vfav1n_error */
extern void T36f236(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav4h_error */
extern void T53f332(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vfav4h */
extern T0* T785c966(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vfav4h_error */
extern void T36f332(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_EXTERNAL_PROCEDURE.is_infixable */
extern T1 T1516f74(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_infixable */
extern T1 T476f63(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_infixable */
extern T1 T1515f67(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_infixable */
extern T1 T469f65(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1m_error */
extern void T53f233(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1m */
extern T0* T785c883(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vfav1m_error */
extern void T36f233(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav4g_error */
extern void T53f331(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vfav4g */
extern T0* T785c965(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vfav4g_error */
extern void T36f331(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_EXTERNAL_PROCEDURE.is_prefixable */
extern T1 T1516f73(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_prefixable */
extern T1 T476f62(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_prefixable */
extern T1 T1515f66(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_prefixable */
extern T1 T469f64(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav3e_error */
extern void T53f230(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vfav3e */
extern T0* T785c880(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vfav3e_error */
extern void T36f230(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav4f_error */
extern void T53f330(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vfav4f */
extern T0* T785c964(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vfav4f_error */
extern void T36f330(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav2e_error */
extern void T53f227(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vfav2e */
extern T0* T785c877(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vfav2e_error */
extern void T36f227(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_EXTERNAL_PROCEDURE.is_bracketable */
extern T1 T1516f72(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_bracketable */
extern T1 T476f61(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_bracketable */
extern T1 T1515f65(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_bracketable */
extern T1 T469f63(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav4e_error */
extern void T53f329(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vfav4e */
extern T0* T785c963(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vfav4e_error */
extern void T36f329(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_INHERITED_FEATURE.is_replicated */
extern T1 T1345f20(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.inherited_clients */
extern T0* T459f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].wipe_out */
extern void T1316f14(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_CLIENT_LIST].keep_head */
extern void T1674f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CLIENT_LIST].set_count */
extern void T1674f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].move_all_cursors_after */
extern void T1316f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLIENT_LIST].set_next_cursor */
extern void T1675f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLIENT_LIST].set_after */
extern void T1675f6(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.aggregated_clients */
extern T0* T459f31(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_FEATURE_FLATTENER.aggregated_clients */
extern T0* T459f31ac1(GE_context* ac, T0* a1);
/* [detachable] PREDICATE [[attached] TUPLE [[attached] ET_CLASS]].set_rout_disp_final */
extern void T870f8(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].wipe_out */
extern void T1317f53(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].slots_wipe_out */
extern void T1317f66(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].clashes_wipe_out */
extern void T1317f65(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].key_storage_wipe_out */
extern void T1317f64(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_CLASS].keep_head */
extern void T1346f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CLASS].set_count */
extern void T1346f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].item_storage_wipe_out */
extern void T1317f63(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_CLIENT].keep_head */
extern void T1679f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CLIENT].set_count */
extern void T1679f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].unset_found_item */
extern void T1317f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].move_all_cursors_after */
extern void T1317f62(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].set_next_cursor */
extern void T1677f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].set_after */
extern void T1677f13(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].back */
extern void T1317f52(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_back */
extern void T1317f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].add_traversing_cursor */
extern void T1317f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].set_position */
extern void T1677f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].remove_traversing_cursor */
extern void T1317f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].set_before */
extern void T1677f11(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].clashes_item */
extern T6 T1317f31(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CLIENTS.put_first */
extern void T1065f18(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLIENTS.fixed_array */
extern T0* T1065f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].item_for_iteration */
extern T0* T1317f22(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_item */
extern T0* T1317f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].item_storage_item */
extern T0* T1317f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].before */
extern T1 T1317f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_before */
extern T1 T1317f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].before */
extern T1 T1677f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].finish */
extern void T1317f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_finish */
extern void T1317f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_off */
extern T1 T1317f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].off */
extern T1 T1677f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].is_empty */
extern T1 T1317f36(GE_context* ac, T0* C);
/* detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS].for_all */
extern T1 T1853f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].forth */
extern void T1990f8(GE_context* ac, T0* C);
/* detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_forth */
extern void T1853f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].forth */
extern void T1677f17(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_forth */
extern void T1317f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].go_after */
extern void T1990f7(GE_context* ac, T0* C);
/* detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_go_after */
extern void T1853f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].go_after */
extern void T1677f16(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_go_after */
extern void T1317f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].item */
extern T0* T1990f4(GE_context* ac, T0* C);
/* detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_item */
extern T0* T1853f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].key */
extern T0* T1677f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_key */
extern T0* T1317f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].key_storage_item */
extern T0* T1317f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].after */
extern T1 T1990f3(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].after */
extern T1 T1677f8(GE_context* ac, T0* C);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].start */
extern void T1990f6(GE_context* ac, T0* C);
/* detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_start */
extern void T1853f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].start */
extern void T1677f15(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].cursor_start */
extern void T1317f73(GE_context* ac, T0* C, T0* a1);
/* detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS].new_cursor */
extern T0* T1853f4(GE_context* ac, T0* C);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].make */
extern T0* T1990c5(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].new_cursor */
extern T0* T1317f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].make */
extern T0* T1677c10(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].keys */
extern T0* T1317f39(GE_context* ac, T0* C);
/* detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS].make */
extern T0* T1853c8(GE_context* ac, T0* a1);
/* detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS].set_internal_cursor */
extern void T1853f9(GE_context* ac, T0* C, T0* a1);
/* detachable DS_SPARSE_TABLE_KEYS [[attached] ET_CLIENT, [attached] ET_CLASS].new_cursor_with_table_cursor */
extern T0* T1853f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [[attached] ET_CLIENT, [attached] ET_CLASS].make_with_table_cursor */
extern T0* T1990c9(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].force */
extern void T1317f50(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].key_storage_put */
extern void T1317f59(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].slots_put */
extern void T1317f58(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].clashes_put */
extern void T1317f57(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].slots_item */
extern T6 T1317f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].hash_position */
extern T6 T1317f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].resize */
extern void T1317f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].clashes_resize */
extern void T1317f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].special_integer_ */
extern T0* T1317f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].key_storage_resize */
extern void T1317f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].item_storage_resize */
extern void T1317f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT].aliased_resized_area */
extern T0* T1678f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_CLIENT].aliased_resized_area */
extern T0* T1679f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].slots_resize */
extern void T1317f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].new_modulus */
extern T6 T1317f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].new_capacity */
extern T6 T1317f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].item_storage_put */
extern void T1317f55(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT].force */
extern void T1678f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLIENT].fill_with */
extern void T1679f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLIENT].extend */
extern void T1679f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].search_position */
extern void T1317f54(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLIENTS.client */
extern T0* T1065f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLIENT.client */
extern T0* T403f4(GE_context* ac, T0* C);
/* detachable ET_CLIENTS.item */
extern T0* T1065f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLIENT_LIST.client */
extern T0* T402f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLIENT_LIST.item */
extern T0* T402f9(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CLIENTS.is_none */
extern T1 T1065f7(GE_context* ac, T0* C);
/* [detachable] ET_CLIENT_LIST.is_none */
extern T1 T402f4(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].item */
extern T0* T1316f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].first */
extern T0* T1316f7(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.add_inherited_clients_to_list */
extern void T459f61(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_FEATURE_EXPORT.clients */
extern T0* T1145f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALL_EXPORT.clients */
extern T0* T1506f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FEATURE_EXPORT.has_feature_name */
extern T1 T1145f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALL_EXPORT.has_feature_name */
extern T1 T1506f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FEATURE_EXPORT.is_all */
extern T1 T1145f6(GE_context* ac, T0* C);
/* detachable ET_EXPORT_LIST.item */
extern T0* T573f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].force_last */
extern void T1316f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_LIST].force */
extern void T1673f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLIENT_LIST].fill_with */
extern void T1674f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLIENT_LIST].extend */
extern void T1674f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].resize */
extern void T1316f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_LIST].aliased_resized_area */
extern T0* T1673f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_CLIENT_LIST].aliased_resized_area */
extern T0* T1674f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].new_capacity */
extern T6 T1316f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].extendible */
extern T1 T1316f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_FLATTENER.flatten_redeclared_feature */
extern void T459f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.check_precursor_validity */
extern void T459f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.check_feature_validity */
extern void T1327f11(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_EXTERNAL_PROCEDURE.process */
extern void T1516f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_external_procedure */
extern void T1327f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_external_procedure */
extern void T1326f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_external_feature_validity */
extern void T1326f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_feature_validity */
extern void T1326f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvkbu1a_error */
extern void T53f533(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_gvkbu1a */
extern T0* T785c1155(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvkbu1_error */
extern T1 T53f161(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvkbu1a_error */
extern void T36f533(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_gvkbu1_error */
extern T1 T36f161(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.error_handler */
extern T0* T1326f9(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.set_fatal_error */
extern void T1326f48(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.unknown_builtin_reported */
extern T1 T1326f8(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.real_64_type */
extern T0* T187f74(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.real_64_type */
extern T0* T58f25(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.real_64_type */
extern T0* T40f81(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.real_64_ref_class_name */
extern T0* T73f940(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_real_n_feature_validity */
extern void T1326f47(GE_context* ac, T0* C, T0* a1, T8 a2);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_expected_builtin_feature_validity */
extern void T1326f50(GE_context* ac, T0* C, T0* a1, T8 a2, T0* a3);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_signature_validity */
extern void T1326f51(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvkbs0a_error */
extern void T53f534(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_gvkbs0a */
extern T0* T785c1156(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_LIKE_N.to_text */
extern T0* T1474f12(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_N.append_to_string */
extern void T1474f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvkbs_error */
extern T1 T53f162(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvkbs0a_error */
extern void T36f534(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_gvkbs_error */
extern T1 T36f162(GE_context* ac, T0* C, T0* a1);
/* [detachable] ARRAY [[attached] ET_TYPE].item */
extern T0* T244f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [[attached] ET_TYPE].count */
extern T6 T244f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].found_item */
extern T0* T1937f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].item_storage_item */
extern T0* T1937f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].found */
extern T1 T1937f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].search */
extern void T1937f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].search_position */
extern void T1937f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].hash_position */
extern T6 T1937f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].key_storage_item */
extern T0* T1937f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].clashes_item */
extern T6 T1937f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].slots_item */
extern T6 T1937f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.truncated_to_real_feature_name */
extern T0* T73f1524(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.truncated_to_integer_64_feature_name */
extern T0* T73f1522(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.truncated_to_integer_feature_name */
extern T0* T73f1520(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.to_double_feature_name */
extern T0* T73f1467(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.quotient_feature_name */
extern T0* T73f1463(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.product_feature_name */
extern T0* T73f1461(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.power_feature_name */
extern T0* T73f1459(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.plus_feature_name */
extern T0* T73f1039(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.string_type */
extern T0* T187f80(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.string_type */
extern T0* T58f79(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.string_type */
extern T0* T40f87(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.out_feature_name */
extern T0* T73f1037(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.opposite_feature_name */
extern T0* T73f1456(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.minus_feature_name */
extern T0* T73f1454(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_positive_infinity_feature_name */
extern T0* T73f1510(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_negative_infinity_feature_name */
extern T0* T73f1508(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_nan_feature_name */
extern T0* T73f1506(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_less_feature_name */
extern T0* T73f1452(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ieee_minimum_number_feature_name */
extern T0* T73f1503(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ieee_maximum_number_feature_name */
extern T0* T73f1501(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ieee_is_less_equal_feature_name */
extern T0* T73f1499(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ieee_is_less_feature_name */
extern T0* T73f1497(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ieee_is_greater_equal_feature_name */
extern T0* T73f1495(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ieee_is_greater_feature_name */
extern T0* T73f1493(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.boolean_type */
extern T0* T187f33(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.boolean_type */
extern T0* T58f13(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.boolean_type */
extern T0* T40f2(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ieee_is_equal_feature_name */
extern T0* T73f1491(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.identity_feature_name */
extern T0* T73f1446(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.floor_real_64_feature_name */
extern T0* T73f1488(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.floor_real_32_feature_name */
extern T0* T73f1486(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ceiling_real_64_feature_name */
extern T0* T73f1484(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.register_builtin_feature */
extern void T1326f49(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T8 a4, T0* a5);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].force_last */
extern void T1937f36(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].key_storage_put */
extern void T1937f45(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME].force */
extern void T1668f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FEATURE_NAME].fill_with */
extern void T1666f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FEATURE_NAME].extend */
extern void T1666f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].slots_put */
extern void T1937f44(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].clashes_put */
extern void T1937f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].resize */
extern void T1937f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].clashes_resize */
extern void T1937f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].special_integer_ */
extern T0* T1937f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].key_storage_resize */
extern void T1937f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME].aliased_resized_area */
extern T0* T1668f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_FEATURE_NAME].aliased_resized_area */
extern T0* T1666f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].item_storage_resize */
extern void T1937f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]].aliased_resized_area */
extern T0* T1997f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]].aliased_resized_area */
extern T0* T1994f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].slots_resize */
extern void T1937f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].new_modulus */
extern T6 T1937f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].new_capacity */
extern T6 T1937f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].item_storage_put */
extern void T1937f41(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]].force */
extern void T1997f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]].fill_with */
extern void T1994f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]].extend */
extern void T1994f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].unset_found_item */
extern void T1937f39(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ceiling_real_32_feature_name */
extern T0* T73f1482(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].force_last */
extern void T1697f43(GE_context* ac, T0* C, T0* a1, T8 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].key_storage_put */
extern void T1697f49(GE_context* ac, T0* C, T8 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].slots_put */
extern void T1697f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].clashes_put */
extern void T1697f47(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].slots_item */
extern T6 T1697f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].hash_position */
extern T6 T1697f23(GE_context* ac, T0* C, T8 a1);
/* NATURAL_8.hash_code */
extern T6 T8f22(GE_context* ac, T8* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].resize */
extern void T1697f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].clashes_resize */
extern void T1697f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].special_integer_ */
extern T0* T1697f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].key_storage_resize */
extern void T1697f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].item_storage_resize */
extern void T1697f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]].aliased_resized_area */
extern T0* T1936f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]].aliased_resized_area */
extern T0* T1938f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].key_storage_item */
extern T8 T1697f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].clashes_item */
extern T6 T1697f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].slots_resize */
extern void T1697f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].new_modulus */
extern T6 T1697f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].new_capacity */
extern T6 T1697f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].item_storage_put */
extern void T1697f45(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]].force */
extern void T1936f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]].fill_with */
extern void T1938f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]].extend */
extern void T1938f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].search_position */
extern void T1697f44(GE_context* ac, T0* C, T8 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].unset_found_item */
extern void T1697f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].set_key_equality_tester */
extern void T1937f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.feature_name_tester */
extern T0* T1326f10(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_NAME_TESTER.default_create */
extern T0* T1311c2(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].make_map */
extern T0* T1937c34(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].make_with_equality_testers */
extern void T1937f38(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].make_sparse_container */
extern void T1937f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].set_internal_cursor */
extern void T1937f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].new_cursor */
extern T0* T1937f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].make */
extern T0* T1998c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].make_slots */
extern void T1937f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].make_clashes */
extern void T1937f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].make_key_storage */
extern void T1937f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME].make */
extern T0* T1668f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_FEATURE_NAME].make_empty */
extern T0* T1666c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME].default_create */
extern T0* T1668c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME].make_item_storage */
extern void T1937f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]].make */
extern T0* T1997f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]].make_empty */
extern T0* T1994c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8]].default_create */
extern T0* T1997c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].found_item */
extern T0* T1697f20(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].item_storage_item */
extern T0* T1697f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].found */
extern T1 T1697f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].search */
extern void T1697f42(GE_context* ac, T0* C, T8 a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_real_n_ref_feature_validity */
extern void T1326f46(GE_context* ac, T0* C, T0* a1, T0* a2, T8 a3);
/* [detachable] ET_TOKEN_CONSTANTS.set_item_feature_name */
extern T0* T73f1005(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.positive_infinity_feature_name */
extern T0* T73f1479(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.negative_infinity_feature_name */
extern T0* T73f1477(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.nan_feature_name */
extern T0* T73f1475(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.real_32_type */
extern T0* T187f72(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.real_32_type */
extern T0* T58f24(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.real_32_type */
extern T0* T40f79(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.real_32_ref_class_name */
extern T0* T73f937(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.natural_64_type */
extern T0* T187f64(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.natural_64_type */
extern T0* T58f23(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.natural_64_type */
extern T0* T40f72(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_64_ref_class_name */
extern T0* T73f934(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.natural_32_type */
extern T0* T187f62(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.natural_32_type */
extern T0* T58f22(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.natural_32_type */
extern T0* T40f70(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_32_ref_class_name */
extern T0* T73f931(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.natural_16_type */
extern T0* T187f60(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.natural_16_type */
extern T0* T58f21(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.natural_16_type */
extern T0* T40f68(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_16_ref_class_name */
extern T0* T73f928(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.natural_8_type */
extern T0* T187f58(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.natural_8_type */
extern T0* T58f20(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.natural_8_type */
extern T0* T40f66(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_8_ref_class_name */
extern T0* T73f925(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.integer_64_type */
extern T0* T187f53(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.integer_64_type */
extern T0* T58f19(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.integer_64_type */
extern T0* T40f61(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_64_ref_class_name */
extern T0* T73f922(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.integer_32_type */
extern T0* T187f51(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.integer_32_type */
extern T0* T58f18(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.integer_32_type */
extern T0* T40f59(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_32_ref_class_name */
extern T0* T73f919(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.integer_16_type */
extern T0* T187f49(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.integer_16_type */
extern T0* T58f17(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.integer_16_type */
extern T0* T40f57(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_16_ref_class_name */
extern T0* T73f916(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_integer_n_feature_validity */
extern void T1326f45(GE_context* ac, T0* C, T0* a1, T8 a2);
/* [detachable] ET_TOKEN_CONSTANTS.to_real_64_feature_name */
extern T0* T73f1473(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.to_real_32_feature_name */
extern T0* T73f1471(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.to_real_feature_name */
extern T0* T73f1469(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.to_character_32_feature_name */
extern T0* T73f1013(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.to_character_8_feature_name */
extern T0* T73f1011(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_remainder_feature_name */
extern T0* T73f1450(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_quotient_feature_name */
extern T0* T73f1448(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.bit_xor_feature_name */
extern T0* T73f1444(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.bit_shift_right_feature_name */
extern T0* T73f1442(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.integer_type */
extern T0* T187f46(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.integer_type */
extern T0* T58f49(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.integer_type */
extern T0* T40f5(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.bit_shift_left_feature_name */
extern T0* T73f1440(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.bit_or_feature_name */
extern T0* T73f1438(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.bit_not_feature_name */
extern T0* T73f1436(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.bit_and_feature_name */
extern T0* T73f1434(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.as_natural_64_feature_name */
extern T0* T73f1432(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.as_natural_32_feature_name */
extern T0* T73f1430(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.as_natural_16_feature_name */
extern T0* T73f1428(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.as_natural_8_feature_name */
extern T0* T73f1426(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.as_integer_64_feature_name */
extern T0* T73f1424(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.as_integer_32_feature_name */
extern T0* T73f1422(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.as_integer_16_feature_name */
extern T0* T73f1420(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.as_integer_8_feature_name */
extern T0* T73f1418(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_integer_n_ref_feature_validity */
extern void T1326f44(GE_context* ac, T0* C, T0* a1, T0* a2, T8 a3);
/* [detachable] ET_ECF_LIBRARY.integer_8_type */
extern T0* T187f47(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.integer_8_type */
extern T0* T58f16(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.integer_8_type */
extern T0* T40f55(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_8_ref_class_name */
extern T0* T73f913(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_tuple_feature_validity */
extern void T1326f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.set_object_comparison_feature_name */
extern T0* T73f1415(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.double_type */
extern T0* T187f39(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.double_type */
extern T0* T58f52(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.double_type */
extern T0* T40f48(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_real_64_feature_name */
extern T0* T73f1412(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.real_type */
extern T0* T187f71(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.real_type */
extern T0* T58f51(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.real_type */
extern T0* T40f78(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_real_32_feature_name */
extern T0* T73f1410(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_pointer_feature_name */
extern T0* T73f1408(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_natural_64_feature_name */
extern T0* T73f1406(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_natural_32_feature_name */
extern T0* T73f1404(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_natural_16_feature_name */
extern T0* T73f1402(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_natural_8_feature_name */
extern T0* T73f1400(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_integer_64_feature_name */
extern T0* T73f1398(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_integer_32_feature_name */
extern T0* T73f1396(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_integer_16_feature_name */
extern T0* T73f1394(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_integer_8_feature_name */
extern T0* T73f1392(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_character_32_feature_name */
extern T0* T73f1390(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_character_8_feature_name */
extern T0* T73f1388(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_boolean_feature_name */
extern T0* T73f1386(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.detachable_separate_any_type */
extern T0* T187f21(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.detachable_separate_any_type */
extern T0* T58f85(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.detachable_separate_any_type */
extern T0* T40f35(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.real_64_item_feature_name */
extern T0* T73f1383(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.real_32_item_feature_name */
extern T0* T73f1381(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.pointer_type */
extern T0* T187f68(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.pointer_type */
extern T0* T58f26(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.pointer_type */
extern T0* T40f8(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.pointer_item_feature_name */
extern T0* T73f10(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.object_comparison_feature_name */
extern T0* T73f1379(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_64_item_feature_name */
extern T0* T73f1377(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_32_item_feature_name */
extern T0* T73f1375(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_16_item_feature_name */
extern T0* T73f1373(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_8_item_feature_name */
extern T0* T73f1371(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.item_code_feature_name */
extern T0* T73f1369(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_64_item_feature_name */
extern T0* T73f1367(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_32_item_feature_name */
extern T0* T73f1365(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_16_item_feature_name */
extern T0* T73f1363(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_8_item_feature_name */
extern T0* T73f1361(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.count_feature_name */
extern T0* T73f4(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_32_item_feature_name */
extern T0* T73f1358(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_8_item_feature_name */
extern T0* T73f1356(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.boolean_item_feature_name */
extern T0* T73f1354(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_function_feature_validity */
extern void T1326f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.detachable_separate_formal_parameter_type */
extern T0* T1326f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.name */
extern T0* T1501f8(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.name */
extern T0* T891f1(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.name */
extern T0* T242f2(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.fast_item_feature_name */
extern T0* T73f1352(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.formal_parameter_type */
extern T0* T74f103(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [detachable ET_FORMAL_PARAMETER_TYPE].replace */
extern void T890f11(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable DS_ARRAYED_LIST [detachable ET_FORMAL_PARAMETER_TYPE].item */
extern T0* T890f7(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [detachable ET_FORMAL_PARAMETER_TYPE].put_last */
extern void T890f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_TYPE].force */
extern void T1622f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_TYPE].fill_with */
extern void T1621f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_TYPE].extend */
extern void T1621f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [detachable ET_FORMAL_PARAMETER_TYPE].resize */
extern void T890f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_TYPE].aliased_resized_area */
extern T0* T1622f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_TYPE].aliased_resized_area */
extern T0* T1621f4(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [detachable ET_FORMAL_PARAMETER_TYPE].make */
extern T0* T890c8(GE_context* ac, T6 a1);
/* detachable DS_ARRAYED_LIST [detachable ET_FORMAL_PARAMETER_TYPE].set_internal_cursor */
extern void T890f12(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [detachable ET_FORMAL_PARAMETER_TYPE].new_cursor */
extern T0* T890f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_FORMAL_PARAMETER_TYPE].make */
extern T0* T1623c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_TYPE].make */
extern T0* T1622f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_TYPE].make_empty */
extern T0* T1621c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_TYPE].default_create */
extern T0* T1622c3(GE_context* ac);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_procedure_feature_validity */
extern void T1326f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.fast_call_feature_name */
extern T0* T73f1350(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_platform_feature_validity */
extern void T1326f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.wide_character_bytes_feature_name */
extern T0* T73f1348(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.real_bytes_feature_name */
extern T0* T73f1346(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.pointer_bytes_feature_name */
extern T0* T73f1344(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_windows_feature_name */
extern T0* T73f1342(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_vxworks_feature_name */
extern T0* T73f1340(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_vms_feature_name */
extern T0* T73f1338(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_unix_feature_name */
extern T0* T73f1336(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_thread_capable_feature_name */
extern T0* T73f1334(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_scoop_capable_feature_name */
extern T0* T73f1332(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_mac_feature_name */
extern T0* T73f1330(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_dotnet_feature_name */
extern T0* T73f1328(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_64_bits_feature_name */
extern T0* T73f1326(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_bytes_feature_name */
extern T0* T73f1324(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.double_bytes_feature_name */
extern T0* T73f1322(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_bytes_feature_name */
extern T0* T73f1320(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.boolean_bytes_feature_name */
extern T0* T73f1318(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.platform_class_name */
extern T0* T73f912(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_ise_runtime_feature_validity */
extern void T1326f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.unmark_object_feature_name */
extern T0* T73f1315(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.unlock_marking_feature_name */
extern T0* T73f1313(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_reference_field_at_feature_name */
extern T0* T73f1312(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_reference_field_feature_name */
extern T0* T73f1311(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_real_64_field_at_feature_name */
extern T0* T73f1309(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_real_64_field_feature_name */
extern T0* T73f1307(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_real_32_field_at_feature_name */
extern T0* T73f1305(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_real_32_field_feature_name */
extern T0* T73f1303(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_pre_ecma_mapping_feature_name */
extern T0* T73f1301(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_pointer_field_at_feature_name */
extern T0* T73f1299(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_pointer_field_feature_name */
extern T0* T73f1297(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_natural_64_field_at_feature_name */
extern T0* T73f1295(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_natural_64_field_feature_name */
extern T0* T73f1293(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_natural_32_field_at_feature_name */
extern T0* T73f1291(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_natural_32_field_feature_name */
extern T0* T73f1289(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_natural_16_field_at_feature_name */
extern T0* T73f1287(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_natural_16_field_feature_name */
extern T0* T73f1285(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_natural_8_field_at_feature_name */
extern T0* T73f1283(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_natural_8_field_feature_name */
extern T0* T73f1281(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_integer_64_field_at_feature_name */
extern T0* T73f1279(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_integer_64_field_feature_name */
extern T0* T73f1277(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_integer_32_field_at_feature_name */
extern T0* T73f1275(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_integer_32_field_feature_name */
extern T0* T73f1273(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_integer_16_field_at_feature_name */
extern T0* T73f1271(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_integer_16_field_feature_name */
extern T0* T73f1269(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_integer_8_field_at_feature_name */
extern T0* T73f1267(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_integer_8_field_feature_name */
extern T0* T73f1265(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_character_32_field_at_feature_name */
extern T0* T73f1263(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_character_32_field_feature_name */
extern T0* T73f1261(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_character_8_field_at_feature_name */
extern T0* T73f1259(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_character_8_field_feature_name */
extern T0* T73f1257(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_boolean_field_at_feature_name */
extern T0* T73f1255(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_boolean_field_feature_name */
extern T0* T73f1253(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.mark_object_feature_name */
extern T0* T73f1251(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.lock_marking_feature_name */
extern T0* T73f1249(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.type_id_from_name_feature_name */
extern T0* T73f1247(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.type_conforms_to_feature_name */
extern T0* T73f1245(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.detachable_string_8_type */
extern T0* T187f83(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.detachable_string_8_type */
extern T0* T58f113(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.detachable_string_8_type */
extern T0* T40f90(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.storable_version_of_type_feature_name */
extern T0* T73f1243(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.reference_field_at_offset_feature_name */
extern T0* T73f1242(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.reference_field_at_feature_name */
extern T0* T73f1241(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.reference_field_feature_name */
extern T0* T73f1240(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.real_64_field_at_feature_name */
extern T0* T73f1238(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.real_64_field_feature_name */
extern T0* T73f1236(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.real_32_field_at_feature_name */
extern T0* T73f1234(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.real_32_field_feature_name */
extern T0* T73f1232(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.raw_reference_field_at_offset_feature_name */
extern T0* T73f1230(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.raw_reference_field_at_feature_name */
extern T0* T73f1228(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.pre_ecma_mapping_status_feature_name */
extern T0* T73f1226(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.pointer_field_at_feature_name */
extern T0* T73f1224(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.pointer_field_feature_name */
extern T0* T73f1222(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.persistent_field_count_of_type_feature_name */
extern T0* T73f1220(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.special_any_type */
extern T0* T187f77(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.special_any_type */
extern T0* T58f27(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.special_any_type */
extern T0* T40f84(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.once_objects_feature_name */
extern T0* T73f1218(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.object_size_feature_name */
extern T0* T73f1216(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.type_detachable_any_type */
extern T0* T187f90(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.type_detachable_any_type */
extern T0* T58f119(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.type_detachable_any_type */
extern T0* T40f95(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.new_type_instance_of_feature_name */
extern T0* T73f1215(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.new_tuple_instance_of_feature_name */
extern T0* T73f1214(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.special_detachable_any_type */
extern T0* T187f78(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.special_detachable_any_type */
extern T0* T58f111(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.special_detachable_any_type */
extern T0* T40f85(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.new_special_of_reference_instance_of_feature_name */
extern T0* T73f1213(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.new_instance_of_feature_name */
extern T0* T73f1212(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_64_field_at_feature_name */
extern T0* T73f1210(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_64_field_feature_name */
extern T0* T73f1208(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_32_field_at_feature_name */
extern T0* T73f1206(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_32_field_feature_name */
extern T0* T73f1204(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_16_field_at_feature_name */
extern T0* T73f1202(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_16_field_feature_name */
extern T0* T73f1200(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_8_field_at_feature_name */
extern T0* T73f1198(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.natural_8_field_feature_name */
extern T0* T73f1196(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_tuple_type_feature_name */
extern T0* T73f1194(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_tuple_feature_name */
extern T0* T73f1192(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_special_of_reference_type_feature_name */
extern T0* T73f1190(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_special_of_reference_or_basic_type_feature_name */
extern T0* T73f1188(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_special_of_reference_feature_name */
extern T0* T73f1186(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_special_of_expanded_feature_name */
extern T0* T73f1184(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_special_copy_semantics_item_feature_name */
extern T0* T73f1182(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_special_feature_name */
extern T0* T73f1180(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_object_marked_feature_name */
extern T0* T73f1178(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_field_transient_of_type_feature_name */
extern T0* T73f1176(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_field_expanded_of_type_feature_name */
extern T0* T73f1174(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_expanded_feature_name */
extern T0* T73f981(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_copy_semantics_field_feature_name */
extern T0* T73f1171(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_attached_type_feature_name */
extern T0* T73f1169(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_64_field_at_feature_name */
extern T0* T73f1167(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_64_field_feature_name */
extern T0* T73f1165(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_32_field_at_feature_name */
extern T0* T73f1163(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_32_field_feature_name */
extern T0* T73f1161(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_16_field_at_feature_name */
extern T0* T73f1159(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_16_field_feature_name */
extern T0* T73f1157(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_8_field_at_feature_name */
extern T0* T73f1155(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.integer_8_field_feature_name */
extern T0* T73f1153(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.in_assertion_feature_name */
extern T0* T73f1151(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.generic_parameter_count_feature_name */
extern T0* T73f973(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.generator_8_of_type_feature_name */
extern T0* T73f1148(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.generator_of_type_feature_name */
extern T0* T73f1146(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.string_8_type */
extern T0* T187f82(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.string_8_type */
extern T0* T58f29(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.string_8_type */
extern T0* T40f89(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.generating_type_8_of_type_feature_name */
extern T0* T73f1144(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.generating_type_of_type_feature_name */
extern T0* T73f1142(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.field_type_of_type_feature_name */
extern T0* T73f1140(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.field_static_type_of_type_feature_name */
extern T0* T73f1138(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.field_offset_of_type_feature_name */
extern T0* T73f1136(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.field_name_of_type_feature_name */
extern T0* T73f1134(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.field_count_of_type_feature_name */
extern T0* T73f1132(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.eif_gen_param_id_feature_name */
extern T0* T73f1130(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.dynamic_type_at_offset_feature_name */
extern T0* T73f1128(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.any_type */
extern T0* T187f19(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.any_type */
extern T0* T58f10(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.any_type */
extern T0* T40f12(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.dynamic_type_feature_name */
extern T0* T73f1126(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.detachable_type_feature_name */
extern T0* T73f1124(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.compiler_version_feature_name */
extern T0* T73f1122(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.check_assert_feature_name */
extern T0* T73f1120(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_32_field_at_feature_name */
extern T0* T73f1118(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_32_field_feature_name */
extern T0* T73f1116(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_8_field_at_feature_name */
extern T0* T73f1114(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_8_field_feature_name */
extern T0* T73f1112(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.boolean_field_at_feature_name */
extern T0* T73f1110(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.boolean_field_feature_name */
extern T0* T73f1108(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.attached_type_feature_name */
extern T0* T73f1106(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ise_runtime_class_name */
extern T0* T73f911(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_identified_routines_feature_validity */
extern void T1326f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.eif_object_id_free_feature_name */
extern T0* T73f1104(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.eif_object_id_feature_name */
extern T0* T73f1103(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.eif_is_object_id_of_current_feature_name */
extern T0* T73f1101(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.eif_id_object_feature_name */
extern T0* T73f1100(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.eif_current_object_id_feature_name */
extern T0* T73f1098(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.identified_routines_class_name */
extern T0* T73f910(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_ise_exception_manager_feature_validity */
extern void T1326f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.developer_raise_feature_name */
extern T0* T73f1096(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_exception_manager_factory_feature_validity */
extern void T1326f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.exception_manager_type */
extern T0* T187f42(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.exception_manager_type */
extern T0* T58f96(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.exception_manager_type */
extern T0* T40f51(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.exception_manager_feature_name */
extern T0* T73f1093(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.exception_manager_factory_class_name */
extern T0* T73f909(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_exception_manager_feature_validity */
extern void T1326f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.set_is_ignored_feature_name */
extern T0* T73f1090(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.exception_type */
extern T0* T187f40(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.exception_type */
extern T0* T58f78(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.exception_type */
extern T0* T40f49(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.raise_feature_name */
extern T0* T73f1088(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ignore_feature_name */
extern T0* T73f1086(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.catch_feature_name */
extern T0* T73f1084(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.detachable_type_exception_type */
extern T0* T187f91(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.detachable_type_exception_type */
extern T0* T58f120(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.detachable_type_exception_type */
extern T0* T40f96(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.type_of_code_feature_name */
extern T0* T73f1082(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.last_exception_feature_name */
extern T0* T73f18(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_raisable_feature_name */
extern T0* T73f1079(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_ignored_feature_name */
extern T0* T73f1077(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_ignorable_feature_name */
extern T0* T73f1075(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.type_detachable_exception_type */
extern T0* T187f92(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.type_detachable_exception_type */
extern T0* T58f121(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.type_detachable_exception_type */
extern T0* T40f97(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_caught_feature_name */
extern T0* T73f1073(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.detachable_exception_type */
extern T0* T187f41(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.detachable_exception_type */
extern T0* T58f77(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.detachable_exception_type */
extern T0* T40f50(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.exception_from_code_feature_name */
extern T0* T73f1071(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_memory_feature_validity */
extern void T1326f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.free_feature_name */
extern T0* T73f1068(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.find_referers_feature_name */
extern T0* T73f1066(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.memory_class_name */
extern T0* T73f908(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_com_failure_feature_validity */
extern void T1326f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.cwin_local_free_feature_name */
extern T0* T73f1063(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.cwin_error_text_feature_name */
extern T0* T73f1061(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_size_feature_name */
extern T0* T73f1059(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ccom_hresult_facility_feature_name */
extern T0* T73f1057(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ccom_hresult_code_feature_name */
extern T0* T73f1055(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.ccom_hresult_feature_name */
extern T0* T73f1053(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.c_strlen_feature_name */
extern T0* T73f1051(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.com_failure_class_name */
extern T0* T73f907(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_arguments_32_feature_validity */
extern void T1326f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.immutable_string_32_type */
extern T0* T187f45(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.immutable_string_32_type */
extern T0* T58f32(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.immutable_string_32_type */
extern T0* T40f54(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.i_th_argument_string_feature_name */
extern T0* T73f1048(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.i_th_argument_pointer_feature_name */
extern T0* T73f1046(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.argument_count_feature_name */
extern T0* T73f1044(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.arguments_32_class_name */
extern T0* T73f906(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_pointer_feature_validity */
extern void T1326f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.to_integer_32_feature_name */
extern T0* T73f1041(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_default_pointer_feature_name */
extern T0* T73f1035(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.hash_code_feature_name */
extern T0* T73f1033(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_pointer_ref_feature_validity */
extern void T1326f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.pointer_ref_class_name */
extern T0* T73f905(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_boolean_feature_validity */
extern void T1326f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.negated_feature_name */
extern T0* T73f1029(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implication_feature_name */
extern T0* T73f1027(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.disjuncted_semistrict_feature_name */
extern T0* T73f1025(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.disjuncted_exclusive_feature_name */
extern T0* T73f1023(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.disjuncted_feature_name */
extern T0* T73f1021(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.conjuncted_semistrict_feature_name */
extern T0* T73f1019(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.conjuncted_feature_name */
extern T0* T73f1017(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_boolean_ref_feature_validity */
extern void T1326f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.boolean_ref_class_name */
extern T0* T73f904(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_32_ref_class_name */
extern T0* T73f901(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_character_n_feature_validity */
extern void T1326f27(GE_context* ac, T0* C, T0* a1, T8 a2);
/* [detachable] ET_TOKEN_CONSTANTS.natural_32_code_feature_name */
extern T0* T73f1009(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.code_feature_name */
extern T0* T73f1007(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_character_n_ref_feature_validity */
extern void T1326f26(GE_context* ac, T0* C, T0* a1, T0* a2, T8 a3);
/* [detachable] ET_ECF_LIBRARY.character_8_type */
extern T0* T187f35(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.character_8_type */
extern T0* T58f14(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.character_8_type */
extern T0* T40f45(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.current_universe */
extern T0* T1326f7(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.character_8_ref_class_name */
extern T0* T73f898(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_special_feature_validity */
extern void T1326f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.set_count_feature_name */
extern T0* T73f1003(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.put_feature_name */
extern T0* T73f1001(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.make_empty_feature_name */
extern T0* T73f999(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.extend_feature_name */
extern T0* T73f997(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.element_size_feature_name */
extern T0* T73f994(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.capacity_feature_name */
extern T0* T73f5(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.base_address_feature_name */
extern T0* T73f990(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.like_current */
extern T0* T73f947(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.aliased_resized_area_feature_name */
extern T0* T73f988(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_type_feature_validity */
extern void T1326f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.type_id_feature_name */
extern T0* T73f985(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.runtime_name_feature_name */
extern T0* T73f983(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_deferred_feature_name */
extern T0* T73f979(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_attached_feature_name */
extern T0* T73f977(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.has_default_feature_name */
extern T0* T73f975(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.generic_parameter_type_feature_name */
extern T0* T73f971(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.default_feature_name */
extern T0* T73f969(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_builtin_any_feature_validity */
extern void T1326f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.standard_copy_feature_name */
extern T0* T73f966(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.tagged_out_feature_name */
extern T0* T73f962(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.standard_twin_feature_name */
extern T0* T73f960(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.standard_is_equal_feature_name */
extern T0* T73f958(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.same_type_feature_name */
extern T0* T73f956(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.is_deep_equal_feature_name */
extern T0* T73f953(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.generator_feature_name */
extern T0* T73f951(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.type_detachable_like_current_type */
extern T0* T187f93(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.type_detachable_like_current_type */
extern T0* T58f122(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.type_detachable_like_current_type */
extern T0* T40f98(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.generating_type_feature_name */
extern T0* T73f949(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.deep_twin_feature_name */
extern T0* T73f946(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.conforms_to_feature_name */
extern T0* T73f944(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.tokens */
extern T0* T1326f5(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.string_ */
extern T0* T1326f6(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.process_external_procedure */
extern void T380f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_external_procedure_validity */
extern void T380f110(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_iteration_components_validity */
extern void T380f118(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_QUANTIFIER_EXPRESSION.unfolded_cursor_name */
extern T0* T1254f3(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.unfolded_cursor_name */
extern T0* T1008f4(GE_context* ac, T0* C);
/* detachable ET_REPEAT_INSTRUCTION.unfolded_cursor_name */
extern T0* T1264f2(GE_context* ac, T0* C);
/* detachable ET_ACROSS_INSTRUCTION.unfolded_cursor_name */
extern T0* T1011f3(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.set_iteration_item */
extern void T184f128(GE_context* ac, T0* C, T1 a1);
/* detachable ET_QUANTIFIER_EXPRESSION.item_name */
extern T0* T1254f2(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.item_name */
extern T0* T1008f2(GE_context* ac, T0* C);
/* detachable ET_REPEAT_INSTRUCTION.item_name */
extern T0* T1264f1(GE_context* ac, T0* C);
/* detachable ET_ACROSS_INSTRUCTION.item_name */
extern T0* T1011f1(GE_context* ac, T0* C);
/* detachable ET_ITERATION_COMPONENT_LIST.iteration_component */
extern T0* T365f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_CHECKER.check_object_tests_validity */
extern void T380f117(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuot3a_error */
extern void T53f246(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vuot3a */
extern T0* T785c896(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vuot3_error */
extern T1 T53f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vuot3a_error */
extern void T36f246(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vuot3_error */
extern T1 T36f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.error_handler */
extern T0* T380f62(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.set_fatal_error */
extern void T380f78(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.older_ise_version */
extern T1 T64f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_VERSION.is_less */
extern T1 T46f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_VERSION.build */
extern T6 T46f9(GE_context* ac, T0* C);
/* [detachable] UT_VERSION.has_build */
extern T1 T46f8(GE_context* ac, T0* C);
/* [detachable] UT_VERSION.revision */
extern T6 T46f7(GE_context* ac, T0* C);
/* [detachable] UT_VERSION.has_revision */
extern T1 T46f6(GE_context* ac, T0* C);
/* [detachable] UT_VERSION.minor */
extern T6 T46f5(GE_context* ac, T0* C);
/* [detachable] UT_VERSION.has_minor */
extern T1 T46f13(GE_context* ac, T0* C);
/* [detachable] UT_VERSION.major */
extern T6 T46f12(GE_context* ac, T0* C);
/* [detachable] UT_VERSION.has_major */
extern T1 T46f11(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.older_ise_version */
extern T1 T375f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.ise_6_4_7_7252 */
extern T0* T380f53(GE_context* ac, T0* C);
/* [detachable] UT_VERSION.make */
extern T0* T46c19(GE_context* ac, T6 a1, T6 a2, T6 a3, T6 a4);
/* [detachable] ET_IDENTIFIER.set_object_test_local */
extern void T184f130(GE_context* ac, T0* C, T1 a1);
/* detachable ET_OBJECT_TEST_LIST.object_test */
extern T0* T1481f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_CHECKER.check_formal_arguments_validity */
extern void T380f116(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_argument_supplier */
extern void T380f131(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SUPPLIER_BUILDER.report_argument_supplier */
extern void T1341f10(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_formal_argument_declaration */
extern void T380f130(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_signature_type_validity */
extern void T380f129(GE_context* ac, T0* C, T0* a1);
extern T1 T380f129ot1(T0* a1, T0** a2);
/* [detachable] ET_TYPE_CHECKER.check_creation_type_validity */
extern void T384f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T384f25ot1(T0* a1, T0** a2);
extern T1 T384f25ot2(T0* a1, T0** a2);
extern T1 T384f25ot3(T0* a1, T0** a2);
extern T1 T384f25ot4(T0* a1, T0** a2);
extern T1 T384f25ot5(T0* a1, T0** a2);
/* [detachable] ET_TYPE_CHECKER.ise_6_0_6_7358 */
extern T0* T384f17(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_like_current */
extern T1 T1501f55(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.is_like_current */
extern T1 T891f53(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_like_current */
extern T1 T1526f44(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_like_current */
extern T1 T1525f47(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.is_like_current */
extern T1 T1524f45(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_like_current */
extern T1 T242f44(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.is_like_current */
extern T1 T449f41(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_like_current */
extern T1 T74f201(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.is_like_current */
extern T1 T254f74(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.is_like_current */
extern T1 T72f75(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.shallow_named_type */
extern T0* T1501f36(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.shallow_named_type_with_type_mark */
extern T0* T1501f38(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1501f38ot1(T0* a1, T0** a2);
extern T1 T1501f38ot2(T0* a1, T0** a2);
extern T1 T1501f38ot3(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.shallow_named_type_with_type_mark */
extern T0* T1526f46(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1526f46ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.shallow_named_type_with_type_mark */
extern T0* T1525f49(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1525f49ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.shallow_named_type_with_type_mark */
extern T0* T1524f47(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.shallow_named_type_with_type_mark */
extern T0* T242f46(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T242f46ot1(T0* a1, T0** a2);
extern T1 T242f46ot2(T0* a1, T0** a2);
extern T1 T242f46ot3(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.shallow_named_type_with_type_mark */
extern T0* T449f43(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.shallow_named_type_with_type_mark */
extern T0* T74f204(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.shallow_base_type_with_type_mark */
extern T0* T74f203(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.shallow_named_type_with_type_mark */
extern T0* T254f76(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.shallow_base_type_with_type_mark */
extern T0* T254f75(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.shallow_named_type_with_type_mark */
extern T0* T72f74(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.shallow_base_type_with_type_mark */
extern T0* T72f73(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER.shallow_named_type_with_type_mark */
extern T0* T891f56(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T891f56ot1(T0* a1, T0** a2);
extern T1 T891f56ot2(T0* a1, T0** a2);
extern T1 T891f56ot3(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.shallow_named_type */
extern T0* T891f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.actual_parameters */
extern T0* T254f1(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.formal_parameters */
extern T0* T74f29(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.actual_parameters */
extern T0* T72f3(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.shallow_named_type */
extern T0* T74f200(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.shallow_named_type */
extern T0* T1526f43(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.shallow_named_type */
extern T0* T1525f46(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.shallow_named_type */
extern T0* T1524f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.shallow_named_type */
extern T0* T254f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.shallow_named_type */
extern T0* T242f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.shallow_named_type */
extern T0* T72f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.shallow_named_type */
extern T0* T449f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.free_context */
extern void T380f77(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.tokens */
extern T0* T380f49(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].force_last */
extern void T391f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NESTED_TYPE_CONTEXT].force */
extern void T731f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_NESTED_TYPE_CONTEXT].fill_with */
extern void T730f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_NESTED_TYPE_CONTEXT].extend */
extern void T730f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].resize */
extern void T391f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NESTED_TYPE_CONTEXT].aliased_resized_area */
extern T0* T731f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_NESTED_TYPE_CONTEXT].aliased_resized_area */
extern T0* T730f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].new_capacity */
extern T6 T391f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].extendible */
extern T1 T391f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TYPE_CHECKER.check_type_validity */
extern void T384f24(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.new_context */
extern T0* T380f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].remove_last */
extern void T391f14(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_NESTED_TYPE_CONTEXT].keep_head */
extern void T730f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_NESTED_TYPE_CONTEXT].set_count */
extern void T730f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].move_last_cursors_after */
extern void T391f17(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT].set_next_cursor */
extern void T732f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT].set_after */
extern void T732f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].last */
extern T0* T391f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].is_empty */
extern T1 T391f10(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_external_procedure */
extern void T71f154(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_external_procedure_validity */
extern void T71f159(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_external_builtin_procedure_validity */
extern void T71f164(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_tuple_put_reference */
extern void T71f171(GE_context* ac, T0* C, T0* a1);
extern T1 T71f171ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.is_expanded */
extern T1 T1466f8(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.is_expanded */
extern T1 T269f10(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.is_expanded */
extern T1 T475f8(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_procedure */
extern void T71f170(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_set_is_ignored */
extern void T71f169(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.argument_type_set */
extern T0* T71f76(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_raise */
extern void T71f168(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_ignore */
extern void T71f167(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_catch */
extern void T71f166(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.is_builtin_unknown */
extern T1 T1516f70(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_builtin */
extern T1 T1516f46(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_external_procedure_validity */
extern void T71f159p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_iteration_components_validity */
extern void T71f139(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_object_tests_validity */
extern void T71f138(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.ise_6_4_7_7252 */
extern T0* T71f69(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_formal_arguments_validity */
extern void T71f137(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_argument_supplier */
extern void T71f145(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_formal_argument_declaration */
extern void T71f144(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_COMMA_ARGUMENT.set_attached_index */
extern void T1767f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FORMAL_ARGUMENT.set_attached_index */
extern void T256f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.set_dynamic_type_set */
extern void T71f149(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.put */
extern void T259f17(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_INFIX_CAST_EXPRESSION.set_index */
extern void T1980f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT.set_index */
extern void T2027f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.set_index */
extern void T2046f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CONVERT_TO_EXPRESSION.set_index */
extern void T2048f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.set_index */
extern void T2047f26(GE_context* ac, T0* C, T6 a1);
/* detachable ET_AGENT_TYPED_OPEN_ARGUMENT.set_index */
extern void T1559f13(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ITERATION_CURSOR.set_index */
extern void T1551f24(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_MANIFEST_STRING.set_index */
extern void T1550f27(GE_context* ac, T0* C, T6 a1);
/* detachable ET_OLD_EXPRESSION.set_index */
extern void T1548f25(GE_context* ac, T0* C, T6 a1);
/* detachable ET_QUESTION_MARK_SYMBOL.set_index */
extern void T1005f47(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.set_index */
extern void T1557f52(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_index */
extern void T1556f60(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_index */
extern void T1555f55(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.set_index */
extern void T1554f53(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_index */
extern void T1553f61(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_index */
extern void T1552f56(GE_context* ac, T0* C, T6 a1);
/* detachable ET_AGENT_OPEN_TARGET.set_index */
extern void T1558f11(GE_context* ac, T0* C, T6 a1);
/* detachable ET_PARENTHESIZED_EXPRESSION.set_index */
extern void T1232f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AGENT_IMPLICIT_CURRENT_TARGET.set_index */
extern void T1811f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CALL_AGENT.set_index */
extern void T1041f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CREATE_EXPRESSION.set_index */
extern void T328f29(GE_context* ac, T0* C, T6 a1);
/* detachable ET_QUANTIFIER_EXPRESSION.set_index */
extern void T1254f43(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ACROSS_EXPRESSION.set_index */
extern void T1008f51(GE_context* ac, T0* C, T6 a1);
/* detachable ET_MANIFEST_ARRAY.set_index */
extern void T1218f35(GE_context* ac, T0* C, T6 a1);
/* detachable ET_PRECURSOR_EXPRESSION.set_index */
extern void T1538f31(GE_context* ac, T0* C, T6 a1);
/* detachable ET_STATIC_CALL_EXPRESSION.set_index */
extern void T1267f35(GE_context* ac, T0* C, T6 a1);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.set_index */
extern void T1491f47(GE_context* ac, T0* C, T6 a1);
/* detachable ET_REGULAR_REAL_CONSTANT.set_index */
extern void T1490f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NAMED_OBJECT_TEST.set_index */
extern void T736f31(GE_context* ac, T0* C, T6 a1);
/* detachable ET_OLD_OBJECT_TEST.set_index */
extern void T1549f33(GE_context* ac, T0* C, T6 a1);
/* detachable ET_STRIP_EXPRESSION.set_index */
extern void T1270f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_MANIFEST_TUPLE.set_index */
extern void T288f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_OBJECT_TEST.set_index */
extern void T311f28(GE_context* ac, T0* C, T6 a1);
/* detachable ET_MANIFEST_TYPE.set_index */
extern void T1546f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.set_index */
extern void T312f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EQUALITY_EXPRESSION.set_index */
extern void T313f26(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FEATURE_ADDRESS.set_index */
extern void T1539f24(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXPRESSION_ADDRESS.set_index */
extern void T1542f25(GE_context* ac, T0* C, T6 a1);
/* detachable ET_RESULT_ADDRESS.set_index */
extern void T1541f24(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CURRENT_ADDRESS.set_index */
extern void T1540f24(GE_context* ac, T0* C, T6 a1);
/* detachable ET_INSPECT_EXPRESSION.set_index */
extern void T1189f28(GE_context* ac, T0* C, T6 a1);
/* detachable ET_IF_EXPRESSION.set_index */
extern void T1167f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_RESULT.set_index */
extern void T309f35(GE_context* ac, T0* C, T6 a1);
/* detachable ET_VOID.set_index */
extern void T997f35(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FALSE_CONSTANT.set_index */
extern void T1494f44(GE_context* ac, T0* C, T6 a1);
/* detachable ET_TRUE_CONSTANT.set_index */
extern void T1493f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CURRENT.set_index */
extern void T282f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.set_index */
extern void T635f30(GE_context* ac, T0* C, T6 a1);
/* detachable ET_BRACKET_EXPRESSION.set_index */
extern void T1038f28(GE_context* ac, T0* C, T6 a1);
/* detachable ET_PREFIX_EXPRESSION.set_index */
extern void T1547f30(GE_context* ac, T0* C, T6 a1);
/* detachable ET_INFIX_EXPRESSION.set_index */
extern void T1543f30(GE_context* ac, T0* C, T6 a1);
/* detachable ET_PARENTHESIS_EXPRESSION.set_index */
extern void T888f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.set_index */
extern void T287f27(GE_context* ac, T0* C, T6 a1);
/* detachable ET_VERBATIM_STRING.set_index */
extern void T1485f52(GE_context* ac, T0* C, T6 a1);
/* detachable ET_SPECIAL_MANIFEST_STRING.set_index */
extern void T1484f48(GE_context* ac, T0* C, T6 a1);
/* detachable ET_REGULAR_MANIFEST_STRING.set_index */
extern void T1483f50(GE_context* ac, T0* C, T6 a1);
/* detachable ET_BINARY_INTEGER_CONSTANT.set_index */
extern void T1489f66(GE_context* ac, T0* C, T6 a1);
/* detachable ET_OCTAL_INTEGER_CONSTANT.set_index */
extern void T1488f66(GE_context* ac, T0* C, T6 a1);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.set_index */
extern void T1487f66(GE_context* ac, T0* C, T6 a1);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.set_index */
extern void T1486f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.set_index */
extern void T372f68(GE_context* ac, T0* C, T6 a1);
/* detachable ET_C3_CHARACTER_CONSTANT.set_index */
extern void T1482f49(GE_context* ac, T0* C, T6 a1);
/* detachable ET_C2_CHARACTER_CONSTANT.set_index */
extern void T1496f52(GE_context* ac, T0* C, T6 a1);
/* detachable ET_C1_CHARACTER_CONSTANT.set_index */
extern void T1495f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.force_last */
extern void T259f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE_SET].force */
extern void T627f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE_SET].fill_with */
extern void T626f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE_SET].extend */
extern void T626f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.fixed_array */
extern T0* T259f5(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE_SET].default_create */
extern T0* T627c3(GE_context* ac);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.resize */
extern void T259f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE_SET].aliased_resized_area */
extern T0* T627f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE_SET].aliased_resized_area */
extern T0* T626f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.new_capacity */
extern T6 T259f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.capacity */
extern T6 T259f6(GE_context* ac, T0* C);
/* detachable ET_AGENT_OPEN_TARGET.index */
extern T6 T1558f4(GE_context* ac, T0* C);
/* [detachable] ET_AGENT_IMPLICIT_CURRENT_TARGET.index */
extern T6 T1811f1(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_EXPRESSION.index */
extern T6 T888f4(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.dynamic_type_set */
extern T0* T71f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_signature_type_validity */
extern void T71f143(GE_context* ac, T0* C, T0* a1);
extern T1 T71f143ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.free_context */
extern void T71f112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.tokens */
extern T0* T71f70(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.new_context */
extern T0* T71f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_external_procedure */
extern void T67f1185(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_procedure */
extern void T67f1228(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_address_routine */
extern void T67f1118(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.dedent */
extern void T67f649(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_routine_name */
extern void T67f666(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T67f666ot1(T0* a1, T0** a2);
/* [detachable] KL_TEXT_OUTPUT_FILE.put_integer */
extern void T300f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.put_integer_64 */
extern void T300f46(GE_context* ac, T0* C, T7 a1);
/* [detachable] KL_STRING_OUTPUT_STREAM.put_integer */
extern void T260f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_STRING_OUTPUT_STREAM.put_integer_64 */
extern void T260f11(GE_context* ac, T0* C, T7 a1);
/* [detachable] ET_C_GENERATOR.print_type_name */
extern void T67f670(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.id */
extern T6 T451f1(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.id */
extern T6 T450f1(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.id */
extern T6 T302f1(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.id */
extern T6 T319f1(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.id */
extern T6 T223f4(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_indentation */
extern void T67f640(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.indent */
extern void T67f639(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_argument_name */
extern void T67f704(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.argument_type_set */
extern T0* T67f495(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST.item */
extern T0* T274f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST.valid_index */
extern T1 T274f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_current_name */
extern void T67f706(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_address_routine_name */
extern void T67f1143(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] KL_STRING_OUTPUT_STREAM.put_character */
extern void T260f7(GE_context* ac, T0* C, T2 a1);
/* [detachable] KL_STRING_OUTPUT_STREAM.put_string */
extern void T260f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STRING_OUTPUT_STREAM.string_ */
extern T0* T260f3(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_type_declaration */
extern void T67f638(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_eif_any_type_name */
extern void T67f607(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_feature_name_comment */
extern void T67f1119(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_TUPLE_TYPE.unaliased_to_text */
extern T0* T254f13(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.append_unaliased_to_string */
extern void T254f121(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.append_unaliased_to_string */
extern void T1957f24(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.append_unaliased_to_string */
extern void T1501f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.append_unaliased_to_string */
extern void T891f87(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.append_unaliased_to_string */
extern void T1526f87(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.append_unaliased_to_string */
extern void T1525f88(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.append_unaliased_to_string */
extern void T1524f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.append_unaliased_to_string */
extern void T242f79(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.append_unaliased_to_string */
extern void T449f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.append_unaliased_to_string */
extern void T74f264(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.append_unaliased_to_string */
extern void T603f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.append_unaliased_to_string */
extern void T72f128(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.append_unaliased_to_string */
extern void T577f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.append_unaliased_to_string */
extern void T1956f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.append_unaliased_to_string */
extern void T240f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.unaliased_to_text */
extern T0* T72f19(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.unaliased_to_text */
extern T0* T74f125(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_external_routine */
extern void T67f1260(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] ET_C_GENERATOR.flush_to_cpp_file */
extern void T67f1277(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.open_append */
extern void T300f36(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.old_open_append */
extern void T300f45(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.open_append */
extern void T300f45p1(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.reset_temp_variables */
extern void T67f650(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE].wipe_out */
extern void T262f12(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_PRIMARY_TYPE].keep_head */
extern void T629f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_PRIMARY_TYPE].set_count */
extern void T629f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE].move_all_cursors_after */
extern void T262f15(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_DYNAMIC_PRIMARY_TYPE].set_next_cursor */
extern void T631f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_DYNAMIC_PRIMARY_TYPE].set_after */
extern void T631f6(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_warning_message */
extern void T53f366(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_warning_message */
extern void T36f366(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_dllwin_body */
extern void T67f1276(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] STRING_8].forth */
extern void T499f12(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].cursor_forth */
extern void T152f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].add_traversing_cursor */
extern void T152f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] STRING_8].set_next_cursor */
extern void T499f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].remove_traversing_cursor */
extern void T152f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] STRING_8].set */
extern void T499f9(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] STRING_8].item */
extern T0* T499f7(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] STRING_8].start */
extern void T499f11(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].cursor_start */
extern void T152f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].cursor_off */
extern T1 T152f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] STRING_8].off */
extern T1 T499f6(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].new_cursor */
extern T0* T152f9(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] STRING_8].make */
extern T0* T499c8(GE_context* ac, T0* a1);
/* [detachable] ET_C_GENERATOR.print_declaration_type_cast_from_c */
extern void T67f1090(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_declaration_type_cast */
extern void T67f710(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.is_boolean */
extern T1 T451f57(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.is_boolean */
extern T1 T450f56(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.is_boolean */
extern T1 T319f31(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.is_boolean */
extern T1 T302f31(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_boolean */
extern T1 T223f49(GE_context* ac, T0* C);
/* [detachable] KL_STRING_OUTPUT_STREAM.put_line */
extern void T260f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_cpp_body */
extern void T67f1275(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_C_GENERATOR.dynamic_type_set */
extern T0* T67f256(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONVERT_TO_EXPRESSION.is_current */
extern T1 T2048f7(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.is_current */
extern T1 T2047f8(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.is_current */
extern T1 T2046f6(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_CAST_EXPRESSION.is_current */
extern T1 T1980f5(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_EXPRESSION.is_current */
extern T1 T888f5(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_current */
extern T1 T1489f9(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_current */
extern T1 T1488f9(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_current */
extern T1 T1487f9(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_current */
extern T1 T1486f9(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_current */
extern T1 T372f11(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_current */
extern T1 T1485f22(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_current */
extern T1 T1484f20(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_current */
extern T1 T1482f16(GE_context* ac, T0* C);
/* detachable ET_ITERATION_CURSOR.is_current */
extern T1 T1551f5(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.is_current */
extern T1 T1254f17(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.is_current */
extern T1 T1008f20(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.is_current */
extern T1 T1550f5(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.is_current */
extern T1 T736f8(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.is_current */
extern T1 T1549f13(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.is_current */
extern T1 T1539f5(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.is_current */
extern T1 T635f5(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.is_current */
extern T1 T1557f34(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.is_current */
extern T1 T1556f17(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.is_current */
extern T1 T1555f17(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.is_current */
extern T1 T1554f18(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.is_current */
extern T1 T1553f20(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.is_current */
extern T1 T1552f36(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.is_current */
extern T1 T1041f10(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.is_current */
extern T1 T1270f10(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.is_current */
extern T1 T1218f10(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.is_current */
extern T1 T1232f5(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.is_current */
extern T1 T1038f5(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.is_current */
extern T1 T311f7(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.is_current */
extern T1 T1548f5(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.is_current */
extern T1 T1547f5(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.is_current */
extern T1 T1546f5(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.is_current */
extern T1 T312f7(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.is_current */
extern T1 T313f7(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.is_current */
extern T1 T1543f15(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.is_current */
extern T1 T1542f5(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.is_current */
extern T1 T1541f5(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.is_current */
extern T1 T1540f5(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.is_current */
extern T1 T1538f8(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.is_current */
extern T1 T1267f8(GE_context* ac, T0* C);
/* detachable ET_INSPECT_EXPRESSION.is_current */
extern T1 T1189f9(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.is_current */
extern T1 T1167f9(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_current */
extern T1 T1496f25(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_current */
extern T1 T1495f25(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.is_current */
extern T1 T1494f8(GE_context* ac, T0* C);
/* detachable ET_VOID.is_current */
extern T1 T997f8(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.is_current */
extern T1 T1493f8(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.is_current */
extern T1 T1491f24(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.is_current */
extern T1 T1490f24(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.is_current */
extern T1 T1483f6(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.is_current */
extern T1 T288f13(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.is_current */
extern T1 T287f9(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.is_current */
extern T1 T309f7(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.is_current */
extern T1 T282f7(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.is_current */
extern T1 T328f7(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_current */
extern T1 T184f7(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].item */
extern T0* T152f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] ST_SPLITTER.split */
extern T0* T1953f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ST_SPLITTER.do_split */
extern T0* T1953f7(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].force_last */
extern void T152f18(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] STRING_8].put_right */
extern void T500f4(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] STRING_8].make */
extern T0* T500c3(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].has */
extern T1 T745f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].search_position */
extern void T745f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].hash_position */
extern T6 T745f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].key_storage_item */
extern T6 T745f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].item_storage_item */
extern T6 T745f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].clashes_item */
extern T6 T745f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].key_equality_tester */
extern T0* T745f23(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].slots_item */
extern T6 T745f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] ST_SPLITTER.string_ */
extern T0* T1953f5(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].make */
extern T0* T152c17(GE_context* ac);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].set_internal_cursor */
extern void T152f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ST_SPLITTER.make_with_separators */
extern T0* T1953c9(GE_context* ac, T0* a1);
/* [detachable] ST_SPLITTER.set_separators */
extern void T1953f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].put */
extern void T745f83(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].slots_put */
extern void T745f67(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].clashes_put */
extern void T745f68(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].item_storage_put */
extern void T745f79(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].unset_found_item */
extern void T745f54(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].make */
extern T0* T745c43(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].set_internal_cursor */
extern void T745f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].new_cursor */
extern T0* T745f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32].make */
extern T0* T1429c9(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].make_slots */
extern void T745f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].new_modulus */
extern T6 T745f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].make_clashes */
extern void T745f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].make_key_storage */
extern void T745f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].make_item_storage */
extern void T745f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_external_c_struct_body */
extern void T67f1274(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_C_GENERATOR.print_external_c_body */
extern void T67f1273(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T1 a7);
/* [detachable] ET_C_GENERATOR.print_external_c_prototype */
extern void T67f1272(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_C_GENERATOR.print_external_builtin_body */
extern void T67f1271(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_procedure_body */
extern void T67f1282(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_tuple_procedure_body */
extern void T67f1320(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].wipe_out */
extern void T272f13(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION].keep_head */
extern void T650f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION].set_count */
extern void T650f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].move_all_cursors_after */
extern void T272f19(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EXPRESSION].set_next_cursor */
extern void T652f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EXPRESSION].set_after */
extern void T652f6(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_set_object_comparison_call */
extern void T67f867(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_attachment_expression */
extern void T67f712(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.print_boxed_attribute_item_name */
extern void T67f669(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.is_attached */
extern T1 T451f37(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.is_attached */
extern T1 T450f37(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.is_attached */
extern T1 T302f41(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.is_attached */
extern T1 T319f41(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.is_attached */
extern T1 T323f10(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_attached */
extern T1 T223f31(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_expression */
extern void T67f698(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INFIX_CAST_EXPRESSION.process */
extern void T1980f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_infix_cast_expression */
extern void T1327f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_infix_cast_expression */
extern void T71f400(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_infix_cast_expression_validity */
extern void T71f449(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_expression_validity */
extern void T71f241(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_expression_supplier */
extern void T71f307(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_infix_cast_expression */
extern void T398f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_infix_cast_expression */
extern void T394f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_infix_cast_expression */
extern void T380f396(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_infix_cast_expression_validity */
extern void T380f398(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_expression_validity */
extern void T380f79(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_expression_supplier */
extern void T380f80(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SUPPLIER_BUILDER.report_expression_supplier */
extern void T1341f7(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_infix_cast_expression */
extern void T67f1444(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_infix_cast_expression */
extern void T67f1446(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONVERT_TO_EXPRESSION.process */
extern void T2048f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_convert_to_expression */
extern void T1327f119(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_convert_to_expression */
extern void T398f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_convert_to_expression */
extern void T394f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_convert_to_expression */
extern void T71f568(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_convert_to_expression_validity */
extern void T71f571(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_call_expression_validity */
extern void T71f332(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T71f332ot1(T0* a1, T0** a2);
extern T1 T71f332ot2(T0* a1, T0** a2);
extern T1 T71f332ot3(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.free_adapted_base_classes */
extern void T71f248(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].force_last */
extern void T404f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].force */
extern void T760f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].fill_with */
extern void T761f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].extend */
extern void T761f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].resize */
extern void T404f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].aliased_resized_area */
extern T0* T760f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].aliased_resized_area */
extern T0* T761f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].new_capacity */
extern T6 T404f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].extendible */
extern T1 T404f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgmc0f_error */
extern void T53f413(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vgmc0f */
extern T0* T785c1036(GE_context* ac, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.report_vgmc0f_error */
extern void T36f413(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuex2a_error */
extern void T53f412(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vuex2a */
extern T0* T785c1035(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vuex2_error */
extern T1 T53f114(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vuex2a_error */
extern void T36f412(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vuex2_error */
extern T1 T36f114(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_tuple_label_call_expression_validity */
extern void T71f317(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_tuple_label_expression */
extern void T71f381(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T71f381ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_qualified_call_target_dynamic_types */
extern void T71f376(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.put_query_call */
extern void T451f83(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_PRECURSOR.put_query_call */
extern void T306f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].force */
extern void T616f35(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].key_storage_put */
extern void T616f43(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CALL_NAME].force */
extern void T1407f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CALL_NAME].fill_with */
extern void T1404f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CALL_NAME].extend */
extern void T1404f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].slots_put */
extern void T616f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].clashes_put */
extern void T616f41(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].slots_item */
extern T6 T616f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].clashes_item */
extern T6 T616f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].hash_position */
extern T6 T616f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENTHESIS_SYMBOL.hash_code */
extern T6 T2020f33(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.hash_code */
extern T6 T1545f16(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.hash_code */
extern T6 T1544f16(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.hash_code */
extern T6 T973f2(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.hash_code */
extern T6 T981f25(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.hash_code */
extern T6 T994f96(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.hash_code */
extern T6 T622f96(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.hash_code */
extern T6 T184f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].resize */
extern void T616f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].clashes_resize */
extern void T616f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].special_integer_ */
extern T0* T616f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].key_storage_resize */
extern void T616f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CALL_NAME].aliased_resized_area */
extern T0* T1407f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_CALL_NAME].aliased_resized_area */
extern T0* T1404f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].item_storage_resize */
extern void T616f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL].aliased_resized_area */
extern T0* T1406f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL].aliased_resized_area */
extern T0* T1402f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].key_storage_item */
extern T0* T616f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].slots_resize */
extern void T616f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].new_modulus */
extern T6 T616f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].new_capacity */
extern T6 T616f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].item_storage_put */
extern void T616f39(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL].force */
extern void T1406f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL].fill_with */
extern void T1402f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL].extend */
extern void T1402f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].search_position */
extern void T616f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].unset_found_item */
extern void T616f37(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.put_query_call */
extern void T236f49(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.set_next */
extern void T314f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.put_query_call */
extern void T450f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.put_query_call */
extern void T302f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.put_query_call */
extern void T319f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_query_call */
extern void T223f73(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.make */
extern T0* T314c15(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuar1c_error */
extern void T53f460(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vuar1c */
extern T0* T785c1083(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vuar1_error */
extern T1 T53f135(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vuar1c_error */
extern void T36f460(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vuar1_error */
extern T1 T36f135(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONVERT_TO_EXPRESSION.arguments_count */
extern T6 T2048f21(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.arguments_count */
extern T6 T1547f20(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.arguments_count */
extern T6 T1543f20(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_EXPRESSION.arguments_count */
extern T6 T888f9(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuta2b_error */
extern void T53f459(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vuta2b */
extern T0* T785c1082(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vuta2_error */
extern T1 T53f134(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vuta2b_error */
extern void T36f459(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vuta2_error */
extern T1 T36f134(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.named_type */
extern T0* T241f14(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_tuple_label_parenthesis_call_validity */
extern void T71f247(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T71f247ot1(T0* a1, T0** a2);
extern T1 T71f247ot2(T0* a1, T0** a2);
extern T1 T71f247ot3(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_feature_call_expression_validity */
extern void T71f316(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T71f316ot1(T0* a1, T0** a2);
extern T1 T71f316ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_feature_call_instruction_validity */
extern void T71f315(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T71f315ot1(T0* a1, T0** a2);
extern T1 T71f315ot2(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vkcn1a_error */
extern void T53f502(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vkcn1a */
extern T0* T785c1125(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_PRECURSOR_KEYWORD.lower_name */
extern T0* T955f22(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.tokens */
extern T0* T955f7(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.position */
extern T0* T955f9(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vkcn1_error */
extern T1 T53f148(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vkcn1a_error */
extern void T36f502(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vkcn1_error */
extern T1 T36f148(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_procedure_call_instruction_validity */
extern void T71f245(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_qualified_call_instruction */
extern void T71f311(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.put_procedure_call */
extern void T451f86(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_PRECURSOR.put_procedure_call */
extern void T306f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].force */
extern void T617f43(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].key_storage_put */
extern void T617f49(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].slots_put */
extern void T617f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].clashes_put */
extern void T617f47(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].slots_item */
extern T6 T617f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].clashes_item */
extern T6 T617f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].hash_position */
extern T6 T617f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].resize */
extern void T617f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].clashes_resize */
extern void T617f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].special_integer_ */
extern T0* T617f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].key_storage_resize */
extern void T617f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].item_storage_resize */
extern void T617f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL].aliased_resized_area */
extern T0* T1411f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL].aliased_resized_area */
extern T0* T1412f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].key_storage_item */
extern T0* T617f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].slots_resize */
extern void T617f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].new_modulus */
extern T6 T617f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].new_capacity */
extern T6 T617f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].item_storage_put */
extern void T617f45(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL].force */
extern void T1411f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL].fill_with */
extern void T1412f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL].extend */
extern void T1412f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].search_position */
extern void T617f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].unset_found_item */
extern void T617f41(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.put_procedure_call */
extern void T236f56(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.set_next */
extern void T318f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.put_procedure_call */
extern void T450f86(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.put_procedure_call */
extern void T302f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.put_procedure_call */
extern void T319f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_procedure_call */
extern void T223f80(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.make */
extern T0* T318c13(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ASSIGNER_INSTRUCTION.target */
extern T0* T1527f8(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_INSTRUCTION.target */
extern T0* T2002f1(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.target */
extern T0* T284f1(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_feature_call_validity */
extern void T71f310(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.reset_fatal_error */
extern void T71f375(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_actual_arguments_validity */
extern void T71f374(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T71f374ot1(T0* a1, T0** a2);
extern T1 T71f374ot2(T0* a1);
extern T1 T71f374ot3(T0* a1, T0** a2);
extern T1 T71f374ot4(T0* a1);
extern T1 T71f374ot5(T0* a1, T0** a2);
extern T1 T71f374ot6(T0* a1, T0** a2);
extern T1 T71f374ot7(T0* a1, T0** a2);
extern T1 T71f374ot8(T0* a1, T0** a2);
extern T1 T71f374ot9(T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuar2b_error */
extern void T53f510(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T6 a5, T0* a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vuar2b */
extern T0* T785c1133(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T6 a5, T0* a6, T0* a7);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vuar2_error */
extern T1 T53f150(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vuar2b_error */
extern void T36f510(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T6 a5, T0* a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.reportable_vuar2_error */
extern T1 T36f150(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuar2a_error */
extern void T53f509(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* [detachable] ET_VALIDITY_ERROR.make_vuar2a */
extern T0* T785c1132(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* [detachable] ET_ERROR_HANDLER.report_vuar2a_error */
extern void T36f509(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdpr4b_error */
extern void T53f508(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* [detachable] ET_VALIDITY_ERROR.make_vdpr4b */
extern T0* T785c1131(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdpr4_error */
extern T1 T53f149(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdpr4b_error */
extern void T36f508(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* [detachable] ET_ERROR_HANDLER.reportable_vdpr4_error */
extern T1 T36f149(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_converted_target_infix_expression_validity */
extern void T71f434(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_INFIX_EXPRESSION.set_left */
extern void T1543f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.conforms_to_type */
extern T1 T241f45(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_INFIX_CAST_EXPRESSION.make */
extern T0* T1980c17(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_INFIX_EXPRESSION.set_right */
extern void T1543f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.put */
extern void T1977f16(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.put */
extern void T283f22(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.convert_expression */
extern T0* T71f90(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_conversion */
extern void T71f519(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.make */
extern T0* T2046c18(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_qualified_call_expression */
extern void T71f377(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T71f377ot1(T0* a1, T0** a2);
/* [detachable] ET_AST_FACTORY.new_convert_to_expression */
extern T0* T465f369(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CONVERT_TO_EXPRESSION.make */
extern T0* T2048c23(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_CONVERT_TO_EXPRESSION.position */
extern T0* T2048f18(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.position */
extern T0* T2047f17(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.position */
extern T0* T2046f14(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_CAST_EXPRESSION.position */
extern T0* T1980f6(GE_context* ac, T0* C);
/* detachable ET_ITERATION_CURSOR.position */
extern T0* T1551f7(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.position */
extern T0* T1254f14(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.position */
extern T0* T1008f21(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.position */
extern T0* T1550f6(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.position */
extern T0* T736f9(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_null_position */
extern T1 T973f27(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.is_null_position */
extern T1 T955f13(GE_context* ac, T0* C);
/* detachable ET_AGENT_KEYWORD.is_null_position */
extern T1 T952f12(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_null_position */
extern T1 T981f6(GE_context* ac, T0* C);
/* detachable ET_VOID.is_null_position */
extern T1 T997f7(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.is_null_position */
extern T1 T309f12(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.position */
extern T0* T1549f10(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.position */
extern T0* T1539f7(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.position */
extern T0* T635f6(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.position */
extern T0* T1557f16(GE_context* ac, T0* C);
/* detachable ET_AGENT_KEYWORD.position */
extern T0* T952f9(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.position */
extern T0* T1556f18(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.position */
extern T0* T1555f35(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.position */
extern T0* T1554f34(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.position */
extern T0* T1553f41(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.position */
extern T0* T1552f18(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.position */
extern T0* T1041f11(GE_context* ac, T0* C);
/* detachable ET_AGENT_OPEN_TARGET.position */
extern T0* T1558f5(GE_context* ac, T0* C);
/* [detachable] ET_AGENT_IMPLICIT_CURRENT_TARGET.position */
extern T0* T1811f3(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.implicit_target_position */
extern T0* T1557f26(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.first_leaf */
extern T0* T1557f32(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.implicit_target_position */
extern T0* T1556f33(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.first_leaf */
extern T0* T1556f38(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.implicit_target_position */
extern T0* T1555f29(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.first_leaf */
extern T0* T1555f34(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.implicit_target_position */
extern T0* T1554f27(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.first_leaf */
extern T0* T1554f33(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.implicit_target_position */
extern T0* T1553f34(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.first_leaf */
extern T0* T1553f39(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.implicit_target_position */
extern T0* T1552f30(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.first_leaf */
extern T0* T1552f35(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.implicit_target_position */
extern T0* T1041f19(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.first_leaf */
extern T0* T184f92(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.position */
extern T0* T1270f12(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.item */
extern T0* T1270f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_STRIP_EXPRESSION.is_empty */
extern T1 T1270f7(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.position */
extern T0* T288f16(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.is_empty */
extern T1 T288f17(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.position */
extern T0* T1218f12(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.position */
extern T0* T1218f12p1(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.item */
extern T0* T1218f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_MANIFEST_ARRAY.is_empty */
extern T1 T1218f7(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.position */
extern T0* T1232f6(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.position */
extern T0* T1038f6(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.position */
extern T0* T311f8(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.position */
extern T0* T1548f6(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.position */
extern T0* T1547f6(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.position */
extern T0* T1546f6(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.position */
extern T0* T312f8(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.position */
extern T0* T313f8(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.position */
extern T0* T1543f6(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.position */
extern T0* T1542f7(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.position */
extern T0* T1541f7(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.position */
extern T0* T1540f7(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.position */
extern T0* T1538f10(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.position */
extern T0* T1267f9(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.position */
extern T0* T287f10(GE_context* ac, T0* C);
/* detachable ET_INSPECT_EXPRESSION.position */
extern T0* T1189f6(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD_EXPRESSION.position */
extern T0* T1771f3(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.position */
extern T0* T1167f10(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.position */
extern T0* T309f9(GE_context* ac, T0* C);
/* detachable ET_VOID.position */
extern T0* T997f10(GE_context* ac, T0* C);
/* detachable ET_CONVERT_PROCEDURE.is_convert_to */
extern T1 T1509f7(GE_context* ac, T0* C);
/* detachable ET_CONVERT_FUNCTION.is_convert_to */
extern T1 T1508f6(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_CONVERT_FEATURE.is_convert_to */
extern T1 T569f5(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_creation_expression */
extern void T71f483(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_argument_operand_dynamic_types */
extern void T71f436(GE_context* ac, T0* C, T0* a1, T6 a2, T0* a3);
/* [detachable] ET_CREATE_EXPRESSION.arguments */
extern T0* T328f13(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.set_creation */
extern void T306f53(GE_context* ac, T0* C, T1 a1);
/* detachable ET_DYNAMIC_FEATURE.set_creation */
extern void T236f48(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.dynamic_procedure */
extern T0* T451f46(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.dynamic_procedure */
extern T0* T450f46(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.dynamic_procedure */
extern T0* T302f47(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.dynamic_procedure */
extern T0* T319f49(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.dynamic_procedure */
extern T0* T223f37(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_creation_vape_validity */
extern void T71f494(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vape2a_error */
extern void T53f527(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vape2a */
extern T0* T785c1150(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vape2_error */
extern T1 T53f160(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vape2a_error */
extern void T36f527(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.reportable_vape2_error */
extern T1 T36f160(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.as_feature */
extern T0* T1516f80(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.as_feature */
extern T0* T253f81(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.as_feature */
extern T0* T1515f75(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.as_feature */
extern T0* T1513f72(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.as_feature */
extern T0* T1512f76(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.as_feature */
extern T0* T1511f74(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.as_feature */
extern T0* T371f73(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.as_feature */
extern T0* T369f72(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.as_feature */
extern T0* T359f74(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.as_feature */
extern T0* T360f69(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.as_feature */
extern T0* T469f73(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.as_feature */
extern T0* T476f71(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.add_non_descendant_caller_clients_to */
extern void T71f432(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CLIENT_LIST.force_first */
extern void T402f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLIENT_LIST.resize */
extern void T402f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_ITEM].aliased_resized_area */
extern T0* T759f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_CLIENT_ITEM].aliased_resized_area */
extern T0* T758f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLIENT_LIST.new_capacity */
extern T6 T402f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLIENT_LIST.capacity */
extern T6 T402f10(GE_context* ac, T0* C);
/* detachable ET_CLIENTS.is_none_or_unknown */
extern T1 T1065f10(GE_context* ac, T0* C);
/* [detachable] ET_CLIENT_LIST.is_none_or_unknown */
extern T1 T402f7(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.add_creation_clients_to */
extern void T1516f122(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CREATOR_LIST.add_creation_clients_to */
extern void T1107f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CREATOR.add_creation_clients_to */
extern void T1104f20(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLIENT_LIST.append_first */
extern void T402f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.any_clients */
extern T0* T187f24(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.any_clients */
extern T0* T58f88(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.any_clients */
extern T0* T40f37(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.add_creation_clients_to */
extern void T476f112(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ONCE_PROCEDURE.add_creation_clients_to */
extern void T1515f118(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DO_PROCEDURE.add_creation_clients_to */
extern void T469f115(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CLIENT_LIST.wipe_out */
extern void T402f16(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_ITEM].keep_head */
extern void T759f5(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLIENT_ITEM].keep_head */
extern void T758f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CLIENT_ITEM].set_count */
extern void T758f10(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_PROCEDURE.is_feature */
extern T1 T1516f79(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_feature */
extern T1 T253f80(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_feature */
extern T1 T1515f74(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_feature */
extern T1 T1513f71(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_feature */
extern T1 T1512f75(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_feature */
extern T1 T1511f73(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_feature */
extern T1 T371f72(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_feature */
extern T1 T369f71(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_feature */
extern T1 T359f73(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_feature */
extern T1 T360f68(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_feature */
extern T1 T469f72(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_feature */
extern T1 T476f70(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcc6c_error */
extern void T53f525(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vgcc6c */
extern T0* T785c1148(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vgcc6c_error */
extern void T36f525(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CONVERT_PROCEDURE.name */
extern T0* T1509f1(GE_context* ac, T0* C);
/* detachable ET_CONVERT_FUNCTION.name */
extern T0* T1508f1(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_CONVERT_FEATURE.name */
extern T0* T569f1(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_convert_from_expression */
extern T0* T465f368(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.make */
extern T0* T2047c23(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_CONVERT_PROCEDURE.is_convert_from */
extern T1 T1509f8(GE_context* ac, T0* C);
/* detachable ET_CONVERT_FUNCTION.is_convert_from */
extern T1 T1508f7(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_CONVERT_FEATURE.is_convert_from */
extern T1 T569f4(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.manifest_constant_convert_feature */
extern T0* T2048f8(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.manifest_constant_convert_feature */
extern T0* T2047f9(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.manifest_constant_convert_feature */
extern T0* T2046f7(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_INFIX_CAST_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1980f7(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_IDENTIFIER.manifest_constant_convert_feature */
extern T0* T184f104(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_BINARY_INTEGER_CONSTANT.manifest_constant_convert_feature */
extern T0* T1489f53(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_LIBRARY.real_64_convert_feature */
extern T0* T187f75(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.real_64_convert_feature */
extern T0* T58f110(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.real_64_convert_feature */
extern T0* T40f82(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.real_32_convert_feature */
extern T0* T187f73(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.real_32_convert_feature */
extern T0* T58f109(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.real_32_convert_feature */
extern T0* T40f80(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.natural_64_convert_feature */
extern T0* T187f65(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.natural_64_convert_feature */
extern T0* T58f106(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.natural_64_convert_feature */
extern T0* T40f73(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_natural_64 */
extern T1 T1489f17(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_negative */
extern T1 T1489f28(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_minus */
extern T1 T994f99(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.natural_32_convert_feature */
extern T0* T187f63(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.natural_32_convert_feature */
extern T0* T58f105(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.natural_32_convert_feature */
extern T0* T40f71(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_natural_32 */
extern T1 T1489f16(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.natural_16_convert_feature */
extern T0* T187f61(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.natural_16_convert_feature */
extern T0* T58f104(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.natural_16_convert_feature */
extern T0* T40f69(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_natural_16 */
extern T1 T1489f15(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.natural_8_convert_feature */
extern T0* T187f59(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.natural_8_convert_feature */
extern T0* T58f103(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.natural_8_convert_feature */
extern T0* T40f67(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_natural_8 */
extern T1 T1489f14(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.integer_64_convert_feature */
extern T0* T187f54(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.integer_64_convert_feature */
extern T0* T58f101(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.integer_64_convert_feature */
extern T0* T40f62(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_integer_64 */
extern T1 T1489f13(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.integer_64_min_value_abs */
extern T11 T1489f27(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.integer_32_convert_feature */
extern T0* T187f52(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.integer_32_convert_feature */
extern T0* T58f100(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.integer_32_convert_feature */
extern T0* T40f60(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_integer_32 */
extern T1 T1489f12(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.integer_32_min_value_abs */
extern T11 T1489f26(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.integer_16_convert_feature */
extern T0* T187f50(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.integer_16_convert_feature */
extern T0* T58f99(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.integer_16_convert_feature */
extern T0* T40f58(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_integer_16 */
extern T1 T1489f11(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.integer_16_min_value_abs */
extern T11 T1489f25(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.integer_8_convert_feature */
extern T0* T187f48(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.integer_8_convert_feature */
extern T0* T58f98(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.integer_8_convert_feature */
extern T0* T40f56(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_integer_8 */
extern T1 T1489f10(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.integer_8_min_value_abs */
extern T11 T1489f24(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.manifest_constant_convert_feature */
extern T0* T1488f53(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_natural_64 */
extern T1 T1488f17(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_negative */
extern T1 T1488f24(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_natural_32 */
extern T1 T1488f16(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_natural_16 */
extern T1 T1488f15(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_natural_8 */
extern T1 T1488f14(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_integer_64 */
extern T1 T1488f13(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.integer_64_min_value_abs */
extern T11 T1488f28(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_integer_32 */
extern T1 T1488f12(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.integer_32_min_value_abs */
extern T11 T1488f27(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_integer_16 */
extern T1 T1488f11(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.integer_16_min_value_abs */
extern T11 T1488f26(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_integer_8 */
extern T1 T1488f10(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.integer_8_min_value_abs */
extern T11 T1488f25(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.manifest_constant_convert_feature */
extern T0* T1487f53(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_natural_64 */
extern T1 T1487f17(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_negative */
extern T1 T1487f28(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_natural_32 */
extern T1 T1487f16(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_natural_16 */
extern T1 T1487f15(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_natural_8 */
extern T1 T1487f14(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_integer_64 */
extern T1 T1487f13(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.integer_64_min_value_abs */
extern T11 T1487f27(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_integer_32 */
extern T1 T1487f12(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.integer_32_min_value_abs */
extern T11 T1487f26(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_integer_16 */
extern T1 T1487f11(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.integer_16_min_value_abs */
extern T11 T1487f25(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_integer_8 */
extern T1 T1487f10(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.integer_8_min_value_abs */
extern T11 T1487f24(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.manifest_constant_convert_feature */
extern T0* T1486f53(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_natural_64 */
extern T1 T1486f17(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_negative */
extern T1 T1486f24(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_natural_32 */
extern T1 T1486f16(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_natural_16 */
extern T1 T1486f15(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_natural_8 */
extern T1 T1486f14(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_integer_64 */
extern T1 T1486f13(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.integer_64_min_value_abs */
extern T11 T1486f28(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_integer_32 */
extern T1 T1486f12(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.integer_32_min_value_abs */
extern T11 T1486f27(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_integer_16 */
extern T1 T1486f11(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.integer_16_min_value_abs */
extern T11 T1486f26(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_integer_8 */
extern T1 T1486f10(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.integer_8_min_value_abs */
extern T11 T1486f25(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.manifest_constant_convert_feature */
extern T0* T372f53(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_natural_64 */
extern T1 T372f24(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_negative */
extern T1 T372f40(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_natural_32 */
extern T1 T372f23(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_natural_16 */
extern T1 T372f22(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_natural_8 */
extern T1 T372f21(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_integer_64 */
extern T1 T372f20(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.integer_64_min_value_abs */
extern T11 T372f37(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_integer_32 */
extern T1 T372f19(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.integer_32_min_value_abs */
extern T11 T372f16(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_integer_16 */
extern T1 T372f18(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.integer_16_min_value_abs */
extern T11 T372f36(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_integer_8 */
extern T1 T372f17(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.integer_8_min_value_abs */
extern T11 T372f35(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.manifest_constant_convert_feature */
extern T0* T1485f39(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_LIBRARY.string_32_convert_feature */
extern T0* T187f86(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.string_32_convert_feature */
extern T0* T58f115(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.string_32_convert_feature */
extern T0* T40f93(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_string_32 */
extern T1 T1485f12(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.has_invalid_code */
extern T1 T1485f18(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.string_32_type */
extern T0* T187f85(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.string_32_type */
extern T0* T58f30(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.string_32_type */
extern T0* T40f92(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.string_8_convert_feature */
extern T0* T187f84(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.string_8_convert_feature */
extern T0* T58f114(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.string_8_convert_feature */
extern T0* T40f91(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_string_8 */
extern T1 T1485f11(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.tokens */
extern T0* T1485f43(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.manifest_constant_convert_feature */
extern T0* T1484f35(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_string_32 */
extern T1 T1484f12(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_string_8 */
extern T1 T1484f11(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.tokens */
extern T0* T1484f39(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.manifest_constant_convert_feature */
extern T0* T1482f38(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_LIBRARY.character_32_convert_feature */
extern T0* T187f38(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.character_32_convert_feature */
extern T0* T58f95(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.character_32_convert_feature */
extern T0* T40f47(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_character_32 */
extern T1 T1482f10(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.character_8_convert_feature */
extern T0* T187f36(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.character_8_convert_feature */
extern T0* T58f94(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.character_8_convert_feature */
extern T0* T40f46(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_character_8 */
extern T1 T1482f9(GE_context* ac, T0* C);
/* detachable ET_ITERATION_CURSOR.manifest_constant_convert_feature */
extern T0* T1551f18(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_QUANTIFIER_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1254f32(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ACROSS_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1008f33(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ONCE_MANIFEST_STRING.manifest_constant_convert_feature */
extern T0* T1550f21(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NAMED_OBJECT_TEST.manifest_constant_convert_feature */
extern T0* T736f22(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_OLD_OBJECT_TEST.manifest_constant_convert_feature */
extern T0* T1549f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_FEATURE_ADDRESS.manifest_constant_convert_feature */
extern T0* T1539f18(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.manifest_constant_convert_feature */
extern T0* T635f22(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.manifest_constant_convert_feature */
extern T0* T1557f35(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.manifest_constant_convert_feature */
extern T0* T1556f41(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.manifest_constant_convert_feature */
extern T0* T1555f37(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.manifest_constant_convert_feature */
extern T0* T1554f36(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.manifest_constant_convert_feature */
extern T0* T1553f42(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.manifest_constant_convert_feature */
extern T0* T1552f38(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CALL_AGENT.manifest_constant_convert_feature */
extern T0* T1041f27(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_STRIP_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1270f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_MANIFEST_TUPLE.manifest_constant_convert_feature */
extern T0* T288f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_MANIFEST_ARRAY.manifest_constant_convert_feature */
extern T0* T1218f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_PARENTHESIZED_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1232f19(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_BRACKET_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1038f22(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_OBJECT_TEST.manifest_constant_convert_feature */
extern T0* T311f21(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_OLD_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1548f18(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_PREFIX_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1547f23(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_MANIFEST_TYPE.manifest_constant_convert_feature */
extern T0* T1546f19(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.manifest_constant_convert_feature */
extern T0* T312f20(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EQUALITY_EXPRESSION.manifest_constant_convert_feature */
extern T0* T313f18(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_INFIX_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1543f23(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_EXPRESSION_ADDRESS.manifest_constant_convert_feature */
extern T0* T1542f18(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_RESULT_ADDRESS.manifest_constant_convert_feature */
extern T0* T1541f18(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CURRENT_ADDRESS.manifest_constant_convert_feature */
extern T0* T1540f18(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_PRECURSOR_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1538f22(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_STATIC_CALL_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1267f26(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.manifest_constant_convert_feature */
extern T0* T287f23(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_INSPECT_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1189f21(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_IF_EXPRESSION.manifest_constant_convert_feature */
extern T0* T1167f24(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_CREATE_EXPRESSION.manifest_constant_convert_feature */
extern T0* T328f26(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_C2_CHARACTER_CONSTANT.manifest_constant_convert_feature */
extern T0* T1496f38(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_C2_CHARACTER_CONSTANT.is_character_32 */
extern T1 T1496f20(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.has_invalid_code */
extern T1 T1496f23(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_character_8 */
extern T1 T1496f19(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.manifest_constant_convert_feature */
extern T0* T1495f37(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_C1_CHARACTER_CONSTANT.is_character_32 */
extern T1 T1495f20(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.has_invalid_code */
extern T1 T1495f23(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_character_8 */
extern T1 T1495f19(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.manifest_constant_convert_feature */
extern T0* T282f28(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_RESULT.manifest_constant_convert_feature */
extern T0* T309f26(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_FALSE_CONSTANT.manifest_constant_convert_feature */
extern T0* T1494f29(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_VOID.manifest_constant_convert_feature */
extern T0* T997f26(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_TRUE_CONSTANT.manifest_constant_convert_feature */
extern T0* T1493f29(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.manifest_constant_convert_feature */
extern T0* T1491f31(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_REGULAR_REAL_CONSTANT.manifest_constant_convert_feature */
extern T0* T1490f31(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_REGULAR_MANIFEST_STRING.manifest_constant_convert_feature */
extern T0* T1483f34(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_REGULAR_MANIFEST_STRING.is_string_32 */
extern T1 T1483f21(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.has_invalid_code */
extern T1 T1483f24(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.is_string_8 */
extern T1 T1483f20(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.tokens */
extern T0* T1483f38(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.current_universe */
extern T0* T71f86(GE_context* ac, T0* C);
/* [detachable] ET_TYPE_CHECKER.convert_feature */
extern T0* T384f21(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.context_named_type */
extern T0* T72f110(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.convert_from_feature */
extern T0* T74f239(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_BRACED_TYPE_LIST.has_named_type_with_type_marks */
extern T1 T1388f7(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TOKEN_CONSTANTS.attached_separate_type_mark */
extern T0* T73f1541(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.convert_to_feature */
extern T0* T74f238(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.conforms_to_context */
extern T1 T241f46(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.conforms_to_context_with_type_marks */
extern T1 T241f44(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_actual_argument_validity */
extern void T71f433(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T71f433ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.free_attachment_scope */
extern void T71f121(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE].force_last */
extern void T399f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ATTACHMENT_SCOPE].force */
extern void T748f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ATTACHMENT_SCOPE].fill_with */
extern void T747f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ATTACHMENT_SCOPE].extend */
extern void T747f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE].resize */
extern void T399f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ATTACHMENT_SCOPE].aliased_resized_area */
extern T0* T748f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ATTACHMENT_SCOPE].aliased_resized_area */
extern T0* T747f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE].new_capacity */
extern T6 T399f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE].extendible */
extern T1 T399f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_OBJECT_TEST_SCOPE.keep_object_tests */
extern void T393f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST].keep_first */
extern void T739f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_NAMED_OBJECT_TEST].keep_head */
extern void T737f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_NAMED_OBJECT_TEST].set_count */
extern void T737f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST].move_all_cursors_after */
extern void T739f14(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NAMED_OBJECT_TEST].set_next_cursor */
extern void T1427f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NAMED_OBJECT_TEST].set_after */
extern void T1427f6(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.build_negated_scope */
extern void T398f60(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.build_negated_scope */
extern void T394f65(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_PRECURSOR_KEYWORD.is_infix_or_else_symbol */
extern T1 T955f21(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.is_infix_or_else */
extern T1 T955f20(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.build_scope */
extern void T398f4(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE.copy_scope */
extern void T397f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].append_last */
extern void T745f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].extend_last */
extern void T745f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32].forth */
extern void T1429f14(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].cursor_forth */
extern void T745f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].add_traversing_cursor */
extern void T745f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32].set_next_cursor */
extern void T1429f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32].set_position */
extern void T1429f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].remove_traversing_cursor */
extern void T745f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32].set_after */
extern void T1429f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].put_last */
extern void T745f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].compress */
extern void T745f81(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].slots_wipe_out */
extern void T745f60(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].clashes_wipe_out */
extern void T745f59(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].key_storage_keep_head */
extern void T745f78(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].item_storage_keep_head */
extern void T745f77(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].move_all_cursors */
extern void T745f82(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].key_storage_put */
extern void T745f80(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32].item */
extern T6 T1429f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].cursor_item */
extern T6 T745f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32].after */
extern T1 T1429f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32].start */
extern void T1429f13(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].cursor_start */
extern void T745f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].cursor_off */
extern T1 T745f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] INTEGER_32].off */
extern T1 T1429f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].is_empty */
extern T1 T745f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].resize */
extern void T745f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].clashes_resize */
extern void T745f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].special_integer_ */
extern T0* T745f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].key_storage_resize */
extern void T745f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].item_storage_resize */
extern void T745f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].slots_resize */
extern void T745f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].new_capacity */
extern T6 T745f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].extendible */
extern T1 T745f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].wipe_out */
extern void T745f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].key_storage_wipe_out */
extern void T745f58(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].item_storage_wipe_out */
extern void T745f57(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].move_all_cursors_after */
extern void T745f56(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.new_attachment_scope */
extern T0* T71f99(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE].remove_last */
extern void T399f13(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_ATTACHMENT_SCOPE].keep_head */
extern void T747f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ATTACHMENT_SCOPE].set_count */
extern void T747f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE].move_last_cursors_after */
extern void T399f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ATTACHMENT_SCOPE].set_next_cursor */
extern void T749f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ATTACHMENT_SCOPE].set_after */
extern void T749f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE].last */
extern T0* T399f6(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_SCOPE.make */
extern T0* T397c11(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE].is_empty */
extern T1 T399f10(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.build_scope */
extern void T394f9(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_OBJECT_TEST_SCOPE.count */
extern T6 T393f3(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.is_infix_implies_symbol */
extern T1 T955f19(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.is_infix_implies */
extern T1 T955f18(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.is_infix_and_then_symbol */
extern T1 T955f17(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.is_infix_and_then */
extern T1 T955f16(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuar1b_error */
extern void T53f507(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vuar1b */
extern T0* T785c1130(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vuar1b_error */
extern void T36f507(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuar1a_error */
extern void T53f506(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vuar1a */
extern T0* T785c1129(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vuar1a_error */
extern void T36f506(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdpr4a_error */
extern void T53f505(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vdpr4a */
extern T0* T785c1128(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vdpr4a_error */
extern void T36f505(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_PRECURSOR_KEYWORD.precursor_keyword */
extern T0* T955f15(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.precursor_keyword */
extern T0* T622f102(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.precursor_keyword */
extern T0* T1545f18(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.precursor_keyword */
extern T0* T1544f18(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.precursor_keyword */
extern T0* T973f29(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.precursor_keyword */
extern T0* T981f27(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.precursor_keyword */
extern T0* T994f103(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.precursor_keyword */
extern T0* T2020f37(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.precursor_keyword */
extern T0* T184f90(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.is_precursor */
extern T1 T955f14(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_precursor */
extern T1 T622f101(GE_context* ac, T0* C);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.is_precursor */
extern T1 T1545f17(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.is_precursor */
extern T1 T1544f17(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.is_precursor */
extern T1 T973f28(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.is_precursor */
extern T1 T981f26(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_precursor */
extern T1 T994f102(GE_context* ac, T0* C);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_precursor */
extern T1 T2020f36(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_precursor */
extern T1 T184f89(GE_context* ac, T0* C);
/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION.set_arguments */
extern void T1756f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_STATIC_CALL_INSTRUCTION.set_arguments */
extern void T1537f16(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECURSOR_INSTRUCTION.set_arguments */
extern void T1535f16(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENTHESIS_INSTRUCTION.set_arguments */
extern void T2002f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_CALL.set_arguments */
extern void T329f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECURSOR_EXPRESSION.set_arguments */
extern void T1538f34(GE_context* ac, T0* C, T0* a1);
/* detachable ET_STATIC_CALL_EXPRESSION.set_arguments */
extern void T1267f37(GE_context* ac, T0* C, T0* a1);
/* detachable ET_BRACKET_EXPRESSION.set_arguments */
extern void T1038f29(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENTHESIS_EXPRESSION.set_arguments */
extern void T888f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.set_arguments */
extern void T635f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.set_arguments */
extern void T287f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.set_arguments */
extern void T284f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.make */
extern T0* T1977c14(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_MANIFEST_TUPLE.make */
extern T0* T288c36(GE_context* ac);
/* [detachable] ET_MANIFEST_TUPLE.make */
extern void T288f36p1(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION_ITEM].make */
extern T0* T671f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION_ITEM].make_empty */
extern T0* T669c5(GE_context* ac, T6 a1);
/* [detachable] ET_MANIFEST_TUPLE.fixed_array */
extern T0* T288f10(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION_ITEM].default_create */
extern T0* T671c3(GE_context* ac);
/* [detachable] ET_MANIFEST_TUPLE.tokens */
extern T0* T288f8(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.put_first */
extern void T288f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION_ITEM].force */
extern void T671f5(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION_ITEM].fill_with */
extern void T669f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION_ITEM].extend */
extern void T669f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MANIFEST_TUPLE.make_with_capacity */
extern T0* T288c28(GE_context* ac, T6 a1);
/* [detachable] ET_MANIFEST_TUPLE.make_with_capacity */
extern void T288f28p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.put_first */
extern void T1977f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.fixed_array */
extern T0* T1977f11(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.make_with_capacity */
extern void T1977f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.make_with_capacity */
extern void T1977f17p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.right_parenthesis_symbol */
extern T0* T73f195(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_right_parenthesis */
extern T0* T341c94(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.left_parenthesis_symbol */
extern T0* T73f194(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_left_parenthesis */
extern T0* T341c93(GE_context* ac);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.tokens */
extern T0* T1977f10(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.tuple_argument_position */
extern T6 T71f93(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_INFIX_CAST_EXPRESSION.is_empty_actual_argument */
extern T1 T1980f12(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.is_empty_actual_argument */
extern T1 T2048f13(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.is_empty_actual_argument */
extern T1 T2047f14(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.is_empty_actual_argument */
extern T1 T2046f12(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_empty_actual_argument */
extern T1 T184f88(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_empty_actual_argument */
extern T1 T1489f43(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_empty_actual_argument */
extern T1 T1488f43(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_empty_actual_argument */
extern T1 T1487f43(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_empty_actual_argument */
extern T1 T1486f43(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_empty_actual_argument */
extern T1 T372f44(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_empty_actual_argument */
extern T1 T1485f33(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_empty_actual_argument */
extern T1 T1484f29(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_empty_actual_argument */
extern T1 T1482f28(GE_context* ac, T0* C);
/* detachable ET_ITERATION_CURSOR.is_empty_actual_argument */
extern T1 T1551f15(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.is_empty_actual_argument */
extern T1 T1254f29(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.is_empty_actual_argument */
extern T1 T1008f30(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.is_empty_actual_argument */
extern T1 T1550f18(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.is_empty_actual_argument */
extern T1 T736f19(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.is_empty_actual_argument */
extern T1 T1549f22(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.is_empty_actual_argument */
extern T1 T1539f15(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.is_empty_actual_argument */
extern T1 T635f17(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.is_empty_actual_argument */
extern T1 T1557f29(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.is_empty_actual_argument */
extern T1 T1556f35(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.is_empty_actual_argument */
extern T1 T1555f31(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.is_empty_actual_argument */
extern T1 T1554f30(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.is_empty_actual_argument */
extern T1 T1553f36(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.is_empty_actual_argument */
extern T1 T1552f32(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.is_empty_actual_argument */
extern T1 T1041f21(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.is_empty_actual_argument */
extern T1 T1270f22(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.is_empty_actual_argument */
extern T1 T288f23(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.is_empty_actual_argument */
extern T1 T1218f22(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.is_empty_actual_argument */
extern T1 T1232f17(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.is_empty_actual_argument */
extern T1 T1038f18(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.is_empty_actual_argument */
extern T1 T311f19(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.is_empty_actual_argument */
extern T1 T1548f15(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.is_empty_actual_argument */
extern T1 T1547f19(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.is_empty_actual_argument */
extern T1 T1546f16(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.is_empty_actual_argument */
extern T1 T312f17(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.is_empty_actual_argument */
extern T1 T313f16(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.is_empty_actual_argument */
extern T1 T1543f19(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.is_empty_actual_argument */
extern T1 T1542f15(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.is_empty_actual_argument */
extern T1 T1541f15(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.is_empty_actual_argument */
extern T1 T1540f15(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.is_empty_actual_argument */
extern T1 T1538f19(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.is_empty_actual_argument */
extern T1 T1267f19(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.is_empty_actual_argument */
extern T1 T287f20(GE_context* ac, T0* C);
/* detachable ET_INSPECT_EXPRESSION.is_empty_actual_argument */
extern T1 T1189f17(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.is_empty_actual_argument */
extern T1 T1167f21(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.is_empty_actual_argument */
extern T1 T328f21(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_empty_actual_argument */
extern T1 T1496f27(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_empty_actual_argument */
extern T1 T1495f27(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.is_empty_actual_argument */
extern T1 T282f24(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.is_empty_actual_argument */
extern T1 T309f24(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.is_empty_actual_argument */
extern T1 T1494f24(GE_context* ac, T0* C);
/* detachable ET_VOID.is_empty_actual_argument */
extern T1 T997f23(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.is_empty_actual_argument */
extern T1 T1493f24(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.is_empty_actual_argument */
extern T1 T1491f26(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.is_empty_actual_argument */
extern T1 T1490f26(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.is_empty_actual_argument */
extern T1 T1483f28(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.is_empty */
extern T1 T1977f6(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.is_empty */
extern T1 T283f10(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.arguments */
extern T0* T2047f16(GE_context* ac, T0* C);
/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION.arguments */
extern T0* T1756f2(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_INSTRUCTION.arguments */
extern T0* T1537f4(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_INSTRUCTION.arguments */
extern T0* T1535f2(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL.arguments */
extern T0* T329f1(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.arguments */
extern T0* T1538f5(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.arguments */
extern T0* T1267f5(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.arguments */
extern T0* T635f3(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.name */
extern T0* T2047f1(GE_context* ac, T0* C);
/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION.name */
extern T0* T1756f1(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_INSTRUCTION.name */
extern T0* T1537f7(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_INSTRUCTION.precursor_keyword */
extern T0* T1535f3(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL.name */
extern T0* T329f3(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.precursor_keyword */
extern T0* T1538f4(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.name */
extern T0* T1267f11(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.name */
extern T0* T635f2(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_vape_validity */
extern void T71f373(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vape1b_error */
extern void T53f504(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vape1b */
extern T0* T785c1127(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vape1_error */
extern T1 T53f136(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vape1b_error */
extern void T36f504(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.reportable_vape1_error */
extern T1 T36f136(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuex2b_error */
extern void T53f461(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vuex2b */
extern T0* T785c1084(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vuex2b_error */
extern void T36f461(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_EXTERNAL_PROCEDURE.is_exported_to */
extern T1 T1516f84(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DEFERRED_PROCEDURE.is_exported_to */
extern T1 T476f75(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ONCE_PROCEDURE.is_exported_to */
extern T1 T1515f78(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DO_PROCEDURE.is_exported_to */
extern T1 T469f76(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EXTERNAL_FUNCTION.is_exported_to */
extern T1 T253f86(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_DEFERRED_FUNCTION.is_exported_to */
extern T1 T1513f77(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ONCE_FUNCTION.is_exported_to */
extern T1 T1512f80(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_DO_FUNCTION.is_exported_to */
extern T1 T1511f78(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_exported_to */
extern T1 T371f78(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_exported_to */
extern T1 T369f77(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_exported_to */
extern T1 T359f78(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTRIBUTE.is_exported_to */
extern T1 T360f74(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuta2a_error */
extern void T53f503(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vuta2a */
extern T0* T785c1126(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vuta2a_error */
extern void T36f503(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.set_parenthesis_call */
extern void T287f35(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_PARENTHESIS_EXPRESSION.make */
extern T0* T888c11(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.set_parenthesis_call */
extern void T284f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_PARENTHESIS_INSTRUCTION.make */
extern T0* T2002c8(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.make */
extern T0* T287c26(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.make_qualified_call */
extern void T287f32(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.set_parenthesis_call_position */
extern void T71f312(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_PARENTHESIS_SYMBOL.set_position */
extern void T2020f51(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_PARENTHESIS_SYMBOL.make */
extern T0* T2020c49(GE_context* ac);
/* [detachable] ET_PARENTHESIS_SYMBOL.make_leaf */
extern void T2020f52(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.set_tuple_label */
extern void T184f123(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vkcn2a_error */
extern void T53f411(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vkcn2a */
extern T0* T785c1034(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vkcn2_error */
extern T1 T53f113(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vkcn2a_error */
extern void T36f411(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vkcn2_error */
extern T1 T36f113(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_query_call_expression_validity */
extern void T71f313(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_query_call_type_validity */
extern void T71f378(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T71f378ot1(T0* a1, T0** a2);
extern T1 T71f378ot2(T0* a1, T0** a2);
extern T1 T71f378ot3(T0* a1, T0** a2);
/* [detachable] ET_CONVERT_TO_EXPRESSION.convert_feature */
extern T0* T2048f4(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.convert_feature */
extern T0* T2047f5(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.convert_feature */
extern T0* T2046f3(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.parenthesis_call */
extern T0* T2048f20(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.parenthesis_call */
extern T0* T1038f17(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.parenthesis_call */
extern T0* T1547f18(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.parenthesis_call */
extern T0* T1543f18(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_EXPRESSION.parenthesis_call */
extern T0* T888f8(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.parenthesis_call */
extern T0* T287f5(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_query_parenthesis_call_validity */
extern void T71f246(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T71f246ot1(T0* a1, T0** a2);
extern T1 T71f246ot2(T0* a1, T0** a2);
extern T1 T71f246ot3(T0* a1, T0** a2);
/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION.parenthesis_call */
extern T0* T1756f3(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.parenthesis_call */
extern T0* T635f4(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.parenthesis_call */
extern T0* T284f4(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_query_call_expression_validity */
extern void T71f314(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ATTACHMENT_SCOPE.has_attribute */
extern T1 T397f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.is_stable_attribute */
extern T1 T253f85(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_stable_attribute */
extern T1 T1513f76(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_stable_attribute */
extern T1 T1512f79(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_stable_attribute */
extern T1 T1511f77(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_stable_attribute */
extern T1 T371f77(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_stable_attribute */
extern T1 T369f76(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_stable_attribute */
extern T1 T359f77(GE_context* ac, T0* C);
/* detachable ET_INDEXING_LIST.has_tagged_indexing_term_value */
extern T1 T1173f6(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INDEXING_TERM_LIST.has_indexing_term_value */
extern T1 T1184f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER.has_indexing_term_value */
extern T1 T184f106(GE_context* ac, T0* C, T0* a1);
/* detachable ET_BINARY_INTEGER_CONSTANT.has_indexing_term_value */
extern T1 T1489f54(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.is_plus */
extern T1 T994f130(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.string_ */
extern T0* T1489f57(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.has_indexing_term_value */
extern T1 T1488f54(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OCTAL_INTEGER_CONSTANT.string_ */
extern T0* T1488f57(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.has_indexing_term_value */
extern T1 T1487f54(GE_context* ac, T0* C, T0* a1);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.string_ */
extern T0* T1487f57(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.has_indexing_term_value */
extern T1 T1486f54(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.string_ */
extern T0* T1486f57(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.has_indexing_term_value */
extern T1 T372f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.string_ */
extern T0* T372f57(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.has_indexing_term_value */
extern T1 T1485f40(GE_context* ac, T0* C, T0* a1);
/* detachable ET_VERBATIM_STRING.string_ */
extern T0* T1485f44(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.has_indexing_term_value */
extern T1 T1484f36(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SPECIAL_MANIFEST_STRING.string_ */
extern T0* T1484f40(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.has_indexing_term_value */
extern T1 T1482f39(GE_context* ac, T0* C, T0* a1);
/* CHARACTER_32.as_lower */
extern T3 T3f16(GE_context* ac, T3* C);
/* detachable ET_CUSTOM_ATTRIBUTE.has_indexing_term_value */
extern T1 T1500f5(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C2_CHARACTER_CONSTANT.has_indexing_term_value */
extern T1 T1496f39(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C1_CHARACTER_CONSTANT.has_indexing_term_value */
extern T1 T1495f38(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FALSE_CONSTANT.has_indexing_term_value */
extern T1 T1494f30(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FALSE_CONSTANT.string_ */
extern T0* T1494f33(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.has_indexing_term_value */
extern T1 T1493f30(GE_context* ac, T0* C, T0* a1);
/* detachable ET_TRUE_CONSTANT.string_ */
extern T0* T1493f33(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.has_indexing_term_value */
extern T1 T1491f32(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.string_ */
extern T0* T1491f35(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.has_indexing_term_value */
extern T1 T1490f32(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REGULAR_REAL_CONSTANT.string_ */
extern T0* T1490f35(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.has_indexing_term_value */
extern T1 T1483f35(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REGULAR_MANIFEST_STRING.string_ */
extern T0* T1483f39(GE_context* ac, T0* C);
/* detachable ET_INDEXING_LIST.string_ */
extern T0* T1173f8(GE_context* ac, T0* C);
/* detachable ET_INDEXING.indexing_clause */
extern T0* T1499f2(GE_context* ac, T0* C);
/* detachable ET_TAGGED_INDEXING.indexing_clause */
extern T0* T1498f3(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_stable_attribute */
extern T1 T360f73(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_unqualified_call_expression */
extern void T71f380(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T71f380ot1(T0* a1, T0** a2);
extern T1 T71f380ot2(T0* a1, T0** a2);
extern T1 T71f380ot3(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_like_argument_dynamic_types */
extern void T71f437(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_argument_dynamic_types */
extern void T71f435(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_vape_validity */
extern void T71f379(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vape1a_error */
extern void T53f464(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vape1a */
extern T0* T785c1087(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.report_vape1a_error */
extern void T36f464(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vucr0d_error */
extern void T53f494(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vucr0d */
extern T0* T785c1117(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vucr_error */
extern T1 T53f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vucr0d_error */
extern void T36f494(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vucr_error */
extern T1 T36f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.is_static */
extern T1 T253f55(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_implicitly_static */
extern T1 T253f50(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.are_postconditions_instance_free_recursive */
extern T1 T253f66(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.are_postconditions_instance_free_recursive */
extern T1 T1516f50(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.are_postconditions_instance_free_recursive */
extern T1 T476f67(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.are_postconditions_instance_free_recursive */
extern T1 T1515f71(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.are_postconditions_instance_free_recursive */
extern T1 T469f69(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.are_postconditions_instance_free_recursive */
extern T1 T1513f68(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.are_postconditions_instance_free_recursive */
extern T1 T1512f72(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.are_postconditions_instance_free_recursive */
extern T1 T1511f70(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.are_postconditions_instance_free_recursive */
extern T1 T371f60(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.are_postconditions_instance_free_recursive */
extern T1 T369f58(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.are_postconditions_instance_free_recursive */
extern T1 T359f70(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.are_postconditions_instance_free_recursive */
extern T1 T360f64(GE_context* ac, T0* C);
/* detachable ET_POSTCONDITIONS.is_instance_free */
extern T1 T1244f12(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_instance_free */
extern T1 T184f84(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_feature_name */
extern T1 T184f87(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_instance_free */
extern T1 T372f28(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.is_instance_free */
extern T1 T1008f26(GE_context* ac, T0* C);
/* detachable ET_LOOP_INVARIANTS.is_instance_free */
extern T1 T1215f7(GE_context* ac, T0* C);
/* detachable ET_LOOP_INVARIANTS.assertion */
extern T0* T1215f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_IDENTIFIER.assertion */
extern T0* T184f83(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.assertion */
extern T0* T1489f21(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.assertion */
extern T0* T1488f21(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.assertion */
extern T0* T1487f21(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.assertion */
extern T0* T1486f21(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.assertion */
extern T0* T372f27(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.assertion */
extern T0* T1485f13(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.assertion */
extern T0* T1484f13(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.assertion */
extern T0* T1482f11(GE_context* ac, T0* C);
/* detachable ET_ITERATION_CURSOR.assertion */
extern T0* T1551f11(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.assertion */
extern T0* T1254f21(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.assertion */
extern T0* T1008f25(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.assertion */
extern T0* T1550f13(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.assertion */
extern T0* T736f14(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.assertion */
extern T0* T1549f17(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.assertion */
extern T0* T1539f11(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.assertion */
extern T0* T635f13(GE_context* ac, T0* C);
/* detachable ET_TAGGED_ASSERTION.assertion */
extern T0* T1562f3(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.assertion */
extern T0* T1557f21(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.assertion */
extern T0* T1556f26(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.assertion */
extern T0* T1555f24(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.assertion */
extern T0* T1554f22(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.assertion */
extern T0* T1553f27(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.assertion */
extern T0* T1552f25(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.assertion */
extern T0* T1041f15(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.assertion */
extern T0* T1270f17(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.assertion */
extern T0* T288f18(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.assertion */
extern T0* T1218f17(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.assertion */
extern T0* T1232f12(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.assertion */
extern T0* T1038f12(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.assertion */
extern T0* T311f11(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.assertion */
extern T0* T1548f10(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.assertion */
extern T0* T1547f12(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.assertion */
extern T0* T1546f12(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.assertion */
extern T0* T312f10(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.assertion */
extern T0* T313f10(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.assertion */
extern T0* T1543f12(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.assertion */
extern T0* T1542f10(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.assertion */
extern T0* T1541f11(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.assertion */
extern T0* T1540f11(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.assertion */
extern T0* T1538f14(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.assertion */
extern T0* T1267f14(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.assertion */
extern T0* T287f14(GE_context* ac, T0* C);
/* detachable ET_INSPECT_EXPRESSION.assertion */
extern T0* T1189f12(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.assertion */
extern T0* T1167f16(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.assertion */
extern T0* T328f14(GE_context* ac, T0* C);
/* detachable ET_CLASS_ASSERTION.assertion */
extern T0* T1521f3(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.assertion */
extern T0* T1496f21(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.assertion */
extern T0* T1495f21(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.assertion */
extern T0* T282f20(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.assertion */
extern T0* T309f20(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.assertion */
extern T0* T1494f20(GE_context* ac, T0* C);
/* detachable ET_VOID.assertion */
extern T0* T997f19(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.assertion */
extern T0* T1493f20(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.assertion */
extern T0* T1491f21(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.assertion */
extern T0* T1490f21(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.assertion */
extern T0* T1483f22(GE_context* ac, T0* C);
/* detachable ET_LOOP_INVARIANTS.item */
extern T0* T1215f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_KEYWORD_EXPRESSION.expression */
extern T0* T1771f1(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.iteration_expression */
extern T0* T1008f24(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.is_instance_free */
extern T1 T736f15(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.is_instance_free */
extern T1 T635f14(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.is_instance_free */
extern T1 T1977f8(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.is_instance_free */
extern T1 T283f12(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_CAST_EXPRESSION.is_instance_free */
extern T1 T1980f4(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.is_instance_free */
extern T1 T2048f5(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.is_instance_free */
extern T1 T2047f6(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.is_instance_free */
extern T1 T2046f16(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.is_instance_free */
extern T1 T1041f16(GE_context* ac, T0* C);
/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT_LIST.is_instance_free */
extern T1 T2026f4(GE_context* ac, T0* C);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.is_instance_free */
extern T1 T1025f10(GE_context* ac, T0* C);
/* detachable ET_AGENT_TYPED_OPEN_ARGUMENT.is_instance_free */
extern T1 T1559f10(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_instance_free */
extern T1 T1005f41(GE_context* ac, T0* C);
/* detachable ET_AGENT_OPEN_TARGET.is_instance_free */
extern T1 T1558f8(GE_context* ac, T0* C);
/* [detachable] ET_AGENT_IMPLICIT_CURRENT_TARGET.is_instance_free */
extern T1 T1811f5(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.is_instance_free */
extern T1 T288f19(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.is_instance_free */
extern T1 T1038f13(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.is_instance_free */
extern T1 T311f12(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.is_instance_free */
extern T1 T312f11(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.is_instance_free */
extern T1 T313f11(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.is_instance_free */
extern T1 T287f15(GE_context* ac, T0* C);
/* detachable ET_INSPECT_EXPRESSION.is_instance_free */
extern T1 T1189f13(GE_context* ac, T0* C);
/* detachable ET_WHEN_EXPRESSION_LIST.is_instance_free */
extern T1 T1296f5(GE_context* ac, T0* C);
/* detachable ET_WHEN_EXPRESSION.is_instance_free */
extern T1 T1293f5(GE_context* ac, T0* C);
/* detachable ET_WHEN_EXPRESSION_LIST.item */
extern T0* T1296f4(GE_context* ac, T0* C, T6 a1);
/* detachable ET_INSPECT_EXPRESSION.expression */
extern T0* T1189f19(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.is_instance_free */
extern T1 T1167f17(GE_context* ac, T0* C);
/* detachable ET_ELSEIF_EXPRESSION_LIST.is_instance_free */
extern T1 T1116f5(GE_context* ac, T0* C);
/* detachable ET_ELSEIF_EXPRESSION.is_instance_free */
extern T1 T1113f6(GE_context* ac, T0* C);
/* detachable ET_ELSEIF_EXPRESSION.conditional_expression */
extern T0* T1113f5(GE_context* ac, T0* C);
/* detachable ET_ELSEIF_EXPRESSION_LIST.item */
extern T0* T1116f4(GE_context* ac, T0* C, T6 a1);
/* detachable ET_IF_EXPRESSION.conditional_expression */
extern T0* T1167f13(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.is_instance_free */
extern T1 T328f15(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.is_instance_free */
extern T1 T282f21(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.is_instance_free */
extern T1 T309f21(GE_context* ac, T0* C);
/* detachable ET_VOID.is_instance_free */
extern T1 T997f20(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.is_instance_free */
extern T1 T1557f22(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.is_instance_free */
extern T1 T1556f27(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.is_instance_free */
extern T1 T1555f25(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.is_instance_free */
extern T1 T1554f23(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.is_instance_free */
extern T1 T1553f28(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.is_instance_free */
extern T1 T1552f26(GE_context* ac, T0* C);
/* detachable ET_ITERATION_CURSOR.is_instance_free */
extern T1 T1551f12(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.is_instance_free */
extern T1 T1550f14(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.is_instance_free */
extern T1 T1549f18(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.is_instance_free */
extern T1 T1548f11(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.is_instance_free */
extern T1 T1547f13(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.is_instance_free */
extern T1 T1546f13(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.is_instance_free */
extern T1 T1543f13(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.is_instance_free */
extern T1 T1542f11(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.is_instance_free */
extern T1 T1541f12(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.is_instance_free */
extern T1 T1540f12(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.is_instance_free */
extern T1 T1539f12(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.is_instance_free */
extern T1 T1538f15(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_instance_free */
extern T1 T1496f22(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_instance_free */
extern T1 T1495f22(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.is_instance_free */
extern T1 T1494f21(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.is_instance_free */
extern T1 T1493f21(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.is_instance_free */
extern T1 T1491f22(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.is_instance_free */
extern T1 T1490f22(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_instance_free */
extern T1 T1489f22(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_instance_free */
extern T1 T1488f22(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_instance_free */
extern T1 T1487f22(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_instance_free */
extern T1 T1486f22(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_instance_free */
extern T1 T1485f14(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_instance_free */
extern T1 T1484f14(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.is_instance_free */
extern T1 T1483f23(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_instance_free */
extern T1 T1482f12(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.is_instance_free */
extern T1 T1270f18(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.is_instance_free */
extern T1 T1267f15(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.is_instance_free */
extern T1 T1254f23(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.variant_part */
extern T0* T1254f25(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.invariant_part */
extern T0* T1254f24(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.until_conditional */
extern T0* T1254f20(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.is_instance_free */
extern T1 T1218f19(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.expression */
extern T0* T1218f14(GE_context* ac, T0* C, T6 a1);
/* detachable ET_PARENTHESIZED_EXPRESSION.is_instance_free */
extern T1 T1232f14(GE_context* ac, T0* C);
/* detachable ET_TAGGED_ASSERTION.expression */
extern T0* T1562f4(GE_context* ac, T0* C);
extern T1 T1562f4ot1(T0* a1, T0** a2);
/* detachable ET_CLASS_ASSERTION.expression */
extern T0* T1521f4(GE_context* ac, T0* C);
/* detachable ET_POSTCONDITIONS.assertion */
extern T0* T1244f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_POSTCONDITIONS.item */
extern T0* T1244f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EXTERNAL_FUNCTION.are_preconditions_instance_free_recursive */
extern T1 T253f65(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.are_preconditions_instance_free_recursive */
extern T1 T1516f49(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.are_preconditions_instance_free_recursive */
extern T1 T476f66(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.are_preconditions_instance_free_recursive */
extern T1 T1515f70(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.are_preconditions_instance_free_recursive */
extern T1 T469f68(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.are_preconditions_instance_free_recursive */
extern T1 T1513f67(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.are_preconditions_instance_free_recursive */
extern T1 T1512f71(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.are_preconditions_instance_free_recursive */
extern T1 T1511f69(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.are_preconditions_instance_free_recursive */
extern T1 T371f59(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.are_preconditions_instance_free_recursive */
extern T1 T369f57(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.are_preconditions_instance_free_recursive */
extern T1 T359f69(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.are_preconditions_instance_free_recursive */
extern T1 T360f63(GE_context* ac, T0* C);
/* detachable ET_PRECONDITIONS.is_instance_free */
extern T1 T406f8(GE_context* ac, T0* C);
/* detachable ET_PRECONDITIONS.assertion */
extern T0* T406f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_PRECONDITIONS.item */
extern T0* T406f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EXTERNAL_FUNCTION.is_builtin_static */
extern T1 T253f64(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.string_ */
extern T0* T253f68(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_static */
extern T1 T253f55p1(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.has_static_mark */
extern T1 T253f51(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.has_class_postcondition_recursive */
extern T1 T253f67(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.has_class_postcondition_recursive */
extern T1 T1516f51(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.has_class_postcondition_recursive */
extern T1 T476f32(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.has_class_postcondition_recursive */
extern T1 T1515f47(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.has_class_postcondition_recursive */
extern T1 T469f32(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.has_class_postcondition_recursive */
extern T1 T1513f47(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.has_class_postcondition_recursive */
extern T1 T1512f52(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.has_class_postcondition_recursive */
extern T1 T1511f49(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.has_class_postcondition_recursive */
extern T1 T371f61(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.has_class_postcondition_recursive */
extern T1 T369f59(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.has_class_postcondition_recursive */
extern T1 T359f60(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.has_class_postcondition_recursive */
extern T1 T360f54(GE_context* ac, T0* C);
/* detachable ET_POSTCONDITIONS.has_class_assertion */
extern T1 T1244f7(GE_context* ac, T0* C);
/* detachable ET_INSPECT_EXPRESSION.is_class_assertion */
extern T1 T1189f14(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.is_class_assertion */
extern T1 T1167f18(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.is_class_assertion */
extern T1 T1041f17(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.is_class_assertion */
extern T1 T1038f14(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.is_class_assertion */
extern T1 T1008f27(GE_context* ac, T0* C);
/* detachable ET_VOID.is_class_assertion */
extern T1 T997f21(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.is_class_assertion */
extern T1 T736f16(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.is_class_assertion */
extern T1 T635f15(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_class_assertion */
extern T1 T372f29(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.is_class_assertion */
extern T1 T328f16(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.is_class_assertion */
extern T1 T313f12(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.is_class_assertion */
extern T1 T312f12(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.is_class_assertion */
extern T1 T311f13(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.is_class_assertion */
extern T1 T309f22(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.is_class_assertion */
extern T1 T288f20(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.is_class_assertion */
extern T1 T287f16(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.is_class_assertion */
extern T1 T282f22(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_class_assertion */
extern T1 T184f85(GE_context* ac, T0* C);
/* detachable ET_TAGGED_ASSERTION.is_class_assertion */
extern T1 T1562f5(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.is_class_assertion */
extern T1 T1557f23(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.is_class_assertion */
extern T1 T1556f30(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.is_class_assertion */
extern T1 T1555f26(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.is_class_assertion */
extern T1 T1554f24(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.is_class_assertion */
extern T1 T1553f31(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.is_class_assertion */
extern T1 T1552f27(GE_context* ac, T0* C);
/* detachable ET_ITERATION_CURSOR.is_class_assertion */
extern T1 T1551f13(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.is_class_assertion */
extern T1 T1550f15(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.is_class_assertion */
extern T1 T1549f19(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.is_class_assertion */
extern T1 T1548f12(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.is_class_assertion */
extern T1 T1547f14(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.is_class_assertion */
extern T1 T1546f14(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.is_class_assertion */
extern T1 T1543f14(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.is_class_assertion */
extern T1 T1542f12(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.is_class_assertion */
extern T1 T1541f13(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.is_class_assertion */
extern T1 T1540f13(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.is_class_assertion */
extern T1 T1539f13(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.is_class_assertion */
extern T1 T1538f16(GE_context* ac, T0* C);
/* detachable ET_CLASS_ASSERTION.is_class_assertion */
extern T1 T1521f5(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_class_assertion */
extern T1 T1496f24(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_class_assertion */
extern T1 T1495f24(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.is_class_assertion */
extern T1 T1494f22(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.is_class_assertion */
extern T1 T1493f22(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.is_class_assertion */
extern T1 T1491f23(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.is_class_assertion */
extern T1 T1490f23(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_class_assertion */
extern T1 T1489f32(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_class_assertion */
extern T1 T1488f32(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_class_assertion */
extern T1 T1487f32(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_class_assertion */
extern T1 T1486f32(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_class_assertion */
extern T1 T1485f19(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_class_assertion */
extern T1 T1484f19(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.is_class_assertion */
extern T1 T1483f25(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_class_assertion */
extern T1 T1482f18(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.is_class_assertion */
extern T1 T1270f19(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.is_class_assertion */
extern T1 T1267f16(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.is_class_assertion */
extern T1 T1254f22(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.is_class_assertion */
extern T1 T1218f18(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.is_class_assertion */
extern T1 T1232f13(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_static */
extern T1 T1513f41(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.has_static_mark */
extern T1 T1513f36(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_static */
extern T1 T1512f43(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.has_static_mark */
extern T1 T1512f38(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_static */
extern T1 T1511f41(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.has_static_mark */
extern T1 T1511f36(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_static */
extern T1 T371f58(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_implicitly_static */
extern T1 T371f49(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_static */
extern T1 T371f58p1(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.has_static_mark */
extern T1 T371f50(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_static */
extern T1 T369f61(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_implicitly_static */
extern T1 T369f48(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_static */
extern T1 T369f61p1(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.has_static_mark */
extern T1 T369f49(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_static */
extern T1 T359f55(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.has_static_mark */
extern T1 T359f51(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_static */
extern T1 T360f57(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.has_static_mark */
extern T1 T360f47(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.make */
extern T0* T635c28(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION.set_parenthesis_call */
extern void T1756f11(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.set_parenthesis_call */
extern void T635f32(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION.target */
extern T0* T1756f6(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.target */
extern T0* T635f18(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.set_seed */
extern void T622f134(GE_context* ac, T0* C, T6 a1);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.set_seed */
extern void T1545f45(GE_context* ac, T0* C, T6 a1);
/* detachable ET_INFIX_AND_THEN_OPERATOR.set_seed */
extern void T1544f45(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FREE_OPERATOR.set_seed */
extern void T973f63(GE_context* ac, T0* C, T6 a1);
/* detachable ET_KEYWORD_OPERATOR.set_seed */
extern void T981f64(GE_context* ac, T0* C, T6 a1);
/* detachable ET_SYMBOL_OPERATOR.set_seed */
extern void T994f157(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_PARENTHESIS_SYMBOL.set_seed */
extern void T2020f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_dotnet_query_call_expression_validity */
extern void T71f392(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.free_overloaded_queries */
extern void T71f444(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].force_last */
extern void T388f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].force */
extern void T718f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].fill_with */
extern void T720f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].extend */
extern void T720f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].resize */
extern void T388f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].aliased_resized_area */
extern T0* T718f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].aliased_resized_area */
extern T0* T720f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].new_capacity */
extern T6 T388f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].extendible */
extern T1 T388f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].first */
extern T0* T719f14(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.keep_best_overloaded_features */
extern void T71f308(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5, T1 a6);
/* detachable ET_ALIASED_FEATURE_NAME.same_call_name */
extern T1 T1517f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER.same_call_name */
extern T1 T184f105(GE_context* ac, T0* C, T0* a1);
extern T1 T184f105ot1(T0* a1, T0** a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].keep_first */
extern void T719f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].keep_first */
extern void T715f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.free_overloaded_features */
extern void T71f372(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].force_last */
extern void T389f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].force */
extern void T722f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].fill_with */
extern void T724f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].extend */
extern void T724f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].resize */
extern void T389f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].aliased_resized_area */
extern T0* T722f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].aliased_resized_area */
extern T0* T724f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].new_capacity */
extern T6 T389f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].extendible */
extern T1 T389f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].has */
extern T1 T723f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].remove */
extern void T723f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].move_left */
extern void T723f21(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].move_cursors_left */
extern void T723f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_FEATURE].set_position */
extern void T2056f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].remove_last */
extern void T723f19(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].move_last_cursors_after */
extern void T723f22(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_FEATURE].set_next_cursor */
extern void T2056f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_FEATURE].set_after */
extern void T2056f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].item */
extern T0* T723f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].wipe_out */
extern void T723f14(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].move_all_cursors_after */
extern void T723f18(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].force_last */
extern void T723f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].resize */
extern void T723f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].new_capacity */
extern T6 T723f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].extendible */
extern T1 T723f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.same_named_type */
extern T1 T241f13(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.new_overloaded_features */
extern T0* T71f87(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].remove_last */
extern void T389f15(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].keep_head */
extern void T724f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].set_count */
extern void T724f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].move_last_cursors_after */
extern void T389f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].set_next_cursor */
extern void T725f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].set_after */
extern void T725f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].last */
extern T0* T389f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].make */
extern T0* T723c12(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].set_internal_cursor */
extern void T723f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_FEATURE].new_cursor */
extern T0* T723f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_FEATURE].make */
extern T0* T2056c5(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].is_empty */
extern T1 T389f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].remove */
extern void T719f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].move_left */
extern void T719f41(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].move_cursors_left */
extern void T719f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUERY].set_position */
extern void T1426f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].remove_last */
extern void T719f39(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].move_last_cursors_after */
extern void T719f42(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].remove */
extern void T715f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].move_left */
extern void T715f41(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].move_cursors_left */
extern void T715f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PROCEDURE].set_position */
extern void T1425f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].remove_last */
extern void T715f39(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].move_last_cursors_after */
extern void T715f42(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.arguments_count */
extern T6 T1516f32(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.arguments_count */
extern T6 T476f33(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.arguments_count */
extern T6 T1515f33(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.arguments_count */
extern T6 T469f36(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.adapted_name */
extern T0* T71f85(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.renames */
extern T0* T1850f2(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].is_empty */
extern T1 T719f13(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.add_overloaded_queries */
extern void T254f130(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.add_overloaded_queries */
extern void T1850f14(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.add_overloaded_queries */
extern void T72f139(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.add_overloaded_queries */
extern void T74f329(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_QUERY_LIST.add_overloaded_features */
extern void T257f14(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T257f14ot1(T0* a1, T0** a2);
/* [detachable] ET_EXTERNAL_FUNCTION.overloaded_alias_names */
extern T0* T253f88(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.overloaded_extended_name */
extern T0* T253f89(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.overloaded_alias_names */
extern T0* T1513f79(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.overloaded_extended_name */
extern T0* T1513f80(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.overloaded_alias_names */
extern T0* T1512f82(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.overloaded_extended_name */
extern T0* T1512f83(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.overloaded_alias_names */
extern T0* T1511f80(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.overloaded_extended_name */
extern T0* T1511f81(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.overloaded_alias_names */
extern T0* T371f80(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.overloaded_extended_name */
extern T0* T371f81(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.overloaded_alias_names */
extern T0* T369f79(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.overloaded_extended_name */
extern T0* T369f80(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.overloaded_alias_names */
extern T0* T359f81(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.overloaded_extended_name */
extern T0* T359f82(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.overloaded_alias_names */
extern T0* T360f76(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.overloaded_extended_name */
extern T0* T360f77(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.overloaded_name */
extern T0* T253f87(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.overloaded_name */
extern T0* T1513f78(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.overloaded_name */
extern T0* T1512f81(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.overloaded_name */
extern T0* T1511f79(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.overloaded_name */
extern T0* T371f79(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.overloaded_name */
extern T0* T369f78(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.overloaded_name */
extern T0* T359f80(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.overloaded_name */
extern T0* T360f75(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.new_overloaded_queries */
extern T0* T71f94(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].remove_last */
extern void T388f15(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].keep_head */
extern void T720f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].set_count */
extern void T720f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].move_last_cursors_after */
extern void T388f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].set_next_cursor */
extern void T721f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].set_after */
extern void T721f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].last */
extern T0* T388f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].make */
extern T0* T719c15(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].set_internal_cursor */
extern void T719f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].new_cursor */
extern T0* T719f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUERY].make */
extern T0* T1426c9(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].is_empty */
extern T1 T388f9(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.set_boolean_operator */
extern void T1543f29(GE_context* ac, T0* C, T1 a1);
/* detachable ET_PREFIX_EXPRESSION.set_boolean_operator */
extern void T1547f29(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_CLASS.is_boolean_class */
extern T1 T74f186(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.has_implementation_error */
extern T1 T71f80(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.has_validity_error */
extern T1 T1516f27(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.has_validity_error */
extern T1 T1515f27(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.has_validity_error */
extern T1 T469f13(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.has_validity_error */
extern T1 T253f29(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.has_validity_error */
extern T1 T1513f26(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.has_validity_error */
extern T1 T1512f29(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.has_validity_error */
extern T1 T1511f28(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.has_validity_error */
extern T1 T371f23(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.has_validity_error */
extern T1 T369f21(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.has_validity_error */
extern T1 T359f26(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.has_validity_error */
extern T1 T360f19(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.has_validity_error */
extern T1 T476f13(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_orphan_actual_arguments_validity */
extern void T71f243(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_adapted_base_classes_validity */
extern void T71f242(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.new_adapted_base_classes */
extern T0* T71f79(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].remove_last */
extern void T404f15(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].keep_head */
extern void T761f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].set_count */
extern void T761f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].move_last_cursors_after */
extern void T404f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].set_next_cursor */
extern void T762f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].set_after */
extern void T762f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].last */
extern T0* T404f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].make */
extern T0* T710c12(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].set_internal_cursor */
extern void T710f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS].new_cursor */
extern T0* T710f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ADAPTED_CLASS].make */
extern T0* T1424c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_CLASS].make */
extern T0* T1421f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_CLASS].make_empty */
extern T0* T1423c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_CLASS].default_create */
extern T0* T1421c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].is_empty */
extern T1 T404f9(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.current_system */
extern T0* T71f81(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.expression */
extern T0* T2048f3(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.target */
extern T0* T1038f2(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.expression */
extern T0* T1547f3(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.left */
extern T0* T1543f2(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_EXPRESSION.target */
extern T0* T888f1(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.target */
extern T0* T287f1(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.process_convert_to_expression */
extern void T380f457(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_convert_to_expression_validity */
extern void T380f460(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_call_expression_validity */
extern void T380f208(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T380f208ot1(T0* a1, T0** a2);
extern T1 T380f208ot2(T0* a1, T0** a2);
extern T1 T380f208ot3(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.free_adapted_base_classes */
extern void T380f272(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_tuple_label_call_expression_validity */
extern void T380f271(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.report_tuple_label_expression */
extern void T380f338(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_tuple_label_parenthesis_call_validity */
extern void T380f270(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T380f270ot1(T0* a1, T0** a2);
extern T1 T380f270ot2(T0* a1, T0** a2);
extern T1 T380f270ot3(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_feature_call_expression_validity */
extern void T380f333(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T380f333ot1(T0* a1, T0** a2);
extern T1 T380f333ot2(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_feature_call_instruction_validity */
extern void T380f332(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T380f332ot1(T0* a1, T0** a2);
extern T1 T380f332ot2(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_procedure_call_instruction_validity */
extern void T380f375(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.report_qualified_call_instruction */
extern void T380f392(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_feature_call_validity */
extern void T380f334(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FEATURE_CHECKER.reset_fatal_error */
extern void T380f337(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_FEATURE_CHECKER.check_actual_arguments_validity */
extern void T380f341(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T380f341ot1(T0* a1, T0** a2);
extern T1 T380f341ot2(T0* a1);
extern T1 T380f341ot3(T0* a1, T0** a2);
extern T1 T380f341ot4(T0* a1);
extern T1 T380f341ot5(T0* a1, T0** a2);
extern T1 T380f341ot6(T0* a1, T0** a2);
extern T1 T380f341ot7(T0* a1, T0** a2);
extern T1 T380f341ot8(T0* a1, T0** a2);
extern T1 T380f341ot9(T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_converted_target_infix_expression_validity */
extern void T380f377(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FEATURE_CHECKER.convert_expression */
extern T0* T380f64(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_builtin_conversion */
extern void T380f397(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_qualified_call_expression */
extern void T380f335(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_creation_expression */
extern void T380f383(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_creation_vape_validity */
extern void T380f394(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_INVARIANTS.as_feature */
extern T0* T1198f15(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.add_non_descendant_caller_clients_to */
extern void T380f347(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_INVARIANTS.clients */
extern T0* T1198f19(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.tokens */
extern T0* T1198f9(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.is_feature */
extern T1 T1198f14(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.current_universe */
extern T0* T380f67(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.check_actual_argument_validity */
extern void T380f376(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T380f376ot1(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.free_attachment_scope */
extern void T380f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.new_attachment_scope */
extern T0* T380f52(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.tuple_argument_position */
extern T6 T380f69(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_vape_validity */
extern void T380f340(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.set_parenthesis_call_position */
extern void T380f331(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_query_call_expression_validity */
extern void T380f269(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FEATURE_CHECKER.check_query_call_type_validity */
extern void T380f336(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T380f336ot1(T0* a1, T0** a2);
extern T1 T380f336ot2(T0* a1, T0** a2);
extern T1 T380f336ot3(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.check_query_parenthesis_call_validity */
extern void T380f268(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T380f268ot1(T0* a1, T0** a2);
extern T1 T380f268ot2(T0* a1, T0** a2);
extern T1 T380f268ot3(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_query_call_expression_validity */
extern void T380f323(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_unqualified_call_expression */
extern void T380f370(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_vape_validity */
extern void T380f285(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_dotnet_query_call_expression_validity */
extern void T380f267(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.free_overloaded_queries */
extern void T380f330(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.keep_best_overloaded_features */
extern void T380f329(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5, T1 a6);
/* [detachable] ET_FEATURE_CHECKER.free_overloaded_features */
extern void T380f374(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.new_overloaded_features */
extern T0* T380f68(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.adapted_name */
extern T0* T380f66(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.new_overloaded_queries */
extern T0* T380f65(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.has_implementation_error */
extern T1 T380f59(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INVARIANTS.has_validity_error */
extern T1 T1198f8(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.postconditions */
extern T0* T1198f24(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.preconditions */
extern T0* T1198f23(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.implementation_feature */
extern T0* T1198f22(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.check_orphan_actual_arguments_validity */
extern void T380f266(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_adapted_base_classes_validity */
extern void T380f265(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.new_adapted_base_classes */
extern T0* T380f58(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.current_system */
extern T0* T380f56(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.process_convert_to_expression */
extern void T67f1485(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_convert_to_expression */
extern void T67f1488(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_qualified_call_expression */
extern void T67f765(GE_context* ac, T0* C, T0* a1);
extern T1 T67f765ot1(T0* a1, T0** a2);
extern T1 T67f765ot2(T0* a1, T0** a2);
extern T1 T67f765ot3(T0* a1, T0** a2);
/* [detachable] ET_MANIFEST_TUPLE.wipe_out */
extern void T288f32(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION_ITEM].keep_head */
extern void T671f4(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION_ITEM].keep_head */
extern void T669f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION_ITEM].set_count */
extern void T669f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.mark_expressions_unfrozen */
extern void T67f793(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.mark_temp_variable_unfrozen */
extern void T67f869(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.is_temp_variable_frozen */
extern T1 T67f302(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_semicolon_newline */
extern void T67f678(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_call_name */
extern void T67f723(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4, T0* a5);
extern T1 T67f723ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.dynamic_type_set_in_feature */
extern T0* T67f277(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT.is_current */
extern T1 T2027f5(GE_context* ac, T0* C);
/* detachable ET_AGENT_TYPED_OPEN_ARGUMENT.is_current */
extern T1 T1559f9(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_current */
extern T1 T1005f39(GE_context* ac, T0* C);
/* detachable ET_ASSIGNER_INSTRUCTION.is_call_agent */
extern T1 T1527f11(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_INSTRUCTION.is_call_agent */
extern T1 T2002f6(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.is_call_agent */
extern T1 T1041f26(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.is_call_agent */
extern T1 T284f7(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.register_polymorphic_called_features */
extern void T67f792(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.set_generated */
extern void T318f19(GE_context* ac, T0* C, T1 a1);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.set_generated */
extern void T314f18(GE_context* ac, T0* C, T1 a1);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.set_force_result_boxing */
extern void T314f17(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_C_GENERATOR.print_attribute_type_id_access */
extern void T67f713(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_attribute_type_id_name */
extern void T67f688(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_non_void_expression */
extern void T67f699(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] ET_C_GENERATOR.print_check_void_expression */
extern void T67f739(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_INFIX_CAST_EXPRESSION.is_never_void */
extern T1 T1980f8(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.is_never_void */
extern T1 T2048f6(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.is_never_void */
extern T1 T2047f7(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.is_never_void */
extern T1 T2046f5(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_never_void */
extern T1 T1489f42(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_never_void */
extern T1 T1488f42(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_never_void */
extern T1 T1487f42(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_never_void */
extern T1 T1486f42(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_never_void */
extern T1 T372f31(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_never_void */
extern T1 T1485f32(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_never_void */
extern T1 T1484f28(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_never_void */
extern T1 T1482f27(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.is_never_void */
extern T1 T1254f27(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.is_never_void */
extern T1 T1008f29(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.is_never_void */
extern T1 T736f18(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.is_never_void */
extern T1 T1549f21(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.is_never_void */
extern T1 T1557f27(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.is_never_void */
extern T1 T1556f34(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.is_never_void */
extern T1 T1555f30(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.is_never_void */
extern T1 T1554f28(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.is_never_void */
extern T1 T1553f35(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.is_never_void */
extern T1 T1552f31(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.is_never_void */
extern T1 T1041f20(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.is_never_void */
extern T1 T1218f21(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.is_never_void */
extern T1 T1232f16(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.is_never_void */
extern T1 T1038f16(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.is_never_void */
extern T1 T311f17(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.is_never_void */
extern T1 T1548f14(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.is_never_void */
extern T1 T1547f17(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.is_never_void */
extern T1 T312f14(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.is_never_void */
extern T1 T313f14(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.is_never_void */
extern T1 T1543f17(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.is_never_void */
extern T1 T1542f14(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.is_never_void */
extern T1 T1538f18(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.is_never_void */
extern T1 T1267f18(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.is_never_void */
extern T1 T287f18(GE_context* ac, T0* C);
/* detachable ET_INSPECT_EXPRESSION.is_never_void */
extern T1 T1189f16(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.is_never_void */
extern T1 T1167f20(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.is_never_void */
extern T1 T328f18(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.is_never_void */
extern T1 T1483f18(GE_context* ac, T0* C);
/* detachable ET_ITERATION_CURSOR.is_never_void */
extern T1 T1551f10(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.is_never_void */
extern T1 T1550f10(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.is_never_void */
extern T1 T1546f11(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.is_never_void */
extern T1 T1541f8(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.is_never_void */
extern T1 T1540f8(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.is_never_void */
extern T1 T1539f8(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_never_void */
extern T1 T1496f15(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_never_void */
extern T1 T1495f15(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.is_never_void */
extern T1 T1494f16(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.is_never_void */
extern T1 T1493f16(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.is_never_void */
extern T1 T1491f16(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.is_never_void */
extern T1 T1490f16(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.is_never_void */
extern T1 T1270f14(GE_context* ac, T0* C);
/* detachable ET_VOID.is_never_void */
extern T1 T997f16(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.is_never_void */
extern T1 T635f8(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.is_never_void */
extern T1 T288f14(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.is_never_void */
extern T1 T309f6(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.is_never_void */
extern T1 T282f6(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_never_void */
extern T1 T184f8(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_object_test_local */
extern T1 T184f16(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.can_be_void */
extern T1 T1466f10(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.can_be_void */
extern T1 T269f9(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.can_be_void */
extern T1 T475f11(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.can_be_void */
extern T1 T323f8(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.can_be_void */
extern T1 T451f49(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.can_be_void */
extern T1 T450f52(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.can_be_void */
extern T1 T319f53(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.can_be_void */
extern T1 T302f52(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.can_be_void */
extern T1 T223f61(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_adapted_named_query_call */
extern void T67f725(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4);
/* Creation of agent #1 in feature [detachable] ET_C_GENERATOR.print_adapted_named_query_call */
extern T0* T67f725ac1(GE_context* ac, T0* a1, T6 a2, T0* a3, T0* a4, T1 a5);
extern T1 T67f725ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_attribute_tuple_item_access */
extern void T67f660(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3, T1 a4);
/* [detachable] ET_C_GENERATOR.print_attribute_tuple_item_name */
extern void T67f691(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.print_type_cast */
extern void T67f643(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_adapted_query_call */
extern void T67f745(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4);
/* Creation of agent #1 in feature [detachable] ET_C_GENERATOR.print_adapted_query_call */
extern T0* T67f745ac1(GE_context* ac, T0* a1, T0* a2, T0* a3, T1 a4);
extern T1 T67f745ot1(T0* a1, T0** a2);
extern T1 T67f745ot2(T0* a1);
/* [detachable] ET_C_GENERATOR.print_query_call */
extern void T67f761(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
extern T1 T67f761ot1(T0* a1, T0** a2);
extern T1 T67f761ot2(T0* a1, T0** a2);
extern T1 T67f761ot3(T0* a1, T0** a2);
extern T1 T67f761ot4(T0* a1, T0** a2);
extern T1 T67f761ot5(T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_query_call */
extern void T67f772(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_type_query_call */
extern void T67f818(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_type_runtime_name_call */
extern void T67f1078(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_type_type_id_call */
extern void T67f1077(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_type_is_expanded_call */
extern void T67f1076(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_type_is_deferred_call */
extern void T67f1075(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_type_is_attached_call */
extern void T67f1074(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_type_has_default_call */
extern void T67f1073(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_type_detachable */
extern T1 T1501f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.is_type_detachable */
extern T1 T891f44(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_type_detachable */
extern T1 T1526f22(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_type_detachable */
extern T1 T1525f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.is_type_detachable */
extern T1 T1524f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_type_detachable */
extern T1 T242f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.is_type_detachable */
extern T1 T449f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_type_detachable */
extern T1 T74f135(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.is_type_detachable */
extern T1 T254f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.is_type_detachable */
extern T1 T72f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_type_generic_parameter_count_call */
extern void T67f1072(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_type_default_call */
extern void T67f1071(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_query_call */
extern void T67f817(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_object_comparison_call */
extern void T67f1070(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_count_call */
extern void T67f1069(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
extern T1 T67f1069ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_special_query_call */
extern void T67f816(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_special_item_call */
extern void T67f1068(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_attribute_special_item_access */
extern void T67f659(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_attribute_special_item_name */
extern void T67f672(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_builtin_special_element_size_call */
extern void T67f1067(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
extern T1 T67f1067ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_special_count_call */
extern void T67f1066(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_attribute_special_count_access */
extern void T67f646(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_attribute_special_count_name */
extern void T67f701(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_attribute_name */
extern void T67f671(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.first */
extern T0* T224f4(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_builtin_special_capacity_call */
extern void T67f1065(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_attribute_special_capacity_access */
extern void T67f645(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_attribute_special_capacity_name */
extern void T67f700(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_builtin_special_base_address_call */
extern void T67f1064(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ref_query_call */
extern void T67f815(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ref_positive_infinity_call */
extern void T67f1063(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ref_negative_infinity_call */
extern void T67f1062(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ref_nan_call */
extern void T67f1061(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ref_item_call */
extern void T67f1060(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_unboxed_expression */
extern void T67f717(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_boxed_attribute_pointer_access */
extern void T67f683(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_boxed_attribute_pointer_name */
extern void T67f692(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_query_call */
extern void T67f814(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_truncated_to_real_call */
extern void T67f1059(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_truncated_to_integer_64_call */
extern void T67f1058(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_truncated_to_integer_call */
extern void T67f1057(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_to_double_call */
extern void T67f1056(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_quotient_call */
extern void T67f1055(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_product_call */
extern void T67f1054(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_power_call */
extern void T67f1053(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_plus_call */
extern void T67f1052(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_opposite_call */
extern void T67f1051(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_minus_call */
extern void T67f1050(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_is_positive_infinity_call */
extern void T67f1049(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_is_negative_infinity_call */
extern void T67f1048(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_is_nan_call */
extern void T67f1047(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_is_less_call */
extern void T67f1046(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_query_c_call */
extern void T67f1084(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3, T0* a4, T1 a5);
/* [detachable] ET_C_GENERATOR.total_order_on_reals_mode */
extern T1 T67f294(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ieee_minimum_number_call */
extern void T67f1045(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ieee_maximum_number_call */
extern void T67f1044(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ieee_is_less_equal_call */
extern void T67f1043(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ieee_is_less_call */
extern void T67f1042(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ieee_is_greater_equal_call */
extern void T67f1041(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ieee_is_greater_call */
extern void T67f1040(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ieee_is_equal_call */
extern void T67f1039(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_identity_call */
extern void T67f1038(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_floor_real_64_call */
extern void T67f1037(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_floor_real_32_call */
extern void T67f1036(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ceiling_real_64_call */
extern void T67f1035(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ceiling_real_32_call */
extern void T67f1034(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_pointer_ref_query_call */
extern void T67f813(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_pointer_ref_item_call */
extern void T67f1033(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_pointer_query_call */
extern void T67f812(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_pointer_to_integer_32_call */
extern void T67f1032(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_pointer_plus_call */
extern void T67f1031(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_pointer_is_default_pointer_call */
extern void T67f1030(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.current_universe_impl */
extern T0* T67f298(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_builtin_pointer_hash_code_call */
extern void T67f1029(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_query_call */
extern void T67f811(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_wide_character_bytes_call */
extern void T67f1028(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_real_bytes_call */
extern void T67f1027(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_pointer_bytes_call */
extern void T67f1026(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_is_windows_call */
extern void T67f1025(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_is_vxworks_call */
extern void T67f1024(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_is_vms_call */
extern void T67f1023(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_is_unix_call */
extern void T67f1022(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_is_thread_capable_call */
extern void T67f1021(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_is_scoop_capable_call */
extern void T67f1020(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_is_mac_call */
extern void T67f1019(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_is_dotnet_call */
extern void T67f1018(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_is_64_bits_call */
extern void T67f1017(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_integer_bytes_call */
extern void T67f1016(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_double_bytes_call */
extern void T67f1015(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_character_bytes_call */
extern void T67f1014(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_platform_boolean_bytes_call */
extern void T67f1013(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_memory_query_call */
extern void T67f810(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_memory_find_referers_call */
extern void T67f1012(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_query_call */
extern void T67f809(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_type_id_from_name_call */
extern void T67f1011(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_type_conforms_to_call */
extern void T67f1010(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_storable_version_of_type_call */
extern void T67f1009(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_reference_field_at_offset_call */
extern void T67f1008(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_reference_field_at_call */
extern void T67f1007(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_reference_field_call */
extern void T67f1006(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_real_64_field_at_call */
extern void T67f1005(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_real_64_field_call */
extern void T67f1004(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_real_32_field_at_call */
extern void T67f1003(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_real_32_field_call */
extern void T67f1002(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_raw_reference_field_at_offset_call */
extern void T67f1001(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_raw_reference_field_at_call */
extern void T67f1000(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_pointer_field_at_call */
extern void T67f999(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_pointer_field_call */
extern void T67f998(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_persistent_field_count_of_type_call */
extern void T67f997(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_once_objects_call */
extern void T67f996(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_object_size_call */
extern void T67f995(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_new_type_instance_of_call */
extern void T67f994(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_new_tuple_instance_of_call */
extern void T67f993(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_new_special_of_reference_instance_of_call */
extern void T67f992(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_new_instance_of_call */
extern void T67f991(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_natural_64_field_at_call */
extern void T67f990(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_natural_64_field_call */
extern void T67f989(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_natural_32_field_at_call */
extern void T67f988(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_natural_32_field_call */
extern void T67f987(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_natural_16_field_at_call */
extern void T67f986(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_natural_16_field_call */
extern void T67f985(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_natural_8_field_at_call */
extern void T67f984(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_natural_8_field_call */
extern void T67f983(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_tuple_type_call */
extern void T67f982(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_tuple_call */
extern void T67f981(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_special_of_reference_type_call */
extern void T67f980(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_special_of_reference_or_basic_type_call */
extern void T67f979(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_special_of_reference_call */
extern void T67f978(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_special_of_expanded_call */
extern void T67f977(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_special_copy_semantics_item_call */
extern void T67f976(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_special_call */
extern void T67f975(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_object_marked_call */
extern void T67f974(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_field_transient_of_type_call */
extern void T67f973(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_field_expanded_of_type_call */
extern void T67f972(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_expanded_call */
extern void T67f971(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_copy_semantics_field_call */
extern void T67f970(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_is_attached_type_call */
extern void T67f969(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_integer_64_field_at_call */
extern void T67f968(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_integer_64_field_call */
extern void T67f967(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_integer_32_field_at_call */
extern void T67f966(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_integer_32_field_call */
extern void T67f965(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_integer_16_field_at_call */
extern void T67f964(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_integer_16_field_call */
extern void T67f963(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_integer_8_field_at_call */
extern void T67f962(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_integer_8_field_call */
extern void T67f961(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_generic_parameter_count_call */
extern void T67f960(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_generator_8_of_type_call */
extern void T67f959(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_generator_of_type_call */
extern void T67f958(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_generating_type_8_of_type_call */
extern void T67f957(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_generating_type_of_type_call */
extern void T67f956(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_field_type_of_type_call */
extern void T67f955(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_field_static_type_of_type_call */
extern void T67f954(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_field_offset_of_type_call */
extern void T67f953(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_field_name_of_type_call */
extern void T67f952(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_field_count_of_type_call */
extern void T67f951(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_eif_gen_param_id_call */
extern void T67f950(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_dynamic_type_at_offset_call */
extern void T67f949(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_dynamic_type_call */
extern void T67f948(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_detachable_type_call */
extern void T67f947(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_compiler_version_call */
extern void T67f946(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_check_assert_call */
extern void T67f945(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_character_32_field_at_call */
extern void T67f944(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_character_32_field_call */
extern void T67f943(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_character_8_field_at_call */
extern void T67f942(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_character_8_field_call */
extern void T67f941(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_boolean_field_at_call */
extern void T67f940(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_boolean_field_call */
extern void T67f939(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_attached_type_call */
extern void T67f938(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_ref_query_call */
extern void T67f808(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_ref_item_call */
extern void T67f937(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_query_call */
extern void T67f807(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_to_real_64_call */
extern void T67f936(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_to_real_32_call */
extern void T67f935(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_to_real_call */
extern void T67f934(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_to_double_call */
extern void T67f933(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_to_character_32_call */
extern void T67f932(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_to_character_8_call */
extern void T67f931(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_quotient_call */
extern void T67f930(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_product_call */
extern void T67f929(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_power_call */
extern void T67f928(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_plus_call */
extern void T67f927(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_opposite_call */
extern void T67f926(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_minus_call */
extern void T67f925(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_is_less_call */
extern void T67f924(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_integer_remainder_call */
extern void T67f923(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_integer_quotient_call */
extern void T67f922(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_identity_call */
extern void T67f921(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_bit_xor_call */
extern void T67f920(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_bit_shift_right_call */
extern void T67f919(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_bit_shift_left_call */
extern void T67f918(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_bit_or_call */
extern void T67f917(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_bit_not_call */
extern void T67f916(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_bit_and_call */
extern void T67f915(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_as_natural_64_call */
extern void T67f914(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_as_natural_32_call */
extern void T67f913(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_as_natural_16_call */
extern void T67f912(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_as_natural_8_call */
extern void T67f911(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_as_integer_64_call */
extern void T67f910(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_as_integer_32_call */
extern void T67f909(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_as_integer_16_call */
extern void T67f908(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_as_integer_8_call */
extern void T67f907(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_identified_routines_query_call */
extern void T67f806(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_identified_routines_eif_object_id_call */
extern void T67f906(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_identified_routines_eif_is_object_id_of_current_call */
extern void T67f905(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_identified_routines_eif_id_object_call */
extern void T67f904(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_identified_routines_eif_current_object_id_call */
extern void T67f903(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_function_query_call */
extern void T67f805(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_function_item_call */
extern void T67f902(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_routine_call_call */
extern void T67f1082(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
extern T1 T67f1082ot1(T0* a1, T0** a2);
extern T1 T67f1082ot2(T0* a1, T0** a2);
extern T1 T67f1082ot3(T0* a1, T0** a2);
extern T1 T67f1082ot4(T0* a1, T0** a2);
extern T1 T67f1082ot5(T0* a1, T0** a2);
extern T1 T67f1082ot6(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.free_standalone_type_set */
extern void T67f714(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.reset */
extern void T269f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.wipe_out */
extern void T268f20(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRIMARY_TYPE].keep_head */
extern void T416f5(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRIMARY_TYPE].keep_head */
extern void T415f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRIMARY_TYPE].set_count */
extern void T415f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].force_last */
extern void T270f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].force */
extern void T648f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].fill_with */
extern void T647f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].extend */
extern void T647f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].resize */
extern void T270f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].aliased_resized_area */
extern T0* T648f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].aliased_resized_area */
extern T0* T647f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].new_capacity */
extern T6 T270f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].extendible */
extern T1 T270f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.mark_call_operands_unfrozen */
extern void T67f1089(GE_context* ac, T0* C);
extern T1 T67f1089ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.keep_first */
extern void T259f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE_SET].keep_head */
extern void T627f5(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE_SET].keep_head */
extern void T626f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE_SET].set_count */
extern void T626f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.free_tuple_label */
extern void T67f1088(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.free_identifier */
extern void T67f764(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER].force_last */
extern void T261f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_IDENTIFIER].force */
extern void T599f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_IDENTIFIER].fill_with */
extern void T600f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_IDENTIFIER].extend */
extern void T600f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER].resize */
extern void T261f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_IDENTIFIER].aliased_resized_area */
extern T0* T599f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_IDENTIFIER].aliased_resized_area */
extern T0* T600f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER].new_capacity */
extern T6 T261f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER].extendible */
extern T1 T261f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.free_qualified_call_expression */
extern void T67f750(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.set_target */
extern void T287f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUALIFIED_CALL_EXPRESSION].force_last */
extern void T265f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_QUALIFIED_CALL_EXPRESSION].force */
extern void T639f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_QUALIFIED_CALL_EXPRESSION].fill_with */
extern void T638f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_QUALIFIED_CALL_EXPRESSION].extend */
extern void T638f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUALIFIED_CALL_EXPRESSION].resize */
extern void T265f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_QUALIFIED_CALL_EXPRESSION].aliased_resized_area */
extern T0* T639f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_QUALIFIED_CALL_EXPRESSION].aliased_resized_area */
extern T0* T638f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUALIFIED_CALL_EXPRESSION].new_capacity */
extern T6 T265f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUALIFIED_CALL_EXPRESSION].extendible */
extern T1 T265f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.make */
extern T0* T269c17(GE_context* ac, T0* a1);
/* [detachable] ET_C_GENERATOR.new_qualified_call_expression */
extern T0* T67f301(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.set_name */
extern void T287f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUALIFIED_CALL_EXPRESSION].remove_last */
extern void T265f13(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_QUALIFIED_CALL_EXPRESSION].keep_head */
extern void T638f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_QUALIFIED_CALL_EXPRESSION].set_count */
extern void T638f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUALIFIED_CALL_EXPRESSION].move_last_cursors_after */
extern void T265f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUALIFIED_CALL_EXPRESSION].set_next_cursor */
extern void T640f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUALIFIED_CALL_EXPRESSION].set_after */
extern void T640f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUALIFIED_CALL_EXPRESSION].last */
extern T0* T265f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUALIFIED_CALL_EXPRESSION].is_empty */
extern T1 T265f10(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.new_tuple_label */
extern T0* T67f484(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.new_identifier */
extern T0* T67f493(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER.set_name */
extern void T184f124(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER.new_hash_code */
extern T6 T184f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER].remove_last */
extern void T261f16(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_IDENTIFIER].keep_head */
extern void T600f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_IDENTIFIER].set_count */
extern void T600f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER].move_last_cursors_after */
extern void T261f18(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_IDENTIFIER].set_next_cursor */
extern void T628f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_IDENTIFIER].set_after */
extern void T628f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER].last */
extern T0* T261f11(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER].is_empty */
extern T1 T261f7(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.put_types */
extern void T269f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.make_empty */
extern T0* T269c13(GE_context* ac, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.result_type_set */
extern T0* T450f47(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_comma */
extern void T67f758(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.mark_temp_variable_free */
extern void T67f648(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE].replace */
extern void T262f11(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE].item */
extern T0* T262f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_assign_to */
extern void T67f681(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.conforming_type_set */
extern T0* T67f483(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.is_never_void */
extern T1 T1466f3(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.is_never_void */
extern T1 T269f4(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.is_never_void */
extern T1 T475f3(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.new_standalone_type_set */
extern T0* T67f264(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].remove_last */
extern void T270f13(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].keep_head */
extern void T647f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].set_count */
extern void T647f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].move_last_cursors_after */
extern void T270f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].set_next_cursor */
extern void T649f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].set_after */
extern void T649f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].last */
extern T0* T270f6(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.reset_with_types */
extern void T269f21(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].is_empty */
extern T1 T270f10(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.mark_call_operands_frozen */
extern void T67f1087(GE_context* ac, T0* C);
extern T1 T67f1087ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.mark_temp_variable_frozen */
extern void T67f868(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_PRECURSOR.is_procedure */
extern T1 T306f39(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.is_procedure */
extern T1 T236f29(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_builtin_function_fast_item_call */
extern void T67f901(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_com_failure_query_call */
extern void T67f804(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_com_failure_cwin_error_text_call */
extern void T67f900(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_com_failure_character_size_call */
extern void T67f899(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_com_failure_ccom_hresult_facility_call */
extern void T67f898(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_com_failure_ccom_hresult_code_call */
extern void T67f897(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_com_failure_ccom_hresult_call */
extern void T67f896(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_com_failure_c_strlen_call */
extern void T67f895(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_character_n_ref_query_call */
extern void T67f803(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_character_n_ref_item_call */
extern void T67f894(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_character_n_query_call */
extern void T67f802(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_character_n_to_character_32_call */
extern void T67f893(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_character_n_to_character_8_call */
extern void T67f892(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_character_n_natural_32_code_call */
extern void T67f891(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_character_n_code_call */
extern void T67f890(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_ref_query_call */
extern void T67f801(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_ref_item_call */
extern void T67f889(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_query_call */
extern void T67f800(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_xor_call */
extern void T67f888(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_or_else_call */
extern void T67f887(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_or_call */
extern void T67f886(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_not_call */
extern void T67f885(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_implies_call */
extern void T67f884(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_and_then_call */
extern void T67f883(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_and_call */
extern void T67f882(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_arguments_32_query_call */
extern void T67f799(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_arguments_32_i_th_argument_string_call */
extern void T67f881(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_arguments_32_i_th_argument_pointer_call */
extern void T67f880(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_arguments_32_argument_count_call */
extern void T67f879(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_any_query_call */
extern void T67f798(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_any_twin_call */
extern void T67f878(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_CLASS.is_type_class */
extern T1 T74f183(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.has_redefined_copy_routine */
extern T1 T451f11(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.has_redefined_copy_routine */
extern T1 T450f11(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.has_redefined_copy_routine */
extern T1 T319f14(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.has_redefined_copy_routine */
extern T1 T302f13(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_redefined_copy_routine */
extern T1 T223f13(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_builtin_any_standard_twin_call */
extern void T67f877(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.has_nested_custom_standard_copy_routine */
extern T1 T451f28(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.has_expanded */
extern T1 T1466f9(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.has_expanded */
extern T1 T269f11(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.has_expanded */
extern T1 T475f9(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.has_expanded */
extern T1 T451f53(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.has_expanded */
extern T1 T450f54(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.has_expanded */
extern T1 T302f38(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.has_expanded */
extern T1 T319f39(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_expanded */
extern T1 T223f27(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.has_expanded */
extern T1 T323f9(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.has_nested_custom_standard_copy_routine */
extern T1 T450f28(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.has_nested_custom_standard_copy_routine */
extern T1 T319f44(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.has_item_nested_custom_standard_copy_routine */
extern T1 T319f47(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.has_nested_custom_standard_copy_routine */
extern T1 T319f44p1(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.has_nested_custom_standard_copy_routine */
extern T1 T302f44(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.has_nested_custom_standard_copy_routine */
extern T1 T302f44p1(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_nested_custom_standard_copy_routine */
extern T1 T223f40(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_builtin_any_standard_is_equal_call */
extern void T67f876(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_any_standard_is_equal_call_with_reference */
extern void T67f1086(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T1 a7);
extern T1 T67f1086ot1(T0* a1, T0** a2);
extern T1 T67f1086ot2(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_minus */
extern void T67f1080(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_times */
extern void T67f1092(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_attribute_special_offset_access */
extern void T67f1091(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_attribute_special_offset_name */
extern void T67f1093(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_plus */
extern void T67f1079(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.is_special */
extern T1 T451f30(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.is_special */
extern T1 T450f30(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.is_special */
extern T1 T319f43(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.is_special */
extern T1 T302f43(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_special */
extern T1 T223f39(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_builtin_any_standard_is_equal_call_with_expanded */
extern void T67f1085(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T1 a7);
/* [detachable] ET_C_GENERATOR.print_equal_to */
extern void T67f1083(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.has_nested_custom_standard_is_equal_routine */
extern T1 T451f64(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.has_redefined_is_equal_routine */
extern T1 T451f10(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.has_redefined_is_equal_routine */
extern T1 T450f10(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.has_redefined_is_equal_routine */
extern T1 T302f15(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.has_redefined_is_equal_routine */
extern T1 T319f15(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_redefined_is_equal_routine */
extern T1 T223f14(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.has_nested_custom_standard_is_equal_routine */
extern T1 T450f65(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.has_nested_custom_standard_is_equal_routine */
extern T1 T319f42(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.has_item_nested_custom_standard_is_equal_routine */
extern T1 T319f45(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.has_nested_custom_standard_is_equal_routine */
extern T1 T319f42p1(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.has_nested_custom_standard_is_equal_routine */
extern T1 T302f42(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.has_nested_custom_standard_is_equal_routine */
extern T1 T302f42p1(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_nested_custom_standard_is_equal_routine */
extern T1 T223f38(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.has_type */
extern T1 T269f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.has_type */
extern T1 T269f7p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.index_of */
extern T6 T269f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.index_of */
extern T6 T269f8p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_any_same_type_call */
extern void T67f875(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_any_is_equal_call */
extern void T67f874(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_any_generator_call */
extern void T67f873(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_any_generating_type_call */
extern void T67f872(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_any_deep_twin_call */
extern void T67f871(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].force_last */
extern void T275f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].slots_put */
extern void T275f59(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_put */
extern void T275f58(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].resize */
extern void T275f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_resize */
extern void T275f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].special_integer_ */
extern T0* T275f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_resize */
extern void T275f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_resize */
extern void T275f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].slots_resize */
extern void T275f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].new_modulus */
extern T6 T275f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].new_capacity */
extern T6 T275f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_put */
extern void T275f56(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].unset_found_item */
extern void T275f47(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_non_inlined_query_call */
extern void T67f771(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.is_self_initializing */
extern T1 T451f56(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.is_self_initializing */
extern T1 T450f49(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.is_self_initializing */
extern T1 T323f12(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.is_self_initializing */
extern T1 T319f30(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.is_self_initializing */
extern T1 T302f30(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_self_initializing */
extern T1 T223f48(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.has_self_initializing_code */
extern T1 T359f49(GE_context* ac, T0* C);
/* detachable ET_COMPOUND.has_non_null_instruction */
extern T1 T355f8(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.to_integer_32 */
extern T6 T372f12(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.remove_last */
extern void T259f12(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_once_value_name */
extern void T67f663(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].force_last */
extern void T292f43(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].key_storage_put */
extern void T292f58(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].slots_put */
extern void T292f57(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].clashes_put */
extern void T292f56(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].slots_item */
extern T6 T292f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].hash_position */
extern T6 T292f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].resize */
extern void T292f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].clashes_resize */
extern void T292f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].special_integer_ */
extern T0* T292f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].key_storage_resize */
extern void T292f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].item_storage_resize */
extern void T292f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTANT].aliased_resized_area */
extern T0* T682f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_CONSTANT].aliased_resized_area */
extern T0* T680f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].key_storage_item */
extern T0* T292f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].clashes_item */
extern T6 T292f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].slots_resize */
extern void T292f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].new_modulus */
extern T6 T292f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].new_capacity */
extern T6 T292f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].item_storage_put */
extern void T292f54(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTANT].force */
extern void T682f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CONSTANT].fill_with */
extern void T680f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CONSTANT].extend */
extern void T680f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].search_position */
extern void T292f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].unset_found_item */
extern void T292f46(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.is_builtin_special_item */
extern T1 T306f33(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_adapted_once_function_call */
extern void T67f760(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4);
/* Creation of agent #1 in feature [detachable] ET_C_GENERATOR.print_adapted_once_function_call */
extern T0* T67f760ac1(GE_context* ac, T0* a1, T0* a2, T6 a3, T6 a4);
/* Creation of agent #2 in feature [detachable] ET_C_GENERATOR.print_adapted_once_function_call */
extern T0* T67f760ac2(GE_context* ac, T0* a1, T0* a2, T0* a3, T1 a4);
/* [detachable] ET_C_GENERATOR.print_once_value */
extern void T67f770(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_once_per_thread_value */
extern void T67f797(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.once_prefixes */
extern T0* T67f311(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.is_once_per_thread */
extern T1 T306f24(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_once_per_thread */
extern T1 T1516f44(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_once_per_thread */
extern T1 T476f46(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_once_per_thread */
extern T1 T1515f48(GE_context* ac, T0* C);
/* detachable ET_INDEXING_LIST.has_indexing_term_with_tag */
extern T1 T1173f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_STANDARD_ONCE_KEYS.has_thread_key */
extern T1 T1857f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.is_once_per_thread */
extern T1 T469f46(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_once_per_thread */
extern T1 T253f60(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_once_per_thread */
extern T1 T1513f32(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_once_per_thread */
extern T1 T1512f57(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_once_per_thread */
extern T1 T1511f32(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_once_per_thread */
extern T1 T371f51(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_once_per_thread */
extern T1 T369f50(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_once_per_thread */
extern T1 T359f47(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_once_per_thread */
extern T1 T360f48(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.is_once_per_thread */
extern T1 T236f35(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_once_per_process_value */
extern void T67f796(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* detachable ET_DYNAMIC_PRECURSOR.is_once_per_process */
extern T1 T306f23(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_once_per_process */
extern T1 T1516f41(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_once_per_process */
extern T1 T476f45(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_once_per_process */
extern T1 T1515f44(GE_context* ac, T0* C);
/* [detachable] ET_STANDARD_ONCE_KEYS.has_process_key */
extern T1 T1857f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.is_once_per_process */
extern T1 T469f45(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_once_per_process */
extern T1 T253f57(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_once_per_process */
extern T1 T1513f46(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_once_per_process */
extern T1 T1512f48(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_once_per_process */
extern T1 T1511f46(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_once_per_process */
extern T1 T371f48(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_once_per_process */
extern T1 T369f47(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_once_per_process */
extern T1 T359f46(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_once_per_process */
extern T1 T360f46(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.is_once_per_process */
extern T1 T236f31(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_once_per_object_value */
extern void T67f795(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_once_per_object_value_name */
extern void T67f694(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.print_attribute_onces_access */
extern void T67f870(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_attribute_onces_name */
extern void T67f690(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.once_kind */
extern T6 T67f309(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_real_64_class */
extern T1 T74f65(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_real_32_class */
extern T1 T74f64(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_pointer_class */
extern T1 T74f63(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_natural_64_class */
extern T1 T74f62(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_natural_32_class */
extern T1 T74f61(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_natural_16_class */
extern T1 T74f60(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_natural_8_class */
extern T1 T74f59(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_integer_64_class */
extern T1 T74f58(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_integer_32_class */
extern T1 T74f57(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_integer_16_class */
extern T1 T74f56(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_integer_8_class */
extern T1 T74f189(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_character_32_class */
extern T1 T74f188(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_character_8_class */
extern T1 T74f187(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].found_item */
extern T6 T290f26(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].item_storage_item */
extern T6 T290f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].found */
extern T1 T290f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].search */
extern void T290f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].search_position */
extern void T290f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].hash_position */
extern T6 T290f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].key_storage_item */
extern T0* T290f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].clashes_item */
extern T6 T290f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].slots_item */
extern T6 T290f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.register_once_feature */
extern void T67f769(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].force_last */
extern void T290f39(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].key_storage_put */
extern void T290f52(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].slots_put */
extern void T290f51(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].clashes_put */
extern void T290f50(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].resize */
extern void T290f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].clashes_resize */
extern void T290f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].special_integer_ */
extern T0* T290f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].key_storage_resize */
extern void T290f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].item_storage_resize */
extern void T290f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].slots_resize */
extern void T290f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].new_modulus */
extern T6 T290f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].new_capacity */
extern T6 T290f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].item_storage_put */
extern void T290f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].unset_found_item */
extern void T290f43(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.is_once */
extern T1 T306f32(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.is_once */
extern T1 T236f30(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_adapted_extended_attribute_call */
extern void T67f759(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #1 in feature [detachable] ET_C_GENERATOR.print_adapted_extended_attribute_call */
extern T0* T67f759ac1(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #2 in feature [detachable] ET_C_GENERATOR.print_adapted_extended_attribute_call */
extern T0* T67f759ac2(GE_context* ac, T0* a1, T0* a2, T0* a3, T1 a4);
/* [detachable] ET_C_GENERATOR.print_adapted_expression_with_agent */
extern void T67f744(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4);
/* [detachable] ET_C_GENERATOR.print_typed_default_entity_value */
extern void T67f791(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_default_name */
extern void T67f644(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_target_expression */
extern void T67f718(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_temp_name */
extern void T67f642(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].put_last */
extern void T272f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION].force */
extern void T651f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION].fill_with */
extern void T650f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION].extend */
extern void T650f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.mark_expressions_frozen */
extern void T67f790(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MANIFEST_TUPLE.resize */
extern void T288f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION_ITEM].aliased_resized_area */
extern T0* T671f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION_ITEM].aliased_resized_area */
extern T0* T669f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_MANIFEST_TUPLE.capacity */
extern T6 T288f11(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_attachment_operand */
extern void T67f733(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_boxed_attribute_type_id_access */
extern void T67f680(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_boxed_type_declaration */
extern void T67f677(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_and_then */
extern void T67f752(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.put_expanded_types */
extern void T269f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.free_twin_feature_name */
extern void T67f751(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.new_twin_feature_name */
extern T0* T67f300(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.set_feature_name */
extern void T184f122(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_C_GENERATOR.print_target_operand */
extern void T67f742(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CONVERT_TO_EXPRESSION.is_call_agent */
extern T1 T2048f17(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_EXPRESSION.is_call_agent */
extern T1 T888f7(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.is_call_agent */
extern T1 T1547f11(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.is_call_agent */
extern T1 T1543f11(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.is_call_agent */
extern T1 T1038f9(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.is_call_agent */
extern T1 T287f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_EXPRESSION].force */
extern void T271f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_EXPRESSION].resize */
extern void T271f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION].aliased_resized_area */
extern T0* T651f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION].aliased_resized_area */
extern T0* T650f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_EXPRESSION].new_capacity */
extern T6 T271f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_EXPRESSION].extendible */
extern T1 T271f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_writable */
extern void T67f735(GE_context* ac, T0* C, T0* a1);
extern T1 T67f735ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_attribute_access */
extern void T67f647(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].remove */
extern void T297f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].remove_position */
extern void T297f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].key_storage_put */
extern void T297f55(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].item_storage_put */
extern void T297f42(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].key_storage_keep_head */
extern void T297f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].item_storage_keep_head */
extern void T297f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].clashes_put */
extern void T297f44(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].slots_put */
extern void T297f45(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].clashes_item */
extern T6 T297f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].move_cursors_forth */
extern void T297f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_IDENTIFIER].set_position */
extern void T695f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_IDENTIFIER].set_next_cursor */
extern void T695f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_IDENTIFIER].set_after */
extern void T695f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].internal_set_key_equality_tester */
extern void T297f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].key_equality_tester */
extern T0* T297f17(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].slots_item */
extern T6 T297f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].hash_position */
extern T6 T297f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].key_storage_item */
extern T0* T297f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].item_storage_item */
extern T0* T297f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].search_position */
extern void T297f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER_TESTER.test */
extern T1 T298f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].unset_found_item */
extern void T297f40(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_local_variable */
extern void T67f756(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_local_name */
extern void T67f767(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].has */
extern T1 T297f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].force_last */
extern void T297f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].resize */
extern void T297f43(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].clashes_resize */
extern void T297f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].special_integer_ */
extern T0* T297f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].key_storage_resize */
extern void T297f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].item_storage_resize */
extern void T297f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].slots_resize */
extern void T297f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].new_modulus */
extern T6 T297f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].new_capacity */
extern T6 T297f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_IDENTIFIER.is_local */
extern T1 T184f30(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_iteration_item */
extern void T67f755(GE_context* ac, T0* C, T0* a1);
extern T1 T67f755ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_iteration_cursor_name */
extern void T67f766(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_QUANTIFIER_EXPRESSION.cursor_item_expression */
extern T0* T1254f11(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.cursor_item_expression */
extern T0* T1008f15(GE_context* ac, T0* C);
/* detachable ET_REPEAT_INSTRUCTION.cursor_item_expression */
extern T0* T1264f10(GE_context* ac, T0* C);
/* detachable ET_ACROSS_INSTRUCTION.cursor_item_expression */
extern T0* T1011f14(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.has_cursor_name */
extern T1 T1254f28(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.has_cursor_name */
extern T1 T1008f11(GE_context* ac, T0* C);
/* detachable ET_REPEAT_INSTRUCTION.has_cursor_name */
extern T1 T1264f14(GE_context* ac, T0* C);
/* detachable ET_ACROSS_INSTRUCTION.has_cursor_name */
extern T1 T1011f10(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.iteration_components */
extern T0* T1557f5(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.iteration_components */
extern T0* T1556f6(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.iteration_components */
extern T0* T1555f6(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.iteration_components */
extern T0* T1554f5(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.iteration_components */
extern T0* T1553f8(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.iteration_components */
extern T0* T1552f8(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.iteration_components */
extern T0* T1516f2(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.iteration_components */
extern T0* T253f2(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.iteration_components */
extern T0* T1515f2(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.iteration_components */
extern T0* T1513f2(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.iteration_components */
extern T0* T1512f2(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.iteration_components */
extern T0* T1511f2(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.iteration_components */
extern T0* T371f4(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.iteration_components */
extern T0* T369f2(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.iteration_components */
extern T0* T359f2(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.iteration_components */
extern T0* T360f2(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.iteration_components */
extern T0* T476f2(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.iteration_components */
extern T0* T469f1(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_iteration_item */
extern T1 T184f27(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_temporary_variable */
extern void T67f754(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.temp_variable_type */
extern T0* T67f308(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.is_temp_variable_known */
extern T1 T67f307(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_formal_argument */
extern void T67f753(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_assignment_operand */
extern void T67f789(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_C_GENERATOR.is_twin_needed_in_attachment */
extern T1 T67f310(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.new_temp_variable */
extern T0* T67f147(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE].force_last */
extern void T262f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_PRIMARY_TYPE].force */
extern void T630f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_PRIMARY_TYPE].fill_with */
extern void T629f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_PRIMARY_TYPE].extend */
extern void T629f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE].resize */
extern void T262f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_PRIMARY_TYPE].aliased_resized_area */
extern T0* T630f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_PRIMARY_TYPE].aliased_resized_area */
extern T0* T629f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE].new_capacity */
extern T6 T262f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE].extendible */
extern T1 T262f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_IDENTIFIER.set_temporary */
extern void T184f120(GE_context* ac, T0* C, T1 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER].item */
extern T0* T261f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.same_declaration_types */
extern T1 T67f492(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.mark_temp_variable_used */
extern void T67f788(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER.is_temporary */
extern T1 T184f12(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.fill_call_operands */
extern void T67f734(GE_context* ac, T0* C, T6 a1);
extern T1 T67f734ot1(T0* a1, T0** a2);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_EXPRESSION].prune */
extern void T271f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_EXPRESSION].i_th */
extern T0* T271f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].resize */
extern void T272f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_operand */
extern void T67f743(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.has_field_access */
extern T1 T314f14(GE_context* ac, T0* C, T0* a1);
extern T1 T314f14ot1(T0* a1);
/* [detachable] ET_C_GENERATOR.in_call_target */
extern T1 T67f299(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.query_call */
extern T0* T306f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].value */
extern T0* T616f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].item_storage_item */
extern T0* T616f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].position_of_key */
extern T6 T616f25(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_FEATURE.query_call */
extern T0* T236f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.process */
extern void T2047f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_convert_from_expression */
extern void T1327f118(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_convert_from_expression */
extern void T398f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_convert_from_expression */
extern void T394f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_convert_from_expression */
extern void T71f567(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_convert_from_expression_validity */
extern void T71f570(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_creation_expression_validity */
extern void T71f387(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcc6b_error */
extern void T53f479(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vgcc6b */
extern T0* T785c1102(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vgcc6b_error */
extern void T36f479(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_dotnet_creation_procedure_call_expression_validity */
extern void T71f441(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4, T0* a5);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.free_overloaded_procedures */
extern void T71f309(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].force_last */
extern void T387f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].force */
extern void T714f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].fill_with */
extern void T716f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].extend */
extern void T716f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].resize */
extern void T387f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].aliased_resized_area */
extern T0* T714f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].aliased_resized_area */
extern T0* T716f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].new_capacity */
extern T6 T387f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].extendible */
extern T1 T387f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].first */
extern T0* T715f14(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].is_empty */
extern T1 T715f13(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.add_overloaded_procedures */
extern void T254f131(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.add_overloaded_procedures */
extern void T1850f15(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.add_overloaded_procedures */
extern void T72f140(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.add_overloaded_procedures */
extern void T74f330(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_PROCEDURE_LIST.add_overloaded_features */
extern void T258f14(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T258f14ot1(T0* a1, T0** a2);
/* detachable ET_EXTERNAL_PROCEDURE.overloaded_alias_names */
extern T0* T1516f86(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.overloaded_extended_name */
extern T0* T1516f87(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.overloaded_alias_names */
extern T0* T476f77(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.overloaded_extended_name */
extern T0* T476f78(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.overloaded_alias_names */
extern T0* T1515f80(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.overloaded_extended_name */
extern T0* T1515f81(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.overloaded_alias_names */
extern T0* T469f78(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.overloaded_extended_name */
extern T0* T469f79(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.overloaded_name */
extern T0* T1516f85(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.overloaded_name */
extern T0* T476f76(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.overloaded_name */
extern T0* T1515f79(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.overloaded_name */
extern T0* T469f77(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.new_overloaded_procedures */
extern T0* T71f84(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].remove_last */
extern void T387f15(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].keep_head */
extern void T716f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].set_count */
extern void T716f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].move_last_cursors_after */
extern void T387f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].set_next_cursor */
extern void T717f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].set_after */
extern void T717f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].last */
extern T0* T387f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].make */
extern T0* T715c15(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].set_internal_cursor */
extern void T715f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].new_cursor */
extern T0* T715f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PROCEDURE].make */
extern T0* T1425c9(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].is_empty */
extern T1 T387f9(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_creation_procedure_call_expression_validity */
extern void T71f440(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5, T0* a6);
extern T1 T71f440ot1(T0* a1, T0** a2);
extern T1 T71f440ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_creation_procedure_call_validity */
extern void T71f482(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_LIKE_N.is_like_current */
extern T1 T1474f30(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.first */
extern T0* T241f47(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_formal_creation_procedure_call_validity */
extern void T71f481(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_formal_parameter_creation_vape_validity */
extern void T71f493(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vape2b_error */
extern void T53f526(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vape2b */
extern T0* T785c1149(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.report_vape2b_error */
extern void T36f526(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_CLIENT.reset */
extern void T403f7(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcc8a_error */
extern void T53f524(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vgcc8a */
extern T0* T785c1147(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vgcc8_error */
extern T1 T53f159(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vgcc8a_error */
extern void T36f524(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.reportable_vgcc8_error */
extern T1 T36f159(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_no_creation_procedure_call_validity */
extern void T71f480(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcc1a_error */
extern void T53f523(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vgcc1a */
extern T0* T785c1146(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_CREATE_INSTRUCTION.position */
extern T0* T1533f8(GE_context* ac, T0* C);
/* detachable ET_BANG_INSTRUCTION.position */
extern T0* T1532f7(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vgcc1_error */
extern T1 T53f158(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vgcc1a_error */
extern void T36f523(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vgcc1_error */
extern T1 T36f158(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcc5a_error */
extern void T53f522(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vgcc5a */
extern T0* T785c1145(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vgcc5_error */
extern T1 T53f157(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vgcc5a_error */
extern void T36f522(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vgcc5_error */
extern T1 T36f157(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_creation_type_validity */
extern void T71f479(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.type_position */
extern T0* T2047f21(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.type_position */
extern T0* T328f23(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_create_supplier */
extern void T71f478(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_LIKE_N.shallow_named_type */
extern T0* T1474f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_N.shallow_named_type_with_type_mark */
extern T0* T1474f29(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_type_validity */
extern void T71f216(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.current_closure_impl */
extern T0* T71f74(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.type */
extern T0* T328f20(GE_context* ac, T0* C);
/* [detachable] ET_AST_NULL_LEAF.column */
extern T6 T2094f3(GE_context* ac, T0* C);
/* [detachable] ET_AST_NULL_LEAF.line */
extern T6 T2094f2(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.last_position */
extern T0* T2047f20(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.last_position */
extern T0* T622f129(GE_context* ac, T0* C);
/* [detachable] ET_COMPRESSED_POSITION.make */
extern T0* T193c11(GE_context* ac, T6 a1, T6 a2);
/* [detachable] ET_COMPRESSED_POSITION.set_position */
extern void T193f13(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_BRACKET_SYMBOL.text */
extern T0* T622f13(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.xor_symbol_name */
extern T0* T73f750(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.there_exists_symbol_name */
extern T0* T73f686(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.or_else_symbol_name */
extern T0* T73f749(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.or_symbol_name */
extern T0* T73f748(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.open_repeat_symbol_name */
extern T0* T73f677(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.not_symbol_name */
extern T0* T73f746(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.implies_symbol_name */
extern T0* T73f745(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.for_all_symbol_name */
extern T0* T73f665(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.close_repeat_symbol_name */
extern T0* T73f656(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.bar_symbol_name */
extern T0* T73f655(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.and_then_symbol_name */
extern T0* T73f743(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.and_symbol_name */
extern T0* T73f742(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.last_position */
extern T0* T955f23(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.last_position */
extern T0* T1496f46(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.last_position */
extern T0* T1495f45(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.last_position */
extern T0* T1494f38(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.last_position */
extern T0* T1493f38(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.last_position */
extern T0* T1491f40(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.last_position */
extern T0* T1490f40(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.last_position */
extern T0* T1489f62(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.last_position */
extern T0* T1488f62(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.last_position */
extern T0* T1487f62(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.last_position */
extern T0* T1486f62(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.last_position */
extern T0* T1485f49(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.occurrences */
extern T6 T775f37(GE_context* ac, T0* C, T2 a1);
/* [detachable] UC_UTF8_STRING.code_occurrences */
extern T6 T775f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] UC_STRING.occurrences */
extern T6 T429f28(GE_context* ac, T0* C, T2 a1);
/* [detachable] UC_STRING.code_occurrences */
extern T6 T429f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] STRING_8.occurrences */
extern T6 T17f38(GE_context* ac, T0* C, T2 a1);
/* detachable ET_SPECIAL_MANIFEST_STRING.last_position */
extern T0* T1484f45(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.last_position */
extern T0* T1483f44(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.last_position */
extern T0* T1482f46(GE_context* ac, T0* C);
/* detachable ET_VOID.last_position */
extern T0* T997f29(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.last_position */
extern T0* T372f62(GE_context* ac, T0* C);
/* [detachable] ET_AST_NULL_LEAF.last_position */
extern T0* T2094f4(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.last_position */
extern T0* T184f110(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.last_position */
extern T0* T413f42(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.last_position */
extern T0* T341f91(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.text */
extern T0* T341f9(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.last_position */
extern T0* T282f29(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.last_position */
extern T0* T309f30(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.last_leaf */
extern T0* T2047f22(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.last_leaf */
extern T0* T2048f22(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.last_leaf */
extern T0* T2046f17(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_CAST_EXPRESSION.last_leaf */
extern T0* T1980f16(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.last_leaf */
extern T0* T184f103(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.last_leaf */
extern T0* T1489f61(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.last_leaf */
extern T0* T1488f61(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.last_leaf */
extern T0* T1487f61(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.last_leaf */
extern T0* T1486f61(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.last_leaf */
extern T0* T372f61(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.last_leaf */
extern T0* T1485f48(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.last_leaf */
extern T0* T1484f44(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.last_leaf */
extern T0* T1482f45(GE_context* ac, T0* C);
/* detachable ET_ITERATION_CURSOR.last_leaf */
extern T0* T1551f20(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.last_leaf */
extern T0* T1254f35(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.last_leaf */
extern T0* T1008f36(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.last_leaf */
extern T0* T1550f23(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.last_leaf */
extern T0* T736f25(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.last_leaf */
extern T0* T1549f28(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.last_leaf */
extern T0* T1539f20(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.last_leaf */
extern T0* T635f27(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.last_leaf */
extern T0* T1977f13(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.last_leaf */
extern T0* T283f14(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.last_leaf */
extern T0* T1557f37(GE_context* ac, T0* C);
extern T1 T1557f37ot1(T0* a1, T0** a2);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.last_leaf */
extern T0* T1025f11(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.last_leaf */
extern T0* T1556f43(GE_context* ac, T0* C);
extern T1 T1556f43ot1(T0* a1, T0** a2);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.last_leaf */
extern T0* T1555f39(GE_context* ac, T0* C);
extern T1 T1555f39ot1(T0* a1, T0** a2);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.last_leaf */
extern T0* T1554f38(GE_context* ac, T0* C);
extern T1 T1554f38ot1(T0* a1, T0** a2);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.last_leaf */
extern T0* T1553f44(GE_context* ac, T0* C);
extern T1 T1553f44ot1(T0* a1, T0** a2);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.last_leaf */
extern T0* T1552f40(GE_context* ac, T0* C);
extern T1 T1552f40ot1(T0* a1, T0** a2);
/* detachable ET_CALL_AGENT.last_leaf */
extern T0* T1041f29(GE_context* ac, T0* C);
extern T1 T1041f29ot1(T0* a1, T0** a2);
/* detachable ET_STRIP_EXPRESSION.last_leaf */
extern T0* T1270f27(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.last_leaf */
extern T0* T288f27(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.last_leaf */
extern T0* T1218f27(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.last_leaf */
extern T0* T1232f21(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.last_leaf */
extern T0* T1038f24(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.last_leaf */
extern T0* T622f128(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.last_leaf */
extern T0* T311f23(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.last_leaf */
extern T0* T1548f20(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.last_leaf */
extern T0* T1547f25(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.last_leaf */
extern T0* T1546f21(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.last_leaf */
extern T0* T312f22(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.last_leaf */
extern T0* T313f20(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.last_leaf */
extern T0* T1543f25(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.last_leaf */
extern T0* T1542f20(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.last_leaf */
extern T0* T1541f20(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.last_leaf */
extern T0* T1540f20(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.last_leaf */
extern T0* T1538f24(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.last_leaf */
extern T0* T1267f28(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.last_leaf */
extern T0* T287f25(GE_context* ac, T0* C);
/* detachable ET_INSPECT_EXPRESSION.last_leaf */
extern T0* T1189f23(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.last_leaf */
extern T0* T1167f26(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.last_leaf */
extern T0* T328f25(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.last_leaf */
extern T0* T1524f82(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.last_leaf */
extern T0* T242f75(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.last_leaf */
extern T0* T1526f79(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.last_leaf */
extern T0* T1525f82(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.last_leaf */
extern T0* T449f75(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.last_leaf */
extern T0* T254f112(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.last_leaf */
extern T0* T1957f21(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.null_leaf */
extern T0* T73f1544(GE_context* ac, T0* C);
/* [detachable] ET_AST_NULL_LEAF.make */
extern T0* T2094c5(GE_context* ac);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.last_leaf */
extern T0* T1501f75(GE_context* ac, T0* C);
/* detachable ET_TYPE_CONSTRAINT_LIST.last_leaf */
extern T0* T1281f10(GE_context* ac, T0* C);
/* detachable ET_TYPE_RENAME_CONSTRAINT.last_leaf */
extern T0* T1502f7(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_RENAME_LIST.last_leaf */
extern T0* T1084f11(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.renames */
extern T0* T1501f76(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_CREATOR.last_leaf */
extern T0* T1081f11(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.last_leaf */
extern T0* T891f77(GE_context* ac, T0* C);
/* [detachable] ET_LABELED_COMMA_ACTUAL_PARAMETER.last_leaf */
extern T0* T1769f9(GE_context* ac, T0* C);
/* detachable ET_LABELED_ACTUAL_PARAMETER.last_leaf */
extern T0* T1522f9(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.last_leaf */
extern T0* T240f25(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.last_leaf */
extern T0* T72f123(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.last_leaf */
extern T0* T577f16(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.last_leaf */
extern T0* T1956f18(GE_context* ac, T0* C);
/* detachable ET_FORMAL_PARAMETER_LIST.last_leaf */
extern T0* T603f25(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.last_leaf */
extern T0* T74f251(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL.last_leaf */
extern T0* T329f4(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.last_leaf */
extern T0* T1496f45(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.last_leaf */
extern T0* T1495f44(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.last_leaf */
extern T0* T282f31(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.last_leaf */
extern T0* T309f29(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.last_leaf */
extern T0* T1494f37(GE_context* ac, T0* C);
/* detachable ET_VOID.last_leaf */
extern T0* T997f28(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.last_leaf */
extern T0* T1493f37(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.last_leaf */
extern T0* T1491f39(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.last_leaf */
extern T0* T1490f39(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.last_leaf */
extern T0* T1483f43(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.last_position */
extern T0* T328f22(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.creation_call */
extern T0* T2047f19(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.process_convert_from_expression */
extern void T380f456(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_convert_from_expression_validity */
extern void T380f459(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_creation_expression_validity */
extern void T380f306(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_dotnet_creation_procedure_call_expression_validity */
extern void T380f357(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4, T0* a5);
/* [detachable] ET_FEATURE_CHECKER.free_overloaded_procedures */
extern void T380f384(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.new_overloaded_procedures */
extern T0* T380f70(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.check_creation_procedure_call_expression_validity */
extern void T380f356(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5, T0* a6);
extern T1 T380f356ot1(T0* a1, T0** a2);
extern T1 T380f356ot2(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.check_creation_procedure_call_validity */
extern void T380f382(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FEATURE_CHECKER.check_formal_creation_procedure_call_validity */
extern void T380f381(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_FEATURE_CHECKER.check_formal_parameter_creation_vape_validity */
extern void T380f393(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_no_creation_procedure_call_validity */
extern void T380f380(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.check_creation_type_validity */
extern void T380f379(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_create_supplier */
extern void T380f378(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SUPPLIER_BUILDER.report_create_supplier */
extern void T1341f17(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_type_validity */
extern void T380f135(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.current_closure_impl */
extern T0* T380f57(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.process_convert_from_expression */
extern void T67f1484(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_convert_from_expression */
extern void T67f1487(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_creation_expression */
extern void T67f674(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_creation_procedure_call */
extern void T67f736(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CREATE_EXPRESSION.name */
extern T0* T328f6(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.process */
extern void T2046f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_convert_builtin_expression */
extern void T1327f117(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_convert_builtin_expression */
extern void T398f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_convert_builtin_expression */
extern void T394f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_convert_builtin_expression */
extern void T71f566(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_convert_builtin_expression_validity */
extern void T71f569(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.process_convert_builtin_expression */
extern void T380f455(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_convert_builtin_expression_validity */
extern void T380f458(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_convert_builtin_expression */
extern void T67f1483(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_convert_builtin_expression */
extern void T67f1486(GE_context* ac, T0* C, T0* a1);
/* detachable ET_BINARY_INTEGER_CONSTANT.process */
extern void T1489f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_binary_integer_constant */
extern void T1327f87(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_binary_integer_constant */
extern void T71f306(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_binary_integer_constant_validity */
extern void T71f371(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_integer_constant_validity */
extern void T71f186(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_BINARY_INTEGER_CONSTANT.set_type */
extern void T1489f67(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OCTAL_INTEGER_CONSTANT.set_type */
extern void T1488f67(GE_context* ac, T0* C, T0* a1);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.set_type */
extern void T1487f67(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.set_type */
extern void T1486f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.set_type */
extern void T372f69(GE_context* ac, T0* C, T0* a1);
/* detachable ET_BINARY_INTEGER_CONSTANT.value */
extern T11 T1489f4(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.value */
extern T11 T1488f4(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.value */
extern T11 T1487f4(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.value */
extern T11 T1486f4(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.value */
extern T11 T372f1(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.sign */
extern T0* T1489f2(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.sign */
extern T0* T1488f2(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.sign */
extern T0* T1487f2(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.sign */
extern T0* T1486f2(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.sign */
extern T0* T372f2(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_binary */
extern T1 T1489f19(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_binary */
extern T1 T1488f19(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_binary */
extern T1 T1487f19(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_binary */
extern T1 T1486f19(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_binary */
extern T1 T372f26(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_hexadecimal */
extern T1 T1489f18(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_hexadecimal */
extern T1 T1488f18(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_hexadecimal */
extern T1 T1487f18(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_hexadecimal */
extern T1 T1486f18(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_hexadecimal */
extern T1 T372f25(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_natural_64_constant */
extern void T71f226(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_constant_expression */
extern void T71f231(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].force_last */
extern void T381f49(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].key_storage_put */
extern void T381f55(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE].force */
extern void T709f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE].fill_with */
extern void T708f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE].extend */
extern void T708f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].slots_put */
extern void T381f54(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].clashes_put */
extern void T381f53(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].slots_item */
extern T6 T381f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].hash_position */
extern T6 T381f24(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.hash_code */
extern T6 T323f22(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].resize */
extern void T381f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].clashes_resize */
extern void T381f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].special_integer_ */
extern T0* T381f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].key_storage_resize */
extern void T381f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE].aliased_resized_area */
extern T0* T709f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE].aliased_resized_area */
extern T0* T708f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].item_storage_resize */
extern void T381f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].key_storage_item */
extern T0* T381f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].clashes_item */
extern T6 T381f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].slots_resize */
extern void T381f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].new_modulus */
extern T6 T381f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].new_capacity */
extern T6 T381f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].item_storage_put */
extern void T381f51(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].search_position */
extern void T381f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].unset_found_item */
extern void T381f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].found_item */
extern T6 T381f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].item_storage_item */
extern T6 T381f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].found */
extern T1 T381f20(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].search */
extern void T381f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_natural_32_constant */
extern void T71f225(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_natural_16_constant */
extern void T71f224(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_natural_8_constant */
extern void T71f223(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_integer_64_constant */
extern void T71f222(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_integer_32_constant */
extern void T71f221(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_integer_16_constant */
extern void T71f220(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvwmc2a_error */
extern void T53f389(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_gvwmc2a */
extern T0* T785c1019(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_BINARY_INTEGER_CONSTANT.literal */
extern T0* T1489f7(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.literal */
extern T0* T1488f7(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.literal */
extern T0* T1487f7(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.literal */
extern T0* T1486f7(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.literal */
extern T0* T372f3(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.value_position */
extern T0* T1489f45(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.value_position */
extern T0* T1488f45(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.value_position */
extern T0* T1487f45(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.value_position */
extern T0* T1486f45(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.value_position */
extern T0* T372f46(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvwmc2_error */
extern T1 T53f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvwmc2a_error */
extern void T36f389(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_gvwmc2_error */
extern T1 T36f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_integer_8_constant */
extern void T71f219(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.current_universe_impl */
extern T0* T71f77(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vwmq0a_error */
extern void T53f388(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vwmq0a */
extern T0* T785c1018(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.type_name */
extern T0* T785f430(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vwmq_error */
extern T1 T53f101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vwmq0a_error */
extern void T36f388(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vwmq_error */
extern T1 T36f101(GE_context* ac, T0* C, T0* a1);
/* detachable ET_BINARY_INTEGER_CONSTANT.cast_type */
extern T0* T1489f3(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.cast_type */
extern T0* T1488f3(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.cast_type */
extern T0* T1487f3(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.cast_type */
extern T0* T1486f3(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.cast_type */
extern T0* T372f7(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_binary_integer_constant */
extern void T398f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_binary_integer_constant */
extern void T394f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_binary_integer_constant */
extern void T380f162(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_binary_integer_constant_validity */
extern void T380f218(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_integer_constant_validity */
extern void T380f127(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_natural_64_constant */
extern void T380f145(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_natural_32_constant */
extern void T380f144(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_natural_16_constant */
extern void T380f143(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_natural_8_constant */
extern void T380f142(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_integer_64_constant */
extern void T380f141(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_integer_32_constant */
extern void T380f140(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_integer_16_constant */
extern void T380f139(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_integer_8_constant */
extern void T380f138(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.current_universe_impl */
extern T0* T380f51(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.process_binary_integer_constant */
extern void T67f1427(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_binary_integer_constant */
extern void T67f1439(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_integer_constant */
extern void T67f1441(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_integer_value */
extern void T67f1256(GE_context* ac, T0* C, T11 a1, T1 a2, T0* a3, T1 a4);
/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM.put_natural_64 */
extern void T161f12(GE_context* ac, T0* C, T11 a1);
/* [detachable] KL_STRING_OUTPUT_STREAM.put_natural_64 */
extern void T260f13(GE_context* ac, T0* C, T11 a1);
/* detachable ET_OCTAL_INTEGER_CONSTANT.process */
extern void T1488f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_octal_integer_constant */
extern void T1327f86(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_octal_integer_constant */
extern void T71f305(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_octal_integer_constant_validity */
extern void T71f370(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_octal_integer_constant */
extern void T398f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_octal_integer_constant */
extern void T394f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_octal_integer_constant */
extern void T380f161(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_octal_integer_constant_validity */
extern void T380f217(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_octal_integer_constant */
extern void T67f1426(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_octal_integer_constant */
extern void T67f1438(GE_context* ac, T0* C, T0* a1);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.process */
extern void T1487f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_hexadecimal_integer_constant */
extern void T1327f85(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_hexadecimal_integer_constant */
extern void T71f304(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_hexadecimal_integer_constant_validity */
extern void T71f369(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_hexadecimal_integer_constant */
extern void T398f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_hexadecimal_integer_constant */
extern void T394f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_hexadecimal_integer_constant */
extern void T380f160(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_hexadecimal_integer_constant_validity */
extern void T380f216(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_hexadecimal_integer_constant */
extern void T67f1425(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_hexadecimal_integer_constant */
extern void T67f1437(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.process */
extern void T1486f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_underscored_integer_constant */
extern void T1327f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_underscored_integer_constant */
extern void T71f303(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_underscored_integer_constant_validity */
extern void T71f368(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_underscored_integer_constant */
extern void T398f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_underscored_integer_constant */
extern void T394f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_underscored_integer_constant */
extern void T380f159(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_underscored_integer_constant_validity */
extern void T380f215(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_underscored_integer_constant */
extern void T67f1424(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_underscored_integer_constant */
extern void T67f1436(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.process */
extern void T372f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_regular_integer_constant */
extern void T1327f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_regular_integer_constant */
extern void T71f302(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_regular_integer_constant_validity */
extern void T71f367(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_regular_integer_constant */
extern void T398f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_regular_integer_constant */
extern void T394f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_regular_integer_constant */
extern void T380f204(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_regular_integer_constant_validity */
extern void T380f261(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_regular_integer_constant */
extern void T67f1443(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_regular_integer_constant */
extern void T67f1445(GE_context* ac, T0* C, T0* a1);
/* detachable ET_VERBATIM_STRING.process */
extern void T1485f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_verbatim_string */
extern void T1327f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_verbatim_string */
extern void T71f301(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_verbatim_string_validity */
extern void T71f366(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_manifest_string_validity */
extern void T71f187(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_VERBATIM_STRING.set_type */
extern void T1485f53(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SPECIAL_MANIFEST_STRING.set_type */
extern void T1484f49(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REGULAR_MANIFEST_STRING.set_type */
extern void T1483f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_immutable_string_32_constant */
extern void T71f230(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_string_constant */
extern void T71f232(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_manifest_string_area_dynamic_type */
extern void T71f233(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_string_32_constant */
extern void T71f229(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_immutable_string_8_constant */
extern void T71f228(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_LIBRARY.immutable_string_8_type */
extern T0* T187f44(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.immutable_string_8_type */
extern T0* T58f31(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.immutable_string_8_type */
extern T0* T40f53(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvwmc2c_error */
extern void T53f391(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_gvwmc2c */
extern T0* T785c1021(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_VERBATIM_STRING.value_position */
extern T0* T1485f35(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.value_position */
extern T0* T1484f31(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.value_position */
extern T0* T1483f30(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_gvwmc2c_error */
extern void T36f391(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_string_8_constant */
extern void T71f227(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vwmq0d_error */
extern void T53f390(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vwmq0d */
extern T0* T785c1020(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vwmq0d_error */
extern void T36f390(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_VERBATIM_STRING.cast_type */
extern T0* T1485f3(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.cast_type */
extern T0* T1484f3(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.cast_type */
extern T0* T1483f3(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_verbatim_string */
extern void T398f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_verbatim_string */
extern void T394f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_verbatim_string */
extern void T380f158(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_verbatim_string_validity */
extern void T380f214(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_manifest_string_validity */
extern void T380f128(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_immutable_string_32_constant */
extern void T380f149(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_string_32_constant */
extern void T380f148(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_immutable_string_8_constant */
extern void T380f147(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_string_8_constant */
extern void T380f146(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_verbatim_string */
extern void T67f1423(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_verbatim_string */
extern void T67f1435(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_manifest_string */
extern void T67f1258(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_utf8_as_escaped_string_8 */
extern void T67f1265(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_INTEGER_FORMATTER.put_octal_natural_32 */
extern void T351f3(GE_context* ac, T0* C, T0* a1, T10 a2);
/* [detachable] ET_C_GENERATOR.integer_formatter_ */
extern T0* T67f282(GE_context* ac, T0* C);
/* [detachable] UT_INTEGER_FORMATTER.default_create */
extern T0* T351c1(GE_context* ac);
/* [detachable] ET_C_GENERATOR.print_utf8_as_escaped_string_32 */
extern void T67f1264(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SPECIAL_MANIFEST_STRING.process */
extern void T1484f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_special_manifest_string */
extern void T1327f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_special_manifest_string */
extern void T71f300(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_special_manifest_string_validity */
extern void T71f365(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_special_manifest_string */
extern void T398f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_special_manifest_string */
extern void T394f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_special_manifest_string */
extern void T380f157(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_special_manifest_string_validity */
extern void T380f213(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_special_manifest_string */
extern void T67f1422(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_special_manifest_string */
extern void T67f1434(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C3_CHARACTER_CONSTANT.process */
extern void T1482f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_c3_character_constant */
extern void T1327f80(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_c3_character_constant */
extern void T71f299(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_c3_character_constant_validity */
extern void T71f364(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_character_constant_validity */
extern void T71f185(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_C3_CHARACTER_CONSTANT.set_type */
extern void T1482f50(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C2_CHARACTER_CONSTANT.set_type */
extern void T1496f53(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C1_CHARACTER_CONSTANT.set_type */
extern void T1495f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_character_32_constant */
extern void T71f218(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvwmc2b_error */
extern void T53f387(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_gvwmc2b */
extern T0* T785c1017(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T785f1017ot1(T0* a1, T0** a2);
extern T1 T785f1017ot2(T0* a1, T0** a2);
/* detachable ET_C3_CHARACTER_CONSTANT.value */
extern T3 T1482f4(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.value */
extern T3 T1496f4(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.value */
extern T3 T1495f4(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.literal */
extern T2 T1496f29(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.value_position */
extern T0* T1482f30(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.value_position */
extern T0* T1496f30(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.value_position */
extern T0* T1495f29(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_gvwmc2b_error */
extern void T36f387(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_character_8_constant */
extern void T71f217(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vwmq0c_error */
extern void T53f386(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vwmq0c */
extern T0* T785c1016(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vwmq0c_error */
extern void T36f386(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_C3_CHARACTER_CONSTANT.cast_type */
extern T0* T1482f2(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.cast_type */
extern T0* T1496f2(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.cast_type */
extern T0* T1495f2(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_c3_character_constant */
extern void T398f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_c3_character_constant */
extern void T394f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_c3_character_constant */
extern void T380f155(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_c3_character_constant_validity */
extern void T380f211(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_character_constant_validity */
extern void T380f126(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_character_32_constant */
extern void T380f137(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_character_8_constant */
extern void T380f136(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_c3_character_constant */
extern void T67f1421(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_character_constant */
extern void T67f1223(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_escaped_character_8 */
extern void T67f1257(GE_context* ac, T0* C, T2 a1);
/* [detachable] UT_INTEGER_FORMATTER.put_octal_integer */
extern void T351f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM.put_natural_32 */
extern void T161f11(GE_context* ac, T0* C, T10 a1);
/* [detachable] KL_STRING_OUTPUT_STREAM.put_natural_32 */
extern void T260f12(GE_context* ac, T0* C, T10 a1);
/* [detachable] ET_CREATE_EXPRESSION.process */
extern void T328f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_create_expression */
extern void T1327f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_actual_argument_list */
extern void T1327f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_create_expression */
extern void T71f263(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_create_expression_validity */
extern void T71f329(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_create_expression */
extern void T398f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_create_expression */
extern void T394f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_create_expression */
extern void T380f190(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_create_expression_validity */
extern void T380f246(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_create_expression */
extern void T67f1416(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_create_expression */
extern void T67f1429(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EQUALITY_EXPRESSION.process */
extern void T313f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_equality_expression */
extern void T1327f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_equality_expression */
extern void T71f273(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_equality_expression_validity */
extern void T71f339(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_equality_expression */
extern void T71f402(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_equality_expression_target_dynamic_types */
extern void T71f450(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.put_equality_expression */
extern void T451f87(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_EQUALITY_EXPRESSION.set_next */
extern void T2042f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.put_equality_expression */
extern void T450f87(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.put_equality_expression */
extern void T302f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.put_equality_expression */
extern void T319f85(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_equality_expression */
extern void T223f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_EQUALITY_EXPRESSION.make */
extern T0* T2042c6(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vweq0a_error */
extern void T53f452(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vweq0a */
extern T0* T785c1075(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_SYMBOL_OPERATOR.text */
extern T0* T994f6(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vweq_error */
extern T1 T53f130(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vweq0a_error */
extern void T36f452(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vweq_error */
extern T1 T36f130(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EQUALITY_EXPRESSION.set_left */
extern void T313f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EQUALITY_EXPRESSION.set_right */
extern void T313f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_equality_expression */
extern void T398f49(GE_context* ac, T0* C, T0* a1);
extern T1 T398f49ot1(T0* a1);
extern T1 T398f49ot2(T0* a1);
extern T1 T398f49ot3(T0* a1, T0** a2);
extern T1 T398f49ot4(T0* a1, T0** a2);
extern T1 T398f49ot5(T0* a1);
extern T1 T398f49ot6(T0* a1);
extern T1 T398f49ot7(T0* a1, T0** a2);
extern T1 T398f49ot8(T0* a1, T0** a2);
/* [detachable] ET_ATTACHMENT_SCOPE.add_name */
extern void T397f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE.add_attribute */
extern void T397f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].force_last */
extern void T745f84(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ATTACHMENT_SCOPE.add_formal_argument */
extern void T397f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE.add_local_variable */
extern void T397f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE.add_result */
extern void T397f15(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_CAST_EXPRESSION.unparenthesized_expression */
extern T0* T1980f9(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.unparenthesized_expression */
extern T0* T2048f9(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.unparenthesized_expression */
extern T0* T2047f10(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.unparenthesized_expression */
extern T0* T2046f8(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.unparenthesized_expression */
extern T0* T184f91(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.unparenthesized_expression */
extern T0* T1489f44(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.unparenthesized_expression */
extern T0* T1488f44(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.unparenthesized_expression */
extern T0* T1487f44(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.unparenthesized_expression */
extern T0* T1486f44(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.unparenthesized_expression */
extern T0* T372f45(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.unparenthesized_expression */
extern T0* T1485f34(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.unparenthesized_expression */
extern T0* T1484f30(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.unparenthesized_expression */
extern T0* T1482f29(GE_context* ac, T0* C);
/* detachable ET_ITERATION_CURSOR.unparenthesized_expression */
extern T0* T1551f16(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.unparenthesized_expression */
extern T0* T1254f30(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.unparenthesized_expression */
extern T0* T1008f31(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.unparenthesized_expression */
extern T0* T1550f19(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.unparenthesized_expression */
extern T0* T736f20(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.unparenthesized_expression */
extern T0* T1549f23(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.unparenthesized_expression */
extern T0* T1539f16(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.unparenthesized_expression */
extern T0* T635f19(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.unparenthesized_expression */
extern T0* T1557f30(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.unparenthesized_expression */
extern T0* T1556f36(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.unparenthesized_expression */
extern T0* T1555f32(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.unparenthesized_expression */
extern T0* T1554f31(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.unparenthesized_expression */
extern T0* T1553f37(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.unparenthesized_expression */
extern T0* T1552f33(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.unparenthesized_expression */
extern T0* T1041f23(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.unparenthesized_expression */
extern T0* T1270f23(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.unparenthesized_expression */
extern T0* T288f24(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.unparenthesized_expression */
extern T0* T1218f23(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.unparenthesized_expression */
extern T0* T1232f18(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.unparenthesized_expression */
extern T0* T1038f20(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.unparenthesized_expression */
extern T0* T311f20(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.unparenthesized_expression */
extern T0* T1548f16(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.unparenthesized_expression */
extern T0* T1547f21(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.unparenthesized_expression */
extern T0* T1546f17(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.unparenthesized_expression */
extern T0* T312f18(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.unparenthesized_expression */
extern T0* T313f17(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.unparenthesized_expression */
extern T0* T1543f21(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.unparenthesized_expression */
extern T0* T1542f16(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.unparenthesized_expression */
extern T0* T1541f16(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.unparenthesized_expression */
extern T0* T1540f16(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.unparenthesized_expression */
extern T0* T1538f20(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.unparenthesized_expression */
extern T0* T1267f20(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.unparenthesized_expression */
extern T0* T287f22(GE_context* ac, T0* C);
/* detachable ET_INSPECT_EXPRESSION.unparenthesized_expression */
extern T0* T1189f18(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.unparenthesized_expression */
extern T0* T1167f22(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.unparenthesized_expression */
extern T0* T328f24(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.unparenthesized_expression */
extern T0* T1496f28(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.unparenthesized_expression */
extern T0* T1495f28(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.unparenthesized_expression */
extern T0* T282f25(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.unparenthesized_expression */
extern T0* T309f25(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.unparenthesized_expression */
extern T0* T1494f25(GE_context* ac, T0* C);
/* detachable ET_VOID.unparenthesized_expression */
extern T0* T997f24(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.unparenthesized_expression */
extern T0* T1493f25(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.unparenthesized_expression */
extern T0* T1491f27(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.unparenthesized_expression */
extern T0* T1490f27(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.unparenthesized_expression */
extern T0* T1483f29(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.is_equal_symbol */
extern T1 T994f106(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_equal_symbol */
extern T1 T622f103(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_equal_symbol */
extern T1 T341f88(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_equality_expression */
extern void T394f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_equality_expression */
extern void T380f194(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_equality_expression_validity */
extern void T380f249(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_equality_expression */
extern void T380f311(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_equality_expression */
extern void T67f1415(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_equality_expression */
extern void T67f720(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_equality_function_name */
extern void T67f719(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3, T0* a4);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION].force_last */
extern void T279f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EQUALITY_EXPRESSION].force */
extern void T664f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_EQUALITY_EXPRESSION].fill_with */
extern void T663f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_EQUALITY_EXPRESSION].extend */
extern void T663f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION].resize */
extern void T279f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EQUALITY_EXPRESSION].aliased_resized_area */
extern T0* T664f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_EQUALITY_EXPRESSION].aliased_resized_area */
extern T0* T663f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION].new_capacity */
extern T6 T279f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION].extendible */
extern T1 T279f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_SYMBOL_OPERATOR.is_not_equal */
extern T1 T994f101(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_not_equal */
extern T1 T622f100(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_not_equal */
extern T1 T341f87(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.process */
extern void T312f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_object_equality_expression */
extern void T1327f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_object_equality_expression */
extern void T71f274(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_object_equality_expression_validity */
extern void T71f340(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_object_equality_expression */
extern void T71f403(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_object_equality_expression_target_dynamic_types */
extern void T71f451(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.put_object_equality_expression */
extern void T451f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_OBJECT_EQUALITY_EXPRESSION.set_next */
extern void T2043f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.put_object_equality_expression */
extern void T450f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.put_object_equality_expression */
extern void T302f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.put_object_equality_expression */
extern void T319f86(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_object_equality_expression */
extern void T223f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_OBJECT_EQUALITY_EXPRESSION.make */
extern T0* T2043c6(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vweq0b_error */
extern void T53f453(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vweq0b */
extern T0* T785c1076(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vweq0b_error */
extern void T36f453(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.set_left */
extern void T312f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.set_right */
extern void T312f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.name */
extern T0* T312f16(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.twin */
extern T0* T184f102(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.tokens */
extern T0* T312f19(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_object_equality_expression */
extern void T398f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_object_equality_expression */
extern void T394f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_object_equality_expression */
extern void T380f195(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_object_equality_expression_validity */
extern void T380f250(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_object_equality_expression */
extern void T380f312(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_object_equality_expression */
extern void T67f1414(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_object_equality_expression */
extern void T67f716(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_object_equality_function_name */
extern void T67f715(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3, T0* a4);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION].force_last */
extern void T278f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_EQUALITY_EXPRESSION].force */
extern void T661f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_OBJECT_EQUALITY_EXPRESSION].fill_with */
extern void T660f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_OBJECT_EQUALITY_EXPRESSION].extend */
extern void T660f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION].resize */
extern void T278f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_EQUALITY_EXPRESSION].aliased_resized_area */
extern T0* T661f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_OBJECT_EQUALITY_EXPRESSION].aliased_resized_area */
extern T0* T660f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION].new_capacity */
extern T6 T278f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION].extendible */
extern T1 T278f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_SYMBOL_OPERATOR.is_not_tilde */
extern T1 T994f100(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.is_not_tilde */
extern T1 T622f99(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.is_not_tilde */
extern T1 T341f86(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.process */
extern void T311f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_object_test */
extern void T1327f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_object_test */
extern void T71f278(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_object_test_validity */
extern void T71f344(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_object_test */
extern void T71f406(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_object_test_type */
extern void T71f405(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_OBJECT_TEST.type */
extern T0* T311f16(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_object_test */
extern void T398f51(GE_context* ac, T0* C, T0* a1);
extern T1 T398f51ot1(T0* a1);
extern T1 T398f51ot2(T0* a1, T0** a2);
extern T1 T398f51ot3(T0* a1, T0** a2);
/* detachable ET_OLD_OBJECT_TEST.expression */
extern T0* T1549f9(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.expression */
extern T0* T736f5(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.expression */
extern T0* T311f4(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_object_test */
extern void T394f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_object_test */
extern void T380f196(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_object_test_validity */
extern void T380f251(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_object_test */
extern void T380f314(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.report_object_test_type */
extern void T380f313(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_object_test */
extern void T67f1413(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_object_test */
extern void T67f1205(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_object_test_function_name */
extern void T67f711(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3, T0* a4);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST].force_last */
extern void T277f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST].force */
extern void T658f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST].fill_with */
extern void T657f9(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST].extend */
extern void T657f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST].resize */
extern void T277f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST].aliased_resized_area */
extern T0* T658f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST].aliased_resized_area */
extern T0* T657f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST].new_capacity */
extern T6 T277f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST].extendible */
extern T1 T277f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_object_test_local_name */
extern void T67f1103(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_OLD_OBJECT_TEST.type */
extern T0* T1549f14(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.type */
extern T0* T736f11(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.name */
extern T0* T311f15(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.name */
extern T0* T1549f2(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.name */
extern T0* T736f2(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.process */
extern void T1557f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_external_procedure_inline_agent */
extern void T1327f71(GE_context* ac, T0* C, T0* a1);
extern T1 T1327f71ot1(T0* a1, T0** a2);
/* [detachable] ET_PRECURSOR_CHECKER.process_agent_argument_operand_list */
extern void T1327f96(GE_context* ac, T0* C, T0* a1);
/* detachable ET_AGENT_TYPED_OPEN_ARGUMENT.process */
extern void T1559f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_agent_typed_open_argument */
extern void T1327f110(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUESTION_MARK_SYMBOL.process */
extern void T1005f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_question_mark_symbol */
extern void T1327f115(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_external_procedure_inline_agent */
extern void T71f290(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_external_procedure_inline_agent_validity */
extern void T71f356(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpir3b_error */
extern void T53f445(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vpir3b */
extern T0* T785c1068(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vpir3_error */
extern T1 T53f127(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vpir3b_error */
extern void T36f445(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vpir3_error */
extern T1 T36f127(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_procedure_inline_agent_validity */
extern void T71f421(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_procedure_inline_agent */
extern void T71f468(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_inline_agent */
extern void T71f489(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T71f489ot1(T0* a1, T0** a2);
extern T1 T71f489ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_agent_closed_operands_dynamic_types */
extern void T71f499(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T71f499ot1(T0* a1, T0** a2);
/* detachable ET_CALL_AGENT.target */
extern T0* T1041f3(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_CAST_EXPRESSION.is_open_operand */
extern T1 T1980f15(GE_context* ac, T0* C);
/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT.is_open_operand */
extern T1 T2027f4(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_TO_EXPRESSION.is_open_operand */
extern T1 T2048f19(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.is_open_operand */
extern T1 T2047f18(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.is_open_operand */
extern T1 T2046f15(GE_context* ac, T0* C);
/* detachable ET_AGENT_OPEN_TARGET.is_open_operand */
extern T1 T1558f6(GE_context* ac, T0* C);
/* [detachable] ET_AGENT_IMPLICIT_CURRENT_TARGET.is_open_operand */
extern T1 T1811f4(GE_context* ac, T0* C);
/* detachable ET_AGENT_TYPED_OPEN_ARGUMENT.is_open_operand */
extern T1 T1559f7(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.is_open_operand */
extern T1 T1557f31(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.is_open_operand */
extern T1 T1556f37(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.is_open_operand */
extern T1 T1555f33(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.is_open_operand */
extern T1 T1554f32(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.is_open_operand */
extern T1 T1553f38(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.is_open_operand */
extern T1 T1552f34(GE_context* ac, T0* C);
/* detachable ET_ITERATION_CURSOR.is_open_operand */
extern T1 T1551f17(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.is_open_operand */
extern T1 T1550f20(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.is_open_operand */
extern T1 T1549f24(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.is_open_operand */
extern T1 T1548f17(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.is_open_operand */
extern T1 T1547f22(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.is_open_operand */
extern T1 T1546f18(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.is_open_operand */
extern T1 T1543f22(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.is_open_operand */
extern T1 T1542f17(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.is_open_operand */
extern T1 T1541f17(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.is_open_operand */
extern T1 T1540f17(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.is_open_operand */
extern T1 T1539f17(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.is_open_operand */
extern T1 T1538f21(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_open_operand */
extern T1 T1496f32(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_open_operand */
extern T1 T1495f31(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.is_open_operand */
extern T1 T1494f26(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.is_open_operand */
extern T1 T1493f26(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.is_open_operand */
extern T1 T1491f28(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.is_open_operand */
extern T1 T1490f28(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_open_operand */
extern T1 T1489f47(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_open_operand */
extern T1 T1488f47(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_open_operand */
extern T1 T1487f47(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_open_operand */
extern T1 T1486f47(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_open_operand */
extern T1 T1485f36(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_open_operand */
extern T1 T1484f32(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.is_open_operand */
extern T1 T1483f31(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_open_operand */
extern T1 T1482f32(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.is_open_operand */
extern T1 T1270f24(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.is_open_operand */
extern T1 T1267f22(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.is_open_operand */
extern T1 T1254f31(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.is_open_operand */
extern T1 T1232f11(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.is_open_operand */
extern T1 T1218f24(GE_context* ac, T0* C);
/* detachable ET_INSPECT_EXPRESSION.is_open_operand */
extern T1 T1189f20(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.is_open_operand */
extern T1 T1167f23(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.is_open_operand */
extern T1 T1041f24(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.is_open_operand */
extern T1 T1038f21(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.is_open_operand */
extern T1 T1008f32(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.is_open_operand */
extern T1 T1005f36(GE_context* ac, T0* C);
/* detachable ET_VOID.is_open_operand */
extern T1 T997f25(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.is_open_operand */
extern T1 T736f21(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.is_open_operand */
extern T1 T635f12(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_open_operand */
extern T1 T372f38(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.is_open_operand */
extern T1 T328f19(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.is_open_operand */
extern T1 T313f15(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.is_open_operand */
extern T1 T312f15(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.is_open_operand */
extern T1 T311f18(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.is_open_operand */
extern T1 T309f19(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.is_open_operand */
extern T1 T288f22(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.is_open_operand */
extern T1 T287f19(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.is_open_operand */
extern T1 T282f19(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_open_operand */
extern T1 T184f62(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.formal_arguments */
extern T0* T1554f10(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.formal_arguments */
extern T0* T1553f3(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.formal_arguments */
extern T0* T1552f3(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.actual_arguments */
extern T0* T1557f6(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.actual_arguments */
extern T0* T1556f7(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.actual_arguments */
extern T0* T1555f7(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.actual_arguments */
extern T0* T1554f6(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.actual_arguments */
extern T0* T1553f9(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.actual_arguments */
extern T0* T1552f9(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_inline_agent_result_dynamic_types */
extern void T71f498(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.implicit_result */
extern T0* T1557f20(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.implicit_result */
extern T0* T1556f23(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.implicit_result */
extern T0* T1555f21(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.implicit_result */
extern T0* T1554f13(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.implicit_result */
extern T0* T1553f6(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.implicit_result */
extern T0* T1552f6(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.target */
extern T0* T1557f12(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.target */
extern T0* T1556f14(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.target */
extern T0* T1555f13(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.target */
extern T0* T1554f14(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.target */
extern T0* T1553f16(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.target */
extern T0* T1552f15(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.procedure_identity_type */
extern T0* T187f70(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.procedure_identity_type */
extern T0* T58f108(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.procedure_identity_type */
extern T0* T40f77(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.tuple_identity_type */
extern T0* T187f27(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.tuple_identity_type */
extern T0* T58f89(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.tuple_identity_type */
extern T0* T40f39(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_agent_arguments_validity */
extern void T71f466(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T71f466ot1(T0* a1, T0** a2);
extern T1 T71f466ot2(T0* a1, T0** a2);
extern T1 T71f466ot3(T0* a1, T0** a2);
extern T1 T71f466ot4(T0* a1, T0** a2);
extern T1 T71f466ot5(T0* a1, T0** a2);
extern T1 T71f466ot6(T0* a1, T0** a2);
extern T1 T71f466ot7(T0* a1, T0** a2);
extern T1 T71f466ot8(T0* a1, T0** a2);
extern T1 T71f466ot9(T0* a1, T0** a2);
extern T1 T71f466ot10(T0* a1, T0** a2);
extern T1 T71f466ot11(T0* a1, T0** a2);
extern T1 T71f466ot12(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpca5b_error */
extern void T53f519(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T6 a5, T0* a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vpca5b */
extern T0* T785c1142(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T6 a5, T0* a6, T0* a7);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vpca5_error */
extern T1 T53f155(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vpca5b_error */
extern void T36f519(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T6 a5, T0* a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.reportable_vpca5_error */
extern T1 T36f155(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpca5a_error */
extern void T53f518(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* [detachable] ET_VALIDITY_ERROR.make_vpca5a */
extern T0* T785c1141(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* [detachable] ET_ERROR_HANDLER.report_vpca5a_error */
extern void T36f518(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpca4b_error */
extern void T53f517(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T6 a5, T0* a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vpca4b */
extern T0* T785c1140(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T6 a5, T0* a6, T0* a7);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vpca4_error */
extern T1 T53f154(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vpca4b_error */
extern void T36f517(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T6 a5, T0* a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.reportable_vpca4_error */
extern T1 T36f154(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpca4a_error */
extern void T53f516(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* [detachable] ET_VALIDITY_ERROR.make_vpca4a */
extern T0* T785c1139(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* [detachable] ET_ERROR_HANDLER.report_vpca4a_error */
extern void T36f516(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T6 a6, T0* a7, T0* a8);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.put */
extern void T1025f15(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpca3b_error */
extern void T53f515(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vpca3b */
extern T0* T785c1138(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vpca3_error */
extern T1 T53f153(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vpca3b_error */
extern void T36f515(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vpca3_error */
extern T1 T36f153(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpca3a_error */
extern void T53f514(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vpca3a */
extern T0* T785c1137(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vpca3a_error */
extern void T36f514(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_CALL_AGENT.is_qualified_call */
extern T1 T1041f14(GE_context* ac, T0* C);
extern T1 T1041f14ot1(T0* a1);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.is_empty */
extern T1 T1025f9(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.force_first */
extern void T240f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.resize */
extern void T240f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACTUAL_PARAMETER_ITEM].aliased_resized_area */
extern T0* T624f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ACTUAL_PARAMETER_ITEM].aliased_resized_area */
extern T0* T621f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.new_capacity */
extern T6 T240f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.capacity */
extern T6 T240f23(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.set_implicit_agent_open_arguments */
extern void T71f465(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CALL_AGENT.set_arguments */
extern void T1041f36(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.set_actual_arguments */
extern void T1557f42(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_actual_arguments */
extern void T1556f48(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_actual_arguments */
extern void T1555f44(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.set_actual_arguments */
extern void T1554f43(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_actual_arguments */
extern void T1553f49(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_actual_arguments */
extern void T1552f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT_LIST.put_last */
extern void T2026f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT].force */
extern void T2078f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT].fill_with */
extern void T2077f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT].extend */
extern void T2077f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT_LIST.fixed_array */
extern T0* T2026f6(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT].default_create */
extern T0* T2078c2(GE_context* ac);
/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT.make */
extern T0* T2027c6(GE_context* ac, T0* a1, T6 a2);
/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT_LIST.make_with_capacity */
extern T0* T2026c7(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT].make */
extern T0* T2078f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_AGENT_IMPLICIT_OPEN_ARGUMENT].make_empty */
extern T0* T2077c4(GE_context* ac, T6 a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.formal_arguments */
extern T0* T1557f10(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.formal_arguments */
extern T0* T1556f3(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.formal_arguments */
extern T0* T1555f3(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].remove_last */
extern void T386f15(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_INLINE_AGENT].keep_head */
extern void T712f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_INLINE_AGENT].set_count */
extern void T712f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].move_last_cursors_after */
extern void T386f17(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INLINE_AGENT].set_next_cursor */
extern void T713f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INLINE_AGENT].set_after */
extern void T713f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].last */
extern T0* T386f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].is_empty */
extern T1 T386f8(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_inline_agent_iteration_components_validity */
extern void T71f418(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_inline_agent_object_tests_validity */
extern void T71f417(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuot3b_error */
extern void T53f471(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vuot3b */
extern T0* T785c1094(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vuot3b_error */
extern void T36f471(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_inline_agent_formal_arguments_validity */
extern void T71f414(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_inline_agent_argument_supplier */
extern void T71f462(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_inline_agent_formal_argument_declaration */
extern void T71f461(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpir1g_error */
extern void T53f470(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vpir1g */
extern T0* T785c1093(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vpir1_error */
extern T1 T53f137(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vpir1g_error */
extern void T36f470(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vpir1_error */
extern T1 T36f137(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ITERATION_ITEM_SCOPE.hidden_iteration_component */
extern T0* T396f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ITERATION_COMPONENT].item */
extern T0* T744f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpir1e_error */
extern void T53f469(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vpir1e */
extern T0* T785c1092(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vpir1e_error */
extern void T36f469(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_OBJECT_TEST_SCOPE.hidden_object_test */
extern T0* T393f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST].item */
extern T0* T739f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_PROCEDURE.locals */
extern T0* T1516f81(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.locals */
extern T0* T253f82(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.locals */
extern T0* T1515f18(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.locals */
extern T0* T1513f73(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.locals */
extern T0* T1512f19(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.locals */
extern T0* T1511f17(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.locals */
extern T0* T371f74(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.locals */
extern T0* T369f73(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.locals */
extern T0* T359f12(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.locals */
extern T0* T360f70(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.locals */
extern T0* T469f19(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.locals */
extern T0* T476f72(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpir1b_error */
extern void T53f468(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vpir1b */
extern T0* T785c1091(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vpir1b_error */
extern void T36f468(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_LOCAL_VARIABLE_LIST.local_variable */
extern T0* T1212f4(GE_context* ac, T0* C, T6 a1);
/* detachable ET_LOCAL_VARIABLE.local_variable */
extern T0* T1206f7(GE_context* ac, T0* C);
/* [detachable] ET_LOCAL_COMMA_VARIABLE.local_variable */
extern T0* T1768f7(GE_context* ac, T0* C);
/* detachable ET_LOCAL_VARIABLE_LIST.item */
extern T0* T1212f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_LOCAL_VARIABLE_LIST.index_of */
extern T6 T1212f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LOCAL_COMMA_VARIABLE.name */
extern T0* T1768f8(GE_context* ac, T0* C);
/* detachable ET_LOCAL_VARIABLE.name */
extern T0* T1206f8(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.locals */
extern T0* T1557f28(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.locals */
extern T0* T1556f9(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.locals */
extern T0* T1555f9(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.locals */
extern T0* T1554f29(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.locals */
extern T0* T1553f11(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.locals */
extern T0* T1552f11(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpir1a_error */
extern void T53f467(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vpir1a */
extern T0* T785c1090(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vpir1a_error */
extern void T36f467(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].item */
extern T0* T386f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vrfa0b_error */
extern void T53f466(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vrfa0b */
extern T0* T785c1089(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vrfa0b_error */
extern void T36f466(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vred0c_error */
extern void T53f465(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vred0c */
extern T0* T785c1088(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vred0c_error */
extern void T36f465(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vucr0i_error */
extern void T53f442(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vucr0i */
extern T0* T785c1065(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vucr0i_error */
extern void T36f442(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ITERATION_ITEM_SCOPE.hide_iteration_components */
extern void T396f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ITERATION_ITEM_SCOPE.count */
extern T6 T396f4(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST_SCOPE.hide_object_tests */
extern void T393f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].force_last */
extern void T386f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_AGENT].force */
extern void T711f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INLINE_AGENT].fill_with */
extern void T712f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INLINE_AGENT].extend */
extern void T712f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].resize */
extern void T386f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_AGENT].aliased_resized_area */
extern T0* T711f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_INLINE_AGENT].aliased_resized_area */
extern T0* T712f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].new_capacity */
extern T6 T386f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].extendible */
extern T1 T386f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_external_procedure_inline_agent */
extern void T398f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_external_procedure_inline_agent */
extern void T394f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_external_procedure_inline_agent */
extern void T380f186(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_external_procedure_inline_agent_validity */
extern void T380f242(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_procedure_inline_agent_validity */
extern void T380f302(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_procedure_inline_agent */
extern void T380f355(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_agent_arguments_validity */
extern void T380f353(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T380f353ot1(T0* a1, T0** a2);
extern T1 T380f353ot2(T0* a1, T0** a2);
extern T1 T380f353ot3(T0* a1, T0** a2);
extern T1 T380f353ot4(T0* a1, T0** a2);
extern T1 T380f353ot5(T0* a1, T0** a2);
extern T1 T380f353ot6(T0* a1, T0** a2);
extern T1 T380f353ot7(T0* a1, T0** a2);
extern T1 T380f353ot8(T0* a1, T0** a2);
extern T1 T380f353ot9(T0* a1, T0** a2);
extern T1 T380f353ot10(T0* a1, T0** a2);
extern T1 T380f353ot11(T0* a1, T0** a2);
extern T1 T380f353ot12(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.set_implicit_agent_open_arguments */
extern void T380f352(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_inline_agent_iteration_components_validity */
extern void T380f299(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_inline_agent_object_tests_validity */
extern void T380f298(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_inline_agent_formal_arguments_validity */
extern void T380f295(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_inline_agent_argument_supplier */
extern void T380f349(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SUPPLIER_BUILDER.report_inline_agent_argument_supplier */
extern void T1341f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_inline_agent_formal_argument_declaration */
extern void T380f348(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INVARIANTS.locals */
extern T0* T1198f17(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.arguments */
extern T0* T1198f16(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.process_external_procedure_inline_agent */
extern void T67f1203(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_procedure_inline_agent */
extern void T67f1251(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_agent */
extern void T67f1255(GE_context* ac, T0* C, T0* a1);
extern T1 T67f1255ot1(T0* a1, T0** a2);
extern T1 T67f1255ot2(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_agent_creation_name */
extern void T67f708(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3, T0* a4);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT].force_last */
extern void T280f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT].force */
extern void T667f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_AGENT].fill_with */
extern void T666f9(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_AGENT].extend */
extern void T666f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT].resize */
extern void T280f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT].aliased_resized_area */
extern T0* T667f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_AGENT].aliased_resized_area */
extern T0* T666f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT].new_capacity */
extern T6 T280f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT].extendible */
extern T1 T280f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.is_qualified_call */
extern T1 T1557f19(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.is_qualified_call */
extern T1 T1556f22(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.is_qualified_call */
extern T1 T1555f20(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.is_qualified_call */
extern T1 T1554f21(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.is_qualified_call */
extern T1 T1553f24(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.is_qualified_call */
extern T1 T1552f22(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_external_procedure_inline_agent_body_declaration */
extern void T67f1250(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.process */
extern void T1556f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_once_procedure_inline_agent */
extern void T1327f70(GE_context* ac, T0* C, T0* a1);
extern T1 T1327f70ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_once_procedure_inline_agent */
extern void T71f289(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_once_procedure_inline_agent_validity */
extern void T71f355(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpir3a_error */
extern void T53f444(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vpir3a */
extern T0* T785c1067(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vpir3a_error */
extern void T36f444(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_rescue_validity */
extern void T71f142(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_instructions_validity */
extern void T71f141(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION.process */
extern void T1756f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_unqualified_call_instruction */
extern void T1327f112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_unqualified_call_instruction */
extern void T380f414(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_call_instruction_validity */
extern void T380f430(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuex1a_error */
extern void T53f455(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vuex1a */
extern T0* T785c1078(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vuex1_error */
extern T1 T53f131(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vuex1a_error */
extern void T36f455(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vuex1_error */
extern T1 T36f131(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vkcn1c_error */
extern void T53f546(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vkcn1c */
extern T0* T785c1168(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vkcn1c_error */
extern void T36f546(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_procedure_call_instruction_validity */
extern void T380f445(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_unqualified_call_instruction */
extern void T380f453(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_EXTERNAL_PROCEDURE.is_static */
extern T1 T1516f39(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_implicitly_static */
extern T1 T1516f33(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_builtin_static */
extern T1 T1516f48(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.string_ */
extern T0* T1516f52(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_static */
extern T1 T1516f39p1(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.has_static_mark */
extern T1 T1516f34(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_static */
extern T1 T476f28(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.has_static_mark */
extern T1 T476f31(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_static */
extern T1 T1515f41(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.has_static_mark */
extern T1 T1515f34(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_static */
extern T1 T469f49(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.has_static_mark */
extern T1 T469f31(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_object_test_local_call_instruction_validity */
extern void T380f444(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuio0c_error */
extern void T53f563(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuio0c */
extern T0* T785c1185(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvuio_error */
extern T1 T53f176(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvuio0c_error */
extern void T36f563(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_gvuio_error */
extern T1 T36f176(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuio0a_error */
extern void T53f562(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuio0a */
extern T0* T785c1184(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvuio0a_error */
extern void T36f562(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuio0b_error */
extern void T53f561(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuio0b */
extern T0* T785c1183(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvuio0b_error */
extern void T36f561(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuac0c_error */
extern void T53f490(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuac0c */
extern T0* T785c1113(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvuac_error */
extern T1 T53f143(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvuac0c_error */
extern void T36f490(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_gvuac_error */
extern T1 T36f143(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.as_invariants */
extern T0* T371f76(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.as_invariants */
extern T0* T369f75(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.as_invariants */
extern T0* T360f72(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.as_invariants */
extern T0* T359f76(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.as_invariants */
extern T0* T253f84(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.as_invariants */
extern T0* T1198f21(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.as_invariants */
extern T0* T1516f83(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.as_invariants */
extern T0* T1515f77(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.as_invariants */
extern T0* T1513f75(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.as_invariants */
extern T0* T1512f78(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.as_invariants */
extern T0* T1511f76(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.as_invariants */
extern T0* T469f75(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.as_invariants */
extern T0* T476f74(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_invariants */
extern T1 T371f75(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_invariants */
extern T1 T369f74(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_invariants */
extern T1 T360f71(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_invariants */
extern T1 T359f75(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_invariants */
extern T1 T253f83(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.is_invariants */
extern T1 T1198f20(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_invariants */
extern T1 T1516f82(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_invariants */
extern T1 T1515f76(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_invariants */
extern T1 T1513f74(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_invariants */
extern T1 T1512f77(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_invariants */
extern T1 T1511f75(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_invariants */
extern T1 T469f74(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_invariants */
extern T1 T476f73(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuac0a_error */
extern void T53f489(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuac0a */
extern T0* T785c1112(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvuac0a_error */
extern void T36f489(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuac0b_error */
extern void T53f488(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuac0b */
extern T0* T785c1111(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvuac0b_error */
extern void T36f488(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_object_test_local_parenthesis_call_validity */
extern void T380f369(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_identifier_parenthesis_call_validity */
extern void T380f390(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T380f390ot1(T0* a1, T0** a2);
extern T1 T380f390ot2(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.check_object_test_local_validity */
extern void T380f259(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.object_tests */
extern T0* T1557f4(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.object_tests */
extern T0* T1556f5(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.object_tests */
extern T0* T1555f5(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.object_tests */
extern T0* T1554f4(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.object_tests */
extern T0* T1553f7(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.object_tests */
extern T0* T1552f7(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.object_tests */
extern T0* T371f21(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.object_tests */
extern T0* T369f19(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.object_tests */
extern T0* T360f17(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.object_tests */
extern T0* T359f24(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.object_tests */
extern T0* T253f27(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.object_tests */
extern T0* T1198f5(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.object_tests */
extern T0* T1516f25(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.object_tests */
extern T0* T1515f25(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.object_tests */
extern T0* T1513f24(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.object_tests */
extern T0* T1512f27(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.object_tests */
extern T0* T1511f26(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.object_tests */
extern T0* T469f20(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.object_tests */
extern T0* T476f18(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.report_object_test_local */
extern void T380f281(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].found_item */
extern T0* T392f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].item_storage_item */
extern T0* T392f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].found */
extern T1 T392f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].search */
extern void T392f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].search_position */
extern void T392f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].hash_position */
extern T6 T392f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NAMED_OBJECT_TEST.hash_code */
extern T6 T736f23(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.hash_code */
extern T6 T1549f26(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].key_storage_item */
extern T0* T392f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].clashes_item */
extern T6 T392f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].slots_item */
extern T6 T392f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen8b_error */
extern void T53f421(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_veen8b */
extern T0* T785c1044(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_veen8_error */
extern T1 T53f119(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_veen8b_error */
extern void T36f421(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_veen8_error */
extern T1 T36f119(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen8a_error */
extern void T53f420(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_veen8a */
extern T0* T785c1043(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_veen8a_error */
extern void T36f420(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_OBJECT_TEST_SCOPE.object_test */
extern T0* T393f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_iteration_item_call_instruction_validity */
extern void T380f443(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuic0c_error */
extern void T53f560(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuic0c */
extern T0* T785c1182(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvuic_error */
extern T1 T53f175(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvuic0c_error */
extern void T36f560(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_gvuic_error */
extern T1 T36f175(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuic0a_error */
extern void T53f559(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuic0a */
extern T0* T785c1181(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvuic0a_error */
extern void T36f559(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuic0b_error */
extern void T53f558(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuic0b */
extern T0* T785c1180(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvuic0b_error */
extern void T36f558(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_iteration_item_parenthesis_call_validity */
extern void T380f368(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_iteration_item_validity */
extern void T380f260(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INVARIANTS.iteration_components */
extern T0* T1198f6(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.report_iteration_item */
extern void T380f284(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].found_item */
extern T0* T395f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].item_storage_item */
extern T0* T395f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].found */
extern T1 T395f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].search */
extern void T395f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].search_position */
extern void T395f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].hash_position */
extern T6 T395f35(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REPEAT_INSTRUCTION.hash_code */
extern T6 T1264f15(GE_context* ac, T0* C);
/* detachable ET_ACROSS_INSTRUCTION.hash_code */
extern T6 T1011f18(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.hash_code */
extern T6 T1008f34(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.hash_code */
extern T6 T1254f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].key_storage_item */
extern T0* T395f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].clashes_item */
extern T6 T395f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].slots_item */
extern T6 T395f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen9b_error */
extern void T53f423(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_veen9b */
extern T0* T785c1046(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_veen9_error */
extern T1 T53f120(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_veen9b_error */
extern void T36f423(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_veen9_error */
extern T1 T36f120(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen9a_error */
extern void T53f422(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_veen9a */
extern T0* T785c1045(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_veen9a_error */
extern void T36f422(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ITERATION_ITEM_SCOPE.iteration_component */
extern T0* T396f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_local_variable_call_instruction_validity */
extern void T380f442(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuil0a_error */
extern void T53f557(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuil0a */
extern T0* T785c1179(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvuil_error */
extern T1 T53f174(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvuil0a_error */
extern void T36f557(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_gvuil_error */
extern T1 T36f174(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuil0b_error */
extern void T53f556(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuil0b */
extern T0* T785c1178(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvuil0b_error */
extern void T36f556(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvual0a_error */
extern void T53f487(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvual0a */
extern T0* T785c1110(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvual_error */
extern T1 T53f142(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvual0a_error */
extern void T36f487(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_gvual_error */
extern T1 T36f142(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvual0b_error */
extern void T53f486(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvual0b */
extern T0* T785c1109(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvual0b_error */
extern void T36f486(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_local_variable_parenthesis_call_validity */
extern void T380f367(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_local_variable_validity */
extern void T380f258(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_local_variable */
extern void T380f290(GE_context* ac, T0* C, T0* a1, T1 a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vevi0a_error */
extern void T53f456(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vevi0a */
extern T0* T785c1079(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vevi_error */
extern T1 T53f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vevi0a_error */
extern void T36f456(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vevi_error */
extern T1 T36f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.is_type_expanded */
extern T1 T241f43(GE_context* ac, T0* C);
/* [detachable] ET_NESTED_TYPE_CONTEXT.is_type_expanded_with_type_mark */
extern T1 T241f33(GE_context* ac, T0* C, T0* a1);
extern T1 T241f33ot1(T0* a1, T0** a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_type_expanded_with_type_mark */
extern T1 T1501f25(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1501f25ot1(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.is_type_expanded_with_type_mark */
extern T1 T891f23(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T891f23ot1(T0* a1, T0** a2);
/* detachable ET_LIKE_FEATURE.is_type_expanded_with_type_mark */
extern T1 T1524f27(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_type_expanded_with_type_mark */
extern T1 T1526f29(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1526f29ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_type_expanded_with_type_mark */
extern T1 T1525f22(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1525f22ot1(T0* a1);
/* [detachable] ET_LIKE_CURRENT.is_type_expanded_with_type_mark */
extern T1 T449f6(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.context_is_type_expanded_with_type_mark */
extern T1 T254f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.context_is_type_expanded_with_type_mark */
extern T1 T72f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.context_is_type_expanded_with_type_mark */
extern T1 T74f111(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_N.is_type_expanded_with_type_mark */
extern T1 T1474f22(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_type_expanded_with_type_mark */
extern T1 T242f34(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T242f34ot1(T0* a1, T0** a2);
/* [detachable] ET_ATTACHMENT_SCOPE.has_local_variable */
extern T1 T397f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LOCAL_COMMA_VARIABLE.type */
extern T0* T1768f6(GE_context* ac, T0* C);
/* detachable ET_LOCAL_VARIABLE.type */
extern T0* T1206f6(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen2c_error */
extern void T53f427(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_veen2c */
extern T0* T785c1050(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_veen2_error */
extern T1 T53f122(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_veen2c_error */
extern void T36f427(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_veen2_error */
extern T1 T36f122(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen2e_error */
extern void T53f426(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_veen2e */
extern T0* T785c1049(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_veen2e_error */
extern void T36f426(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_formal_argument_call_instruction_validity */
extern void T380f441(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuia0a_error */
extern void T53f555(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuia0a */
extern T0* T785c1177(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvuia_error */
extern T1 T53f173(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvuia0a_error */
extern void T36f555(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_gvuia_error */
extern T1 T36f173(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuia0b_error */
extern void T53f554(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuia0b */
extern T0* T785c1176(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvuia0b_error */
extern void T36f554(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuaa0a_error */
extern void T53f485(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuaa0a */
extern T0* T785c1108(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvuaa_error */
extern T1 T53f141(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvuaa0a_error */
extern void T36f485(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_gvuaa_error */
extern T1 T36f141(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuaa0b_error */
extern void T53f484(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuaa0b */
extern T0* T785c1107(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvuaa0b_error */
extern void T36f484(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_formal_argument_parenthesis_call_validity */
extern void T380f366(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_formal_argument_validity */
extern void T380f257(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_formal_argument */
extern void T380f289(GE_context* ac, T0* C, T0* a1, T1 a2, T0* a3);
/* [detachable] ET_ATTACHMENT_SCOPE.has_formal_argument */
extern T1 T397f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_call_instruction_validity */
extern void T380f372(GE_context* ac, T0* C, T0* a1);
extern T1 T380f372ot1(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vkcn1b_error */
extern void T53f521(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vkcn1b */
extern T0* T785c1144(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vkcn1b_error */
extern void T36f521(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_PARENTHESIS_INSTRUCTION.parenthesis_call */
extern T0* T2002f7(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_dotnet_procedure_call_instruction_validity */
extern void T380f391(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_unqualified_call_instruction */
extern void T71f518(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_call_instruction_validity */
extern void T71f537(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_procedure_call_instruction_validity */
extern void T71f552(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_unqualified_call_instruction */
extern void T71f562(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T71f562ot1(T0* a1, T0** a2);
extern T1 T71f562ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_object_test_local_call_instruction_validity */
extern void T71f551(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_object_test_local_parenthesis_call_validity */
extern void T71f472(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_identifier_parenthesis_call_validity */
extern void T71f490(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T71f490ot1(T0* a1, T0** a2);
extern T1 T71f490ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_object_test_local_validity */
extern void T71f239(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_object_test_local */
extern void T71f251(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_iteration_item_call_instruction_validity */
extern void T71f550(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_iteration_item_parenthesis_call_validity */
extern void T71f471(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_iteration_item_validity */
extern void T71f240(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_iteration_item */
extern void T71f252(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_local_variable_call_instruction_validity */
extern void T71f549(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_local_variable_parenthesis_call_validity */
extern void T71f470(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_local_variable_validity */
extern void T71f238(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_local_variable */
extern void T71f250(GE_context* ac, T0* C, T0* a1, T1 a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_cap_dynamic_types */
extern void T71f318(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_SYSTEM.attached_type */
extern T0* T63f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_formal_argument_call_instruction_validity */
extern void T71f548(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_formal_argument_parenthesis_call_validity */
extern void T71f469(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_formal_argument_validity */
extern void T71f237(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_formal_argument */
extern void T71f249(GE_context* ac, T0* C, T0* a1, T1 a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_call_instruction_validity */
extern void T71f236(GE_context* ac, T0* C, T0* a1);
extern T1 T71f236ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_dotnet_procedure_call_instruction_validity */
extern void T71f244(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_unqualified_call_instruction */
extern void T67f1462(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_unqualified_call_instruction */
extern void T67f1478(GE_context* ac, T0* C, T0* a1);
extern T1 T67f1478ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_named_procedure_call */
extern void T67f726(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_procedure_call */
extern void T67f747(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_non_inlined_procedure_call */
extern void T67f763(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_procedure_target_expression */
extern void T67f787(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_procedure_call */
extern void T67f762(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_procedure_call */
extern void T67f786(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_special_procedure_call */
extern void T67f785(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_special_put_call */
extern void T67f866(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ref_procedure_call */
extern void T67f784(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_ref_set_item_call */
extern void T67f865(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_procedure_procedure_call */
extern void T67f783(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_procedure_fast_call_call */
extern void T67f864(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_procedure_call_call */
extern void T67f863(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_pointer_ref_procedure_call */
extern void T67f782(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_pointer_ref_set_item_call */
extern void T67f862(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_memory_procedure_call */
extern void T67f781(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_memory_free_call */
extern void T67f861(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_procedure_c_call */
extern void T67f1081(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3, T0* a4, T1 a5);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_procedure_call */
extern void T67f780(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_unmark_object_call */
extern void T67f860(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_unlock_marking_call */
extern void T67f859(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_reference_field_at_call */
extern void T67f858(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_reference_field_call */
extern void T67f857(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_real_64_field_at_call */
extern void T67f856(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_real_64_field_call */
extern void T67f855(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_real_32_field_at_call */
extern void T67f854(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_real_32_field_call */
extern void T67f853(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_pointer_field_at_call */
extern void T67f852(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_pointer_field_call */
extern void T67f851(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_natural_64_field_at_call */
extern void T67f850(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_natural_64_field_call */
extern void T67f849(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_natural_32_field_at_call */
extern void T67f848(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_natural_32_field_call */
extern void T67f847(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_natural_16_field_at_call */
extern void T67f846(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_natural_16_field_call */
extern void T67f845(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_natural_8_field_at_call */
extern void T67f844(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_natural_8_field_call */
extern void T67f843(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_integer_64_field_at_call */
extern void T67f842(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_integer_64_field_call */
extern void T67f841(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_integer_32_field_at_call */
extern void T67f840(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_integer_32_field_call */
extern void T67f839(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_integer_16_field_at_call */
extern void T67f838(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_integer_16_field_call */
extern void T67f837(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_integer_8_field_at_call */
extern void T67f836(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_integer_8_field_call */
extern void T67f835(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_character_32_field_at_call */
extern void T67f834(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_character_32_field_call */
extern void T67f833(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_character_8_field_at_call */
extern void T67f832(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_character_8_field_call */
extern void T67f831(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_boolean_field_at_call */
extern void T67f830(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_boolean_field_call */
extern void T67f829(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_mark_object_call */
extern void T67f828(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_lock_marking_call */
extern void T67f827(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_exception_manager_procedure_call */
extern void T67f779(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_exception_manager_developer_raise_call */
extern void T67f826(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_ref_procedure_call */
extern void T67f778(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_integer_n_ref_set_item_call */
extern void T67f825(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_identified_routines_procedure_call */
extern void T67f777(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_identified_routines_eif_object_id_free_call */
extern void T67f824(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_com_failure_procedure_call */
extern void T67f776(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_com_failure_cwin_local_free_call */
extern void T67f823(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_character_n_ref_procedure_call */
extern void T67f775(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_character_n_ref_set_item_call */
extern void T67f822(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_ref_procedure_call */
extern void T67f774(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_boolean_ref_set_item_call */
extern void T67f821(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_any_procedure_call */
extern void T67f773(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_any_standard_copy_call */
extern void T67f820(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
extern T1 T67f820ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_any_copy_call */
extern void T67f819(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_tuple_label_setter_call */
extern void T67f746(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
extern T1 T67f746ot1(T0* a1, T0** a2);
/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION.is_call_agent */
extern T1 T1756f5(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_instruction_static_call */
extern void T67f1481(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.register_called_static_feature */
extern void T67f1261(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE].force_last */
extern void T295f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE].resize */
extern void T295f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE].new_capacity */
extern T6 T295f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE].extendible */
extern T1 T295f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DYNAMIC_PRECURSOR.set_static_generated */
extern void T306f46(GE_context* ac, T0* C, T1 a1);
/* detachable ET_DYNAMIC_FEATURE.set_static_generated */
extern void T236f54(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_C_GENERATOR.print_position */
extern void T67f1263(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION.position */
extern T0* T1756f4(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.line_generation_mode */
extern T1 T67f528(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_qualified_call_instruction */
extern void T67f1254(GE_context* ac, T0* C, T0* a1);
extern T1 T67f1254ot1(T0* a1, T0** a2);
/* detachable ET_DYNAMIC_PRECURSOR.procedure_call */
extern T0* T306f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].value */
extern T0* T617f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].item_storage_item */
extern T0* T617f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].position_of_key */
extern T6 T617f20(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_FEATURE.procedure_call */
extern T0* T236f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_procedure_target_operand */
extern void T67f1107(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ASSIGNER_INSTRUCTION.position */
extern T0* T1527f7(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_INSTRUCTION.position */
extern T0* T2002f4(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.position */
extern T0* T284f8(GE_context* ac, T0* C);
/* detachable ET_REPEAT_INSTRUCTION.process */
extern void T1264f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_repeat_instruction */
extern void T1327f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_compound */
extern void T1327f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_repeat_instruction */
extern void T380f405(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_repeat_instruction_validity */
extern void T380f421(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_iteration_instruction_validity */
extern void T380f431(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].remove_found_item */
extern void T395f65(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].unset_found_item */
extern void T395f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].remove_position */
extern void T395f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].key_storage_put */
extern void T395f73(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ITERATION_COMPONENT].force */
extern void T743f5(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ITERATION_COMPONENT].fill_with */
extern void T741f9(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ITERATION_COMPONENT].extend */
extern void T741f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].item_storage_put */
extern void T395f69(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].key_storage_keep_head */
extern void T395f77(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ITERATION_COMPONENT].keep_head */
extern void T741f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ITERATION_COMPONENT].set_count */
extern void T741f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].item_storage_keep_head */
extern void T395f76(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].clashes_put */
extern void T395f71(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].slots_put */
extern void T395f72(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].move_cursors_forth */
extern void T395f75(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].set_position */
extern void T740f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].set_next_cursor */
extern void T740f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].set_after */
extern void T740f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].internal_set_key_equality_tester */
extern void T395f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ITERATION_ITEM_SCOPE.remove_iteration_components */
extern void T396f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ITERATION_COMPONENT].prune_last */
extern void T744f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ITERATION_COMPONENT].move_all_cursors_after */
extern void T744f13(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ITERATION_COMPONENT].set_next_cursor */
extern void T1428f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ITERATION_COMPONENT].set_after */
extern void T1428f6(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.check_loop_component_no_from_validity */
extern void T380f436(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE.is_subset */
extern T1 T397f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].forth */
extern void T745f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].go_after */
extern void T745f85(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].cursor_go_after */
extern void T745f86(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].item_for_iteration */
extern T6 T745f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].after */
extern T1 T745f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].cursor_after */
extern T1 T745f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].start */
extern void T745f46(GE_context* ac, T0* C);
/* detachable ET_COMPOUND.is_empty */
extern T1 T355f4(GE_context* ac, T0* C);
/* detachable ET_REPEAT_INSTRUCTION.loop_compound */
extern T0* T1264f7(GE_context* ac, T0* C);
/* detachable ET_ACROSS_INSTRUCTION.loop_compound */
extern T0* T1011f8(GE_context* ac, T0* C);
/* detachable ET_LOOP_INSTRUCTION.loop_compound */
extern T0* T1530f3(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vwbe0a_error */
extern void T53f207(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vwbe0a */
extern T0* T785c857(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vwbe_error */
extern T1 T53f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vwbe0a_error */
extern void T36f207(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vwbe_error */
extern T1 T36f25(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REPEAT_INSTRUCTION.until_conditional */
extern T0* T1264f19(GE_context* ac, T0* C);
/* detachable ET_ACROSS_INSTRUCTION.until_conditional */
extern T0* T1011f7(GE_context* ac, T0* C);
/* detachable ET_LOOP_INSTRUCTION.until_conditional */
extern T0* T1530f2(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.check_loop_variant_validity */
extern void T380f327(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vave0a_error */
extern void T53f501(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vave0a */
extern T0* T785c1124(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vave_error */
extern T1 T53f147(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vave0a_error */
extern void T36f501(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vave_error */
extern T1 T36f147(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REPEAT_INSTRUCTION.variant_part */
extern T0* T1264f18(GE_context* ac, T0* C);
/* detachable ET_ACROSS_INSTRUCTION.variant_part */
extern T0* T1011f12(GE_context* ac, T0* C);
/* detachable ET_LOOP_INSTRUCTION.variant_part */
extern T0* T1530f6(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.check_loop_invariant_validity */
extern void T380f326(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REPEAT_INSTRUCTION.invariant_part */
extern T0* T1264f17(GE_context* ac, T0* C);
/* detachable ET_ACROSS_INSTRUCTION.invariant_part */
extern T0* T1011f11(GE_context* ac, T0* C);
/* detachable ET_LOOP_INSTRUCTION.invariant_part */
extern T0* T1530f5(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.check_instructions_validity */
extern void T380f120(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REPEAT_INSTRUCTION.from_compound */
extern T0* T1264f16(GE_context* ac, T0* C);
/* [detachable] ET_ITERATION_ITEM_SCOPE.add_iteration_component */
extern void T396f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ITERATION_COMPONENT].force_last */
extern void T744f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ITERATION_COMPONENT].resize */
extern void T744f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ITERATION_COMPONENT].aliased_resized_area */
extern T0* T743f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ITERATION_COMPONENT].aliased_resized_area */
extern T0* T741f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ITERATION_COMPONENT].new_capacity */
extern T6 T744f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ITERATION_COMPONENT].extendible */
extern T1 T744f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_CHECKER.check_iteration_component_header_validity */
extern void T380f325(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.report_iteration_item_declaration */
extern void T380f373(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_REPEAT_INSTRUCTION.cursor_forth_instruction */
extern T0* T1264f12(GE_context* ac, T0* C);
/* detachable ET_ACROSS_INSTRUCTION.cursor_forth_instruction */
extern T0* T1011f16(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.cursor_forth_instruction */
extern T0* T1008f17(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.cursor_forth_instruction */
extern T0* T1254f13(GE_context* ac, T0* C);
/* detachable ET_REPEAT_INSTRUCTION.cursor_after_expression */
extern T0* T1264f11(GE_context* ac, T0* C);
/* detachable ET_ACROSS_INSTRUCTION.cursor_after_expression */
extern T0* T1011f15(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.cursor_after_expression */
extern T0* T1008f16(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.cursor_after_expression */
extern T0* T1254f12(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].force_last */
extern void T395f66(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].resize */
extern void T395f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].clashes_resize */
extern void T395f81(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].special_integer_ */
extern T0* T395f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].key_storage_resize */
extern void T395f80(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].item_storage_resize */
extern void T395f79(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].slots_resize */
extern void T395f78(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].new_modulus */
extern T6 T395f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].new_capacity */
extern T6 T395f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_CHECKER.report_iteration_cursor_declaration */
extern void T380f371(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_REPEAT_INSTRUCTION.new_cursor_expression */
extern T0* T1264f9(GE_context* ac, T0* C);
/* detachable ET_ACROSS_INSTRUCTION.new_cursor_expression */
extern T0* T1011f13(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.new_cursor_expression */
extern T0* T1008f14(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.new_cursor_expression */
extern T0* T1254f10(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_voit2e_error */
extern void T53f500(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_voit2e */
extern T0* T785c1123(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_voit2_error */
extern T1 T53f146(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_voit2e_error */
extern void T36f500(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_voit2_error */
extern T1 T36f146(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_voit2d_error */
extern void T53f499(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_voit2d */
extern T0* T785c1122(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_voit2d_error */
extern void T36f499(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_voit2c_error */
extern void T53f498(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_voit2c */
extern T0* T785c1121(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_voit2c_error */
extern void T36f498(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_voit2b_error */
extern void T53f497(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_voit2b */
extern T0* T785c1120(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_voit2b_error */
extern void T36f497(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_voit2a_error */
extern void T53f496(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_voit2a */
extern T0* T785c1119(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_voit2a_error */
extern void T36f496(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_voit1a_error */
extern void T53f495(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_voit1a */
extern T0* T785c1118(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_voit1_error */
extern T1 T53f145(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_voit1a_error */
extern void T36f495(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_voit1_error */
extern T1 T36f145(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.iterable_detachable_separate_any_type */
extern T0* T187f56(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.iterable_detachable_separate_any_type */
extern T0* T58f102(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.iterable_detachable_separate_any_type */
extern T0* T40f64(GE_context* ac, T0* C);
/* detachable ET_REPEAT_INSTRUCTION.iterable_expression */
extern T0* T1264f5(GE_context* ac, T0* C);
/* detachable ET_ACROSS_INSTRUCTION.iterable_expression */
extern T0* T1011f5(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.iterable_expression */
extern T0* T1008f6(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.iterable_expression */
extern T0* T1254f6(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_repeat_instruction */
extern void T71f509(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_repeat_instruction_validity */
extern void T71f528(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_iteration_instruction_validity */
extern void T71f538(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_loop_component_no_from_validity */
extern void T71f543(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_loop_variant_validity */
extern void T71f476(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_loop_variant_validity */
extern void T71f476p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_loop_invariant_validity */
extern void T71f475(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_loop_invariant_validity */
extern void T71f475p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_iteration_component_header_validity */
extern void T71f474(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_iteration_item_declaration */
extern void T71f492(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_iteration_cursor_declaration */
extern void T71f491(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_repeat_instruction */
extern void T67f1453(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_repeat_instruction */
extern void T67f1469(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_iteration_instruction */
extern void T67f1479(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_instruction */
extern void T67f1160(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_compound */
extern void T67f1153(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REPEAT_INSTRUCTION.position */
extern T0* T1264f20(GE_context* ac, T0* C);
/* detachable ET_ACROSS_INSTRUCTION.position */
extern T0* T1011f19(GE_context* ac, T0* C);
/* detachable ET_ACROSS_INSTRUCTION.process */
extern void T1011f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_across_instruction */
extern void T1327f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_loop_invariants */
extern void T1327f89(GE_context* ac, T0* C, T0* a1);
/* detachable ET_TAGGED_ASSERTION.process */
extern void T1562f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_tagged_assertion */
extern void T1327f111(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLASS_ASSERTION.process */
extern void T1521f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_class_assertion */
extern void T1327f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_across_instruction */
extern void T380f400(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_across_instruction_validity */
extern void T380f416(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_across_instruction */
extern void T71f504(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_across_instruction_validity */
extern void T71f523(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_across_instruction */
extern void T67f1448(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_across_instruction */
extern void T67f1464(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CHECK_INSTRUCTION.process */
extern void T1046f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_check_instruction */
extern void T1327f29(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CHECK_INSTRUCTION.assertion */
extern T0* T1046f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CHECK_INSTRUCTION.item */
extern T0* T1046f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_CHECKER.process_check_instruction */
extern void T380f401(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_check_instruction_validity */
extern void T380f417(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE.set_code_unreachable */
extern void T397f20(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_IDENTIFIER.is_false */
extern T1 T184f112(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_false */
extern T1 T1489f60(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_false */
extern T1 T1488f60(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_false */
extern T1 T1487f60(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_false */
extern T1 T1486f60(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_false */
extern T1 T372f60(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_false */
extern T1 T1485f47(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_false */
extern T1 T1484f43(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_false */
extern T1 T1482f44(GE_context* ac, T0* C);
/* detachable ET_ITERATION_CURSOR.is_false */
extern T1 T1551f19(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.is_false */
extern T1 T1254f34(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.is_false */
extern T1 T1008f35(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.is_false */
extern T1 T1550f22(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.is_false */
extern T1 T736f24(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.is_false */
extern T1 T1549f27(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.is_false */
extern T1 T1539f19(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.is_false */
extern T1 T635f26(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.is_false */
extern T1 T1557f36(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.is_false */
extern T1 T1556f42(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.is_false */
extern T1 T1555f38(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.is_false */
extern T1 T1554f37(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.is_false */
extern T1 T1553f43(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.is_false */
extern T1 T1552f39(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.is_false */
extern T1 T1041f28(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.is_false */
extern T1 T1270f26(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.is_false */
extern T1 T288f26(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.is_false */
extern T1 T1218f26(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.is_false */
extern T1 T1232f20(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.is_false */
extern T1 T1038f23(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST.is_false */
extern T1 T311f22(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.is_false */
extern T1 T1548f19(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.is_false */
extern T1 T1547f24(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.is_false */
extern T1 T1546f20(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.is_false */
extern T1 T312f21(GE_context* ac, T0* C);
/* [detachable] ET_EQUALITY_EXPRESSION.is_false */
extern T1 T313f19(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.is_false */
extern T1 T1543f24(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.is_false */
extern T1 T1542f19(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.is_false */
extern T1 T1541f19(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.is_false */
extern T1 T1540f19(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.is_false */
extern T1 T1538f23(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.is_false */
extern T1 T1267f27(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.is_false */
extern T1 T287f24(GE_context* ac, T0* C);
/* detachable ET_INSPECT_EXPRESSION.is_false */
extern T1 T1189f22(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.is_false */
extern T1 T1167f25(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.is_false */
extern T1 T328f27(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_false */
extern T1 T1496f44(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_false */
extern T1 T1495f43(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.is_false */
extern T1 T282f30(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.is_false */
extern T1 T309f28(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.is_false */
extern T1 T1494f36(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.tokens */
extern T0* T1494f11(GE_context* ac, T0* C);
/* detachable ET_VOID.is_false */
extern T1 T997f27(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.is_false */
extern T1 T1493f36(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.tokens */
extern T0* T1493f11(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.is_false */
extern T1 T1491f38(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.is_false */
extern T1 T1490f38(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.is_false */
extern T1 T1483f42(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_check_instruction */
extern void T71f505(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_check_instruction_validity */
extern void T71f524(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_check_instruction_validity */
extern void T71f524p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_check_instruction */
extern void T67f1449(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_check_instruction */
extern void T67f1465(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CHECK_INSTRUCTION.position */
extern T0* T1046f9(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_INSTRUCTION.process */
extern void T1537f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_static_call_instruction */
extern void T1327f108(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_static_call_instruction */
extern void T380f413(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_static_call_instruction_validity */
extern void T380f429(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_static_parenthesis_call_validity */
extern void T380f274(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T380f274ot1(T0* a1, T0** a2);
extern T1 T380f274ot2(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.check_static_query_call_expression_validity */
extern void T380f275(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.report_static_call_expression */
extern void T380f343(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_static_feature_validity */
extern void T380f342(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T380f342ot1(T0* a1, T0** a2);
extern T1 T380f342ot2(T0* a1, T0** a2);
extern T1 T380f342ot3(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuno5a_error */
extern void T53f513(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vuno5a */
extern T0* T785c1136(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vuno5_error */
extern T1 T53f152(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vuno5a_error */
extern void T36f513(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vuno5_error */
extern T1 T36f152(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuno5b_error */
extern void T53f512(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vuno5b */
extern T0* T785c1135(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vuno5b_error */
extern void T36f512(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuno3a_error */
extern void T53f511(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vuno3a */
extern T0* T785c1134(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vuno3_error */
extern T1 T53f151(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vuno3a_error */
extern void T36f511(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vuno3_error */
extern T1 T36f151(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.report_static_supplier */
extern void T380f339(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SUPPLIER_BUILDER.report_static_supplier */
extern void T1341f14(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_STATIC_CALL_EXPRESSION.type */
extern T0* T1267f10(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_INSTRUCTION.set_parenthesis_call */
extern void T1537f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_STATIC_CALL_EXPRESSION.set_parenthesis_call */
extern void T1267f36(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_STATIC_CALL_EXPRESSION.make */
extern T0* T1267c29(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_STATIC_CALL_EXPRESSION.make_qualified_call */
extern void T1267f32(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_static_procedure_call_instruction_validity */
extern void T380f440(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.report_static_call_instruction */
extern void T380f452(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_static_dotnet_procedure_call_instruction_validity */
extern void T380f439(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_STATIC_CALL_INSTRUCTION.type */
extern T0* T1537f6(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_static_call_instruction */
extern void T71f517(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_static_call_instruction_validity */
extern void T71f536(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_static_parenthesis_call_validity */
extern void T71f394(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T71f394ot1(T0* a1, T0** a2);
extern T1 T71f394ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_static_query_call_expression_validity */
extern void T71f395(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_static_call_expression */
extern void T71f447(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T71f447ot1(T0* a1, T0** a2);
extern T1 T71f447ot2(T0* a1, T0** a2);
extern T1 T71f447ot3(T0* a1, T0** a2);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.first */
extern T0* T1977f12(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.first */
extern T0* T283f13(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.is_builtin_ise_runtime_new_tuple_instance_of */
extern T1 T306f43(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.is_builtin_ise_runtime_new_tuple_instance_of */
extern T1 T236f43(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.set_static */
extern void T451f89(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.set_static */
extern void T450f89(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.set_static */
extern void T302f85(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.set_static */
extern void T319f87(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_static */
extern void T223f83(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_static_feature_validity */
extern void T71f446(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T71f446ot1(T0* a1, T0** a2);
extern T1 T71f446ot2(T0* a1, T0** a2);
extern T1 T71f446ot3(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_static_supplier */
extern void T71f445(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_static_procedure_call_instruction_validity */
extern void T71f547(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_static_call_instruction */
extern void T71f561(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_static_dotnet_procedure_call_instruction_validity */
extern void T71f546(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_static_call_instruction */
extern void T67f1461(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_static_call_instruction */
extern void T67f1477(GE_context* ac, T0* C, T0* a1);
/* detachable ET_STATIC_CALL_INSTRUCTION.position */
extern T0* T1537f8(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_INSTRUCTION.process */
extern void T1535f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_precursor_instruction */
extern void T1327f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.check_precursor_validity */
extern void T1327f94(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENT_CLAUSE_LIST.item */
extern T0* T1237f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdpr3c_error */
extern void T53f569(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdpr3c */
extern T0* T785c1190(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.tokens */
extern T0* T785f843(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdpr3_error */
extern T1 T53f118(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdpr3c_error */
extern void T36f569(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdpr3_error */
extern T1 T36f118(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdpr3b_error */
extern void T53f568(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vdpr3b */
extern T0* T785c1189(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_PRECURSOR_INSTRUCTION.position */
extern T0* T1535f7(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_vdpr3b_error */
extern void T36f568(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_PRECURSOR_INSTRUCTION.set_parent_type */
extern void T1535f17(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECURSOR_EXPRESSION.set_parent_type */
extern void T1538f32(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECURSOR_KEYWORD.set_seed */
extern void T955f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdpr3a_error */
extern void T53f567(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vdpr3a */
extern T0* T785c1188(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vdpr3a_error */
extern void T36f567(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdpr2a_error */
extern void T53f566(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vdpr2a */
extern T0* T785c1187(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdpr2_error */
extern T1 T53f179(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdpr2a_error */
extern void T36f566(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vdpr2_error */
extern T1 T36f179(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.error_handler */
extern T0* T1327f9(GE_context* ac, T0* C);
/* [detachable] ET_PRECURSOR_CHECKER.set_fatal_error */
extern void T1327f99(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.has_master_class */
extern T1 T187f109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.has_master_class */
extern T1 T58f143(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.has_master_class */
extern T1 T40f120(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.current_universe */
extern T0* T1327f8(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.class_name */
extern T0* T184f108(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.process_precursor_instruction */
extern void T380f412(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_precursor_instruction_validity */
extern void T380f428(GE_context* ac, T0* C, T0* a1);
extern T1 T380f428ot1(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.check_precursor_parenthesis_call_validity */
extern void T380f279(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T380f279ot1(T0* a1, T0** a2);
extern T1 T380f279ot2(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.check_precursor_query_expression_validity */
extern void T380f280(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].force_last */
extern void T1340f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].slots_put */
extern void T1340f59(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].clashes_put */
extern void T1340f58(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].slots_item */
extern T6 T1340f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].hash_position */
extern T6 T1340f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].resize */
extern void T1340f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].clashes_resize */
extern void T1340f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].special_integer_ */
extern T0* T1340f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].key_storage_resize */
extern void T1340f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].item_storage_resize */
extern void T1340f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].key_storage_item */
extern T0* T1340f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].item_storage_item */
extern T0* T1340f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].clashes_item */
extern T6 T1340f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].slots_resize */
extern void T1340f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].new_modulus */
extern T6 T1340f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].new_capacity */
extern T6 T1340f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].item_storage_put */
extern void T1340f56(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].search_position */
extern void T1340f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].key_equality_tester */
extern T0* T1340f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].unset_found_item */
extern void T1340f43(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.report_precursor_expression */
extern void T380f346(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vucr0h_error */
extern void T53f463(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vucr0h */
extern T0* T785c1086(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vucr0h_error */
extern void T36f463(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_PRECURSOR_INSTRUCTION.set_parenthesis_call */
extern void T1535f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_PRECURSOR_EXPRESSION.set_parenthesis_call */
extern void T1538f33(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_PRECURSOR_EXPRESSION.set_precursor_keyword */
extern void T1538f28(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECURSOR_EXPRESSION.make */
extern T0* T1538c25(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.precursor_keyword */
extern T0* T73f1530(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_KEYWORD.make */
extern T0* T955c24(GE_context* ac);
/* detachable ET_PRECURSOR_KEYWORD.make_token */
extern void T955f26(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECURSOR_KEYWORD.make_leaf */
extern void T955f27(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.tokens */
extern T0* T1538f9(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.check_precursor_procedure_instruction_validity */
extern void T380f438(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].force_last */
extern void T1339f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].slots_put */
extern void T1339f59(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].clashes_put */
extern void T1339f58(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].slots_item */
extern T6 T1339f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].hash_position */
extern T6 T1339f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].resize */
extern void T1339f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].clashes_resize */
extern void T1339f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].special_integer_ */
extern T0* T1339f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].key_storage_resize */
extern void T1339f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].item_storage_resize */
extern void T1339f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].key_storage_item */
extern T0* T1339f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].item_storage_item */
extern T0* T1339f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].clashes_item */
extern T6 T1339f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].slots_resize */
extern void T1339f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].new_modulus */
extern T6 T1339f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].new_capacity */
extern T6 T1339f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].item_storage_put */
extern void T1339f56(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].search_position */
extern void T1339f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].key_equality_tester */
extern T0* T1339f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].unset_found_item */
extern void T1339f43(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.report_precursor_instruction */
extern void T380f451(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdpr3d_error */
extern void T53f419(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdpr3d */
extern T0* T785c1042(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdpr3d_error */
extern void T36f419(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_INVARIANTS.first_precursor */
extern T0* T1198f13(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.first_precursor */
extern T0* T1516f6(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.first_precursor */
extern T0* T253f6(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.first_precursor */
extern T0* T1515f7(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.first_precursor */
extern T0* T1513f6(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.first_precursor */
extern T0* T1512f6(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.first_precursor */
extern T0* T1511f6(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.first_precursor */
extern T0* T371f8(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.first_precursor */
extern T0* T369f7(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.first_precursor */
extern T0* T359f6(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.first_precursor */
extern T0* T360f6(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.first_precursor */
extern T0* T469f15(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.first_precursor */
extern T0* T476f15(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdpr3e_error */
extern void T53f418(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vdpr3e */
extern T0* T785c1041(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vdpr3e_error */
extern void T36f418(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdpr1a_error */
extern void T53f545(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vdpr1a */
extern T0* T785c1167(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdpr1_error */
extern T1 T53f117(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdpr1a_error */
extern void T36f545(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vdpr1_error */
extern T1 T36f117(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_precursor_instruction */
extern void T71f516(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_precursor_instruction_validity */
extern void T71f535(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_precursor_parenthesis_call_validity */
extern void T71f396(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T71f396ot1(T0* a1, T0** a2);
extern T1 T71f396ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_precursor_query_expression_validity */
extern void T71f397(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_precursor_expression */
extern void T71f448(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DYNAMIC_PRECURSOR.dynamic_precursor */
extern T0* T306f35(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DYNAMIC_FEATURE.dynamic_precursor */
extern T0* T236f39(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DYNAMIC_PRECURSOR_LIST.force_last */
extern void T307f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRECURSOR].force */
extern void T1965f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRECURSOR].fill_with */
extern void T1964f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRECURSOR].extend */
extern void T1964f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_PRECURSOR_LIST.fixed_array */
extern T0* T307f4(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRECURSOR].default_create */
extern T0* T1965c3(GE_context* ac);
/* detachable ET_DYNAMIC_PRECURSOR_LIST.resize */
extern void T307f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRECURSOR].aliased_resized_area */
extern T0* T1965f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRECURSOR].aliased_resized_area */
extern T0* T1964f4(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DYNAMIC_PRECURSOR_LIST.new_capacity */
extern T6 T307f6(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DYNAMIC_PRECURSOR_LIST.capacity */
extern T6 T307f5(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR_LIST.put_last */
extern void T307f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_PRECURSOR_LIST.make_with_capacity */
extern T0* T307c7(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_PRECURSOR].make */
extern T0* T1965f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_PRECURSOR].make_empty */
extern T0* T1964c5(GE_context* ac, T6 a1);
/* detachable ET_DYNAMIC_PRECURSOR.make */
extern T0* T306c45(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_DYNAMIC_PRECURSOR.make_feature */
extern void T306f54(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T306f54ot1(T0* a1, T0** a2);
/* detachable ET_DYNAMIC_PRECURSOR.tokens */
extern T0* T306f41(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.empty_dynamic_type_sets */
extern T0* T306f40(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.make */
extern T0* T259c15(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE_SET].make */
extern T0* T627f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE_SET].make_empty */
extern T0* T626c5(GE_context* ac, T6 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_precursor_procedure_instruction_validity */
extern void T71f545(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_precursor_instruction */
extern void T71f560(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_precursor_instruction */
extern void T67f1460(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_precursor_instruction */
extern void T67f1476(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.process */
extern void T284f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_qualified_call_instruction */
extern void T1327f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_qualified_call_instruction */
extern void T380f395(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_qualified_call_instruction */
extern void T71f234(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_qualified_call_instruction */
extern void T67f1412(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEBUG_INSTRUCTION.process */
extern void T1110f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_debug_instruction */
extern void T1327f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_debug_instruction */
extern void T380f402(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_debug_instruction_validity */
extern void T380f418(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_debug_instruction */
extern void T71f506(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_debug_instruction_validity */
extern void T71f525(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_debug_instruction_validity */
extern void T71f525p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_debug_instruction */
extern void T67f1450(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_debug_instruction */
extern void T67f1466(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INSPECT_INSTRUCTION.process */
extern void T1192f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_inspect_instruction */
extern void T1327f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_when_part_list */
extern void T1327f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_when_part */
extern void T1327f97(GE_context* ac, T0* C, T0* a1);
/* detachable ET_WHEN_PART_LIST.item */
extern T0* T1302f4(GE_context* ac, T0* C, T6 a1);
/* detachable ET_INSPECT_INSTRUCTION.expression */
extern T0* T1192f6(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.process_inspect_instruction */
extern void T380f404(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_inspect_instruction_validity */
extern void T380f420(GE_context* ac, T0* C, T0* a1);
extern T1 T380f420ot1(T0* a1, T0** a2);
extern T1 T380f420ot2(T0* a1, T0** a2);
extern T1 T380f420ot3(T0* a1, T0** a2);
extern T1 T380f420ot4(T0* a1, T0** a2);
/* [detachable] ET_ATTACHMENT_SCOPE.merge_scope */
extern void T397f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].remove */
extern void T745f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].remove_position */
extern void T745f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].move_cursors_forth */
extern void T745f76(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].internal_set_key_equality_tester */
extern void T745f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.upper */
extern T0* T635f25(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.upper */
extern T0* T372f50(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.upper */
extern T0* T184f99(GE_context* ac, T0* C);
/* detachable ET_CHOICE_RANGE.upper */
extern T0* T1534f2(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.upper */
extern T0* T1496f35(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.upper */
extern T0* T1495f34(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.upper */
extern T0* T1489f50(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.upper */
extern T0* T1488f50(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.upper */
extern T0* T1487f50(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.upper */
extern T0* T1486f50(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.upper */
extern T0* T1482f35(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.upper */
extern T0* T1267f25(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.is_range */
extern T1 T635f23(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_range */
extern T1 T372f48(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_range */
extern T1 T184f97(GE_context* ac, T0* C);
/* detachable ET_CHOICE_RANGE.is_range */
extern T1 T1534f6(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_range */
extern T1 T1496f33(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_range */
extern T1 T1495f32(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_range */
extern T1 T1489f48(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_range */
extern T1 T1488f48(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_range */
extern T1 T1487f48(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_range */
extern T1 T1486f48(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_range */
extern T1 T1482f33(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.is_range */
extern T1 T1267f23(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.set_cast_type */
extern void T1496f48(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C1_CHARACTER_CONSTANT.set_cast_type */
extern void T1495f47(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C3_CHARACTER_CONSTANT.set_cast_type */
extern void T1482f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vomb2a_error */
extern void T53f451(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vomb2a */
extern T0* T785c1074(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vomb2_error */
extern T1 T53f129(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vomb2a_error */
extern void T36f451(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vomb2_error */
extern T1 T36f129(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.set_cast_type */
extern void T372f70(GE_context* ac, T0* C, T0* a1);
/* detachable ET_BINARY_INTEGER_CONSTANT.set_cast_type */
extern void T1489f68(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OCTAL_INTEGER_CONSTANT.set_cast_type */
extern void T1488f68(GE_context* ac, T0* C, T0* a1);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.set_cast_type */
extern void T1487f68(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.set_cast_type */
extern void T1486f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vomb2b_error */
extern void T53f450(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vomb2b */
extern T0* T785c1073(GE_context* ac, T0* a1, T0* a2, T0* a3);
extern T1 T785f1073ot1(T0* a1, T0** a2);
extern T1 T785f1073ot2(T0* a1, T0** a2);
/* [detachable] ET_ERROR_HANDLER.report_vomb2b_error */
extern void T36f450(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.choice_constant */
extern T0* T380f63(GE_context* ac, T0* C, T0* a1);
extern T1 T380f63ot1(T0* a1, T0** a2);
extern T1 T380f63ot2(T0* a1, T0** a2);
extern T1 T380f63ot3(T0* a1, T0** a2);
extern T1 T380f63ot4(T0* a1, T0** a2);
extern T1 T380f63ot5(T0* a1, T0** a2);
extern T1 T380f63ot6(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.integer_choice_constant */
extern T0* T380f72(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_integer_constant */
extern T1 T1485f38(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_integer_constant */
extern T1 T1484f34(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.is_integer_constant */
extern T1 T1494f28(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.is_integer_constant */
extern T1 T1493f28(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.is_integer_constant */
extern T1 T1491f30(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.is_integer_constant */
extern T1 T1490f30(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.is_integer_constant */
extern T1 T1483f33(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_integer_constant */
extern T1 T372f52(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_integer_constant */
extern T1 T1496f37(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_integer_constant */
extern T1 T1495f36(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_integer_constant */
extern T1 T1489f52(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_integer_constant */
extern T1 T1488f52(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_integer_constant */
extern T1 T1487f52(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_integer_constant */
extern T1 T1486f52(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_integer_constant */
extern T1 T1482f37(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.character_choice_constant */
extern T0* T380f71(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.make */
extern T0* T1495c46(GE_context* ac, T3 a1);
/* detachable ET_C1_CHARACTER_CONSTANT.make_leaf */
extern void T1495f50(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_character_constant */
extern T1 T1485f37(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_character_constant */
extern T1 T1484f33(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.is_character_constant */
extern T1 T1494f27(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.is_character_constant */
extern T1 T1493f27(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.is_character_constant */
extern T1 T1491f29(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.is_character_constant */
extern T1 T1490f29(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.is_character_constant */
extern T1 T1483f32(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_character_constant */
extern T1 T372f51(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_character_constant */
extern T1 T1496f36(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_character_constant */
extern T1 T1495f35(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_character_constant */
extern T1 T1489f51(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_character_constant */
extern T1 T1488f51(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_character_constant */
extern T1 T1487f51(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_character_constant */
extern T1 T1486f51(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_character_constant */
extern T1 T1482f36(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.arguments_count */
extern T6 T184f95(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.arguments */
extern T0* T184f100(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.arguments_count */
extern T6 T635f21(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.lower */
extern T0* T635f24(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.lower */
extern T0* T372f49(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.lower */
extern T0* T184f98(GE_context* ac, T0* C);
/* detachable ET_CHOICE_RANGE.lower */
extern T0* T1534f1(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.lower */
extern T0* T1496f34(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.lower */
extern T0* T1495f33(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.lower */
extern T0* T1489f49(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.lower */
extern T0* T1488f49(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.lower */
extern T0* T1487f49(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.lower */
extern T0* T1486f49(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.lower */
extern T0* T1482f34(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.lower */
extern T0* T1267f24(GE_context* ac, T0* C);
/* detachable ET_CHOICE_LIST.choice */
extern T0* T1058f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_IDENTIFIER.choice */
extern T0* T184f93(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.choice */
extern T0* T1489f46(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.choice */
extern T0* T1488f46(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.choice */
extern T0* T1487f46(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.choice */
extern T0* T1486f46(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.choice */
extern T0* T372f47(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.choice */
extern T0* T1482f31(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.choice */
extern T0* T635f20(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.choice */
extern T0* T1267f21(GE_context* ac, T0* C);
/* detachable ET_CHOICE_RANGE.choice */
extern T0* T1534f5(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.choice */
extern T0* T1496f31(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.choice */
extern T0* T1495f30(GE_context* ac, T0* C);
/* detachable ET_CHOICE_LIST.item */
extern T0* T1058f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vomb1a_error */
extern void T53f449(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vomb1a */
extern T0* T785c1072(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vomb1_error */
extern T1 T53f128(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vomb1a_error */
extern void T36f449(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vomb1_error */
extern T1 T36f128(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_inspect_instruction */
extern void T71f508(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_inspect_instruction_validity */
extern void T71f527(GE_context* ac, T0* C, T0* a1);
extern T1 T71f527ot1(T0* a1, T0** a2);
extern T1 T71f527ot2(T0* a1, T0** a2);
extern T1 T71f527ot3(T0* a1, T0** a2);
extern T1 T71f527ot4(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.choice_constant */
extern T0* T71f88(GE_context* ac, T0* C, T0* a1);
extern T1 T71f88ot1(T0* a1, T0** a2);
extern T1 T71f88ot2(T0* a1, T0** a2);
extern T1 T71f88ot3(T0* a1, T0** a2);
extern T1 T71f88ot4(T0* a1, T0** a2);
extern T1 T71f88ot5(T0* a1, T0** a2);
extern T1 T71f88ot6(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.integer_choice_constant */
extern T0* T71f97(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.character_choice_constant */
extern T0* T71f96(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.process_inspect_instruction */
extern void T67f1452(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_inspect_instruction */
extern void T67f1468(GE_context* ac, T0* C, T0* a1);
extern T1 T67f1468ot1(T0* a1, T0** a2);
extern T1 T67f1468ot2(T0* a1, T0** a2);
extern T1 T67f1468ot3(T0* a1, T0** a2);
extern T1 T67f1468ot4(T0* a1, T0** a2);
extern T1 T67f1468ot5(T0* a1, T0** a2);
extern T1 T67f1468ot6(T0* a1, T0** a2);
/* INTEGER_32.abs */
extern T6 T6f34(GE_context* ac, T6* C);
/* INTEGER_32.abs_ref */
extern T6 T6f35(GE_context* ac, T6* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.to_integer_32 */
extern T6 T1489f41(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.to_integer_32 */
extern T6 T1488f41(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.to_integer_32 */
extern T6 T1487f41(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.to_integer_32 */
extern T6 T1486f41(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.choice_constant */
extern T0* T67f522(GE_context* ac, T0* C, T0* a1);
extern T1 T67f522ot1(T0* a1, T0** a2);
extern T1 T67f522ot2(T0* a1, T0** a2);
extern T1 T67f522ot3(T0* a1, T0** a2);
extern T1 T67f522ot4(T0* a1, T0** a2);
extern T1 T67f522ot5(T0* a1, T0** a2);
/* detachable ET_WHEN_PART.position */
extern T0* T1299f3(GE_context* ac, T0* C);
/* detachable ET_CHOICE_LIST.position */
extern T0* T1058f8(GE_context* ac, T0* C);
/* detachable ET_CHOICE_RANGE.position */
extern T0* T1534f7(GE_context* ac, T0* C);
/* detachable ET_CHOICE_LIST.first */
extern T0* T1058f10(GE_context* ac, T0* C);
/* detachable ET_CHOICE_LIST.is_empty */
extern T1 T1058f9(GE_context* ac, T0* C);
/* detachable ET_INSPECT_INSTRUCTION.position */
extern T0* T1192f7(GE_context* ac, T0* C);
/* detachable ET_IF_INSTRUCTION.process */
extern void T1170f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_if_instruction */
extern void T1327f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_elseif_part_list */
extern void T1327f90(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ELSEIF_PART.process */
extern void T1119f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_elseif_part */
extern void T1327f120(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ELSEIF_PART.expression */
extern T0* T1119f4(GE_context* ac, T0* C);
/* detachable ET_ELSEIF_PART_LIST.item */
extern T0* T1122f4(GE_context* ac, T0* C, T6 a1);
/* detachable ET_IF_INSTRUCTION.expression */
extern T0* T1170f7(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.process_if_instruction */
extern void T380f403(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_if_instruction_validity */
extern void T380f419(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_if_instruction */
extern void T71f507(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_if_instruction_validity */
extern void T71f526(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_if_instruction */
extern void T67f1451(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_if_instruction */
extern void T67f1467(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ELSEIF_PART.position */
extern T0* T1119f3(GE_context* ac, T0* C);
/* detachable ET_IF_INSTRUCTION.position */
extern T0* T1170f8(GE_context* ac, T0* C);
/* detachable ET_CREATE_INSTRUCTION.process */
extern void T1533f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_create_instruction */
extern void T1327f106(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_create_instruction */
extern void T380f411(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_create_instruction_validity */
extern void T380f427(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_creation_instruction_validity */
extern void T380f437(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcc6d_error */
extern void T53f553(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vgcc6d */
extern T0* T785c1175(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vgcc6d_error */
extern void T36f553(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.check_dotnet_creation_procedure_call_instruction_validity */
extern void T380f450(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4, T0* a5, T0* a6);
/* [detachable] ET_FEATURE_CHECKER.check_creation_procedure_call_instruction_validity */
extern void T380f449(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5, T0* a6, T0* a7);
extern T1 T380f449ot1(T0* a1, T0** a2);
extern T1 T380f449ot2(T0* a1, T0** a2);
extern T1 T380f449ot3(T0* a1);
extern T1 T380f449ot4(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.report_creation_instruction */
extern void T380f454(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.type */
extern T0* T1557f24(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.type */
extern T0* T1556f31(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.type */
extern T0* T1555f27(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.type */
extern T0* T1554f25(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.type */
extern T0* T1553f32(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.type */
extern T0* T1552f28(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.type */
extern T0* T1198f18(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.is_type_valid */
extern T1 T380f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.set_class_interface_error_only */
extern void T384f45(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcc3a_error */
extern void T53f564(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vgcc3a */
extern T0* T785c1186(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vgcc3_error */
extern T1 T53f177(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vgcc3a_error */
extern void T36f564(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vgcc3_error */
extern T1 T36f177(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_writable_validity */
extern void T380f433(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T380f433ot1(T0* a1, T0** a2);
extern T1 T380f433ot2(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.report_attribute_assignment_target */
extern void T380f448(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vucr0c_error */
extern void T53f552(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vucr0c */
extern T0* T785c1174(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vucr0c_error */
extern void T36f552(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vjaw0a_error */
extern void T53f551(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vjaw0a */
extern T0* T785c1173(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vjaw_error */
extern T1 T53f171(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vjaw0a_error */
extern void T36f551(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vjaw_error */
extern T1 T36f171(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen0a_error */
extern void T53f550(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_veen0a */
extern T0* T785c1172(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_veen_error */
extern T1 T53f172(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_veen0a_error */
extern void T36f550(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_veen_error */
extern T1 T36f172(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen0b_error */
extern void T53f549(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_veen0b */
extern T0* T785c1171(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_veen0b_error */
extern void T36f549(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vjaw0b_error */
extern void T53f548(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vjaw0b */
extern T0* T785c1170(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vjaw0b_error */
extern void T36f548(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vjaw0c_error */
extern void T53f547(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vjaw0c */
extern T0* T785c1169(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vjaw0c_error */
extern void T36f547(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_local_assignment_target */
extern void T380f447(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_result_assignment_target */
extern void T380f446(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen2a_error */
extern void T53f432(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_veen2a */
extern T0* T785c1055(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_veen2a_error */
extern void T36f432(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen2g_error */
extern void T53f446(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_veen2g */
extern T0* T785c1069(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_veen2g_error */
extern void T36f446(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CREATE_INSTRUCTION.type */
extern T0* T1533f7(GE_context* ac, T0* C);
/* detachable ET_CREATE_INSTRUCTION.last_position */
extern T0* T1533f9(GE_context* ac, T0* C);
/* detachable ET_CREATE_INSTRUCTION.last_leaf */
extern T0* T1533f11(GE_context* ac, T0* C);
/* detachable ET_BANG_INSTRUCTION.last_position */
extern T0* T1532f8(GE_context* ac, T0* C);
/* detachable ET_BANG_INSTRUCTION.last_leaf */
extern T0* T1532f10(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_create_instruction */
extern void T71f515(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_create_instruction_validity */
extern void T71f534(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_creation_instruction_validity */
extern void T71f544(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_dotnet_creation_procedure_call_instruction_validity */
extern void T71f559(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4, T0* a5, T0* a6);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_creation_procedure_call_instruction_validity */
extern void T71f558(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5, T0* a6, T0* a7);
extern T1 T71f558ot1(T0* a1, T0** a2);
extern T1 T71f558ot2(T0* a1, T0** a2);
extern T1 T71f558ot3(T0* a1);
extern T1 T71f558ot4(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_creation_instruction */
extern void T71f563(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_creation_dynamic_type */
extern void T71f564(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CREATE_INSTRUCTION.arguments */
extern T0* T1533f10(GE_context* ac, T0* C);
/* detachable ET_BANG_INSTRUCTION.arguments */
extern T0* T1532f9(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.is_type_valid */
extern T1 T71f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_writable_validity */
extern void T71f540(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T71f540ot1(T0* a1, T0** a2);
extern T1 T71f540ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_attribute_assignment_target */
extern void T71f555(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_local_assignment_target */
extern void T71f554(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_result_assignment_target */
extern void T71f553(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_result */
extern void T71f386(GE_context* ac, T0* C, T0* a1, T1 a2);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.result_index */
extern T6 T1557f14(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.result_index */
extern T6 T1556f16(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.result_index */
extern T6 T1555f15(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.result_index */
extern T6 T1554f16(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.result_index */
extern T6 T1553f18(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.result_index */
extern T6 T1552f17(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.attached_result_index */
extern T6 T1557f13(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.attached_result_index */
extern T6 T1556f15(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.attached_result_index */
extern T6 T1555f14(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.attached_result_index */
extern T6 T1554f15(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.attached_result_index */
extern T6 T1553f17(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.attached_result_index */
extern T6 T1552f16(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.process_create_instruction */
extern void T67f1459(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_create_instruction */
extern void T67f1475(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_creation_instruction */
extern void T67f1480(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_assign_result_to_once_value */
extern void T67f1138(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_IDENTIFIER.is_result */
extern T1 T184f109(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_adapted_creation_procedure_call */
extern void T67f1482(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Creation of agent #1 in feature [detachable] ET_C_GENERATOR.print_adapted_creation_procedure_call */
extern T0* T67f1482ac1(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_BANG_INSTRUCTION.process */
extern void T1532f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_bang_instruction */
extern void T1327f105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_bang_instruction */
extern void T380f410(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_bang_instruction_validity */
extern void T380f426(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_bang_instruction */
extern void T71f514(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_bang_instruction_validity */
extern void T71f533(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_bang_instruction */
extern void T67f1458(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_bang_instruction */
extern void T67f1474(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LOOP_INSTRUCTION.process */
extern void T1530f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_loop_instruction */
extern void T1327f104(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LOOP_INSTRUCTION.until_expression */
extern T0* T1530f9(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.process_loop_instruction */
extern void T380f409(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_loop_instruction_validity */
extern void T380f425(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_loop_instruction */
extern void T71f513(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_loop_instruction_validity */
extern void T71f532(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_loop_instruction */
extern void T67f1457(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_loop_instruction */
extern void T67f1473(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LOOP_INSTRUCTION.position */
extern T0* T1530f10(GE_context* ac, T0* C);
/* detachable ET_VARIANT.position */
extern T0* T1290f5(GE_context* ac, T0* C);
/* detachable ET_LOOP_INVARIANTS.position */
extern T0* T1215f9(GE_context* ac, T0* C);
/* detachable ET_TAGGED_ASSERTION.position */
extern T0* T1562f6(GE_context* ac, T0* C);
/* detachable ET_CLASS_ASSERTION.position */
extern T0* T1521f6(GE_context* ac, T0* C);
/* detachable ET_LOOP_INVARIANTS.is_empty */
extern T1 T1215f10(GE_context* ac, T0* C);
/* detachable ET_COMPOUND.position */
extern T0* T355f9(GE_context* ac, T0* C);
/* detachable ET_DEBUG_INSTRUCTION.position */
extern T0* T1110f5(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_STRING_LIST.position */
extern T0* T1224f9(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_STRING_LIST.is_empty */
extern T1 T1224f10(GE_context* ac, T0* C);
/* detachable ET_ASSIGNMENT_ATTEMPT.position */
extern T0* T1529f5(GE_context* ac, T0* C);
/* detachable ET_ASSIGNMENT.position */
extern T0* T1528f5(GE_context* ac, T0* C);
/* detachable ET_RETRY_INSTRUCTION.position */
extern T0* T991f9(GE_context* ac, T0* C);
/* detachable ET_COMPOUND.first */
extern T0* T355f10(GE_context* ac, T0* C);
/* detachable ET_RETRY_INSTRUCTION.is_null_position */
extern T1 T991f11(GE_context* ac, T0* C);
/* detachable ET_ASSIGNMENT_ATTEMPT.process */
extern void T1529f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_assignment_attempt */
extern void T1327f103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_assignment_attempt */
extern void T380f408(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_assignment_attempt_validity */
extern void T380f424(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.report_assignment_attempt */
extern void T380f435(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vjrv0a_error */
extern void T53f544(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vjrv0a */
extern T0* T785c1166(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vjrv_error */
extern T1 T53f170(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vjrv0a_error */
extern void T36f544(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vjrv_error */
extern T1 T36f170(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NESTED_TYPE_CONTEXT.is_type_reference */
extern T1 T241f50(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.newer_or_same_ise_version */
extern T1 T64f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_VERSION.is_greater_equal */
extern T1 T46f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.newer_or_same_ise_version */
extern T1 T375f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.ise_5_7_0 */
extern T0* T380f73(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_assignment_attempt */
extern void T71f512(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_assignment_attempt_validity */
extern void T71f531(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_assignment_attempt */
extern void T71f542(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_assignment_attempt_dynamic_types */
extern void T71f557(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.ise_5_7_0 */
extern T0* T71f98(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.process_assignment_attempt */
extern void T67f1456(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_assignment_attempt */
extern void T67f1472(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_HASH_LIST.is_empty */
extern T1 T268f9(GE_context* ac, T0* C);
/* detachable ET_ASSIGNMENT.process */
extern void T1528f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_assignment */
extern void T1327f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_assignment */
extern void T380f407(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_assignment_validity */
extern void T380f423(GE_context* ac, T0* C, T0* a1);
extern T1 T380f423ot1(T0* a1);
extern T1 T380f423ot2(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.report_assignment */
extern void T380f434(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE.remove_name */
extern void T397f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE.remove_attribute */
extern void T397f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE.remove_formal_argument */
extern void T397f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE.remove_local_variable */
extern void T397f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vjar0b_error */
extern void T53f543(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vjar0b */
extern T0* T785c1165(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vjar_error */
extern T1 T53f169(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vjar0b_error */
extern void T36f543(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vjar_error */
extern T1 T36f169(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE.remove_result */
extern void T397f21(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_type_attached */
extern T1 T1526f81(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_type_attached */
extern T1 T1525f84(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.is_type_attached */
extern T1 T1524f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.is_type_attached */
extern T1 T254f117(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_type_attached */
extern T1 T242f76(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.is_type_attached */
extern T1 T72f124(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.is_type_attached */
extern T1 T449f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vjar0a_error */
extern void T53f542(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vjar0a */
extern T0* T785c1164(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vjar0a_error */
extern void T36f542(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_ASSIGNMENT.set_source */
extern void T1528f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_assignment */
extern void T71f511(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_assignment_validity */
extern void T71f530(GE_context* ac, T0* C, T0* a1);
extern T1 T71f530ot1(T0* a1);
extern T1 T71f530ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_assignment */
extern void T71f541(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_assignment_dynamic_types */
extern void T71f556(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_assignment */
extern void T67f1455(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_assignment */
extern void T67f1471(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ASSIGNER_INSTRUCTION.process */
extern void T1527f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_assigner_instruction */
extern void T1327f101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_assigner_instruction */
extern void T380f406(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_assigner_instruction_validity */
extern void T380f422(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.report_tuple_label_setter */
extern void T380f432(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vbac1a_error */
extern void T53f541(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vbac1a */
extern T0* T785c1163(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vbac1_error */
extern T1 T53f168(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vbac1a_error */
extern void T36f541(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vbac1_error */
extern T1 T36f168(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ASSIGNER_INSTRUCTION.set_source */
extern void T1527f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vbac2a_error */
extern void T53f540(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vbac2a */
extern T0* T785c1162(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vbac2_error */
extern T1 T53f167(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vbac2a_error */
extern void T36f540(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vbac2_error */
extern T1 T36f167(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ASSIGNER_INSTRUCTION.set_name */
extern void T1527f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.free_call_info */
extern void T380f291(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].force_last */
extern void T390f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].force */
extern void T726f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].fill_with */
extern void T728f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].extend */
extern void T728f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].resize */
extern void T390f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].aliased_resized_area */
extern T0* T726f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].aliased_resized_area */
extern T0* T728f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].new_capacity */
extern T6 T390f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].extendible */
extern T1 T390f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_CHECKER.new_call_info */
extern T0* T380f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].remove_last */
extern void T390f15(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].keep_head */
extern void T728f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].set_count */
extern void T728f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].move_last_cursors_after */
extern void T390f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].set_next_cursor */
extern void T729f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].set_after */
extern void T729f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].last */
extern T0* T390f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].is_empty */
extern T1 T390f9(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_assigner_instruction */
extern void T71f510(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_assigner_instruction_validity */
extern void T71f529(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_tuple_label_setter */
extern void T71f539(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.free_call_info */
extern void T71f401(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.new_call_info */
extern T0* T71f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_assigner_instruction */
extern void T67f1454(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_assigner_instruction */
extern void T67f1470(GE_context* ac, T0* C, T0* a1);
/* detachable ET_RETRY_INSTRUCTION.process */
extern void T991f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_retry_instruction */
extern void T1327f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_retry_instruction */
extern void T380f399(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_retry_instruction_validity */
extern void T380f415(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vxrt0a_error */
extern void T53f539(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vxrt0a */
extern T0* T785c1161(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vxrt_error */
extern T1 T53f166(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vxrt0a_error */
extern void T36f539(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vxrt_error */
extern T1 T36f166(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_retry_instruction */
extern void T71f503(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_retry_instruction_validity */
extern void T71f522(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_retry_instruction */
extern void T67f1447(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_retry_instruction */
extern void T67f1463(GE_context* ac, T0* C, T0* a1);
/* detachable ET_COMPOUND.item */
extern T0* T355f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_once_keys_validity */
extern void T71f163(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vvok2a_error */
extern void T53f373(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vvok2a */
extern T0* T785c1003(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vvok2_error */
extern T1 T53f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vvok2a_error */
extern void T36f373(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vvok2_error */
extern T1 T36f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_STANDARD_ONCE_KEYS.is_process_key */
extern T1 T1857f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_STANDARD_ONCE_KEYS.is_thread_key */
extern T1 T1857f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vvok1b_error */
extern void T53f372(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vvok1b */
extern T0* T785c1002(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.indexing_term_value */
extern T0* T372f59(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.indexing_term_value */
extern T0* T184f111(GE_context* ac, T0* C);
/* detachable ET_CUSTOM_ATTRIBUTE.indexing_term_value */
extern T0* T1500f8(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.indexing_term_value */
extern T0* T1496f43(GE_context* ac, T0* C);
/* CHARACTER_32.out */
extern T0* T3f17(GE_context* ac, T3* C);
/* detachable ET_C1_CHARACTER_CONSTANT.indexing_term_value */
extern T0* T1495f42(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.indexing_term_value */
extern T0* T1494f35(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.indexing_term_value */
extern T0* T1493f35(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.indexing_term_value */
extern T0* T1491f37(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.indexing_term_value */
extern T0* T1490f37(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.indexing_term_value */
extern T0* T1489f59(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.indexing_term_value */
extern T0* T1488f59(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.indexing_term_value */
extern T0* T1487f59(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.indexing_term_value */
extern T0* T1486f59(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.indexing_term_value */
extern T0* T1485f46(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.indexing_term_value */
extern T0* T1484f42(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.indexing_term_value */
extern T0* T1483f41(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.indexing_term_value */
extern T0* T1482f43(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vvok1_error */
extern T1 T53f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vvok1b_error */
extern void T36f372(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vvok1_error */
extern T1 T36f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vvok1a_error */
extern void T53f371(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vvok1a */
extern T0* T785c1001(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vvok1a_error */
extern void T36f371(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_STANDARD_ONCE_KEYS.is_object_key */
extern T1 T1857f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.standard_once_keys */
extern T0* T71f75(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM].wipe_out */
extern void T401f12(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_INDEXING_TERM].keep_head */
extern void T756f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_INDEXING_TERM].set_count */
extern void T756f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM].move_all_cursors_after */
extern void T401f13(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INDEXING_TERM].set_next_cursor */
extern void T757f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INDEXING_TERM].set_after */
extern void T757f6(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vvok2b_error */
extern void T53f370(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vvok2b */
extern T0* T785c1000(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vvok2b_error */
extern void T36f370(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vvok1c_error */
extern void T53f369(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vvok1c */
extern T0* T785c999(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vvok1c_error */
extern void T36f369(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM].item */
extern T0* T401f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_INDEXING_LIST.append_tagged_indexing_terms_to_list */
extern void T1173f15(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INDEXING_TERM_LIST.append_indexing_terms_to_list */
extern void T1184f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM].force_last */
extern void T401f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM].force */
extern void T754f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INDEXING_TERM].fill_with */
extern void T756f9(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INDEXING_TERM].extend */
extern void T756f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM].resize */
extern void T401f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM].aliased_resized_area */
extern T0* T754f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_INDEXING_TERM].aliased_resized_area */
extern T0* T756f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM].new_capacity */
extern T6 T401f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM].extendible */
extern T1 T401f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_IDENTIFIER.indexing_term */
extern T0* T184f107(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.indexing_term */
extern T0* T1489f56(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.indexing_term */
extern T0* T1488f56(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.indexing_term */
extern T0* T1487f56(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.indexing_term */
extern T0* T1486f56(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.indexing_term */
extern T0* T372f56(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.indexing_term */
extern T0* T1485f42(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.indexing_term */
extern T0* T1484f38(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.indexing_term */
extern T0* T1482f41(GE_context* ac, T0* C);
/* detachable ET_CUSTOM_ATTRIBUTE.indexing_term */
extern T0* T1500f7(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.indexing_term */
extern T0* T1496f41(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.indexing_term */
extern T0* T1495f40(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.indexing_term */
extern T0* T1494f32(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.indexing_term */
extern T0* T1493f32(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.indexing_term */
extern T0* T1491f34(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.indexing_term */
extern T0* T1490f34(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.indexing_term */
extern T0* T1483f37(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.first_indexing */
extern T0* T1556f29(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_inline_agent_locals_validity */
extern void T71f419(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_inline_agent_local_supplier */
extern void T71f464(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_inline_agent_local_variable_declaration */
extern void T71f463(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LOCAL_COMMA_VARIABLE.set_attached_index */
extern void T1768f13(GE_context* ac, T0* C, T6 a1);
/* detachable ET_LOCAL_VARIABLE.set_attached_index */
extern void T1206f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_LOCAL_COMMA_VARIABLE.set_index */
extern void T1768f12(GE_context* ac, T0* C, T6 a1);
/* detachable ET_LOCAL_VARIABLE.set_index */
extern void T1206f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_local_type_validity */
extern void T71f146(GE_context* ac, T0* C, T0* a1);
extern T1 T71f146ot1(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpir1h_error */
extern void T53f478(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vpir1h */
extern T0* T785c1101(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vpir1h_error */
extern void T36f478(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpir1f_error */
extern void T53f477(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vpir1f */
extern T0* T785c1100(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vpir1f_error */
extern void T36f477(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpir1d_error */
extern void T53f476(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vpir1d */
extern T0* T785c1099(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vpir1d_error */
extern void T36f476(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpir1c_error */
extern void T53f475(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vpir1c */
extern T0* T785c1098(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vpir1c_error */
extern void T36f475(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vrlv1b_error */
extern void T53f474(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vrlv1b */
extern T0* T785c1097(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vrlv1_error */
extern T1 T53f139(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vrlv1b_error */
extern void T36f474(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vrlv1_error */
extern T1 T36f139(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vrlv2b_error */
extern void T53f473(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vrlv2b */
extern T0* T785c1096(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vrlv2_error */
extern T1 T53f138(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vrlv2b_error */
extern void T36f473(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vrlv2_error */
extern T1 T36f138(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vred0d_error */
extern void T53f472(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vred0d */
extern T0* T785c1095(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vred0d_error */
extern void T36f472(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_IDENTIFIER.set_local */
extern void T184f129(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_once_procedure_inline_agent */
extern void T398f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_once_procedure_inline_agent */
extern void T394f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_once_procedure_inline_agent */
extern void T380f185(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_once_procedure_inline_agent_validity */
extern void T380f241(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_rescue_validity */
extern void T380f121(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_once_keys_validity */
extern void T380f125(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.standard_once_keys */
extern T0* T380f55(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.check_inline_agent_locals_validity */
extern void T380f300(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_inline_agent_local_supplier */
extern void T380f351(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SUPPLIER_BUILDER.report_inline_agent_local_supplier */
extern void T1341f16(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_inline_agent_local_variable_declaration */
extern void T380f350(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_local_type_validity */
extern void T380f132(GE_context* ac, T0* C, T0* a1);
extern T1 T380f132ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.process_once_procedure_inline_agent */
extern void T67f1202(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_once_procedure_inline_agent */
extern void T67f1249(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.is_once_per_object */
extern T1 T1556f21(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.standard_once_keys */
extern T0* T1556f28(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_once_procedure_inline_agent_body_declaration */
extern void T67f1248(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_internal_routine_inline_agent_body_declaration */
extern void T67f1262(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_internal_feature_body_declaration */
extern void T67f1141(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EXTENDED_ATTRIBUTE.compound */
extern T0* T359f14(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.compound */
extern T0* T1556f11(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.compound */
extern T0* T1555f10(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.compound */
extern T0* T1553f13(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.compound */
extern T0* T1552f12(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.compound */
extern T0* T1512f20(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.compound */
extern T0* T1511f18(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.compound */
extern T0* T1515f6(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.compound */
extern T0* T469f7(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_once_mutex_unlock */
extern void T67f1130(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_once_mutex */
extern void T67f1150(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_once_per_process_mutex */
extern void T67f1159(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_once_per_object_mutex */
extern void T67f1158(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_once_per_object_mutex_name */
extern void T67f696(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.print_assign_completed_to_once_status */
extern void T67f1142(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_once_status */
extern void T67f1122(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_once_per_thread_status */
extern void T67f1146(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_once_per_process_status */
extern void T67f1145(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_once_per_object_status */
extern void T67f1144(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_once_per_object_status_name */
extern void T67f693(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.print_assign_called_to_thread_safe_once_status */
extern void T67f1134(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_thread_safe_once_per_process_status */
extern void T67f1125(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_assign_once_exception_to_thread_safe_once_exception */
extern void T67f1133(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_once_exception */
extern void T67f1147(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_once_per_thread_exception */
extern void T67f1157(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_once_per_process_exception */
extern void T67f1156(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_once_per_object_exception */
extern void T67f1155(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_once_per_object_exception_name */
extern void T67f695(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.print_thread_safe_once_per_process_exception */
extern void T67f1148(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.is_once_per_process */
extern T1 T1556f24(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.tokens */
extern T0* T1556f40(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.is_once_per_process */
extern T1 T1555f22(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.is_once_per_process */
extern T1 T1553f25(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.tokens */
extern T0* T1553f40(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.first_indexing */
extern T0* T1553f30(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.standard_once_keys */
extern T0* T1553f29(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.is_once_per_process */
extern T1 T1552f23(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_assign_last_exception_to_once_exception */
extern void T67f1154(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.reset_rescue_data */
extern void T67f1152(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].wipe_out */
extern void T297f56(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].slots_wipe_out */
extern void T297f61(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].clashes_wipe_out */
extern void T297f60(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].key_storage_wipe_out */
extern void T297f59(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].item_storage_wipe_out */
extern void T297f58(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].move_all_cursors_after */
extern void T297f57(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.rescue_clause */
extern T0* T359f13(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.rescue_clause */
extern T0* T1556f13(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.rescue_clause */
extern T0* T1555f12(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.rescue_clause */
extern T0* T1553f15(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.rescue_clause */
extern T0* T1552f14(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.rescue_clause */
extern T0* T1512f21(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.rescue_clause */
extern T0* T1511f19(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.rescue_clause */
extern T0* T1515f19(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.rescue_clause */
extern T0* T469f18(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_agent_trace_message_call */
extern void T67f1278(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] ET_C_GENERATOR.print_unindented_feature_info_message_call */
extern void T67f1151(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.is_inline_agent */
extern T1 T1556f25(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.is_inline_agent */
extern T1 T1555f23(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.is_inline_agent */
extern T1 T1553f26(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.is_inline_agent */
extern T1 T1552f24(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.process */
extern void T1555f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_do_procedure_inline_agent */
extern void T1327f69(GE_context* ac, T0* C, T0* a1);
extern T1 T1327f69ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_do_procedure_inline_agent */
extern void T71f288(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_do_procedure_inline_agent_validity */
extern void T71f354(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_do_procedure_inline_agent */
extern void T398f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_do_procedure_inline_agent */
extern void T394f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_do_procedure_inline_agent */
extern void T380f184(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_do_procedure_inline_agent_validity */
extern void T380f240(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_do_procedure_inline_agent */
extern void T67f1201(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_do_procedure_inline_agent */
extern void T67f1247(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_do_procedure_inline_agent_body_declaration */
extern void T67f1246(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.process */
extern void T1554f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_external_function_inline_agent */
extern void T1327f68(GE_context* ac, T0* C, T0* a1);
extern T1 T1327f68ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_external_function_inline_agent */
extern void T71f287(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_external_function_inline_agent_validity */
extern void T71f353(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_query_inline_agent_validity */
extern void T71f420(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_query_inline_agent */
extern void T71f467(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_LIBRARY.function_identity_any_type */
extern T0* T187f43(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.function_identity_any_type */
extern T0* T58f97(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.function_identity_any_type */
extern T0* T40f52(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.predicate_identity_type */
extern T0* T187f69(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.predicate_identity_type */
extern T0* T58f107(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.predicate_identity_type */
extern T0* T40f76(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_inline_agent_result_supplier */
extern void T71f416(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_inline_agent_result_declaration */
extern void T71f415(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.set_attached_result_index */
extern void T1557f51(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_attached_result_index */
extern void T1556f59(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_attached_result_index */
extern void T1555f54(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.set_attached_result_index */
extern void T1554f52(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_attached_result_index */
extern void T1553f60(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_attached_result_index */
extern void T1552f55(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.set_result_index */
extern void T1557f50(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_result_index */
extern void T1556f58(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_result_index */
extern void T1555f53(GE_context* ac, T0* C, T6 a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.set_result_index */
extern void T1554f51(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_result_index */
extern void T1553f59(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_result_index */
extern void T1552f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_external_function_inline_agent */
extern void T398f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_external_function_inline_agent */
extern void T394f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_external_function_inline_agent */
extern void T380f183(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_external_function_inline_agent_validity */
extern void T380f239(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_query_inline_agent_validity */
extern void T380f301(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_query_inline_agent */
extern void T380f354(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_inline_agent_result_supplier */
extern void T380f297(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SUPPLIER_BUILDER.report_inline_agent_result_supplier */
extern void T1341f13(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_inline_agent_result_declaration */
extern void T380f296(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_external_function_inline_agent */
extern void T67f1200(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_function_inline_agent */
extern void T67f1245(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_function_inline_agent_body_declaration */
extern void T67f1244(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.process */
extern void T1553f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_once_function_inline_agent */
extern void T1327f67(GE_context* ac, T0* C, T0* a1);
extern T1 T1327f67ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_once_function_inline_agent */
extern void T71f286(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_once_function_inline_agent_validity */
extern void T71f352(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vevi0d_error */
extern void T53f443(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vevi0d */
extern T0* T785c1066(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_PARENTHESIS_SYMBOL.is_null_position */
extern T1 T2020f41(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_vevi0d_error */
extern void T36f443(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ATTACHMENT_SCOPE.has_result */
extern T1 T397f6(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_once_function_inline_agent */
extern void T398f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_once_function_inline_agent */
extern void T394f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_once_function_inline_agent */
extern void T380f182(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_once_function_inline_agent_validity */
extern void T380f238(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_once_function_inline_agent */
extern void T67f1199(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_once_function_inline_agent */
extern void T67f1243(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.is_once_per_object */
extern T1 T1553f23(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_once_function_inline_agent_body_declaration */
extern void T67f1242(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.process */
extern void T1552f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_do_function_inline_agent */
extern void T1327f66(GE_context* ac, T0* C, T0* a1);
extern T1 T1327f66ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_do_function_inline_agent */
extern void T71f285(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_do_function_inline_agent_validity */
extern void T71f351(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_do_function_inline_agent */
extern void T398f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_do_function_inline_agent */
extern void T394f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_do_function_inline_agent */
extern void T380f181(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_do_function_inline_agent_validity */
extern void T380f237(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_do_function_inline_agent */
extern void T67f1198(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_do_function_inline_agent */
extern void T67f1241(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_do_function_inline_agent_body_declaration */
extern void T67f1240(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ITERATION_CURSOR.process */
extern void T1551f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_iteration_cursor */
extern void T1327f79(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_iteration_cursor */
extern void T71f298(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_iteration_cursor_validity */
extern void T71f363(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_iteration_cursor */
extern void T71f431(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_voit3a_error */
extern void T53f441(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_voit3a */
extern T0* T785c1064(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_voit3_error */
extern T1 T53f126(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_voit3a_error */
extern void T36f441(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_voit3_error */
extern T1 T36f126(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_iteration_cursor */
extern void T398f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_iteration_cursor */
extern void T394f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_iteration_cursor */
extern void T380f180(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_iteration_cursor_validity */
extern void T380f236(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_iteration_cursor */
extern void T380f294(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_iteration_cursor */
extern void T67f1197(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_iteration_cursor */
extern void T67f1239(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_MANIFEST_STRING.process */
extern void T1550f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_once_manifest_string */
extern void T1327f76(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_once_manifest_string */
extern void T71f295(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_once_manifest_string_validity */
extern void T71f360(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_once_manifest_string */
extern void T398f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_once_manifest_string */
extern void T394f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_once_manifest_string */
extern void T380f179(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_once_manifest_string_validity */
extern void T380f235(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_once_manifest_string */
extern void T67f1196(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_once_manifest_string */
extern void T67f1238(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_inline_constant_name */
extern void T67f664(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].force_last */
extern void T293f60(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].key_storage_put */
extern void T293f66(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_CONSTANT].force */
extern void T686f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INLINE_CONSTANT].fill_with */
extern void T684f9(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INLINE_CONSTANT].extend */
extern void T684f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].slots_put */
extern void T293f65(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].clashes_put */
extern void T293f64(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].slots_item */
extern T6 T293f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].hash_position */
extern T6 T293f33(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_MANIFEST_STRING.hash_code */
extern T6 T1550f17(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].resize */
extern void T293f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].clashes_resize */
extern void T293f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].special_integer_ */
extern T0* T293f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].key_storage_resize */
extern void T293f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_CONSTANT].aliased_resized_area */
extern T0* T686f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_INLINE_CONSTANT].aliased_resized_area */
extern T0* T684f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].item_storage_resize */
extern void T293f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].key_storage_item */
extern T0* T293f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].clashes_item */
extern T6 T293f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].slots_resize */
extern void T293f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].new_modulus */
extern T6 T293f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].new_capacity */
extern T6 T293f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].item_storage_put */
extern void T293f62(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].search_position */
extern void T293f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].unset_found_item */
extern void T293f45(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.process */
extern void T1549f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_old_object_test */
extern void T1327f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_old_object_test */
extern void T71f293(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_named_object_test_validity */
extern void T71f359(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_named_object_test */
extern void T71f429(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_named_object_test_dynamic_types */
extern void T71f473(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuot4b_error */
extern void T53f440(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vuot4b */
extern T0* T785c1063(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vuot4_error */
extern T1 T53f125(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vuot4b_error */
extern void T36f440(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vuot4_error */
extern T1 T36f125(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuot4a_error */
extern void T53f439(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vuot4a */
extern T0* T785c1062(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vuot4a_error */
extern void T36f439(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.ise_6_3_7_5660 */
extern T0* T71f91(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuot1e_error */
extern void T53f438(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vuot1e */
extern T0* T785c1061(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vuot1_error */
extern T1 T53f124(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vuot1e_error */
extern void T36f438(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vuot1_error */
extern T1 T36f124(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuot1d_error */
extern void T53f437(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vuot1d */
extern T0* T785c1060(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vuot1d_error */
extern void T36f437(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuot1c_error */
extern void T53f436(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vuot1c */
extern T0* T785c1059(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vuot1c_error */
extern void T36f436(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuot1b_error */
extern void T53f435(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vuot1b */
extern T0* T785c1058(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vuot1b_error */
extern void T36f435(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuot1a_error */
extern void T53f434(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vuot1a */
extern T0* T785c1057(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vuot1a_error */
extern void T36f434(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].force_last */
extern void T392f61(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].key_storage_put */
extern void T392f67(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_OBJECT_TEST].force */
extern void T738f5(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_NAMED_OBJECT_TEST].fill_with */
extern void T737f9(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_NAMED_OBJECT_TEST].extend */
extern void T737f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].slots_put */
extern void T392f66(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].clashes_put */
extern void T392f65(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].resize */
extern void T392f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].clashes_resize */
extern void T392f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].special_integer_ */
extern T0* T392f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].key_storage_resize */
extern void T392f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_OBJECT_TEST].aliased_resized_area */
extern T0* T738f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_NAMED_OBJECT_TEST].aliased_resized_area */
extern T0* T737f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].item_storage_resize */
extern void T392f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].slots_resize */
extern void T392f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].new_modulus */
extern T6 T392f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].new_capacity */
extern T6 T392f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].item_storage_put */
extern void T392f63(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].unset_found_item */
extern void T392f47(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_old_object_test */
extern void T398f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_named_object_test */
extern void T398f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_old_object_test */
extern void T394f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_named_object_test */
extern void T394f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE.add_object_test */
extern void T393f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST].force_last */
extern void T739f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST].resize */
extern void T739f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST].new_capacity */
extern T6 T739f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST].extendible */
extern T1 T739f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuot1f_error */
extern void T53f535(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vuot1f */
extern T0* T785c1157(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vuot1f_error */
extern void T36f535(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.error_handler */
extern T0* T394f7(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.set_fatal_error */
extern void T394f66(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.process_old_object_test */
extern void T380f178(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_named_object_test_validity */
extern void T380f234(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_named_object_test */
extern void T380f293(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.ise_6_3_7_5660 */
extern T0* T380f61(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.process_old_object_test */
extern void T67f1195(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OLD_EXPRESSION.process */
extern void T1548f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_old_expression */
extern void T1327f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_old_expression */
extern void T71f277(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_old_expression_validity */
extern void T71f343(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vaol1a_error */
extern void T53f433(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vaol1a */
extern T0* T785c1056(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vaol1_error */
extern T1 T53f123(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vaol1a_error */
extern void T36f433(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vaol1_error */
extern T1 T36f123(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_old_expression */
extern void T398f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_old_expression */
extern void T394f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_old_expression */
extern void T380f177(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_old_expression_validity */
extern void T380f233(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_old_expression */
extern void T67f1194(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_old_expression */
extern void T67f1237(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PREFIX_EXPRESSION.process */
extern void T1547f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_prefix_expression */
extern void T1327f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_prefix_expression */
extern void T71f276(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_prefix_expression_validity */
extern void T71f342(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_prefix_expression */
extern void T398f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_prefix_expression */
extern void T394f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_prefix_expression */
extern void T380f176(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_prefix_expression_validity */
extern void T380f232(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_prefix_expression */
extern void T67f1193(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_prefix_expression */
extern void T67f1236(GE_context* ac, T0* C, T0* a1);
/* detachable ET_MANIFEST_TYPE.process */
extern void T1546f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_manifest_type */
extern void T1327f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_manifest_type */
extern void T71f275(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_manifest_type_validity */
extern void T71f341(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_manifest_type */
extern void T71f404(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_LIBRARY.type_identity_type */
extern T0* T187f94(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.type_identity_type */
extern T0* T58f123(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.type_identity_type */
extern T0* T40f99(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_manifest_type */
extern void T398f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_manifest_type */
extern void T394f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_manifest_type */
extern void T380f175(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_manifest_type_validity */
extern void T380f231(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_manifest_type */
extern void T380f292(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_manifest_type */
extern void T67f1192(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_manifest_type */
extern void T67f1235(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INFIX_EXPRESSION.process */
extern void T1543f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_infix_expression */
extern void T1327f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_infix_expression */
extern void T71f272(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_infix_expression_validity */
extern void T71f338(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_infix_expression */
extern void T398f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_infix_expression */
extern void T394f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_infix_expression */
extern void T380f174(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_infix_expression_validity */
extern void T380f230(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_infix_expression */
extern void T67f1191(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_infix_expression */
extern void T67f1234(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXPRESSION_ADDRESS.process */
extern void T1542f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_expression_address */
extern void T1327f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_expression_address */
extern void T71f271(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_expression_address_validity */
extern void T71f337(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_pointer_expression */
extern void T71f399(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_typed_pointer_expression */
extern void T71f398(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_LIBRARY.typed_pointer_identity_type */
extern T0* T187f95(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.typed_pointer_identity_type */
extern T0* T58f35(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.typed_pointer_identity_type */
extern T0* T40f100(GE_context* ac, T0* C);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_expression_address */
extern void T398f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_expression_address */
extern void T394f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_expression_address */
extern void T380f173(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_expression_address_validity */
extern void T380f229(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_pointer_expression */
extern void T380f283(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_typed_pointer_expression */
extern void T380f282(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_expression_address */
extern void T67f1190(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_expression_address */
extern void T67f1233(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_feature_address */
extern void T67f1230(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.special_type */
extern T0* T1466f14(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.special_type */
extern T0* T269f12(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.special_type */
extern T0* T475f14(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.special_type */
extern T0* T323f24(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.special_type */
extern T0* T451f62(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.special_type */
extern T0* T450f62(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.special_type */
extern T0* T319f61(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.special_type */
extern T0* T302f59(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.special_type */
extern T0* T223f58(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.iteration_item_name */
extern T0* T184f58(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.object_test_local_name */
extern T0* T184f57(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.local_name */
extern T0* T184f56(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.argument_name */
extern T0* T184f55(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.is_empty */
extern T1 T224f9(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.make */
extern T0* T1539c21(GE_context* ac, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.dollar_symbol */
extern T0* T73f1531(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_dollar */
extern T0* T341c99(GE_context* ac);
/* detachable ET_FEATURE_ADDRESS.tokens */
extern T0* T1539f6(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.process */
extern void T1541f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_result_address */
extern void T1327f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_result_address */
extern void T71f270(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_result_address_validity */
extern void T71f336(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen2f_error */
extern void T53f431(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_veen2f */
extern T0* T785c1054(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_veen2f_error */
extern void T36f431(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen2d_error */
extern void T53f430(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_veen2d */
extern T0* T785c1053(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_veen2d_error */
extern void T36f430(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_veen2b_error */
extern void T53f429(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_veen2b */
extern T0* T785c1052(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_veen2b_error */
extern void T36f429(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_result_address */
extern void T398f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_result_address */
extern void T394f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_result_address */
extern void T380f172(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_result_address_validity */
extern void T380f228(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_result_address */
extern void T67f1189(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_result_address */
extern void T67f1232(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.result_keyword */
extern T0* T73f26(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.make */
extern T0* T309c31(GE_context* ac);
/* [detachable] ET_RESULT.make_token */
extern void T309f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_RESULT.make_leaf */
extern void T309f34(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.tokens */
extern T0* T309f5(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.process */
extern void T1540f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_current_address */
extern void T1327f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_current_address */
extern void T71f269(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_current_address_validity */
extern void T71f335(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vucr0g_error */
extern void T53f428(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vucr0g */
extern T0* T785c1051(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vucr0g_error */
extern void T36f428(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_current_address */
extern void T398f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_current_address */
extern void T394f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_current_address */
extern void T380f171(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_current_address_validity */
extern void T380f227(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_current_address */
extern void T67f1188(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_current_address */
extern void T67f1231(GE_context* ac, T0* C, T0* a1);
extern T1 T67f1231ot1(T0* a1, T0** a2);
/* detachable ET_FEATURE_ADDRESS.process */
extern void T1539f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_feature_address */
extern void T1327f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_feature_address */
extern void T71f292(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_feature_address_validity */
extern void T71f358(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vuar4a_error */
extern void T53f425(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vuar4a */
extern T0* T785c1048(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vuar4_error */
extern T1 T53f121(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vuar4a_error */
extern void T36f425(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vuar4_error */
extern T1 T36f121(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_function_address */
extern void T71f428(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_DYNAMIC_PRECURSOR.set_address */
extern void T306f55(GE_context* ac, T0* C, T1 a1);
/* detachable ET_DYNAMIC_FEATURE.set_address */
extern void T236f55(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_CLASS.is_base_type */
extern T1 T74f247(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_base_type */
extern T1 T1501f73(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.is_base_type */
extern T1 T891f75(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.is_base_type */
extern T1 T1524f70(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.is_base_type */
extern T1 T254f99(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_base_type */
extern T1 T242f66(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.is_base_type */
extern T1 T72f109(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_base_type */
extern T1 T1526f69(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_base_type */
extern T1 T1525f72(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.is_base_type */
extern T1 T449f65(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_attribute_address */
extern void T71f427(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vucr0e_error */
extern void T53f424(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vucr0e */
extern T0* T785c1047(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vucr0e_error */
extern void T36f424(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_procedure_address */
extern void T71f426(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_feature_address */
extern void T398f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_feature_address */
extern void T394f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_feature_address */
extern void T380f170(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_feature_address_validity */
extern void T380f226(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_function_address */
extern void T380f288(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_attribute_address */
extern void T380f287(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_procedure_address */
extern void T380f286(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_feature_address */
extern void T67f1187(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECURSOR_EXPRESSION.process */
extern void T1538f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_precursor_expression */
extern void T1327f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_precursor_expression */
extern void T71f268(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_precursor_expression_validity */
extern void T71f334(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdpr1b_error */
extern void T53f417(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vdpr1b */
extern T0* T785c1040(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vdpr1b_error */
extern void T36f417(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_precursor_expression */
extern void T398f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_precursor_expression */
extern void T394f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_precursor_expression */
extern void T380f169(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_precursor_expression_validity */
extern void T380f225(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T380f225ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.process_precursor_expression */
extern void T67f1186(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_precursor_expression */
extern void T67f1229(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C2_CHARACTER_CONSTANT.process */
extern void T1496f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_c2_character_constant */
extern void T1327f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_c2_character_constant */
extern void T71f262(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_c2_character_constant_validity */
extern void T71f328(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_c2_character_constant */
extern void T398f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_c2_character_constant */
extern void T394f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_c2_character_constant */
extern void T380f168(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_c2_character_constant_validity */
extern void T380f224(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_c2_character_constant */
extern void T67f1180(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C1_CHARACTER_CONSTANT.process */
extern void T1495f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_c1_character_constant */
extern void T1327f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_c1_character_constant */
extern void T71f261(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_c1_character_constant_validity */
extern void T71f327(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_c1_character_constant */
extern void T398f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_c1_character_constant */
extern void T394f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_c1_character_constant */
extern void T380f167(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_c1_character_constant_validity */
extern void T380f223(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_c1_character_constant */
extern void T67f1179(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FALSE_CONSTANT.process */
extern void T1494f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_false_constant */
extern void T1327f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_false_constant */
extern void T71f259(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_false_constant_validity */
extern void T71f325(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_boolean_constant */
extern void T71f384(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_false_constant */
extern void T398f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_false_constant */
extern void T394f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_false_constant */
extern void T380f166(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_false_constant_validity */
extern void T380f222(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_boolean_constant */
extern void T380f278(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_false_constant */
extern void T67f1178(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_false_constant */
extern void T67f1222(GE_context* ac, T0* C, T0* a1);
/* detachable ET_TRUE_CONSTANT.process */
extern void T1493f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_true_constant */
extern void T1327f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_true_constant */
extern void T71f257(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_true_constant_validity */
extern void T71f323(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_true_constant */
extern void T398f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_true_constant */
extern void T394f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_true_constant */
extern void T380f165(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_true_constant_validity */
extern void T380f221(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_true_constant */
extern void T67f1177(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_true_constant */
extern void T67f1221(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.process */
extern void T1491f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_underscored_real_constant */
extern void T1327f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_underscored_real_constant */
extern void T71f256(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_underscored_real_constant_validity */
extern void T71f322(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_real_constant_validity */
extern void T71f383(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.set_type */
extern void T1491f48(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REGULAR_REAL_CONSTANT.set_type */
extern void T1490f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_real_64_constant */
extern void T71f439(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_real_32_constant */
extern void T71f438(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vwmq0b_error */
extern void T53f462(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vwmq0b */
extern T0* T785c1085(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vwmq0b_error */
extern void T36f462(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_underscored_real_constant */
extern void T398f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_underscored_real_constant */
extern void T394f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_underscored_real_constant */
extern void T380f164(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_underscored_real_constant_validity */
extern void T380f220(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_real_constant_validity */
extern void T380f277(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_real_64_constant */
extern void T380f345(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_real_32_constant */
extern void T380f344(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_underscored_real_constant */
extern void T67f1176(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_underscored_real_constant */
extern void T67f1220(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.is_negative */
extern T1 T1491f17(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.process */
extern void T1490f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_regular_real_constant */
extern void T1327f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_regular_real_constant */
extern void T71f255(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_regular_real_constant_validity */
extern void T71f321(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_regular_real_constant */
extern void T398f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_regular_real_constant */
extern void T394f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_regular_real_constant */
extern void T380f163(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_regular_real_constant_validity */
extern void T380f219(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_regular_real_constant */
extern void T67f1175(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_regular_real_constant */
extern void T67f1219(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REGULAR_REAL_CONSTANT.is_negative */
extern T1 T1490f17(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.process */
extern void T1483f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_regular_manifest_string */
extern void T1327f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_regular_manifest_string */
extern void T71f254(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_regular_manifest_string_validity */
extern void T71f320(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_regular_manifest_string */
extern void T398f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_regular_manifest_string */
extern void T394f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_regular_manifest_string */
extern void T380f156(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_regular_manifest_string_validity */
extern void T380f212(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_regular_manifest_string */
extern void T67f1174(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_regular_manifest_string */
extern void T67f1218(GE_context* ac, T0* C, T0* a1);
/* detachable ET_STRIP_EXPRESSION.process */
extern void T1270f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_strip_expression */
extern void T1327f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_strip_expression */
extern void T71f283(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_strip_expression_validity */
extern void T71f349(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_strip_expression */
extern void T71f410(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vwst2a_error */
extern void T53f416(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vwst2a */
extern T0* T785c1039(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vwst2_error */
extern T1 T53f116(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vwst2a_error */
extern void T36f416(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vwst2_error */
extern T1 T36f116(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vwst1a_error */
extern void T53f415(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vwst1a */
extern T0* T785c1038(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vwst1_error */
extern T1 T53f115(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vwst1a_error */
extern void T36f415(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vwst1_error */
extern T1 T36f115(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vwst1b_error */
extern void T53f414(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vwst1b */
extern T0* T785c1037(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vwst1b_error */
extern void T36f414(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_STRIP_EXPRESSION.feature_name */
extern T0* T1270f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_strip_expression */
extern void T398f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_strip_expression */
extern void T394f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_strip_expression */
extern void T380f154(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_strip_expression_validity */
extern void T380f210(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_strip_expression */
extern void T380f276(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_strip_expression */
extern void T67f1173(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_strip_expression */
extern void T67f1217(GE_context* ac, T0* C, T0* a1);
/* detachable ET_STATIC_CALL_EXPRESSION.process */
extern void T1267f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_static_call_expression */
extern void T1327f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_static_call_expression */
extern void T71f267(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_static_call_expression_validity */
extern void T71f333(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_static_dotnet_query_call_expression_validity */
extern void T71f393(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_static_call_expression */
extern void T398f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_static_call_expression */
extern void T394f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_static_call_expression */
extern void T380f153(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_static_call_expression_validity */
extern void T380f209(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_static_dotnet_query_call_expression_validity */
extern void T380f273(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_static_call_expression */
extern void T67f1172(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_static_call_expression */
extern void T67f1216(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_expression_static_call */
extern void T67f1252(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T67f1252ot1(T0* a1, T0** a2);
extern T1 T67f1252ot2(T0* a1, T0** a2);
extern T1 T67f1252ot3(T0* a1, T0** a2);
/* detachable ET_QUANTIFIER_EXPRESSION.process */
extern void T1254f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_quantifier_expression */
extern void T1327f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_quantifier_expression */
extern void T71f297(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_quantifier_expression_validity */
extern void T71f362(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_iteration_expression_validity */
extern void T71f430(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_iteration_expression */
extern void T71f477(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_quantifier_expression */
extern void T398f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_quantifier_expression */
extern void T394f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_quantifier_expression */
extern void T380f152(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_quantifier_expression_validity */
extern void T380f207(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_iteration_expression_validity */
extern void T380f264(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_iteration_expression */
extern void T380f328(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_quantifier_expression */
extern void T67f1171(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_quantifier_expression */
extern void T67f1215(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_iteration_expression */
extern void T67f1253(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENTHESIZED_EXPRESSION.process */
extern void T1232f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_parenthesized_expression */
extern void T1327f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_parenthesized_expression */
extern void T71f280(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_parenthesized_expression_validity */
extern void T71f346(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_parenthesized_expression */
extern void T398f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_parenthesized_expression */
extern void T394f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_parenthesized_expression */
extern void T380f151(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_parenthesized_expression_validity */
extern void T380f206(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_parenthesized_expression */
extern void T67f1170(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_parenthesized_expression */
extern void T67f1214(GE_context* ac, T0* C, T0* a1);
/* detachable ET_MANIFEST_ARRAY.process */
extern void T1218f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_manifest_array */
extern void T1327f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_manifest_array */
extern void T71f281(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_manifest_array_validity */
extern void T71f347(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_untyped_manifest_array_validity */
extern void T71f408(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T71f408ot1(T0* a1, T0** a2);
extern T1 T71f408ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.free_common_ancestor_types */
extern void T71f390(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].keep_first */
extern void T391f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].move_all_cursors_after */
extern void T391f22(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].item */
extern T0* T391f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.expression */
extern T0* T2047f4(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.expression */
extern T0* T2046f2(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.update_common_ancestor_type_list */
extern void T71f388(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].remove */
extern void T391f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].move_left */
extern void T391f21(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].move_cursors_left */
extern void T391f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT].set_position */
extern void T732f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_MANIFEST_ARRAY.put */
extern void T1218f34(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_manifest_array */
extern void T71f452(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T71f452ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS_TYPE.conforms_to_context */
extern T1 T72f111(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.conforms_to_context_with_type_marks */
extern T1 T72f121(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS.context_named_type */
extern T0* T74f237(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_typed_manifest_array_validity */
extern void T71f407(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T71f407ot1(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vwma2a_error */
extern void T53f458(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vwma2a */
extern T0* T785c1081(GE_context* ac, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vwma2_error */
extern T1 T53f133(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vwma2a_error */
extern void T36f458(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.reportable_vwma2_error */
extern T1 T36f133(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vwma1a_error */
extern void T53f457(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vwma1a */
extern T0* T785c1080(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vwma1_error */
extern T1 T53f132(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vwma1a_error */
extern void T36f457(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vwma1_error */
extern T1 T36f132(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_manifest_array */
extern void T398f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_manifest_array */
extern void T394f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_manifest_array */
extern void T380f150(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_manifest_array_validity */
extern void T380f205(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_untyped_manifest_array_validity */
extern void T380f263(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T380f263ot1(T0* a1, T0** a2);
extern T1 T380f263ot2(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.free_common_ancestor_types */
extern void T380f309(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_FEATURE_CHECKER.update_common_ancestor_type_list */
extern void T380f307(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] ET_FEATURE_CHECKER.report_manifest_array */
extern void T380f324(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_typed_manifest_array_validity */
extern void T380f262(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T380f262ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.process_manifest_array */
extern void T67f1169(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_manifest_array */
extern void T67f1213(GE_context* ac, T0* C, T0* a1);
extern T1 T67f1213ot1(T0* a1, T0** a2);
/* detachable ET_INSPECT_EXPRESSION.process */
extern void T1189f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_inspect_expression */
extern void T1327f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_when_expression_list */
extern void T1327f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_when_expression */
extern void T1327f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_inspect_expression */
extern void T71f265(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_inspect_expression_validity */
extern void T71f331(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T71f331ot1(T0* a1, T0** a2);
extern T1 T71f331ot2(T0* a1, T0** a2);
extern T1 T71f331ot3(T0* a1, T0** a2);
extern T1 T71f331ot4(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_inspect_expression */
extern void T71f391(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_inspect_expression_dynamic_types */
extern void T71f443(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_inspect_expression */
extern void T398f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_inspect_expression */
extern void T394f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_inspect_expression */
extern void T380f192(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_inspect_expression_validity */
extern void T380f248(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T380f248ot1(T0* a1, T0** a2);
extern T1 T380f248ot2(T0* a1, T0** a2);
extern T1 T380f248ot3(T0* a1, T0** a2);
extern T1 T380f248ot4(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.report_inspect_expression */
extern void T380f310(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_inspect_expression */
extern void T67f1168(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_inspect_expression */
extern void T67f1212(GE_context* ac, T0* C, T0* a1);
extern T1 T67f1212ot1(T0* a1, T0** a2);
extern T1 T67f1212ot2(T0* a1, T0** a2);
extern T1 T67f1212ot3(T0* a1, T0** a2);
extern T1 T67f1212ot4(T0* a1, T0** a2);
extern T1 T67f1212ot5(T0* a1, T0** a2);
extern T1 T67f1212ot6(T0* a1, T0** a2);
/* detachable ET_IF_EXPRESSION.process */
extern void T1167f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_if_expression */
extern void T1327f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_elseif_expression_list */
extern void T1327f92(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ELSEIF_EXPRESSION.process */
extern void T1113f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_elseif_expression */
extern void T1327f116(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_if_expression */
extern void T71f264(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_if_expression_validity */
extern void T71f330(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_if_expression */
extern void T71f389(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_if_expression_dynamic_types */
extern void T71f442(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_if_expression */
extern void T398f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_if_expression */
extern void T394f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_if_expression */
extern void T380f191(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_if_expression_validity */
extern void T380f247(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_if_expression */
extern void T380f308(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_if_expression */
extern void T67f1167(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_if_expression */
extern void T67f1211(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CALL_AGENT.process */
extern void T1041f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_call_agent */
extern void T1327f65(GE_context* ac, T0* C, T0* a1);
extern T1 T1327f65ot1(T0* a1, T0** a2);
/* detachable ET_AGENT_OPEN_TARGET.process */
extern void T1558f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_agent_open_target */
extern void T1327f109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AGENT_IMPLICIT_CURRENT_TARGET.process */
extern void T1811f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_agent_implicit_current_target */
extern void T1327f113(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_call_agent */
extern void T71f284(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_call_agent_validity */
extern void T71f350(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T71f350ot1(T0* a1, T0** a2);
extern T1 T71f350ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_typed_call_agent_validity */
extern void T71f413(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgmc0g_error */
extern void T53f483(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_VALIDITY_ERROR.make_vgmc0g */
extern T0* T785c1106(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_ERROR_HANDLER.report_vgmc0g_error */
extern void T36f483(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* detachable ET_CALL_AGENT.is_procedure */
extern T1 T1041f22(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpca1b_error */
extern void T53f482(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vpca1b */
extern T0* T785c1105(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vpca1_error */
extern T1 T53f140(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vpca1b_error */
extern void T36f482(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vpca1_error */
extern T1 T36f140(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_typed_tuple_label_call_agent_validity */
extern void T71f460(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_tuple_label_call_agent */
extern void T71f488(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T71f488ot1(T0* a1, T0** a2);
extern T1 T71f488ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_agent_qualified_query_call */
extern void T71f497(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_AGENT_IMPLICIT_OPEN_ARGUMENT_LIST.is_empty */
extern T1 T2026f3(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_typed_query_call_agent_validity */
extern void T71f459(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_qualified_query_call_agent */
extern void T71f487(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_qualified_call_agent */
extern void T71f496(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T71f496ot1(T0* a1, T0** a2);
extern T1 T71f496ot2(T0* a1, T0** a2);
extern T1 T71f496ot3(T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpca2a_error */
extern void T53f520(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vpca2a */
extern T0* T785c1143(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vpca2_error */
extern T1 T53f156(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vpca2a_error */
extern void T36f520(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.reportable_vpca2_error */
extern T1 T36f156(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_typed_procedure_call_agent_validity */
extern void T71f458(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_qualified_procedure_call_agent */
extern void T71f486(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_CALL_AGENT.set_procedure */
extern void T1041f35(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_call_agent_validity */
extern void T71f412(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_tuple_label_call_agent_validity */
extern void T71f457(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_query_call_agent_validity */
extern void T71f456(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_qualified_procedure_call_agent_validity */
extern void T71f455(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_call_agent_validity */
extern void T71f411(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vpca1a_error */
extern void T53f481(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vpca1a */
extern T0* T785c1104(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vpca1a_error */
extern void T36f481(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_query_call_agent_validity */
extern void T71f454(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_unqualified_query_call_agent */
extern void T71f485(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_unqualified_call_agent */
extern void T71f495(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T71f495ot1(T0* a1, T0** a2);
extern T1 T71f495ot2(T0* a1, T0** a2);
extern T1 T71f495ot3(T0* a1, T0** a2);
extern T1 T71f495ot4(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_call_agent_result_dynamic_types */
extern void T71f500(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_procedure_call_agent_validity */
extern void T71f453(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_unqualified_procedure_call_agent */
extern void T71f484(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vucr0j_error */
extern void T53f480(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vucr0j */
extern T0* T785c1103(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vucr0j_error */
extern void T36f480(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_call_agent */
extern void T398f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_call_agent */
extern void T394f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_call_agent */
extern void T380f199(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_call_agent_validity */
extern void T380f254(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T380f254ot1(T0* a1, T0** a2);
extern T1 T380f254ot2(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.check_typed_call_agent_validity */
extern void T380f318(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_typed_tuple_label_call_agent_validity */
extern void T380f365(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_tuple_label_call_agent */
extern void T380f389(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_typed_query_call_agent_validity */
extern void T380f364(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_qualified_query_call_agent */
extern void T380f388(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.check_typed_procedure_call_agent_validity */
extern void T380f363(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_qualified_procedure_call_agent */
extern void T380f387(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_call_agent_validity */
extern void T380f317(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_tuple_label_call_agent_validity */
extern void T380f362(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_query_call_agent_validity */
extern void T380f361(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_qualified_procedure_call_agent_validity */
extern void T380f360(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_call_agent_validity */
extern void T380f316(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_query_call_agent_validity */
extern void T380f359(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_unqualified_query_call_agent */
extern void T380f386(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_procedure_call_agent_validity */
extern void T380f358(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_unqualified_procedure_call_agent */
extern void T380f385(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_C_GENERATOR.process_call_agent */
extern void T67f1166(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_call_agent */
extern void T67f1210(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_call_agent_body_declaration */
extern void T67f1209(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.set_name */
extern void T284f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.set_target */
extern void T284f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_indentation_assign_to_result */
extern void T67f679(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.process */
extern void T1038f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_bracket_expression */
extern void T1327f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.process */
extern void T1977f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_unfolded_tuple_actual_argument_list */
extern void T1327f114(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.process */
extern void T283f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_bracket_expression */
extern void T71f279(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_bracket_expression_validity */
extern void T71f345(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_bracket_expression */
extern void T398f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_bracket_expression */
extern void T394f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_bracket_expression */
extern void T380f197(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_bracket_expression_validity */
extern void T380f252(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_bracket_expression */
extern void T67f1165(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_bracket_expression */
extern void T67f1208(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_EXPRESSION.process */
extern void T1008f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_across_expression */
extern void T1327f77(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_across_expression */
extern void T71f296(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_across_expression_validity */
extern void T71f361(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_across_expression */
extern void T398f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_across_expression */
extern void T394f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_across_expression */
extern void T380f202(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_across_expression_validity */
extern void T380f256(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_across_expression */
extern void T67f1164(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_across_expression */
extern void T67f1207(GE_context* ac, T0* C, T0* a1);
/* detachable ET_VOID.process */
extern void T997f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_void */
extern void T1327f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_void */
extern void T71f258(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_void_validity */
extern void T71f324(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_void_constant */
extern void T71f385(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_void */
extern void T398f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_void */
extern void T394f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_void */
extern void T380f187(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_void_validity */
extern void T380f243(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_void_constant */
extern void T380f303(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_void */
extern void T67f1163(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_void */
extern void T67f1206(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NAMED_OBJECT_TEST.process */
extern void T736f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_named_object_test */
extern void T1327f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_named_object_test */
extern void T71f294(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_named_object_test */
extern void T380f201(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_named_object_test */
extern void T67f1162(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.process */
extern void T635f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_unqualified_call_expression */
extern void T1327f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_unqualified_call_expression */
extern void T71f291(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_call_expression_validity */
extern void T71f357(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vkcn2c_error */
extern void T53f454(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vkcn2c */
extern T0* T785c1077(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vkcn2c_error */
extern void T36f454(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_object_test_local_call_expression_validity */
extern void T71f425(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuao0c_error */
extern void T53f493(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuao0c */
extern T0* T785c1116(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvuao_error */
extern T1 T53f144(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvuao0c_error */
extern void T36f493(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_gvuao_error */
extern T1 T36f144(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuao0a_error */
extern void T53f492(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuao0a */
extern T0* T785c1115(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvuao0a_error */
extern void T36f492(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvuao0b_error */
extern void T53f491(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvuao0b */
extern T0* T785c1114(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvuao0b_error */
extern void T36f491(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_iteration_item_call_expression_validity */
extern void T71f424(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_local_variable_call_expression_validity */
extern void T71f423(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unqualified_formal_argument_call_expression_validity */
extern void T71f422(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_unqualified_call_expression */
extern void T398f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_unqualified_call_expression */
extern void T394f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_unqualified_call_expression */
extern void T380f200(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_call_expression_validity */
extern void T380f255(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_object_test_local_call_expression_validity */
extern void T380f322(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_iteration_item_call_expression_validity */
extern void T380f321(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_local_variable_call_expression_validity */
extern void T380f320(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.check_unqualified_formal_argument_call_expression_validity */
extern void T380f319(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_unqualified_call_expression */
extern void T67f1161(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_unqualified_call_expression */
extern void T67f1204(GE_context* ac, T0* C, T0* a1);
extern T1 T67f1204ot1(T0* a1, T0** a2);
extern T1 T67f1204ot2(T0* a1, T0** a2);
extern T1 T67f1204ot3(T0* a1, T0** a2);
extern T1 T67f1204ot4(T0* a1, T0** a2);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.is_call_agent */
extern T1 T635f9(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.is_constant_attribute */
extern T1 T306f34(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.is_constant_attribute */
extern T1 T236f38(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_boxed_attribute_item_access */
extern void T67f684(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].replace */
extern void T272f15(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].force_last */
extern void T272f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].new_capacity */
extern T6 T272f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].extendible */
extern T1 T272f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].is_empty */
extern T1 T272f11(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.process */
extern void T288f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_manifest_tuple */
extern void T1327f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_manifest_tuple */
extern void T71f282(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_manifest_tuple_validity */
extern void T71f348(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T71f348ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_manifest_tuple */
extern void T71f409(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T71f409ot1(T0* a1, T0** a2);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_manifest_tuple */
extern void T398f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_manifest_tuple */
extern void T394f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_manifest_tuple */
extern void T380f198(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_manifest_tuple_validity */
extern void T380f253(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T380f253ot1(T0* a1, T0** a2);
/* [detachable] ET_FEATURE_CHECKER.report_manifest_tuple */
extern void T380f315(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.process_manifest_tuple */
extern void T67f1110(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_manifest_tuple */
extern void T67f1113(GE_context* ac, T0* C, T0* a1);
extern T1 T67f1113ot1(T0* a1, T0** a2);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].force_last */
extern void T289f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].slots_put */
extern void T289f59(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].clashes_put */
extern void T289f58(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].slots_item */
extern T6 T289f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].hash_position */
extern T6 T289f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].resize */
extern void T289f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].clashes_resize */
extern void T289f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].special_integer_ */
extern T0* T289f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].key_storage_resize */
extern void T289f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].item_storage_resize */
extern void T289f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TUPLE_TYPE].aliased_resized_area */
extern T0* T674f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TUPLE_TYPE].aliased_resized_area */
extern T0* T673f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].key_storage_item */
extern T0* T289f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].item_storage_item */
extern T0* T289f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].clashes_item */
extern T6 T289f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].slots_resize */
extern void T289f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].new_modulus */
extern T6 T289f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].new_capacity */
extern T6 T289f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].item_storage_put */
extern void T289f56(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TUPLE_TYPE].force */
extern void T674f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TUPLE_TYPE].fill_with */
extern void T673f9(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TUPLE_TYPE].extend */
extern void T673f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].search_position */
extern void T289f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].key_equality_tester */
extern T0* T289f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].unset_found_item */
extern void T289f43(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.process */
extern void T309f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_result */
extern void T1327f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_result */
extern void T71f260(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_result_validity */
extern void T71f326(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vevi0b_error */
extern void T53f447(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vevi0b */
extern T0* T785c1070(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vevi0b_error */
extern void T36f447(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_result */
extern void T398f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_result */
extern void T394f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_result */
extern void T380f188(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_result_validity */
extern void T380f244(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_result */
extern void T380f304(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] ET_C_GENERATOR.process_result */
extern void T67f1097(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_result */
extern void T67f1102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CURRENT.process */
extern void T282f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_current */
extern void T1327f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_current */
extern void T71f253(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_current_validity */
extern void T71f319(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_current */
extern void T71f382(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vucr0f_error */
extern void T53f448(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vucr0f */
extern T0* T785c1071(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vucr0f_error */
extern void T36f448(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_current */
extern void T398f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_current */
extern void T394f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_current */
extern void T380f189(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_current_validity */
extern void T380f245(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_current */
extern void T380f305(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_current */
extern void T67f1095(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_current */
extern void T67f1101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.process */
extern void T287f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_qualified_call_expression */
extern void T1327f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_qualified_call_expression */
extern void T71f266(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_qualified_call_expression */
extern void T398f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_qualified_call_expression */
extern void T394f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_qualified_call_expression */
extern void T380f193(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_qualified_call_expression */
extern void T67f1096(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER.process */
extern void T184f119(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_identifier */
extern void T1327f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_identifier */
extern void T71f235(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.process_identifier */
extern void T398f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.process_identifier */
extern void T394f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_identifier */
extern void T380f203(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_identifier */
extern void T67f1094(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_agent_closed_operand */
extern void T67f1100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_agent_closed_operand_access */
extern void T67f709(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.formal_argument */
extern T0* T67f257(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER].put_last */
extern void T261f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER.is_agent_closed_operand */
extern T1 T184f34(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_agent_open_operand */
extern void T67f1099(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_agent_open_operand_access */
extern void T67f705(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_agent_open_target_operand */
extern void T67f1104(GE_context* ac, T0* C, T0* a1);
extern T1 T67f1104ot1(T0* a1, T0** a2);
extern T1 T67f1104ot2(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_indentation_end_newline */
extern void T67f1109(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_EXPRESSION].item */
extern T0* T271f8(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_indentation_else_newline */
extern void T67f1108(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_then_newline */
extern void T67f1106(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_indentation_if */
extern void T67f1105(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.is_agent_open_operand */
extern T1 T184f33(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_object_test_local */
extern void T67f1098(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_boxed_expression */
extern void T67f741(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_boxed_type_cast */
extern void T67f737(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_boxed_type_name */
extern void T67f668(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].force_last_new */
extern void T216f56(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].sort */
extern void T263f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_QUICK_SORTER [[attached] INTEGER_32].sort */
extern void T344f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_QUICK_SORTER [[attached] INTEGER_32].sort_with_comparator */
extern void T344f4(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_QUICK_SORTER [[attached] INTEGER_32].subsort_with_comparator */
extern void T344f5(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].swap */
extern void T263f26(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] INTEGER_32].less_than */
extern T1 T472f1(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] INTEGER_32].detachable_less_than */
extern T1 T472f2(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] INTEGER_32].attached_less_than */
extern T1 T472f3(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_C_GENERATOR.dynamic_type_id_sorter */
extern T0* T67f276(GE_context* ac, T0* C);
/* [detachable] DS_QUICK_SORTER [[attached] INTEGER_32].make */
extern T0* T344c2(GE_context* ac, T0* a1);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] INTEGER_32].make */
extern T0* T472c4(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].is_empty */
extern T1 T263f6(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.dynamic_type */
extern T0* T269f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.count */
extern T6 T1466f2(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.count */
extern T6 T269f2(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_adapted_attribute_access */
extern void T67f768(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4);
/* [detachable] ET_C_GENERATOR.print_boxed_attribute_access */
extern void T67f794(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4);
/* [detachable] ET_C_GENERATOR.argument_type_set_in_feature */
extern T0* T67f273(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].item */
extern T0* T272f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].first */
extern T0* T272f10(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.fill_call_formal_arguments */
extern void T67f721(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_reference_body */
extern void T67f1381(GE_context* ac, T0* C, T0* a1);
extern T1 T67f1381ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_real_64_body */
extern void T67f1380(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_basic_expanded_item_body */
extern void T67f1395(GE_context* ac, T0* C, T0* a1);
extern T1 T67f1395ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_real_32_body */
extern void T67f1379(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_pointer_body */
extern void T67f1378(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_natural_64_body */
extern void T67f1377(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_natural_32_body */
extern void T67f1376(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_natural_16_body */
extern void T67f1375(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_natural_8_body */
extern void T67f1374(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_integer_64_body */
extern void T67f1373(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_integer_32_body */
extern void T67f1372(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_integer_16_body */
extern void T67f1371(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_integer_8_body */
extern void T67f1370(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_character_32_body */
extern void T67f1369(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_character_8_body */
extern void T67f1368(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_put_boolean_body */
extern void T67f1367(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_special_procedure_body */
extern void T67f1319(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_special_set_count_body */
extern void T67f1366(GE_context* ac, T0* C, T0* a1);
extern T1 T67f1366ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.has_nested_reference_fields */
extern T1 T319f54(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.has_nested_reference_fields */
extern T1 T451f38(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.has_reference_fields */
extern T1 T451f52(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.has_nested_reference_fields */
extern T1 T450f38(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.has_reference_fields */
extern T1 T450f55(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.has_nested_reference_fields */
extern T1 T302f53(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.has_reference_fields */
extern T1 T302f56(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_nested_reference_fields */
extern T1 T223f51(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_reference_fields */
extern T1 T223f55(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.has_reference_fields */
extern T1 T319f58(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.result_type_set_in_feature */
extern T0* T67f244(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_special_extend_body */
extern void T67f1365(GE_context* ac, T0* C, T0* a1);
extern T1 T67f1365ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_external_builtin_real_n_ref_procedure_body */
extern void T67f1318(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_procedure_procedure_body */
extern void T67f1317(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_pointer_ref_procedure_body */
extern void T67f1316(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_memory_procedure_body */
extern void T67f1315(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_ise_runtime_procedure_body */
extern void T67f1314(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_set_pre_ecma_mapping_body */
extern void T67f1364(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_ise_exception_manager_procedure_body */
extern void T67f1313(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_integer_n_ref_procedure_body */
extern void T67f1312(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_identified_routines_procedure_body */
extern void T67f1311(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_exception_manager_procedure_body */
extern void T67f1310(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_set_is_ignored_body */
extern void T67f1363(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_raise_body */
extern void T67f1362(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_ignore_body */
extern void T67f1361(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_catch_body */
extern void T67f1360(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_com_failure_procedure_body */
extern void T67f1309(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_character_n_ref_procedure_body */
extern void T67f1308(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_boolean_ref_procedure_body */
extern void T67f1307(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_any_procedure_body */
extern void T67f1306(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_any_standard_copy_body */
extern void T67f1359(GE_context* ac, T0* C, T0* a1);
extern T1 T67f1359ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_any_standard_copy_body_with_no_special */
extern void T67f1394(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T67f1394ot1(T0* a1, T0** a2);
extern T1 T67f1394ot2(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_any_standard_copy_custom_tuple_items */
extern void T67f1409(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.print_builtin_any_standard_copy_custom_attributes */
extern void T67f1407(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_assign_temp_variable_to_onces_attribute */
extern void T67f1404(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.has_once_per_object_routines */
extern T1 T451f13(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.has_once_per_object_routines */
extern T1 T450f13(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.has_once_per_object_routines */
extern T1 T302f16(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_once_per_object_routines */
extern T1 T223f15(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.has_once_per_object_routines */
extern T1 T319f16(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_assign_temp_variable_to_flags_attribute */
extern void T67f1403(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_attribute_flags_access */
extern void T67f1410(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_attribute_flags_name */
extern void T67f689(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_assign_onces_attribute_to_temp_variable */
extern void T67f1401(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_once_per_object_data_type_name */
extern void T67f687(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_assign_flags_attribute_to_temp_variable */
extern void T67f1400(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_builtin_any_standard_copy_body_with_special */
extern void T67f1391(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.print_builtin_any_standard_copy_custom_special_items */
extern void T67f1408(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_C_GENERATOR.free_actual_argument_list */
extern void T67f1399(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.wipe_out */
extern void T283f16(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.fixed_array */
extern T0* T283f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACTUAL_ARGUMENT_LIST].force_last */
extern void T266f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACTUAL_ARGUMENT_LIST].force */
extern void T641f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ACTUAL_ARGUMENT_LIST].fill_with */
extern void T642f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ACTUAL_ARGUMENT_LIST].extend */
extern void T642f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACTUAL_ARGUMENT_LIST].resize */
extern void T266f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACTUAL_ARGUMENT_LIST].aliased_resized_area */
extern T0* T641f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ACTUAL_ARGUMENT_LIST].aliased_resized_area */
extern T0* T642f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACTUAL_ARGUMENT_LIST].new_capacity */
extern T6 T266f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACTUAL_ARGUMENT_LIST].extendible */
extern T1 T266f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.force_first */
extern void T283f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.resize */
extern void T283f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.new_capacity */
extern T6 T283f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.capacity */
extern T6 T283f8(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.new_actual_argument_list */
extern T0* T67f550(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACTUAL_ARGUMENT_LIST].remove_last */
extern void T266f15(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_ACTUAL_ARGUMENT_LIST].keep_head */
extern void T642f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ACTUAL_ARGUMENT_LIST].set_count */
extern void T642f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACTUAL_ARGUMENT_LIST].move_last_cursors_after */
extern void T266f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ACTUAL_ARGUMENT_LIST].set_next_cursor */
extern void T643f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ACTUAL_ARGUMENT_LIST].set_after */
extern void T643f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACTUAL_ARGUMENT_LIST].last */
extern T0* T266f10(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.make_with_capacity */
extern void T283f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.make_with_capacity */
extern T0* T283c15(GE_context* ac, T6 a1);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.make_with_capacity */
extern void T283f15p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.tokens */
extern T0* T283f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACTUAL_ARGUMENT_LIST].is_empty */
extern T1 T266f9(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_greater_than_or_equal */
extern void T67f1398(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_assign_temp_variable_to_special_capacity_attribute */
extern void T67f1406(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4);
/* [detachable] ET_C_GENERATOR.print_assign_temp_variable_to_special_count_attribute */
extern void T67f1405(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4);
/* [detachable] ET_C_GENERATOR.print_assign_special_capacity_attribute_to_temp_variable */
extern void T67f1402(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4);
/* [detachable] ET_C_GENERATOR.print_greater_than */
extern void T67f1390(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_assign_special_count_attribute_to_temp_variable */
extern void T67f1397(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4);
/* [detachable] ET_C_GENERATOR.print_builtin_any_copy_body */
extern void T67f1358(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_function_body */
extern void T67f1281(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_type_function_body */
extern void T67f1305(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_type_generic_parameter_type_body */
extern void T67f1357(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.shallow_base_type */
extern T0* T1501f54(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.shallow_base_type_with_type_mark */
extern T0* T1501f56(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1501f56ot1(T0* a1, T0** a2);
extern T1 T1501f56ot2(T0* a1, T0** a2);
extern T1 T1501f56ot3(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.shallow_base_type_with_type_mark */
extern T0* T1526f45(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1526f45ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.shallow_base_type_with_type_mark */
extern T0* T1525f48(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1525f48ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.shallow_base_type_with_type_mark */
extern T0* T1524f46(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.shallow_base_type_with_type_mark */
extern T0* T242f45(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T242f45ot1(T0* a1, T0** a2);
extern T1 T242f45ot2(T0* a1, T0** a2);
extern T1 T242f45ot3(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.shallow_base_type_with_type_mark */
extern T0* T449f42(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER.shallow_base_type_with_type_mark */
extern T0* T891f55(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T891f55ot1(T0* a1, T0** a2);
extern T1 T891f55ot2(T0* a1, T0** a2);
extern T1 T891f55ot3(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.shallow_base_type */
extern T0* T891f52(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.shallow_base_type */
extern T0* T1526f42(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.shallow_base_type */
extern T0* T1525f45(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.shallow_base_type */
extern T0* T1524f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.shallow_base_type */
extern T0* T242f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.shallow_base_type */
extern T0* T449f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.shallow_base_type */
extern T0* T74f199(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.shallow_base_type */
extern T0* T254f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.shallow_base_type */
extern T0* T72f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_tuple_function_body */
extern void T67f1304(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_reference_item_body */
extern void T67f1356(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_C_GENERATOR.print_builtin_tuple_reference_item_body */
extern T0* T67f1356ac1(GE_context* ac, T0* a1, T6 a2, T0* a3, T0* a4, T1 a5);
extern T1 T67f1356ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_real_64_item_body */
extern void T67f1355(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_basic_expanded_item_body */
extern void T67f1393(GE_context* ac, T0* C, T0* a1);
extern T1 T67f1393ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_real_32_item_body */
extern void T67f1354(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_pointer_item_body */
extern void T67f1353(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_natural_64_item_body */
extern void T67f1352(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_natural_32_item_body */
extern void T67f1351(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_natural_16_item_body */
extern void T67f1350(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_natural_8_item_body */
extern void T67f1349(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_item_code_body */
extern void T67f1348(GE_context* ac, T0* C, T0* a1);
extern T1 T67f1348ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_integer_64_item_body */
extern void T67f1347(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_integer_32_item_body */
extern void T67f1346(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_integer_16_item_body */
extern void T67f1345(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_integer_8_item_body */
extern void T67f1344(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_character_32_item_body */
extern void T67f1343(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_character_8_item_body */
extern void T67f1342(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_tuple_boolean_item_body */
extern void T67f1341(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_special_function_body */
extern void T67f1303(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_special_aliased_resized_area_body */
extern void T67f1340(GE_context* ac, T0* C, T0* a1);
extern T1 T67f1340ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.free_result_expression */
extern void T67f1392(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_RESULT].force_last */
extern void T267f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RESULT].force */
extern void T644f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_RESULT].fill_with */
extern void T645f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_RESULT].extend */
extern void T645f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_RESULT].resize */
extern void T267f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RESULT].aliased_resized_area */
extern T0* T644f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_RESULT].aliased_resized_area */
extern T0* T645f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_RESULT].new_capacity */
extern T6 T267f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_RESULT].extendible */
extern T1 T267f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.new_result_expression */
extern T0* T67f549(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_RESULT].remove_last */
extern void T267f15(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_RESULT].keep_head */
extern void T645f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_RESULT].set_count */
extern void T645f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_RESULT].move_last_cursors_after */
extern void T267f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_RESULT].set_next_cursor */
extern void T646f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_RESULT].set_after */
extern void T646f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_RESULT].last */
extern T0* T267f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_RESULT].is_empty */
extern T1 T267f9(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_external_builtin_real_n_ref_function_body */
extern void T67f1302(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_real_n_function_body */
extern void T67f1301(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_real_n_out_body */
extern void T67f1339(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_pointer_ref_function_body */
extern void T67f1300(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_pointer_function_body */
extern void T67f1299(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_pointer_out_body */
extern void T67f1338(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_platform_function_body */
extern void T67f1298(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_memory_function_body */
extern void T67f1297(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_ise_runtime_function_body */
extern void T67f1296(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_pre_ecma_mapping_status_body */
extern void T67f1337(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_ise_runtime_in_assertion_body */
extern void T67f1336(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_integer_n_ref_function_body */
extern void T67f1295(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_integer_n_function_body */
extern void T67f1294(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_identified_routines_function_body */
extern void T67f1293(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_function_function_body */
extern void T67f1292(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_exception_manager_factory_function_body */
extern void T67f1291(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_factory_exception_manager_body */
extern void T67f1335(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_exception_manager_function_body */
extern void T67f1290(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_type_of_code_body */
extern void T67f1334(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_last_exception_body */
extern void T67f1333(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_is_raisable_body */
extern void T67f1332(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_is_ignored_body */
extern void T67f1331(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_is_ignorable_body */
extern void T67f1330(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_is_caught_body */
extern void T67f1329(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_exception_manager_exception_from_code_body */
extern void T67f1328(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_com_failure_function_body */
extern void T67f1289(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_character_n_ref_function_body */
extern void T67f1288(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_character_n_function_body */
extern void T67f1287(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_boolean_ref_function_body */
extern void T67f1286(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_boolean_function_body */
extern void T67f1285(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_arguments_32_function_body */
extern void T67f1284(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_builtin_any_function_body */
extern void T67f1283(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_any_twin_body */
extern void T67f1327(GE_context* ac, T0* C, T0* a1);
extern T1 T67f1327ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_any_twin_body_with_reference */
extern void T67f1389(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_any_twin_body_with_expanded */
extern void T67f1388(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_any_twin_body_with_special */
extern void T67f1387(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_builtin_any_tagged_out_body */
extern void T67f1326(GE_context* ac, T0* C, T0* a1);
extern T1 T67f1326ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_any_standard_twin_body */
extern void T67f1325(GE_context* ac, T0* C, T0* a1);
extern T1 T67f1325ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_any_standard_twin_body_with_no_special */
extern void T67f1386(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_builtin_any_standard_twin_body_with_special */
extern void T67f1385(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_any_standard_is_equal_body */
extern void T67f1324(GE_context* ac, T0* C, T0* a1);
extern T1 T67f1324ot1(T0* a1, T0** a2);
extern T1 T67f1324ot2(T0* a1, T0** a2);
extern T1 T67f1324ot3(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_builtin_any_standard_is_equal_tuple_items */
extern void T67f1384(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.print_builtin_any_standard_is_equal_field */
extern void T67f1396(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.print_builtin_any_standard_is_equal_special_items */
extern void T67f1383(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.print_builtin_any_standard_is_equal_attributes */
extern void T67f1382(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_builtin_any_is_equal_body */
extern void T67f1323(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_any_is_deep_equal_body */
extern void T67f1322(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_builtin_any_conforms_to_body */
extern void T67f1321(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.is_function */
extern T1 T1516f69(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_external_c_inline_body */
extern void T67f1270(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_CHARACTER_ROUTINES.as_lower */
extern T2 T1954f1(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_C_GENERATOR.character_ */
extern T0* T67f544(GE_context* ac, T0* C);
/* [detachable] KL_CHARACTER_ROUTINES.default_create */
extern T0* T1954c2(GE_context* ac);
/* [detachable] UC_UTF8_STRING.has_substring */
extern T1 T775f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_STRING.has_substring */
extern T1 T429f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.has_substring */
extern T1 T17f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.same_characters */
extern T1 T17f21(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] ET_C_GENERATOR.print_external_c_includes */
extern void T67f1269(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.include_header_filename */
extern void T67f1280(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] UC_UTF8_STRING.same_string */
extern T1 T775f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_STRING.same_string */
extern T1 T429f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.same_string */
extern T1 T17f12(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].force_last */
extern void T70f44(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].slots_put */
extern void T70f52(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_HASH_SET [[attached] STRING_8].clashes_put */
extern void T70f51(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_HASH_SET [[attached] STRING_8].slots_item */
extern T6 T70f22(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].hash_position */
extern T6 T70f36(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].resize */
extern void T70f50(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].clashes_resize */
extern void T70f69(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].special_integer_ */
extern T0* T70f35(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].key_storage_resize */
extern void T70f68(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].item_storage_resize */
extern void T70f67(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].key_storage_item */
extern T0* T70f29(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].item_storage_item */
extern T0* T70f26(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].clashes_item */
extern T6 T70f21(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].slots_resize */
extern void T70f66(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].new_modulus */
extern T6 T70f24(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].new_capacity */
extern T6 T70f33(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].item_storage_put */
extern void T70f49(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable DS_HASH_SET [[attached] STRING_8].search_position */
extern void T70f47(GE_context* ac, T0* C, T0* a1);
/* detachable KL_EQUALITY_TESTER [[attached] STRING_8].test */
extern T1 T154f1(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T154f1oe1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] UC_UTF8_STRING.is_equal */
extern T1 T775f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_STRING.is_equal */
extern T1 T429f24(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].key_equality_tester */
extern T0* T70f27(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].unset_found_item */
extern void T70f48(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].remove */
extern void T70f43(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].remove_position */
extern void T70f59(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].key_storage_put */
extern void T70f76(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable DS_HASH_SET [[attached] STRING_8].key_storage_keep_head */
extern void T70f75(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].item_storage_keep_head */
extern void T70f74(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].move_cursors_forth */
extern void T70f73(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8].set_position */
extern void T376f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8].set_next_cursor */
extern void T376f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8].set_after */
extern void T376f10(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].internal_set_key_equality_tester */
extern void T70f72(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].has */
extern T1 T70f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_cpp_includes */
extern void T67f1268(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.include_cpp_header_filename */
extern void T67f1279(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.captured_substring_count */
extern T6 T81f308(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_malloc_current */
extern void T67f1140(GE_context* ac, T0* C, T0* a1);
extern T1 T67f1140ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.current_universe */
extern T0* T67f504(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_feature_info_message_call */
extern void T67f1267(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_show_console_call */
extern void T67f673(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.is_empty */
extern T1 T1466f7(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.is_empty */
extern T1 T269f5(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.is_empty */
extern T1 T475f10(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.is_empty */
extern T1 T451f33(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.is_empty */
extern T1 T450f33(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.is_empty */
extern T1 T302f32(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.is_empty */
extern T1 T319f32(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_empty */
extern T1 T223f26(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.is_empty */
extern T1 T323f15(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_feature_trace_message_call */
extern void T67f1139(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_C_GENERATOR.print_default_entity_value */
extern void T67f727(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_result_name */
extern void T67f641(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_escaped_string */
extern void T67f667(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_creation_procedure_name */
extern void T67f757(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.print_static_routine_name */
extern void T67f740(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T67f740ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_inline_routine_name */
extern void T67f1266(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T67f1266ot1(T0* a1, T0** a2);
/* detachable ET_ONCE_PROCEDURE.process */
extern void T1515f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_once_procedure */
extern void T1327f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_once_procedure */
extern void T1326f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_once_procedure */
extern void T380f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_once_procedure_validity */
extern void T380f109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vucr0b_error */
extern void T53f368(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vucr0b */
extern T0* T785c998(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vucr0b_error */
extern void T36f368(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.check_locals_validity */
extern void T380f119(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_CHECKER.report_local_supplier */
extern void T380f134(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SUPPLIER_BUILDER.report_local_supplier */
extern void T1341f11(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_local_variable_declaration */
extern void T380f133(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vrle1a_error */
extern void T53f249(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vrle1a */
extern T0* T785c899(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vrle1_error */
extern T1 T53f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vrle1a_error */
extern void T36f249(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vrle1_error */
extern T1 T36f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vrle2a_error */
extern void T53f248(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vrle2a */
extern T0* T785c898(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vrle2_error */
extern T1 T53f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vrle2a_error */
extern void T36f248(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vrle2_error */
extern T1 T36f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vreg0b_error */
extern void T53f247(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vreg0b */
extern T0* T785c897(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vreg0b_error */
extern void T36f247(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_once_procedure */
extern void T71f153(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_once_procedure_validity */
extern void T71f158(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_locals_validity */
extern void T71f140(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_local_supplier */
extern void T71f148(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_local_variable_declaration */
extern void T71f147(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_once_procedure */
extern void T67f1184(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_once_procedure */
extern void T67f1227(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_internal_procedure */
extern void T67f1116(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_internal_feature */
extern void T67f1117(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] ET_C_GENERATOR.print_assign_called_to_once_status */
extern void T67f1137(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_once_status_is_completed */
extern void T67f1136(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_return_statement */
extern void T67f1135(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_assign_result_to_thread_safe_once_value */
extern void T67f1132(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_thread_safe_once_per_process_value */
extern void T67f1149(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_assign_once_value_to_result */
extern void T67f1131(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_once_mutex_lock */
extern void T67f1129(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_once_mutex_try_lock */
extern void T67f1128(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_return_thread_safe_once_value */
extern void T67f1127(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_raise_thread_safe_once_exception_if_any */
extern void T67f1126(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_return_once_value */
extern void T67f1124(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_raise_once_exception_if_any */
extern void T67f1123(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_C_GENERATOR.print_once_per_object_global_mutex_unlock */
extern void T67f1121(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_once_per_object_global_mutex_lock */
extern void T67f1120(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].value */
extern T6 T290f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].position_of_key */
extern T6 T290f35(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.process */
extern void T1513f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_deferred_function */
extern void T1327f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_deferred_function */
extern void T1326f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_deferred_function */
extern void T380f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_deferred_function_validity */
extern void T380f108(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.report_result_supplier */
extern void T380f124(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_SUPPLIER_BUILDER.report_result_supplier */
extern void T1341f9(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_CHECKER.report_result_declaration */
extern void T380f123(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_query_type_validity */
extern void T380f122(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_deferred_function */
extern void T71f152(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_deferred_function_validity */
extern void T71f157(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_result_supplier */
extern void T71f162(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_result_declaration */
extern void T71f161(GE_context* ac, T0* C, T0* a1);
extern T1 T71f161ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.propagate_extended_attribute_result_dynamic_types */
extern void T71f165(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_query_type_validity */
extern void T71f160(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.process_deferred_function */
extern void T67f1183(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_deferred_function */
extern void T67f1226(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.process */
extern void T1512f86(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_once_function */
extern void T1327f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_once_function */
extern void T1326f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_once_function */
extern void T380f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_once_function_validity */
extern void T380f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vevi0c_error */
extern void T53f367(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vevi0c */
extern T0* T785c997(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vevi0c_error */
extern void T36f367(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_once_function */
extern void T71f151(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_once_function_validity */
extern void T71f156(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_once_function */
extern void T67f1182(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_once_function */
extern void T67f1225(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_internal_function */
extern void T67f1259(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.process */
extern void T1511f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_do_function */
extern void T1327f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_do_function */
extern void T1326f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_do_function */
extern void T380f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_do_function_validity */
extern void T380f106(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_do_function */
extern void T71f150(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_do_function_validity */
extern void T71f155(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_do_function */
extern void T67f1181(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_do_function */
extern void T67f1224(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.process */
extern void T469f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_do_procedure */
extern void T1327f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_do_procedure */
extern void T1326f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_do_procedure */
extern void T380f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_do_procedure_validity */
extern void T380f104(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_do_procedure */
extern void T71f133(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_do_procedure_validity */
extern void T71f135(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_do_procedure */
extern void T67f1111(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_do_procedure */
extern void T67f1114(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.process */
extern void T371f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_unique_attribute */
extern void T1327f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_unique_attribute */
extern void T1326f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_unique_attribute */
extern void T380f103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_unique_attribute_validity */
extern void T380f115(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vqui0a_error */
extern void T53f385(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vqui0a */
extern T0* T785c1015(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vqui_error */
extern T1 T53f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vqui0a_error */
extern void T36f385(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vqui_error */
extern T1 T36f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_unique_attribute */
extern void T71f177(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_unique_attribute_validity */
extern void T71f182(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_unique_attribute */
extern void T67f1420(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_unique_attribute */
extern void T67f1433(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_attribute_wrapper */
extern void T67f1440(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.free_unqualified_call_expression */
extern void T67f1442(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].force_last */
extern void T264f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].force */
extern void T634f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].fill_with */
extern void T636f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].extend */
extern void T636f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].resize */
extern void T264f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].aliased_resized_area */
extern T0* T634f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].aliased_resized_area */
extern T0* T636f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].new_capacity */
extern T6 T264f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].extendible */
extern T1 T264f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.new_unqualified_call_expression */
extern T0* T67f551(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.set_name */
extern void T635f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].remove_last */
extern void T264f15(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].keep_head */
extern void T636f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].set_count */
extern void T636f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].move_last_cursors_after */
extern void T264f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].set_next_cursor */
extern void T637f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].set_after */
extern void T637f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].last */
extern T0* T264f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].is_empty */
extern T1 T264f9(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.process */
extern void T369f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_constant_attribute */
extern void T1327f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_constant_attribute */
extern void T1326f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_constant_attribute */
extern void T380f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_constant_attribute_validity */
extern void T380f114(GE_context* ac, T0* C, T0* a1);
extern T1 T380f114ot1(T0* a1, T0** a2);
extern T1 T380f114ot2(T0* a1, T0** a2);
extern T1 T380f114ot3(T0* a1, T0** a2);
extern T1 T380f114ot4(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vqmc5a_error */
extern void T53f384(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vqmc5a */
extern T0* T785c1014(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vqmc5_error */
extern T1 T53f99(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vqmc5a_error */
extern void T36f384(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vqmc5_error */
extern T1 T36f99(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.system_string_type */
extern T0* T187f89(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.system_string_type */
extern T0* T58f118(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.system_string_type */
extern T0* T40f94(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vqmc5b_error */
extern void T53f383(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vqmc5b */
extern T0* T785c1013(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vqmc5b_error */
extern void T36f383(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vqmc4a_error */
extern void T53f382(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vqmc4a */
extern T0* T785c1012(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vqmc4_error */
extern T1 T53f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vqmc4a_error */
extern void T36f382(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vqmc4_error */
extern T1 T36f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vqmc3a_error */
extern void T53f381(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vqmc3a */
extern T0* T785c1011(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vqmc3_error */
extern T1 T53f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vqmc3a_error */
extern void T36f381(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vqmc3_error */
extern T1 T36f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vqmc3b_error */
extern void T53f380(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vqmc3b */
extern T0* T785c1010(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vqmc3b_error */
extern void T36f380(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vqmc2a_error */
extern void T53f379(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vqmc2a */
extern T0* T785c1009(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vqmc2_error */
extern T1 T53f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vqmc2a_error */
extern void T36f379(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vqmc2_error */
extern T1 T36f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vqmc2b_error */
extern void T53f378(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vqmc2b */
extern T0* T785c1008(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T785f1008ot1(T0* a1, T0** a2);
extern T1 T785f1008ot2(T0* a1, T0** a2);
/* [detachable] ET_ERROR_HANDLER.report_vqmc2b_error */
extern void T36f378(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vqmc1a_error */
extern void T53f377(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vqmc1a */
extern T0* T785c1007(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vqmc1_error */
extern T1 T53f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vqmc1a_error */
extern void T36f377(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vqmc1_error */
extern T1 T36f95(GE_context* ac, T0* C, T0* a1);
/* detachable ET_BINARY_INTEGER_CONSTANT.is_boolean_constant */
extern T1 T1489f36(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.is_boolean_constant */
extern T1 T1488f36(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.is_boolean_constant */
extern T1 T1487f36(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.is_boolean_constant */
extern T1 T1486f36(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.is_boolean_constant */
extern T1 T372f15(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.is_boolean_constant */
extern T1 T1485f26(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.is_boolean_constant */
extern T1 T1484f9(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.is_boolean_constant */
extern T1 T1482f22(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.is_boolean_constant */
extern T1 T1496f18(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.is_boolean_constant */
extern T1 T1495f18(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.is_boolean_constant */
extern T1 T1494f19(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.is_boolean_constant */
extern T1 T1493f19(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.is_boolean_constant */
extern T1 T1491f20(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.is_boolean_constant */
extern T1 T1490f20(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.is_boolean_constant */
extern T1 T1483f19(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_constant_attribute */
extern void T71f176(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_constant_attribute_validity */
extern void T71f181(GE_context* ac, T0* C, T0* a1);
extern T1 T71f181ot1(T0* a1, T0** a2);
extern T1 T71f181ot2(T0* a1, T0** a2);
extern T1 T71f181ot3(T0* a1, T0** a2);
extern T1 T71f181ot4(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.process_constant_attribute */
extern void T67f1419(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_constant_attribute */
extern void T67f1432(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.process */
extern void T360f80(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_attribute */
extern void T1327f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_attribute */
extern void T1326f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_attribute */
extern void T380f101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_attribute_validity */
extern void T380f113(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vucr0a_error */
extern void T53f376(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vucr0a */
extern T0* T785c1006(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vucr0a_error */
extern void T36f376(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_attribute */
extern void T71f175(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_attribute_validity */
extern void T71f180(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_attribute */
extern void T67f1418(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_attribute */
extern void T67f1431(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.process */
extern void T359f85(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_extended_attribute */
extern void T1327f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_extended_attribute */
extern void T1326f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_extended_attribute */
extern void T380f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_extended_attribute_validity */
extern void T380f112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vwab0a_error */
extern void T53f375(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vwab0a */
extern T0* T785c1005(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EXTENDED_ATTRIBUTE.position */
extern T0* T359f79(GE_context* ac, T0* C);
/* detachable ET_ALIASED_FEATURE_NAME.position */
extern T0* T1517f4(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vwab_error */
extern T1 T53f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vwab0a_error */
extern void T36f375(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vwab_error */
extern T1 T36f94(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.is_type_self_initializing */
extern T1 T1524f48(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.is_type_self_initializing_with_type_mark */
extern T1 T1524f58(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.is_type_self_initializing */
extern T1 T254f77(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.is_type_self_initializing_with_type_mark */
extern T1 T254f95(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_type_self_initializing */
extern T1 T242f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_type_self_initializing_with_type_mark */
extern T1 T242f61(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS_TYPE.is_type_self_initializing */
extern T1 T72f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.is_type_self_initializing_with_type_mark */
extern T1 T72f99(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_type_self_initializing */
extern T1 T1526f47(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_type_self_initializing_with_type_mark */
extern T1 T1526f59(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_type_self_initializing */
extern T1 T1525f50(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_type_self_initializing_with_type_mark */
extern T1 T1525f61(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIKE_CURRENT.is_type_self_initializing */
extern T1 T449f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.is_type_self_initializing_with_type_mark */
extern T1 T449f49(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vevi0e_error */
extern void T53f374(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vevi0e */
extern T0* T785c1004(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vevi0e_error */
extern void T36f374(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_extended_attribute */
extern void T71f174(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_extended_attribute_validity */
extern void T71f179(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_extended_attribute */
extern void T67f1417(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_extended_attribute */
extern void T67f1430(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.process */
extern void T253f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_external_function */
extern void T1327f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_external_function */
extern void T1326f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_external_function */
extern void T380f99(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_external_function_validity */
extern void T380f111(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_external_function */
extern void T71f173(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_external_function_validity */
extern void T71f178(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.mark_string_type_alive */
extern void T71f184(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.result_type_set */
extern T0* T71f78(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.same_base_type_with_type_marks */
extern T1 T1524f63(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS.same_base_type_with_type_marks */
extern T1 T74f235(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.same_base_class_type_with_type_marks */
extern T1 T72f118(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.same_base_type_with_type_marks */
extern T1 T254f90(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_CLASS_TYPE.same_base_tuple_type_with_type_marks */
extern T1 T72f119(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.same_base_type_with_type_marks */
extern T1 T242f57(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T242f57ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS_TYPE.same_base_formal_parameter_type_with_type_marks */
extern T1 T72f120(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.same_base_type_with_type_marks */
extern T1 T1501f47(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1501f47ot1(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.same_base_type_with_type_marks */
extern T1 T891f48(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T891f48ot1(T0* a1, T0** a2);
/* [detachable] ET_CLASS_TYPE.same_base_type_with_type_marks */
extern T1 T72f98(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_QUALIFIED_LIKE_TYPE.same_base_type_with_type_marks */
extern T1 T1526f62(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1526f62ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.same_base_type_with_type_marks */
extern T1 T1525f65(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T1525f65ot1(T0* a1);
/* [detachable] ET_LIKE_CURRENT.same_base_type_with_type_marks */
extern T1 T449f58(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NESTED_TYPE_CONTEXT.same_base_type_with_type_marks */
extern T1 T241f48(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T241f48ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.same_base_type_with_type_marks */
extern T1 T1474f31(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TUPLE_TYPE.context_same_base_type_with_type_marks */
extern T1 T254f102(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS_TYPE.context_same_base_type_with_type_marks */
extern T1 T72f114(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CLASS.context_same_base_type_with_type_marks */
extern T1 T74f243(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_external_builtin_function_validity */
extern void T71f183(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_type_generic_parameter_type */
extern void T71f215(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_type_default */
extern void T71f214(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_tuple_reference_item */
extern void T71f213(GE_context* ac, T0* C, T0* a1);
extern T1 T71f213ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_special_item */
extern void T71f212(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_ise_runtime_type_conforms_to */
extern void T71f211(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.set_ise_runtime_type_conforms_to_feature */
extern void T63f110(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_ise_runtime_storable_version_of_type */
extern void T71f210(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_ise_runtime_reference_field_at_offset */
extern void T71f209(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_ise_runtime_reference_field_at */
extern void T71f208(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_ise_runtime_reference_field */
extern void T71f207(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_ise_runtime_new_type_instance_of */
extern void T71f206(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_ise_runtime_new_tuple_instance_of */
extern void T71f205(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_ise_runtime_new_special_of_reference_instance_of */
extern void T71f204(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_ise_runtime_new_instance_of */
extern void T71f203(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_identified_routines_eif_object_id */
extern void T71f202(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_identified_routines_eif_is_object_id_of_current */
extern void T71f201(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_identified_routines_eif_id_object */
extern void T71f200(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_identified_routines_eif_current_object_id */
extern void T71f199(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_function_item */
extern void T71f198(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_factory_exception_manager */
extern void T71f197(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_type_of_code */
extern void T71f196(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_last_exception */
extern void T71f195(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_is_raisable */
extern void T71f194(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_is_ignored */
extern void T71f193(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_is_ignorable */
extern void T71f192(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_is_caught */
extern void T71f191(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_exception_manager_exception_from_code */
extern void T71f190(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_function */
extern void T71f189(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_builtin_any_twin */
extern void T71f188(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.is_builtin_unknown */
extern T1 T253f79(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_external_function_validity */
extern void T71f178p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_external_function */
extern void T67f1411(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_external_function */
extern void T67f1428(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.process */
extern void T476f80(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.process_deferred_procedure */
extern void T1327f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.process_deferred_procedure */
extern void T1326f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.process_deferred_procedure */
extern void T380f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.check_deferred_procedure_validity */
extern void T380f105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.process_deferred_procedure */
extern void T71f134(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_deferred_procedure_validity */
extern void T71f136(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.process_deferred_procedure */
extern void T67f1112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_deferred_procedure */
extern void T67f1115(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd3b_error */
extern void T53f312(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd3b */
extern T0* T785c945(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_POSTCONDITIONS.position */
extern T0* T1244f14(GE_context* ac, T0* C);
/* detachable ET_POSTCONDITIONS.first */
extern T0* T1244f16(GE_context* ac, T0* C);
/* detachable ET_POSTCONDITIONS.is_empty */
extern T1 T1244f15(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdrd3_error */
extern T1 T53f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdrd3b_error */
extern void T36f312(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdrd3_error */
extern T1 T36f60(GE_context* ac, T0* C, T0* a1);
/* detachable ET_POSTCONDITIONS.is_ensure_then */
extern T1 T1244f13(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.postconditions */
extern T0* T1516f13(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.postconditions */
extern T0* T1515f12(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.postconditions */
extern T0* T1513f11(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.postconditions */
extern T0* T1512f11(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.postconditions */
extern T0* T1511f11(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.postconditions */
extern T0* T469f17(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.postconditions */
extern T0* T359f10(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.postconditions */
extern T0* T253f13(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.postconditions */
extern T0* T476f17(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd3a_error */
extern void T53f311(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd3a */
extern T0* T785c944(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_PRECONDITIONS.position */
extern T0* T406f13(GE_context* ac, T0* C);
/* detachable ET_PRECONDITIONS.first */
extern T0* T406f15(GE_context* ac, T0* C);
/* detachable ET_PRECONDITIONS.is_empty */
extern T1 T406f14(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_vdrd3a_error */
extern void T36f311(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_PRECONDITIONS.is_require_else */
extern T1 T406f12(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.preconditions */
extern T0* T1516f8(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.preconditions */
extern T0* T1515f9(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.preconditions */
extern T0* T1513f8(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.preconditions */
extern T0* T1512f8(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.preconditions */
extern T0* T1511f8(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.preconditions */
extern T0* T469f14(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.preconditions */
extern T0* T359f8(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.preconditions */
extern T0* T253f8(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.preconditions */
extern T0* T476f14(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.redeclared_clients */
extern T0* T459f28(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FEATURE_LIST.item */
extern T0* T408f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_FLATTENER.check_redeclaration_validity */
extern void T459f55(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd6a_error */
extern void T53f327(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd6a */
extern T0* T785c961(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdrd6a_error */
extern void T36f327(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd4b_error */
extern void T53f326(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd4b */
extern T0* T785c960(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdrd4_error */
extern T1 T53f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdrd4b_error */
extern void T36f326(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdrd4_error */
extern T1 T36f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd4c_error */
extern void T53f325(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd4c */
extern T0* T785c959(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdrd4c_error */
extern void T36f325(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vmfn0b_error */
extern void T53f324(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.set_ge_reported */
extern void T785f958(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_VALIDITY_ERROR.make_vmfn0b */
extern T0* T785c957(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vmfn0b_error */
extern void T36f324(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd4a_error */
extern void T53f323(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd4a */
extern T0* T785c956(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdrd4a_error */
extern void T36f323(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrs4b_error */
extern void T53f322(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrs4b */
extern T0* T785c955(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdrs4b_error */
extern void T36f322(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FEATURE_FLATTENER.ise_19_01_0 */
extern T0* T459f30(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd5a_error */
extern void T53f321(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd5a */
extern T0* T785c954(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdrd5_error */
extern T1 T53f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdrd5a_error */
extern void T36f321(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdrd5_error */
extern T1 T36f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd7c_error */
extern void T53f320(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd7c */
extern T0* T785c953(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.first_alias_name */
extern T0* T785f265(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdrd7_error */
extern T1 T53f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdrd7c_error */
extern void T36f320(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdrd7_error */
extern T1 T36f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd7a_error */
extern void T53f319(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd7a */
extern T0* T785c952(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdrd7a_error */
extern void T36f319(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrd7b_error */
extern void T53f318(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrd7b */
extern T0* T785c951(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vdrd7b_error */
extern void T36f318(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_EXTERNAL_PROCEDURE.same_version */
extern T1 T1516f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.same_version */
extern T1 T476f69(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.same_version */
extern T1 T1515f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.same_version */
extern T1 T469f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.same_version */
extern T1 T253f77(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.same_version */
extern T1 T1513f70(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.same_version */
extern T1 T1512f74(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.same_version */
extern T1 T1511f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.same_version */
extern T1 T371f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.same_version */
extern T1 T369f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.same_version */
extern T1 T359f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.same_version */
extern T1 T360f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REDECLARED_FEATURE.is_replicated */
extern T1 T1569f19(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.implementation_feature */
extern T0* T1516f4(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.implementation_feature */
extern T0* T1515f4(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.implementation_feature */
extern T0* T1513f4(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.implementation_feature */
extern T0* T1512f4(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.implementation_feature */
extern T0* T1511f4(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.implementation_feature */
extern T0* T469f3(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.implementation_feature */
extern T0* T371f6(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.implementation_feature */
extern T0* T369f5(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.implementation_feature */
extern T0* T360f4(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.implementation_feature */
extern T0* T359f4(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.implementation_feature */
extern T0* T253f4(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.implementation_feature */
extern T0* T476f1(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_dotnet */
extern T1 T1516f77(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_dotnet */
extern T1 T1515f72(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_dotnet */
extern T1 T1513f69(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_dotnet */
extern T1 T1512f73(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_dotnet */
extern T1 T1511f71(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_dotnet */
extern T1 T469f70(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_dotnet */
extern T1 T371f69(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_dotnet */
extern T1 T369f68(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_dotnet */
extern T1 T360f65(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_dotnet */
extern T1 T359f71(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_dotnet */
extern T1 T253f76(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_dotnet */
extern T1 T476f68(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.check_builtin_feature_validity */
extern void T459f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.check_feature_validity */
extern void T1326f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vmss3a_error */
extern void T53f297(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vmss3a */
extern T0* T785c930(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vmss3_error */
extern T1 T53f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vmss3a_error */
extern void T36f297(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vmss3_error */
extern T1 T36f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REDECLARED_FEATURE.selected_feature */
extern T0* T1569f17(GE_context* ac, T0* C);
/* detachable ET_PARENT_FEATURE.has_select */
extern T1 T1344f22(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.selected_feature */
extern T0* T1345f18(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.flatten_immediate_feature */
extern void T459f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INHERITED_FEATURE.immediate_feature */
extern T0* T1345f11(GE_context* ac, T0* C);
/* [detachable] ET_REDECLARED_FEATURE.immediate_feature */
extern T0* T1569f10(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.flattened_immediate_feature */
extern T0* T1516f60(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.flattened_immediate_feature */
extern T0* T253f39(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.flattened_immediate_feature */
extern T0* T476f51(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.flattened_immediate_feature */
extern T0* T1515f55(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.flattened_immediate_feature */
extern T0* T469f53(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.flattened_immediate_feature */
extern T0* T1513f56(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.flattened_immediate_feature */
extern T0* T1512f60(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.flattened_immediate_feature */
extern T0* T1511f58(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.flattened_immediate_feature */
extern T0* T371f38(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.flattened_immediate_feature */
extern T0* T369f37(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.flattened_immediate_feature */
extern T0* T359f37(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.flattened_immediate_feature */
extern T0* T360f36(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.is_immediate */
extern T1 T1345f10(GE_context* ac, T0* C);
/* [detachable] ET_REDECLARED_FEATURE.is_immediate */
extern T1 T1569f9(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_flattened_immediate */
extern T1 T1516f59(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_flattened_immediate */
extern T1 T253f38(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_flattened_immediate */
extern T1 T476f50(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_flattened_immediate */
extern T1 T1515f54(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_flattened_immediate */
extern T1 T469f52(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_flattened_immediate */
extern T1 T1513f55(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_flattened_immediate */
extern T1 T1512f59(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_flattened_immediate */
extern T1 T1511f57(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_flattened_immediate */
extern T1 T371f37(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_flattened_immediate */
extern T1 T369f36(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_flattened_immediate */
extern T1 T359f36(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_flattened_immediate */
extern T1 T360f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].item_for_iteration */
extern T0* T1310f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].cursor_item */
extern T0* T1310f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].before */
extern T1 T1310f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].cursor_before */
extern T1 T1310f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].before */
extern T1 T1662f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].finish */
extern void T1310f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].cursor_finish */
extern void T1310f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.resolve_feature_adaptations */
extern void T459f39(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.resolve_feature_adaptations */
extern void T1318f23(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.process_replication */
extern void T1318f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].wipe_out */
extern void T1682f45(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].slots_wipe_out */
extern void T1682f57(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].clashes_wipe_out */
extern void T1682f56(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].key_storage_wipe_out */
extern void T1682f55(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].item_storage_wipe_out */
extern void T1682f54(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_REPLICABLE_FEATURE].keep_head */
extern void T1912f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_REPLICABLE_FEATURE].set_count */
extern void T1912f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].unset_found_item */
extern void T1682f53(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].move_all_cursors_after */
extern void T1682f52(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].set_next_cursor */
extern void T1911f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].set_after */
extern void T1911f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].forth */
extern void T1682f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].cursor_forth */
extern void T1682f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].add_traversing_cursor */
extern void T1682f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].set_position */
extern void T1911f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].remove_traversing_cursor */
extern void T1682f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].clashes_item */
extern T6 T1682f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_REDECLARED_FEATURE.selected_count */
extern T6 T1569f25(GE_context* ac, T0* C);
/* [detachable] ET_REDECLARED_FEATURE.has_selected_feature */
extern T1 T1569f18(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.selected_count */
extern T6 T1345f27(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.has_selected_feature */
extern T1 T1345f19(GE_context* ac, T0* C);
/* [detachable] ET_REPLICATED_FEATURE.selected_count */
extern T6 T1684f2(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.process_replicated_feature */
extern void T1318f34(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vmrc2b_error */
extern void T53f354(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vmrc2b */
extern T0* T785c987(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].after */
extern T1 T1919f4(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].forth */
extern void T1919f11(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE].cursor_forth */
extern void T1686f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].set_position */
extern void T1919f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE].add_traversing_cursor */
extern void T1686f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].set_next_cursor */
extern void T1919f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE].remove_traversing_cursor */
extern void T1686f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].item */
extern T0* T1919f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE].cursor_item */
extern T0* T1686f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].start */
extern void T1919f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE].cursor_start */
extern void T1686f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].set_after */
extern void T1919f12(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE].is_empty */
extern T1 T1686f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].off */
extern T1 T1919f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].before */
extern T1 T1919f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE].new_cursor */
extern T0* T1686f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PARENT_FEATURE].make */
extern T0* T1919c9(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE].first */
extern T0* T1686f7(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vmrc2_error */
extern T1 T53f86(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vmrc2b_error */
extern void T36f354(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vmrc2_error */
extern T1 T36f86(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REDECLARED_FEATURE.add_replicated_feature */
extern void T1569f33(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE].force_last */
extern void T1696f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] ET_PARENT_FEATURE].put_right */
extern void T2071f4(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] ET_PARENT_FEATURE].make */
extern T0* T2071c3(GE_context* ac, T0* a1);
/* detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE].make */
extern T0* T1696c7(GE_context* ac);
/* detachable DS_LINKED_LIST [[attached] ET_PARENT_FEATURE].set_internal_cursor */
extern void T1696f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INHERITED_FEATURE.add_replicated_feature */
extern void T1345f37(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENT_FEATURE.has_seed */
extern T1 T1344f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_REDECLARED_FEATURE.set_selected */
extern void T1569f35(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.set_selected */
extern void T1345f39(GE_context* ac, T0* C);
/* [detachable] ET_REDECLARED_FEATURE.set_replicated */
extern void T1569f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_INHERITED_FEATURE.set_replicated */
extern void T1345f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_REPLICATED_FEATURE.first_feature */
extern T0* T1684f5(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].first */
extern T0* T1685f8(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vmrc2a_error */
extern void T53f353(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vmrc2a */
extern T0* T785c986(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vmrc2a_error */
extern void T36f353(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.error_handler */
extern T0* T1318f18(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.set_fatal_error */
extern void T1318f27(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].forth */
extern void T1685f16(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].cursor_forth */
extern void T1685f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].add_traversing_cursor */
extern void T1685f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_ADAPTED_FEATURE].set_next_cursor */
extern void T1915f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].remove_traversing_cursor */
extern void T1685f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_ADAPTED_FEATURE].set */
extern void T1915f8(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE].put_last */
extern void T1686f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_FEATURE].force */
extern void T1917f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_PARENT_FEATURE].fill_with */
extern void T1918f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_PARENT_FEATURE].extend */
extern void T1918f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REDECLARED_FEATURE.seeded_feature */
extern T0* T1569f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_INHERITED_FEATURE.seeded_feature */
extern T0* T1345f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].item_for_iteration */
extern T0* T1685f7(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].cursor_item */
extern T0* T1685f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].after */
extern T1 T1685f6(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].cursor_after */
extern T1 T1685f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].start */
extern void T1685f15(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].cursor_start */
extern void T1685f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].cursor_off */
extern T1 T1685f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_ADAPTED_FEATURE].off */
extern T1 T1915f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE].make */
extern T0* T1686c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PARENT_FEATURE].set_internal_cursor */
extern void T1686f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_FEATURE].make */
extern T0* T1917f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_PARENT_FEATURE].make_empty */
extern T0* T1918c4(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_FEATURE].default_create */
extern T0* T1917c2(GE_context* ac);
/* [detachable] ET_REDECLARED_FEATURE.replicated_feature */
extern T0* T1569f23(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.replicated_feature */
extern T0* T1345f25(GE_context* ac, T0* C);
/* [detachable] ET_REPLICATED_FEATURE.replicated_feature */
extern T0* T1684f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].key_for_iteration */
extern T6 T1682f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].cursor_key */
extern T6 T1682f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].key_storage_item */
extern T6 T1682f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_REDECLARED_FEATURE.has_replication */
extern T1 T1569f22(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.has_replication */
extern T1 T1345f24(GE_context* ac, T0* C);
/* [detachable] ET_REPLICATED_FEATURE.has_replication */
extern T1 T1684f3(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].item_for_iteration */
extern T0* T1682f26(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].cursor_item */
extern T0* T1682f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].item_storage_item */
extern T0* T1682f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].after */
extern T1 T1682f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].cursor_after */
extern T1 T1682f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].after */
extern T1 T1911f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].start */
extern void T1682f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].cursor_start */
extern void T1682f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].cursor_off */
extern T1 T1682f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].off */
extern T1 T1911f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].is_empty */
extern T1 T1682f28(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.record_replicable_feature */
extern void T1318f33(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].force_new */
extern void T1682f48(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].key_storage_put */
extern void T1682f63(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].item_storage_put */
extern void T1682f59(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_REPLICABLE_FEATURE].force */
extern void T1914f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_REPLICABLE_FEATURE].fill_with */
extern void T1912f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_REPLICABLE_FEATURE].extend */
extern void T1912f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].slots_put */
extern void T1682f62(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].clashes_put */
extern void T1682f61(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].slots_item */
extern T6 T1682f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].hash_position */
extern T6 T1682f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].resize */
extern void T1682f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].clashes_resize */
extern void T1682f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].special_integer_ */
extern T0* T1682f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].key_storage_resize */
extern void T1682f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].item_storage_resize */
extern void T1682f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_REPLICABLE_FEATURE].aliased_resized_area */
extern T0* T1914f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_REPLICABLE_FEATURE].aliased_resized_area */
extern T0* T1912f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].slots_resize */
extern void T1682f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].new_modulus */
extern T6 T1682f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].new_capacity */
extern T6 T1682f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].replace_found_item */
extern void T1682f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REPLICATED_FEATURE.make */
extern T0* T1684c6(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].make */
extern T0* T1685c12(GE_context* ac);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].set_internal_cursor */
extern void T1685f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].new_cursor */
extern T0* T1685f5(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_ADAPTED_FEATURE].make */
extern T0* T1915c7(GE_context* ac, T0* a1);
/* [detachable] ET_REDECLARED_FEATURE.first_feature */
extern T0* T1569f24(GE_context* ac, T0* C);
/* [detachable] ET_INHERITED_FEATURE.first_feature */
extern T0* T1345f26(GE_context* ac, T0* C);
/* [detachable] ET_REPLICATED_FEATURE.put_feature */
extern void T1684f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_ADAPTED_FEATURE].force_last */
extern void T1685f13(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] ET_ADAPTED_FEATURE].put_right */
extern void T1916f4(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] ET_ADAPTED_FEATURE].make */
extern T0* T1916c3(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].found_item */
extern T0* T1682f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].found */
extern T1 T1682f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].search */
extern void T1682f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].search_position */
extern void T1682f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.add_inherited_features */
extern void T1318f25(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vmss1a_error */
extern void T53f346(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vmss1a */
extern T0* T785c979(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vmss1_error */
extern T1 T53f79(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vmss1a_error */
extern void T36f346(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vmss1_error */
extern T1 T36f79(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrs1a_error */
extern void T53f345(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdrs1a */
extern T0* T785c978(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdrs1_error */
extern T1 T53f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdrs1a_error */
extern void T36f345(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdrs1_error */
extern T1 T36f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].wipe_out */
extern void T1681f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].slots_wipe_out */
extern void T1681f62(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].clashes_wipe_out */
extern void T1681f61(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].key_storage_wipe_out */
extern void T1681f60(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].item_storage_wipe_out */
extern void T1681f59(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].unset_found_item */
extern void T1681f58(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].move_all_cursors_after */
extern void T1681f57(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].set_next_cursor */
extern void T1909f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].set_after */
extern void T1909f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].forth */
extern void T1681f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].cursor_forth */
extern void T1681f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].add_traversing_cursor */
extern void T1681f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].set_position */
extern void T1909f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].remove_traversing_cursor */
extern void T1681f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].clashes_item */
extern T6 T1681f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdus1a_error */
extern void T53f344(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vdus1a */
extern T0* T785c977(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdus1_error */
extern T1 T53f77(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdus1a_error */
extern void T36f344(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vdus1_error */
extern T1 T36f77(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].key_for_iteration */
extern T0* T1681f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].cursor_key */
extern T0* T1681f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].key_storage_item */
extern T0* T1681f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].item_for_iteration */
extern T1 T1681f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].cursor_item */
extern T1 T1681f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].item_storage_item */
extern T1 T1681f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].after */
extern T1 T1681f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].cursor_after */
extern T1 T1681f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].after */
extern T1 T1909f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].start */
extern void T1681f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].cursor_start */
extern void T1681f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].cursor_off */
extern T1 T1681f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].off */
extern T1 T1909f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].is_empty */
extern T1 T1681f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].wipe_out */
extern void T1680f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].slots_wipe_out */
extern void T1680f60(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].clashes_wipe_out */
extern void T1680f59(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].key_storage_wipe_out */
extern void T1680f58(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].item_storage_wipe_out */
extern void T1680f57(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].unset_found_item */
extern void T1680f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].move_all_cursors_after */
extern void T1680f56(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_FEATURE_NAME].set_next_cursor */
extern void T1908f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_FEATURE_NAME].set_after */
extern void T1908f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].forth */
extern void T1680f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].cursor_forth */
extern void T1680f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].add_traversing_cursor */
extern void T1680f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_FEATURE_NAME].set_position */
extern void T1908f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].remove_traversing_cursor */
extern void T1680f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].clashes_item */
extern T6 T1680f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vlel2a_error */
extern void T53f343(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vlel2a */
extern T0* T785c976(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vlel2_error */
extern T1 T53f76(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vlel2a_error */
extern void T36f343(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vlel2_error */
extern T1 T36f76(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].item_for_iteration */
extern T0* T1680f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].cursor_item */
extern T0* T1680f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].item_storage_item */
extern T0* T1680f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].after */
extern T1 T1680f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].cursor_after */
extern T1 T1680f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_FEATURE_NAME].after */
extern T1 T1908f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].start */
extern void T1680f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].cursor_start */
extern void T1680f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].cursor_off */
extern T1 T1680f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_FEATURE_NAME].off */
extern T1 T1908f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].wipe_out */
extern void T1332f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].slots_wipe_out */
extern void T1332f57(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].clashes_wipe_out */
extern void T1332f56(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].key_storage_wipe_out */
extern void T1332f55(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].item_storage_wipe_out */
extern void T1332f54(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_RENAME].keep_head */
extern void T1704f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_RENAME].set_count */
extern void T1704f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].unset_found_item */
extern void T1332f53(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].move_all_cursors_after */
extern void T1332f52(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].set_next_cursor */
extern void T1703f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].set_after */
extern void T1703f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].forth */
extern void T1332f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].cursor_forth */
extern void T1332f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].add_traversing_cursor */
extern void T1332f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].set_position */
extern void T1703f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].remove_traversing_cursor */
extern void T1332f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].clashes_item */
extern T6 T1332f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vhrc1a_error */
extern void T53f342(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vhrc1a */
extern T0* T785c975(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vhrc1_error */
extern T1 T53f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vhrc1a_error */
extern void T36f342(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vhrc1_error */
extern T1 T36f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].item_for_iteration */
extern T0* T1332f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].cursor_item */
extern T0* T1332f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].item_storage_item */
extern T0* T1332f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].after */
extern T1 T1332f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].cursor_after */
extern T1 T1332f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].after */
extern T1 T1703f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].start */
extern void T1332f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].cursor_start */
extern void T1332f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].cursor_off */
extern T1 T1332f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].off */
extern T1 T1703f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].put_last_new */
extern void T1310f75(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].key_storage_put */
extern void T1310f86(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].item_storage_put */
extern void T1310f85(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].slots_put */
extern void T1310f79(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].clashes_put */
extern void T1310f80(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].compress */
extern void T1310f84(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].key_storage_keep_head */
extern void T1310f89(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].item_storage_keep_head */
extern void T1310f88(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].move_all_cursors */
extern void T1310f87(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.new_inherited_feature */
extern T0* T1318f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INHERITED_FEATURE.set_next */
extern void T1345f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INHERITED_FEATURE.make */
extern T0* T1345c29(GE_context* ac, T0* a1);
/* [detachable] ET_INHERITED_FEATURE.reset */
extern void T1345f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REDECLARED_FEATURE.put_parent_feature */
extern void T1569f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REDECLARED_FEATURE.has_seed */
extern T1 T1569f21(GE_context* ac, T0* C, T6 a1);
/* detachable ET_PARENT_FEATURE.set_merged_feature */
extern void T1344f33(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENT_FEATURE.set_has_other_effective */
extern void T1344f32(GE_context* ac, T0* C, T1 a1);
/* detachable ET_PARENT_FEATURE.set_has_other_deferred */
extern void T1344f31(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_INHERITED_FEATURE.put_parent_feature */
extern void T1345f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INHERITED_FEATURE.has_seed */
extern T1 T1345f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].replace_found_item */
extern void T1310f76(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.new_redeclared_feature */
extern T0* T1318f20(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_REDECLARED_FEATURE.set_next */
extern void T1569f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REDECLARED_FEATURE.make */
extern T0* T1569c27(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_REDECLARED_FEATURE.reset */
extern void T1569f28(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_PARENT_FEATURE.set_select_name */
extern void T1344f28(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENT_FEATURE.set_redefine_name */
extern void T1344f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].replace_found_item */
extern void T1681f46(GE_context* ac, T0* C, T1 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].item_storage_put */
extern void T1681f54(GE_context* ac, T0* C, T1 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].found_item */
extern T1 T1681f25(GE_context* ac, T0* C);
/* detachable ET_PARENT_FEATURE.set_undefine_name */
extern void T1344f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].found_key */
extern T0* T1681f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].found */
extern T1 T1681f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].search */
extern void T1681f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].search_position */
extern void T1681f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].hash_position */
extern T6 T1681f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].slots_item */
extern T6 T1681f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].remove_found_item */
extern void T1680f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].remove_position */
extern void T1680f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].key_storage_put */
extern void T1680f72(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].item_storage_put */
extern void T1680f67(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].key_storage_keep_head */
extern void T1680f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].item_storage_keep_head */
extern void T1680f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].clashes_put */
extern void T1680f63(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].slots_put */
extern void T1680f62(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].move_cursors_forth */
extern void T1680f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].search_position */
extern void T1680f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].internal_set_key_equality_tester */
extern void T1680f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].key_equality_tester */
extern T0* T1680f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].slots_item */
extern T6 T1680f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].hash_position */
extern T6 T1680f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].key_storage_item */
extern T0* T1680f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].found_item */
extern T0* T1680f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].found */
extern T1 T1680f17(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].search */
extern void T1680f39(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENT_FEATURE.set_new_name */
extern void T1344f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].remove_found_item */
extern void T1332f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].remove_position */
extern void T1332f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].key_storage_put */
extern void T1332f65(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].item_storage_put */
extern void T1332f64(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME].force */
extern void T1707f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_RENAME].fill_with */
extern void T1704f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_RENAME].extend */
extern void T1704f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].key_storage_keep_head */
extern void T1332f73(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].item_storage_keep_head */
extern void T1332f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].clashes_put */
extern void T1332f60(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].slots_put */
extern void T1332f59(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].move_cursors_forth */
extern void T1332f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].search_position */
extern void T1332f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].internal_set_key_equality_tester */
extern void T1332f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].slots_item */
extern T6 T1332f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].hash_position */
extern T6 T1332f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].key_storage_item */
extern T0* T1332f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].found_item */
extern T0* T1332f22(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].found */
extern T1 T1332f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].search */
extern void T1332f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.new_parent_feature */
extern T0* T1318f19(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_PARENT_FEATURE.set_next */
extern void T1344f30(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENT_FEATURE.make */
extern T0* T1344c24(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_PARENT_FEATURE.reset */
extern void T1344f29(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].resize */
extern void T1310f74(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].clashes_resize */
extern void T1310f83(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].special_integer_ */
extern T0* T1310f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].key_storage_resize */
extern void T1310f82(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].item_storage_resize */
extern void T1310f81(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].slots_resize */
extern void T1310f78(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].new_modulus */
extern T6 T1310f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.fill_select_table */
extern void T1318f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vmss2a_error */
extern void T53f352(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vmss2a */
extern T0* T785c985(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vmss2_error */
extern T1 T53f85(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vmss2a_error */
extern void T36f352(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vmss2_error */
extern T1 T36f85(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].put_new */
extern void T1681f51(GE_context* ac, T0* C, T1 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].key_storage_put */
extern void T1681f69(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].slots_put */
extern void T1681f64(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].clashes_put */
extern void T1681f65(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_KEYWORD_FEATURE_NAME_LIST.feature_name */
extern T0* T574f6(GE_context* ac, T0* C, T6 a1);
/* detachable ET_KEYWORD_FEATURE_NAME_LIST.item */
extern T0* T574f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].resize */
extern void T1681f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].clashes_resize */
extern void T1681f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].special_integer_ */
extern T0* T1681f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].key_storage_resize */
extern void T1681f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].item_storage_resize */
extern void T1681f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].slots_resize */
extern void T1681f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].new_modulus */
extern T6 T1681f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.fill_redefine_table */
extern void T1318f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdrs3a_error */
extern void T53f351(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vdrs3a */
extern T0* T785c984(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdrs3_error */
extern T1 T53f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdrs3a_error */
extern void T36f351(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vdrs3_error */
extern T1 T36f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.fill_undefine_table */
extern void T1318f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vdus4a_error */
extern void T53f350(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vdus4a */
extern T0* T785c983(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vdus4_error */
extern T1 T53f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vdus4a_error */
extern void T36f350(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vdus4_error */
extern T1 T36f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].is_empty */
extern T1 T1680f32(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.fill_export_table */
extern void T1318f29(GE_context* ac, T0* C, T0* a1);
extern T1 T1318f29ot1(T0* a1, T0** a2);
extern T1 T1318f29ot2(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vlel3a_error */
extern void T53f349(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vlel3a */
extern T0* T785c982(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vlel3_error */
extern T1 T53f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vlel3a_error */
extern void T36f349(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vlel3_error */
extern T1 T36f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].put_new */
extern void T1680f45(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FEATURE_EXPORT.feature_name */
extern T0* T1145f5(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FEATURE_EXPORT.item */
extern T0* T1145f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].resize */
extern void T1680f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].clashes_resize */
extern void T1680f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].special_integer_ */
extern T0* T1680f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].key_storage_resize */
extern void T1680f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].item_storage_resize */
extern void T1680f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].slots_resize */
extern void T1680f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].new_modulus */
extern T6 T1680f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vlel1a_error */
extern void T53f348(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vlel1a */
extern T0* T785c981(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ALL_EXPORT.position */
extern T0* T1506f4(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vlel1_error */
extern T1 T53f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vlel1a_error */
extern void T36f348(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vlel1_error */
extern T1 T36f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].is_empty */
extern T1 T1332f30(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.fill_rename_table */
extern void T1318f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vhrc2a_error */
extern void T53f347(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vhrc2a */
extern T0* T785c980(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vhrc2_error */
extern T1 T53f80(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vhrc2a_error */
extern void T36f347(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vhrc2_error */
extern T1 T36f80(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALIASED_FEATURE_NAME.same_extended_feature_name */
extern T1 T1517f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IDENTIFIER.same_extended_feature_name */
extern T1 T184f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].put_new */
extern void T1332f46(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_RENAME_LIST.rename_pair */
extern T0* T572f6(GE_context* ac, T0* C, T6 a1);
/* detachable ET_RENAME_LIST.item */
extern T0* T572f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].resize */
extern void T1332f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].clashes_resize */
extern void T1332f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].special_integer_ */
extern T0* T1332f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].key_storage_resize */
extern void T1332f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].item_storage_resize */
extern void T1332f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME].aliased_resized_area */
extern T0* T1707f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_RENAME].aliased_resized_area */
extern T0* T1704f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].slots_resize */
extern void T1332f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].new_modulus */
extern T6 T1332f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.add_current_features */
extern void T1318f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vmfn0a_error */
extern void T53f341(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vmfn0a */
extern T0* T785c974(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vmfn0a_error */
extern void T36f341(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.resolve_feature_adaptations */
extern void T1319f20(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].wipe_out */
extern void T1690f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].slots_wipe_out */
extern void T1690f45(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].clashes_wipe_out */
extern void T1690f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].key_storage_wipe_out */
extern void T1690f43(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_DOTNET_FEATURE].keep_head */
extern void T1920f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DOTNET_FEATURE].set_count */
extern void T1920f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].item_storage_wipe_out */
extern void T1690f42(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]].keep_head */
extern void T1923f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]].set_count */
extern void T1923f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].unset_found_item */
extern void T1690f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].move_all_cursors_after */
extern void T1690f40(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].set_next_cursor */
extern void T1926f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE_CURSOR [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].set_after */
extern void T1926f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].wipe_out */
extern void T1688f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].slots_wipe_out */
extern void T1688f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].clashes_wipe_out */
extern void T1688f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].key_storage_wipe_out */
extern void T1688f45(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].item_storage_wipe_out */
extern void T1688f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].unset_found_item */
extern void T1688f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].move_all_cursors_after */
extern void T1688f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DOTNET_FEATURE].set_next_cursor */
extern void T1921f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DOTNET_FEATURE].set_after */
extern void T1921f7(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.add_any_features */
extern void T1319f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.new_inherited_feature */
extern T0* T1319f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_RENAME.make */
extern T0* T1337c6(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.as_keyword */
extern T0* T73f705(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_as */
extern T0* T413c66(GE_context* ac);
/* [detachable] ET_RENAME.tokens */
extern T0* T1337f4(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.new_parent_feature */
extern T0* T1319f16(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.current_system */
extern T0* T1319f15(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.system_object_parents */
extern T0* T187f88(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.system_object_parents */
extern T0* T58f117(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.system_object_parents */
extern T0* T40f9(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.current_universe */
extern T0* T1319f13(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.add_inherited_features_from_parents */
extern void T1319f22(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.add_inherited_features */
extern void T1319f26(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.error_handler */
extern T0* T1319f14(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.set_fatal_error */
extern void T1319f25(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].forth */
extern void T1692f15(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].cursor_forth */
extern void T1692f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].add_traversing_cursor */
extern void T1692f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_DOTNET_FEATURE].set_next_cursor */
extern void T1928f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].remove_traversing_cursor */
extern void T1692f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_DOTNET_FEATURE].set */
extern void T1928f8(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].item_for_iteration */
extern T0* T1692f7(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].cursor_item */
extern T0* T1692f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].after */
extern T1 T1692f6(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].cursor_after */
extern T1 T1692f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].start */
extern void T1692f14(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].cursor_start */
extern void T1692f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].cursor_off */
extern T1 T1692f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_DOTNET_FEATURE].off */
extern T1 T1928f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].found_item */
extern T0* T1690f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].item_storage_item */
extern T0* T1690f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].found */
extern T1 T1690f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].search */
extern void T1690f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].search_position */
extern void T1690f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].hash_position */
extern T6 T1690f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_SIGNATURE_TESTER.test */
extern T1 T1687f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].key_storage_item */
extern T0* T1690f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].clashes_item */
extern T6 T1690f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].slots_item */
extern T6 T1690f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.add_inherited_feature */
extern void T1319f27(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.new_redeclared_feature */
extern T0* T1319f18(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].found_item */
extern T0* T1688f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].item_storage_item */
extern T0* T1688f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].found */
extern T1 T1688f26(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].search */
extern void T1688f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].search_position */
extern void T1688f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].hash_position */
extern T6 T1688f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].key_storage_item */
extern T0* T1688f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].clashes_item */
extern T6 T1688f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].key_equality_tester */
extern T0* T1688f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].slots_item */
extern T6 T1688f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_PARENT_LIST.has_conforming_parent */
extern T1 T568f10(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.add_current_features */
extern void T1319f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.add_current_feature */
extern void T1319f24(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].force_last */
extern void T1688f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].slots_put */
extern void T1688f49(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].clashes_put */
extern void T1688f50(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].new_capacity */
extern T6 T1688f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].item_storage_put */
extern void T1688f55(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_FEATURE].force */
extern void T1922f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DOTNET_FEATURE].fill_with */
extern void T1920f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DOTNET_FEATURE].extend */
extern void T1920f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].force_last_new */
extern void T1690f38(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].key_storage_put */
extern void T1690f51(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].item_storage_put */
extern void T1690f50(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]].force */
extern void T1927f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]].fill_with */
extern void T1923f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]].extend */
extern void T1923f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].slots_put */
extern void T1690f49(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].clashes_put */
extern void T1690f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].resize */
extern void T1690f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].clashes_resize */
extern void T1690f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].special_integer_ */
extern T0* T1690f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].key_storage_resize */
extern void T1690f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_FEATURE].aliased_resized_area */
extern T0* T1922f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_DOTNET_FEATURE].aliased_resized_area */
extern T0* T1920f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].item_storage_resize */
extern void T1690f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]].aliased_resized_area */
extern T0* T1927f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]].aliased_resized_area */
extern T0* T1923f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].slots_resize */
extern void T1690f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].new_modulus */
extern T6 T1690f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].new_capacity */
extern T6 T1690f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].make */
extern T0* T1692c11(GE_context* ac);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].set_internal_cursor */
extern void T1692f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].new_cursor */
extern T0* T1692f5(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] ET_DOTNET_FEATURE].make */
extern T0* T1928c7(GE_context* ac, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE].force_last */
extern void T1692f12(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] ET_DOTNET_FEATURE].put_right */
extern void T1929f4(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] ET_DOTNET_FEATURE].make */
extern T0* T1929c3(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].force_last_new */
extern void T1310f77(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].new_capacity */
extern T6 T1310f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].resize */
extern void T1688f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].clashes_resize */
extern void T1688f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].special_integer_ */
extern T0* T1688f27(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].key_storage_resize */
extern void T1688f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].item_storage_resize */
extern void T1688f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].slots_resize */
extern void T1688f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].new_modulus */
extern T6 T1688f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_FLATTENER.check_parents_validity */
extern void T459f37(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER2.check_parents_validity */
extern void T1324f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.check_formal_parameters_validity */
extern void T459f36(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.check_formal_parameters_validity */
extern void T1325f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.check_constraint_validity */
extern void T1325f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_LIST.parent */
extern T0* T568f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_PARENT.parent */
extern T0* T567f10(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_LIST.item */
extern T0* T568f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS.parents */
extern T0* T74f117(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS.implicit_parent_clause */
extern T0* T74f131(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.any_parents */
extern T0* T187f23(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.any_parents */
extern T0* T58f87(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.any_parents */
extern T0* T40f10(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.parents_count */
extern T6 T74f116(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.has_explicit_conforming_parent */
extern T1 T74f130(GE_context* ac, T0* C);
/* detachable ET_PARENT_CLAUSE_LIST.has_conforming_parent */
extern T1 T1237f5(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.set_fatal_error */
extern void T459f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_flattening_error */
extern void T74f257(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_FLATTENER.make */
extern T0* T459c32(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].make_map */
extern T0* T1328c39(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].make_with_equality_testers */
extern void T1328f45(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].make_sparse_container */
extern void T1328f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].set_internal_cursor */
extern void T1328f76(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].new_cursor */
extern T0* T1328f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FEATURE, [attached] INTEGER_32].make */
extern T0* T1698c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].make_slots */
extern void T1328f75(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].make_clashes */
extern void T1328f74(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].make_key_storage */
extern void T1328f73(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FEATURE, [attached] INTEGER_32].make_item_storage */
extern void T1328f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_PRECURSOR_CHECKER.make */
extern T0* T1327c10(GE_context* ac, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.dummy_feature */
extern T0* T1327f6(GE_context* ac, T0* C);
/* [detachable] ET_PRECURSOR_CHECKER.current_system */
extern T0* T1327f7(GE_context* ac, T0* C);
/* [detachable] ET_PRECURSOR_CHECKER.make */
extern void T1327f10p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PRECURSOR_CHECKER.tokens */
extern T0* T1327f5(GE_context* ac, T0* C);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.make */
extern T0* T1326c12(GE_context* ac, T0* a1);
/* [detachable] ET_BUILTIN_FEATURE_CHECKER.make */
extern void T1326f12p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].make_map */
extern T0* T1697c33(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].make_with_equality_testers */
extern void T1697f34(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].make_sparse_container */
extern void T1697f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].set_internal_cursor */
extern void T1697f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].new_cursor */
extern T0* T1697f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].make */
extern T0* T1935c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].make_slots */
extern void T1697f39(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].make_clashes */
extern void T1697f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].make_key_storage */
extern void T1697f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME], [attached] NATURAL_8].make_item_storage */
extern void T1697f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]].make */
extern T0* T1936f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]].make_empty */
extern T0* T1938c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] TUPLE [detachable ARRAY [[attached] ET_TYPE], detachable ET_TYPE, [attached] NATURAL_8], [attached] ET_FEATURE_NAME]].default_create */
extern T0* T1936c3(GE_context* ac);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.make */
extern T0* T1325c8(GE_context* ac, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.make */
extern void T1325f8p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER2.tokens */
extern T0* T1325f5(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER2.make */
extern T0* T1324c9(GE_context* ac, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.make */
extern void T1324f9p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER2.tokens */
extern T0* T1324f5(GE_context* ac, T0* C);
/* [detachable] ET_SIGNATURE_CHECKER.make */
extern T0* T1323c10(GE_context* ac, T0* a1);
/* [detachable] ET_SIGNATURE_CHECKER.make */
extern void T1323f10p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.make */
extern T0* T1322c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].make_default */
extern T0* T1693c10(GE_context* ac);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_LIKE_FEATURE].make */
extern void T1693f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable DS_LINKABLE [[attached] INTEGER_32]].make */
extern T0* T699c11(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable DS_LINKABLE [[attached] INTEGER_32]].make */
extern T0* T1415f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable DS_LINKABLE [[attached] INTEGER_32]].default_create */
extern T0* T1415c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].make */
extern void T263f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].make */
extern T0* T263c16(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].make_map */
extern T0* T1930c33(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].make_with_equality_testers */
extern void T1930f37(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].make_sparse_container */
extern void T1930f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].set_internal_cursor */
extern void T1930f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].new_cursor */
extern T0* T1930f26(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].make */
extern T0* T1993c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].make_slots */
extern void T1930f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].make_clashes */
extern void T1930f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].make_key_storage */
extern void T1930f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_LIKE_FEATURE].make_item_storage */
extern void T1930f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.make */
extern void T1322f8p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCHORED_TYPE_CHECKER.tokens */
extern T0* T1322f7(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.make */
extern T0* T1321c6(GE_context* ac, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.make */
extern void T1321f6p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER1.tokens */
extern T0* T1321f4(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.make */
extern T0* T1320c7(GE_context* ac, T0* a1);
/* [detachable] ET_IDENTIFIER_TYPE_RESOLVER.tokens */
extern T0* T1320f5(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.make */
extern T0* T1319c19(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].set_key_equality_tester */
extern void T1690f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].make_map */
extern T0* T1690c34(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].make_with_equality_testers */
extern void T1690f39(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].make_sparse_container */
extern void T1690f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].set_internal_cursor */
extern void T1690f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].new_cursor */
extern T0* T1690f31(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].make */
extern T0* T1926c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].make_slots */
extern void T1690f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].make_clashes */
extern void T1690f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].make_key_storage */
extern void T1690f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_FEATURE].make */
extern T0* T1922f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DOTNET_FEATURE].make_empty */
extern T0* T1920c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_FEATURE].default_create */
extern T0* T1922c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE], [attached] ET_DOTNET_FEATURE].make_item_storage */
extern void T1690f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]].make */
extern T0* T1927f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]].make_empty */
extern T0* T1923c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_LINKED_LIST [[attached] ET_DOTNET_FEATURE]].default_create */
extern T0* T1927c3(GE_context* ac);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].set_equality_tester */
extern void T1688f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_SIGNATURE_TESTER.make */
extern T0* T1687c2(GE_context* ac);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].make */
extern T0* T1688c31(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].set_internal_cursor */
extern void T1688f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].new_cursor */
extern T0* T1688f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DOTNET_FEATURE].make */
extern T0* T1921c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].make_slots */
extern void T1688f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].make_clashes */
extern void T1688f39(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].make_key_storage */
extern void T1688f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_FEATURE].make_item_storage */
extern void T1688f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.make */
extern void T1319f19p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_FEATURE_ADAPTATION_RESOLVER.tokens */
extern T0* T1319f12(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.make */
extern T0* T1318c22(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].make_map */
extern T0* T1682c42(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].make_with_equality_testers */
extern void T1682f49(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].make_sparse_container */
extern void T1682f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].set_internal_cursor */
extern void T1682f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].new_cursor */
extern T0* T1682f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].make */
extern T0* T1911c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].make_slots */
extern void T1682f74(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].make_clashes */
extern void T1682f73(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].make_key_storage */
extern void T1682f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_REPLICABLE_FEATURE, [attached] INTEGER_32].make_item_storage */
extern void T1682f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_REPLICABLE_FEATURE].make */
extern T0* T1914f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_REPLICABLE_FEATURE].make_empty */
extern T0* T1912c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_REPLICABLE_FEATURE].default_create */
extern T0* T1914c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].set_key_equality_tester */
extern void T1681f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].make_map */
extern T0* T1681c43(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].make_with_equality_testers */
extern void T1681f52(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].make_sparse_container */
extern void T1681f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].set_internal_cursor */
extern void T1681f77(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].new_cursor */
extern T0* T1681f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].make */
extern T0* T1909c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].make_slots */
extern void T1681f76(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].make_clashes */
extern void T1681f75(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].make_key_storage */
extern void T1681f74(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] ET_FEATURE_NAME].make_item_storage */
extern void T1681f73(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].set_equality_tester */
extern void T1680f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].make */
extern T0* T1680c37(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].set_internal_cursor */
extern void T1680f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].new_cursor */
extern T0* T1680f20(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_FEATURE_NAME].make */
extern T0* T1908c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].make_slots */
extern void T1680f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].make_clashes */
extern void T1680f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].make_key_storage */
extern void T1680f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_FEATURE_NAME].make_item_storage */
extern void T1680f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].set_key_equality_tester */
extern void T1332f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.feature_name_tester */
extern T0* T1318f17(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].make_map */
extern T0* T1332c41(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].make_with_equality_testers */
extern void T1332f51(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].make_sparse_container */
extern void T1332f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].set_internal_cursor */
extern void T1332f80(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].new_cursor */
extern T0* T1332f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].make */
extern T0* T1703c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].make_slots */
extern void T1332f79(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].make_clashes */
extern void T1332f78(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].make_key_storage */
extern void T1332f77(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].make_item_storage */
extern void T1332f76(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME].make */
extern T0* T1707f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_RENAME].make_empty */
extern T0* T1704c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME].default_create */
extern T0* T1707c3(GE_context* ac);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.make */
extern void T1318f22p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_ADAPTATION_RESOLVER.tokens */
extern T0* T1318f16(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].make_map */
extern T0* T1317c41(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].make_with_equality_testers */
extern void T1317f42(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].make_sparse_container */
extern void T1317f43(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].set_internal_cursor */
extern void T1317f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].make_slots */
extern void T1317f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].make_clashes */
extern void T1317f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].make_key_storage */
extern void T1317f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CLIENT, [attached] ET_CLASS].make_item_storage */
extern void T1317f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT].make */
extern T0* T1678f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CLIENT].make_empty */
extern T0* T1679c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT].default_create */
extern T0* T1678c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].make */
extern T0* T1316c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].set_internal_cursor */
extern void T1316f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLIENT_LIST].new_cursor */
extern T0* T1316f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLIENT_LIST].make */
extern T0* T1675c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_LIST].make */
extern T0* T1673f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CLIENT_LIST].make_empty */
extern T0* T1674c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLIENT_LIST].default_create */
extern T0* T1673c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].set_key_equality_tester */
extern void T1313f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.alias_name_tester */
extern T0* T459f24(GE_context* ac, T0* C);
/* [detachable] ET_ALIAS_NAME_TESTER.default_create */
extern T0* T1314c2(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].make_map */
extern T0* T1313c35(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].make_with_equality_testers */
extern void T1313f40(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].make_sparse_container */
extern void T1313f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].set_internal_cursor */
extern void T1313f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].new_cursor */
extern T0* T1313f32(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].make */
extern T0* T1672c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].make_slots */
extern void T1313f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].make_clashes */
extern void T1313f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].make_key_storage */
extern void T1313f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ALIAS_NAME].make */
extern T0* T1671f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ALIAS_NAME].make_empty */
extern T0* T1669c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ALIAS_NAME].default_create */
extern T0* T1671c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_ALIAS_NAME].make_item_storage */
extern void T1313f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FLATTENED_FEATURE].make */
extern T0* T1667f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_FLATTENED_FEATURE].make_empty */
extern T0* T1663c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FLATTENED_FEATURE].default_create */
extern T0* T1667c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].set_key_equality_tester */
extern void T1310f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.feature_name_tester */
extern T0* T459f22(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].make_map */
extern T0* T1310c44(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].make_with_equality_testers */
extern void T1310f54(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].make_sparse_container */
extern void T1310f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].set_internal_cursor */
extern void T1310f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].new_cursor */
extern T0* T1310f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].make */
extern T0* T1662c9(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].make_slots */
extern void T1310f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].make_clashes */
extern void T1310f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].make_key_storage */
extern void T1310f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_FLATTENED_FEATURE, [attached] ET_FEATURE_NAME].make_item_storage */
extern void T1310f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FEATURE_FLATTENER.make */
extern void T459f32p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_FLATTENER.error_handler */
extern T0* T459f23(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_features_flattened */
extern void T74f278(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.process_class */
extern void T456f1050(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.internal_process_class */
extern void T456f1231(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.report_class_processed */
extern void T375f111(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.report_class_processed */
extern void T64f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.reset_preparsed */
extern void T74f277(GE_context* ac, T0* C);
/* [detachable] KL_STRING_INPUT_STREAM.make */
extern T0* T1477c9(GE_context* ac, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY_CLASSIC_CONSUMER.consume_class */
extern void T467f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gaaaa_error */
extern void T53f338(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_assembly_error */
extern void T53f365(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY_ERROR.make_gaaaa */
extern T0* T1942c14(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.full_lower_name */
extern T0* T211f98(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.relative_lower_name */
extern T0* T211f102(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.string_ */
extern T0* T211f104(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.full_lower_name */
extern T0* T596f98(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_DOTNET_ASSEMBLY.relative_lower_name */
extern T0* T596f100(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_DOTNET_ASSEMBLY.string_ */
extern T0* T596f101(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gaaaa_error */
extern T1 T53f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gaaaa_error */
extern void T36f338(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_assembly_error */
extern void T36f365(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_gaaaa_error */
extern T1 T36f72(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.dotnet_assembly */
extern T0* T802f47(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.dotnet_assembly */
extern T0* T887f6(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.dotnet_assembly */
extern T0* T410f8(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY_CLASSIC_CONSUMER.error_handler */
extern T0* T467f2(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_syntax_error */
extern void T74f255(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gcaab_error */
extern void T53f251(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLUSTER_ERROR.make_gcaab */
extern T0* T1940c23(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gcaab_error */
extern T1 T53f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gcaab_error */
extern void T36f251(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_gcaab_error */
extern T1 T36f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvscn1b_error */
extern void T53f250(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_gvscn1b */
extern T0* T785c900(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvscn1_error */
extern T1 T53f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvscn1b_error */
extern void T36f250(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_gvscn1_error */
extern T1 T36f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.syntax_error */
extern T1 T456f735(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_filename */
extern void T74f276(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.parse_file */
extern void T456f1234(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T0* a4);
/* [detachable] ET_CLASS.set_has_utf8_bom */
extern void T74f303(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_EIFFEL_PARSER.yyparse */
extern void T456f1235(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_pop_last_value */
extern void T456f1245(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_push_error_value */
extern void T456f1244(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action */
extern void T456f1243(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_801_1000 */
extern void T456f1271(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_946 */
extern void T456f2233(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.add_counter */
extern void T456f2241(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32].force */
extern void T937f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32].resize */
extern void T937f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32].new_capacity */
extern T6 T937f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32].extendible */
extern T1 T937f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_945 */
extern void T456f2232(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IDENTIFIER].force */
extern void T975f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_IDENTIFIER].fill_with */
extern void T976f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_IDENTIFIER].extend */
extern void T976f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_944 */
extern void T456f2231(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_943 */
extern void T456f2230(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_942 */
extern void T456f2229(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_REAL_CONSTANT].force */
extern void T985f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_REAL_CONSTANT].fill_with */
extern void T987f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_REAL_CONSTANT].extend */
extern void T987f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.set_cast_type */
extern void T1491f43(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REGULAR_REAL_CONSTANT.set_cast_type */
extern void T1490f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_target_type */
extern T0* T465f264(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.ast_factory */
extern T0* T456f756(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_941 */
extern void T456f2228(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.set_sign */
extern void T1491f42(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REGULAR_REAL_CONSTANT.set_sign */
extern void T1490f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_940 */
extern void T456f2227(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_939 */
extern void T456f2226(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_938 */
extern void T456f2225(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_937 */
extern void T456f2224(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_936 */
extern void T456f2223(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_935 */
extern void T456f2222(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INTEGER_CONSTANT].force */
extern void T977f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INTEGER_CONSTANT].fill_with */
extern void T979f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INTEGER_CONSTANT].extend */
extern void T979f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_934 */
extern void T456f2221(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.set_sign */
extern void T1489f65(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OCTAL_INTEGER_CONSTANT.set_sign */
extern void T1488f65(GE_context* ac, T0* C, T0* a1);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.set_sign */
extern void T1487f65(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.set_sign */
extern void T1486f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.set_sign */
extern void T372f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_933 */
extern void T456f2220(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_932 */
extern void T456f2219(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_931 */
extern void T456f2218(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_930 */
extern void T456f2217(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_929 */
extern void T456f2216(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_928 */
extern void T456f2215(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BOOLEAN_CONSTANT].force */
extern void T961f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_BOOLEAN_CONSTANT].fill_with */
extern void T963f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_BOOLEAN_CONSTANT].extend */
extern void T963f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_927 */
extern void T456f2214(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_926 */
extern void T456f2213(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHARACTER_CONSTANT].force */
extern void T967f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHARACTER_CONSTANT].fill_with */
extern void T969f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHARACTER_CONSTANT].extend */
extern void T969f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_925 */
extern void T456f2212(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHARACTER_CONSTANT].aliased_resized_area */
extern T0* T967f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CHARACTER_CONSTANT].aliased_resized_area */
extern T0* T969f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_924 */
extern void T456f2211(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_923 */
extern void T456f2210(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_922 */
extern void T456f2209(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_921 */
extern void T456f2208(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING].force */
extern void T983f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING].fill_with */
extern void T984f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING].extend */
extern void T984f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_VERBATIM_STRING.set_cast_type */
extern void T1485f54(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SPECIAL_MANIFEST_STRING.set_cast_type */
extern void T1484f50(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REGULAR_MANIFEST_STRING.set_cast_type */
extern void T1483f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_920 */
extern void T456f2207(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING].aliased_resized_area */
extern T0* T983f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING].aliased_resized_area */
extern T0* T984f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_919 */
extern void T456f2206(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_918 */
extern void T456f2205(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_917 */
extern void T456f2204(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_916 */
extern void T456f2203(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_915 */
extern void T456f2202(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_914 */
extern void T456f2201(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_913 */
extern void T456f2200(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_912 */
extern void T456f2199(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_911 */
extern void T456f2198(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_910 */
extern void T456f2197(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_909 */
extern void T456f2196(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_908 */
extern void T456f2195(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_907 */
extern void T456f2194(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_906 */
extern void T456f2193(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_905 */
extern void T456f2192(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_904 */
extern void T456f2191(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_903 */
extern void T456f2190(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_902 */
extern void T456f2189(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_901 */
extern void T456f2188(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_900 */
extern void T456f2187(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_899 */
extern void T456f2186(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_898 */
extern void T456f2185(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_897 */
extern void T456f2184(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_896 */
extern void T456f2183(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_895 */
extern void T456f2182(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_894 */
extern void T456f2181(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_893 */
extern void T456f2180(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_892 */
extern void T456f2179(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_891 */
extern void T456f2178(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_890 */
extern void T456f2177(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_889 */
extern void T456f2176(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_888 */
extern void T456f2175(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_887 */
extern void T456f2174(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND].force */
extern void T1019f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND].fill_with */
extern void T1020f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND].extend */
extern void T1020f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND].aliased_resized_area */
extern T0* T1019f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND].aliased_resized_area */
extern T0* T1020f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_agent_typed_open_argument */
extern T0* T465f331(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_AGENT_TYPED_OPEN_ARGUMENT.make */
extern T0* T1559c11(GE_context* ac, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.question_mark_symbol */
extern T0* T73f1536(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.make */
extern T0* T1005c42(GE_context* ac);
/* detachable ET_QUESTION_MARK_SYMBOL.make_leaf */
extern void T1005f45(GE_context* ac, T0* C);
/* detachable ET_AGENT_TYPED_OPEN_ARGUMENT.tokens */
extern T0* T1559f8(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_886 */
extern void T456f2173(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_885 */
extern void T456f2172(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_884 */
extern void T456f2171(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM].force */
extern void T1021f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM].fill_with */
extern void T1023f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM].extend */
extern void T1023f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM].aliased_resized_area */
extern T0* T1021f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM].aliased_resized_area */
extern T0* T1023f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.increment_counter */
extern void T456f2242(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32].replace */
extern void T937f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_agent_argument_operand_comma */
extern T0* T465f330(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_883 */
extern void T456f2170(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_LIST].force */
extern void T1024f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_LIST].fill_with */
extern void T1026f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_LIST].extend */
extern void T1026f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.put_first */
extern void T1025f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM].force */
extern void T1810f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM].fill_with */
extern void T1809f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM].extend */
extern void T1809f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.fixed_array */
extern T0* T1025f6(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM].default_create */
extern T0* T1810c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_882 */
extern void T456f2169(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_LIST].aliased_resized_area */
extern T0* T1024f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_LIST].aliased_resized_area */
extern T0* T1026f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_agent_argument_operands */
extern T0* T465f329(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.make_with_capacity */
extern T0* T1025c12(GE_context* ac, T6 a1);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.make_with_capacity */
extern void T1025f12p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM].make */
extern T0* T1810f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_AGENT_ARGUMENT_OPERAND_ITEM].make_empty */
extern T0* T1809c4(GE_context* ac, T6 a1);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.tokens */
extern T0* T1025f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.counter_value */
extern T6 T456f948(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.last_symbol */
extern T0* T456f950(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_SYMBOL].item */
extern T0* T941f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_881 */
extern void T456f2168(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_880 */
extern void T456f2167(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.add_symbol */
extern void T456f2245(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_SYMBOL].force */
extern void T941f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL].force */
extern void T957f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_SYMBOL].fill_with */
extern void T958f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_SYMBOL].extend */
extern void T958f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_SYMBOL].resize */
extern void T941f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL].aliased_resized_area */
extern T0* T957f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_SYMBOL].aliased_resized_area */
extern T0* T958f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_SYMBOL].new_capacity */
extern T6 T941f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_SYMBOL].extendible */
extern T1 T941f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_879 */
extern void T456f2166(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.remove_counter */
extern void T456f2239(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.remove_symbol */
extern void T456f2244(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_SYMBOL].remove */
extern void T941f10(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable ET_SYMBOL].keep_head */
extern void T958f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_SYMBOL].set_count */
extern void T958f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_878 */
extern void T456f2165(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_877 */
extern void T456f2164(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_876 */
extern void T456f2163(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_TARGET].force */
extern void T1027f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_AGENT_TARGET].fill_with */
extern void T1028f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_AGENT_TARGET].extend */
extern void T1028f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_TARGET].aliased_resized_area */
extern T0* T1027f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_AGENT_TARGET].aliased_resized_area */
extern T0* T1028f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_agent_open_target */
extern T0* T465f328(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_AGENT_OPEN_TARGET.make */
extern T0* T1558c9(GE_context* ac, T0* a1);
/* detachable ET_AGENT_OPEN_TARGET.tokens */
extern T0* T1558f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_875 */
extern void T456f2162(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_874 */
extern void T456f2161(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_873 */
extern void T456f2160(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_872 */
extern void T456f2159(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_agent_identifier_target */
extern T0* T456f1000(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_unqualified_call_expression */
extern T0* T465f353(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_OBJECT_TEST_LIST.index_of_name */
extern T6 T1481f4(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ITERATION_COMPONENT_LIST.index_of_name */
extern T6 T365f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LOCAL_COMMA_VARIABLE.set_used */
extern void T1768f11(GE_context* ac, T0* C, T1 a1);
/* detachable ET_LOCAL_VARIABLE.set_used */
extern void T1206f10(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_FORMAL_COMMA_ARGUMENT.set_used */
extern void T1767f13(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_FORMAL_ARGUMENT.set_used */
extern void T256f12(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_871 */
extern void T456f2158(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.set_start_closure */
extern void T456f2253(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_ITERATION_COMPONENT_LIST].force */
extern void T944f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ITERATION_COMPONENT_LIST].force */
extern void T1644f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ITERATION_COMPONENT_LIST].fill_with */
extern void T1643f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ITERATION_COMPONENT_LIST].extend */
extern void T1643f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_ITERATION_COMPONENT_LIST].resize */
extern void T944f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ITERATION_COMPONENT_LIST].aliased_resized_area */
extern T0* T1644f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ITERATION_COMPONENT_LIST].aliased_resized_area */
extern T0* T1643f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_ITERATION_COMPONENT_LIST].new_capacity */
extern T6 T944f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_ITERATION_COMPONENT_LIST].extendible */
extern T1 T944f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST].force */
extern void T942f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBJECT_TEST_LIST].force */
extern void T1640f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_OBJECT_TEST_LIST].fill_with */
extern void T1639f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_OBJECT_TEST_LIST].extend */
extern void T1639f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST].resize */
extern void T942f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBJECT_TEST_LIST].aliased_resized_area */
extern T0* T1640f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_OBJECT_TEST_LIST].aliased_resized_area */
extern T0* T1639f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST].new_capacity */
extern T6 T942f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST].extendible */
extern T1 T942f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST].force */
extern void T939f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_LIST].force */
extern void T1211f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_LIST].fill_with */
extern void T1213f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_LIST].extend */
extern void T1213f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST].resize */
extern void T939f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_LIST].aliased_resized_area */
extern T0* T1211f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_LIST].aliased_resized_area */
extern T0* T1213f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST].new_capacity */
extern T6 T939f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST].extendible */
extern T1 T939f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST].force */
extern void T938f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_LIST].force */
extern void T1157f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_LIST].fill_with */
extern void T1158f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_LIST].extend */
extern void T1158f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST].resize */
extern void T938f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_LIST].aliased_resized_area */
extern T0* T1157f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_LIST].aliased_resized_area */
extern T0* T1158f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST].new_capacity */
extern T6 T938f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST].extendible */
extern T1 T938f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_870 */
extern void T456f2157(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_869 */
extern void T456f2156(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INLINE_AGENT].force */
extern void T1186f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INLINE_AGENT].fill_with */
extern void T1187f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INLINE_AGENT].extend */
extern void T1187f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INLINE_AGENT].aliased_resized_area */
extern T0* T1186f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_INLINE_AGENT].aliased_resized_area */
extern T0* T1187f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.register_inline_agent */
extern void T456f2264(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.set_end_closure */
extern void T456f2268(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_ITERATION_COMPONENT_LIST].remove */
extern void T944f13(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable ET_ITERATION_COMPONENT_LIST].keep_head */
extern void T1643f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ITERATION_COMPONENT_LIST].set_count */
extern void T1643f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_ITERATION_COMPONENT_LIST].item */
extern T0* T944f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_ITERATION_COMPONENT_LIST].is_empty */
extern T1 T944f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_ITERATION_COMPONENT_LIST].force */
extern void T945f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ITERATION_COMPONENT_LIST].force */
extern void T1645f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ITERATION_COMPONENT_LIST].fill_with */
extern void T1646f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ITERATION_COMPONENT_LIST].extend */
extern void T1646f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_ITERATION_COMPONENT_LIST].resize */
extern void T945f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ITERATION_COMPONENT_LIST].aliased_resized_area */
extern T0* T1645f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ITERATION_COMPONENT_LIST].aliased_resized_area */
extern T0* T1646f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_ITERATION_COMPONENT_LIST].new_capacity */
extern T6 T945f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_ITERATION_COMPONENT_LIST].extendible */
extern T1 T945f5(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ITERATION_COMPONENT_LIST.wipe_out */
extern void T365f10(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ITERATION_COMPONENT].keep_head */
extern void T743f4(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* detachable ET_ITERATION_COMPONENT_LIST.fixed_array */
extern T0* T365f6(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ITERATION_COMPONENT].default_create */
extern T0* T743c3(GE_context* ac);
/* [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST].remove */
extern void T942f13(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable ET_OBJECT_TEST_LIST].keep_head */
extern void T1639f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_OBJECT_TEST_LIST].set_count */
extern void T1639f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST].item */
extern T0* T942f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST].is_empty */
extern T1 T942f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_OBJECT_TEST_LIST].force */
extern void T943f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST_LIST].force */
extern void T1641f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST_LIST].fill_with */
extern void T1642f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST_LIST].extend */
extern void T1642f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_OBJECT_TEST_LIST].resize */
extern void T943f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST_LIST].aliased_resized_area */
extern T0* T1641f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST_LIST].aliased_resized_area */
extern T0* T1642f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_OBJECT_TEST_LIST].new_capacity */
extern T6 T943f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_OBJECT_TEST_LIST].extendible */
extern T1 T943f5(GE_context* ac, T0* C, T6 a1);
/* detachable ET_OBJECT_TEST_LIST.wipe_out */
extern void T1481f10(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_OBJECT_TEST].keep_head */
extern void T738f4(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* detachable ET_OBJECT_TEST_LIST.fixed_array */
extern T0* T1481f6(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_OBJECT_TEST].default_create */
extern T0* T738c3(GE_context* ac);
/* [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST].remove */
extern void T939f12(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_LIST].keep_head */
extern void T1213f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_LIST].set_count */
extern void T1213f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST].item */
extern T0* T939f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST].is_empty */
extern T1 T939f5(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST].remove */
extern void T938f12(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_LIST].keep_head */
extern void T1158f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_LIST].set_count */
extern void T1158f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST].item */
extern T0* T938f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST].is_empty */
extern T1 T938f5(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.set_iteration_components */
extern void T1557f40(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_iteration_components */
extern void T1556f46(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_iteration_components */
extern void T1555f42(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.set_iteration_components */
extern void T1554f41(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_iteration_components */
extern void T1553f47(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_iteration_components */
extern void T1552f43(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ITERATION_COMPONENT_LIST.cloned_iteration_component_list */
extern T0* T365f3(GE_context* ac, T0* C);
/* detachable ET_ITERATION_COMPONENT_LIST.put_last */
extern void T365f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ITERATION_COMPONENT_LIST.make_with_capacity */
extern T0* T365c9(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ITERATION_COMPONENT].make */
extern T0* T743f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ITERATION_COMPONENT].make_empty */
extern T0* T741c5(GE_context* ac, T6 a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.set_object_tests */
extern void T1557f39(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_object_tests */
extern void T1556f45(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_object_tests */
extern void T1555f41(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.set_object_tests */
extern void T1554f40(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_object_tests */
extern void T1553f46(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_object_tests */
extern void T1552f42(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OBJECT_TEST_LIST.cloned_object_test_list */
extern T0* T1481f3(GE_context* ac, T0* C);
/* detachable ET_OBJECT_TEST_LIST.put_last */
extern void T1481f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OBJECT_TEST_LIST.make_with_capacity */
extern T0* T1481c9(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_OBJECT_TEST].make */
extern T0* T738f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_NAMED_OBJECT_TEST].make_empty */
extern T0* T737c5(GE_context* ac, T6 a1);
/* [detachable] ET_AST_FACTORY.new_external_procedure_inline_agent */
extern T0* T465f327(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.set_postconditions */
extern void T1557f46(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.set_alias_clause */
extern void T1557f45(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.set_preconditions */
extern void T1557f44(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.set_agent_keyword */
extern void T1557f43(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.make */
extern T0* T1557c38(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.make_procedure */
extern void T1557f47(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.make_inline_agent */
extern void T1557f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AGENT_IMPLICIT_CURRENT_TARGET.make */
extern T0* T1811c6(GE_context* ac, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.agent_keyword */
extern T0* T73f892(GE_context* ac, T0* C);
/* detachable ET_AGENT_KEYWORD.make */
extern T0* T952c13(GE_context* ac);
/* detachable ET_AGENT_KEYWORD.make_token */
extern void T952f15(GE_context* ac, T0* C, T0* a1);
/* detachable ET_AGENT_KEYWORD.make_leaf */
extern void T952f16(GE_context* ac, T0* C);
/* detachable ET_AGENT_KEYWORD.tokens */
extern T0* T952f7(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.tokens */
extern T0* T1557f15(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_external_language */
extern T0* T465f190(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_868 */
extern void T456f2155(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_867 */
extern void T456f2154(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_once_procedure_inline_agent */
extern T0* T465f326(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_rescue_clause */
extern void T1556f55(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_postconditions */
extern void T1556f54(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_compound */
extern void T1556f53(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_keys */
extern void T1556f52(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_locals */
extern void T1556f51(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_preconditions */
extern void T1556f50(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.set_agent_keyword */
extern void T1556f49(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.make */
extern T0* T1556c44(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.make_inline_agent */
extern void T1556f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_once_compound */
extern T0* T465f187(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_COMPOUND.set_keyword */
extern void T355f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.once_keyword */
extern T0* T73f740(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_once */
extern T0* T413c55(GE_context* ac);
/* [detachable] ET_AST_FACTORY.tokens */
extern T0* T465f367(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_866 */
extern void T456f2153(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_865 */
extern void T456f2152(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_do_procedure_inline_agent */
extern T0* T465f325(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_rescue_clause */
extern void T1555f50(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_postconditions */
extern void T1555f49(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_compound */
extern void T1555f48(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_locals */
extern void T1555f47(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_preconditions */
extern void T1555f46(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.set_agent_keyword */
extern void T1555f45(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.make */
extern T0* T1555c40(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.make_inline_agent */
extern void T1555f51(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.tokens */
extern T0* T1555f16(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_864 */
extern void T456f2151(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_863 */
extern void T456f2150(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_external_function_inline_agent */
extern T0* T465f324(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.set_postconditions */
extern void T1554f47(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.set_alias_clause */
extern void T1554f46(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.set_preconditions */
extern void T1554f45(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.set_agent_keyword */
extern void T1554f44(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.make */
extern T0* T1554c39(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.make_function */
extern void T1554f48(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.make_inline_agent */
extern void T1554f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_RESULT.set_position */
extern void T309f36(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.tokens */
extern T0* T1554f35(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_colon_type */
extern T0* T465f181(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_862 */
extern void T456f2149(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_861 */
extern void T456f2148(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_once_function_inline_agent */
extern T0* T465f323(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_rescue_clause */
extern void T1553f56(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_postconditions */
extern void T1553f55(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_compound */
extern void T1553f54(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_keys */
extern void T1553f53(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_locals */
extern void T1553f52(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_preconditions */
extern void T1553f51(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.set_agent_keyword */
extern void T1553f50(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.make */
extern T0* T1553c45(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.make_inline_agent */
extern void T1553f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_860 */
extern void T456f2147(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_859 */
extern void T456f2146(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_do_function_inline_agent */
extern T0* T465f322(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_rescue_clause */
extern void T1552f51(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_postconditions */
extern void T1552f50(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_compound */
extern void T1552f49(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_locals */
extern void T1552f48(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_preconditions */
extern void T1552f47(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.set_agent_keyword */
extern void T1552f46(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.make */
extern T0* T1552c41(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.make_inline_agent */
extern void T1552f52(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.tokens */
extern T0* T1552f37(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_858 */
extern void T456f2145(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_857 */
extern void T456f2144(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.set_inline_agent_actual_arguments */
extern void T456f2263(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_856 */
extern void T456f2143(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CALL_AGENT].force */
extern void T1040f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CALL_AGENT].fill_with */
extern void T1042f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CALL_AGENT].extend */
extern void T1042f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CALL_AGENT].aliased_resized_area */
extern T0* T1040f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CALL_AGENT].aliased_resized_area */
extern T0* T1042f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_call_agent */
extern T0* T465f320(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_CALL_AGENT.set_agent_keyword */
extern void T1041f33(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CALL_AGENT.make */
extern T0* T1041c32(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_CALL_AGENT.tokens */
extern T0* T1041f9(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.make_unqualified */
extern T0* T1041c30(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_dot_feature_name */
extern T0* T465f321(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_855 */
extern void T456f2142(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_854 */
extern void T456f2141(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION].force */
extern void T1129f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXPRESSION].fill_with */
extern void T1130f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXPRESSION].extend */
extern void T1130f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION].aliased_resized_area */
extern T0* T1129f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_EXPRESSION].aliased_resized_area */
extern T0* T1130f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_iteration_cursor */
extern T0* T456f999(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_iteration_cursor */
extern T0* T465f366(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ITERATION_CURSOR.set_at_symbol */
extern void T1551f23(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ITERATION_CURSOR.make */
extern T0* T1551c21(GE_context* ac, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.at_symbol */
extern T0* T73f1535(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_at */
extern T0* T341c101(GE_context* ac);
/* detachable ET_ITERATION_CURSOR.tokens */
extern T0* T1551f6(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_prefix_expression */
extern T0* T465f309(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_PREFIX_EXPRESSION.make */
extern T0* T1547c26(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_prefix_free_operator_from_symbol */
extern T0* T465f319(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FREE_OPERATOR.set_break */
extern void T973f58(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.break */
extern T0* T994f1(GE_context* ac, T0* C);
/* [detachable] ET_BRACKET_SYMBOL.break */
extern T0* T622f3(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.break */
extern T0* T341f3(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.set_position */
extern void T973f55(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_FREE_OPERATOR.make_prefix */
extern T0* T973c54(GE_context* ac, T0* a1);
/* [detachable] KL_STRING_ROUTINES.case_insensitive_hash_code */
extern T6 T82f7(GE_context* ac, T0* C, T0* a1);
/* CHARACTER_32.upper */
extern T3 T3f6(GE_context* ac, T3* C);
/* [detachable] CHARACTER_PROPERTY.to_upper */
extern T3 T780f2(GE_context* ac, T0* C, T3 a1);
/* [detachable] CHARACTER_PROPERTY.to_upper_table_17 */
extern T0* T780f36(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_upper_table_16 */
extern T0* T780f35(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_upper_table_15 */
extern T0* T780f34(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_upper_table_14 */
extern T0* T780f33(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_upper_table_13 */
extern T0* T780f32(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_upper_table_12 */
extern T0* T780f31(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_upper_table_11 */
extern T0* T780f30(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_upper_table_10 */
extern T0* T780f29(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_upper_table_9 */
extern T0* T780f28(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_upper_table_8 */
extern T0* T780f27(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_upper_table_7 */
extern T0* T780f26(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_upper_table_6 */
extern T0* T780f25(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_upper_table_5 */
extern T0* T780f24(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_upper_table_4 */
extern T0* T780f23(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_upper_table_3 */
extern T0* T780f22(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_upper_table_2 */
extern T0* T780f21(GE_context* ac, T0* C);
/* [detachable] CHARACTER_PROPERTY.to_upper_table_1 */
extern T0* T780f20(GE_context* ac, T0* C);
/* CHARACTER_32.minus */
extern T3 T3f14(GE_context* ac, T3* C, T10 a1);
/* detachable ET_FREE_OPERATOR.string_ */
extern T0* T973f10(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.make_token */
extern void T973f60(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FREE_OPERATOR.make_leaf */
extern void T973f61(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.use_obsolete_syntax_mode */
extern T1 T187f101(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.use_obsolete_syntax_mode */
extern T1 T58f128(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.use_obsolete_syntax_mode */
extern T1 T40f104(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.current_universe */
extern T0* T456f754(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_853 */
extern void T456f2140(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUANTIFIER_EXPRESSION].force */
extern void T1253f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_QUANTIFIER_EXPRESSION].fill_with */
extern void T1255f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_QUANTIFIER_EXPRESSION].extend */
extern void T1255f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUANTIFIER_EXPRESSION].aliased_resized_area */
extern T0* T1253f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_QUANTIFIER_EXPRESSION].aliased_resized_area */
extern T0* T1255f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_there_exists_quantifier_expression_header */
extern T0* T456f998(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_ITERATION_COMPONENT_LIST.force_last */
extern void T365f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ITERATION_COMPONENT_LIST.resize */
extern void T365f13(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ITERATION_COMPONENT_LIST.new_capacity */
extern T6 T365f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ITERATION_COMPONENT_LIST.capacity */
extern T6 T365f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_iteration_component_list */
extern T0* T456f1015(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_ITERATION_COMPONENT_LIST].remove */
extern void T945f12(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_ITERATION_COMPONENT_LIST].keep_head */
extern void T1646f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ITERATION_COMPONENT_LIST].set_count */
extern void T1646f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_ITERATION_COMPONENT_LIST].item */
extern T0* T945f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_ITERATION_COMPONENT_LIST].is_empty */
extern T1 T945f7(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_there_exists_quantifier_expression */
extern T0* T465f365(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_QUANTIFIER_EXPRESSION.make_there_exists */
extern T0* T1254c39(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_QUANTIFIER_EXPRESSION.create_unfolded_form */
extern void T1254f40(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.make */
extern T0* T284c9(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.make_qualified_call */
extern void T284f10(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.bar_symbol */
extern T0* T73f697(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_bar */
extern T0* T341c113(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.colon_symbol */
extern T0* T73f692(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_colon */
extern T0* T341c108(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.there_exists_symbol */
extern T0* T73f696(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_there_exists */
extern T0* T341c112(GE_context* ac);
/* detachable ET_QUANTIFIER_EXPRESSION.tokens */
extern T0* T1254f16(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.true_keyword */
extern T0* T73f266(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.make */
extern T0* T1493c39(GE_context* ac);
/* detachable ET_TRUE_CONSTANT.make_token */
extern void T1493f42(GE_context* ac, T0* C, T0* a1);
/* detachable ET_TRUE_CONSTANT.make_leaf */
extern void T1493f43(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.tokens */
extern T0* T456f685(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_852 */
extern void T456f2139(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_for_all_quantifier_expression_header */
extern T0* T456f997(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_AST_FACTORY.new_for_all_quantifier_expression */
extern T0* T465f364(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_QUANTIFIER_EXPRESSION.make_for_all */
extern T0* T1254c36(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_851 */
extern void T456f2138(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_quantifier_expression */
extern T0* T456f996(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_QUANTIFIER_EXPRESSION.set_iteration_expression */
extern void T1254f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_850 */
extern void T456f2137(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_EXPRESSION].force */
extern void T1007f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ACROSS_EXPRESSION].fill_with */
extern void T1009f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ACROSS_EXPRESSION].extend */
extern void T1009f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_EXPRESSION].aliased_resized_area */
extern T0* T1007f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ACROSS_EXPRESSION].aliased_resized_area */
extern T0* T1009f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_across_expression_header */
extern T0* T456f995(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ACROSS_EXPRESSION.set_has_cursor_name */
extern void T1008f46(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_KEYWORD.is_as */
extern T1 T413f30(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.obsolete_iteration_mode */
extern T1 T187f98(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.obsolete_iteration_mode */
extern T1 T58f129(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.obsolete_iteration_mode */
extern T1 T40f105(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_across_all_expression */
extern T0* T465f363(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9);
/* detachable ET_ACROSS_EXPRESSION.set_as_keyword */
extern void T1008f47(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_EXPRESSION.set_variant_part */
extern void T1008f43(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_EXPRESSION.set_invariant_part */
extern void T1008f42(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_EXPRESSION.make_all */
extern T0* T1008c37(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ACROSS_EXPRESSION.create_unfolded_form */
extern void T1008f48(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.across_keyword */
extern T0* T73f721(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_across */
extern T0* T413c82(GE_context* ac);
/* detachable ET_ACROSS_EXPRESSION.tokens */
extern T0* T1008f18(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_849 */
extern void T456f2136(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_848 */
extern void T456f2135(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_across_all_expression */
extern T0* T456f994(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_ACROSS_EXPRESSION.set_end_keyword */
extern void T1008f44(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_EXPRESSION.set_all */
extern void T1008f45(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ACROSS_EXPRESSION.set_iteration_conditional */
extern void T1008f40(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_EXPRESSION.set_until_conditional */
extern void T1008f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_conditional */
extern T0* T465f258(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_KEYWORD_EXPRESSION.make */
extern T0* T1771c4(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_KEYWORD.is_inspect */
extern T1 T413f29(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.is_if */
extern T1 T413f28(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_847 */
extern void T456f2134(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_across_some_expression */
extern T0* T456f993(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_ACROSS_EXPRESSION.set_some */
extern void T1008f41(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_846 */
extern void T456f2133(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTANT].force */
extern void T1072f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTANT].fill_with */
extern void T1073f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTANT].extend */
extern void T1073f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTANT].aliased_resized_area */
extern T0* T1072f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CONSTANT].aliased_resized_area */
extern T0* T1073f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_845 */
extern void T456f2132(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_844 */
extern void T456f2131(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_843 */
extern void T456f2130(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_842 */
extern void T456f2129(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_841 */
extern void T456f2128(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_STRIP_EXPRESSION].force */
extern void T1269f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_STRIP_EXPRESSION].fill_with */
extern void T1271f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_STRIP_EXPRESSION].extend */
extern void T1271f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_STRIP_EXPRESSION.put_first */
extern void T1270f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME_ITEM].force */
extern void T1783f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FEATURE_NAME_ITEM].fill_with */
extern void T1782f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FEATURE_NAME_ITEM].extend */
extern void T1782f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_STRIP_EXPRESSION.fixed_array */
extern T0* T1270f13(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME_ITEM].default_create */
extern T0* T1783c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_840 */
extern void T456f2127(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_STRIP_EXPRESSION].aliased_resized_area */
extern T0* T1269f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_STRIP_EXPRESSION].aliased_resized_area */
extern T0* T1271f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_strip_expression */
extern T0* T465f318(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4);
/* detachable ET_STRIP_EXPRESSION.set_strip_keyword */
extern void T1270f31(GE_context* ac, T0* C, T0* a1);
/* detachable ET_STRIP_EXPRESSION.make_with_capacity */
extern T0* T1270c28(GE_context* ac, T6 a1);
/* detachable ET_STRIP_EXPRESSION.make_with_capacity */
extern void T1270f28p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_NAME_ITEM].make */
extern T0* T1783f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_FEATURE_NAME_ITEM].make_empty */
extern T0* T1782c4(GE_context* ac, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.strip_keyword */
extern T0* T73f897(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_strip */
extern T0* T413c59(GE_context* ac);
/* detachable ET_STRIP_EXPRESSION.tokens */
extern T0* T1270f11(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.last_keyword */
extern T0* T456f947(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_KEYWORD].item */
extern T0* T940f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_839 */
extern void T456f2126(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_838 */
extern void T456f2125(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.add_keyword */
extern void T456f2240(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_KEYWORD].force */
extern void T940f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD].force */
extern void T949f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_KEYWORD].fill_with */
extern void T950f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_KEYWORD].extend */
extern void T950f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_KEYWORD].resize */
extern void T940f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD].aliased_resized_area */
extern T0* T949f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_KEYWORD].aliased_resized_area */
extern T0* T950f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_KEYWORD].new_capacity */
extern T6 T940f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_KEYWORD].extendible */
extern T1 T940f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_837 */
extern void T456f2124(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.remove_keyword */
extern void T456f2238(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_KEYWORD].remove */
extern void T940f10(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable ET_KEYWORD].keep_head */
extern void T950f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_KEYWORD].set_count */
extern void T950f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_836 */
extern void T456f2123(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_835 */
extern void T456f2122(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_TUPLE].force */
extern void T1226f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_MANIFEST_TUPLE].fill_with */
extern void T1227f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_MANIFEST_TUPLE].extend */
extern void T1227f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_834 */
extern void T456f2121(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_TUPLE].aliased_resized_area */
extern T0* T1226f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_MANIFEST_TUPLE].aliased_resized_area */
extern T0* T1227f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_manifest_tuple */
extern T0* T465f317(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] ET_MANIFEST_TUPLE.set_left_symbol */
extern void T288f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_833 */
extern void T456f2120(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_832 */
extern void T456f2119(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_831 */
extern void T456f2118(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_830 */
extern void T456f2117(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_829 */
extern void T456f2116(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_ARRAY].force */
extern void T1217f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_MANIFEST_ARRAY].fill_with */
extern void T1219f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_MANIFEST_ARRAY].extend */
extern void T1219f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_MANIFEST_ARRAY.put_first */
extern void T1218f31(GE_context* ac, T0* C, T0* a1);
/* detachable ET_MANIFEST_ARRAY.fixed_array */
extern T0* T1218f13(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_828 */
extern void T456f2115(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_ARRAY].aliased_resized_area */
extern T0* T1217f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_MANIFEST_ARRAY].aliased_resized_area */
extern T0* T1219f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_manifest_array */
extern T0* T465f316(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_MANIFEST_ARRAY.set_left_symbol */
extern void T1218f32(GE_context* ac, T0* C, T0* a1);
/* detachable ET_MANIFEST_ARRAY.make_with_capacity */
extern T0* T1218c28(GE_context* ac, T6 a1);
/* detachable ET_MANIFEST_ARRAY.make_with_capacity */
extern void T1218f28p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.right_array_symbol */
extern T0* T73f700(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_right_array */
extern T0* T341c116(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.left_array_symbol */
extern T0* T73f895(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_left_array */
extern T0* T341c104(GE_context* ac);
/* detachable ET_MANIFEST_ARRAY.tokens */
extern T0* T1218f11(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_827 */
extern void T456f2114(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_826 */
extern void T456f2113(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.set_cast_type */
extern void T1218f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_825 */
extern void T456f2112(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_824 */
extern void T456f2111(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_823 */
extern void T456f2110(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_822 */
extern void T456f2109(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENTHESIZED_EXPRESSION].force */
extern void T1231f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PARENTHESIZED_EXPRESSION].fill_with */
extern void T1233f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PARENTHESIZED_EXPRESSION].extend */
extern void T1233f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENTHESIZED_EXPRESSION].aliased_resized_area */
extern T0* T1231f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_PARENTHESIZED_EXPRESSION].aliased_resized_area */
extern T0* T1233f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_parenthesized_expression */
extern T0* T465f315(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_PARENTHESIZED_EXPRESSION.set_left_parenthesis */
extern void T1232f24(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENTHESIZED_EXPRESSION.make */
extern T0* T1232c22(GE_context* ac, T0* a1);
/* detachable ET_PARENTHESIZED_EXPRESSION.tokens */
extern T0* T1232f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_821 */
extern void T456f2108(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_ARGUMENT_LIST].force */
extern void T1013f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ACTUAL_ARGUMENT_LIST].fill_with */
extern void T1014f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ACTUAL_ARGUMENT_LIST].extend */
extern void T1014f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.put_first */
extern void T283f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_820 */
extern void T456f2107(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_ARGUMENT_LIST].aliased_resized_area */
extern T0* T1013f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ACTUAL_ARGUMENT_LIST].aliased_resized_area */
extern T0* T1014f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_bracket_arguments */
extern T0* T465f314(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.make_bracketed_with_capacity */
extern T0* T283c19(GE_context* ac, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_819 */
extern void T456f2106(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_818 */
extern void T456f2105(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_EXPRESSION].force */
extern void T1037f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_BRACKET_EXPRESSION].fill_with */
extern void T1039f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_BRACKET_EXPRESSION].extend */
extern void T1039f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_EXPRESSION].aliased_resized_area */
extern T0* T1037f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_BRACKET_EXPRESSION].aliased_resized_area */
extern T0* T1039f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_817 */
extern void T456f2104(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_bracket_expression */
extern T0* T465f313(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_BRACKET_EXPRESSION.make */
extern T0* T1038c25(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_816 */
extern void T456f2103(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_815 */
extern void T456f2102(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_814 */
extern void T456f2101(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_813 */
extern void T456f2100(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_812 */
extern void T456f2099(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_811 */
extern void T456f2098(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_810 */
extern void T456f2097(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_809 */
extern void T456f2096(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_808 */
extern void T456f2095(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_807 */
extern void T456f2094(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.raise_error */
extern void T456f2235(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.ise_17_11_0 */
extern T0* T456f992(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_806 */
extern void T456f2093(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_805 */
extern void T456f2092(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_804 */
extern void T456f2091(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_803 */
extern void T456f2090(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_802 */
extern void T456f2089(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_801 */
extern void T456f2088(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_once_manifest_string */
extern T0* T456f991(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.register_inline_constant */
extern void T74f319(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_MANIFEST_STRING.set_id */
extern void T1550f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_once_manifest_string */
extern T0* T465f362(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ONCE_MANIFEST_STRING.set_once_keyword */
extern void T1550f26(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_MANIFEST_STRING.make */
extern T0* T1550c24(GE_context* ac, T0* a1);
/* detachable ET_ONCE_MANIFEST_STRING.tokens */
extern T0* T1550f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_601_800 */
extern void T456f1270(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_800 */
extern void T456f2087(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_799 */
extern void T456f2086(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_798 */
extern void T456f2085(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_797 */
extern void T456f2084(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_796 */
extern void T456f2083(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_795 */
extern void T456f2082(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_794 */
extern void T456f2081(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_793 */
extern void T456f2080(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_792 */
extern void T456f2079(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_791 */
extern void T456f2078(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_790 */
extern void T456f2077(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_789 */
extern void T456f2076(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_788 */
extern void T456f2075(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_787 */
extern void T456f2074(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_named_object_test */
extern T0* T456f990(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_OBJECT_TEST_LIST.force_last */
extern void T1481f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OBJECT_TEST_LIST.resize */
extern void T1481f13(GE_context* ac, T0* C, T6 a1);
/* detachable ET_OBJECT_TEST_LIST.new_capacity */
extern T6 T1481f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_OBJECT_TEST_LIST.capacity */
extern T6 T1481f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_object_test_list */
extern T0* T456f1016(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_OBJECT_TEST_LIST].remove */
extern void T943f12(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST_LIST].keep_head */
extern void T1642f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST_LIST].set_count */
extern void T1642f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_OBJECT_TEST_LIST].item */
extern T0* T943f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_OBJECT_TEST_LIST].is_empty */
extern T1 T943f7(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_named_object_test */
extern T0* T465f361(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NAMED_OBJECT_TEST.set_attached_keyword */
extern void T736f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NAMED_OBJECT_TEST.make */
extern T0* T736c26(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NAMED_OBJECT_TEST.make_object_test */
extern void T736f29(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NAMED_OBJECT_TEST.tokens */
extern T0* T736f10(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_786 */
extern void T456f2073(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_785 */
extern void T456f2072(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_object_test */
extern T0* T465f312(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_OBJECT_TEST.set_attached_keyword */
extern void T311f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_OBJECT_TEST.make */
extern T0* T311c24(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_OBJECT_TEST.tokens */
extern T0* T311f10(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_784 */
extern void T456f2071(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_783 */
extern void T456f2070(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_old_object_test */
extern T0* T456f989(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_AST_FACTORY.new_old_object_test */
extern T0* T465f360(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_OLD_OBJECT_TEST.set_left_brace */
extern void T1549f31(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OLD_OBJECT_TEST.make */
extern T0* T1549c29(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_OLD_OBJECT_TEST.tokens */
extern T0* T1549f12(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.ise_6_1_0 */
extern T0* T456f975(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_782 */
extern void T456f2069(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_old_expression */
extern T0* T465f311(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_OLD_EXPRESSION.set_old_keyword */
extern void T1548f23(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OLD_EXPRESSION.make */
extern T0* T1548c21(GE_context* ac, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.old_keyword */
extern T0* T73f1534(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_old */
extern T0* T413c58(GE_context* ac);
/* detachable ET_OLD_EXPRESSION.tokens */
extern T0* T1548f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_781 */
extern void T456f2068(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_prefix_free_operator */
extern T0* T465f310(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FREE_OPERATOR.set_prefix */
extern void T973f59(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_780 */
extern void T456f2067(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_779 */
extern void T456f2066(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_778 */
extern void T456f2065(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_prefix_minus_expression */
extern T0* T456f988(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T456f988ot1(T0* a1, T0** a2);
extern T1 T456f988ot2(T0* a1, T0** a2);
/* [detachable] ET_AST_FACTORY.new_prefix_minus_operator */
extern T0* T465f359(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.set_prefix_minus */
extern void T994f154(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_777 */
extern void T456f2064(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_prefix_plus_expression */
extern T0* T456f987(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T456f987ot1(T0* a1, T0** a2);
extern T1 T456f987ot2(T0* a1, T0** a2);
/* [detachable] ET_AST_FACTORY.new_prefix_plus_operator */
extern T0* T465f358(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.set_prefix_plus */
extern void T994f153(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_776 */
extern void T456f2063(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_775 */
extern void T456f2062(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_774 */
extern void T456f2061(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_773 */
extern void T456f2060(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_772 */
extern void T456f2059(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_771 */
extern void T456f2058(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_770 */
extern void T456f2057(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_769 */
extern void T456f2056(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_768 */
extern void T456f2055(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_767 */
extern void T456f2054(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_766 */
extern void T456f2053(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_765 */
extern void T456f2052(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_manifest_type */
extern T0* T465f308(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_MANIFEST_TYPE.make */
extern T0* T1546c22(GE_context* ac, T0* a1);
/* detachable ET_MANIFEST_TYPE.tokens */
extern T0* T1546f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_764 */
extern void T456f2051(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_763 */
extern void T456f2050(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_762 */
extern void T456f2049(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_761 */
extern void T456f2048(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_760 */
extern void T456f2047(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_759 */
extern void T456f2046(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_object_equality_expression */
extern T0* T465f307(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.make */
extern T0* T312c23(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_758 */
extern void T456f2045(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_757 */
extern void T456f2044(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_equality_expression */
extern T0* T465f306(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EQUALITY_EXPRESSION.make */
extern T0* T313c21(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_756 */
extern void T456f2043(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_755 */
extern void T456f2042(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_expression */
extern T0* T465f300(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_INFIX_EXPRESSION.make */
extern T0* T1543c26(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_754 */
extern void T456f2041(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_753 */
extern void T456f2040(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_752 */
extern void T456f2039(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_or_else_operator */
extern T0* T465f305(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.set_or_keyword */
extern void T1545f43(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.make */
extern T0* T1545c42(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.else_keyword */
extern T0* T73f709(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_else */
extern T0* T413c70(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.or_keyword */
extern T0* T73f1533(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.make_or */
extern T0* T981c55(GE_context* ac);
/* detachable ET_KEYWORD_OPERATOR.make_token */
extern void T981f61(GE_context* ac, T0* C, T0* a1);
/* detachable ET_KEYWORD_OPERATOR.make_leaf */
extern void T981f62(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_751 */
extern void T456f2038(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_750 */
extern void T456f2037(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_and_then_operator */
extern T0* T465f304(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_INFIX_AND_THEN_OPERATOR.set_and_keyword */
extern void T1544f43(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INFIX_AND_THEN_OPERATOR.make */
extern T0* T1544c42(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.then_keyword */
extern T0* T73f714(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_then */
extern T0* T413c75(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.and_keyword */
extern T0* T73f1532(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.make_and */
extern T0* T981c57(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_749 */
extern void T456f2036(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_748 */
extern void T456f2035(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_747 */
extern void T456f2034(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_746 */
extern void T456f2033(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_745 */
extern void T456f2032(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_744 */
extern void T456f2031(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_743 */
extern void T456f2030(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_742 */
extern void T456f2029(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_741 */
extern void T456f2028(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_740 */
extern void T456f2027(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_739 */
extern void T456f2026(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_738 */
extern void T456f2025(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_737 */
extern void T456f2024(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_736 */
extern void T456f2023(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_735 */
extern void T456f2022(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_734 */
extern void T456f2021(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_minus_operator */
extern T0* T465f303(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.set_infix_minus */
extern void T994f152(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_733 */
extern void T456f2020(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_plus_operator */
extern T0* T465f302(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.set_infix_plus */
extern void T994f151(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_732 */
extern void T456f2019(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_free_operator_from_symbol */
extern T0* T465f301(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FREE_OPERATOR.make_infix */
extern T0* T973c57(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_731 */
extern void T456f2018(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_infix_free_operator */
extern T0* T465f299(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FREE_OPERATOR.set_infix */
extern void T973f56(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_730 */
extern void T456f2017(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_729 */
extern void T456f2016(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_728 */
extern void T456f2015(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WRITABLE].force */
extern void T1304f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_WRITABLE].fill_with */
extern void T1305f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_WRITABLE].extend */
extern void T1305f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WRITABLE].aliased_resized_area */
extern T0* T1304f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_WRITABLE].aliased_resized_area */
extern T0* T1305f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_727 */
extern void T456f2014(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_writable */
extern T0* T456f986(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_726 */
extern void T456f2013(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_expression_address */
extern T0* T465f298(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_EXPRESSION_ADDRESS.set_dollar */
extern void T1542f23(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXPRESSION_ADDRESS.make */
extern T0* T1542c21(GE_context* ac, T0* a1);
/* detachable ET_EXPRESSION_ADDRESS.tokens */
extern T0* T1542f6(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_725 */
extern void T456f2012(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_result_address */
extern T0* T465f297(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_RESULT_ADDRESS.set_dollar */
extern void T1541f23(GE_context* ac, T0* C, T0* a1);
/* detachable ET_RESULT_ADDRESS.make */
extern T0* T1541c21(GE_context* ac);
/* detachable ET_RESULT_ADDRESS.tokens */
extern T0* T1541f6(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_724 */
extern void T456f2011(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_current_address */
extern T0* T465f296(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CURRENT_ADDRESS.set_dollar */
extern void T1540f23(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CURRENT_ADDRESS.make */
extern T0* T1540c21(GE_context* ac);
/* detachable ET_CURRENT_ADDRESS.tokens */
extern T0* T1540f6(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_723 */
extern void T456f2010(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_feature_address */
extern T0* T456f985(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_feature_address */
extern T0* T465f357(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_FEATURE_ADDRESS.set_dollar */
extern void T1539f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_722 */
extern void T456f2009(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION_ITEM].force */
extern void T1131f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXPRESSION_ITEM].fill_with */
extern void T1132f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXPRESSION_ITEM].extend */
extern void T1132f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION_ITEM].aliased_resized_area */
extern T0* T1131f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_EXPRESSION_ITEM].aliased_resized_area */
extern T0* T1132f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_expression_comma */
extern T0* T465f295(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_721 */
extern void T456f2008(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_720 */
extern void T456f2007(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_actual_arguments */
extern T0* T465f294(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_719 */
extern void T456f2006(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_718 */
extern void T456f2005(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_717 */
extern void T456f2004(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_716 */
extern void T456f2003(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_715 */
extern void T456f2002(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_714 */
extern void T456f2001(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_qualified_call_expression */
extern T0* T465f291(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.new_dot_feature_name */
extern T0* T456f974(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_713 */
extern void T456f2000(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.ise_5_7_59914 */
extern T0* T456f959(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_712 */
extern void T456f1999(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_711 */
extern void T456f1998(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_710 */
extern void T456f1997(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_709 */
extern void T456f1996(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_708 */
extern void T456f1995(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_707 */
extern void T456f1994(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_706 */
extern void T456f1993(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_705 */
extern void T456f1992(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_704 */
extern void T456f1991(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_unqualified_call_expression */
extern T0* T456f984(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_703 */
extern void T456f1990(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_precursor_expression */
extern T0* T465f293(GE_context* ac, T0* C, T1 a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_PRECURSOR_EXPRESSION.set_parent_prefixed */
extern void T1538f27(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_AST_FACTORY.new_precursor_class_name */
extern T0* T465f289(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_702 */
extern void T456f1989(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_701 */
extern void T456f1988(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_STATIC_CALL_EXPRESSION].force */
extern void T1266f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_STATIC_CALL_EXPRESSION].fill_with */
extern void T1268f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_STATIC_CALL_EXPRESSION].extend */
extern void T1268f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_STATIC_CALL_EXPRESSION].aliased_resized_area */
extern T0* T1266f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_STATIC_CALL_EXPRESSION].aliased_resized_area */
extern T0* T1268f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_static_call_expression */
extern T0* T465f292(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_STATIC_CALL_EXPRESSION.set_feature_keyword */
extern void T1267f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_700 */
extern void T456f1987(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_699 */
extern void T456f1986(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_CALL_EXPRESSION].force */
extern void T1043f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_CALL_EXPRESSION].fill_with */
extern void T1044f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_CALL_EXPRESSION].extend */
extern void T1044f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_CALL_EXPRESSION].aliased_resized_area */
extern T0* T1043f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_CALL_EXPRESSION].aliased_resized_area */
extern T0* T1044f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_698 */
extern void T456f1985(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_697 */
extern void T456f1984(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_696 */
extern void T456f1983(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_695 */
extern void T456f1982(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_694 */
extern void T456f1981(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSTRUCTION].force */
extern void T1194f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INSTRUCTION].fill_with */
extern void T1196f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INSTRUCTION].extend */
extern void T1196f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSTRUCTION].aliased_resized_area */
extern T0* T1194f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_INSTRUCTION].aliased_resized_area */
extern T0* T1196f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_static_call_instruction */
extern T0* T465f290(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_STATIC_CALL_INSTRUCTION.set_feature_keyword */
extern void T1537f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_STATIC_CALL_INSTRUCTION.make */
extern T0* T1537c9(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_STATIC_CALL_INSTRUCTION.make_qualified_call */
extern void T1537f11(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_693 */
extern void T456f1980(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_692 */
extern void T456f1979(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_precursor_instruction */
extern T0* T465f288(GE_context* ac, T0* C, T1 a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_PRECURSOR_INSTRUCTION.set_precursor_keyword */
extern void T1535f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECURSOR_INSTRUCTION.set_parent_prefixed */
extern void T1535f10(GE_context* ac, T0* C, T1 a1);
/* detachable ET_PRECURSOR_INSTRUCTION.make */
extern T0* T1535c9(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_PRECURSOR_INSTRUCTION.tokens */
extern T0* T1535f8(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_691 */
extern void T456f1978(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_690 */
extern void T456f1977(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_qualified_call_instruction */
extern T0* T465f287(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_689 */
extern void T456f1976(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_688 */
extern void T456f1975(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_unqualified_call_instruction */
extern T0* T456f983(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_unqualified_call_instruction */
extern T0* T465f356(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION.make */
extern T0* T1756c7(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_687 */
extern void T456f1974(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_ITEM].force */
extern void T1220f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_ITEM].fill_with */
extern void T1222f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_ITEM].extend */
extern void T1222f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_ITEM].aliased_resized_area */
extern T0* T1220f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_ITEM].aliased_resized_area */
extern T0* T1222f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_manifest_string_comma */
extern T0* T465f286(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_686 */
extern void T456f1973(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_LIST].force */
extern void T1223f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_LIST].fill_with */
extern void T1225f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_LIST].extend */
extern void T1225f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_MANIFEST_STRING_LIST.put_first */
extern void T1224f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MANIFEST_STRING_ITEM].force */
extern void T1839f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_MANIFEST_STRING_ITEM].fill_with */
extern void T1838f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_MANIFEST_STRING_ITEM].extend */
extern void T1838f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_MANIFEST_STRING_LIST.fixed_array */
extern T0* T1224f6(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MANIFEST_STRING_ITEM].default_create */
extern T0* T1839c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_685 */
extern void T456f1972(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_LIST].aliased_resized_area */
extern T0* T1223f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_LIST].aliased_resized_area */
extern T0* T1225f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_manifest_string_list */
extern T0* T465f285(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_MANIFEST_STRING_LIST.make_with_capacity */
extern T0* T1224c11(GE_context* ac, T6 a1);
/* detachable ET_MANIFEST_STRING_LIST.tokens */
extern T0* T1224f5(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_STRING_LIST.make_with_capacity */
extern void T1224f11p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_MANIFEST_STRING_ITEM].make */
extern T0* T1839f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_MANIFEST_STRING_ITEM].make_empty */
extern T0* T1838c4(GE_context* ac, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_684 */
extern void T456f1971(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_683 */
extern void T456f1970(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_682 */
extern void T456f1969(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_681 */
extern void T456f1968(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_680 */
extern void T456f1967(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DEBUG_INSTRUCTION].force */
extern void T1109f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_DEBUG_INSTRUCTION].fill_with */
extern void T1111f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_DEBUG_INSTRUCTION].extend */
extern void T1111f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DEBUG_INSTRUCTION].aliased_resized_area */
extern T0* T1109f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_DEBUG_INSTRUCTION].aliased_resized_area */
extern T0* T1111f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_debug_instruction */
extern T0* T465f284(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_DEBUG_INSTRUCTION.make */
extern T0* T1110c6(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_DEBUG_INSTRUCTION.tokens */
extern T0* T1110f4(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_debug_compound */
extern T0* T465f283(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.debug_keyword */
extern T0* T73f718(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_debug */
extern T0* T413c79(GE_context* ac);
/* detachable ET_COMPOUND.make */
extern T0* T355c11(GE_context* ac);
/* detachable ET_COMPOUND.make */
extern void T355f11p1(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INSTRUCTION].make */
extern T0* T1968f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_INSTRUCTION].make_empty */
extern T0* T1966c4(GE_context* ac, T6 a1);
/* detachable ET_COMPOUND.fixed_array */
extern T0* T355f7(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INSTRUCTION].default_create */
extern T0* T1968c2(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.do_keyword */
extern T0* T73f706(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_do */
extern T0* T413c67(GE_context* ac);
/* detachable ET_COMPOUND.tokens */
extern T0* T355f6(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_679 */
extern void T456f1966(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONDITIONAL].force */
extern void T1069f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONDITIONAL].fill_with */
extern void T1071f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONDITIONAL].extend */
extern void T1071f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONDITIONAL].aliased_resized_area */
extern T0* T1069f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CONDITIONAL].aliased_resized_area */
extern T0* T1071f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_678 */
extern void T456f1965(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_677 */
extern void T456f1964(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_REPEAT_INSTRUCTION].force */
extern void T1263f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_REPEAT_INSTRUCTION].fill_with */
extern void T1265f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_REPEAT_INSTRUCTION].extend */
extern void T1265f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_REPEAT_INSTRUCTION].aliased_resized_area */
extern T0* T1263f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_REPEAT_INSTRUCTION].aliased_resized_area */
extern T0* T1265f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_repeat_instruction_header */
extern T0* T456f982(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_AST_FACTORY.new_repeat_instruction */
extern T0* T465f355(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* detachable ET_REPEAT_INSTRUCTION.make */
extern T0* T1264c21(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_REPEAT_INSTRUCTION.create_unfolded_form */
extern void T1264f24(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.close_repeat_symbol */
extern T0* T73f699(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_close_repeat */
extern T0* T341c115(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.open_repeat_symbol */
extern T0* T73f698(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_open_repeat */
extern T0* T341c114(GE_context* ac);
/* detachable ET_REPEAT_INSTRUCTION.tokens */
extern T0* T1264f13(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_676 */
extern void T456f1963(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_INSTRUCTION].force */
extern void T1010f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ACROSS_INSTRUCTION].fill_with */
extern void T1012f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ACROSS_INSTRUCTION].extend */
extern void T1012f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_INSTRUCTION].aliased_resized_area */
extern T0* T1010f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ACROSS_INSTRUCTION].aliased_resized_area */
extern T0* T1012f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_across_instruction_header */
extern T0* T456f981(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ACROSS_INSTRUCTION.set_has_cursor_name */
extern void T1011f27(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_AST_FACTORY.new_across_instruction */
extern T0* T465f354(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10);
/* detachable ET_ACROSS_INSTRUCTION.set_as_keyword */
extern void T1011f28(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_INSTRUCTION.set_variant_part */
extern void T1011f25(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_INSTRUCTION.set_invariant_part */
extern void T1011f24(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_INSTRUCTION.make */
extern T0* T1011c20(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_ACROSS_INSTRUCTION.create_unfolded_form */
extern void T1011f29(GE_context* ac, T0* C);
/* detachable ET_ACROSS_INSTRUCTION.tokens */
extern T0* T1011f17(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_675 */
extern void T456f1962(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_674 */
extern void T456f1961(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_EXPRESSION].force */
extern void T1292f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_WHEN_EXPRESSION].fill_with */
extern void T1294f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_WHEN_EXPRESSION].extend */
extern void T1294f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_EXPRESSION].aliased_resized_area */
extern T0* T1292f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_WHEN_EXPRESSION].aliased_resized_area */
extern T0* T1294f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_when_expression */
extern T0* T465f282(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_WHEN_EXPRESSION.make */
extern T0* T1293c6(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_WHEN_EXPRESSION.tokens */
extern T0* T1293f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_673 */
extern void T456f1960(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_EXPRESSION_LIST].force */
extern void T1295f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_WHEN_EXPRESSION_LIST].fill_with */
extern void T1297f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_WHEN_EXPRESSION_LIST].extend */
extern void T1297f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_WHEN_EXPRESSION_LIST.put_first */
extern void T1296f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_WHEN_EXPRESSION].force */
extern void T1846f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_WHEN_EXPRESSION].fill_with */
extern void T1845f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_WHEN_EXPRESSION].extend */
extern void T1845f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_WHEN_EXPRESSION_LIST.fixed_array */
extern T0* T1296f3(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_WHEN_EXPRESSION].default_create */
extern T0* T1846c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_672 */
extern void T456f1959(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_EXPRESSION_LIST].aliased_resized_area */
extern T0* T1295f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_WHEN_EXPRESSION_LIST].aliased_resized_area */
extern T0* T1297f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_when_expression_list */
extern T0* T465f281(GE_context* ac, T0* C, T6 a1);
/* detachable ET_WHEN_EXPRESSION_LIST.make_with_capacity */
extern T0* T1296c6(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_WHEN_EXPRESSION].make */
extern T0* T1846f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_WHEN_EXPRESSION].make_empty */
extern T0* T1845c4(GE_context* ac, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_671 */
extern void T456f1958(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_670 */
extern void T456f1957(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_669 */
extern void T456f1956(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSPECT_EXPRESSION].force */
extern void T1188f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INSPECT_EXPRESSION].fill_with */
extern void T1190f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INSPECT_EXPRESSION].extend */
extern void T1190f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSPECT_EXPRESSION].aliased_resized_area */
extern T0* T1188f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_INSPECT_EXPRESSION].aliased_resized_area */
extern T0* T1190f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_inspect_expression */
extern T0* T465f280(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_INSPECT_EXPRESSION.set_else_part */
extern void T1189f25(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INSPECT_EXPRESSION.make */
extern T0* T1189c24(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_INSPECT_EXPRESSION.tokens */
extern T0* T1189f8(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_668 */
extern void T456f1955(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_667 */
extern void T456f1954(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_CONSTANT].force */
extern void T1051f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHOICE_CONSTANT].fill_with */
extern void T1053f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHOICE_CONSTANT].extend */
extern void T1053f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_CONSTANT].aliased_resized_area */
extern T0* T1051f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CHOICE_CONSTANT].aliased_resized_area */
extern T0* T1053f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_666 */
extern void T456f1953(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_choice_attribute_constant */
extern T0* T456f980(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_665 */
extern void T456f1952(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_664 */
extern void T456f1951(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_663 */
extern void T456f1950(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE].force */
extern void T1048f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHOICE].fill_with */
extern void T1050f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHOICE].extend */
extern void T1050f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE].aliased_resized_area */
extern T0* T1048f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CHOICE].aliased_resized_area */
extern T0* T1050f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_choice_range */
extern T0* T465f279(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CHOICE_RANGE.make */
extern T0* T1534c8(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.dotdot_symbol */
extern T0* T73f702(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_dotdot */
extern T0* T341c118(GE_context* ac);
/* detachable ET_CHOICE_RANGE.tokens */
extern T0* T1534f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_662 */
extern void T456f1949(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_661 */
extern void T456f1948(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_ITEM].force */
extern void T1054f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHOICE_ITEM].fill_with */
extern void T1056f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHOICE_ITEM].extend */
extern void T1056f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_ITEM].aliased_resized_area */
extern T0* T1054f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CHOICE_ITEM].aliased_resized_area */
extern T0* T1056f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_choice_comma */
extern T0* T465f278(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_660 */
extern void T456f1947(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_LIST].force */
extern void T1057f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHOICE_LIST].fill_with */
extern void T1059f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHOICE_LIST].extend */
extern void T1059f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CHOICE_LIST.put_first */
extern void T1058f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CHOICE_ITEM].force */
extern void T1814f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CHOICE_ITEM].fill_with */
extern void T1813f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CHOICE_ITEM].extend */
extern void T1813f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CHOICE_LIST.fixed_array */
extern T0* T1058f5(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CHOICE_ITEM].default_create */
extern T0* T1814c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_659 */
extern void T456f1946(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_LIST].aliased_resized_area */
extern T0* T1057f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CHOICE_LIST].aliased_resized_area */
extern T0* T1059f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_choice_list */
extern T0* T465f277(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ET_CHOICE_LIST.set_when_keyword */
extern void T1058f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CHOICE_LIST.make_with_capacity */
extern T0* T1058c11(GE_context* ac, T6 a1);
/* detachable ET_CHOICE_LIST.make_with_capacity */
extern void T1058f11p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CHOICE_ITEM].make */
extern T0* T1814f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CHOICE_ITEM].make_empty */
extern T0* T1813c4(GE_context* ac, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.when_keyword */
extern T0* T73f715(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_when */
extern T0* T413c76(GE_context* ac);
/* detachable ET_CHOICE_LIST.tokens */
extern T0* T1058f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_658 */
extern void T456f1945(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_657 */
extern void T456f1944(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_656 */
extern void T456f1943(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_655 */
extern void T456f1942(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_654 */
extern void T456f1941(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART].force */
extern void T1298f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_WHEN_PART].fill_with */
extern void T1300f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_WHEN_PART].extend */
extern void T1300f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART].aliased_resized_area */
extern T0* T1298f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_WHEN_PART].aliased_resized_area */
extern T0* T1300f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_when_part */
extern T0* T465f276(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_WHEN_PART.make */
extern T0* T1299c4(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_653 */
extern void T456f1940(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART_LIST].force */
extern void T1301f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_WHEN_PART_LIST].fill_with */
extern void T1303f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_WHEN_PART_LIST].extend */
extern void T1303f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_WHEN_PART_LIST.put_first */
extern void T1302f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_WHEN_PART].force */
extern void T1848f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_WHEN_PART].fill_with */
extern void T1847f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_WHEN_PART].extend */
extern void T1847f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_WHEN_PART_LIST.fixed_array */
extern T0* T1302f3(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_WHEN_PART].default_create */
extern T0* T1848c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_652 */
extern void T456f1939(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART_LIST].aliased_resized_area */
extern T0* T1301f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_WHEN_PART_LIST].aliased_resized_area */
extern T0* T1303f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_when_part_list */
extern T0* T465f275(GE_context* ac, T0* C, T6 a1);
/* detachable ET_WHEN_PART_LIST.make_with_capacity */
extern T0* T1302c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_WHEN_PART].make */
extern T0* T1848f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_WHEN_PART].make_empty */
extern T0* T1847c4(GE_context* ac, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_651 */
extern void T456f1938(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_650 */
extern void T456f1937(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_649 */
extern void T456f1936(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSPECT_INSTRUCTION].force */
extern void T1191f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INSPECT_INSTRUCTION].fill_with */
extern void T1193f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INSPECT_INSTRUCTION].extend */
extern void T1193f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSPECT_INSTRUCTION].aliased_resized_area */
extern T0* T1191f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_INSPECT_INSTRUCTION].aliased_resized_area */
extern T0* T1193f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_inspect_instruction */
extern T0* T465f274(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_INSPECT_INSTRUCTION.set_else_compound */
extern void T1192f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INSPECT_INSTRUCTION.make */
extern T0* T1192c8(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_INSPECT_INSTRUCTION.tokens */
extern T0* T1192f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_648 */
extern void T456f1935(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_647 */
extern void T456f1934(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION].force */
extern void T1112f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION].fill_with */
extern void T1114f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION].extend */
extern void T1114f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION].aliased_resized_area */
extern T0* T1112f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION].aliased_resized_area */
extern T0* T1114f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_elseif_expression */
extern T0* T465f273(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ELSEIF_EXPRESSION.make */
extern T0* T1113c7(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_ELSEIF_EXPRESSION.tokens */
extern T0* T1113f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_646 */
extern void T456f1933(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION_LIST].force */
extern void T1115f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION_LIST].fill_with */
extern void T1117f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION_LIST].extend */
extern void T1117f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ELSEIF_EXPRESSION_LIST.put_first */
extern void T1116f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_EXPRESSION].force */
extern void T1824f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ELSEIF_EXPRESSION].fill_with */
extern void T1823f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ELSEIF_EXPRESSION].extend */
extern void T1823f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ELSEIF_EXPRESSION_LIST.fixed_array */
extern T0* T1116f3(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_EXPRESSION].default_create */
extern T0* T1824c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_645 */
extern void T456f1932(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION_LIST].aliased_resized_area */
extern T0* T1115f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION_LIST].aliased_resized_area */
extern T0* T1117f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_elseif_expression_list */
extern T0* T465f272(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ELSEIF_EXPRESSION_LIST.make_with_capacity */
extern T0* T1116c6(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_EXPRESSION].make */
extern T0* T1824f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ELSEIF_EXPRESSION].make_empty */
extern T0* T1823c4(GE_context* ac, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_644 */
extern void T456f1931(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_643 */
extern void T456f1930(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_EXPRESSION].force */
extern void T1166f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_IF_EXPRESSION].fill_with */
extern void T1168f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_IF_EXPRESSION].extend */
extern void T1168f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_EXPRESSION].aliased_resized_area */
extern T0* T1166f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_IF_EXPRESSION].aliased_resized_area */
extern T0* T1168f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_if_expression */
extern T0* T465f271(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* detachable ET_IF_EXPRESSION.set_elseif_parts */
extern void T1167f28(GE_context* ac, T0* C, T0* a1);
/* detachable ET_IF_EXPRESSION.make */
extern T0* T1167c27(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_IF_EXPRESSION.tokens */
extern T0* T1167f11(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_642 */
extern void T456f1929(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_641 */
extern void T456f1928(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART].force */
extern void T1118f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART].fill_with */
extern void T1120f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART].extend */
extern void T1120f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART].aliased_resized_area */
extern T0* T1118f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART].aliased_resized_area */
extern T0* T1120f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_elseif_part */
extern T0* T465f270(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ELSEIF_PART.make */
extern T0* T1119c5(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_640 */
extern void T456f1927(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART_LIST].force */
extern void T1121f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART_LIST].fill_with */
extern void T1123f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART_LIST].extend */
extern void T1123f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ELSEIF_PART_LIST.put_first */
extern void T1122f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_PART].force */
extern void T1826f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ELSEIF_PART].fill_with */
extern void T1825f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ELSEIF_PART].extend */
extern void T1825f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ELSEIF_PART_LIST.fixed_array */
extern T0* T1122f3(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_PART].default_create */
extern T0* T1826c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_639 */
extern void T456f1926(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART_LIST].aliased_resized_area */
extern T0* T1121f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART_LIST].aliased_resized_area */
extern T0* T1123f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_elseif_part_list */
extern T0* T465f269(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ELSEIF_PART_LIST.make_with_capacity */
extern T0* T1122c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ELSEIF_PART].make */
extern T0* T1826f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ELSEIF_PART].make_empty */
extern T0* T1825c4(GE_context* ac, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_638 */
extern void T456f1925(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_637 */
extern void T456f1924(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_INSTRUCTION].force */
extern void T1169f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_IF_INSTRUCTION].fill_with */
extern void T1171f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_IF_INSTRUCTION].extend */
extern void T1171f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_INSTRUCTION].aliased_resized_area */
extern T0* T1169f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_IF_INSTRUCTION].aliased_resized_area */
extern T0* T1171f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_if_instruction */
extern T0* T465f268(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_IF_INSTRUCTION.set_else_compound */
extern void T1170f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_IF_INSTRUCTION.set_elseif_parts */
extern void T1170f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_IF_INSTRUCTION.make */
extern T0* T1170c9(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_IF_INSTRUCTION.tokens */
extern T0* T1170f6(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_636 */
extern void T456f1923(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_635 */
extern void T456f1922(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_634 */
extern void T456f1921(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_633 */
extern void T456f1920(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATION_REGION].force */
extern void T1100f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CREATION_REGION].fill_with */
extern void T1102f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CREATION_REGION].extend */
extern void T1102f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATION_REGION].aliased_resized_area */
extern T0* T1100f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CREATION_REGION].aliased_resized_area */
extern T0* T1102f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_creation_region */
extern T0* T465f267(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CREATION_REGION.set_right_symbol */
extern void T1101f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CREATION_REGION.set_left_symbol */
extern void T1101f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CREATION_REGION.make */
extern T0* T1101c5(GE_context* ac, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.greater_than_symbol */
extern T0* T73f894(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.make_gt */
extern T0* T994c137(GE_context* ac);
/* detachable ET_SYMBOL_OPERATOR.make_leaf */
extern void T994f155(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.less_than_symbol */
extern T0* T73f893(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.make_lt */
extern T0* T994c138(GE_context* ac);
/* detachable ET_CREATION_REGION.tokens */
extern T0* T1101f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_632 */
extern void T456f1919(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_631 */
extern void T456f1918(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATE_EXPRESSION].force */
extern void T1098f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CREATE_EXPRESSION].fill_with */
extern void T1099f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CREATE_EXPRESSION].extend */
extern void T1099f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATE_EXPRESSION].aliased_resized_area */
extern T0* T1098f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CREATE_EXPRESSION].aliased_resized_area */
extern T0* T1099f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_create_expression */
extern T0* T465f266(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_CREATE_EXPRESSION.set_creation_region */
extern void T328f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CREATE_EXPRESSION.set_create_keyword */
extern void T328f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CREATE_EXPRESSION.make */
extern T0* T328c28(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.create_keyword */
extern T0* T73f209(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_create */
extern T0* T413c50(GE_context* ac);
/* [detachable] ET_CREATE_EXPRESSION.tokens */
extern T0* T328f8(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_qualified_call */
extern T0* T465f263(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_QUALIFIED_CALL.make */
extern T0* T329c5(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_630 */
extern void T456f1917(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_629 */
extern void T456f1916(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_create_instruction */
extern T0* T465f265(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_CREATE_INSTRUCTION.set_creation_region */
extern void T1533f14(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CREATE_INSTRUCTION.set_create_keyword */
extern void T1533f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CREATE_INSTRUCTION.make */
extern T0* T1533c12(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_CREATE_INSTRUCTION.tokens */
extern T0* T1533f6(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_628 */
extern void T456f1915(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_627 */
extern void T456f1914(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_626 */
extern void T456f1913(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_625 */
extern void T456f1912(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_bang_instruction */
extern T0* T465f262(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_BANG_INSTRUCTION.set_left_bang */
extern void T1532f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_BANG_INSTRUCTION.make */
extern T0* T1532c11(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_TOKEN_CONSTANTS.bang_symbol */
extern T0* T73f1529(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_bang */
extern T0* T341c98(GE_context* ac);
/* detachable ET_BANG_INSTRUCTION.tokens */
extern T0* T1532f6(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_624 */
extern void T456f1911(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_623 */
extern void T456f1910(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_622 */
extern void T456f1909(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_621 */
extern void T456f1908(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHECK_INSTRUCTION].force */
extern void T1045f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHECK_INSTRUCTION].fill_with */
extern void T1047f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CHECK_INSTRUCTION].extend */
extern void T1047f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHECK_INSTRUCTION].aliased_resized_area */
extern T0* T1045f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CHECK_INSTRUCTION].aliased_resized_area */
extern T0* T1047f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_check_instruction */
extern T0* T456f979(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].remove_last */
extern void T946f15(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_ASSERTION_ITEM].keep_head */
extern void T1647f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ASSERTION_ITEM].set_count */
extern void T1647f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].move_last_cursors_after */
extern void T946f19(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ASSERTION_ITEM].set_next_cursor */
extern void T1649f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ASSERTION_ITEM].set_after */
extern void T1649f6(GE_context* ac, T0* C);
/* detachable ET_CHECK_INSTRUCTION.put_first */
extern void T1046f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTION_ITEM].force */
extern void T1648f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ASSERTION_ITEM].fill_with */
extern void T1647f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ASSERTION_ITEM].extend */
extern void T1647f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CHECK_INSTRUCTION.fixed_array */
extern T0* T1046f6(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTION_ITEM].default_create */
extern T0* T1648c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].item */
extern T0* T946f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_check_instruction */
extern T0* T465f352(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4);
/* detachable ET_CHECK_INSTRUCTION.set_check_keyword */
extern void T1046f14(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CHECK_INSTRUCTION.set_then_compound */
extern void T1046f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CHECK_INSTRUCTION.make_with_capacity */
extern T0* T1046c11(GE_context* ac, T6 a1);
/* detachable ET_CHECK_INSTRUCTION.make_with_capacity */
extern void T1046f11p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTION_ITEM].make */
extern T0* T1648f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ASSERTION_ITEM].make_empty */
extern T0* T1647c5(GE_context* ac, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.check_keyword */
extern T0* T73f717(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_check */
extern T0* T413c78(GE_context* ac);
/* detachable ET_CHECK_INSTRUCTION.tokens */
extern T0* T1046f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].remove_last */
extern void T263f29(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].move_last_cursors_after */
extern void T263f30(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].last */
extern T6 T263f15(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_620 */
extern void T456f1907(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_619 */
extern void T456f1906(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_618 */
extern void T456f1905(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_617 */
extern void T456f1904(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.start_check_instruction */
extern void T456f2262(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.start_assertions */
extern void T456f2267(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_616 */
extern void T456f1903(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_null_instruction */
extern T0* T465f261(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_615 */
extern void T456f1902(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_614 */
extern void T456f1901(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_613 */
extern void T456f1900(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_612 */
extern void T456f1899(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_repeat_instruction */
extern T0* T456f978(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_REPEAT_INSTRUCTION.set_close_repeat_symbol */
extern void T1264f23(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REPEAT_INSTRUCTION.set_loop_compound */
extern void T1264f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_611 */
extern void T456f1898(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_across_instruction */
extern T0* T456f977(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* detachable ET_ACROSS_INSTRUCTION.set_end_keyword */
extern void T1011f26(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_INSTRUCTION.set_loop_compound */
extern void T1011f23(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_INSTRUCTION.set_until_conditional */
extern void T1011f22(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ACROSS_INSTRUCTION.set_from_compound */
extern void T1011f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_610 */
extern void T456f1897(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_loop_instruction */
extern T0* T465f260(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_LOOP_INSTRUCTION.set_has_old_variant_syntax */
extern void T1530f14(GE_context* ac, T0* C, T1 a1);
/* detachable ET_LOOP_INSTRUCTION.set_variant_part */
extern void T1530f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LOOP_INSTRUCTION.set_invariant_part */
extern void T1530f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LOOP_INSTRUCTION.make */
extern T0* T1530c11(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_LOOP_INSTRUCTION.tokens */
extern T0* T1530f8(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.ise_6_3_7_4554 */
extern T0* T456f976(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_609 */
extern void T456f1896(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_loop_instruction_old_syntax */
extern T0* T465f259(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_608 */
extern void T456f1895(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_607 */
extern void T456f1894(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_606 */
extern void T456f1893(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_605 */
extern void T456f1892(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_assignment_attempt */
extern T0* T465f257(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ASSIGNMENT_ATTEMPT.make */
extern T0* T1529c6(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.assign_attempt_symbol */
extern T0* T73f704(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_assign_attempt */
extern T0* T341c120(GE_context* ac);
/* detachable ET_ASSIGNMENT_ATTEMPT.tokens */
extern T0* T1529f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_604 */
extern void T456f1891(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_assignment */
extern T0* T465f256(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ASSIGNMENT.make */
extern T0* T1528c6(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.assign_symbol */
extern T0* T73f703(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_assign */
extern T0* T341c119(GE_context* ac);
/* detachable ET_ASSIGNMENT.tokens */
extern T0* T1528f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_603 */
extern void T456f1890(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_assigner_instruction */
extern T0* T465f255(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ASSIGNER_INSTRUCTION.make */
extern T0* T1527c15(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_ASSIGNER_INSTRUCTION.dummy_name */
extern T0* T1527f6(GE_context* ac, T0* C);
/* detachable ET_ASSIGNER_INSTRUCTION.tokens */
extern T0* T1527f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_602 */
extern void T456f1889(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_601 */
extern void T456f1888(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_401_600 */
extern void T456f1269(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_600 */
extern void T456f1887(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_599 */
extern void T456f1886(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_598 */
extern void T456f1885(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_COMPOUND].force */
extern void T1067f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_COMPOUND].fill_with */
extern void T1068f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_COMPOUND].extend */
extern void T1068f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_COMPOUND].aliased_resized_area */
extern T0* T1067f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_COMPOUND].aliased_resized_area */
extern T0* T1068f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_597 */
extern void T456f1884(GE_context* ac, T0* C);
/* detachable ET_COMPOUND.put_first */
extern void T355f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INSTRUCTION].force */
extern void T1968f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INSTRUCTION].fill_with */
extern void T1966f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INSTRUCTION].extend */
extern void T1966f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_596 */
extern void T456f1883(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_compound */
extern T0* T465f254(GE_context* ac, T0* C, T6 a1);
/* detachable ET_COMPOUND.make_with_capacity */
extern T0* T355c14(GE_context* ac, T6 a1);
/* detachable ET_COMPOUND.make_with_capacity */
extern void T355f14p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_595 */
extern void T456f1882(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_594 */
extern void T456f1881(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_593 */
extern void T456f1880(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_592 */
extern void T456f1879(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_empty_compound */
extern T0* T465f253(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_591 */
extern void T456f1878(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_590 */
extern void T456f1877(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_loop_compound */
extern T0* T465f252(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.loop_keyword */
extern T0* T73f711(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_loop */
extern T0* T413c72(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_589 */
extern void T456f1876(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_588 */
extern void T456f1875(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_587 */
extern void T456f1874(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_from_compound */
extern T0* T465f251(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.from_keyword */
extern T0* T73f710(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_from */
extern T0* T413c71(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_586 */
extern void T456f1873(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_rescue_compound */
extern T0* T465f250(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.rescue_keyword */
extern T0* T73f726(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_rescue */
extern T0* T413c87(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_585 */
extern void T456f1872(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_else_compound */
extern T0* T465f249(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_584 */
extern void T456f1871(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_583 */
extern void T456f1870(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_then_compound */
extern T0* T465f248(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_582 */
extern void T456f1869(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_581 */
extern void T456f1868(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_attribute_compound */
extern T0* T465f247(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.attribute_keyword */
extern T0* T73f738(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_attribute */
extern T0* T413c99(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_580 */
extern void T456f1867(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_do_compound */
extern T0* T465f246(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_579 */
extern void T456f1866(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_LIKE_IDENTIFIER].force */
extern void T1250f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_LIKE_IDENTIFIER].fill_with */
extern void T1252f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_LIKE_IDENTIFIER].extend */
extern void T1252f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_LIKE_IDENTIFIER].aliased_resized_area */
extern T0* T1250f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_LIKE_IDENTIFIER].aliased_resized_area */
extern T0* T1252f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_qualified_like_type */
extern T0* T465f245(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_QUALIFIED_LIKE_TYPE.make */
extern T0* T1526c82(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_AST_FACTORY.new_attachment_mark_separate_keyword */
extern T0* T465f149(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_578 */
extern void T456f1865(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_577 */
extern void T456f1864(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_576 */
extern void T456f1863(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_575 */
extern void T456f1862(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_574 */
extern void T456f1861(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_573 */
extern void T456f1860(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_572 */
extern void T456f1859(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_571 */
extern void T456f1858(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_570 */
extern void T456f1857(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_qualified_like_braced_type */
extern T0* T465f244(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.set_like_keyword */
extern void T1525f86(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.make */
extern T0* T1525c85(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_569 */
extern void T456f1856(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_568 */
extern void T456f1855(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_567 */
extern void T456f1854(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_566 */
extern void T456f1853(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_565 */
extern void T456f1852(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_564 */
extern void T456f1851(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_563 */
extern void T456f1850(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_562 */
extern void T456f1849(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_561 */
extern void T456f1848(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_560 */
extern void T456f1847(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_559 */
extern void T456f1846(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_558 */
extern void T456f1845(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LIKE_TYPE].force */
extern void T1202f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_LIKE_TYPE].fill_with */
extern void T1204f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_LIKE_TYPE].extend */
extern void T1204f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LIKE_TYPE].aliased_resized_area */
extern T0* T1202f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_LIKE_TYPE].aliased_resized_area */
extern T0* T1204f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_557 */
extern void T456f1844(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_like_current */
extern T0* T465f243(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_LIKE_CURRENT.set_like_keyword */
extern void T449f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_556 */
extern void T456f1843(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_555 */
extern void T456f1842(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_554 */
extern void T456f1841(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_553 */
extern void T456f1840(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_552 */
extern void T456f1839(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_551 */
extern void T456f1838(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_550 */
extern void T456f1837(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_549 */
extern void T456f1836(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_548 */
extern void T456f1835(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_547 */
extern void T456f1834(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_like_feature */
extern T0* T456f973(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_AST_FACTORY.new_like_feature */
extern T0* T465f351(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_LIKE_FEATURE.set_like_keyword */
extern void T1524f86(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.make */
extern T0* T1524c85(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_546 */
extern void T456f1833(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_545 */
extern void T456f1832(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_544 */
extern void T456f1831(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_543 */
extern void T456f1830(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_542 */
extern void T456f1829(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_541 */
extern void T456f1828(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_540 */
extern void T456f1827(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_539 */
extern void T456f1826(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_538 */
extern void T456f1825(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_537 */
extern void T456f1824(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_536 */
extern void T456f1823(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_535 */
extern void T456f1822(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_534 */
extern void T456f1821(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_ITEM].force */
extern void T1015f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_ITEM].fill_with */
extern void T1016f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_ITEM].extend */
extern void T1016f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_ITEM].aliased_resized_area */
extern T0* T1015f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_ITEM].aliased_resized_area */
extern T0* T1016f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_labeled_actual_parameter_semicolon */
extern T0* T465f242(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_labeled_actual_parameter */
extern T0* T465f239(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_LABELED_ACTUAL_PARAMETER.make */
extern T0* T1522c10(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_533 */
extern void T456f1820(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_532 */
extern void T456f1819(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_LIST].force */
extern void T1017f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_LIST].fill_with */
extern void T1018f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_LIST].extend */
extern void T1018f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.add_to_actual_parameter_list */
extern void T456f2261(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_labeled_comma_actual_parameter */
extern T0* T465f241(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LABELED_COMMA_ACTUAL_PARAMETER.make */
extern T0* T1769c10(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.first */
extern T0* T240f15(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_label_comma */
extern T0* T465f240(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_531 */
extern void T456f1818(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_530 */
extern void T456f1817(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_529 */
extern void T456f1816(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_528 */
extern void T456f1815(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_527 */
extern void T456f1814(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_LIST].aliased_resized_area */
extern T0* T1017f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_LIST].aliased_resized_area */
extern T0* T1018f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_actual_parameters */
extern T0* T465f237(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.set_left_bracket */
extern void T240f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_526 */
extern void T456f1813(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_525 */
extern void T456f1812(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_524 */
extern void T456f1811(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_523 */
extern void T456f1810(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_522 */
extern void T456f1809(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_521 */
extern void T456f1808(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_actual_parameter_comma */
extern T0* T465f238(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_520 */
extern void T456f1807(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_519 */
extern void T456f1806(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_tuple_type */
extern T0* T456f972(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T456f972ot1(T0* a1, T0** a2);
/* [detachable] ET_AST_FACTORY.new_tuple_type */
extern T0* T465f350(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_EIFFEL_PARSER.providers_enabled */
extern T1 T456f1001(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_518 */
extern void T456f1805(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_named_type */
extern T0* T456f971(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T456f971ot1(T0* a1, T0** a2);
/* [detachable] ET_AST_FACTORY.new_class_type */
extern T0* T465f337(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_AST_FACTORY.new_generic_class_type */
extern T0* T465f336(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_AST_FACTORY.new_formal_parameter_type */
extern T0* T465f349(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T0* a4);
/* [detachable] ET_CLASS.formal_parameter */
extern T0* T74f195(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.formal_parameter_by_name */
extern T0* T603f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_517 */
extern void T456f1804(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_516 */
extern void T456f1803(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_515 */
extern void T456f1802(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_514 */
extern void T456f1801(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_513 */
extern void T456f1800(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_512 */
extern void T456f1799(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_511 */
extern void T456f1798(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_510 */
extern void T456f1797(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_509 */
extern void T456f1796(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_508 */
extern void T456f1795(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_507 */
extern void T456f1794(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE].force */
extern void T1272f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE].fill_with */
extern void T1273f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE].extend */
extern void T1273f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE].aliased_resized_area */
extern T0* T1272f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_TYPE].aliased_resized_area */
extern T0* T1273f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_506 */
extern void T456f1793(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_505 */
extern void T456f1792(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_504 */
extern void T456f1791(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_503 */
extern void T456f1790(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_502 */
extern void T456f1789(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_501 */
extern void T456f1788(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_500 */
extern void T456f1787(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_499 */
extern void T456f1786(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_498 */
extern void T456f1785(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_497 */
extern void T456f1784(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_496 */
extern void T456f1783(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_495 */
extern void T456f1782(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_494 */
extern void T456f1781(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_493 */
extern void T456f1780(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_492 */
extern void T456f1779(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_491 */
extern void T456f1778(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_490 */
extern void T456f1777(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_489 */
extern void T456f1776(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_488 */
extern void T456f1775(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_487 */
extern void T456f1774(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_486 */
extern void T456f1773(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_485 */
extern void T456f1772(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_484 */
extern void T456f1771(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_483 */
extern void T456f1770(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_482 */
extern void T456f1769(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_481 */
extern void T456f1768(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_480 */
extern void T456f1767(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_479 */
extern void T456f1766(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_478 */
extern void T456f1765(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_477 */
extern void T456f1764(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_476 */
extern void T456f1763(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_475 */
extern void T456f1762(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_474 */
extern void T456f1761(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_473 */
extern void T456f1760(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_472 */
extern void T456f1759(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_471 */
extern void T456f1758(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_470 */
extern void T456f1757(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_469 */
extern void T456f1756(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_468 */
extern void T456f1755(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_467 */
extern void T456f1754(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_466 */
extern void T456f1753(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_465 */
extern void T456f1752(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_464 */
extern void T456f1751(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_463 */
extern void T456f1750(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_462 */
extern void T456f1749(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_461 */
extern void T456f1748(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_460 */
extern void T456f1747(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_459 */
extern void T456f1746(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_458 */
extern void T456f1745(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_457 */
extern void T456f1744(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_456 */
extern void T456f1743(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_455 */
extern void T456f1742(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_454 */
extern void T456f1741(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_453 */
extern void T456f1740(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_452 */
extern void T456f1739(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_451 */
extern void T456f1738(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_450 */
extern void T456f1737(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_449 */
extern void T456f1736(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_448 */
extern void T456f1735(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_447 */
extern void T456f1734(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_446 */
extern void T456f1733(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_445 */
extern void T456f1732(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_444 */
extern void T456f1731(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_443 */
extern void T456f1730(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_442 */
extern void T456f1729(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_441 */
extern void T456f1728(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_440 */
extern void T456f1727(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_439 */
extern void T456f1726(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_438 */
extern void T456f1725(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_437 */
extern void T456f1724(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_436 */
extern void T456f1723(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_VARIANT].force */
extern void T1289f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_VARIANT].fill_with */
extern void T1291f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_VARIANT].extend */
extern void T1291f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_435 */
extern void T456f1722(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_VARIANT].aliased_resized_area */
extern T0* T1289f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_VARIANT].aliased_resized_area */
extern T0* T1291f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_434 */
extern void T456f1721(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_variant */
extern T0* T465f236(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_VARIANT.set_variant_keyword */
extern void T1290f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_VARIANT.make */
extern T0* T1290c6(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.variant_keyword */
extern T0* T73f731(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_variant */
extern T0* T413c92(GE_context* ac);
/* detachable ET_VARIANT.tokens */
extern T0* T1290f4(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_tag */
extern T0* T465f133(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_433 */
extern void T456f1720(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_432 */
extern void T456f1719(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOOP_INVARIANTS].force */
extern void T1214f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_LOOP_INVARIANTS].fill_with */
extern void T1216f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_LOOP_INVARIANTS].extend */
extern void T1216f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOOP_INVARIANTS].aliased_resized_area */
extern T0* T1214f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_LOOP_INVARIANTS].aliased_resized_area */
extern T0* T1216f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_loop_invariants */
extern T0* T456f970(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LOOP_INVARIANTS.put_first */
extern void T1215f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LOOP_INVARIANTS.fixed_array */
extern T0* T1215f5(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_loop_invariants */
extern T0* T465f348(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ET_LOOP_INVARIANTS.set_invariant_keyword */
extern void T1215f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LOOP_INVARIANTS.make_with_capacity */
extern T0* T1215c11(GE_context* ac, T6 a1);
/* detachable ET_LOOP_INVARIANTS.make_with_capacity */
extern void T1215f11p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.invariant_keyword */
extern T0* T73f739(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_invariant */
extern T0* T413c100(GE_context* ac);
/* detachable ET_LOOP_INVARIANTS.tokens */
extern T0* T1215f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_431 */
extern void T456f1718(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_430 */
extern void T456f1717(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_429 */
extern void T456f1716(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_428 */
extern void T456f1715(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.start_loop_invariant */
extern void T456f2260(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_427 */
extern void T456f1714(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_426 */
extern void T456f1713(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INVARIANTS].force */
extern void T1197f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INVARIANTS].fill_with */
extern void T1199f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INVARIANTS].extend */
extern void T1199f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INVARIANTS].aliased_resized_area */
extern T0* T1197f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_INVARIANTS].aliased_resized_area */
extern T0* T1199f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_invariants */
extern T0* T456f969(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.wipe_out_last_iteration_components_stack */
extern void T456f1250(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_ITERATION_COMPONENT_LIST].wipe_out */
extern void T944f11(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_ITERATION_COMPONENT_LIST].i_th */
extern T0* T944f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.wipe_out_last_object_tests_stack */
extern void T456f1249(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST].wipe_out */
extern void T942f11(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST].i_th */
extern T0* T942f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.wipe_out_last_local_variables_stack */
extern void T456f1248(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST].wipe_out */
extern void T939f10(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.wipe_out_last_formal_arguments_stack */
extern void T456f1247(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST].wipe_out */
extern void T938f10(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.set_iteration_components */
extern void T1198f28(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INVARIANTS.set_object_tests */
extern void T1198f27(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INVARIANTS.put_first */
extern void T1198f26(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INVARIANTS.fixed_array */
extern T0* T1198f10(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_invariants */
extern T0* T465f347(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_INVARIANTS.set_invariant_keyword */
extern void T1198f29(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INVARIANTS.make_with_capacity */
extern T0* T1198c25(GE_context* ac, T0* a1, T6 a2);
/* detachable ET_INVARIANTS.make_assertions_with_capacity */
extern void T1198f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_425 */
extern void T456f1712(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_424 */
extern void T456f1711(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_423 */
extern void T456f1710(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_422 */
extern void T456f1709(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.start_invariant */
extern void T456f2259(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_421 */
extern void T456f1708(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSTCONDITIONS].force */
extern void T1243f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_POSTCONDITIONS].fill_with */
extern void T1245f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_POSTCONDITIONS].extend */
extern void T1245f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSTCONDITIONS].aliased_resized_area */
extern T0* T1243f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_POSTCONDITIONS].aliased_resized_area */
extern T0* T1245f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_postconditions */
extern T0* T456f968(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_POSTCONDITIONS.put_first */
extern void T1244f18(GE_context* ac, T0* C, T0* a1);
/* detachable ET_POSTCONDITIONS.fixed_array */
extern T0* T1244f10(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_postconditions */
extern T0* T465f346(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_POSTCONDITIONS.set_then_keyword */
extern void T1244f20(GE_context* ac, T0* C, T0* a1);
/* detachable ET_POSTCONDITIONS.set_ensure_keyword */
extern void T1244f19(GE_context* ac, T0* C, T0* a1);
/* detachable ET_POSTCONDITIONS.make_with_capacity */
extern T0* T1244c17(GE_context* ac, T6 a1);
/* detachable ET_POSTCONDITIONS.make_with_capacity */
extern void T1244f17p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.ensure_keyword */
extern T0* T73f896(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_ensure */
extern T0* T413c60(GE_context* ac);
/* detachable ET_POSTCONDITIONS.tokens */
extern T0* T1244f9(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_420 */
extern void T456f1707(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_419 */
extern void T456f1706(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_418 */
extern void T456f1705(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_417 */
extern void T456f1704(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_416 */
extern void T456f1703(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.start_postcondition */
extern void T456f2258(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_415 */
extern void T456f1702(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECONDITIONS].force */
extern void T1246f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PRECONDITIONS].fill_with */
extern void T1247f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PRECONDITIONS].extend */
extern void T1247f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECONDITIONS].aliased_resized_area */
extern T0* T1246f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_PRECONDITIONS].aliased_resized_area */
extern T0* T1247f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_preconditions */
extern T0* T456f967(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_PRECONDITIONS.put_first */
extern void T406f17(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECONDITIONS.fixed_array */
extern T0* T406f10(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_preconditions */
extern T0* T465f345(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_PRECONDITIONS.set_else_keyword */
extern void T406f19(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECONDITIONS.set_require_keyword */
extern void T406f18(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECONDITIONS.make_with_capacity */
extern T0* T406c16(GE_context* ac, T6 a1);
/* detachable ET_PRECONDITIONS.make_with_capacity */
extern void T406f16p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.require_keyword */
extern T0* T73f1538(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_require */
extern T0* T413c63(GE_context* ac);
/* detachable ET_PRECONDITIONS.tokens */
extern T0* T406f9(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_414 */
extern void T456f1701(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_413 */
extern void T456f1700(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_412 */
extern void T456f1699(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_411 */
extern void T456f1698(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_410 */
extern void T456f1697(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.start_precondition */
extern void T456f2257(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_409 */
extern void T456f1696(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.add_class_assertion */
extern void T456f2256(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.add_untagged_assertion */
extern void T456f2266(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T456f2266ot1(T0* a1, T0** a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].force_last */
extern void T946f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].resize */
extern void T946f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTION_ITEM].aliased_resized_area */
extern T0* T1648f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ASSERTION_ITEM].aliased_resized_area */
extern T0* T1647f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].new_capacity */
extern T6 T946f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].extendible */
extern T1 T946f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].replace */
extern void T946f14(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_AST_FACTORY.new_assertion_semicolon */
extern T0* T465f333(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_TAGGED_ASSERTION.set_untagged_assertion */
extern void T1562f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].last */
extern T0* T946f9(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_class_assertion */
extern T0* T465f235(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLASS_ASSERTION.set_class_keyword */
extern void T1521f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLASS_ASSERTION.make */
extern T0* T1521c7(GE_context* ac);
/* detachable ET_CLASS_ASSERTION.tokens */
extern T0* T1521f2(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_408 */
extern void T456f1695(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_407 */
extern void T456f1694(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.add_tagged_assertion */
extern void T456f2255(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T456f2255ot1(T0* a1, T0** a2);
/* [detachable] ET_AST_FACTORY.new_tagged_assertion */
extern T0* T465f332(GE_context* ac, T0* C, T0* a1);
/* detachable ET_TAGGED_ASSERTION.make */
extern T0* T1562c7(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.report_syntax_error */
extern void T456f1273(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_syntax_error */
extern void T53f266(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_SYNTAX_ERROR.make */
extern T0* T1941c18(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_SYNTAX_ERROR.set_compilers */
extern void T1941f19(GE_context* ac, T0* C, T1 a1);
/* detachable ET_BINARY_INTEGER_CONSTANT.position_to_text */
extern T0* T1489f55(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.append_position_to_string */
extern void T1489f72(GE_context* ac, T0* C, T0* a1);
/* detachable ET_BINARY_INTEGER_CONSTANT.integer_ */
extern T0* T1489f58(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.position_to_text */
extern T0* T1488f55(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.append_position_to_string */
extern void T1488f72(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OCTAL_INTEGER_CONSTANT.integer_ */
extern T0* T1488f58(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.position_to_text */
extern T0* T1487f55(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.append_position_to_string */
extern void T1487f72(GE_context* ac, T0* C, T0* a1);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.integer_ */
extern T0* T1487f58(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.position_to_text */
extern T0* T1486f55(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.append_position_to_string */
extern void T1486f72(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.integer_ */
extern T0* T1486f58(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.position_to_text */
extern T0* T372f55(GE_context* ac, T0* C);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.append_position_to_string */
extern void T372f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.integer_ */
extern T0* T372f58(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.position_to_text */
extern T0* T1485f41(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.append_position_to_string */
extern void T1485f58(GE_context* ac, T0* C, T0* a1);
/* detachable ET_VERBATIM_STRING.integer_ */
extern T0* T1485f45(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.position_to_text */
extern T0* T1484f37(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.append_position_to_string */
extern void T1484f54(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SPECIAL_MANIFEST_STRING.integer_ */
extern T0* T1484f41(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.position_to_text */
extern T0* T1482f40(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.append_position_to_string */
extern void T1482f55(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C3_CHARACTER_CONSTANT.integer_ */
extern T0* T1482f42(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.position_to_text */
extern T0* T1496f40(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.append_position_to_string */
extern void T1496f55(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C2_CHARACTER_CONSTANT.integer_ */
extern T0* T1496f42(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.position_to_text */
extern T0* T1495f39(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.append_position_to_string */
extern void T1495f54(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C1_CHARACTER_CONSTANT.integer_ */
extern T0* T1495f41(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.position_to_text */
extern T0* T1494f31(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.append_position_to_string */
extern void T1494f46(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FALSE_CONSTANT.integer_ */
extern T0* T1494f34(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.position_to_text */
extern T0* T1493f31(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.append_position_to_string */
extern void T1493f46(GE_context* ac, T0* C, T0* a1);
/* detachable ET_TRUE_CONSTANT.integer_ */
extern T0* T1493f34(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.position_to_text */
extern T0* T1491f33(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.append_position_to_string */
extern void T1491f50(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.integer_ */
extern T0* T1491f36(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.position_to_text */
extern T0* T1490f33(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.append_position_to_string */
extern void T1490f50(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REGULAR_REAL_CONSTANT.integer_ */
extern T0* T1490f36(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.position_to_text */
extern T0* T1483f36(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.append_position_to_string */
extern void T1483f53(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REGULAR_MANIFEST_STRING.integer_ */
extern T0* T1483f40(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.position_to_text */
extern T0* T282f26(GE_context* ac, T0* C);
/* [detachable] ET_CURRENT.append_position_to_string */
extern void T282f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CURRENT.integer_ */
extern T0* T282f27(GE_context* ac, T0* C);
/* [detachable] ET_FILE_POSITION.to_text */
extern T0* T1564f10(GE_context* ac, T0* C);
/* [detachable] ET_FILE_POSITION.append_to_string */
extern void T1564f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FILE_POSITION.append_to_string_with_filename */
extern void T1564f16(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FILE_POSITION.append_context_to_string */
extern void T1564f17(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FILE_POSITION.append_line_to_string */
extern void T1564f19(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FILE_POSITION.skip_lines */
extern void T1564f18(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_FILE_POSITION.integer_ */
extern T0* T1564f11(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_syntax_error */
extern void T36f266(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.set_syntax_error */
extern void T456f1277(GE_context* ac, T0* C);
/* [detachable] ET_FILE_POSITION.make */
extern T0* T1564c12(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_FILE_POSITION.make_compressed_position */
extern void T1564f13(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_FILE_POSITION.set_position */
extern void T1564f14(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_FILE_POSITION.column */
extern T6 T1564f3(GE_context* ac, T0* C);
/* [detachable] ET_FILE_POSITION.line */
extern T6 T1564f6(GE_context* ac, T0* C);
/* [detachable] ET_FILE_POSITION.is_null */
extern T1 T1564f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.current_position */
extern T0* T456f745(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_406 */
extern void T456f1693(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_405 */
extern void T456f1692(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.add_expression_assertion */
extern void T456f2254(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_404 */
extern void T456f1691(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_403 */
extern void T456f1690(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_402 */
extern void T456f1689(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_401 */
extern void T456f1688(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_201_400 */
extern void T456f1268(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_400 */
extern void T456f1687(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_399 */
extern void T456f1686(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_398 */
extern void T456f1685(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_397 */
extern void T456f1684(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_ITEM].force */
extern void T1208f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_ITEM].fill_with */
extern void T1210f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_ITEM].extend */
extern void T1210f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_ITEM].aliased_resized_area */
extern T0* T1208f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_ITEM].aliased_resized_area */
extern T0* T1210f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_local_variable_semicolon */
extern T0* T465f234(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_local_variable */
extern T0* T465f233(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_LOCAL_VARIABLE.make */
extern T0* T1206c9(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_396 */
extern void T456f1683(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_395 */
extern void T456f1682(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE].force */
extern void T1205f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE].fill_with */
extern void T1207f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE].extend */
extern void T1207f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE].aliased_resized_area */
extern T0* T1205f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE].aliased_resized_area */
extern T0* T1207f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_local_comma_variable */
extern T0* T465f232(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LOCAL_COMMA_VARIABLE.make */
extern T0* T1768c9(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.dummy_type */
extern T0* T456f964(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_394 */
extern void T456f1681(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_local_name_comma */
extern T0* T465f231(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_393 */
extern void T456f1680(GE_context* ac, T0* C);
/* detachable ET_LOCAL_VARIABLE_LIST.put_first */
extern void T1212f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LOCAL_VARIABLE_ITEM].force */
extern void T1836f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_LOCAL_VARIABLE_ITEM].fill_with */
extern void T1834f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_LOCAL_VARIABLE_ITEM].extend */
extern void T1834f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LOCAL_VARIABLE_LIST.fixed_array */
extern T0* T1212f10(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LOCAL_VARIABLE_ITEM].default_create */
extern T0* T1836c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_392 */
extern void T456f1679(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_391 */
extern void T456f1678(GE_context* ac, T0* C);
/* [detachable] ET_LOCAL_COMMA_VARIABLE.set_declared_type */
extern void T1768f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LOCAL_VARIABLE_LIST.first */
extern T0* T1212f6(GE_context* ac, T0* C);
/* detachable ET_LOCAL_VARIABLE_LIST.is_empty */
extern T1 T1212f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_390 */
extern void T456f1677(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_389 */
extern void T456f1676(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_local_variables */
extern T0* T456f965(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_AST_FACTORY.new_local_variables */
extern T0* T465f344(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ET_LOCAL_VARIABLE_LIST.set_local_keyword */
extern void T1212f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LOCAL_VARIABLE_LIST.make_with_capacity */
extern T0* T1212c11(GE_context* ac, T6 a1);
/* detachable ET_LOCAL_VARIABLE_LIST.make_with_capacity */
extern void T1212f11p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LOCAL_VARIABLE_ITEM].make */
extern T0* T1836f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_LOCAL_VARIABLE_ITEM].make_empty */
extern T0* T1834c4(GE_context* ac, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.local_keyword */
extern T0* T73f719(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_local */
extern T0* T413c80(GE_context* ac);
/* detachable ET_LOCAL_VARIABLE_LIST.tokens */
extern T0* T1212f9(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_388 */
extern void T456f1675(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_387 */
extern void T456f1674(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_386 */
extern void T456f1673(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_385 */
extern void T456f1672(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_384 */
extern void T456f1671(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_383 */
extern void T456f1670(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_ITEM].force */
extern void T1154f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_ITEM].fill_with */
extern void T1156f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_ITEM].extend */
extern void T1156f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_ITEM].aliased_resized_area */
extern T0* T1154f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_ITEM].aliased_resized_area */
extern T0* T1156f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_formal_argument_semicolon */
extern T0* T465f230(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_formal_argument */
extern T0* T465f229(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_ARGUMENT.make */
extern T0* T256c9(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_382 */
extern void T456f1669(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_381 */
extern void T456f1668(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT].force */
extern void T1152f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT].fill_with */
extern void T1153f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT].extend */
extern void T1153f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT].aliased_resized_area */
extern T0* T1152f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT].aliased_resized_area */
extern T0* T1153f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_formal_comma_argument */
extern T0* T465f228(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_COMMA_ARGUMENT.make */
extern T0* T1767c9(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_380 */
extern void T456f1667(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_argument_name_comma */
extern T0* T465f227(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_379 */
extern void T456f1666(GE_context* ac, T0* C);
/* detachable ET_FORMAL_ARGUMENT_LIST.put_first */
extern void T255f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_ARGUMENT_ITEM].force */
extern void T1962f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FORMAL_ARGUMENT_ITEM].fill_with */
extern void T1960f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FORMAL_ARGUMENT_ITEM].extend */
extern void T1960f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_ARGUMENT_LIST.fixed_array */
extern T0* T255f11(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_ARGUMENT_ITEM].default_create */
extern T0* T1962c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_378 */
extern void T456f1665(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_377 */
extern void T456f1664(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_COMMA_ARGUMENT.set_declared_type */
extern void T1767f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_ARGUMENT_LIST.first */
extern T0* T255f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_376 */
extern void T456f1663(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_375 */
extern void T456f1662(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_formal_arguments */
extern T0* T456f963(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] ET_AST_FACTORY.new_formal_arguments */
extern T0* T465f343(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_FORMAL_ARGUMENT_LIST.make_with_capacity */
extern T0* T255c12(GE_context* ac, T6 a1);
/* detachable ET_FORMAL_ARGUMENT_LIST.make_with_capacity */
extern void T255f12p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_ARGUMENT_ITEM].make */
extern T0* T1962f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_FORMAL_ARGUMENT_ITEM].make_empty */
extern T0* T1960c4(GE_context* ac, T6 a1);
/* detachable ET_FORMAL_ARGUMENT_LIST.tokens */
extern T0* T255f9(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_374 */
extern void T456f1661(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_373 */
extern void T456f1660(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_372 */
extern void T456f1659(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_371 */
extern void T456f1658(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_370 */
extern void T456f1657(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_369 */
extern void T456f1656(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_368 */
extern void T456f1655(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_367 */
extern void T456f1654(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ALIAS_NAME].force */
extern void T1029f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ALIAS_NAME].fill_with */
extern void T1031f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ALIAS_NAME].extend */
extern void T1031f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ALIAS_NAME].aliased_resized_area */
extern T0* T1029f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ALIAS_NAME].aliased_resized_area */
extern T0* T1031f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_invalid_alias_name */
extern T0* T456f962(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.new_alias_free_name */
extern T0* T456f1014(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_AST_FACTORY.new_alias_free_name */
extern T0* T465f224(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_FREE_NAME.set_convert_keyword */
extern void T1518f56(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALIAS_FREE_NAME.set_alias_keyword */
extern void T1518f55(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALIAS_FREE_NAME.make_infix */
extern T0* T1518c54(GE_context* ac, T0* a1);
/* detachable ET_ALIAS_FREE_NAME.default_keyword */
extern T0* T1518f53(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.alias_keyword */
extern T0* T73f716(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_alias */
extern T0* T413c77(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_366 */
extern void T456f1653(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_parenthesis_name */
extern T0* T465f226(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.set_convert_keyword */
extern void T1030f82(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALIAS_NAME.set_alias_keyword */
extern void T1030f81(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALIAS_NAME.make_parenthesis */
extern T0* T1030c109(GE_context* ac, T0* a1);
/* detachable ET_ALIAS_NAME.default_keyword */
extern T0* T1030f6(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_365 */
extern void T456f1652(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_bracket_name */
extern T0* T465f225(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_bracket */
extern T0* T1030c108(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_364 */
extern void T456f1651(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_363 */
extern void T456f1650(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_dotdot_name */
extern T0* T465f223(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_dotdot */
extern T0* T1030c107(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_362 */
extern void T456f1649(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_xor_symbol_name */
extern T0* T465f222(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_xor_symbol */
extern T0* T1030c106(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_361 */
extern void T456f1648(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_xor_name */
extern T0* T465f221(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_xor */
extern T0* T1030c105(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_360 */
extern void T456f1647(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_implies_symbol_name */
extern T0* T465f220(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_implies_symbol */
extern T0* T1030c104(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_359 */
extern void T456f1646(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_implies_name */
extern T0* T465f219(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_implies */
extern T0* T1030c103(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_358 */
extern void T456f1645(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_or_else_symbol_name */
extern T0* T465f218(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_or_else_symbol */
extern T0* T1030c102(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_357 */
extern void T456f1644(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_or_else_name */
extern T0* T465f217(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_or_else */
extern T0* T1030c101(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_356 */
extern void T456f1643(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_or_symbol_name */
extern T0* T465f216(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_or_symbol */
extern T0* T1030c100(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_355 */
extern void T456f1642(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_or_name */
extern T0* T465f215(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_or */
extern T0* T1030c99(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_354 */
extern void T456f1641(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_and_then_symbol_name */
extern T0* T465f214(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_and_then_symbol */
extern T0* T1030c98(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_353 */
extern void T456f1640(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_and_then_name */
extern T0* T465f213(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_and_then */
extern T0* T1030c97(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_352 */
extern void T456f1639(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_and_symbol_name */
extern T0* T465f212(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_and_symbol */
extern T0* T1030c96(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_351 */
extern void T456f1638(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_and_name */
extern T0* T465f211(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_and */
extern T0* T1030c95(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_350 */
extern void T456f1637(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_ge_name */
extern T0* T465f210(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_ge */
extern T0* T1030c94(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_349 */
extern void T456f1636(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_gt_name */
extern T0* T465f209(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_gt */
extern T0* T1030c93(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_348 */
extern void T456f1635(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_le_name */
extern T0* T465f208(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_le */
extern T0* T1030c92(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_347 */
extern void T456f1634(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_lt_name */
extern T0* T465f207(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_lt */
extern T0* T1030c91(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_346 */
extern void T456f1633(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_power_name */
extern T0* T465f206(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_power */
extern T0* T1030c90(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_345 */
extern void T456f1632(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_mod_name */
extern T0* T465f205(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_mod */
extern T0* T1030c89(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_344 */
extern void T456f1631(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_div_name */
extern T0* T465f204(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_div */
extern T0* T1030c88(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_343 */
extern void T456f1630(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_divide_name */
extern T0* T465f203(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_divide */
extern T0* T1030c87(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_342 */
extern void T456f1629(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_times_name */
extern T0* T465f202(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_times */
extern T0* T1030c86(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_341 */
extern void T456f1628(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_minus_name */
extern T0* T465f201(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_minus */
extern T0* T1030c85(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_340 */
extern void T456f1627(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_plus_name */
extern T0* T465f200(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_plus */
extern T0* T1030c84(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_339 */
extern void T456f1626(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_not_symbol_name */
extern T0* T465f199(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_not_symbol */
extern T0* T1030c83(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_338 */
extern void T456f1625(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_not_name */
extern T0* T465f198(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ALIAS_NAME.make_not */
extern T0* T1030c80(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_337 */
extern void T456f1624(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ALIAS_NAME_LIST].force */
extern void T1032f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ALIAS_NAME_LIST].fill_with */
extern void T1033f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ALIAS_NAME_LIST].extend */
extern void T1033f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ALIAS_NAME_LIST].aliased_resized_area */
extern T0* T1032f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ALIAS_NAME_LIST].aliased_resized_area */
extern T0* T1033f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_336 */
extern void T456f1623(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME_LIST.put_first */
extern void T911f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ALIAS_NAME_LIST.fixed_array */
extern T0* T911f3(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_335 */
extern void T456f1622(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_alias_name_list */
extern T0* T465f197(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ALIAS_NAME_LIST.make_with_capacity */
extern T0* T911c11(GE_context* ac, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_334 */
extern void T456f1621(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTENDED_FEATURE_NAME].force */
extern void T1133f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXTENDED_FEATURE_NAME].fill_with */
extern void T1134f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXTENDED_FEATURE_NAME].extend */
extern void T1134f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTENDED_FEATURE_NAME].aliased_resized_area */
extern T0* T1133f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_EXTENDED_FEATURE_NAME].aliased_resized_area */
extern T0* T1134f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_aliased_feature_name */
extern T0* T465f196(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_333 */
extern void T456f1620(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_332 */
extern void T456f1619(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME].force */
extern void T1147f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME].fill_with */
extern void T1148f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME].extend */
extern void T1148f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME].aliased_resized_area */
extern T0* T1147f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME].aliased_resized_area */
extern T0* T1148f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_331 */
extern void T456f1618(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ASSIGNER].force */
extern void T1034f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ASSIGNER].fill_with */
extern void T1036f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_ASSIGNER].extend */
extern void T1036f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ASSIGNER].aliased_resized_area */
extern T0* T1034f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_ASSIGNER].aliased_resized_area */
extern T0* T1036f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_assigner */
extern T0* T465f195(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_330 */
extern void T456f1617(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_329 */
extern void T456f1616(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTERNAL_ALIAS].force */
extern void T1135f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXTERNAL_ALIAS].fill_with */
extern void T1137f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXTERNAL_ALIAS].extend */
extern void T1137f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTERNAL_ALIAS].aliased_resized_area */
extern T0* T1135f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_EXTERNAL_ALIAS].aliased_resized_area */
extern T0* T1137f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_external_alias */
extern T0* T465f194(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_328 */
extern void T456f1615(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_327 */
extern void T456f1614(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SEMICOLON_SYMBOL].force */
extern void T999f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_SEMICOLON_SYMBOL].fill_with */
extern void T1001f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_SEMICOLON_SYMBOL].extend */
extern void T1001f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_326 */
extern void T456f1613(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SEMICOLON_SYMBOL].aliased_resized_area */
extern T0* T999f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_SEMICOLON_SYMBOL].aliased_resized_area */
extern T0* T1001f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_325 */
extern void T456f1612(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_324 */
extern void T456f1611(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_323 */
extern void T456f1610(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PROCEDURE].force */
extern void T1248f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PROCEDURE].fill_with */
extern void T1249f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PROCEDURE].extend */
extern void T1249f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PROCEDURE].aliased_resized_area */
extern T0* T1248f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_PROCEDURE].aliased_resized_area */
extern T0* T1249f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_external_procedure */
extern T0* T456f961(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11, T0* a12, T0* a13, T0* a14);
/* [detachable] ET_AST_FACTORY.new_external_procedure */
extern T0* T465f342(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11, T0* a12, T0* a13, T0* a14);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_322 */
extern void T456f1609(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_321 */
extern void T456f1608(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_deferred_procedure */
extern T0* T465f193(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11, T0* a12, T0* a13);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_320 */
extern void T456f1607(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_319 */
extern void T456f1606(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_once_procedure */
extern T0* T465f192(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11, T0* a12, T0* a13, T0* a14, T0* a15, T0* a16);
/* detachable ET_ONCE_PROCEDURE.set_keys */
extern void T1515f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_318 */
extern void T456f1605(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_317 */
extern void T456f1604(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_do_procedure */
extern T0* T465f191(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11, T0* a12, T0* a13, T0* a14, T0* a15);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_316 */
extern void T456f1603(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_315 */
extern void T456f1602(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUERY].force */
extern void T1256f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_QUERY].fill_with */
extern void T1257f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_QUERY].extend */
extern void T1257f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUERY].aliased_resized_area */
extern T0* T1256f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_QUERY].aliased_resized_area */
extern T0* T1257f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_external_function */
extern T0* T456f960(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11, T0* a12, T0* a13, T0* a14, T0* a15, T0* a16);
/* [detachable] ET_AST_FACTORY.new_external_function */
extern T0* T465f341(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11, T0* a12, T0* a13, T0* a14, T0* a15, T0* a16);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_314 */
extern void T456f1601(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_313 */
extern void T456f1600(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_312 */
extern void T456f1599(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_311 */
extern void T456f1598(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_deferred_function */
extern T0* T465f189(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11, T0* a12, T0* a13, T0* a14, T0* a15);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_310 */
extern void T456f1597(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_309 */
extern void T456f1596(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_308 */
extern void T456f1595(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_307 */
extern void T456f1594(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_once_function */
extern T0* T465f188(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11, T0* a12, T0* a13, T0* a14, T0* a15, T0* a16, T0* a17, T0* a18);
/* detachable ET_ONCE_FUNCTION.set_keys */
extern void T1512f101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_306 */
extern void T456f1593(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_305 */
extern void T456f1592(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_304 */
extern void T456f1591(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_303 */
extern void T456f1590(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_do_function */
extern T0* T465f186(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11, T0* a12, T0* a13, T0* a14, T0* a15, T0* a16, T0* a17);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_302 */
extern void T456f1589(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_301 */
extern void T456f1588(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_300 */
extern void T456f1587(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_299 */
extern void T456f1586(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_unique_attribute */
extern T0* T465f185(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_298 */
extern void T456f1585(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_297 */
extern void T456f1584(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_constant_attribute */
extern T0* T465f184(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_296 */
extern void T456f1583(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_295 */
extern void T456f1582(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_extended_attribute */
extern T0* T465f183(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T0* a11, T0* a12, T0* a13, T0* a14, T0* a15);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_294 */
extern void T456f1581(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_attribute */
extern T0* T465f182(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_293 */
extern void T456f1580(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_292 */
extern void T456f1579(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.register_procedure_synonym */
extern void T456f2252(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].back */
extern void T715f22(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].cursor_back */
extern void T715f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].add_traversing_cursor */
extern void T715f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].remove_traversing_cursor */
extern void T715f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].force_left */
extern void T715f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].force_left_cursor */
extern void T715f27(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].put */
extern void T715f31(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].move_cursors_right */
extern void T715f34(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].move_right */
extern void T715f33(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].put_last */
extern void T715f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PROCEDURE].index */
extern T6 T1425f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].cursor_index */
extern T6 T715f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].forth */
extern void T715f20(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].cursor_forth */
extern void T715f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].before */
extern T1 T715f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].cursor_before */
extern T1 T715f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PROCEDURE].before */
extern T1 T1425f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_procedure_synonym */
extern T0* T456f958(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_EXTERNAL_PROCEDURE.new_synonym */
extern T0* T1516f31(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.set_synonym */
extern void T1516f110(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.new_synonym */
extern T0* T476f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DEFERRED_PROCEDURE.set_synonym */
extern void T476f101(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.new_synonym */
extern T0* T1515f52(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_PROCEDURE.set_synonym */
extern void T1515f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.new_synonym */
extern T0* T469f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DO_PROCEDURE.set_synonym */
extern void T469f104(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_291 */
extern void T456f1578(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_extended_feature_name_comma */
extern T0* T465f180(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_290 */
extern void T456f1577(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_289 */
extern void T456f1576(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_288 */
extern void T456f1575(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.register_procedure */
extern void T456f2251(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].finish */
extern void T715f19(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_PROCEDURE].cursor_finish */
extern void T715f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PROCEDURE].off */
extern T1 T1425f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_PROCEDURE].after */
extern T1 T1425f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_287 */
extern void T456f1574(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_286 */
extern void T456f1573(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.register_query_synonym */
extern void T456f2250(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].back */
extern void T719f22(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].cursor_back */
extern void T719f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].add_traversing_cursor */
extern void T719f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].remove_traversing_cursor */
extern void T719f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].force_left */
extern void T719f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].force_left_cursor */
extern void T719f27(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].put */
extern void T719f31(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].move_cursors_right */
extern void T719f34(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].move_right */
extern void T719f33(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].put_last */
extern void T719f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUERY].index */
extern T6 T1426f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].cursor_index */
extern T6 T719f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].forth */
extern void T719f20(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].cursor_forth */
extern void T719f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].before */
extern T1 T719f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].cursor_before */
extern T1 T719f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUERY].before */
extern T1 T1426f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_query_synonym */
extern T0* T456f957(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EXTERNAL_FUNCTION.new_synonym */
extern T0* T253f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTERNAL_FUNCTION.set_synonym */
extern void T253f113(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.new_synonym */
extern T0* T1513f33(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DEFERRED_FUNCTION.set_synonym */
extern void T1513f104(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.new_synonym */
extern T0* T1512f33(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ONCE_FUNCTION.set_synonym */
extern void T1512f110(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.new_synonym */
extern T0* T1511f33(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DO_FUNCTION.set_synonym */
extern void T1511f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.new_synonym */
extern T0* T371f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_synonym */
extern void T371f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.new_synonym */
extern T0* T369f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_synonym */
extern void T369f99(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.new_synonym */
extern T0* T359f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_synonym */
extern void T359f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.new_synonym */
extern T0* T360f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ATTRIBUTE.set_synonym */
extern void T360f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_285 */
extern void T456f1572(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_284 */
extern void T456f1571(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_283 */
extern void T456f1570(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_282 */
extern void T456f1569(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.register_query */
extern void T456f2249(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].finish */
extern void T719f19(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUERY].cursor_finish */
extern void T719f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUERY].off */
extern T1 T1426f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUERY].after */
extern T1 T1426f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_281 */
extern void T456f1568(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_280 */
extern void T456f1567(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_279 */
extern void T456f1566(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_278 */
extern void T456f1565(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_277 */
extern void T456f1564(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_276 */
extern void T456f1563(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE].force */
extern void T1138f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE].fill_with */
extern void T1140f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE].extend */
extern void T1140f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE].aliased_resized_area */
extern T0* T1138f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE].aliased_resized_area */
extern T0* T1140f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_feature_clause */
extern T0* T465f179(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_FEATURE_CLAUSE.make */
extern T0* T1139c4(GE_context* ac, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.feature_keyword */
extern T0* T73f729(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_feature */
extern T0* T413c90(GE_context* ac);
/* detachable ET_FEATURE_CLAUSE.tokens */
extern T0* T1139f3(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_any_clients */
extern T0* T456f956(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.new_client */
extern T0* T456f954(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_client */
extern T0* T465f339(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.current_system */
extern T0* T456f1013(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_275 */
extern void T456f1562(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_274 */
extern void T456f1561(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_273 */
extern void T456f1560(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_272 */
extern void T456f1559(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE_LIST].force */
extern void T1141f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE_LIST].fill_with */
extern void T1143f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE_LIST].extend */
extern void T1143f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FEATURE_CLAUSE_LIST.put_first */
extern void T1142f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_CLAUSE].force */
extern void T1828f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FEATURE_CLAUSE].fill_with */
extern void T1827f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FEATURE_CLAUSE].extend */
extern void T1827f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FEATURE_CLAUSE_LIST.fixed_array */
extern T0* T1142f3(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_CLAUSE].default_create */
extern T0* T1828c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_271 */
extern void T456f1558(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE_LIST].aliased_resized_area */
extern T0* T1141f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE_LIST].aliased_resized_area */
extern T0* T1143f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_feature_clauses */
extern T0* T465f178(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FEATURE_CLAUSE_LIST.make_with_capacity */
extern T0* T1142c4(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FEATURE_CLAUSE].make */
extern T0* T1828f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_FEATURE_CLAUSE].make_empty */
extern T0* T1827c4(GE_context* ac, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_270 */
extern void T456f1557(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.set_class_features */
extern void T456f2248(GE_context* ac, T0* C);
extern T1 T456f2248ot1(T0* a1, T0** a2);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.set_value */
extern void T372f64(GE_context* ac, T0* C, T11 a1);
/* NATURAL_64.out */
extern T0* T11f20(GE_context* ac, T11* C);
/* [detachable] STRING_8.append_natural_64 */
extern void T17f85(GE_context* ac, T0* C, T11 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_269 */
extern void T456f1556(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_268 */
extern void T456f1555(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_267 */
extern void T456f1554(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_ITEM].force */
extern void T1283f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE_ITEM].fill_with */
extern void T1285f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE_ITEM].extend */
extern void T1285f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_ITEM].aliased_resized_area */
extern T0* T1283f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_TYPE_ITEM].aliased_resized_area */
extern T0* T1285f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_type_comma */
extern T0* T465f177(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_266 */
extern void T456f1553(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_LIST].force */
extern void T1286f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE_LIST].fill_with */
extern void T1288f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE_LIST].extend */
extern void T1288f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BRACED_TYPE_LIST.put_first */
extern void T1388f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_ITEM].force */
extern void T1729f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_TYPE_ITEM].fill_with */
extern void T1728f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_TYPE_ITEM].extend */
extern void T1728f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_265 */
extern void T456f1552(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_LIST].aliased_resized_area */
extern T0* T1286f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_TYPE_LIST].aliased_resized_area */
extern T0* T1288f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_convert_types */
extern T0* T465f176(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] ET_BRACED_TYPE_LIST.set_left_brace */
extern void T1388f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BRACED_TYPE_LIST.make_with_capacity */
extern T0* T1388c9(GE_context* ac, T6 a1);
/* [detachable] ET_BRACED_TYPE_LIST.make_with_capacity */
extern void T1388f9p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_264 */
extern void T456f1551(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_263 */
extern void T456f1550(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_262 */
extern void T456f1549(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_261 */
extern void T456f1548(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE].force */
extern void T1089f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE].fill_with */
extern void T1091f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE].extend */
extern void T1091f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE].aliased_resized_area */
extern T0* T1089f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE].aliased_resized_area */
extern T0* T1091f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_convert_procedure */
extern T0* T465f175(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_CONVERT_PROCEDURE.make */
extern T0* T1509c9(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_CONVERT_PROCEDURE.tokens */
extern T0* T1509f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_260 */
extern void T456f1547(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_convert_function */
extern T0* T465f174(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CONVERT_FUNCTION.make */
extern T0* T1508c8(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_CONVERT_FUNCTION.tokens */
extern T0* T1508f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_259 */
extern void T456f1546(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_ITEM].force */
extern void T1092f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_ITEM].fill_with */
extern void T1094f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_ITEM].extend */
extern void T1094f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_ITEM].aliased_resized_area */
extern T0* T1092f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_ITEM].aliased_resized_area */
extern T0* T1094f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_convert_feature_comma */
extern T0* T465f173(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_258 */
extern void T456f1545(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_LIST].force */
extern void T1095f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_LIST].fill_with */
extern void T1097f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_LIST].extend */
extern void T1097f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONVERT_FEATURE_LIST.put_first */
extern void T1096f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONVERT_FEATURE_ITEM].force */
extern void T1820f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CONVERT_FEATURE_ITEM].fill_with */
extern void T1819f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CONVERT_FEATURE_ITEM].extend */
extern void T1819f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONVERT_FEATURE_LIST.fixed_array */
extern T0* T1096f5(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONVERT_FEATURE_ITEM].default_create */
extern T0* T1820c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_257 */
extern void T456f1544(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_LIST].aliased_resized_area */
extern T0* T1095f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_LIST].aliased_resized_area */
extern T0* T1097f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_convert_features */
extern T0* T465f172(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ET_CONVERT_FEATURE_LIST.make_with_capacity */
extern T0* T1096c8(GE_context* ac, T6 a1);
/* detachable ET_CONVERT_FEATURE_LIST.make_with_capacity */
extern void T1096f8p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONVERT_FEATURE_ITEM].make */
extern T0* T1820f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CONVERT_FEATURE_ITEM].make_empty */
extern T0* T1819c4(GE_context* ac, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.convert_keyword */
extern T0* T73f728(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_convert */
extern T0* T413c89(GE_context* ac);
/* detachable ET_CONVERT_FEATURE_LIST.tokens */
extern T0* T1096f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_256 */
extern void T456f1543(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_255 */
extern void T456f1542(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_254 */
extern void T456f1541(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_253 */
extern void T456f1540(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_252 */
extern void T456f1539(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_251 */
extern void T456f1538(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME_ITEM].force */
extern void T1149f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME_ITEM].fill_with */
extern void T1151f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME_ITEM].extend */
extern void T1151f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME_ITEM].aliased_resized_area */
extern T0* T1149f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME_ITEM].aliased_resized_area */
extern T0* T1151f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_feature_name_comma */
extern T0* T465f169(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_250 */
extern void T456f1537(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR].force */
extern void T1103f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CREATOR].fill_with */
extern void T1105f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CREATOR].extend */
extern void T1105f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CREATOR.put_first */
extern void T1104f17(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CREATOR.fixed_array */
extern T0* T1104f6(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_249 */
extern void T456f1536(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR].aliased_resized_area */
extern T0* T1103f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CREATOR].aliased_resized_area */
extern T0* T1105f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_creator */
extern T0* T465f171(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_CREATOR.make_with_capacity */
extern T0* T1104c16(GE_context* ac, T0* a1, T6 a2);
/* detachable ET_CREATOR.make_feature_name_list_with_capacity */
extern void T1104f18(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CREATOR.tokens */
extern T0* T1104f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_248 */
extern void T456f1535(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_247 */
extern void T456f1534(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_246 */
extern void T456f1533(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_245 */
extern void T456f1532(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_244 */
extern void T456f1531(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_243 */
extern void T456f1530(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_242 */
extern void T456f1529(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_241 */
extern void T456f1528(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_240 */
extern void T456f1527(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_239 */
extern void T456f1526(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_238 */
extern void T456f1525(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_237 */
extern void T456f1524(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_236 */
extern void T456f1523(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_235 */
extern void T456f1522(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR_LIST].force */
extern void T1106f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CREATOR_LIST].fill_with */
extern void T1108f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CREATOR_LIST].extend */
extern void T1108f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR_LIST].aliased_resized_area */
extern T0* T1106f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CREATOR_LIST].aliased_resized_area */
extern T0* T1108f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_234 */
extern void T456f1521(GE_context* ac, T0* C);
/* detachable ET_CREATOR_LIST.put_first */
extern void T1107f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CREATOR].force */
extern void T1822f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CREATOR].fill_with */
extern void T1821f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CREATOR].extend */
extern void T1821f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CREATOR_LIST.fixed_array */
extern T0* T1107f3(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CREATOR].default_create */
extern T0* T1822c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_233 */
extern void T456f1520(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_creators */
extern T0* T465f170(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CREATOR_LIST.make_with_capacity */
extern T0* T1107c9(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CREATOR].make */
extern T0* T1822f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CREATOR].make_empty */
extern T0* T1821c4(GE_context* ac, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_232 */
extern void T456f1519(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_231 */
extern void T456f1518(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_230 */
extern void T456f1517(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_229 */
extern void T456f1516(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_FEATURE_NAME_LIST].force */
extern void T1200f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_KEYWORD_FEATURE_NAME_LIST].fill_with */
extern void T1201f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_KEYWORD_FEATURE_NAME_LIST].extend */
extern void T1201f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_KEYWORD_FEATURE_NAME_LIST.put_first */
extern void T574f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_KEYWORD_FEATURE_NAME_LIST.fixed_array */
extern T0* T574f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_228 */
extern void T456f1515(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_FEATURE_NAME_LIST].aliased_resized_area */
extern T0* T1200f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_KEYWORD_FEATURE_NAME_LIST].aliased_resized_area */
extern T0* T1201f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_keyword_feature_name_list */
extern T0* T465f168(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ET_KEYWORD_FEATURE_NAME_LIST.make_with_capacity */
extern T0* T574c8(GE_context* ac, T6 a1);
/* detachable ET_KEYWORD_FEATURE_NAME_LIST.make_with_capacity */
extern void T574f8p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.redefine_keyword */
extern T0* T73f736(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_redefine */
extern T0* T413c97(GE_context* ac);
/* detachable ET_KEYWORD_FEATURE_NAME_LIST.tokens */
extern T0* T574f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_227 */
extern void T456f1514(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_226 */
extern void T456f1513(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_225 */
extern void T456f1512(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_224 */
extern void T456f1511(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_223 */
extern void T456f1510(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_222 */
extern void T456f1509(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_221 */
extern void T456f1508(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_220 */
extern void T456f1507(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_219 */
extern void T456f1506(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_218 */
extern void T456f1505(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_217 */
extern void T456f1504(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_216 */
extern void T456f1503(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_215 */
extern void T456f1502(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_214 */
extern void T456f1501(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_213 */
extern void T456f1500(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_212 */
extern void T456f1499(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_211 */
extern void T456f1498(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENTS].force */
extern void T1064f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CLIENTS].fill_with */
extern void T1066f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CLIENTS].extend */
extern void T1066f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENTS].aliased_resized_area */
extern T0* T1064f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CLIENTS].aliased_resized_area */
extern T0* T1066f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_clients */
extern T0* T465f167(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_CLIENTS.set_left_brace */
extern void T1065f19(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLIENTS.make_with_capacity */
extern T0* T1065c20(GE_context* ac, T6 a1);
/* detachable ET_CLIENTS.make_with_capacity */
extern void T1065f20p1(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CLIENTS.tokens */
extern T0* T1065f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_210 */
extern void T456f1497(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_209 */
extern void T456f1496(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_208 */
extern void T456f1495(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENT_ITEM].force */
extern void T1062f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CLIENT_ITEM].fill_with */
extern void T1063f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CLIENT_ITEM].extend */
extern void T1063f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENT_ITEM].aliased_resized_area */
extern T0* T1062f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CLIENT_ITEM].aliased_resized_area */
extern T0* T1063f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_client_comma */
extern T0* T456f955(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_client_comma */
extern T0* T465f340(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_207 */
extern void T456f1494(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_206 */
extern void T456f1493(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_205 */
extern void T456f1492(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_204 */
extern void T456f1491(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_203 */
extern void T456f1490(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_202 */
extern void T456f1489(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_none_clients */
extern T0* T465f166(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CLIENTS.make */
extern T0* T1065c17(GE_context* ac);
/* detachable ET_CLIENTS.make */
extern void T1065f17p1(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_201 */
extern void T456f1488(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_1_200 */
extern void T456f1267(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_200 */
extern void T456f1487(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_199 */
extern void T456f1486(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_EXPORT].force */
extern void T1144f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FEATURE_EXPORT].fill_with */
extern void T1146f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FEATURE_EXPORT].extend */
extern void T1146f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FEATURE_EXPORT.put_first */
extern void T1145f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FEATURE_EXPORT.fixed_array */
extern T0* T1145f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_198 */
extern void T456f1485(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_EXPORT].aliased_resized_area */
extern T0* T1144f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FEATURE_EXPORT].aliased_resized_area */
extern T0* T1146f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_feature_export */
extern T0* T465f164(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ET_FEATURE_EXPORT.make_with_capacity */
extern T0* T1145c10(GE_context* ac, T0* a1, T6 a2);
/* detachable ET_FEATURE_EXPORT.make_feature_name_list_with_capacity */
extern void T1145f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_197 */
extern void T456f1484(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_196 */
extern void T456f1483(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT].force */
extern void T1124f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXPORT].fill_with */
extern void T1126f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXPORT].extend */
extern void T1126f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT].aliased_resized_area */
extern T0* T1124f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_EXPORT].aliased_resized_area */
extern T0* T1126f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_null_export */
extern T0* T465f165(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_195 */
extern void T456f1482(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_194 */
extern void T456f1481(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_193 */
extern void T456f1480(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_192 */
extern void T456f1479(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_all_export */
extern T0* T465f163(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ALL_EXPORT.make */
extern T0* T1506c8(GE_context* ac, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.all_keyword */
extern T0* T73f1528(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_all */
extern T0* T413c57(GE_context* ac);
/* detachable ET_ALL_EXPORT.tokens */
extern T0* T1506f3(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_191 */
extern void T456f1478(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT_LIST].force */
extern void T1127f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXPORT_LIST].fill_with */
extern void T1128f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_EXPORT_LIST].extend */
extern void T1128f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT_LIST].aliased_resized_area */
extern T0* T1127f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_EXPORT_LIST].aliased_resized_area */
extern T0* T1128f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_190 */
extern void T456f1477(GE_context* ac, T0* C);
/* detachable ET_EXPORT_LIST.put_first */
extern void T573f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPORT].force */
extern void T1780f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_EXPORT].fill_with */
extern void T1779f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_EXPORT].extend */
extern void T1779f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXPORT_LIST.fixed_array */
extern T0* T573f5(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPORT].default_create */
extern T0* T1780c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_189 */
extern void T456f1476(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_exports */
extern T0* T465f162(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ET_EXPORT_LIST.make_with_capacity */
extern T0* T573c7(GE_context* ac, T6 a1);
/* detachable ET_EXPORT_LIST.make_with_capacity */
extern void T573f7p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPORT].make */
extern T0* T1780f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_EXPORT].make_empty */
extern T0* T1779c4(GE_context* ac, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.export_keyword */
extern T0* T73f724(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_export */
extern T0* T413c85(GE_context* ac);
/* detachable ET_EXPORT_LIST.tokens */
extern T0* T573f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_188 */
extern void T456f1475(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_187 */
extern void T456f1474(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_186 */
extern void T456f1473(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_185 */
extern void T456f1472(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_184 */
extern void T456f1471(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_183 */
extern void T456f1470(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_ITEM].force */
extern void T1258f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_RENAME_ITEM].fill_with */
extern void T1260f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_RENAME_ITEM].extend */
extern void T1260f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_ITEM].aliased_resized_area */
extern T0* T1258f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_RENAME_ITEM].aliased_resized_area */
extern T0* T1260f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_rename_comma */
extern T0* T465f161(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_182 */
extern void T456f1469(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_rename */
extern T0* T465f160(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_181 */
extern void T456f1468(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_LIST].force */
extern void T1261f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_RENAME_LIST].fill_with */
extern void T1262f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_RENAME_LIST].extend */
extern void T1262f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_RENAME_LIST.put_first */
extern void T572f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME_ITEM].force */
extern void T1777f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_RENAME_ITEM].fill_with */
extern void T1776f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_RENAME_ITEM].extend */
extern void T1776f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_RENAME_LIST.fixed_array */
extern T0* T572f5(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME_ITEM].default_create */
extern T0* T1777c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_180 */
extern void T456f1467(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_LIST].aliased_resized_area */
extern T0* T1261f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_RENAME_LIST].aliased_resized_area */
extern T0* T1262f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_renames */
extern T0* T465f159(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ET_RENAME_LIST.make_with_capacity */
extern T0* T572c8(GE_context* ac, T6 a1);
/* detachable ET_RENAME_LIST.make_with_capacity */
extern void T572f8p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RENAME_ITEM].make */
extern T0* T1777f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_RENAME_ITEM].make_empty */
extern T0* T1776c4(GE_context* ac, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.rename_keyword */
extern T0* T73f725(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_rename */
extern T0* T413c86(GE_context* ac);
/* detachable ET_RENAME_LIST.tokens */
extern T0* T572f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_179 */
extern void T456f1466(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_178 */
extern void T456f1465(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_177 */
extern void T456f1464(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_176 */
extern void T456f1463(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_175 */
extern void T456f1462(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_ITEM].force */
extern void T1239f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PARENT_ITEM].fill_with */
extern void T1240f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PARENT_ITEM].extend */
extern void T1240f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_ITEM].aliased_resized_area */
extern T0* T1239f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_PARENT_ITEM].aliased_resized_area */
extern T0* T1240f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_parent_semicolon */
extern T0* T465f158(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_174 */
extern void T456f1461(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_LIST].force */
extern void T1241f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PARENT_LIST].fill_with */
extern void T1242f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PARENT_LIST].extend */
extern void T1242f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_173 */
extern void T456f1460(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_172 */
extern void T456f1459(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_LIST].aliased_resized_area */
extern T0* T1241f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_PARENT_LIST].aliased_resized_area */
extern T0* T1242f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_parents */
extern T0* T465f157(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_171 */
extern void T456f1458(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_170 */
extern void T456f1457(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT].force */
extern void T1234f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PARENT].fill_with */
extern void T1235f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PARENT].extend */
extern void T1235f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT].aliased_resized_area */
extern T0* T1234f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_PARENT].aliased_resized_area */
extern T0* T1235f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.new_parent */
extern T0* T456f953(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8);
/* [detachable] ET_AST_FACTORY.new_parent */
extern T0* T465f338(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_CLASS.has_formal_parameter */
extern T1 T74f194(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.has_formal_parameter */
extern T1 T603f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_169 */
extern void T456f1456(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_168 */
extern void T456f1455(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_167 */
extern void T456f1454(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_166 */
extern void T456f1453(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_165 */
extern void T456f1452(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_164 */
extern void T456f1451(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_163 */
extern void T456f1450(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_LIST.set_clients_clause */
extern void T568f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_162 */
extern void T456f1449(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_161 */
extern void T456f1448(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_CLAUSE_LIST].force */
extern void T1236f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PARENT_CLAUSE_LIST].fill_with */
extern void T1238f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PARENT_CLAUSE_LIST].extend */
extern void T1238f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_CLAUSE_LIST].aliased_resized_area */
extern T0* T1236f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_PARENT_CLAUSE_LIST].aliased_resized_area */
extern T0* T1238f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_160 */
extern void T456f1447(GE_context* ac, T0* C);
/* detachable ET_PARENT_CLAUSE_LIST.put_first */
extern void T1237f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_LIST].force */
extern void T1841f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_PARENT_LIST].fill_with */
extern void T1840f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_PARENT_LIST].extend */
extern void T1840f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PARENT_CLAUSE_LIST.fixed_array */
extern T0* T1237f3(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_LIST].default_create */
extern T0* T1841c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_159 */
extern void T456f1446(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_parent_clauses */
extern T0* T465f156(GE_context* ac, T0* C, T6 a1);
/* detachable ET_PARENT_CLAUSE_LIST.make_with_capacity */
extern T0* T1237c6(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_PARENT_LIST].make */
extern T0* T1841f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_PARENT_LIST].make_empty */
extern T0* T1840c4(GE_context* ac, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_158 */
extern void T456f1445(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_157 */
extern void T456f1444(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_156 */
extern void T456f1443(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBSOLETE].force */
extern void T1228f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_OBSOLETE].fill_with */
extern void T1230f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_OBSOLETE].extend */
extern void T1230f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBSOLETE].aliased_resized_area */
extern T0* T1228f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_OBSOLETE].aliased_resized_area */
extern T0* T1230f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_obsolete_message */
extern T0* T465f155(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_155 */
extern void T456f1442(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_154 */
extern void T456f1441(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].force */
extern void T1074f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].fill_with */
extern void T1076f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].extend */
extern void T1076f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].aliased_resized_area */
extern T0* T1074f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].aliased_resized_area */
extern T0* T1076f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_constraint_labeled_actual_parameter_semicolon */
extern T0* T465f154(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_constraint_labeled_actual_parameter */
extern T0* T465f152(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CONSTRAINT_LABELED_ACTUAL_PARAMETER.make */
extern T0* T1504c6(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_153 */
extern void T456f1440(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_152 */
extern void T456f1439(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST].force */
extern void T1077f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST].fill_with */
extern void T1079f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST].extend */
extern void T1079f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.add_to_constraint_actual_parameter_list */
extern void T456f2247(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST.put_first */
extern void T1078f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].force */
extern void T1817f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].fill_with */
extern void T1815f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].extend */
extern void T1815f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST.fixed_array */
extern T0* T1078f8(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].default_create */
extern T0* T1817c2(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_constraint_labeled_comma_actual_parameter */
extern T0* T465f153(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CONSTRAINT_LABELED_COMMA_ACTUAL_PARAMETER.make */
extern T0* T1505c6(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_CONSTRAINT_NAMED_TYPE.type */
extern T0* T1503f4(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_GENERIC_NAMED_TYPE.type */
extern T0* T1755f5(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_LABELED_COMMA_ACTUAL_PARAMETER.type */
extern T0* T1505f3(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_LABELED_ACTUAL_PARAMETER.type */
extern T0* T1504f3(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST.first */
extern T0* T1078f7(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST.is_empty */
extern T1 T1078f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_151 */
extern void T456f1438(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_150 */
extern void T456f1437(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_149 */
extern void T456f1436(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST].aliased_resized_area */
extern T0* T1077f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST].aliased_resized_area */
extern T0* T1079f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_constraint_actual_parameters */
extern T0* T465f150(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST.set_left_bracket */
extern void T1078f14(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST.make_with_capacity */
extern T0* T1078c12(GE_context* ac, T6 a1);
/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST.make_with_capacity */
extern void T1078f12p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].make */
extern T0* T1817f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].make_empty */
extern T0* T1815c4(GE_context* ac, T6 a1);
/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST.tokens */
extern T0* T1078f6(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_148 */
extern void T456f1435(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_147 */
extern void T456f1434(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_146 */
extern void T456f1433(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_145 */
extern void T456f1432(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_144 */
extern void T456f1431(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_143 */
extern void T456f1430(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_constraint_actual_parameter_comma */
extern T0* T465f151(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_142 */
extern void T456f1429(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.new_constraint_named_type */
extern T0* T456f952(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_AST_FACTORY.new_constraint_named_type */
extern T0* T465f335(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CONSTRAINT_NAMED_TYPE.make */
extern T0* T1503c7(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_constraint_generic_named_type */
extern T0* T465f334(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CONSTRAINT_GENERIC_NAMED_TYPE.make */
extern T0* T1755c8(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_141 */
extern void T456f1428(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_140 */
extern void T456f1427(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_139 */
extern void T456f1426(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_138 */
extern void T456f1425(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_137 */
extern void T456f1424(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_136 */
extern void T456f1423(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_135 */
extern void T456f1422(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_TYPE].force */
extern void T1086f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_TYPE].fill_with */
extern void T1088f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_TYPE].extend */
extern void T1088f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_TYPE].aliased_resized_area */
extern T0* T1086f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_TYPE].aliased_resized_area */
extern T0* T1088f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_134 */
extern void T456f1421(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_133 */
extern void T456f1420(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_132 */
extern void T456f1419(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_131 */
extern void T456f1418(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_130 */
extern void T456f1417(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_129 */
extern void T456f1416(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_128 */
extern void T456f1415(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_127 */
extern void T456f1414(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_126 */
extern void T456f1413(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_125 */
extern void T456f1412(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_124 */
extern void T456f1411(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_123 */
extern void T456f1410(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_122 */
extern void T456f1409(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_121 */
extern void T456f1408(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_120 */
extern void T456f1407(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_119 */
extern void T456f1406(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_118 */
extern void T456f1405(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_117 */
extern void T456f1404(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_116 */
extern void T456f1403(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_115 */
extern void T456f1402(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_114 */
extern void T456f1401(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_113 */
extern void T456f1400(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_112 */
extern void T456f1399(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_111 */
extern void T456f1398(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_110 */
extern void T456f1397(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_109 */
extern void T456f1396(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_108 */
extern void T456f1395(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_107 */
extern void T456f1394(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_106 */
extern void T456f1393(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_105 */
extern void T456f1392(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_104 */
extern void T456f1391(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_103 */
extern void T456f1390(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_102 */
extern void T456f1389(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_101 */
extern void T456f1388(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_100 */
extern void T456f1387(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_99 */
extern void T456f1386(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_98 */
extern void T456f1385(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_97 */
extern void T456f1384(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_96 */
extern void T456f1383(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_95 */
extern void T456f1382(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_94 */
extern void T456f1381(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_93 */
extern void T456f1380(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_92 */
extern void T456f1379(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_91 */
extern void T456f1378(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_CREATOR].force */
extern void T1080f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_CREATOR].fill_with */
extern void T1082f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_CREATOR].extend */
extern void T1082f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINT_CREATOR.put_first */
extern void T1081f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINT_CREATOR.fixed_array */
extern T0* T1081f6(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_90 */
extern void T456f1377(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_CREATOR].aliased_resized_area */
extern T0* T1080f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_CREATOR].aliased_resized_area */
extern T0* T1082f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_constraint_creator */
extern T0* T465f148(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_CONSTRAINT_CREATOR.make_with_capacity */
extern T0* T1081c12(GE_context* ac, T6 a1);
/* detachable ET_CONSTRAINT_CREATOR.make_with_capacity */
extern void T1081f12p1(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONSTRAINT_CREATOR.tokens */
extern T0* T1081f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_89 */
extern void T456f1376(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_88 */
extern void T456f1375(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_87 */
extern void T456f1374(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_86 */
extern void T456f1373(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_85 */
extern void T456f1372(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_84 */
extern void T456f1371(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_83 */
extern void T456f1370(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_RENAME_LIST].force */
extern void T1083f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_RENAME_LIST].fill_with */
extern void T1085f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_RENAME_LIST].extend */
extern void T1085f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINT_RENAME_LIST.put_first */
extern void T1084f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINT_RENAME_LIST.fixed_array */
extern T0* T1084f6(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_82 */
extern void T456f1369(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_RENAME_LIST].aliased_resized_area */
extern T0* T1083f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_RENAME_LIST].aliased_resized_area */
extern T0* T1085f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_constraint_renames */
extern T0* T465f147(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_CONSTRAINT_RENAME_LIST.make_with_capacity */
extern T0* T1084c12(GE_context* ac, T6 a1);
/* detachable ET_CONSTRAINT_RENAME_LIST.make_with_capacity */
extern void T1084f12p1(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONSTRAINT_RENAME_LIST.make_with_capacity */
extern void T1084f12p0(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONSTRAINT_RENAME_LIST.tokens */
extern T0* T1084f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_81 */
extern void T456f1368(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_80 */
extern void T456f1367(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_79 */
extern void T456f1366(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_78 */
extern void T456f1365(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_77 */
extern void T456f1364(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_ITEM].force */
extern void T1277f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_ITEM].fill_with */
extern void T1279f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_ITEM].extend */
extern void T1279f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_ITEM].aliased_resized_area */
extern T0* T1277f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_ITEM].aliased_resized_area */
extern T0* T1279f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_type_constraint_comma */
extern T0* T465f146(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_76 */
extern void T456f1363(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_LIST].force */
extern void T1280f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_LIST].fill_with */
extern void T1282f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_LIST].extend */
extern void T1282f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_TYPE_CONSTRAINT_LIST.put_first */
extern void T1281f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_CONSTRAINT_ITEM].force */
extern void T1844f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_TYPE_CONSTRAINT_ITEM].fill_with */
extern void T1843f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_TYPE_CONSTRAINT_ITEM].extend */
extern void T1843f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_TYPE_CONSTRAINT_LIST.fixed_array */
extern T0* T1281f6(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_CONSTRAINT_ITEM].default_create */
extern T0* T1844c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_75 */
extern void T456f1362(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_LIST].aliased_resized_area */
extern T0* T1280f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_LIST].aliased_resized_area */
extern T0* T1282f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_type_constraint_list */
extern T0* T465f145(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_TYPE_CONSTRAINT_LIST.set_left_brace */
extern void T1281f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_TYPE_CONSTRAINT_LIST.make_with_capacity */
extern T0* T1281c11(GE_context* ac, T6 a1);
/* detachable ET_TYPE_CONSTRAINT_LIST.make_with_capacity */
extern void T1281f11p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_CONSTRAINT_ITEM].make */
extern T0* T1844f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_TYPE_CONSTRAINT_ITEM].make_empty */
extern T0* T1843c4(GE_context* ac, T6 a1);
/* detachable ET_TYPE_CONSTRAINT_LIST.tokens */
extern T0* T1281f5(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_74 */
extern void T456f1361(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_73 */
extern void T456f1360(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_72 */
extern void T456f1359(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT].force */
extern void T1274f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT].fill_with */
extern void T1276f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT].extend */
extern void T1276f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT].aliased_resized_area */
extern T0* T1274f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT].aliased_resized_area */
extern T0* T1276f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.register_constraint */
extern void T456f2246(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE].force_last */
extern void T947f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE].resize */
extern void T947f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE].new_capacity */
extern T6 T947f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE].extendible */
extern T1 T947f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_type_rename_constraint */
extern T0* T465f144(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_TYPE_RENAME_CONSTRAINT.make */
extern T0* T1502c8(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.dummy_constraint */
extern T0* T456f951(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_71 */
extern void T456f1358(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_70 */
extern void T456f1357(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER].force */
extern void T1159f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER].fill_with */
extern void T1160f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER].extend */
extern void T1160f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER].aliased_resized_area */
extern T0* T1159f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER].aliased_resized_area */
extern T0* T1160f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_constrained_formal_parameter */
extern T0* T465f143(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.set_type_mark */
extern void T1501f78(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.make */
extern T0* T1501c77(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.make_unconstrained */
extern void T1501f80(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TOKEN_CONSTANTS.arrow_symbol */
extern T0* T73f701(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_arrow */
extern T0* T341c117(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_69 */
extern void T456f1356(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_68 */
extern void T456f1355(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_67 */
extern void T456f1354(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_66 */
extern void T456f1353(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_65 */
extern void T456f1352(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_64 */
extern void T456f1351(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_formal_parameter */
extern T0* T465f142(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_FORMAL_PARAMETER.set_type_mark */
extern void T891f79(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.make */
extern T0* T891c78(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_63 */
extern void T456f1350(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_62 */
extern void T456f1349(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_61 */
extern void T456f1348(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_ITEM].force */
extern void T1161f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_ITEM].fill_with */
extern void T1163f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_ITEM].extend */
extern void T1163f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_ITEM].aliased_resized_area */
extern T0* T1161f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_ITEM].aliased_resized_area */
extern T0* T1163f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_formal_parameter_comma */
extern T0* T465f141(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_60 */
extern void T456f1347(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_LIST].force */
extern void T1164f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_LIST].fill_with */
extern void T1165f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_LIST].extend */
extern void T1165f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.put_first */
extern void T603f28(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.set_index */
extern void T1501f79(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FORMAL_PARAMETER.set_index */
extern void T891f80(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FORMAL_PARAMETER_LIST.capacity */
extern T6 T603f9(GE_context* ac, T0* C);
/* detachable ET_FORMAL_PARAMETER_LIST.put_first */
extern void T603f28p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_PARAMETER_ITEM].force */
extern void T1785f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FORMAL_PARAMETER_ITEM].fill_with */
extern void T1784f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FORMAL_PARAMETER_ITEM].extend */
extern void T1784f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.fixed_array */
extern T0* T603f8(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_PARAMETER_ITEM].default_create */
extern T0* T1785c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_59 */
extern void T456f1346(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_LIST].aliased_resized_area */
extern T0* T1164f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_LIST].aliased_resized_area */
extern T0* T1165f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_formal_parameters */
extern T0* T465f140(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* detachable ET_FORMAL_PARAMETER_LIST.set_left_bracket */
extern void T603f29(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.make_with_capacity */
extern T0* T603c26(GE_context* ac, T6 a1);
/* detachable ET_FORMAL_PARAMETER_LIST.make_with_capacity */
extern void T603f26p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_PARAMETER_ITEM].make */
extern T0* T1785f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_FORMAL_PARAMETER_ITEM].make_empty */
extern T0* T1784c4(GE_context* ac, T6 a1);
/* detachable ET_FORMAL_PARAMETER_LIST.tokens */
extern T0* T603f6(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_58 */
extern void T456f1345(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_57 */
extern void T456f1344(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.set_formal_parameters */
extern void T456f2243(GE_context* ac, T0* C, T0* a1);
extern T1 T456f2243ot1(T0* a1, T0** a2);
extern T1 T456f2243ot2(T0* a1, T0** a2);
extern T1 T456f2243ot3(T0* a1, T0** a2);
extern T1 T456f2243ot4(T0* a1, T0** a2);
extern T1 T456f2243ot5(T0* a1, T0** a2);
/* [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE].wipe_out */
extern void T947f11(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_TYPE].keep_head */
extern void T1088f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_TYPE].set_count */
extern void T1088f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE].move_all_cursors_after */
extern void T947f14(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_CONSTRAINT_TYPE].set_next_cursor */
extern void T1650f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_CONSTRAINT_TYPE].set_after */
extern void T1650f6(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_formal_parameters */
extern void T74f305(GE_context* ac, T0* C, T0* a1);
extern T1 T74f305ot1(T0* a1);
/* detachable ET_TYPE_CONSTRAINT_LIST.put */
extern void T1281f15(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ET_TYPE_CONSTRAINT_LIST.remove */
extern void T1281f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_TYPE_CONSTRAINT_ITEM].keep_head */
extern void T1844f4(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_TYPE_CONSTRAINT_ITEM].keep_head */
extern void T1843f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_TYPE_CONSTRAINT_ITEM].set_count */
extern void T1843f9(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.set_constraint */
extern void T1501f81(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.remove */
extern void T603f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_FORMAL_PARAMETER_ITEM].keep_head */
extern void T1785f3(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_FORMAL_PARAMETER_ITEM].keep_head */
extern void T1784f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_FORMAL_PARAMETER_ITEM].set_count */
extern void T1784f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_TYPE_RENAME_CONSTRAINT.set_type */
extern void T1502f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.resolved_syntactical_constraint */
extern T0* T1524f23(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CONSTRAINT_NAMED_TYPE.resolved_syntactical_constraint */
extern T0* T1503f3(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.resolved_constraint_named_type */
extern T0* T456f1032(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T456f1032ot1(T0* a1, T0** a2);
/* detachable ET_CONSTRAINT_GENERIC_NAMED_TYPE.resolved_syntactical_constraint */
extern T0* T1755f4(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.resolved_constraint_generic_named_type */
extern T0* T456f1036(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T456f1036ot1(T0* a1, T0** a2);
/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST.resolved_syntactical_constraint */
extern T0* T1078f10(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.resolved_constraint_actual_parameter_list */
extern T0* T456f1037(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_LIKE_FEATURE.resolved_syntactical_constraint_with_type */
extern T0* T1524f83(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CONSTRAINT_NAMED_TYPE.resolved_syntactical_constraint_with_type */
extern T0* T1503f6(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CONSTRAINT_GENERIC_NAMED_TYPE.resolved_syntactical_constraint_with_type */
extern T0* T1755f7(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.resolved_syntactical_constraint_with_type */
extern T0* T1526f80(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.resolved_syntactical_constraint_with_type */
extern T0* T1525f83(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIKE_CURRENT.resolved_syntactical_constraint_with_type */
extern T0* T449f77(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CONSTRAINT_LABELED_COMMA_ACTUAL_PARAMETER.resolved_syntactical_constraint_with_type */
extern T0* T1505f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.resolved_constraint_labeled_comma_actual_parameter */
extern T0* T456f1039(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CONSTRAINT_LABELED_ACTUAL_PARAMETER.resolved_syntactical_constraint_with_type */
extern T0* T1504f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.resolved_constraint_labeled_actual_parameter */
extern T0* T456f1038(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST.item */
extern T0* T1078f11(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST.position */
extern T0* T1078f9(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_NAMED_TYPE.position */
extern T0* T1503f5(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_GENERIC_NAMED_TYPE.position */
extern T0* T1755f6(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_LABELED_COMMA_ACTUAL_PARAMETER.position */
extern T0* T1505f4(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_LABELED_ACTUAL_PARAMETER.position */
extern T0* T1504f4(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.resolved_syntactical_constraint */
extern T0* T1526f33(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.resolved_syntactical_constraint */
extern T0* T1525f29(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_LIKE_CURRENT.resolved_syntactical_constraint */
extern T0* T449f31(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE].item */
extern T0* T947f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_56 */
extern void T456f1343(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_55 */
extern void T456f1342(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_54 */
extern void T456f1341(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_53 */
extern void T456f1340(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_52 */
extern void T456f1339(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_51 */
extern void T456f1338(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_50 */
extern void T456f1337(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLASS].force */
extern void T1060f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CLASS].fill_with */
extern void T1061f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CLASS].extend */
extern void T1061f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLASS].aliased_resized_area */
extern T0* T1060f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CLASS].aliased_resized_area */
extern T0* T1061f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS.set_external_keyword */
extern void T74f302(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_frozen_keyword */
extern void T74f301(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_class_mark */
extern void T74f316(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_class_keyword */
extern void T74f300(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.new_class */
extern T0* T456f949(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_time_stamp */
extern void T74f318(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_MASTER_CLASS.add_last_local_class */
extern void T570f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.add_last_local_non_override_class */
extern void T570f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].force_last */
extern void T462f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.add_last_local_override_class */
extern void T570f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.add_last_local_ignored_class */
extern void T570f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.register_class */
extern void T58f240(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_name */
extern void T74f317(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.reset */
extern void T74f290(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_parsed */
extern void T74f291(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_ancestors_built */
extern void T74f299(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_features_flattened */
extern void T74f298(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_interface_checked */
extern void T74f297(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_implementation_checked */
extern void T74f296(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.set_modified */
extern void T570f61(GE_context* ac, T0* C, T1 a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.same_pathnames */
extern T1 T126f27(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_PATHNAME.same_pathname */
extern T1 T483f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_PATHNAME.is_subpathname */
extern T1 T483f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_PATHNAME.string_ */
extern T0* T483f18(GE_context* ac, T0* C);
/* [detachable] KL_PATHNAME.string_equality_tester */
extern T0* T483f19(GE_context* ac, T0* C);
/* [detachable] UC_STRING_EQUALITY_TESTER.default_create */
extern T0* T896c3(GE_context* ac);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.same_pathnames */
extern T1 T125f30(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_PATHNAME.same_case_insensitive */
extern T1 T483f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_PATHNAME.is_case_insensitive_subpathname */
extern T1 T483f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_PATHNAME.case_insensitive_string_equality_tester */
extern T0* T483f17(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.file_system */
extern T0* T456f1012(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.unix_file_system */
extern T0* T456f1020(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.windows_file_system */
extern T0* T456f1019(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.operating_system */
extern T0* T456f1018(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvscn1a_error */
extern void T53f269(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvscn1a */
extern T0* T785c901(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_gvscn1a_error */
extern void T36f269(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_MASTER_CLASS.first_local_class */
extern T0* T570f28(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.has_local_class */
extern T1 T570f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.has_local_hidden_class */
extern T1 T570f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.has_local_ignored_class */
extern T1 T570f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].has */
extern T1 T462f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.has_local_non_override_class */
extern T1 T570f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.has_local_override_class */
extern T1 T570f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_49 */
extern void T456f1336(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_48 */
extern void T456f1335(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_47 */
extern void T456f1334(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_46 */
extern void T456f1333(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_ITEM].force */
extern void T1180f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_ITEM].fill_with */
extern void T1182f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_ITEM].extend */
extern void T1182f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_ITEM].aliased_resized_area */
extern T0* T1180f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_ITEM].aliased_resized_area */
extern T0* T1182f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_indexing_term_comma */
extern T0* T465f139(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_45 */
extern void T456f1332(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM].force */
extern void T1178f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM].fill_with */
extern void T1179f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM].extend */
extern void T1179f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM].aliased_resized_area */
extern T0* T1178f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM].aliased_resized_area */
extern T0* T1179f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_custom_attribute */
extern T0* T465f138(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_CUSTOM_ATTRIBUTE.make */
extern T0* T1500c9(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_CUSTOM_ATTRIBUTE.tokens */
extern T0* T1500f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_44 */
extern void T456f1331(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_43 */
extern void T456f1330(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_42 */
extern void T456f1329(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_41 */
extern void T456f1328(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_40 */
extern void T456f1327(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_39 */
extern void T456f1326(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_38 */
extern void T456f1325(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_37 */
extern void T456f1324(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_LIST].force */
extern void T1183f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_LIST].fill_with */
extern void T1185f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_LIST].extend */
extern void T1185f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INDEXING_TERM_LIST.put_first */
extern void T1184f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM_ITEM].force */
extern void T1833f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INDEXING_TERM_ITEM].fill_with */
extern void T1832f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INDEXING_TERM_ITEM].extend */
extern void T1832f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INDEXING_TERM_LIST.fixed_array */
extern T0* T1184f3(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM_ITEM].default_create */
extern T0* T1833c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_36 */
extern void T456f1323(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_LIST].aliased_resized_area */
extern T0* T1183f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_LIST].aliased_resized_area */
extern T0* T1185f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_indexing_terms */
extern T0* T465f137(GE_context* ac, T0* C, T6 a1);
/* detachable ET_INDEXING_TERM_LIST.make_with_capacity */
extern T0* T1184c9(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM_ITEM].make */
extern T0* T1833f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_INDEXING_TERM_ITEM].make_empty */
extern T0* T1832c4(GE_context* ac, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_35 */
extern void T456f1322(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_ITEM].force */
extern void T1175f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INDEXING_ITEM].fill_with */
extern void T1177f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INDEXING_ITEM].extend */
extern void T1177f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_indexing_semicolon */
extern T0* T465f135(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_34 */
extern void T456f1321(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_33 */
extern void T456f1320(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_ITEM].aliased_resized_area */
extern T0* T1175f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_INDEXING_ITEM].aliased_resized_area */
extern T0* T1177f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_tagged_indexing */
extern T0* T465f134(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_TAGGED_INDEXING.make */
extern T0* T1498c5(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_32 */
extern void T456f1319(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_indexing */
extern T0* T465f136(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INDEXING.make */
extern T0* T1499c5(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_31 */
extern void T456f1318(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_30 */
extern void T456f1317(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_LIST].force */
extern void T1172f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INDEXING_LIST].fill_with */
extern void T1174f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_INDEXING_LIST].extend */
extern void T1174f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_LIST].aliased_resized_area */
extern T0* T1172f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_INDEXING_LIST].aliased_resized_area */
extern T0* T1174f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_29 */
extern void T456f1316(GE_context* ac, T0* C);
/* detachable ET_INDEXING_LIST.put_first */
extern void T1173f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_ITEM].force */
extern void T1831f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INDEXING_ITEM].fill_with */
extern void T1830f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INDEXING_ITEM].extend */
extern void T1830f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INDEXING_LIST.fixed_array */
extern T0* T1173f5(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_ITEM].default_create */
extern T0* T1831c2(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_28 */
extern void T456f1315(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_27 */
extern void T456f1314(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_26 */
extern void T456f1313(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_indexings */
extern T0* T465f132(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ET_INDEXING_LIST.set_indexing_keyword */
extern void T1173f14(GE_context* ac, T0* C, T0* a1);
/* detachable ET_INDEXING_LIST.make_with_capacity */
extern T0* T1173c12(GE_context* ac, T6 a1);
/* detachable ET_INDEXING_LIST.make_with_capacity */
extern void T1173f12p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_ITEM].make */
extern T0* T1831f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_INDEXING_ITEM].make_empty */
extern T0* T1830c4(GE_context* ac, T6 a1);
/* [detachable] ET_TOKEN_CONSTANTS.indexing_keyword */
extern T0* T73f734(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_indexing */
extern T0* T413c95(GE_context* ac);
/* detachable ET_INDEXING_LIST.tokens */
extern T0* T1173f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_25 */
extern void T456f1312(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_24 */
extern void T456f1311(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_23 */
extern void T456f1310(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_22 */
extern void T456f1309(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_21 */
extern void T456f1308(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_20 */
extern void T456f1307(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_19 */
extern void T456f1306(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_18 */
extern void T456f1305(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_17 */
extern void T456f1304(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_16 */
extern void T456f1303(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_15 */
extern void T456f1302(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_14 */
extern void T456f1301(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_13 */
extern void T456f1300(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_12 */
extern void T456f1299(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_11 */
extern void T456f1298(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_10 */
extern void T456f1297(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_9 */
extern void T456f1296(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_8 */
extern void T456f1295(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.set_class_providers */
extern void T456f2237(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].wipe_out */
extern void T948f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].slots_wipe_out */
extern void T948f53(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].clashes_wipe_out */
extern void T948f52(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].key_storage_wipe_out */
extern void T948f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].item_storage_wipe_out */
extern void T948f50(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_NAMED_CLASS].keep_head */
extern void T1651f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_NAMED_CLASS].set_count */
extern void T1651f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].move_all_cursors_after */
extern void T948f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].forth */
extern void T948f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].put_last */
extern void T948f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].compress */
extern void T948f57(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].key_storage_keep_head */
extern void T948f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].item_storage_keep_head */
extern void T948f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].move_all_cursors */
extern void T948f64(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].key_storage_put */
extern void T948f63(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].item_for_iteration */
extern T0* T948f17(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].after */
extern T1 T948f20(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].cursor_after */
extern T1 T948f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].start */
extern void T948f39(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_providers */
extern void T74f304(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].make */
extern T0* T948c36(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].set_internal_cursor */
extern void T948f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].make_slots */
extern void T948f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].make_clashes */
extern void T948f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].make_key_storage */
extern void T948f43(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].make_item_storage */
extern void T948f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_CLASS].make */
extern T0* T1653f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_NAMED_CLASS].make_empty */
extern T0* T1651c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NAMED_CLASS].default_create */
extern T0* T1653c3(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_7 */
extern void T456f1294(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.set_class_to_end */
extern void T456f2236(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9);
/* [detachable] ET_CLASS.set_parsed */
extern void T74f274(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_end_keyword */
extern void T74f315(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_second_indexing */
extern void T74f314(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_invariants */
extern void T74f313(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_feature_clauses */
extern void T74f312(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_convert_features */
extern void T74f311(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_creators */
extern void T74f310(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_parent_clauses */
extern void T74f309(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_obsolete_message */
extern void T74f308(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_6 */
extern void T456f1293(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_5 */
extern void T456f1292(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_4 */
extern void T456f1291(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_3 */
extern void T456f1290(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_first_indexing */
extern void T74f307(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_2 */
extern void T456f1289(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_leading_break */
extern void T74f306(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_action_1 */
extern void T456f1288(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_push_last_value */
extern void T456f1242(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUESTION_MARK_SYMBOL].force */
extern void T1004f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_QUESTION_MARK_SYMBOL].fill_with */
extern void T1006f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_QUESTION_MARK_SYMBOL].extend */
extern void T1006f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUESTION_MARK_SYMBOL].aliased_resized_area */
extern T0* T1004f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_QUESTION_MARK_SYMBOL].aliased_resized_area */
extern T0* T1006f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_SYMBOL].force */
extern void T1002f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_BRACKET_SYMBOL].fill_with */
extern void T1003f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_BRACKET_SYMBOL].extend */
extern void T1003f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_SYMBOL].aliased_resized_area */
extern T0* T1002f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_BRACKET_SYMBOL].aliased_resized_area */
extern T0* T1003f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_VOID].force */
extern void T996f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_VOID].fill_with */
extern void T998f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_VOID].extend */
extern void T998f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_VOID].aliased_resized_area */
extern T0* T996f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_VOID].aliased_resized_area */
extern T0* T998f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL_OPERATOR].force */
extern void T993f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_SYMBOL_OPERATOR].fill_with */
extern void T995f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_SYMBOL_OPERATOR].extend */
extern void T995f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL_OPERATOR].aliased_resized_area */
extern T0* T993f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_SYMBOL_OPERATOR].aliased_resized_area */
extern T0* T995f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RETRY_INSTRUCTION].force */
extern void T990f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_RETRY_INSTRUCTION].fill_with */
extern void T992f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_RETRY_INSTRUCTION].extend */
extern void T992f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RETRY_INSTRUCTION].aliased_resized_area */
extern T0* T990f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_RETRY_INSTRUCTION].aliased_resized_area */
extern T0* T992f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RESULT].force */
extern void T988f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_RESULT].fill_with */
extern void T989f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_RESULT].extend */
extern void T989f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RESULT].aliased_resized_area */
extern T0* T988f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_RESULT].aliased_resized_area */
extern T0* T989f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_REAL_CONSTANT].aliased_resized_area */
extern T0* T985f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_REAL_CONSTANT].aliased_resized_area */
extern T0* T987f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_OPERATOR].force */
extern void T980f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_KEYWORD_OPERATOR].fill_with */
extern void T982f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_KEYWORD_OPERATOR].extend */
extern void T982f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_OPERATOR].aliased_resized_area */
extern T0* T980f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_KEYWORD_OPERATOR].aliased_resized_area */
extern T0* T982f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INTEGER_CONSTANT].aliased_resized_area */
extern T0* T977f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_INTEGER_CONSTANT].aliased_resized_area */
extern T0* T979f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IDENTIFIER].aliased_resized_area */
extern T0* T975f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_IDENTIFIER].aliased_resized_area */
extern T0* T976f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FREE_OPERATOR].force */
extern void T972f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FREE_OPERATOR].fill_with */
extern void T974f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_FREE_OPERATOR].extend */
extern void T974f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FREE_OPERATOR].aliased_resized_area */
extern T0* T972f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_FREE_OPERATOR].aliased_resized_area */
extern T0* T974f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CURRENT].force */
extern void T970f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CURRENT].fill_with */
extern void T971f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_CURRENT].extend */
extern void T971f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CURRENT].aliased_resized_area */
extern T0* T970f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_CURRENT].aliased_resized_area */
extern T0* T971f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BREAK].force */
extern void T964f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_BREAK].fill_with */
extern void T966f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_BREAK].extend */
extern void T966f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BREAK].aliased_resized_area */
extern T0* T964f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_BREAK].aliased_resized_area */
extern T0* T966f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BOOLEAN_CONSTANT].aliased_resized_area */
extern T0* T961f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_BOOLEAN_CONSTANT].aliased_resized_area */
extern T0* T963f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSITION].force */
extern void T959f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_POSITION].fill_with */
extern void T960f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_POSITION].extend */
extern void T960f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSITION].aliased_resized_area */
extern T0* T959f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_POSITION].aliased_resized_area */
extern T0* T960f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECURSOR_KEYWORD].force */
extern void T954f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PRECURSOR_KEYWORD].fill_with */
extern void T956f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_PRECURSOR_KEYWORD].extend */
extern void T956f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECURSOR_KEYWORD].aliased_resized_area */
extern T0* T954f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_PRECURSOR_KEYWORD].aliased_resized_area */
extern T0* T956f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_KEYWORD].force */
extern void T951f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_AGENT_KEYWORD].fill_with */
extern void T953f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_AGENT_KEYWORD].extend */
extern void T953f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_KEYWORD].aliased_resized_area */
extern T0* T951f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_AGENT_KEYWORD].aliased_resized_area */
extern T0* T953f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.accept */
extern void T456f1241(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_error_action */
extern void T456f1239(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.report_error */
extern void T456f1266(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_error_action_1800_1999 */
extern void T456f1265(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.report_eof_expected_error */
extern void T456f1287(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_do_error_action_1600_1799 */
extern void T456f1264(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_error_action_1400_1599 */
extern void T456f1263(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_error_action_1200_1399 */
extern void T456f1262(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_error_action_1000_1199 */
extern void T456f1261(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_error_action_800_999 */
extern void T456f1260(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_error_action_600_799 */
extern void T456f1259(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_error_action_400_599 */
extern void T456f1258(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_error_action_200_399 */
extern void T456f1257(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_do_error_action_0_199 */
extern void T456f1256(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.read_token */
extern void T456f1238(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_execute_action */
extern void T456f1255(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.less */
extern void T456f1286(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.process_break */
extern void T456f1275(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_comment */
extern T0* T465f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_break */
extern T0* T465f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_underscored_real_constant */
extern T0* T465f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.set_position */
extern void T1491f44(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.make */
extern T0* T1491c41(GE_context* ac, T0* a1);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.make_leaf */
extern void T1491f46(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.last_literal */
extern T0* T456f1030(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.text_substring */
extern T0* T456f1035(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].force_new */
extern void T1744f41(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].key_storage_put */
extern void T1744f50(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].item_storage_put */
extern void T1744f46(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].slots_put */
extern void T1744f49(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].clashes_put */
extern void T1744f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].slots_item */
extern T6 T1744f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].hash_position */
extern T6 T1744f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].clashes_item */
extern T6 T1744f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].resize */
extern void T1744f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].clashes_resize */
extern void T1744f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].special_integer_ */
extern T0* T1744f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].key_storage_resize */
extern void T1744f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].item_storage_resize */
extern void T1744f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].key_storage_item */
extern T0* T1744f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].slots_resize */
extern void T1744f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].new_modulus */
extern T6 T1744f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].new_capacity */
extern T6 T1744f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].unset_found_item */
extern void T1744f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].found_key */
extern T0* T1744f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].found */
extern T1 T1744f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].search */
extern void T1744f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].search_position */
extern void T1744f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.strings */
extern T0* T456f1011(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].set_key_equality_tester */
extern void T1744f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.string_equality_tester */
extern T0* T456f1017(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].make_map */
extern T0* T1744c36(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].make_with_equality_testers */
extern void T1744f43(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].make_sparse_container */
extern void T1744f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].set_internal_cursor */
extern void T1744f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].new_cursor */
extern T0* T1744f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] STRING_8].make */
extern T0* T1984c5(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].make_slots */
extern void T1744f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].make_clashes */
extern void T1744f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].make_key_storage */
extern void T1744f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].make_item_storage */
extern void T1744f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.append_text_substring_to_string */
extern void T456f2272(GE_context* ac, T0* C, T6 a1, T6 a2, T0* a3);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.append_substring_to_string */
extern void T1479f21(GE_context* ac, T0* C, T6 a1, T6 a2, T0* a3);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.item */
extern T2 T1479f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.unicode_item */
extern T3 T1479f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_CHARACTER_BUFFER.append_substring_to_string */
extern void T1446f18(GE_context* ac, T0* C, T6 a1, T6 a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.string_ */
extern T0* T456f1010(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.string_buffer */
extern T0* T456f1009(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_regular_real_constant */
extern T0* T465f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REGULAR_REAL_CONSTANT.set_position */
extern void T1490f44(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_REGULAR_REAL_CONSTANT.make */
extern T0* T1490c41(GE_context* ac, T0* a1);
/* detachable ET_REGULAR_REAL_CONSTANT.make_leaf */
extern void T1490f46(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_binary_integer_constant */
extern T0* T465f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_BINARY_INTEGER_CONSTANT.set_position */
extern void T1489f69(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.last_binary_integer_constant */
extern T0* T456f1029(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.make */
extern T0* T1489c63(GE_context* ac, T0* a1, T11 a2, T1 a3);
/* detachable ET_BINARY_INTEGER_CONSTANT.make_leaf */
extern void T1489f71(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.compute_natural_64_value */
extern void T456f2271(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ET_AST_FACTORY.new_octal_integer_constant */
extern T0* T465f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_OCTAL_INTEGER_CONSTANT.set_position */
extern void T1488f69(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.last_octal_integer_constant */
extern T0* T456f1028(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.make */
extern T0* T1488c63(GE_context* ac, T0* a1, T11 a2, T1 a3);
/* detachable ET_OCTAL_INTEGER_CONSTANT.make_leaf */
extern void T1488f71(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_hexadecimal_integer_constant */
extern T0* T465f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.set_position */
extern void T1487f69(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.last_hexadecimal_integer_constant */
extern T0* T456f1027(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.make */
extern T0* T1487c63(GE_context* ac, T0* a1, T11 a2, T1 a3);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.make_leaf */
extern void T1487f71(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_silu_error */
extern void T53f265(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_silu_error */
extern void T36f265(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_sifu_error */
extern void T53f264(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_sifu_error */
extern void T36f264(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_underscored_integer_constant */
extern T0* T465f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.set_position */
extern void T1486f69(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.last_underscored_integer_constant */
extern T0* T456f1026(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.make */
extern T0* T1486c63(GE_context* ac, T0* a1, T11 a2, T1 a3);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.make_leaf */
extern void T1486f71(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_regular_integer_constant */
extern T0* T465f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.set_position */
extern void T372f71(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.last_regular_integer_constant */
extern T0* T456f1025(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_ssnp_error */
extern void T53f253(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_ssnp_error */
extern void T36f253(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_ssel_error */
extern void T53f263(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_ssel_error */
extern void T36f263(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_ssnl_error */
extern void T53f262(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_ssnl_error */
extern void T36f262(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_sssc_error */
extern void T53f261(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_sssc_error */
extern void T36f261(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_sscu_error */
extern void T53f260(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_sscu_error */
extern void T36f260(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_ssac_error */
extern void T53f259(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_ssac_error */
extern void T36f259(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_ssas_error */
extern void T53f258(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_ssas_error */
extern void T36f258(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_verbatim_string */
extern T0* T465f6(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4, T0* a5);
/* detachable ET_VERBATIM_STRING.set_position */
extern void T1485f55(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.last_verbatim_string */
extern T0* T456f1024(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4);
/* detachable ET_VERBATIM_STRING.make */
extern T0* T1485c50(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T1 a6);
/* detachable ET_VERBATIM_STRING.make_leaf */
extern void T1485f57(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.last_utf8_literal */
extern T0* T456f1021(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.is_verbatim_string_closer */
extern T1 T456f775(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.utf8_text */
extern T0* T456f773(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.append_utf8_text_to_string */
extern void T456f2270(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.append_utf8_substring_to_string */
extern void T1479f16(GE_context* ac, T0* C, T6 a1, T6 a2, T0* a3);
/* [detachable] KL_CHARACTER_BUFFER.append_utf8_substring_to_string */
extern void T1446f16(GE_context* ac, T0* C, T6 a1, T6 a2, T0* a3);
/* [detachable] ET_EIFFEL_PARSER.utf8_text_substring */
extern T0* T456f770(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.append_utf8_text_substring_to_string */
extern void T456f2269(GE_context* ac, T0* C, T6 a1, T6 a2, T0* a3);
/* [detachable] ET_AST_FACTORY.new_special_manifest_string */
extern T0* T465f5(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SPECIAL_MANIFEST_STRING.set_position */
extern void T1484f51(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.last_special_manifest_string */
extern T0* T456f1023(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.make */
extern T0* T1484c46(GE_context* ac, T0* a1, T0* a2, T1 a3);
/* detachable ET_SPECIAL_MANIFEST_STRING.make_leaf */
extern void T1484f53(GE_context* ac, T0* C);
/* NATURAL_64.to_natural_32 */
extern T10 T11f21(GE_context* ac, T11* C);
/* [detachable] ET_AST_FACTORY.new_regular_manifest_string */
extern T0* T465f4(GE_context* ac, T0* C, T0* a1);
/* detachable ET_REGULAR_MANIFEST_STRING.set_position */
extern void T1483f47(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_REGULAR_MANIFEST_STRING.make */
extern T0* T1483c45(GE_context* ac, T0* a1);
/* detachable ET_REGULAR_MANIFEST_STRING.make_leaf */
extern void T1483f49(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.process_regular_manifest_string */
extern void T456f1285(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_sceq_error */
extern void T53f257(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_sceq_error */
extern void T36f257(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_scqq_error */
extern void T53f256(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_scqq_error */
extern void T36f256(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_scac_error */
extern void T53f255(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_scac_error */
extern void T36f255(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_scas_error */
extern void T53f254(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_scas_error */
extern void T36f254(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_AST_FACTORY.new_c3_character_constant */
extern T0* T465f3(GE_context* ac, T0* C, T0* a1);
/* detachable ET_C3_CHARACTER_CONSTANT.set_position */
extern void T1482f52(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.last_c3_character_constant */
extern T0* T456f1022(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.make */
extern T0* T1482c47(GE_context* ac, T0* a1, T3 a2, T1 a3);
/* detachable ET_C3_CHARACTER_CONSTANT.make_leaf */
extern void T1482f54(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.process_c2_character_constant */
extern void T456f1284(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_AST_FACTORY.new_c2_character_constant */
extern T0* T465f131(GE_context* ac, T0* C, T2 a1, T0* a2);
/* detachable ET_C2_CHARACTER_CONSTANT.set_position */
extern void T1496f49(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_C2_CHARACTER_CONSTANT.make */
extern T0* T1496c47(GE_context* ac, T2 a1);
/* detachable ET_C2_CHARACTER_CONSTANT.make_leaf */
extern void T1496f51(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_scsc_error */
extern void T53f268(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_scsc_error */
extern void T36f268(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_sccu_error */
extern void T53f267(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_sccu_error */
extern void T36f267(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.text_item */
extern T2 T456f762(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.process_c1_character_constant */
extern void T456f1283(GE_context* ac, T0* C, T3 a1);
/* [detachable] ET_AST_FACTORY.new_c1_character_constant */
extern T0* T465f130(GE_context* ac, T0* C, T3 a1, T0* a2);
/* detachable ET_C1_CHARACTER_CONSTANT.set_position */
extern void T1495f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.set_start_condition */
extern void T456f1276(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.more */
extern void T456f1282(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.process_identifier */
extern void T456f1281(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_AST_FACTORY.new_detachable_keyword */
extern T0* T465f129(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_KEYWORD.set_position */
extern void T413f56(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PARSER.use_detachable_keyword */
extern T1 T456f916(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_reference_keyword */
extern T0* T465f128(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_KEYWORD.make_reference */
extern T0* T413c65(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.use_reference_keyword */
extern T1 T456f914(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_precursor_keyword */
extern T0* T465f127(GE_context* ac, T0* C, T0* a1);
/* detachable ET_PRECURSOR_KEYWORD.set_position */
extern void T955f25(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_AST_FACTORY.new_invariant_keyword */
extern T0* T465f126(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_attribute_keyword */
extern T0* T465f125(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.use_attribute_keyword */
extern T1 T456f910(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.use_obsolete_syntax */
extern T1 T802f26(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.use_obsolete_syntax */
extern T1 T887f3(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.use_obsolete_syntax */
extern T1 T410f2(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_undefine_keyword */
extern T0* T465f124(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.undefine_keyword */
extern T0* T73f737(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_undefine */
extern T0* T413c98(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_separate_keyword */
extern T0* T465f123(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_redefine_keyword */
extern T0* T465f122(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_obsolete_keyword */
extern T0* T465f121(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.obsolete_keyword */
extern T0* T73f735(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_obsolete */
extern T0* T413c96(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_indexing_keyword */
extern T0* T465f120(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_external_keyword */
extern T0* T465f119(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.external_keyword */
extern T0* T73f733(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_external */
extern T0* T413c94(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_expanded_keyword */
extern T0* T465f118(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_KEYWORD.make_expanded */
extern T0* T413c64(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_deferred_keyword */
extern T0* T465f117(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_creation_keyword */
extern T0* T465f116(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.creation_keyword */
extern T0* T73f732(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_creation */
extern T0* T413c93(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_attached_keyword */
extern T0* T465f115(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.use_attached_keyword */
extern T1 T456f899(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_variant_keyword */
extern T0* T465f114(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_require_keyword */
extern T0* T465f113(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_inspect_keyword */
extern T0* T465f112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.inspect_keyword */
extern T0* T73f730(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_inspect */
extern T0* T413c91(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_inherit_keyword */
extern T0* T465f111(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_implies_keyword */
extern T0* T465f110(GE_context* ac, T0* C, T0* a1);
/* detachable ET_KEYWORD_OPERATOR.set_position */
extern void T981f56(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_KEYWORD_OPERATOR.make_implies */
extern T0* T981c60(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_feature_keyword */
extern T0* T465f109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_convert_keyword */
extern T0* T465f108(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_current_keyword */
extern T0* T465f107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CURRENT.set_position */
extern void T282f37(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_AST_FACTORY.new_unique_keyword */
extern T0* T465f106(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_select_keyword */
extern T0* T465f105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.select_keyword */
extern T0* T73f727(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_select */
extern T0* T413c88(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_result_keyword */
extern T0* T465f104(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_rescue_keyword */
extern T0* T465f103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_rename_keyword */
extern T0* T465f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_frozen_keyword */
extern T0* T465f101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_KEYWORD.make_frozen */
extern T0* T413c61(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_export_keyword */
extern T0* T465f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_ensure_keyword */
extern T0* T465f99(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_elseif_keyword */
extern T0* T465f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.elseif_keyword */
extern T0* T73f723(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_elseif */
extern T0* T413c84(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_create_keyword */
extern T0* T465f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_assign_keyword */
extern T0* T465f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.assign_keyword */
extern T0* T73f722(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_assign */
extern T0* T413c83(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_across_keyword */
extern T0* T465f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_until_keyword */
extern T0* T465f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.until_keyword */
extern T0* T73f720(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_until */
extern T0* T413c81(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_strip_keyword */
extern T0* T465f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_retry_keyword */
extern T0* T465f92(GE_context* ac, T0* C, T0* a1);
/* detachable ET_RETRY_INSTRUCTION.set_position */
extern void T991f14(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_RETRY_INSTRUCTION.make */
extern T0* T991c13(GE_context* ac);
/* detachable ET_RETRY_INSTRUCTION.make_token */
extern void T991f15(GE_context* ac, T0* C, T0* a1);
/* detachable ET_RETRY_INSTRUCTION.make_leaf */
extern void T991f16(GE_context* ac, T0* C);
/* detachable ET_RETRY_INSTRUCTION.tokens */
extern T0* T991f7(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_local_keyword */
extern T0* T465f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_false_keyword */
extern T0* T465f90(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FALSE_CONSTANT.set_position */
extern void T1494f40(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_FALSE_CONSTANT.make */
extern T0* T1494c39(GE_context* ac);
/* detachable ET_FALSE_CONSTANT.make_token */
extern void T1494f42(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FALSE_CONSTANT.make_leaf */
extern void T1494f43(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_debug_keyword */
extern T0* T465f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_class_keyword */
extern T0* T465f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_check_keyword */
extern T0* T465f87(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_alias_keyword */
extern T0* T465f86(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_agent_keyword */
extern T0* T465f85(GE_context* ac, T0* C, T0* a1);
/* detachable ET_AGENT_KEYWORD.set_position */
extern void T952f14(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_AST_FACTORY.new_when_keyword */
extern T0* T465f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_void_keyword */
extern T0* T465f83(GE_context* ac, T0* C, T0* a1);
/* detachable ET_VOID.set_position */
extern void T997f31(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_VOID.make */
extern T0* T997c30(GE_context* ac);
/* detachable ET_VOID.make_token */
extern void T997f33(GE_context* ac, T0* C, T0* a1);
/* detachable ET_VOID.make_leaf */
extern void T997f34(GE_context* ac, T0* C);
/* detachable ET_VOID.tokens */
extern T0* T997f11(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_true_keyword */
extern T0* T465f82(GE_context* ac, T0* C, T0* a1);
/* detachable ET_TRUE_CONSTANT.set_position */
extern void T1493f40(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_AST_FACTORY.new_then_keyword */
extern T0* T465f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_some_keyword */
extern T0* T465f80(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.some_keyword */
extern T0* T73f713(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_some */
extern T0* T413c74(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_note_keyword */
extern T0* T465f79(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.note_keyword */
extern T0* T73f712(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_note */
extern T0* T413c73(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.use_note_keyword */
extern T1 T456f860(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_identifier */
extern T0* T465f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.last_identifier */
extern T0* T454f131(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].replace_found_item */
extern void T1744f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_IDENTIFIER.make_with_hash_code */
extern void T184f125(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_IDENTIFIER.make_with_hash_code */
extern T0* T184c125(GE_context* ac, T0* a1, T6 a2);
/* [detachable] ET_IDENTIFIER.make_leaf */
extern void T184f126(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].found_item */
extern T6 T1744f26(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].item_storage_item */
extern T6 T1744f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PREPARSER.strings */
extern T0* T454f134(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.tokens */
extern T0* T454f93(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.string_equality_tester */
extern T0* T454f135(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.append_text_substring_to_string */
extern void T454f206(GE_context* ac, T0* C, T6 a1, T6 a2, T0* a3);
/* [detachable] ET_EIFFEL_PREPARSER.string_ */
extern T0* T454f133(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.string_buffer */
extern T0* T454f132(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.last_identifier */
extern T0* T456f1031(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_loop_keyword */
extern T0* T465f77(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_like_keyword */
extern T0* T465f76(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_from_keyword */
extern T0* T465f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_else_keyword */
extern T0* T465f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_xor_keyword */
extern T0* T465f73(GE_context* ac, T0* C, T0* a1);
/* detachable ET_KEYWORD_OPERATOR.make_xor */
extern T0* T981c59(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_old_keyword */
extern T0* T465f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_not_keyword */
extern T0* T465f71(GE_context* ac, T0* C, T0* a1);
/* detachable ET_KEYWORD_OPERATOR.make_not */
extern T0* T981c58(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_end_keyword */
extern T0* T465f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_all_keyword */
extern T0* T465f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_and_keyword */
extern T0* T465f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_or_keyword */
extern T0* T465f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_is_keyword */
extern T0* T465f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_if_keyword */
extern T0* T465f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.if_keyword */
extern T0* T73f707(GE_context* ac, T0* C);
/* [detachable] ET_KEYWORD.make_if */
extern T0* T413c68(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_do_keyword */
extern T0* T465f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_as_keyword */
extern T0* T465f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_once_keyword */
extern T0* T465f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.process_two_char_symbol */
extern void T456f1280(GE_context* ac, T0* C, T3 a1, T3 a2);
/* [detachable] ET_AST_FACTORY.new_or_else_symbol */
extern T0* T465f62(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.set_position */
extern void T994f132(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable ET_SYMBOL_OPERATOR.make_or_else_symbol */
extern T0* T994c149(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_and_then_symbol */
extern T0* T465f61(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.make_and_then_symbol */
extern T0* T994c148(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_assign_attempt_symbol */
extern T0* T465f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_assign_symbol */
extern T0* T465f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_dotdot_symbol */
extern T0* T465f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_arrow_symbol */
extern T0* T465f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_left_array_symbol */
extern T0* T465f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYMBOL.set_position */
extern void T341f97(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_AST_FACTORY.new_le_symbol */
extern T0* T465f55(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.make_le */
extern T0* T994c147(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_right_array_symbol */
extern T0* T465f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_ge_symbol */
extern T0* T465f53(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.make_ge */
extern T0* T994c146(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_mod_symbol */
extern T0* T465f52(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.make_mod */
extern T0* T994c145(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_not_tilde_symbol */
extern T0* T465f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYMBOL.make_not_tilde */
extern T0* T341c103(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_not_equal_symbol */
extern T0* T465f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYMBOL.make_not_equal */
extern T0* T341c102(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_div_symbol */
extern T0* T465f49(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.make_div */
extern T0* T994c144(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_free_operator */
extern T0* T465f1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.text_count */
extern T6 T456f753(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_set_line_column */
extern void T456f1279(GE_context* ac, T0* C);
/* [detachable] KL_CHARACTER_BUFFER.unicode_item */
extern T3 T1446f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.process_one_char_symbol */
extern void T456f1278(GE_context* ac, T0* C, T3 a1);
/* [detachable] ET_AST_FACTORY.new_xor_symbol */
extern T0* T465f48(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.make_xor_symbol */
extern T0* T994c143(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_or_symbol */
extern T0* T465f47(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.make_or_symbol */
extern T0* T994c142(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_not_symbol */
extern T0* T465f46(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.make_not_symbol */
extern T0* T994c141(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_implies_symbol */
extern T0* T465f45(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.make_implies_symbol */
extern T0* T994c140(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_and_symbol */
extern T0* T465f44(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.make_and_symbol */
extern T0* T994c139(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_close_repeat_symbol */
extern T0* T465f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_open_repeat_symbol */
extern T0* T465f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_bar_symbol */
extern T0* T465f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_there_exists_symbol */
extern T0* T465f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_for_all_symbol */
extern T0* T465f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.for_all_symbol */
extern T0* T73f695(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_for_all */
extern T0* T341c111(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_at_symbol */
extern T0* T465f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_tilde_symbol */
extern T0* T465f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYMBOL.make_tilde */
extern T0* T341c100(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_dollar_symbol */
extern T0* T465f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_right_bracket_symbol */
extern T0* T465f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_left_bracket_symbol */
extern T0* T465f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BRACKET_SYMBOL.set_position */
extern void T622f132(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_AST_FACTORY.new_right_brace_symbol */
extern T0* T465f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_left_brace_symbol */
extern T0* T465f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_right_parenthesis_symbol */
extern T0* T465f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_left_parenthesis_symbol */
extern T0* T465f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_question_mark_symbol */
extern T0* T465f29(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUESTION_MARK_SYMBOL.set_position */
extern void T1005f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_AST_FACTORY.new_bang_symbol */
extern T0* T465f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_colon_symbol */
extern T0* T465f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_comma_symbol */
extern T0* T465f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.comma_symbol */
extern T0* T73f691(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_comma */
extern T0* T341c107(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_semicolon_symbol */
extern T0* T465f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.semicolon_symbol */
extern T0* T73f690(GE_context* ac, T0* C);
/* detachable ET_SEMICOLON_SYMBOL.make */
extern T0* T1000c4(GE_context* ac);
/* detachable ET_SEMICOLON_SYMBOL.make_leaf */
extern void T1000f5(GE_context* ac, T0* C);
/* detachable ET_SEMICOLON_SYMBOL.tokens */
extern T0* T1000f3(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.new_dot_symbol */
extern T0* T465f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.dot_symbol */
extern T0* T73f689(GE_context* ac, T0* C);
/* [detachable] ET_SYMBOL.make_dot */
extern T0* T341c106(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_lt_symbol */
extern T0* T465f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_gt_symbol */
extern T0* T465f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_FACTORY.new_equal_symbol */
extern T0* T465f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYMBOL.make_equal */
extern T0* T341c96(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_power_symbol */
extern T0* T465f20(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.make_power */
extern T0* T994c136(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_divide_symbol */
extern T0* T465f19(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.make_divide */
extern T0* T994c135(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_times_symbol */
extern T0* T465f18(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.make_times */
extern T0* T994c134(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_plus_symbol */
extern T0* T465f17(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.make_plus */
extern T0* T994c133(GE_context* ac);
/* [detachable] ET_AST_FACTORY.new_minus_symbol */
extern T0* T465f16(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYMBOL_OPERATOR.make_minus */
extern T0* T994c131(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.unicode_text_item */
extern T3 T456f752(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_execute_eof_action */
extern void T456f1254(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_sseq_error */
extern void T53f252(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_sseq_error */
extern void T36f252(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.terminate */
extern void T456f1274(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.wrap */
extern T1 T456f715(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_refill_input_buffer */
extern void T456f1253(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_set_content */
extern void T456f1049(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.count */
extern T6 T1479f6(GE_context* ac, T0* C);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.as_special */
extern T0* T1479f12(GE_context* ac, T0* C);
/* [detachable] YY_UNICODE_FILE_BUFFER.fill */
extern void T936f26(GE_context* ac, T0* C);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.put */
extern void T1479f17(GE_context* ac, T0* C, T2 a1, T6 a2);
/* CHARACTER_8.to_character_8 */
extern T2 T2f31(GE_context* ac, T2* C);
/* [detachable] KL_STRING_INPUT_STREAM.end_of_input */
extern T1 T1477f1(GE_context* ac, T0* C);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.fill_from_iso_8859_1_stream */
extern T6 T1479f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.fill_from_iso_8859_1_substring */
extern void T1479f20(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] STRING_32.put */
extern void T18f38(GE_context* ac, T0* C, T3 a1, T6 a2);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.fill_from_utf8_stream */
extern T6 T1479f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] YY_UNICODE_FILE_BUFFER.compact_left */
extern void T936f28(GE_context* ac, T0* C);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.move_left */
extern void T1479f19(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] STRING_32.subcopy */
extern void T18f39(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.resize */
extern void T1479f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] YY_UNICODE_FILE_BUFFER.default_capacity */
extern T6 T936f19(GE_context* ac, T0* C);
/* [detachable] YY_UNICODE_FILE_BUFFER.set_index */
extern void T936f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_null_trans_state */
extern T6 T456f714(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_previous_state */
extern T6 T456f713(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.fatal_error */
extern void T456f1252(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.report_invalid_unicode_character_error */
extern void T456f1251(GE_context* ac, T0* C, T10 a1);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.item_code */
extern T10 T1479f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.yy_at_beginning_of_line */
extern T6 T456f703(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.special_integer_ */
extern T0* T456f742(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_init_value_stacks */
extern void T456f1237(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_clear_all */
extern void T456f1246(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.clear_all */
extern void T456f1272(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.clear_stacks */
extern void T456f2234(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_clear_value_stacks */
extern void T456f2265(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable ET_WRITABLE].keep_head */
extern void T1305f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_WRITABLE].set_count */
extern void T1305f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_WHEN_PART_LIST].keep_head */
extern void T1303f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_WHEN_PART_LIST].set_count */
extern void T1303f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_WHEN_PART].keep_head */
extern void T1300f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_WHEN_PART].set_count */
extern void T1300f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_WHEN_EXPRESSION_LIST].keep_head */
extern void T1297f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_WHEN_EXPRESSION_LIST].set_count */
extern void T1297f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_WHEN_EXPRESSION].keep_head */
extern void T1294f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_WHEN_EXPRESSION].set_count */
extern void T1294f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_VARIANT].keep_head */
extern void T1291f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_VARIANT].set_count */
extern void T1291f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_LIST].keep_head */
extern void T1288f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_LIST].set_count */
extern void T1288f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_ITEM].keep_head */
extern void T1285f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_ITEM].set_count */
extern void T1285f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_LIST].keep_head */
extern void T1282f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_LIST].set_count */
extern void T1282f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_ITEM].keep_head */
extern void T1279f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_ITEM].set_count */
extern void T1279f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT].keep_head */
extern void T1276f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT].set_count */
extern void T1276f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE].keep_head */
extern void T1273f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE].set_count */
extern void T1273f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_STRIP_EXPRESSION].keep_head */
extern void T1271f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_STRIP_EXPRESSION].set_count */
extern void T1271f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_STATIC_CALL_EXPRESSION].keep_head */
extern void T1268f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_STATIC_CALL_EXPRESSION].set_count */
extern void T1268f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_REPEAT_INSTRUCTION].keep_head */
extern void T1265f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_REPEAT_INSTRUCTION].set_count */
extern void T1265f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RENAME_LIST].keep_head */
extern void T1262f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RENAME_LIST].set_count */
extern void T1262f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RENAME_ITEM].keep_head */
extern void T1260f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RENAME_ITEM].set_count */
extern void T1260f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUERY].keep_head */
extern void T1257f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUERY].set_count */
extern void T1257f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUANTIFIER_EXPRESSION].keep_head */
extern void T1255f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUANTIFIER_EXPRESSION].set_count */
extern void T1255f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_LIKE_IDENTIFIER].keep_head */
extern void T1252f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_LIKE_IDENTIFIER].set_count */
extern void T1252f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PROCEDURE].keep_head */
extern void T1249f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PROCEDURE].set_count */
extern void T1249f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PRECONDITIONS].keep_head */
extern void T1247f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PRECONDITIONS].set_count */
extern void T1247f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_POSTCONDITIONS].keep_head */
extern void T1245f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_POSTCONDITIONS].set_count */
extern void T1245f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT_LIST].keep_head */
extern void T1242f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT_LIST].set_count */
extern void T1242f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT_ITEM].keep_head */
extern void T1240f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT_ITEM].set_count */
extern void T1240f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT_CLAUSE_LIST].keep_head */
extern void T1238f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT_CLAUSE_LIST].set_count */
extern void T1238f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT].keep_head */
extern void T1235f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT].set_count */
extern void T1235f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENTHESIZED_EXPRESSION].keep_head */
extern void T1233f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENTHESIZED_EXPRESSION].set_count */
extern void T1233f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_OBSOLETE].keep_head */
extern void T1230f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_OBSOLETE].set_count */
extern void T1230f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_TUPLE].keep_head */
extern void T1227f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_TUPLE].set_count */
extern void T1227f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_LIST].keep_head */
extern void T1225f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_LIST].set_count */
extern void T1225f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_ITEM].keep_head */
extern void T1222f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_ITEM].set_count */
extern void T1222f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_ARRAY].keep_head */
extern void T1219f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_ARRAY].set_count */
extern void T1219f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LOOP_INVARIANTS].keep_head */
extern void T1216f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LOOP_INVARIANTS].set_count */
extern void T1216f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_ITEM].keep_head */
extern void T1210f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_ITEM].set_count */
extern void T1210f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE].keep_head */
extern void T1207f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE].set_count */
extern void T1207f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LIKE_TYPE].keep_head */
extern void T1204f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LIKE_TYPE].set_count */
extern void T1204f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_KEYWORD_FEATURE_NAME_LIST].keep_head */
extern void T1201f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_KEYWORD_FEATURE_NAME_LIST].set_count */
extern void T1201f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INVARIANTS].keep_head */
extern void T1199f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INVARIANTS].set_count */
extern void T1199f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INSTRUCTION].keep_head */
extern void T1196f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INSTRUCTION].set_count */
extern void T1196f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INSPECT_INSTRUCTION].keep_head */
extern void T1193f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INSPECT_INSTRUCTION].set_count */
extern void T1193f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INSPECT_EXPRESSION].keep_head */
extern void T1190f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INSPECT_EXPRESSION].set_count */
extern void T1190f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INLINE_AGENT].keep_head */
extern void T1187f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INLINE_AGENT].set_count */
extern void T1187f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_LIST].keep_head */
extern void T1185f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_LIST].set_count */
extern void T1185f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_ITEM].keep_head */
extern void T1182f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_ITEM].set_count */
extern void T1182f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM].keep_head */
extern void T1179f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM].set_count */
extern void T1179f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_ITEM].keep_head */
extern void T1177f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_ITEM].set_count */
extern void T1177f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_LIST].keep_head */
extern void T1174f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_LIST].set_count */
extern void T1174f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_IF_INSTRUCTION].keep_head */
extern void T1171f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_IF_INSTRUCTION].set_count */
extern void T1171f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_IF_EXPRESSION].keep_head */
extern void T1168f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_IF_EXPRESSION].set_count */
extern void T1168f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_LIST].keep_head */
extern void T1165f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_LIST].set_count */
extern void T1165f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_ITEM].keep_head */
extern void T1163f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_ITEM].set_count */
extern void T1163f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER].keep_head */
extern void T1160f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER].set_count */
extern void T1160f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_ITEM].keep_head */
extern void T1156f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_ITEM].set_count */
extern void T1156f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT].keep_head */
extern void T1153f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT].set_count */
extern void T1153f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME_ITEM].keep_head */
extern void T1151f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME_ITEM].set_count */
extern void T1151f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME].keep_head */
extern void T1148f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME].set_count */
extern void T1148f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_EXPORT].keep_head */
extern void T1146f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_EXPORT].set_count */
extern void T1146f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE_LIST].keep_head */
extern void T1143f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE_LIST].set_count */
extern void T1143f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE].keep_head */
extern void T1140f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE].set_count */
extern void T1140f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXTERNAL_ALIAS].keep_head */
extern void T1137f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXTERNAL_ALIAS].set_count */
extern void T1137f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXTENDED_FEATURE_NAME].keep_head */
extern void T1134f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXTENDED_FEATURE_NAME].set_count */
extern void T1134f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPRESSION_ITEM].keep_head */
extern void T1132f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPRESSION_ITEM].set_count */
extern void T1132f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPRESSION].keep_head */
extern void T1130f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPRESSION].set_count */
extern void T1130f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPORT_LIST].keep_head */
extern void T1128f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPORT_LIST].set_count */
extern void T1128f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPORT].keep_head */
extern void T1126f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPORT].set_count */
extern void T1126f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART_LIST].keep_head */
extern void T1123f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART_LIST].set_count */
extern void T1123f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART].keep_head */
extern void T1120f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART].set_count */
extern void T1120f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION_LIST].keep_head */
extern void T1117f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION_LIST].set_count */
extern void T1117f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION].keep_head */
extern void T1114f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION].set_count */
extern void T1114f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_DEBUG_INSTRUCTION].keep_head */
extern void T1111f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_DEBUG_INSTRUCTION].set_count */
extern void T1111f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATOR_LIST].keep_head */
extern void T1108f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATOR_LIST].set_count */
extern void T1108f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATOR].keep_head */
extern void T1105f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATOR].set_count */
extern void T1105f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATION_REGION].keep_head */
extern void T1102f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATION_REGION].set_count */
extern void T1102f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATE_EXPRESSION].keep_head */
extern void T1099f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATE_EXPRESSION].set_count */
extern void T1099f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_LIST].keep_head */
extern void T1097f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_LIST].set_count */
extern void T1097f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_ITEM].keep_head */
extern void T1094f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_ITEM].set_count */
extern void T1094f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE].keep_head */
extern void T1091f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE].set_count */
extern void T1091f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_RENAME_LIST].keep_head */
extern void T1085f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_RENAME_LIST].set_count */
extern void T1085f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_CREATOR].keep_head */
extern void T1082f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_CREATOR].set_count */
extern void T1082f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST].keep_head */
extern void T1079f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST].set_count */
extern void T1079f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].keep_head */
extern void T1076f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].set_count */
extern void T1076f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTANT].keep_head */
extern void T1073f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTANT].set_count */
extern void T1073f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONDITIONAL].keep_head */
extern void T1071f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONDITIONAL].set_count */
extern void T1071f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_COMPOUND].keep_head */
extern void T1068f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_COMPOUND].set_count */
extern void T1068f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CLIENTS].keep_head */
extern void T1066f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CLIENTS].set_count */
extern void T1066f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CLIENT_ITEM].keep_head */
extern void T1063f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CLIENT_ITEM].set_count */
extern void T1063f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CLASS].keep_head */
extern void T1061f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CLASS].set_count */
extern void T1061f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE_LIST].keep_head */
extern void T1059f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE_LIST].set_count */
extern void T1059f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE_ITEM].keep_head */
extern void T1056f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE_ITEM].set_count */
extern void T1056f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE_CONSTANT].keep_head */
extern void T1053f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE_CONSTANT].set_count */
extern void T1053f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE].keep_head */
extern void T1050f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE].set_count */
extern void T1050f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHECK_INSTRUCTION].keep_head */
extern void T1047f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHECK_INSTRUCTION].set_count */
extern void T1047f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_CALL_EXPRESSION].keep_head */
extern void T1044f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_CALL_EXPRESSION].set_count */
extern void T1044f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CALL_AGENT].keep_head */
extern void T1042f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CALL_AGENT].set_count */
extern void T1042f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BRACKET_EXPRESSION].keep_head */
extern void T1039f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BRACKET_EXPRESSION].set_count */
extern void T1039f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ASSIGNER].keep_head */
extern void T1036f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ASSIGNER].set_count */
extern void T1036f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ALIAS_NAME_LIST].keep_head */
extern void T1033f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ALIAS_NAME_LIST].set_count */
extern void T1033f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ALIAS_NAME].keep_head */
extern void T1031f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ALIAS_NAME].set_count */
extern void T1031f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_TARGET].keep_head */
extern void T1028f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_TARGET].set_count */
extern void T1028f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_LIST].keep_head */
extern void T1026f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_LIST].set_count */
extern void T1026f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM].keep_head */
extern void T1023f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM].set_count */
extern void T1023f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND].keep_head */
extern void T1020f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND].set_count */
extern void T1020f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_LIST].keep_head */
extern void T1018f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_LIST].set_count */
extern void T1018f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_ITEM].keep_head */
extern void T1016f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_ITEM].set_count */
extern void T1016f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACTUAL_ARGUMENT_LIST].keep_head */
extern void T1014f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACTUAL_ARGUMENT_LIST].set_count */
extern void T1014f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACROSS_INSTRUCTION].keep_head */
extern void T1012f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACROSS_INSTRUCTION].set_count */
extern void T1012f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACROSS_EXPRESSION].keep_head */
extern void T1009f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACROSS_EXPRESSION].set_count */
extern void T1009f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUESTION_MARK_SYMBOL].keep_head */
extern void T1006f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUESTION_MARK_SYMBOL].set_count */
extern void T1006f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BRACKET_SYMBOL].keep_head */
extern void T1003f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BRACKET_SYMBOL].set_count */
extern void T1003f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_SEMICOLON_SYMBOL].keep_head */
extern void T1001f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_SEMICOLON_SYMBOL].set_count */
extern void T1001f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_VOID].keep_head */
extern void T998f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_VOID].set_count */
extern void T998f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_SYMBOL_OPERATOR].keep_head */
extern void T995f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_SYMBOL_OPERATOR].set_count */
extern void T995f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RETRY_INSTRUCTION].keep_head */
extern void T992f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RETRY_INSTRUCTION].set_count */
extern void T992f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RESULT].keep_head */
extern void T989f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RESULT].set_count */
extern void T989f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_REAL_CONSTANT].keep_head */
extern void T987f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_REAL_CONSTANT].set_count */
extern void T987f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING].keep_head */
extern void T984f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING].set_count */
extern void T984f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_KEYWORD_OPERATOR].keep_head */
extern void T982f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_KEYWORD_OPERATOR].set_count */
extern void T982f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INTEGER_CONSTANT].keep_head */
extern void T979f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INTEGER_CONSTANT].set_count */
extern void T979f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_IDENTIFIER].keep_head */
extern void T976f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_IDENTIFIER].set_count */
extern void T976f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FREE_OPERATOR].keep_head */
extern void T974f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FREE_OPERATOR].set_count */
extern void T974f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CURRENT].keep_head */
extern void T971f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CURRENT].set_count */
extern void T971f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHARACTER_CONSTANT].keep_head */
extern void T969f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHARACTER_CONSTANT].set_count */
extern void T969f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BREAK].keep_head */
extern void T966f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BREAK].set_count */
extern void T966f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BOOLEAN_CONSTANT].keep_head */
extern void T963f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BOOLEAN_CONSTANT].set_count */
extern void T963f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_POSITION].keep_head */
extern void T960f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_POSITION].set_count */
extern void T960f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PRECURSOR_KEYWORD].keep_head */
extern void T956f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PRECURSOR_KEYWORD].set_count */
extern void T956f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_KEYWORD].keep_head */
extern void T953f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_KEYWORD].set_count */
extern void T953f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PARSER.abort */
extern void T456f1240(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_load_input_buffer */
extern void T456f1047(GE_context* ac, T0* C);
/* [detachable] YY_UNICODE_FILE_BUFFER.set_file */
extern void T936f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] YY_UNICODE_FILE_BUFFER.set_file_with_size */
extern void T936f27(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] YY_UNICODE_FILE_BUFFER.flush */
extern void T936f29(GE_context* ac, T0* C);
/* [detachable] YY_UNICODE_FILE_BUFFER.flush */
extern void T936f29p1(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.reset */
extern void T456f1236(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].wipe_out */
extern void T946f12(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].move_all_cursors_after */
extern void T946f17(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_SYMBOL].wipe_out */
extern void T941f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [detachable ET_KEYWORD].wipe_out */
extern void T940f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] INTEGER_32].wipe_out */
extern void T937f10(GE_context* ac, T0* C);
/* [detachable] YY_UNICODE_FILE_BUFFER.set_end_of_file */
extern void T936f24(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.reset */
extern void T456f1236p1(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.reset */
extern void T456f1236p0(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.reset */
extern void T456f1236p2(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.time_stamp */
extern T6 T37f43(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.date */
extern T6 T37f51(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.eif_file_date */
extern T6 T37f57(GE_context* ac, T0* C, T14 a1);
/* [detachable] ET_EIFFEL_PARSER.tmp_file */
extern T0* T456f734(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.is_closed */
extern T1 T37f17(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.shared_file */
extern T0* T456f1007(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_after_preparsed */
extern void T74f275(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_in_cluster */
extern T1 T74f158(GE_context* ac, T0* C);
/* [detachable] MUTEX.try_lock */
extern T1 T131f7(GE_context* ac, T0* C);
/* [detachable] MUTEX.eif_thr_mutex_trylock */
extern T1 T131f8(GE_context* ac, T0* C, T14 a1);
/* [detachable] ET_EIFFEL_PARSER.error_handler */
extern T0* T456f733(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.set_fatal_error */
extern void T456f1230(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.process_none_class */
extern void T456f1229(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCESTOR_BUILDER.process_class */
extern void T458f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCESTOR_BUILDER.internal_process_class */
extern void T458f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vhpr1a_error */
extern void T53f209(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vhpr1a */
extern T0* T785c859(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS].after */
extern T1 T1348f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS].forth */
extern void T1348f14(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].cursor_forth */
extern void T462f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].add_traversing_cursor */
extern void T462f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS].set_next_cursor */
extern void T1348f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].remove_traversing_cursor */
extern void T462f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS].item */
extern T0* T1348f5(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].cursor_item */
extern T0* T462f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS].start */
extern void T1348f13(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].cursor_start */
extern void T462f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS].set_after */
extern void T1348f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS].off */
extern T1 T1348f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_CLASS].before */
extern T1 T1348f8(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vhpr1_error */
extern T1 T53f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vhpr1a_error */
extern void T36f209(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vhpr1_error */
extern T1 T36f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCESTOR_BUILDER.set_parents_inheritance_error */
extern void T458f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].wipe_out */
extern void T1306f13(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].wipe_out */
extern void T1655f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].slots_wipe_out */
extern void T1655f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].clashes_wipe_out */
extern void T1655f45(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].key_storage_wipe_out */
extern void T1655f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].item_storage_wipe_out */
extern void T1655f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].unset_found_item */
extern void T1655f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].move_all_cursors_after */
extern void T1655f41(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_CLASS].set_next_cursor */
extern void T1903f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_CLASS].set_after */
extern void T1903f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].wipe_out */
extern void T462f29(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].move_all_cursors_after */
extern void T462f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].reset */
extern void T1306f17(GE_context* ac, T0* C);
/* [detachable] ET_ANCESTOR_BUILDER.check_parents_validity */
extern void T458f20(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER1.check_parents_validity */
extern void T1308f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCESTOR_BUILDER.check_formal_parameters_validity */
extern void T458f19(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.check_formal_parameters_validity */
extern void T1309f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT].wipe_out */
extern void T1659f12(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE_CONSTRAINT].keep_head */
extern void T1904f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE_CONSTRAINT].set_count */
extern void T1904f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT].move_all_cursors_after */
extern void T1659f15(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_BASE_TYPE_CONSTRAINT].set_next_cursor */
extern void T1906f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_BASE_TYPE_CONSTRAINT].set_after */
extern void T1906f6(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.set_constraint_base_types */
extern void T1309f28(GE_context* ac, T0* C, T0* a1);
extern T1 T1309f28ot1(T0* a1, T0** a2);
/* [detachable] ET_BASE_TYPE_CONSTRAINT_LIST.put_first */
extern void T1661f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BASE_TYPE_CONSTRAINT_LIST.fixed_array */
extern T0* T1661f6(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE_CONSTRAINT].default_create */
extern T0* T1905c3(GE_context* ac);
/* [detachable] ET_BASE_TYPE_CONSTRAINT_LIST.make_with_capacity */
extern T0* T1661c11(GE_context* ac, T0* a1, T6 a2);
/* [detachable] ET_BASE_TYPE_CONSTRAINT_LIST.make_empty_with_capacity */
extern void T1661f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE_CONSTRAINT].make */
extern T0* T1905f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE_CONSTRAINT].make_empty */
extern T0* T1904c5(GE_context* ac, T6 a1);
/* [detachable] ET_BASE_TYPE_CONSTRAINT_LIST.tokens */
extern T0* T1661f5(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT].item */
extern T0* T1659f6(GE_context* ac, T0* C, T6 a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.set_constraint_base_types */
extern void T1501f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT].first */
extern T0* T1659f10(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.base_type_constraint */
extern T0* T1309f19(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.tokens */
extern T0* T1309f14(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.set_recursive_formal_constraints */
extern void T1309f27(GE_context* ac, T0* C, T0* a1);
extern T1 T1309f27ot1(T0* a1, T0** a2);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.set_recursive_formal_constraints */
extern void T1501f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32].item */
extern T10 T1660f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32].replace */
extern void T1660f12(GE_context* ac, T0* C, T10 a1, T6 a2);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.add_n_flags_to_list */
extern void T1309f30(GE_context* ac, T0* C, T10 a1, T6 a2, T0* a3);
/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32].force_last */
extern void T1660f13(GE_context* ac, T0* C, T10 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32].resize */
extern void T1660f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32].new_capacity */
extern T6 T1660f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32].extendible */
extern T1 T1660f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32].wipe_out */
extern void T1660f11(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32].move_all_cursors_after */
extern void T1660f15(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] NATURAL_32].set_next_cursor */
extern void T1907f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] NATURAL_32].set_after */
extern void T1907f6(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.propagate_recursive_formal_dependencies */
extern void T1309f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.check_constraint_validity */
extern void T1309f25(GE_context* ac, T0* C, T0* a1);
extern T1 T1309f25ot1(T0* a1, T0** a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vcfg2a_error */
extern void T53f340(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vcfg2a */
extern T0* T785c973(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vcfg2_error */
extern T1 T53f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vcfg2a_error */
extern void T36f340(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vcfg2_error */
extern T1 T36f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vcfg1a_error */
extern void T53f339(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vcfg1a */
extern T0* T785c972(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vcfg1_error */
extern T1 T53f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vcfg1a_error */
extern void T36f339(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vcfg1_error */
extern T1 T36f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.initalize_base_type_dependencies */
extern void T1309f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.initalize_formal_dependencies */
extern void T1309f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ANCESTOR_BUILDER.set_ancestors */
extern void T458f18(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_conforming_ancestors */
extern void T74f269(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_ancestors */
extern void T74f268(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].wipe_out */
extern void T1307f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].slots_wipe_out */
extern void T1307f54(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].clashes_wipe_out */
extern void T1307f53(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].key_storage_wipe_out */
extern void T1307f52(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].item_storage_wipe_out */
extern void T1307f51(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE].keep_head */
extern void T923f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE].set_count */
extern void T923f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].unset_found_item */
extern void T1307f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].move_all_cursors_after */
extern void T1307f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_BASE_TYPE, [attached] ET_CLASS].set_next_cursor */
extern void T1656f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_BASE_TYPE, [attached] ET_CLASS].set_after */
extern void T1656f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].forth */
extern void T1307f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].cursor_forth */
extern void T1307f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].add_traversing_cursor */
extern void T1307f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_BASE_TYPE, [attached] ET_CLASS].set_position */
extern void T1656f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].remove_traversing_cursor */
extern void T1307f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].clashes_item */
extern T6 T1307f39(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_BASE_TYPE_LIST.put_last */
extern void T412f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE].force */
extern void T924f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE].fill_with */
extern void T923f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE].extend */
extern void T923f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].item_for_iteration */
extern T0* T1307f22(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].cursor_item */
extern T0* T1307f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].item_storage_item */
extern T0* T1307f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].after */
extern T1 T1307f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].cursor_after */
extern T1 T1307f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_BASE_TYPE, [attached] ET_CLASS].after */
extern T1 T1656f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].start */
extern void T1307f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].cursor_start */
extern void T1307f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].cursor_off */
extern T1 T1307f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_BASE_TYPE, [attached] ET_CLASS].off */
extern T1 T1656f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].is_empty */
extern T1 T1307f24(GE_context* ac, T0* C);
/* [detachable] ET_ANCESTOR_BUILDER.add_parent_to_ancestors */
extern void T458f23(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] ET_CLASS_TYPE.resolved_formal_parameters */
extern T0* T72f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.resolved_formal_parameters_with_type_mark */
extern T0* T72f103(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.resolved_formal_parameters */
extern T0* T577f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.resolved_formal_parameters_with_type */
extern T0* T254f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.resolved_formal_parameters */
extern T0* T254f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.resolved_formal_parameters_with_type_mark */
extern T0* T254f98(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.resolved_formal_parameters */
extern T0* T1956f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.resolved_formal_parameters_with_type */
extern T0* T1501f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.resolved_formal_parameters_with_type */
extern T0* T891f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LABELED_COMMA_ACTUAL_PARAMETER.resolved_formal_parameters_with_type */
extern T0* T1769f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.resolved_formal_parameters_with_type */
extern T0* T1526f65(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.resolved_formal_parameters_with_type */
extern T0* T1525f68(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.resolved_formal_parameters_with_type */
extern T0* T1524f66(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LABELED_ACTUAL_PARAMETER.resolved_formal_parameters_with_type */
extern T0* T1522f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.resolved_formal_parameters_with_type */
extern T0* T242f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.resolved_formal_parameters_with_type */
extern T0* T449f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.resolved_formal_parameters_with_type */
extern T0* T74f228(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.resolved_formal_parameters_with_type */
extern T0* T72f106(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.resolved_formal_parameters */
extern T0* T1501f67(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.resolved_formal_parameters_with_type_mark */
extern T0* T1501f72(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1501f72ot1(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.type_with_type_mark */
extern T0* T1526f67(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.type_with_type_mark */
extern T0* T1525f70(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.set_right_brace */
extern void T1525f97(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.set_left_brace */
extern void T1525f96(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.type_with_type_mark */
extern T0* T1524f68(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.set_type_mark */
extern void T1524f97(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.twin */
extern T0* T1524f78(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.type_with_type_mark */
extern T0* T449f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.set_current_keyword */
extern void T449f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.resolved_formal_parameters */
extern T0* T891f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.resolved_formal_parameters_with_type_mark */
extern T0* T891f74(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T891f74ot1(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.resolved_formal_parameters */
extern T0* T1526f64(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.resolved_formal_parameters_with_type_mark */
extern T0* T1526f76(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.resolved_formal_parameters */
extern T0* T1525f67(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.resolved_formal_parameters_with_type_mark */
extern T0* T1525f79(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_LIKE_FEATURE.resolved_formal_parameters */
extern T0* T1524f65(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.resolved_formal_parameters_with_type_mark */
extern T0* T1524f77(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.resolved_formal_parameters */
extern T0* T242f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.resolved_formal_parameters_with_type_mark */
extern T0* T242f64(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T242f64ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_CURRENT.resolved_formal_parameters */
extern T0* T449f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.resolved_formal_parameters_with_type_mark */
extern T0* T449f67(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLASS.resolved_formal_parameters */
extern T0* T74f227(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.resolved_formal_parameters_with_type_mark */
extern T0* T74f249(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.resolved_formal_parameters */
extern T0* T1957f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.resolved_formal_parameters */
extern T0* T603f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.resolved_formal_parameters */
extern T0* T240f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BASE_TYPE_LIST.item */
extern T0* T412f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].force_new */
extern void T1307f45(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].key_storage_put */
extern void T1307f60(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].item_storage_put */
extern void T1307f59(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].slots_put */
extern void T1307f58(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].clashes_put */
extern void T1307f57(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].slots_item */
extern T6 T1307f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].hash_position */
extern T6 T1307f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].resize */
extern void T1307f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].clashes_resize */
extern void T1307f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].special_integer_ */
extern T0* T1307f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].key_storage_resize */
extern void T1307f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].item_storage_resize */
extern void T1307f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_BASE_TYPE].aliased_resized_area */
extern T0* T924f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_BASE_TYPE].aliased_resized_area */
extern T0* T923f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].key_storage_item */
extern T0* T1307f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].slots_resize */
extern void T1307f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].new_modulus */
extern T6 T1307f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].new_capacity */
extern T6 T1307f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvagp0a_error */
extern void T53f216(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvagp0a */
extern T0* T785c866(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvagp_error */
extern T1 T53f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvagp0a_error */
extern void T36f216(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_gvagp_error */
extern T1 T36f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].found_item */
extern T0* T1307f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].found */
extern T1 T1307f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].search */
extern void T1307f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].search_position */
extern void T1307f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vhpr2a_error */
extern void T53f215(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vhpr2a */
extern T0* T785c865(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vhpr2_error */
extern T1 T53f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vhpr2a_error */
extern void T36f215(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_vhpr2_error */
extern T1 T36f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_frozen */
extern T1 T74f217(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.has_frozen_mark */
extern T1 T74f224(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvhso2a_error */
extern void T53f214(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_VALIDITY_ERROR.make_gvhso2a */
extern T0* T785c864(GE_context* ac, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvhso2_error */
extern T1 T53f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvhso2a_error */
extern void T36f214(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_gvhso2_error */
extern T1 T36f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvhso1a_error */
extern void T53f212(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_VALIDITY_ERROR.make_gvhso1a */
extern T0* T785c862(GE_context* ac, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvhso1_error */
extern T1 T53f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvhso1a_error */
extern void T36f212(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_gvhso1_error */
extern T1 T36f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vhay0a_error */
extern void T53f211(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_VALIDITY_ERROR.make_vhay0a */
extern T0* T785c861(GE_context* ac, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vhay_error */
extern T1 T53f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vhay0a_error */
extern void T36f211(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_vhay_error */
extern T1 T36f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vhpr1b_error */
extern void T53f210(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_vhpr1b */
extern T0* T785c860(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_vhpr1b_error */
extern void T36f210(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_PARENT_LIST.is_conforming */
extern T1 T568f7(GE_context* ac, T0* C);
/* [detachable] ET_ANCESTOR_BUILDER.current_system */
extern T0* T458f12(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].item */
extern T0* T462f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].sort */
extern void T1306f12(GE_context* ac, T0* C);
/* [detachable] ET_ANCESTOR_BUILDER.add_class_to_sorter */
extern void T458f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.system_object_type */
extern T0* T187f87(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.system_object_type */
extern T0* T58f116(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.system_object_type */
extern T0* T40f11(GE_context* ac, T0* C);
/* [detachable] ET_ANCESTOR_BUILDER.current_universe */
extern T0* T458f11(GE_context* ac, T0* C);
/* [detachable] ET_ANCESTOR_BUILDER.add_parents_to_sorter */
extern void T458f22(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].force_relation */
extern void T1306f15(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].put_indexed_relation */
extern void T1306f19(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].count */
extern T6 T1306f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].index_of */
extern T6 T1306f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].found_item */
extern T6 T1655f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].item_storage_item */
extern T6 T1655f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].found */
extern T1 T1655f23(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].search */
extern void T1655f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].search_position */
extern void T1655f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].hash_position */
extern T6 T1655f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].key_storage_item */
extern T0* T1655f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].clashes_item */
extern T6 T1655f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].slots_item */
extern T6 T1655f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].force */
extern void T1306f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].put */
extern void T1306f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].put_new */
extern void T1655f37(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].key_storage_put */
extern void T1655f54(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].item_storage_put */
extern void T1655f53(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].slots_put */
extern void T1655f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].clashes_put */
extern void T1655f49(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].resize */
extern void T1655f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].clashes_resize */
extern void T1655f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].special_integer_ */
extern T0* T1655f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].key_storage_resize */
extern void T1655f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].item_storage_resize */
extern void T1655f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].slots_resize */
extern void T1655f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].new_modulus */
extern T6 T1655f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].capacity */
extern T6 T1306f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].has */
extern T1 T1306f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].has */
extern T1 T1655f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCESTOR_BUILDER.set_fatal_error */
extern void T458f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_ancestors_error */
extern void T74f256(GE_context* ac, T0* C);
/* [detachable] ET_ANCESTOR_BUILDER.make */
extern T0* T458c13(GE_context* ac, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.make */
extern T0* T1309c20(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32].make */
extern T0* T1660c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32].set_internal_cursor */
extern void T1660f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] NATURAL_32].new_cursor */
extern T0* T1660f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] NATURAL_32].make */
extern T0* T1907c5(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT].make */
extern T0* T1659c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT].set_internal_cursor */
extern void T1659f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_BASE_TYPE_CONSTRAINT].new_cursor */
extern T0* T1659f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_BASE_TYPE_CONSTRAINT].make */
extern T0* T1906c5(GE_context* ac, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_CHECKER1.make */
extern void T1309f20p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.make */
extern T0* T1308c8(GE_context* ac, T0* a1);
/* [detachable] ET_PARENT_CHECKER1.tokens */
extern T0* T1308f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].make_map */
extern T0* T1307c40(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].make_with_equality_testers */
extern void T1307f46(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].make_sparse_container */
extern void T1307f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].set_internal_cursor */
extern void T1307f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].new_cursor */
extern T0* T1307f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_BASE_TYPE, [attached] ET_CLASS].make */
extern T0* T1656c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].make_slots */
extern void T1307f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].make_clashes */
extern void T1307f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].make_key_storage */
extern void T1307f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_BASE_TYPE, [attached] ET_CLASS].make_item_storage */
extern void T1307f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].make_default */
extern T0* T1306c11(GE_context* ac);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_CLASS].make */
extern void T1306f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].make_map */
extern T0* T1655c34(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].make_with_equality_testers */
extern void T1655f40(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].make_sparse_container */
extern void T1655f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].set_internal_cursor */
extern void T1655f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].new_cursor */
extern T0* T1655f32(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_CLASS].make */
extern T0* T1903c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].make_slots */
extern void T1655f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].make_clashes */
extern void T1655f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].make_key_storage */
extern void T1655f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_CLASS].make_item_storage */
extern void T1655f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ANCESTOR_BUILDER.make */
extern void T458f13p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCESTOR_BUILDER.tokens */
extern T0* T458f9(GE_context* ac, T0* C);
/* [detachable] ET_ANCESTOR_BUILDER.error_handler */
extern T0* T458f10(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_ancestors_built */
extern void T74f267(GE_context* ac, T0* C);
/* [detachable] ET_IMPLEMENTATION_CHECKER.process_class */
extern void T461f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.internal_process_class */
extern void T461f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.post_action */
extern void T461f32(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_implementation_checked */
extern void T74f272(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_suppliers */
extern void T74f271(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].extend */
extern void T948f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].put */
extern void T948f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.check_invariants_validity */
extern void T461f31(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.flat_mode */
extern T1 T461f17(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.check_invariants_validity */
extern void T380f82(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ATTACHMENT_SCOPE.wipe_out */
extern void T397f12(GE_context* ac, T0* C);
/* [detachable] ET_ITERATION_ITEM_SCOPE.wipe_out */
extern void T396f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ITERATION_COMPONENT].wipe_out */
extern void T744f11(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].key_for_iteration */
extern T0* T395f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].cursor_key */
extern T0* T395f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].wipe_out */
extern void T395f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].slots_wipe_out */
extern void T395f55(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].clashes_wipe_out */
extern void T395f54(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].key_storage_wipe_out */
extern void T395f53(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].item_storage_wipe_out */
extern void T395f52(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].move_all_cursors_after */
extern void T395f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].forth */
extern void T395f45(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].cursor_forth */
extern void T395f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].add_traversing_cursor */
extern void T395f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].remove_traversing_cursor */
extern void T395f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].item_for_iteration */
extern T0* T395f23(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].cursor_item */
extern T0* T395f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].after */
extern T1 T395f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].cursor_after */
extern T1 T395f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].after */
extern T1 T740f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].start */
extern void T395f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].cursor_start */
extern void T395f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].cursor_off */
extern T1 T395f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].off */
extern T1 T740f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].is_empty */
extern T1 T395f40(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST_SCOPE.wipe_out */
extern void T393f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST].wipe_out */
extern void T739f11(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].wipe_out */
extern void T392f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].slots_wipe_out */
extern void T392f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].clashes_wipe_out */
extern void T392f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].key_storage_wipe_out */
extern void T392f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].item_storage_wipe_out */
extern void T392f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].move_all_cursors_after */
extern void T392f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].set_next_cursor */
extern void T733f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].set_after */
extern void T733f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].forth */
extern void T392f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].cursor_forth */
extern void T392f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].add_traversing_cursor */
extern void T392f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].set_position */
extern void T733f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].remove_traversing_cursor */
extern void T392f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].item_for_iteration */
extern T0* T392f23(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].cursor_item */
extern T0* T392f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].after */
extern T1 T392f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].cursor_after */
extern T1 T392f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].after */
extern T1 T733f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].start */
extern void T392f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].cursor_start */
extern void T392f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].cursor_off */
extern T1 T392f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].off */
extern T1 T733f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].is_empty */
extern T1 T392f27(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.set_validity_error */
extern void T1198f33(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.set_validity_checked */
extern void T1198f32(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.assertion */
extern T0* T1198f11(GE_context* ac, T0* C, T6 a1);
/* detachable ET_INVARIANTS.item */
extern T0* T1198f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_CLASS.same_as_base_class */
extern T1 T74f71(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.reset_validity_checked */
extern void T1198f31(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.set_supplier_handler */
extern void T380f81(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TYPE_CHECKER.set_supplier_handler */
extern void T384f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.check_features_validity */
extern void T461f30(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.check_procedure_validity */
extern void T461f36(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].wipe_out */
extern void T1339f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].slots_wipe_out */
extern void T1339f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].clashes_wipe_out */
extern void T1339f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].key_storage_wipe_out */
extern void T1339f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].item_storage_wipe_out */
extern void T1339f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].move_all_cursors_after */
extern void T1339f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_PROCEDURE].set_next_cursor */
extern void T1712f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_PROCEDURE].set_after */
extern void T1712f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].forth */
extern void T1339f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].cursor_forth */
extern void T1339f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].add_traversing_cursor */
extern void T1339f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_PROCEDURE].set_position */
extern void T1712f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].remove_traversing_cursor */
extern void T1339f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.check_precursor_procedure_validity */
extern void T461f40(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] ET_FEATURE_CHECKER.check_precursor_feature_validity */
extern void T380f87(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].item_for_iteration */
extern T0* T1339f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].cursor_item */
extern T0* T1339f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].after */
extern T1 T1339f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].cursor_after */
extern T1 T1339f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_PROCEDURE].after */
extern T1 T1712f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].start */
extern void T1339f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].cursor_start */
extern void T1339f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].cursor_off */
extern T1 T1339f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_PROCEDURE].off */
extern T1 T1712f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].is_empty */
extern T1 T1339f21(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.check_feature_validity */
extern void T380f84(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_EXTERNAL_PROCEDURE.set_validity_error */
extern void T1516f92(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.set_validity_error */
extern void T253f94(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.set_validity_error */
extern void T476f82(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.set_validity_error */
extern void T1515f86(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.set_validity_error */
extern void T469f83(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.set_validity_error */
extern void T1513f85(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.set_validity_error */
extern void T1512f88(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.set_validity_error */
extern void T1511f86(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_validity_error */
extern void T371f86(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_validity_error */
extern void T369f85(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_validity_error */
extern void T359f87(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.set_validity_error */
extern void T360f82(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.set_validity_checked */
extern void T1516f91(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.set_validity_checked */
extern void T253f93(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.set_validity_checked */
extern void T476f81(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.set_validity_checked */
extern void T1515f85(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.set_validity_checked */
extern void T469f82(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.set_validity_checked */
extern void T1513f84(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.set_validity_checked */
extern void T1512f87(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.set_validity_checked */
extern void T1511f85(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.set_validity_checked */
extern void T371f85(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.set_validity_checked */
extern void T369f84(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.set_validity_checked */
extern void T359f86(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.set_validity_checked */
extern void T360f81(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.build_preconditions_attachment_scope */
extern void T380f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].wipe_out */
extern void T400f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].slots_wipe_out */
extern void T400f57(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].clashes_wipe_out */
extern void T400f56(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].key_storage_wipe_out */
extern void T400f55(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].item_storage_wipe_out */
extern void T400f54(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_ASSERTIONS].keep_head */
extern void T751f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ASSERTIONS].set_count */
extern void T751f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].unset_found_item */
extern void T400f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].move_all_cursors_after */
extern void T400f53(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].set_next_cursor */
extern void T750f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].set_after */
extern void T750f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].forth */
extern void T400f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].cursor_forth */
extern void T400f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].add_traversing_cursor */
extern void T400f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].set_position */
extern void T750f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].remove_traversing_cursor */
extern void T400f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].clashes_item */
extern T6 T400f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].is_last */
extern T1 T400f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].cursor_is_last */
extern T1 T400f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].is_empty */
extern T1 T400f26(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].item_for_iteration */
extern T0* T400f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].cursor_item */
extern T0* T400f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].item_storage_item */
extern T0* T400f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].start */
extern void T400f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].cursor_start */
extern void T400f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].cursor_off */
extern T1 T400f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].off */
extern T1 T750f5(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.build_assertions_attachment_scope */
extern void T380f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].first */
extern T0* T400f20(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].force_last */
extern void T400f39(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].key_storage_put */
extern void T400f50(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].slots_put */
extern void T400f49(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].clashes_put */
extern void T400f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].slots_item */
extern T6 T400f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].hash_position */
extern T6 T400f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].resize */
extern void T400f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].clashes_resize */
extern void T400f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].special_integer_ */
extern T0* T400f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].key_storage_resize */
extern void T400f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].item_storage_resize */
extern void T400f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTIONS].aliased_resized_area */
extern T0* T753f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ASSERTIONS].aliased_resized_area */
extern T0* T751f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].key_storage_item */
extern T0* T400f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].slots_resize */
extern void T400f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].new_modulus */
extern T6 T400f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].new_capacity */
extern T6 T400f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].item_storage_put */
extern void T400f46(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTIONS].force */
extern void T753f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ASSERTIONS].fill_with */
extern void T751f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ASSERTIONS].extend */
extern void T751f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].search_position */
extern void T400f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.add_precursors_with_preconditions_recursive */
extern void T380f89(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_EXTERNAL_PROCEDURE.other_precursors */
extern T0* T1516f7(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.other_precursors */
extern T0* T253f7(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.other_precursors */
extern T0* T476f16(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.other_precursors */
extern T0* T1515f8(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.other_precursors */
extern T0* T469f16(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.other_precursors */
extern T0* T1513f7(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.other_precursors */
extern T0* T1512f7(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.other_precursors */
extern T0* T1511f7(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.other_precursors */
extern T0* T371f9(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.other_precursors */
extern T0* T369f8(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.other_precursors */
extern T0* T359f7(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.other_precursors */
extern T0* T360f7(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.check_preconditions_validity */
extern void T380f76(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_PRECONDITIONS.set_validity_error */
extern void T406f21(GE_context* ac, T0* C);
/* detachable ET_PRECONDITIONS.set_validity_checked */
extern void T406f20(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.feature_checker */
extern T0* T380f48(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.validity_checked */
extern T1 T371f22(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.validity_checked */
extern T1 T369f20(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.validity_checked */
extern T1 T360f18(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.validity_checked */
extern T1 T359f25(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.validity_checked */
extern T1 T253f28(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.validity_checked */
extern T1 T1516f26(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.validity_checked */
extern T1 T1515f26(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.validity_checked */
extern T1 T1513f25(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.validity_checked */
extern T1 T1512f28(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.validity_checked */
extern T1 T1511f27(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.validity_checked */
extern T1 T476f12(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.validity_checked */
extern T1 T469f12(GE_context* ac, T0* C);
/* [detachable] ET_IMPLEMENTATION_CHECKER.check_assertions_validity */
extern void T461f38(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_IMPLEMENTATION_CHECKER.flat_dbc_mode */
extern T1 T461f18(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.check_postconditions_validity */
extern void T380f86(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_POSTCONDITIONS.set_validity_error */
extern void T1244f23(GE_context* ac, T0* C);
/* detachable ET_POSTCONDITIONS.set_validity_checked */
extern void T1244f22(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.set_precursor_procedures */
extern void T380f85(GE_context* ac, T0* C, T0* a1);
/* detachable ET_EXTERNAL_PROCEDURE.reset_validity_checked */
extern void T1516f93(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.reset_validity_checked */
extern void T476f84(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.reset_validity_checked */
extern void T1515f87(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.reset_validity_checked */
extern void T469f85(GE_context* ac, T0* C);
/* [detachable] ET_IMPLEMENTATION_CHECKER.check_query_validity */
extern void T461f35(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].wipe_out */
extern void T1340f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].slots_wipe_out */
extern void T1340f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].clashes_wipe_out */
extern void T1340f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].key_storage_wipe_out */
extern void T1340f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].item_storage_wipe_out */
extern void T1340f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].move_all_cursors_after */
extern void T1340f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_QUERY].set_next_cursor */
extern void T1713f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_QUERY].set_after */
extern void T1713f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].forth */
extern void T1340f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].cursor_forth */
extern void T1340f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].add_traversing_cursor */
extern void T1340f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_QUERY].set_position */
extern void T1713f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].remove_traversing_cursor */
extern void T1340f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.check_precursor_query_validity */
extern void T461f39(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].item_for_iteration */
extern T0* T1340f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].cursor_item */
extern T0* T1340f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].after */
extern T1 T1340f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].cursor_after */
extern T1 T1340f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_QUERY].after */
extern T1 T1713f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].start */
extern void T1340f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].cursor_start */
extern void T1340f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].cursor_off */
extern T1 T1340f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_QUERY].off */
extern T1 T1713f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].is_empty */
extern T1 T1340f21(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.set_precursor_queries */
extern void T380f83(GE_context* ac, T0* C, T0* a1);
/* detachable ET_POSTCONDITIONS.reset_validity_checked */
extern void T1244f21(GE_context* ac, T0* C);
/* detachable ET_PRECONDITIONS.reset_validity_checked */
extern void T406f22(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.reset_validity_checked */
extern void T253f95(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.reset_validity_checked */
extern void T1513f86(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.reset_validity_checked */
extern void T1512f89(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.reset_validity_checked */
extern void T1511f87(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.reset_validity_checked */
extern void T371f87(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.reset_validity_checked */
extern void T369f86(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.reset_validity_checked */
extern void T359f88(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.reset_validity_checked */
extern void T360f83(GE_context* ac, T0* C);
/* [detachable] ET_SUPPLIER_BUILDER.set */
extern void T1341f6(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_IMPLEMENTATION_CHECKER.suppliers_enabled */
extern T1 T461f15(GE_context* ac, T0* C);
/* [detachable] ET_IMPLEMENTATION_CHECKER.check_signatures_validity */
extern void T461f29(GE_context* ac, T0* C);
/* [detachable] ET_IMPLEMENTATION_CHECKER.check_signature_validity */
extern void T461f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.resolve_feature_adaptations */
extern void T461f33(GE_context* ac, T0* C);
/* [detachable] ET_IMPLEMENTATION_CHECKER.resolve_inherited_features */
extern void T461f37(GE_context* ac, T0* C, T0* a1);
extern T1 T461f37ot1(T0* a1, T0** a2);
/* [detachable] ET_IMPLEMENTATION_CHECKER.check_parents_validity */
extern void T461f28(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_CHECKER3.check_parents_validity */
extern void T1338f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.pre_action */
extern void T461f27(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.report_class_postponed */
extern void T64f108(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.report_class_postponed */
extern void T375f112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.process_parent_class */
extern void T461f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_checking_implementation */
extern void T74f273(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.set_fatal_error */
extern void T461f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_implementation_error */
extern void T74f259(GE_context* ac, T0* C);
/* [detachable] ET_IMPLEMENTATION_CHECKER.make */
extern T0* T461c19(GE_context* ac, T0* a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.make_with_feature_checker */
extern void T461f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SUPPLIER_BUILDER.make */
extern T0* T1341c5(GE_context* ac);
/* [detachable] ET_SUPPLIER_BUILDER.dummy_suppliers */
extern T0* T1341f4(GE_context* ac, T0* C);
/* [detachable] ET_SUPPLIER_BUILDER.tokens */
extern T0* T1341f3(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].make */
extern T0* T1340c35(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].set_internal_cursor */
extern void T1340f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].new_cursor */
extern T0* T1340f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_QUERY].make */
extern T0* T1713c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].make_slots */
extern void T1340f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].make_clashes */
extern void T1340f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].make_key_storage */
extern void T1340f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_QUERY].make_item_storage */
extern void T1340f39(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].make */
extern T0* T1339c35(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].set_internal_cursor */
extern void T1339f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].new_cursor */
extern T0* T1339f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_PROCEDURE].make */
extern T0* T1712c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].make_slots */
extern void T1339f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].make_clashes */
extern void T1339f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].make_key_storage */
extern void T1339f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_PROCEDURE].make_item_storage */
extern void T1339f39(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_PARENT_CHECKER3.make */
extern T0* T1338c6(GE_context* ac, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.make */
extern void T1338f6p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_PARENT_CHECKER3.tokens */
extern T0* T1338f4(GE_context* ac, T0* C);
/* [detachable] ET_IMPLEMENTATION_CHECKER.feature_name_tester */
extern T0* T461f14(GE_context* ac, T0* C);
/* [detachable] ET_IMPLEMENTATION_CHECKER.make_class_processor */
extern void T461f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.tokens */
extern T0* T461f16(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.make */
extern T0* T380c75(GE_context* ac, T0* a1);
/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER.make */
extern T0* T405c9(GE_context* ac, T0* a1);
/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER.make */
extern void T405f9p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].make */
extern T0* T404c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].set_internal_cursor */
extern void T404f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].new_cursor */
extern T0* T404f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].make */
extern T0* T762c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].make */
extern T0* T760f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].make_empty */
extern T0* T761c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_ADAPTED_CLASS]].default_create */
extern T0* T760c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM].make */
extern T0* T401c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM].set_internal_cursor */
extern void T401f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INDEXING_TERM].new_cursor */
extern T0* T401f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INDEXING_TERM].make */
extern T0* T757c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM].make */
extern T0* T754f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_INDEXING_TERM].make_empty */
extern T0* T756c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INDEXING_TERM].default_create */
extern T0* T754c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].make_map */
extern T0* T400c38(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].make_with_equality_testers */
extern void T400f43(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].make_sparse_container */
extern void T400f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].set_internal_cursor */
extern void T400f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].new_cursor */
extern T0* T400f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].make */
extern T0* T750c7(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].make_slots */
extern void T400f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].make_clashes */
extern void T400f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].make_key_storage */
extern void T400f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ASSERTIONS, [attached] ET_FEATURE].make_item_storage */
extern void T400f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTIONS].make */
extern T0* T753f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ASSERTIONS].make_empty */
extern T0* T751c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ASSERTIONS].default_create */
extern T0* T753c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE].make */
extern T0* T399c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE].set_internal_cursor */
extern void T399f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ATTACHMENT_SCOPE].new_cursor */
extern T0* T399f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ATTACHMENT_SCOPE].make */
extern T0* T749c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ATTACHMENT_SCOPE].make */
extern T0* T748f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ATTACHMENT_SCOPE].make_empty */
extern T0* T747c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ATTACHMENT_SCOPE].default_create */
extern T0* T748c3(GE_context* ac);
/* [detachable] ET_ATTACHMENT_SCOPE_BUILDER.make */
extern T0* T398c3(GE_context* ac);
/* [detachable] ET_ITERATION_ITEM_SCOPE.make */
extern T0* T396c6(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ITERATION_COMPONENT].make */
extern T0* T744c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ITERATION_COMPONENT].set_internal_cursor */
extern void T744f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ITERATION_COMPONENT].new_cursor */
extern T0* T744f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ITERATION_COMPONENT].make */
extern T0* T1428c5(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].make_map */
extern T0* T395c43(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].make_with_equality_testers */
extern void T395f47(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].make_sparse_container */
extern void T395f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].set_internal_cursor */
extern void T395f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].new_cursor */
extern T0* T395f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].make */
extern T0* T740c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].make_slots */
extern void T395f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].make_clashes */
extern void T395f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].make_key_storage */
extern void T395f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_ITERATION_COMPONENT].make_item_storage */
extern void T395f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NESTED_TYPE_CONTEXT].make */
extern T0* T731f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_NESTED_TYPE_CONTEXT].make_empty */
extern T0* T730c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_NESTED_TYPE_CONTEXT].default_create */
extern T0* T731c3(GE_context* ac);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.make */
extern T0* T394c8(GE_context* ac, T0* a1);
/* [detachable] ET_OBJECT_TEST_SCOPE_BUILDER.tokens */
extern T0* T394f6(GE_context* ac, T0* C);
/* [detachable] ET_OBJECT_TEST_SCOPE.make */
extern T0* T393c6(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST].make */
extern T0* T739c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST].set_internal_cursor */
extern void T739f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NAMED_OBJECT_TEST].new_cursor */
extern T0* T739f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NAMED_OBJECT_TEST].make */
extern T0* T1427c5(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].make_map */
extern T0* T392c39(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].make_with_equality_testers */
extern void T392f43(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].make_sparse_container */
extern void T392f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].set_internal_cursor */
extern void T392f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].new_cursor */
extern T0* T392f26(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].make */
extern T0* T733c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].make_slots */
extern void T392f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].make_clashes */
extern void T392f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].make_key_storage */
extern void T392f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_NESTED_TYPE_CONTEXT, [attached] ET_NAMED_OBJECT_TEST].make_item_storage */
extern void T392f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].make */
extern T0* T391c12(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].set_internal_cursor */
extern void T391f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_NESTED_TYPE_CONTEXT].new_cursor */
extern T0* T391f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_NESTED_TYPE_CONTEXT].make */
extern T0* T732c5(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].make */
extern T0* T390c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].set_internal_cursor */
extern void T390f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].new_cursor */
extern T0* T390f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].make */
extern T0* T729c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].make */
extern T0* T726f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].make_empty */
extern T0* T728c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] TUPLE [detachable ET_QUERY, [attached] ET_CLASS, [attached] ET_NESTED_TYPE_CONTEXT]].default_create */
extern T0* T726c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].make */
extern T0* T389c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].set_internal_cursor */
extern void T389f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].new_cursor */
extern T0* T389f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].make */
extern T0* T725c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].make */
extern T0* T722f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].make_empty */
extern T0* T724c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_FEATURE]].default_create */
extern T0* T722c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].make */
extern T0* T388c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].set_internal_cursor */
extern void T388f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].new_cursor */
extern T0* T388f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].make */
extern T0* T721c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].make */
extern T0* T718f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].make_empty */
extern T0* T720c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_QUERY]].default_create */
extern T0* T718c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].make */
extern T0* T387c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].set_internal_cursor */
extern void T387f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].new_cursor */
extern T0* T387f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].make */
extern T0* T717c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].make */
extern T0* T714f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].make_empty */
extern T0* T716c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] ET_PROCEDURE]].default_create */
extern T0* T714c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].make */
extern T0* T386c12(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].set_internal_cursor */
extern void T386f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_INLINE_AGENT].new_cursor */
extern T0* T386f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_INLINE_AGENT].make */
extern T0* T713c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_AGENT].make */
extern T0* T711f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_INLINE_AGENT].make_empty */
extern T0* T712c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_AGENT].default_create */
extern T0* T711c3(GE_context* ac);
/* detachable ET_INVARIANTS.implementation_class */
extern T0* T1198f1(GE_context* ac, T0* C);
/* [detachable] ET_FEATURE_CHECKER.dummy_feature */
extern T0* T380f54(GE_context* ac, T0* C);
/* [detachable] ET_TYPE_CHECKER.make */
extern T0* T384c22(GE_context* ac, T0* a1);
/* [detachable] ET_TYPE_CHECKER.dummy_feature */
extern T0* T384f16(GE_context* ac, T0* C);
/* [detachable] ET_TYPE_CHECKER.make */
extern void T384f22p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FEATURE_CHECKER.make */
extern void T380f75p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_CHECKER.error_handler */
extern T0* T461f13(GE_context* ac, T0* C);
/* [detachable] ET_IMPLEMENTATION_CHECKER.process_none_class */
extern void T461f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_NULL_PROCESSOR.process_class */
extern void T464f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERFACE_CHECKER.process_class */
extern void T460f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].remove */
extern void T1329f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].remove_position */
extern void T1329f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].key_storage_put */
extern void T1329f49(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].key_storage_keep_head */
extern void T1329f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].item_storage_keep_head */
extern void T1329f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].move_cursors_forth */
extern void T1329f43(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_CLASS].set_position */
extern void T1701f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_CLASS].set_next_cursor */
extern void T1701f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_CLASS].set_after */
extern void T1701f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].internal_set_key_equality_tester */
extern void T1329f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].last */
extern T0* T1329f17(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].is_empty */
extern T1 T1329f28(GE_context* ac, T0* C);
/* [detachable] ET_INTERFACE_CHECKER.internal_process_class */
extern void T460f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERFACE_CHECKER.check_constraint_creations_validity */
extern void T460f20(GE_context* ac, T0* C);
/* [detachable] ET_INTERFACE_CHECKER.check_constraint_creation_validity */
extern void T460f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vggc3b_error */
extern void T53f223(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vggc3b */
extern T0* T785c873(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.types_to_text */
extern T0* T1850f5(GE_context* ac, T0* C);
/* [detachable] ET_BASE_TYPE_RENAME_CONSTRAINT.append_types_to_string */
extern void T1850f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_BASE_TYPE_CONSTRAINT_LIST.types_to_text */
extern T0* T1661f8(GE_context* ac, T0* C);
/* [detachable] ET_BASE_TYPE_CONSTRAINT_LIST.append_types_to_string */
extern void T1661f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vggc3_error */
extern T1 T53f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vggc3b_error */
extern void T36f223(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_vggc3_error */
extern T1 T36f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vggc3a_error */
extern void T53f222(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vggc3a */
extern T0* T785c872(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vggc3a_error */
extern void T36f222(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vggc3c_error */
extern void T53f221(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vggc3c */
extern T0* T785c871(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vggc3c_error */
extern void T36f221(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vggc3d_error */
extern void T53f220(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vggc3d */
extern T0* T785c870(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.report_vggc3d_error */
extern void T36f220(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vgcp3c_error */
extern void T53f219(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_vgcp3c */
extern T0* T785c869(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vgcp3c_error */
extern void T36f219(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_INTERFACE_CHECKER.check_constraint_renamings_validity */
extern void T460f19(GE_context* ac, T0* C);
/* [detachable] ET_INTERFACE_CHECKER.check_constraint_renaming_validity */
extern void T460f22(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T460f22ot1(T0* a1, T0** a2);
extern T1 T460f22ot2(T0* a1, T0** a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].wipe_out */
extern void T1333f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].slots_wipe_out */
extern void T1333f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].clashes_wipe_out */
extern void T1333f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].key_storage_wipe_out */
extern void T1333f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].item_storage_wipe_out */
extern void T1333f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].unset_found_item */
extern void T1333f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].move_all_cursors_after */
extern void T1333f45(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].set_next_cursor */
extern void T1711f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].set_after */
extern void T1711f7(GE_context* ac, T0* C);
/* [detachable] ET_INTERFACE_CHECKER.check_constraint_not_renamed_names_validity */
extern void T460f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1t_error */
extern void T53f245(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1t */
extern T0* T785c895(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.report_vfav1t_error */
extern void T36f245(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1s_error */
extern void T53f244(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1s */
extern T0* T785c894(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.report_vfav1s_error */
extern void T36f244(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav3g_error */
extern void T53f243(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vfav3g */
extern T0* T785c893(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.report_vfav3g_error */
extern void T36f243(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav2g_error */
extern void T53f242(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vfav2g */
extern T0* T785c892(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.report_vfav2g_error */
extern void T36f242(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].found_key */
extern T0* T1333f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].key_storage_item */
extern T0* T1333f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].found_item */
extern T0* T1333f27(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].item_storage_item */
extern T0* T1333f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].found */
extern T1 T1333f26(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].search */
extern void T1333f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].search_position */
extern void T1333f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].hash_position */
extern T6 T1333f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].clashes_item */
extern T6 T1333f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].slots_item */
extern T6 T1333f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vggc2f_error */
extern void T53f241(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vggc2f */
extern T0* T785c891(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vggc2_error */
extern T1 T53f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vggc2f_error */
extern void T36f241(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_vggc2_error */
extern T1 T36f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_FEATURE_NAME].has */
extern T1 T1332f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERFACE_CHECKER.check_constraint_renamed_names_validity */
extern void T460f24(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vggc2e_error */
extern void T53f240(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vggc2e */
extern T0* T785c890(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vggc2e_error */
extern void T36f240(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1r_error */
extern void T53f237(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1r */
extern T0* T785c887(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.report_vfav1r_error */
extern void T36f237(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav4l_error */
extern void T53f235(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vfav4l */
extern T0* T785c885(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vfav4l_error */
extern void T36f235(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav1q_error */
extern void T53f234(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vfav1q */
extern T0* T785c884(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.report_vfav1q_error */
extern void T36f234(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav4k_error */
extern void T53f232(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vfav4k */
extern T0* T785c882(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vfav4k_error */
extern void T36f232(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav3f_error */
extern void T53f231(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vfav3f */
extern T0* T785c881(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.report_vfav3f_error */
extern void T36f231(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav4j_error */
extern void T53f229(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vfav4j */
extern T0* T785c879(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vfav4j_error */
extern void T36f229(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].put_new */
extern void T1333f41(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].key_storage_put */
extern void T1333f58(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].item_storage_put */
extern void T1333f57(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].slots_put */
extern void T1333f52(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].clashes_put */
extern void T1333f53(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav2f_error */
extern void T53f228(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_VALIDITY_ERROR.make_vfav2f */
extern T0* T785c878(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ERROR_HANDLER.report_vfav2f_error */
extern void T36f228(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].item */
extern T0* T1333f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].has */
extern T1 T1333f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vfav4i_error */
extern void T53f226(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vfav4i */
extern T0* T785c876(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vfav4i_error */
extern void T36f226(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vggc2a_error */
extern void T53f225(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vggc2a */
extern T0* T785c875(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vggc2a_error */
extern void T36f225(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vggc2d_error */
extern void T53f224(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_VALIDITY_ERROR.make_vggc2d */
extern T0* T785c874(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ERROR_HANDLER.report_vggc2d_error */
extern void T36f224(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].resize */
extern void T1333f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].clashes_resize */
extern void T1333f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].special_integer_ */
extern T0* T1333f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].key_storage_resize */
extern void T1333f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].item_storage_resize */
extern void T1333f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].slots_resize */
extern void T1333f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].new_modulus */
extern T6 T1333f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vggc2c_error */
extern void T53f218(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vggc2c */
extern T0* T785c868(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_TUPLE_TYPE.tuple_keyword */
extern T0* T254f4(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.name */
extern T0* T72f2(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.name */
extern T0* T74f13(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_vggc2c_error */
extern void T36f218(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vggc2b_error */
extern void T53f217(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_vggc2b */
extern T0* T785c867(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vggc2b_error */
extern void T36f217(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_INTERFACE_CHECKER.resolve_signatures_unfolded_tuple_actual_parameters */
extern void T460f18(GE_context* ac, T0* C);
/* [detachable] ET_INTERFACE_CHECKER.resolve_signature_unfolded_tuple_actual_parameters */
extern void T460f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.resolve_type */
extern void T1331f7(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_INTERFACE_CHECKER.check_qualified_anchored_signatures_validity */
extern void T460f17(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.check_signatures_validity */
extern void T1330f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERFACE_CHECKER.set_fatal_error */
extern void T460f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_interface_error */
extern void T74f258(GE_context* ac, T0* C);
/* [detachable] ET_INTERFACE_CHECKER.make */
extern T0* T460c13(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].set_key_equality_tester */
extern void T1333f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERFACE_CHECKER.alias_name_tester */
extern T0* T460f11(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].make_map */
extern T0* T1333c37(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].make_with_equality_testers */
extern void T1333f44(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].make_sparse_container */
extern void T1333f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].set_internal_cursor */
extern void T1333f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].new_cursor */
extern T0* T1333f35(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].make */
extern T0* T1711c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].make_slots */
extern void T1333f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].make_clashes */
extern void T1333f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].make_key_storage */
extern void T1333f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_RENAME, [attached] ET_ALIAS_NAME].make_item_storage */
extern void T1333f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_INTERFACE_CHECKER.feature_name_tester */
extern T0* T460f9(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.make */
extern T0* T1331c6(GE_context* ac, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.make */
extern void T1331f6p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS_RESOLVER2.tokens */
extern T0* T1331f4(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.set_classes_to_be_processed */
extern void T1330f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.make */
extern T0* T1330c13(GE_context* ac, T0* a1);
/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER.set_feature_flattening_error_only */
extern void T405f10(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_CHECKER.make */
extern void T1330f13p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].make */
extern T0* T1329c32(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].set_internal_cursor */
extern void T1329f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].new_cursor */
extern T0* T1329f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_CLASS].make */
extern T0* T1701c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].make_slots */
extern void T1329f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].make_clashes */
extern void T1329f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].make_key_storage */
extern void T1329f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_CLASS].make_item_storage */
extern void T1329f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_INTERFACE_CHECKER.make */
extern void T460f13p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERFACE_CHECKER.tokens */
extern T0* T460f10(GE_context* ac, T0* C);
/* [detachable] ET_INTERFACE_CHECKER.error_handler */
extern T0* T460f12(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.set_interface_checked */
extern void T74f270(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].force_last */
extern void T226f35(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].key_storage_put */
extern void T226f43(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].slots_put */
extern void T226f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].clashes_put */
extern void T226f41(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].slots_item */
extern T6 T226f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].hash_position */
extern T6 T226f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].resize */
extern void T226f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].clashes_resize */
extern void T226f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].special_integer_ */
extern T0* T226f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].key_storage_resize */
extern void T226f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].item_storage_resize */
extern void T226f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].key_storage_item */
extern T0* T226f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].clashes_item */
extern T6 T226f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].slots_resize */
extern void T226f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].new_modulus */
extern T6 T226f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].new_capacity */
extern T6 T226f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].item_storage_put */
extern void T226f39(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].search_position */
extern void T226f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].unset_found_item */
extern void T226f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].found_item */
extern T0* T226f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].item_storage_item */
extern T0* T226f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].found */
extern T1 T226f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].search */
extern void T226f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.append_runtime_name_to_string */
extern void T74f265(GE_context* ac, T0* C, T0* a1);
/* detachable ET_FORMAL_PARAMETER_LIST.append_runtime_name_to_string */
extern void T603f30(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.append_runtime_name_to_string */
extern void T1501f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.append_runtime_name_to_string */
extern void T891f86(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.append_runtime_name_to_string */
extern void T254f120(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.append_runtime_name_to_string */
extern void T1957f23(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.append_runtime_name_to_string */
extern void T1526f86(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.append_runtime_name_to_string */
extern void T1525f87(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.append_runtime_name_to_string */
extern void T1524f87(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.append_runtime_name_to_string */
extern void T242f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.append_runtime_name_to_string */
extern void T449f80(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.append_runtime_name_to_string */
extern void T240f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.append_runtime_name_to_string */
extern void T72f126(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.append_runtime_name_to_string */
extern void T577f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.append_runtime_name_to_string */
extern void T1956f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.type_name_buffer */
extern T0* T63f64(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.new_dynamic_primary_type */
extern T0* T63f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.register_dynamic_type */
extern void T63f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.propagate_conforming_ancestors */
extern void T63f105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.null_system_processor */
extern T0* T73f24(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.make_null */
extern void T64f59(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.make_null */
extern T0* T64c59(GE_context* ac);
/* [detachable] ET_SYSTEM_PROCESSOR.set_qualified_anchored_types_cycle_detection_enabled_only */
extern void T64f74(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_unknown_builtin_reported_only */
extern void T64f73(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_preparse_shallow_mode_only */
extern void T64f72(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.set_default_keyword_usage_only */
extern void T64f71(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.set_use_detachable_keyword_only */
extern void T64f79(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_use_attached_keyword_only */
extern void T64f78(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_use_reference_keyword_only */
extern void T64f77(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_use_note_keyword_only */
extern void T64f76(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_use_attribute_keyword_only */
extern void T64f75(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DOTNET_ASSEMBLY_CLASSIC_CONSUMER.make */
extern T0* T467c3(GE_context* ac, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.null_error_handler */
extern T0* T73f124(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.make_null */
extern T0* T53c180(GE_context* ac);
/* [detachable] ET_NULL_ERROR_HANDLER.make_null */
extern void T53f180p1(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.null_output_stream */
extern T0* T53f24(GE_context* ac, T0* C);
/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM.make */
extern T0* T161c3(GE_context* ac, T0* a1);
/* [detachable] ET_TOKEN_CONSTANTS.null_ast_processor */
extern T0* T73f123(GE_context* ac, T0* C);
/* [detachable] ET_AST_NULL_PROCESSOR.make */
extern T0* T464c1(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.default_ast_factory */
extern T0* T73f122(GE_context* ac, T0* C);
/* [detachable] ET_AST_FACTORY.make */
extern T0* T465c370(GE_context* ac);
/* [detachable] ET_SYSTEM_PROCESSOR.tokens */
extern T0* T64f45(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].make_default */
extern T0* T263c23(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] INTEGER_32].default_capacity */
extern T6 T263f10(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.conforming_ancestors */
extern T0* T451f9(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.conforming_ancestors */
extern T0* T450f9(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.conforming_ancestors */
extern T0* T302f9(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.conforming_ancestors */
extern T0* T319f9(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.conforming_ancestors */
extern T0* T223f9(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.set_next_type */
extern void T451f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.set_next_type */
extern void T450f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.set_next_type */
extern void T302f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.set_next_type */
extern void T319f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_next_type */
extern void T223f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.set_index */
extern void T74f262(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.set_hash_code */
extern void T451f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.set_hash_code */
extern void T450f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.set_hash_code */
extern void T302f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.set_hash_code */
extern void T319f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_hash_code */
extern void T223f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].force_last */
extern void T76f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].resize */
extern void T76f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].new_capacity */
extern T6 T76f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].extendible */
extern T1 T76f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.make */
extern T0* T223c62(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T275c37(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].set_internal_cursor */
extern void T275f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].new_cursor */
extern T0* T275f18(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T334c9(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].make_slots */
extern void T275f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].make_clashes */
extern void T275f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].make_key_storage */
extern void T275f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].make_item_storage */
extern void T275f43(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.empty_features */
extern T0* T223f43(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FEATURE_LIST.make */
extern T0* T224c14(GE_context* ac);
/* [detachable] ET_DYNAMIC_SYSTEM.new_predicate_type */
extern T0* T63f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.make */
extern T0* T451c66(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.make_type */
extern void T451f85(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.empty_features */
extern T0* T451f59(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.empty_dynamic_type_sets */
extern T0* T63f71(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.new_function_type */
extern T0* T63f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.new_procedure_type */
extern T0* T63f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.make */
extern T0* T450c66(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.make_type */
extern void T450f85(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.empty_features */
extern T0* T450f59(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.new_type_type */
extern T0* T63f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.new_tuple_type */
extern T0* T63f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.make */
extern T0* T302c63(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.make_type */
extern void T302f75(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.empty_features */
extern T0* T302f50(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.new_special_type */
extern T0* T63f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.make */
extern T0* T319c65(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.make_type */
extern void T319f77(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.empty_features */
extern T0* T319f52(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.same_as_base_class */
extern T1 T254f12(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.same_as_base_class */
extern T1 T72f12(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.base_type_with_type_mark */
extern T0* T1501f24(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1501f24ot1(T0* a1, T0** a2);
extern T1 T1501f24ot2(T0* a1, T0** a2);
extern T1 T1501f24ot3(T0* a1, T0** a2);
/* [detachable] ET_FORMAL_PARAMETER.base_type_with_type_mark */
extern T0* T891f22(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T891f22ot1(T0* a1, T0** a2);
extern T1 T891f22ot2(T0* a1, T0** a2);
extern T1 T891f22ot3(T0* a1, T0** a2);
/* detachable ET_QUALIFIED_LIKE_TYPE.base_type_with_type_mark */
extern T0* T1526f21(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1526f21ot1(T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.base_type_with_type_mark */
extern T0* T1525f43(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T1525f43ot1(T0* a1);
/* detachable ET_LIKE_FEATURE.base_type_with_type_mark */
extern T0* T1524f7(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.base_type_with_type_mark */
extern T0* T242f22(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T242f22ot1(T0* a1, T0** a2);
extern T1 T242f22ot2(T0* a1, T0** a2);
extern T1 T242f22ot3(T0* a1, T0** a2);
/* [detachable] ET_CLASS.base_type_with_type_mark */
extern T0* T74f134(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_FORMAL_PARAMETER_LIST.named_types */
extern T0* T603f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.base_type_with_type_mark */
extern T0* T449f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NESTED_TYPE_CONTEXT.base_type_with_type_mark */
extern T0* T241f32(GE_context* ac, T0* C, T0* a1);
extern T1 T241f32ot1(T0* a1, T0** a2);
/* [detachable] ET_LIKE_N.base_type_with_type_mark */
extern T0* T1474f21(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_TUPLE_TYPE.context_base_type_with_type_mark */
extern T0* T254f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.context_base_type_with_type_mark */
extern T0* T72f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.context_base_type_with_type_mark */
extern T0* T74f110(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.base_type_with_type_mark */
extern T0* T254f9(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.named_types */
extern T0* T1957f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.named_parameter_with_type */
extern T0* T1501f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.named_parameter_with_type */
extern T0* T891f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LABELED_COMMA_ACTUAL_PARAMETER.named_parameter_with_type */
extern T0* T1769f4(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.named_parameter_with_type */
extern T0* T1526f31(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.named_parameter_with_type */
extern T0* T1525f27(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.named_parameter_with_type */
extern T0* T1524f21(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LABELED_ACTUAL_PARAMETER.named_parameter_with_type */
extern T0* T1522f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.named_parameter_with_type */
extern T0* T242f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.named_parameter_with_type */
extern T0* T449f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.named_parameter_with_type */
extern T0* T254f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.named_parameter_with_type */
extern T0* T74f141(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.named_parameter_with_type */
extern T0* T72f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.named_types */
extern T0* T240f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.base_type_with_type_mark */
extern T0* T72f7(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.named_types */
extern T0* T577f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.named_types */
extern T0* T1956f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.next_type */
extern T0* T451f12(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.next_type */
extern T0* T450f12(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.next_type */
extern T0* T302f19(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.next_type */
extern T0* T319f19(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.next_type */
extern T0* T223f17(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.is_type_expanded */
extern T1 T1501f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.is_type_expanded */
extern T1 T891f43(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.is_type_expanded */
extern T1 T1526f20(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.is_type_expanded */
extern T1 T1525f42(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.is_type_expanded */
extern T1 T1524f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.is_type_expanded */
extern T1 T242f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_type_expanded */
extern T1 T74f133(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.is_type_expanded */
extern T1 T449f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.is_type_expanded */
extern T1 T254f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.is_type_expanded */
extern T1 T72f77(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.is_tuple_class */
extern T1 T74f98(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.is_generic */
extern T1 T74f85(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.base_class */
extern T0* T1501f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER.base_class */
extern T0* T891f42(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_TYPE.base_class */
extern T0* T1526f19(GE_context* ac, T0* C, T0* a1);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.base_class */
extern T0* T1525f37(GE_context* ac, T0* C, T0* a1);
/* detachable ET_LIKE_FEATURE.base_class */
extern T0* T1524f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.base_class */
extern T0* T242f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS.base_class_in_context */
extern T0* T74f132(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIKE_CURRENT.base_class */
extern T0* T449f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_TUPLE_TYPE.base_class_in_context */
extern T0* T254f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.base_class_in_context */
extern T0* T72f76(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.tokens */
extern T0* T63f74(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.base_type */
extern T0* T323f11(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_COMMA_ARGUMENT.set_index */
extern void T1767f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FORMAL_ARGUMENT.set_index */
extern void T256f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_IDENTIFIER.set_index */
extern void T184f116(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_FORMAL_COMMA_ARGUMENT.name */
extern T0* T1767f8(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_ARGUMENT.name */
extern T0* T256f8(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.put_last */
extern void T259f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FORMAL_COMMA_ARGUMENT.type */
extern T0* T1767f7(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_ARGUMENT.type */
extern T0* T256f7(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.argument_type_set */
extern T0* T306f38(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DYNAMIC_FEATURE.argument_type_set */
extern T0* T236f28(GE_context* ac, T0* C, T6 a1);
/* detachable ET_FORMAL_ARGUMENT_LIST.formal_argument */
extern T0* T255f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.make_with_capacity */
extern T0* T259c9(GE_context* ac, T6 a1);
/* detachable ET_EXTERNAL_PROCEDURE.arguments */
extern T0* T1516f12(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.arguments */
extern T0* T253f12(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.arguments */
extern T0* T1515f11(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.arguments */
extern T0* T1513f10(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.arguments */
extern T0* T1512f10(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.arguments */
extern T0* T1511f10(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.arguments */
extern T0* T371f65(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.arguments */
extern T0* T369f64(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.arguments */
extern T0* T359f58(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.arguments */
extern T0* T360f60(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.arguments */
extern T0* T476f6(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.arguments */
extern T0* T469f6(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_AGENT_OPERAND_PUSH_TYPE_SET.set_never_void */
extern void T1466f16(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.set_never_void */
extern void T269f16(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.set_never_void */
extern void T475f16(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.set_never_void */
extern void T451f72(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.set_never_void */
extern void T450f72(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.set_never_void */
extern void T302f69(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.set_never_void */
extern void T319f71(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.set_never_void */
extern void T323f27(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_never_void */
extern void T223f68(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.new_dynamic_type_set */
extern T0* T230f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.new_dynamic_type_set */
extern T0* T71f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.make */
extern T0* T475c15(GE_context* ac, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET.put_type */
extern void T475f20(GE_context* ac, T0* C, T0* a1);
extern T1 T475f20ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.alive_conforming_descendants */
extern T0* T230f11(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].force_last_new */
extern void T382f44(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].key_storage_put */
extern void T382f54(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].item_storage_put */
extern void T382f53(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].slots_put */
extern void T382f52(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].clashes_put */
extern void T382f51(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].slots_item */
extern T6 T382f37(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].hash_position */
extern T6 T382f35(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].resize */
extern void T382f50(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].clashes_resize */
extern void T382f61(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].special_integer_ */
extern T0* T382f36(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].key_storage_resize */
extern void T382f60(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].item_storage_resize */
extern void T382f59(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].key_storage_item */
extern T0* T382f28(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].slots_resize */
extern void T382f58(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].new_modulus */
extern T6 T382f29(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].new_capacity */
extern T6 T382f34(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].unset_found_item */
extern void T382f49(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].found_item */
extern T0* T382f31(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].found */
extern T1 T382f27(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].search */
extern void T382f43(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].search_position */
extern void T382f48(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].make_map_default */
extern T0* T382c40(GE_context* ac);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].make_map */
extern void T382f45(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].make_with_equality_testers */
extern void T382f55(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].make_sparse_container */
extern void T382f62(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].set_internal_cursor */
extern void T382f67(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].new_cursor */
extern T0* T382f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].make */
extern T0* T921c8(GE_context* ac, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].make_slots */
extern void T382f66(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].make_clashes */
extern void T382f65(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].make_key_storage */
extern void T382f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE].make */
extern T0* T709f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE].make_empty */
extern T0* T708c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TYPE].default_create */
extern T0* T709c3(GE_context* ac);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].make_item_storage */
extern void T382f63(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_DYNAMIC_TYPE_SET, [attached] ET_DYNAMIC_TYPE].default_capacity */
extern T6 T382f30(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.alive_conforming_descendants */
extern T0* T71f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.object_id_dynamic_type_set */
extern T0* T230f10(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.mark_type_alive */
extern void T230f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.current_system */
extern T0* T230f8(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.set_alive */
extern void T451f67(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.set_alive */
extern void T450f67(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.set_alive */
extern void T302f64(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.set_alive */
extern void T319f66(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_alive */
extern void T223f63(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.mark_type_alive */
extern void T71f104(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.primary_type */
extern T0* T451f36(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.primary_type */
extern T0* T450f36(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.primary_type */
extern T0* T302f39(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.primary_type */
extern T0* T319f38(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.primary_type */
extern T0* T223f30(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.primary_type */
extern T0* T323f1(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.dynamic_type */
extern T0* T63f60(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.set_attached_type */
extern void T451f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.set_attached_type */
extern void T450f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.set_attached_type */
extern void T302f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.set_attached_type */
extern void T319f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_attached_type */
extern void T223f64(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.make */
extern T0* T323c25(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.base_type */
extern T0* T451f3(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.base_type */
extern T0* T450f3(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.base_type */
extern T0* T319f3(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.base_type */
extern T0* T302f3(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.base_type */
extern T0* T223f5(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_unique_attribute */
extern T1 T1516f43(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_unique_attribute */
extern T1 T253f59(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_unique_attribute */
extern T1 T1515f43(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_unique_attribute */
extern T1 T1513f43(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_unique_attribute */
extern T1 T1512f45(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_unique_attribute */
extern T1 T1511f43(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_unique_attribute */
extern T1 T371f64(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_unique_attribute */
extern T1 T369f63(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_unique_attribute */
extern T1 T359f57(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_unique_attribute */
extern T1 T360f59(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_unique_attribute */
extern T1 T476f30(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_unique_attribute */
extern T1 T469f30(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.is_constant_attribute */
extern T1 T1516f42(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.is_constant_attribute */
extern T1 T253f58(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.is_constant_attribute */
extern T1 T1515f42(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.is_constant_attribute */
extern T1 T1513f42(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.is_constant_attribute */
extern T1 T1512f44(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.is_constant_attribute */
extern T1 T1511f42(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.is_constant_attribute */
extern T1 T371f63(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.is_constant_attribute */
extern T1 T369f62(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.is_constant_attribute */
extern T1 T359f56(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.is_constant_attribute */
extern T1 T360f58(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.is_constant_attribute */
extern T1 T476f29(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.is_constant_attribute */
extern T1 T469f50(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.type */
extern T0* T1516f38(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.type */
extern T0* T253f54(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.type */
extern T0* T1515f40(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.type */
extern T0* T1513f40(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.type */
extern T0* T1512f42(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.type */
extern T0* T1511f40(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.type */
extern T0* T371f57(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.type */
extern T0* T369f56(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.type */
extern T0* T359f54(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.type */
extern T0* T360f56(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.type */
extern T0* T476f48(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.type */
extern T0* T469f48(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_FEATURE.tokens */
extern T0* T236f27(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].make_map */
extern T0* T617c34(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].make_with_equality_testers */
extern void T617f35(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].make_sparse_container */
extern void T617f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].set_internal_cursor */
extern void T617f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].new_cursor */
extern T0* T617f26(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].make */
extern T0* T1410c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].make_slots */
extern void T617f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].make_clashes */
extern void T617f39(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].make_key_storage */
extern void T617f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CALL_NAME].make */
extern T0* T1407f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CALL_NAME].make_empty */
extern T0* T1404c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CALL_NAME].default_create */
extern T0* T1407c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL, [attached] ET_CALL_NAME].make_item_storage */
extern void T617f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL].make */
extern T0* T1411f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL].make_empty */
extern T0* T1412c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL].default_create */
extern T0* T1411c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].make_map */
extern T0* T616c34(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].make_with_equality_testers */
extern void T616f36(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].make_sparse_container */
extern void T616f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].set_internal_cursor */
extern void T616f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].new_cursor */
extern T0* T616f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].make */
extern T0* T1408c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].make_slots */
extern void T616f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].make_clashes */
extern void T616f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].make_key_storage */
extern void T616f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL, [attached] ET_CALL_NAME].make_item_storage */
extern void T616f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL].make */
extern T0* T1406f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL].make_empty */
extern T0* T1402c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_QUALIFIED_QUERY_CALL].default_create */
extern T0* T1406c3(GE_context* ac);
/* detachable ET_DYNAMIC_FEATURE.empty_dynamic_type_sets */
extern T0* T236f25(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.make */
extern T0* T469c80(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DO_PROCEDURE.tokens */
extern T0* T469f44(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.base_class */
extern T0* T451f2(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.base_class */
extern T0* T450f2(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.base_class */
extern T0* T319f2(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.base_class */
extern T0* T302f2(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.base_class */
extern T0* T223f7(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.make */
extern T0* T184c113(GE_context* ac, T0* a1);
/* [detachable] ET_C_GENERATOR.close_cpp_file */
extern void T67f601(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.close_c_file */
extern void T67f600(GE_context* ac, T0* C);
/* [detachable] KL_UNIX_FILE_SYSTEM.delete_file */
extern void T126f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.delete */
extern void T37f76(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.old_delete */
extern void T37f79(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.file_unlink */
extern void T37f81(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.delete_file */
extern void T125f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.include_file */
extern void T67f599(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM.append */
extern void T161f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_NULL_TEXT_OUTPUT_STREAM.append */
extern void T161f9p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.read_string */
extern void T37f77(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_STRING_OUTPUT_STREAM.append */
extern void T260f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STRING_OUTPUT_STREAM.append */
extern void T260f10p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.append */
extern void T300f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_OUTPUT_FILE.append */
extern void T300f38p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.rename_file */
extern void T126f32(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_TEXT_INPUT_FILE.change_name */
extern void T37f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.old_change_name */
extern void T37f78(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.rename_file */
extern void T37f80(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.set_name */
extern void T37f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.file_rename */
extern void T37f82(GE_context* ac, T0* C, T14 a1, T14 a2);
/* [detachable] KL_TEXT_INPUT_FILE.same_physical_file */
extern T1 T37f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.file_system */
extern T0* T37f44(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.unix_file_system */
extern T0* T37f54(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.windows_file_system */
extern T0* T37f53(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.operating_system */
extern T0* T37f52(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.count */
extern T6 T37f42(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.old_count */
extern T6 T37f50(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.file_size */
extern T6 T37f56(GE_context* ac, T0* C, T14 a1);
/* [detachable] FILE_INFO.size */
extern T6 T141f18(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.old_is_open_write */
extern T1 T37f55(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.inode */
extern T6 T37f41(GE_context* ac, T0* C);
/* [detachable] FILE_INFO.inode */
extern T6 T141f17(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.tmp_file1 */
extern T0* T37f40(GE_context* ac, T0* C);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.rename_file */
extern void T125f35(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].wipe_out */
extern void T294f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].slots_wipe_out */
extern void T294f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_wipe_out */
extern void T294f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_wipe_out */
extern void T294f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_wipe_out */
extern void T294f40(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_FEATURE].keep_head */
extern void T687f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_FEATURE].set_count */
extern void T687f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].unset_found_item */
extern void T294f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].move_all_cursors_after */
extern void T294f38(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].set_next_cursor */
extern void T689f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE_CURSOR [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].set_after */
extern void T689f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].wipe_out */
extern void T293f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].slots_wipe_out */
extern void T293f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].clashes_wipe_out */
extern void T293f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].key_storage_wipe_out */
extern void T293f47(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_INLINE_CONSTANT].keep_head */
extern void T684f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_INLINE_CONSTANT].set_count */
extern void T684f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].item_storage_wipe_out */
extern void T293f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].move_all_cursors_after */
extern void T293f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].set_next_cursor */
extern void T683f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].set_after */
extern void T683f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].wipe_out */
extern void T292f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].slots_wipe_out */
extern void T292f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].clashes_wipe_out */
extern void T292f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].key_storage_wipe_out */
extern void T292f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].item_storage_wipe_out */
extern void T292f47(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_CONSTANT].keep_head */
extern void T680f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CONSTANT].set_count */
extern void T680f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].move_all_cursors_after */
extern void T292f45(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CONSTANT, [attached] ET_FEATURE].set_next_cursor */
extern void T679f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CONSTANT, [attached] ET_FEATURE].set_after */
extern void T679f9(GE_context* ac, T0* C);
/* [detachable] ARRAY [[attached] INTEGER_32].fill_with */
extern void T291f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].wipe_out */
extern void T290f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].slots_wipe_out */
extern void T290f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].clashes_wipe_out */
extern void T290f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].key_storage_wipe_out */
extern void T290f45(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].item_storage_wipe_out */
extern void T290f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].move_all_cursors_after */
extern void T290f42(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_FEATURE].set_next_cursor */
extern void T676f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_FEATURE].set_after */
extern void T676f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].wipe_out */
extern void T296f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].slots_wipe_out */
extern void T296f58(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].clashes_wipe_out */
extern void T296f57(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].key_storage_wipe_out */
extern void T296f56(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].item_storage_wipe_out */
extern void T296f55(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].unset_found_item */
extern void T296f54(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].move_all_cursors_after */
extern void T296f53(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] STRING_8].set_next_cursor */
extern void T692f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] STRING_8].set_after */
extern void T692f9(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].wipe_out */
extern void T70f45(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].slots_wipe_out */
extern void T70f64(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].clashes_wipe_out */
extern void T70f63(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].key_storage_wipe_out */
extern void T70f62(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].item_storage_wipe_out */
extern void T70f61(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].move_all_cursors_after */
extern void T70f60(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].is_empty */
extern T1 T70f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].forth */
extern void T296f45(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].cursor_forth */
extern void T296f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].add_traversing_cursor */
extern void T296f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] STRING_8].set_position */
extern void T692f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].remove_traversing_cursor */
extern void T296f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].clashes_item */
extern T6 T296f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].key_for_iteration */
extern T0* T296f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].cursor_key */
extern T0* T296f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].key_storage_item */
extern T0* T296f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].item_for_iteration */
extern T1 T296f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].cursor_item */
extern T1 T296f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].item_storage_item */
extern T1 T296f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].after */
extern T1 T296f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].cursor_after */
extern T1 T296f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] STRING_8].after */
extern T1 T692f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].start */
extern void T296f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].cursor_start */
extern void T296f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].cursor_off */
extern T1 T296f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] STRING_8].off */
extern T1 T692f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].is_empty */
extern T1 T296f31(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].forth */
extern void T70f46(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].cursor_forth */
extern void T70f65(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].add_traversing_cursor */
extern void T70f71(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].remove_traversing_cursor */
extern void T70f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.include_runtime_file */
extern void T67f598(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable DS_HASH_SET [[attached] STRING_8].item_for_iteration */
extern T0* T70f17(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].cursor_item */
extern T0* T70f18(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].after */
extern T1 T70f23(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].cursor_after */
extern T1 T70f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8].after */
extern T1 T376f8(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].start */
extern void T70f42(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].cursor_start */
extern void T70f58(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].cursor_off */
extern T1 T70f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8].off */
extern T1 T376f5(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_end_extern_c */
extern void T67f597(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_main_function */
extern void T67f596(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.temp_variable */
extern T0* T67f158(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_types_array */
extern void T67f595(GE_context* ac, T0* C);
extern T1 T67f595ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_SYSTEM.is_new_instance_type */
extern T1 T63f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.replace_substring_all */
extern void T17f74(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] SPECIAL [CHARACTER_8].resized_area_with_default */
extern T0* T15f8(GE_context* ac, T0* C, T2 a1, T6 a2);
/* detachable ARRAYED_LIST [[attached] INTEGER_32].area */
extern T0* T782f3(GE_context* ac, T0* C);
/* [detachable] STRING_8_SEARCHER.substring_index_list_with_deltas */
extern T0* T781f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
/* detachable ARRAYED_LIST [[attached] INTEGER_32].extend */
extern void T782f9(GE_context* ac, T0* C, T6 a1);
/* detachable ARRAYED_LIST [[attached] INTEGER_32].additional_space */
extern T6 T782f5(GE_context* ac, T0* C);
/* detachable ARRAYED_LIST [[attached] INTEGER_32].capacity */
extern T6 T782f6(GE_context* ac, T0* C);
/* detachable ARRAYED_LIST [[attached] INTEGER_32].count */
extern T6 T782f4(GE_context* ac, T0* C);
/* detachable ARRAYED_LIST [[attached] INTEGER_32].make */
extern T0* T782c8(GE_context* ac, T6 a1);
/* [detachable] STRING_8_SEARCHER.initialize_deltas */
extern void T781f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.is_tuple */
extern T1 T451f44(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.is_tuple */
extern T1 T450f44(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.is_tuple */
extern T1 T302f55(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.is_tuple */
extern T1 T319f56(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_tuple */
extern T1 T223f53(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.type_id */
extern T6 T451f39(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.type_id */
extern T6 T450f39(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.type_id */
extern T6 T302f45(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.type_id */
extern T6 T319f46(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.type_id */
extern T6 T223f42(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_SECONDARY_TYPE.type_id */
extern T6 T323f5(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].wipe_out */
extern void T76f12(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].move_all_cursors_after */
extern void T76f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].set_next_cursor */
extern void T417f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].set_after */
extern void T417f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].sort */
extern void T76f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_QUICK_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].sort */
extern void T322f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_QUICK_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].sort_with_comparator */
extern void T322f4(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_QUICK_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].subsort_with_comparator */
extern void T322f5(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].swap */
extern void T76f22(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].replace */
extern void T76f21(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_COMPARATOR_BY_ID.less_than */
extern T1 T324f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].is_empty */
extern T1 T76f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].append_last */
extern void T76f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].extend_last */
extern void T76f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].forth */
extern void T334f14(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].cursor_forth */
extern void T275f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].add_traversing_cursor */
extern void T275f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].set_next_cursor */
extern void T334f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].set_position */
extern void T334f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].remove_traversing_cursor */
extern void T275f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].set_after */
extern void T334f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].item */
extern T0* T334f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].cursor_item */
extern T0* T275f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].after */
extern T1 T334f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].start */
extern void T334f13(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].cursor_start */
extern void T275f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].cursor_off */
extern T1 T275f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].off */
extern T1 T334f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].is_empty */
extern T1 T275f21(GE_context* ac, T0* C);
/* [detachable] DS_QUICK_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T322c2(GE_context* ac, T0* a1);
/* [detachable] ET_C_GENERATOR.dynamic_primary_type_comparator */
extern T0* T67f166(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE_COMPARATOR_BY_ID.make */
extern T0* T324c2(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T76c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].set_internal_cursor */
extern void T76f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].new_cursor */
extern T0* T76f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T417c5(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].found_item */
extern T0* T294f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_item */
extern T0* T294f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].found */
extern T1 T294f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].search */
extern void T294f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].search_position */
extern void T294f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].hash_position */
extern T6 T294f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_item */
extern T0* T294f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_item */
extern T6 T294f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].slots_item */
extern T6 T294f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_dynamic_type_id_set_constants */
extern void T67f594(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].forth */
extern void T216f54(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].key_for_iteration */
extern T0* T216f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].item_for_iteration */
extern T0* T216f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].after */
extern T1 T216f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].cursor_after */
extern T1 T216f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].start */
extern void T216f53(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_init_const_function */
extern void T67f593(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].forth */
extern void T293f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].cursor_forth */
extern void T293f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].add_traversing_cursor */
extern void T293f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].set_position */
extern void T683f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].remove_traversing_cursor */
extern void T293f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].item_for_iteration */
extern T0* T293f20(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].cursor_item */
extern T0* T293f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].item_storage_item */
extern T0* T293f36(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ONCE_MANIFEST_STRING.constant */
extern T0* T1550f12(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].key_for_iteration */
extern T0* T293f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].cursor_key */
extern T0* T293f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].after */
extern T1 T293f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].cursor_after */
extern T1 T293f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].after */
extern T1 T683f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].start */
extern void T293f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].cursor_start */
extern void T293f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].cursor_off */
extern T1 T293f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].off */
extern T1 T683f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].is_empty */
extern T1 T293f23(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].forth */
extern void T292f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].cursor_forth */
extern void T292f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].add_traversing_cursor */
extern void T292f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CONSTANT, [attached] ET_FEATURE].set_position */
extern void T679f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].remove_traversing_cursor */
extern void T292f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].item_for_iteration */
extern T0* T292f23(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].cursor_item */
extern T0* T292f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].item_storage_item */
extern T0* T292f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_once_status_name */
extern void T67f665(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].has */
extern T1 T290f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].key_for_iteration */
extern T0* T292f22(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].cursor_key */
extern T0* T292f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].after */
extern T1 T292f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].cursor_after */
extern T1 T292f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CONSTANT, [attached] ET_FEATURE].after */
extern T1 T679f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].start */
extern void T292f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].cursor_start */
extern void T292f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].cursor_off */
extern T1 T292f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CONSTANT, [attached] ET_FEATURE].off */
extern T1 T679f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].is_empty */
extern T1 T292f28(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_constants_declaration */
extern void T67f592(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_deep_twin_functions */
extern void T67f591(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].wipe_out */
extern void T276f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].slots_wipe_out */
extern void T276f53(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_wipe_out */
extern void T276f52(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_wipe_out */
extern void T276f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_wipe_out */
extern void T276f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].unset_found_item */
extern void T276f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].move_all_cursors_after */
extern void T276f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].set_next_cursor */
extern void T655f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].set_after */
extern void T655f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].forth */
extern void T276f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].cursor_forth */
extern void T276f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].add_traversing_cursor */
extern void T276f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].set_position */
extern void T655f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].remove_traversing_cursor */
extern void T276f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_item */
extern T6 T276f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_deep_twin_polymorphic_call_function */
extern void T67f662(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].wipe_out */
extern void T273f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].slots_wipe_out */
extern void T273f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].clashes_wipe_out */
extern void T273f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].key_storage_wipe_out */
extern void T273f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].item_storage_wipe_out */
extern void T273f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].unset_found_item */
extern void T273f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].move_all_cursors_after */
extern void T273f45(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].set_next_cursor */
extern void T654f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].set_after */
extern void T654f7(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_deep_twin_binary_search_polymorphic_call */
extern void T67f732(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T0* a4, T0* a5);
/* [detachable] ET_C_GENERATOR.print_adapted_deep_twin_call */
extern void T67f731(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Creation of agent #1 in feature [detachable] ET_C_GENERATOR.print_adapted_deep_twin_call */
extern T0* T67f731ac1(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.print_deep_twin_call */
extern void T67f749(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].item */
extern T0* T273f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].item_storage_item */
extern T0* T273f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].search_position */
extern void T273f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].hash_position */
extern T6 T273f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].key_storage_item */
extern T6 T273f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].clashes_item */
extern T6 T273f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].slots_item */
extern T6 T273f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].force_last */
extern void T273f36(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].key_storage_put */
extern void T273f44(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].slots_put */
extern void T273f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].clashes_put */
extern void T273f41(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].resize */
extern void T273f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].clashes_resize */
extern void T273f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].special_integer_ */
extern T0* T273f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].key_storage_resize */
extern void T273f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].item_storage_resize */
extern void T273f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].slots_resize */
extern void T273f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].new_modulus */
extern T6 T273f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].new_capacity */
extern T6 T273f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].item_storage_put */
extern void T273f43(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_for_iteration */
extern T0* T276f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].cursor_item */
extern T0* T276f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_item */
extern T0* T276f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].after */
extern T1 T276f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].cursor_after */
extern T1 T276f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].after */
extern T1 T655f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].start */
extern void T276f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].cursor_start */
extern void T276f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].cursor_off */
extern T1 T276f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].off */
extern T1 T655f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].is_empty */
extern T1 T276f22(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern void T67f661(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T67f661ac1(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #2 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T67f661ac2(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #3 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T67f661ac3(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #4 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T67f661ac4(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #5 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T67f661ac5(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #6 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T67f661ac6(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #7 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T67f661ac7(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #8 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T67f661ac8(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #9 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T67f661ac9(GE_context* ac, T0* a1, T6 a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #10 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T67f661ac10(GE_context* ac, T0* a1, T6 a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #11 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T67f661ac11(GE_context* ac, T0* a1, T6 a2, T0* a3, T0* a4, T1 a5);
/* Creation of agent #12 in feature [detachable] ET_C_GENERATOR.print_deep_twin_function */
extern T0* T67f661ac12(GE_context* ac, T0* a1, T6 a2, T0* a3, T0* a4, T1 a5);
extern T1 T67f661ot1(T0* a1, T0** a2);
extern T1 T67f661ot2(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_attribute_special_indexed_item_access */
extern void T67f729(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4);
/* [detachable] ET_C_GENERATOR.print_set_deep_twined_attribute */
extern void T67f730(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_C_GENERATOR.print_deep_twined_attribute */
extern void T67f748(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].force_last_new */
extern void T276f44(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_put */
extern void T276f59(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_put */
extern void T276f58(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].slots_put */
extern void T276f57(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_put */
extern void T276f56(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].slots_item */
extern T6 T276f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].hash_position */
extern T6 T276f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].resize */
extern void T276f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_resize */
extern void T276f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].special_integer_ */
extern T0* T276f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_resize */
extern void T276f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_resize */
extern void T276f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_item */
extern T0* T276f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].slots_resize */
extern void T276f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].new_modulus */
extern T6 T276f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].new_capacity */
extern T6 T276f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST.force_last */
extern void T274f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST.fixed_array */
extern T0* T274f5(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].default_create */
extern T0* T648c3(GE_context* ac);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST.resize */
extern void T274f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST.new_capacity */
extern T6 T274f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST.capacity */
extern T6 T274f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].found_item */
extern T0* T276f27(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].found */
extern T1 T276f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].search */
extern void T276f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].search_position */
extern void T276f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.has_nested_reference_attributes */
extern T1 T451f32(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.has_nested_reference_attributes */
extern T1 T450f32(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.has_nested_reference_attributes */
extern T1 T302f51(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.has_nested_reference_attributes */
extern T1 T319f57(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_nested_reference_attributes */
extern T1 T223f54(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].wipe_out */
extern void T289f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].slots_wipe_out */
extern void T289f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].clashes_wipe_out */
extern void T289f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].key_storage_wipe_out */
extern void T289f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].item_storage_wipe_out */
extern void T289f48(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TUPLE_TYPE].keep_head */
extern void T673f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TUPLE_TYPE].set_count */
extern void T673f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].move_all_cursors_after */
extern void T289f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_TUPLE_TYPE].set_next_cursor */
extern void T672f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_TUPLE_TYPE].set_after */
extern void T672f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].forth */
extern void T289f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].cursor_forth */
extern void T289f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].add_traversing_cursor */
extern void T289f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_TUPLE_TYPE].set_position */
extern void T672f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].remove_traversing_cursor */
extern void T289f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_manifest_tuple_function */
extern void T67f590(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].item_for_iteration */
extern T0* T289f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].cursor_item */
extern T0* T289f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].after */
extern T1 T289f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].cursor_after */
extern T1 T289f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_TUPLE_TYPE].after */
extern T1 T672f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].start */
extern void T289f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].cursor_start */
extern void T289f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].cursor_off */
extern T1 T289f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_TUPLE_TYPE].off */
extern T1 T672f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].is_empty */
extern T1 T289f21(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_big_manifest_array_function */
extern void T67f589(GE_context* ac, T0* C, T0* a1);
extern T1 T67f589ot1(T0* a1, T0** a2);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].wipe_out */
extern void T275f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].slots_wipe_out */
extern void T275f55(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_wipe_out */
extern void T275f54(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_wipe_out */
extern void T275f53(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_wipe_out */
extern void T275f52(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].move_all_cursors_after */
extern void T275f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].forth */
extern void T275f39(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_manifest_array_function */
extern void T67f588(GE_context* ac, T0* C, T0* a1);
extern T1 T67f588ot1(T0* a1, T0** a2);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].item_for_iteration */
extern T0* T275f17(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].after */
extern T1 T275f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].cursor_after */
extern T1 T275f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].start */
extern void T275f38(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_polymorphic_procedure_call_functions */
extern void T67f587(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_polymorphic_call_function */
extern void T67f658(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T67f658ot1(T0* a1, T0** a2);
extern T1 T67f658ot2(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_binary_search_polymorphic_calls */
extern void T67f728(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T6 a5, T0* a6, T0* a7);
/* [detachable] ET_C_GENERATOR.set_polymorphic_call_argument_type_sets */
extern void T67f724(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T67f724ot1(T0* a1, T0** a2);
extern T1 T67f724ot2(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET.set_static_type */
extern void T269f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST.put_last */
extern void T274f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.is_call_agent */
extern T1 T318f7(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.is_call_agent */
extern T1 T314f10(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.force_result_boxing */
extern T1 T318f9(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_call_name_comment */
extern void T67f722(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T67f722ot1(T0* a1, T0** a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].force_last_new */
extern void T273f34(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].has */
extern T1 T273f25(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.static_feature */
extern T0* T318f8(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.static_feature */
extern T0* T314f11(GE_context* ac, T0* C);
/* detachable ET_ASSIGNER_INSTRUCTION.is_empty */
extern T1 T1527f14(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.same_declared_signature */
extern T1 T67f156(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T67f156ot1(T0* a1);
extern T1 T67f156ot2(T0* a1);
extern T1 T67f156ot3(T0* a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.procedure_calls */
extern T0* T451f17(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.procedure_calls */
extern T0* T450f17(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.procedure_calls */
extern T0* T302f22(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.procedure_calls */
extern T0* T319f22(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.procedure_calls */
extern T0* T223f21(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_polymorphic_tuple_label_call_functions */
extern void T67f586(GE_context* ac, T0* C);
extern T1 T67f586ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_polymorphic_query_call_functions */
extern void T67f585(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.query_calls */
extern T0* T451f16(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.query_calls */
extern T0* T450f16(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.query_calls */
extern T0* T302f20(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.query_calls */
extern T0* T319f20(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.query_calls */
extern T0* T223f19(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_feature */
extern void T67f584(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION].wipe_out */
extern void T279f11(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_EQUALITY_EXPRESSION].keep_head */
extern void T663f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_EQUALITY_EXPRESSION].set_count */
extern void T663f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION].move_all_cursors_after */
extern void T279f14(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EQUALITY_EXPRESSION].set_next_cursor */
extern void T665f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EQUALITY_EXPRESSION].set_after */
extern void T665f6(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_equality_function */
extern void T67f657(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION].item */
extern T0* T279f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION].wipe_out */
extern void T278f11(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_OBJECT_EQUALITY_EXPRESSION].keep_head */
extern void T660f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_OBJECT_EQUALITY_EXPRESSION].set_count */
extern void T660f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION].move_all_cursors_after */
extern void T278f14(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_EQUALITY_EXPRESSION].set_next_cursor */
extern void T662f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_EQUALITY_EXPRESSION].set_after */
extern void T662f6(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_object_equality_function */
extern void T67f656(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION].item */
extern T0* T278f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST].wipe_out */
extern void T277f11(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST].keep_head */
extern void T657f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST].set_count */
extern void T657f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST].move_all_cursors_after */
extern void T277f13(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_TEST].set_next_cursor */
extern void T659f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_TEST].set_after */
extern void T659f6(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_object_test_function */
extern void T67f655(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST].item */
extern T0* T277f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT].wipe_out */
extern void T280f11(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_AGENT].keep_head */
extern void T666f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_AGENT].set_count */
extern void T666f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT].move_all_cursors_after */
extern void T280f13(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_AGENT].set_next_cursor */
extern void T668f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_AGENT].set_after */
extern void T668f6(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_agent_declaration */
extern void T67f654(GE_context* ac, T0* C, T6 a1, T0* a2);
extern T1 T67f654ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_agent_body_declaration */
extern void T67f707(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_agent_function_name */
extern void T67f703(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_IDENTIFIER.set_agent_closed_operand */
extern void T184f117(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_IDENTIFIER.set_agent_open_operand */
extern void T184f114(GE_context* ac, T0* C, T1 a1);
/* detachable ET_CALL_AGENT.implicit_result */
extern T0* T1041f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT].item */
extern T0* T280f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DYNAMIC_PRECURSOR.is_semistrict */
extern T1 T306f25(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_FEATURE.is_semistrict */
extern T1 T236f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE].remove_last */
extern void T295f12(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_FEATURE].keep_head */
extern void T605f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_FEATURE].set_count */
extern void T605f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE].move_last_cursors_after */
extern void T295f15(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_FEATURE].set_next_cursor */
extern void T691f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_FEATURE].set_after */
extern void T691f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE].last */
extern T0* T295f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE].is_empty */
extern T1 T295f10(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.register_called_feature */
extern void T67f583(GE_context* ac, T0* C, T0* a1);
/* detachable ET_DYNAMIC_PRECURSOR.set_generated */
extern void T306f52(GE_context* ac, T0* C, T1 a1);
/* detachable ET_DYNAMIC_FEATURE.set_generated */
extern void T236f47(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_C_GENERATOR.print_once_per_object_data_dispose_functions */
extern void T67f582(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_once_per_object_data_dispose_function */
extern void T67f653(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_once_per_object_data_allocation_functions */
extern void T67f581(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_once_per_object_data_allocation_function */
extern void T67f652(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_object_allocation_functions */
extern void T67f580(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_object_allocation_function */
extern void T67f651(GE_context* ac, T0* C, T0* a1);
extern T1 T67f651ot1(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_dispose_registration */
extern void T67f702(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T67f702ot1(T0* a1, T0** a2);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].force_last_new */
extern void T294f36(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_put */
extern void T294f49(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_put */
extern void T294f48(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_FEATURE].force */
extern void T690f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_FEATURE].fill_with */
extern void T687f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_FEATURE].extend */
extern void T687f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].slots_put */
extern void T294f47(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_put */
extern void T294f46(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].resize */
extern void T294f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_resize */
extern void T294f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].special_integer_ */
extern T0* T294f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_resize */
extern void T294f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_resize */
extern void T294f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_FEATURE].aliased_resized_area */
extern T0* T690f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_FEATURE].aliased_resized_area */
extern T0* T687f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].slots_resize */
extern void T294f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].new_modulus */
extern T6 T294f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].new_capacity */
extern T6 T294f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_new_immutable_string_32_function */
extern void T67f579(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_new_immutable_string_8_function */
extern void T67f578(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_new_string_32_function */
extern void T67f577(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_new_string_8_function */
extern void T67f576(GE_context* ac, T0* C);
/* [detachable] KL_STRING_OUTPUT_STREAM.put_new_line */
extern void T260f5(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_default_declarations */
extern void T67f575(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_default_declaration */
extern void T67f637(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_default_object_value */
extern void T67f697(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T67f697ot1(T0* a1, T0** a2);
extern T1 T67f697ot2(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_default_attribute_value */
extern void T67f738(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.flush_to_c_file */
extern void T67f574(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_types */
extern void T67f573(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_once_per_object_data_struct */
extern void T67f636(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_once_per_object_data_struct_name */
extern void T67f686(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_boxed_type_struct */
extern void T67f635(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_boxed_struct_name */
extern void T67f685(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_type_struct */
extern void T67f634(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T67f634ot1(T0* a1, T0** a2);
extern T1 T67f634ot2(T0* a1, T0** a2);
/* [detachable] ET_C_GENERATOR.print_struct_name */
extern void T67f676(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].has_cycle */
extern T1 T308f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].sort */
extern void T308f15(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_PRIMARY_TYPE].put_last */
extern void T76f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].reset */
extern void T308f20(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.print_once_per_object_data_type_definition */
extern void T67f633(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_boxed_type_definition */
extern void T67f632(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_type_type_definition */
extern void T67f631(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_boxed_pointer_function */
extern void T67f630(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.print_boxed_attribute_flags_access */
extern void T67f682(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] ET_C_GENERATOR.print_boxed_function */
extern void T67f629(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].force_relation */
extern void T308f14(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].put_indexed_relation */
extern void T308f19(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].count */
extern T6 T308f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].index_of */
extern T6 T308f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].found_item */
extern T6 T698f20(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_item */
extern T6 T698f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].found */
extern T1 T698f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].search */
extern void T698f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].search_position */
extern void T698f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].hash_position */
extern T6 T698f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_item */
extern T0* T698f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_item */
extern T6 T698f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].slots_item */
extern T6 T698f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_type_definition */
extern void T67f628(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_pointer_type_definition */
extern void T67f627(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_real_64_type_definition */
extern void T67f626(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_real_32_type_definition */
extern void T67f625(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_natural_64_type_definition */
extern void T67f624(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_natural_32_type_definition */
extern void T67f623(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_natural_16_type_definition */
extern void T67f622(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_natural_8_type_definition */
extern void T67f621(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_integer_64_type_definition */
extern void T67f620(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_integer_32_type_definition */
extern void T67f619(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_integer_16_type_definition */
extern void T67f618(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_integer_8_type_definition */
extern void T67f617(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_character_32_type_definition */
extern void T67f616(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_character_8_type_definition */
extern void T67f615(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_boolean_type_definition */
extern void T67f614(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].force */
extern void T308f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].put */
extern void T308f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].put_new */
extern void T698f36(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_put */
extern void T698f48(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_put */
extern void T698f47(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].slots_put */
extern void T698f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_put */
extern void T698f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].unset_found_item */
extern void T698f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].resize */
extern void T698f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].clashes_resize */
extern void T698f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].special_integer_ */
extern T0* T698f26(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].key_storage_resize */
extern void T698f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].item_storage_resize */
extern void T698f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].slots_resize */
extern void T698f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].new_modulus */
extern T6 T698f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].capacity */
extern T6 T308f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].has */
extern T1 T308f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].has */
extern T1 T698f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].make_default */
extern T0* T308c12(GE_context* ac);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].make */
extern void T308f16(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T308c16(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_map */
extern T0* T698c34(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_with_equality_testers */
extern void T698f39(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_sparse_container */
extern void T698f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].set_internal_cursor */
extern void T698f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].new_cursor */
extern T0* T698f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T1413c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_slots */
extern void T698f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_clashes */
extern void T698f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_key_storage */
extern void T698f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_item_storage */
extern void T698f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.print_aliased_double_type_definition */
extern void T67f613(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_aliased_real_type_definition */
extern void T67f612(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_aliased_natural_type_definition */
extern void T67f611(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_aliased_integer_type_definition */
extern void T67f610(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_aliased_wide_character_type_definition */
extern void T67f609(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_aliased_character_type_definition */
extern void T67f608(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.print_start_extern_c */
extern void T67f572(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.include_runtime_header_file */
extern void T67f571(GE_context* ac, T0* C, T0* a1, T1 a2, T0* a3);
/* [detachable] ET_C_GENERATOR.include_runtime_c_file */
extern void T67f606(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].force_last */
extern void T296f48(GE_context* ac, T0* C, T1 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].key_storage_put */
extern void T296f63(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].slots_put */
extern void T296f62(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].clashes_put */
extern void T296f61(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].slots_item */
extern T6 T296f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].hash_position */
extern T6 T296f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].resize */
extern void T296f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].clashes_resize */
extern void T296f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].special_integer_ */
extern T0* T296f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].key_storage_resize */
extern void T296f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].item_storage_resize */
extern void T296f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].slots_resize */
extern void T296f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].new_modulus */
extern T6 T296f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].new_capacity */
extern T6 T296f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].item_storage_put */
extern void T296f59(GE_context* ac, T0* C, T1 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].search_position */
extern void T296f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].replace */
extern void T296f47(GE_context* ac, T0* C, T1 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].item */
extern T1 T296f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].has */
extern T1 T296f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE.same_named_type_with_type_marks */
extern T1 T72f24(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_TOKEN_CONSTANTS.implicit_detachable_type_mark */
extern T0* T73f22(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.tokens */
extern T0* T67f291(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.character_32_type */
extern T0* T187f37(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.character_32_type */
extern T0* T58f15(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.character_32_type */
extern T0* T40f4(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.character_type */
extern T0* T187f34(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.character_type */
extern T0* T58f47(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.character_type */
extern T0* T40f3(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.universe */
extern T0* T74f96(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.exception_trace_mode */
extern T1 T67f202(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.trace_mode */
extern T1 T67f125(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.generate_ids */
extern void T67f570(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.set_id */
extern void T306f48(GE_context* ac, T0* C, T6 a1);
/* detachable ET_DYNAMIC_FEATURE.set_id */
extern void T236f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.has_static */
extern T1 T451f15(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.has_static */
extern T1 T450f15(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.has_static */
extern T1 T302f21(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.has_static */
extern T1 T319f21(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_static */
extern T1 T223f20(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_FUNCTION_TYPE.set_id */
extern void T451f73(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PROCEDURE_TYPE.set_id */
extern void T450f73(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_TUPLE_TYPE.set_id */
extern void T302f76(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_SPECIAL_TYPE.set_id */
extern void T319f78(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_id */
extern void T223f75(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.sort_types */
extern void T67f605(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TOPOLOGICAL_SORTER [[attached] ET_DYNAMIC_PRIMARY_TYPE].put_relation */
extern void T308f17(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_PRIMARY_TYPE].new_iterator */
extern T0* T275f34(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.open_cpp_file */
extern void T67f569(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.open_c_file */
extern void T67f568(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.set_split_threshold */
extern void T67f560(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_C_GENERATOR.set_split_mode */
extern void T67f559(GE_context* ac, T0* C, T1 a1);
/* [detachable] GEC.no_split */
extern T1 T25f48(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.parameter */
extern T1 T48f12(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] BOOLEAN].last */
extern T1 T157f7(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] BOOLEAN].is_empty */
extern T1 T157f6(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.set_finalize_mode */
extern void T67f558(GE_context* ac, T0* C, T1 a1);
/* [detachable] GEC.is_finalize */
extern T1 T25f49(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.set_use_boehm_gc */
extern void T67f557(GE_context* ac, T0* C, T1 a1);
/* [detachable] GEC.use_boehm_gc */
extern T1 T25f47(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.was_found */
extern T1 T47f18(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.make */
extern T0* T67c556(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_C_GENERATOR.make_external_regexps */
extern void T67f563(GE_context* ac, T0* C);
/* [detachable] ET_C_GENERATOR.make_rescue_data */
extern void T67f562(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].set_equality_tester */
extern void T297f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_C_GENERATOR.identifier_tester */
extern T0* T67f121(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER_TESTER.default_create */
extern T0* T298c2(GE_context* ac);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].make */
extern T0* T297c31(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].set_internal_cursor */
extern void T297f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].new_cursor */
extern T0* T297f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_IDENTIFIER].make */
extern T0* T695c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].make_slots */
extern void T297f39(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].make_clashes */
extern void T297f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].make_key_storage */
extern void T297f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_IDENTIFIER].make_item_storage */
extern void T297f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_IDENTIFIER].make */
extern T0* T599f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_IDENTIFIER].make_empty */
extern T0* T600c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_IDENTIFIER].default_create */
extern T0* T599c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].set_key_equality_tester */
extern void T296f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].make */
extern T0* T296c42(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].make_with_equality_testers */
extern void T296f50(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].make_sparse_container */
extern void T296f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].set_internal_cursor */
extern void T296f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].new_cursor */
extern T0* T296f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] BOOLEAN, [attached] STRING_8].make */
extern T0* T692c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].make_slots */
extern void T296f74(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].make_clashes */
extern void T296f73(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].make_key_storage */
extern void T296f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] BOOLEAN, [attached] STRING_8].make_item_storage */
extern void T296f71(GE_context* ac, T0* C, T6 a1);
/* detachable KL_EQUALITY_TESTER [[attached] STRING_8].default_create */
extern T0* T154c2(GE_context* ac);
/* detachable DS_HASH_SET [[attached] STRING_8].set_equality_tester */
extern void T70f41(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].make */
extern void T70f40(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].make */
extern T0* T70c40(GE_context* ac, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].set_internal_cursor */
extern void T70f57(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].new_cursor */
extern T0* T70f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8].make */
extern T0* T376c9(GE_context* ac, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].make_slots */
extern void T70f56(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].make_clashes */
extern void T70f55(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].make_key_storage */
extern void T70f54(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_SET [[attached] STRING_8].make_item_storage */
extern void T70f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE].make */
extern T0* T295c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE].set_internal_cursor */
extern void T295f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_FEATURE].new_cursor */
extern T0* T295f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_FEATURE].make */
extern T0* T691c5(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_map */
extern T0* T294c33(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_with_equality_testers */
extern void T294f37(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_sparse_container */
extern void T294f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].set_internal_cursor */
extern void T294f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].new_cursor */
extern T0* T294f30(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T689c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_slots */
extern void T294f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_clashes */
extern void T294f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_key_storage */
extern void T294f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable ET_DYNAMIC_FEATURE, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_item_storage */
extern void T294f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_FEATURE].make */
extern T0* T690f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_FEATURE].make_empty */
extern T0* T687c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_FEATURE].default_create */
extern T0* T690c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].make_map */
extern T0* T293c39(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].make_with_equality_testers */
extern void T293f43(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].make_sparse_container */
extern void T293f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].set_internal_cursor */
extern void T293f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].new_cursor */
extern T0* T293f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].make */
extern T0* T683c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].make_slots */
extern void T293f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].make_clashes */
extern void T293f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].make_key_storage */
extern void T293f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_CONSTANT].make */
extern T0* T686f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_INLINE_CONSTANT].make_empty */
extern T0* T684c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INLINE_CONSTANT].default_create */
extern T0* T686c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] ET_INLINE_CONSTANT].make_item_storage */
extern void T293f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].make_map */
extern T0* T292c39(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].make_with_equality_testers */
extern void T292f44(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].make_sparse_container */
extern void T292f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].set_internal_cursor */
extern void T292f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].new_cursor */
extern T0* T292f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_CONSTANT, [attached] ET_FEATURE].make */
extern T0* T679c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].make_slots */
extern void T292f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].make_clashes */
extern void T292f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].make_key_storage */
extern void T292f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_CONSTANT, [attached] ET_FEATURE].make_item_storage */
extern void T292f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTANT].make */
extern T0* T682f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_CONSTANT].make_empty */
extern T0* T680c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CONSTANT].default_create */
extern T0* T682c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].make_map */
extern T0* T290c36(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].make_with_equality_testers */
extern void T290f41(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].make_sparse_container */
extern void T290f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].set_internal_cursor */
extern void T290f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].new_cursor */
extern T0* T290f31(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_FEATURE].make */
extern T0* T676c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].make_slots */
extern void T290f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].make_clashes */
extern void T290f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].make_key_storage */
extern void T290f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_FEATURE].make_item_storage */
extern void T290f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].make */
extern T0* T289c35(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].set_internal_cursor */
extern void T289f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].new_cursor */
extern T0* T289f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DYNAMIC_TUPLE_TYPE].make */
extern T0* T672c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].make_slots */
extern void T289f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].make_clashes */
extern void T289f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].make_key_storage */
extern void T289f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DYNAMIC_TUPLE_TYPE].make_item_storage */
extern void T289f39(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TUPLE_TYPE].make */
extern T0* T674f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TUPLE_TYPE].make_empty */
extern T0* T673c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_TUPLE_TYPE].default_create */
extern T0* T674c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT].make */
extern T0* T280c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT].set_internal_cursor */
extern void T280f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_AGENT].new_cursor */
extern T0* T280f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_AGENT].make */
extern T0* T668c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT].make */
extern T0* T667f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_AGENT].make_empty */
extern T0* T666c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_AGENT].default_create */
extern T0* T667c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION].make */
extern T0* T279c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION].set_internal_cursor */
extern void T279f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EQUALITY_EXPRESSION].new_cursor */
extern T0* T279f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EQUALITY_EXPRESSION].make */
extern T0* T665c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EQUALITY_EXPRESSION].make */
extern T0* T664f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_EQUALITY_EXPRESSION].make_empty */
extern T0* T663c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EQUALITY_EXPRESSION].default_create */
extern T0* T664c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION].make */
extern T0* T278c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION].set_internal_cursor */
extern void T278f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_EQUALITY_EXPRESSION].new_cursor */
extern T0* T278f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_EQUALITY_EXPRESSION].make */
extern T0* T662c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_EQUALITY_EXPRESSION].make */
extern T0* T661f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_OBJECT_EQUALITY_EXPRESSION].make_empty */
extern T0* T660c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_EQUALITY_EXPRESSION].default_create */
extern T0* T661c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST].make */
extern T0* T277c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST].set_internal_cursor */
extern void T277f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_OBJECT_TEST].new_cursor */
extern T0* T277f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_OBJECT_TEST].make */
extern T0* T659c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST].make */
extern T0* T658f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST].make_empty */
extern T0* T657c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST].default_create */
extern T0* T658c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_map */
extern T0* T276c39(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_with_equality_testers */
extern void T276f45(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_sparse_container */
extern void T276f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].set_internal_cursor */
extern void T276f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].new_cursor */
extern T0* T276f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T655c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_slots */
extern void T276f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_clashes */
extern void T276f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_key_storage */
extern void T276f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET, [attached] ET_DYNAMIC_PRIMARY_TYPE].make_item_storage */
extern void T276f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].make */
extern T0* T648f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].make_empty */
extern T0* T647c5(GE_context* ac, T6 a1);
/* [detachable] ET_DYNAMIC_STANDALONE_TYPE_SET_LIST.make_with_capacity */
extern T0* T274c8(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].make_map */
extern T0* T273c33(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].make_with_equality_testers */
extern void T273f38(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].make_sparse_container */
extern void T273f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].set_internal_cursor */
extern void T273f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].new_cursor */
extern T0* T273f30(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].make */
extern T0* T654c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].make_slots */
extern void T273f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].make_clashes */
extern void T273f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].make_key_storage */
extern void T273f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] INTEGER_32].make_item_storage */
extern void T273f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].make */
extern T0* T272c12(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].set_internal_cursor */
extern void T272f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_EXPRESSION].new_cursor */
extern T0* T272f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_EXPRESSION].make */
extern T0* T652c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION].make */
extern T0* T651f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_EXPRESSION].make_empty */
extern T0* T650c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_EXPRESSION].default_create */
extern T0* T651c3(GE_context* ac);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_EXPRESSION].make */
extern T0* T271c9(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].make */
extern T0* T270c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].set_internal_cursor */
extern void T270f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].new_cursor */
extern T0* T270f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DYNAMIC_STANDALONE_TYPE_SET].make */
extern T0* T649c5(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_RESULT].make */
extern T0* T267c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_RESULT].set_internal_cursor */
extern void T267f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_RESULT].new_cursor */
extern T0* T267f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_RESULT].make */
extern T0* T646c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RESULT].make */
extern T0* T644f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_RESULT].make_empty */
extern T0* T645c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_RESULT].default_create */
extern T0* T644c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACTUAL_ARGUMENT_LIST].make */
extern T0* T266c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACTUAL_ARGUMENT_LIST].set_internal_cursor */
extern void T266f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ACTUAL_ARGUMENT_LIST].new_cursor */
extern T0* T266f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ACTUAL_ARGUMENT_LIST].make */
extern T0* T643c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACTUAL_ARGUMENT_LIST].make */
extern T0* T641f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ACTUAL_ARGUMENT_LIST].make_empty */
extern T0* T642c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ACTUAL_ARGUMENT_LIST].default_create */
extern T0* T641c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUALIFIED_CALL_EXPRESSION].make */
extern T0* T265c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUALIFIED_CALL_EXPRESSION].set_internal_cursor */
extern void T265f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_QUALIFIED_CALL_EXPRESSION].new_cursor */
extern T0* T265f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_QUALIFIED_CALL_EXPRESSION].make */
extern T0* T640c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_QUALIFIED_CALL_EXPRESSION].make */
extern T0* T639f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_QUALIFIED_CALL_EXPRESSION].make_empty */
extern T0* T638c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_QUALIFIED_CALL_EXPRESSION].default_create */
extern T0* T639c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].make */
extern T0* T264c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].set_internal_cursor */
extern void T264f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].new_cursor */
extern T0* T264f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].make */
extern T0* T637c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].make */
extern T0* T634f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].make_empty */
extern T0* T636c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_UNQUALIFIED_CALL_EXPRESSION].default_create */
extern T0* T634c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T262c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE].set_internal_cursor */
extern void T262f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_DYNAMIC_PRIMARY_TYPE].new_cursor */
extern T0* T262f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T631c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_PRIMARY_TYPE].make */
extern T0* T630f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_DYNAMIC_PRIMARY_TYPE].make_empty */
extern T0* T629c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DYNAMIC_PRIMARY_TYPE].default_create */
extern T0* T630c3(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER].make */
extern T0* T261c12(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER].set_internal_cursor */
extern void T261f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_IDENTIFIER].new_cursor */
extern T0* T261f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_IDENTIFIER].make */
extern T0* T628c5(GE_context* ac, T0* a1);
/* [detachable] KL_STRING_OUTPUT_STREAM.make */
extern T0* T260c4(GE_context* ac, T0* a1);
/* [detachable] ET_C_GENERATOR.null_output_stream */
extern T0* T67f514(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.lower_name */
extern T0* T74f160(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.base_class */
extern T0* T72f11(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.same_named_type */
extern T1 T72f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_class */
extern T0* T73f1(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.make_unknown */
extern T0* T74c253(GE_context* ac);
/* [detachable] GEC.tokens */
extern T0* T25f46(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.record_end_time */
extern void T64f57(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_SYSTEM_PROCESSOR.print_time */
extern void T64f69(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DT_DATE_TIME_DURATION.out */
extern T0* T468f9(GE_context* ac, T0* C);
/* [detachable] DT_DATE_TIME_DURATION.append_to_string */
extern void T468f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DT_DATE_TIME_DURATION.append_time_to_string */
extern void T468f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DT_DATE_TIME_DURATION.integer_ */
extern T0* T468f14(GE_context* ac, T0* C);
/* [detachable] DT_DATE_TIME_DURATION.append_date_to_string */
extern void T468f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DT_DATE_TIME_DURATION.set_canonical */
extern void T468f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DT_DATE_TIME_DURATION.make_canonical_from_dates */
extern void T468f21(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DT_DATE_TIME_DURATION.make_precise */
extern void T468f16(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3, T6 a4, T6 a5, T6 a6, T6 a7);
/* [detachable] DT_DATE_TIME_DURATION.make_precise */
extern T0* T468c16(GE_context* ac, T6 a1, T6 a2, T6 a3, T6 a4, T6 a5, T6 a6, T6 a7);
/* [detachable] DT_DATE_TIME_DURATION.make_precise_time_duration */
extern void T468f20(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3, T6 a4);
/* [detachable] DT_DATE_TIME_DURATION.make_date_duration */
extern void T468f19(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* detachable DT_DATE_TIME.days_in_current_month */
extern T6 T68f43(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.days_in_month */
extern T6 T68f45(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DT_DATE_TIME.leap_year */
extern T1 T68f38(GE_context* ac, T0* C, T6 a1);
/* detachable DT_DATE_TIME.days_in_previous_month */
extern T6 T68f42(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.is_less */
extern T1 T68f41(GE_context* ac, T0* C, T0* a1);
/* detachable DT_DATE_TIME.millisecond_count */
extern T6 T68f40(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.day */
extern T6 T68f10(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.month */
extern T6 T68f9(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.year */
extern T6 T68f8(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.add_duration */
extern void T68f62(GE_context* ac, T0* C, T0* a1);
/* detachable DT_DATE_TIME.add_milliseconds */
extern void T68f66(GE_context* ac, T0* C, T6 a1);
/* detachable DT_DATE_TIME.set_normalized_time_storage */
extern void T68f69(GE_context* ac, T0* C, T7 a1);
/* detachable DT_DATE_TIME.add_days */
extern void T68f70(GE_context* ac, T0* C, T6 a1);
/* detachable DT_DATE_TIME.set_from_epoch_days */
extern void T68f68(GE_context* ac, T0* C, T6 a1);
/* detachable DT_DATE_TIME.set_year_month_day */
extern void T68f57(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* detachable DT_DATE_TIME.epoch_to_y2k_days */
extern T6 T68f46(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.epoch_days */
extern T6 T68f11(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* detachable DT_DATE_TIME.epoch_days_at_month */
extern T6 T68f24(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.days_at_month */
extern T6 T68f23(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DT_DATE_TIME.epoch_leap_years */
extern T6 T68f22(GE_context* ac, T0* C);
/* [detachable] KL_INTEGER_ROUTINES.div */
extern T6 T772f4(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DT_DATE_TIME.integer_ */
extern T0* T68f21(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.set_day */
extern void T68f67(GE_context* ac, T0* C, T6 a1);
/* INTEGER_64.to_integer */
extern T6 T7f21(GE_context* ac, T7* C);
/* [detachable] DT_DATE_TIME_DURATION.millisecond_count */
extern T6 T468f15(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.add_years_months_days */
extern void T68f65(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* detachable DT_DATE_TIME.set_time */
extern void T68f61(GE_context* ac, T0* C, T0* a1);
/* detachable DT_DATE_TIME.set_date */
extern void T68f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] DT_DATE_TIME_DURATION.tmp_date_time */
extern T0* T468f8(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.make */
extern T0* T68c59(GE_context* ac, T6 a1, T6 a2, T6 a3, T6 a4, T6 a5, T6 a6);
/* detachable DT_DATE_TIME.make_time */
extern void T68f64(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* detachable DT_DATE_TIME.make_date */
extern void T68f63(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* detachable DT_DATE_TIME.minus */
extern T0* T68f12(GE_context* ac, T0* C, T0* a1);
/* detachable DT_DATE_TIME.duration */
extern T0* T68f16(GE_context* ac, T0* C, T0* a1);
/* detachable DT_DATE_TIME.millisecond */
extern T6 T68f7(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.second */
extern T6 T68f6(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.minute */
extern T6 T68f5(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.hour */
extern T6 T68f4(GE_context* ac, T0* C);
/* detachable DT_DATE_TIME.day_count */
extern T6 T68f3(GE_context* ac, T0* C);
/* [detachable] DT_SYSTEM_CLOCK.date_time_now */
extern T0* T453f10(GE_context* ac, T0* C);
/* [detachable] DT_SYSTEM_CLOCK.set_date_time_to_now */
extern void T453f12(GE_context* ac, T0* C, T0* a1);
/* detachable DT_DATE_TIME.set_precise_hour_minute_second */
extern void T68f58(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3, T6 a4);
/* [detachable] DT_SYSTEM_CLOCK.set_local_time */
extern void T453f13(GE_context* ac, T0* C);
/* detachable C_DATE.second_now */
extern T6 T929f6(GE_context* ac, T0* C);
/* detachable C_DATE.get_tm_sec */
extern T6 T929f13(GE_context* ac, T0* C, T14 a1);
/* detachable C_DATE.minute_now */
extern T6 T929f5(GE_context* ac, T0* C);
/* detachable C_DATE.get_tm_min */
extern T6 T929f12(GE_context* ac, T0* C, T14 a1);
/* detachable C_DATE.hour_now */
extern T6 T929f4(GE_context* ac, T0* C);
/* detachable C_DATE.get_tm_hour */
extern T6 T929f11(GE_context* ac, T0* C, T14 a1);
/* detachable C_DATE.day_now */
extern T6 T929f15(GE_context* ac, T0* C);
/* detachable C_DATE.get_tm_mday */
extern T6 T929f10(GE_context* ac, T0* C, T14 a1);
/* detachable C_DATE.month_now */
extern T6 T929f8(GE_context* ac, T0* C);
/* detachable C_DATE.get_tm_mon */
extern T6 T929f9(GE_context* ac, T0* C, T14 a1);
/* detachable C_DATE.year_now */
extern T6 T929f7(GE_context* ac, T0* C);
/* detachable C_DATE.get_tm_year */
extern T6 T929f14(GE_context* ac, T0* C, T14 a1);
/* detachable C_DATE.update */
extern void T929f23(GE_context* ac, T0* C);
/* POINTER.memory_free */
extern void T14f14(GE_context* ac, T14* C);
/* POINTER.default_pointer */
extern T14 T14f10(GE_context* ac, T14* C);
/* POINTER.c_free */
extern void T14f16(GE_context* ac, T14* C, T14 a1);
/* detachable C_DATE.get_millitm */
extern T6 T929f21(GE_context* ac, T0* C, T14 a1);
/* detachable C_DATE.tm_structure_size */
extern T6 T929f20(GE_context* ac, T0* C);
/* detachable C_DATE.localtime */
extern T14 T929f19(GE_context* ac, T0* C, T14 a1);
/* detachable C_DATE.gmtime */
extern T14 T929f18(GE_context* ac, T0* C, T14 a1);
/* detachable C_DATE.get_time */
extern void T929f25(GE_context* ac, T0* C, T14 a1, T14 a2);
/* detachable C_DATE.ftime */
extern void T929f24(GE_context* ac, T0* C, T14 a1);
/* detachable C_DATE.time_t_structure_size */
extern T6 T929f17(GE_context* ac, T0* C);
/* detachable C_DATE.timeb_structure_size */
extern T6 T929f16(GE_context* ac, T0* C);
/* detachable C_DATE.default_create */
extern T0* T929c22(GE_context* ac);
/* detachable DT_DATE_TIME.make_from_storage */
extern T0* T68c56(GE_context* ac, T6 a1, T6 a2);
/* [detachable] DT_SHARED_SYSTEM_CLOCK.system_clock */
extern T0* T452f1(GE_context* ac, T0* C);
/* [detachable] DT_SYSTEM_CLOCK.make */
extern T0* T453c11(GE_context* ac);
/* [detachable] DT_SHARED_SYSTEM_CLOCK.default_create */
extern T0* T452c2(GE_context* ac);
/* [detachable] ET_SYSTEM_PROCESSOR.stop_requested */
extern T1 T64f39(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.record_end_time */
extern void T375f59(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.print_time */
extern void T375f71(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.stop_requested */
extern T1 T375f40(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.compile */
extern void T63f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.compile_system */
extern void T63f87(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.build_dynamic_type_sets */
extern void T63f91(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.build_dynamic_type_sets */
extern void T230f19(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.build_dynamic_type_sets */
extern void T71f107(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_catcall_validity */
extern void T71f111(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_catcall_call_validity */
extern void T71f116(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_catcall_target_validity */
extern void T71f118(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T71f118ot1(T0* a1, T0** a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.report_catcall_error */
extern void T71f122(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T0* a4, T0* a5, T0* a6, T0* a7);
/* [detachable] ET_NULL_ERROR_HANDLER.report_catcall_error */
extern void T53f201(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_catcall_error */
extern void T36f201(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.append_catcall_error_message */
extern void T71f123(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T6 a4, T0* a5, T0* a6, T0* a7, T0* a8);
/* detachable ET_DYNAMIC_QUALIFIED_PROCEDURE_CALL.position */
extern T0* T318f11(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_QUALIFIED_QUERY_CALL.position */
extern T0* T314f12(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIS_EXPRESSION.position */
extern T0* T888f10(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.string_ */
extern T0* T71f67(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.shared_error_message */
extern T0* T71f66(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.build_feature_dynamic_type_sets */
extern void T71f110(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_TYPE_SET_LIST.wipe_out */
extern void T259f14(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].wipe_out */
extern void T381f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].slots_wipe_out */
extern void T381f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].clashes_wipe_out */
extern void T381f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].key_storage_wipe_out */
extern void T381f39(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE].keep_head */
extern void T708f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DYNAMIC_TYPE].set_count */
extern void T708f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].item_storage_wipe_out */
extern void T381f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].move_all_cursors_after */
extern void T381f36(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].set_next_cursor */
extern void T707f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].set_after */
extern void T707f7(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.set_built */
extern void T306f50(GE_context* ac, T0* C, T1 a1);
/* detachable ET_DYNAMIC_FEATURE.set_built */
extern void T236f53(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_feature_validity */
extern void T71f114(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.build_preconditions_attachment_scope */
extern void T71f117(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.build_assertions_attachment_scope */
extern void T71f120(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.add_precursors_with_preconditions_recursive */
extern void T71f119(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.check_precursor_feature_validity */
extern void T71f113(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_DYNAMIC_FEATURE.is_precursor */
extern T1 T236f36(GE_context* ac, T0* C);
/* detachable ET_DYNAMIC_PRECURSOR.set_dynamic_type_sets */
extern void T306f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.set_catcall_warning_mode */
extern void T230f15(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.set_catcall_warning_mode */
extern void T71f103(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.set_catcall_error_mode */
extern void T230f14(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.set_catcall_error_mode */
extern void T71f102(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.set_no_assertion */
extern void T230f18(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.set_no_assertion */
extern void T71f106(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.set_no_debug */
extern void T230f17(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.set_no_debug */
extern void T71f105(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_CLASS.is_creation_directly_exported_to */
extern T1 T74f198(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CREATOR_LIST.is_directly_exported_to */
extern T1 T1107f7(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_CREATOR.is_directly_exported_to */
extern T1 T1104f9(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvsrc5a_error */
extern void T53f192(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_system_error */
extern void T53f202(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_ERROR.make_gvsrc5a */
extern T0* T783c32(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvsrc5_error */
extern T1 T53f16(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_gvsrc5a_error */
extern void T36f192(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.report_system_error */
extern void T36f202(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_gvsrc5_error */
extern T1 T36f16(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vsrc1a_error */
extern void T53f191(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_ERROR.make_vsrc1a */
extern T0* T783c31(GE_context* ac, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vsrc1_error */
extern T1 T53f15(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_vsrc1a_error */
extern void T36f191(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_vsrc1_error */
extern T1 T36f15(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvsrc4a_error */
extern void T53f190(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_ERROR.make_gvsrc4a */
extern T0* T783c30(GE_context* ac, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvsrc4_error */
extern T1 T53f14(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_gvsrc4a_error */
extern void T36f190(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_gvsrc4_error */
extern T1 T36f14(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvsrc6a_error */
extern void T53f189(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_SYSTEM_ERROR.make_gvsrc6a */
extern T0* T783c29(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvsrc6_error */
extern T1 T53f13(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_gvsrc6a_error */
extern void T36f189(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_gvsrc6_error */
extern T1 T36f13(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvsrc3a_error */
extern void T53f188(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_ERROR.make_gvsrc3a */
extern T0* T783c28(GE_context* ac);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvsrc3_error */
extern T1 T53f12(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_gvsrc3a_error */
extern void T36f188(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.reportable_gvsrc3_error */
extern T1 T36f12(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.compile_kernel */
extern void T63f89(GE_context* ac, T0* C, T0* a1);
extern T1 T63f89ot1(T0* a1, T0** a2);
/* [detachable] ET_TOKEN_CONSTANTS.set_exception_data_feature_name */
extern T0* T73f20(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.once_raise_feature_name */
extern T0* T73f19(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.init_exception_manager_feature_name */
extern T0* T73f17(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.set_rout_disp_final_feature_name */
extern T0* T73f16(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.rout_disp_feature_name */
extern T0* T73f15(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.closed_operands_feature_name */
extern T0* T73f14(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.to_pointer_feature_name */
extern T0* T73f11(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.ise_6_0_6_7057 */
extern T0* T63f77(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.upper_feature_name */
extern T0* T73f9(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.lower_feature_name */
extern T0* T73f8(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.area_feature_name */
extern T0* T73f7(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvkfe6a_error */
extern void T53f199(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_VALIDITY_ERROR.make_gvkfe6a */
extern T0* T785c855(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvkfe6_error */
extern T1 T53f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvkfe6a_error */
extern void T36f199(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.reportable_gvkfe6_error */
extern T1 T36f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvkfe3a_error */
extern void T53f197(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_VALIDITY_ERROR.make_gvkfe3a */
extern T0* T785c853(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvkfe3_error */
extern T1 T53f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvkfe3a_error */
extern void T36f197(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.reportable_gvkfe3_error */
extern T1 T36f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvkfe2a_error */
extern void T53f195(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_VALIDITY_ERROR.make_gvkfe2a */
extern T0* T785c851(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvkfe2_error */
extern T1 T53f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gvkfe2a_error */
extern void T36f195(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_gvkfe2_error */
extern T1 T36f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gvknl1a_error */
extern void T53f194(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_ERROR.make_gvknl1a */
extern T0* T783c33(GE_context* ac, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gvknl1_error */
extern T1 T53f17(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_gvknl1a_error */
extern void T36f194(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_gvknl1_error */
extern T1 T36f17(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.compile_degree_6 */
extern void T64f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.report_degree_6_metrics */
extern void T64f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.class_count_recursive */
extern T6 T58f139(GE_context* ac, T0* C);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.class_count_recursive */
extern T0* T58f139ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.class_count_recursive */
extern T0* T58f139ac2(GE_context* ac, T0* a1, T0* a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_CLASS]].set_rout_disp_final */
extern void T228f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] KL_AGENT_ROUTINES [[attached] ET_CLASS].call */
extern void T834f2(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] UT_COUNTER.increment */
extern void T833f3(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.classes_do_recursive */
extern void T58f173(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.classes_do_recursive */
extern T0* T58f173ac1(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.classes_do_all */
extern void T211f158(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.classes_do_all */
extern void T211f158p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].forth */
extern void T609f12(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].cursor_forth */
extern void T225f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].add_traversing_cursor */
extern void T225f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].set_next_cursor */
extern void T609f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].set_position */
extern void T609f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].remove_traversing_cursor */
extern void T225f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].set_after */
extern void T609f14(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.is_mapped */
extern T1 T570f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].item */
extern T0* T609f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].cursor_item */
extern T0* T225f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].after */
extern T1 T609f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].start */
extern void T609f11(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].cursor_start */
extern void T225f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].cursor_off */
extern T1 T225f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].off */
extern T1 T609f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].is_empty */
extern T1 T225f35(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.classes_do_all */
extern void T596f158(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.classes_do_all */
extern void T596f158p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.classes_do_all */
extern void T187f188(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.classes_do_all */
extern void T852f159(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.classes_do_all */
extern void T40f199(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.classes_do_all */
extern void T58f272(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.universes_do_recursive */
extern void T58f177(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].do_all */
extern void T855f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].item_storage_item */
extern T0* T855f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].clashes_item */
extern T6 T855f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.add_universe_recursive */
extern void T58f190(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.add_universe_recursive */
extern T0* T58f190ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.add_universe_recursive */
extern T0* T58f190ac2(GE_context* ac, T0* a1);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]].set_rout_disp_final */
extern void T845f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.add_universe_recursive */
extern void T211f156(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_DOTNET_ASSEMBLY.add_universe_recursive */
extern T0* T211f156ac1(GE_context* ac, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLIES.do_all */
extern void T595f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.add_universe_recursive */
extern void T596f156(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_DOTNET_ASSEMBLY.add_universe_recursive */
extern T0* T596f156ac1(GE_context* ac, T0* a1);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_LIBRARY]].set_rout_disp_final */
extern void T854f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_ECF_LIBRARY.add_universe_recursive */
extern void T187f185(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_LIBRARY.add_universe_recursive */
extern T0* T187f185ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_LIBRARY.add_universe_recursive */
extern T0* T187f185ac2(GE_context* ac, T0* a1);
/* [detachable] ET_LIBRARY.add_universe_recursive */
extern void T852f156(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_LIBRARY.add_universe_recursive */
extern T0* T852f156ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_LIBRARY.add_universe_recursive */
extern T0* T852f156ac2(GE_context* ac, T0* a1);
/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.do_all */
extern void T210f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ADAPTED_LIBRARIES.do_all */
extern void T188f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].force_last */
extern void T855f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].slots_put */
extern void T855f55(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].clashes_put */
extern void T855f54(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].slots_item */
extern T6 T855f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].hash_position */
extern T6 T855f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].resize */
extern void T855f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].clashes_resize */
extern void T855f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].special_integer_ */
extern T0* T855f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].key_storage_resize */
extern void T855f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].item_storage_resize */
extern void T855f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].key_storage_item */
extern T0* T855f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].slots_resize */
extern void T855f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].new_modulus */
extern T6 T855f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].new_capacity */
extern T6 T855f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].item_storage_put */
extern void T855f52(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].search_position */
extern void T855f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].key_equality_tester */
extern T0* T855f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].unset_found_item */
extern void T855f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].has */
extern T1 T855f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].make */
extern T0* T855c36(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].set_internal_cursor */
extern void T855f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].new_cursor */
extern T0* T855f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_UNIVERSE].make */
extern T0* T1608c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].make_slots */
extern void T855f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].make_clashes */
extern void T855f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].make_key_storage */
extern void T855f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].make_item_storage */
extern void T855f43(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.initial_universes_capacity */
extern T6 T58f149(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.class_actions */
extern T0* T58f147(GE_context* ac, T0* C);
/* [detachable] KL_AGENT_ROUTINES [[attached] ET_CLASS].default_create */
extern T0* T834c1(GE_context* ac);
/* [detachable] UT_COUNTER.make */
extern T0* T833c2(GE_context* ac, T6 a1);
/* [detachable] ET_ECF_SYSTEM.preparse_recursive */
extern void T58f175(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.build_scm_write_mappings */
extern void T58f180(GE_context* ac, T0* C);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.build_scm_write_mappings */
extern T0* T58f180ac1(GE_context* ac, T0* a1);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_CLUSTER]].set_rout_disp_final */
extern void T858f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_ECF_SYSTEM.build_scm_write_mapping */
extern void T58f193(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_scm_write_mapping */
extern void T802f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.clusters_do_explicit_recursive */
extern void T58f192(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.clusters_do_explicit_recursive */
extern T0* T58f192ac1(GE_context* ac, T0* a1);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_INTERNAL_UNIVERSE]].set_rout_disp_final */
extern void T848f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_ECF_LIBRARY.clusters_do_explicit */
extern void T187f197(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLUSTERS.do_explicit */
extern void T564f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTERS.do_recursive */
extern void T787f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].item */
extern T0* T1579f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLUSTER].item */
extern T0* T1384f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_LIBRARY.clusters_do_explicit */
extern void T852f169(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.clusters_do_explicit */
extern void T58f279(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.internal_universes_do_recursive */
extern void T58f181(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].do_all */
extern void T860f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].item_storage_item */
extern T0* T860f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].clashes_item */
extern T6 T860f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.add_internal_universe_recursive */
extern void T58f194(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.add_internal_universe_recursive */
extern T0* T58f194ac1(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_LIBRARY.add_internal_universe_recursive */
extern void T187f186(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_LIBRARY.add_internal_universe_recursive */
extern T0* T187f186ac1(GE_context* ac, T0* a1);
/* [detachable] ET_LIBRARY.add_internal_universe_recursive */
extern void T852f157(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_LIBRARY.add_internal_universe_recursive */
extern T0* T852f157ac1(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].force_last */
extern void T860f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].slots_put */
extern void T860f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].clashes_put */
extern void T860f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].slots_item */
extern T6 T860f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].hash_position */
extern T6 T860f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].resize */
extern void T860f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].clashes_resize */
extern void T860f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].special_integer_ */
extern T0* T860f27(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].key_storage_resize */
extern void T860f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].item_storage_resize */
extern void T860f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INTERNAL_UNIVERSE].aliased_resized_area */
extern T0* T1614f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_INTERNAL_UNIVERSE].aliased_resized_area */
extern T0* T1615f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].key_storage_item */
extern T0* T860f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].slots_resize */
extern void T860f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].new_modulus */
extern T6 T860f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].new_capacity */
extern T6 T860f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].item_storage_put */
extern void T860f40(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INTERNAL_UNIVERSE].force */
extern void T1614f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INTERNAL_UNIVERSE].fill_with */
extern void T1615f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_INTERNAL_UNIVERSE].extend */
extern void T1615f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].search_position */
extern void T860f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].key_equality_tester */
extern T0* T860f19(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].unset_found_item */
extern void T860f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].has */
extern T1 T860f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].make */
extern T0* T860c30(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].set_internal_cursor */
extern void T860f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].new_cursor */
extern T0* T860f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_INTERNAL_UNIVERSE].make */
extern T0* T1612c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].make_slots */
extern void T860f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].make_clashes */
extern void T860f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].make_key_storage */
extern void T860f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_INTERNAL_UNIVERSE].make_item_storage */
extern void T860f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INTERNAL_UNIVERSE].make */
extern T0* T1614f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_INTERNAL_UNIVERSE].make_empty */
extern T0* T1615c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_INTERNAL_UNIVERSE].default_create */
extern T0* T1614c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM.build_scm_read_mappings */
extern void T58f179(GE_context* ac, T0* C);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.build_scm_read_mappings */
extern T0* T58f179ac1(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM.build_scm_read_mapping */
extern void T58f191(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_scm_read_mapping */
extern void T802f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.check_master_class_validity */
extern void T375f110(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_SYSTEM_MULTIPROCESSOR.check_master_class_validity */
extern T0* T375f110ac1(GE_context* ac, T0* a1);
/* [detachable] ET_MASTER_CLASS.process */
extern void T570f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_NULL_PROCESSOR.process_master_class */
extern void T464f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS_CHECKER.process_master_class */
extern void T455f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS_CHECKER.check_master_class_validity */
extern void T455f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS_CHECKER.check_overriding_classes_validity */
extern void T455f12(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vscn0d_error */
extern void T53f405(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_universe_error */
extern void T53f410(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_UNIVERSE_ERROR.make_vscn0d */
extern T0* T1999c21(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_MASTER_CLASS.upper_name */
extern T0* T570f49(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.full_lower_name */
extern T0* T852f97(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_LIBRARY.relative_lower_name */
extern T0* T852f100(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_LIBRARY.string_ */
extern T0* T852f103(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.lower_name */
extern T0* T852f102(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.full_lower_name */
extern T0* T40f133(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_ECF_SYSTEM.full_lower_name */
extern T0* T58f159(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_ECF_LIBRARY.full_lower_name */
extern T0* T187f107(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_LIBRARY.kind_name */
extern T0* T852f99(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.kind_name */
extern T0* T596f102(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.kind_name */
extern T0* T40f132(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.kind_name */
extern T0* T58f158(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.kind_name */
extern T0* T211f99(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.kind_name */
extern T0* T187f114(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_vscn_error */
extern T1 T53f110(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_vscn0d_error */
extern void T36f405(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_universe_error */
extern void T36f410(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.reportable_vscn_error */
extern T1 T36f110(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS_CHECKER.current_universe */
extern T0* T455f6(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS_CHECKER.error_handler */
extern T0* T455f5(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS_CHECKER.set_fatal_error */
extern void T455f13(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS_CHECKER.check_intrinsic_classes_validity */
extern void T455f11(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS_CHECKER.check_overridden_class_validity */
extern void T455f14(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vscn0c_error */
extern void T53f407(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_UNIVERSE_ERROR.make_vscn0c */
extern T0* T1999c23(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_CLUSTER.kind_name */
extern T0* T802f64(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.kind_name */
extern T0* T887f20(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.kind_name */
extern T0* T410f21(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_vscn0c_error */
extern void T36f407(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vscn0b_error */
extern void T53f406(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_UNIVERSE_ERROR.make_vscn0b */
extern T0* T1999c22(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ERROR_HANDLER.report_vscn0b_error */
extern void T36f406(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_MASTER_CLASS.actual_intrinsic_class */
extern T0* T570f26(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.actual_intrinsic_class */
extern T0* T74f193(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.report_vscn0a_error */
extern void T53f404(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_UNIVERSE_ERROR.make_vscn0a */
extern T0* T1999c20(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ERROR_HANDLER.report_vscn0a_error */
extern void T36f404(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_MASTER_CLASS_CHECKER.reset_fatal_error */
extern void T455f10(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_SYSTEM.master_classes_do_recursive_until */
extern void T58f241(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.master_classes_do_recursive_until */
extern T0* T58f241ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.master_classes_do_recursive_until */
extern T0* T58f241ac2(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.master_classes_do_until */
extern void T211f163(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].go_after */
extern void T609f13(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].cursor_go_after */
extern void T225f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.master_classes_do_until */
extern void T596f163(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_LIBRARY.master_classes_do_until */
extern void T187f193(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIBRARY.master_classes_do_until */
extern void T852f164(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_SYSTEM.master_classes_do_until */
extern void T40f203(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.master_classes_do_until */
extern void T58f275(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.master_classes_do_all */
extern void T211f160(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].do_all */
extern void T225f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.master_classes_do_all */
extern void T596f160(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.master_classes_do_all */
extern void T187f190(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.master_classes_do_all */
extern void T852f161(GE_context* ac, T0* C, T0* a1);
/* detachable ET_SYSTEM.master_classes_do_all */
extern void T40f201(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.master_classes_do_all */
extern void T58f195(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.universes_do_recursive_until */
extern void T58f178(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].forth */
extern void T855f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].cursor_forth */
extern void T855f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].add_traversing_cursor */
extern void T855f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_UNIVERSE].set_next_cursor */
extern void T1608f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_UNIVERSE].set_position */
extern void T1608f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].remove_traversing_cursor */
extern void T855f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_UNIVERSE].set_after */
extern void T1608f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].item_for_iteration */
extern T0* T855f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].cursor_item */
extern T0* T855f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].go_after */
extern void T855f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].cursor_go_after */
extern void T855f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].cursor_off */
extern T1 T855f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_UNIVERSE].off */
extern T1 T1608f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].after */
extern T1 T855f27(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].cursor_after */
extern T1 T855f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_UNIVERSE].after */
extern T1 T1608f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].start */
extern void T855f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].cursor_start */
extern void T855f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_UNIVERSE].is_empty */
extern T1 T855f31(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.check_master_class_validity */
extern void T64f106(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_SYSTEM_PROCESSOR.check_master_class_validity */
extern T0* T64f106ac1(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM.preparse_recursive */
extern void T58f175p1(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.preparse_recursive */
extern T0* T58f175p1ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.preparse_recursive */
extern T0* T58f175p1ac2(GE_context* ac);
/* Creation of agent #3 in feature [detachable] ET_ECF_SYSTEM.preparse_recursive */
extern T0* T58f175p1ac3(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #4 in feature [detachable] ET_ECF_SYSTEM.preparse_recursive */
extern T0* T58f175p1ac4(GE_context* ac, T0* a1);
/* Creation of agent #5 in feature [detachable] ET_ECF_SYSTEM.preparse_recursive */
extern T0* T58f175p1ac5(GE_context* ac, T0* a1);
/* Creation of agent #6 in feature [detachable] ET_ECF_SYSTEM.preparse_recursive */
extern T0* T58f175p1ac6(GE_context* ac, T0* a1);
/* Creation of agent #7 in feature [detachable] ET_ECF_SYSTEM.preparse_recursive */
extern T0* T58f175p1ac7(GE_context* ac);
/* Creation of agent #8 in feature [detachable] ET_ECF_SYSTEM.preparse_recursive */
extern T0* T58f175p1ac8(GE_context* ac);
/* [detachable] ET_ECF_LIBRARY.import_classes */
extern void T187f184(GE_context* ac, T0* C);
/* Creation of agent #1 in feature [detachable] ET_ECF_LIBRARY.import_classes */
extern T0* T187f184ac1(GE_context* ac);
/* Creation of agent #2 in feature [detachable] ET_ECF_LIBRARY.import_classes */
extern T0* T187f184ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_ECF_LIBRARY.import_classes */
extern T0* T187f184ac3(GE_context* ac, T0* a1);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ADAPTED_DOTNET_ASSEMBLY]].set_rout_disp_final */
extern void T865f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.export_classes */
extern void T213f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.add_last_imported_class */
extern void T570f85(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.has_imported_class */
extern T1 T570f47(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].has */
extern T1 T1389f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.master_class */
extern T0* T852f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].item */
extern T0* T216f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].key */
extern T0* T609f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].cursor_key */
extern T0* T225f37(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.is_hidden */
extern T1 T802f7(GE_context* ac, T0* C);
/* [detachable] ET_NONE_GROUP.is_hidden */
extern T1 T887f16(GE_context* ac, T0* C);
/* [detachable] ET_UNKNOWN_GROUP.is_hidden */
extern T1 T410f14(GE_context* ac, T0* C);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ADAPTED_LIBRARY]].set_rout_disp_final */
extern void T191f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.export_classes */
extern void T207f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.remove_unknown_imported_classes */
extern void T570f59(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.remove_imported_class */
extern void T570f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.is_declared_locally */
extern T1 T570f24(GE_context* ac, T0* C);
/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.do_adapted */
extern void T210f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].do_all */
extern void T594f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ADAPTED_LIBRARIES.do_adapted */
extern void T188f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY].do_all */
extern void T578f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.import_classes */
extern void T852f155(GE_context* ac, T0* C);
/* Creation of agent #1 in feature [detachable] ET_LIBRARY.import_classes */
extern T0* T852f155ac1(GE_context* ac);
/* Creation of agent #2 in feature [detachable] ET_LIBRARY.import_classes */
extern T0* T852f155ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_LIBRARY.import_classes */
extern T0* T852f155ac3(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.import_classes */
extern void T211f155(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.import_kernel_class */
extern void T211f167(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.master_class */
extern T0* T211f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.tokens */
extern T0* T211f96(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.import_kernel_classes */
extern void T211f166(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.native_array_class_name */
extern T0* T73f1539(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.import_classes */
extern void T596f155(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.import_kernel_class */
extern void T596f167(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.master_class */
extern T0* T596f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.tokens */
extern T0* T596f95(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.import_kernel_classes */
extern void T596f166(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.preparse */
extern void T187f183(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.preparse */
extern void T187f183p1(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_LIBRARY.preparse */
extern T0* T187f183p1ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_LIBRARY.preparse */
extern T0* T187f183p1ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_ECF_LIBRARY.preparse */
extern T0* T187f183p1ac3(GE_context* ac);
/* Creation of agent #4 in feature [detachable] ET_ECF_LIBRARY.preparse */
extern T0* T187f183p1ac4(GE_context* ac, T0* a1);
/* [detachable] ET_MASTER_CLASS.remove_unknown_local_classes */
extern void T570f58(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.remove_local_ignored_class */
extern void T570f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].remove_first */
extern void T462f26(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].delete */
extern void T462f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.remove_local_non_override_class */
extern void T570f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.remove_local_override_class */
extern void T570f71(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].remove */
extern void T462f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].move_left */
extern void T462f32(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].move_cursors_left */
extern void T462f31(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].remove_last */
extern void T462f30(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].move_last_cursors_after */
extern void T462f34(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.reset_local_modified_classes */
extern void T570f57(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_MASTER_CLASS.reset_local_modified_classes */
extern T0* T570f57ac1(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #2 in feature [detachable] ET_MASTER_CLASS.reset_local_modified_classes */
extern T0* T570f57ac2(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #3 in feature [detachable] ET_MASTER_CLASS.reset_local_modified_classes */
extern T0* T570f57ac3(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #4 in feature [detachable] ET_MASTER_CLASS.reset_local_modified_classes */
extern T0* T570f57ac4(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #5 in feature [detachable] ET_MASTER_CLASS.reset_local_modified_classes */
extern T0* T570f57ac5(GE_context* ac);
/* [detachable] ET_MASTER_CLASS.reset_local_modified_class */
extern void T570f66(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_UNIX_FILE_SYSTEM.file_time_stamp */
extern T6 T126f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.file_time_stamp */
extern T6 T125f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.file_system */
extern T0* T570f23(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.unix_file_system */
extern T0* T570f31(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.windows_file_system */
extern T0* T570f30(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.operating_system */
extern T0* T570f29(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_after_parsed_and_errors */
extern void T74f285(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_after_parsed */
extern void T74f286(GE_context* ac, T0* C);
/* detachable ET_CONVERT_FEATURE_LIST.reset */
extern void T1096f10(GE_context* ac, T0* C);
/* detachable ET_CONVERT_PROCEDURE.reset */
extern void T1509f10(GE_context* ac, T0* C);
/* [detachable] ET_BRACED_TYPE_LIST.reset */
extern void T1388f12(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.reset */
extern void T1524f91(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.reset */
extern void T254f124(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.reset */
extern void T1957f26(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.reset */
extern void T1501f91(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.reset_constraint_creation_procedures */
extern void T1501f85(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_CREATOR.reset */
extern void T1081f14(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.reset_constraint_base_types */
extern void T1501f93(GE_context* ac, T0* C);
/* detachable ET_TYPE_CONSTRAINT_LIST.reset */
extern void T1281f18(GE_context* ac, T0* C);
/* detachable ET_TYPE_RENAME_CONSTRAINT.reset */
extern void T1502f11(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINT_RENAME_LIST.reset */
extern void T1084f14(GE_context* ac, T0* C);
/* [detachable] ET_RENAME.reset */
extern void T1337f7(GE_context* ac, T0* C);
/* detachable ET_ALIASED_FEATURE_NAME.reset */
extern void T1517f7(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME_LIST.reset */
extern void T911f13(GE_context* ac, T0* C);
/* detachable ET_ALIAS_FREE_NAME.reset */
extern void T1518f59(GE_context* ac, T0* C);
/* detachable ET_ALIAS_NAME.reset */
extern void T1030f112(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.reset */
extern void T891f88(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_type */
extern void T74f320(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.reset */
extern void T240f33(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.reset */
extern void T242f82(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.reset */
extern void T72f131(GE_context* ac, T0* C);
extern T1 T72f131ot1(T0* a1, T0** a2);
extern T1 T72f131ot2(T0* a1, T0** a2);
extern T1 T72f131ot3(T0* a1, T0** a2);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.reset */
extern void T577f21(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.reset */
extern void T1956f23(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.reset */
extern void T1526f84(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.reset */
extern void T1525f91(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.reset */
extern void T449f85(GE_context* ac, T0* C);
/* [detachable] ET_IDENTIFIER.reset */
extern void T184f127(GE_context* ac, T0* C);
/* detachable ET_CONVERT_FUNCTION.reset */
extern void T1508f9(GE_context* ac, T0* C);
/* detachable ET_CREATOR_LIST.reset */
extern void T1107f11(GE_context* ac, T0* C);
/* detachable ET_CREATOR.reset */
extern void T1104f19(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.reset */
extern void T1198f34(GE_context* ac, T0* C);
/* detachable ET_INVARIANTS.reset */
extern void T1198f34p1(GE_context* ac, T0* C);
/* detachable ET_INSPECT_EXPRESSION.reset */
extern void T1189f27(GE_context* ac, T0* C);
/* detachable ET_WHEN_EXPRESSION_LIST.reset */
extern void T1296f8(GE_context* ac, T0* C);
/* detachable ET_WHEN_EXPRESSION.reset */
extern void T1293f7(GE_context* ac, T0* C);
/* detachable ET_CHOICE_LIST.reset */
extern void T1058f14(GE_context* ac, T0* C);
/* detachable ET_CHOICE_RANGE.reset */
extern void T1534f9(GE_context* ac, T0* C);
/* detachable ET_IF_EXPRESSION.reset */
extern void T1167f30(GE_context* ac, T0* C);
/* detachable ET_ELSEIF_EXPRESSION_LIST.reset */
extern void T1116f8(GE_context* ac, T0* C);
/* detachable ET_ELSEIF_EXPRESSION.reset */
extern void T1113f8(GE_context* ac, T0* C);
/* detachable ET_CALL_AGENT.reset */
extern void T1041f34(GE_context* ac, T0* C);
extern T1 T1041f34ot1(T0* a1, T0** a2);
/* detachable ET_AGENT_ARGUMENT_OPERAND_LIST.reset */
extern void T1025f14(GE_context* ac, T0* C);
extern T1 T1025f14ot1(T0* a1, T0** a2);
/* [detachable] ET_CONVERT_TO_EXPRESSION.reset */
extern void T2048f25(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_FROM_EXPRESSION.reset */
extern void T2047f25(GE_context* ac, T0* C);
/* [detachable] ET_INFIX_CAST_EXPRESSION.reset */
extern void T1980f20(GE_context* ac, T0* C);
/* [detachable] ET_CONVERT_BUILTIN_EXPRESSION.reset */
extern void T2046f20(GE_context* ac, T0* C);
/* detachable ET_AGENT_TYPED_OPEN_ARGUMENT.reset */
extern void T1559f12(GE_context* ac, T0* C);
/* detachable ET_QUESTION_MARK_SYMBOL.reset */
extern void T1005f46(GE_context* ac, T0* C);
/* detachable ET_AGENT_OPEN_TARGET.reset */
extern void T1558f10(GE_context* ac, T0* C);
/* [detachable] ET_AGENT_IMPLICIT_CURRENT_TARGET.reset */
extern void T1811f7(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.reset */
extern void T1038f27(GE_context* ac, T0* C);
/* detachable ET_BRACKET_EXPRESSION.reset_arguments */
extern void T1038f30(GE_context* ac, T0* C);
extern T1 T1038f30ot1(T0* a1, T0** a2);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_ARGUMENT_LIST.reset */
extern void T1977f15(GE_context* ac, T0* C);
extern T1 T1977f15ot1(T0* a1, T0** a2);
/* [detachable] ET_ACTUAL_ARGUMENT_LIST.reset */
extern void T283f20(GE_context* ac, T0* C);
extern T1 T283f20ot1(T0* a1, T0** a2);
/* [detachable] ET_BRACKET_SYMBOL.reset */
extern void T622f135(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.reset */
extern void T1008f49(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.reset */
extern void T1008f49p1(GE_context* ac, T0* C);
/* detachable ET_ACROSS_EXPRESSION.reset_unfolded_form */
extern void T1008f50(GE_context* ac, T0* C);
/* detachable ET_VARIANT.reset */
extern void T1290f8(GE_context* ac, T0* C);
/* detachable ET_LOOP_INVARIANTS.reset */
extern void T1215f14(GE_context* ac, T0* C);
/* detachable ET_VOID.reset */
extern void T997f36(GE_context* ac, T0* C);
/* [detachable] ET_NAMED_OBJECT_TEST.reset */
extern void T736f30(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.reset */
extern void T635f31(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.reset */
extern void T635f31p1(GE_context* ac, T0* C);
/* [detachable] ET_UNQUALIFIED_CALL_EXPRESSION.reset_arguments */
extern void T635f34(GE_context* ac, T0* C);
extern T1 T635f34ot1(T0* a1, T0** a2);
/* [detachable] ET_REGULAR_INTEGER_CONSTANT.reset */
extern void T372f72(GE_context* ac, T0* C);
/* [detachable] ET_CREATE_EXPRESSION.reset */
extern void T328f33(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL.reset */
extern void T329f6(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL.reset_arguments */
extern void T329f7(GE_context* ac, T0* C);
extern T1 T329f7ot1(T0* a1, T0** a2);
/* [detachable] ET_EQUALITY_EXPRESSION.reset */
extern void T313f23(GE_context* ac, T0* C);
extern T1 T313f23ot1(T0* a1, T0** a2);
extern T1 T313f23ot2(T0* a1, T0** a2);
/* [detachable] ET_OBJECT_EQUALITY_EXPRESSION.reset */
extern void T312f25(GE_context* ac, T0* C);
extern T1 T312f25ot1(T0* a1, T0** a2);
extern T1 T312f25ot2(T0* a1, T0** a2);
/* [detachable] ET_OBJECT_TEST.reset */
extern void T311f27(GE_context* ac, T0* C);
/* [detachable] ET_RESULT.reset */
extern void T309f37(GE_context* ac, T0* C);
/* [detachable] ET_MANIFEST_TUPLE.reset */
extern void T288f35(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.reset */
extern void T287f33(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.reset */
extern void T287f33p1(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.reset */
extern void T287f33p0(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_EXPRESSION.reset_arguments */
extern void T287f34(GE_context* ac, T0* C);
extern T1 T287f34ot1(T0* a1, T0** a2);
/* [detachable] ET_CURRENT.reset */
extern void T282f38(GE_context* ac, T0* C);
/* detachable ET_TAGGED_ASSERTION.reset */
extern void T1562f9(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE_INLINE_AGENT.reset */
extern void T1557f49(GE_context* ac, T0* C);
extern T1 T1557f49ot1(T0* a1, T0** a2);
/* detachable ET_POSTCONDITIONS.reset */
extern void T1244f24(GE_context* ac, T0* C);
/* detachable ET_POSTCONDITIONS.reset */
extern void T1244f24p1(GE_context* ac, T0* C);
/* detachable ET_PRECONDITIONS.reset */
extern void T406f23(GE_context* ac, T0* C);
/* detachable ET_PRECONDITIONS.reset */
extern void T406f23p1(GE_context* ac, T0* C);
/* detachable ET_FORMAL_ARGUMENT_LIST.reset */
extern void T255f14(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_COMMA_ARGUMENT.reset */
extern void T1767f14(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_ARGUMENT.reset */
extern void T256f13(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.reset */
extern void T1556f57(GE_context* ac, T0* C);
/* detachable ET_COMPOUND.reset */
extern void T355f15(GE_context* ac, T0* C);
/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION.reset */
extern void T1756f9(GE_context* ac, T0* C);
/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION.reset */
extern void T1756f9p1(GE_context* ac, T0* C);
/* detachable ET_UNQUALIFIED_CALL_INSTRUCTION.reset_arguments */
extern void T1756f10(GE_context* ac, T0* C);
extern T1 T1756f10ot1(T0* a1, T0** a2);
/* detachable ET_REPEAT_INSTRUCTION.reset */
extern void T1264f26(GE_context* ac, T0* C);
/* detachable ET_REPEAT_INSTRUCTION.reset */
extern void T1264f26p1(GE_context* ac, T0* C);
/* detachable ET_REPEAT_INSTRUCTION.reset_unfolded_form */
extern void T1264f27(GE_context* ac, T0* C);
/* detachable ET_ACROSS_INSTRUCTION.reset */
extern void T1011f31(GE_context* ac, T0* C);
/* detachable ET_ACROSS_INSTRUCTION.reset */
extern void T1011f31p1(GE_context* ac, T0* C);
/* detachable ET_ACROSS_INSTRUCTION.reset_unfolded_form */
extern void T1011f32(GE_context* ac, T0* C);
/* detachable ET_CHECK_INSTRUCTION.reset */
extern void T1046f16(GE_context* ac, T0* C);
/* detachable ET_CHECK_INSTRUCTION.reset */
extern void T1046f16p1(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_INSTRUCTION.reset */
extern void T1537f13(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_INSTRUCTION.reset */
extern void T1537f13p1(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_INSTRUCTION.reset_arguments */
extern void T1537f14(GE_context* ac, T0* C);
extern T1 T1537f14ot1(T0* a1, T0** a2);
/* detachable ET_PRECURSOR_INSTRUCTION.reset */
extern void T1535f13(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_INSTRUCTION.reset_arguments */
extern void T1535f14(GE_context* ac, T0* C);
extern T1 T1535f14ot1(T0* a1, T0** a2);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.reset */
extern void T284f16(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.reset */
extern void T284f16p1(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.reset */
extern void T284f16p0(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_CALL_INSTRUCTION.reset_arguments */
extern void T284f17(GE_context* ac, T0* C);
extern T1 T284f17ot1(T0* a1, T0** a2);
/* detachable ET_DEBUG_INSTRUCTION.reset */
extern void T1110f8(GE_context* ac, T0* C);
/* detachable ET_INSPECT_INSTRUCTION.reset */
extern void T1192f11(GE_context* ac, T0* C);
/* detachable ET_WHEN_PART_LIST.reset */
extern void T1302f7(GE_context* ac, T0* C);
/* detachable ET_WHEN_PART.reset */
extern void T1299f5(GE_context* ac, T0* C);
/* detachable ET_IF_INSTRUCTION.reset */
extern void T1170f13(GE_context* ac, T0* C);
/* detachable ET_ELSEIF_PART_LIST.reset */
extern void T1122f7(GE_context* ac, T0* C);
/* detachable ET_ELSEIF_PART.reset */
extern void T1119f7(GE_context* ac, T0* C);
/* detachable ET_CREATE_INSTRUCTION.reset */
extern void T1533f16(GE_context* ac, T0* C);
/* detachable ET_BANG_INSTRUCTION.reset */
extern void T1532f14(GE_context* ac, T0* C);
/* detachable ET_LOOP_INSTRUCTION.reset */
extern void T1530f16(GE_context* ac, T0* C);
/* detachable ET_ASSIGNMENT_ATTEMPT.reset */
extern void T1529f8(GE_context* ac, T0* C);
/* detachable ET_ASSIGNMENT.reset */
extern void T1528f8(GE_context* ac, T0* C);
extern T1 T1528f8ot1(T0* a1, T0** a2);
/* detachable ET_ASSIGNER_INSTRUCTION.reset */
extern void T1527f17(GE_context* ac, T0* C);
extern T1 T1527f17ot1(T0* a1, T0** a2);
/* detachable ET_RETRY_INSTRUCTION.reset */
extern void T991f18(GE_context* ac, T0* C);
/* detachable ET_LOCAL_VARIABLE_LIST.reset */
extern void T1212f14(GE_context* ac, T0* C);
/* [detachable] ET_LOCAL_COMMA_VARIABLE.reset */
extern void T1768f14(GE_context* ac, T0* C);
/* detachable ET_LOCAL_VARIABLE.reset */
extern void T1206f13(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE_INLINE_AGENT.reset */
extern void T1556f57p1(GE_context* ac, T0* C);
extern T1 T1556f57p1ot1(T0* a1, T0** a2);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.reset */
extern void T1555f52(GE_context* ac, T0* C);
/* detachable ET_DO_PROCEDURE_INLINE_AGENT.reset */
extern void T1555f52p1(GE_context* ac, T0* C);
extern T1 T1555f52p1ot1(T0* a1, T0** a2);
/* detachable ET_EXTERNAL_FUNCTION_INLINE_AGENT.reset */
extern void T1554f50(GE_context* ac, T0* C);
extern T1 T1554f50ot1(T0* a1, T0** a2);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.reset */
extern void T1553f58(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION_INLINE_AGENT.reset */
extern void T1553f58p1(GE_context* ac, T0* C);
extern T1 T1553f58p1ot1(T0* a1, T0** a2);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.reset */
extern void T1552f53(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION_INLINE_AGENT.reset */
extern void T1552f53p1(GE_context* ac, T0* C);
extern T1 T1552f53p1ot1(T0* a1, T0** a2);
/* detachable ET_ITERATION_CURSOR.reset */
extern void T1551f25(GE_context* ac, T0* C);
/* detachable ET_ONCE_MANIFEST_STRING.reset */
extern void T1550f29(GE_context* ac, T0* C);
/* detachable ET_OLD_OBJECT_TEST.reset */
extern void T1549f32(GE_context* ac, T0* C);
/* detachable ET_OLD_EXPRESSION.reset */
extern void T1548f24(GE_context* ac, T0* C);
/* detachable ET_PREFIX_EXPRESSION.reset */
extern void T1547f28(GE_context* ac, T0* C);
/* detachable ET_FREE_OPERATOR.reset */
extern void T973f62(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_OPERATOR.reset */
extern void T981f63(GE_context* ac, T0* C);
/* detachable ET_SYMBOL_OPERATOR.reset */
extern void T994f156(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_TYPE.reset */
extern void T1546f25(GE_context* ac, T0* C);
/* detachable ET_INFIX_EXPRESSION.reset */
extern void T1543f28(GE_context* ac, T0* C);
extern T1 T1543f28ot1(T0* a1, T0** a2);
extern T1 T1543f28ot2(T0* a1, T0** a2);
extern T1 T1543f28ot3(T0* a1, T0** a2);
/* detachable ET_INFIX_OR_ELSE_OPERATOR.reset */
extern void T1545f44(GE_context* ac, T0* C);
/* detachable ET_INFIX_AND_THEN_OPERATOR.reset */
extern void T1544f44(GE_context* ac, T0* C);
/* detachable ET_EXPRESSION_ADDRESS.reset */
extern void T1542f24(GE_context* ac, T0* C);
/* detachable ET_RESULT_ADDRESS.reset */
extern void T1541f25(GE_context* ac, T0* C);
/* detachable ET_CURRENT_ADDRESS.reset */
extern void T1540f25(GE_context* ac, T0* C);
/* detachable ET_FEATURE_ADDRESS.reset */
extern void T1539f25(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.reset */
extern void T1538f29(GE_context* ac, T0* C);
/* detachable ET_PRECURSOR_EXPRESSION.reset_arguments */
extern void T1538f30(GE_context* ac, T0* C);
extern T1 T1538f30ot1(T0* a1, T0** a2);
/* detachable ET_CLASS_ASSERTION.reset */
extern void T1521f9(GE_context* ac, T0* C);
/* detachable ET_C2_CHARACTER_CONSTANT.reset */
extern void T1496f54(GE_context* ac, T0* C);
/* detachable ET_C1_CHARACTER_CONSTANT.reset */
extern void T1495f53(GE_context* ac, T0* C);
/* detachable ET_FALSE_CONSTANT.reset */
extern void T1494f45(GE_context* ac, T0* C);
/* detachable ET_TRUE_CONSTANT.reset */
extern void T1493f45(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_REAL_CONSTANT.reset */
extern void T1491f49(GE_context* ac, T0* C);
/* detachable ET_REGULAR_REAL_CONSTANT.reset */
extern void T1490f49(GE_context* ac, T0* C);
/* detachable ET_BINARY_INTEGER_CONSTANT.reset */
extern void T1489f70(GE_context* ac, T0* C);
/* detachable ET_OCTAL_INTEGER_CONSTANT.reset */
extern void T1488f70(GE_context* ac, T0* C);
/* detachable ET_HEXADECIMAL_INTEGER_CONSTANT.reset */
extern void T1487f70(GE_context* ac, T0* C);
/* detachable ET_UNDERSCORED_INTEGER_CONSTANT.reset */
extern void T1486f70(GE_context* ac, T0* C);
/* detachable ET_VERBATIM_STRING.reset */
extern void T1485f56(GE_context* ac, T0* C);
/* detachable ET_SPECIAL_MANIFEST_STRING.reset */
extern void T1484f52(GE_context* ac, T0* C);
/* detachable ET_REGULAR_MANIFEST_STRING.reset */
extern void T1483f52(GE_context* ac, T0* C);
/* detachable ET_C3_CHARACTER_CONSTANT.reset */
extern void T1482f53(GE_context* ac, T0* C);
/* detachable ET_STRIP_EXPRESSION.reset */
extern void T1270f33(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.reset */
extern void T1267f33(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.reset */
extern void T1267f33p1(GE_context* ac, T0* C);
/* detachable ET_STATIC_CALL_EXPRESSION.reset_arguments */
extern void T1267f34(GE_context* ac, T0* C);
extern T1 T1267f34ot1(T0* a1, T0** a2);
/* detachable ET_QUANTIFIER_EXPRESSION.reset */
extern void T1254f41(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.reset */
extern void T1254f41p1(GE_context* ac, T0* C);
/* detachable ET_QUANTIFIER_EXPRESSION.reset_unfolded_form */
extern void T1254f42(GE_context* ac, T0* C);
/* detachable ET_PARENTHESIZED_EXPRESSION.reset */
extern void T1232f25(GE_context* ac, T0* C);
/* detachable ET_MANIFEST_ARRAY.reset */
extern void T1218f33(GE_context* ac, T0* C);
extern T1 T1218f33ot1(T0* a1, T0** a2);
/* detachable ET_PARENT_CLAUSE_LIST.reset */
extern void T1237f8(GE_context* ac, T0* C);
/* [detachable] ET_PARENT_LIST.reset */
extern void T568f14(GE_context* ac, T0* C);
/* [detachable] ET_PARENT.reset */
extern void T567f12(GE_context* ac, T0* C);
/* detachable ET_KEYWORD_FEATURE_NAME_LIST.reset */
extern void T574f10(GE_context* ac, T0* C);
/* detachable ET_EXPORT_LIST.reset */
extern void T573f9(GE_context* ac, T0* C);
/* detachable ET_FEATURE_EXPORT.reset */
extern void T1145f13(GE_context* ac, T0* C);
/* detachable ET_ALL_EXPORT.reset */
extern void T1506f9(GE_context* ac, T0* C);
/* detachable ET_RENAME_LIST.reset */
extern void T572f10(GE_context* ac, T0* C);
/* detachable ET_FORMAL_PARAMETER_LIST.reset */
extern void T603f33(GE_context* ac, T0* C);
/* [detachable] ET_PROCEDURE_LIST.reset */
extern void T258f11(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.reset */
extern void T1516f94(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.reset_after_features_flattened */
extern void T1516f95(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.reset_signature_qualified_anchored_types */
extern void T1516f114(GE_context* ac, T0* C);
/* detachable ET_FORMAL_ARGUMENT_LIST.reset_qualified_anchored_types */
extern void T255f15(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_qualified_anchored_types */
extern void T74f321(GE_context* ac, T0* C);
/* detachable ET_FORMAL_PARAMETER_LIST.reset_qualified_anchored_types */
extern void T603f34(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.reset_qualified_anchored_types */
extern void T1501f92(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.reset_qualified_anchored_types */
extern void T891f89(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.reset_qualified_anchored_types */
extern void T1524f92(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.reset_qualified_anchored_types */
extern void T254f125(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_SUBLIST.reset_qualified_anchored_types */
extern void T1957f27(GE_context* ac, T0* C);
/* [detachable] ET_ACTUAL_PARAMETER_LIST.reset_qualified_anchored_types */
extern void T240f34(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.reset_qualified_anchored_types */
extern void T242f83(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.reset_qualified_anchored_types */
extern void T72f132(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_EMPTY_TUPLE_ACTUAL_PARAMETERS.reset_qualified_anchored_types */
extern void T577f22(GE_context* ac, T0* C);
/* [detachable] ET_UNFOLDED_TUPLE_ACTUAL_PARAMETERS.reset_qualified_anchored_types */
extern void T1956f24(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.reset_qualified_anchored_types */
extern void T1526f85(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.reset_qualified_anchored_types */
extern void T1525f92(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.reset_qualified_anchored_types */
extern void T449f86(GE_context* ac, T0* C);
/* detachable ET_EXTERNAL_PROCEDURE.reset_after_interface_checked */
extern void T1516f96(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.reset */
extern void T476f85(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.reset_after_features_flattened */
extern void T476f86(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.reset_signature_qualified_anchored_types */
extern void T476f104(GE_context* ac, T0* C);
/* [detachable] ET_DEFERRED_PROCEDURE.reset_after_interface_checked */
extern void T476f87(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.reset */
extern void T1515f88(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.reset_after_features_flattened */
extern void T1515f89(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.reset_signature_qualified_anchored_types */
extern void T1515f110(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.reset_after_interface_checked */
extern void T1515f90(GE_context* ac, T0* C);
/* detachable ET_ONCE_PROCEDURE.reset_after_interface_checked */
extern void T1515f90p1(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.reset */
extern void T469f86(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.reset_after_features_flattened */
extern void T469f87(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.reset_signature_qualified_anchored_types */
extern void T469f107(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.reset_after_interface_checked */
extern void T469f88(GE_context* ac, T0* C);
/* [detachable] ET_DO_PROCEDURE.reset_after_interface_checked */
extern void T469f88p1(GE_context* ac, T0* C);
/* [detachable] ET_QUERY_LIST.reset */
extern void T257f11(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.reset */
extern void T253f96(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.reset */
extern void T253f96p1(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.reset_after_features_flattened */
extern void T253f97(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.reset_signature_qualified_anchored_types */
extern void T253f117(GE_context* ac, T0* C);
/* [detachable] ET_EXTERNAL_FUNCTION.reset_after_interface_checked */
extern void T253f98(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.reset */
extern void T1513f87(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.reset */
extern void T1513f87p1(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.reset_after_features_flattened */
extern void T1513f88(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.reset_signature_qualified_anchored_types */
extern void T1513f107(GE_context* ac, T0* C);
/* detachable ET_DEFERRED_FUNCTION.reset_after_interface_checked */
extern void T1513f89(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.reset */
extern void T1512f90(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.reset */
extern void T1512f90p1(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.reset_after_features_flattened */
extern void T1512f91(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.reset_signature_qualified_anchored_types */
extern void T1512f113(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.reset_after_interface_checked */
extern void T1512f92(GE_context* ac, T0* C);
/* detachable ET_ONCE_FUNCTION.reset_after_interface_checked */
extern void T1512f92p1(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.reset */
extern void T1511f88(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.reset */
extern void T1511f88p1(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.reset_after_features_flattened */
extern void T1511f89(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.reset_signature_qualified_anchored_types */
extern void T1511f110(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.reset_after_interface_checked */
extern void T1511f90(GE_context* ac, T0* C);
/* detachable ET_DO_FUNCTION.reset_after_interface_checked */
extern void T1511f90p1(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.reset */
extern void T371f88(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.reset */
extern void T371f88p1(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.reset_after_features_flattened */
extern void T371f89(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.reset_signature_qualified_anchored_types */
extern void T371f105(GE_context* ac, T0* C);
/* [detachable] ET_UNIQUE_ATTRIBUTE.reset_after_interface_checked */
extern void T371f90(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.reset */
extern void T369f87(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.reset */
extern void T369f87p1(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.reset_after_features_flattened */
extern void T369f88(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.reset_signature_qualified_anchored_types */
extern void T369f102(GE_context* ac, T0* C);
/* [detachable] ET_CONSTANT_ATTRIBUTE.reset_after_interface_checked */
extern void T369f89(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.reset */
extern void T359f89(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.reset */
extern void T359f89p1(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.reset_after_features_flattened */
extern void T359f90(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.reset_signature_qualified_anchored_types */
extern void T359f110(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.reset_after_interface_checked */
extern void T359f91(GE_context* ac, T0* C);
/* [detachable] ET_EXTENDED_ATTRIBUTE.reset_after_interface_checked */
extern void T359f91p1(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.reset */
extern void T360f84(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.reset */
extern void T360f84p1(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.reset_after_features_flattened */
extern void T360f85(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.reset_signature_qualified_anchored_types */
extern void T360f98(GE_context* ac, T0* C);
/* [detachable] ET_ATTRIBUTE.reset_after_interface_checked */
extern void T360f86(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.local_ignored_classes_do_if */
extern void T570f70(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].do_if */
extern void T462f24(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_MASTER_CLASS.local_ignored_classes_do_all */
extern void T570f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].do_all */
extern void T462f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.local_non_override_classes_do_all */
extern void T570f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.local_override_classes_do_all */
extern void T570f67(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.process */
extern void T802f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_AST_NULL_PROCESSOR.process_cluster */
extern void T464f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.process_cluster */
extern void T454f184(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.preparse_cluster */
extern void T454f185(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_EIFFEL_PREPARSER.preparse_cluster */
extern T0* T454f185ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_EIFFEL_PREPARSER.preparse_cluster */
extern T0* T454f185ac2(GE_context* ac, T0* a1);
/* [detachable] ET_CLASS.is_in_group */
extern T1 T74f252(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.preparse_clusters */
extern void T454f189(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].remove */
extern void T1579f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_CLUSTER].keep_head */
extern void T1859f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_CLUSTER].set_count */
extern void T1859f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].move_left */
extern void T1579f18(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].move_cursors_left */
extern void T1579f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_CLUSTER].set_position */
extern void T1861f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].remove_last */
extern void T1579f16(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].move_last_cursors_after */
extern void T1579f19(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_CLUSTER].set_next_cursor */
extern void T1861f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_CLUSTER].set_after */
extern void T1861f7(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.build_dependant_constraint */
extern void T454f188(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_dependant_constraint */
extern void T802f97(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTERS.set_dependant_constraint */
extern void T787f14(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.make */
extern T0* T1758c17(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].set_key_equality_tester */
extern void T2040f35(GE_context* ac, T0* C, T0* a1);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.string_equality_tester */
extern T0* T1758f11(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].make_map */
extern T0* T2040c34(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].make_with_equality_testers */
extern void T2040f37(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].make_sparse_container */
extern void T2040f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].set_internal_cursor */
extern void T2040f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].unset_found_item */
extern void T2040f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].make_slots */
extern void T2040f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].new_modulus */
extern T6 T2040f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].make_clashes */
extern void T2040f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].make_key_storage */
extern void T2040f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].make_item_storage */
extern void T2040f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_WILDCARD].make */
extern T0* T2082f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_WILDCARD].make_empty */
extern T0* T2079c4(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_WILDCARD].default_create */
extern T0* T2082c2(GE_context* ac);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.make_with_pathnames */
extern T0* T1758c16(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].put_last */
extern void T2040f36(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].key_storage_put */
extern void T2040f44(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].slots_put */
extern void T2040f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].clashes_put */
extern void T2040f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].slots_item */
extern T6 T2040f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].compress */
extern void T2040f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].hash_position */
extern T6 T2040f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].slots_wipe_out */
extern void T2040f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].clashes_wipe_out */
extern void T2040f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].key_storage_keep_head */
extern void T2040f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].item_storage_keep_head */
extern void T2040f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_WILDCARD].keep_head */
extern void T2079f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_WILDCARD].set_count */
extern void T2079f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].move_all_cursors */
extern void T2040f46(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].item_storage_put */
extern void T2040f40(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_WILDCARD].force */
extern void T2082f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_WILDCARD].fill_with */
extern void T2079f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_WILDCARD].extend */
extern void T2079f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_WILDCARD, [attached] STRING_8].search_position */
extern void T2040f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_DFA_WILDCARD.compile */
extern void T2068f20(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] LX_DFA_WILDCARD.compile */
extern T0* T2068c20(GE_context* ac, T0* a1, T1 a2);
/* [detachable] LX_FULL_DFA.make */
extern T0* T2097c38(GE_context* ac, T0* a1);
/* [detachable] LX_FULL_DFA.build */
extern void T2097f41(GE_context* ac, T0* C);
/* [detachable] LX_FULL_DFA.build_accept_table */
extern void T2097f47(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].first */
extern T0* T2105f6(GE_context* ac, T0* C);
/* [detachable] LX_FULL_DFA.build_nxt_table */
extern void T2097f46(GE_context* ac, T0* C);
/* [detachable] LX_TRANSITION_TABLE [[attached] LX_DFA_STATE].target */
extern T0* T2144f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].value */
extern T0* T2194f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].item_storage_item */
extern T0* T2194f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].position_of_key */
extern T6 T2194f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].clashes_item */
extern T6 T2194f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].key_storage_item */
extern T6 T2194f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].hash_position */
extern T6 T2194f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].slots_item */
extern T6 T2194f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_DFA_STATE.is_accepting */
extern T1 T2136f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].is_empty */
extern T1 T2105f11(GE_context* ac, T0* C);
/* [detachable] LX_FULL_DFA.build_transitions */
extern void T2097f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_TRANSITION_TABLE [[attached] LX_DFA_STATE].set_target */
extern void T2144f9(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].force_new */
extern void T2194f46(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].key_storage_put */
extern void T2194f52(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].item_storage_put */
extern void T2194f48(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_DFA_STATE].force */
extern void T2191f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_DFA_STATE].fill_with */
extern void T2190f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_DFA_STATE].extend */
extern void T2190f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].slots_put */
extern void T2194f51(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].clashes_put */
extern void T2194f50(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].resize */
extern void T2194f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].clashes_resize */
extern void T2194f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].special_integer_ */
extern T0* T2194f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].key_storage_resize */
extern void T2194f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].item_storage_resize */
extern void T2194f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_DFA_STATE].aliased_resized_area */
extern T0* T2191f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] LX_DFA_STATE].aliased_resized_area */
extern T0* T2190f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].slots_resize */
extern void T2194f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].new_modulus */
extern T6 T2194f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].new_capacity */
extern T6 T2194f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].unset_found_item */
extern void T2194f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].replace_found_item */
extern void T2194f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].found */
extern T1 T2194f20(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].search */
extern void T2194f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].search_position */
extern void T2194f47(GE_context* ac, T0* C, T6 a1);
/* detachable LX_SYMBOL_PARTITIONS.previous_symbol */
extern T6 T2139f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [[attached] DS_BILINKABLE [[attached] INTEGER_32]].item */
extern T0* T2178f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_FULL_DFA.new_state */
extern T0* T2097f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_DFA_STATE.set_id */
extern void T2136f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_DFA_STATE].put_last */
extern void T2138f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_DFA_STATE.is_equal */
extern T1 T2136f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].is_equal */
extern T1 T2137f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].any_ */
extern T0* T2137f17(GE_context* ac, T0* C);
/* [detachable] LX_DFA_STATE.any_ */
extern T0* T2136f14(GE_context* ac, T0* C);
/* [detachable] LX_DFA_STATE.new_state */
extern T0* T2136f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_DFA_STATE.make */
extern T0* T2136c16(GE_context* ac, T0* a1, T6 a2, T6 a3);
extern T1 T2136f16ot1(T0* a1, T0** a2);
/* detachable LX_RULE.set_useful */
extern void T2122f22(GE_context* ac, T0* C, T1 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].sort */
extern void T2105f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BUBBLE_SORTER [[attached] LX_RULE].sort */
extern void T2183f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BUBBLE_SORTER [[attached] LX_RULE].sort_with_comparator */
extern void T2183f4(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_BUBBLE_SORTER [[attached] LX_RULE].subsort_with_comparator */
extern void T2183f5(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].replace */
extern void T2105f22(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_RULE].less_than */
extern T1 T2201f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_RULE].detachable_less_than */
extern T1 T2201f2(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_RULE].attached_less_than */
extern T1 T2201f3(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable LX_RULE.is_less */
extern T1 T2122f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].item */
extern T0* T2105f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_DFA_STATE.rule_sorter */
extern T0* T2136f7(GE_context* ac, T0* C);
/* [detachable] DS_BUBBLE_SORTER [[attached] LX_RULE].make */
extern T0* T2183c2(GE_context* ac, T0* a1);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_RULE].make */
extern T0* T2201c4(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].sort */
extern void T2137f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BUBBLE_SORTER [[attached] LX_NFA_STATE].sort */
extern void T2181f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BUBBLE_SORTER [[attached] LX_NFA_STATE].sort_with_comparator */
extern void T2181f4(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_BUBBLE_SORTER [[attached] LX_NFA_STATE].subsort_with_comparator */
extern void T2181f5(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].replace */
extern void T2137f22(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_NFA_STATE].less_than */
extern T1 T2204f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_NFA_STATE].detachable_less_than */
extern T1 T2204f2(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_NFA_STATE].attached_less_than */
extern T1 T2204f3(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] LX_NFA_STATE.is_less */
extern T1 T2147f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_DFA_STATE.bubble_sorter */
extern T0* T2136f8(GE_context* ac, T0* C);
/* [detachable] DS_BUBBLE_SORTER [[attached] LX_NFA_STATE].make */
extern T0* T2181c2(GE_context* ac, T0* a1);
/* [detachable] KL_COMPARABLE_COMPARATOR [[attached] LX_NFA_STATE].make */
extern T0* T2204c4(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].last */
extern T0* T2137f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].is_empty */
extern T1 T2137f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].force_last */
extern void T2105f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_RULE].force */
extern void T2155f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_RULE].fill_with */
extern void T2135f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_RULE].extend */
extern void T2135f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].resize */
extern void T2105f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_RULE].aliased_resized_area */
extern T0* T2155f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] LX_RULE].aliased_resized_area */
extern T0* T2135f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].new_capacity */
extern T6 T2105f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].extendible */
extern T1 T2105f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_NFA_STATE.is_accepting_head */
extern T1 T2147f6(GE_context* ac, T0* C);
/* [detachable] LX_NFA_STATE.has_transition */
extern T1 T2147f8(GE_context* ac, T0* C);
/* [detachable] LX_NFA_STATE.is_accepting */
extern T1 T2147f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].has */
extern T1 T2137f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].force_last */
extern void T2137f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA_STATE].force */
extern void T2187f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_NFA_STATE].fill_with */
extern void T2185f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_NFA_STATE].extend */
extern void T2185f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].resize */
extern void T2137f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA_STATE].aliased_resized_area */
extern T0* T2187f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] LX_NFA_STATE].aliased_resized_area */
extern T0* T2185f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].new_capacity */
extern T6 T2137f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].extendible */
extern T1 T2137f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].make */
extern T0* T2105c12(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].set_internal_cursor */
extern void T2105f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].new_cursor */
extern T0* T2105f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_RULE].make */
extern T0* T2156c4(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_RULE].make */
extern T0* T2155f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_RULE].make_empty */
extern void T2135f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_RULE].make_empty */
extern T0* T2135c7(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_RULE].default_create */
extern T0* T2155c3(GE_context* ac);
/* [detachable] LX_TRANSITION_TABLE [[attached] LX_DFA_STATE].make */
extern T0* T2144c8(GE_context* ac, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].make_map */
extern T0* T2194c35(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].make_with_equality_testers */
extern void T2194f36(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].make_sparse_container */
extern void T2194f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].set_internal_cursor */
extern void T2194f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].new_cursor */
extern T0* T2194f27(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_DFA_STATE, [attached] INTEGER_32].make */
extern T0* T2213c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].make_slots */
extern void T2194f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].make_clashes */
extern void T2194f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].make_key_storage */
extern void T2194f39(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_DFA_STATE, [attached] INTEGER_32].make_item_storage */
extern void T2194f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_DFA_STATE].make */
extern T0* T2191f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_DFA_STATE].make_empty */
extern T0* T2190c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_DFA_STATE].default_create */
extern T0* T2191c3(GE_context* ac);
/* [detachable] LX_DFA_STATE.maximum_symbol */
extern T6 T2136f13(GE_context* ac, T0* C);
/* [detachable] LX_DFA_STATE.minimum_symbol */
extern T6 T2136f12(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].put_last */
extern void T2137f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_SYMBOL_TRANSITION [[attached] LX_NFA_STATE].target */
extern T0* T2175f2(GE_context* ac, T0* C);
/* [detachable] LX_SYMBOL_CLASS_TRANSITION [[attached] LX_NFA_STATE].target */
extern T0* T2174f2(GE_context* ac, T0* C);
/* [detachable] LX_EPSILON_TRANSITION [[attached] LX_NFA_STATE].target */
extern T0* T2171f1(GE_context* ac, T0* C);
/* [detachable] LX_SYMBOL_TRANSITION [[attached] LX_NFA_STATE].labeled */
extern T1 T2175f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_SYMBOL_CLASS_TRANSITION [[attached] LX_NFA_STATE].labeled */
extern T1 T2174f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_SYMBOL_CLASS.has */
extern T1 T2109f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_SYMBOL_CLASS.added */
extern T1 T2109f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_SYMBOL_CLASS.index_in_chunk */
extern T6 T2109f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_SYMBOL_CLASS.chunk */
extern T0* T2109f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_SYMBOL_CLASS.chunk_index */
extern T6 T2109f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_SYMBOL_CLASS.masks */
extern T0* T2109f10(GE_context* ac, T0* C);
/* [detachable] LX_EPSILON_TRANSITION [[attached] LX_NFA_STATE].labeled */
extern T1 T2171f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].item */
extern T0* T2137f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].make */
extern T0* T2137c19(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].set_internal_cursor */
extern void T2137f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].new_cursor */
extern T0* T2137f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE].make */
extern T0* T2188c9(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA_STATE].make */
extern T0* T2187f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_NFA_STATE].make_empty */
extern T0* T2185c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA_STATE].default_create */
extern T0* T2187c3(GE_context* ac);
/* detachable LX_SYMBOL_PARTITIONS.is_representative */
extern T1 T2139f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_DFA_STATE.partition */
extern void T2136f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_SYMBOL_TRANSITION [[attached] LX_NFA_STATE].record */
extern void T2175f5(GE_context* ac, T0* C, T0* a1);
/* detachable LX_SYMBOL_PARTITIONS.put */
extern void T2139f13(GE_context* ac, T0* C, T6 a1);
/* detachable LX_SYMBOL_PARTITIONS.put */
extern void T2139f13p1(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_BILINKABLE [[attached] INTEGER_32].forget_left */
extern void T2179f7(GE_context* ac, T0* C);
/* [detachable] DS_BILINKABLE [[attached] INTEGER_32].forget_right */
extern void T2179f6(GE_context* ac, T0* C);
/* [detachable] DS_BILINKABLE [[attached] INTEGER_32].put_right */
extern void T2179f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKABLE [[attached] INTEGER_32].attach_left */
extern void T2179f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].force */
extern void T745f87(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_SYMBOL_CLASS_TRANSITION [[attached] LX_NFA_STATE].record */
extern void T2174f5(GE_context* ac, T0* C, T0* a1);
/* detachable LX_SYMBOL_PARTITIONS.add */
extern void T2139f12(GE_context* ac, T0* C, T0* a1);
/* detachable LX_SYMBOL_PARTITIONS.add */
extern void T2139f12p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_EPSILON_TRANSITION [[attached] LX_NFA_STATE].record */
extern void T2171f4(GE_context* ac, T0* C, T0* a1);
/* detachable LX_SYMBOL_PARTITIONS.initialize */
extern void T2139f11(GE_context* ac, T0* C);
/* detachable LX_SYMBOL_PARTITIONS.initialize */
extern void T2139f11p1(GE_context* ac, T0* C);
/* detachable LX_SYMBOL_PARTITIONS.new_lower */
extern T6 T2139f9(GE_context* ac, T0* C);
/* [detachable] DS_BILINKABLE [[attached] INTEGER_32].put_left */
extern void T2179f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKABLE [[attached] INTEGER_32].attach_right */
extern void T2179f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKABLE [[attached] INTEGER_32].put */
extern void T2179f8(GE_context* ac, T0* C, T6 a1);
/* detachable LX_SYMBOL_PARTITIONS.lower */
extern T6 T2139f7(GE_context* ac, T0* C);
/* detachable LX_SYMBOL_PARTITIONS.upper */
extern T6 T2139f6(GE_context* ac, T0* C);
/* [detachable] LX_FULL_DFA.resize */
extern void T2097f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_DFA_STATE].resize */
extern void T2138f10(GE_context* ac, T0* C, T6 a1);
/* detachable LX_SYMBOL_PARTITIONS.capacity */
extern T6 T2139f5(GE_context* ac, T0* C);
/* [detachable] ARRAY [[attached] DS_BILINKABLE [[attached] INTEGER_32]].count */
extern T6 T2178f5(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_DFA_STATE].item */
extern T0* T2138f7(GE_context* ac, T0* C, T6 a1);
/* detachable LX_SYMBOL_PARTITIONS.make */
extern T0* T2139c10(GE_context* ac, T6 a1, T6 a2);
/* detachable LX_SYMBOL_PARTITIONS.make */
extern void T2139f10p1(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ARRAY [[attached] DS_BILINKABLE [[attached] INTEGER_32]].put */
extern void T2178f7(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ARRAY [[attached] DS_BILINKABLE [[attached] INTEGER_32]].make_filled */
extern T0* T2178c6(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] ARRAY [[attached] DS_BILINKABLE [[attached] INTEGER_32]].make_filled_area */
extern void T2178f8(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_BILINKABLE [[attached] INTEGER_32]].make_filled */
extern T0* T2180c4(GE_context* ac, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_BILINKABLE [[attached] INTEGER_32]].fill_with */
extern void T2180f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_BILINKABLE [[attached] INTEGER_32]].extend */
extern void T2180f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [[attached] DS_BILINKABLE [[attached] INTEGER_32]].make_empty */
extern void T2180f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_BILINKABLE [[attached] INTEGER_32].make */
extern T0* T2179c4(GE_context* ac, T6 a1);
/* [detachable] LX_FULL_DFA.put_eob_state */
extern void T2097f40(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].force_first */
extern void T2105f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].put */
extern void T2105f18(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].move_cursors_right */
extern void T2105f21(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_RULE].set_position */
extern void T2156f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].move_right */
extern void T2105f20(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_RULE].put_last */
extern void T2105f19(GE_context* ac, T0* C, T0* a1);
/* detachable LX_RULE.make_default */
extern T0* T2122c14(GE_context* ac, T6 a1);
/* detachable LX_RULE.dummy_action */
extern T0* T2122f11(GE_context* ac, T0* C);
/* [detachable] LX_ACTION.make */
extern T0* T2177c2(GE_context* ac, T0* a1);
/* detachable LX_RULE.dummy_pattern */
extern T0* T2122f12(GE_context* ac, T0* C);
/* [detachable] LX_NFA.make_epsilon */
extern T0* T2120c6(GE_context* ac, T1 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].put_first */
extern void T2137f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].put */
extern void T2137f27(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].move_cursors_right */
extern void T2137f31(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE].set_position */
extern void T2188f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].move_right */
extern void T2137f30(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] LX_NFA_STATE.set_transition */
extern void T2147f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_EPSILON_TRANSITION [[attached] LX_NFA_STATE].make */
extern T0* T2171c3(GE_context* ac, T0* a1);
/* [detachable] LX_NFA_STATE.make */
extern T0* T2147c10(GE_context* ac, T1 a1);
/* [detachable] LX_FULL_DFA.initialize */
extern void T2097f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_FULL_DFA.initialize_dfa */
extern void T2097f44(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] LX_FULL_DFA.put_start_condition */
extern void T2097f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_NFA.start_state */
extern T0* T2120f4(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].first */
extern T0* T2137f14(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA].item */
extern T0* T2145f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_START_CONDITIONS.item */
extern T0* T2106f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_FULL_DFA.set_nfa_state_ids */
extern void T2097f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_NFA_STATE.set_id */
extern void T2147f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].force */
extern void T2146f35(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].key_storage_put */
extern void T2146f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].slots_put */
extern void T2146f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].clashes_put */
extern void T2146f41(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].slots_item */
extern T6 T2146f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].clashes_item */
extern T6 T2146f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].hash_position */
extern T6 T2146f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].resize */
extern void T2146f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].clashes_resize */
extern void T2146f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].special_integer_ */
extern T0* T2146f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].key_storage_resize */
extern void T2146f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].item_storage_resize */
extern void T2146f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA].aliased_resized_area */
extern T0* T2119f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] LX_NFA].aliased_resized_area */
extern T0* T2121f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].key_storage_item */
extern T6 T2146f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].slots_resize */
extern void T2146f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].new_modulus */
extern T6 T2146f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].new_capacity */
extern T6 T2146f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].item_storage_put */
extern void T2146f39(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA].force */
extern void T2119f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_NFA].fill_with */
extern void T2121f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_NFA].extend */
extern void T2121f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].search_position */
extern void T2146f36(GE_context* ac, T0* C, T6 a1);
/* detachable KL_EQUALITY_TESTER [[attached] INTEGER_32].test */
extern T1 T653f1(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].unset_found_item */
extern void T2146f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].item */
extern T0* T2146f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].item_storage_item */
extern T0* T2146f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].has */
extern T1 T2146f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].make */
extern T0* T2146c34(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].make_with_equality_testers */
extern void T2146f37(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].make_sparse_container */
extern void T2146f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].set_internal_cursor */
extern void T2146f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].new_cursor */
extern T0* T2146f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_NFA, [attached] INTEGER_32].make */
extern T0* T2197c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].make_slots */
extern void T2146f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].make_clashes */
extern void T2146f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].make_key_storage */
extern void T2146f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_NFA, [attached] INTEGER_32].make_item_storage */
extern void T2146f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA].make */
extern T0* T2119f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_NFA].make_empty */
extern T0* T2121c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_NFA].default_create */
extern T0* T2119c3(GE_context* ac);
/* detachable KL_EQUALITY_TESTER [[attached] INTEGER_32].default_create */
extern T0* T653c2(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_DFA_STATE].make */
extern T0* T2138c8(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_DFA_STATE].set_internal_cursor */
extern void T2138f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_DFA_STATE].new_cursor */
extern T0* T2138f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_DFA_STATE].make */
extern T0* T2192c3(GE_context* ac, T0* a1);
/* [detachable] LX_EQUIVALENCE_CLASSES.to_array */
extern T0* T2124f6(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] LX_EQUIVALENCE_CLASSES.lower */
extern T6 T2124f5(GE_context* ac, T0* C);
/* [detachable] LX_EQUIVALENCE_CLASSES.upper */
extern T6 T2124f4(GE_context* ac, T0* C);
/* [detachable] LX_EQUIVALENCE_CLASSES.built */
extern T1 T2124f3(GE_context* ac, T0* C);
/* [detachable] LX_EQUIVALENCE_CLASSES.new_lower */
extern T6 T2124f7(GE_context* ac, T0* C);
/* [detachable] LX_FULL_DFA.build_eof_rules */
extern void T2097f43(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] ARRAY [detachable LX_RULE].put */
extern void T2141f5(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ARRAY [detachable LX_RULE].make_filled */
extern T0* T2141c4(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] ARRAY [detachable LX_RULE].make_filled_area */
extern void T2141f6(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable LX_RULE].make_filled */
extern T0* T2142c4(GE_context* ac, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable LX_RULE].fill_with */
extern void T2142f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable LX_RULE].extend */
extern void T2142f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [detachable LX_RULE].make_empty */
extern void T2142f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_FULL_DFA.build_rules */
extern void T2097f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ARRAY [[attached] LX_RULE].put */
extern void T2134f5(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ARRAY [[attached] LX_RULE].make_filled */
extern T0* T2134c4(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] ARRAY [[attached] LX_RULE].make_filled_area */
extern void T2134f6(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] LX_RULE].make_filled */
extern T0* T2135c5(GE_context* ac, T0* a1, T6 a2);
/* [detachable] KL_ARRAY_ROUTINES [[attached] LX_RULE].make_empty_with_lower */
extern T0* T2140f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [[attached] LX_RULE].make_from_array */
extern T0* T2134c8(GE_context* ac, T0* a1);
/* [detachable] ARRAY [[attached] LX_RULE].set_area */
extern void T2134f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_ARRAY [[attached] LX_RULE].new_rebase */
extern void T2193f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_ARRAY [[attached] LX_RULE].make_empty */
extern T0* T2193c4(GE_context* ac);
/* [detachable] KL_ARRAY [[attached] LX_RULE].make_empty_area */
extern void T2193f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [[attached] LX_RULE].make_empty */
extern T0* T2134c7(GE_context* ac);
/* [detachable] ARRAY [[attached] LX_RULE].make_empty_area */
extern void T2134f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_ARRAY_ROUTINES [[attached] LX_RULE].default_create */
extern T0* T2140c2(GE_context* ac);
/* [detachable] LX_START_CONDITIONS.names */
extern T0* T2106f7(GE_context* ac, T0* C);
/* [detachable] LX_START_CONDITIONS.first */
extern T0* T2106f6(GE_context* ac, T0* C);
/* [detachable] KL_ARRAY_ROUTINES [[attached] STRING_8].make_empty_with_lower */
extern T0* T2157f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [[attached] STRING_8].make_from_array */
extern T0* T33c11(GE_context* ac, T0* a1);
/* [detachable] ARRAY [[attached] STRING_8].set_area */
extern void T33f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_ARRAY [[attached] STRING_8].new_rebase */
extern void T2211f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_ARRAY [[attached] STRING_8].make_empty */
extern T0* T2211c4(GE_context* ac);
/* [detachable] KL_ARRAY [[attached] STRING_8].make_empty_area */
extern void T2211f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ARRAY [[attached] STRING_8].make_empty */
extern T0* T33c10(GE_context* ac);
/* [detachable] ARRAY [[attached] STRING_8].make_empty_area */
extern void T33f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_START_CONDITIONS.string_array_routines */
extern T0* T2106f11(GE_context* ac, T0* C);
/* [detachable] KL_ARRAY_ROUTINES [[attached] STRING_8].default_create */
extern T0* T2157c2(GE_context* ac);
/* [detachable] LX_WILDCARD_PARSER.parse_string */
extern void T2095f247(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.parse */
extern void T2095f251(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_pop_last_value */
extern void T2095f264(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.yy_push_error_value */
extern void T2095f263(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_do_action */
extern void T2095f262(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_SYMBOL_CLASS.twin */
extern T0* T2109f21(GE_context* ac, T0* C);
/* [detachable] LX_SYMBOL_CLASS.copy */
extern void T2109f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [[attached] NATURAL_64].twin */
extern T0* T767f7(GE_context* ac, T0* C);
/* [detachable] LX_SYMBOL_CLASS.chunk_of_ones */
extern T0* T2109f20(GE_context* ac, T0* C);
/* detachable SPECIAL [detachable SPECIAL [[attached] NATURAL_64]].make_filled */
extern T0* T2166c5(GE_context* ac, T0* a1, T6 a2);
/* detachable SPECIAL [detachable SPECIAL [[attached] NATURAL_64]].fill_with */
extern void T2166f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* detachable SPECIAL [detachable SPECIAL [[attached] NATURAL_64]].extend */
extern void T2166f9(GE_context* ac, T0* C, T0* a1);
/* detachable SPECIAL [detachable SPECIAL [[attached] NATURAL_64]].make_empty */
extern void T2166f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_SYMBOL_CLASS.remove_symbol_class */
extern void T2109f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_SYMBOL_CLASS.add_symbol_class */
extern void T2109f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [[attached] NATURAL_64].filled_with */
extern T1 T767f6(GE_context* ac, T0* C, T11 a1, T6 a2, T6 a3);
/* [detachable] LX_SYMBOL_CLASS.attached_other_sets */
extern T0* T2109f19(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.append_character_set_to_character_class */
extern T0* T2095f188(GE_context* ac, T0* C, T6 a1, T6 a2, T0* a3);
/* [detachable] LX_SYMBOL_CLASS.add_symbol */
extern void T2109f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_SYMBOL_CLASS.attached_chunk */
extern T0* T2109f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.report_negative_range_in_character_class_error */
extern void T2095f304(GE_context* ac, T0* C);
/* [detachable] LX_NEGATIVE_RANGE_IN_CHARACTER_CLASS_ERROR.make */
extern T0* T2154c8(GE_context* ac, T0* a1, T6 a2);
/* [detachable] LX_WILDCARD_PARSER.filename */
extern T0* T2095f209(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_SYMBOL_CLASS].aliased_resized_area */
extern T0* T2117f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] LX_SYMBOL_CLASS].aliased_resized_area */
extern T0* T2118f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.append_character_to_character_class */
extern T0* T2095f187(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] LX_WILDCARD_PARSER.new_character_class */
extern T0* T2095f186(GE_context* ac, T0* C);
/* [detachable] LX_SYMBOL_CLASS.make */
extern T0* T2109c25(GE_context* ac, T6 a1, T6 a2);
/* [detachable] LX_SYMBOL_CLASS.set_negated */
extern void T2109f27(GE_context* ac, T0* C, T1 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].force */
extern void T2112f37(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].key_storage_put */
extern void T2112f47(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_32].force */
extern void T2115f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] STRING_32].fill_with */
extern void T2116f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] STRING_32].extend */
extern void T2116f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].slots_put */
extern void T2112f46(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].clashes_put */
extern void T2112f45(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].slots_item */
extern T6 T2112f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].clashes_item */
extern T6 T2112f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].hash_position */
extern T6 T2112f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_32.hash_code */
extern T6 T18f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].resize */
extern void T2112f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].clashes_resize */
extern void T2112f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].special_integer_ */
extern T0* T2112f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].key_storage_resize */
extern void T2112f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_32].aliased_resized_area */
extern T0* T2115f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] STRING_32].aliased_resized_area */
extern T0* T2116f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].item_storage_resize */
extern void T2112f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].key_storage_item */
extern T0* T2112f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].slots_resize */
extern void T2112f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].new_modulus */
extern T6 T2112f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].new_capacity */
extern T6 T2112f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].item_storage_put */
extern void T2112f43(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].search_position */
extern void T2112f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_EQUALITY_TESTER [[attached] STRING_32].test */
extern T1 T2113f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] STRING_32.is_equal */
extern T1 T18f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [CHARACTER_32].same_items */
extern T1 T16f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].unset_found_item */
extern void T2112f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].force_new */
extern void T2111f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].item_storage_put */
extern void T2111f50(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].slots_put */
extern void T2111f49(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].clashes_put */
extern void T2111f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].slots_item */
extern T6 T2111f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].hash_position */
extern T6 T2111f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_SYMBOL_CLASS.hash_code */
extern T6 T2109f22(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].clashes_item */
extern T6 T2111f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].resize */
extern void T2111f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].clashes_resize */
extern void T2111f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].special_integer_ */
extern T0* T2111f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].key_storage_resize */
extern void T2111f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].item_storage_resize */
extern void T2111f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].key_storage_item */
extern T0* T2111f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].item_storage_item */
extern T0* T2111f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].slots_resize */
extern void T2111f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].new_modulus */
extern T6 T2111f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].new_capacity */
extern T6 T2111f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].unset_found_item */
extern void T2111f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].found_item */
extern T0* T2111f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].found */
extern T1 T2111f26(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].search */
extern void T2111f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].search_position */
extern void T2111f46(GE_context* ac, T0* C, T0* a1);
/* detachable KL_EQUALITY_TESTER [[attached] LX_SYMBOL_CLASS].test */
extern T1 T2162f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] LX_SYMBOL_CLASS.is_equal */
extern T1 T2109f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_SYMBOL_CLASS.same_symbol_class */
extern T1 T2109f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [[attached] NATURAL_64].same_items */
extern T1 T767f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].key_equality_tester */
extern T0* T2111f27(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_SYMBOL_CLASS].force */
extern void T2117f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_SYMBOL_CLASS].fill_with */
extern void T2118f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_SYMBOL_CLASS].extend */
extern void T2118f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.new_nfa_from_character_class */
extern T0* T2095f184(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.new_symbol_class_nfa */
extern T0* T2095f238(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_NFA.make_symbol_class */
extern T0* T2120c12(GE_context* ac, T0* a1, T1 a2);
/* [detachable] LX_SYMBOL_CLASS_TRANSITION [[attached] LX_NFA_STATE].make */
extern T0* T2174c4(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].force */
extern void T2111f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.new_nfa_from_utf8_character_class */
extern T0* T2095f240(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.new_nfa_from_unions_of_concatenations_of_symbol_classes */
extern T0* T2095f244(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS].item */
extern T0* T2153f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS].first */
extern T0* T2153f7(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].item */
extern T0* T2152f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].force_new */
extern void T2108f36(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].key_storage_put */
extern void T2108f44(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].item_storage_put */
extern void T2108f43(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]].force */
extern void T2164f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]].fill_with */
extern void T2161f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]].extend */
extern void T2161f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].slots_put */
extern void T2108f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].clashes_put */
extern void T2108f41(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].slots_item */
extern T6 T2108f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].hash_position */
extern T6 T2108f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].clashes_item */
extern T6 T2108f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].resize */
extern void T2108f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].clashes_resize */
extern void T2108f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].special_integer_ */
extern T0* T2108f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].key_storage_resize */
extern void T2108f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].item_storage_resize */
extern void T2108f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]].aliased_resized_area */
extern T0* T2164f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]].aliased_resized_area */
extern T0* T2161f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].key_storage_item */
extern T0* T2108f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].slots_resize */
extern void T2108f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].new_modulus */
extern T6 T2108f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].new_capacity */
extern T6 T2108f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].unset_found_item */
extern void T2108f39(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].is_empty */
extern T1 T2152f6(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.append_concatenations_of_symbol_classes_from_utf8_n_byte_character_class */
extern void T2095f313(GE_context* ac, T0* C, T0* a1, T6 a2, T0* a3, T0* a4);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].new_iterator */
extern T0* T745f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].is_equal */
extern T1 T745f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].same_equality_tester */
extern T1 T745f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] INTEGER_32].any_ */
extern T0* T745f40(GE_context* ac, T0* C);
/* detachable ARRAY [detachable DS_HASH_SET [[attached] INTEGER_32]].put */
extern void T2150f6(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ARRAY [detachable DS_HASH_SET [[attached] INTEGER_32]].item */
extern T0* T2150f4(GE_context* ac, T0* C, T6 a1);
/* detachable ARRAY [detachable DS_HASH_SET [[attached] INTEGER_32]].make_filled */
extern T0* T2150c5(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* detachable ARRAY [detachable DS_HASH_SET [[attached] INTEGER_32]].make_filled_area */
extern void T2150f7(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable DS_HASH_SET [[attached] INTEGER_32]].make_filled */
extern T0* T2151c4(GE_context* ac, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable DS_HASH_SET [[attached] INTEGER_32]].fill_with */
extern void T2151f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable DS_HASH_SET [[attached] INTEGER_32]].extend */
extern void T2151f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [detachable DS_HASH_SET [[attached] INTEGER_32]].make_empty */
extern void T2151f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.append_concatenations_of_symbol_classes_from_utf8_2_byte_character_class */
extern void T2095f312(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ARRAY [detachable LX_SYMBOL_CLASS].put */
extern void T2148f6(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable ARRAY [detachable LX_SYMBOL_CLASS].item */
extern T0* T2148f4(GE_context* ac, T0* C, T6 a1);
/* detachable ARRAY [detachable LX_SYMBOL_CLASS].make_filled */
extern T0* T2148c5(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* detachable ARRAY [detachable LX_SYMBOL_CLASS].make_filled_area */
extern void T2148f7(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable LX_SYMBOL_CLASS].make_filled */
extern T0* T2149c4(GE_context* ac, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable LX_SYMBOL_CLASS].fill_with */
extern void T2149f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable LX_SYMBOL_CLASS].extend */
extern void T2149f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] SPECIAL [detachable LX_SYMBOL_CLASS].make_empty */
extern void T2149f6(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].force_last */
extern void T2152f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].force */
extern void T2208f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].fill_with */
extern void T2207f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].extend */
extern void T2207f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].resize */
extern void T2152f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].aliased_resized_area */
extern T0* T2208f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].aliased_resized_area */
extern T0* T2207f4(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].new_capacity */
extern T6 T2152f9(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].extendible */
extern T1 T2152f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS].put_last */
extern void T2153f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS].make */
extern T0* T2153c9(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS].set_internal_cursor */
extern void T2153f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS].new_cursor */
extern T0* T2153f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_SYMBOL_CLASS].make */
extern T0* T2210c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_SYMBOL_CLASS].make */
extern T0* T2117f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_SYMBOL_CLASS].make_empty */
extern T0* T2118c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_SYMBOL_CLASS].default_create */
extern T0* T2117c3(GE_context* ac);
/* detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].make */
extern T0* T2152c11(GE_context* ac, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].set_internal_cursor */
extern void T2152f13(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].new_cursor */
extern T0* T2152f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].make */
extern T0* T2209c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].make */
extern T0* T2208f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].make_empty */
extern T0* T2207c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]].default_create */
extern T0* T2208c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].found_item */
extern T0* T2108f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].item_storage_item */
extern T0* T2108f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].found */
extern T1 T2108f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].search */
extern void T2108f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].search_position */
extern void T2108f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.question_character_class */
extern T0* T2095f183(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].force_new */
extern void T2112f39(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].found_item */
extern T0* T2112f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].item_storage_item */
extern T0* T2112f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].found */
extern T1 T2112f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].search */
extern void T2112f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_NFA.build_optional */
extern void T2120f11(GE_context* ac, T0* C);
/* [detachable] LX_NFA_STATE.set_epsilon_transition */
extern void T2147f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_NFA.final_state */
extern T0* T2120f5(GE_context* ac, T0* C);
/* [detachable] LX_NFA.build_positive_closure */
extern void T2120f10(GE_context* ac, T0* C);
/* [detachable] LX_NFA.build_closure */
extern void T2120f9(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.new_nfa_from_character */
extern T0* T2095f182(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.new_symbol_nfa */
extern T0* T2095f239(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_NFA.make_symbol */
extern T0* T2120c13(GE_context* ac, T6 a1, T1 a2);
/* [detachable] LX_SYMBOL_TRANSITION [[attached] LX_NFA_STATE].make */
extern T0* T2175c4(GE_context* ac, T6 a1, T0* a2);
/* [detachable] STRING_32.append_integer */
extern void T18f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.append_character_to_string */
extern T0* T2095f181(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] LX_WILDCARD_PARSER.new_epsilon_nfa */
extern T0* T2095f179(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_STACK [[attached] BOOLEAN].item */
extern T1 T2114f5(GE_context* ac, T0* C);
/* [detachable] LX_NFA.build_concatenation */
extern void T2120f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].append_last */
extern void T2137f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].extend_last */
extern void T2137f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE].forth */
extern void T2188f12(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].cursor_forth */
extern void T2137f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].add_traversing_cursor */
extern void T2137f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE].set_next_cursor */
extern void T2188f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].remove_traversing_cursor */
extern void T2137f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE].item */
extern T0* T2188f4(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].cursor_item */
extern T0* T2137f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE].after */
extern T1 T2188f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE].start */
extern void T2188f11(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA_STATE].cursor_start */
extern void T2137f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE].set_after */
extern void T2188f13(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE].off */
extern T1 T2188f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA_STATE].before */
extern T1 T2188f8(GE_context* ac, T0* C);
/* [detachable] LX_NFA.build_union */
extern void T2120f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.report_unrecognized_rule_error */
extern void T2095f279(GE_context* ac, T0* C);
/* [detachable] LX_UNRECOGNIZED_RULE_ERROR.make */
extern T0* T2125c8(GE_context* ac, T0* a1, T6 a2);
/* [detachable] LX_WILDCARD_PARSER.process_rule */
extern void T2095f278(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_START_CONDITIONS.add_nfa_to_all */
extern void T2106f13(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] LX_START_CONDITION.add_nfa */
extern void T2143f7(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] LX_START_CONDITION.put_nfa */
extern void T2143f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA].force_last */
extern void T2145f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA].resize */
extern void T2145f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA].new_capacity */
extern T6 T2145f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA].extendible */
extern T1 T2145f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_START_CONDITION.put_bol_nfa */
extern void T2143f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_START_CONDITIONS.add_nfa_to_non_exclusive */
extern void T2106f15(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] LX_START_CONDITIONS.is_empty */
extern T1 T2106f8(GE_context* ac, T0* C);
/* [detachable] LX_DESCRIPTION.set_bol_needed */
extern void T2096f40(GE_context* ac, T0* C, T1 a1);
/* [detachable] LX_DESCRIPTION.set_variable_trail_context */
extern void T2096f39(GE_context* ac, T0* C, T1 a1);
/* detachable LX_RULE.set_column_count */
extern void T2122f21(GE_context* ac, T0* C, T6 a1);
/* detachable LX_RULE.set_line_count */
extern void T2122f20(GE_context* ac, T0* C, T6 a1);
/* detachable LX_RULE.set_trail_count */
extern void T2122f19(GE_context* ac, T0* C, T6 a1);
/* detachable LX_RULE.set_head_count */
extern void T2122f18(GE_context* ac, T0* C, T6 a1);
/* detachable LX_RULE.set_trail_context */
extern void T2122f17(GE_context* ac, T0* C, T1 a1);
/* detachable LX_RULE.set_line_nb */
extern void T2122f16(GE_context* ac, T0* C, T6 a1);
/* detachable LX_RULE.set_pattern */
extern void T2122f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_NFA.set_accepted_rule */
extern void T2120f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_NFA_STATE.set_accepted_rule */
extern void T2147f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.check_options */
extern void T2095f277(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.report_full_and_variable_trailing_context_error */
extern void T2095f296(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_VARIABLE_TRAILING_CONTEXT_ERROR.make */
extern T0* T2132c8(GE_context* ac);
/* [detachable] LX_WILDCARD_PARSER.report_full_and_reject_error */
extern void T2095f295(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_REJECT_ERROR.make */
extern T0* T2131c8(GE_context* ac);
/* [detachable] LX_WILDCARD_PARSER.report_full_and_meta_equiv_classes_error */
extern void T2095f294(GE_context* ac, T0* C);
/* [detachable] LX_FULL_AND_META_ERROR.make */
extern T0* T2130c8(GE_context* ac);
/* [detachable] LX_WILDCARD_PARSER.build_equiv_classes */
extern void T2095f276(GE_context* ac, T0* C);
/* [detachable] LX_DESCRIPTION.set_equiv_classes */
extern void T2096f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_SYMBOL_CLASS.convert_to_equivalence */
extern void T2109f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_EQUIVALENCE_CLASSES.equivalence_class */
extern T6 T2124f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_EQUIVALENCE_CLASSES.is_representative */
extern T1 T2124f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_EQUIVALENCE_CLASSES.build */
extern void T2124f12(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS].forth */
extern void T2123f11(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].cursor_forth */
extern void T2111f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].add_traversing_cursor */
extern void T2111f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS].set_next_cursor */
extern void T2123f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS].set_position */
extern void T2123f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].remove_traversing_cursor */
extern void T2111f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS].set_after */
extern void T2123f12(GE_context* ac, T0* C);
/* [detachable] LX_EQUIVALENCE_CLASSES.add */
extern void T2124f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS].item */
extern T0* T2123f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].cursor_item */
extern T0* T2111f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS].after */
extern T1 T2123f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS].start */
extern void T2123f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].cursor_start */
extern void T2111f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].cursor_off */
extern T1 T2111f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS].off */
extern T1 T2123f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].is_empty */
extern T1 T2111f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].new_cursor */
extern T0* T2111f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] LX_SYMBOL_CLASS].make */
extern T0* T2123c9(GE_context* ac, T0* a1);
/* [detachable] LX_EQUIVALENCE_CLASSES.make */
extern T0* T2124c10(GE_context* ac, T6 a1, T6 a2);
/* [detachable] LX_EQUIVALENCE_CLASSES.initialize */
extern void T2124f13(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.set_maximum_symbol_equivalence_class */
extern void T2095f275(GE_context* ac, T0* C);
/* [detachable] LX_SYMBOL_CLASS.set_maximum_symbol_equivalence_class */
extern void T2109f32(GE_context* ac, T0* C, T6 a1);
/* detachable SPECIAL [detachable SPECIAL [[attached] NATURAL_64]].upper */
extern T6 T2166f4(GE_context* ac, T0* C);
/* [detachable] LX_DESCRIPTION.set_maximum_symbol */
extern void T2096f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.yy_push_last_value */
extern void T2095f261(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.accept */
extern void T2095f260(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_do_error_action */
extern void T2095f258(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.report_error */
extern void T2095f274(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.report_eof_expected_error */
extern void T2095f273(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.read_token */
extern void T2095f257(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_execute_action */
extern void T2095f272(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.character_text */
extern T0* T2095f208(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.unicode_text_item */
extern T3 T2095f221(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.text */
extern T0* T2095f241(GE_context* ac, T0* C);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.substring */
extern T0* T1479f14(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] LX_WILDCARD_PARSER.text_count */
extern T6 T2095f227(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.report_bad_character_error */
extern void T2095f293(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_BAD_CHARACTER_ERROR.make */
extern T0* T2129c8(GE_context* ac, T0* a1, T6 a2, T0* a3);
/* [detachable] LX_WILDCARD_PARSER.report_bad_character_class_error */
extern void T2095f292(GE_context* ac, T0* C);
/* [detachable] LX_BAD_CHARACTER_CLASS_ERROR.make */
extern T0* T2128c8(GE_context* ac, T0* a1, T6 a2);
/* [detachable] LX_WILDCARD_PARSER.process_unicode_character */
extern void T2095f291(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.process_character */
extern void T2095f299(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.report_invalid_character_error */
extern void T2095f282(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_INVALID_UNICODE_CHARACTER_ERROR.make */
extern T0* T2126c8(GE_context* ac, T0* a1, T6 a2, T0* a3);
/* [detachable] LX_WILDCARD_PARSER.report_character_out_of_range_error */
extern void T2095f301(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_CHARACTER_OUT_OF_RANGE_ERROR.make */
extern T0* T2133c8(GE_context* ac, T0* a1, T6 a2, T0* a3);
/* [detachable] LX_WILDCARD_PARSER.process_hexadecimal_character */
extern void T2095f290(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.process_octal_character */
extern void T2095f289(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.process_escaped_character */
extern void T2095f288(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.report_missing_quote_error */
extern void T2095f287(GE_context* ac, T0* C);
/* [detachable] LX_MISSING_QUOTE_ERROR.make */
extern T0* T2127c8(GE_context* ac, T0* a1, T6 a2);
/* [detachable] LX_WILDCARD_PARSER.process_single_character */
extern void T2095f286(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.text_item */
extern T2 T2095f201(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.less */
extern void T2095f285(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.yy_set_line_column */
extern void T2095f298(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.character_class_with_name */
extern T0* T2095f195(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.unicode_text */
extern T0* T2095f194(GE_context* ac, T0* C);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.unicode_substring */
extern T0* T1479f13(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] KL_CHARACTER_BUFFER.unicode_substring */
extern T0* T1446f10(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] KL_CHARACTER_BUFFER.append_unicode_substring_to_string */
extern void T1446f19(GE_context* ac, T0* C, T6 a1, T6 a2, T0* a3);
/* [detachable] STRING_32.append_substring_general */
extern void T18f42(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] LX_WILDCARD_PARSER.set_start_condition */
extern void T2095f284(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.yy_execute_eof_action */
extern void T2095f271(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.terminate */
extern void T2095f283(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.wrap */
extern T1 T2095f165(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_refill_input_buffer */
extern void T2095f270(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_set_content */
extern void T2095f267(GE_context* ac, T0* C, T0* a1);
/* [detachable] YY_UNICODE_BUFFER.fill */
extern void T1638f15(GE_context* ac, T0* C);
/* [detachable] YY_UNICODE_BUFFER.set_index */
extern void T1638f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.yy_null_trans_state */
extern T6 T2095f164(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.yy_previous_state */
extern T6 T2095f163(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.fatal_error */
extern void T2095f269(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STDERR_FILE.put_character */
extern void T134f9(GE_context* ac, T0* C, T2 a1);
/* [detachable] LX_WILDCARD_PARSER.std */
extern T0* T2095f189(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.report_invalid_unicode_character_error */
extern void T2095f268(GE_context* ac, T0* C, T10 a1);
/* [detachable] LX_WILDCARD_PARSER.yy_at_beginning_of_line */
extern T6 T2095f132(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.special_integer_ */
extern T0* T2095f129(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_init_value_stacks */
extern void T2095f256(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_clear_all */
extern void T2095f265(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.clear_all */
extern void T2095f280(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.clear_stacks */
extern void T2095f297(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_clear_value_stacks */
extern void T2095f300(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] LX_NFA].keep_head */
extern void T2121f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_NFA].set_count */
extern void T2121f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_SYMBOL_CLASS].keep_head */
extern void T2118f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_SYMBOL_CLASS].set_count */
extern void T2118f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] STRING_32].keep_head */
extern void T2116f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] STRING_32].set_count */
extern void T2116f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.abort */
extern void T2095f259(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.set_input_buffer */
extern void T2095f250(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yy_load_input_buffer */
extern void T2095f255(GE_context* ac, T0* C);
/* [detachable] YY_UNICODE_BUFFER.set_position */
extern void T1638f14(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] LX_WILDCARD_PARSER.new_unicode_string_buffer */
extern T0* T2095f147(GE_context* ac, T0* C, T0* a1);
/* [detachable] YY_UNICODE_BUFFER.make */
extern T0* T1638c12(GE_context* ac, T0* a1);
/* [detachable] YY_UNICODE_BUFFER.make_from_buffer */
extern void T1638f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.fill_from_string */
extern void T1479f22(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.fill_from_substring */
extern void T1479f23(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] YY_UNICODE_BUFFER.new_default_buffer */
extern T0* T1638f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_UNICODE_CHARACTER_BUFFER.make */
extern T0* T1479c15(GE_context* ac, T6 a1);
/* [detachable] STRING_32.make_filled */
extern T0* T18c37(GE_context* ac, T3 a1, T6 a2);
/* [detachable] STRING_32.fill_character */
extern void T18f40(GE_context* ac, T0* C, T3 a1);
/* [detachable] LX_WILDCARD_PARSER.make_from_description */
extern T0* T2095c246(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].make */
extern T0* T2108c34(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].make_with_equality_testers */
extern void T2108f37(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].make_sparse_container */
extern void T2108f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].set_internal_cursor */
extern void T2108f54(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].new_cursor */
extern T0* T2108f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].make */
extern T0* T2165c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].make_slots */
extern void T2108f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].make_clashes */
extern void T2108f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].make_key_storage */
extern void T2108f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]], [attached] LX_SYMBOL_CLASS].make_item_storage */
extern void T2108f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]].make */
extern T0* T2164f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]].make_empty */
extern T0* T2161c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_ARRAYED_LIST [[attached] DS_ARRAYED_LIST [[attached] LX_SYMBOL_CLASS]]].default_create */
extern T0* T2164c3(GE_context* ac);
/* detachable KL_EQUALITY_TESTER [[attached] LX_SYMBOL_CLASS].default_create */
extern T0* T2162c2(GE_context* ac);
/* [detachable] LX_ACTION_FACTORY.make */
extern T0* T2107c1(GE_context* ac);
/* [detachable] LX_START_CONDITIONS.make */
extern void T2106f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_START_CONDITIONS.make */
extern T0* T2106c12(GE_context* ac, T6 a1);
/* [detachable] LX_START_CONDITIONS.set_internal_cursor */
extern void T2106f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_START_CONDITIONS.new_cursor */
extern T0* T2106f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_START_CONDITION].make */
extern T0* T2160c4(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_START_CONDITION].make */
extern T0* T2159f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] LX_START_CONDITION].make_empty */
extern T0* T2158c4(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_START_CONDITION].default_create */
extern T0* T2159c2(GE_context* ac);
/* [detachable] LX_WILDCARD_PARSER.make_parser_skeleton */
extern void T2095f249(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_build_parser_tables */
extern void T2095f254(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yycheck_template */
extern T0* T2095f178(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yyfixed_array */
extern T0* T2095f232(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yytable_template */
extern T0* T2095f177(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yypgoto_template */
extern T0* T2095f176(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yypact_template */
extern T0* T2095f174(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yydefgoto_template */
extern T0* T2095f173(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yydefact_template */
extern T0* T2095f172(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yytypes2_template */
extern T0* T2095f171(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yytypes1_template */
extern T0* T2095f169(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yyr1_template */
extern T0* T2095f160(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yytranslate_template */
extern T0* T2095f159(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yytranslate_template_2 */
extern void T2095f303(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yyarray_subcopy */
extern void T2095f314(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] LX_WILDCARD_PARSER.integer_array_ */
extern T0* T2095f245(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yytranslate_template_1 */
extern void T2095f302(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yy_create_value_stacks */
extern void T2095f253(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_32].make */
extern T0* T2115f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] STRING_32].make_empty */
extern T0* T2116c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] STRING_32].default_create */
extern T0* T2115c3(GE_context* ac);
/* [detachable] LX_WILDCARD_PARSER.make_lex_scanner_from_description */
extern void T2095f248(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_STACK [[attached] BOOLEAN].force */
extern void T2114f9(GE_context* ac, T0* C, T1 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] BOOLEAN].resize */
extern void T2114f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] BOOLEAN].new_capacity */
extern T6 T2114f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] BOOLEAN].extendible */
extern T1 T2114f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_STACK [[attached] BOOLEAN].make */
extern T0* T2114c8(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].make */
extern void T2111f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].make */
extern T0* T2111c36(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].set_internal_cursor */
extern void T2111f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].make_slots */
extern void T2111f43(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].make_clashes */
extern void T2111f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].make_key_storage */
extern void T2111f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].make_item_storage */
extern void T2111f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].set_key_equality_tester */
extern void T2112f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.string_32_equality_tester */
extern T0* T2095f143(GE_context* ac, T0* C);
/* [detachable] KL_EQUALITY_TESTER [[attached] STRING_32].default_create */
extern T0* T2113c2(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].make_map */
extern T0* T2112c35(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].make_with_equality_testers */
extern void T2112f40(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].make_sparse_container */
extern void T2112f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].set_internal_cursor */
extern void T2112f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].new_cursor */
extern T0* T2112f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].make */
extern T0* T2170c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].make_slots */
extern void T2112f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].make_clashes */
extern void T2112f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].make_key_storage */
extern void T2112f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] LX_SYMBOL_CLASS, [attached] STRING_32].make_item_storage */
extern void T2112f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] LX_SYMBOL_CLASS].make_equal */
extern T0* T2111c35(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].set_key_equality_tester */
extern void T2110f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.string_equality_tester */
extern T0* T2095f126(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].make_map */
extern T0* T2110c20(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].make_with_equality_testers */
extern void T2110f22(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].make_sparse_container */
extern void T2110f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].set_internal_cursor */
extern void T2110f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].new_cursor */
extern T0* T2110f17(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] STRING_32, [attached] STRING_8].make */
extern T0* T2168c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].unset_found_item */
extern void T2110f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].make_slots */
extern void T2110f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].new_modulus */
extern T6 T2110f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].make_clashes */
extern void T2110f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].make_key_storage */
extern void T2110f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] STRING_32, [attached] STRING_8].make_item_storage */
extern void T2110f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_WILDCARD_PARSER.make_with_buffer */
extern void T2095f252(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yy_initialize */
extern void T2095f266(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_initialize */
extern void T2095f266p1(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_build_tables */
extern void T2095f281(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_accept_template */
extern T0* T2095f218(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_fixed_array */
extern T0* T2095f242(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yy_meta_template */
extern T0* T2095f217(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_ec_template */
extern T0* T2095f216(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_ec_template_1 */
extern void T2095f311(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yy_array_subcopy */
extern void T2095f315(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] LX_WILDCARD_PARSER.yy_def_template */
extern T0* T2095f215(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_base_template */
extern T0* T2095f214(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_chk_template */
extern T0* T2095f213(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_chk_template_3 */
extern void T2095f310(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yy_chk_template_2 */
extern void T2095f309(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yy_chk_template_1 */
extern void T2095f308(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yy_nxt_template */
extern T0* T2095f212(GE_context* ac, T0* C);
/* [detachable] LX_WILDCARD_PARSER.yy_nxt_template_3 */
extern void T2095f307(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yy_nxt_template_2 */
extern void T2095f306(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.yy_nxt_template_1 */
extern void T2095f305(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_WILDCARD_PARSER.empty_buffer */
extern T0* T2095f139(GE_context* ac, T0* C);
/* [detachable] LX_DESCRIPTION.set_case_insensitive */
extern void T2096f36(GE_context* ac, T0* C, T1 a1);
/* [detachable] LX_DESCRIPTION.set_full_table */
extern void T2096f35(GE_context* ac, T0* C, T1 a1);
/* [detachable] LX_DESCRIPTION.set_meta_equiv_classes_used */
extern void T2096f34(GE_context* ac, T0* C, T1 a1);
/* [detachable] LX_DESCRIPTION.set_equiv_classes_used */
extern void T2096f33(GE_context* ac, T0* C, T1 a1);
/* [detachable] LX_DESCRIPTION.make */
extern T0* T2096c32(GE_context* ac);
/* [detachable] LX_START_CONDITIONS.make_with_initial */
extern T0* T2106c14(GE_context* ac, T6 a1);
/* [detachable] LX_START_CONDITIONS.put_first */
extern void T2106f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_START_CONDITIONS.put */
extern void T2106f18(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] LX_START_CONDITIONS.move_cursors_right */
extern void T2106f21(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_START_CONDITION].set_position */
extern void T2160f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] LX_START_CONDITIONS.move_right */
extern void T2106f20(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] LX_START_CONDITION].force */
extern void T2159f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_START_CONDITION].fill_with */
extern void T2158f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] LX_START_CONDITION].extend */
extern void T2158f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_START_CONDITIONS.put_last */
extern void T2106f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_START_CONDITION.make */
extern T0* T2143c6(GE_context* ac, T0* a1, T6 a2, T1 a3);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA].make */
extern T0* T2145c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA].set_internal_cursor */
extern void T2145f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] LX_NFA].new_cursor */
extern T0* T2145f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] LX_NFA].make */
extern T0* T2195c3(GE_context* ac, T0* a1);
/* [detachable] UT_ERROR_HANDLER.make_null */
extern T0* T132c12(GE_context* ac);
/* [detachable] UT_ERROR_HANDLER.null_output_stream */
extern T0* T132f6(GE_context* ac, T0* C);
/* [detachable] LX_DFA_WILDCARD.wipe_out */
extern void T2068f22(GE_context* ac, T0* C);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.file_system */
extern T0* T1758f13(GE_context* ac, T0* C);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.unix_file_system */
extern T0* T1758f15(GE_context* ac, T0* C);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.windows_file_system */
extern T0* T1758f14(GE_context* ac, T0* C);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.execution_environment */
extern T0* T1758f12(GE_context* ac, T0* C);
/* detachable ET_CLUSTER_DEPENDENCE_CONSTRAINT.operating_system */
extern T0* T1758f10(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.current_system */
extern T0* T454f62(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] STRING_8].make_from_linear */
extern T0* T147c40(GE_context* ac, T0* a1);
/* [detachable] ST_SPLITTER.make */
extern T0* T1953c11(GE_context* ac);
/* [detachable] ET_EIFFEL_PREPARSER.build_provider_constraint */
extern void T454f187(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_EIFFEL_PREPARSER.build_provider_constraint */
extern T0* T454f187ac1(GE_context* ac, T0* a1, T0* a2);
extern T1 T454f187ot1(T0* a1);
/* detachable ET_ECF_CLUSTER.set_provider_constraint */
extern void T802f96(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTERS.set_provider_constraint */
extern void T787f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.universes_do_all */
extern void T187f198(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.universes_do_all */
extern void T210f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ADAPTED_LIBRARIES.universes_do_all */
extern void T188f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.universes_do_all */
extern void T58f281(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.add_universe_full_name */
extern void T454f193(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.full_name */
extern T0* T211f97(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.relative_name */
extern T0* T211f101(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_DOTNET_ASSEMBLY.full_name */
extern T0* T596f97(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_DOTNET_ASSEMBLY.relative_name */
extern T0* T596f99(GE_context* ac, T0* C, T0* a1, T2 a2);
/* [detachable] ET_LIBRARY.full_name */
extern T0* T852f98(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_LIBRARY.relative_name */
extern T0* T852f101(GE_context* ac, T0* C, T0* a1, T2 a2);
/* detachable ET_SYSTEM.full_name */
extern T0* T40f131(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_ECF_SYSTEM.full_name */
extern T0* T58f157(GE_context* ac, T0* C, T2 a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_gcaaa_error */
extern void T53f565(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_CLUSTER_ERROR.make_gcaaa */
extern T0* T1940c26(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_NULL_ERROR_HANDLER.reportable_gcaaa_error */
extern T1 T53f178(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.report_gcaaa_error */
extern void T36f565(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ERROR_HANDLER.reportable_gcaaa_error */
extern T1 T36f178(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.add_recursive_cluster */
extern void T802f100(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.add_subcluster */
extern void T802f95(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_parent */
extern void T802f94(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTERS.put_last */
extern void T787f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].force_last */
extern void T1579f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CLUSTER].force */
extern void T1860f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_CLUSTER].fill_with */
extern void T1859f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_CLUSTER].extend */
extern void T1859f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].resize */
extern void T1579f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CLUSTER].aliased_resized_area */
extern T0* T1860f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_CLUSTER].aliased_resized_area */
extern T0* T1859f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].new_capacity */
extern T6 T1579f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].extendible */
extern T1 T1579f7(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_CLUSTER.set_subclusters */
extern void T802f98(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTERS.set_parent */
extern void T787f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTERS.make_empty */
extern T0* T787c9(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].make */
extern T0* T1579c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].set_internal_cursor */
extern void T1579f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].new_cursor */
extern T0* T1579f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_CLUSTER].make */
extern T0* T1861c5(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CLUSTER].make */
extern T0* T1860f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_CLUSTER].make_empty */
extern T0* T1859c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CLUSTER].default_create */
extern T0* T1860c3(GE_context* ac);
/* detachable ET_ECF_CLUSTER.new_recursive_cluster */
extern T0* T802f61(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_overridden_group */
extern void T802f86(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_provider_groups */
extern void T802f82(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_visible_classes */
extern void T802f81(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_class_options */
extern void T802f78(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_options */
extern void T802f77(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_class_mappings */
extern void T802f80(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_class_renamings */
extern void T802f79(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_classname_prefix */
extern void T802f72(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_use_obsolete_syntax */
extern void T802f88(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_CLUSTER.set_hidden */
extern void T802f71(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_CLUSTER.set_read_only */
extern void T802f73(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_CLUSTER.set_override */
extern void T802f85(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_CLUSTER.set_implicit */
extern void T802f103(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_CLUSTER.set_recursive */
extern void T802f74(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_CLUSTER.set_file_rules */
extern void T802f102(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.make_relative */
extern T0* T802c101(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] UT_VERSION.is_less_equal */
extern T1 T46f16(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.ecf_1_4_0 */
extern T0* T802f37(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.set_scm_mapping_constraint_enabled */
extern void T802f87(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_CLUSTERS.set_scm_mapping_constraint_enabled */
extern void T787f10(GE_context* ac, T0* C, T1 a1);
/* [detachable] KL_UNIX_FILE_SYSTEM.directory_exists */
extern T1 T126f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_DIRECTORY.exists */
extern T1 T333f31(GE_context* ac, T0* C);
/* [detachable] KL_UNIX_FILE_SYSTEM.tmp_directory */
extern T0* T126f30(GE_context* ac, T0* C);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.directory_exists */
extern T1 T125f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_WINDOWS_FILE_SYSTEM.tmp_directory */
extern T0* T125f33(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.is_valid_directory_name */
extern T1 T802f40(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_FILE_RULES.is_included */
extern T1 T790f3(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature detachable ET_ECF_FILE_RULES.is_included */
extern T0* T790f3ac1(GE_context* ac, T0* a1);
/* [detachable] PREDICATE [[attached] TUPLE [[attached] ET_ECF_FILE_RULE]].set_rout_disp_final */
extern void T2060f8(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable ET_ECF_FILE_RULE.is_included */
extern T1 T810f10(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature detachable ET_ECF_FILE_RULE.is_included */
extern T0* T810f10ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature detachable ET_ECF_FILE_RULE.is_included */
extern T0* T810f10ac2(GE_context* ac, T0* a1);
/* [detachable] PREDICATE [[attached] TUPLE [[attached] RX_PCRE_REGULAR_EXPRESSION]].set_rout_disp_final */
extern void T2102f8(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.matches */
extern T1 T81f306(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] RX_PCRE_REGULAR_EXPRESSION].there_exists */
extern T1 T1593f7(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] RX_PCRE_REGULAR_EXPRESSION].item */
extern T0* T1593f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE].for_all */
extern T1 T1582f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE].item */
extern T0* T1582f9(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_CLUSTER.implicit_relative_name */
extern T0* T802f52(GE_context* ac, T0* C, T2 a1);
/* detachable ET_ECF_CLUSTER.is_valid_directory_name */
extern T1 T802f40p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.preparse_file */
extern void T454f186(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PREPARSER.reset */
extern void T454f192(GE_context* ac, T0* C);
/* [detachable] YY_FILE_BUFFER.set_end_of_file */
extern void T930f28(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.reset */
extern void T454f192p1(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.reset */
extern void T454f192p0(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.reset */
extern void T454f192p2(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.current_position */
extern T0* T454f61(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.read_token */
extern void T454f191(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_execute_action */
extern void T454f198(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PREPARSER.more */
extern void T454f204(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.is_verbatim_string_closer */
extern T1 T454f116(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PREPARSER.report_syntax_error */
extern void T454f199(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.set_start_condition */
extern void T454f202(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PREPARSER.unicode_text_item */
extern T3 T454f113(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PREPARSER.utf8_text_substring */
extern T0* T454f111(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] ET_EIFFEL_PREPARSER.append_utf8_text_substring_to_string */
extern void T454f205(GE_context* ac, T0* C, T6 a1, T6 a2, T0* a3);
/* [detachable] ET_EIFFEL_PREPARSER.set_syntax_error */
extern void T454f201(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.text_count */
extern T6 T454f109(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.use_reference_keyword */
extern T1 T454f108(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.use_note_keyword */
extern T1 T454f107(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.use_detachable_keyword */
extern T1 T454f106(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.use_attribute_keyword */
extern T1 T454f105(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.ast_factory */
extern T0* T454f104(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.use_attached_keyword */
extern T1 T454f100(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_set_line_column */
extern void T454f203(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_execute_eof_action */
extern void T454f197(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PREPARSER.terminate */
extern void T454f200(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.wrap */
extern T1 T454f92(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_refill_input_buffer */
extern void T454f196(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_set_content */
extern void T454f142(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_null_trans_state */
extern T6 T454f91(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_previous_state */
extern T6 T454f90(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.fatal_error */
extern void T454f195(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.report_invalid_unicode_character_error */
extern void T454f194(GE_context* ac, T0* C, T10 a1);
/* [detachable] ET_EIFFEL_PREPARSER.special_integer_ */
extern T0* T454f60(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_at_beginning_of_line */
extern T6 T454f72(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_load_input_buffer */
extern void T454f140(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.tmp_file */
extern T0* T454f58(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.shared_file */
extern T0* T454f129(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.preparse_class */
extern void T454f190(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] ET_ECF_LIBRARY.classes_in_group */
extern T0* T187f119(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_LIBRARY.classes_in_group */
extern T0* T187f119ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_LIBRARY.classes_in_group */
extern T0* T187f119ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_ECF_LIBRARY.classes_in_group */
extern T0* T187f119ac3(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_LIBRARY.initial_classes_in_group_capacity */
extern T6 T187f120(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.classes_in_group */
extern T0* T58f161(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.classes_in_group */
extern T0* T58f161ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.classes_in_group */
extern T0* T58f161ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_ECF_SYSTEM.classes_in_group */
extern T0* T58f161ac3(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.initial_classes_in_group_capacity */
extern T6 T58f162(GE_context* ac, T0* C);
/* detachable ET_SYSTEM.classes_in_group */
extern T0* T40f136(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature detachable ET_SYSTEM.classes_in_group */
extern T0* T40f136ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature detachable ET_SYSTEM.classes_in_group */
extern T0* T40f136ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature detachable ET_SYSTEM.classes_in_group */
extern T0* T40f136ac3(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_SYSTEM.initial_classes_in_group_capacity */
extern T6 T40f137(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.local_classes_do_if */
extern void T570f86(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_MASTER_CLASS.local_non_override_classes_do_if */
extern void T570f88(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_MASTER_CLASS.local_override_classes_do_if */
extern void T570f87(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PREPARSER.current_universe */
extern T0* T454f57(GE_context* ac, T0* C);
/* detachable DS_CELL [detachable ET_CLASS].put */
extern void T2051f3(GE_context* ac, T0* C, T0* a1);
/* detachable DS_CELL [detachable ET_CLASS].make */
extern T0* T2051c2(GE_context* ac, T0* a1);
/* detachable ET_ECF_CLUSTER.is_valid_eiffel_filename */
extern T1 T802f39(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.is_valid_eiffel_filename */
extern T1 T802f39p1(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.has_eiffel_extension */
extern T1 T802f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_DIRECTORY.reset */
extern void T333f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.tmp_directory */
extern T0* T454f56(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.shared_directory */
extern T0* T454f127(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.file_system */
extern T0* T454f55(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.unix_file_system */
extern T0* T454f126(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.windows_file_system */
extern T0* T454f125(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.operating_system */
extern T0* T454f124(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.execution_environment */
extern T0* T454f112(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.error_handler */
extern T0* T454f103(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.set_preparsed */
extern void T802f99(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_CLUSTERS.do_all */
extern void T564f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_NULL_ERROR_HANDLER.report_preparsing_status */
extern void T53f532(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.kind_lower_name */
extern T0* T802f65(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.kind_lower_name */
extern T0* T187f106(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.kind_lower_name */
extern T0* T852f96(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.report_preparsing_status */
extern void T36f532(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.preparse */
extern void T852f154(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.preparse */
extern void T852f154p1(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_LIBRARY.preparse */
extern T0* T852f154p1ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_LIBRARY.preparse */
extern T0* T852f154p1ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_LIBRARY.preparse */
extern T0* T852f154p1ac3(GE_context* ac);
/* Creation of agent #4 in feature [detachable] ET_LIBRARY.preparse */
extern T0* T852f154p1ac4(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.preparse */
extern void T211f154(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.preparse */
extern void T596f154(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLIES.put_last */
extern void T595f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY].force_last */
extern void T1398f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_ASSEMBLY].force */
extern void T1605f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DOTNET_ASSEMBLY].fill_with */
extern void T1606f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_DOTNET_ASSEMBLY].extend */
extern void T1606f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY].resize */
extern void T1398f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_ASSEMBLY].aliased_resized_area */
extern T0* T1605f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_DOTNET_ASSEMBLY].aliased_resized_area */
extern T0* T1606f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY].new_capacity */
extern T6 T1398f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY].extendible */
extern T1 T1398f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_LIBRARY.dotnet_assemblies_do_if */
extern void T187f196(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.do_if */
extern void T210f11(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIBRARY.dotnet_assemblies_do_if */
extern void T852f168(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.dotnet_assemblies_do_if */
extern void T58f278(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] PREDICATE [[attached] TUPLE [[attached] ET_DOTNET_ASSEMBLY]].set_rout_disp_final */
extern void T844f8(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.is_consumable */
extern T1 T211f93(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.is_consumable */
extern T1 T596f93(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].force_last */
extern void T842f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].slots_put */
extern void T842f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].clashes_put */
extern void T842f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].slots_item */
extern T6 T842f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].hash_position */
extern T6 T842f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].resize */
extern void T842f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].clashes_resize */
extern void T842f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].special_integer_ */
extern T0* T842f26(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].key_storage_resize */
extern void T842f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].item_storage_resize */
extern void T842f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].key_storage_item */
extern T0* T842f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].item_storage_item */
extern T0* T842f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].clashes_item */
extern T6 T842f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].slots_resize */
extern void T842f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].new_modulus */
extern T6 T842f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].new_capacity */
extern T6 T842f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].item_storage_put */
extern void T842f40(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].search_position */
extern void T842f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].key_equality_tester */
extern T0* T842f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].unset_found_item */
extern void T842f37(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern void T58f184(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac1(GE_context* ac);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac3(GE_context* ac);
/* Creation of agent #4 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac4(GE_context* ac);
/* Creation of agent #5 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac5(GE_context* ac);
/* Creation of agent #6 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac6(GE_context* ac);
/* Creation of agent #7 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac7(GE_context* ac);
/* Creation of agent #8 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac8(GE_context* ac);
/* Creation of agent #9 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac9(GE_context* ac);
/* Creation of agent #10 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac10(GE_context* ac);
/* Creation of agent #11 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac11(GE_context* ac);
/* Creation of agent #12 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac12(GE_context* ac);
/* Creation of agent #13 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac13(GE_context* ac);
/* Creation of agent #14 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac14(GE_context* ac);
/* Creation of agent #15 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac15(GE_context* ac);
/* Creation of agent #16 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac16(GE_context* ac);
/* Creation of agent #17 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac17(GE_context* ac);
/* Creation of agent #18 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac18(GE_context* ac);
/* Creation of agent #19 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac19(GE_context* ac);
/* Creation of agent #20 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac20(GE_context* ac);
/* Creation of agent #21 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac21(GE_context* ac);
/* Creation of agent #22 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac22(GE_context* ac);
/* Creation of agent #23 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac23(GE_context* ac, T0* a1);
/* Creation of agent #24 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac24(GE_context* ac);
/* Creation of agent #25 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac25(GE_context* ac, T0* a1);
/* Creation of agent #26 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac26(GE_context* ac);
/* Creation of agent #27 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac27(GE_context* ac, T0* a1);
/* Creation of agent #28 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac28(GE_context* ac);
/* Creation of agent #29 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac29(GE_context* ac, T0* a1);
/* Creation of agent #30 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac30(GE_context* ac);
/* Creation of agent #31 in feature [detachable] ET_ECF_SYSTEM.reset_classes_incremental_recursive */
extern T0* T58f184ac31(GE_context* ac, T1 a1);
/* [detachable] ET_IMPLEMENTATION_STATUS_CHECKER.process_class */
extern void T869f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_STATUS_CHECKER.internal_process_class */
extern void T869f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_STATUS_CHECKER.check_providers_validity */
extern void T869f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_NAMED_CLASS].go_after */
extern void T1620f11(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_NAMED_CLASS].cursor_go_after */
extern void T948f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.has_interface_error */
extern T1 T570f41(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.interface_checked */
extern T1 T570f40(GE_context* ac, T0* C);
/* [detachable] ET_IMPLEMENTATION_STATUS_CHECKER.check_suppliers_validity */
extern void T869f9(GE_context* ac, T0* C);
/* [detachable] ET_IMPLEMENTATION_STATUS_CHECKER.set_fatal_error */
extern void T869f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_STATUS_CHECKER.error_handler */
extern T0* T869f4(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.unset_implementation_error */
extern void T74f295(GE_context* ac, T0* C);
/* [detachable] ET_INTERFACE_STATUS_CHECKER.process_class */
extern void T868f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERFACE_STATUS_CHECKER.internal_process_class */
extern void T868f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERFACE_STATUS_CHECKER.check_parents_validity */
extern void T868f13(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.check_type_validity */
extern void T1619f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERFACE_STATUS_CHECKER.check_formal_parameters_validity */
extern void T868f12(GE_context* ac, T0* C);
/* [detachable] ET_INTERFACE_STATUS_CHECKER.check_qualified_anchored_signatures_validity */
extern void T868f11(GE_context* ac, T0* C);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.check_signatures_validity */
extern void T1618f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.check_feature_signatures_validity */
extern void T1618f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERFACE_STATUS_CHECKER.set_fatal_error */
extern void T868f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERFACE_STATUS_CHECKER.error_handler */
extern T0* T868f6(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.unset_interface_error */
extern void T74f294(GE_context* ac, T0* C);
/* [detachable] ET_FLATTENING_STATUS_CHECKER.process_class */
extern void T867f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FLATTENING_STATUS_CHECKER.internal_process_class */
extern void T867f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FLATTENING_STATUS_CHECKER.check_signatures_validity */
extern void T867f12(GE_context* ac, T0* C);
/* [detachable] ET_FLATTENING_STATUS_CHECKER.check_feature_signatures_validity */
extern void T867f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.check_type_validity */
extern void T1617f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FLATTENING_STATUS_CHECKER.check_parents_validity */
extern void T867f11(GE_context* ac, T0* C);
/* [detachable] ET_FLATTENING_STATUS_CHECKER.check_formal_parameters_validity */
extern void T867f10(GE_context* ac, T0* C);
/* [detachable] ET_FLATTENING_STATUS_CHECKER.set_fatal_error */
extern void T867f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FLATTENING_STATUS_CHECKER.error_handler */
extern T0* T867f5(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.unset_flattening_error */
extern void T74f293(GE_context* ac, T0* C);
/* [detachable] ET_ANCESTORS_STATUS_CHECKER.process_class */
extern void T866f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCESTORS_STATUS_CHECKER.internal_process_class */
extern void T866f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCESTORS_STATUS_CHECKER.check_parents_validity */
extern void T866f11(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.check_type_validity */
extern void T1616f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCESTORS_STATUS_CHECKER.check_formal_parameters_validity */
extern void T866f10(GE_context* ac, T0* C);
/* [detachable] ET_ANCESTORS_STATUS_CHECKER.set_fatal_error */
extern void T866f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCESTORS_STATUS_CHECKER.error_handler */
extern T0* T866f5(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.unset_ancestors_error */
extern void T74f292(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.not_implementation_checked */
extern T1 T74f172(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_after_interface_checked */
extern void T74f289(GE_context* ac, T0* C);
/* [detachable] ET_PROCEDURE_LIST.reset_after_interface_checked */
extern void T258f13(GE_context* ac, T0* C);
/* [detachable] ET_QUERY_LIST.reset_after_interface_checked */
extern void T257f13(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_after_features_flattened */
extern void T74f288(GE_context* ac, T0* C);
/* detachable ET_FORMAL_PARAMETER_LIST.reset_constraint_creation_procedures */
extern void T603f36(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.reset_constraint_creation_procedures */
extern void T891f82(GE_context* ac, T0* C);
/* detachable ET_FORMAL_PARAMETER_LIST.reset_constraint_renames */
extern void T603f35(GE_context* ac, T0* C);
/* detachable ET_CONSTRAINED_FORMAL_PARAMETER.reset_constraint_renames */
extern void T1501f84(GE_context* ac, T0* C);
/* [detachable] ET_CLASS_TYPE.reset_renames */
extern void T72f138(GE_context* ac, T0* C);
/* [detachable] ET_TUPLE_TYPE.reset_renames */
extern void T254f128(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER_TYPE.reset_renames */
extern void T242f85(GE_context* ac, T0* C);
/* [detachable] ET_LIKE_CURRENT.reset_renames */
extern void T449f89(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_TYPE.reset_renames */
extern void T1526f91(GE_context* ac, T0* C);
/* detachable ET_QUALIFIED_LIKE_BRACED_TYPE.reset_renames */
extern void T1525f95(GE_context* ac, T0* C);
/* detachable ET_LIKE_FEATURE.reset_renames */
extern void T1524f96(GE_context* ac, T0* C);
/* detachable ET_TYPE_CONSTRAINT_LIST.reset_renames */
extern void T1281f17(GE_context* ac, T0* C);
/* detachable ET_TYPE_RENAME_CONSTRAINT.reset_renames */
extern void T1502f10(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_renames */
extern void T74f327(GE_context* ac, T0* C);
/* [detachable] ET_FORMAL_PARAMETER.reset_constraint_renames */
extern void T891f81(GE_context* ac, T0* C);
/* [detachable] ET_PROCEDURE_LIST.reset_after_features_flattened */
extern void T258f12(GE_context* ac, T0* C);
/* [detachable] ET_QUERY_LIST.reset_after_features_flattened */
extern void T257f12(GE_context* ac, T0* C);
/* [detachable] ET_CLASS.reset_after_ancestors_built */
extern void T74f287(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.local_classes_do_unless_actual */
extern void T570f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.local_non_override_classes_do_unless_actual */
extern void T570f76(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS.local_override_classes_do_unless_actual */
extern void T570f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_IMPLEMENTATION_STATUS_CHECKER.make */
extern T0* T869c5(GE_context* ac, T0* a1);
/* [detachable] ET_IMPLEMENTATION_STATUS_CHECKER.tokens */
extern T0* T869f3(GE_context* ac, T0* C);
/* [detachable] ET_INTERFACE_STATUS_CHECKER.make */
extern T0* T868c7(GE_context* ac, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER3.make */
extern T0* T1619c2(GE_context* ac);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.make */
extern T0* T1618c5(GE_context* ac, T0* a1);
/* [detachable] ET_QUALIFIED_ANCHORED_TYPE_STATUS_CHECKER.tokens */
extern T0* T1618f4(GE_context* ac, T0* C);
/* [detachable] ET_INTERFACE_STATUS_CHECKER.make */
extern void T868f7p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_INTERFACE_STATUS_CHECKER.tokens */
extern T0* T868f5(GE_context* ac, T0* C);
/* [detachable] ET_FLATTENING_STATUS_CHECKER.make */
extern T0* T867c6(GE_context* ac, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER2.make */
extern T0* T1617c2(GE_context* ac);
/* [detachable] ET_FLATTENING_STATUS_CHECKER.make */
extern void T867f6p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_FLATTENING_STATUS_CHECKER.tokens */
extern T0* T867f4(GE_context* ac, T0* C);
/* [detachable] ET_ANCESTORS_STATUS_CHECKER.make */
extern T0* T866c6(GE_context* ac, T0* a1);
/* [detachable] ET_CLASS_TYPE_STATUS_CHECKER1.make */
extern T0* T1616c2(GE_context* ac);
/* [detachable] ET_ANCESTORS_STATUS_CHECKER.make */
extern void T866f6p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ANCESTORS_STATUS_CHECKER.tokens */
extern T0* T866f4(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.classes_do_if_recursive */
extern void T58f197(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.classes_do_if_recursive */
extern T0* T58f197ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.classes_do_if */
extern void T211f161(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.classes_do_if */
extern void T211f161p1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_ASSEMBLY.classes_do_if */
extern void T596f161(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_ASSEMBLY.classes_do_if */
extern void T596f161p1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_LIBRARY.classes_do_if */
extern void T187f191(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIBRARY.classes_do_if */
extern void T852f162(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_SYSTEM.classes_do_if */
extern void T40f202(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.classes_do_if */
extern void T58f274(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.master_classes_do_recursive */
extern void T58f196(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.master_classes_do_recursive */
extern T0* T58f196ac1(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM.classes_modified_recursive */
extern T1 T58f140(GE_context* ac, T0* C);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.classes_modified_recursive */
extern T0* T58f140ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.classes_modified_recursive */
extern T0* T58f140ac2(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #3 in feature [detachable] ET_ECF_SYSTEM.classes_modified_recursive */
extern T0* T58f140ac3(GE_context* ac);
/* Creation of agent #4 in feature [detachable] ET_ECF_SYSTEM.classes_modified_recursive */
extern T0* T58f140ac4(GE_context* ac, T0* a1);
/* [detachable] PREDICATE [[attached] TUPLE].set_rout_disp_final */
extern void T1453f8(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] PREDICATE [[attached] TUPLE [[attached] ET_MASTER_CLASS]].set_rout_disp_final */
extern void T1452f8(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] KL_AGENT_ROUTINES [[attached] ANY].call */
extern void T1449f2(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] UT_TRISTATE.set_true */
extern void T1448f4(GE_context* ac, T0* C);
/* [detachable] UT_TRISTATE.is_true */
extern T1 T1448f2(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.master_classes_do_if_recursive_until */
extern void T58f242(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.master_classes_do_if_recursive_until */
extern T0* T58f242ac1(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.master_classes_do_if_recursive_until */
extern T0* T58f242ac2(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.master_classes_do_if_until */
extern void T211f165(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DOTNET_ASSEMBLY.master_classes_do_if_until */
extern void T596f165(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_LIBRARY.master_classes_do_if_until */
extern void T187f195(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_LIBRARY.master_classes_do_if_until */
extern void T852f166(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_SYSTEM.master_classes_do_if_until */
extern void T40f205(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_SYSTEM.master_classes_do_if_until */
extern void T58f277(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.master_classes_do_if */
extern void T211f164(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_MASTER_CLASS, [attached] ET_CLASS_NAME].do_if */
extern void T225f61(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_ASSEMBLY.master_classes_do_if */
extern void T596f164(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_LIBRARY.master_classes_do_if */
extern void T187f194(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIBRARY.master_classes_do_if */
extern void T852f165(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_SYSTEM.master_classes_do_if */
extern void T40f204(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.master_classes_do_if */
extern void T58f276(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.any_actions */
extern T0* T58f151(GE_context* ac, T0* C);
/* [detachable] KL_AGENT_ROUTINES [[attached] ANY].default_create */
extern T0* T1449c1(GE_context* ac);
/* [detachable] UT_TRISTATE.make_false */
extern T0* T1448c3(GE_context* ac);
/* [detachable] UT_TRISTATE.set_false */
extern void T1448f5(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.import_classes */
extern void T58f183(GE_context* ac, T0* C);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.import_classes */
extern T0* T58f183ac1(GE_context* ac);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.import_classes */
extern T0* T58f183ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_ECF_SYSTEM.import_classes */
extern T0* T58f183ac3(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM.preparse */
extern void T58f182(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.preparse */
extern T0* T58f182ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.preparse */
extern T0* T58f182ac2(GE_context* ac, T0* a1);
/* Creation of agent #3 in feature [detachable] ET_ECF_SYSTEM.preparse */
extern T0* T58f182ac3(GE_context* ac);
/* Creation of agent #4 in feature [detachable] ET_ECF_SYSTEM.preparse */
extern T0* T58f182ac4(GE_context* ac, T0* a1);
/* [detachable] ET_ADAPTED_LIBRARIES.do_recursive */
extern void T188f8(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ADAPTED_LIBRARIES.do_recursive */
extern T0* T188f8ac1(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_LIBRARY.add_library_recursive */
extern void T187f187(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_LIBRARY.add_library_recursive */
extern T0* T187f187ac1(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].force_last */
extern void T903f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].slots_put */
extern void T903f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].clashes_put */
extern void T903f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].slots_item */
extern T6 T903f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].hash_position */
extern T6 T903f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].resize */
extern void T903f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].clashes_resize */
extern void T903f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].special_integer_ */
extern T0* T903f27(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].key_storage_resize */
extern void T903f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].item_storage_resize */
extern void T903f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIBRARY].aliased_resized_area */
extern T0* T1629f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_LIBRARY].aliased_resized_area */
extern T0* T1630f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].key_storage_item */
extern T0* T903f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].item_storage_item */
extern T0* T903f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].clashes_item */
extern T6 T903f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].slots_resize */
extern void T903f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].new_modulus */
extern T6 T903f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].new_capacity */
extern T6 T903f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].item_storage_put */
extern void T903f40(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIBRARY].force */
extern void T1629f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_LIBRARY].fill_with */
extern void T1630f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_LIBRARY].extend */
extern void T1630f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].search_position */
extern void T903f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].key_equality_tester */
extern T0* T903f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].unset_found_item */
extern void T903f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].has */
extern T1 T903f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_LIBRARY.add_library_recursive */
extern void T852f158(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_LIBRARY.add_library_recursive */
extern T0* T852f158ac1(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].do_all */
extern void T903f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].make */
extern T0* T903c30(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].set_internal_cursor */
extern void T903f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].new_cursor */
extern T0* T903f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_LIBRARY].make */
extern T0* T1628c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].make_slots */
extern void T903f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].make_clashes */
extern void T903f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].make_key_storage */
extern void T903f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_LIBRARY].make_item_storage */
extern void T903f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIBRARY].make */
extern T0* T1629f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_LIBRARY].make_empty */
extern T0* T1630c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_LIBRARY].default_create */
extern T0* T1629c3(GE_context* ac);
/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.do_recursive */
extern void T210f7(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.do_recursive */
extern T0* T210f7ac1(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.add_dotnet_assembly_recursive */
extern void T211f157(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_DOTNET_ASSEMBLY.add_dotnet_assembly_recursive */
extern T0* T211f157ac1(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].has */
extern T1 T842f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.add_dotnet_assembly_recursive */
extern void T596f157(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_DOTNET_ASSEMBLY.add_dotnet_assembly_recursive */
extern T0* T596f157ac1(GE_context* ac, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY_CLASSIC_CONSUMER.consume_assemblies */
extern void T467f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.consume */
extern void T211f153(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY_CLASSIC_CONSUMER.consume_assembly */
extern void T467f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLY.consume */
extern void T596f153(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLIES.dotnet_assembly */
extern T0* T595f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DOTNET_ASSEMBLIES.count */
extern T6 T595f3(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].do_all */
extern void T842f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DOTNET_ASSEMBLIES.make_empty */
extern T0* T595c5(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY].make */
extern T0* T1398c12(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY].set_internal_cursor */
extern void T1398f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_DOTNET_ASSEMBLY].new_cursor */
extern T0* T1398f11(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_DOTNET_ASSEMBLY].make */
extern T0* T1731c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_ASSEMBLY].make */
extern T0* T1605f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_DOTNET_ASSEMBLY].make_empty */
extern T0* T1606c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_DOTNET_ASSEMBLY].default_create */
extern T0* T1605c3(GE_context* ac);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].make */
extern T0* T842c30(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].set_internal_cursor */
extern void T842f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].new_cursor */
extern T0* T842f17(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] ET_DOTNET_ASSEMBLY].make */
extern T0* T1604c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].make_slots */
extern void T842f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].make_clashes */
extern void T842f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].make_key_storage */
extern void T842f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] ET_DOTNET_ASSEMBLY].make_item_storage */
extern void T842f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.count */
extern T6 T210f3(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.compile_degree_6 */
extern void T375f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.report_degree_6_metrics */
extern void T375f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.activate_dynamic_type_set_builder */
extern void T63f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.compile_all */
extern void T63f86(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_DYNAMIC_SYSTEM.compile_all */
extern T0* T63f86ac1(GE_context* ac, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.compile_all_features */
extern void T63f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.classes_do_recursive_until */
extern void T58f174(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM.classes_do_recursive_until */
extern T0* T58f174ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM.classes_do_recursive_until */
extern T0* T58f174ac2(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.classes_do_until */
extern void T211f159(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_ASSEMBLY.classes_do_until */
extern void T596f159(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_LIBRARY.classes_do_until */
extern void T187f189(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIBRARY.classes_do_until */
extern void T852f160(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_SYSTEM.classes_do_until */
extern void T40f200(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.classes_do_until */
extern void T58f273(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_SYSTEM_PROCESSOR.compile_all */
extern void T64f56(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_SYSTEM_PROCESSOR.compile_all */
extern T0* T64f56ac1(GE_context* ac, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.compile_classes */
extern void T64f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.compile_degree_3 */
extern void T64f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.report_degree_3_metrics */
extern void T64f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.report_degree_metrics */
extern void T64f103(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_SYSTEM_PROCESSOR.report_processor_metrics */
extern void T64f104(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.report_processed_class_count_stack */
extern void T64f105(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.total_processed_class_count */
extern T6 T64f40(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.set_implementation_internal_error */
extern void T64f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.check_implementation_validity */
extern void T64f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.reset_processed_class_count */
extern void T64f102(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.reset_total_processed_class_count */
extern void T64f101(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.compile_degree_4 */
extern void T64f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.report_degree_4_metrics */
extern void T64f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_interface_internal_error */
extern void T64f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.check_interface_validity */
extern void T64f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_flattening_internal_error */
extern void T64f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.flatten_features */
extern void T64f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_ancestors_internal_error */
extern void T64f87(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.build_ancestors */
extern void T64f86(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.compile_degree_5 */
extern void T64f81(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] ET_SYSTEM_PROCESSOR.check_provider_validity */
extern void T64f85(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.compile_degree_5_2 */
extern void T64f84(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] ET_SYSTEM_PROCESSOR.report_degree_5_2_metrics */
extern void T64f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_syntax_internal_error */
extern void T64f99(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.parse_classes */
extern void T64f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.remove_unmarked_classes */
extern void T64f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].keep_first */
extern void T462f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLASS].replace */
extern void T462f17(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] ET_CLASS.is_marked */
extern T1 T74f175(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.parse_marked_classes */
extern void T64f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.nested_benchmark_start_time */
extern T0* T64f43(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.compile_all */
extern void T375f58(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_SYSTEM_MULTIPROCESSOR.compile_all */
extern T0* T375f58ac1(GE_context* ac, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.compile_classes */
extern void T375f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.compile_degree_3 */
extern void T375f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.report_degree_3_metrics */
extern void T375f98(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.report_degree_metrics */
extern void T375f106(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.report_processor_metrics_recursive */
extern void T375f107(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.report_processor_metrics */
extern void T375f108(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.report_processed_class_count_stack */
extern void T375f109(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.total_processed_class_count */
extern T6 T375f42(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_SYSTEM_PROCESSOR].item */
extern T0* T703f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.total_processed_class_count_recursive */
extern T6 T375f41(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_implementation_internal_error */
extern void T375f97(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.check_implementation_validity */
extern void T375f96(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_SYSTEM_MULTIPROCESSOR.check_implementation_validity */
extern T0* T375f96ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.check_implementation_validity */
extern void T375f96p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.reset_processed_class_count */
extern void T375f105(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.reset_total_processed_class_count */
extern void T375f104(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.tokens */
extern T0* T375f47(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.compile_degree_4 */
extern void T375f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.report_degree_4_metrics */
extern void T375f95(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_interface_internal_error */
extern void T375f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.check_interface_validity */
extern void T375f93(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_flattening_internal_error */
extern void T375f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.flatten_features */
extern void T375f91(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_ancestors_internal_error */
extern void T375f90(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.build_ancestors */
extern void T375f89(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.compile_degree_5 */
extern void T375f81(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.check_provider_validity */
extern void T375f88(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.compile_degree_5_2 */
extern void T375f87(GE_context* ac, T0* C, T0* a1, T1 a2);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.report_degree_5_2_metrics */
extern void T375f103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_syntax_internal_error */
extern void T375f102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.parse_classes */
extern void T375f101(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_SYSTEM_MULTIPROCESSOR.parse_classes */
extern T0* T375f101ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.parse_classes */
extern void T375f101p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.remove_unmarked_classes */
extern void T375f100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.parse_marked_classes */
extern void T375f99(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_SYSTEM_MULTIPROCESSOR.parse_marked_classes */
extern T0* T375f99ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.parse_marked_classes */
extern void T375f99p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.nested_benchmark_start_time */
extern T0* T375f45(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.set_dynamic_type_set_builder */
extern void T63f83(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.make */
extern T0* T71c101(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].make_map */
extern T0* T381c33(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].make_with_equality_testers */
extern void T381f35(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].make_sparse_container */
extern void T381f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].set_internal_cursor */
extern void T381f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].new_cursor */
extern T0* T381f25(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].make */
extern T0* T707c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].make_slots */
extern void T381f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].make_clashes */
extern void T381f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].make_key_storage */
extern void T381f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] ET_DYNAMIC_TYPE].make_item_storage */
extern void T381f43(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.dummy_dynamic_feature */
extern T0* T71f82(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.dummy_feature */
extern T0* T71f95(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.dummy_dynamic_type */
extern T0* T71f71(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.make_feature_checker */
extern void T71f108(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_PUSH_TYPE_SET_BUILDER.make */
extern void T71f108p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.set_new_instance_types */
extern void T63f82(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_DYNAMIC_SYSTEM.set_catcall_warning_mode */
extern void T63f81(GE_context* ac, T0* C, T1 a1);
/* [detachable] GEC.catcall_warning_mode */
extern T1 T25f44(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.set_catcall_error_mode */
extern void T63f80(GE_context* ac, T0* C, T1 a1);
/* [detachable] GEC.catcall_error_mode */
extern T1 T25f43(GE_context* ac, T0* C);
/* [detachable] ET_DYNAMIC_SYSTEM.set_full_class_checking */
extern void T63f79(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_DYNAMIC_SYSTEM.make */
extern T0* T63c78(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #1 in feature [detachable] ET_DYNAMIC_SYSTEM.make */
extern T0* T63f78ac1(GE_context* ac, T6 a1);
/* [detachable] ET_DYNAMIC_NULL_TYPE_SET_BUILDER.make */
extern T0* T230c13(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].make */
extern T0* T226c33(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].make_with_equality_testers */
extern void T226f36(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].make_sparse_container */
extern void T226f44(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].set_internal_cursor */
extern void T226f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].new_cursor */
extern T0* T226f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].make */
extern T0* T615c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].make_slots */
extern void T226f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].make_clashes */
extern void T226f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].make_key_storage */
extern void T226f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_DYNAMIC_PRIMARY_TYPE, [attached] STRING_8].make_item_storage */
extern void T226f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_DYNAMIC_SYSTEM.make_basic_types */
extern void T63f85(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.compile */
extern void T64f55(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_SYSTEM_PROCESSOR.compile */
extern T0* T64f55ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_SYSTEM_PROCESSOR.compile */
extern T0* T64f55ac2(GE_context* ac, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.compile_marked_classes */
extern void T64f68(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.compile */
extern void T375f57(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_SYSTEM_MULTIPROCESSOR.compile */
extern T0* T375f57ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_SYSTEM_MULTIPROCESSOR.compile */
extern T0* T375f57ac2(GE_context* ac, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.compile_marked_classes */
extern void T375f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.benchmark_start_time */
extern T0* T64f41(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.benchmark_start_time */
extern T0* T375f43(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.set_flat_dbc_mode */
extern void T64f54(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_flat_dbc_mode_only */
extern void T64f66(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_flat_dbc_mode */
extern void T375f56(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_flat_dbc_mode_only */
extern void T375f68(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_flat_mode */
extern void T64f53(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_flat_mode_only */
extern void T64f65(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_flat_mode */
extern void T375f55(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_flat_mode_only */
extern void T375f67(GE_context* ac, T0* C, T1 a1);
/* [detachable] GEC.is_gelint */
extern T1 T25f42(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.set_metrics_shown */
extern void T64f52(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_metrics_shown_only */
extern void T64f64(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_metrics_shown */
extern void T375f54(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_metrics_shown_only */
extern void T375f66(GE_context* ac, T0* C, T1 a1);
/* [detachable] GEC.is_metrics */
extern T1 T25f41(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.set_nested_benchmark_shown */
extern void T64f51(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_nested_benchmark_shown_only */
extern void T64f63(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_nested_benchmark_shown */
extern void T375f53(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_nested_benchmark_shown_only */
extern void T375f65(GE_context* ac, T0* C, T1 a1);
/* [detachable] GEC.is_nested_benchmark */
extern T1 T25f40(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.set_benchmark_shown */
extern void T64f50(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_benchmark_shown_only */
extern void T64f62(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_benchmark_shown */
extern void T375f52(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_benchmark_shown_only */
extern void T375f64(GE_context* ac, T0* C, T1 a1);
/* [detachable] GEC.is_silent */
extern T1 T25f45(GE_context* ac, T0* C);
/* [detachable] GEC.is_no_benchmark */
extern T1 T25f39(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_PROCESSOR.set_ise_version */
extern void T64f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_ise_version_only */
extern void T64f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_ecma_version_only */
extern void T64f80(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_ise_version */
extern void T375f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_ise_version_only */
extern void T375f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_default_keyword_usage_only */
extern void T375f80(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_ecma_version_only */
extern void T375f79(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_use_detachable_keyword_only */
extern void T375f78(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_use_attached_keyword_only */
extern void T375f77(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_use_reference_keyword_only */
extern void T375f76(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_use_note_keyword_only */
extern void T375f75(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_use_attribute_keyword_only */
extern void T375f74(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_error_handler */
extern void T64f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_PROCESSOR.set_error_handler_only */
extern void T64f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_error_handler */
extern void T375f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_error_handler_only */
extern void T375f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] GEC.thread_count */
extern T6 T25f38(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.set_verbose */
extern void T53f182(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ERROR_HANDLER.set_verbose */
extern void T36f182(GE_context* ac, T0* C, T1 a1);
/* [detachable] GEC.is_verbose */
extern T1 T25f37(GE_context* ac, T0* C);
/* [detachable] ET_NULL_ERROR_HANDLER.set_ise */
extern void T53f181(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.set_ise */
extern void T36f181(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.close */
extern void T37f63(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.old_close */
extern void T37f65(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.default_pointer */
extern T14 T37f21(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.file_close */
extern void T37f70(GE_context* ac, T0* C, T14 a1);
/* [detachable] GEC.parse_ecf_file */
extern void T25f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.to_boolean */
extern T1 T775f68(GE_context* ac, T0* C);
/* [detachable] UC_STRING.to_boolean */
extern T1 T429f68(GE_context* ac, T0* C);
/* [detachable] STRING_8.to_boolean */
extern T1 T17f46(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.is_boolean */
extern T1 T775f66(GE_context* ac, T0* C);
/* [detachable] UC_STRING.is_boolean */
extern T1 T429f67(GE_context* ac, T0* C);
/* [detachable] STRING_8.is_boolean */
extern T1 T17f45(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ERROR_HANDLER.has_error */
extern T1 T56f5(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.parse_file */
extern void T55f214(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM_PARSER.parse_file */
extern T0* T55f214ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature [detachable] ET_ECF_SYSTEM_PARSER.parse_file */
extern T0* T55f214ac2(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] XM_ELEMENT, detachable XM_POSITION_TABLE, [attached] STRING_8]].set_rout_disp_final */
extern void T171f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_ECF_SYSTEM_PARSER.build_system */
extern void T55f217(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM_PARSER.build_system */
extern T0* T55f217ac1(GE_context* ac);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.propagate_read_only */
extern void T207f35(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_read_only */
extern void T187f127(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_LIBRARY.set_read_only */
extern void T852f167(GE_context* ac, T0* C, T1 a1);
/* [detachable] DS_CELL [detachable ET_ECF_SYSTEM].put */
extern void T163f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].wipe_out */
extern void T176f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].slots_wipe_out */
extern void T176f45(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].clashes_wipe_out */
extern void T176f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].key_storage_wipe_out */
extern void T176f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].item_storage_wipe_out */
extern void T176f42(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_ECF_DOTNET_ASSEMBLY].keep_head */
extern void T511f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_DOTNET_ASSEMBLY].set_count */
extern void T511f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].unset_found_item */
extern void T176f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].move_all_cursors_after */
extern void T176f40(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].set_next_cursor */
extern void T514f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].set_after */
extern void T514f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].wipe_out */
extern void T174f44(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].slots_wipe_out */
extern void T174f55(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].clashes_wipe_out */
extern void T174f54(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].key_storage_wipe_out */
extern void T174f53(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].item_storage_wipe_out */
extern void T174f52(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] ET_ECF_LIBRARY].keep_head */
extern void T507f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_LIBRARY].set_count */
extern void T507f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].unset_found_item */
extern void T174f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].move_all_cursors_after */
extern void T174f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_LIBRARY, [attached] STRING_8].set_next_cursor */
extern void T506f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_LIBRARY, [attached] STRING_8].set_after */
extern void T506f9(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.parse_dotnet_assemblies */
extern void T55f221(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].force_last_new */
extern void T176f38(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].key_storage_put */
extern void T176f51(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].item_storage_put */
extern void T176f50(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_DOTNET_ASSEMBLY].force */
extern void T515f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_DOTNET_ASSEMBLY].fill_with */
extern void T511f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_DOTNET_ASSEMBLY].extend */
extern void T511f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].slots_put */
extern void T176f49(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].clashes_put */
extern void T176f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].slots_item */
extern T6 T176f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].hash_position */
extern T6 T176f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].resize */
extern void T176f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].clashes_resize */
extern void T176f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].special_integer_ */
extern T0* T176f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].key_storage_resize */
extern void T176f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].item_storage_resize */
extern void T176f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_DOTNET_ASSEMBLY].aliased_resized_area */
extern T0* T515f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_DOTNET_ASSEMBLY].aliased_resized_area */
extern T0* T511f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].key_storage_item */
extern T0* T176f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].clashes_item */
extern T6 T176f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].slots_resize */
extern void T176f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].new_modulus */
extern T6 T176f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].new_capacity */
extern T6 T176f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_dotnet_assembly */
extern T0* T162f2(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.make */
extern T0* T211c105(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.make_from_system */
extern void T211f106(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_kernel_types */
extern void T211f110(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_wide_character_type */
extern void T211f152(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_typed_pointer_type */
extern void T211f151(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_type_type */
extern void T211f150(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_system_string_type */
extern void T211f149(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_system_object_type */
extern void T211f148(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_string_type */
extern void T211f147(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_string_32_type */
extern void T211f146(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_string_8_type */
extern void T211f145(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_special_type */
extern void T211f144(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_routine_type */
extern void T211f143(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_real_64_type */
extern void T211f142(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_real_32_type */
extern void T211f141(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_real_type */
extern void T211f140(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_procedure_type */
extern void T211f139(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_predicate_type */
extern void T211f138(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_pointer_type */
extern void T211f137(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_natural_64_type */
extern void T211f136(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_natural_32_type */
extern void T211f135(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_natural_16_type */
extern void T211f134(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_natural_8_type */
extern void T211f133(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_natural_type */
extern void T211f132(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_iterable_type */
extern void T211f131(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_ise_exception_manager_type */
extern void T211f130(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_integer_64_type */
extern void T211f129(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_integer_32_type */
extern void T211f128(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_integer_16_type */
extern void T211f127(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_integer_8_type */
extern void T211f126(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_integer_type */
extern void T211f125(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_immutable_string_32_type */
extern void T211f124(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_immutable_string_8_type */
extern void T211f123(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_function_type */
extern void T211f122(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_exception_manager_type */
extern void T211f121(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_exception_type */
extern void T211f120(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_double_type */
extern void T211f119(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_character_32_type */
extern void T211f118(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_character_8_type */
extern void T211f117(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_character_type */
extern void T211f116(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_boolean_type */
extern void T211f115(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_array_type */
extern void T211f114(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_tuple_type */
extern void T211f113(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_none_type */
extern void T211f112(GE_context* ac, T0* C);
/* [detachable] ET_MASTER_CLASS.add_first_imported_class */
extern void T570f52(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].force_first */
extern void T1389f15(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].put */
extern void T1389f23(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].move_cursors_right */
extern void T1389f30(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].move_right */
extern void T1389f29(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_ARRAYED_LIST [[attached] ET_MASTER_CLASS].put_last */
extern void T1389f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_any_type */
extern void T211f111(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.make_adapted */
extern void T211f109(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.default_read_only_value */
extern T1 T211f90(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.set_unknown_kernel_types */
extern void T211f108(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.initialize */
extern void T211f107(GE_context* ac, T0* C);
/* [detachable] ET_ECF_DOTNET_ASSEMBLY.class_name_tester */
extern T0* T211f89(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_dotnet_assembly */
extern void T213f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].found_item */
extern T0* T176f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].item_storage_item */
extern T0* T176f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].found */
extern T1 T176f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].search */
extern void T176f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].search_position */
extern void T176f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.operating_system */
extern T0* T55f26(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.full_filename */
extern T0* T55f27(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM_PARSER.windows_file_system */
extern T0* T55f19(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.file_system */
extern T0* T55f18(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.unix_file_system */
extern T0* T55f32(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.pathname */
extern T0* T213f22(GE_context* ac, T0* C);
/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.dotnet_assembly */
extern T0* T210f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_TARGET.fill_class_mappings */
extern void T59f56(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.set_class_mapping */
extern void T58f249(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_MASTER_CLASS.set_mapped_class */
extern void T570f80(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.set_class_mapping */
extern void T187f182(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].new_iterator */
extern T0* T216f43(GE_context* ac, T0* C);
/* [detachable] ET_ECF_TARGET.fill_options */
extern void T59f55(GE_context* ac, T0* C, T0* a1);
extern T1 T59f55ot1(T0* a1, T0** a2);
/* [detachable] ET_ECF_SYSTEM.set_trace_mode */
extern void T58f248(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_SYSTEM.set_obsolete_routine_type_mode */
extern void T58f247(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_LIBRARY.set_obsolete_routine_type_mode */
extern void T187f181(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_SYSTEM.set_obsolete_iteration_mode */
extern void T58f246(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_LIBRARY.set_obsolete_iteration_mode */
extern void T187f180(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_SYSTEM.set_implicit_attachment_type_mark */
extern void T58f245(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.set_implicit_attachment_type_mark */
extern void T187f179(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.tokens */
extern T0* T59f29(GE_context* ac, T0* C);
/* [detachable] ET_ECF_OPTIONS.value */
extern T0* T444f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.fill_capabilities */
extern void T59f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_target_type_attachment_mode */
extern void T58f256(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_SYSTEM.set_attachment_type_conformance_mode */
extern void T58f255(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_SYSTEM.set_multithreaded_mode */
extern void T58f254(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_TARGET.string_ */
extern T0* T59f30(GE_context* ac, T0* C);
/* detachable ET_ECF_CAPABILITIES.support_value */
extern T0* T61f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CAPABILITIES.use_value */
extern T0* T61f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].found_item */
extern T0* T221f32(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].item_storage_item */
extern T0* T221f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].found */
extern T1 T221f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].search */
extern void T221f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].search_position */
extern void T221f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].hash_position */
extern T6 T221f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].key_storage_item */
extern T0* T221f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].clashes_item */
extern T6 T221f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].slots_item */
extern T6 T221f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_TARGET.fill_settings */
extern void T59f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_line_generation_mode */
extern void T58f253(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_SYSTEM.set_total_order_on_reals_mode */
extern void T58f252(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_SYSTEM.set_exception_trace_mode */
extern void T58f251(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_SYSTEM.set_console_application_mode */
extern void T58f250(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_TARGET.fill_root */
extern void T59f60(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ROOT_CLASS.fill_root */
extern void T816f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_root_creation */
extern void T58f270(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_root_type */
extern void T58f269(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ROOT_ALL_CLASSES.fill_root */
extern void T817f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.unset_root_type */
extern void T58f271(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].forth */
extern void T174f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].cursor_forth */
extern void T174f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].add_traversing_cursor */
extern void T174f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_LIBRARY, [attached] STRING_8].set_position */
extern void T506f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].remove_traversing_cursor */
extern void T174f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].clashes_item */
extern T6 T174f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].item_for_iteration */
extern T0* T174f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].cursor_item */
extern T0* T174f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].item_storage_item */
extern T0* T174f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].after */
extern T1 T174f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].cursor_after */
extern T1 T174f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_LIBRARY, [attached] STRING_8].after */
extern T1 T506f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].start */
extern void T174f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].cursor_start */
extern void T174f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].cursor_off */
extern T1 T174f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_LIBRARY, [attached] STRING_8].off */
extern T1 T506f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].is_empty */
extern T1 T174f23(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.parse_libraries */
extern void T55f220(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM_PARSER.parse_libraries */
extern T0* T55f220ac1(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_SYSTEM_PARSER.build_library */
extern void T55f223(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] DS_CELL [detachable ET_ECF_LIBRARY].put */
extern void T206f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_library */
extern T0* T55f30(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].force_last_new */
extern void T174f46(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].key_storage_put */
extern void T174f61(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].item_storage_put */
extern void T174f60(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_LIBRARY].force */
extern void T510f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_LIBRARY].fill_with */
extern void T507f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_LIBRARY].extend */
extern void T507f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].slots_put */
extern void T174f59(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].clashes_put */
extern void T174f58(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].slots_item */
extern T6 T174f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].hash_position */
extern T6 T174f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].resize */
extern void T174f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].clashes_resize */
extern void T174f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].special_integer_ */
extern T0* T174f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].key_storage_resize */
extern void T174f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].item_storage_resize */
extern void T174f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_LIBRARY].aliased_resized_area */
extern T0* T510f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_LIBRARY].aliased_resized_area */
extern T0* T507f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].key_storage_item */
extern T0* T174f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].slots_resize */
extern void T174f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].new_modulus */
extern T6 T174f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].new_capacity */
extern T6 T174f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.fill_system_config */
extern void T55f224(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T55f224ot1(T0* a1, T0** a2);
extern T1 T55f224ot2(T0* a1, T0** a2);
/* [detachable] ET_ECF_LIBRARY.set_library_target */
extern void T187f129(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_CONFIG.set_library_target */
extern void T185f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_library_target */
extern void T58f172(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eltm_error */
extern void T56f37(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR_HANDLER.report_error */
extern void T56f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ERROR_HANDLER.report_error_message */
extern void T56f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ERROR_HANDLER.report_error_message */
extern void T56f32p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ERROR_HANDLER.message */
extern T0* T56f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ERROR.default_message */
extern T0* T214f9(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ERROR.message */
extern T0* T214f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ERROR.arguments */
extern T0* T214f6(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ERROR.string_ */
extern T0* T214f8(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ERROR.make_eltm */
extern T0* T214c105(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.filename */
extern T0* T214f14(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ERROR.default_message_template */
extern T0* T214f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ERROR.add_dependent_universes */
extern void T214f116(GE_context* ac, T0* C, T0* a1);
extern T1 T214f116ot1(T0* a1, T0** a2);
/* [detachable] ET_ECF_TARGET.parent_path */
extern T0* T59f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].put_last */
extern void T909f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET_PARENT].force */
extern void T1636f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_TARGET_PARENT].fill_with */
extern void T1635f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_TARGET_PARENT].extend */
extern void T1635f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].make */
extern void T909f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].make */
extern T0* T909c13(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].set_internal_cursor */
extern void T909f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].new_cursor */
extern T0* T909f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_TARGET_PARENT].make */
extern T0* T1637c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET_PARENT].make */
extern T0* T1636f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_TARGET_PARENT].make_empty */
extern T0* T1635c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET_PARENT].default_create */
extern T0* T1636c3(GE_context* ac);
/* [detachable] ET_ECF_ERROR.add_parent_path */
extern void T214f117(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].item */
extern T0* T909f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].force_last */
extern void T909f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].resize */
extern void T909f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET_PARENT].aliased_resized_area */
extern T0* T1636f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_TARGET_PARENT].aliased_resized_area */
extern T0* T1635f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].new_capacity */
extern T6 T909f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].extendible */
extern T1 T909f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].make_default */
extern T0* T909c11(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET_PARENT].default_capacity */
extern T6 T909f8(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.library */
extern T0* T187f12(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_CONFIG.universe */
extern T0* T185f3(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.universe */
extern T0* T58f28(GE_context* ac, T0* C);
/* detachable ET_ECF_TARGETS.target_by_name */
extern T0* T203f2(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_TARGETS.string_ */
extern T0* T203f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET].item */
extern T0* T906f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_ERROR_HANDLER.report_etam_error */
extern void T56f36(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR.make_etam */
extern T0* T214c104(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_ECF_TARGETS.is_empty */
extern T1 T203f5(GE_context* ac, T0* C);
/* detachable ET_ECF_TARGETS.count */
extern T6 T203f3(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].forth */
extern void T447f12(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].cursor_forth */
extern void T894f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].add_traversing_cursor */
extern void T894f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].set_next_cursor */
extern void T447f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].remove_traversing_cursor */
extern void T894f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].set */
extern void T447f15(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] ET_ECF_LIBRARY.set_uuid */
extern void T187f128(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_CONFIG.set_uuid */
extern void T185f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_uuid */
extern void T58f171(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eatb_error */
extern void T56f35(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ERROR.make_eatb */
extern T0* T214c103(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_SYSTEM_PARSER.attribute_value */
extern T0* T55f51(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable XM_POSITION_TABLE.item */
extern T0* T169f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].forth */
extern void T898f11(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].cursor_forth */
extern void T899f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].add_traversing_cursor */
extern void T899f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].set_next_cursor */
extern void T898f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].remove_traversing_cursor */
extern void T899f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].set */
extern void T898f12(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].go_after */
extern void T898f10(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].cursor_go_after */
extern void T899f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].set_after */
extern void T898f13(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].cursor_off */
extern T1 T899f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].off */
extern T1 T898f7(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].item */
extern T0* T898f6(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].start */
extern void T898f9(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].cursor_start */
extern void T899f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].new_cursor */
extern T0* T899f5(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].make */
extern T0* T898c8(GE_context* ac, T0* a1);
/* detachable XM_POSITION_TABLE.has */
extern T1 T169f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.is_false */
extern T1 T55f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_CONFIG.set_read_only */
extern void T185f18(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_SYSTEM.set_read_only */
extern void T58f170(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.is_true */
extern T1 T55f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.set_targets */
extern void T187f126(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_CONFIG.set_targets */
extern void T185f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_targets */
extern void T58f169(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_targets */
extern T0* T55f46(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8);
/* detachable ET_ECF_TARGETS.put_last */
extern void T203f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET].force_last */
extern void T906f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET].force */
extern void T589f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_TARGET].fill_with */
extern void T587f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_TARGET].extend */
extern void T587f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET].resize */
extern void T906f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET].aliased_resized_area */
extern T0* T589f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_TARGET].aliased_resized_area */
extern T0* T587f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET].new_capacity */
extern T6 T906f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET].extendible */
extern T1 T906f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_targets */
extern T0* T162f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_TARGETS.make */
extern T0* T203c8(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET].put_last */
extern void T906f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET].make */
extern T0* T906c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET].set_internal_cursor */
extern void T906f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET].new_cursor */
extern T0* T906f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_TARGET].make */
extern T0* T1631c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET].make */
extern T0* T589f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_TARGET].make_empty */
extern T0* T587c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_TARGET].default_create */
extern T0* T589c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_target */
extern T0* T55f77(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6, T0* a7);
extern T1 T55f77ot1(T0* a1, T0** a2);
extern T1 T55f77ot2(T0* a1, T0** a2);
/* [detachable] ET_ECF_TARGET.set_notes */
extern void T59f53(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_post_compile_actions */
extern void T59f52(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_pre_compile_actions */
extern void T59f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_external_resources */
extern void T59f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_external_objects */
extern void T59f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_external_makes */
extern void T59f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_external_linker_flags */
extern void T59f47(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_external_libraries */
extern void T59f46(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_external_includes */
extern void T59f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_external_cflags */
extern void T59f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_class_mappings */
extern void T59f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_file_rules */
extern void T59f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_dotnet_assemblies */
extern void T59f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_libraries */
extern void T59f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_clusters */
extern void T59f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.adapt_target_options */
extern void T55f244(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.adapt_options */
extern void T55f246(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_OPTIONS.set_primary_value */
extern void T444f15(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_OPTIONS.primary_value */
extern T0* T444f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_SETTINGS.primary_value */
extern T0* T57f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.adapt_capabilities */
extern void T55f243(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM_PARSER.ecf_1_5_0 */
extern T0* T55f73(GE_context* ac, T0* C);
/* detachable ET_ECF_CAPABILITIES.set_primary_use_value */
extern void T61f14(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_CAPABILITIES.set_primary_use_value_id */
extern void T61f15(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].force_last */
extern void T221f42(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].key_storage_put */
extern void T221f50(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].slots_put */
extern void T221f49(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].clashes_put */
extern void T221f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].resize */
extern void T221f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].clashes_resize */
extern void T221f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].special_integer_ */
extern T0* T221f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].key_storage_resize */
extern void T221f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].item_storage_resize */
extern void T221f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].slots_resize */
extern void T221f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].new_modulus */
extern T6 T221f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].new_capacity */
extern T6 T221f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].item_storage_put */
extern void T221f46(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].unset_found_item */
extern void T221f44(GE_context* ac, T0* C);
/* detachable ET_ECF_CAPABILITIES.set_primary_support_value */
extern void T61f16(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM_PARSER.ecf_1_11_0 */
extern T0* T55f59(GE_context* ac, T0* C);
/* detachable ET_ECF_CAPABILITIES.primary_use_value */
extern T0* T61f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CAPABILITIES.primary_use_value_id */
extern T0* T61f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CAPABILITIES.primary_support_value */
extern T0* T61f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.set_abstract */
extern void T59f38(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_TARGET.set_version */
extern void T59f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_version */
extern T0* T55f120(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_VERSION.set_copyright */
extern void T801f14(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VERSION.set_company */
extern void T801f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VERSION.set_product */
extern void T801f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VERSION.make_major */
extern T0* T801c11(GE_context* ac, T6 a1);
/* detachable ET_ECF_VERSION.make_major_minor */
extern T0* T801c10(GE_context* ac, T6 a1, T6 a2);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eats_error */
extern void T56f39(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ERROR.make_eats */
extern T0* T214c107(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_VERSION.make_major_minor_release */
extern T0* T801c9(GE_context* ac, T6 a1, T6 a2, T6 a3);
/* detachable ET_ECF_VERSION.make */
extern T0* T801c8(GE_context* ac, T6 a1, T6 a2, T6 a3, T6 a4);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eati_error */
extern void T56f38(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ERROR.make_eati */
extern T0* T214c106(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_SYSTEM_PARSER.add_variable */
extern void T55f242(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_VARIABLES.set_primary_value */
extern void T62f21(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_VARIABLES.primary_value */
extern T0* T62f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.add_setting */
extern void T55f241(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_TARGET.set_root */
extern void T59f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_root */
extern T0* T55f116(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_ROOT_CLASS.set_cluster_name */
extern void T816f6(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ROOT_CLASS.set_creation_procedure_name */
extern void T816f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_root_class */
extern T0* T162f24(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ROOT_CLASS.make */
extern T0* T816c4(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_root_all_classes */
extern T0* T162f23(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ROOT_ALL_CLASSES.make */
extern T0* T817c1(GE_context* ac);
/* [detachable] ET_ECF_TARGET.set_precompiled_library */
extern void T59f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_adapted_precompiled_library */
extern T0* T55f114(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM_PARSER.new_adapted_precompiled_library */
extern T0* T55f114ac1(GE_context* ac, T0* a1);
extern T1 T55f114ot1(T0* a1, T0** a2);
extern T1 T55f114ot2(T0* a1, T0** a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_OPTIONS]].set_rout_disp_final */
extern void T815f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.set_conditions */
extern void T799f29(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.set_visible_classes */
extern void T799f28(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.set_class_renamings */
extern void T799f27(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.set_class_options */
extern void T799f26(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].do_all */
extern void T812f35(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].item_storage_item */
extern T0* T812f33(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].clashes_item */
extern T6 T812f28(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.set_options */
extern void T799f25(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.set_use_application_options */
extern void T799f24(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.set_read_only */
extern void T799f23(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.set_classname_prefix */
extern void T799f22(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.set_eifgens_location */
extern void T799f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_visible_classes */
extern T0* T55f143(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_VISIBLE_CLASS].force_last */
extern void T813f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_VISIBLE_CLASS].force */
extern void T1599f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_VISIBLE_CLASS].fill_with */
extern void T1600f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_VISIBLE_CLASS].extend */
extern void T1600f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_VISIBLE_CLASS].resize */
extern void T813f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_VISIBLE_CLASS].aliased_resized_area */
extern T0* T1599f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_VISIBLE_CLASS].aliased_resized_area */
extern T0* T1600f4(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_VISIBLE_CLASS].new_capacity */
extern T6 T813f7(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_VISIBLE_CLASS].extendible */
extern T1 T813f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_visible_classes */
extern T0* T162f41(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_VISIBLE_CLASS].make */
extern T0* T813c9(GE_context* ac, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_VISIBLE_CLASS].set_internal_cursor */
extern void T813f11(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_VISIBLE_CLASS].new_cursor */
extern T0* T813f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_VISIBLE_CLASS].make */
extern T0* T1601c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_VISIBLE_CLASS].make */
extern T0* T1599f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_VISIBLE_CLASS].make_empty */
extern T0* T1600c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_VISIBLE_CLASS].default_create */
extern T0* T1599c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_visible_class */
extern T0* T55f175(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_VISIBLE_CLASS.set_new_feature_name */
extern void T820f8(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VISIBLE_CLASS.set_new_class_name */
extern void T820f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VISIBLE_CLASS.set_feature_name */
extern void T820f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_visible_class */
extern T0* T162f44(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VISIBLE_CLASS.make */
extern T0* T820c5(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_renamings */
extern T0* T55f141(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_SYSTEM_PARSER.add_renaming */
extern void T55f247(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eatd_error */
extern void T56f47(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ERROR.make_eatd */
extern T0* T214c115(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_AST_FACTORY.new_class_renamings */
extern T0* T162f40(GE_context* ac, T0* C);
/* [detachable] ET_ECF_AST_FACTORY.case_insensitive_string_equality_tester */
extern T0* T162f52(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_options */
extern T0* T55f139(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ECF_AST_FACTORY.new_options */
extern T0* T162f39(GE_context* ac, T0* C);
/* [detachable] ET_ECF_OPTIONS.make */
extern T0* T444c14(GE_context* ac);
/* Creation of agent #1 in feature [detachable] ET_ECF_OPTIONS.make */
extern T0* T444f14ac1(GE_context* ac, T0* a1);
/* [detachable] FUNCTION [[attached] TUPLE [[attached] STRING_8], [attached] INTEGER_32].set_rout_disp_final */
extern void T218f8(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].set_hash_function */
extern void T216f50(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_AGENT_HASH_FUNCTION [[attached] STRING_8].make */
extern T0* T215c3(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_OPTIONS.string_ */
extern T0* T444f10(GE_context* ac, T0* C);
/* [detachable] ET_ECF_OPTIONS.case_insensitive_string_equality_tester */
extern T0* T444f9(GE_context* ac, T0* C);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.set_description */
extern void T799f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_class_options */
extern T0* T55f138(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].force_last */
extern void T812f37(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].key_storage_put */
extern void T812f46(GE_context* ac, T0* C, T0* a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].slots_put */
extern void T812f45(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].clashes_put */
extern void T812f44(GE_context* ac, T0* C, T6 a1, T6 a2);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].slots_item */
extern T6 T812f26(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].hash_position */
extern T6 T812f23(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].resize */
extern void T812f43(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].clashes_resize */
extern void T812f51(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].special_integer_ */
extern T0* T812f31(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].key_storage_resize */
extern void T812f50(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].item_storage_resize */
extern void T812f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_OPTIONS].aliased_resized_area */
extern T0* T1597f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_OPTIONS].aliased_resized_area */
extern T0* T1594f4(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].key_storage_item */
extern T0* T812f22(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].slots_resize */
extern void T812f48(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].new_modulus */
extern T6 T812f24(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].new_capacity */
extern T6 T812f21(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].item_storage_put */
extern void T812f42(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_OPTIONS].force */
extern void T1597f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_OPTIONS].fill_with */
extern void T1594f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_OPTIONS].extend */
extern void T1594f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].search_position */
extern void T812f40(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].unset_found_item */
extern void T812f41(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].found_item */
extern T0* T812f25(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].found */
extern T1 T812f27(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].search */
extern void T812f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_class_options */
extern T0* T162f38(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].set_key_equality_tester */
extern void T812f38(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].make_map */
extern T0* T812c34(GE_context* ac, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].make_with_equality_testers */
extern void T812f39(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].make_sparse_container */
extern void T812f47(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].set_internal_cursor */
extern void T812f56(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].new_cursor */
extern T0* T812f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_OPTIONS, [attached] STRING_8].make */
extern T0* T1598c4(GE_context* ac, T0* a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].make_slots */
extern void T812f55(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].make_clashes */
extern void T812f54(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].make_key_storage */
extern void T812f53(GE_context* ac, T0* C, T6 a1);
/* detachable DS_HASH_TABLE [[attached] ET_ECF_OPTIONS, [attached] STRING_8].make_item_storage */
extern void T812f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_OPTIONS].make */
extern T0* T1597f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_OPTIONS].make_empty */
extern T0* T1594c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_OPTIONS].default_create */
extern T0* T1597c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_conditions */
extern T0* T55f136(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T55f136ot1(T0* a1, T0** a2);
extern T1 T55f136ot2(T0* a1, T0** a2);
/* [detachable] ET_ECF_ORED_CONDITIONS.put_last */
extern void T818f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS].force_last */
extern void T1602f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ANDED_CONDITIONS].force */
extern void T1895f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_ANDED_CONDITIONS].fill_with */
extern void T1896f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_ANDED_CONDITIONS].extend */
extern void T1896f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS].resize */
extern void T1602f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ANDED_CONDITIONS].aliased_resized_area */
extern T0* T1895f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_ANDED_CONDITIONS].aliased_resized_area */
extern T0* T1896f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS].new_capacity */
extern T6 T1602f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS].extendible */
extern T1 T1602f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_ored_conditions */
extern T0* T162f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ORED_CONDITIONS.make */
extern T0* T818c6(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS].put_last */
extern void T1602f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS].make */
extern T0* T1602c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS].set_internal_cursor */
extern void T1602f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS].new_cursor */
extern T0* T1602f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ANDED_CONDITIONS].make */
extern T0* T1897c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ANDED_CONDITIONS].make */
extern T0* T1895f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_ANDED_CONDITIONS].make_empty */
extern T0* T1896c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ANDED_CONDITIONS].default_create */
extern T0* T1895c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_condition */
extern T0* T55f174(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T55f174ot1(T0* a1, T0** a2);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_version_condition */
extern T0* T55f192(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eabj_error */
extern void T56f46(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_eabj */
extern T0* T214c114(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_MSIL_CLR_VERSION_CONDITION.make */
extern T0* T831c7(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_COMPILER_VERSION_CONDITION.make */
extern T0* T830c7(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eabi_error */
extern void T56f45(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR.make_eabi */
extern T0* T214c113(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eabh_error */
extern void T56f44(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ECF_ERROR.make_eabh */
extern T0* T214c112(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eabg_error */
extern void T56f43(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_eabg */
extern T0* T214c111(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eabf_error */
extern void T56f42(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_eabf */
extern T0* T214c110(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] UT_VERSION.make_major_minor_revision */
extern T0* T46c21(GE_context* ac, T6 a1, T6 a2, T6 a3);
/* [detachable] UT_VERSION.make_major_minor */
extern T0* T46c18(GE_context* ac, T6 a1, T6 a2);
/* [detachable] UT_VERSION.make_major */
extern T0* T46c17(GE_context* ac, T6 a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_platform_condition */
extern T0* T55f191(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eatn_error */
extern void T56f41(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ERROR.make_eatn */
extern T0* T214c109(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_AST_FACTORY.new_platform_condition */
extern T0* T162f51(GE_context* ac, T0* C, T0* a1, T1 a2);
/* detachable ET_ECF_PLATFORM_CONDITION.make */
extern T0* T828c9(GE_context* ac, T0* a1);
/* detachable ET_ECF_PLATFORM_CONDITION.make_excluded */
extern T0* T828c8(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eatx_error */
extern void T56f40(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ERROR.make_eatx */
extern T0* T214c108(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_multithreaded_condition */
extern T0* T55f189(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_AST_FACTORY.new_concurrency_condition */
extern T0* T162f46(GE_context* ac, T0* C, T0* a1, T1 a2);
/* detachable ET_ECF_CONCURRENCY_CONDITION.make */
extern T0* T823c18(GE_context* ac, T0* a1);
/* detachable ET_ECF_CONCURRENCY_CONDITION.make_excluded */
extern T0* T823c17(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_dynamic_runtime_condition */
extern T0* T55f187(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_AST_FACTORY.new_dynamic_runtime_condition */
extern T0* T162f50(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_DYNAMIC_RUNTIME_CONDITION.make */
extern T0* T827c5(GE_context* ac, T1 a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_dotnet_condition */
extern T0* T55f185(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_AST_FACTORY.new_dotnet_condition */
extern T0* T162f49(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_DOTNET_CONDITION.make */
extern T0* T826c5(GE_context* ac, T1 a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_custom_condition */
extern T0* T55f183(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_AST_FACTORY.new_custom_condition */
extern T0* T162f48(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T1 a4);
/* detachable ET_ECF_CUSTOM_CONDITION.make */
extern T0* T825c13(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_CUSTOM_CONDITION.make_excluded */
extern T0* T825c12(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_void_safety_condition */
extern T0* T55f181(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_AST_FACTORY.new_void_safety_condition */
extern T0* T162f47(GE_context* ac, T0* C, T0* a1, T1 a2);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.make */
extern T0* T824c18(GE_context* ac, T0* a1);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.make_excluded */
extern T0* T824c17(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_concurrency_condition */
extern T0* T55f179(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_ANDED_CONDITIONS.put_last */
extern void T819f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CONDITION_ITEM].force_last */
extern void T1603f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CONDITION_ITEM].force */
extern void T1898f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_CONDITION_ITEM].fill_with */
extern void T1899f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_CONDITION_ITEM].extend */
extern void T1899f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CONDITION_ITEM].resize */
extern void T1603f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CONDITION_ITEM].aliased_resized_area */
extern T0* T1898f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_CONDITION_ITEM].aliased_resized_area */
extern T0* T1899f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CONDITION_ITEM].new_capacity */
extern T6 T1603f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CONDITION_ITEM].extendible */
extern T1 T1603f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_build_condition */
extern T0* T55f177(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_AST_FACTORY.new_build_condition */
extern T0* T162f45(GE_context* ac, T0* C, T0* a1, T1 a2);
/* detachable ET_ECF_BUILD_CONDITION.make */
extern T0* T822c8(GE_context* ac, T0* a1);
/* detachable ET_ECF_BUILD_CONDITION.make_excluded */
extern T0* T822c7(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_anded_conditions */
extern T0* T162f43(GE_context* ac, T0* C);
/* detachable ET_ECF_ANDED_CONDITIONS.make_empty */
extern T0* T819c6(GE_context* ac);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CONDITION_ITEM].make */
extern T0* T1603c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CONDITION_ITEM].set_internal_cursor */
extern void T1603f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CONDITION_ITEM].new_cursor */
extern T0* T1603f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_CONDITION_ITEM].make */
extern T0* T1900c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CONDITION_ITEM].make */
extern T0* T1898f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_CONDITION_ITEM].make_empty */
extern T0* T1899c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_CONDITION_ITEM].default_create */
extern T0* T1898c3(GE_context* ac);
/* [detachable] ET_ECF_AST_FACTORY.new_adapted_precompiled_library */
extern T0* T162f22(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.make */
extern T0* T799c19(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.make_adapted_library */
extern void T799f30(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.default_read_only_value */
extern T1 T799f18(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_library */
extern T0* T73f231(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.make */
extern T0* T852c106(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_LIBRARY.make_from_system */
extern void T852f107(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIBRARY.set_kernel_types */
extern void T852f111(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_wide_character_type */
extern void T852f153(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.tokens */
extern T0* T852f95(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_typed_pointer_type */
extern void T852f152(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_type_type */
extern void T852f151(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_system_string_type */
extern void T852f150(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_system_object_type */
extern void T852f149(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_string_type */
extern void T852f148(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_string_32_type */
extern void T852f147(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_string_8_type */
extern void T852f146(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_special_type */
extern void T852f145(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_routine_type */
extern void T852f144(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_real_64_type */
extern void T852f143(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_real_32_type */
extern void T852f142(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_real_type */
extern void T852f141(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_procedure_type */
extern void T852f140(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_predicate_type */
extern void T852f139(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_pointer_type */
extern void T852f138(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_natural_64_type */
extern void T852f137(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_natural_32_type */
extern void T852f136(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_natural_16_type */
extern void T852f135(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_natural_8_type */
extern void T852f134(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_natural_type */
extern void T852f133(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_iterable_type */
extern void T852f132(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_ise_exception_manager_type */
extern void T852f131(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_integer_64_type */
extern void T852f130(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_integer_32_type */
extern void T852f129(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_integer_16_type */
extern void T852f128(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_integer_8_type */
extern void T852f127(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_integer_type */
extern void T852f126(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_immutable_string_32_type */
extern void T852f125(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_immutable_string_8_type */
extern void T852f124(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_function_type */
extern void T852f123(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_exception_manager_type */
extern void T852f122(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_exception_type */
extern void T852f121(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_double_type */
extern void T852f120(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_character_32_type */
extern void T852f119(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_character_8_type */
extern void T852f118(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_character_type */
extern void T852f117(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_boolean_type */
extern void T852f116(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_array_type */
extern void T852f115(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_tuple_type */
extern void T852f114(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_none_type */
extern void T852f113(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_any_type */
extern void T852f112(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.make_adapted */
extern void T852f110(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_LIBRARY.default_read_only_value */
extern T1 T852f91(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.set_unknown_kernel_types */
extern void T852f109(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.initialize */
extern void T852f108(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.initialize */
extern void T852f108p1(GE_context* ac, T0* C);
/* [detachable] ET_LIBRARY.class_name_tester */
extern T0* T852f90(GE_context* ac, T0* C);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.tokens */
extern T0* T799f17(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_actions */
extern T0* T55f111(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_ACTION].force_last */
extern void T798f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ACTION].force */
extern void T1590f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_ACTION].fill_with */
extern void T1591f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_ACTION].extend */
extern void T1591f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_ACTION].resize */
extern void T798f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ACTION].aliased_resized_area */
extern T0* T1590f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_ACTION].aliased_resized_area */
extern T0* T1591f4(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_ACTION].new_capacity */
extern T6 T798f7(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_ACTION].extendible */
extern T1 T798f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_actions */
extern T0* T162f21(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_ACTION].make */
extern T0* T798c9(GE_context* ac, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_ACTION].set_internal_cursor */
extern void T798f11(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_ACTION].new_cursor */
extern T0* T798f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ACTION].make */
extern T0* T1592c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ACTION].make */
extern T0* T1590f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_ACTION].make_empty */
extern T0* T1591c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ACTION].default_create */
extern T0* T1590c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_action */
extern T0* T55f134(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T55f134ot1(T0* a1, T0** a2);
extern T1 T55f134ot2(T0* a1, T0** a2);
/* detachable ET_ECF_ACTION.set_conditions */
extern void T811f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ACTION.set_working_directory */
extern void T811f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ACTION.set_must_succeed */
extern void T811f8(GE_context* ac, T0* C, T1 a1);
/* detachable ET_ECF_ACTION.set_description */
extern void T811f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_action */
extern T0* T162f36(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ACTION.make */
extern T0* T811c6(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.add_options */
extern void T55f240(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T55f240ot1(T0* a1, T0** a2);
extern T1 T55f240ot2(T0* a1, T0** a2);
/* [detachable] ET_ECF_OPTIONS.set_description */
extern void T444f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.add_warning */
extern void T55f254(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_OPTIONS.set_primary_warning_value */
extern void T444f16(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_OPTIONS.primary_warning_value */
extern T0* T444f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.ecf_1_21_0 */
extern T0* T55f64(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.ecf_version */
extern T0* T187f4(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_CONFIG.ecf_version */
extern T0* T185f5(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.ecf_version */
extern T0* T58f69(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.add_debug */
extern void T55f253(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_OPTIONS.set_primary_debug_value */
extern void T444f18(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_OPTIONS.primary_debug_value */
extern T0* T444f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.add_assertions */
extern void T55f252(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
extern T1 T55f252ot1(T0* a1, T0** a2);
/* [detachable] ET_ECF_OPTIONS.set_primary_assertion_value */
extern void T444f17(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_OPTIONS.primary_assertion_value */
extern T0* T444f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_mappings */
extern T0* T55f108(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_SYSTEM_PARSER.add_mapping */
extern void T55f245(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_AST_FACTORY.new_class_mappings */
extern T0* T162f20(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_adapted_libraries */
extern T0* T55f106(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_ECF_ADAPTED_LIBRARIES.put_last */
extern void T788f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY].force_last */
extern void T1580f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_LIBRARY].force */
extern void T1863f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_LIBRARY].fill_with */
extern void T1862f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_LIBRARY].extend */
extern void T1862f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY].resize */
extern void T1580f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_LIBRARY].aliased_resized_area */
extern T0* T1863f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_LIBRARY].aliased_resized_area */
extern T0* T1862f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY].new_capacity */
extern T6 T1580f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY].extendible */
extern T1 T1580f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_adapted_libraries */
extern T0* T162f19(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ADAPTED_LIBRARIES.make */
extern T0* T788c5(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY].put_last */
extern void T1580f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY].make */
extern T0* T1580c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY].set_internal_cursor */
extern void T1580f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY].new_cursor */
extern T0* T1580f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ADAPTED_LIBRARY].make */
extern T0* T1864c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_LIBRARY].make */
extern T0* T1863f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_LIBRARY].make_empty */
extern T0* T1862c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_LIBRARY].default_create */
extern T0* T1863c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_adapted_library */
extern T0* T55f133(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM_PARSER.new_adapted_library */
extern T0* T55f133ac1(GE_context* ac, T0* a1);
extern T1 T55f133ot1(T0* a1, T0** a2);
extern T1 T55f133ot2(T0* a1, T0** a2);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.set_notes */
extern void T207f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.set_conditions */
extern void T207f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.set_visible_classes */
extern void T207f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.set_class_renamings */
extern void T207f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.set_class_options */
extern void T207f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.set_options */
extern void T207f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.set_use_application_options */
extern void T207f26(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.set_read_only */
extern void T207f25(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.set_classname_prefix */
extern void T207f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.set_description */
extern void T207f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_adapted_library */
extern T0* T162f35(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.make */
extern T0* T207c22(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.make_adapted_library */
extern void T207f33(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.default_read_only_value */
extern T1 T207f18(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.tokens */
extern T0* T207f17(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_override_clusters */
extern T0* T55f104(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ECF_AST_FACTORY.new_clusters */
extern T0* T162f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTERS.make */
extern T0* T787c5(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CLUSTER].put_last */
extern void T1579f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_override_cluster */
extern T0* T55f132(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM_PARSER.new_override_cluster */
extern T0* T55f132ac1(GE_context* ac, T0* a1);
extern T1 T55f132ot1(T0* a1, T0** a2);
extern T1 T55f132ot2(T0* a1, T0** a2);
/* detachable ET_ECF_CLUSTER.set_conditions */
extern void T802f83(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_conditioned_file_rules */
extern void T802f76(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTER.set_conditioned_subclusters */
extern void T802f75(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_provider_groups */
extern T0* T55f163(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_SYSTEM_PARSER.add_provider_group */
extern void T55f248(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_AST_FACTORY.new_provider_groups */
extern T0* T162f42(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_overridden_group */
extern T0* T55f170(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_CLUSTER.set_description */
extern void T802f70(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_cluster */
extern T0* T162f26(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_CLUSTER.make */
extern T0* T802c69(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_file_rules */
extern T0* T55f102(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_FILE_RULES.put_last */
extern void T790f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE].force_last */
extern void T1582f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_FILE_RULE].force */
extern void T1868f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_FILE_RULE].fill_with */
extern void T1869f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_FILE_RULE].extend */
extern void T1869f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE].resize */
extern void T1582f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_FILE_RULE].aliased_resized_area */
extern T0* T1868f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_FILE_RULE].aliased_resized_area */
extern T0* T1869f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE].new_capacity */
extern T6 T1582f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE].extendible */
extern T1 T1582f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_file_rules */
extern T0* T162f18(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_FILE_RULES.make */
extern T0* T790c4(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE].put_last */
extern void T1582f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE].make */
extern T0* T1582c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE].set_internal_cursor */
extern void T1582f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE].new_cursor */
extern T0* T1582f8(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_FILE_RULE].make */
extern T0* T1870c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_FILE_RULE].make */
extern T0* T1868f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_FILE_RULE].make_empty */
extern T0* T1869c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_FILE_RULE].default_create */
extern T0* T1868c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_file_rule */
extern T0* T55f131(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T55f131ot1(T0* a1, T0** a2);
/* detachable ET_ECF_FILE_RULE.set_description */
extern void T810f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_FILE_RULE.set_conditions */
extern void T810f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_file_rule */
extern T0* T162f34(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_FILE_RULE.make */
extern T0* T810c11(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_ECF_FILE_RULE.compiled_regexp */
extern T0* T810f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8].forth */
extern void T376f14(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] RX_PCRE_REGULAR_EXPRESSION].put_last */
extern void T1593f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] RX_PCRE_REGULAR_EXPRESSION].force */
extern void T1892f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] RX_PCRE_REGULAR_EXPRESSION].fill_with */
extern void T1893f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] RX_PCRE_REGULAR_EXPRESSION].extend */
extern void T1893f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] RX_PCRE_REGULAR_EXPRESSION.is_compiled */
extern T1 T81f307(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8].item */
extern T0* T376f6(GE_context* ac, T0* C);
/* detachable ET_ECF_FILE_RULE.operating_system */
extern T0* T810f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] STRING_8].start */
extern void T376f13(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] RX_PCRE_REGULAR_EXPRESSION].make */
extern T0* T1593c9(GE_context* ac, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] RX_PCRE_REGULAR_EXPRESSION].set_internal_cursor */
extern void T1593f11(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] RX_PCRE_REGULAR_EXPRESSION].new_cursor */
extern T0* T1593f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] RX_PCRE_REGULAR_EXPRESSION].make */
extern T0* T1894c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] RX_PCRE_REGULAR_EXPRESSION].make */
extern T0* T1892f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] RX_PCRE_REGULAR_EXPRESSION].make_empty */
extern T0* T1893c4(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] RX_PCRE_REGULAR_EXPRESSION].default_create */
extern T0* T1892c2(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_resources */
extern T0* T55f100(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_EXTERNAL_RESOURCES.put_last */
extern void T797f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE].force_last */
extern void T1589f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_RESOURCE].force */
extern void T1889f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_RESOURCE].fill_with */
extern void T1890f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_RESOURCE].extend */
extern void T1890f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE].resize */
extern void T1589f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_RESOURCE].aliased_resized_area */
extern T0* T1889f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_RESOURCE].aliased_resized_area */
extern T0* T1890f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE].new_capacity */
extern T6 T1589f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE].extendible */
extern T1 T1589f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_resources */
extern T0* T162f17(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_RESOURCES.make */
extern T0* T797c3(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE].put_last */
extern void T1589f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE].make */
extern T0* T1589c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE].set_internal_cursor */
extern void T1589f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE].new_cursor */
extern T0* T1589f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_RESOURCE].make */
extern T0* T1891c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_RESOURCE].make */
extern T0* T1889f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_RESOURCE].make_empty */
extern T0* T1890c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_RESOURCE].default_create */
extern T0* T1889c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_resource */
extern T0* T55f130(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T55f130ot1(T0* a1, T0** a2);
/* detachable ET_ECF_EXTERNAL_RESOURCE.set_conditions */
extern void T809f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_RESOURCE.set_description */
extern void T809f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_resource */
extern T0* T162f33(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_RESOURCE.make */
extern T0* T809c9(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_objects */
extern T0* T55f98(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_EXTERNAL_OBJECTS.put_last */
extern void T796f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT].force_last */
extern void T1588f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_OBJECT].force */
extern void T1886f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_OBJECT].fill_with */
extern void T1887f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_OBJECT].extend */
extern void T1887f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT].resize */
extern void T1588f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_OBJECT].aliased_resized_area */
extern T0* T1886f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_OBJECT].aliased_resized_area */
extern T0* T1887f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT].new_capacity */
extern T6 T1588f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT].extendible */
extern T1 T1588f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_objects */
extern T0* T162f16(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_OBJECTS.make */
extern T0* T796c3(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT].put_last */
extern void T1588f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT].make */
extern T0* T1588c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT].set_internal_cursor */
extern void T1588f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT].new_cursor */
extern T0* T1588f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_OBJECT].make */
extern T0* T1888c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_OBJECT].make */
extern T0* T1886f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_OBJECT].make_empty */
extern T0* T1887c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_OBJECT].default_create */
extern T0* T1886c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_object */
extern T0* T55f129(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T55f129ot1(T0* a1, T0** a2);
/* detachable ET_ECF_EXTERNAL_OBJECT.set_conditions */
extern void T808f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_OBJECT.set_description */
extern void T808f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_object */
extern T0* T162f32(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_OBJECT.make */
extern T0* T808c9(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_makes */
extern T0* T55f96(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_EXTERNAL_MAKES.put_last */
extern void T795f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE].force_last */
extern void T1587f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_MAKE].force */
extern void T1883f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_MAKE].fill_with */
extern void T1884f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_MAKE].extend */
extern void T1884f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE].resize */
extern void T1587f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_MAKE].aliased_resized_area */
extern T0* T1883f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_MAKE].aliased_resized_area */
extern T0* T1884f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE].new_capacity */
extern T6 T1587f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE].extendible */
extern T1 T1587f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_makes */
extern T0* T162f15(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_MAKES.make */
extern T0* T795c3(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE].put_last */
extern void T1587f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE].make */
extern T0* T1587c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE].set_internal_cursor */
extern void T1587f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE].new_cursor */
extern T0* T1587f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_MAKE].make */
extern T0* T1885c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_MAKE].make */
extern T0* T1883f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_MAKE].make_empty */
extern T0* T1884c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_MAKE].default_create */
extern T0* T1883c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_make */
extern T0* T55f128(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T55f128ot1(T0* a1, T0** a2);
/* detachable ET_ECF_EXTERNAL_MAKE.set_conditions */
extern void T807f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_MAKE.set_description */
extern void T807f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_make */
extern T0* T162f31(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_MAKE.make */
extern T0* T807c9(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_linker_flags */
extern T0* T55f94(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_EXTERNAL_LINKER_FLAGS.put_last */
extern void T794f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].force_last */
extern void T1586f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].force */
extern void T1880f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].fill_with */
extern void T1881f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].extend */
extern void T1881f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].resize */
extern void T1586f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].aliased_resized_area */
extern T0* T1880f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].aliased_resized_area */
extern T0* T1881f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].new_capacity */
extern T6 T1586f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].extendible */
extern T1 T1586f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_linker_flags */
extern T0* T162f14(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_LINKER_FLAGS.make */
extern T0* T794c3(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].put_last */
extern void T1586f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].make */
extern T0* T1586c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].set_internal_cursor */
extern void T1586f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].new_cursor */
extern T0* T1586f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].make */
extern T0* T1882c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].make */
extern T0* T1880f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].make_empty */
extern T0* T1881c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].default_create */
extern T0* T1880c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_linker_flag */
extern T0* T55f127(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T55f127ot1(T0* a1, T0** a2);
/* detachable ET_ECF_EXTERNAL_LINKER_FLAG.set_conditions */
extern void T806f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_LINKER_FLAG.set_description */
extern void T806f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_linker_flag */
extern T0* T162f30(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_LINKER_FLAG.make */
extern T0* T806c5(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_libraries */
extern T0* T55f92(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_EXTERNAL_LIBRARIES.put_last */
extern void T793f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY].force_last */
extern void T1585f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LIBRARY].force */
extern void T1877f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LIBRARY].fill_with */
extern void T1878f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LIBRARY].extend */
extern void T1878f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY].resize */
extern void T1585f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LIBRARY].aliased_resized_area */
extern T0* T1877f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LIBRARY].aliased_resized_area */
extern T0* T1878f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY].new_capacity */
extern T6 T1585f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY].extendible */
extern T1 T1585f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_libraries */
extern T0* T162f13(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_LIBRARIES.make */
extern T0* T793c3(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY].put_last */
extern void T1585f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY].make */
extern T0* T1585c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY].set_internal_cursor */
extern void T1585f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY].new_cursor */
extern T0* T1585f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_LIBRARY].make */
extern T0* T1879c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LIBRARY].make */
extern T0* T1877f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_LIBRARY].make_empty */
extern T0* T1878c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_LIBRARY].default_create */
extern T0* T1877c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_library */
extern T0* T55f126(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T55f126ot1(T0* a1, T0** a2);
/* detachable ET_ECF_EXTERNAL_LIBRARY.set_conditions */
extern void T805f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_LIBRARY.set_description */
extern void T805f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_library */
extern T0* T162f29(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_LIBRARY.make */
extern T0* T805c9(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_includes */
extern T0* T55f90(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_EXTERNAL_INCLUDES.put_last */
extern void T792f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE].force_last */
extern void T1584f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_INCLUDE].force */
extern void T1874f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_INCLUDE].fill_with */
extern void T1875f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_INCLUDE].extend */
extern void T1875f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE].resize */
extern void T1584f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_INCLUDE].aliased_resized_area */
extern T0* T1874f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_INCLUDE].aliased_resized_area */
extern T0* T1875f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE].new_capacity */
extern T6 T1584f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE].extendible */
extern T1 T1584f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_includes */
extern T0* T162f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_INCLUDES.make */
extern T0* T792c3(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE].put_last */
extern void T1584f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE].make */
extern T0* T1584c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE].set_internal_cursor */
extern void T1584f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE].new_cursor */
extern T0* T1584f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_INCLUDE].make */
extern T0* T1876c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_INCLUDE].make */
extern T0* T1874f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_INCLUDE].make_empty */
extern T0* T1875c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_INCLUDE].default_create */
extern T0* T1874c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_include */
extern T0* T55f125(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T55f125ot1(T0* a1, T0** a2);
/* detachable ET_ECF_EXTERNAL_INCLUDE.set_conditions */
extern void T804f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_INCLUDE.set_description */
extern void T804f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_include */
extern T0* T162f28(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_INCLUDE.make */
extern T0* T804c9(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_cflags */
extern T0* T55f88(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_EXTERNAL_CFLAGS.put_last */
extern void T791f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG].force_last */
extern void T1583f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_CFLAG].force */
extern void T1871f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_CFLAG].fill_with */
extern void T1872f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_CFLAG].extend */
extern void T1872f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG].resize */
extern void T1583f15(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_CFLAG].aliased_resized_area */
extern T0* T1871f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_CFLAG].aliased_resized_area */
extern T0* T1872f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG].new_capacity */
extern T6 T1583f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG].extendible */
extern T1 T1583f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_cflags */
extern T0* T162f11(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_CFLAGS.make */
extern T0* T791c3(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG].put_last */
extern void T1583f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG].make */
extern T0* T1583c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG].set_internal_cursor */
extern void T1583f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG].new_cursor */
extern T0* T1583f9(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_EXTERNAL_CFLAG].make */
extern T0* T1873c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_CFLAG].make */
extern T0* T1871f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_EXTERNAL_CFLAG].make_empty */
extern T0* T1872c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_EXTERNAL_CFLAG].default_create */
extern T0* T1871c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_external_cflag */
extern T0* T55f124(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T55f124ot1(T0* a1, T0** a2);
/* detachable ET_ECF_EXTERNAL_CFLAG.set_conditions */
extern void T803f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_CFLAG.set_description */
extern void T803f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_external_cflag */
extern T0* T162f27(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_EXTERNAL_CFLAG.make */
extern T0* T803c5(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_TARGET.set_description */
extern void T59f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_clusters */
extern T0* T55f86(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_cluster */
extern T0* T55f123(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM_PARSER.new_cluster */
extern T0* T55f123ac1(GE_context* ac, T0* a1);
extern T1 T55f123ot1(T0* a1, T0** a2);
extern T1 T55f123ot2(T0* a1, T0** a2);
/* detachable ET_ECF_CLUSTER.set_notes */
extern void T802f84(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.add_capabilities */
extern void T55f239(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
extern T1 T55f239ot1(T0* a1, T0** a2);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_adapted_dotnet_assemblies */
extern T0* T55f83(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* detachable ET_ECF_ADAPTED_DOTNET_ASSEMBLIES.put_last */
extern void T789f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].force_last */
extern void T1581f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].force */
extern void T1866f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].fill_with */
extern void T1865f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].extend */
extern void T1865f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].resize */
extern void T1581f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].aliased_resized_area */
extern T0* T1866f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].aliased_resized_area */
extern T0* T1865f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].new_capacity */
extern T6 T1581f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].extendible */
extern T1 T1581f7(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_adapted_dotnet_assemblies */
extern T0* T162f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ADAPTED_DOTNET_ASSEMBLIES.make */
extern T0* T789c5(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].put_last */
extern void T1581f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].make */
extern T0* T1581c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].set_internal_cursor */
extern void T1581f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].new_cursor */
extern T0* T1581f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].make */
extern T0* T1867c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].make */
extern T0* T1866f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].make_empty */
extern T0* T1865c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].default_create */
extern T0* T1866c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_adapted_dotnet_assembly */
extern T0* T55f122(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM_PARSER.new_adapted_dotnet_assembly */
extern T0* T55f122ac1(GE_context* ac, T0* a1);
extern T1 T55f122ot1(T0* a1, T0** a2);
extern T1 T55f122ot2(T0* a1, T0** a2);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_notes */
extern void T213f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_conditions */
extern void T213f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_class_renamings */
extern void T213f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_class_options */
extern void T213f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_options */
extern void T213f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_read_only */
extern void T213f31(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_classname_prefix */
extern void T213f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_assembly_version */
extern void T213f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_assembly_name */
extern void T213f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_assembly_key */
extern void T213f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_assembly_culture */
extern void T213f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.set_description */
extern void T213f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_adapted_dotnet_assembly */
extern T0* T162f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.make */
extern T0* T213c24(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.make_adapted_dotnet_assembly */
extern void T213f37(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.default_read_only_value */
extern T1 T213f20(GE_context* ac, T0* C);
/* [detachable] ET_TOKEN_CONSTANTS.unknown_dotnet_assembly */
extern T0* T73f232(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.make */
extern T0* T596c105(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_DOTNET_ASSEMBLY.make_from_system */
extern void T596f106(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_ASSEMBLY.set_kernel_types */
extern void T596f110(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_wide_character_type */
extern void T596f152(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_typed_pointer_type */
extern void T596f151(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_type_type */
extern void T596f150(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_system_string_type */
extern void T596f149(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_system_object_type */
extern void T596f148(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_string_type */
extern void T596f147(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_string_32_type */
extern void T596f146(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_string_8_type */
extern void T596f145(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_special_type */
extern void T596f144(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_routine_type */
extern void T596f143(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_real_64_type */
extern void T596f142(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_real_32_type */
extern void T596f141(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_real_type */
extern void T596f140(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_procedure_type */
extern void T596f139(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_predicate_type */
extern void T596f138(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_pointer_type */
extern void T596f137(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_natural_64_type */
extern void T596f136(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_natural_32_type */
extern void T596f135(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_natural_16_type */
extern void T596f134(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_natural_8_type */
extern void T596f133(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_natural_type */
extern void T596f132(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_iterable_type */
extern void T596f131(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_ise_exception_manager_type */
extern void T596f130(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_integer_64_type */
extern void T596f129(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_integer_32_type */
extern void T596f128(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_integer_16_type */
extern void T596f127(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_integer_8_type */
extern void T596f126(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_integer_type */
extern void T596f125(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_immutable_string_32_type */
extern void T596f124(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_immutable_string_8_type */
extern void T596f123(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_function_type */
extern void T596f122(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_exception_manager_type */
extern void T596f121(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_exception_type */
extern void T596f120(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_double_type */
extern void T596f119(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_character_32_type */
extern void T596f118(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_character_8_type */
extern void T596f117(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_character_type */
extern void T596f116(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_boolean_type */
extern void T596f115(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_array_type */
extern void T596f114(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_tuple_type */
extern void T596f113(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_none_type */
extern void T596f112(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_any_type */
extern void T596f111(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.make_adapted */
extern void T596f109(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_DOTNET_ASSEMBLY.default_read_only_value */
extern T1 T596f90(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.set_unknown_kernel_types */
extern void T596f108(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.initialize */
extern void T596f107(GE_context* ac, T0* C);
/* [detachable] ET_DOTNET_ASSEMBLY.class_name_tester */
extern T0* T596f89(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.tokens */
extern T0* T213f19(GE_context* ac, T0* C);
/* [detachable] ET_ECF_TARGET.set_parent */
extern void T59f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.execution_environment */
extern T0* T59f28(GE_context* ac, T0* C);
/* detachable ET_ECF_VARIABLES.set_secondary_variables */
extern void T62f22(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_TARGET_PARENT.make */
extern T0* T200c4(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_OPTIONS.set_secondary_options */
extern void T444f20(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CAPABILITIES.set_secondary_capabilities */
extern void T61f17(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_SETTINGS.set_secondary_settings */
extern void T57f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_AST_FACTORY.new_target */
extern T0* T162f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_TARGET.make */
extern T0* T59c32(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_ECF_CAPABILITIES.make */
extern T0* T61c13(GE_context* ac);
/* Creation of agent #1 in feature detachable ET_ECF_CAPABILITIES.make */
extern T0* T61f13ac1(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].set_hash_function */
extern void T221f41(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CAPABILITIES.string_ */
extern T0* T61f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].set_key_equality_tester */
extern void T221f40(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CAPABILITIES.case_insensitive_string_equality_tester */
extern T0* T61f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].make_map */
extern T0* T221c39(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].make_with_equality_testers */
extern void T221f43(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].make_sparse_container */
extern void T221f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].set_internal_cursor */
extern void T221f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].new_cursor */
extern T0* T221f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8].make */
extern T0* T601c10(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].make_slots */
extern void T221f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].make_clashes */
extern void T221f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].make_key_storage */
extern void T221f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].make_item_storage */
extern void T221f56(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_VARIABLES.make */
extern T0* T62c20(GE_context* ac);
/* Creation of agent #1 in feature detachable ET_ECF_VARIABLES.make */
extern T0* T62f20ac1(GE_context* ac, T0* a1);
/* detachable ET_ECF_VARIABLES.case_insensitive_string_equality_tester */
extern T0* T62f4(GE_context* ac, T0* C);
/* detachable ET_ECF_VARIABLES.execution_environment */
extern T0* T62f3(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_notes */
extern void T187f125(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_CONFIG.set_notes */
extern void T185f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_notes */
extern void T58f168(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_notes */
extern T0* T55f44(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT].force_last */
extern void T445f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_NOTE_ELEMENT].force */
extern void T926f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_NOTE_ELEMENT].fill_with */
extern void T927f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_NOTE_ELEMENT].extend */
extern void T927f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT].resize */
extern void T445f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_NOTE_ELEMENT].aliased_resized_area */
extern T0* T926f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_NOTE_ELEMENT].aliased_resized_area */
extern T0* T927f4(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT].new_capacity */
extern T6 T445f7(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT].extendible */
extern T1 T445f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_AST_FACTORY.new_notes */
extern T0* T162f5(GE_context* ac, T0* C);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT].make */
extern void T445f10(GE_context* ac, T0* C, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT].make */
extern T0* T445c10(GE_context* ac, T6 a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT].set_internal_cursor */
extern void T445f12(GE_context* ac, T0* C, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT].new_cursor */
extern T0* T445f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ECF_NOTE_ELEMENT].make */
extern T0* T928c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_NOTE_ELEMENT].make */
extern T0* T926f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_NOTE_ELEMENT].make_empty */
extern T0* T927c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_NOTE_ELEMENT].default_create */
extern T0* T926c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_note_element */
extern T0* T55f76(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
extern T1 T55f76ot1(T0* a1, T0** a2);
extern T1 T55f76ot2(T0* a1, T0** a2);
/* detachable ET_ECF_NOTE_ELEMENT.set_content */
extern void T786f10(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_NOTE_ELEMENT.add_element */
extern void T786f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_NOTE_ELEMENT.add_attribute */
extern void T786f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_AST_FACTORY.new_note */
extern T0* T162f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_NOTE_ELEMENT.make */
extern T0* T786c7(GE_context* ac, T0* a1);
/* Creation of agent #1 in feature detachable ET_ECF_NOTE_ELEMENT.make */
extern T0* T786f7ac1(GE_context* ac, T0* a1);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT].make_default */
extern T0* T445c14(GE_context* ac);
/* detachable DS_ARRAYED_LIST [[attached] ET_ECF_NOTE_ELEMENT].default_capacity */
extern T6 T445f9(GE_context* ac, T0* C);
/* detachable ET_ECF_NOTE_ELEMENT.string_ */
extern T0* T786f6(GE_context* ac, T0* C);
/* detachable ET_ECF_NOTE_ELEMENT.case_insensitive_string_equality_tester */
extern T0* T786f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].make_map_default */
extern T0* T216c93(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].default_capacity */
extern T6 T216f46(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_description */
extern void T187f124(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_CONFIG.set_description */
extern void T185f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_description */
extern void T58f167(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_ELEMENT.text */
extern T0* T168f14(GE_context* ac, T0* C);
extern T1 T168f14ot1(T0* a1, T0** a2);
/* [detachable] XM_ELEMENT.string_ */
extern T0* T168f15(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.string_ */
extern T0* T55f20(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].item */
extern T0* T447f8(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].start */
extern void T447f11(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].cursor_start */
extern void T894f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].cursor_off */
extern T1 T894f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].off */
extern T1 T447f9(GE_context* ac, T0* C);
/* [detachable] XM_ELEMENT.new_cursor */
extern T0* T168f6(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].new_cursor */
extern T0* T894f5(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].make */
extern T0* T447c10(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_options */
extern T0* T55f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_options_1_0_0 */
extern T0* T55f75(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_options_1_0_0 */
extern void T55f238(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_options_1_5_0 */
extern void T55f237(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_options_1_9_0 */
extern void T55f236(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_options_1_11_0 */
extern void T55f235(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_options_1_15_0 */
extern void T55f234(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_options_1_18_0 */
extern void T55f233(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_options_1_21_0 */
extern void T55f232(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_options_1_22_0 */
extern void T55f231(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_options_latest */
extern void T55f251(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_options_1_5_0 */
extern T0* T55f74(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_options_1_9_0 */
extern T0* T55f72(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.ecf_1_9_0 */
extern T0* T55f71(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_options_1_11_0 */
extern T0* T55f70(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_options_1_15_0 */
extern T0* T55f69(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.ecf_1_15_0 */
extern T0* T55f68(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_options_1_18_0 */
extern T0* T55f67(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.ecf_1_18_0 */
extern T0* T55f66(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_options_1_21_0 */
extern T0* T55f65(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_options_1_22_0 */
extern T0* T55f63(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.ecf_1_22_0 */
extern T0* T55f62(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_capabilities */
extern T0* T55f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_capabilities_1_0_0 */
extern T0* T55f61(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_capabilities_1_0_0 */
extern void T55f230(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_capabilities_1_11_0 */
extern void T55f229(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_capabilities_1_13_0 */
extern void T55f228(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_capabilities_1_16_0 */
extern void T55f227(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_capabilities_latest */
extern void T55f250(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_capabilities_1_11_0 */
extern T0* T55f60(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_capabilities_1_13_0 */
extern T0* T55f58(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.ecf_1_13_0 */
extern T0* T55f57(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_capabilities_1_16_0 */
extern T0* T55f56(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.ecf_1_16_0 */
extern T0* T55f55(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_settings */
extern T0* T55f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_settings_1_0_0 */
extern T0* T55f54(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_settings_1_0_0 */
extern void T55f226(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_settings_1_20_0 */
extern void T55f225(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_default_settings_latest */
extern void T55f249(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.default_settings_1_20_0 */
extern T0* T55f53(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.ecf_1_20_0 */
extern T0* T55f52(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.ecf_latest */
extern T0* T55f38(GE_context* ac, T0* C);
/* [detachable] UT_VERSION.make_latest */
extern T0* T46c20(GE_context* ac);
/* [detachable] ET_ECF_LIBRARY.set_ecf_version */
extern void T187f123(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_CONFIG.set_ecf_version */
extern void T185f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_ecf_version */
extern void T58f166(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.ecf_version */
extern T0* T55f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_LIBRARY.set_ecf_namespace */
extern void T187f122(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_CONFIG.set_ecf_namespace */
extern void T185f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_ecf_namespace */
extern void T58f165(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eate_error */
extern void T56f34(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_eate */
extern T0* T214c102(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_SYSTEM_PARSER.attribute_name */
extern T0* T55f24(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].found_item */
extern T0* T174f27(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].found */
extern T1 T174f22(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].search */
extern void T174f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].search_position */
extern void T174f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eatm_error */
extern void T56f33(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_eatm */
extern T0* T214c101(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_AST_FACTORY.new_library */
extern T0* T162f4(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_LIBRARY.make */
extern T0* T187c121(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_LIBRARY.make_library */
extern void T187f130(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_LIBRARY.make_from_system */
extern void T187f131(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_LIBRARY.set_kernel_types */
extern void T187f135(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_wide_character_type */
extern void T187f177(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.tokens */
extern T0* T187f102(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_typed_pointer_type */
extern void T187f176(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_type_type */
extern void T187f175(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_system_string_type */
extern void T187f174(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_system_object_type */
extern void T187f173(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_string_type */
extern void T187f172(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_string_32_type */
extern void T187f171(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_string_8_type */
extern void T187f170(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_special_type */
extern void T187f169(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_routine_type */
extern void T187f168(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_real_64_type */
extern void T187f167(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_real_32_type */
extern void T187f166(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_real_type */
extern void T187f165(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_procedure_type */
extern void T187f164(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_predicate_type */
extern void T187f163(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_pointer_type */
extern void T187f162(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_natural_64_type */
extern void T187f161(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_natural_32_type */
extern void T187f160(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_natural_16_type */
extern void T187f159(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_natural_8_type */
extern void T187f158(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_natural_type */
extern void T187f157(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_iterable_type */
extern void T187f156(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_ise_exception_manager_type */
extern void T187f155(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_integer_64_type */
extern void T187f154(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_integer_32_type */
extern void T187f153(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_integer_16_type */
extern void T187f152(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_integer_8_type */
extern void T187f151(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_integer_type */
extern void T187f150(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_immutable_string_32_type */
extern void T187f149(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_immutable_string_8_type */
extern void T187f148(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_function_type */
extern void T187f147(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_exception_manager_type */
extern void T187f146(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_exception_type */
extern void T187f145(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_double_type */
extern void T187f144(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_character_32_type */
extern void T187f143(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_character_8_type */
extern void T187f142(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_character_type */
extern void T187f141(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_boolean_type */
extern void T187f140(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_array_type */
extern void T187f139(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_tuple_type */
extern void T187f138(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_none_type */
extern void T187f137(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_any_type */
extern void T187f136(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.make_adapted */
extern void T187f134(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_LIBRARY.default_read_only_value */
extern T1 T187f103(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.set_unknown_kernel_types */
extern void T187f133(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.initialize */
extern void T187f132(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.initialize */
extern void T187f132p1(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.class_name_tester */
extern T0* T187f110(GE_context* ac, T0* C);
/* [detachable] XM_ELEMENT.attribute_by_name */
extern T0* T168f5(GE_context* ac, T0* C, T0* a1);
extern T1 T168f5ot1(T0* a1, T0** a2);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].go_after */
extern void T447f13(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].cursor_go_after */
extern void T894f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].set_after */
extern void T447f16(GE_context* ac, T0* C);
/* [detachable] XM_ELEMENT.attribute_same_name */
extern T1 T168f7(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_ELEMENT.same_string */
extern T1 T168f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_ELEMENT.string_equality_tester */
extern T0* T168f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].forth */
extern void T205f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].cursor_forth */
extern void T204f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].add_traversing_cursor */
extern void T204f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].set_next_cursor */
extern void T205f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].set_position */
extern void T205f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].remove_traversing_cursor */
extern void T204f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].set_after */
extern void T205f12(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].clashes_item */
extern T6 T204f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eadp_error */
extern void T56f30(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ERROR.make_eadp */
extern T0* T214c100(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* detachable ET_ECF_ANDED_CONDITIONS.is_capability_supported */
extern T1 T819f3(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_ANDED_CONDITIONS.is_capability_supported */
extern T1 T819f3p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_PLATFORM_CONDITION.is_capability_supported */
extern T1 T828f4(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_DYNAMIC_RUNTIME_CONDITION.is_capability_supported */
extern T1 T827f3(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_DOTNET_CONDITION.is_capability_supported */
extern T1 T826f3(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_CUSTOM_CONDITION.is_capability_supported */
extern T1 T825f6(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.is_capability_supported */
extern T1 T824f4(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Creation of agent #1 in feature detachable ET_ECF_VOID_SAFETY_CONDITION.is_capability_supported */
extern T0* T824f4ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature detachable ET_ECF_VOID_SAFETY_CONDITION.is_capability_supported */
extern T0* T824f4ac2(GE_context* ac, T0* a1);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] STRING_8]].set_rout_disp_final */
extern void T1742f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable DS_HASH_SET [[attached] STRING_8].go_after */
extern void T70f80(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].cursor_go_after */
extern void T70f83(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].put_last */
extern void T70f79(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].compress */
extern void T70f82(GE_context* ac, T0* C);
/* detachable DS_HASH_SET [[attached] STRING_8].move_all_cursors */
extern void T70f84(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].do_all */
extern void T152f27(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].append_last */
extern void T70f78(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].extend_last */
extern void T70f81(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].extendible */
extern T1 T70f37(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.supported_void_safety_capability_values */
extern T0* T824f8(GE_context* ac, T0* C);
/* Creation of agent #1 in feature detachable ET_ECF_VOID_SAFETY_CONDITION.supported_void_safety_capability_values */
extern T0* T824f8ac1(GE_context* ac, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].set_hash_function */
extern void T70f77(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.string_ */
extern T0* T824f6(GE_context* ac, T0* C);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.case_insensitive_string_equality_tester */
extern T0* T824f7(GE_context* ac, T0* C);
/* detachable ET_ECF_CONCURRENCY_CONDITION.is_capability_supported */
extern T1 T823f4(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Creation of agent #1 in feature detachable ET_ECF_CONCURRENCY_CONDITION.is_capability_supported */
extern T0* T823f4ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature detachable ET_ECF_CONCURRENCY_CONDITION.is_capability_supported */
extern T0* T823f4ac2(GE_context* ac, T0* a1);
/* detachable ET_ECF_CONCURRENCY_CONDITION.supported_concurrency_capability_values */
extern T0* T823f8(GE_context* ac, T0* C);
/* Creation of agent #1 in feature detachable ET_ECF_CONCURRENCY_CONDITION.supported_concurrency_capability_values */
extern T0* T823f8ac1(GE_context* ac, T0* a1);
/* detachable ET_ECF_CONCURRENCY_CONDITION.string_ */
extern T0* T823f6(GE_context* ac, T0* C);
/* detachable ET_ECF_CONCURRENCY_CONDITION.case_insensitive_string_equality_tester */
extern T0* T823f7(GE_context* ac, T0* C);
/* detachable ET_ECF_BUILD_CONDITION.is_capability_supported */
extern T1 T822f4(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_MSIL_CLR_VERSION_CONDITION.is_capability_supported */
extern T1 T831f4(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_COMPILER_VERSION_CONDITION.is_capability_supported */
extern T1 T830f4(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_PLATFORM_CONDITION.is_capability_aware */
extern T1 T828f3(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_DYNAMIC_RUNTIME_CONDITION.is_capability_aware */
extern T1 T827f2(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_DOTNET_CONDITION.is_capability_aware */
extern T1 T826f2(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CUSTOM_CONDITION.is_capability_aware */
extern T1 T825f5(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.is_capability_aware */
extern T1 T824f3(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CONCURRENCY_CONDITION.is_capability_aware */
extern T1 T823f3(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_BUILD_CONDITION.is_capability_aware */
extern T1 T822f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_MSIL_CLR_VERSION_CONDITION.is_capability_aware */
extern T1 T831f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_COMPILER_VERSION_CONDITION.is_capability_aware */
extern T1 T830f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_CONDITION_ITEM].item */
extern T0* T1603f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_ORED_CONDITIONS.is_capability_supported */
extern T1 T818f3(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ANDED_CONDITIONS].item */
extern T0* T1602f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_ORED_CONDITIONS.is_capability_supported */
extern T1 T818f3p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* detachable ET_ECF_CAPABILITIES.is_capability_supported */
extern T1 T61f9(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].found_item */
extern T0* T1745f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].item_storage_item */
extern T0* T1745f27(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].found */
extern T1 T1745f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].search */
extern void T1745f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].search_position */
extern void T1745f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].hash_position */
extern T6 T1745f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].key_storage_item */
extern T0* T1745f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].clashes_item */
extern T6 T1745f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].slots_item */
extern T6 T1745f23(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_CAPABILITIES.capability_order */
extern T0* T61f12(GE_context* ac, T0* C);
/* Creation of agent #1 in feature detachable ET_ECF_CAPABILITIES.capability_order */
extern T0* T61f12ac1(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].put_last */
extern void T1745f36(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].key_storage_put */
extern void T1745f44(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].slots_put */
extern void T1745f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].clashes_put */
extern void T1745f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].compress */
extern void T1745f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].slots_wipe_out */
extern void T1745f50(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].clashes_wipe_out */
extern void T1745f49(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].key_storage_keep_head */
extern void T1745f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].item_storage_keep_head */
extern void T1745f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]].keep_head */
extern void T1985f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]].set_count */
extern void T1985f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].move_all_cursors */
extern void T1745f46(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].set_position */
extern void T1989f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].item_storage_put */
extern void T1745f40(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]].force */
extern void T1988f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]].fill_with */
extern void T1985f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]].extend */
extern void T1985f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].unset_found_item */
extern void T1745f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].put_last */
extern void T1744f39(GE_context* ac, T0* C, T6 a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].compress */
extern void T1744f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].slots_wipe_out */
extern void T1744f57(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].clashes_wipe_out */
extern void T1744f56(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].key_storage_keep_head */
extern void T1744f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].item_storage_keep_head */
extern void T1744f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].move_all_cursors */
extern void T1744f53(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] INTEGER_32, [attached] STRING_8].set_position */
extern void T1984f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8].set_hash_function */
extern void T1744f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].set_hash_function */
extern void T1745f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].set_key_equality_tester */
extern void T1745f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].make_map */
extern T0* T1745c32(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].make_with_equality_testers */
extern void T1745f37(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].make_sparse_container */
extern void T1745f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].set_internal_cursor */
extern void T1745f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].new_cursor */
extern T0* T1745f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].make */
extern T0* T1989c5(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].make_slots */
extern void T1745f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].new_modulus */
extern T6 T1745f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].make_clashes */
extern void T1745f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].make_key_storage */
extern void T1745f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8], [attached] STRING_8].make_item_storage */
extern void T1745f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]].make */
extern T0* T1988f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]].make_empty */
extern T0* T1985c4(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_TABLE [[attached] INTEGER_32, [attached] STRING_8]].default_create */
extern T0* T1988c2(GE_context* ac);
/* [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].key */
extern T0* T205f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].cursor_key */
extern T0* T204f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].key_storage_item */
extern T0* T204f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].after */
extern T1 T205f7(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].new_iterator */
extern T0* T204f23(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].start */
extern void T205f11(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].cursor_start */
extern void T204f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].cursor_off */
extern T1 T204f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].off */
extern T1 T205f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].is_empty */
extern T1 T204f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].new_cursor */
extern T0* T204f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].make */
extern T0* T205c9(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.valid_capabilities_latest */
extern T0* T55f28(GE_context* ac, T0* C);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM_PARSER.valid_capabilities_latest */
extern T0* T55f28ac1(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.void_safety_detection_capability_value_regexp */
extern T0* T55f35(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.concurrency_capability_value_regexp */
extern T0* T55f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].force_last */
extern void T204f42(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].key_storage_put */
extern void T204f50(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].slots_put */
extern void T204f49(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].clashes_put */
extern void T204f48(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].slots_item */
extern T6 T204f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].hash_position */
extern T6 T204f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].resize */
extern void T204f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].clashes_resize */
extern void T204f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].special_integer_ */
extern T0* T204f30(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].key_storage_resize */
extern void T204f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].item_storage_resize */
extern void T204f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable RX_REGULAR_EXPRESSION].aliased_resized_area */
extern T0* T592f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].aliased_resized_area */
extern T0* T593f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].slots_resize */
extern void T204f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].new_modulus */
extern T6 T204f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].new_capacity */
extern T6 T204f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].item_storage_put */
extern void T204f46(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [detachable RX_REGULAR_EXPRESSION].force */
extern void T592f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].fill_with */
extern void T593f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].extend */
extern void T593f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].search_position */
extern void T204f45(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].unset_found_item */
extern void T204f44(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.catcall_detection_capability_value_regexp */
extern T0* T55f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].set_hash_function */
extern void T204f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].set_key_equality_tester */
extern void T204f40(GE_context* ac, T0* C, T0* a1);
/* detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].internal_set_equality_tester */
extern void T591f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.case_insensitive_string_equality_tester */
extern T0* T55f14(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].make_map */
extern T0* T204c39(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].make_with_equality_testers */
extern void T204f43(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].make_sparse_container */
extern void T204f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].set_internal_cursor */
extern void T204f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].make_slots */
extern void T204f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].make_clashes */
extern void T204f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].make_key_storage */
extern void T204f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].make_item_storage */
extern void T204f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable RX_REGULAR_EXPRESSION].make */
extern T0* T592f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].make_empty */
extern T0* T593c6(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable RX_REGULAR_EXPRESSION].default_create */
extern T0* T592c3(GE_context* ac);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eabw_error */
extern void T56f29(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_eabw */
extern T0* T214c99(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.set_library */
extern void T207f34(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eabv_error */
extern void T56f28(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_eabv */
extern T0* T214c98(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.pathname */
extern T0* T207f20(GE_context* ac, T0* C);
/* [detachable] ET_ADAPTED_LIBRARIES.library */
extern T0* T188f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ADAPTED_LIBRARIES.count */
extern T6 T188f3(GE_context* ac, T0* C);
/* [detachable] DS_CELL [detachable ET_ECF_LIBRARY].make */
extern T0* T206c2(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.select_target */
extern void T55f219(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Creation of agent #1 in feature [detachable] ET_ECF_SYSTEM_PARSER.select_target */
extern T0* T55f219ac1(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_SYSTEM_PARSER.build_system_config */
extern void T55f222(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] DS_CELL [detachable ET_ECF_SYSTEM_CONFIG].put */
extern void T201f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_system_config */
extern T0* T55f29(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_AST_FACTORY.new_system_config */
extern T0* T162f3(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_SYSTEM_CONFIG.make */
extern T0* T185c12(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_TARGET.fill_universe */
extern void T59f54(GE_context* ac, T0* C, T0* a1, T0* a2);
extern T1 T59f54ot1(T0* a1);
/* detachable ET_ECF_EXTERNAL_LINKER_FLAGS.fill_external_linker_flags */
extern void T794f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature detachable ET_ECF_EXTERNAL_LINKER_FLAGS.fill_external_linker_flags */
extern T0* T794f5ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_LINKER_FLAG]].set_rout_disp_final */
extern void T1796f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable ET_ECF_EXTERNAL_LINKER_FLAG.fill_external_linker_flags */
extern void T806f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_LINKER_FLAG.add_external_value */
extern void T806f9(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_LINKER_FLAG.is_enabled */
extern T1 T806f4(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ANDED_CONDITIONS.is_enabled */
extern T1 T819f4(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_PLATFORM_CONDITION.is_enabled */
extern T1 T828f5(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature detachable ET_ECF_PLATFORM_CONDITION.is_enabled */
extern T0* T828f5ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].there_exists */
extern T1 T152f16(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_PLATFORM_CONDITION.string_ */
extern T0* T828f7(GE_context* ac, T0* C);
/* detachable ET_ECF_PLATFORM_CONDITION.operating_system */
extern T0* T828f6(GE_context* ac, T0* C);
/* detachable ET_ECF_DYNAMIC_RUNTIME_CONDITION.is_enabled */
extern T1 T827f4(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_DOTNET_CONDITION.is_enabled */
extern T1 T826f4(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CUSTOM_CONDITION.is_enabled */
extern T1 T825f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] LX_DFA_WILDCARD.compile_case_sensitive */
extern T0* T2068c19(GE_context* ac, T0* a1);
/* detachable ET_ECF_CUSTOM_CONDITION.string_ */
extern T0* T825f8(GE_context* ac, T0* C);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.is_enabled */
extern T1 T824f5(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature detachable ET_ECF_VOID_SAFETY_CONDITION.is_enabled */
extern T0* T824f5ac1(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.default_capabilities */
extern T0* T824f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.default_capabilities_1_0_0 */
extern T0* T824f16(GE_context* ac, T0* C);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.set_default_capabilities_1_0_0 */
extern void T824f22(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.set_default_capabilities_1_11_0 */
extern void T824f21(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.set_default_capabilities_1_13_0 */
extern void T824f20(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.set_default_capabilities_1_16_0 */
extern void T824f19(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.set_default_capabilities_latest */
extern void T824f23(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.default_capabilities_1_11_0 */
extern T0* T824f15(GE_context* ac, T0* C);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.ecf_1_11_0 */
extern T0* T824f14(GE_context* ac, T0* C);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.default_capabilities_1_13_0 */
extern T0* T824f13(GE_context* ac, T0* C);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.ecf_1_13_0 */
extern T0* T824f12(GE_context* ac, T0* C);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.default_capabilities_1_16_0 */
extern T0* T824f11(GE_context* ac, T0* C);
/* detachable ET_ECF_VOID_SAFETY_CONDITION.ecf_1_16_0 */
extern T0* T824f10(GE_context* ac, T0* C);
/* detachable ET_ECF_CONCURRENCY_CONDITION.is_enabled */
extern T1 T823f5(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature detachable ET_ECF_CONCURRENCY_CONDITION.is_enabled */
extern T0* T823f5ac1(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_ECF_CONCURRENCY_CONDITION.default_capabilities */
extern T0* T823f9(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CONCURRENCY_CONDITION.default_capabilities_1_0_0 */
extern T0* T823f16(GE_context* ac, T0* C);
/* detachable ET_ECF_CONCURRENCY_CONDITION.set_default_capabilities_1_0_0 */
extern void T823f22(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CONCURRENCY_CONDITION.set_default_capabilities_1_11_0 */
extern void T823f21(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CONCURRENCY_CONDITION.set_default_capabilities_1_13_0 */
extern void T823f20(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CONCURRENCY_CONDITION.set_default_capabilities_1_16_0 */
extern void T823f19(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CONCURRENCY_CONDITION.set_default_capabilities_latest */
extern void T823f23(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CONCURRENCY_CONDITION.default_capabilities_1_11_0 */
extern T0* T823f15(GE_context* ac, T0* C);
/* detachable ET_ECF_CONCURRENCY_CONDITION.ecf_1_11_0 */
extern T0* T823f14(GE_context* ac, T0* C);
/* detachable ET_ECF_CONCURRENCY_CONDITION.default_capabilities_1_13_0 */
extern T0* T823f13(GE_context* ac, T0* C);
/* detachable ET_ECF_CONCURRENCY_CONDITION.ecf_1_13_0 */
extern T0* T823f12(GE_context* ac, T0* C);
/* detachable ET_ECF_CONCURRENCY_CONDITION.default_capabilities_1_16_0 */
extern T0* T823f11(GE_context* ac, T0* C);
/* detachable ET_ECF_CONCURRENCY_CONDITION.ecf_1_16_0 */
extern T0* T823f10(GE_context* ac, T0* C);
/* detachable ET_ECF_BUILD_CONDITION.is_enabled */
extern T1 T822f5(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature detachable ET_ECF_BUILD_CONDITION.is_enabled */
extern T0* T822f5ac1(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_ECF_BUILD_CONDITION.string_ */
extern T0* T822f6(GE_context* ac, T0* C);
/* [detachable] ET_ECF_MSIL_CLR_VERSION_CONDITION.is_enabled */
extern T1 T831f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_MSIL_CLR_VERSION_CONDITION.is_included */
extern T1 T831f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_COMPILER_VERSION_CONDITION.is_enabled */
extern T1 T830f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_COMPILER_VERSION_CONDITION.is_included */
extern T1 T830f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ORED_CONDITIONS.is_enabled */
extern T1 T818f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].do_all */
extern void T1586f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LINKER_FLAG].item */
extern T0* T1586f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_EXTERNAL_CFLAGS.fill_external_cflags */
extern void T791f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature detachable ET_ECF_EXTERNAL_CFLAGS.fill_external_cflags */
extern T0* T791f5ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_CFLAG]].set_rout_disp_final */
extern void T1789f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable ET_ECF_EXTERNAL_CFLAG.fill_external_cflags */
extern void T803f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_CFLAG.add_external_value */
extern void T803f9(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_CFLAG.is_enabled */
extern T1 T803f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG].do_all */
extern void T1583f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_CFLAG].item */
extern T0* T1583f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_EXTERNAL_MAKES.fill_external_makes */
extern void T795f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature detachable ET_ECF_EXTERNAL_MAKES.fill_external_makes */
extern T0* T795f5ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_MAKE]].set_rout_disp_final */
extern void T1798f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable ET_ECF_EXTERNAL_MAKE.fill_external_makes */
extern void T807f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_MAKE.add_external_value */
extern void T807f13(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_MAKE.file_system */
extern T0* T807f5(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_MAKE.unix_file_system */
extern T0* T807f8(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_MAKE.windows_file_system */
extern T0* T807f7(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_MAKE.operating_system */
extern T0* T807f6(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_MAKE.is_enabled */
extern T1 T807f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE].do_all */
extern void T1587f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_MAKE].item */
extern T0* T1587f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_EXTERNAL_RESOURCES.fill_external_resources */
extern void T797f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature detachable ET_ECF_EXTERNAL_RESOURCES.fill_external_resources */
extern T0* T797f5ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_RESOURCE]].set_rout_disp_final */
extern void T1802f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable ET_ECF_EXTERNAL_RESOURCE.fill_external_resources */
extern void T809f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_RESOURCE.add_external_value */
extern void T809f13(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_RESOURCE.file_system */
extern T0* T809f5(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_RESOURCE.unix_file_system */
extern T0* T809f8(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_RESOURCE.windows_file_system */
extern T0* T809f7(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_RESOURCE.operating_system */
extern T0* T809f6(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_RESOURCE.is_enabled */
extern T1 T809f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE].do_all */
extern void T1589f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_RESOURCE].item */
extern T0* T1589f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_EXTERNAL_LIBRARIES.fill_external_libraries */
extern void T793f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature detachable ET_ECF_EXTERNAL_LIBRARIES.fill_external_libraries */
extern T0* T793f5ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_LIBRARY]].set_rout_disp_final */
extern void T1794f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable ET_ECF_EXTERNAL_LIBRARY.fill_external_libraries */
extern void T805f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_LIBRARY.add_external_value */
extern void T805f13(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_LIBRARY.file_system */
extern T0* T805f5(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_LIBRARY.unix_file_system */
extern T0* T805f8(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_LIBRARY.windows_file_system */
extern T0* T805f7(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_LIBRARY.operating_system */
extern T0* T805f6(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_LIBRARY.is_enabled */
extern T1 T805f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY].do_all */
extern void T1585f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_LIBRARY].item */
extern T0* T1585f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_EXTERNAL_OBJECTS.fill_external_objects */
extern void T796f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature detachable ET_ECF_EXTERNAL_OBJECTS.fill_external_objects */
extern T0* T796f5ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_OBJECT]].set_rout_disp_final */
extern void T1800f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable ET_ECF_EXTERNAL_OBJECT.fill_external_objects */
extern void T808f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_OBJECT.add_external_value */
extern void T808f13(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_OBJECT.file_system */
extern T0* T808f5(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_OBJECT.unix_file_system */
extern T0* T808f8(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_OBJECT.windows_file_system */
extern T0* T808f7(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_OBJECT.operating_system */
extern T0* T808f6(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_OBJECT.is_enabled */
extern T1 T808f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT].do_all */
extern void T1588f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_OBJECT].item */
extern T0* T1588f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_EXTERNAL_INCLUDES.fill_external_includes */
extern void T792f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature detachable ET_ECF_EXTERNAL_INCLUDES.fill_external_includes */
extern T0* T792f5ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_EXTERNAL_INCLUDE]].set_rout_disp_final */
extern void T1792f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable ET_ECF_EXTERNAL_INCLUDE.fill_external_includes */
extern void T804f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_INCLUDE.add_external_value */
extern void T804f13(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_EXTERNAL_INCLUDE.file_system */
extern T0* T804f5(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_INCLUDE.unix_file_system */
extern T0* T804f8(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_INCLUDE.windows_file_system */
extern T0* T804f7(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_INCLUDE.operating_system */
extern T0* T804f6(GE_context* ac, T0* C);
/* detachable ET_ECF_EXTERNAL_INCLUDE.is_enabled */
extern T1 T804f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE].do_all */
extern void T1584f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_EXTERNAL_INCLUDE].item */
extern T0* T1584f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM.set_dotnet */
extern void T58f244(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_ADAPTED_DOTNET_ASSEMBLIES.put_last */
extern void T210f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].force_last */
extern void T594f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].force */
extern void T1396f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].fill_with */
extern void T1395f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].extend */
extern void T1395f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].resize */
extern void T594f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].aliased_resized_area */
extern T0* T1396f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].aliased_resized_area */
extern T0* T1395f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].new_capacity */
extern T6 T594f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_DOTNET_ASSEMBLY].extendible */
extern T1 T594f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.is_enabled */
extern T1 T213f21(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ADAPTED_DOTNET_ASSEMBLIES.dotnet_assembly */
extern T0* T789f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY].item */
extern T0* T1581f9(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_ADAPTED_DOTNET_ASSEMBLIES.count */
extern T6 T789f3(GE_context* ac, T0* C);
/* [detachable] ET_ECF_STATE.is_dotnet */
extern T1 T183f3(GE_context* ac, T0* C);
/* [detachable] ET_ADAPTED_LIBRARIES.put_last */
extern void T188f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY].force_last */
extern void T578f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_LIBRARY].force */
extern void T1391f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_LIBRARY].fill_with */
extern void T1390f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_LIBRARY].extend */
extern void T1390f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY].resize */
extern void T578f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ADAPTED_LIBRARY].aliased_resized_area */
extern T0* T1391f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ADAPTED_LIBRARY].aliased_resized_area */
extern T0* T1390f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY].new_capacity */
extern T6 T578f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ADAPTED_LIBRARY].extendible */
extern T1 T578f8(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.is_enabled */
extern T1 T207f19(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ADAPTED_LIBRARIES.library */
extern T0* T788f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_ADAPTED_LIBRARY].item */
extern T0* T1580f9(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_ADAPTED_LIBRARIES.count */
extern T6 T788f3(GE_context* ac, T0* C);
/* detachable ET_ECF_CLUSTER.fill_file_rules */
extern void T802f93(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature detachable ET_ECF_CLUSTER.fill_file_rules */
extern T0* T802f93ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_CLUSTER]].set_rout_disp_final */
extern void T1733f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* detachable ET_ECF_CLUSTERS.do_all */
extern void T787f12(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_FILE_RULES.fill_file_rules */
extern void T790f7(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature detachable ET_ECF_FILE_RULES.fill_file_rules */
extern T0* T790f7ac1(GE_context* ac, T0* a1);
/* Creation of agent #2 in feature detachable ET_ECF_FILE_RULES.fill_file_rules */
extern T0* T790f7ac2(GE_context* ac, T0* a1);
/* detachable ET_ECF_FILE_RULE.is_enabled */
extern T1 T810f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_FILE_RULE]].set_rout_disp_final */
extern void T2058f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_FILE_RULE].do_if */
extern void T1582f16(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_TARGET.fill_file_rules */
extern void T59f67(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_FILE_RULES.make_empty */
extern T0* T790c6(GE_context* ac);
/* detachable ET_ECF_CLUSTER.fill_subclusters */
extern void T802f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_CLUSTERS.put_last */
extern void T564f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLUSTER].force_last */
extern void T1384f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLUSTER].force */
extern void T1726f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLUSTER].fill_with */
extern void T1725f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_CLUSTER].extend */
extern void T1725f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLUSTER].resize */
extern void T1384f13(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_CLUSTER].aliased_resized_area */
extern T0* T1726f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_CLUSTER].aliased_resized_area */
extern T0* T1725f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLUSTER].new_capacity */
extern T6 T1384f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_CLUSTER].extendible */
extern T1 T1384f8(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_CLUSTER.is_enabled */
extern T1 T802f67(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTERS.cluster */
extern T0* T787f4(GE_context* ac, T0* C, T6 a1);
/* detachable ET_ECF_CLUSTERS.count */
extern T6 T787f3(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ERROR_HANDLER.report_epcd_error */
extern void T56f27(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ECF_ERROR.make_epcd */
extern T0* T214c97(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* [detachable] ET_ECF_ERROR_HANDLER.report_epcl_error */
extern void T56f26(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ERROR.make_epcl */
extern T0* T214c96(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] ET_ECF_ERROR_HANDLER.report_epcr_error */
extern void T56f25(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ECF_ERROR.make_epcr */
extern T0* T214c95(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eadq_error */
extern void T56f24(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_eadq */
extern T0* T214c94(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].before */
extern T1 T199f42(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].cursor_before */
extern T1 T199f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_TARGET, [attached] STRING_8].before */
extern T1 T586f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].back */
extern void T199f48(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].cursor_back */
extern void T199f60(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].add_traversing_cursor */
extern void T199f67(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_TARGET, [attached] STRING_8].set_next_cursor */
extern void T586f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_TARGET, [attached] STRING_8].set_position */
extern void T586f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].remove_traversing_cursor */
extern void T199f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_TARGET, [attached] STRING_8].set_before */
extern void T586f13(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].clashes_item */
extern T6 T199f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].finish */
extern void T199f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].cursor_finish */
extern void T199f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].cursor_off */
extern T1 T199f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_TARGET, [attached] STRING_8].off */
extern T1 T586f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].is_empty */
extern T1 T199f36(GE_context* ac, T0* C);
/* detachable ET_ECF_TARGET_PARENT.set_target */
extern void T200f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_ERROR_HANDLER.report_epcy_error */
extern void T56f23(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_epcy */
extern T0* T214c93(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].item_for_iteration */
extern T0* T199f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].cursor_item */
extern T0* T199f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].item_storage_item */
extern T0* T199f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].forth */
extern void T199f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].cursor_forth */
extern void T199f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_TARGET, [attached] STRING_8].set_after */
extern void T586f10(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].key_for_iteration */
extern T0* T199f35(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].cursor_key */
extern T0* T199f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].key_storage_item */
extern T0* T199f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].after */
extern T1 T199f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].cursor_after */
extern T1 T199f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_TARGET, [attached] STRING_8].after */
extern T1 T586f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].start */
extern void T199f45(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].cursor_start */
extern void T199f57(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].has */
extern T1 T199f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].search_position */
extern void T199f49(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].hash_position */
extern T6 T199f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].slots_item */
extern T6 T199f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_ERROR_HANDLER.report_internal_error */
extern void T56f22(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR.make_internal */
extern T0* T214c92(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR_HANDLER.report_epud_error */
extern void T56f21(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_epud */
extern T0* T214c91(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_LIBRARY.library_target */
extern T0* T187f10(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_CONFIG.library_target */
extern T0* T185f11(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.library_target */
extern T0* T58f3(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ERROR_HANDLER.report_epul_error */
extern void T56f20(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR.make_epul */
extern T0* T214c90(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR_HANDLER.report_epur_error */
extern void T56f19(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_epur */
extern T0* T214c89(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_LIBRARY.targets */
extern T0* T187f7(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_CONFIG.targets */
extern T0* T185f8(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.targets */
extern T0* T58f2(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ERROR_HANDLER.report_epfe_error */
extern void T56f18(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_epfe */
extern T0* T214c88(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].found_item */
extern T0* T198f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].item_storage_item */
extern T0* T198f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].found */
extern T1 T198f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].search */
extern void T198f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].search_position */
extern void T198f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].hash_position */
extern T6 T198f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].key_storage_item */
extern T0* T198f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].clashes_item */
extern T6 T198f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].slots_item */
extern T6 T198f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_CELL [detachable ET_ECF_SYSTEM_CONFIG].make */
extern T0* T201c2(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].force_last */
extern void T199f44(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].key_storage_put */
extern void T199f56(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].slots_put */
extern void T199f55(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].clashes_put */
extern void T199f54(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].resize */
extern void T199f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].clashes_resize */
extern void T199f65(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].special_integer_ */
extern T0* T199f39(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].key_storage_resize */
extern void T199f64(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].item_storage_resize */
extern void T199f63(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].slots_resize */
extern void T199f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].new_modulus */
extern T6 T199f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].new_capacity */
extern T6 T199f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].item_storage_put */
extern void T199f52(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].unset_found_item */
extern void T199f51(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].make_default */
extern T0* T199c43(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].make */
extern void T199f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].make_with_equality_testers */
extern void T199f61(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].make_sparse_container */
extern void T199f68(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].set_internal_cursor */
extern void T199f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].new_cursor */
extern T0* T199f41(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_TARGET, [attached] STRING_8].make */
extern T0* T586c9(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].make_slots */
extern void T199f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].make_clashes */
extern void T199f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].make_key_storage */
extern void T199f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].make_item_storage */
extern void T199f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_TARGET, [attached] STRING_8].default_capacity */
extern T6 T199f40(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].force_last */
extern void T198f35(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].key_storage_put */
extern void T198f44(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].slots_put */
extern void T198f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].clashes_put */
extern void T198f42(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].resize */
extern void T198f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].clashes_resize */
extern void T198f49(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].special_integer_ */
extern T0* T198f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].key_storage_resize */
extern void T198f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].item_storage_resize */
extern void T198f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_SYSTEM_CONFIG].aliased_resized_area */
extern T0* T584f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] ET_ECF_SYSTEM_CONFIG].aliased_resized_area */
extern T0* T582f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].slots_resize */
extern void T198f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].new_modulus */
extern T6 T198f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].new_capacity */
extern T6 T198f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].item_storage_put */
extern void T198f40(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_SYSTEM_CONFIG].force */
extern void T584f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_SYSTEM_CONFIG].fill_with */
extern void T582f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_ECF_SYSTEM_CONFIG].extend */
extern void T582f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].unset_found_item */
extern void T198f38(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].make_default */
extern T0* T198c34(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].make */
extern void T198f37(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].make_with_equality_testers */
extern void T198f45(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].make_sparse_container */
extern void T198f50(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].set_internal_cursor */
extern void T198f55(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].new_cursor */
extern T0* T198f31(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].make */
extern T0* T585c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].make_slots */
extern void T198f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].make_clashes */
extern void T198f53(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].make_key_storage */
extern void T198f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].make_item_storage */
extern void T198f51(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_SYSTEM_CONFIG].make */
extern T0* T584f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_SYSTEM_CONFIG].make_empty */
extern T0* T582c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_SYSTEM_CONFIG].default_create */
extern T0* T584c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_SYSTEM_CONFIG, [attached] STRING_8].default_capacity */
extern T6 T198f27(GE_context* ac, T0* C);
/* [detachable] ET_ECF_LIBRARY.select_target */
extern void T187f178(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.select_target */
extern void T58f164(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_STATE.make */
extern T0* T183c4(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR_HANDLER.report_etnu_error */
extern void T56f11(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_etnu */
extern T0* T214c81(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_SYSTEM_PARSER.element_name */
extern T0* T55f17(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.target_with_name */
extern T0* T58f141(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_TARGETS.target */
extern T0* T203f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.new_system */
extern T0* T55f16(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_AST_FACTORY.new_system */
extern T0* T162f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.make */
extern T0* T58c163(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.make_system */
extern void T58f176(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.set_kernel_types */
extern void T58f189(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_wide_character_type */
extern void T58f239(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.tokens */
extern T0* T58f144(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_typed_pointer_type */
extern void T58f238(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_type_type */
extern void T58f237(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_system_string_type */
extern void T58f236(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_system_object_type */
extern void T58f235(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_string_type */
extern void T58f234(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_string_32_type */
extern void T58f233(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_string_8_type */
extern void T58f232(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_special_type */
extern void T58f231(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_routine_type */
extern void T58f230(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_real_64_type */
extern void T58f229(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_real_32_type */
extern void T58f228(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_real_type */
extern void T58f227(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_procedure_type */
extern void T58f226(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_predicate_type */
extern void T58f225(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_pointer_type */
extern void T58f224(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_natural_64_type */
extern void T58f223(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_natural_32_type */
extern void T58f222(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_natural_16_type */
extern void T58f221(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_natural_8_type */
extern void T58f220(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_natural_type */
extern void T58f219(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_iterable_type */
extern void T58f218(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_ise_exception_manager_type */
extern void T58f217(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_integer_64_type */
extern void T58f216(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_integer_32_type */
extern void T58f215(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_integer_16_type */
extern void T58f214(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_integer_8_type */
extern void T58f213(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_integer_type */
extern void T58f212(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_immutable_string_32_type */
extern void T58f211(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_immutable_string_8_type */
extern void T58f210(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_function_type */
extern void T58f209(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_exception_manager_type */
extern void T58f208(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_exception_type */
extern void T58f207(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_double_type */
extern void T58f206(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_character_32_type */
extern void T58f205(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_character_8_type */
extern void T58f204(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_character_type */
extern void T58f203(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_boolean_type */
extern void T58f202(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_array_type */
extern void T58f201(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_tuple_type */
extern void T58f200(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_none_type */
extern void T58f199(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_any_type */
extern void T58f198(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.make_adapted */
extern void T58f188(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM.default_read_only_value */
extern T1 T58f145(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_unknown_kernel_types */
extern void T58f187(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.set_system_name */
extern void T58f186(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM.initialize */
extern void T58f185(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.initialize */
extern void T58f185p1(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM.class_name_tester */
extern T0* T58f146(GE_context* ac, T0* C);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_TARGET]].set_rout_disp_final */
extern void T165f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_ECF_SYSTEM_PARSER.override_target */
extern void T55f216(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.override_variables */
extern void T59f59(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.override_capabilities */
extern void T59f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8].forth */
extern void T601f11(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].cursor_forth */
extern void T221f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].add_traversing_cursor */
extern void T221f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8].set_next_cursor */
extern void T601f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8].set_position */
extern void T601f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].remove_traversing_cursor */
extern void T221f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8].set_after */
extern void T601f13(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8].item */
extern T0* T601f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].cursor_item */
extern T0* T221f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8].key */
extern T0* T601f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].cursor_key */
extern T0* T221f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8].after */
extern T1 T601f4(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].new_iterator */
extern T0* T221f34(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8].start */
extern void T601f12(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].cursor_start */
extern void T221f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].cursor_off */
extern T1 T221f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_IDENTIFIER, [attached] STRING_8].off */
extern T1 T601f9(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_IDENTIFIER, [attached] STRING_8].is_empty */
extern T1 T221f37(GE_context* ac, T0* C);
/* [detachable] ET_ECF_TARGET.override_settings */
extern void T59f57(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_TARGET.override_settings */
extern T0* T59f57ac1(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_TARGET.override_all_assertions */
extern void T59f63(GE_context* ac, T0* C, T0* a1);
/* Creation of agent #1 in feature [detachable] ET_ECF_TARGET.override_all_assertions */
extern T0* T59f63ac1(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #2 in feature [detachable] ET_ECF_TARGET.override_all_assertions */
extern T0* T59f63ac2(GE_context* ac, T0* a1, T0* a2);
/* Creation of agent #3 in feature [detachable] ET_ECF_TARGET.override_all_assertions */
extern T0* T59f63ac3(GE_context* ac, T0* a1, T0* a2);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_ADAPTED_DOTNET_ASSEMBLY]].set_rout_disp_final */
extern void T1739f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] PROCEDURE [[attached] TUPLE [[attached] ET_ECF_ADAPTED_LIBRARY]].set_rout_disp_final */
extern void T1737f7(GE_context* ac, T0* C, T14 a1, T14 a2, T14 a3, T0* a4, T1 a5, T6 a6);
/* [detachable] ET_ECF_TARGET.override_all_assertions_in_cluster */
extern void T59f65(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature [detachable] ET_ECF_TARGET.override_all_assertions_in_cluster */
extern T0* T59f65ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_TARGET.override_all_assertions_in_adapted_group */
extern void T59f66(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature [detachable] ET_ECF_TARGET.override_all_assertions_in_adapted_group */
extern T0* T59f66ac1(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.class_options */
extern T0* T213f13(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.class_options */
extern T0* T207f10(GE_context* ac, T0* C);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.class_options */
extern T0* T799f11(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ADAPTED_DOTNET_ASSEMBLY.options */
extern T0* T213f12(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ADAPTED_LIBRARY.options */
extern T0* T207f9(GE_context* ac, T0* C);
/* detachable ET_ECF_ADAPTED_PRECOMPILED_LIBRARY.options */
extern T0* T799f10(GE_context* ac, T0* C);
/* detachable ET_ECF_ADAPTED_DOTNET_ASSEMBLIES.do_adapted */
extern void T789f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_ADAPTED_LIBRARIES.do_adapted */
extern void T788f7(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_CLUSTERS.do_explicit */
extern void T787f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_TARGET.override_all_assertions_in_option */
extern void T59f64(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Creation of agent #1 in feature [detachable] ET_ECF_TARGET.override_all_assertions_in_option */
extern T0* T59f64ac1(GE_context* ac, T0* a1, T0* a2);
/* detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].do_all */
extern void T591f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].forth */
extern void T2034f7(GE_context* ac, T0* C);
/* detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].cursor_forth */
extern void T591f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].item */
extern T0* T2034f4(GE_context* ac, T0* C);
/* detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].cursor_item */
extern T0* T591f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].after */
extern T1 T2034f3(GE_context* ac, T0* C);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].start */
extern void T2034f6(GE_context* ac, T0* C);
/* detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].cursor_start */
extern void T591f11(GE_context* ac, T0* C, T0* a1);
/* detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].new_cursor */
extern T0* T591f4(GE_context* ac, T0* C);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].make */
extern T0* T2034c5(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].keys */
extern T0* T204f35(GE_context* ac, T0* C);
/* detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].make */
extern T0* T591c7(GE_context* ac, T0* a1);
/* detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].set_internal_cursor */
extern void T591f10(GE_context* ac, T0* C, T0* a1);
/* detachable DS_SPARSE_TABLE_KEYS [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].new_cursor_with_table_cursor */
extern T0* T591f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_SPARSE_TABLE_KEYS_CURSOR [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].make_with_table_cursor */
extern T0* T2034c8(GE_context* ac, T0* a1, T0* a2);
/* detachable ET_ECF_TARGETS.do_all */
extern void T203f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ECF_TARGET].do_all */
extern void T906f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.parse_file_with_action */
extern void T55f218(GE_context* ac, T0* C, T0* a1, T1 a2, T0* a3);
/* [detachable] XM_EIFFEL_PARSER.last_error_description */
extern T0* T178f120(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.position */
extern T0* T178f114(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] XM_POSITION].first */
extern T0* T519f6(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.positions */
extern T0* T178f118(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.new_positions */
extern T0* T178f88(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_STACK [[attached] XM_EIFFEL_SCANNER].remove */
extern void T520f7(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_STACK [[attached] XM_EIFFEL_SCANNER].item */
extern T0* T520f4(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_STACK [[attached] XM_EIFFEL_SCANNER].is_empty */
extern T1 T520f3(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] XM_POSITION].force_last */
extern void T519f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] XM_POSITION].put_right */
extern void T1358f5(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] XM_POSITION].attach_left */
extern void T1358f6(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] XM_POSITION].make */
extern T0* T1358c4(GE_context* ac, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] XM_POSITION].make */
extern T0* T519c7(GE_context* ac);
/* [detachable] DS_BILINKED_LIST [[attached] XM_POSITION].set_internal_cursor */
extern void T519f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] XM_POSITION].new_cursor */
extern T0* T519f5(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_POSITION].make */
extern T0* T1359c3(GE_context* ac, T0* a1);
/* [detachable] DS_LINKED_STACK [[attached] XM_EIFFEL_SCANNER].copy */
extern void T520f6(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] XM_EIFFEL_SCANNER].put_right */
extern void T1360f4(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] XM_EIFFEL_SCANNER].make */
extern T0* T1360c3(GE_context* ac, T0* a1);
/* [detachable] DS_LINKED_STACK [[attached] XM_EIFFEL_SCANNER].make */
extern T0* T520c5(GE_context* ac);
/* [detachable] XM_EIFFEL_ENTITY_DEF.error_position */
extern T0* T556f59(GE_context* ac, T0* C);
/* [detachable] XM_DEFAULT_POSITION.make */
extern T0* T517c5(GE_context* ac, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.error_position */
extern T0* T559f61(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.error_position */
extern T0* T557f59(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.error_position */
extern T0* T516f70(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ERROR_HANDLER.report_syntax_error */
extern void T56f17(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_syntax */
extern T0* T214c87(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_TREE_CALLBACKS_PIPE.last_error */
extern T0* T179f5(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eabx_error */
extern void T56f16(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR.make_eabx */
extern T0* T214c86(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eadf_error */
extern void T56f15(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_eadf */
extern T0* T214c85(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eadi_error */
extern void T56f14(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_ERROR.make_eadi */
extern T0* T214c84(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] ET_ECF_SYSTEM_PARSER.string_equality_tester */
extern T0* T55f13(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.execution_environment */
extern T0* T55f25(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eadh_error */
extern void T56f13(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR.make_eadh */
extern T0* T214c83(GE_context* ac, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR_HANDLER.report_eadg_error */
extern void T56f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_ERROR.make_eadg */
extern T0* T214c82(GE_context* ac, T0* a1, T0* a2);
/* [detachable] XM_TREE_CALLBACKS_PIPE.document */
extern T0* T179f8(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.is_correct */
extern T1 T178f90(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.parse_from_stream */
extern void T178f208(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.parse_with_events */
extern void T178f213(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.on_finish */
extern void T178f219(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.on_finish */
extern void T560f29(GE_context* ac, T0* C);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_finish */
extern void T195f10(GE_context* ac, T0* C);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_finish */
extern void T195f10p1(GE_context* ac, T0* C);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.on_finish */
extern void T182f15(GE_context* ac, T0* C);
/* [detachable] XM_CALLBACKS_NULL.on_finish */
extern void T521f3(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.parse */
extern void T178f218(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yy_pop_last_value */
extern void T178f228(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PARSER.yy_push_error_value */
extern void T178f227(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yy_do_action */
extern void T178f226(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PARSER.on_notation_declaration */
extern void T178f256(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_DTD_CALLBACKS_NULL.on_notation_declaration */
extern void T526f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_DTD_EXTERNAL_ID.set_public */
extern void T536f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_EXTERNAL_ID].aliased_resized_area */
extern T0* T535f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] XM_DTD_EXTERNAL_ID].aliased_resized_area */
extern T0* T537f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_DTD_EXTERNAL_ID.set_system */
extern void T536f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.new_dtd_external_id */
extern T0* T178f168(GE_context* ac, T0* C);
/* [detachable] XM_DTD_EXTERNAL_ID.make */
extern T0* T536c3(GE_context* ac);
/* [detachable] XM_EIFFEL_PARSER.when_pe_entity_declared */
extern void T178f255(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].force_new */
extern void T522f38(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].key_storage_put */
extern void T522f52(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].item_storage_put */
extern void T522f51(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_ENTITY_DEF].force */
extern void T1365f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_ENTITY_DEF].fill_with */
extern void T1361f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_ENTITY_DEF].extend */
extern void T1361f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].slots_put */
extern void T522f50(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].clashes_put */
extern void T522f49(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].slots_item */
extern T6 T522f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].hash_position */
extern T6 T522f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].clashes_item */
extern T6 T522f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].resize */
extern void T522f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].clashes_resize */
extern void T522f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].special_integer_ */
extern T0* T522f33(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].key_storage_resize */
extern void T522f56(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].item_storage_resize */
extern void T522f55(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_ENTITY_DEF].aliased_resized_area */
extern T0* T1365f1(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_ENTITY_DEF].aliased_resized_area */
extern T0* T1361f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].key_storage_item */
extern T0* T522f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].slots_resize */
extern void T522f54(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].new_modulus */
extern T6 T522f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].new_capacity */
extern T6 T522f24(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].unset_found_item */
extern void T522f43(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].has */
extern T1 T522f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].search_position */
extern void T522f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.make_def */
extern T0* T559c214(GE_context* ac, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.make_literal */
extern void T559f216(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_NULL_EXTERNAL_RESOLVER.default_create */
extern T0* T524c3(GE_context* ac);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.make_scanner */
extern void T559f217(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.reset */
extern void T559f219(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.reset_sent */
extern void T559f222(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.reset */
extern void T559f219p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.reset */
extern void T559f219p0(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.reset */
extern void T559f219p2(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.reset */
extern void T559f219p3(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.make_with_buffer */
extern void T559f218(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_load_input_buffer */
extern void T559f221(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_set_content */
extern void T559f224(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_initialize */
extern void T559f220(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_initialize */
extern void T559f220p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_build_tables */
extern void T559f223(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_accept_template */
extern T0* T559f54(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_fixed_array */
extern T0* T559f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_accept_template_4 */
extern void T559f255(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_array_subcopy */
extern void T559f279(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.integer_array_ */
extern T0* T559f60(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_accept_template_3 */
extern void T559f254(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_accept_template_2 */
extern void T559f253(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_accept_template_1 */
extern void T559f252(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_meta_template */
extern T0* T559f52(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_ec_template */
extern T0* T559f73(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_ec_template_2 */
extern void T559f231(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_ec_template_1 */
extern void T559f230(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_def_template */
extern T0* T559f68(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_def_template_5 */
extern void T559f229(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_def_template_4 */
extern void T559f228(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_def_template_3 */
extern void T559f227(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_def_template_2 */
extern void T559f226(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_def_template_1 */
extern void T559f225(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_base_template */
extern T0* T559f57(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_base_template_5 */
extern void T559f278(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_base_template_4 */
extern void T559f277(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_base_template_3 */
extern void T559f276(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_base_template_2 */
extern void T559f275(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_base_template_1 */
extern void T559f274(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template */
extern T0* T559f55(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_18 */
extern void T559f273(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_17 */
extern void T559f272(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_16 */
extern void T559f271(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_15 */
extern void T559f270(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_14 */
extern void T559f269(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_13 */
extern void T559f268(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_12 */
extern void T559f267(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_11 */
extern void T559f266(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_10 */
extern void T559f265(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_9 */
extern void T559f264(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_8 */
extern void T559f263(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_7 */
extern void T559f262(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_6 */
extern void T559f261(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_5 */
extern void T559f260(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_4 */
extern void T559f259(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_3 */
extern void T559f258(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_2 */
extern void T559f257(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_chk_template_1 */
extern void T559f256(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template */
extern T0* T559f97(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_20 */
extern void T559f251(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_19 */
extern void T559f250(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_18 */
extern void T559f249(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_17 */
extern void T559f248(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_16 */
extern void T559f247(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_15 */
extern void T559f246(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_14 */
extern void T559f245(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_13 */
extern void T559f244(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_12 */
extern void T559f243(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_11 */
extern void T559f242(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_10 */
extern void T559f241(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_9 */
extern void T559f240(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_8 */
extern void T559f239(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_7 */
extern void T559f238(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_6 */
extern void T559f237(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_5 */
extern void T559f236(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_4 */
extern void T559f235(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_3 */
extern void T559f234(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_2 */
extern void T559f233(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_nxt_template_1 */
extern void T559f232(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.special_integer_ */
extern T0* T559f56(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.empty_buffer */
extern T0* T559f53(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_CHARACTER_ENTITY.make */
extern T0* T1356c8(GE_context* ac);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.make_external */
extern void T559f215(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_EIFFEL_PARSER.new_external_entity */
extern T0* T178f167(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.make_external */
extern T0* T556c213(GE_context* ac, T0* a1, T0* a2);
/* [detachable] XM_EIFFEL_ENTITY_DEF.make_scanner */
extern void T556f214(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.reset */
extern void T556f216(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.reset */
extern void T556f216p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.reset */
extern void T556f216p0(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.reset */
extern void T556f216p2(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.make_with_buffer */
extern void T556f215(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_load_input_buffer */
extern void T556f218(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_set_content */
extern void T556f220(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_initialize */
extern void T556f217(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_initialize */
extern void T556f217p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_build_tables */
extern void T556f219(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_accept_template */
extern T0* T556f52(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_fixed_array */
extern T0* T556f56(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_accept_template_4 */
extern void T556f251(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_array_subcopy */
extern void T556f275(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] XM_EIFFEL_ENTITY_DEF.integer_array_ */
extern T0* T556f58(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_accept_template_3 */
extern void T556f250(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_accept_template_2 */
extern void T556f249(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_accept_template_1 */
extern void T556f248(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_meta_template */
extern T0* T556f50(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_ec_template */
extern T0* T556f71(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_ec_template_2 */
extern void T556f227(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_ec_template_1 */
extern void T556f226(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_def_template */
extern T0* T556f66(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_def_template_5 */
extern void T556f225(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_def_template_4 */
extern void T556f224(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_def_template_3 */
extern void T556f223(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_def_template_2 */
extern void T556f222(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_def_template_1 */
extern void T556f221(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_base_template */
extern T0* T556f55(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_base_template_5 */
extern void T556f274(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_base_template_4 */
extern void T556f273(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_base_template_3 */
extern void T556f272(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_base_template_2 */
extern void T556f271(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_base_template_1 */
extern void T556f270(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template */
extern T0* T556f53(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_18 */
extern void T556f269(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_17 */
extern void T556f268(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_16 */
extern void T556f267(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_15 */
extern void T556f266(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_14 */
extern void T556f265(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_13 */
extern void T556f264(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_12 */
extern void T556f263(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_11 */
extern void T556f262(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_10 */
extern void T556f261(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_9 */
extern void T556f260(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_8 */
extern void T556f259(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_7 */
extern void T556f258(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_6 */
extern void T556f257(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_5 */
extern void T556f256(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_4 */
extern void T556f255(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_3 */
extern void T556f254(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_2 */
extern void T556f253(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_chk_template_1 */
extern void T556f252(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template */
extern T0* T556f94(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_20 */
extern void T556f247(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_19 */
extern void T556f246(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_18 */
extern void T556f245(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_17 */
extern void T556f244(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_16 */
extern void T556f243(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_15 */
extern void T556f242(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_14 */
extern void T556f241(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_13 */
extern void T556f240(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_12 */
extern void T556f239(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_11 */
extern void T556f238(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_10 */
extern void T556f237(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_9 */
extern void T556f236(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_8 */
extern void T556f235(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_7 */
extern void T556f234(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_6 */
extern void T556f233(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_5 */
extern void T556f232(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_4 */
extern void T556f231(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_3 */
extern void T556f230(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_2 */
extern void T556f229(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_nxt_template_1 */
extern void T556f228(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.special_integer_ */
extern T0* T556f70(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.empty_buffer */
extern T0* T556f60(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.on_entity_declaration */
extern void T178f254(GE_context* ac, T0* C, T0* a1, T1 a2, T0* a3, T0* a4, T0* a5);
/* [detachable] XM_DTD_CALLBACKS_NULL.on_entity_declaration */
extern void T526f7(GE_context* ac, T0* C, T0* a1, T1 a2, T0* a3, T0* a4, T0* a5);
/* [detachable] XM_EIFFEL_PARSER.when_entity_declared */
extern void T178f253(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_EIFFEL_PARSER.new_literal_entity */
extern T0* T178f166(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_EIFFEL_ENTITY_DEF.make_literal */
extern T0* T556c212(GE_context* ac, T0* a1, T0* a2);
/* [detachable] XM_EIFFEL_ENTITY_DEF.push_start_condition_dtd_ignore */
extern void T556f281(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.push_start_condition */
extern void T556f290(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.set_start_condition */
extern void T556f277(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.start_condition */
extern T6 T556f63(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.push_start_condition_dtd_ignore */
extern void T559f285(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.push_start_condition */
extern void T559f295(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.set_start_condition */
extern void T559f281(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.start_condition */
extern T6 T559f65(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.push_start_condition_dtd_ignore */
extern void T557f209(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.push_start_condition */
extern void T557f223(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.set_start_condition */
extern void T557f212(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.start_condition */
extern T6 T557f53(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.special_integer_ */
extern T0* T557f65(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.push_start_condition_dtd_ignore */
extern void T516f207(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.push_start_condition */
extern void T516f218(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER.set_start_condition */
extern void T516f226(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER.start_condition */
extern T6 T516f46(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.special_integer_ */
extern T0* T516f54(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_default_value */
extern void T545f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_value_fixed */
extern void T545f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_value_implied */
extern void T545f24(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_value_required */
extern void T545f23(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] STRING_8]].aliased_resized_area */
extern T0* T547f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] STRING_8]].aliased_resized_area */
extern T0* T549f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].force_last */
extern void T548f11(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] STRING_8].put_right */
extern void T1374f5(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] STRING_8].attach_left */
extern void T1374f6(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] STRING_8].make */
extern T0* T1374c4(GE_context* ac, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.new_string_bilinked_list */
extern T0* T178f163(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].set_equality_tester */
extern void T548f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.string_equality_tester */
extern T0* T178f177(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].make */
extern T0* T548c10(GE_context* ac);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].set_internal_cursor */
extern void T548f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].new_cursor */
extern T0* T548f6(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] STRING_8].make */
extern T0* T1373c8(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] STRING_8]].force */
extern void T547f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] STRING_8]].fill_with */
extern void T549f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] STRING_8]].extend */
extern void T549f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_enumeration_list */
extern void T545f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_enumeration */
extern void T545f27(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.default_enumeration_list */
extern T0* T545f11(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_notation */
extern void T545f21(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_token */
extern void T545f20(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_entity */
extern void T545f19(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_list_type */
extern void T545f18(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_id_ref */
extern void T545f17(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_id */
extern void T545f16(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_data */
extern void T545f15(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.new_dtd_attribute_content */
extern T0* T178f162(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ATTRIBUTE_CONTENT].aliased_resized_area */
extern T0* T544f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] XM_DTD_ATTRIBUTE_CONTENT].aliased_resized_area */
extern T0* T546f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.make */
extern T0* T545c12(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ATTRIBUTE_CONTENT].force */
extern void T544f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_DTD_ATTRIBUTE_CONTENT].fill_with */
extern void T546f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_DTD_ATTRIBUTE_CONTENT].extend */
extern void T546f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.copy_default */
extern void T545f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.is_value_implied */
extern T1 T545f10(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.is_value_required */
extern T1 T545f9(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.has_default_value */
extern T1 T545f8(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.is_value_fixed */
extern T1 T545f7(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ATTRIBUTE_CONTENT.set_name */
extern void T545f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]].force */
extern void T541f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]].fill_with */
extern void T543f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]].extend */
extern void T543f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]].aliased_resized_area */
extern T0* T541f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]].aliased_resized_area */
extern T0* T543f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT].force_last */
extern void T542f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] XM_DTD_ATTRIBUTE_CONTENT].put_right */
extern void T1372f5(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] XM_DTD_ATTRIBUTE_CONTENT].attach_left */
extern void T1372f6(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] XM_DTD_ATTRIBUTE_CONTENT].make */
extern T0* T1372c4(GE_context* ac, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.new_dtd_attribute_content_list */
extern T0* T178f160(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT].make */
extern T0* T542c7(GE_context* ac);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT].set_internal_cursor */
extern void T542f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT].new_cursor */
extern T0* T542f5(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ATTRIBUTE_CONTENT].make */
extern T0* T1371c8(GE_context* ac, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.on_attribute_declarations */
extern void T178f252(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ATTRIBUTE_CONTENT].forth */
extern void T1371f10(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT].cursor_forth */
extern void T542f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT].add_traversing_cursor */
extern void T542f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ATTRIBUTE_CONTENT].set_next_cursor */
extern void T1371f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT].remove_traversing_cursor */
extern void T542f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ATTRIBUTE_CONTENT].set */
extern void T1371f11(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] XM_EIFFEL_PARSER.on_attribute_declaration */
extern void T178f262(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_DTD_CALLBACKS_NULL.on_attribute_declaration */
extern void T526f9(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ATTRIBUTE_CONTENT].item */
extern T0* T1371f6(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ATTRIBUTE_CONTENT].start */
extern void T1371f9(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT].cursor_start */
extern void T542f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT].cursor_off */
extern T1 T542f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ATTRIBUTE_CONTENT].off */
extern T1 T1371f7(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ELEMENT_CONTENT.set_zero_or_more */
extern void T539f11(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ELEMENT_CONTENT.make_mixed */
extern T0* T539c10(GE_context* ac);
/* [detachable] XM_DTD_ELEMENT_CONTENT.set_content_mixed */
extern void T539f21(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ELEMENT_CONTENT.set_choice */
extern void T539f19(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ELEMENT_CONTENT.make_list */
extern void T539f16(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ELEMENT_CONTENT.set_default */
extern void T539f22(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ELEMENT_CONTENT.set_one */
extern void T539f12(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ELEMENT_CONTENT.set_sequence */
extern void T539f20(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ELEMENT_CONTENT].make */
extern T0* T1370c6(GE_context* ac);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ELEMENT_CONTENT].set_internal_cursor */
extern void T1370f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ELEMENT_CONTENT].new_cursor */
extern T0* T1370f5(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] XM_DTD_ELEMENT_CONTENT].make */
extern T0* T1719c3(GE_context* ac, T0* a1);
/* [detachable] XM_DTD_ELEMENT_CONTENT.make_sequence */
extern T0* T539c9(GE_context* ac);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ELEMENT_CONTENT].force_last */
extern void T1370f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] XM_DTD_ELEMENT_CONTENT].put_right */
extern void T1720f5(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] XM_DTD_ELEMENT_CONTENT].attach_left */
extern void T1720f6(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] XM_DTD_ELEMENT_CONTENT].make */
extern T0* T1720c4(GE_context* ac, T0* a1);
/* [detachable] XM_DTD_ELEMENT_CONTENT.make_choice */
extern T0* T539c8(GE_context* ac);
/* [detachable] DS_BILINKED_LIST [[attached] XM_DTD_ELEMENT_CONTENT].force_first */
extern void T1370f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.element_name */
extern T0* T178f155(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DTD_ELEMENT_CONTENT.make_name */
extern T0* T539c15(GE_context* ac, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.set_element_repetition */
extern void T178f251(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_DTD_ELEMENT_CONTENT.set_zero_or_one */
extern void T539f14(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ELEMENT_CONTENT.set_one_or_more */
extern void T539f13(GE_context* ac, T0* C);
/* [detachable] XM_DTD_ELEMENT_CONTENT.make_any */
extern T0* T539c7(GE_context* ac);
/* [detachable] XM_DTD_ELEMENT_CONTENT.set_content_any */
extern void T539f18(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ELEMENT_CONTENT].force */
extern void T538f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_DTD_ELEMENT_CONTENT].fill_with */
extern void T540f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_DTD_ELEMENT_CONTENT].extend */
extern void T540f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ELEMENT_CONTENT].aliased_resized_area */
extern T0* T538f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] XM_DTD_ELEMENT_CONTENT].aliased_resized_area */
extern T0* T540f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_DTD_ELEMENT_CONTENT.make_empty */
extern T0* T539c6(GE_context* ac);
/* [detachable] XM_DTD_ELEMENT_CONTENT.set_content_empty */
extern void T539f17(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.on_element_declaration */
extern void T178f250(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_DTD_CALLBACKS_NULL.on_element_declaration */
extern void T526f6(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_EIFFEL_PARSER.on_attribute */
extern void T178f249(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] XM_NAMESPACE_RESOLVER.on_attribute */
extern void T560f36(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.add */
extern void T1375f14(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].force_new */
extern void T216f95(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].last */
extern T0* T1722f5(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.on_error */
extern void T560f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CALLBACKS_NULL.on_error */
extern void T521f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_error */
extern void T195f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_error */
extern void T195f17p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.on_error */
extern void T182f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.shallow_has */
extern T1 T1375f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER.attributes_force */
extern void T560f41(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] DS_LINKED_QUEUE [[attached] STRING_8].force */
extern void T1379f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_QUEUE [detachable STRING_8].force */
extern void T1378f7(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [detachable STRING_8].put_right */
extern void T1724f4(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [detachable STRING_8].make */
extern T0* T1724c3(GE_context* ac, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.add_default */
extern void T1375f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.default_pseudo_prefix */
extern T0* T1375f7(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.is_xmlns */
extern T1 T560f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER.same_string */
extern T1 T560f23(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_NAMESPACE_RESOLVER.string_equality_tester */
extern T0* T560f26(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.xmlns */
extern T0* T560f22(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.has_prefix */
extern T1 T560f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CALLBACKS_NULL.on_attribute */
extern void T521f10(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].has */
extern T1 T533f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].search_position */
extern void T533f33(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].hash_position */
extern T6 T533f24(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER_NAME.hash_code */
extern T6 T530f16(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER_NAME.item */
extern T0* T530f17(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].item */
extern T0* T548f8(GE_context* ac, T0* C, T6 a1);
/* detachable KL_EQUALITY_TESTER [[attached] XM_EIFFEL_PARSER_NAME].test */
extern T1 T1368f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].key_storage_item */
extern T0* T533f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].item_storage_item */
extern T0* T533f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].clashes_item */
extern T6 T533f23(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].key_equality_tester */
extern T0* T533f17(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].slots_item */
extern T6 T533f25(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]].force */
extern void T532f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]].fill_with */
extern void T534f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]].extend */
extern void T534f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]].aliased_resized_area */
extern T0* T532f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]].aliased_resized_area */
extern T0* T534f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].force_new */
extern void T533f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].item_storage_put */
extern void T533f39(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].slots_put */
extern void T533f38(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].clashes_put */
extern void T533f37(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].resize */
extern void T533f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].clashes_resize */
extern void T533f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].special_integer_ */
extern T0* T533f28(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].key_storage_resize */
extern void T533f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].item_storage_resize */
extern void T533f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].slots_resize */
extern void T533f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].new_modulus */
extern T6 T533f20(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].new_capacity */
extern T6 T533f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].unset_found_item */
extern void T533f35(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.new_name_set */
extern T0* T178f148(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].make_equal */
extern T0* T533c31(GE_context* ac, T6 a1);
/* detachable KL_EQUALITY_TESTER [[attached] XM_EIFFEL_PARSER_NAME].default_create */
extern T0* T1368c2(GE_context* ac);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].make */
extern void T533f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].set_internal_cursor */
extern void T533f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].new_cursor */
extern T0* T533f21(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] XM_EIFFEL_PARSER_NAME].make */
extern T0* T1369c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].make_slots */
extern void T533f43(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].make_clashes */
extern void T533f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].make_key_storage */
extern void T533f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME].make_item_storage */
extern void T533f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_PARSER_NAME].make */
extern T0* T529f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_PARSER_NAME].make_empty */
extern T0* T531c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_PARSER_NAME].default_create */
extern T0* T529c3(GE_context* ac);
/* [detachable] XM_EIFFEL_PARSER.on_start_tag */
extern void T178f248(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_NAMESPACE_RESOLVER.on_start_tag */
extern void T560f35(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.push */
extern void T1375f12(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].force_last */
extern void T1722f10(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].put_right */
extern void T1939f5(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].attach_left */
extern void T1939f6(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].make */
extern T0* T1939c4(GE_context* ac, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.new_string_string_table */
extern T0* T1375f8(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8].set_equality_tester */
extern void T216f94(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.string_equality_tester */
extern T0* T1375f9(GE_context* ac, T0* C);
/* [detachable] XM_CALLBACKS_NULL.on_start_tag */
extern void T521f9(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_EIFFEL_PARSER.on_end_tag */
extern void T178f247(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_NAMESPACE_RESOLVER.on_end_tag */
extern void T560f34(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.pop */
extern void T1375f11(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].remove_last */
extern void T1722f9(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].set_last_cell */
extern void T1722f14(GE_context* ac, T0* C, T0* a1);
/* detachable DS_BILINKABLE [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].forget_right */
extern void T1939f7(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].move_last_cursors_after */
extern void T1722f13(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].set_next_cursor */
extern void T1723f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].set_after */
extern void T1723f12(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].wipe_out */
extern void T1722f12(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].move_all_cursors_after */
extern void T1722f15(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.resolve_default */
extern T0* T1375f3(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.on_end_tag */
extern void T560f34p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_end_tag */
extern void T195f16(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_end_tag */
extern void T195f16p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.on_end_tag */
extern void T182f21(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_ELEMENT.parent_element */
extern T0* T168f10(GE_context* ac, T0* C);
extern T1 T168f10ot1(T0* a1, T0** a2);
/* [detachable] XM_ELEMENT.is_root_node */
extern T1 T168f11(GE_context* ac, T0* C);
/* [detachable] XM_DOCUMENT.is_root_node */
extern T1 T192f8(GE_context* ac, T0* C);
/* [detachable] XM_DOCUMENT.parent */
extern T0* T192f9(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.resolve */
extern T0* T1375f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].back */
extern void T1723f11(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].cursor_back */
extern void T1722f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].add_traversing_cursor */
extern void T1722f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].remove_traversing_cursor */
extern void T1722f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].set */
extern void T1723f14(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].go_before */
extern void T1723f10(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].cursor_go_before */
extern void T1722f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].set_before */
extern void T1723f15(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].cursor_off */
extern T1 T1722f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].off */
extern T1 T1723f7(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].item */
extern T0* T1723f6(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].finish */
extern void T1723f9(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].cursor_finish */
extern void T1722f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].new_cursor */
extern T0* T1722f6(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].make */
extern T0* T1723c8(GE_context* ac, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.default_namespace */
extern T0* T1375f6(GE_context* ac, T0* C);
/* [detachable] XM_CALLBACKS_NULL.on_end_tag */
extern void T521f8(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_EIFFEL_PARSER_NAME.local_part */
extern T0* T530f11(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] STRING_8].forth */
extern void T1373f10(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].cursor_forth */
extern void T548f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].add_traversing_cursor */
extern void T548f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] STRING_8].set_next_cursor */
extern void T1373f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].remove_traversing_cursor */
extern void T548f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] STRING_8].set */
extern void T1373f11(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] STRING_8].item */
extern T0* T1373f6(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] STRING_8].start */
extern void T1373f9(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].cursor_start */
extern void T548f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].cursor_off */
extern T1 T548f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_BILINKED_LIST_CURSOR [[attached] STRING_8].off */
extern T1 T1373f7(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER_NAME.string_ */
extern T0* T530f10(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER_NAME.last */
extern T0* T530f9(GE_context* ac, T0* C);
/* [detachable] DS_BILINKED_LIST [[attached] STRING_8].last */
extern T0* T548f7(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER_NAME.ns_prefix */
extern T0* T530f8(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER_NAME.is_namespace_name */
extern T1 T530f7(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER_NAME.make_no_namespaces */
extern void T530f18(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER_NAME.make_no_namespaces */
extern T0* T530c18(GE_context* ac);
/* [detachable] XM_EIFFEL_PARSER_NAME.new_string_bilinked_list */
extern T0* T530f15(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER_NAME.string_equality_tester */
extern T0* T530f14(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.on_start_tag_finish */
extern void T178f246(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.on_start_tag_finish */
extern void T560f33(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.on_start_tag_finish */
extern void T560f33p1(GE_context* ac, T0* C);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_start_tag_finish */
extern void T195f15(GE_context* ac, T0* C);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_start_tag_finish */
extern void T195f15p1(GE_context* ac, T0* C);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.on_start_tag_finish */
extern void T182f20(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.string_ */
extern T0* T560f14(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.on_delayed_attributes */
extern void T560f40(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.attributes_remove */
extern void T560f44(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_QUEUE [[attached] STRING_8].remove */
extern void T1379f7(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_QUEUE [[attached] STRING_8].wipe_out */
extern void T1379f8(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_QUEUE [detachable STRING_8].remove */
extern void T1378f8(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_QUEUE [detachable STRING_8].wipe_out */
extern void T1378f9(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.unprefixed_attribute_namespace */
extern T0* T560f19(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.default_namespace */
extern T0* T560f25(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.xmlns_namespace */
extern T0* T560f18(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.xml_prefix_namespace */
extern T0* T560f17(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.is_xml */
extern T1 T560f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER.xml_prefix */
extern T0* T560f24(GE_context* ac, T0* C);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_attribute */
extern void T195f18(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_attribute */
extern void T195f18p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.on_attribute */
extern void T182f23(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.handle_position */
extern void T182f24(GE_context* ac, T0* C, T0* a1);
/* detachable XM_POSITION_TABLE.put */
extern void T169f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].put_last */
extern void T899f9(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].put_right */
extern void T1806f4(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].make */
extern T0* T1806c3(GE_context* ac, T0* a1);
/* [detachable] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE].make */
extern T0* T900c3(GE_context* ac, T0* a1, T0* a2);
/* detachable XM_ATTRIBUTE.make_last */
extern T0* T196c5(GE_context* ac, T0* a1, T0* a2, T0* a3, T0* a4);
/* [detachable] XM_ELEMENT.force_last */
extern void T168f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].force_last */
extern void T894f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].force_last */
extern void T894f13p1(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] XM_ELEMENT_NODE].put_right */
extern void T1750f4(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] XM_ELEMENT_NODE].make */
extern T0* T1750c3(GE_context* ac, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].before_addition */
extern void T894f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_ELEMENT.before_addition */
extern void T168f22(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CHARACTER_DATA.node_set_parent */
extern void T1471f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_PROCESSING_INSTRUCTION.node_set_parent */
extern void T1470f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_COMMENT.node_set_parent */
extern void T1469f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_ELEMENT.node_set_parent */
extern void T168f17(GE_context* ac, T0* C, T0* a1);
/* detachable XM_ATTRIBUTE.node_set_parent */
extern void T196f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DOCUMENT.equality_delete */
extern void T192f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].forth */
extern void T905f13(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].cursor_forth */
extern void T579f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].add_traversing_cursor */
extern void T579f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].set_next_cursor */
extern void T905f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].remove_traversing_cursor */
extern void T579f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].set */
extern void T905f15(GE_context* ac, T0* C, T0* a1, T1 a2, T1 a3);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].remove */
extern void T905f12(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].remove_at_cursor */
extern void T579f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].move_all_cursors */
extern void T579f22(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].set_current_cell */
extern void T905f17(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] XM_DOCUMENT_NODE].put */
extern void T1393f6(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] XM_DOCUMENT_NODE].put_right */
extern void T1393f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].set_last_cell */
extern void T579f21(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] XM_DOCUMENT_NODE].forget_right */
extern void T1393f5(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].remove_last */
extern void T579f20(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].move_last_cursors_after */
extern void T579f26(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].set_after */
extern void T905f18(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].wipe_out */
extern void T579f24(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].move_all_cursors_after */
extern void T579f27(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].is_last */
extern T1 T905f7(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].cursor_is_last */
extern T1 T579f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].remove_first */
extern void T579f19(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].set_first_cell */
extern void T579f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].is_first */
extern T1 T905f6(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].cursor_is_first */
extern T1 T579f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].item */
extern T0* T905f8(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].start */
extern void T905f11(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].cursor_start */
extern void T579f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].cursor_off */
extern T1 T579f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].off */
extern T1 T905f9(GE_context* ac, T0* C);
/* [detachable] XM_DOCUMENT.new_cursor */
extern T0* T192f7(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].new_cursor */
extern T0* T579f5(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].make */
extern T0* T905c10(GE_context* ac, T0* a1);
/* [detachable] XM_ELEMENT.equality_delete */
extern void T168f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].remove */
extern void T447f14(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].remove_at_cursor */
extern void T894f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].move_all_cursors */
extern void T894f24(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].set_current_cell */
extern void T447f18(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] XM_ELEMENT_NODE].put */
extern void T1750f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].set_last_cell */
extern void T894f23(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] XM_ELEMENT_NODE].forget_right */
extern void T1750f5(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].remove_last */
extern void T894f22(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].move_last_cursors_after */
extern void T894f27(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].wipe_out */
extern void T894f25(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].move_all_cursors_after */
extern void T894f28(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].is_last */
extern T1 T447f7(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].cursor_is_last */
extern T1 T894f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].remove_first */
extern void T894f21(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].set_first_cell */
extern void T894f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_ELEMENT_NODE].is_first */
extern T1 T447f6(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].cursor_is_first */
extern T1 T894f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CHARACTER_DATA.parent */
extern T0* T1471f2(GE_context* ac, T0* C);
/* [detachable] XM_PROCESSING_INSTRUCTION.parent */
extern T0* T1470f3(GE_context* ac, T0* C);
/* [detachable] XM_COMMENT.parent */
extern T0* T1469f2(GE_context* ac, T0* C);
/* [detachable] XM_ELEMENT.parent */
extern T0* T168f3(GE_context* ac, T0* C);
/* detachable XM_ATTRIBUTE.parent */
extern T0* T196f4(GE_context* ac, T0* C);
/* [detachable] XM_ELEMENT.last */
extern T0* T168f13(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].last */
extern T0* T894f8(GE_context* ac, T0* C);
/* [detachable] XM_ELEMENT.is_empty */
extern T1 T168f12(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].is_empty */
extern T1 T894f7(GE_context* ac, T0* C);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.new_namespace */
extern T0* T182f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].force_last */
extern void T563f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].slots_put */
extern void T563f44(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].clashes_put */
extern void T563f43(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].slots_item */
extern T6 T563f22(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].hash_position */
extern T6 T563f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE.hash_code */
extern T6 T446f6(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].resize */
extern void T563f42(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].clashes_resize */
extern void T563f48(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].special_integer_ */
extern T0* T563f29(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].key_storage_resize */
extern void T563f47(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].item_storage_resize */
extern void T563f46(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_NAMESPACE].aliased_resized_area */
extern T0* T1382f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] XM_NAMESPACE].aliased_resized_area */
extern T0* T1383f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].key_storage_item */
extern T0* T563f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].item_storage_item */
extern T0* T563f18(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].clashes_item */
extern T6 T563f26(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].slots_resize */
extern void T563f45(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].new_modulus */
extern T6 T563f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].new_capacity */
extern T6 T563f19(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].item_storage_put */
extern void T563f41(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_NAMESPACE].force */
extern void T1382f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_NAMESPACE].fill_with */
extern void T1383f7(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_NAMESPACE].extend */
extern void T1383f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].search_position */
extern void T563f40(GE_context* ac, T0* C, T0* a1);
/* detachable KL_EQUALITY_TESTER [[attached] XM_NAMESPACE].test */
extern T1 T1380f1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_NAMESPACE.is_equal */
extern T1 T446f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE.string_ */
extern T0* T446f5(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].key_equality_tester */
extern T0* T563f25(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].unset_found_item */
extern void T563f37(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE.same_prefix */
extern T1 T446f3(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].item */
extern T0* T563f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].has */
extern T1 T563f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE.make */
extern void T446f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_NAMESPACE.make */
extern T0* T446c8(GE_context* ac, T0* a1, T0* a2);
/* [detachable] DS_LINKED_QUEUE [[attached] STRING_8].item */
extern T0* T1379f4(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_QUEUE [detachable STRING_8].item */
extern T0* T1378f5(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.attributes_is_empty */
extern T1 T560f12(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_QUEUE [detachable STRING_8].is_empty */
extern T1 T1378f4(GE_context* ac, T0* C);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_start_tag */
extern void T195f14(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_start_tag */
extern void T195f14p1(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.on_start_tag */
extern void T182f19(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_ELEMENT.make_root */
extern T0* T168c19(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_DOCUMENT.set_root_element */
extern void T192f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DOCUMENT.force_last */
extern void T192f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].force_last */
extern void T579f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].force_last */
extern void T579f12p1(GE_context* ac, T0* C, T0* a1);
/* detachable DS_LINKABLE [[attached] XM_DOCUMENT_NODE].make */
extern T0* T1393c3(GE_context* ac, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].before_addition */
extern void T579f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DOCUMENT.before_addition */
extern void T192f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DOCUMENT.last */
extern T0* T192f6(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].last */
extern T0* T579f7(GE_context* ac, T0* C);
/* [detachable] XM_DOCUMENT.is_empty */
extern T1 T192f5(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].is_empty */
extern T1 T579f6(GE_context* ac, T0* C);
/* [detachable] XM_DOCUMENT.remove_previous_root_element */
extern void T192f16(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST_CURSOR [[attached] XM_DOCUMENT_NODE].go_after */
extern void T905f14(GE_context* ac, T0* C);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].cursor_go_after */
extern void T579f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].make */
extern T0* T894c11(GE_context* ac);
/* [detachable] XM_LINKED_LIST [[attached] XM_ELEMENT_NODE].set_internal_cursor */
extern void T894f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_ELEMENT.make_last */
extern T0* T168c18(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.has */
extern T1 T1375f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CALLBACKS_NULL.on_start_tag_finish */
extern void T521f7(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER_NAME.is_equal */
extern T1 T530f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER_NAME.same_string */
extern T1 T530f13(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_EIFFEL_PARSER.when_external_dtd */
extern void T178f245(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.set_input_from_resolver */
extern void T556f283(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NULL_EXTERNAL_RESOLVER.last_stream */
extern T0* T524f2(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.set_input_from_resolver */
extern void T559f287(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.set_input_from_resolver */
extern void T557f211(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.set_input_from_resolver */
extern void T516f209(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.make_scanner */
extern T0* T557c205(GE_context* ac);
/* [detachable] XM_EIFFEL_SCANNER_DTD.make_scanner */
extern void T557f205p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.reset */
extern void T557f214(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.reset */
extern void T557f214p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.reset */
extern void T557f214p0(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.make_with_buffer */
extern void T557f213(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_load_input_buffer */
extern void T557f225(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_set_content */
extern void T557f226(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_initialize */
extern void T557f224(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_initialize */
extern void T557f224p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_build_tables */
extern void T557f229(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_accept_template */
extern T0* T557f199(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_fixed_array */
extern T0* T557f202(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_accept_template_4 */
extern void T557f283(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_array_subcopy */
extern void T557f284(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] XM_EIFFEL_SCANNER_DTD.integer_array_ */
extern T0* T557f204(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_accept_template_3 */
extern void T557f282(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_accept_template_2 */
extern void T557f281(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_accept_template_1 */
extern void T557f280(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_meta_template */
extern T0* T557f198(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_ec_template */
extern T0* T557f197(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_ec_template_2 */
extern void T557f279(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_ec_template_1 */
extern void T557f278(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_def_template */
extern T0* T557f196(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_def_template_5 */
extern void T557f277(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_def_template_4 */
extern void T557f276(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_def_template_3 */
extern void T557f275(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_def_template_2 */
extern void T557f274(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_def_template_1 */
extern void T557f273(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_base_template */
extern T0* T557f195(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_base_template_5 */
extern void T557f272(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_base_template_4 */
extern void T557f271(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_base_template_3 */
extern void T557f270(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_base_template_2 */
extern void T557f269(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_base_template_1 */
extern void T557f268(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template */
extern T0* T557f194(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_18 */
extern void T557f267(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_17 */
extern void T557f266(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_16 */
extern void T557f265(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_15 */
extern void T557f264(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_14 */
extern void T557f263(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_13 */
extern void T557f262(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_12 */
extern void T557f261(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_11 */
extern void T557f260(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_10 */
extern void T557f259(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_9 */
extern void T557f258(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_8 */
extern void T557f257(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_7 */
extern void T557f256(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_6 */
extern void T557f255(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_5 */
extern void T557f254(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_4 */
extern void T557f253(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_3 */
extern void T557f252(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_2 */
extern void T557f251(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_chk_template_1 */
extern void T557f250(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template */
extern T0* T557f193(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_20 */
extern void T557f249(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_19 */
extern void T557f248(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_18 */
extern void T557f247(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_17 */
extern void T557f246(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_16 */
extern void T557f245(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_15 */
extern void T557f244(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_14 */
extern void T557f243(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_13 */
extern void T557f242(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_12 */
extern void T557f241(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_11 */
extern void T557f240(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_10 */
extern void T557f239(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_9 */
extern void T557f238(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_8 */
extern void T557f237(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_7 */
extern void T557f236(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_6 */
extern void T557f235(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_5 */
extern void T557f234(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_4 */
extern void T557f233(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_3 */
extern void T557f232(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_2 */
extern void T557f231(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_nxt_template_1 */
extern void T557f230(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.empty_buffer */
extern T0* T557f52(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_STACK [[attached] XM_EIFFEL_SCANNER].force */
extern void T520f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.null_resolver */
extern T0* T178f112(GE_context* ac, T0* C);
/* [detachable] XM_NULL_EXTERNAL_RESOLVER.last_error */
extern T0* T524f1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.resolve_external_id */
extern void T178f261(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_NULL_EXTERNAL_RESOLVER.resolve */
extern void T524f5(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NULL_EXTERNAL_RESOLVER.resolve_public */
extern void T524f4(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_EXTERNAL_ID].force */
extern void T535f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_DTD_EXTERNAL_ID].fill_with */
extern void T537f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_DTD_EXTERNAL_ID].extend */
extern void T537f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.on_dtd_end */
extern void T178f244(GE_context* ac, T0* C);
/* [detachable] XM_DTD_CALLBACKS_NULL.on_dtd_end */
extern void T526f5(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.on_doctype */
extern void T178f243(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] XM_DTD_CALLBACKS_NULL.on_doctype */
extern void T526f4(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] XM_EIFFEL_DECLARATION.set_encoding */
extern void T552f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_DECLARATION.set_stand_alone */
extern void T552f8(GE_context* ac, T0* C, T1 a1);
/* [detachable] XM_EIFFEL_DECLARATION.set_version */
extern void T552f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_DECLARATION].force */
extern void T551f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_DECLARATION].fill_with */
extern void T553f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_DECLARATION].extend */
extern void T553f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_DECLARATION].aliased_resized_area */
extern T0* T551f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_DECLARATION].aliased_resized_area */
extern T0* T553f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_DECLARATION.make */
extern T0* T552c7(GE_context* ac);
/* [detachable] XM_EIFFEL_DECLARATION.process */
extern void T552f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.on_xml_declaration */
extern void T178f287(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] XM_NAMESPACE_RESOLVER.on_xml_declaration */
extern void T560f45(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_xml_declaration */
extern void T195f19(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.on_xml_declaration */
extern void T182f25(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] XM_CALLBACKS_NULL.on_xml_declaration */
extern void T521f12(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* [detachable] XM_EIFFEL_PARSER.apply_encoding */
extern void T178f242(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.set_encoding */
extern void T556f282(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_INPUT_STREAM.set_encoding */
extern void T901f31(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.set_encoding */
extern void T559f286(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.set_encoding */
extern void T557f210(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.set_encoding */
extern void T516f208(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.is_applicable_encoding */
extern T1 T556f62(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_INPUT_STREAM.is_applicable_encoding */
extern T1 T901f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_INPUT_STREAM.is_valid_encoding */
extern T1 T901f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.is_applicable_encoding */
extern T1 T559f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.is_applicable_encoding */
extern T1 T557f51(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.is_applicable_encoding */
extern T1 T516f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.on_content */
extern void T178f241(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER.on_content */
extern void T560f32(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_content */
extern void T195f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_content */
extern void T195f13p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.on_content */
extern void T182f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CHARACTER_DATA.make_last */
extern T0* T1471c3(GE_context* ac, T0* a1, T0* a2);
/* [detachable] XM_CALLBACKS_NULL.on_content */
extern void T521f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.on_dtd_processing_instruction */
extern void T178f240(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_DTD_CALLBACKS_NULL.on_dtd_processing_instruction */
extern void T526f3(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_EIFFEL_PARSER.on_processing_instruction */
extern void T178f239(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_NAMESPACE_RESOLVER.on_processing_instruction */
extern void T560f31(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_processing_instruction */
extern void T195f12(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_processing_instruction */
extern void T195f12p1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.on_processing_instruction */
extern void T182f17(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_PROCESSING_INSTRUCTION.make_last_in_document */
extern T0* T1470c5(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_PROCESSING_INSTRUCTION.make_last */
extern T0* T1470c4(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_CALLBACKS_NULL.on_processing_instruction */
extern void T521f5(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_EIFFEL_PARSER.on_dtd_comment */
extern void T178f238(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DTD_CALLBACKS_NULL.on_dtd_comment */
extern void T526f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.on_comment */
extern void T178f237(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER.on_comment */
extern void T560f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_comment */
extern void T195f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_comment */
extern void T195f11p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.on_comment */
extern void T182f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_COMMENT.make_last_in_document */
extern T0* T1469c4(GE_context* ac, T0* a1, T0* a2);
/* [detachable] XM_COMMENT.make_last */
extern T0* T1469c3(GE_context* ac, T0* a1, T0* a2);
/* [detachable] XM_CALLBACKS_NULL.on_comment */
extern void T521f4(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.entity_referenced_in_entity_value */
extern T0* T178f138(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.defined_entity_referenced */
extern T0* T178f191(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.external_entity_to_string */
extern T0* T178f198(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.remove */
extern void T17f80(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_NULL_EXTERNAL_RESOLVER.resolve_finish */
extern void T524f6(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.read_string */
extern void T901f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_INTEGER_ROUTINES.platform */
extern T0* T772f3(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.integer_ */
extern T0* T178f203(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_INPUT_STREAM.make_from_stream */
extern T0* T901c29(GE_context* ac, T0* a1);
/* [detachable] DS_LINKED_QUEUE [[attached] CHARACTER_8].make */
extern T0* T1626c5(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].item */
extern T0* T522f21(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].item_storage_item */
extern T0* T522f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PARSER.string_ */
extern T0* T178f137(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.force_error */
extern void T178f236(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.report_error */
extern void T178f234(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.on_error */
extern void T178f260(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.last_error_extended_description */
extern T0* T178f173(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.safe_error_component */
extern T0* T178f193(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.is_safe_error_character */
extern T1 T178f201(GE_context* ac, T0* C, T2 a1);
/* [detachable] XM_EIFFEL_PARSER.setup_error_state */
extern void T178f259(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.close_input */
extern void T556f280(GE_context* ac, T0* C);
/* [detachable] KL_STRING_INPUT_STREAM.close */
extern void T1477f11(GE_context* ac, T0* C);
/* [detachable] KL_STRING_INPUT_STREAM.is_closable */
extern T1 T1477f7(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.close_input */
extern void T559f284(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.close_input */
extern void T557f208(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.close_input */
extern void T516f206(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.onstring_utf8 */
extern T0* T178f126(GE_context* ac, T0* C, T0* a1);
/* [detachable] STRING_8.string */
extern T0* T17f43(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.maximum_item_code */
extern T6 T178f186(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.is_string_mode_latin1 */
extern T1 T178f185(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.new_unicode_string_from_utf8 */
extern T0* T178f181(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_STRING.make_from_utf8 */
extern T0* T775c71(GE_context* ac, T0* a1);
/* [detachable] UC_UTF8_STRING.append_utf8 */
extern void T775f92(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_ROUTINES.valid_utf8 */
extern T1 T434f23(GE_context* ac, T0* C, T0* a1);
/* [detachable] UC_UTF8_ROUTINES.valid_utf8_substring */
extern T1 T434f24(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] UC_UTF8_ROUTINES.is_encoded_next_byte */
extern T1 T434f30(GE_context* ac, T0* C, T2 a1);
/* [detachable] UC_UTF8_ROUTINES.is_encoded_second_byte */
extern T1 T434f26(GE_context* ac, T0* C, T2 a1, T2 a2);
/* [detachable] UC_UTF8_ROUTINES.is_encoded_first_byte */
extern T1 T434f25(GE_context* ac, T0* C, T2 a1);
/* [detachable] XM_EIFFEL_PARSER.utf8 */
extern T0* T178f184(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.is_string_mode_ascii */
extern T1 T178f182(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.onstring_ascii */
extern T0* T178f124(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.is_string_mode_unicode */
extern T1 T178f180(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.shared_empty_string */
extern T0* T178f135(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.shared_empty_string_string */
extern T0* T178f190(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.shared_empty_string_uc */
extern T0* T178f189(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.new_unicode_string_empty */
extern T0* T178f197(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_STRING.make_empty */
extern T0* T775c81(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_PARSER_NAME].force */
extern void T529f4(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_PARSER_NAME].fill_with */
extern void T531f8(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_PARSER_NAME].extend */
extern void T531f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_PARSER_NAME].aliased_resized_area */
extern T0* T529f2(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_PARSER_NAME].aliased_resized_area */
extern T0* T531f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PARSER.namespace_force_last */
extern void T178f235(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_EIFFEL_PARSER_NAME.force_last */
extern void T530f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER_NAME.can_force_last */
extern T1 T530f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.new_namespace_name */
extern T0* T178f134(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER_NAME.make_namespaces */
extern T0* T530c20(GE_context* ac);
/* [detachable] XM_EIFFEL_PARSER.yy_push_last_value */
extern void T178f225(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PARSER.accept */
extern void T178f224(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yy_do_error_action */
extern void T178f222(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PARSER.report_eof_expected_error */
extern void T178f233(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.read_token */
extern void T178f221(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.process_attribute_entity */
extern void T178f232(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.process_entity_scanner */
extern void T178f258(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.apply_input_buffer */
extern void T556f276(GE_context* ac, T0* C);
/* [detachable] UC_UTF8_ROUTINES.to_utf8 */
extern T0* T434f49(GE_context* ac, T0* C, T0* a1);
extern T1 T434f49ot1(T0* a1, T0** a2);
/* [detachable] XM_EIFFEL_ENTITY_DEF.utf8 */
extern T0* T556f69(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.fatal_error */
extern void T556f284(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.apply_input_buffer */
extern void T559f280(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.utf8 */
extern T0* T559f71(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.fatal_error */
extern void T559f288(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.is_literal */
extern T1 T556f64(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.is_external */
extern T1 T556f67(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.is_literal */
extern T1 T559f66(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.is_external */
extern T1 T559f69(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.process_entity */
extern void T178f231(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.process_pe_entity */
extern void T178f230(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.end_of_file */
extern T1 T556f61(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.end_of_file */
extern T1 T559f63(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.end_of_file */
extern T1 T557f69(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.end_of_file */
extern T1 T516f72(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.last_value */
extern T0* T556f10(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.last_value */
extern T0* T559f10(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.last_value */
extern T0* T557f2(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.last_value */
extern T0* T516f2(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.last_token */
extern T6 T556f45(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.last_token */
extern T6 T559f47(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.last_token */
extern T6 T557f1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.last_token */
extern T6 T516f1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.read_token */
extern void T556f279(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.read_token */
extern void T556f279p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_execute_action */
extern void T556f289(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_CHARACTER_ENTITY.from_hexadecimal */
extern void T1356f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_STRING_ROUTINES.hexadecimal_to_integer */
extern T6 T82f20(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_CHARACTER_ENTITY.string_ */
extern T0* T1356f7(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_CHARACTER_ENTITY.to_utf8 */
extern T0* T1356f3(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_CHARACTER_ENTITY.utf8 */
extern T0* T1356f6(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_CHARACTER_ENTITY.is_ascii */
extern T1 T1356f2(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_CHARACTER_ENTITY.is_valid */
extern T1 T1356f5(GE_context* ac, T0* C);
/* [detachable] UC_UNICODE_ROUTINES.valid_non_surrogate_code */
extern T1 T426f30(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_CHARACTER_ENTITY.unicode */
extern T0* T1356f4(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_CHARACTER_ENTITY.from_decimal */
extern void T1356f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.has_normalized_newline */
extern T1 T556f204(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.text_substring */
extern T0* T556f184(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] XM_EIFFEL_ENTITY_DEF.text_count */
extern T6 T556f183(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.system_literal_text */
extern T0* T556f165(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.text_item */
extern T2 T556f211(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.normalized_newline */
extern T0* T556f124(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.normalized_newline */
extern T0* T556f124p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_set_line_column */
extern void T556f294(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.text */
extern T0* T556f108(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.pop_start_condition */
extern void T556f293(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.set_last_token */
extern void T556f292(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_execute_eof_action */
extern void T556f288(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.terminate */
extern void T556f291(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.wrap */
extern T1 T556f102(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_refill_input_buffer */
extern void T556f287(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_null_trans_state */
extern T6 T556f101(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_previous_state */
extern T6 T556f100(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.report_invalid_unicode_character_error */
extern void T556f286(GE_context* ac, T0* C, T10 a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.yy_at_beginning_of_line */
extern T6 T556f89(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.read_token */
extern void T559f283(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.read_token */
extern void T559f283p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.read_token */
extern void T559f283p0(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_execute_action */
extern void T559f294(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.has_normalized_newline */
extern T1 T559f207(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.text_substring */
extern T0* T559f187(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.text_count */
extern T6 T559f186(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.system_literal_text */
extern T0* T559f168(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.text_item */
extern T2 T559f213(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.normalized_newline */
extern T0* T559f127(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.normalized_newline */
extern T0* T559f127p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_set_line_column */
extern void T559f298(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.text */
extern T0* T559f111(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.pop_start_condition */
extern void T559f297(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.set_last_token */
extern void T559f296(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_execute_eof_action */
extern void T559f293(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.wrap */
extern T1 T559f105(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_refill_input_buffer */
extern void T559f292(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_null_trans_state */
extern T6 T559f104(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_previous_state */
extern T6 T559f103(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.report_invalid_unicode_character_error */
extern void T559f291(GE_context* ac, T0* C, T10 a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.yy_at_beginning_of_line */
extern T6 T559f92(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.terminate */
extern void T559f290(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.read_token */
extern void T557f207(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.read_token */
extern void T557f207p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_execute_action */
extern void T557f222(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.has_normalized_newline */
extern T1 T557f187(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.text_substring */
extern T0* T557f167(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] XM_EIFFEL_SCANNER_DTD.text_count */
extern T6 T557f166(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.system_literal_text */
extern T0* T557f148(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.text_item */
extern T2 T557f201(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.normalized_newline */
extern T0* T557f110(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_set_line_column */
extern void T557f228(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.text */
extern T0* T557f84(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.pop_start_condition */
extern void T557f227(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_execute_eof_action */
extern void T557f221(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.wrap */
extern T1 T557f72(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_refill_input_buffer */
extern void T557f220(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_null_trans_state */
extern T6 T557f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_previous_state */
extern T6 T557f70(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.fatal_error */
extern void T557f219(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.report_invalid_unicode_character_error */
extern void T557f218(GE_context* ac, T0* C, T10 a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.yy_at_beginning_of_line */
extern T6 T557f62(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.terminate */
extern void T557f217(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER_DTD.set_last_token */
extern void T557f216(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER.read_token */
extern void T516f205(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_execute_action */
extern void T516f217(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER.has_normalized_newline */
extern T1 T516f185(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.text_substring */
extern T0* T516f165(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] XM_EIFFEL_SCANNER.text_count */
extern T6 T516f164(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.system_literal_text */
extern T0* T516f146(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.text_item */
extern T2 T516f199(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER.normalized_newline */
extern T0* T516f105(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_set_line_column */
extern void T516f225(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.text */
extern T0* T516f79(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.pop_start_condition */
extern void T516f224(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.set_last_token */
extern void T516f223(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_execute_eof_action */
extern void T516f216(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER.terminate */
extern void T516f222(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.wrap */
extern T1 T516f67(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_refill_input_buffer */
extern void T516f215(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_set_content */
extern void T516f221(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_null_trans_state */
extern T6 T516f66(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_previous_state */
extern T6 T516f65(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.fatal_error */
extern void T516f214(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.report_invalid_unicode_character_error */
extern void T516f213(GE_context* ac, T0* C, T10 a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_at_beginning_of_line */
extern T6 T516f50(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.special_integer_ */
extern T0* T178f97(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yy_init_value_stacks */
extern void T178f220(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yy_clear_all */
extern void T178f229(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.clear_all */
extern void T178f257(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.clear_stacks */
extern void T178f263(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yy_clear_value_stacks */
extern void T178f264(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_DECLARATION].keep_head */
extern void T553f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_DECLARATION].set_count */
extern void T553f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] STRING_8]].keep_head */
extern void T549f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] STRING_8]].set_count */
extern void T549f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_DTD_ATTRIBUTE_CONTENT].keep_head */
extern void T546f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_DTD_ATTRIBUTE_CONTENT].set_count */
extern void T546f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]].keep_head */
extern void T543f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]].set_count */
extern void T543f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_DTD_ELEMENT_CONTENT].keep_head */
extern void T540f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_DTD_ELEMENT_CONTENT].set_count */
extern void T540f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_DTD_EXTERNAL_ID].keep_head */
extern void T537f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_DTD_EXTERNAL_ID].set_count */
extern void T537f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]].keep_head */
extern void T534f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]].set_count */
extern void T534f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_PARSER_NAME].keep_head */
extern void T531f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_PARSER_NAME].set_count */
extern void T531f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_EIFFEL_PARSER.abort */
extern void T178f223(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.on_start */
extern void T178f217(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.on_start */
extern void T560f28(GE_context* ac, T0* C);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_start */
extern void T195f9(GE_context* ac, T0* C);
/* [detachable] XM_STOP_ON_ERROR_FILTER.on_start */
extern void T195f9p1(GE_context* ac, T0* C);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.on_start */
extern void T182f14(GE_context* ac, T0* C);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.initialize */
extern void T182f12(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].make_equal */
extern T0* T563c31(GE_context* ac, T6 a1);
/* detachable KL_EQUALITY_TESTER [[attached] XM_NAMESPACE].default_create */
extern T0* T1380c2(GE_context* ac);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].make */
extern void T563f32(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].set_internal_cursor */
extern void T563f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].new_cursor */
extern T0* T563f24(GE_context* ac, T0* C);
/* [detachable] DS_HASH_SET_CURSOR [[attached] XM_NAMESPACE].make */
extern T0* T1381c4(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].make_slots */
extern void T563f36(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].make_clashes */
extern void T563f35(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].make_key_storage */
extern void T563f34(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_SET [[attached] XM_NAMESPACE].make_item_storage */
extern void T563f33(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_NAMESPACE].make */
extern T0* T1382f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_NAMESPACE].make_empty */
extern T0* T1383c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_NAMESPACE].default_create */
extern T0* T1382c3(GE_context* ac);
/* [detachable] XM_DOCUMENT.make */
extern T0* T192c10(GE_context* ac);
/* [detachable] XM_DOCUMENT.make_with_root_named */
extern void T192f11(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_ELEMENT.make */
extern T0* T168c16(GE_context* ac, T0* a1, T0* a2, T0* a3);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].make */
extern T0* T579c11(GE_context* ac);
/* [detachable] XM_LINKED_LIST [[attached] XM_DOCUMENT_NODE].set_internal_cursor */
extern void T579f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_DOCUMENT.default_ns */
extern T0* T192f3(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE.make_default */
extern T0* T446c7(GE_context* ac);
/* [detachable] XM_NAMESPACE_RESOLVER.initialize */
extern void T560f39(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.attributes_make */
extern void T560f43(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.new_string_queue */
extern T0* T560f21(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_QUEUE [[attached] STRING_8].make */
extern T0* T1379c5(GE_context* ac);
/* [detachable] XM_NAMESPACE_RESOLVER.new_detachable_string_queue */
extern T0* T560f20(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_QUEUE [detachable STRING_8].make */
extern T0* T1378c6(GE_context* ac);
/* [detachable] XM_NAMESPACE_RESOLVER_CONTEXT.make */
extern T0* T1375c10(GE_context* ac);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].make */
extern T0* T1722c8(GE_context* ac);
/* [detachable] DS_BILINKED_LIST [[attached] DS_HASH_TABLE [[attached] STRING_8, [attached] STRING_8]].set_internal_cursor */
extern void T1722f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CALLBACKS_NULL.on_start */
extern void T521f2(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.reset_error_state */
extern void T178f216(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.set_input_stream */
extern void T556f278(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.set_input_buffer */
extern void T556f285(GE_context* ac, T0* C, T0* a1);
/* [detachable] YY_FILE_BUFFER.name */
extern T0* T930f14(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_ENTITY_DEF.set_input_buffer */
extern void T556f285p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_ENTITY_DEF.new_file_buffer */
extern T0* T556f72(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.set_input_stream */
extern void T559f282(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.set_input_buffer */
extern void T559f289(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.set_input_buffer */
extern void T559f289p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PE_ENTITY_DEF.new_file_buffer */
extern T0* T559f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.set_input_stream */
extern void T557f206(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.set_input_buffer */
extern void T557f215(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.set_input_buffer */
extern void T557f215p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER_DTD.new_file_buffer */
extern T0* T557f58(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.set_input_stream */
extern void T516f204(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.set_input_buffer */
extern void T516f212(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.set_input_buffer */
extern void T516f212p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_load_input_buffer */
extern void T516f220(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.new_file_buffer */
extern T0* T516f48(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.reset */
extern void T178f212(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].wipe_out */
extern void T522f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].slots_wipe_out */
extern void T522f47(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].clashes_wipe_out */
extern void T522f46(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].key_storage_wipe_out */
extern void T522f45(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].item_storage_wipe_out */
extern void T522f44(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_ENTITY_DEF].keep_head */
extern void T1361f6(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_ENTITY_DEF].set_count */
extern void T1361f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].move_all_cursors_after */
extern void T522f42(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].set_next_cursor */
extern void T1364f8(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_TABLE_CURSOR [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].set_after */
extern void T1364f7(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.make_scanner */
extern void T178f210(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.make_scanner */
extern T0* T516c203(GE_context* ac);
/* [detachable] XM_EIFFEL_SCANNER.reset */
extern void T516f211(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.reset */
extern void T516f211p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.reset */
extern void T516f211p0(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.make_with_buffer */
extern void T516f210(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_initialize */
extern void T516f219(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_initialize */
extern void T516f219p1(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_build_tables */
extern void T516f227(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_accept_template */
extern T0* T516f197(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_fixed_array */
extern T0* T516f200(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_accept_template_4 */
extern void T516f281(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_array_subcopy */
extern void T516f282(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] XM_EIFFEL_SCANNER.integer_array_ */
extern T0* T516f202(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_accept_template_3 */
extern void T516f280(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_accept_template_2 */
extern void T516f279(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_accept_template_1 */
extern void T516f278(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_meta_template */
extern T0* T516f196(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_ec_template */
extern T0* T516f195(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_ec_template_2 */
extern void T516f277(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_ec_template_1 */
extern void T516f276(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_def_template */
extern T0* T516f194(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_def_template_5 */
extern void T516f275(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_def_template_4 */
extern void T516f274(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_def_template_3 */
extern void T516f273(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_def_template_2 */
extern void T516f272(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_def_template_1 */
extern void T516f271(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_base_template */
extern T0* T516f193(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_base_template_5 */
extern void T516f270(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_base_template_4 */
extern void T516f269(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_base_template_3 */
extern void T516f268(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_base_template_2 */
extern void T516f267(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_base_template_1 */
extern void T516f266(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template */
extern T0* T516f192(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_18 */
extern void T516f265(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_17 */
extern void T516f264(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_16 */
extern void T516f263(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_15 */
extern void T516f262(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_14 */
extern void T516f261(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_13 */
extern void T516f260(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_12 */
extern void T516f259(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_11 */
extern void T516f258(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_10 */
extern void T516f257(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_9 */
extern void T516f256(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_8 */
extern void T516f255(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_7 */
extern void T516f254(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_6 */
extern void T516f253(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_5 */
extern void T516f252(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_4 */
extern void T516f251(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_3 */
extern void T516f250(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_2 */
extern void T516f249(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_chk_template_1 */
extern void T516f248(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template */
extern T0* T516f191(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_20 */
extern void T516f247(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_19 */
extern void T516f246(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_18 */
extern void T516f245(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_17 */
extern void T516f244(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_16 */
extern void T516f243(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_15 */
extern void T516f242(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_14 */
extern void T516f241(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_13 */
extern void T516f240(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_12 */
extern void T516f239(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_11 */
extern void T516f238(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_10 */
extern void T516f237(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_9 */
extern void T516f236(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_8 */
extern void T516f235(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_7 */
extern void T516f234(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_6 */
extern void T516f233(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_5 */
extern void T516f232(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_4 */
extern void T516f231(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_3 */
extern void T516f230(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_2 */
extern void T516f229(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.yy_nxt_template_1 */
extern void T516f228(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_SCANNER.empty_buffer */
extern T0* T516f45(GE_context* ac, T0* C);
/* [detachable] DS_CELL [detachable ET_ECF_SYSTEM].make */
extern T0* T163c2(GE_context* ac, T0* a1);
/* [detachable] GEC.target_name */
extern T0* T25f35(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.parameter */
extern T0* T44f15(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_override_variables */
extern void T55f213(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_override_capabilities */
extern void T55f212(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.set_override_settings */
extern void T55f211(GE_context* ac, T0* C, T0* a1);
/* detachable ET_ECF_SETTINGS.set_primary_value */
extern void T57f8(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable ET_ECF_SETTINGS.make */
extern T0* T57c7(GE_context* ac);
/* Creation of agent #1 in feature detachable ET_ECF_SETTINGS.make */
extern T0* T57f7ac1(GE_context* ac, T0* a1);
/* detachable ET_ECF_SETTINGS.string_ */
extern T0* T57f3(GE_context* ac, T0* C);
/* detachable ET_ECF_SETTINGS.case_insensitive_string_equality_tester */
extern T0* T57f6(GE_context* ac, T0* C);
/* [detachable] ET_ECF_SYSTEM_PARSER.make */
extern T0* T55c210(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_SYSTEM_PARSER.make_with_factory */
extern void T55f215(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.enable_position_table */
extern void T182f10(GE_context* ac, T0* C, T0* a1);
/* detachable XM_POSITION_TABLE.make */
extern T0* T169c4(GE_context* ac);
/* [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].make */
extern T0* T899c7(GE_context* ac);
/* [detachable] DS_LINKED_LIST [[attached] DS_PAIR [[attached] XM_POSITION, [attached] XM_NODE]].set_internal_cursor */
extern void T899f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.set_callbacks */
extern void T178f207(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_TREE_CALLBACKS_PIPE.make */
extern T0* T179c10(GE_context* ac);
/* [detachable] XM_TREE_CALLBACKS_PIPE.callbacks_pipe */
extern T0* T179f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_STOP_ON_ERROR_FILTER.set_next */
extern void T195f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER.set_next */
extern void T560f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.set_next */
extern void T182f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ARRAY [[attached] XM_CALLBACKS_FILTER].item */
extern T0* T561f4(GE_context* ac, T0* C, T6 a1);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.make_null */
extern T0* T182c9(GE_context* ac);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.make_next */
extern void T182f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_CALLBACKS_TO_TREE_FILTER.null_callbacks */
extern T0* T182f7(GE_context* ac, T0* C);
/* [detachable] XM_CALLBACKS_NULL.make */
extern T0* T521c1(GE_context* ac);
/* [detachable] XM_TREE_CALLBACKS_PIPE.new_stop_on_error */
extern T0* T179f7(GE_context* ac, T0* C);
/* [detachable] XM_STOP_ON_ERROR_FILTER.make_null */
extern T0* T195c5(GE_context* ac);
/* [detachable] XM_STOP_ON_ERROR_FILTER.make_next */
extern void T195f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_STOP_ON_ERROR_FILTER.initialize */
extern void T195f7(GE_context* ac, T0* C);
/* [detachable] XM_STOP_ON_ERROR_FILTER.null_callbacks */
extern T0* T195f4(GE_context* ac, T0* C);
/* [detachable] XM_TREE_CALLBACKS_PIPE.new_namespace_resolver */
extern T0* T179f6(GE_context* ac, T0* C);
/* [detachable] XM_NAMESPACE_RESOLVER.make_null */
extern T0* T560c27(GE_context* ac);
/* [detachable] XM_NAMESPACE_RESOLVER.make_next */
extern void T560f38(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_NAMESPACE_RESOLVER.null_callbacks */
extern T0* T560f13(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.set_string_mode_mixed */
extern void T178f206(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.make */
extern T0* T178c205(GE_context* ac);
/* [detachable] XM_EIFFEL_PARSER.new_entities_table */
extern T0* T178f85(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].set_key_equality_tester */
extern void T522f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].make_map_default */
extern T0* T522c36(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].make_map */
extern void T522f41(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].make_with_equality_testers */
extern void T522f53(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].make_sparse_container */
extern void T522f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].set_internal_cursor */
extern void T522f63(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].new_cursor */
extern T0* T522f34(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].make */
extern T0* T1364c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].make_slots */
extern void T522f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].make_clashes */
extern void T522f61(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].make_key_storage */
extern void T522f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].make_item_storage */
extern void T522f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_ENTITY_DEF].make */
extern T0* T1365f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_ENTITY_DEF].make_empty */
extern T0* T1361c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_ENTITY_DEF].default_create */
extern T0* T1365c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] XM_EIFFEL_ENTITY_DEF, [attached] STRING_8].default_capacity */
extern T6 T522f30(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.make_parser */
extern void T178f211(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yy_build_parser_tables */
extern void T178f215(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yycheck_template */
extern T0* T178f121(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yyfixed_array */
extern T0* T178f178(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yycheck_template_4 */
extern void T178f285(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yyarray_subcopy */
extern void T178f286(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] XM_EIFFEL_PARSER.integer_array_ */
extern T0* T178f204(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yycheck_template_3 */
extern void T178f284(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yycheck_template_2 */
extern void T178f283(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yycheck_template_1 */
extern void T178f282(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yytable_template */
extern T0* T178f119(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yytable_template_4 */
extern void T178f281(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yytable_template_3 */
extern void T178f280(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yytable_template_2 */
extern void T178f279(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yytable_template_1 */
extern void T178f278(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yypgoto_template */
extern T0* T178f117(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yypact_template */
extern T0* T178f115(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yypact_template_3 */
extern void T178f277(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yypact_template_2 */
extern void T178f276(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yypact_template_1 */
extern void T178f275(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yydefgoto_template */
extern T0* T178f113(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yydefact_template */
extern T0* T178f105(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yydefact_template_3 */
extern void T178f274(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yydefact_template_2 */
extern void T178f273(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yydefact_template_1 */
extern void T178f272(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yytypes2_template */
extern T0* T178f98(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yytypes1_template */
extern T0* T178f93(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yytypes1_template_3 */
extern void T178f271(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yytypes1_template_2 */
extern void T178f270(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yytypes1_template_1 */
extern void T178f269(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yyr1_template */
extern T0* T178f89(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yyr1_template_2 */
extern void T178f268(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yyr1_template_1 */
extern void T178f267(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yytranslate_template */
extern T0* T178f87(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.yytranslate_template_2 */
extern void T178f266(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yytranslate_template_1 */
extern void T178f265(GE_context* ac, T0* C, T0* a1);
/* [detachable] XM_EIFFEL_PARSER.yy_create_value_stacks */
extern void T178f214(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_DECLARATION].make */
extern T0* T551f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_EIFFEL_DECLARATION].make_empty */
extern T0* T553c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_EIFFEL_DECLARATION].default_create */
extern T0* T551c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] STRING_8]].make */
extern T0* T547f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] STRING_8]].make_empty */
extern T0* T549c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] STRING_8]].default_create */
extern T0* T547c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ATTRIBUTE_CONTENT].make */
extern T0* T544f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_DTD_ATTRIBUTE_CONTENT].make_empty */
extern T0* T546c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ATTRIBUTE_CONTENT].default_create */
extern T0* T544c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]].make */
extern T0* T541f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]].make_empty */
extern T0* T543c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_BILINKED_LIST [[attached] XM_DTD_ATTRIBUTE_CONTENT]].default_create */
extern T0* T541c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ELEMENT_CONTENT].make */
extern T0* T538f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_DTD_ELEMENT_CONTENT].make_empty */
extern T0* T540c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_ELEMENT_CONTENT].default_create */
extern T0* T538c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_EXTERNAL_ID].make */
extern T0* T535f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] XM_DTD_EXTERNAL_ID].make_empty */
extern T0* T537c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] XM_DTD_EXTERNAL_ID].default_create */
extern T0* T535c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]].make */
extern T0* T532f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]].make_empty */
extern T0* T534c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] DS_HASH_SET [[attached] XM_EIFFEL_PARSER_NAME]].default_create */
extern T0* T532c3(GE_context* ac);
/* [detachable] XM_EIFFEL_PARSER.initialize */
extern void T178f209(GE_context* ac, T0* C);
/* [detachable] XM_EIFFEL_PARSER.null_dtd_callbacks */
extern T0* T178f86(GE_context* ac, T0* C);
/* [detachable] XM_DTD_CALLBACKS_NULL.make */
extern T0* T526c1(GE_context* ac);
/* [detachable] XM_EIFFEL_PARSER.null_callbacks */
extern T0* T178f84(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].set_key_equality_tester */
extern void T176f35(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].make_map */
extern T0* T176c34(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].make_with_equality_testers */
extern void T176f39(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].make_sparse_container */
extern void T176f52(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].set_internal_cursor */
extern void T176f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].new_cursor */
extern T0* T176f31(GE_context* ac, T0* C);
/* detachable DS_HASH_TABLE_CURSOR [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].make */
extern T0* T514c6(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].make_slots */
extern void T176f60(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].make_clashes */
extern void T176f59(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].make_key_storage */
extern void T176f58(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_DOTNET_ASSEMBLY, [attached] STRING_8].make_item_storage */
extern void T176f57(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_DOTNET_ASSEMBLY].make */
extern T0* T515f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_DOTNET_ASSEMBLY].make_empty */
extern T0* T511c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_DOTNET_ASSEMBLY].default_create */
extern T0* T515c3(GE_context* ac);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].set_key_equality_tester */
extern void T174f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].make_map */
extern T0* T174c40(GE_context* ac, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].make_with_equality_testers */
extern void T174f47(GE_context* ac, T0* C, T6 a1, T0* a2, T0* a3);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].make_sparse_container */
extern void T174f62(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].set_internal_cursor */
extern void T174f73(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].new_cursor */
extern T0* T174f37(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE_CURSOR [[attached] ET_ECF_LIBRARY, [attached] STRING_8].make */
extern T0* T506c8(GE_context* ac, T0* a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].make_slots */
extern void T174f72(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].make_clashes */
extern void T174f71(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].make_key_storage */
extern void T174f70(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [[attached] ET_ECF_LIBRARY, [attached] STRING_8].make_item_storage */
extern void T174f69(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_LIBRARY].make */
extern T0* T510f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ECF_LIBRARY].make_empty */
extern T0* T507c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ECF_LIBRARY].default_create */
extern T0* T510c3(GE_context* ac);
/* [detachable] ET_ECF_SYSTEM_PARSER.make_with_factory */
extern void T55f215p1(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_ECF_SYSTEM_PARSER.ise_latest */
extern T0* T55f15(GE_context* ac, T0* C);
/* [detachable] ET_ECF_AST_FACTORY.make */
extern T0* T162c53(GE_context* ac);
/* [detachable] ET_ECF_ERROR_HANDLER.make_standard */
extern T0* T56c10(GE_context* ac);
/* [detachable] ET_ECF_ERROR_HANDLER.std */
extern T0* T56f7(GE_context* ac, T0* C);
/* [detachable] ET_ECF_ERROR_HANDLER.make_null */
extern T0* T56c9(GE_context* ac);
/* [detachable] ET_ECF_ERROR_HANDLER.null_output_stream */
extern T0* T56f6(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.is_open_read */
extern T1 T37f14(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.old_is_open_read */
extern T1 T37f27(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.open_read */
extern void T37f62(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.old_open_read */
extern void T37f67(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.open_read */
extern void T37f67p1(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_INPUT_FILE.file_open */
extern T14 T37f25(GE_context* ac, T0* C, T14 a1, T6 a2);
/* [detachable] KL_TEXT_INPUT_FILE.old_is_readable */
extern T1 T37f16(GE_context* ac, T0* C);
/* [detachable] FILE_INFO.is_readable */
extern T1 T141f8(GE_context* ac, T0* C);
/* [detachable] FILE_INFO.file_eaccess */
extern T1 T141f6(GE_context* ac, T0* C, T14 a1, T6 a2);
/* [detachable] KL_TEXT_INPUT_FILE.make */
extern void T37f61(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.make */
extern T0* T37c61(GE_context* ac, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.make */
extern void T37f61p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.old_make */
extern void T37f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_TEXT_INPUT_FILE.make_with_name */
extern void T37f66p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_FLAG.was_found */
extern T1 T39f9(GE_context* ac, T0* C);
/* [detachable] GEC.parse_arguments */
extern void T25f64(GE_context* ac, T0* C, T0* a1);
/* [detachable] GEC.set_new_instance_types */
extern void T25f75(GE_context* ac, T0* C, T0* a1, T0* a2);
/* detachable DS_HASH_SET [[attached] STRING_8].force */
extern void T70f39(GE_context* ac, T0* C, T0* a1);
/* detachable DS_HASH_SET [[attached] STRING_8].make_equal */
extern T0* T70c38(GE_context* ac, T6 a1);
/* [detachable] GEC.set_split_size */
extern void T25f74(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] GEC.set_override_variables */
extern void T25f73(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8].forth */
extern void T151f10(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].cursor_forth */
extern void T150f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8].set_position */
extern void T151f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].add_traversing_cursor */
extern void T150f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8].set_next_cursor */
extern void T151f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].remove_traversing_cursor */
extern void T150f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8].item */
extern T0* T151f5(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].cursor_item */
extern T0* T150f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8].after */
extern T1 T151f7(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].new_iterator */
extern T0* T150f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8].start */
extern void T151f11(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable STRING_8].cursor_start */
extern void T150f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8].set_after */
extern void T151f13(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8].off */
extern T1 T151f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable STRING_8].before */
extern T1 T151f8(GE_context* ac, T0* C);
/* [detachable] KL_EXECUTION_ENVIRONMENT.set_variable_value */
extern void T86f17(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] EXECUTION_ENVIRONMENT.put */
extern void T124f11(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] EXECUTION_ENVIRONMENT.eif_putenv */
extern T6 T124f8(GE_context* ac, T0* C, T14 a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].force */
extern void T480f35(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] SPECIAL [[attached] NATIVE_STRING].force */
extern void T1354f5(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] NATIVE_STRING].extend */
extern void T1354f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].position */
extern T6 T480f22(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] IMMUTABLE_STRING_32].force */
extern void T485f5(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] BOOLEAN].force */
extern void T129f6(GE_context* ac, T0* C, T1 a1, T6 a2);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].deleted_position */
extern T6 T480f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].add_space */
extern void T480f37(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].accommodate */
extern void T480f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].set_indexes_map */
extern void T480f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].set_deleted_marks */
extern void T480f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].set_keys */
extern void T480f40(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].set_content */
extern void T480f39(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].put */
extern void T480f43(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].set_conflict */
extern void T480f45(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].found */
extern T1 T480f31(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].occupied */
extern T1 T480f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].empty_duplicate */
extern T0* T480f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].compare_objects */
extern void T480f44(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].make */
extern T0* T480c34(GE_context* ac, T6 a1);
/* [detachable] SPECIAL [[attached] NATIVE_STRING].make_empty */
extern T0* T1354c4(GE_context* ac, T6 a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].soon_full */
extern T1 T480f20(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].not_found */
extern T1 T480f19(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].internal_search */
extern void T480f36(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].same_keys */
extern T1 T480f26(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] IMMUTABLE_STRING_32.is_equal */
extern T1 T20f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] NATIVE_STRING, [attached] IMMUTABLE_STRING_32].hash_code_of */
extern T6 T480f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] IMMUTABLE_STRING_32.hash_code */
extern T6 T20f11(GE_context* ac, T0* C);
/* [detachable] IMMUTABLE_STRING_32.make_from_string_general */
extern T0* T20c13(GE_context* ac, T0* a1);
extern T1 T20f13ot1(T0* a1, T0** a2);
extern T1 T20f13ot2(T0* a1, T0** a2);
/* [detachable] STRING_32.to_string_32 */
extern T0* T18f11(GE_context* ac, T0* C);
/* [detachable] IMMUTABLE_STRING_32.make_from_string_32 */
extern void T20f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] IMMUTABLE_STRING_32.make_from_string_8 */
extern void T20f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] EXECUTION_ENVIRONMENT.environ */
extern T0* T124f6(GE_context* ac, T0* C);
/* [detachable] GEC.execution_environment */
extern T0* T25f53(GE_context* ac, T0* C);
/* [detachable] GEC.set_override_capabilities */
extern void T25f72(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] GEC.set_override_settings */
extern void T25f71(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] GEC.set_ise_version */
extern void T25f70(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] GEC.report_usage_message */
extern void T25f69(GE_context* ac, T0* C, T0* a1);
/* [detachable] GEC.report_version_number */
extern void T25f68(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.set_parameter_description */
extern void T48f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_BOOLEAN_OPTION.set_description */
extern void T48f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_BOOLEAN_OPTION.make_with_long_form */
extern T0* T48c26(GE_context* ac, T0* a1);
/* [detachable] AP_BOOLEAN_OPTION.set_long_form */
extern void T48f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_BOOLEAN_OPTION.initialize */
extern void T48f29(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.initialize */
extern void T48f29p1(GE_context* ac, T0* C);
/* [detachable] AP_BOOLEAN_OPTION.initialize */
extern void T48f29p0(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.set_parameter_description */
extern void T47f28(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ENUMERATION_OPTION.extend */
extern void T47f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ENUMERATION_OPTION.update_parameter_description */
extern void T47f32(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].forth */
extern void T152f21(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].is_last */
extern T1 T152f10(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].cursor_is_last */
extern T1 T152f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].item_for_iteration */
extern T0* T152f8(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].cursor_item */
extern T0* T152f6(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].after */
extern T1 T152f11(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].cursor_after */
extern T1 T152f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].start */
extern void T152f20(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.set_description */
extern void T47f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ENUMERATION_OPTION.make_with_long_form */
extern T0* T47c25(GE_context* ac, T0* a1);
/* [detachable] AP_ENUMERATION_OPTION.set_long_form */
extern void T47f30(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ENUMERATION_OPTION.initialize */
extern void T47f29(GE_context* ac, T0* C);
/* [detachable] DS_LINKED_LIST [[attached] STRING_8].set_equality_tester */
extern void T152f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_ENUMERATION_OPTION.string_equality_tester */
extern T0* T47f14(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.initialize */
extern void T47f29p1(GE_context* ac, T0* C);
/* [detachable] AP_ENUMERATION_OPTION.initialize */
extern void T47f29p0(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.set_default_parameter */
extern void T44f27(GE_context* ac, T0* C, T0* a1);
/* [detachable] UT_VERSION.out */
extern T0* T46f10(GE_context* ac, T0* C);
/* [detachable] GEC.ise_latest */
extern T0* T25f34(GE_context* ac, T0* C);
/* [detachable] AP_FLAG.make_with_long_form */
extern T0* T39c17(GE_context* ac, T0* a1);
/* [detachable] AP_STRING_OPTION.set_parameter_description */
extern void T44f26(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_STRING_OPTION.set_description */
extern void T44f25(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_STRING_OPTION.make_with_long_form */
extern T0* T44c24(GE_context* ac, T0* a1);
/* [detachable] AP_STRING_OPTION.set_long_form */
extern void T44f29(GE_context* ac, T0* C, T0* a1);
/* [detachable] AP_STRING_OPTION.initialize */
extern void T44f28(GE_context* ac, T0* C);
/* [detachable] AP_STRING_OPTION.initialize */
extern void T44f28p1(GE_context* ac, T0* C);
/* [detachable] ET_ISE_VARIABLES.set_ise_variables */
extern void T38f18(GE_context* ac, T0* C);
/* [detachable] ET_ISE_VARIABLES.set_ise_c_compiler_variable */
extern void T38f22(GE_context* ac, T0* C);
/* [detachable] ET_ISE_VARIABLES.string_ */
extern T0* T38f11(GE_context* ac, T0* C);
/* [detachable] ET_ISE_VARIABLES.file_system */
extern T0* T38f10(GE_context* ac, T0* C);
/* [detachable] ET_ISE_VARIABLES.unix_file_system */
extern T0* T38f16(GE_context* ac, T0* C);
/* [detachable] ET_ISE_VARIABLES.windows_file_system */
extern T0* T38f15(GE_context* ac, T0* C);
/* [detachable] ET_ISE_VARIABLES.operating_system */
extern T0* T38f14(GE_context* ac, T0* C);
/* [detachable] ET_ISE_VARIABLES.ise_c_compiler_value */
extern T0* T38f9(GE_context* ac, T0* C);
/* [detachable] ET_ISE_VARIABLES.set_ise_platform_variable */
extern void T38f21(GE_context* ac, T0* C);
/* [detachable] ET_ISE_VARIABLES.set_eiffel_library_variable */
extern void T38f20(GE_context* ac, T0* C);
/* [detachable] ET_ISE_VARIABLES.ise_eiffel_value */
extern T0* T38f2(GE_context* ac, T0* C);
/* [detachable] ET_ISE_VARIABLES.eiffel_library_value */
extern T0* T38f5(GE_context* ac, T0* C);
/* [detachable] ET_ISE_VARIABLES.set_ise_library_variable */
extern void T38f19(GE_context* ac, T0* C);
/* [detachable] ET_ISE_VARIABLES.ise_library_value */
extern T0* T38f1(GE_context* ac, T0* C);
/* [detachable] GEC.ise_variables */
extern T0* T25f36(GE_context* ac, T0* C);
/* [detachable] KL_ARGUMENTS.set_program_name */
extern void T32f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_ERROR_HANDLER.make_standard */
extern T0* T36c180(GE_context* ac);
/* [detachable] ET_ERROR_HANDLER.make_standard */
extern void T36f180p1(GE_context* ac, T0* C);
/* [detachable] ET_ERROR_HANDLER.std */
extern T0* T36f24(GE_context* ac, T0* C);
/* [detachable] KL_ARGUMENTS.to_array */
extern T0* T32f5(GE_context* ac, T0* C);
/* [detachable] KL_ARGUMENTS.argument_count */
extern T6 T32f2(GE_context* ac, T0* C);
/* [detachable] GEC.arguments */
extern T0* T25f33(GE_context* ac, T0* C);
/* [detachable] KL_EXCEPTIONS.die */
extern void T35f2(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_EXCEPTIONS.exceptions_die */
extern void T35f3(GE_context* ac, T0* C, T6 a1);
/* [detachable] GEC.exceptions */
extern T0* T25f32(GE_context* ac, T0* C);
/* [detachable] ISE_EXCEPTION_MANAGER.set_exception_data */
extern void T21f16(GE_context* ac, T0* C, T6 a1, T1 a2, T6 a3, T6 a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T0* a10, T6 a11, T1 a12);
/* [detachable] NO_MORE_MEMORY.set_type_name */
extern void T99f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] SERIALIZATION_FAILURE.set_type_name */
extern void T123f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] OLD_VIOLATION.set_type_name */
extern void T97f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] COM_FAILURE.set_type_name */
extern void T104f19(GE_context* ac, T0* C, T0* a1);
/* [detachable] ADDRESS_APPLIED_TO_MELTED_FEATURE.set_type_name */
extern void T122f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] DEVELOPER_EXCEPTION.set_type_name */
extern void T121f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] MISMATCH_FAILURE.set_type_name */
extern void T120f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] OPERATING_SYSTEM_FAILURE.set_type_name */
extern void T103f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] IO_FAILURE.set_type_name */
extern void T102f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] EXCEPTION_IN_SIGNAL_HANDLER_FAILURE.set_type_name */
extern void T119f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] VOID_ASSIGNED_TO_EXPANDED.set_type_name */
extern void T118f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] EXTERNAL_FAILURE.set_type_name */
extern void T117f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] CREATE_ON_DEFERRED.set_type_name */
extern void T116f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] RESUMPTION_FAILURE.set_type_name */
extern void T115f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] RESCUE_FAILURE.set_type_name */
extern void T114f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] EIFFEL_RUNTIME_PANIC.set_type_name */
extern void T105f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] OPERATING_SYSTEM_SIGNAL_FAILURE.set_type_name */
extern void T101f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] LOOP_INVARIANT_VIOLATION.set_type_name */
extern void T113f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] VARIANT_VIOLATION.set_type_name */
extern void T112f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] BAD_INSPECT_VALUE.set_type_name */
extern void T111f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] ROUTINE_FAILURE.set_type_name */
extern void T95f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] CHECK_VIOLATION.set_type_name */
extern void T110f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] INVARIANT_VIOLATION.set_type_name */
extern void T100f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] FLOATING_POINT_FAILURE.set_type_name */
extern void T109f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] POSTCONDITION_VIOLATION.set_type_name */
extern void T108f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] PRECONDITION_VIOLATION.set_type_name */
extern void T107f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] VOID_TARGET.set_type_name */
extern void T93f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] NO_MORE_MEMORY.set_recipient_name */
extern void T99f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] SERIALIZATION_FAILURE.set_recipient_name */
extern void T123f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] OLD_VIOLATION.set_recipient_name */
extern void T97f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] COM_FAILURE.set_recipient_name */
extern void T104f18(GE_context* ac, T0* C, T0* a1);
/* [detachable] ADDRESS_APPLIED_TO_MELTED_FEATURE.set_recipient_name */
extern void T122f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DEVELOPER_EXCEPTION.set_recipient_name */
extern void T121f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] MISMATCH_FAILURE.set_recipient_name */
extern void T120f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] OPERATING_SYSTEM_FAILURE.set_recipient_name */
extern void T103f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] IO_FAILURE.set_recipient_name */
extern void T102f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] EXCEPTION_IN_SIGNAL_HANDLER_FAILURE.set_recipient_name */
extern void T119f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] VOID_ASSIGNED_TO_EXPANDED.set_recipient_name */
extern void T118f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] EXTERNAL_FAILURE.set_recipient_name */
extern void T117f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] CREATE_ON_DEFERRED.set_recipient_name */
extern void T116f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] RESUMPTION_FAILURE.set_recipient_name */
extern void T115f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] RESCUE_FAILURE.set_recipient_name */
extern void T114f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] EIFFEL_RUNTIME_PANIC.set_recipient_name */
extern void T105f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] OPERATING_SYSTEM_SIGNAL_FAILURE.set_recipient_name */
extern void T101f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] LOOP_INVARIANT_VIOLATION.set_recipient_name */
extern void T113f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] VARIANT_VIOLATION.set_recipient_name */
extern void T112f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] BAD_INSPECT_VALUE.set_recipient_name */
extern void T111f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ROUTINE_FAILURE.set_recipient_name */
extern void T95f14(GE_context* ac, T0* C, T0* a1);
/* [detachable] CHECK_VIOLATION.set_recipient_name */
extern void T110f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] INVARIANT_VIOLATION.set_recipient_name */
extern void T100f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] FLOATING_POINT_FAILURE.set_recipient_name */
extern void T109f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] POSTCONDITION_VIOLATION.set_recipient_name */
extern void T108f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] PRECONDITION_VIOLATION.set_recipient_name */
extern void T107f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] VOID_TARGET.set_recipient_name */
extern void T93f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] NO_MORE_MEMORY.set_exception_trace */
extern void T99f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] SERIALIZATION_FAILURE.set_exception_trace */
extern void T123f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] OLD_VIOLATION.set_exception_trace */
extern void T97f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] COM_FAILURE.set_exception_trace */
extern void T104f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] ADDRESS_APPLIED_TO_MELTED_FEATURE.set_exception_trace */
extern void T122f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] DEVELOPER_EXCEPTION.set_exception_trace */
extern void T121f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] MISMATCH_FAILURE.set_exception_trace */
extern void T120f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] OPERATING_SYSTEM_FAILURE.set_exception_trace */
extern void T103f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] IO_FAILURE.set_exception_trace */
extern void T102f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] EXCEPTION_IN_SIGNAL_HANDLER_FAILURE.set_exception_trace */
extern void T119f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] VOID_ASSIGNED_TO_EXPANDED.set_exception_trace */
extern void T118f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] EXTERNAL_FAILURE.set_exception_trace */
extern void T117f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] CREATE_ON_DEFERRED.set_exception_trace */
extern void T116f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] RESUMPTION_FAILURE.set_exception_trace */
extern void T115f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] RESCUE_FAILURE.set_exception_trace */
extern void T114f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] EIFFEL_RUNTIME_PANIC.set_exception_trace */
extern void T105f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] OPERATING_SYSTEM_SIGNAL_FAILURE.set_exception_trace */
extern void T101f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] LOOP_INVARIANT_VIOLATION.set_exception_trace */
extern void T113f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] VARIANT_VIOLATION.set_exception_trace */
extern void T112f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] BAD_INSPECT_VALUE.set_exception_trace */
extern void T111f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ROUTINE_FAILURE.set_exception_trace */
extern void T95f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] CHECK_VIOLATION.set_exception_trace */
extern void T110f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] INVARIANT_VIOLATION.set_exception_trace */
extern void T100f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] FLOATING_POINT_FAILURE.set_exception_trace */
extern void T109f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] POSTCONDITION_VIOLATION.set_exception_trace */
extern void T108f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] PRECONDITION_VIOLATION.set_exception_trace */
extern void T107f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] VOID_TARGET.set_exception_trace */
extern void T93f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] ISE_EXCEPTION_MANAGER.exception_from_data */
extern T0* T21f10(GE_context* ac, T0* C);
extern T1 T21f10ot1(T0* a1, T0** a2);
extern T1 T21f10ot2(T0* a1, T0** a2);
extern T1 T21f10ot3(T0* a1, T0** a2);
extern T1 T21f10ot4(T0* a1, T0** a2);
extern T1 T21f10ot5(T0* a1, T0** a2);
extern T1 T21f10ot6(T0* a1, T0** a2);
extern T1 T21f10ot7(T0* a1, T0** a2);
/* [detachable] NO_MORE_MEMORY.set_description */
extern void T99f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] SERIALIZATION_FAILURE.set_description */
extern void T123f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] OLD_VIOLATION.set_description */
extern void T97f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] COM_FAILURE.set_description */
extern void T104f17(GE_context* ac, T0* C, T0* a1);
/* [detachable] ADDRESS_APPLIED_TO_MELTED_FEATURE.set_description */
extern void T122f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] MISMATCH_FAILURE.set_description */
extern void T120f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] OPERATING_SYSTEM_FAILURE.set_description */
extern void T103f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] IO_FAILURE.set_description */
extern void T102f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] EXCEPTION_IN_SIGNAL_HANDLER_FAILURE.set_description */
extern void T119f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] VOID_ASSIGNED_TO_EXPANDED.set_description */
extern void T118f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] EXTERNAL_FAILURE.set_description */
extern void T117f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] CREATE_ON_DEFERRED.set_description */
extern void T116f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] RESUMPTION_FAILURE.set_description */
extern void T115f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] RESCUE_FAILURE.set_description */
extern void T114f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] EIFFEL_RUNTIME_PANIC.set_description */
extern void T105f12(GE_context* ac, T0* C, T0* a1);
/* [detachable] OPERATING_SYSTEM_SIGNAL_FAILURE.set_description */
extern void T101f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] LOOP_INVARIANT_VIOLATION.set_description */
extern void T113f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] VARIANT_VIOLATION.set_description */
extern void T112f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] BAD_INSPECT_VALUE.set_description */
extern void T111f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] ROUTINE_FAILURE.set_description */
extern void T95f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] CHECK_VIOLATION.set_description */
extern void T110f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] INVARIANT_VIOLATION.set_description */
extern void T100f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] FLOATING_POINT_FAILURE.set_description */
extern void T109f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] POSTCONDITION_VIOLATION.set_description */
extern void T108f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] PRECONDITION_VIOLATION.set_description */
extern void T107f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] VOID_TARGET.set_description */
extern void T93f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] COM_FAILURE.set_exception_information */
extern void T104f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] COM_FAILURE.ccom_hresult_to_string */
extern T0* T104f11(GE_context* ac, T0* C, T6 a1);
/* INTEGER_32.to_hex_string */
extern T0* T6f28(GE_context* ac, T6* C);
/* [detachable] COM_FAILURE.set_hresult_code */
extern void T104f14(GE_context* ac, T0* C, T6 a1);
/* [detachable] OPERATING_SYSTEM_FAILURE.set_error_code */
extern void T103f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] IO_FAILURE.set_error_code */
extern void T102f11(GE_context* ac, T0* C, T6 a1);
/* [detachable] OPERATING_SYSTEM_SIGNAL_FAILURE.set_signal_code */
extern void T101f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] INVARIANT_VIOLATION.set_is_entry */
extern void T100f9(GE_context* ac, T0* C, T1 a1);
/* [detachable] ROUTINE_FAILURE.set_class_name */
extern void T95f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] ROUTINE_FAILURE.set_routine_name */
extern void T95f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] NO_MORE_MEMORY.set_throwing_exception */
extern void T99f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] SERIALIZATION_FAILURE.set_throwing_exception */
extern void T123f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] OLD_VIOLATION.set_throwing_exception */
extern void T97f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] COM_FAILURE.set_throwing_exception */
extern void T104f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] ADDRESS_APPLIED_TO_MELTED_FEATURE.set_throwing_exception */
extern void T122f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] MISMATCH_FAILURE.set_throwing_exception */
extern void T120f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] OPERATING_SYSTEM_FAILURE.set_throwing_exception */
extern void T103f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] IO_FAILURE.set_throwing_exception */
extern void T102f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] EXCEPTION_IN_SIGNAL_HANDLER_FAILURE.set_throwing_exception */
extern void T119f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] VOID_ASSIGNED_TO_EXPANDED.set_throwing_exception */
extern void T118f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] EXTERNAL_FAILURE.set_throwing_exception */
extern void T117f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] CREATE_ON_DEFERRED.set_throwing_exception */
extern void T116f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] RESUMPTION_FAILURE.set_throwing_exception */
extern void T115f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] RESCUE_FAILURE.set_throwing_exception */
extern void T114f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] EIFFEL_RUNTIME_PANIC.set_throwing_exception */
extern void T105f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] OPERATING_SYSTEM_SIGNAL_FAILURE.set_throwing_exception */
extern void T101f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] LOOP_INVARIANT_VIOLATION.set_throwing_exception */
extern void T113f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] VARIANT_VIOLATION.set_throwing_exception */
extern void T112f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] BAD_INSPECT_VALUE.set_throwing_exception */
extern void T111f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ROUTINE_FAILURE.set_throwing_exception */
extern void T95f11(GE_context* ac, T0* C, T0* a1);
/* [detachable] CHECK_VIOLATION.set_throwing_exception */
extern void T110f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] INVARIANT_VIOLATION.set_throwing_exception */
extern void T100f8(GE_context* ac, T0* C, T0* a1);
/* [detachable] FLOATING_POINT_FAILURE.set_throwing_exception */
extern void T109f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] POSTCONDITION_VIOLATION.set_throwing_exception */
extern void T108f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] PRECONDITION_VIOLATION.set_throwing_exception */
extern void T107f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] VOID_TARGET.set_throwing_exception */
extern void T93f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] ISE_EXCEPTION_MANAGER.exception_from_code */
extern T0* T21f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] SERIALIZATION_FAILURE.default_create */
extern T0* T123c6(GE_context* ac);
/* [detachable] OLD_VIOLATION.default_create */
extern T0* T97c6(GE_context* ac);
/* [detachable] COM_FAILURE.default_create */
extern T0* T104c12(GE_context* ac);
/* [detachable] ADDRESS_APPLIED_TO_MELTED_FEATURE.default_create */
extern T0* T122c6(GE_context* ac);
/* [detachable] MISMATCH_FAILURE.default_create */
extern T0* T120c6(GE_context* ac);
/* [detachable] OPERATING_SYSTEM_FAILURE.default_create */
extern T0* T103c7(GE_context* ac);
/* [detachable] IO_FAILURE.set_code */
extern void T102f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] IO_FAILURE.default_create */
extern T0* T102c8(GE_context* ac);
/* [detachable] EXCEPTION_IN_SIGNAL_HANDLER_FAILURE.default_create */
extern T0* T119c6(GE_context* ac);
/* [detachable] VOID_ASSIGNED_TO_EXPANDED.default_create */
extern T0* T118c6(GE_context* ac);
/* [detachable] EXTERNAL_FAILURE.default_create */
extern T0* T117c6(GE_context* ac);
/* [detachable] CREATE_ON_DEFERRED.default_create */
extern T0* T116c6(GE_context* ac);
/* [detachable] RESUMPTION_FAILURE.default_create */
extern T0* T115c6(GE_context* ac);
/* [detachable] RESCUE_FAILURE.default_create */
extern T0* T114c6(GE_context* ac);
/* [detachable] EIFFEL_RUNTIME_PANIC.set_code */
extern void T105f9(GE_context* ac, T0* C, T6 a1);
/* [detachable] EIFFEL_RUNTIME_PANIC.code */
extern T6 T105f7(GE_context* ac, T0* C);
/* [detachable] EIFFEL_RUNTIME_PANIC.default_create */
extern T0* T105c8(GE_context* ac);
/* [detachable] OPERATING_SYSTEM_SIGNAL_FAILURE.default_create */
extern T0* T101c7(GE_context* ac);
/* [detachable] LOOP_INVARIANT_VIOLATION.default_create */
extern T0* T113c6(GE_context* ac);
/* [detachable] VARIANT_VIOLATION.default_create */
extern T0* T112c6(GE_context* ac);
/* [detachable] BAD_INSPECT_VALUE.default_create */
extern T0* T111c6(GE_context* ac);
/* [detachable] ROUTINE_FAILURE.default_create */
extern T0* T95c8(GE_context* ac);
/* [detachable] CHECK_VIOLATION.default_create */
extern T0* T110c6(GE_context* ac);
/* [detachable] INVARIANT_VIOLATION.default_create */
extern T0* T100c7(GE_context* ac);
/* [detachable] FLOATING_POINT_FAILURE.default_create */
extern T0* T109c6(GE_context* ac);
/* [detachable] POSTCONDITION_VIOLATION.default_create */
extern T0* T108c6(GE_context* ac);
/* [detachable] PRECONDITION_VIOLATION.default_create */
extern T0* T107c6(GE_context* ac);
/* [detachable] NO_MORE_MEMORY.set_code */
extern void T99f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] NO_MORE_MEMORY.code */
extern T6 T99f7(GE_context* ac, T0* C);
/* [detachable] ISE_EXCEPTION_MANAGER.no_memory_exception_object_cell */
extern T0* T21f7(GE_context* ac, T0* C);
/* [detachable] CELL [[attached] NO_MORE_MEMORY].put */
extern T0* T29c2(GE_context* ac, T0* a1);
/* [detachable] NO_MORE_MEMORY.default_create */
extern T0* T99c8(GE_context* ac);
/* [detachable] VOID_TARGET.default_create */
extern T0* T93c6(GE_context* ac);
/* [detachable] ISE_EXCEPTION_MANAGER.exception_data */
extern T0* T21f11(GE_context* ac, T0* C);
/* [detachable] CELL [detachable TUPLE [[attached] INTEGER_32, [attached] INTEGER_32, [attached] INTEGER_32, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] INTEGER_32, [attached] BOOLEAN]].put */
extern void T28f2(GE_context* ac, T0* C, T0* a1);
/* [detachable] CELL [detachable TUPLE [[attached] INTEGER_32, [attached] INTEGER_32, [attached] INTEGER_32, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] INTEGER_32, [attached] BOOLEAN]].put */
extern T0* T28c2(GE_context* ac, T0* a1);
/* [detachable] ISE_EXCEPTION_MANAGER.exception_data_cell */
extern T0* T21f6(GE_context* ac, T0* C);
/* [detachable] ISE_EXCEPTION_MANAGER.once_raise */
extern void T21f15(GE_context* ac, T0* C, T0* a1);
/* [detachable] ISE_EXCEPTION_MANAGER.last_exception */
extern T0* T21f1(GE_context* ac, T0* C);
/* [detachable] ISE_EXCEPTION_MANAGER.init_exception_manager */
extern void T21f14(GE_context* ac, T0* C);
/* [detachable] CELL [[attached] NO_MORE_MEMORY].do_nothing */
extern void T29f3(GE_context* ac, T0* C);
/* [detachable] CELL [detachable EXCEPTION].do_nothing */
extern void T26f3(GE_context* ac, T0* C);
/* [detachable] CELL [detachable TUPLE [[attached] INTEGER_32, [attached] INTEGER_32, [attached] INTEGER_32, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] STRING_8, [attached] INTEGER_32, [attached] BOOLEAN]].do_nothing */
extern void T28f3(GE_context* ac, T0* C);
/* [detachable] ISE_EXCEPTION_MANAGER.unraisable_exceptions */
extern T0* T21f5(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].force */
extern void T27f37(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].position */
extern T6 T27f22(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].deleted_position */
extern T6 T27f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].add_space */
extern void T27f39(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].accommodate */
extern void T27f40(GE_context* ac, T0* C, T6 a1);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].set_indexes_map */
extern void T27f44(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].set_deleted_marks */
extern void T27f43(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].set_keys */
extern void T27f42(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].set_content */
extern void T27f41(GE_context* ac, T0* C, T0* a1);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].put */
extern void T27f45(GE_context* ac, T0* C, T6 a1, T6 a2);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].set_conflict */
extern void T27f47(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].found */
extern T1 T27f31(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].occupied */
extern T1 T27f29(GE_context* ac, T0* C, T6 a1);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].empty_duplicate */
extern T0* T27f28(GE_context* ac, T0* C, T6 a1);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].compare_objects */
extern void T27f46(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].soon_full */
extern T1 T27f20(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].not_found */
extern T1 T27f19(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].internal_search */
extern void T27f38(GE_context* ac, T0* C, T6 a1);
/* [detachable] ISE_EXCEPTION_MANAGER.unignorable_exceptions */
extern T0* T21f4(GE_context* ac, T0* C);
/* [detachable] HASH_TABLE [[attached] INTEGER_32, [attached] INTEGER_32].do_nothing */
extern void T27f36(GE_context* ac, T0* C);
/* [detachable] KL_DIRECTORY.dispose */
extern void T333f32(GE_context* ac, T0* C);
/* [detachable] KL_TEXT_OUTPUT_FILE.dispose */
extern void T300f28(GE_context* ac, T0* C);
/* [detachable] MANAGED_POINTER.dispose */
extern void T140f9(GE_context* ac, T0* C);
/* [detachable] MUTEX.dispose */
extern void T131f9(GE_context* ac, T0* C);
/* [detachable] MUTEX.destroy */
extern void T131f13(GE_context* ac, T0* C);
/* [detachable] MUTEX.eif_thr_mutex_destroy */
extern void T131f16(GE_context* ac, T0* C, T14 a1);
/* [detachable] KL_TEXT_INPUT_FILE.dispose */
extern void T37f60(GE_context* ac, T0* C);
/* [detachable] PLATFORM.is_mac */
extern T1 T50s4(GE_context* ac);
/* [detachable] PLATFORM.is_64_bits */
extern T1 T50s3(GE_context* ac);
/* [detachable] ET_ECF_OPTION_DEFAULTS.valid_assertions_latest */
extern T0* T1740s2(GE_context* ac);
/* Creation of agent #1 in feature [detachable] ET_ECF_OPTION_DEFAULTS.valid_assertions_latest */
extern T0* T1740s2ac1(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_OPTION_DEFAULTS.boolean_option_value_regexp */
extern T0* T1740s8(GE_context* ac);
/* [detachable] ET_ECF_OPTION_DEFAULTS.string_ */
extern T0* T1740s7(GE_context* ac);
/* [detachable] ET_ECF_OPTION_DEFAULTS.case_insensitive_string_equality_tester */
extern T0* T1740s6(GE_context* ac);
/* [detachable] ET_ECF_OPTION_DEFAULTS.valid_assertions */
extern T0* T1740s1(GE_context* ac, T0* a1);
/* [detachable] ET_ECF_OPTION_DEFAULTS.valid_assertions_1_0_0 */
extern T0* T1740s5(GE_context* ac);
/* [detachable] ET_ECF_OPTION_DEFAULTS.valid_assertions_1_2_0 */
extern T0* T1740s4(GE_context* ac);
/* [detachable] ET_ECF_OPTION_DEFAULTS.ecf_1_2_0 */
extern T0* T1740s3(GE_context* ac);
/* [detachable] UT_SHARED_ECF_VERSIONS.ecf_last_known */
extern T0* T2067s1(GE_context* ac);
/* [detachable] UT_SHARED_ECF_VERSIONS.ecf_1_22_0 */
extern T0* T2067s2(GE_context* ac);
/* [detachable] UC_UTF8_ROUTINES.append_code_to_utf8 */
extern void T434s56(GE_context* ac, T0* a1, T6 a2);
/* [detachable] UC_UTF8_ROUTINES.is_encoded_next_byte */
extern T1 T434s30(GE_context* ac, T2 a1);
/* [detachable] UC_UTF8_ROUTINES.is_encoded_second_byte */
extern T1 T434s26(GE_context* ac, T2 a1, T2 a2);
/* [detachable] UC_UTF8_ROUTINES.is_encoded_first_byte */
extern T1 T434s25(GE_context* ac, T2 a1);
/* [detachable] UC_UTF8_ROUTINES.is_endian_detection_character */
extern T1 T434s39(GE_context* ac, T2 a1, T2 a2, T2 a3);
/* [detachable] UC_UTF8_ROUTINES.is_endian_detection_character_start */
extern T1 T434s40(GE_context* ac, T2 a1, T2 a2);
/* [detachable] UC_UNICODE_CONSTANTS.bom_character */
extern T3 T779s6(GE_context* ac);
/* [detachable] UC_UTF8_ROUTINES.append_substring_to_utf8 */
extern void T434s54(GE_context* ac, T0* a1, T0* a2, T6 a3, T6 a4);
/* [detachable] UC_UTF8_ROUTINES.append_natural_32_code_to_utf8 */
extern void T434s55(GE_context* ac, T0* a1, T10 a2);
/* [detachable] UC_UTF8_ROUTINES.four_byte_character_code */
extern T10 T434s46(GE_context* ac, T2 a1, T2 a2, T2 a3, T2 a4);
/* [detachable] UC_UTF8_ROUTINES.three_byte_character_code */
extern T10 T434s45(GE_context* ac, T2 a1, T2 a2, T2 a3);
/* [detachable] UC_UTF8_ROUTINES.two_byte_character_code */
extern T10 T434s44(GE_context* ac, T2 a1, T2 a2);
/* [detachable] UC_UTF8_ROUTINES.encoded_byte_count */
extern T6 T434s4(GE_context* ac, T2 a1);
/* [detachable] UC_UTF8_ROUTINES.unicode_substring_character_count */
extern T6 T434s48(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] UC_UTF8_ROUTINES.unicode_character_count */
extern T6 T434s47(GE_context* ac, T0* a1);
/* [detachable] UC_UTF8_ROUTINES.is_string_8 */
extern T1 T434s50(GE_context* ac, T0* a1);
/* [detachable] UC_UTF8_ROUTINES.is_string_32 */
extern T1 T434s51(GE_context* ac, T0* a1);
/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER.reset_context_if_multiple_constraints */
extern void T405s11(GE_context* ac, T1 a1, T0* a2, T0* a3);
/* [detachable] ET_ADAPTED_BASE_CLASS_CHECKER.tokens */
extern T0* T405s7(GE_context* ac);
/* [detachable] UC_UNICODE_ROUTINES.valid_non_surrogate_natural_32_code */
extern T1 T426s1(GE_context* ac, T10 a1);
/* [detachable] UC_UTF8_ROUTINES.string_to_utf8 */
extern T0* T434s1(GE_context* ac, T0* a1);
/* [detachable] UC_UTF8_ROUTINES.substring_to_utf8 */
extern T0* T434s8(GE_context* ac, T0* a1, T6 a2, T6 a3);
/* [detachable] UC_UTF8_ROUTINES.substring_byte_count */
extern T6 T434s2(GE_context* ac, T0* a1, T6 a2, T6 a3);
extern T1 T434s2ot1(T0* a1, T0** a2);
extern T1 T434s2ot2(T0* a1, T0** a2);
extern T1 T434s2ot3(T0* a1, T0** a2);
/* [detachable] UC_UTF8_ROUTINES.character_32_byte_count */
extern T6 T434s13(GE_context* ac, T3 a1);
/* [detachable] UC_UTF8_ROUTINES.natural_32_code_byte_count */
extern T6 T434s18(GE_context* ac, T10 a1);
/* [detachable] UC_UTF8_ROUTINES.dummy_uc_string */
extern T0* T434s12(GE_context* ac);
/* [detachable] UC_UTF8_ROUTINES.character_8_byte_count */
extern T6 T434s11(GE_context* ac, T2 a1);
/* [detachable] UC_UTF8_ROUTINES.any_ */
extern T0* T434s9(GE_context* ac);
/* [detachable] EXECUTION_ENVIRONMENT.available_cpu_count */
extern T10 T124s7(GE_context* ac);
/* [detachable] PLATFORM.is_thread_capable */
extern T1 T50s1(GE_context* ac);
/* [detachable] THREAD_CONTROL.join_all */
extern void T433s1(GE_context* ac);
/* [detachable] EXECUTION_ENVIRONMENT.sleep */
extern void T124s9(GE_context* ac, T7 a1);
/* [detachable] EXECUTION_ENVIRONMENT.eif_sleep */
extern void T124s12(GE_context* ac, T7 a1);
/* [detachable] ET_SYSTEM_PROCESSOR_FACTORY.new_system_processor */
extern T0* T69s1(GE_context* ac, T6 a1);
/* [detachable] PLATFORM.is_windows */
extern T1 T50s2(GE_context* ac);
/* [detachable] ISE_RUNTIME.check_assert */
extern T1 T774s1(GE_context* ac, T1 a1);
/* [detachable] ET_SYSTEM_PROCESSOR.make */
extern T0* T64c47(GE_context* ac);
/* [detachable] ET_TOKEN_CONSTANTS.standard_error_handler */
extern T0* T73f121(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.make */
extern T0* T456c1040(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.make_parser_skeleton */
extern void T456f1042(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_build_parser_tables */
extern void T456f1045(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template */
extern T0* T456f666(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yyfixed_array */
extern T0* T456f729(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_43 */
extern void T456f1160(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yyarray_subcopy */
extern void T456f1232(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] ET_EIFFEL_PARSER.integer_array_ */
extern T0* T456f736(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_42 */
extern void T456f1159(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_41 */
extern void T456f1158(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_40 */
extern void T456f1157(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_39 */
extern void T456f1156(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_38 */
extern void T456f1155(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_37 */
extern void T456f1154(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_36 */
extern void T456f1153(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_35 */
extern void T456f1152(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_34 */
extern void T456f1151(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_33 */
extern void T456f1150(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_32 */
extern void T456f1149(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_31 */
extern void T456f1148(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_30 */
extern void T456f1147(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_29 */
extern void T456f1146(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_28 */
extern void T456f1145(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_27 */
extern void T456f1144(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_26 */
extern void T456f1143(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_25 */
extern void T456f1142(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_24 */
extern void T456f1141(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_23 */
extern void T456f1140(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_22 */
extern void T456f1139(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_21 */
extern void T456f1138(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_20 */
extern void T456f1137(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_19 */
extern void T456f1136(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_18 */
extern void T456f1135(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_17 */
extern void T456f1134(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_16 */
extern void T456f1133(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_15 */
extern void T456f1132(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_14 */
extern void T456f1131(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_13 */
extern void T456f1130(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_12 */
extern void T456f1129(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_11 */
extern void T456f1128(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_10 */
extern void T456f1127(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_9 */
extern void T456f1126(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_8 */
extern void T456f1125(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_7 */
extern void T456f1124(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_6 */
extern void T456f1123(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_5 */
extern void T456f1122(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_4 */
extern void T456f1121(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_3 */
extern void T456f1120(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_2 */
extern void T456f1119(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yycheck_template_1 */
extern void T456f1118(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template */
extern T0* T456f749(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_43 */
extern void T456f1117(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_42 */
extern void T456f1116(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_41 */
extern void T456f1115(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_40 */
extern void T456f1114(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_39 */
extern void T456f1113(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_38 */
extern void T456f1112(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_37 */
extern void T456f1111(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_36 */
extern void T456f1110(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_35 */
extern void T456f1109(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_34 */
extern void T456f1108(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_33 */
extern void T456f1107(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_32 */
extern void T456f1106(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_31 */
extern void T456f1105(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_30 */
extern void T456f1104(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_29 */
extern void T456f1103(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_28 */
extern void T456f1102(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_27 */
extern void T456f1101(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_26 */
extern void T456f1100(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_25 */
extern void T456f1099(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_24 */
extern void T456f1098(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_23 */
extern void T456f1097(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_22 */
extern void T456f1096(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_21 */
extern void T456f1095(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_20 */
extern void T456f1094(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_19 */
extern void T456f1093(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_18 */
extern void T456f1092(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_17 */
extern void T456f1091(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_16 */
extern void T456f1090(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_15 */
extern void T456f1089(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_14 */
extern void T456f1088(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_13 */
extern void T456f1087(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_12 */
extern void T456f1086(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_11 */
extern void T456f1085(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_10 */
extern void T456f1084(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_9 */
extern void T456f1083(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_8 */
extern void T456f1082(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_7 */
extern void T456f1081(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_6 */
extern void T456f1080(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_5 */
extern void T456f1079(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_4 */
extern void T456f1078(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_3 */
extern void T456f1077(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_2 */
extern void T456f1076(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytable_template_1 */
extern void T456f1075(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yypgoto_template */
extern T0* T456f746(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yypgoto_template_2 */
extern void T456f1074(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yypgoto_template_1 */
extern void T456f1073(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yypact_template */
extern T0* T456f743(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yypact_template_10 */
extern void T456f1072(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yypact_template_9 */
extern void T456f1071(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yypact_template_8 */
extern void T456f1070(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yypact_template_7 */
extern void T456f1069(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yypact_template_6 */
extern void T456f1068(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yypact_template_5 */
extern void T456f1067(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yypact_template_4 */
extern void T456f1066(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yypact_template_3 */
extern void T456f1065(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yypact_template_2 */
extern void T456f1064(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yypact_template_1 */
extern void T456f1063(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yydefgoto_template */
extern T0* T456f741(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yydefgoto_template_2 */
extern void T456f1062(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yydefgoto_template_1 */
extern void T456f1061(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yydefact_template */
extern T0* T456f738(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yydefact_template_10 */
extern void T456f1060(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yydefact_template_9 */
extern void T456f1059(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yydefact_template_8 */
extern void T456f1058(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yydefact_template_7 */
extern void T456f1057(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yydefact_template_6 */
extern void T456f1056(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yydefact_template_5 */
extern void T456f1055(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yydefact_template_4 */
extern void T456f1054(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yydefact_template_3 */
extern void T456f1053(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yydefact_template_2 */
extern void T456f1052(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yydefact_template_1 */
extern void T456f1051(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytypes2_template */
extern T0* T456f728(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yytypes1_template */
extern T0* T456f726(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yytypes1_template_10 */
extern void T456f1228(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytypes1_template_9 */
extern void T456f1227(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytypes1_template_8 */
extern void T456f1226(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytypes1_template_7 */
extern void T456f1225(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytypes1_template_6 */
extern void T456f1224(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytypes1_template_5 */
extern void T456f1223(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytypes1_template_4 */
extern void T456f1222(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytypes1_template_3 */
extern void T456f1221(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytypes1_template_2 */
extern void T456f1220(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytypes1_template_1 */
extern void T456f1219(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yyr1_template */
extern T0* T456f722(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yyr1_template_5 */
extern void T456f1210(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yyr1_template_4 */
extern void T456f1209(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yyr1_template_3 */
extern void T456f1208(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yyr1_template_2 */
extern void T456f1207(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yyr1_template_1 */
extern void T456f1206(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytranslate_template */
extern T0* T456f718(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yytranslate_template_2 */
extern void T456f1199(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yytranslate_template_1 */
extern void T456f1198(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_create_value_stacks */
extern void T456f1044(GE_context* ac, T0* C);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WRITABLE].make */
extern T0* T1304f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_WRITABLE].make_empty */
extern T0* T1305c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WRITABLE].default_create */
extern T0* T1304c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART_LIST].make */
extern T0* T1301f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_WHEN_PART_LIST].make_empty */
extern T0* T1303c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART_LIST].default_create */
extern T0* T1301c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART].make */
extern T0* T1298f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_WHEN_PART].make_empty */
extern T0* T1300c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_PART].default_create */
extern T0* T1298c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_EXPRESSION_LIST].make */
extern T0* T1295f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_WHEN_EXPRESSION_LIST].make_empty */
extern T0* T1297c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_EXPRESSION_LIST].default_create */
extern T0* T1295c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_EXPRESSION].make */
extern T0* T1292f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_WHEN_EXPRESSION].make_empty */
extern T0* T1294c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_WHEN_EXPRESSION].default_create */
extern T0* T1292c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_VARIANT].make */
extern T0* T1289f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_VARIANT].make_empty */
extern T0* T1291c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_VARIANT].default_create */
extern T0* T1289c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_LIST].make */
extern T0* T1286f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_LIST].make_empty */
extern T0* T1288c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_LIST].default_create */
extern T0* T1286c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_ITEM].make */
extern T0* T1283f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_ITEM].make_empty */
extern T0* T1285c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_ITEM].default_create */
extern T0* T1283c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_LIST].make */
extern T0* T1280f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_LIST].make_empty */
extern T0* T1282c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_LIST].default_create */
extern T0* T1280c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_ITEM].make */
extern T0* T1277f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT_ITEM].make_empty */
extern T0* T1279c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT_ITEM].default_create */
extern T0* T1277c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT].make */
extern T0* T1274f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE_CONSTRAINT].make_empty */
extern T0* T1276c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE_CONSTRAINT].default_create */
extern T0* T1274c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE].make */
extern T0* T1272f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_TYPE].make_empty */
extern T0* T1273c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_TYPE].default_create */
extern T0* T1272c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_STRIP_EXPRESSION].make */
extern T0* T1269f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_STRIP_EXPRESSION].make_empty */
extern T0* T1271c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_STRIP_EXPRESSION].default_create */
extern T0* T1269c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_STATIC_CALL_EXPRESSION].make */
extern T0* T1266f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_STATIC_CALL_EXPRESSION].make_empty */
extern T0* T1268c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_STATIC_CALL_EXPRESSION].default_create */
extern T0* T1266c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_REPEAT_INSTRUCTION].make */
extern T0* T1263f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_REPEAT_INSTRUCTION].make_empty */
extern T0* T1265c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_REPEAT_INSTRUCTION].default_create */
extern T0* T1263c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_LIST].make */
extern T0* T1261f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RENAME_LIST].make_empty */
extern T0* T1262c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_LIST].default_create */
extern T0* T1261c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_ITEM].make */
extern T0* T1258f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RENAME_ITEM].make_empty */
extern T0* T1260c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RENAME_ITEM].default_create */
extern T0* T1258c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUERY].make */
extern T0* T1256f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUERY].make_empty */
extern T0* T1257c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUERY].default_create */
extern T0* T1256c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUANTIFIER_EXPRESSION].make */
extern T0* T1253f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUANTIFIER_EXPRESSION].make_empty */
extern T0* T1255c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUANTIFIER_EXPRESSION].default_create */
extern T0* T1253c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_LIKE_IDENTIFIER].make */
extern T0* T1250f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_LIKE_IDENTIFIER].make_empty */
extern T0* T1252c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_LIKE_IDENTIFIER].default_create */
extern T0* T1250c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PROCEDURE].make */
extern T0* T1248f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PROCEDURE].make_empty */
extern T0* T1249c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PROCEDURE].default_create */
extern T0* T1248c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECONDITIONS].make */
extern T0* T1246f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PRECONDITIONS].make_empty */
extern T0* T1247c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECONDITIONS].default_create */
extern T0* T1246c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSTCONDITIONS].make */
extern T0* T1243f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_POSTCONDITIONS].make_empty */
extern T0* T1245c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSTCONDITIONS].default_create */
extern T0* T1243c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_LIST].make */
extern T0* T1241f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT_LIST].make_empty */
extern T0* T1242c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_LIST].default_create */
extern T0* T1241c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_ITEM].make */
extern T0* T1239f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT_ITEM].make_empty */
extern T0* T1240c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_ITEM].default_create */
extern T0* T1239c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_CLAUSE_LIST].make */
extern T0* T1236f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT_CLAUSE_LIST].make_empty */
extern T0* T1238c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT_CLAUSE_LIST].default_create */
extern T0* T1236c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT].make */
extern T0* T1234f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENT].make_empty */
extern T0* T1235c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENT].default_create */
extern T0* T1234c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENTHESIZED_EXPRESSION].make */
extern T0* T1231f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PARENTHESIZED_EXPRESSION].make_empty */
extern T0* T1233c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PARENTHESIZED_EXPRESSION].default_create */
extern T0* T1231c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBSOLETE].make */
extern T0* T1228f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_OBSOLETE].make_empty */
extern T0* T1230c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBSOLETE].default_create */
extern T0* T1228c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_TUPLE].make */
extern T0* T1226f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_TUPLE].make_empty */
extern T0* T1227c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_TUPLE].default_create */
extern T0* T1226c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_LIST].make */
extern T0* T1223f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_LIST].make_empty */
extern T0* T1225c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_LIST].default_create */
extern T0* T1223c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_ITEM].make */
extern T0* T1220f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING_ITEM].make_empty */
extern T0* T1222c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING_ITEM].default_create */
extern T0* T1220c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_ARRAY].make */
extern T0* T1217f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_ARRAY].make_empty */
extern T0* T1219c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_ARRAY].default_create */
extern T0* T1217c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOOP_INVARIANTS].make */
extern T0* T1214f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LOOP_INVARIANTS].make_empty */
extern T0* T1216c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOOP_INVARIANTS].default_create */
extern T0* T1214c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_LIST].make */
extern T0* T1211f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_LIST].make_empty */
extern T0* T1213c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_LIST].default_create */
extern T0* T1211c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_ITEM].make */
extern T0* T1208f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE_ITEM].make_empty */
extern T0* T1210c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE_ITEM].default_create */
extern T0* T1208c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE].make */
extern T0* T1205f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LOCAL_VARIABLE].make_empty */
extern T0* T1207c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LOCAL_VARIABLE].default_create */
extern T0* T1205c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LIKE_TYPE].make */
extern T0* T1202f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_LIKE_TYPE].make_empty */
extern T0* T1204c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_LIKE_TYPE].default_create */
extern T0* T1202c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_FEATURE_NAME_LIST].make */
extern T0* T1200f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_KEYWORD_FEATURE_NAME_LIST].make_empty */
extern T0* T1201c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_FEATURE_NAME_LIST].default_create */
extern T0* T1200c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INVARIANTS].make */
extern T0* T1197f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INVARIANTS].make_empty */
extern T0* T1199c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INVARIANTS].default_create */
extern T0* T1197c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSTRUCTION].make */
extern T0* T1194f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INSTRUCTION].make_empty */
extern T0* T1196c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSTRUCTION].default_create */
extern T0* T1194c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSPECT_INSTRUCTION].make */
extern T0* T1191f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INSPECT_INSTRUCTION].make_empty */
extern T0* T1193c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSPECT_INSTRUCTION].default_create */
extern T0* T1191c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSPECT_EXPRESSION].make */
extern T0* T1188f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INSPECT_EXPRESSION].make_empty */
extern T0* T1190c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INSPECT_EXPRESSION].default_create */
extern T0* T1188c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INLINE_AGENT].make */
extern T0* T1186f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INLINE_AGENT].make_empty */
extern T0* T1187c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INLINE_AGENT].default_create */
extern T0* T1186c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_LIST].make */
extern T0* T1183f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_LIST].make_empty */
extern T0* T1185c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_LIST].default_create */
extern T0* T1183c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_ITEM].make */
extern T0* T1180f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM_ITEM].make_empty */
extern T0* T1182c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM_ITEM].default_create */
extern T0* T1180c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM].make */
extern T0* T1178f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_TERM].make_empty */
extern T0* T1179c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_TERM].default_create */
extern T0* T1178c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_ITEM].make */
extern T0* T1175f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_ITEM].make_empty */
extern T0* T1177c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_ITEM].default_create */
extern T0* T1175c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_LIST].make */
extern T0* T1172f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INDEXING_LIST].make_empty */
extern T0* T1174c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INDEXING_LIST].default_create */
extern T0* T1172c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_INSTRUCTION].make */
extern T0* T1169f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_IF_INSTRUCTION].make_empty */
extern T0* T1171c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_INSTRUCTION].default_create */
extern T0* T1169c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_EXPRESSION].make */
extern T0* T1166f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_IF_EXPRESSION].make_empty */
extern T0* T1168c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IF_EXPRESSION].default_create */
extern T0* T1166c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_LIST].make */
extern T0* T1164f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_LIST].make_empty */
extern T0* T1165c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_LIST].default_create */
extern T0* T1164c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_ITEM].make */
extern T0* T1161f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER_ITEM].make_empty */
extern T0* T1163c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER_ITEM].default_create */
extern T0* T1161c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER].make */
extern T0* T1159f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_PARAMETER].make_empty */
extern T0* T1160c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_PARAMETER].default_create */
extern T0* T1159c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_LIST].make */
extern T0* T1157f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_LIST].make_empty */
extern T0* T1158c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_LIST].default_create */
extern T0* T1157c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_ITEM].make */
extern T0* T1154f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT_ITEM].make_empty */
extern T0* T1156c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT_ITEM].default_create */
extern T0* T1154c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT].make */
extern T0* T1152f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FORMAL_ARGUMENT].make_empty */
extern T0* T1153c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FORMAL_ARGUMENT].default_create */
extern T0* T1152c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME_ITEM].make */
extern T0* T1149f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME_ITEM].make_empty */
extern T0* T1151c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME_ITEM].default_create */
extern T0* T1149c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME].make */
extern T0* T1147f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_NAME].make_empty */
extern T0* T1148c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_NAME].default_create */
extern T0* T1147c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_EXPORT].make */
extern T0* T1144f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_EXPORT].make_empty */
extern T0* T1146c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_EXPORT].default_create */
extern T0* T1144c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE_LIST].make */
extern T0* T1141f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE_LIST].make_empty */
extern T0* T1143c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE_LIST].default_create */
extern T0* T1141c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE].make */
extern T0* T1138f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FEATURE_CLAUSE].make_empty */
extern T0* T1140c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FEATURE_CLAUSE].default_create */
extern T0* T1138c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTERNAL_ALIAS].make */
extern T0* T1135f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXTERNAL_ALIAS].make_empty */
extern T0* T1137c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTERNAL_ALIAS].default_create */
extern T0* T1135c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTENDED_FEATURE_NAME].make */
extern T0* T1133f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXTENDED_FEATURE_NAME].make_empty */
extern T0* T1134c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXTENDED_FEATURE_NAME].default_create */
extern T0* T1133c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION_ITEM].make */
extern T0* T1131f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPRESSION_ITEM].make_empty */
extern T0* T1132c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION_ITEM].default_create */
extern T0* T1131c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION].make */
extern T0* T1129f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPRESSION].make_empty */
extern T0* T1130c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPRESSION].default_create */
extern T0* T1129c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT_LIST].make */
extern T0* T1127f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPORT_LIST].make_empty */
extern T0* T1128c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT_LIST].default_create */
extern T0* T1127c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT].make */
extern T0* T1124f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_EXPORT].make_empty */
extern T0* T1126c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_EXPORT].default_create */
extern T0* T1124c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART_LIST].make */
extern T0* T1121f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART_LIST].make_empty */
extern T0* T1123c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART_LIST].default_create */
extern T0* T1121c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART].make */
extern T0* T1118f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_PART].make_empty */
extern T0* T1120c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_PART].default_create */
extern T0* T1118c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION_LIST].make */
extern T0* T1115f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION_LIST].make_empty */
extern T0* T1117c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION_LIST].default_create */
extern T0* T1115c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION].make */
extern T0* T1112f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ELSEIF_EXPRESSION].make_empty */
extern T0* T1114c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ELSEIF_EXPRESSION].default_create */
extern T0* T1112c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DEBUG_INSTRUCTION].make */
extern T0* T1109f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_DEBUG_INSTRUCTION].make_empty */
extern T0* T1111c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_DEBUG_INSTRUCTION].default_create */
extern T0* T1109c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR_LIST].make */
extern T0* T1106f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATOR_LIST].make_empty */
extern T0* T1108c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR_LIST].default_create */
extern T0* T1106c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR].make */
extern T0* T1103f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATOR].make_empty */
extern T0* T1105c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATOR].default_create */
extern T0* T1103c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATION_REGION].make */
extern T0* T1100f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATION_REGION].make_empty */
extern T0* T1102c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATION_REGION].default_create */
extern T0* T1100c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATE_EXPRESSION].make */
extern T0* T1098f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CREATE_EXPRESSION].make_empty */
extern T0* T1099c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CREATE_EXPRESSION].default_create */
extern T0* T1098c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_LIST].make */
extern T0* T1095f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_LIST].make_empty */
extern T0* T1097c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_LIST].default_create */
extern T0* T1095c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_ITEM].make */
extern T0* T1092f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE_ITEM].make_empty */
extern T0* T1094c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE_ITEM].default_create */
extern T0* T1092c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE].make */
extern T0* T1089f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONVERT_FEATURE].make_empty */
extern T0* T1091c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONVERT_FEATURE].default_create */
extern T0* T1089c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_TYPE].make */
extern T0* T1086f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_TYPE].make_empty */
extern T0* T1088c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_TYPE].default_create */
extern T0* T1086c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_RENAME_LIST].make */
extern T0* T1083f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_RENAME_LIST].make_empty */
extern T0* T1085c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_RENAME_LIST].default_create */
extern T0* T1083c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_CREATOR].make */
extern T0* T1080f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_CREATOR].make_empty */
extern T0* T1082c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_CREATOR].default_create */
extern T0* T1080c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST].make */
extern T0* T1077f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST].make_empty */
extern T0* T1079c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_LIST].default_create */
extern T0* T1077c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].make */
extern T0* T1074f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].make_empty */
extern T0* T1076c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM].default_create */
extern T0* T1074c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTANT].make */
extern T0* T1072f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONSTANT].make_empty */
extern T0* T1073c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONSTANT].default_create */
extern T0* T1072c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONDITIONAL].make */
extern T0* T1069f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CONDITIONAL].make_empty */
extern T0* T1071c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CONDITIONAL].default_create */
extern T0* T1069c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_COMPOUND].make */
extern T0* T1067f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_COMPOUND].make_empty */
extern T0* T1068c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_COMPOUND].default_create */
extern T0* T1067c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENTS].make */
extern T0* T1064f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CLIENTS].make_empty */
extern T0* T1066c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENTS].default_create */
extern T0* T1064c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENT_ITEM].make */
extern T0* T1062f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CLIENT_ITEM].make_empty */
extern T0* T1063c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLIENT_ITEM].default_create */
extern T0* T1062c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLASS].make */
extern T0* T1060f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CLASS].make_empty */
extern T0* T1061c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CLASS].default_create */
extern T0* T1060c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_LIST].make */
extern T0* T1057f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE_LIST].make_empty */
extern T0* T1059c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_LIST].default_create */
extern T0* T1057c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_ITEM].make */
extern T0* T1054f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE_ITEM].make_empty */
extern T0* T1056c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_ITEM].default_create */
extern T0* T1054c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_CONSTANT].make */
extern T0* T1051f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE_CONSTANT].make_empty */
extern T0* T1053c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE_CONSTANT].default_create */
extern T0* T1051c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE].make */
extern T0* T1048f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHOICE].make_empty */
extern T0* T1050c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHOICE].default_create */
extern T0* T1048c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHECK_INSTRUCTION].make */
extern T0* T1045f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHECK_INSTRUCTION].make_empty */
extern T0* T1047c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHECK_INSTRUCTION].default_create */
extern T0* T1045c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_CALL_EXPRESSION].make */
extern T0* T1043f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUALIFIED_CALL_EXPRESSION].make_empty */
extern T0* T1044c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUALIFIED_CALL_EXPRESSION].default_create */
extern T0* T1043c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CALL_AGENT].make */
extern T0* T1040f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CALL_AGENT].make_empty */
extern T0* T1042c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CALL_AGENT].default_create */
extern T0* T1040c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_EXPRESSION].make */
extern T0* T1037f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BRACKET_EXPRESSION].make_empty */
extern T0* T1039c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_EXPRESSION].default_create */
extern T0* T1037c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ASSIGNER].make */
extern T0* T1034f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ASSIGNER].make_empty */
extern T0* T1036c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ASSIGNER].default_create */
extern T0* T1034c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ALIAS_NAME_LIST].make */
extern T0* T1032f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ALIAS_NAME_LIST].make_empty */
extern T0* T1033c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ALIAS_NAME_LIST].default_create */
extern T0* T1032c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ALIAS_NAME].make */
extern T0* T1029f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ALIAS_NAME].make_empty */
extern T0* T1031c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ALIAS_NAME].default_create */
extern T0* T1029c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_TARGET].make */
extern T0* T1027f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_TARGET].make_empty */
extern T0* T1028c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_TARGET].default_create */
extern T0* T1027c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_LIST].make */
extern T0* T1024f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_LIST].make_empty */
extern T0* T1026c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_LIST].default_create */
extern T0* T1024c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM].make */
extern T0* T1021f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM].make_empty */
extern T0* T1023c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND_ITEM].default_create */
extern T0* T1021c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND].make */
extern T0* T1019f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_ARGUMENT_OPERAND].make_empty */
extern T0* T1020c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_ARGUMENT_OPERAND].default_create */
extern T0* T1019c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_LIST].make */
extern T0* T1017f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_LIST].make_empty */
extern T0* T1018c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_LIST].default_create */
extern T0* T1017c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_ITEM].make */
extern T0* T1015f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACTUAL_PARAMETER_ITEM].make_empty */
extern T0* T1016c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_PARAMETER_ITEM].default_create */
extern T0* T1015c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_ARGUMENT_LIST].make */
extern T0* T1013f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACTUAL_ARGUMENT_LIST].make_empty */
extern T0* T1014c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACTUAL_ARGUMENT_LIST].default_create */
extern T0* T1013c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_INSTRUCTION].make */
extern T0* T1010f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACROSS_INSTRUCTION].make_empty */
extern T0* T1012c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_INSTRUCTION].default_create */
extern T0* T1010c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_EXPRESSION].make */
extern T0* T1007f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ACROSS_EXPRESSION].make_empty */
extern T0* T1009c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ACROSS_EXPRESSION].default_create */
extern T0* T1007c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUESTION_MARK_SYMBOL].make */
extern T0* T1004f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_QUESTION_MARK_SYMBOL].make_empty */
extern T0* T1006c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_QUESTION_MARK_SYMBOL].default_create */
extern T0* T1004c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_SYMBOL].make */
extern T0* T1002f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BRACKET_SYMBOL].make_empty */
extern T0* T1003c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BRACKET_SYMBOL].default_create */
extern T0* T1002c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SEMICOLON_SYMBOL].make */
extern T0* T999f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_SEMICOLON_SYMBOL].make_empty */
extern T0* T1001c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SEMICOLON_SYMBOL].default_create */
extern T0* T999c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_VOID].make */
extern T0* T996f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_VOID].make_empty */
extern T0* T998c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_VOID].default_create */
extern T0* T996c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL_OPERATOR].make */
extern T0* T993f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_SYMBOL_OPERATOR].make_empty */
extern T0* T995c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL_OPERATOR].default_create */
extern T0* T993c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RETRY_INSTRUCTION].make */
extern T0* T990f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RETRY_INSTRUCTION].make_empty */
extern T0* T992c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RETRY_INSTRUCTION].default_create */
extern T0* T990c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RESULT].make */
extern T0* T988f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_RESULT].make_empty */
extern T0* T989c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_RESULT].default_create */
extern T0* T988c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_REAL_CONSTANT].make */
extern T0* T985f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_REAL_CONSTANT].make_empty */
extern T0* T987c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_REAL_CONSTANT].default_create */
extern T0* T985c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING].make */
extern T0* T983f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_MANIFEST_STRING].make_empty */
extern T0* T984c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_MANIFEST_STRING].default_create */
extern T0* T983c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_OPERATOR].make */
extern T0* T980f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_KEYWORD_OPERATOR].make_empty */
extern T0* T982c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD_OPERATOR].default_create */
extern T0* T980c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INTEGER_CONSTANT].make */
extern T0* T977f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_INTEGER_CONSTANT].make_empty */
extern T0* T979c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_INTEGER_CONSTANT].default_create */
extern T0* T977c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IDENTIFIER].make */
extern T0* T975f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_IDENTIFIER].make_empty */
extern T0* T976c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_IDENTIFIER].default_create */
extern T0* T975c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FREE_OPERATOR].make */
extern T0* T972f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_FREE_OPERATOR].make_empty */
extern T0* T974c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_FREE_OPERATOR].default_create */
extern T0* T972c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CURRENT].make */
extern T0* T970f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CURRENT].make_empty */
extern T0* T971c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CURRENT].default_create */
extern T0* T970c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHARACTER_CONSTANT].make */
extern T0* T967f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_CHARACTER_CONSTANT].make_empty */
extern T0* T969c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_CHARACTER_CONSTANT].default_create */
extern T0* T967c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BREAK].make */
extern T0* T964f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BREAK].make_empty */
extern T0* T966c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BREAK].default_create */
extern T0* T964c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BOOLEAN_CONSTANT].make */
extern T0* T961f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_BOOLEAN_CONSTANT].make_empty */
extern T0* T963c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_BOOLEAN_CONSTANT].default_create */
extern T0* T961c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSITION].make */
extern T0* T959f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_POSITION].make_empty */
extern T0* T960c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_POSITION].default_create */
extern T0* T959c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL].make */
extern T0* T957f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_SYMBOL].make_empty */
extern T0* T958c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_SYMBOL].default_create */
extern T0* T957c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECURSOR_KEYWORD].make */
extern T0* T954f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_PRECURSOR_KEYWORD].make_empty */
extern T0* T956c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_PRECURSOR_KEYWORD].default_create */
extern T0* T954c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_KEYWORD].make */
extern T0* T951f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_AGENT_KEYWORD].make_empty */
extern T0* T953c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_AGENT_KEYWORD].default_create */
extern T0* T951c3(GE_context* ac);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD].make */
extern T0* T949f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_KEYWORD].make_empty */
extern T0* T950c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_KEYWORD].default_create */
extern T0* T949c3(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.make_eiffel_scanner */
extern void T456f1041(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PARSER.make_with_buffer */
extern void T456f1043(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_initialize */
extern void T456f1046(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_initialize */
extern void T456f1046p1(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_build_tables */
extern void T456f1048(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_acclist_template */
extern T0* T456f725(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_fixed_array */
extern T0* T456f730(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_acclist_template_5 */
extern void T456f1218(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_array_subcopy */
extern void T456f1233(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] ET_EIFFEL_PARSER.yy_acclist_template_4 */
extern void T456f1217(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_acclist_template_3 */
extern void T456f1216(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_acclist_template_2 */
extern void T456f1215(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_acclist_template_1 */
extern void T456f1214(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_accept_template */
extern T0* T456f723(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_accept_template_3 */
extern void T456f1213(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_accept_template_2 */
extern void T456f1212(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_accept_template_1 */
extern void T456f1211(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_meta_template */
extern T0* T456f721(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_ec_template */
extern T0* T456f719(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_ec_template_6 */
extern void T456f1205(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_ec_template_5 */
extern void T456f1204(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_ec_template_4 */
extern void T456f1203(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_ec_template_3 */
extern void T456f1202(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_ec_template_2 */
extern void T456f1201(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_ec_template_1 */
extern void T456f1200(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_def_template */
extern T0* T456f696(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_def_template_3 */
extern void T456f1197(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_def_template_2 */
extern void T456f1196(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_def_template_1 */
extern void T456f1195(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_base_template */
extern T0* T456f694(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_base_template_3 */
extern void T456f1194(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_base_template_2 */
extern void T456f1193(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_base_template_1 */
extern void T456f1192(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template */
extern T0* T456f692(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_17 */
extern void T456f1191(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_16 */
extern void T456f1190(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_15 */
extern void T456f1189(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_14 */
extern void T456f1188(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_13 */
extern void T456f1187(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_12 */
extern void T456f1186(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_11 */
extern void T456f1185(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_10 */
extern void T456f1184(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_9 */
extern void T456f1183(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_8 */
extern void T456f1182(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_7 */
extern void T456f1181(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_6 */
extern void T456f1180(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_5 */
extern void T456f1179(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_4 */
extern void T456f1178(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_3 */
extern void T456f1177(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_2 */
extern void T456f1176(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_chk_template_1 */
extern void T456f1175(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template */
extern T0* T456f690(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_14 */
extern void T456f1174(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_13 */
extern void T456f1173(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_12 */
extern void T456f1172(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_11 */
extern void T456f1171(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_10 */
extern void T456f1170(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_9 */
extern void T456f1169(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_8 */
extern void T456f1168(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_7 */
extern void T456f1167(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_6 */
extern void T456f1166(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_5 */
extern void T456f1165(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_4 */
extern void T456f1164(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_3 */
extern void T456f1163(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_2 */
extern void T456f1162(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.yy_nxt_template_1 */
extern void T456f1161(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PARSER.empty_buffer */
extern T0* T456f687(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE].make */
extern T0* T947c10(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE].set_internal_cursor */
extern void T947f13(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [detachable ET_CONSTRAINT_TYPE].new_cursor */
extern T0* T947f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [detachable ET_CONSTRAINT_TYPE].make */
extern T0* T1650c5(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].make */
extern T0* T946c11(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].set_internal_cursor */
extern void T946f16(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_ASSERTION_ITEM].new_cursor */
extern T0* T946f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_ASSERTION_ITEM].make */
extern T0* T1649c5(GE_context* ac, T0* a1);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_ITERATION_COMPONENT_LIST].make */
extern T0* T945c9(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ITERATION_COMPONENT_LIST].make */
extern T0* T1645f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_ITERATION_COMPONENT_LIST].make_empty */
extern T0* T1646c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_ITERATION_COMPONENT_LIST].default_create */
extern T0* T1645c3(GE_context* ac);
/* [detachable] DS_ARRAYED_STACK [detachable ET_ITERATION_COMPONENT_LIST].make */
extern T0* T944c10(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ITERATION_COMPONENT_LIST].make */
extern T0* T1644f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_ITERATION_COMPONENT_LIST].make_empty */
extern T0* T1643c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_ITERATION_COMPONENT_LIST].default_create */
extern T0* T1644c3(GE_context* ac);
/* [detachable] DS_ARRAYED_STACK [[attached] ET_OBJECT_TEST_LIST].make */
extern T0* T943c9(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST_LIST].make */
extern T0* T1641f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_OBJECT_TEST_LIST].make_empty */
extern T0* T1642c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_OBJECT_TEST_LIST].default_create */
extern T0* T1641c3(GE_context* ac);
/* [detachable] DS_ARRAYED_STACK [detachable ET_OBJECT_TEST_LIST].make */
extern T0* T942c10(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBJECT_TEST_LIST].make */
extern T0* T1640f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable ET_OBJECT_TEST_LIST].make_empty */
extern T0* T1639c5(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [detachable ET_OBJECT_TEST_LIST].default_create */
extern T0* T1640c3(GE_context* ac);
/* [detachable] DS_ARRAYED_STACK [detachable ET_SYMBOL].make */
extern T0* T941c8(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_KEYWORD].make */
extern T0* T940c8(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_LOCAL_VARIABLE_LIST].make */
extern T0* T939c9(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_STACK [detachable ET_FORMAL_ARGUMENT_LIST].make */
extern T0* T938c9(GE_context* ac, T6 a1);
/* [detachable] YY_UNICODE_FILE_BUFFER.make_with_size */
extern T0* T936c22(GE_context* ac, T0* a1, T6 a2);
/* [detachable] YY_UNICODE_FILE_BUFFER.new_default_buffer */
extern T0* T936f21(GE_context* ac, T0* C, T6 a1);
/* [detachable] KL_STANDARD_FILES.input */
extern T0* T133f3(GE_context* ac, T0* C);
/* [detachable] KL_STDIN_FILE.make */
extern T0* T931c12(GE_context* ac);
/* [detachable] ET_EIFFEL_PARSER.std */
extern T0* T456f739(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PARSER.make */
extern void T456f1040p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_MASTER_CLASS_CHECKER.make */
extern T0* T455c7(GE_context* ac, T0* a1);
/* [detachable] ET_MASTER_CLASS_CHECKER.tokens */
extern T0* T455f4(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.make */
extern T0* T454c136(GE_context* ac, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.make_eiffel_scanner */
extern void T454f137(GE_context* ac, T0* C, T0* a1, T0* a2);
/* [detachable] ET_EIFFEL_PREPARSER.make_with_buffer */
extern void T454f138(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_initialize */
extern void T454f139(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_initialize */
extern void T454f139p1(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_build_tables */
extern void T454f141(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_acclist_template */
extern T0* T454f87(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_fixed_array */
extern T0* T454f96(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_acclist_template_4 */
extern void T454f173(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_array_subcopy */
extern void T454f183(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3, T6 a4, T6 a5);
/* [detachable] ET_EIFFEL_PREPARSER.integer_array_ */
extern T0* T454f102(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_acclist_template_3 */
extern void T454f172(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_acclist_template_2 */
extern void T454f171(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_acclist_template_1 */
extern void T454f170(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_accept_template */
extern T0* T454f85(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_accept_template_3 */
extern void T454f169(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_accept_template_2 */
extern void T454f168(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_accept_template_1 */
extern void T454f167(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_meta_template */
extern T0* T454f80(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_ec_template */
extern T0* T454f74(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_ec_template_6 */
extern void T454f166(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_ec_template_5 */
extern void T454f165(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_ec_template_4 */
extern void T454f164(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_ec_template_3 */
extern void T454f163(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_ec_template_2 */
extern void T454f162(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_ec_template_1 */
extern void T454f161(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_def_template */
extern T0* T454f71(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_def_template_6 */
extern void T454f160(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_def_template_5 */
extern void T454f159(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_def_template_4 */
extern void T454f158(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_def_template_3 */
extern void T454f157(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_def_template_2 */
extern void T454f156(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_def_template_1 */
extern void T454f155(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_base_template */
extern T0* T454f63(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_base_template_3 */
extern void T454f154(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_base_template_2 */
extern void T454f153(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_base_template_1 */
extern void T454f152(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_chk_template */
extern T0* T454f59(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_chk_template_9 */
extern void T454f151(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_chk_template_8 */
extern void T454f150(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_chk_template_7 */
extern void T454f149(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_chk_template_6 */
extern void T454f148(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_chk_template_5 */
extern void T454f147(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_chk_template_4 */
extern void T454f146(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_chk_template_3 */
extern void T454f145(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_chk_template_2 */
extern void T454f144(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_chk_template_1 */
extern void T454f143(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_nxt_template */
extern T0* T454f94(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.yy_nxt_template_9 */
extern void T454f182(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_nxt_template_8 */
extern void T454f181(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_nxt_template_7 */
extern void T454f180(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_nxt_template_6 */
extern void T454f179(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_nxt_template_5 */
extern void T454f178(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_nxt_template_4 */
extern void T454f177(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_nxt_template_3 */
extern void T454f176(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_nxt_template_2 */
extern void T454f175(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.yy_nxt_template_1 */
extern void T454f174(GE_context* ac, T0* C, T0* a1);
/* [detachable] ET_EIFFEL_PREPARSER.empty_buffer */
extern T0* T454f83(GE_context* ac, T0* C);
/* [detachable] ET_EIFFEL_PREPARSER.std */
extern T0* T454f95(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.make */
extern T0* T375c49(GE_context* ac, T6 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.make_single */
extern void T375f61(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.make_null_single */
extern void T375f73(GE_context* ac, T0* C);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_qualified_anchored_types_cycle_detection_enabled_only */
extern void T375f86(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_unknown_builtin_reported_only */
extern void T375f85(GE_context* ac, T0* C, T1 a1);
/* [detachable] ET_SYSTEM_MULTIPROCESSOR.set_preparse_shallow_mode_only */
extern void T375f84(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_SYSTEM_PROCESSOR].put_last */
extern void T703f9(GE_context* ac, T0* C, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_SYSTEM_PROCESSOR].force */
extern void T1419f3(GE_context* ac, T0* C, T0* a1, T0* a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_SYSTEM_PROCESSOR].fill_with */
extern void T1418f6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] ET_SYSTEM_PROCESSOR].extend */
extern void T1418f7(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_SYSTEM_PROCESSOR].make */
extern T0* T703c8(GE_context* ac, T6 a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_SYSTEM_PROCESSOR].set_internal_cursor */
extern void T703f10(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_ARRAYED_LIST [[attached] ET_SYSTEM_PROCESSOR].new_cursor */
extern T0* T703f6(GE_context* ac, T0* C);
/* [detachable] DS_ARRAYED_LIST_CURSOR [[attached] ET_SYSTEM_PROCESSOR].make */
extern T0* T1420c3(GE_context* ac, T0* a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_SYSTEM_PROCESSOR].make */
extern T0* T1419f1(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] ET_SYSTEM_PROCESSOR].make_empty */
extern T0* T1418c4(GE_context* ac, T6 a1);
/* [detachable] KL_SPECIAL_ROUTINES [[attached] ET_SYSTEM_PROCESSOR].default_create */
extern T0* T1419c2(GE_context* ac);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].remove */
extern void T204f65(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].remove_position */
extern void T204f67(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].key_storage_keep_head */
extern void T204f77(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].item_storage_keep_head */
extern void T204f76(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].keep_head */
extern void T593f10(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].set_count */
extern void T593f12(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].move_cursors_forth */
extern void T204f75(GE_context* ac, T0* C, T6 a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].internal_set_key_equality_tester */
extern void T204f74(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].twin */
extern T0* T204f36(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].copy */
extern void T204f66(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].copy */
extern void T204f66p1(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].clone_clashes */
extern void T204f73(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].clone_slots */
extern void T204f72(GE_context* ac, T0* C);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].clone_key_storage */
extern void T204f71(GE_context* ac, T0* C);
/* [detachable] SPECIAL [[attached] STRING_8].resized_area */
extern T0* T34f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [[attached] STRING_8].copy_data */
extern void T34f13(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] SPECIAL [[attached] STRING_8].force */
extern void T34f15(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [[attached] STRING_8].move_data */
extern void T34f14(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] STRING_8].overlapping_move */
extern void T34f17(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [[attached] STRING_8].non_overlapping_move */
extern void T34f16(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].clone_item_storage */
extern void T204f70(GE_context* ac, T0* C);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].resized_area */
extern T0* T593f5(GE_context* ac, T0* C, T6 a1);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].copy_data */
extern void T593f11(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3, T6 a4);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].force */
extern void T593f14(GE_context* ac, T0* C, T0* a1, T6 a2);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].move_data */
extern void T593f13(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].overlapping_move */
extern void T593f16(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] SPECIAL [detachable RX_REGULAR_EXPRESSION].non_overlapping_move */
extern void T593f15(GE_context* ac, T0* C, T6 a1, T6 a2, T6 a3);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].valid_cursor */
extern T1 T204f37(GE_context* ac, T0* C, T0* a1);
/* [detachable] DS_HASH_TABLE [detachable RX_REGULAR_EXPRESSION, [attached] STRING_8].move_all_cursors_after */
extern void T204f68(GE_context* ac, T0* C);
/* Call to [detachable] STRING_8.to_integer */
extern T6 T17x2228289(GE_context* ac, T0* C);
/* Call to [detachable] STRING_8.is_integer */
extern T1 T17x2228255(GE_context* ac, T0* C);
/* Call to [detachable] STRING_8.count */
extern T6 T17x2228263(GE_context* ac, T0* C);
/* Call to [detachable] STRING_8.occurrences */
extern T6 T17x2211855T2(GE_context* ac, T0* C, T2 a1);
/* Call to [detachable] STRING_8.item */
extern T2 T17x2572289T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] STRING_8.is_empty */
extern T1 T17x2228241(GE_context* ac, T0* C);
/* Call to [detachable] STRING_8.hash_code */
extern T6 T17x655361(GE_context* ac, T0* C);
/* Call to [detachable] STRING_8.as_lower */
extern T0* T17x2228285(GE_context* ac, T0* C);
/* Call to [detachable] STRING_8.twin */
extern T0* T17x229387(GE_context* ac, T0* C);
/* Call to [detachable] STRING_8.has */
extern T1 T17x2211867T2(GE_context* ac, T0* C, T2 a1);
/* Call to [detachable] STRING_8.substring */
extern T0* T17x2228304T6T6(GE_context* ac, T0* C, T6 a1, T6 a2);
/* Call to [detachable] STRING_8.ends_with */
extern T1 T17x2211869T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] STRING_8.item_code */
extern T6 T17x2211843T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] STRING_8.alias "+" */
extern T0* T17x2211872T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] STRING_8.starts_with */
extern T1 T17x2211868T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] STRING_8.same_string */
extern T1 T17x2211860T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] STRING_8.has_substring */
extern T1 T17x2228268T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] STRING_8.to_boolean */
extern T1 T17x2228301(GE_context* ac, T0* C);
/* Call to [detachable] STRING_8.is_boolean */
extern T1 T17x2228252(GE_context* ac, T0* C);
/* Call to [detachable] STRING_8.code */
extern T10 T17x2228225T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] STRING_8.as_upper */
extern T0* T17x2228286(GE_context* ac, T0* C);
/* Call to [detachable] STRING_8.same_type */
extern T1 T17x229380T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] STRING_8.index_of */
extern T6 T17x2211845T2T6(GE_context* ac, T0* C, T2 a1, T6 a2);
/* Call to [detachable] ANY.same_type */
extern T1 T23x229380T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] KI_CHARACTER_INPUT_STREAM.read_to_string */
extern T6 T41x6029313T0T6T6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* Call to [detachable] KI_CHARACTER_INPUT_STREAM.end_of_input */
extern T1 T41x6144005(GE_context* ac, T0* C);
/* Call to [detachable] KI_CHARACTER_INPUT_STREAM.last_character */
extern T2 T41x6144008(GE_context* ac, T0* C);
/* Call to [detachable] KI_CHARACTER_INPUT_STREAM.name */
extern T0* T41x6144007(GE_context* ac, T0* C);
/* Call to [detachable] KI_CHARACTER_INPUT_STREAM.is_closable */
extern T1 T41x6144002(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.short_form */
extern T2 T51x5537800(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.long_form */
extern T0* T51x5537795(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.description */
extern T0* T51x5537793(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.names */
extern T0* T51x5537797(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.is_hidden */
extern T1 T51x5537803(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.name */
extern T0* T51x5537796(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.needs_parameter */
extern T1 T51x5537806(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.has_short_form */
extern T1 T51x5537802(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.allows_parameter */
extern T1 T51x5537805(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.maximum_occurrences */
extern T6 T51x5537799(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.occurrences */
extern T6 T51x5537798(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.was_found */
extern T1 T51x5537809(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.is_mandatory */
extern T1 T51x5537804(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.example */
extern T0* T51x5537794(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.conforms_to_type_with_type_marks */
extern T1 T66x15319046T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to [detachable] ET_BASE_TYPE.base_class */
extern T0* T66x22429699(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.index_of_label */
extern T6 T66x21856265T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_BASE_TYPE.base_type_index_of_label */
extern T6 T66x22396941T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.same_base_type_with_type_marks */
extern T1 T66x22396979T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_BASE_TYPE.to_text */
extern T0* T66x22397004(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.type_with_type_mark */
extern T0* T66x22396949T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_BASE_TYPE.actual_parameters */
extern T0* T66x21856260(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.adapted_base_class_with_seeded_feature */
extern T0* T66x22396934T6T0(GE_context* ac, T0* C, T6 a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.adapted_base_class_with_named_feature */
extern T0* T66x22396933T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.is_type_attached_with_type_mark */
extern T1 T66x22396960T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.named_base_class */
extern T0* T66x22429700(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.base_type_actual */
extern T0* T66x22396939T6T0(GE_context* ac, T0* C, T6 a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.actual_parameter_count */
extern T6 T66x21856268(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.shallow_named_type_with_type_mark */
extern T0* T66x22396945T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.is_type_reference_with_type_mark */
extern T1 T66x22396958T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.context_adapted_base_class_with_seeded_feature */
extern T0* T66x22429702T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_BASE_TYPE.context_adapted_base_class_with_named_feature */
extern T0* T66x22429701T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_BASE_TYPE.is_expanded */
extern T1 T66x21856272(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.is_attached */
extern T1 T66x21856273(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.is_type_detachable_with_type_mark */
extern T1 T66x22396962T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.is_type_expanded_with_type_mark */
extern T1 T66x22396956T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.base_type_with_type_mark */
extern T0* T66x22396936T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T66x22396995T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to [detachable] ET_BASE_TYPE.conforms_from_tuple_type_with_type_marks */
extern T1 T66x22396996T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to [detachable] ET_BASE_TYPE.conforms_from_class_type_with_type_marks */
extern T1 T66x22396994T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to [detachable] ET_BASE_TYPE.shallow_base_type_with_type_mark */
extern T0* T66x22396938T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.same_named_type */
extern T1 T66x22396976T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to [detachable] ET_BASE_TYPE.same_named_type_with_type_marks */
extern T1 T66x22396977T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_BASE_TYPE.unaliased_to_text */
extern T0* T66x22397005(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.same_named_formal_parameter_type_with_type_marks */
extern T1 T66x22396987T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_BASE_TYPE.same_named_tuple_type_with_type_marks */
extern T1 T66x22396988T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_BASE_TYPE.same_named_class_type_with_type_marks */
extern T1 T66x22396986T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_BASE_TYPE.base_type_actual_count */
extern T6 T66x22396951T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_BASE_TYPE.named_type_with_type_mark */
extern T0* T66x22396943T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.upper_name */
extern T0* T66x22282242(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.name */
extern T0* T66x22282241(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.same_as_base_class */
extern T1 T66x21856280(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.runtime_name_to_text */
extern T0* T66x22397006(GE_context* ac, T0* C);
/* Call to [detachable] UT_ERROR.default_message */
extern T0* T78x11829251(GE_context* ac, T0* C);
/* Call to [detachable] READABLE_STRING_GENERAL.item */
extern T3 T83x2228226T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] READABLE_STRING_GENERAL.count */
extern T6 T83x2228263(GE_context* ac, T0* C);
/* Call to [detachable] READABLE_STRING_GENERAL.code */
extern T10 T83x2228225T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] READABLE_STRING_GENERAL.is_empty */
extern T1 T83x2228241(GE_context* ac, T0* C);
/* Call to [detachable] READABLE_STRING_GENERAL.substring */
extern T0* T83x2228304T6T6(GE_context* ac, T0* C, T6 a1, T6 a2);
/* Call to [detachable] READABLE_STRING_GENERAL.as_string_8 */
extern T0* T83x2228280(GE_context* ac, T0* C);
/* Call to [detachable] READABLE_STRING_GENERAL.as_string_32 */
extern T0* T83x2228283(GE_context* ac, T0* C);
/* Call to [detachable] READABLE_STRING_GENERAL.to_string_32 */
extern T0* T83x2228284(GE_context* ac, T0* C);
/* Call to [detachable] READABLE_STRING_GENERAL.substring_index */
extern T6 T83x2228274T0T6(GE_context* ac, T0* C, T0* a1, T6 a2);
/* Call to [detachable] READABLE_STRING_GENERAL.same_string */
extern T1 T83x2228269T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] READABLE_STRING_8.area */
extern T0* T92x2211878(GE_context* ac, T0* C);
/* Call to [detachable] READABLE_STRING_8.count */
extern T6 T92x2228263(GE_context* ac, T0* C);
/* Call to [detachable] READABLE_STRING_8.area_lower */
extern T6 T92x2211879(GE_context* ac, T0* C);
/* Call to [detachable] READABLE_STRING_8.to_string_8 */
extern T0* T92x2228277(GE_context* ac, T0* C);
/* Call to detachable KL_EQUALITY_TESTER [[attached] STRING_8].test */
extern T1 T154x7569409T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.ecf_version */
extern T0* T185x13287428(GE_context* ac, T0* C);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.filename */
extern T0* T185x13287425(GE_context* ac, T0* C);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.universe */
extern T0* T185x13287429(GE_context* ac, T0* C);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.targets */
extern T0* T185x14057475(GE_context* ac, T0* C);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.library_target */
extern T0* T185x14057476(GE_context* ac, T0* C);
/* Call to [detachable] ET_POSITION.is_null */
extern T1 T197x20627459(GE_context* ac, T0* C);
/* Call to [detachable] ET_POSITION.column */
extern T6 T197x20627458(GE_context* ac, T0* C);
/* Call to [detachable] ET_POSITION.line */
extern T6 T197x20627457(GE_context* ac, T0* C);
/* Call to [detachable] ET_POSITION.to_text */
extern T0* T197x20627463(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.conforms_to_primary_type */
extern T1 T223x23805970T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.seeded_dynamic_query */
extern T0* T223x23805986T6T0(GE_context* ac, T0* C, T6 a1, T0* a2);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_basic */
extern T1 T223x24215556(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_expanded */
extern T1 T223x24248321(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.dynamic_procedure */
extern T0* T223x23805985T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.base_type */
extern T0* T223x24215560(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_agent_type */
extern T1 T223x23805954(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.dynamic_query */
extern T0* T223x23805984T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.seeded_dynamic_procedure */
extern T0* T223x23805987T6T0(GE_context* ac, T0* C, T6 a1, T0* a2);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.base_class */
extern T0* T223x24215561(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.id */
extern T6 T223x23805968(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.queries */
extern T0* T223x23805982(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.attribute_count */
extern T6 T223x23805972(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_special */
extern T1 T223x23805955(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_builtin_attribute */
extern T1 T223x23805988T0T8T8(GE_context* ac, T0* C, T0* a1, T8 a2, T8 a3);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_nested_custom_standard_copy_routine */
extern T1 T223x23805979(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_redefined_copy_routine */
extern T1 T223x23805978(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_once_per_object_routines */
extern T1 T223x23805975(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_nested_custom_standard_is_equal_routine */
extern T1 T223x23805981(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.primary_type */
extern T0* T223x24215562(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.hash_code */
extern T6 T223x655361(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_nested_reference_fields */
extern T1 T223x23805977(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_nested_reference_attributes */
extern T1 T223x23805974(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_redefined_is_equal_routine */
extern T1 T223x23805980(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_alive */
extern T1 T223x23805953(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.attached_type */
extern T0* T223x23805966(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.next_type */
extern T0* T223x23805995(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.conforming_ancestors */
extern T0* T223x23805971(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.type_mark */
extern T0* T223x24215564(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.procedure_calls */
extern T0* T223x23805992(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.query_calls */
extern T0* T223x23805991(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.procedures */
extern T0* T223x23805983(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_boolean */
extern T1 T223x23805957(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.meta_type */
extern T0* T223x24215563(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.is_tuple */
extern T1 T223x23805956(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.has_static */
extern T1 T223x23805960(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.hash_code */
extern T6 T234x655361(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.first_seed */
extern T6 T234x18268171(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.is_static */
extern T1 T234x18186278(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.is_exported_to */
extern T1 T234x18186295T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to detachable ET_PROCEDURE.arguments_count */
extern T6 T234x18186294(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.arguments */
extern T0* T234x18268170(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.overloaded_alias_names */
extern T0* T234x18186247(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.alias_names */
extern T0* T234x18186243(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.overloaded_name */
extern T0* T234x18186246(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.name */
extern T0* T234x18268168(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.is_creation_exported_to */
extern T1 T234x18186297T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to detachable ET_PROCEDURE.lower_name */
extern T0* T234x18923523(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.new_synonym */
extern T0* T234x18186301T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_PROCEDURE.has_seed */
extern T1 T234x18268166T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_PROCEDURE.implementation_class */
extern T0* T234x18923525(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.is_parenthesisable */
extern T1 T234x18186285(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.is_frozen */
extern T1 T234x18186267(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.is_deferred */
extern T1 T234x18186268(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.postconditions */
extern T0* T234x17776650(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.preconditions */
extern T0* T234x17776649(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.lower_name */
extern T0* T235x18923523(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.type */
extern T0* T235x18268169(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.first_seed */
extern T6 T235x18268171(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_attribute */
extern T1 T235x18186271(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_stable_attribute */
extern T1 T235x18186281(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.assigner */
extern T0* T235x18726915(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.overloaded_alias_names */
extern T0* T235x18186247(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.alias_names */
extern T0* T235x18186243(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.overloaded_name */
extern T0* T235x18186246(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.name */
extern T0* T235x18268168(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.hash_code */
extern T6 T235x655361(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_exported_to */
extern T1 T235x18186295T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to detachable ET_QUERY.arguments */
extern T0* T235x18268170(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_static */
extern T1 T235x18186278(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.arguments_count */
extern T6 T235x18186294(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.has_seed */
extern T1 T235x18268166T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_QUERY.implementation_class */
extern T0* T235x18923525(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.implementation_feature */
extern T0* T235x17776653(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_unique_attribute */
extern T1 T235x18186273(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_constant_attribute */
extern T1 T235x18186272(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.new_synonym */
extern T0* T235x18186301T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_QUERY.is_prefixable */
extern T1 T235x18186283(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_infixable */
extern T1 T235x18186282(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_once_per_object */
extern T1 T235x17776644(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_once */
extern T1 T235x17776641(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_parenthesisable */
extern T1 T235x18186285(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_bracketable */
extern T1 T235x18186284(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_frozen */
extern T1 T235x18186267(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.is_deferred */
extern T1 T235x18186268(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.postconditions */
extern T0* T235x17776650(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.preconditions */
extern T0* T235x17776649(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE.position */
extern T0* T238x20414465(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE.is_type_detachable */
extern T1 T238x22396961T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.shallow_named_type */
extern T0* T238x22396944T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.is_type_attached */
extern T1 T238x22396959T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.base_type_index_of_label */
extern T6 T238x22396941T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_TYPE.adapted_base_class_with_seeded_feature */
extern T0* T238x22396934T6T0(GE_context* ac, T0* C, T6 a1, T0* a2);
/* Call to [detachable] ET_TYPE.adapted_base_class_with_named_feature */
extern T0* T238x22396933T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_TYPE.same_base_type_with_type_marks */
extern T1 T238x22396979T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.same_named_type_with_type_marks */
extern T1 T238x22396977T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.to_text */
extern T0* T238x22397004(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE.is_base_type */
extern T1 T238x22396954(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE.same_syntactical_like_current_with_type_marks */
extern T1 T238x22396982T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.same_syntactical_tuple_type_with_type_marks */
extern T1 T238x22396985T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.same_syntactical_class_type_with_type_marks */
extern T1 T238x22396980T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.is_like_current */
extern T1 T238x22396952(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE.named_type */
extern T0* T238x22396942T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.conforms_to_type */
extern T1 T238x15319045T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Call to [detachable] ET_TYPE.same_named_type */
extern T1 T238x22396976T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to [detachable] ET_TYPE.is_named_type */
extern T1 T238x15319043(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE.is_type_expanded */
extern T1 T238x22396955T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.same_syntactical_type */
extern T1 T238x22396974T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to [detachable] ET_TYPE.resolved_formal_parameters */
extern T0* T238x22396997T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.has_identifier_anchored_type */
extern T1 T238x22396967(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE.depends_on_qualified_anchored_type */
extern T1 T238x22396968T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.has_formal_types */
extern T1 T238x22396969T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.has_unqualified_anchored_type */
extern T1 T238x22396966(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE.type_with_type_mark */
extern T0* T238x22396949T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.shallow_base_type_with_type_mark */
extern T0* T238x22396938T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_TYPE.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T238x22396981T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T238x22396984T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.same_syntactical_like_feature_with_type_marks */
extern T1 T238x22396983T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.same_syntactical_type_with_type_marks */
extern T1 T238x22396975T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.shallow_named_type_with_type_mark */
extern T0* T238x22396945T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_TYPE.is_type_reference_with_type_mark */
extern T1 T238x22396958T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_TYPE.is_type_attached_with_type_mark */
extern T1 T238x22396960T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_TYPE.is_type_detachable_with_type_mark */
extern T1 T238x22396962T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_TYPE.is_type_expanded_with_type_mark */
extern T1 T238x22396956T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_TYPE.base_type_with_type_mark */
extern T0* T238x22396936T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_TYPE.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T238x22396995T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to [detachable] ET_TYPE.conforms_from_tuple_type_with_type_marks */
extern T1 T238x22396996T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to [detachable] ET_TYPE.conforms_from_class_type_with_type_marks */
extern T1 T238x22396994T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to [detachable] ET_TYPE.named_base_class */
extern T0* T238x22396932T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.conforms_to_type_with_type_marks */
extern T1 T238x15319046T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to [detachable] ET_TYPE.base_type_actual */
extern T0* T238x22396939T6T0(GE_context* ac, T0* C, T6 a1, T0* a2);
/* Call to [detachable] ET_TYPE.base_type_actual_count */
extern T6 T238x22396951T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.named_type_with_type_mark */
extern T0* T238x22396943T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_TYPE.same_named_formal_parameter_type_with_type_marks */
extern T1 T238x22396987T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.same_named_tuple_type_with_type_marks */
extern T1 T238x22396988T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.same_named_class_type_with_type_marks */
extern T1 T238x22396986T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE.base_class */
extern T0* T238x22396931T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.conforms_to_constraint */
extern T1 T238x22396992T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Call to [detachable] ET_TYPE.is_type_reference */
extern T1 T238x22396957T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.shallow_base_type */
extern T0* T238x22396937T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.is_type_self_initializing */
extern T1 T238x22396963T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.type */
extern T0* T238x22446081(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE_CONTEXT.to_nested_type_context */
extern T0* T239x22429747(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE_CONTEXT.as_nested_type_context */
extern T0* T239x22429746(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE_CONTEXT.root_context */
extern T0* T239x22429697(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE_CONTEXT.is_root_context */
extern T1 T239x22429712(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE_CONTEXT.base_type_actual */
extern T0* T239x22429705T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_TYPE_CONTEXT.base_type_actual_count */
extern T6 T239x22429710(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE_CONTEXT.base_class */
extern T0* T239x22429699(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE_CONTEXT.scoop_mode */
extern T1 T239x22429726(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE_CONTEXT.attachment_type_conformance_mode */
extern T1 T239x22429725(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE_CONTEXT.named_type */
extern T0* T239x22429708(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE_CONTEXT.same_base_type_with_type_marks */
extern T1 T239x22429732T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Call to [detachable] ET_TYPE_CONTEXT.is_type_attached_with_type_mark */
extern T1 T239x22429719T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE_CONTEXT.is_type_reference_with_type_mark */
extern T1 T239x22429717T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE_CONTEXT.adapted_base_class_with_seeded_feature */
extern T0* T239x22429702T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_TYPE_CONTEXT.adapted_base_class_with_named_feature */
extern T0* T239x22429701T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE_CONTEXT.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T239x22429744T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE_CONTEXT.same_named_formal_parameter_type_with_type_marks */
extern T1 T239x22429734T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Call to [detachable] ET_TYPE_CONTEXT.named_type_with_type_mark */
extern T0* T239x22429709T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE_CONTEXT.conforms_to_type_with_type_marks */
extern T1 T239x22429740T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE_CONTEXT.same_named_type_with_type_marks */
extern T1 T239x22429728T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Call to [detachable] ET_TYPE_CONTEXT.conforms_from_tuple_type_with_type_marks */
extern T1 T239x22429745T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE_CONTEXT.conforms_from_class_type_with_type_marks */
extern T1 T239x22429743T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to [detachable] ET_TYPE_CONTEXT.same_named_tuple_type_with_type_marks */
extern T1 T239x22429735T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Call to [detachable] ET_TYPE_CONTEXT.same_named_class_type_with_type_marks */
extern T1 T239x22429733T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Call to [detachable] ET_TYPE_CONTEXT.is_type_detachable_with_type_mark */
extern T1 T239x22429721T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE_CONTEXT.is_type_expanded_with_type_mark */
extern T1 T239x22429715T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE_CONTEXT.named_base_class */
extern T0* T239x22429700(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE_CONTEXT.base_type_with_type_mark */
extern T0* T239x22429704T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FORMAL_PARAMETER_TYPE.index */
extern T6 T242x22134787(GE_context* ac, T0* C);
/* Call to [detachable] ET_FORMAL_PARAMETER_TYPE.implementation_class */
extern T0* T242x22134788(GE_context* ac, T0* C);
/* Call to [detachable] ET_FORMAL_PARAMETER_TYPE.is_type_detachable_with_type_mark */
extern T1 T242x22396962T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_FORMAL_PARAMETER_TYPE.is_type_attached_with_type_mark */
extern T1 T242x22396960T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_FORMAL_PARAMETER_TYPE.type_with_type_mark */
extern T0* T242x22396949T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FORMAL_PARAMETER_TYPE.overridden_type_mark */
extern T0* T242x22396930T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FORMAL_PARAMETER_TYPE.name */
extern T0* T242x22282241(GE_context* ac, T0* C);
/* Call to [detachable] ET_FORMAL_PARAMETER_TYPE.type */
extern T0* T242x22446081(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.adapted_universe */
extern T0* T243x19447826T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_UNIVERSE.boolean_type */
extern T0* T243x19447849(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.character_32_type */
extern T0* T243x19447852(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.character_8_type */
extern T0* T243x19447851(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.natural_64_type */
extern T0* T243x19447871(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.natural_32_type */
extern T0* T243x19447870(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.natural_16_type */
extern T0* T243x19447869(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.natural_8_type */
extern T0* T243x19447868(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.integer_64_type */
extern T0* T243x19447864(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.integer_32_type */
extern T0* T243x19447863(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.integer_16_type */
extern T0* T243x19447862(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.integer_8_type */
extern T0* T243x19447861(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.real_64_convert_feature */
extern T0* T243x19447918(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.real_64_type */
extern T0* T243x19447879(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.real_32_convert_feature */
extern T0* T243x19447917(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.real_32_type */
extern T0* T243x19447878(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.natural_64_convert_feature */
extern T0* T243x19447916(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.natural_32_convert_feature */
extern T0* T243x19447915(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.natural_16_convert_feature */
extern T0* T243x19447914(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.natural_8_convert_feature */
extern T0* T243x19447913(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.integer_64_convert_feature */
extern T0* T243x19447912(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.integer_32_convert_feature */
extern T0* T243x19447911(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.integer_16_convert_feature */
extern T0* T243x19447910(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.integer_8_convert_feature */
extern T0* T243x19447909(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.shortest_path */
extern T0* T243x19447827T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_UNIVERSE.full_lower_name */
extern T0* T243x19447831T2(GE_context* ac, T0* C, T2 a1);
/* Call to [detachable] ET_UNIVERSE.kind_name */
extern T0* T243x19447833(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.any_clients */
extern T0* T243x19447844(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.character_32_convert_feature */
extern T0* T243x19447908(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.character_8_convert_feature */
extern T0* T243x19447907(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.string_32_convert_feature */
extern T0* T243x19447920(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.string_32_type */
extern T0* T243x19447888(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.string_8_convert_feature */
extern T0* T243x19447919(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.string_8_type */
extern T0* T243x19447886(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.master_class */
extern T0* T243x19447815T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_UNIVERSE.has_master_class */
extern T1 T243x19447809T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_UNIVERSE.full_name */
extern T0* T243x19447830T2(GE_context* ac, T0* C, T2 a1);
/* Call to [detachable] ET_UNIVERSE.classes_in_group */
extern T0* T243x19447820T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_UNIVERSE.tuple_type */
extern T0* T243x19447892(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.integer_type */
extern T0* T243x19447860(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.iterable_detachable_separate_any_type */
extern T0* T243x19447866(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.function_identity_any_type */
extern T0* T243x19447857(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.predicate_identity_type */
extern T0* T243x19447875(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.tuple_identity_type */
extern T0* T243x19447894(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.procedure_identity_type */
extern T0* T243x19447876(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.pointer_type */
extern T0* T243x19447874(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.typed_pointer_identity_type */
extern T0* T243x19447901(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.type_identity_type */
extern T0* T243x19447900(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.detachable_any_type */
extern T0* T243x19447840(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.implicit_attachment_type_mark */
extern T0* T243x19447903(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.unfolded_empty_tuple_actual_parameters */
extern T0* T243x19447895(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.obsolete_routine_type_mode */
extern T1 T243x19447905(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.immutable_string_32_type */
extern T0* T243x19447859(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.immutable_string_8_type */
extern T0* T243x19447858(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.system_string_type */
extern T0* T243x19447891(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.string_type */
extern T0* T243x19447884(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.hash_code */
extern T6 T243x655361(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.detachable_separate_any_type */
extern T0* T243x19447841(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.double_type */
extern T0* T243x19447853(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.real_type */
extern T0* T243x19447877(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.detachable_string_8_type */
extern T0* T243x19447887(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.any_type */
extern T0* T243x19447839(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.special_any_type */
extern T0* T243x19447881(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.type_detachable_any_type */
extern T0* T243x19447896(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.special_detachable_any_type */
extern T0* T243x19447882(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.exception_manager_type */
extern T0* T243x19447856(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.type_detachable_exception_type */
extern T0* T243x19447898(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.exception_type */
extern T0* T243x19447854(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.detachable_type_exception_type */
extern T0* T243x19447897(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.detachable_exception_type */
extern T0* T243x19447855(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.type_detachable_like_current_type */
extern T0* T243x19447899(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.obsolete_iteration_mode */
extern T1 T243x19447904(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.system_object_parents */
extern T0* T243x19447890(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.use_obsolete_syntax_mode */
extern T1 T243x19447906(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.any_parents */
extern T0* T243x19447843(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.system_object_type */
extern T0* T243x19447889(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.current_system */
extern T0* T243x19447825(GE_context* ac, T0* C);
/* Call to [detachable] ET_UNIVERSE.character_type */
extern T0* T243x19447850(GE_context* ac, T0* C);
/* Call to [detachable] ET_ACTUAL_PARAMETER_ITEM.resolved_formal_parameters_with_type */
extern T0* T246x21807109T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ACTUAL_PARAMETER_ITEM.type */
extern T0* T246x21807106(GE_context* ac, T0* C);
/* Call to [detachable] ET_ACTUAL_PARAMETER_ITEM.actual_parameter */
extern T0* T246x21807105(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_MARK.is_detachable_mark */
extern T1 T248x20840452(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_MARK.is_attached_mark */
extern T1 T248x20840451(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_MARK.position */
extern T0* T248x20414465(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_MARK.is_expandedness_mark */
extern T1 T248x20840461(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_MARK.is_attachment_mark */
extern T1 T248x20840449(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_MARK.is_implicit_mark */
extern T1 T248x20840467(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_MARK.is_separateness_mark */
extern T1 T248x20840457(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_MARK.is_expanded_mark */
extern T1 T248x20840463(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_MARK.overridden_type_mark */
extern T0* T248x20840468T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_TYPE_MARK.is_explicit_attachment_mark */
extern T1 T248x20840450(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_MARK.is_separate_mark */
extern T1 T248x20840459(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_MARK.is_reference_mark */
extern T1 T248x20840464(GE_context* ac, T0* C);
/* Call to detachable ET_ACTUAL_PARAMETERS.last_leaf */
extern T0* T249x20414469(GE_context* ac, T0* C);
/* Call to detachable ET_ACTUAL_PARAMETERS.type */
extern T0* T249x21774339T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_ACTUAL_PARAMETERS.count */
extern T6 T249x21774337(GE_context* ac, T0* C);
/* Call to detachable ET_ACTUAL_PARAMETERS.is_empty */
extern T1 T249x21774343(GE_context* ac, T0* C);
/* Call to detachable ET_ACTUAL_PARAMETERS.same_named_types */
extern T1 T249x21774351T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to detachable ET_ACTUAL_PARAMETERS.index_of_label */
extern T6 T249x21774341T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_ACTUAL_PARAMETERS.actual_parameter */
extern T0* T249x21774338T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_ACTUAL_PARAMETERS.same_syntactical_types */
extern T1 T249x21774350T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to detachable ET_ACTUAL_PARAMETERS.resolved_formal_parameters */
extern T0* T249x21774355T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_ACTUAL_PARAMETERS.has_identifier_anchored_type */
extern T1 T249x21774346(GE_context* ac, T0* C);
/* Call to detachable ET_ACTUAL_PARAMETERS.depends_on_qualified_anchored_type */
extern T1 T249x21774347T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_ACTUAL_PARAMETERS.has_formal_types */
extern T1 T249x21774348T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_ACTUAL_PARAMETERS.has_unqualified_anchored_type */
extern T1 T249x21774345(GE_context* ac, T0* C);
/* Call to detachable ET_ACTUAL_PARAMETERS.conforms_to_types */
extern T1 T249x21774352T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Call to detachable ET_ACTUAL_PARAMETERS.named_types */
extern T0* T249x21774340T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_DYNAMIC_TYPE_SET.static_type */
extern T0* T251x24248325(GE_context* ac, T0* C);
/* Call to detachable ET_DYNAMIC_TYPE_SET.is_never_void */
extern T1 T251x24248322(GE_context* ac, T0* C);
/* Call to detachable ET_DYNAMIC_TYPE_SET.can_be_void */
extern T1 T251x24248323(GE_context* ac, T0* C);
/* Call to detachable ET_DYNAMIC_TYPE_SET.dynamic_type */
extern T0* T251x23822337T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_DYNAMIC_TYPE_SET.count */
extern T6 T251x23822340(GE_context* ac, T0* C);
/* Call to detachable ET_DYNAMIC_TYPE_SET.is_expanded */
extern T1 T251x24248321(GE_context* ac, T0* C);
/* Call to detachable ET_DYNAMIC_TYPE_SET.has_expanded */
extern T1 T251x23822344(GE_context* ac, T0* C);
/* Call to detachable ET_DYNAMIC_TYPE_SET.has_type */
extern T1 T251x23822342T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_DYNAMIC_TYPE_SET.is_empty */
extern T1 T251x23822341(GE_context* ac, T0* C);
/* Call to detachable ET_DYNAMIC_TYPE_SET.special_type */
extern T0* T251x23822338(GE_context* ac, T0* C);
/* Call to detachable ET_DYNAMIC_TYPE_SET.dynamic_types */
extern T0* T251x24248327(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_TYPE.primary_type */
extern T0* T252x24215562(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_TYPE.type_mark */
extern T0* T252x24215564(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_TYPE.is_basic */
extern T1 T252x24215556(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_TYPE.is_expanded */
extern T1 T252x24248321(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_TYPE.hash_code */
extern T6 T252x655361(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_TYPE.conforms_to_type */
extern T1 T252x24215570T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_TYPE.base_class */
extern T0* T252x24215561(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_TYPE.base_type */
extern T0* T252x24215560(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_TYPE.is_self_initializing */
extern T1 T252x24215557(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_TYPE.meta_type */
extern T0* T252x24215563(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_TYPE.is_attached */
extern T1 T252x24215558(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_TYPE.type_id */
extern T6 T252x24215565(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION.last_leaf */
extern T0* T281x20414469(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION.position */
extern T0* T281x20414465(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION.is_instance_free */
extern T1 T281x16629764(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION.is_false */
extern T1 T281x16957446(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION.manifest_constant_convert_feature */
extern T0* T281x16957450T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to [detachable] ET_EXPRESSION.index */
extern T6 T281x17383428(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION.is_instance_free */
extern T1 T281x17383426(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION.unparenthesized_expression */
extern T0* T281x16957444(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION.is_current */
extern T1 T281x17383427(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION.is_never_void */
extern T1 T281x16957445(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.lower_name */
extern T0* T285x18923523(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.name */
extern T0* T285x18268168(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.implementation_feature */
extern T0* T285x17776653(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.implementation_class */
extern T0* T285x18923525(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.arguments */
extern T0* T285x18268170(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.has_seed */
extern T1 T285x18268166T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_FEATURE.first_seed */
extern T6 T285x18268171(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_deferred */
extern T1 T285x18186268(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_static */
extern T1 T285x18186278(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.clients */
extern T0* T285x18923527(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_exported_to */
extern T1 T285x18186295T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_FEATURE.extended_name */
extern T0* T285x18186244(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.arguments_count */
extern T6 T285x18186294(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.version */
extern T6 T285x18186252(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.are_postconditions_instance_free_recursive */
extern T1 T285x18186287(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.are_preconditions_instance_free_recursive */
extern T1 T285x18186286(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.alias_names */
extern T0* T285x18186243(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_procedure */
extern T1 T285x18186275(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.other_seeds */
extern T0* T285x18268172(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.type */
extern T0* T285x18268169(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.has_class_postcondition_recursive */
extern T1 T285x18186288(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_unique_attribute */
extern T1 T285x18186273(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_constant_attribute */
extern T1 T285x18186272(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_once */
extern T1 T285x17776641(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_attribute */
extern T1 T285x18186271(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_once_per_thread */
extern T1 T285x17776643(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_once_per_process */
extern T1 T285x17776642(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_once_per_object */
extern T1 T285x17776644(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_frozen */
extern T1 T285x18186267(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_infixable */
extern T1 T285x18186282(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_prefixable */
extern T1 T285x18186283(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_parenthesisable */
extern T1 T285x18186285(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_bracketable */
extern T1 T285x18186284(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.id */
extern T6 T285x18186251(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.renamed_feature */
extern T0* T285x18186302T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FEATURE.undefined_feature */
extern T0* T285x18186303T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FEATURE.postconditions */
extern T0* T285x17776650(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.preconditions */
extern T0* T285x17776649(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.same_version */
extern T1 T285x18186300T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FEATURE.is_dotnet */
extern T1 T285x18268165(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.is_query */
extern T1 T285x18186274(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.other_precursors */
extern T0* T285x18186254(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.first_precursor */
extern T0* T285x18923524(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.has_validity_error */
extern T1 T285x18923529(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.validity_checked */
extern T1 T285x18923528(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.hash_code */
extern T6 T285x655361(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT.target */
extern T0* T310x14417922(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT.arguments */
extern T0* T310x14417923(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT.implicit_target_position */
extern T0* T310x14417930(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT.position */
extern T0* T310x20414465(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT.is_inline_agent */
extern T1 T310x14417928(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT.index */
extern T6 T310x17383428(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT.is_qualified_call */
extern T1 T310x14417925(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT.implicit_result */
extern T0* T310x14417924(GE_context* ac, T0* C);
/* Call to [detachable] ET_OBJECT_TEST.expression */
extern T0* T311x17334276(GE_context* ac, T0* C);
/* Call to [detachable] ET_OBJECT_TEST.index */
extern T6 T311x17383428(GE_context* ac, T0* C);
/* Call to [detachable] ET_OBJECT_TEST.type */
extern T0* T311x17334275(GE_context* ac, T0* C);
/* Call to [detachable] ET_OBJECT_TEST.name */
extern T0* T311x17334277(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_COMPONENT.name */
extern T0* T315x16678914(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_COMPONENT.arguments */
extern T0* T315x16678915(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_COMPONENT.position */
extern T0* T315x20414465(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_COMPONENT.is_tuple_label */
extern T1 T315x16678917(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_COMPONENT.is_call_agent */
extern T1 T315x16678918(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_tuple_label */
extern T1 T316x21004297(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.seed */
extern T6 T316x21004291(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.lower_name */
extern T0* T316x21004290(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.position */
extern T0* T316x20414465(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_or_else_symbol */
extern T1 T316x21004318(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_or_else */
extern T1 T316x21004317(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_implies_symbol */
extern T1 T316x21004310(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_implies */
extern T1 T316x21004309(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_and_then_symbol */
extern T1 T316x21004303(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_and_then */
extern T1 T316x21004302(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.precursor_keyword */
extern T0* T316x21004336(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_precursor */
extern T1 T316x21004334(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_prefix_not_symbol */
extern T1 T316x21004330(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_prefix_not */
extern T1 T316x21004329(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_prefix_plus */
extern T1 T316x21004327(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_prefix_minus */
extern T1 T316x21004326(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_dotdot */
extern T1 T316x21004324(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_xor_symbol */
extern T1 T316x21004323(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_xor */
extern T1 T316x21004322(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_times */
extern T1 T316x21004321(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_power */
extern T1 T316x21004320(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_plus */
extern T1 T316x21004319(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_or_symbol */
extern T1 T316x21004316(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_or */
extern T1 T316x21004315(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_mod */
extern T1 T316x21004314(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_minus */
extern T1 T316x21004313(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_lt */
extern T1 T316x21004312(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_le */
extern T1 T316x21004311(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_gt */
extern T1 T316x21004308(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_ge */
extern T1 T316x21004307(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_divide */
extern T1 T316x21004305(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_div */
extern T1 T316x21004304(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_and_symbol */
extern T1 T316x21004301(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_infix_and */
extern T1 T316x21004300(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_parenthesis */
extern T1 T316x21004332(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.is_bracket */
extern T1 T316x21004331(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_NAME.hash_code */
extern T6 T316x655361(GE_context* ac, T0* C);
/* Call to [detachable] ET_CONSTANT.is_integer_constant */
extern T1 T321x16252931(GE_context* ac, T0* C);
/* Call to [detachable] ET_CONSTANT.is_character_constant */
extern T1 T321x16252930(GE_context* ac, T0* C);
/* Call to [detachable] ET_CONSTANT.is_boolean_constant */
extern T1 T321x16252929(GE_context* ac, T0* C);
/* Call to [detachable] ET_CONSTANT.index */
extern T6 T321x17383428(GE_context* ac, T0* C);
/* Call to [detachable] ET_TARGET_TYPE.position */
extern T0* T330x20414465(GE_context* ac, T0* C);
/* Call to [detachable] ET_TARGET_TYPE.type */
extern T0* T330x22364161(GE_context* ac, T0* C);
/* Call to [detachable] ET_TARGET_TYPE.last_leaf */
extern T0* T330x20414469(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT_TARGET.index */
extern T6 T335x17383428(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT_TARGET.is_instance_free */
extern T1 T335x17383426(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT_TARGET.position */
extern T0* T335x20414465(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT_TARGET.is_open_operand */
extern T1 T335x17383425(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT_ARGUMENT_OPERAND.is_instance_free */
extern T1 T337x17383426(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT_ARGUMENT_OPERAND.is_open_operand */
extern T1 T337x17383425(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT_ARGUMENT_OPERAND.index */
extern T6 T337x17383428(GE_context* ac, T0* C);
/* Call to [detachable] ET_OPERAND.is_open_operand */
extern T1 T338x17383425(GE_context* ac, T0* C);
/* Call to [detachable] ET_OPERAND.index */
extern T6 T338x17383428(GE_context* ac, T0* C);
/* Call to [detachable] ET_OPERAND.is_current */
extern T1 T338x17383427(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION_ITEM.expression */
extern T0* T340x17022977(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION_ITEM.position */
extern T0* T340x20414465(GE_context* ac, T0* C);
/* Call to [detachable] ET_SYMBOL.position */
extern T0* T341x20414465(GE_context* ac, T0* C);
/* Call to [detachable] ET_SYMBOL.text */
extern T0* T341x20496386(GE_context* ac, T0* C);
/* Call to [detachable] ET_SYMBOL.is_equal_symbol */
extern T1 T341x20742163(GE_context* ac, T0* C);
/* Call to [detachable] ET_SYMBOL.break */
extern T0* T341x20414470(GE_context* ac, T0* C);
/* Call to [detachable] ET_SYMBOL.column */
extern T6 T341x20627458(GE_context* ac, T0* C);
/* Call to [detachable] ET_SYMBOL.line */
extern T6 T341x20627457(GE_context* ac, T0* C);
/* Call to [detachable] ET_SYMBOL.is_not_equal */
extern T1 T341x20742177(GE_context* ac, T0* C);
/* Call to [detachable] ET_SYMBOL.is_not_tilde */
extern T1 T341x20742178(GE_context* ac, T0* C);
/* Call to detachable ET_ARGUMENT_OPERANDS.actual_argument */
extern T0* T342x16629765T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_ARGUMENT_OPERANDS.count */
extern T6 T342x16629766(GE_context* ac, T0* C);
/* Call to detachable ET_ARGUMENT_OPERANDS.is_empty */
extern T1 T342x16629762(GE_context* ac, T0* C);
/* Call to detachable ET_ACTUAL_ARGUMENTS.actual_argument */
extern T0* T353x16629765T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_ACTUAL_ARGUMENTS.count */
extern T6 T353x16629766(GE_context* ac, T0* C);
/* Call to detachable ET_ACTUAL_ARGUMENTS.is_empty */
extern T1 T353x16629762(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPES.dynamic_type */
extern T0* T356x23822337T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPES.count */
extern T6 T356x23822340(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPES.index_of */
extern T6 T356x23822339T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPES.has_type */
extern T1 T356x23822342T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_CLOSURE.type */
extern T0* T362x17776645(GE_context* ac, T0* C);
/* Call to [detachable] ET_CLOSURE.locals */
extern T0* T362x17776648(GE_context* ac, T0* C);
/* Call to [detachable] ET_CLOSURE.iteration_components */
extern T0* T362x17776652(GE_context* ac, T0* C);
/* Call to [detachable] ET_CLOSURE.object_tests */
extern T0* T362x17776651(GE_context* ac, T0* C);
/* Call to [detachable] ET_ITERATION_COMPONENT.item_name */
extern T0* T363x19972099(GE_context* ac, T0* C);
/* Call to [detachable] ET_ITERATION_COMPONENT.hash_code */
extern T6 T363x655361(GE_context* ac, T0* C);
/* Call to [detachable] ET_ITERATION_COMPONENT.cursor_item_expression */
extern T0* T363x19972106(GE_context* ac, T0* C);
/* Call to [detachable] ET_ITERATION_COMPONENT.new_cursor_expression */
extern T0* T363x19972105(GE_context* ac, T0* C);
/* Call to [detachable] ET_ITERATION_COMPONENT.has_cursor_name */
extern T1 T363x19972097(GE_context* ac, T0* C);
/* Call to [detachable] ET_ITERATION_COMPONENT.cursor_forth_instruction */
extern T0* T363x19972108(GE_context* ac, T0* C);
/* Call to [detachable] ET_ITERATION_COMPONENT.cursor_after_expression */
extern T0* T363x19972107(GE_context* ac, T0* C);
/* Call to [detachable] ET_ITERATION_COMPONENT.unfolded_cursor_name */
extern T0* T363x19972104(GE_context* ac, T0* C);
/* Call to [detachable] ET_ITERATION_COMPONENT.iterable_expression */
extern T0* T363x19972098(GE_context* ac, T0* C);
/* Call to [detachable] ET_INLINE_AGENT.position */
extern T0* T364x20414465(GE_context* ac, T0* C);
/* Call to [detachable] ET_INLINE_AGENT.implicit_result */
extern T0* T364x14417924(GE_context* ac, T0* C);
/* Call to [detachable] ET_INLINE_AGENT.formal_arguments */
extern T0* T364x17776646(GE_context* ac, T0* C);
/* Call to [detachable] ET_INLINE_AGENT.actual_arguments */
extern T0* T364x14417923(GE_context* ac, T0* C);
/* Call to [detachable] ET_INLINE_AGENT.target */
extern T0* T364x14417922(GE_context* ac, T0* C);
/* Call to [detachable] ET_INLINE_AGENT.locals */
extern T0* T364x17776648(GE_context* ac, T0* C);
/* Call to [detachable] ET_INLINE_AGENT.attached_result_index */
extern T6 T364x14729218(GE_context* ac, T0* C);
/* Call to [detachable] ET_INLINE_AGENT.result_index */
extern T6 T364x14729217(GE_context* ac, T0* C);
/* Call to [detachable] ET_INLINE_AGENT.type */
extern T0* T364x17776645(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_EXPRESSION.parenthesis_call */
extern T0* T366x19890179(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_EXPRESSION.arguments */
extern T0* T366x16678915(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_EXPRESSION.name */
extern T0* T366x16678914(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_EXPRESSION.target */
extern T0* T366x16678913(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_EXPRESSION.arguments_count */
extern T6 T366x19595267(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_EXPRESSION.index */
extern T6 T366x17383428(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_EXPRESSION.is_call_agent */
extern T1 T366x16678918(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_EXPRESSION.is_tuple_label */
extern T1 T366x16678917(GE_context* ac, T0* C);
/* Call to [detachable] ET_MANIFEST_STRING.index */
extern T6 T370x17383428(GE_context* ac, T0* C);
/* Call to [detachable] ET_MANIFEST_STRING.value_position */
extern T0* T370x16334859(GE_context* ac, T0* C);
/* Call to [detachable] ET_MANIFEST_STRING.cast_type */
extern T0* T370x16334856(GE_context* ac, T0* C);
/* Call to [detachable] ET_MANIFEST_STRING.value */
extern T0* T370x16334854(GE_context* ac, T0* C);
/* Call to [detachable] ET_MANIFEST_STRING.position */
extern T0* T370x20414465(GE_context* ac, T0* C);
/* Call to [detachable] ET_MANIFEST_STRING.is_string_32 */
extern T1 T370x16334850(GE_context* ac, T0* C);
/* Call to [detachable] ET_MANIFEST_STRING.is_string_8 */
extern T1 T370x16334849(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.as_invariants */
extern T0* T385x18923531(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.is_invariants */
extern T1 T385x18923522(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.as_feature */
extern T0* T385x18923530(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.is_feature */
extern T1 T385x18923521(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.first_precursor */
extern T0* T385x18923524(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.has_validity_error */
extern T1 T385x18923529(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.implementation_feature */
extern T0* T385x17776653(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.postconditions */
extern T0* T385x17776650(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.preconditions */
extern T0* T385x17776649(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.clients */
extern T0* T385x18923527(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.locals */
extern T0* T385x17776648(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.arguments */
extern T0* T385x17776646(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.type */
extern T0* T385x17776645(GE_context* ac, T0* C);
/* Call to [detachable] ET_STANDALONE_CLOSURE.implementation_class */
extern T0* T385x18923525(GE_context* ac, T0* C);
/* Call to [detachable] ET_ASSERTION.expression */
extern T0* T409x14893058(GE_context* ac, T0* C);
/* Call to [detachable] ET_ASSERTION.is_class_assertion */
extern T1 T409x14893057(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.relative_lower_name */
extern T0* T411x19300373T0T2(GE_context* ac, T0* C, T0* a1, T2 a2);
/* Call to [detachable] ET_PRIMARY_GROUP.kind_name */
extern T0* T411x19300379(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.full_lower_name */
extern T0* T411x19300375T2(GE_context* ac, T0* C, T2 a1);
/* Call to [detachable] ET_PRIMARY_GROUP.is_hidden */
extern T1 T411x19300362(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.is_cluster */
extern T1 T411x19300355(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.current_system */
extern T0* T411x19300369(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.use_obsolete_syntax */
extern T1 T411x19300363(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.universe */
extern T0* T411x19300368(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.is_override */
extern T1 T411x19300360(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.is_unknown */
extern T1 T411x19300359(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.cluster */
extern T0* T411x19300392(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.dotnet_assembly */
extern T0* T411x19300394(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.is_dotnet_assembly */
extern T1 T411x19300357(GE_context* ac, T0* C);
/* Call to [detachable] ET_PRIMARY_GROUP.is_none */
extern T1 T411x19300358(GE_context* ac, T0* C);
/* Call to [detachable] STRING_GENERAL.item */
extern T3 T420x2228226T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] STRING_GENERAL.count */
extern T6 T420x2228263(GE_context* ac, T0* C);
/* Call to [detachable] XM_EIFFEL_SCANNER.start_condition */
extern T6 T516x12582922(GE_context* ac, T0* C);
/* Call to [detachable] XM_EIFFEL_SCANNER.is_applicable_encoding */
extern T1 T516x29016068T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] XM_EIFFEL_SCANNER.end_of_file */
extern T1 T516x12582928(GE_context* ac, T0* C);
/* Call to [detachable] XM_EIFFEL_SCANNER.last_value */
extern T0* T516x29016077(GE_context* ac, T0* C);
/* Call to [detachable] XM_EIFFEL_SCANNER.last_token */
extern T6 T516x12582913(GE_context* ac, T0* C);
/* Call to [detachable] XM_EIFFEL_SCANNER.error_position */
extern T0* T516x29016075(GE_context* ac, T0* C);
/* Call to [detachable] XM_DOCUMENT_NODE.parent */
extern T0* T581x29933569(GE_context* ac, T0* C);
/* Call to [detachable] ET_ACTUAL_PARAMETER.last_leaf */
extern T0* T619x20414469(GE_context* ac, T0* C);
/* Call to [detachable] ET_ACTUAL_PARAMETER.label */
extern T0* T619x21807107(GE_context* ac, T0* C);
/* Call to [detachable] ET_ACTUAL_PARAMETER.position */
extern T0* T619x20414465(GE_context* ac, T0* C);
/* Call to [detachable] ET_ACTUAL_PARAMETER.named_parameter_with_type */
extern T0* T619x21757956T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ACTUAL_PARAMETER.type */
extern T0* T619x21807106(GE_context* ac, T0* C);
/* Call to detachable ET_NAMED_TYPE.to_text */
extern T0* T620x22397004(GE_context* ac, T0* C);
/* Call to detachable ET_NAMED_TYPE.type_with_type_mark */
extern T0* T620x22396949T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_NAMED_TYPE.same_syntactical_like_current_with_type_marks */
extern T1 T620x22396982T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.same_syntactical_tuple_type_with_type_marks */
extern T1 T620x22396985T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.has_formal_types */
extern T1 T620x22396969T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_NAMED_TYPE.same_syntactical_formal_parameter_type_with_type_marks */
extern T1 T620x22396981T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.same_syntactical_class_type_with_type_marks */
extern T1 T620x22396980T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.same_base_type_with_type_marks */
extern T1 T620x22396979T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.same_syntactical_qualified_like_identifier_with_type_marks */
extern T1 T620x22396984T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.same_syntactical_like_feature_with_type_marks */
extern T1 T620x22396983T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.same_syntactical_type_with_type_marks */
extern T1 T620x22396975T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.is_type_reference_with_type_mark */
extern T1 T620x22396958T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to detachable ET_NAMED_TYPE.is_type_attached_with_type_mark */
extern T1 T620x22396960T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to detachable ET_NAMED_TYPE.is_type_detachable_with_type_mark */
extern T1 T620x22396962T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to detachable ET_NAMED_TYPE.is_type_expanded_with_type_mark */
extern T1 T620x22396956T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to detachable ET_NAMED_TYPE.conforms_from_formal_parameter_type_with_type_marks */
extern T1 T620x22396995T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to detachable ET_NAMED_TYPE.conforms_from_tuple_type_with_type_marks */
extern T1 T620x22396996T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to detachable ET_NAMED_TYPE.conforms_from_class_type_with_type_marks */
extern T1 T620x22396994T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to detachable ET_NAMED_TYPE.conforms_to_type_with_type_marks */
extern T1 T620x15319046T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to detachable ET_NAMED_TYPE.same_named_type_with_type_marks */
extern T1 T620x22396977T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.same_named_formal_parameter_type_with_type_marks */
extern T1 T620x22396987T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.same_named_tuple_type_with_type_marks */
extern T1 T620x22396988T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.same_named_class_type_with_type_marks */
extern T1 T620x22396986T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5);
/* Call to detachable ET_NAMED_TYPE.base_class */
extern T0* T620x22396931T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_INDEXING_TERM.indexing_term_value */
extern T0* T755x15777794(GE_context* ac, T0* C);
/* Call to [detachable] ET_INDEXING_TERM.position */
extern T0* T755x20414465(GE_context* ac, T0* C);
/* Call to [detachable] ET_INDEXING_TERM.has_indexing_term_value */
extern T1 T755x15777795T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_ECF_CONDITION_ITEM.is_enabled */
extern T1 T821x13221889T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_ECF_CONDITION_ITEM.is_capability_supported */
extern T1 T821x13221890T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to detachable ET_ECF_CONDITION_ITEM.is_capability_aware */
extern T1 T821x13271041T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_INTERNAL_UNIVERSE.hash_code */
extern T6 T841x655361(GE_context* ac, T0* C);
/* Call to [detachable] ET_ADAPTED_UNIVERSE.lower_name */
extern T0* T908x19087363(GE_context* ac, T0* C);
/* Call to [detachable] ET_ADAPTED_UNIVERSE.name */
extern T0* T908x19087362(GE_context* ac, T0* C);
/* Call to detachable ET_CHARACTER_CONSTANT.cast_type */
extern T0* T968x16236551(GE_context* ac, T0* C);
/* Call to detachable ET_CHARACTER_CONSTANT.index */
extern T6 T968x17383428(GE_context* ac, T0* C);
/* Call to detachable ET_CHARACTER_CONSTANT.value */
extern T3 T968x16236550(GE_context* ac, T0* C);
/* Call to detachable ET_CHARACTER_CONSTANT.value_position */
extern T0* T968x16236555(GE_context* ac, T0* C);
/* Call to detachable ET_CHARACTER_CONSTANT.is_character_32 */
extern T1 T968x16236546(GE_context* ac, T0* C);
/* Call to detachable ET_CHARACTER_CONSTANT.is_character_8 */
extern T1 T968x16236545(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.cast_type */
extern T0* T978x16318476(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.index */
extern T6 T978x17383428(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.is_negative */
extern T1 T978x16318483(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.value */
extern T11 T978x16318466(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.to_integer_32 */
extern T6 T978x16318469(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.literal */
extern T0* T978x16318465(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.sign */
extern T0* T978x16318475(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.value_position */
extern T0* T978x16318480(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.is_natural_64 */
extern T1 T978x16318495(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.is_binary */
extern T1 T978x16318486(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.is_hexadecimal */
extern T1 T978x16318485(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.is_integer_64 */
extern T1 T978x16318491(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.is_integer_32 */
extern T1 T978x16318490(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.is_natural_32 */
extern T1 T978x16318494(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.is_natural_16 */
extern T1 T978x16318493(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.is_natural_8 */
extern T1 T978x16318492(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.is_integer_16 */
extern T1 T978x16318489(GE_context* ac, T0* C);
/* Call to detachable ET_INTEGER_CONSTANT.is_integer_8 */
extern T1 T978x16318488(GE_context* ac, T0* C);
/* Call to detachable ET_AGENT_ARGUMENT_OPERAND_ITEM.agent_actual_argument */
extern T0* T1022x14483457(GE_context* ac, T0* C);
/* Call to detachable ET_CHOICE.upper */
extern T0* T1049x19644419(GE_context* ac, T0* C);
/* Call to detachable ET_CHOICE.is_range */
extern T1 T1049x19644417(GE_context* ac, T0* C);
/* Call to detachable ET_CHOICE.lower */
extern T0* T1049x19644418(GE_context* ac, T0* C);
/* Call to detachable ET_CHOICE_CONSTANT.position */
extern T0* T1052x20414465(GE_context* ac, T0* C);
/* Call to detachable ET_CHOICE_ITEM.position */
extern T0* T1055x20414465(GE_context* ac, T0* C);
/* Call to detachable ET_CHOICE_ITEM.choice */
extern T0* T1055x19693569(GE_context* ac, T0* C);
/* Call to detachable ET_CONDITIONAL.expression */
extern T0* T1070x16711681(GE_context* ac, T0* C);
/* Call to detachable ET_CONDITIONAL.position */
extern T0* T1070x20414465(GE_context* ac, T0* C);
/* Call to detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM.resolved_syntactical_constraint_with_type */
extern T0* T1075x21987332T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM.type */
extern T0* T1075x21987330(GE_context* ac, T0* C);
/* Call to detachable ET_CONSTRAINT_ACTUAL_PARAMETER_ITEM.position */
extern T0* T1075x21987331(GE_context* ac, T0* C);
/* Call to detachable ET_CONSTRAINT_TYPE.resolved_syntactical_constraint */
extern T0* T1087x22102018T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to detachable ET_CONVERT_FEATURE.name */
extern T0* T1090x15384579(GE_context* ac, T0* C);
/* Call to detachable ET_CONVERT_FEATURE.is_convert_to */
extern T1 T1090x15384578(GE_context* ac, T0* C);
/* Call to detachable ET_CONVERT_FEATURE.is_convert_from */
extern T1 T1090x15384577(GE_context* ac, T0* C);
/* Call to detachable ET_EXTERNAL_ALIAS.manifest_string */
extern T0* T1136x18087937(GE_context* ac, T0* C);
/* Call to detachable ET_INDEXING_TERM_ITEM.indexing_term */
extern T0* T1181x15777793(GE_context* ac, T0* C);
/* Call to detachable ET_INDEXING_TERM_ITEM.position */
extern T0* T1181x20414465(GE_context* ac, T0* C);
/* Call to detachable ET_INDEXING_TERM_ITEM.has_indexing_term_value */
extern T1 T1181x15777795T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_INSTRUCTION.position */
extern T0* T1195x20414465(GE_context* ac, T0* C);
/* Call to detachable ET_LIKE_TYPE.to_text */
extern T0* T1203x22397004(GE_context* ac, T0* C);
/* Call to detachable ET_LIKE_TYPE.position */
extern T0* T1203x20414465(GE_context* ac, T0* C);
/* Call to detachable ET_LIKE_TYPE.resolved_formal_parameters */
extern T0* T1203x22396997T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_LIKE_TYPE.like_keyword */
extern T0* T1203x22265858(GE_context* ac, T0* C);
/* Call to detachable ET_LIKE_TYPE.has_formal_types */
extern T1 T1203x22396969T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_LIKE_TYPE.has_unqualified_anchored_type */
extern T1 T1203x22396966(GE_context* ac, T0* C);
/* Call to detachable ET_MANIFEST_STRING_ITEM.position */
extern T0* T1221x20414465(GE_context* ac, T0* C);
/* Call to detachable ET_MANIFEST_STRING_ITEM.manifest_string */
extern T0* T1221x17252353(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_CONSTRAINT.type */
extern T0* T1275x16105474(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_CONSTRAINT.renames */
extern T0* T1275x16105475(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_CONSTRAINT_ITEM.type_constraint */
extern T0* T1278x16105473(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_CONSTRAINT_ITEM.type */
extern T0* T1278x16105474(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_ITEM.type */
extern T0* T1284x22446081(GE_context* ac, T0* C);
/* Call to detachable ET_CONSTRAINT.last_leaf */
extern T0* T1334x20414469(GE_context* ac, T0* C);
/* Call to detachable ET_CONSTRAINT.type_constraint */
extern T0* T1334x15319041T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_CONSTRAINT.count */
extern T6 T1334x15319044(GE_context* ac, T0* C);
/* Call to detachable ET_CONSTRAINT.types_to_text */
extern T0* T1334x15319047(GE_context* ac, T0* C);
/* Call to [detachable] ET_CONSTRAINT_BASE_TYPES.type_constraint */
extern T0* T1335x15319041T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_CONSTRAINT_BASE_TYPES.count */
extern T6 T1335x15319044(GE_context* ac, T0* C);
/* Call to [detachable] ET_CONSTRAINT_BASE_TYPES.conforms_to_type_with_type_marks */
extern T1 T1335x15319046T0T0T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4, T0* a5, T0* a6);
/* Call to [detachable] ET_CONSTRAINT_BASE_TYPES.types_to_text */
extern T0* T1335x15319047(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE_CONSTRAINT.renames */
extern T0* T1336x16105475(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE_CONSTRAINT.type */
extern T0* T1336x16105474(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE_CONSTRAINT.base_class */
extern T0* T1336x15089665(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE_CONSTRAINT.named_feature */
extern T0* T1336x15089669T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FLATTENED_FEATURE.adapted_feature */
extern T0* T1343x18268177(GE_context* ac, T0* C);
/* Call to [detachable] ET_FLATTENED_FEATURE.is_adapted */
extern T1 T1343x18268164(GE_context* ac, T0* C);
/* Call to [detachable] ET_FLATTENED_FEATURE.inherited_feature */
extern T0* T1343x18268175(GE_context* ac, T0* C);
/* Call to [detachable] ET_FLATTENED_FEATURE.is_inherited */
extern T1 T1343x18268162(GE_context* ac, T0* C);
/* Call to [detachable] ET_FLATTENED_FEATURE.redeclared_feature */
extern T0* T1343x18268176(GE_context* ac, T0* C);
/* Call to [detachable] ET_FLATTENED_FEATURE.is_redeclared */
extern T1 T1343x18268163(GE_context* ac, T0* C);
/* Call to [detachable] ET_FLATTENED_FEATURE.is_immediate */
extern T1 T1343x18268161(GE_context* ac, T0* C);
/* Call to [detachable] ET_FLATTENED_FEATURE.flattened_feature */
extern T0* T1343x18268173(GE_context* ac, T0* C);
/* Call to [detachable] ET_FLATTENED_FEATURE.immediate_feature */
extern T0* T1343x18268174(GE_context* ac, T0* C);
/* Call to [detachable] ET_ADAPTED_CLASS.named_query */
extern T0* T1422x15089667T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ADAPTED_CLASS.base_class */
extern T0* T1422x15089665(GE_context* ac, T0* C);
/* Call to [detachable] ET_ADAPTED_CLASS.named_procedure */
extern T0* T1422x15089668T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ADAPTED_CLASS.base_type */
extern T0* T1422x15089666(GE_context* ac, T0* C);
/* Call to [detachable] ET_ADAPTED_CLASS.named_feature */
extern T0* T1422x15089669T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ADAPTED_CLASS.base_type_index_of_label */
extern T6 T1422x15089670T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] XM_ELEMENT_NODE.parent */
extern T0* T1472x29933569(GE_context* ac, T0* C);
/* Call to detachable ET_DECLARED_TYPE.last_leaf */
extern T0* T1510x20414469(GE_context* ac, T0* C);
/* Call to detachable ET_DECLARED_TYPE.type */
extern T0* T1510x22118401(GE_context* ac, T0* C);
/* Call to detachable ET_DECLARED_TYPE.position */
extern T0* T1510x20414465(GE_context* ac, T0* C);
/* Call to detachable ET_EXTERNAL_LANGUAGE.manifest_string */
extern T0* T1514x18120705(GE_context* ac, T0* C);
/* Call to detachable ET_UNTAGGED_ASSERTION.is_class_assertion */
extern T1 T1561x14893057(GE_context* ac, T0* C);
/* Call to detachable ET_ASSERTION_ITEM.position */
extern T0* T1563x20414465(GE_context* ac, T0* C);
/* Call to detachable ET_ASSERTION_ITEM.assertion */
extern T0* T1563x14925825(GE_context* ac, T0* C);
/* Call to [detachable] ET_REPLICABLE_FEATURE.first_feature */
extern T0* T1683x18776067(GE_context* ac, T0* C);
/* Call to [detachable] ET_REPLICABLE_FEATURE.replicated_feature */
extern T0* T1683x18776066(GE_context* ac, T0* C);
/* Call to [detachable] ET_REPLICABLE_FEATURE.has_replication */
extern T1 T1683x18776065(GE_context* ac, T0* C);
/* Call to [detachable] ET_REPLICABLE_FEATURE.selected_count */
extern T6 T1683x18776068(GE_context* ac, T0* C);
/* Call to [detachable] ET_ECF_GROUP.class_options */
extern T0* T1735x13746182(GE_context* ac, T0* C);
/* Call to [detachable] ET_ECF_GROUP.options */
extern T0* T1735x13746181(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_FEATURE.is_attribute */
extern T1 T1746x18186271(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_FEATURE.implementation_feature */
extern T0* T1746x17776653(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_FEATURE.is_once_per_object */
extern T1 T1746x17776644(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_FEATURE.is_once_per_thread */
extern T1 T1746x17776643(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_FEATURE.is_once_per_process */
extern T1 T1746x17776642(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_FEATURE.is_once */
extern T1 T1746x17776641(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_FEATURE.lower_name */
extern T0* T1746x18923523(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_FEATURE.rescue_clause */
extern T0* T1746x18415619(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_FEATURE.compound */
extern T0* T1746x18415617(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_FEATURE.arguments */
extern T0* T1746x18268170(GE_context* ac, T0* C);
/* Call to [detachable] ET_ROUTINE.implementation_feature */
extern T0* T1747x17776653(GE_context* ac, T0* C);
/* Call to [detachable] ET_ROUTINE.arguments */
extern T0* T1747x18268170(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_FEATURE_CLOSURE.locals */
extern T0* T1748x17776648(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_FEATURE_CLOSURE.compound */
extern T0* T1748x18415617(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_FEATURE_CLOSURE.is_once_per_process */
extern T1 T1748x17776642(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_FEATURE_CLOSURE.rescue_clause */
extern T0* T1748x18415619(GE_context* ac, T0* C);
/* Call to detachable ET_AST_LEAF.last_position */
extern T0* T1752x20414467(GE_context* ac, T0* C);
/* Call to detachable ET_OPERATOR.is_prefix_not_symbol */
extern T1 T1753x21004330(GE_context* ac, T0* C);
/* Call to detachable ET_OPERATOR.is_prefix_not */
extern T1 T1753x21004329(GE_context* ac, T0* C);
/* Call to detachable ET_OPERATOR.is_infix_and_then_symbol */
extern T1 T1753x21004303(GE_context* ac, T0* C);
/* Call to detachable ET_OPERATOR.is_infix_and_symbol */
extern T1 T1753x21004301(GE_context* ac, T0* C);
/* Call to detachable ET_OPERATOR.is_infix_and_then */
extern T1 T1753x21004302(GE_context* ac, T0* C);
/* Call to detachable ET_OPERATOR.is_infix_and */
extern T1 T1753x21004300(GE_context* ac, T0* C);
/* Call to detachable ET_OPERATOR.is_infix_implies_symbol */
extern T1 T1753x21004310(GE_context* ac, T0* C);
/* Call to detachable ET_OPERATOR.is_infix_implies */
extern T1 T1753x21004309(GE_context* ac, T0* C);
/* Call to detachable ET_OPERATOR.is_infix_or_else_symbol */
extern T1 T1753x21004318(GE_context* ac, T0* C);
/* Call to detachable ET_OPERATOR.is_infix_or_symbol */
extern T1 T1753x21004316(GE_context* ac, T0* C);
/* Call to detachable ET_OPERATOR.is_infix_or_else */
extern T1 T1753x21004317(GE_context* ac, T0* C);
/* Call to detachable ET_OPERATOR.is_infix_or */
extern T1 T1753x21004315(GE_context* ac, T0* C);
/* Call to detachable ET_OPERATOR.is_tuple_label */
extern T1 T1753x21004297(GE_context* ac, T0* C);
/* Call to detachable ET_OPERATOR.position */
extern T0* T1753x20414465(GE_context* ac, T0* C);
/* Call to [detachable] ET_GROUP.is_cluster */
extern T1 T1757x19300355(GE_context* ac, T0* C);
/* Call to [detachable] ET_GROUP.absolute_pathname */
extern T0* T1757x19300378(GE_context* ac, T0* C);
/* Call to [detachable] ET_GROUP.universe */
extern T0* T1757x19300368(GE_context* ac, T0* C);
/* Call to [detachable] ET_GROUP.parent */
extern T0* T1757x19300395(GE_context* ac, T0* C);
/* Call to [detachable] ET_GROUP.full_name */
extern T0* T1757x19300374T2(GE_context* ac, T0* C, T2 a1);
/* Call to [detachable] ET_GROUP.full_lower_name */
extern T0* T1757x19300375T2(GE_context* ac, T0* C, T2 a1);
/* Call to [detachable] ET_GROUP.kind_lower_name */
extern T0* T1757x19300380(GE_context* ac, T0* C);
/* Call to [detachable] ET_GROUP.cluster */
extern T0* T1757x19300392(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_INSTRUCTION.target */
extern T0* T1949x16678913(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_INSTRUCTION.name */
extern T0* T1949x16678914(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_INSTRUCTION.is_call_agent */
extern T1 T1949x16678918(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_INSTRUCTION.is_tuple_label */
extern T1 T1949x16678917(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_INSTRUCTION.arguments */
extern T0* T1949x16678915(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_INSTRUCTION.position */
extern T0* T1949x20414465(GE_context* ac, T0* C);
/* Call to [detachable] ET_INTERNAL_ROUTINE_INLINE_AGENT.implicit_result */
extern T0* T1952x14417924(GE_context* ac, T0* C);
/* Call to [detachable] ET_LIKE_IDENTIFIER.name */
extern T0* T1970x22233089(GE_context* ac, T0* C);
/* Call to [detachable] ET_LIKE_IDENTIFIER.to_text */
extern T0* T1970x22397004(GE_context* ac, T0* C);
/* Call to detachable ET_AST_NODE.position */
extern T0* T1972x20414465(GE_context* ac, T0* C);
/* Call to [detachable] ET_CONVERT_EXPRESSION.expression */
extern T0* T1975x16744449(GE_context* ac, T0* C);
/* Call to [detachable] ET_CONVERT_EXPRESSION.index */
extern T6 T1975x17383428(GE_context* ac, T0* C);
/* Call to [detachable] ET_CONVERT_EXPRESSION.convert_feature */
extern T0* T1975x16744450(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_WITH_ACTUAL_ARGUMENTS.name */
extern T0* T2009x19595265(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_WITH_ACTUAL_ARGUMENTS.arguments */
extern T0* T2009x19595266(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL.name */
extern T0* T2011x16678914(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUERY_INLINE_AGENT.type */
extern T0* T2016x17776645(GE_context* ac, T0* C);
/* Call to [detachable] ET_QUERY_INLINE_AGENT.formal_arguments */
extern T0* T2016x17776646(GE_context* ac, T0* C);
/* Call to [detachable] ET_PROCEDURE_INLINE_AGENT.formal_arguments */
extern T0* T2019x17776646(GE_context* ac, T0* C);
/* Call to [detachable] ET_REGULAR_FEATURE_CALL.parenthesis_call */
extern T0* T2021x19890179(GE_context* ac, T0* C);
/* Call to [detachable] ET_REGULAR_FEATURE_CALL.name */
extern T0* T2021x16678914(GE_context* ac, T0* C);
/* Call to [detachable] ET_REGULAR_FEATURE_CALL.target */
extern T0* T2021x16678913(GE_context* ac, T0* C);
/* Call to [detachable] ET_REGULAR_FEATURE_CALL.arguments */
extern T0* T2021x16678915(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE_CALL_EXPRESSION.arguments */
extern T0* T2023x16678915(GE_context* ac, T0* C);
/* Call to [detachable] ET_CALL_WITH_ACTUAL_ARGUMENT_LIST.arguments */
extern T0* T2025x19595266(GE_context* ac, T0* C);
/* Call to [detachable] ET_CREATION_COMPONENT.position */
extern T0* T2029x20414465(GE_context* ac, T0* C);
/* Call to [detachable] ET_LOOP_COMPONENT.loop_compound */
extern T0* T2087x20021253(GE_context* ac, T0* C);
/* Call to [detachable] ET_LOOP_COMPONENT.until_conditional */
extern T0* T2087x20021252(GE_context* ac, T0* C);
/* Call to [detachable] ET_LOOP_COMPONENT.variant_part */
extern T0* T2087x20021251(GE_context* ac, T0* C);
/* Call to [detachable] ET_LOOP_COMPONENT.invariant_part */
extern T0* T2087x20021250(GE_context* ac, T0* C);
/* Call to detachable LX_TRANSITION [[attached] LX_NFA_STATE].target */
extern T0* T2172x26558465(GE_context* ac, T0* C);
/* Call to detachable LX_TRANSITION [[attached] LX_NFA_STATE].labeled */
extern T1 T2172x26558466T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] STRING_8.append_character */
extern void T17x2293809T2(GE_context* ac, T0* C, T2 a1);
/* Call to [detachable] STRING_8.append_string */
extern void T17x2293820T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] STRING_8.put */
extern void T17x2588678T2T6(GE_context* ac, T0* C, T2 a1, T6 a2);
/* Call to [detachable] STRING_8.remove_head */
extern void T17x2326533T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] STRING_8.append_code */
extern void T17x2326558T10(GE_context* ac, T0* C, T10 a1);
/* Call to [detachable] STRING_8.append_substring */
extern void T17x2293821T0T6T6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* Call to [detachable] STRING_8.append_string_general */
extern void T17x2326547T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] STRING_8.append_substring_general */
extern void T17x2326546T0T6T6(GE_context* ac, T0* C, T0* a1, T6 a2, T6 a3);
/* Call to [detachable] STRING_8.wipe_out */
extern void T17x2326530(GE_context* ac, T0* C);
/* Call to [detachable] STRING_8.remove_tail */
extern void T17x2326531T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] STRING_8.keep_head */
extern void T17x2326539T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable EXCEPTION.set_type_name */
extern void T22x1753112T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable EXCEPTION.set_recipient_name */
extern void T22x1753115T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable EXCEPTION.set_description */
extern void T22x1753117T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable EXCEPTION.set_exception_trace */
extern void T22x1753111T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable EXCEPTION.set_throwing_exception */
extern void T22x1753116T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] KI_CHARACTER_INPUT_STREAM.read_character */
extern void T41x6144012(GE_context* ac, T0* C);
/* Call to [detachable] KI_CHARACTER_INPUT_STREAM.close */
extern void T41x6144010(GE_context* ac, T0* C);
/* Call to [detachable] AP_OPTION.record_occurrence */
extern void T51x5537811T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] AP_OPTION.reset */
extern void T51x5537810(GE_context* ac, T0* C);
/* Call to [detachable] ET_BASE_TYPE.append_to_string */
extern void T66x22397010T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_BASE_TYPE.add_adapted_base_classes_to_list */
extern void T66x22397011T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_BASE_TYPE.context_add_adapted_base_classes_to_list */
extern void T66x22429749T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_BASE_TYPE.process */
extern void T66x20414474T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_BASE_TYPE.append_runtime_name_to_string */
extern void T66x22397008T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] KI_TEXT_OUTPUT_STREAM.put_integer */
extern void T75x6045710T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] KI_TEXT_OUTPUT_STREAM.put_character */
extern void T75x6176774T2(GE_context* ac, T0* C, T2 a1);
/* Call to [detachable] KI_TEXT_OUTPUT_STREAM.put_new_line */
extern void T75x6258690(GE_context* ac, T0* C);
/* Call to [detachable] KI_TEXT_OUTPUT_STREAM.put_string */
extern void T75x6045712T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] KI_TEXT_OUTPUT_STREAM.append */
extern void T75x6176773T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] KI_TEXT_OUTPUT_STREAM.put_line */
extern void T75x6258691T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] XM_CALLBACKS_FILTER.set_next */
extern void T180x28065794T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] XM_CALLBACKS.on_xml_declaration */
extern void T181x28000269T0T0T1(GE_context* ac, T0* C, T0* a1, T0* a2, T1 a3);
/* Call to [detachable] XM_CALLBACKS.on_attribute */
extern void T181x28000264T0T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3, T0* a4);
/* Call to [detachable] XM_CALLBACKS.on_error */
extern void T181x28000268T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] XM_CALLBACKS.on_end_tag */
extern void T181x28000262T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to [detachable] XM_CALLBACKS.on_start_tag_finish */
extern void T181x28000263(GE_context* ac, T0* C);
/* Call to [detachable] XM_CALLBACKS.on_start_tag */
extern void T181x28000265T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to [detachable] XM_CALLBACKS.on_content */
extern void T181x28000261T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] XM_CALLBACKS.on_processing_instruction */
extern void T181x28000267T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] XM_CALLBACKS.on_comment */
extern void T181x28000266T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] XM_CALLBACKS.on_finish */
extern void T181x28000270(GE_context* ac, T0* C);
/* Call to [detachable] XM_CALLBACKS.on_start */
extern void T181x28000271(GE_context* ac, T0* C);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.set_library_target */
extern void T185x14057483T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.set_uuid */
extern void T185x13287433T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.set_read_only */
extern void T185x14057486T1(GE_context* ac, T0* C, T1 a1);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.set_targets */
extern void T185x14057484T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.set_notes */
extern void T185x14057481T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.set_description */
extern void T185x14057482T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.set_ecf_version */
extern void T185x13287431T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ECF_SYSTEM_CONFIG.set_ecf_namespace */
extern void T185x13287432T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_static */
extern void T223x23806013T1(GE_context* ac, T0* C, T1 a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_procedure_call */
extern void T223x23806002T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_query_call */
extern void T223x23806003T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_object_equality_expression */
extern void T223x23806000T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.put_equality_expression */
extern void T223x23806001T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.use_all_attributes */
extern void T223x23806008T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_alive */
extern void T223x23806014(GE_context* ac, T0* C);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_attribute_position */
extern void T223x23806009T0T6(GE_context* ac, T0* C, T0* a1, T6 a2);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_next_type */
extern void T223x23805999T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_hash_code */
extern void T223x23806010T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_attached_type */
extern void T223x23806012T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_PRIMARY_TYPE.set_id */
extern void T223x23806011T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_PROCEDURE.add_creation_clients_to */
extern void T234x18186325T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to detachable ET_PROCEDURE.reset_after_interface_checked */
extern void T234x18186327(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.reset_after_features_flattened */
extern void T234x18186328(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.reset */
extern void T234x18186329(GE_context* ac, T0* C);
/* Call to detachable ET_PROCEDURE.set_iteration_components */
extern void T234x17776654T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_PROCEDURE.set_object_tests */
extern void T234x17776655T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_PROCEDURE.set_frozen_keyword */
extern void T234x18186313T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_PROCEDURE.reset_validity_checked */
extern void T234x18923533(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.reset_after_interface_checked */
extern void T235x18186327(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.reset_after_features_flattened */
extern void T235x18186328(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.reset */
extern void T235x18186329(GE_context* ac, T0* C);
/* Call to detachable ET_QUERY.set_iteration_components */
extern void T235x17776654T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_QUERY.set_object_tests */
extern void T235x17776655T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_QUERY.set_frozen_keyword */
extern void T235x18186313T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_QUERY.reset_validity_checked */
extern void T235x18923533(GE_context* ac, T0* C);
/* Call to [detachable] ET_AST_PROCESSOR.process_convert_to_expression */
extern void T237x24821928T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_convert_from_expression */
extern void T237x24821930T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_convert_builtin_expression */
extern void T237x24821934T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_unqualified_call_instruction */
extern void T237x24821770T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_static_call_instruction */
extern void T237x24821789T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_precursor_instruction */
extern void T237x24821812T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_create_instruction */
extern void T237x24821926T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_bang_instruction */
extern void T237x24821963T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_loop_instruction */
extern void T237x24821841T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_assignment_attempt */
extern void T237x24821967T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_assignment */
extern void T237x24821968T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_assigner_instruction */
extern void T237x24821969T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_repeat_instruction */
extern void T237x24821796T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_inspect_instruction */
extern void T237x24821859T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_if_instruction */
extern void T237x24821871T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_debug_instruction */
extern void T237x24821919T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_check_instruction */
extern void T237x24821950T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_across_instruction */
extern void T237x24821988T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_retry_instruction */
extern void T237x24821793T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_infix_cast_expression */
extern void T237x24821864T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_constrained_formal_parameter */
extern void T237x24821937T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_formal_parameter */
extern void T237x24821881T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_regular_integer_constant */
extern void T237x24821802T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_binary_integer_constant */
extern void T237x24821960T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_octal_integer_constant */
extern void T237x24821831T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_hexadecimal_integer_constant */
extern void T237x24821876T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_underscored_integer_constant */
extern void T237x24821777T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_verbatim_string */
extern void T237x24821768T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_special_manifest_string */
extern void T237x24821791T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_c3_character_constant */
extern void T237x24821952T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_like_current */
extern void T237x24821848T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_unique_attribute */
extern void T237x24821772T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_constant_attribute */
extern void T237x24821938T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_attribute */
extern void T237x24821964T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_extended_attribute */
extern void T237x24821897T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_create_expression */
extern void T237x24821927T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_equality_expression */
extern void T237x24821903T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_object_equality_expression */
extern void T237x24821833T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_object_test */
extern void T237x24821832T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_qualified_call_instruction */
extern void T237x24821807T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_tuple_type */
extern void T237x24821782T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_external_function */
extern void T237x24821895T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_formal_parameter_type */
extern void T237x24821878T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_class_type */
extern void T237x24821944T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_external_procedure_inline_agent */
extern void T237x24821892T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_once_procedure_inline_agent */
extern void T237x24821824T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_do_procedure_inline_agent */
extern void T237x24821913T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_external_function_inline_agent */
extern void T237x24821894T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_once_function_inline_agent */
extern void T237x24821827T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_do_function_inline_agent */
extern void T237x24821915T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_iteration_cursor */
extern void T237x24821857T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_once_manifest_string */
extern void T237x24821826T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_old_object_test */
extern void T237x24821829T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_old_expression */
extern void T237x24821830T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_prefix_expression */
extern void T237x24821810T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_manifest_type */
extern void T237x24821835T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_infix_expression */
extern void T237x24821862T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_expression_address */
extern void T237x24821899T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_result_address */
extern void T237x24821794T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_current_address */
extern void T237x24821921T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_feature_address */
extern void T237x24821890T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_precursor_expression */
extern void T237x24821813T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_qualified_like_type */
extern void T237x24821805T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_qualified_like_braced_type */
extern void T237x24821806T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_like_feature */
extern void T237x24821847T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_external_procedure */
extern void T237x24821893T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_once_procedure */
extern void T237x24821825T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_deferred_function */
extern void T237x24821918T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_once_function */
extern void T237x24821828T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_do_function */
extern void T237x24821916T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_c2_character_constant */
extern void T237x24821953T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_c1_character_constant */
extern void T237x24821954T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_false_constant */
extern void T237x24821891T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_true_constant */
extern void T237x24821783T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_underscored_real_constant */
extern void T237x24821776T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_regular_real_constant */
extern void T237x24821800T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_regular_manifest_string */
extern void T237x24821801T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_strip_expression */
extern void T237x24821788T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_static_call_expression */
extern void T237x24821790T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_quantifier_expression */
extern void T237x24821804T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_parenthesized_expression */
extern void T237x24821816T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_manifest_array */
extern void T237x24821839T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_inspect_expression */
extern void T237x24821860T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_if_expression */
extern void T237x24821872T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_call_agent */
extern void T237x24821951T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_bracket_expression */
extern void T237x24821956T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_across_expression */
extern void T237x24821989T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_void */
extern void T237x24821767T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_named_object_test */
extern void T237x24821834T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_unqualified_call_expression */
extern void T237x24821771T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_deferred_procedure */
extern void T237x24821917T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_do_procedure */
extern void T237x24821914T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_manifest_tuple */
extern void T237x24821836T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_result */
extern void T237x24821795T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_qualified_call_expression */
extern void T237x24821808T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_current */
extern void T237x24821922T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_identifier */
extern void T237x24821875T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AST_PROCESSOR.process_class */
extern void T237x24821946T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.add_adapted_base_classes_to_list */
extern void T238x22397011T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_TYPE.reset */
extern void T238x15319050(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE.process */
extern void T238x20414474T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.reset_qualified_anchored_types */
extern void T238x22397012(GE_context* ac, T0* C);
/* Call to [detachable] ET_TYPE.append_to_string */
extern void T238x22397010T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.append_unaliased_to_string */
extern void T238x22397009T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_TYPE.append_runtime_name_to_string */
extern void T238x22397008T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_UNIVERSE.universes_do_if */
extern void T243x19447942T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_UNIVERSE.master_classes_do_if_until */
extern void T243x19447948T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to [detachable] ET_UNIVERSE.master_classes_do_if */
extern void T243x19447949T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_UNIVERSE.master_classes_do_until */
extern void T243x19447950T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_UNIVERSE.master_classes_do_all */
extern void T243x19447951T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_UNIVERSE.classes_do_if */
extern void T243x19447959T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_UNIVERSE.classes_do_until */
extern void T243x19447960T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_UNIVERSE.classes_do_all */
extern void T243x19447961T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_TYPE_MARK.append_to_string_with_space */
extern void T248x20840472T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_ACTUAL_PARAMETERS.reset_qualified_anchored_types */
extern void T249x21774359(GE_context* ac, T0* C);
/* Call to detachable ET_ACTUAL_PARAMETERS.reset */
extern void T249x21774360(GE_context* ac, T0* C);
/* Call to detachable ET_ACTUAL_PARAMETERS.append_to_string */
extern void T249x21774358T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_ACTUAL_PARAMETERS.append_unaliased_to_string */
extern void T249x21774357T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_ACTUAL_PARAMETERS.append_runtime_name_to_string */
extern void T249x21774356T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_DYNAMIC_TYPE_SET.put_target */
extern void T251x24248330T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to detachable ET_DYNAMIC_TYPE_SET.set_never_void */
extern void T251x24248331(GE_context* ac, T0* C);
/* Call to detachable ET_DYNAMIC_TYPE_SET.propagate_can_be_void */
extern void T251x24133633T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_DYNAMIC_TYPE_SET.put_type_from_type_set */
extern void T251x24133635T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to [detachable] ET_DYNAMIC_TYPE.put_target */
extern void T252x24248330T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_DYNAMIC_TYPE.set_meta_type */
extern void T252x24215575T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_EXPRESSION.process */
extern void T281x20414474T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_EXPRESSION.reset */
extern void T281x14893060(GE_context* ac, T0* C);
/* Call to [detachable] ET_EXPRESSION.set_index */
extern void T281x17383429T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_FEATURE.set_id */
extern void T285x18186323T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_FEATURE.process */
extern void T285x20414474T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FEATURE.set_other_seeds */
extern void T285x18186316T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FEATURE.set_first_seed */
extern void T285x18186317T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_FEATURE.set_clients */
extern void T285x18186324T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FEATURE.set_version */
extern void T285x18186320T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_FEATURE.set_other_precursors */
extern void T285x18186314T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FEATURE.set_first_precursor */
extern void T285x18186315T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FEATURE.reset_postconditions */
extern void T285x18186309(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.reset_preconditions */
extern void T285x18186310(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.set_implementation_class */
extern void T285x18186319T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FEATURE.set_implementation_feature */
extern void T285x18186318T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_FEATURE.set_validity_error */
extern void T285x18923534(GE_context* ac, T0* C);
/* Call to [detachable] ET_FEATURE.set_validity_checked */
extern void T285x18923535(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT.set_arguments */
extern void T310x14417932T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AGENT.process */
extern void T310x20414474T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_CALL_NAME.set_seed */
extern void T316x21004342T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_CONSTANT.set_index */
extern void T321x17383429T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_CONSTANT.process */
extern void T321x20414474T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AGENT_TARGET.process */
extern void T335x20414474T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AGENT_TARGET.set_index */
extern void T335x17383429T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_AGENT_TARGET.reset */
extern void T335x17383430(GE_context* ac, T0* C);
/* Call to [detachable] ET_AGENT_ARGUMENT_OPERAND.process */
extern void T337x20414474T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_AGENT_ARGUMENT_OPERAND.reset */
extern void T337x17383430(GE_context* ac, T0* C);
/* Call to [detachable] ET_OPERAND.set_index */
extern void T338x17383429T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_INLINE_AGENT.set_attached_result_index */
extern void T364x14729227T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_INLINE_AGENT.set_result_index */
extern void T364x14729228T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_INLINE_AGENT.set_iteration_components */
extern void T364x17776654T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_INLINE_AGENT.set_object_tests */
extern void T364x17776655T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_INLINE_AGENT.set_actual_arguments */
extern void T364x14417932T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_QUALIFIED_FEATURE_CALL_EXPRESSION.set_index */
extern void T366x17383429T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_MANIFEST_STRING.set_index */
extern void T370x17383429T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] ET_MANIFEST_STRING.set_type */
extern void T370x16334864T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_MANIFEST_STRING.set_cast_type */
extern void T370x16334865T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ASSERTION.process */
extern void T409x20414474T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_ASSERTION.reset */
extern void T409x14893060(GE_context* ac, T0* C);
/* Call to [detachable] STRING_GENERAL.keep_head */
extern void T420x2326539T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] STRING_GENERAL.remove_head */
extern void T420x2326533T6(GE_context* ac, T0* C, T6 a1);
/* Call to [detachable] STRING_GENERAL.append_code */
extern void T420x2326558T10(GE_context* ac, T0* C, T10 a1);
/* Call to [detachable] XM_EIFFEL_SCANNER.close_input */
extern void T516x29016088(GE_context* ac, T0* C);
/* Call to [detachable] XM_EIFFEL_SCANNER.set_input_from_resolver */
extern void T516x29016089T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] XM_EIFFEL_SCANNER.set_encoding */
extern void T516x29016087T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] XM_EIFFEL_SCANNER.push_start_condition_dtd_ignore */
extern void T516x29016092(GE_context* ac, T0* C);
/* Call to [detachable] XM_EIFFEL_SCANNER.read_token */
extern void T516x12582970(GE_context* ac, T0* C);
/* Call to [detachable] XM_EIFFEL_SCANNER.set_input_stream */
extern void T516x29016090T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] XM_DOCUMENT_NODE.node_set_parent */
extern void T581x29786114T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_TARGET.propagate_can_be_void */
extern void T701x24133633T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_DYNAMIC_TARGET.put_type_from_type_set */
extern void T701x24133635T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to [detachable] ET_INTERNAL_UNIVERSE.clusters_do_explicit */
extern void T841x19333162T0(GE_context* ac, T0* C, T0* a1);
/* Call to [detachable] ET_INTERNAL_UNIVERSE.dotnet_assemblies_do_if */
extern void T841x19333155T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to detachable ET_CHARACTER_CONSTANT.set_index */
extern void T968x17383429T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_CHARACTER_CONSTANT.set_cast_type */
extern void T968x16236559T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_CHARACTER_CONSTANT.set_type */
extern void T968x16236558T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_INTEGER_CONSTANT.set_index */
extern void T978x17383429T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_INTEGER_CONSTANT.set_cast_type */
extern void T978x16318505T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_INTEGER_CONSTANT.set_type */
extern void T978x16318504T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_INTEGER_CONSTANT.set_sign */
extern void T978x16318506T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_CHOICE.reset */
extern void T1049x19644421(GE_context* ac, T0* C);
/* Call to detachable ET_CHOICE_CONSTANT.reset */
extern void T1052x19644421(GE_context* ac, T0* C);
/* Call to detachable ET_INSTRUCTION.process */
extern void T1195x20414474T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_INSTRUCTION.reset */
extern void T1195x19955714(GE_context* ac, T0* C);
/* Call to detachable ET_LIKE_TYPE.reset_qualified_anchored_types */
extern void T1203x22397012(GE_context* ac, T0* C);
/* Call to detachable ET_LIKE_TYPE.reset */
extern void T1203x15319050(GE_context* ac, T0* C);
/* Call to detachable ET_LIKE_TYPE.append_to_string */
extern void T1203x22397010T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_TYPE_CONSTRAINT.reset */
extern void T1275x15319050(GE_context* ac, T0* C);
/* Call to detachable ET_TYPE_CONSTRAINT.reset_renames */
extern void T1275x15319049(GE_context* ac, T0* C);
/* Call to detachable ET_CONSTRAINT.reset */
extern void T1334x15319050(GE_context* ac, T0* C);
/* Call to detachable ET_CONSTRAINT.reset_renames */
extern void T1334x15319049(GE_context* ac, T0* C);
/* Call to [detachable] ET_ADAPTED_CLASS.add_overloaded_procedures */
extern void T1422x15089671T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] ET_ADAPTED_CLASS.add_overloaded_queries */
extern void T1422x15089672T0T0(GE_context* ac, T0* C, T0* a1, T0* a2);
/* Call to [detachable] XM_ELEMENT_NODE.node_set_parent */
extern void T1472x29786114T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_UNTAGGED_ASSERTION.reset */
extern void T1561x14893060(GE_context* ac, T0* C);
/* Call to [detachable] ET_ECF_GROUP.set_options */
extern void T1735x13746192T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable ET_OPERATOR.set_seed */
extern void T1753x21004342T6(GE_context* ac, T0* C, T6 a1);
/* Call to detachable ET_OPERATOR.reset */
extern void T1753x21004343(GE_context* ac, T0* C);
/* Call to [detachable] ET_REGULAR_FEATURE_CALL.set_parenthesis_call */
extern void T2021x20217858T0T0T0(GE_context* ac, T0* C, T0* a1, T0* a2, T0* a3);
/* Call to [detachable] ET_CALL_WITH_ACTUAL_ARGUMENT_LIST.set_arguments */
extern void T2025x19611650T0(GE_context* ac, T0* C, T0* a1);
/* Call to detachable LX_TRANSITION [[attached] LX_NFA_STATE].record */
extern void T2172x26558468T0(GE_context* ac, T0* C, T0* a1);
extern T0* GE_ma33(T6 c, T6 n, ...);
extern T0* GE_ma1578(T6 c, T6 n, ...);
extern T0* GE_ma1577(T6 c, T6 n, ...);
extern T0* GE_ma291(T6 c, T6 n, ...);
extern T0* GE_ma427(T6 c, T6 n, ...);
extern T0* GE_ma244(T6 c, T6 n, ...);
extern T0* GE_ma561(T6 c, T6 n, ...);
extern T0* GE_mt1978(T0* a1, T0* a2, T8 a3);
extern T0* GE_mt727(T0* a1, T0* a2, T0* a3);
extern T0* GE_mt166(void);
extern T0* GE_mt857(T0* a1);
extern T0* GE_mt31(T6 a1, T6 a2, T6 a3, T0* a4, T0* a5, T0* a6, T0* a7, T0* a8, T0* a9, T6 a10, T1 a11);
extern T0* ge398ov6258689;
extern T0* ge486ov7962687;
extern T0* ge496ov8126519;
extern T0* ge363ov5947436;
extern T0* ge728ov11829252;
extern T0* ge727ov11829252;
extern T0* ge721ov11829252;
extern T0* ge732ov11829252;
extern T0* ge725ov11829252;
extern T0* ge720ov11829252;
extern T0* ge407ov6258689;
extern T0* ge722ov11829254;
extern T0* ge1645ov11829252;
extern T0* ge1630ov11829252;
extern T0* ge1636ov11829252;
extern T0* ge1635ov11829252;
extern T0* ge1634ov11829252;
extern T0* ge1625ov11829252;
extern T0* ge1624ov11829252;
extern T0* ge1642ov11829252;
extern T0* ge1639ov11829252;
extern T0* ge1655ov11829252;
extern T0* ge329ov5390358;
extern T0* ge436ov7143425;
extern T0* ge331ov5423111;
extern T0* ge334ov5472261;
extern T0* ge334ov5472268;
extern T0* ge334ov5472259;
extern T0* ge334ov5472266;
extern T0* ge329ov5390359;
extern T0* ge329ov5390360;
extern T0* ge329ov5390361;
extern T0* ge334ov5472258;
extern T0* ge334ov5472265;
extern T0* ge334ov5472260;
extern T0* ge334ov5472267;
extern T0* ge334ov5472262;
extern T0* ge334ov5472269;
extern T0* ge334ov5472263;
extern T0* ge334ov5472270;
extern T0* ge331ov5423106;
extern T0* ge331ov5423105;
extern T0* ge331ov5423113;
extern T0* ge331ov5423112;
extern T0* ge409ov6258689;
extern T0* ge1496ov24511015;
extern T0* ge1496ov24511020;
extern T0* ge413ov6258689;
extern T0* ge393ov6438920;
extern T0* ge1496ov24511011;
extern T0* ge1496ov24511019;
extern T0* ge1548ov25362439;
extern T0* ge1496ov24511017;
extern T0* ge1496ov24511016;
extern T0* ge1496ov24511018;
extern T0* ge394ov6455312;
extern T0* ge1496ov24511012;
extern T0* ge1496ov24511014;
extern T0* ge1496ov24510639;
extern T0* ge1496ov24510951;
extern T0* ge1496ov24510964;
extern T0* ge392ov6422547;
extern T0* ge392ov6422548;
extern T0* ge782ov12812319;
extern T0* ge782ov12812305;
extern T0* ge782ov12812303;
extern T0* ge782ov12812289;
extern T0* ge782ov12812307;
extern T0* ge782ov12812304;
extern T0* ge782ov12812308;
extern T0* ge782ov12812313;
extern T0* ge782ov12812310;
extern T0* ge782ov12812306;
extern T0* ge782ov12812300;
extern T0* ge782ov12812301;
extern T0* ge782ov12812309;
extern T0* ge782ov12812302;
extern T0* ge782ov12812314;
extern T0* ge782ov12812290;
extern T0* ge782ov12812316;
extern T0* ge782ov12812317;
extern T0* ge782ov12812292;
extern T0* ge782ov12812291;
extern T0* ge782ov12812315;
extern T0* ge782ov12812318;
extern T0* ge782ov12812298;
extern T0* ge782ov12812299;
extern T0* ge782ov12812295;
extern T0* ge782ov12812312;
extern T0* ge782ov12812311;
extern T0* ge782ov12812296;
extern T0* ge782ov12812297;
extern T0* ge782ov12812294;
extern T0* ge782ov12812293;
extern T0* ge778ov12746770;
extern T0* ge1496ov24511021;
extern T0* ge1496ov24511013;
extern T0* ge1496ov24510949;
extern T0* ge408ov6144007;
extern T0* ge408ov6684682;
extern T0* ge395ov6471690;
extern T0* ge1496ov24510673;
extern T0* ge1496ov24510647;
extern T0* ge1496ov24510643;
extern T0* ge1496ov24510645;
extern T0* ge1496ov24510960;
extern T0* ge1496ov24510637;
extern T0* ge1496ov24510681;
extern T0* ge1496ov24510944;
extern T0* ge1496ov24510948;
extern T0* ge1496ov24510945;
extern T0* ge1496ov24510947;
extern T0* ge1496ov24510946;
extern T0* ge1496ov24510952;
extern T0* ge1496ov24510954;
extern T0* ge1496ov24510953;
extern T0* ge1496ov24510955;
extern T0* ge1265ov20725771;
extern T0* ge1490ov24412163;
extern T0* ge1490ov24412162;
extern T0* ge1490ov24412161;
extern T0* ge1271ov20824883;
extern T0* ge1271ov20824868;
extern T0* ge1495ov24494730;
extern T0* ge1495ov24494486;
extern T0* ge1495ov24494222;
extern T0* ge1359ov22265864;
extern T0* ge1366ov22380571;
extern T0* ge1355ov22200360;
extern T0* ge1271ov20824983;
extern T0* ge1271ov20825030;
extern T0* ge1271ov20825009;
extern T0* ge1271ov20825031;
extern T0* ge1271ov20825033;
extern T0* ge1271ov20825016;
extern T0* ge1271ov20825032;
extern T0* ge1271ov20824967;
extern T0* ge1271ov20825028;
extern T0* ge1271ov20824977;
extern T0* ge1271ov20825029;
extern T0* ge1495ov24494753;
extern T0* ge1495ov24494491;
extern T0* ge1495ov24494244;
extern T0* ge1495ov24494746;
extern T0* ge1495ov24494489;
extern T0* ge1495ov24494237;
extern T0* ge1271ov20825131;
extern T0* ge1271ov20825130;
extern T0* ge1271ov20825127;
extern T0* ge1271ov20825126;
extern T0* ge1298ov21266439;
extern T0* ge1271ov20825104;
extern T0* ge1271ov20825107;
extern T0* ge1271ov20825109;
extern T0* ge1271ov20825112;
extern T0* ge1271ov20825128;
extern T0* ge1271ov20825083;
extern T0* ge1271ov20825114;
extern T0* ge1271ov20825115;
extern T0* ge1271ov20825116;
extern T0* ge1271ov20825117;
extern T0* ge1271ov20825118;
extern T0* ge1271ov20825119;
extern T0* ge1271ov20825120;
extern T0* ge1271ov20825121;
extern T0* ge1271ov20825122;
extern T0* ge1271ov20825123;
extern T0* ge1271ov20825124;
extern T0* ge1271ov20825125;
extern T0* ge1271ov20824630;
extern T0* ge1271ov20824997;
extern T0* ge1271ov20824952;
extern T0* ge1495ov24494834;
extern T0* ge1495ov24494534;
extern T0* ge1495ov24494325;
extern T0* ge1495ov24494836;
extern T0* ge1495ov24494535;
extern T0* ge1495ov24494327;
extern T0* ge1495ov24494742;
extern T0* ge1495ov24494233;
extern T0* ge1495ov24494745;
extern T0* ge1495ov24494236;
extern T0* ge1495ov24494744;
extern T0* ge1495ov24494235;
extern T0* ge1495ov24494743;
extern T0* ge1495ov24494234;
extern T0* ge1495ov24494832;
extern T0* ge1495ov24494323;
extern T0* ge1495ov24494833;
extern T0* ge1495ov24494324;
extern T0* ge1495ov24494901;
extern T0* ge1495ov24494567;
extern T0* ge1495ov24494390;
extern T0* ge1495ov24494837;
extern T0* ge1495ov24494536;
extern T0* ge1495ov24494328;
extern T0* ge1495ov24494914;
extern T0* ge1495ov24494579;
extern T0* ge1495ov24494403;
extern T0* ge1495ov24494915;
extern T0* ge1495ov24494404;
extern T0* ge1495ov24494845;
extern T0* ge1495ov24494540;
extern T0* ge1495ov24494334;
extern T0* ge1495ov24494844;
extern T0* ge1495ov24494539;
extern T0* ge1495ov24494333;
extern T0* ge1495ov24494840;
extern T0* ge1495ov24494538;
extern T0* ge1495ov24494331;
extern T0* ge1495ov24494838;
extern T0* ge1495ov24494537;
extern T0* ge1495ov24494329;
extern T0* ge1495ov24494839;
extern T0* ge1495ov24494330;
extern T0* ge1485ov24330254;
extern T0* ge1485ov24330249;
extern T0* ge1485ov24330244;
extern T0* ge1271ov20824633;
extern T0* ge1271ov20824632;
extern T0* ge1271ov20824631;
extern T0* ge1271ov20824629;
extern T0* ge1271ov20824993;
extern T0* ge1271ov20824981;
extern T0* ge1271ov20824628;
extern T0* ge1271ov20824625;
extern T0* ge1271ov20824627;
extern T0* ge1271ov20824626;
extern T0* ge1271ov20824624;
extern T0* ge1271ov20824623;
extern T0* ge1271ov20824622;
extern T0* ge1271ov20824621;
extern T0* ge1271ov20824620;
extern T0* ge1271ov20824619;
extern T0* ge1271ov20824618;
extern T0* ge1271ov20824617;
extern T0* ge1271ov20824614;
extern T0* ge1271ov20824613;
extern T0* ge1271ov20824612;
extern T0* ge1271ov20824611;
extern T0* ge1271ov20824610;
extern T0* ge1271ov20824606;
extern T0* ge1271ov20824604;
extern T0* ge1271ov20824602;
extern T0* ge1271ov20824601;
extern T0* ge1271ov20824600;
extern T0* ge1271ov20824599;
extern T0* ge1271ov20824598;
extern T0* ge1271ov20824597;
extern T0* ge1271ov20824596;
extern T0* ge1271ov20824594;
extern T0* ge1271ov20824592;
extern T0* ge1271ov20824591;
extern T0* ge1271ov20824590;
extern T0* ge1271ov20824587;
extern T0* ge1271ov20824586;
extern T0* ge1271ov20824585;
extern T0* ge1271ov20824584;
extern T0* ge1271ov20824583;
extern T0* ge1271ov20824634;
extern T0* ge1271ov20824607;
extern T0* ge1271ov20824589;
extern T0* ge1271ov20824581;
extern T0* ge1271ov20824970;
extern T0* ge1271ov20824615;
extern T0* ge1485ov24330255;
extern T0* ge1485ov24330250;
extern T0* ge1485ov24330245;
extern T0* ge415ov6094879;
extern T0* ge415ov6094880;
extern T0* ge400ov6553612;
extern T0* ge400ov6553611;
extern T0* ge418ov6094879;
extern T0* ge418ov6094880;
extern T0* ge855ov14008345;
extern T0* ge1271ov20824725;
extern T0* ge1271ov20824635;
extern T0* ge1271ov20824830;
extern T0* ge1271ov20824696;
extern T0* ge1271ov20824802;
extern T0* ge1271ov20824661;
extern T0* ge1271ov20824775;
extern T0* ge1495ov24494910;
extern T0* ge1495ov24494576;
extern T0* ge1495ov24494399;
extern T0* ge1271ov20824942;
extern T0* ge1271ov20824684;
extern T0* ge1495ov24494906;
extern T0* ge1495ov24494572;
extern T0* ge1495ov24494395;
extern T0* ge1495ov24494909;
extern T0* ge1495ov24494575;
extern T0* ge1495ov24494398;
extern T0* ge1271ov20824690;
extern T0* ge1495ov24494725;
extern T0* ge1495ov24494484;
extern T0* ge1495ov24494217;
extern T0* ge1495ov24494726;
extern T0* ge1495ov24494218;
extern T0* ge1495ov24494719;
extern T0* ge1495ov24494481;
extern T0* ge1495ov24494211;
extern T0* ge1495ov24494728;
extern T0* ge1495ov24494485;
extern T0* ge1495ov24494220;
extern T0* ge1495ov24494727;
extern T0* ge1495ov24494219;
extern T0* ge1495ov24494724;
extern T0* ge1495ov24494483;
extern T0* ge1495ov24494216;
extern T0* ge1495ov24494667;
extern T0* ge1495ov24494468;
extern T0* ge1495ov24494161;
extern T0* ge1495ov24494666;
extern T0* ge1495ov24494467;
extern T0* ge1495ov24494160;
extern T0* ge1495ov24494665;
extern T0* ge1495ov24494466;
extern T0* ge1495ov24494159;
extern T0* ge1495ov24494664;
extern T0* ge1495ov24494465;
extern T0* ge1495ov24494158;
extern T0* ge1495ov24494663;
extern T0* ge1495ov24494157;
extern T0* ge1495ov24494601;
extern T0* ge1495ov24494436;
extern T0* ge1495ov24494095;
extern T0* ge1495ov24494599;
extern T0* ge1495ov24494435;
extern T0* ge1495ov24494093;
extern T0* ge1495ov24494621;
extern T0* ge1495ov24494446;
extern T0* ge1495ov24494115;
extern T0* ge1495ov24494622;
extern T0* ge1495ov24494116;
extern T0* ge1495ov24494605;
extern T0* ge1495ov24494439;
extern T0* ge1495ov24494099;
extern T0* ge1495ov24494604;
extern T0* ge1495ov24494098;
extern T0* ge1495ov24494606;
extern T0* ge1495ov24494100;
extern T0* ge1495ov24494630;
extern T0* ge1495ov24494450;
extern T0* ge1495ov24494124;
extern T0* ge1495ov24494619;
extern T0* ge1495ov24494113;
extern T0* ge1495ov24494620;
extern T0* ge1495ov24494114;
extern T0* ge1495ov24494673;
extern T0* ge1495ov24494469;
extern T0* ge1495ov24494167;
extern T0* ge1275ov20889611;
extern T0* ge1495ov24494670;
extern T0* ge1495ov24494164;
extern T0* ge1495ov24494692;
extern T0* ge1495ov24494471;
extern T0* ge1495ov24494186;
extern T0* ge1495ov24494685;
extern T0* ge1495ov24494470;
extern T0* ge1495ov24494179;
extern T0* ge1495ov24494674;
extern T0* ge1495ov24494168;
extern T0* ge1495ov24494671;
extern T0* ge1495ov24494165;
extern T0* ge1495ov24494693;
extern T0* ge1495ov24494187;
extern T0* ge1495ov24494686;
extern T0* ge1495ov24494180;
extern T0* ge1495ov24494675;
extern T0* ge1495ov24494169;
extern T0* ge1495ov24494672;
extern T0* ge1495ov24494166;
extern T0* ge1495ov24494694;
extern T0* ge1495ov24494188;
extern T0* ge1495ov24494687;
extern T0* ge1495ov24494181;
extern T0* ge1495ov24494835;
extern T0* ge1495ov24494326;
extern T0* ge1495ov24494715;
extern T0* ge1495ov24494475;
extern T0* ge1495ov24494207;
extern T0* ge1495ov24494714;
extern T0* ge1495ov24494206;
extern T0* ge1495ov24494710;
extern T0* ge1495ov24494473;
extern T0* ge1495ov24494204;
extern T0* ge1495ov24494676;
extern T0* ge1495ov24494170;
extern T0* ge1495ov24494669;
extern T0* ge1495ov24494163;
extern T0* ge1495ov24494701;
extern T0* ge1495ov24494472;
extern T0* ge1495ov24494195;
extern T0* ge1495ov24494668;
extern T0* ge1495ov24494162;
extern T0* ge1495ov24494700;
extern T0* ge1495ov24494194;
extern T0* ge1495ov24494691;
extern T0* ge1495ov24494185;
extern T0* ge1495ov24494699;
extern T0* ge1495ov24494193;
extern T0* ge1495ov24494684;
extern T0* ge1495ov24494178;
extern T0* ge1495ov24494698;
extern T0* ge1495ov24494192;
extern T0* ge1495ov24494711;
extern T0* ge1495ov24494474;
extern T0* ge1495ov24494205;
extern T0* ge1495ov24494824;
extern T0* ge1495ov24494529;
extern T0* ge1495ov24494315;
extern T0* ge1495ov24494822;
extern T0* ge1495ov24494528;
extern T0* ge1495ov24494313;
extern T0* ge1495ov24494608;
extern T0* ge1495ov24494440;
extern T0* ge1495ov24494102;
extern T0* ge1495ov24494774;
extern T0* ge1495ov24494503;
extern T0* ge1495ov24494265;
extern T0* ge1495ov24494607;
extern T0* ge1495ov24494101;
extern T0* ge1495ov24494773;
extern T0* ge1495ov24494264;
extern T0* ge1271ov20824976;
extern T0* ge1271ov20825022;
extern T0* ge1271ov20824996;
extern T0* ge1495ov24494772;
extern T0* ge1495ov24494502;
extern T0* ge1495ov24494263;
extern T0* ge1495ov24494638;
extern T0* ge1495ov24494455;
extern T0* ge1495ov24494132;
extern T0* ge1495ov24494636;
extern T0* ge1495ov24494453;
extern T0* ge1495ov24494130;
extern T0* ge1495ov24494635;
extern T0* ge1495ov24494129;
extern T0* ge1495ov24494642;
extern T0* ge1495ov24494457;
extern T0* ge1495ov24494136;
extern T0* ge1495ov24494641;
extern T0* ge1495ov24494135;
extern T0* ge1495ov24494643;
extern T0* ge1495ov24494458;
extern T0* ge1495ov24494137;
extern T0* ge1495ov24494679;
extern T0* ge1495ov24494173;
extern T0* ge1495ov24494678;
extern T0* ge1495ov24494172;
extern T0* ge1495ov24494705;
extern T0* ge1495ov24494199;
extern T0* ge1495ov24494677;
extern T0* ge1495ov24494171;
extern T0* ge1495ov24494704;
extern T0* ge1495ov24494198;
extern T0* ge1495ov24494695;
extern T0* ge1495ov24494189;
extern T0* ge1495ov24494703;
extern T0* ge1495ov24494197;
extern T0* ge1495ov24494688;
extern T0* ge1495ov24494182;
extern T0* ge1495ov24494702;
extern T0* ge1495ov24494196;
extern T0* ge1271ov20825132;
extern T0* ge1271ov20825134;
extern T0* ge1271ov20825133;
extern T0* ge1495ov24494905;
extern T0* ge1495ov24494571;
extern T0* ge1495ov24494394;
extern T0* ge1271ov20825102;
extern T0* ge1495ov24494904;
extern T0* ge1495ov24494570;
extern T0* ge1495ov24494393;
extern T0* ge1271ov20824941;
extern T0* ge1271ov20824940;
extern T0* ge1271ov20824939;
extern T0* ge1271ov20824933;
extern T0* ge1271ov20824869;
extern T0* ge1271ov20824852;
extern T0* ge1271ov20824850;
extern T0* ge1271ov20824844;
extern T0* ge1271ov20824842;
extern T0* ge1271ov20824841;
extern T0* ge1271ov20824812;
extern T0* ge1271ov20824786;
extern T0* ge1271ov20824784;
extern T0* ge1271ov20824783;
extern T0* ge1271ov20824781;
extern T0* ge1271ov20824746;
extern T0* ge1271ov20824745;
extern T0* ge1271ov20824744;
extern T0* ge1271ov20824743;
extern T0* ge1271ov20824742;
extern T0* ge1271ov20824741;
extern T0* ge1271ov20824740;
extern T0* ge1271ov20824739;
extern T0* ge1271ov20824724;
extern T0* ge1271ov20824723;
extern T0* ge1271ov20824668;
extern T0* ge1271ov20824667;
extern T0* ge1271ov20824904;
extern T0* ge1271ov20824849;
extern T0* ge1271ov20824829;
extern T0* ge1271ov20824814;
extern T0* ge1271ov20825101;
extern T0* ge1271ov20825098;
extern T0* ge1271ov20825097;
extern T0* ge1271ov20825096;
extern T0* ge1271ov20825095;
extern T0* ge1271ov20825093;
extern T0* ge1271ov20825092;
extern T0* ge1271ov20825091;
extern T0* ge1271ov20824938;
extern T0* ge1271ov20824937;
extern T0* ge1271ov20824936;
extern T0* ge1271ov20824932;
extern T0* ge1271ov20824931;
extern T0* ge1271ov20824765;
extern T0* ge1271ov20824764;
extern T0* ge1271ov20824655;
extern T0* ge1271ov20824654;
extern T0* ge1271ov20824653;
extern T0* ge1271ov20824652;
extern T0* ge1271ov20824651;
extern T0* ge1271ov20824650;
extern T0* ge1271ov20824646;
extern T0* ge1271ov20824645;
extern T0* ge1271ov20824644;
extern T0* ge1271ov20824643;
extern T0* ge1271ov20824642;
extern T0* ge1271ov20824641;
extern T0* ge1271ov20824640;
extern T0* ge1271ov20824639;
extern T0* ge1271ov20825090;
extern T0* ge1271ov20824913;
extern T0* ge1271ov20824867;
extern T0* ge1271ov20824866;
extern T0* ge1271ov20824865;
extern T0* ge1271ov20824864;
extern T0* ge1271ov20824863;
extern T0* ge1271ov20824862;
extern T0* ge1271ov20824861;
extern T0* ge1271ov20824860;
extern T0* ge1271ov20824859;
extern T0* ge1271ov20824858;
extern T0* ge1271ov20824857;
extern T0* ge1271ov20824856;
extern T0* ge1271ov20824855;
extern T0* ge1271ov20824854;
extern T0* ge1271ov20824878;
extern T0* ge1271ov20824875;
extern T0* ge1271ov20824848;
extern T0* ge1271ov20824837;
extern T0* ge1271ov20824827;
extern T0* ge1271ov20824824;
extern T0* ge1271ov20824820;
extern T0* ge1271ov20824817;
extern T0* ge1271ov20824803;
extern T0* ge1271ov20824762;
extern T0* ge1271ov20824759;
extern T0* ge1271ov20824756;
extern T0* ge1271ov20824753;
extern T0* ge1271ov20824685;
extern T0* ge1271ov20824674;
extern T0* ge1271ov20824671;
extern T0* ge1271ov20824659;
extern T0* ge1271ov20824711;
extern T0* ge1271ov20824710;
extern T0* ge1271ov20824951;
extern T0* ge1271ov20824879;
extern T0* ge1271ov20824845;
extern T0* ge1271ov20824801;
extern T0* ge1271ov20824800;
extern T0* ge1271ov20824799;
extern T0* ge1271ov20824798;
extern T0* ge1271ov20824795;
extern T0* ge1271ov20824788;
extern T0* ge1271ov20824782;
extern T0* ge1271ov20824774;
extern T0* ge1271ov20824766;
extern T0* ge1271ov20824763;
extern T0* ge1271ov20824697;
extern T0* ge1271ov20824675;
extern T0* ge1271ov20824656;
extern T0* ge1271ov20824616;
extern T0* ge1271ov20824949;
extern T0* ge1271ov20824948;
extern T0* ge1271ov20824923;
extern T0* ge1271ov20824922;
extern T0* ge1271ov20824921;
extern T0* ge1271ov20824920;
extern T0* ge1271ov20824919;
extern T0* ge1271ov20824918;
extern T0* ge1271ov20824917;
extern T0* ge1271ov20824916;
extern T0* ge1271ov20824915;
extern T0* ge1271ov20824912;
extern T0* ge1271ov20824911;
extern T0* ge1271ov20824910;
extern T0* ge1271ov20824909;
extern T0* ge1271ov20824908;
extern T0* ge1271ov20824907;
extern T0* ge1271ov20824906;
extern T0* ge1271ov20824905;
extern T0* ge1271ov20824902;
extern T0* ge1271ov20824901;
extern T0* ge1271ov20824900;
extern T0* ge1271ov20824899;
extern T0* ge1271ov20824898;
extern T0* ge1271ov20824897;
extern T0* ge1271ov20824896;
extern T0* ge1271ov20824895;
extern T0* ge1271ov20824892;
extern T0* ge1271ov20824891;
extern T0* ge1271ov20824890;
extern T0* ge1271ov20824889;
extern T0* ge1271ov20824888;
extern T0* ge1271ov20824887;
extern T0* ge1271ov20824810;
extern T0* ge1271ov20824806;
extern T0* ge1271ov20824945;
extern T0* ge1271ov20824943;
extern T0* ge1271ov20824928;
extern T0* ge1271ov20824882;
extern T0* ge1271ov20824881;
extern T0* ge1271ov20824880;
extern T0* ge1271ov20824877;
extern T0* ge1271ov20824876;
extern T0* ge1271ov20824874;
extern T0* ge1271ov20824873;
extern T0* ge1271ov20824872;
extern T0* ge1271ov20824871;
extern T0* ge1271ov20824851;
extern T0* ge1271ov20824847;
extern T0* ge1271ov20824846;
extern T0* ge1271ov20824843;
extern T0* ge1271ov20824839;
extern T0* ge1271ov20824838;
extern T0* ge1271ov20824836;
extern T0* ge1271ov20824835;
extern T0* ge1271ov20824834;
extern T0* ge1271ov20824832;
extern T0* ge1271ov20824826;
extern T0* ge1271ov20824825;
extern T0* ge1271ov20824823;
extern T0* ge1271ov20824822;
extern T0* ge1271ov20824819;
extern T0* ge1271ov20824818;
extern T0* ge1271ov20824816;
extern T0* ge1271ov20824815;
extern T0* ge1271ov20824797;
extern T0* ge1271ov20824796;
extern T0* ge1271ov20824794;
extern T0* ge1271ov20824793;
extern T0* ge1271ov20824792;
extern T0* ge1271ov20824791;
extern T0* ge1271ov20824790;
extern T0* ge1271ov20824789;
extern T0* ge1271ov20824785;
extern T0* ge1271ov20824778;
extern T0* ge1271ov20824777;
extern T0* ge1271ov20824776;
extern T0* ge1271ov20824770;
extern T0* ge1271ov20824768;
extern T0* ge1271ov20824761;
extern T0* ge1271ov20824760;
extern T0* ge1271ov20824758;
extern T0* ge1271ov20824757;
extern T0* ge1271ov20824755;
extern T0* ge1271ov20824754;
extern T0* ge1271ov20824752;
extern T0* ge1271ov20824751;
extern T0* ge1271ov20824749;
extern T0* ge1271ov20824733;
extern T0* ge1271ov20824732;
extern T0* ge1271ov20824731;
extern T0* ge1271ov20824729;
extern T0* ge1271ov20824728;
extern T0* ge1271ov20824721;
extern T0* ge1271ov20824719;
extern T0* ge1271ov20824717;
extern T0* ge1271ov20824716;
extern T0* ge1271ov20824714;
extern T0* ge1271ov20824701;
extern T0* ge1271ov20824699;
extern T0* ge1271ov20824698;
extern T0* ge1271ov20824691;
extern T0* ge1271ov20824680;
extern T0* ge1271ov20824677;
extern T0* ge1271ov20824673;
extern T0* ge1271ov20824672;
extern T0* ge1271ov20824670;
extern T0* ge1271ov20824669;
extern T0* ge1271ov20824658;
extern T0* ge1271ov20824657;
extern T0* ge1271ov20824647;
extern T0* ge1271ov20824605;
extern T0* ge1271ov20824705;
extern T0* ge1271ov20824704;
extern T0* ge1271ov20824703;
extern T0* ge1271ov20824702;
extern T0* ge1271ov20824700;
extern T0* ge1271ov20824595;
extern T0* ge1271ov20824692;
extern T0* ge1271ov20824708;
extern T0* ge1271ov20824593;
extern T0* ge1271ov20824903;
extern T0* ge1271ov20824870;
extern T0* ge1271ov20824747;
extern T0* ge1271ov20824663;
extern T0* ge1271ov20824946;
extern T0* ge1271ov20824804;
extern T0* ge1271ov20824787;
extern T0* ge1271ov20824780;
extern T0* ge1271ov20824779;
extern T0* ge1271ov20824769;
extern T0* ge1271ov20824707;
extern T0* ge1271ov20824726;
extern T0* ge1271ov20824722;
extern T0* ge1271ov20824608;
extern T0* ge1271ov20824687;
extern T0* ge1271ov20824686;
extern T0* ge1271ov20824676;
extern T0* ge1271ov20824666;
extern T0* ge1271ov20824665;
extern T0* ge1271ov20824664;
extern T0* ge1271ov20824660;
extern T0* ge1271ov20824588;
extern T0* ge1271ov20824738;
extern T0* ge1271ov20824737;
extern T0* ge1271ov20824638;
extern T0* ge1271ov20824582;
extern T0* ge1271ov20824934;
extern T0* ge1271ov20824772;
extern T0* ge1271ov20824736;
extern T0* ge1271ov20825099;
extern T0* ge1271ov20824828;
extern T0* ge1271ov20824748;
extern T0* ge1271ov20824695;
extern T0* ge1271ov20824694;
extern T0* ge1271ov20824693;
extern T0* ge1271ov20824683;
extern T0* ge1271ov20824682;
extern T0* ge1271ov20825084;
extern T0* ge1271ov20825087;
extern T0* ge1271ov20824821;
extern T0* ge1271ov20824679;
extern T0* ge1271ov20825086;
extern T0* ge1271ov20824893;
extern T0* ge1271ov20824853;
extern T0* ge1271ov20824809;
extern T0* ge1271ov20824709;
extern T0* ge1271ov20824706;
extern T0* ge1271ov20824662;
extern T0* ge1271ov20824648;
extern T0* ge1271ov20824636;
extern T0* ge1271ov20824944;
extern T0* ge1271ov20824885;
extern T0* ge1271ov20824773;
extern T0* ge1271ov20824767;
extern T0* ge1271ov20824735;
extern T0* ge1271ov20824734;
extern T0* ge1271ov20824689;
extern T0* ge1271ov20824925;
extern T0* ge1271ov20824929;
extern T0* ge1271ov20824927;
extern T0* ge1271ov20824926;
extern T0* ge1271ov20824886;
extern T0* ge1271ov20824771;
extern T0* ge1271ov20824730;
extern T0* ge1271ov20824727;
extern T0* ge1271ov20824688;
extern T0* ge1271ov20824681;
extern T0* ge1495ov24494874;
extern T0* ge1495ov24494550;
extern T0* ge1495ov24494363;
extern T0* ge1496ov24510687;
extern T0* ge1496ov24511007;
extern T0* ge1496ov24510990;
extern T0* ge1496ov24510720;
extern T0* ge1496ov24510719;
extern T0* ge1496ov24510627;
extern T0* ge1496ov24510701;
extern T0* ge1496ov24510996;
extern T0* ge1496ov24510629;
extern T0* ge1496ov24510632;
extern T0* ge1496ov24510997;
extern T0* ge1496ov24510981;
extern T0* ge1496ov24510633;
extern T0* ge1496ov24510842;
extern T0* ge1496ov24510735;
extern T0* ge1496ov24510672;
extern T0* ge1496ov24510660;
extern T0* ge1496ov24510728;
extern T0* ge1496ov24510704;
extern T0* ge1495ov24494747;
extern T0* ge1495ov24494238;
extern T0* ge1495ov24494863;
extern T0* ge1495ov24494546;
extern T0* ge1495ov24494352;
extern T0* ge1495ov24494848;
extern T0* ge1495ov24494541;
extern T0* ge1495ov24494337;
extern T0* ge1495ov24494879;
extern T0* ge1495ov24494552;
extern T0* ge1495ov24494368;
extern T0* ge1495ov24494762;
extern T0* ge1495ov24494497;
extern T0* ge1495ov24494253;
extern T0* ge1271ov20825007;
extern T0* ge1495ov24494850;
extern T0* ge1495ov24494542;
extern T0* ge1495ov24494339;
extern T0* ge1495ov24494849;
extern T0* ge1495ov24494338;
extern T0* ge1495ov24494618;
extern T0* ge1495ov24494445;
extern T0* ge1495ov24494112;
extern T0* ge1495ov24494594;
extern T0* ge1495ov24494431;
extern T0* ge1495ov24494088;
extern T0* ge1495ov24494721;
extern T0* ge1495ov24494213;
extern T0* ge1495ov24494847;
extern T0* ge1495ov24494336;
extern T0* ge1495ov24494846;
extern T0* ge1495ov24494335;
extern T0* ge1495ov24494617;
extern T0* ge1495ov24494444;
extern T0* ge1495ov24494111;
extern T0* ge1495ov24494593;
extern T0* ge1495ov24494430;
extern T0* ge1495ov24494087;
extern T0* ge1495ov24494864;
extern T0* ge1495ov24494353;
extern T0* ge1495ov24494878;
extern T0* ge1495ov24494367;
extern T0* ge1495ov24494765;
extern T0* ge1495ov24494498;
extern T0* ge1495ov24494256;
extern T0* ge1271ov20825137;
extern T0* ge1271ov20825140;
extern T0* ge1495ov24494592;
extern T0* ge1495ov24494086;
extern T0* ge1495ov24494855;
extern T0* ge1495ov24494544;
extern T0* ge1495ov24494344;
extern T0* ge1496ov24510959;
extern T0* ge1496ov24510680;
extern T0* ge1496ov24510674;
extern T0* ge1496ov24510648;
extern T0* ge1496ov24510958;
extern T0* ge1496ov24510956;
extern T0* ge1496ov24510743;
extern T0* ge1496ov24510675;
extern T0* ge1496ov24511000;
extern T0* ge1496ov24510770;
extern T0* ge1496ov24510767;
extern T0* ge1496ov24510991;
extern T0* ge1496ov24510864;
extern T0* ge1496ov24510881;
extern T0* ge1496ov24510863;
extern T0* ge1496ov24510880;
extern T0* ge1496ov24510862;
extern T0* ge1496ov24510879;
extern T0* ge1496ov24510839;
extern T0* ge1496ov24510640;
extern T0* ge1496ov24510861;
extern T0* ge1496ov24510878;
extern T0* ge1496ov24510860;
extern T0* ge1496ov24510877;
extern T0* ge1496ov24510859;
extern T0* ge1496ov24510876;
extern T0* ge1496ov24510858;
extern T0* ge1496ov24510875;
extern T0* ge1496ov24510856;
extern T0* ge1496ov24510873;
extern T0* ge1496ov24510855;
extern T0* ge1496ov24510872;
extern T0* ge1496ov24510854;
extern T0* ge1496ov24510871;
extern T0* ge1496ov24510853;
extern T0* ge1496ov24510870;
extern T0* ge1496ov24510852;
extern T0* ge1496ov24510869;
extern T0* ge1496ov24510851;
extern T0* ge1496ov24510868;
extern T0* ge1496ov24510850;
extern T0* ge1496ov24510867;
extern T0* ge1496ov24510742;
extern T0* ge1496ov24510712;
extern T0* ge1496ov24510635;
extern T0* ge1496ov24510669;
extern T0* ge1496ov24510667;
extern T0* ge1496ov24510658;
extern T0* ge1496ov24510657;
extern T0* ge1496ov24510656;
extern T0* ge1496ov24510655;
extern T0* ge1496ov24510654;
extern T0* ge1496ov24510766;
extern T0* ge1496ov24510652;
extern T0* ge1496ov24510670;
extern T0* ge1496ov24510644;
extern T0* ge1496ov24510642;
extern T0* ge1496ov24510688;
extern T0* ge1496ov24510730;
extern T0* ge1496ov24510729;
extern T0* ge1496ov24510919;
extern T0* ge1496ov24510828;
extern T0* ge1496ov24510883;
extern T0* ge1496ov24510882;
extern T0* ge1496ov24510866;
extern T0* ge1496ov24510865;
extern T0* ge1496ov24510849;
extern T0* ge1496ov24510848;
extern T0* ge1496ov24510844;
extern T0* ge1496ov24510845;
extern T0* ge1496ov24510838;
extern T0* ge1496ov24510837;
extern T0* ge1496ov24510836;
extern T0* ge1496ov24510984;
extern T0* ge1496ov24510833;
extern T0* ge1496ov24510826;
extern T0* ge1496ov24510825;
extern T0* ge1496ov24510822;
extern T0* ge1496ov24510818;
extern T0* ge1496ov24510815;
extern T0* ge1496ov24510814;
extern T0* ge1496ov24510813;
extern T0* ge1496ov24510812;
extern T0* ge1496ov24510811;
extern T0* ge1496ov24510810;
extern T0* ge1496ov24510809;
extern T0* ge1496ov24510808;
extern T0* ge1496ov24510780;
extern T0* ge1496ov24510781;
extern T0* ge1496ov24510777;
extern T0* ge1496ov24510779;
extern T0* ge1496ov24510778;
extern T0* ge1496ov24510776;
extern T0* ge1496ov24510774;
extern T0* ge1496ov24510775;
extern T0* ge1496ov24510773;
extern T0* ge1496ov24510772;
extern T0* ge1496ov24510771;
extern T0* ge1496ov24510769;
extern T0* ge1496ov24510768;
extern T0* ge1496ov24510765;
extern T0* ge1496ov24510764;
extern T0* ge1496ov24510763;
extern T0* ge1496ov24510762;
extern T0* ge1496ov24510761;
extern T0* ge1496ov24510760;
extern T0* ge1496ov24510759;
extern T0* ge1496ov24510758;
extern T0* ge1496ov24510748;
extern T0* ge1496ov24510746;
extern T0* ge1496ov24510745;
extern T0* ge1496ov24510744;
extern T0* ge1496ov24510741;
extern T0* ge1496ov24510740;
extern T0* ge1496ov24510739;
extern T0* ge1496ov24510738;
extern T0* ge1496ov24510737;
extern T0* ge1496ov24510747;
extern T0* ge1496ov24510830;
extern T0* ge1496ov24510829;
extern T0* ge1496ov24510827;
extern T0* ge1496ov24510718;
extern T0* ge1496ov24510717;
extern T0* ge1496ov24510716;
extern T0* ge1496ov24510715;
extern T0* ge1496ov24510714;
extern T0* ge1496ov24510713;
extern T0* ge1496ov24510698;
extern T0* ge1496ov24510697;
extern T0* ge1496ov24510695;
extern T0* ge1496ov24510831;
extern T0* ge1496ov24510749;
extern T0* ge1496ov24510570;
extern T0* ge1496ov24510706;
extern T0* ge1496ov24510705;
extern T0* ge1496ov24510709;
extern T0* ge1496ov24510708;
extern T0* ge1496ov24510707;
extern T0* ge1496ov24510711;
extern T0* ge1496ov24510782;
extern T0* ge1496ov24510694;
extern T0* ge1496ov24510693;
extern T0* ge1496ov24510857;
extern T0* ge1496ov24510874;
extern T0* ge1496ov24510974;
extern T0* ge1496ov24510982;
extern T0* ge1496ov24510796;
extern T0* ge1496ov24510795;
extern T0* ge1496ov24510941;
extern T0* ge1496ov24510723;
extern T0* ge1496ov24510998;
extern T0* ge1496ov24510841;
extern T0* ge1496ov24511006;
extern T0* ge1271ov20825136;
extern T0* ge1271ov20825139;
extern T0* ge1265ov20725769;
extern T0* ge1271ov20825138;
extern T0* ge1265ov20725770;
extern T0* ge1496ov24510985;
extern T0* ge1496ov24510700;
extern T0* ge1496ov24510699;
extern T0* ge1496ov24510724;
extern T0* ge1496ov24510940;
extern T0* ge1496ov24510689;
extern T0* ge1496ov24510925;
extern T0* ge1495ov24494720;
extern T0* ge1495ov24494212;
extern T0* ge1495ov24494595;
extern T0* ge1495ov24494089;
extern T0* ge1495ov24494723;
extern T0* ge1495ov24494482;
extern T0* ge1495ov24494215;
extern T0* ge1495ov24494716;
extern T0* ge1495ov24494478;
extern T0* ge1495ov24494208;
extern T0* ge1495ov24494718;
extern T0* ge1495ov24494480;
extern T0* ge1495ov24494210;
extern T0* ge1271ov20825036;
extern T0* ge1271ov20825037;
extern T0* ge1271ov20825038;
extern T0* ge1271ov20825039;
extern T0* ge1271ov20825040;
extern T0* ge1271ov20825044;
extern T0* ge1271ov20825045;
extern T0* ge1271ov20825046;
extern T0* ge1271ov20825047;
extern T0* ge1271ov20825048;
extern T0* ge1271ov20825049;
extern T0* ge1271ov20825050;
extern T0* ge1271ov20825051;
extern T0* ge1271ov20825053;
extern T0* ge1271ov20825054;
extern T0* ge1271ov20825056;
extern T0* ge1271ov20825057;
extern T0* ge1271ov20825058;
extern T0* ge1271ov20825059;
extern T0* ge1271ov20825060;
extern T0* ge1271ov20825061;
extern T0* ge1271ov20825062;
extern T0* ge1271ov20825063;
extern T0* ge1271ov20825065;
extern T0* ge1271ov20825066;
extern T0* ge1271ov20825071;
extern T0* ge1271ov20825072;
extern T0* ge1271ov20825073;
extern T0* ge1271ov20825074;
extern T0* ge1271ov20825075;
extern T0* ge1271ov20825076;
extern T0* ge1271ov20825077;
extern T0* ge1271ov20825078;
extern T0* ge1271ov20825080;
extern T0* ge1271ov20825081;
extern T0* ge1495ov24494936;
extern T0* ge1495ov24494588;
extern T0* ge1495ov24494425;
extern T0* ge1495ov24494892;
extern T0* ge1495ov24494561;
extern T0* ge1495ov24494381;
extern T0* ge1496ov24510754;
extern T0* ge1496ov24510755;
extern T0* ge1496ov24510793;
extern T0* ge1496ov24510756;
extern T0* ge1496ov24510702;
extern T0* ge1496ov24510794;
extern T0* ge1496ov24510757;
extern T0* ge1496ov24510804;
extern T0* ge1496ov24510805;
extern T0* ge1496ov24510806;
extern T0* ge1496ov24510703;
extern T0* ge1496ov24510807;
extern T0* ge1495ov24494938;
extern T0* ge1495ov24494427;
extern T0* ge1495ov24494895;
extern T0* ge1495ov24494384;
extern T0* ge1496ov24510797;
extern T0* ge1496ov24510751;
extern T0* ge1496ov24510750;
extern T0* ge1495ov24494937;
extern T0* ge1495ov24494426;
extern T0* ge1495ov24494894;
extern T0* ge1495ov24494383;
extern T0* ge1495ov24494888;
extern T0* ge1495ov24494558;
extern T0* ge1495ov24494377;
extern T0* ge1496ov24510986;
extern T0* ge1496ov24510978;
extern T0* ge1496ov24510979;
extern T0* ge1496ov24510682;
extern T0* ge1496ov24510628;
extern T0* ge1496ov24510980;
extern T0* ge1496ov24510677;
extern T0* ge1495ov24494889;
extern T0* ge1495ov24494378;
extern T0* ge1495ov24494808;
extern T0* ge1495ov24494519;
extern T0* ge1495ov24494299;
extern T0* ge1495ov24494798;
extern T0* ge1495ov24494517;
extern T0* ge1495ov24494289;
extern T0* ge1495ov24494797;
extern T0* ge1495ov24494288;
extern T0* ge1495ov24494796;
extern T0* ge1495ov24494516;
extern T0* ge1495ov24494287;
extern T0* ge1495ov24494795;
extern T0* ge1495ov24494286;
extern T0* ge1495ov24494794;
extern T0* ge1495ov24494515;
extern T0* ge1495ov24494285;
extern T0* ge1495ov24494793;
extern T0* ge1495ov24494284;
extern T0* ge1495ov24494875;
extern T0* ge1495ov24494364;
extern T0* ge1495ov24494805;
extern T0* ge1495ov24494518;
extern T0* ge1495ov24494296;
extern T0* ge1495ov24494803;
extern T0* ge1495ov24494294;
extern T0* ge1495ov24494800;
extern T0* ge1495ov24494291;
extern T0* ge1495ov24494799;
extern T0* ge1495ov24494290;
extern T0* ge1495ov24494825;
extern T0* ge1495ov24494316;
extern T0* ge1495ov24494820;
extern T0* ge1495ov24494527;
extern T0* ge1495ov24494311;
extern T0* ge1495ov24494860;
extern T0* ge1495ov24494349;
extern T0* ge1495ov24494807;
extern T0* ge1495ov24494298;
extern T0* ge1495ov24494862;
extern T0* ge1495ov24494545;
extern T0* ge1495ov24494351;
extern T0* ge1495ov24494764;
extern T0* ge1495ov24494255;
extern T0* ge1495ov24494935;
extern T0* ge1495ov24494587;
extern T0* ge1495ov24494424;
extern T0* ge1495ov24494933;
extern T0* ge1495ov24494422;
extern T0* ge1495ov24494934;
extern T0* ge1495ov24494423;
extern T0* ge1495ov24494920;
extern T0* ge1495ov24494581;
extern T0* ge1495ov24494409;
extern T0* ge1495ov24494918;
extern T0* ge1495ov24494407;
extern T0* ge1495ov24494919;
extern T0* ge1495ov24494408;
extern T0* ge1495ov24494655;
extern T0* ge1495ov24494462;
extern T0* ge1495ov24494149;
extern T0* ge1495ov24494654;
extern T0* ge1495ov24494148;
extern T0* ge1495ov24494930;
extern T0* ge1495ov24494585;
extern T0* ge1495ov24494419;
extern T0* ge1495ov24494928;
extern T0* ge1495ov24494417;
extern T0* ge1495ov24494929;
extern T0* ge1495ov24494418;
extern T0* ge1495ov24494657;
extern T0* ge1495ov24494463;
extern T0* ge1495ov24494151;
extern T0* ge1495ov24494656;
extern T0* ge1495ov24494150;
extern T0* ge1495ov24494931;
extern T0* ge1495ov24494586;
extern T0* ge1495ov24494420;
extern T0* ge1495ov24494932;
extern T0* ge1495ov24494421;
extern T0* ge1495ov24494921;
extern T0* ge1495ov24494582;
extern T0* ge1495ov24494410;
extern T0* ge1495ov24494922;
extern T0* ge1495ov24494411;
extern T0* ge1495ov24494658;
extern T0* ge1495ov24494464;
extern T0* ge1495ov24494152;
extern T0* ge1495ov24494649;
extern T0* ge1495ov24494461;
extern T0* ge1495ov24494143;
extern T0* ge1495ov24494651;
extern T0* ge1495ov24494145;
extern T0* ge1495ov24494926;
extern T0* ge1495ov24494584;
extern T0* ge1495ov24494415;
extern T0* ge1495ov24494927;
extern T0* ge1495ov24494416;
extern T0* ge1495ov24494916;
extern T0* ge1495ov24494580;
extern T0* ge1495ov24494405;
extern T0* ge1495ov24494917;
extern T0* ge1495ov24494406;
extern T0* ge1495ov24494763;
extern T0* ge1495ov24494254;
extern T0* ge1496ov24510659;
extern T0* ge1496ov24510943;
extern T0* ge1496ov24510942;
extern T0* ge1496ov24510917;
extern T0* ge1496ov24510916;
extern T0* ge1496ov24510915;
extern T0* ge1496ov24510914;
extern T0* ge1496ov24510913;
extern T0* ge1496ov24510912;
extern T0* ge1496ov24510911;
extern T0* ge1496ov24510910;
extern T0* ge1496ov24510909;
extern T0* ge1496ov24510908;
extern T0* ge1496ov24510907;
extern T0* ge1496ov24510906;
extern T0* ge1496ov24510905;
extern T0* ge1496ov24510904;
extern T0* ge1496ov24510903;
extern T0* ge1496ov24510902;
extern T0* ge1496ov24510901;
extern T0* ge1496ov24510900;
extern T0* ge1496ov24510899;
extern T0* ge1496ov24510898;
extern T0* ge1496ov24510897;
extern T0* ge1496ov24510896;
extern T0* ge1496ov24510895;
extern T0* ge1496ov24510894;
extern T0* ge1496ov24510892;
extern T0* ge1496ov24510891;
extern T0* ge1496ov24510890;
extern T0* ge1496ov24510889;
extern T0* ge1496ov24510888;
extern T0* ge1496ov24510887;
extern T0* ge1496ov24510792;
extern T0* ge1496ov24510786;
extern T0* ge1496ov24510726;
extern T0* ge1496ov24510832;
extern T0* ge1496ov24510710;
extern T0* ge1496ov24510975;
extern T0* ge1496ov24510973;
extern T0* ge1496ov24510995;
extern T0* ge1496ov24510634;
extern T0* ge1496ov24510631;
extern T0* ge1495ov24494886;
extern T0* ge1495ov24494556;
extern T0* ge1495ov24494375;
extern T0* ge1495ov24494596;
extern T0* ge1495ov24494432;
extern T0* ge1495ov24494090;
extern T0* ge1495ov24494785;
extern T0* ge1495ov24494509;
extern T0* ge1495ov24494276;
extern T0* ge1495ov24494784;
extern T0* ge1495ov24494275;
extern T0* ge1495ov24494783;
extern T0* ge1495ov24494274;
extern T0* ge1495ov24494782;
extern T0* ge1495ov24494273;
extern T0* ge1495ov24494781;
extern T0* ge1495ov24494272;
extern T0* ge1495ov24494780;
extern T0* ge1495ov24494508;
extern T0* ge1495ov24494271;
extern T0* ge1496ov24511009;
extern T0* ge1496ov24510732;
extern T0* ge1495ov24494866;
extern T0* ge1495ov24494548;
extern T0* ge1495ov24494355;
extern T0* ge1495ov24494867;
extern T0* ge1495ov24494356;
extern T0* ge1495ov24494865;
extern T0* ge1495ov24494547;
extern T0* ge1495ov24494354;
extern T0* ge1495ov24494613;
extern T0* ge1495ov24494443;
extern T0* ge1495ov24494107;
extern T0* ge1495ov24494612;
extern T0* ge1495ov24494106;
extern T0* ge1495ov24494611;
extern T0* ge1495ov24494442;
extern T0* ge1495ov24494105;
extern T0* ge1495ov24494859;
extern T0* ge1495ov24494348;
extern T0* ge1495ov24494615;
extern T0* ge1495ov24494109;
extern T0* ge1495ov24494616;
extern T0* ge1495ov24494110;
extern T0* ge1495ov24494609;
extern T0* ge1495ov24494441;
extern T0* ge1495ov24494103;
extern T0* ge1495ov24494788;
extern T0* ge1495ov24494512;
extern T0* ge1495ov24494279;
extern T0* ge1495ov24494789;
extern T0* ge1495ov24494280;
extern T0* ge1495ov24494787;
extern T0* ge1495ov24494511;
extern T0* ge1495ov24494278;
extern T0* ge1496ov24510638;
extern T0* ge1496ov24510736;
extern T0* ge1495ov24494722;
extern T0* ge1495ov24494214;
extern T0* ge1495ov24494717;
extern T0* ge1495ov24494479;
extern T0* ge1495ov24494209;
extern T0* ge1495ov24494854;
extern T0* ge1495ov24494343;
extern T0* ge1495ov24494758;
extern T0* ge1495ov24494495;
extern T0* ge1495ov24494249;
extern T0* ge1495ov24494645;
extern T0* ge1495ov24494460;
extern T0* ge1495ov24494139;
extern T0* ge1495ov24494646;
extern T0* ge1495ov24494140;
extern T0* ge1495ov24494759;
extern T0* ge1495ov24494250;
extern T0* ge1495ov24494760;
extern T0* ge1495ov24494251;
extern T0* ge1495ov24494647;
extern T0* ge1495ov24494141;
extern T0* ge1495ov24494653;
extern T0* ge1495ov24494147;
extern T0* ge1495ov24494761;
extern T0* ge1495ov24494496;
extern T0* ge1495ov24494252;
extern T0* ge1495ov24494757;
extern T0* ge1495ov24494494;
extern T0* ge1495ov24494248;
extern T0* ge1495ov24494756;
extern T0* ge1495ov24494247;
extern T0* ge1495ov24494597;
extern T0* ge1495ov24494433;
extern T0* ge1495ov24494091;
extern T0* ge1495ov24494598;
extern T0* ge1495ov24494434;
extern T0* ge1495ov24494092;
extern T0* ge1495ov24494899;
extern T0* ge1495ov24494564;
extern T0* ge1495ov24494388;
extern T0* ge1496ov24510963;
extern T0* ge1496ov24510999;
extern T0* ge1496ov24510683;
extern T0* ge1496ov24510686;
extern T0* ge1496ov24510957;
extern T0* ge1496ov24510885;
extern T0* ge1495ov24494883;
extern T0* ge1495ov24494554;
extern T0* ge1495ov24494372;
extern T0* ge1495ov24494881;
extern T0* ge1495ov24494553;
extern T0* ge1495ov24494370;
extern T0* ge947ov15777794;
extern T0* ge1495ov24494880;
extern T0* ge1495ov24494369;
extern T0* ge1495ov24494884;
extern T0* ge1495ov24494373;
extern T0* ge1495ov24494882;
extern T0* ge1495ov24494371;
extern T0* ge1495ov24494806;
extern T0* ge1495ov24494297;
extern T0* ge1495ov24494804;
extern T0* ge1495ov24494295;
extern T0* ge1495ov24494802;
extern T0* ge1495ov24494293;
extern T0* ge1495ov24494801;
extern T0* ge1495ov24494292;
extern T0* ge1495ov24494829;
extern T0* ge1495ov24494532;
extern T0* ge1495ov24494320;
extern T0* ge1495ov24494831;
extern T0* ge1495ov24494533;
extern T0* ge1495ov24494322;
extern T0* ge1495ov24494821;
extern T0* ge1495ov24494312;
extern T0* ge1496ov24510884;
extern T0* ge1496ov24511003;
extern T0* ge1496ov24510886;
extern T0* ge1496ov24510733;
extern T0* ge1496ov24510783;
extern T0* ge1496ov24510988;
extern T0* ge1496ov24510972;
extern T0* ge1496ov24511008;
extern T0* ge1496ov24510803;
extern T0* ge1496ov24510840;
extern T0* ge1496ov24510977;
extern T0* ge1496ov24510992;
extern T0* ge1496ov24510989;
extern T0* ge1496ov24510685;
extern T0* ge1496ov24510785;
extern T0* ge1496ov24510692;
extern T0* ge1496ov24510993;
extern T0* ge1495ov24494661;
extern T0* ge1495ov24494155;
extern T0* ge1495ov24494786;
extern T0* ge1495ov24494510;
extern T0* ge1495ov24494277;
extern T0* ge1495ov24494877;
extern T0* ge1495ov24494551;
extern T0* ge1495ov24494366;
extern T0* ge1495ov24494876;
extern T0* ge1495ov24494365;
extern T0* ge1495ov24494872;
extern T0* ge1495ov24494549;
extern T0* ge1495ov24494361;
extern T0* ge1495ov24494871;
extern T0* ge1495ov24494360;
extern T0* ge1495ov24494870;
extern T0* ge1495ov24494359;
extern T0* ge1495ov24494869;
extern T0* ge1495ov24494358;
extern T0* ge1495ov24494868;
extern T0* ge1495ov24494357;
extern T0* ge1495ov24494873;
extern T0* ge1495ov24494362;
extern T0* ge1495ov24494591;
extern T0* ge1495ov24494429;
extern T0* ge1495ov24494085;
extern T0* ge1496ov24510971;
extern T0* ge1495ov24494648;
extern T0* ge1495ov24494146;
extern T0* ge1495ov24494650;
extern T0* ge1495ov24494144;
extern T0* ge1495ov24494142;
extern T0* ge1271ov20824955;
extern T0* ge1495ov24494858;
extern T0* ge1495ov24494347;
extern T0* ge1495ov24494851;
extern T0* ge1495ov24494543;
extern T0* ge1495ov24494340;
extern T0* ge1495ov24494856;
extern T0* ge1495ov24494345;
extern T0* ge1495ov24494610;
extern T0* ge1495ov24494104;
extern T0* ge1495ov24494893;
extern T0* ge1495ov24494382;
extern T0* ge1495ov24494898;
extern T0* ge1495ov24494563;
extern T0* ge1495ov24494387;
extern T0* ge1495ov24494896;
extern T0* ge1495ov24494562;
extern T0* ge1495ov24494385;
extern T0* ge1495ov24494891;
extern T0* ge1495ov24494560;
extern T0* ge1495ov24494380;
extern T0* ge1495ov24494890;
extern T0* ge1495ov24494559;
extern T0* ge1495ov24494379;
extern T0* ge1496ov24510787;
extern T0* ge1496ov24510966;
extern T0* ge1496ov24510696;
extern T0* ge1495ov24494748;
extern T0* ge1495ov24494239;
extern T0* ge1495ov24494791;
extern T0* ge1495ov24494513;
extern T0* ge1495ov24494282;
extern T0* ge1495ov24494792;
extern T0* ge1495ov24494514;
extern T0* ge1495ov24494283;
extern T0* ge1495ov24494790;
extern T0* ge1495ov24494281;
extern T0* ge1495ov24494861;
extern T0* ge1495ov24494350;
extern T0* ge1495ov24494766;
extern T0* ge1495ov24494257;
extern T0* ge1495ov24494925;
extern T0* ge1495ov24494583;
extern T0* ge1495ov24494414;
extern T0* ge1495ov24494923;
extern T0* ge1495ov24494412;
extern T0* ge1495ov24494924;
extern T0* ge1495ov24494413;
extern T0* ge1496ov24510798;
extern T0* ge1495ov24494659;
extern T0* ge1495ov24494153;
extern T0* ge1495ov24494857;
extern T0* ge1495ov24494346;
extern T0* ge1496ov24510961;
extern T0* ge1496ov24510790;
extern T0* ge1496ov24510791;
extern T0* ge1496ov24510976;
extern T0* ge1496ov24510987;
extern T0* ge1496ov24510684;
extern T0* ge1496ov24511002;
extern T0* ge1496ov24510691;
extern T0* ge1496ov24510816;
extern T0* ge1496ov24510962;
extern T0* ge1496ov24510918;
extern T0* ge1495ov24494853;
extern T0* ge1495ov24494342;
extern T0* ge1495ov24494826;
extern T0* ge1495ov24494530;
extern T0* ge1495ov24494317;
extern T0* ge1495ov24494827;
extern T0* ge1495ov24494531;
extern T0* ge1495ov24494318;
extern T0* ge1495ov24494823;
extern T0* ge1495ov24494314;
extern T0* ge1496ov24510821;
extern T0* ge1496ov24510801;
extern T0* ge1496ov24510802;
extern T0* ge1496ov24510843;
extern T0* ge1496ov24510834;
extern T0* ge1495ov24494660;
extern T0* ge1495ov24494154;
extern T0* ge1495ov24494817;
extern T0* ge1495ov24494526;
extern T0* ge1495ov24494308;
extern T0* ge1495ov24494815;
extern T0* ge1495ov24494524;
extern T0* ge1495ov24494306;
extern T0* ge1495ov24494816;
extern T0* ge1495ov24494307;
extern T0* ge1495ov24494814;
extern T0* ge1495ov24494523;
extern T0* ge1495ov24494305;
extern T0* ge1495ov24494812;
extern T0* ge1495ov24494522;
extern T0* ge1495ov24494303;
extern T0* ge1495ov24494813;
extern T0* ge1495ov24494304;
extern T0* ge1495ov24494810;
extern T0* ge1495ov24494521;
extern T0* ge1495ov24494301;
extern T0* ge1495ov24494811;
extern T0* ge1495ov24494302;
extern T0* ge1495ov24494809;
extern T0* ge1495ov24494520;
extern T0* ge1495ov24494300;
extern T0* ge1495ov24494852;
extern T0* ge1495ov24494341;
extern T0* ge1495ov24494885;
extern T0* ge1495ov24494555;
extern T0* ge1495ov24494374;
extern T0* ge1495ov24494662;
extern T0* ge1495ov24494156;
extern T0* ge1495ov24494624;
extern T0* ge1495ov24494447;
extern T0* ge1495ov24494118;
extern T0* ge1495ov24494623;
extern T0* ge1495ov24494117;
extern T0* ge1495ov24494629;
extern T0* ge1495ov24494123;
extern T0* ge1495ov24494626;
extern T0* ge1495ov24494448;
extern T0* ge1495ov24494120;
extern T0* ge1495ov24494627;
extern T0* ge1495ov24494121;
extern T0* ge1495ov24494771;
extern T0* ge1495ov24494262;
extern T0* ge1495ov24494625;
extern T0* ge1495ov24494119;
extern T0* ge1495ov24494639;
extern T0* ge1495ov24494133;
extern T0* ge1495ov24494628;
extern T0* ge1495ov24494449;
extern T0* ge1495ov24494122;
extern T0* ge1495ov24494633;
extern T0* ge1495ov24494451;
extern T0* ge1495ov24494127;
extern T0* ge1495ov24494631;
extern T0* ge1495ov24494125;
extern T0* ge1495ov24494632;
extern T0* ge1495ov24494126;
extern T0* ge1495ov24494779;
extern T0* ge1495ov24494507;
extern T0* ge1495ov24494270;
extern T0* ge1495ov24494776;
extern T0* ge1495ov24494504;
extern T0* ge1495ov24494267;
extern T0* ge1495ov24494775;
extern T0* ge1495ov24494266;
extern T0* ge1495ov24494777;
extern T0* ge1495ov24494505;
extern T0* ge1495ov24494268;
extern T0* ge1495ov24494634;
extern T0* ge1495ov24494452;
extern T0* ge1495ov24494128;
extern T0* ge1495ov24494640;
extern T0* ge1495ov24494456;
extern T0* ge1495ov24494134;
extern T0* ge1495ov24494768;
extern T0* ge1495ov24494500;
extern T0* ge1495ov24494259;
extern T0* ge1495ov24494754;
extern T0* ge1495ov24494492;
extern T0* ge1495ov24494245;
extern T0* ge1495ov24494778;
extern T0* ge1495ov24494506;
extern T0* ge1495ov24494269;
extern T0* ge1495ov24494637;
extern T0* ge1495ov24494454;
extern T0* ge1495ov24494131;
extern T0* ge1495ov24494644;
extern T0* ge1495ov24494459;
extern T0* ge1495ov24494138;
extern T0* ge1495ov24494769;
extern T0* ge1495ov24494501;
extern T0* ge1495ov24494260;
extern T0* ge1495ov24494767;
extern T0* ge1495ov24494499;
extern T0* ge1495ov24494258;
extern T0* ge1495ov24494755;
extern T0* ge1495ov24494493;
extern T0* ge1495ov24494246;
extern T0* ge1495ov24494770;
extern T0* ge1495ov24494261;
extern T0* ge1271ov20824965;
extern T0* ge1486ov24346630;
extern T0* ge1486ov24346628;
extern T0* ge1486ov24346626;
extern T0* ge1485ov24330253;
extern T0* ge1485ov24330248;
extern T0* ge1485ov24330243;
extern T0* ge1495ov24494913;
extern T0* ge1495ov24494578;
extern T0* ge1495ov24494402;
extern T0* ge1271ov20824960;
extern T0* ge1271ov20825004;
extern T0* ge1271ov20824956;
extern T0* ge1271ov20824959;
extern T0* ge1271ov20825018;
extern T0* ge1271ov20825003;
extern T0* ge1271ov20824979;
extern T0* ge1271ov20825005;
extern T0* ge1271ov20825019;
extern T0* ge1271ov20824963;
extern T0* ge1271ov20824975;
extern T0* ge1271ov20824978;
extern T0* ge1271ov20825026;
extern T0* ge1271ov20824972;
extern T0* ge1271ov20824969;
extern T0* ge1271ov20824999;
extern T0* ge1271ov20824988;
extern T0* ge1271ov20825012;
extern T0* ge1271ov20824968;
extern T0* ge1271ov20825024;
extern T0* ge1271ov20824995;
extern T0* ge1271ov20824982;
extern T0* ge1271ov20825011;
extern T0* ge1492ov24444931;
extern T0* ge1492ov24444930;
extern T0* ge1271ov20824998;
extern T0* ge1271ov20824961;
extern T0* ge1271ov20824987;
extern T0* ge1271ov20824971;
extern T0* ge1271ov20825008;
extern T0* ge1271ov20824962;
extern T0* ge1271ov20824984;
extern T0* ge1271ov20825010;
extern T0* ge1495ov24494912;
extern T0* ge1495ov24494401;
extern T0* ge1271ov20824992;
extern T0* ge1507ov24690739;
extern T0* ge1271ov20825085;
extern T0* ge1271ov20825088;
extern T0* ge1271ov20825089;
extern T0* ge1271ov20825094;
extern T0* ge1271ov20825100;
extern T0* ge1271ov20825103;
extern T0* ge1271ov20824603;
extern T0* ge1271ov20824609;
extern T0* ge1271ov20824637;
extern T0* ge1271ov20824649;
extern T0* ge1271ov20824678;
extern T0* ge1271ov20824712;
extern T0* ge1271ov20824713;
extern T0* ge1271ov20824715;
extern T0* ge1271ov20824718;
extern T0* ge1271ov20824720;
extern T0* ge1271ov20824750;
extern T0* ge1271ov20824805;
extern T0* ge1271ov20824807;
extern T0* ge1271ov20824808;
extern T0* ge1271ov20824811;
extern T0* ge1271ov20824813;
extern T0* ge1271ov20824831;
extern T0* ge1271ov20824833;
extern T0* ge1271ov20824840;
extern T0* ge1271ov20824884;
extern T0* ge1271ov20824894;
extern T0* ge1271ov20824914;
extern T0* ge1271ov20824924;
extern T0* ge1271ov20824930;
extern T0* ge1271ov20824935;
extern T0* ge1271ov20824947;
extern T0* ge1271ov20824950;
extern T0* ge1271ov20824953;
extern T0* ge1271ov20824954;
extern T0* ge1271ov20824957;
extern T0* ge1271ov20824958;
extern T0* ge1271ov20824966;
extern T0* ge1271ov20824973;
extern T0* ge1271ov20824974;
extern T0* ge1271ov20824980;
extern T0* ge1271ov20824985;
extern T0* ge1271ov20824986;
extern T0* ge1271ov20824989;
extern T0* ge1271ov20824990;
extern T0* ge1271ov20824991;
extern T0* ge1271ov20824994;
extern T0* ge1271ov20825000;
extern T0* ge1271ov20825001;
extern T0* ge1271ov20825002;
extern T0* ge1271ov20825013;
extern T0* ge1271ov20825014;
extern T0* ge1271ov20825015;
extern T0* ge1271ov20825017;
extern T0* ge1271ov20825020;
extern T0* ge1271ov20825021;
extern T0* ge1271ov20825023;
extern T0* ge1271ov20825025;
extern T0* ge1271ov20825027;
extern T0* ge496ov8126476;
extern T0* ge1507ov24690760;
extern T0* ge1495ov24494750;
extern T0* ge1495ov24494476;
extern T0* ge1495ov24494241;
extern T0* ge1495ov24494603;
extern T0* ge1495ov24494438;
extern T0* ge1495ov24494097;
extern T0* ge1495ov24494602;
extern T0* ge1495ov24494437;
extern T0* ge1495ov24494096;
extern T0* ge1495ov24494900;
extern T0* ge1495ov24494565;
extern T0* ge1495ov24494389;
extern T0* ge1495ov24494752;
extern T0* ge1495ov24494477;
extern T0* ge1495ov24494243;
extern T0* ge1495ov24494903;
extern T0* ge1495ov24494569;
extern T0* ge1495ov24494392;
extern T0* ge1495ov24494902;
extern T0* ge1495ov24494568;
extern T0* ge1495ov24494391;
extern T0* ge1495ov24494749;
extern T0* ge1495ov24494490;
extern T0* ge1495ov24494240;
extern T0* ge1495ov24494751;
extern T0* ge1495ov24494242;
extern T0* ge1495ov24494739;
extern T0* ge1495ov24494488;
extern T0* ge1495ov24494230;
extern T0* ge1495ov24494738;
extern T0* ge1495ov24494229;
extern T0* ge1495ov24494740;
extern T0* ge1495ov24494231;
extern T0* ge1495ov24494741;
extern T0* ge1495ov24494232;
extern T0* ge1495ov24494729;
extern T0* ge1495ov24494221;
extern T0* ge1495ov24494683;
extern T0* ge1495ov24494177;
extern T0* ge1495ov24494682;
extern T0* ge1495ov24494176;
extern T0* ge1495ov24494697;
extern T0* ge1495ov24494191;
extern T0* ge1495ov24494690;
extern T0* ge1495ov24494184;
extern T0* ge1495ov24494736;
extern T0* ge1495ov24494487;
extern T0* ge1495ov24494228;
extern T0* ge1495ov24494735;
extern T0* ge1495ov24494227;
extern T0* ge1495ov24494681;
extern T0* ge1495ov24494175;
extern T0* ge1495ov24494709;
extern T0* ge1495ov24494203;
extern T0* ge1495ov24494680;
extern T0* ge1495ov24494174;
extern T0* ge1495ov24494708;
extern T0* ge1495ov24494202;
extern T0* ge1495ov24494696;
extern T0* ge1495ov24494190;
extern T0* ge1495ov24494707;
extern T0* ge1495ov24494201;
extern T0* ge1495ov24494689;
extern T0* ge1495ov24494183;
extern T0* ge1495ov24494706;
extern T0* ge1495ov24494200;
extern T0* ge1495ov24494731;
extern T0* ge1495ov24494223;
extern T0* ge1495ov24494734;
extern T0* ge1495ov24494226;
extern T0* ge1495ov24494733;
extern T0* ge1495ov24494225;
extern T0* ge1495ov24494732;
extern T0* ge1495ov24494224;
extern T0* ge393ov6438922;
extern T0* ge1496ov24510725;
extern T0* ge1496ov24510817;
extern T0* ge1496ov24510753;
extern T0* ge1496ov24510784;
extern T0* ge1496ov24510835;
extern T0* ge1496ov24510893;
extern T0* ge1496ov24510799;
extern T0* ge1496ov24510920;
extern T0* ge1496ov24510983;
extern T0* ge1496ov24510938;
extern T0* ge1496ov24510939;
extern T0* ge1496ov24510934;
extern T0* ge1496ov24510921;
extern T0* ge1496ov24510922;
extern T0* ge1496ov24510923;
extern T0* ge1496ov24510926;
extern T0* ge1496ov24510927;
extern T0* ge1496ov24510928;
extern T0* ge1496ov24510929;
extern T0* ge1496ov24510930;
extern T0* ge1496ov24510931;
extern T0* ge1496ov24510932;
extern T0* ge1496ov24510933;
extern T0* ge1496ov24510935;
extern T0* ge1496ov24510936;
extern T0* ge1496ov24510937;
extern T0* ge1496ov24510924;
extern T0* ge1496ov24510676;
extern T0* ge1496ov24510752;
extern T0* ge1496ov24510722;
extern T0* ge1496ov24510727;
extern T0* ge1496ov24510800;
extern T0* ge1496ov24510788;
extern T0* ge1496ov24510847;
extern T0* ge1496ov24510965;
extern T0* ge1496ov24510789;
extern T0* ge1496ov24510846;
extern T0* ge1496ov24510820;
extern T0* ge1496ov24510819;
extern T0* ge1496ov24510824;
extern T0* ge1496ov24510823;
extern T0* ge411ov6258689;
extern T0* ge1496ov24510641;
extern T0* ge1496ov24510994;
extern T0* ge1496ov24511005;
extern T0* ge1496ov24510671;
extern T0* ge1496ov24511001;
extern T0* ge1496ov24510665;
extern T0* ge1496ov24510664;
extern T0* ge1496ov24510663;
extern T0* ge1496ov24510662;
extern T0* ge1496ov24510653;
extern T0* ge1496ov24510651;
extern T0* ge1496ov24510650;
extern T0* ge1496ov24510646;
extern T0* ge1496ov24510668;
extern T0* ge1496ov24510661;
extern T0* ge1496ov24510649;
extern T0* ge1496ov24510678;
extern T0* ge1493ov24461329;
extern T0* ge1493ov24461323;
extern T0* ge1493ov24461317;
extern T0* ge1493ov24461325;
extern T0* ge1493ov24461319;
extern T0* ge1493ov24461313;
extern T0* ge1493ov24461328;
extern T0* ge1493ov24461322;
extern T0* ge1493ov24461316;
extern T0* ge1493ov24461330;
extern T0* ge1493ov24461324;
extern T0* ge1493ov24461318;
extern T0* ge1493ov24461327;
extern T0* ge1493ov24461321;
extern T0* ge1493ov24461315;
extern T0* ge1495ov24494911;
extern T0* ge1495ov24494577;
extern T0* ge1495ov24494400;
extern T0* ge1495ov24494908;
extern T0* ge1495ov24494574;
extern T0* ge1495ov24494397;
extern T0* ge1495ov24494907;
extern T0* ge1495ov24494573;
extern T0* ge1495ov24494396;
extern T0* ge1493ov24461326;
extern T0* ge1493ov24461320;
extern T0* ge1493ov24461314;
extern T0* ge1494ov24477706;
extern T0* ge1494ov24477702;
extern T0* ge1494ov24477701;
extern T0* ge1494ov24477705;
extern T0* ge1494ov24477700;
extern T0* ge1494ov24477704;
extern T0* ge1494ov24477699;
extern T0* ge1494ov24477703;
extern T0* ge1494ov24477698;
extern T0* ge1660ov27197457;
extern T0* ge1668ov27328535;
extern T0* ge540ov8847364;
extern T0* ge1485ov24330252;
extern T0* ge1485ov24330247;
extern T0* ge1485ov24330242;
extern T0* ge1167ov19120179;
extern T0* ge1167ov19120180;
extern T0* ge855ov14008338;
extern T0* ge855ov14008379;
extern T0* ge846ov13860874;
extern T0* ge846ov13860875;
extern T0* ge846ov13860876;
extern T0* ge846ov13860885;
extern T0* ge802ov13139970;
extern T0* ge802ov13139977;
extern T0* ge802ov13139976;
extern T0* ge802ov13139971;
extern T0* ge802ov13139975;
extern T0* ge802ov13139972;
extern T0* ge802ov13139978;
extern T0* ge855ov14008330;
extern T0* ge855ov14008335;
extern T0* ge855ov14008362;
extern T0* ge855ov14008346;
extern T0* ge874ov14319702;
extern T0* ge874ov14319673;
extern T0* ge874ov14319639;
extern T0* ge874ov14319675;
extern T0* ge874ov14319696;
extern T0* ge874ov14319662;
extern T0* ge874ov14319686;
extern T0* ge868ov14221378;
extern T0* ge868ov14221343;
extern T0* ge868ov14221387;
extern T0* ge868ov14221352;
extern T0* ge868ov14221370;
extern T0* ge868ov14221335;
extern T0* ge877ov14368837;
extern T0* ge877ov14368836;
extern T0* ge874ov14319646;
extern T0* ge874ov14319647;
extern T0* ge874ov14319619;
extern T0* ge874ov14319626;
extern T0* ge874ov14319630;
extern T0* ge874ov14319648;
extern T0* ge874ov14319649;
extern T0* ge874ov14319650;
extern T0* ge874ov14319651;
extern T0* ge874ov14319652;
extern T0* ge874ov14319653;
extern T0* ge874ov14319654;
extern T0* ge874ov14319657;
extern T0* ge874ov14319678;
extern T0* ge874ov14319661;
extern T0* ge874ov14319665;
extern T0* ge874ov14319677;
extern T0* ge874ov14319681;
extern T0* ge874ov14319682;
extern T0* ge874ov14319683;
extern T0* ge874ov14319691;
extern T0* ge874ov14319692;
extern T0* ge874ov14319704;
extern T0* ge874ov14319705;
extern T0* ge874ov14319617;
extern T0* ge855ov14008340;
extern T0* ge846ov13860873;
extern T0* ge846ov13860884;
extern T0* ge846ov13860883;
extern T0* ge846ov13860911;
extern T0* ge846ov13860914;
extern T0* ge846ov13860913;
extern T0* ge846ov13860886;
extern T0* ge846ov13860902;
extern T0* ge846ov13860877;
extern T0* ge846ov13860871;
extern T0* ge802ov13139969;
extern T0* ge855ov14008329;
extern T0* ge855ov14008357;
extern T0* ge874ov14319664;
extern T0* ge874ov14319670;
extern T0* ge874ov14319689;
extern T0* ge874ov14319625;
extern T0* ge874ov14319685;
extern T0* ge874ov14319632;
extern T0* ge874ov14319636;
extern T0* ge874ov14319697;
extern T0* ge868ov14221376;
extern T0* ge868ov14221341;
extern T0* ge868ov14221373;
extern T0* ge868ov14221338;
extern T0* ge874ov14319703;
extern T0* ge855ov14008331;
extern T0* ge855ov14008369;
extern T0* ge855ov14008375;
extern T0* ge855ov14008370;
extern T0* ge874ov14319618;
extern T0* ge874ov14319627;
extern T0* ge874ov14319655;
extern T0* ge874ov14319663;
extern T0* ge874ov14319635;
extern T0* ge874ov14319628;
extern T0* ge874ov14319690;
extern T0* ge874ov14319706;
extern T0* ge874ov14319642;
extern T0* ge874ov14319684;
extern T0* ge874ov14319700;
extern T0* ge874ov14319629;
extern T0* ge874ov14319656;
extern T0* ge874ov14319676;
extern T0* ge874ov14319674;
extern T0* ge868ov14221371;
extern T0* ge868ov14221336;
extern T0* ge874ov14319634;
extern T0* ge874ov14319707;
extern T0* ge874ov14319637;
extern T0* ge874ov14319640;
extern T0* ge874ov14319641;
extern T0* ge874ov14319672;
extern T0* ge874ov14319680;
extern T0* ge874ov14319669;
extern T0* ge874ov14319667;
extern T0* ge874ov14319698;
extern T0* ge874ov14319633;
extern T0* ge874ov14319671;
extern T0* ge868ov14221360;
extern T0* ge868ov14221325;
extern T0* ge868ov14221359;
extern T0* ge868ov14221324;
extern T0* ge868ov14221358;
extern T0* ge868ov14221323;
extern T0* ge868ov14221357;
extern T0* ge868ov14221322;
extern T0* ge868ov14221356;
extern T0* ge868ov14221321;
extern T0* ge874ov14319645;
extern T0* ge868ov14221375;
extern T0* ge868ov14221340;
extern T0* ge868ov14221377;
extern T0* ge868ov14221342;
extern T0* ge874ov14319666;
extern T0* ge874ov14319631;
extern T0* ge874ov14319693;
extern T0* ge874ov14319709;
extern T0* ge874ov14319624;
extern T0* ge874ov14319638;
extern T0* ge874ov14319708;
extern T0* ge846ov13860887;
extern T0* ge846ov13860906;
extern T0* ge846ov13860910;
extern T0* ge846ov13860915;
extern T0* ge846ov13860916;
extern T0* ge846ov13860891;
extern T0* ge874ov14319643;
extern T0* ge846ov13860904;
extern T0* ge874ov14319679;
extern T0* ge874ov14319701;
extern T0* ge874ov14319659;
extern T0* ge874ov14319687;
extern T0* ge874ov14319658;
extern T0* ge874ov14319644;
extern T0* ge874ov14319660;
extern T0* ge874ov14319694;
extern T0* ge874ov14319699;
extern T0* ge874ov14319620;
extern T0* ge874ov14319621;
extern T0* ge874ov14319622;
extern T0* ge874ov14319623;
extern T0* ge846ov13860878;
extern T0* ge846ov13860909;
extern T0* ge846ov13860889;
extern T0* ge846ov13860872;
extern T0* ge846ov13860879;
extern T0* ge846ov13860881;
extern T0* ge846ov13860882;
extern T0* ge846ov13860865;
extern T0* ge846ov13860866;
extern T0* ge846ov13860867;
extern T0* ge846ov13860868;
extern T0* ge846ov13860869;
extern T0* ge846ov13860870;
extern T0* ge846ov13860888;
extern T0* ge846ov13860890;
extern T0* ge846ov13860892;
extern T0* ge846ov13860893;
extern T0* ge846ov13860894;
extern T0* ge846ov13860895;
extern T0* ge846ov13860896;
extern T0* ge846ov13860897;
extern T0* ge846ov13860898;
extern T0* ge846ov13860899;
extern T0* ge846ov13860900;
extern T0* ge846ov13860901;
extern T0* ge855ov14008321;
extern T0* ge855ov14008322;
extern T0* ge855ov14008323;
extern T0* ge855ov14008324;
extern T0* ge855ov14008325;
extern T0* ge855ov14008326;
extern T0* ge855ov14008327;
extern T0* ge855ov14008328;
extern T0* ge855ov14008365;
extern T0* ge855ov14008332;
extern T0* ge855ov14008333;
extern T0* ge855ov14008334;
extern T0* ge855ov14008339;
extern T0* ge855ov14008341;
extern T0* ge855ov14008342;
extern T0* ge855ov14008343;
extern T0* ge855ov14008344;
extern T0* ge855ov14008347;
extern T0* ge855ov14008366;
extern T0* ge855ov14008353;
extern T0* ge855ov14008354;
extern T0* ge855ov14008368;
extern T0* ge855ov14008356;
extern T0* ge855ov14008358;
extern T0* ge855ov14008359;
extern T0* ge855ov14008363;
extern T0* ge855ov14008364;
extern T0* ge868ov14221372;
extern T0* ge868ov14221337;
extern T0* ge868ov14221374;
extern T0* ge868ov14221339;
extern T0* ge868ov14221368;
extern T0* ge868ov14221333;
extern T0* ge802ov13139980;
extern T0* ge802ov13139973;
extern T0* ge802ov13139974;
extern T0* ge868ov14221362;
extern T0* ge868ov14221327;
extern T0* ge868ov14221361;
extern T0* ge868ov14221326;
extern T0* ge855ov14008382;
extern T0* ge855ov14008380;
extern T0* ge812ov13303813;
extern T0* ge812ov13303814;
extern T0* ge812ov13303815;
extern T0* ge855ov14008371;
extern T0* ge855ov14008383;
extern T0* ge868ov14221379;
extern T0* ge868ov14221344;
extern T0* ge868ov14221380;
extern T0* ge868ov14221345;
extern T0* ge868ov14221381;
extern T0* ge868ov14221346;
extern T0* ge868ov14221369;
extern T0* ge868ov14221334;
extern T0* ge868ov14221382;
extern T0* ge868ov14221347;
extern T0* ge868ov14221355;
extern T0* ge868ov14221320;
extern T0* ge868ov14221384;
extern T0* ge868ov14221349;
extern T0* ge868ov14221385;
extern T0* ge868ov14221350;
extern T0* ge868ov14221386;
extern T0* ge868ov14221351;
extern T0* ge868ov14221383;
extern T0* ge868ov14221348;
extern T0* ge874ov14319695;
extern T0* ge874ov14319688;
extern T0* ge868ov14221354;
extern T0* ge868ov14221319;
extern T0* ge868ov14221363;
extern T0* ge868ov14221328;
extern T0* ge868ov14221364;
extern T0* ge868ov14221329;
extern T0* ge868ov14221367;
extern T0* ge868ov14221332;
extern T0* ge868ov14221366;
extern T0* ge868ov14221331;
extern T0* ge868ov14221365;
extern T0* ge868ov14221330;
extern T0* ge1764ov28901418;
extern T0* ge1764ov28901380;
extern T0* ge1764ov28901378;
extern T0* ge1764ov28901419;
extern T0* ge1764ov28901391;
extern T0* ge1764ov28901390;
extern T0* ge1764ov28901401;
extern T0* ge1764ov28901395;
extern T0* ge1764ov28901394;
extern T0* ge1764ov28901393;
extern T0* ge1764ov28901399;
extern T0* ge1764ov28901398;
extern T0* ge1764ov28901400;
extern T0* ge1764ov28901377;
extern T0* ge1764ov28901403;
extern T0* ge1764ov28901412;
extern T0* ge1767ov28950555;
extern T0* ge1767ov28950553;
extern T0* ge1767ov28950554;
extern T0* ge1764ov28901413;
extern T0* ge1764ov28901414;
extern T0* ge1764ov28901417;
extern T0* ge1764ov28901415;
extern T0* ge1764ov28901416;
extern T0* ge1731ov28360717;
extern T0* ge1731ov28360716;
extern T0* ge1764ov28901410;
extern T0* ge1760ov28835841;
extern T0* ge1760ov28835842;
extern T0* ge1764ov28901383;
extern T0* ge1762ov28868612;
extern T0* ge1762ov28868614;
extern T0* ge1762ov28868611;
extern T0* ge1762ov28868613;
extern T0* ge1764ov28901406;
extern T0* ge1775ov29081617;
extern T0* ge1775ov29081616;
extern T0* ge1764ov28901426;
extern T0* ge1764ov28901427;
extern T0* ge1764ov28901428;
extern T0* ge1764ov28901421;
extern T0* ge1764ov28901404;
extern T0* ge1764ov28901405;
extern T0* ge1764ov28901407;
extern T0* ge1764ov28901411;
extern T0* ge1771ov29016084;
extern T0* ge1771ov29016083;
extern T0* ge1819ov29802497;
extern T0* ge1548ov25362441;
extern T0* ge1548ov25362437;
extern T0* ge1548ov25362433;
extern T0* ge1548ov25362435;
extern T0* geic24510465;
extern T0* geic24510466;
extern T0* geic24510467;
extern T0* geic13156353;
extern void GE_init_const1(void);
extern void GE_init_const2(void);
extern void GE_init_const3(void);
extern void GE_init_const4(void);
extern void GE_init_const5(void);
extern void GE_init_const6(void);
extern void GE_init_const7(void);
extern void GE_init_const8(void);
extern void GE_init_const9(void);
extern void GE_init_const10(void);
extern void GE_init_const11(void);
extern void GE_init_const12(void);
extern void GE_init_const13(void);
extern void GE_init_const14(void);
extern void GE_init_const15(void);
extern void GE_init_const16(void);
extern void GE_init_const17(void);
extern void GE_init_const18(void);
extern void GE_init_const19(void);
extern void GE_init_const20(void);
extern void GE_init_const21(void);
extern void GE_init_const22(void);
extern void GE_init_const23(void);
extern void GE_init_const(void);

#ifdef __cplusplus
}
#endif


/*
	description:

		"C functions used to implement class C_DATE"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2006-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef GE_TIME_H
#define GE_TIME_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#if defined _XOPEN_UNIX || defined __USE_XOPEN || defined __USE_BSD
#define HAS_GETTIMEOFDAY
#else
#ifndef WIN32
#define HAS_GETTIMEOFDAY
#endif
#endif

#ifdef HAS_GETTIMEOFDAY
/*
	OpenBSD does not support 'ftime' unless -lcompat is specified.
	Use the recommended XSI (X/Open System Interface) extension instead.
*/
#include <sys/time.h>
#include <sys/types.h>
#define GE_ftime(p) gettimeofday((struct timeval*)(p),((void*)0))
#define GE_timebsz sizeof(struct timeval)
#define GE_timebmillitm(p) (((struct timeval*)(p))->tv_usec/1000)
#define GE_timebtime(p) ((struct timeval*)(p))->tv_sec
#else
#include <sys/timeb.h>
extern void ftime(struct timeb *); /* Needed for lcc-win32 */
#define GE_ftime(p) ftime((struct timeb*)(p))
#define GE_timebsz sizeof(struct timeb)
#define GE_timebmillitm(p) ((struct timeb*)(p))->millitm
#define GE_timebtime(p) ((struct timeb*)(p))->time
#endif

#endif
/*
	description:

		"Part of ISE Eiffel runtime. Needed to compile the EiffelCOM library."

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2010-2018, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef EIF_CECIL_H
#define EIF_CECIL_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

typedef void (*EIF_PROCEDURE)(EIF_REFERENCE, ...);
#define EIF_PROC EIF_PROCEDURE
typedef EIF_INTEGER_32 (*EIF_INTEGER_32_FUNCTION)(EIF_REFERENCE, ...); /* Returns an Eiffel Integer 32 bits */
#define EIF_INTEGER_FUNCTION EIF_INTEGER_32_FUNCTION
typedef EIF_REFERENCE (*EIF_REFERENCE_FUNCTION)(EIF_REFERENCE, ...); /* Returns an Eiffel Reference */
typedef EIF_POINTER (*EIF_POINTER_FUNCTION)(EIF_REFERENCE, ...); /* Returns an Eiffel Pointer */
typedef EIF_BOOLEAN (*EIF_BOOLEAN_FUNCTION)(EIF_REFERENCE, ...); /* Returns an Eiffel Boolean */
typedef EIF_CHARACTER_8 (*EIF_CHARACTER_8_FUNCTION)(EIF_REFERENCE, ...); /* Returns char */
#define EIF_CHARACTER_FUNCTION EIF_CHARACTER_8_FUNCTION
typedef EIF_REAL_64 (*EIF_REAL_64_FUNCTION)(EIF_REFERENCE, ...); /* Returns an Eiffel Double */
#define EIF_DOUBLE_FUNCTION EIF_REAL_64_FUNCTION
typedef EIF_REAL_32 (*EIF_REAL_32_FUNCTION)(EIF_REFERENCE, ...); /* Returns an Eiffel Real */
#define EIF_REAL_FUNCTION EIF_REAL_32_FUNCTION

extern EIF_PROCEDURE eif_procedure(char* rout, EIF_TYPE_ID cid);
#define eif_proc eif_procedure /* Use `eif_procedure' instead */
extern EIF_INTEGER_FUNCTION eif_integer_function(char* rout, EIF_TYPE_ID cid);
extern EIF_REFERENCE_FUNCTION eif_reference_function(char* rout, EIF_TYPE_ID cid);
extern EIF_POINTER_FUNCTION eif_pointer_function(char* rout, EIF_TYPE_ID cid);
extern EIF_BOOLEAN_FUNCTION eif_boolean_function(char* rout, EIF_TYPE_ID cid);
extern EIF_REAL_64_FUNCTION eif_real_64_function(char* rout, EIF_TYPE_ID cid);
#define eif_double_function(rout,cid) eif_real_64_function((rout),(cid))
extern EIF_REAL_32_FUNCTION eif_real_32_function(char* rout, EIF_TYPE_ID cid);
#define eif_real_function(rout,cid) eif_real_32_function((rout),(cid))
extern EIF_CHARACTER_8_FUNCTION eif_character_8_function(char* rout, EIF_TYPE_ID cid);
#define eif_character_function(rout,cid) eif_character_8_function((rout),(cid))
extern EIF_OBJECT eif_create(EIF_TYPE_ID cid);
extern EIF_TYPE_ID eif_type_id(char* type_string);
extern EIF_INTEGER eifaddr_offset(EIF_REFERENCE object, char *name, int * const ret);
#define eif_string(s) RTMS(s)
#define makestr(s,c) RTMS_EX((s),(c))
#define eifaddr(object,name,ret) ((void *) ((char *) object + eifaddr_offset (object, name, ret)))
#define eif_field(object,name,type) *(type *)(eifaddr(object,name, NULL)) /* Obsolete. Use "eif_attribute" instead. */
#define eif_attribute(object,name,type,ret) *(type *)(eifaddr(object,name,ret)) /* Returns the attribute of an object. Return status in "ret".*/

/* Miscellaneous useful functions. */

/* Make an Eiffel array from a C array:
 * `eif_array' is the direct reference to the Eiffel array.
 * `c_array' is the C array.
 * `nelts' the number of elements to copy in the eiffel array, it has to
 * be equal to `eif_array.count'.
 * type is an Eiffel type.
 */
#define eif_make_from_c(eif_array, c_array, nelts, type) \
	{ \
		EIF_REFERENCE area = eif_field(eif_array, "area", EIF_REFERENCE); \
		memcpy((type *)area, c_array, nelts * sizeof(type));\
	}

#define eif_type eiftype /* Dynamic type ID */
extern EIF_TYPE_ID eiftype(EIF_OBJECT object); /* Give dynamic type of EIF_OBJECT. Obsolete, use "eif_type_by_object". */
extern EIF_TYPE_ID eif_type_by_reference(EIF_REFERENCE object);
extern EIF_TYPE_INDEX eif_gen_param_id(EIF_TYPE_INDEX dftype, uint32_t pos);

#ifdef __cplusplus
}
#endif

#endif
/*
	description:

		"Part of ISE Eiffel runtime. Needed to compile the EiffelCOM library."

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2010-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef EIF_PLUG_H
#define EIF_PLUG_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifdef __cplusplus
extern "C" {
#endif

extern int nstcall;	/* Nested call global variable: signals a nested call and
					 * trigger an invariant check in generated C routines  */

#ifdef __cplusplus
}
#endif

#endif
/*
	description:

		"C functions used to implement Thread support"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2007-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef EIF_THREADS_H
#define EIF_THREADS_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef EIF_CECIL_H
#include "eif_cecil.h"
#endif
#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif
#ifdef GE_USE_THREADS
#ifndef GE_THREAD_H
#include "ge_thread.h"
#endif
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifndef GE_USE_THREADS

/*
 * Empty stubs for EiffelThread library so that it may be compiled against a non-multithreaded run-time.
 */

/* THREAD_ENVIRONMENT */
#define eif_thr_thread_id() NULL

/* THREAD_CONTROL */
#define eif_thr_yield()
#define eif_thr_join_all()

/* THREAD */
#define eif_thr_create_with_attr(current_obj, init_func, set_terminated_func, attr)
#define eif_thr_wait(obj,get_terminated)
#define eif_thr_wait_with_timeout(obj,get_terminated,tms) EIF_TRUE
#define eif_thr_last_thread() NULL
#define eif_thr_exit()

/* THREAD_ATTRIBUTES */
#define eif_thr_default_priority() 0
#define eif_thr_min_priority() 0
#define eif_thr_max_priority() 0

/* MUTEX */
#define eif_thr_mutex_create() NULL
#define eif_thr_mutex_lock(a_mutex_pointer)
#define eif_thr_mutex_trylock(a_mutex_pointer) EIF_TRUE
#define eif_thr_mutex_unlock(a_mutex_pointer)
#define eif_thr_mutex_destroy(a_mutex_pointer)

/* CONDITION_VARIABLE */
#define eif_thr_cond_create() NULL
#define eif_thr_cond_broadcast(a_cond_ptr)
#define eif_thr_cond_signal(a_cond_ptr)
#define eif_thr_cond_wait(a_cond_ptr,a_mutex_ptr)
#define eif_thr_cond_wait_with_timeout(a_cond_ptr,a_mutex_ptr,a_timeout) 1
#define eif_thr_cond_destroy(a_cond_ptr)

/* SEMAPHORE */
#define eif_thr_sem_create(a_count) NULL
#define eif_thr_sem_wait(a_sem_pointer)
#define eif_thr_sem_post(a_sem_pointer)
#define eif_thr_sem_trywait(a_sem_pointer) EIF_TRUE
#define eif_thr_sem_destroy(a_sem_pointer)

/* READ_WRITE_LOCK */
#define eif_thr_rwl_create() NULL
#define eif_thr_rwl_rdlock(a_read_write_lock_pointer)
#define eif_thr_rwl_unlock(a_read_write_lock_pointer)
#define eif_thr_rwl_wrlock(a_read_write_lock_pointer)
#define eif_thr_rwl_destroy(a_read_write_lock_pointer)

#else

/* THREAD_ENVIRONMENT */
#define eif_thr_thread_id() GE_thread_id()

/* THREAD_CONTROL */
#define eif_thr_yield() GE_thread_yield()
#define eif_thr_join_all() GE_thread_join_all()

/* THREAD */
#define eif_thr_create_with_attr(current_obj,init_func,set_terminated_func,attr) GE_thread_create_with_attr((current_obj),(void (*)(EIF_REFERENCE))(init_func),(void (*)(EIF_REFERENCE,EIF_BOOLEAN))(set_terminated_func),(EIF_THR_ATTR_TYPE*)(attr))
#define eif_thr_wait(obj,get_terminated) GE_thread_wait((obj),(EIF_BOOLEAN (*)(EIF_REFERENCE))(get_terminated))
#define eif_thr_wait_with_timeout(obj,get_terminated,tms) GE_thread_wait_with_timeout((obj),(EIF_BOOLEAN (*)(EIF_REFERENCE))(get_terminated),(tms))
#define eif_thr_last_thread() GE_last_thread_created()
#define eif_thr_exit() GE_thread_exit()

/* THREAD_ATTRIBUTES */
#define eif_thr_default_priority() GE_thread_default_priority()
#define eif_thr_min_priority() GE_thread_min_priority()
#define eif_thr_max_priority() GE_thread_max_priority()

/* MUTEX */
#define eif_thr_mutex_create() GE_mutex_create()
#define eif_thr_mutex_lock(a_mutex_pointer) GE_mutex_lock(a_mutex_pointer)
#define eif_thr_mutex_trylock(a_mutex_pointer) GE_mutex_try_lock(a_mutex_pointer)
#define eif_thr_mutex_unlock(a_mutex_pointer) GE_mutex_unlock(a_mutex_pointer)
#define eif_thr_mutex_destroy(a_mutex_pointer) GE_mutex_destroy(a_mutex_pointer)

/* CONDITION_VARIABLE */
#define eif_thr_cond_create() GE_condition_variable_create()
#define eif_thr_cond_broadcast(a_cond_ptr) GE_condition_variable_broadcast(a_cond_ptr)
#define eif_thr_cond_signal(a_cond_ptr) GE_condition_variable_signal(a_cond_ptr)
#define eif_thr_cond_wait(a_cond_ptr,a_mutex_ptr) GE_condition_variable_wait((a_cond_ptr),(a_mutex_ptr))
#define eif_thr_cond_wait_with_timeout(a_cond_ptr,a_mutex_ptr,a_timeout) GE_condition_variable_wait_with_timeout((a_cond_ptr),(a_mutex_ptr),(a_timeout))
#define eif_thr_cond_destroy(a_cond_ptr) GE_condition_variable_destroy(a_cond_ptr)

/* SEMAPHORE */
#define eif_thr_sem_create(a_count) GE_semaphore_create(a_count)
#define eif_thr_sem_wait(a_sem_pointer) GE_semaphore_wait(a_sem_pointer)
#define eif_thr_sem_post(a_sem_pointer) GE_semaphore_post(a_sem_pointer)
#define eif_thr_sem_trywait(a_sem_pointer) GE_semaphore_try_wait(a_sem_pointer)
#define eif_thr_sem_destroy(a_sem_pointer) GE_semaphore_destroy(a_sem_pointer)

/* READ_WRITE_LOCK */
#define eif_thr_rwl_create() GE_read_write_lock_create()
#define eif_thr_rwl_rdlock(a_read_write_lock_pointer) GE_read_write_lock_read_lock(a_read_write_lock_pointer)
#define eif_thr_rwl_unlock(a_read_write_lock_pointer) GE_read_write_lock_unlock(a_read_write_lock_pointer)
#define eif_thr_rwl_wrlock(a_read_write_lock_pointer) GE_read_write_lock_write_lock(a_read_write_lock_pointer)
#define eif_thr_rwl_destroy(a_read_write_lock_pointer) GE_read_write_lock_destroy(a_read_write_lock_pointer)

/* WEL */
#ifdef EIF_WINDOWS
#define eif_thr_create_wel_per_thread_data(a_size) GE_thread_create_wel_per_thread_data(a_size)
#endif

#endif

#ifdef __cplusplus
}
#endif

#endif
/*
	description:

		"C functions used to implement class CONSOLE"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2006-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef EIF_CONSOLE_H
#define EIF_CONSOLE_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

extern EIF_POINTER console_def(EIF_INTEGER file);
extern EIF_BOOLEAN console_eof(FILE* fp);
extern EIF_CHARACTER_8 console_separator(FILE* f);
extern void console_ps(FILE* f, char* str, EIF_INTEGER len);
extern void console_pr(FILE* f, EIF_REAL_32 number);
extern void console_pc(FILE* f, EIF_CHARACTER_8 c);
extern void console_pd(FILE* f, EIF_REAL_64 val);
extern void console_pi(FILE* f, EIF_INTEGER number);
extern void console_tnwl(FILE* f);
extern EIF_CHARACTER_8 console_readchar(FILE* f);
extern EIF_REAL_32 console_readreal(FILE* f);
extern EIF_INTEGER console_readint(FILE* f);
extern EIF_REAL_64 console_readdouble(FILE* f);
extern EIF_INTEGER console_readword(FILE* f, char* s, EIF_INTEGER bound, EIF_INTEGER start);
extern EIF_INTEGER console_readline(FILE* f, char* s, EIF_INTEGER bound, EIF_INTEGER start);
extern void console_next_line(FILE* f);
extern EIF_INTEGER console_readstream(FILE* f, char* s, EIF_INTEGER bound);
extern void console_file_close (FILE* f);

#ifdef __cplusplus
}
#endif

#endif
/*
	description:

		"C functions used to implement class FILE"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2006-2019, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef EIF_FILE_H
#define EIF_FILE_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#include <time.h>
#include <sys/stat.h>

#ifdef __cplusplus
extern "C" {
#endif

/* Let's define the stat structure for our platforms. */
/* The definition is the same for both ANSI and Unicode versions on Windows. */
#ifdef EIF_WINDOWS
#	ifdef __LCC__
#		define rt_stat_buf	struct stat
#	elif defined EIF_64_BITS
#		define rt_stat_buf	struct _stat64
#	else
#		define rt_stat_buf	struct _stat64i32
#	endif
#else
#	define rt_stat_buf		struct stat
#endif

/*
 * Create directory `dirname'.
 */
extern void eif_file_mkdir(EIF_FILENAME dirname);

/*
 * Rename file `from' into `to'.
 */
extern void eif_file_rename(EIF_FILENAME from, EIF_FILENAME  to);

/*
 * Link file `from' into `to'.
 */
extern void eif_file_link(EIF_FILENAME from, EIF_FILENAME to);

/*
 * Delete file or directory `name'.
 */
extern void eif_file_unlink(EIF_FILENAME name);

/*
 * Open file `name' with the corresponding type `how'.
 */
extern EIF_POINTER eif_file_open(EIF_FILENAME name, int how);

/*
 * Open file `fd' with the corresponding type `how'.
 */
extern EIF_POINTER eif_file_dopen(int fd, int how);

/*
 * Reopen file `name' with the corresponding type `how' and substitute that
 * to the old stream described by `old'. This is useful to redirect 'stdout'
 * to another place, for instance.
 */
extern EIF_POINTER eif_file_reopen(EIF_FILENAME name, int how, FILE *old);

/*
 * Close the file.
 */
extern void eif_file_close(FILE *fp);

/*
 * Flush data held in stdio buffer.
 */
extern void eif_file_flush(FILE *fp);

/*
 * Return the associated file descriptor.
 */
extern EIF_INTEGER eif_file_fd(FILE *f);

/*
 * Get a character from `f'.
 */
extern EIF_CHARACTER_8 eif_file_gc(FILE *f);

/*
 * Get a string from `f' and fill it into `s' (at most `bound' characters),
 * with `start' being the amount of bytes already stored within s. This
 * means we really have to read (bound - start) characters.
 */
extern EIF_INTEGER eif_file_gs(FILE *f, char *s, EIF_INTEGER bound, EIF_INTEGER start);

/*
 * Read min (bound, remaining bytes in file) characters into `s' and
 * return the number of characters read.
 */
extern EIF_INTEGER eif_file_gss(FILE *f, char *s, EIF_INTEGER bound);

/*
 * Get a word from `f' and fill it into `s' (at most `bound' characters),
 * with `start' being the amount of bytes already stored within s. This
 * means we really have to read (bound - start) characters. Any leading
 * spaces are skipped.
 */
extern EIF_INTEGER eif_file_gw(FILE *f, char *s, EIF_INTEGER bound, EIF_INTEGER start);

/*
 * Look ahead one character. If EOF, return 0.
 */
extern EIF_CHARACTER_8 eif_file_lh(FILE *f);

/*
 * Size of file `fp'.
 */
extern EIF_INTEGER eif_file_size(FILE *fp);

/*
 * Read upto next input line.
 */
extern void eif_file_tnil(FILE *f);

/*
 * Current position within file.
 */
extern EIF_INTEGER eif_file_tell(FILE *f);

/*
 * Touch file `name' by setting both access and modification time to the
 * current time stamp. This external function exists only because there
 * is no way within UNIX_FILE to get the current time stamp. Otherwise,
 * we could simply call file_utime.
 */
extern void eif_file_touch(EIF_FILENAME name);

/*
 * Modify the modification and/or the access time stored in the file's
 * inode. The 'how' parameter tells which attributes should be set.
 */
extern void eif_file_utime(EIF_FILENAME name, time_t stamp, int how);

/*
 * This is an encapsulation of the stat() system call. The routine either
 * succeeds and returns or fails and raises the appropriate exception.
 */
extern int eif_file_stat(EIF_FILENAME path, rt_stat_buf *buf, int follow);

/*
 * Change permissions of file `name', using an interface like chmod(1).
 * The flag is true if permissions are to be added, 0 to remove them.
 */
extern void eif_file_perm(EIF_FILENAME name, char *who, char *what, int flag);

/*
 * Change permission mode on file `path'.
 */
extern void eif_file_chmod(EIF_FILENAME path, int mode);

/*
 * Change the owner of the file to `uid'.
 */
extern void eif_file_chown(EIF_FILENAME name, int uid);

/*
 * Change the group of the file to `gid'.
 */
extern void eif_file_chgrp(EIF_FILENAME name, int gid);

/*
 * Put new_line onto `f'.
 */
extern void eif_file_tnwl(FILE *f);

/*
 * Append a copy of `other' to `f'.
 */
extern void eif_file_append(FILE *f, FILE *other, EIF_INTEGER l);

/*
 * Write string `str' on `f'.
 */
extern void eif_file_ps(FILE *f, char *str, EIF_INTEGER len);

/*
 * Write character `c' on `f'.
 */
extern void eif_file_pc(FILE *f, char c);

/*
 * Go to absolute position `pos' counted from start.
 */
extern void eif_file_go(FILE *f, EIF_INTEGER pos);

/*
 * Go to absolute position `pos' counted from end.
 */
extern void eif_file_recede(FILE *f, EIF_INTEGER pos);

/*
 * Go to absolute position `pos' counted from current position.
 */
extern void eif_file_move(FILE *f, EIF_INTEGER pos);

/*
 * End of file.
 */
extern EIF_BOOLEAN eif_file_feof(FILE *fp);

/*
 * Test whether file exists or not. If `name' represents a symbolic link,
 * it will check that pointed file does exist.
 */
extern EIF_BOOLEAN eif_file_exists(EIF_FILENAME name);

/*
 * Test whether file exists or not without following the symbolic link
 * if `name' represents one.
 */
extern EIF_BOOLEAN eif_file_path_exists(EIF_FILENAME name);

/*
 * Check whether access permission `op' are possible on file `name' using
 * real UID and real GID. This is probably only useful to setuid or setgid
 * programs.
 */
extern EIF_BOOLEAN eif_file_access(EIF_FILENAME name, EIF_INTEGER op);

/*
 * Check whether the file `path' may be created: we need write permissions
 * in the parent directory and there must not be any file bearing that name
 * with no write permissions...
 */
extern EIF_BOOLEAN eif_file_creatable(EIF_FILENAME path, EIF_INTEGER nbytes);

/*
 * Get an integer from `f'.
 */
extern EIF_INTEGER eif_file_gi(FILE *f);

/*
 * Get a real from `f'.
 */
extern EIF_REAL_32 eif_file_gr(FILE *f);

/*
 * Get a double from `f'.
 */
extern EIF_REAL_64 eif_file_gd(FILE *f);

/*
 * Write `number' on `f'.
 */
extern void eif_file_pi(FILE *f, EIF_INTEGER number);

/*
 * Write `number' on `f'.
 */
extern void eif_file_pr(FILE *f, EIF_REAL_32 number);

/*
 * Write double `val' onto `f'.
 */
extern void eif_file_pd(FILE *f, EIF_REAL_64 val);

/*
 * Size of the stat structure. This is used by the Eiffel side to create
 * the area (special object) which will play the role of a stat buffer
 * structure.
 */
extern EIF_INTEGER stat_size(void);

/*
 * Check file permissions using effective UID and effective GID. The
 * current permission mode is held in the st_mode field of the stat()
 * buffer structure `buf'.
 */
extern EIF_BOOLEAN eif_file_eaccess(rt_stat_buf *buf, int op);

/*
 * Perform the field dereferencing from the appropriate stat structure,
 * which Eiffel cannot do directly.
 */
extern EIF_INTEGER eif_file_info(rt_stat_buf *buf, int op);

/*
 * Return the Eiffel string filled in with the name associated with `uid'
 * if found in /etc/passwd. Otherwise, return fill it in with the numeric
 * value.
 */
extern EIF_REFERENCE eif_file_owner(int uid);

/*
 * Return the Eiffel string filled in with the name associated with `gid'
 * if found in /etc/group. Otherwise, return fill it in with the numeric
 * value.
 */
extern EIF_REFERENCE eif_file_group(int gid);

/*
 * Get an integer from `f'.
 */
extern EIF_INTEGER eif_file_gib(FILE* f);

/*
 * Get a real from `f'.
 */
extern EIF_REAL_32 eif_file_grb(FILE* f);

/*
 * Get a double from `f'.
 */
extern EIF_REAL_64 eif_file_gdb(FILE* f);

/*
 * Open file `name' with the corresponding type `how'.
 */
extern EIF_POINTER eif_file_binary_open(EIF_FILENAME name, int how);

/*
 * Open file `fd' with the corresponding type `how'.
 */
extern EIF_POINTER eif_file_binary_dopen(int fd, int how);

/*
 * Reopen file `name' with the corresponding type `how' and substitute that
 * to the old stream described by `old'. This is useful to redirect 'stdout'
 * to another place, for instance.
 */
extern EIF_POINTER eif_file_binary_reopen(EIF_FILENAME name, int how, FILE* old);

/*
 * Write `number' on `f'.
 */
extern void eif_file_pib(FILE* f, EIF_INTEGER number);

/*
 * Write `number' on `f'.
 */
extern void eif_file_prb(FILE* f, EIF_REAL_32 number);

/*
 * Write double `val' onto `f'.
 */
extern void eif_file_pdb(FILE* f, EIF_REAL_64 val);

/*
 * Modification time of a file.
 * Seconds since epoch (01 January 1970) in UTC or 0 if time cannot be retrieved.
 */
extern EIF_INTEGER eif_file_date(EIF_FILENAME  name);

/*
 * Access time of a file.
 * Seconds since epoch (01 January 1970) in UTC or 0 if time cannot be retrieved.
 */
extern EIF_INTEGER eif_file_access_date(EIF_FILENAME name);

/*
 * Generate a temporary file with a name based on a template and
 * return a file descriptor to the file.
 * The template is overwritten with the name of the new file.
 * Return a non-zero file descriptor for the temporary file upon success, -1 otherwise.
 * `template'is the template to use for creating the temporary file.
 * It must match the rules for mk[s]temp (i.e. end in "XXXXXX").
 * On exit, it is updated with the new name.
 * `is_text_mode', if non-zero, then the temporary file is created
 * in text mode, otherwise in binary mode.<param>
 * Note: Code was inspired from https://github.com/mirror/mingw-w64/blob/master/mingw-w64-crt/misc/mkstemp.c
 */
extern EIF_INTEGER eif_file_mkstemp(EIF_FILENAME a_template, EIF_BOOLEAN is_text_mode);

#ifdef __cplusplus
}
#endif

#endif
/*
	description:

		"C functions used to implement class EXCEPTIONS"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2006-2018, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef EIF_EXCEPT_H
#define EIF_EXCEPT_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif
#ifndef GE_EXCEPTION_H
#include "ge_exception.h"
#endif

#include <errno.h>

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Predefined exception codes.
 */
#define EN_VOID		GE_EX_VOID			/* Feature applied to void reference */
#define EN_MEM		GE_EX_MEM			/* No more memory */
#define EN_PRE		GE_EX_PRE			/* Pre-condition violated */
#define EN_POST		GE_EX_POST			/* Post-condition violated */
#define EN_FLOAT	GE_EX_FLOAT			/* Floating point exception (signal SIGFPE) */
#define EN_CINV		GE_EX_CINV			/* Class invariant violated */
#define EN_CHECK	GE_EX_CHECK			/* Assertion violated */
#define EN_FAIL		GE_EX_FAIL			/* Routine failure */
#define EN_WHEN		GE_EX_WHEN			/* Unmatched inspect value */
#define EN_VAR		GE_EX_VAR			/* Non-decreasing loop variant */
#define EN_LINV		GE_EX_LINV			/* Loop invariant violated */
#define EN_SIG		GE_EX_SIG			/* Operating system signal */
#define EN_BYE		GE_EX_BYE			/* Eiffel run-time panic */
#define EN_RESC		GE_EX_RESC			/* Exception in rescue clause */
#define EN_OMEM		GE_EX_OMEM			/* Out of memory (cannot be ignored) */
#define EN_RES		GE_EX_RES			/* Resumption failed (retry did not succeed) */
#define EN_CDEF		GE_EX_CDEF			/* Create on deferred */
#define EN_EXT		GE_EX_EXT			/* External event */
#define EN_VEXP		GE_EX_VEXP			/* Void assigned to expanded */
#define EN_HDLR		GE_EX_HDLR			/* Exception in signal handler */
#define EN_IO		GE_EX_IO			/* I/O error */
#define EN_SYS		GE_EX_SYS			/* Operating system error */
#define EN_RETR		GE_EX_RETR			/* Retrieval error */
#define EN_PROG		GE_EX_PROG			/* Developer exception */
#define EN_FATAL	GE_EX_FATAL			/* Eiffel run-time fatal error */
#define EN_DOL		GE_EX_DOL			/* $ applied to melted feature */
#define EN_ISE_IO	GE_EX_ISE_IO		/* I/O error raised by the ISE Eiffel runtime */
#define EN_COM		GE_EX_COM			/* COM error raised by EiffelCOM runtime */
#define EN_RT_CHECK	GE_EX_RT_CHECK		/* Runtime check error such as out-of-bound array access */
#define EN_OLD		GE_EX_OLD			/* Old violation */
#define EN_SEL		GE_EX_SEL			/* Serialization failure */
#define EN_DIRTY	GE_EX_DIRTY			/* SCOOP processor dirty exception. */
#define EN_NEX		GE_EX_NEX			/* Number of internal exceptions */

/*
 * Raise an Eiffel exception.
 */
extern void eraise(const char* name, long code);

/*
 * Raise an Eiffel exception of the given code with no associated tag.
 */
extern void xraise(int code);

/*
 * Raise an "Operating system error" exception.
 */
extern void esys(void);

/*
 * As a special case, an I/O error is raised when a system call which is I/O bound fails.
 */
extern void eise_io(const char* tag);

/*
 * Raise a "No more memory" exception.
 */
extern void enomem(void);

/*
 * Raise EiffelCOM exception.
 */
extern void com_eraise(const char* tag, long num);

/*
 * Terminate execution with exit status `code',
 * without triggering an exception.
 */
extern void esdie(int code);

/*
 * Exception tag associated with `code'.
 * This is a duplication from Eiffel classes, but still used for trace printing and in EiffelCom.
 */
extern EIF_REFERENCE eename(long except);

/*
 * Is exception `ex' defined?
 * Used in EiffelCOM.
 */
extern char eedefined(long ex);

/*
 * Enable/diable printing of the exception trace.
 * Per thead information.
 */
extern void eetrace(char b);

/*
 * Is current execution during rescue?
 */
extern EIF_BOOLEAN eif_is_in_rescue(void);

#ifdef EIF_WINDOWS
/*
 * Set default exception handler.
 */
extern void set_windows_exception_filter(void);
#endif

/*
 * Used in EiffelCOM.
 */
#define echval GE_current_context()->exception_code
#define echtg GE_current_context()->exception_tag

#ifdef __cplusplus
}
#endif

#endif
/*
	description:

		"C functions used to implement class REAL"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2006-2019, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef GE_REAL_H
#define GE_REAL_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#include <math.h>

#ifndef GE_power
#define GE_power(x,y) pow((x),(y))
#endif
#define GE_ceiling(x) ceil(x)
#define GE_floor(x) floor(x)

extern EIF_NATURAL_32 GE_real_32_to_bits (EIF_REAL_32 v);
extern EIF_NATURAL_64 GE_real_64_to_bits (EIF_REAL_64 v);
extern EIF_REAL_32 GE_real_32_from_bits (EIF_NATURAL_32 v);
extern EIF_REAL_64 GE_real_64_from_bits (EIF_NATURAL_64 v);

#define GE_real_32_is_nan(v) ((v) != (v) ? EIF_TRUE : ((GE_real_32_to_bits(v) & ~GE_nat32(0x80000000)) > GE_nat32(0x7FF00000)))
#define GE_real_64_is_nan(v) ((v) != (v) ? EIF_TRUE : ((GE_real_64_to_bits(v) & ~GE_nat64(0x8000000000000000)) > GE_nat64(0x7FF0000000000000)))
#define GE_real_32_is_negative_infinity(v) (GE_real_32_to_bits(v) == GE_nat32(0xFF800000))
#define GE_real_64_is_negative_infinity(v) (GE_real_64_to_bits(v) == GE_nat64(0xFFF0000000000000))
#define GE_real_32_is_positive_infinity(v) (GE_real_32_to_bits(v) == GE_nat32(0x7F800000))
#define GE_real_64_is_positive_infinity(v) (GE_real_64_to_bits(v) == GE_nat64(0x7FF0000000000000))
#define GE_real_32_nan GE_real_32_from_bits(GE_nat32(0x7FC00000))
#define GE_real_64_nan GE_real_64_from_bits(GE_nat64(0x7FF8000000000000))
#define GE_real_32_negative_infinity GE_real_32_from_bits(GE_nat32(0xFF800000))
#define GE_real_64_negative_infinity GE_real_64_from_bits(GE_nat64(0xFFF0000000000000))
#define GE_real_32_positive_infinity GE_real_32_from_bits(GE_nat32(0x7F800000))
#define GE_real_64_positive_infinity GE_real_64_from_bits(GE_nat64(0x7FF0000000000000))
#define GE_real_32_is_equal(v1,v2) ((v1) == (v1) ? (v1) == (v2) : (v2) != (v2))
#define GE_real_64_is_equal(v1,v2) ((v1) == (v1) ? (v1) == (v2) : (v2) != (v2))
#define GE_real_32_is_less(v1,v2) ((v1) == (v1) ? (v1) < (v2) : (v2) == (v2))
#define GE_real_64_is_less(v1,v2) ((v1) == (v1) ? (v1) < (v2) : (v2) == (v2))
#define GE_real_32_ieee_is_equal(v1,v2) (v1) == (v2)
#define GE_real_64_ieee_is_equal(v1,v2) (v1) == (v2)
#define GE_real_32_ieee_is_greater(v1,v2) isgreater(v1,v2)
#define GE_real_64_ieee_is_greater(v1,v2) isgreater(v1,v2)
#define GE_real_32_ieee_is_greater_equal(v1,v2) isgreaterequal(v1,v2)
#define GE_real_64_ieee_is_greater_equal(v1,v2) isgreaterequal(v1,v2)
#define GE_real_32_ieee_is_less(v1,v2) isless(v1,v2)
#define GE_real_64_ieee_is_less(v1,v2) isless(v1,v2)
#define GE_real_32_ieee_is_less_equal(v1,v2) islessequal(v1,v2)
#define GE_real_64_ieee_is_less_equal(v1,v2) islessequal(v1,v2)
#define GE_real_32_ieee_maximum_number(v1,v2) fmaxf(v1,v2)
#define GE_real_64_ieee_maximum_number(v1,v2) fmax(v1,v2)
#define GE_real_32_ieee_minimum_number(v1,v2) fminf(v1,v2)
#define GE_real_64_ieee_minimum_number(v1,v2) fmin(v1,v2)

#endif
/*
	description:

		"Part of ISE Eiffel runtime. Needed to compile ISE's compiler."

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2019, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef EIF_SIZE_H
#define EIF_SIZE_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* Basic sizes */
#define EIF_ALIGN	8	/* Alignment restrictions */
#define R64SIZ	8	/* Size of double */
#define I64SIZ	8	/* Size of integer 64 bits */
#define R32SIZ	4	/* Size of float */
#define REFSIZ	sizeof(char*)	/* Size of char* */
#define LNGSIZ	4	/* Size of long */
#define I16SIZ	2	/* Size of integer 16 bits */
#define CHRSIZ	1	/* Size of char */
#define PTRSIZ	sizeof(void*)	/* Size of a function pointer */

/* Macros used to access fields in the object */
#define REFACS(n) ((n)*REFSIZ)
#define CHRACS(n) ((n)*CHRSIZ)
#define I16ACS(n) ((n)*I16SIZ)
#define LNGACS(n) ((n)*LNGSIZ)
#define R32ACS(n) ((n)*R32SIZ)
#define I64ACS(n) ((n)*I64SIZ)
#define R64ACS(n) ((n)*R64SIZ)
#define PTRACS(n) ((n)*PTRSIZ)

#ifdef __cplusplus
}
#endif

#endif
/*
	description:

		"Part of ISE Eiffel runtime. Needed to compile the EiffelCOM library."

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2010-2019, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef EIF_GLOBALS_H
#define EIF_GLOBALS_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EXCEPTION_H
#include "ge_exception.h"
#endif
#ifndef EIF_THREADS_H
#include "eif_threads.h"
#endif
#ifndef EIF_SIZE_H
#include "eif_size.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifdef GE_USE_THREADS
#define EIF_GET_CONTEXT \
	GE_context* eif_globals = GE_current_context();
#else
#define EIF_GET_CONTEXT
#endif

#define GTCX EIF_GET_CONTEXT

#ifdef __cplusplus
}
#endif

#endif
/*
	description:

		"C declarations for the ISE Eiffel runtime."

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2006-2016, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef EIF_EIFFEL_H
#define EIF_EIFFEL_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif
#ifndef GE_STRING_H
#include "ge_string.h"
#endif
#ifndef EIF_GLOBALS_H
#include "eif_globals.h"
#endif
#ifndef EIF_EXCEPT_H
#include "eif_except.h"
#endif

#endif
/*
	description:

		"C functions used to implement class DIRECTORY"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2006-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef EIF_DIR_H
#define EIF_DIR_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

extern EIF_POINTER eif_dir_open(EIF_FILENAME dirname);
extern EIF_POINTER eif_dir_next(EIF_POINTER dir);
extern EIF_POINTER eif_dir_rewind(EIF_POINTER d, EIF_FILENAME dir_name);
extern void eif_dir_close(EIF_POINTER dir);
extern EIF_BOOLEAN eif_dir_exists(EIF_FILENAME dirname);
extern EIF_BOOLEAN eif_dir_is_readable(EIF_FILENAME dirname);
extern EIF_BOOLEAN eif_dir_is_writable(EIF_FILENAME dirname);
extern EIF_BOOLEAN eif_dir_is_executable(EIF_FILENAME dirname);
extern EIF_BOOLEAN eif_dir_is_deletable(EIF_FILENAME name);
extern EIF_CHARACTER_8 eif_dir_separator(void);
extern EIF_INTEGER eif_dir_current(EIF_FILENAME a_buffer, EIF_INTEGER a_count);
extern EIF_INTEGER eif_chdir(EIF_FILENAME path);

#ifdef __cplusplus
}
#endif

#endif
/*
	description:

		"C functions used to access DLLs"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2007-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef GE_DLL_H
#define GE_DLL_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifdef __cplusplus
extern "C" {
#endif

#ifdef EIF_WINDOWS

#include <windows.h>

/* DLL declarations */
#define GE_load_dll(name) LoadLibraryA((LPCSTR)(name))

#endif

#ifdef __cplusplus
}
#endif

#endif
/*
	description:

		"C functions used to implement class EXECUTION_ENVIRONMENT"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2006-2017, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef EIF_MISC_H
#define EIF_MISC_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif
#ifndef GE_DLL_H
#include "ge_dll.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

extern EIF_INTEGER eif_system(EIF_NATIVE_CHAR* s);
extern void eif_system_asynchronous(EIF_NATIVE_CHAR* cmd);
extern void eif_sleep(EIF_INTEGER_64);

#ifdef EIF_WINDOWS
/* DLL declarations */
#define eif_load_dll(name) GE_load_dll(name)
#endif

#ifdef __cplusplus
}
#endif

#endif
/*
	description:

		"C functions used to implement class COM_FAILURE"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2016-2018, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef GE_COM_FAILURE_H
#define GE_COM_FAILURE_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifndef GE_EIFFEL_H
#include "ge_eiffel.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

extern EIF_INTEGER GE_ccom_hresult(EIF_POINTER an_exception_code);

extern EIF_INTEGER GE_ccom_hresult_code(EIF_INTEGER an_hresult);

extern EIF_INTEGER GE_ccom_hresult_facility(EIF_INTEGER an_hresult);

/*
 * Get text from error `a_code'. It is up to the caller to free
 * the returned buffer using `GE_ccom_local_free'.
 */
extern EIF_POINTER GE_ccom_error_text(EIF_INTEGER a_code);

/*
 * Number of characters in `ptr'.
 */
extern EIF_INTEGER_32 GE_ccom_strlen(EIF_POINTER ptr);

/*
 * Number of bytes occupied by a TCHAR.
 */
extern EIF_INTEGER_32 GE_ccom_character_size(void);

/*
 * Free `a_ptr' using LocalFree.
 */
extern void GE_ccom_local_free(EIF_POINTER a_ptr);

#ifdef __cplusplus
}
#endif

#endif
/*
	description:

		""Part of ISE Eiffel runtime. Needed to compile class EXECUTION_ENVIRONMENT"

	system: "Gobo Eiffel Compiler"
	copyright: "Copyright (c) 2018-2020, Eric Bezault and others"
	license: "MIT License"
	date: "$Date: 2023-03-09 10:35:58 +0000 (Thu, 09 Mar 2023) $"
	revision: "$Revision: 106643 $"
*/

#ifndef EIF_SYSTEM_H
#define EIF_SYSTEM_H
#if defined(_MSC_VER) && (_MSC_VER >= 1020)
#pragma once
#endif

#ifdef EIF_WINDOWS
#include "Winbase.h"
#elif defined(EIF_VMS) || defined(EIF_MACOSX)
#include "sysctl.conf"
#else
#include <unistd.h>
#endif

#endif
#include <string.h>
#include <stdlib.h>
#include <time.h>
